#ifndef WEFGPRIV_H
#   define WEFGPRIV_H

#include "WEFG.h"


#ifdef USE_ESFLIB
#   define SFL_PREFIX(name)           ESFL3_##name
#   define SFL_DIM_INDEP_PREFIX(name) ESFLIB_##name
#   define SFL_sf_table_t             ESFLIB_sf_table_t
#   define SFL_generic_ptr_t          ESFLIB_generic_ptr_t
#   define SFL_efg_node_id_t          ESFLIB_efg_node_id_t
#   define SFL_shape_func_at          ESFL3_shape_func_at
#   define SFL_num_of_entries         ESFL3_num_of_entries
#   define SFL_inspect_row            ESFL3_inspect_row
#   define SFL_explain_error(sf)      ESFL3_explain_error(ESFL3_sf_lib(sf))
#   define SFL_efg_point_at           ESFL3_efg_node_at 
#   define SFL_search_cast            (ESFLIB_search_range_func)
#   define SFL_first_cast             (ESFLIB_first_in_range_func)
#   define SFL_next_cast              (ESFLIB_next_in_range_func)
#else
#   define SFL_PREFIX(name)           ESP_3_##name
#   define SFL_DIM_INDEP_PREFIX(name) ESPACK_##name
#   define SFL_sf_table_t             ESPACK_sf_table_t
#   define SFL_generic_ptr_t          ESPACK_generic_ptr_t
#   define SFL_efg_node_id_t          ESPACK_efg_node_id_t
#   define SFL_shape_func_at          ESP_3_shape_func_at
#   define SFL_num_of_entries         ESP_3_num_of_entries
#   define SFL_inspect_row            ESP_3_inspect_row
#   define SFL_explain_error(sf)      ESP_3_explain_error()
#   define SFL_efg_point_at           ESP_3_efg_point_at
#   define SFL_search_cast            (ESPACK_search_range_func)
#   define SFL_first_cast             (ESPACK_first_in_range_func)
#   define SFL_next_cast              (ESPACK_next_in_range_func)
#endif


void
bad_search(char *ss, G3D_box_t *range);


#define INIT_EFG_SF_STAT(se)                     \
  {                                              \
    se->efg_sf_stat.sum_nneighbors = 0;          \
    se->efg_sf_stat.nevals = 0;                  \
    se->efg_sf_stat.min_nneighbors = 1000000;    \
    se->efg_sf_stat.max_nneighbors = 0;          \
    se->efg_sf_stat.avg_nneighbors = 0;          \
  } 
#define EVAL_EFG_SF_STAT(se)                                              \
    {                                                                     \
      se->efg_sf_stat.avg_nneighbors =                                    \
        (se->efg_sf_stat.sum_nneighbors / se->efg_sf_stat.nevals);        \
    }
#define UPDATE_EFG_SF_STAT(se, ne)                                            \
  {                                                                           \
    se->efg_sf_stat.sum_nneighbors += ne;                                     \
    se->efg_sf_stat.nevals++;                                                 \
    se->efg_sf_stat.min_nneighbors = min(se->efg_sf_stat.min_nneighbors, ne); \
    se->efg_sf_stat.max_nneighbors = max(se->efg_sf_stat.max_nneighbors, ne); \
  }


/* Private function prototypes:  These functions are NOT expected */
/* to be called from outside the EFG module.  */


/* EFG SUPERELEMENT */

WEFG_vertex_t *
efg_get_vertex(WEFG_superelem_t *t, W_index_t n);

WEFG_node_t *
efg_add_node(WEFG_superelem_t *se, W_loc_t *loc);

WEFG_node_t *
efg_node_in_range(WEFG_superelem_t *se, G3D_box_t *range);

WEFG_node_t *
efg_next_node(WEFG_superelem_t *se, WEFG_node_t *en);

void
attach_b_face(WEFG_superelem_t *se, WEFG_b_face_t *f);

WEFG_b_face_t * /* face or NULL (no memory) */
copy_b_face(WEFG_b_face_t *f /* face to be copied */
           );

void
attach_crack_surf(WEFG_superelem_t *se, WEFG_crack_surf_t *cs);

void
efg_attach_vertex(WEFG_superelem_t *se, WEFG_vertex_t *v);

void
efg_attach_hex (WEFG_superelem_t *se, WEFG_hex_volume_t *h);

void
attach_symm_plane(WEFG_superelem_t *se, WEFG_symm_plane_t *sp);

void
efgss_init_ss_routines(WEFG_superelem_t *se);

void
efg_attach_node(WEFG_superelem_t *se, WEFG_node_t *en);

WEFG_hex_volume_t *
efg_find_enclosing_hex(WEFG_superelem_t *se, W_loc_t *loc,
                       double eps, W_loc_t *param_loc);

void
efg_add_invalid_region(WEFG_superelem_t *se,
                       int ntriangles,
                       G3D_tri_t *triangles);

void
efg_invalid_region_triangles_reliable (WEFG_superelem_t *se,
                                       W_bool_t YES_OR_NO);

W_bool_t
efg_pt_is_inside_invalid_region(WEFG_superelem_t *se,
                                W_loc_t *pt,
                                WEFG_sf_tab_t *sftab,
                                W_bool_t full_check);

WEFG_b_face_t *
efg_find_intersection_with_bface (WEFG_superelem_t *se,
                                  G3D_segment_t *s,
                                  double lower_lambda,
                                  double upper_lambda,
                                  W_loc_t *intersection,
                                  WEFG_which_b_face_enum which);

void
efg_add_ip_to_search_structure(WEFG_superelem_t          *se,
			       W_loc_t                   *ip_loc,
			       WEFG_hex_volume_t         *enclosing_hex,
			       WEFG_integration_point_t  *ip);

void
efg_enable_ip_storage_to_search_structure(WEFG_superelem_t *se);

W_bool_t
efg_average_stress_in_box(WEFG_superelem_t *se,
			  G3D_box_t *box,
			  W_mat_3d_stress_type_enum requested_out_type,
			  double stress[3][3]);

W_mat_3d_state_p
efg_interpolate_mat_state(WEFG_superelem_t *se,
			  W_loc_t *at_point,
			  double initial_box_size);

void
efg_write_ve3d(W_elem_t *el, FILE *stream);

void
efg_write_tecplot(W_elem_t *el, FILE *stream, void *optdata);


#ifdef USE_ESFLIB
int
efg_error_esflib (ESFLIB_lib_t lib,
                  char *file, int line, double x, double y, double z,
                  int kind
                  );
#define SFL_ERROR(sf, X, Y, Z, kind) \
   {efg_error_esflib(ESFL3_sf_lib(sf), __FILE__, __LINE__, X, Y, Z, kind);}
#else
int
efg_error_espack(char *file, int line, double x, double y, double z);
#define SFL_ERROR(sf, X, Y, Z) \
   {efg_error_espack(__FILE__, __LINE__, X, Y, Z);}
#endif


W_bool_t 
efg_probe(W_elem_t *el, void *options,
	  W_loc_t *loc, unsigned long spec,
	  W_probe_package_t *result);


W_bool_t
efg_interpolate_at(WEFG_superelem_t *se,
		   W_loc_t *at,
		   void *options,
		   WEFG_update_field_pkg_proc update_pkg,
		   WEFG_field_package_t *result
	);

W_bool_t 
efg_update_displ_pkg(void *options,
		     WEFG_superelem_t *se,
		     WEFG_ip_ref_package_t *ippkg,
		     double weight,
		     WEFG_field_package_t *pkg);

W_bool_t 
efg_update_stress_pkg(void *options,
		      WEFG_superelem_t *se,
		     WEFG_ip_ref_package_t *ippkg,
		     double weight,
		     WEFG_field_package_t *pkg);

W_bool_t 
efg_update_displ_grad_pkg(void *options,
			  WEFG_superelem_t *se,
			  WEFG_ip_ref_package_t *ippkg,
			  double weight,
			  WEFG_field_package_t *pkg);

W_bool_t 
efg_update_deform_energy_pkg(void *options,
			     WEFG_superelem_t *se,
			     WEFG_ip_ref_package_t *ippkg,
			     double weight,
			     WEFG_field_package_t *pkg);

void
efgj_j_volume(WEFG_superelem_t *se,
              W_loc_t *at_point,
              W_vect_t *forward,
              W_vect_t *tangent,
              W_vect_t *normal,
              double hsize_forward_plus,
              double hsize_forward_minus,
              double hsize_tangent_plus,
              double hsize_tangent_minus,
              double hsize_normal_plus,
              double hsize_normal_minus,
              WEFG_tensor_prod_comp_func func,
              WEFG_tensor_prod_comp_der_func func_der,
              double *J_forward_static,
              double *J_forward_dynamic,
              double *J_normal_static,
              double *J_normal_dynamic
              );
void *
efg_new_crack_tip_sampler(WEFG_superelem_t *se,
			  W_index_t pntidx, W_loc_t *loc);

double
efg_support_size_at(WEFG_superelem_t *se, double x, double y, double z);

void
efg_set_support_size(WEFG_superelem_t *se, double support_size);

void
efg_set_max_support_size(WEFG_superelem_t *se, double support_size);



/* VERTICES */

void
read_vertices(WEFG_superelem_t *t, W_domain_t *d);

void
read_connectivities(WEFG_superelem_t *t, W_domain_t *d);

void
attach_hex_vol(WEFG_vertex_t *v, WEFG_hex_volume_t *h);

WEFG_hex_volume_t *
next_incident_hex_vol(WEFG_vertex_t *v, WEFG_hex_volume_t *h);

void
vertex_ref_loc(WEFG_superelem_t *se, WEFG_vertex_t *v, W_loc_t *loc);

void
vertex_current_loc(WEFG_superelem_t *se, WEFG_vertex_t *v, W_loc_t *loc);

void
vertex_current_vel(WEFG_superelem_t *se, WEFG_vertex_t *v, W_loc_t *vel);

void
vertex_current_acc(WEFG_superelem_t *se, WEFG_vertex_t *v, W_loc_t *acc);

WEFG_vertex_t *
efgv_new_vertex(W_index_t id, W_loc_t *loc);
void
efgv_generate_efgn (WEFG_vertex_t *v, WEFG_superelem_t *se);
WEFG_vertex_t *
efg_vertex_in_range(WEFG_superelem_t *se, G3D_box_t *range);

void
efgv_update_stress(WEFG_superelem_t *se, WEFG_vertex_t *v);

/*
  Retrieve named material state variable.
 */
double
efgv_get_named_var (WEFG_superelem_t *se, /* EFG superelement */
                    WEFG_vertex_t *v /* integration volume vertex */,
                    char *var);
void
efgv_get_stress(WEFG_superelem_t *se, WEFG_vertex_t *v,
                W_output_stress_type_enum stress_type, double stress[3][3]);

void
efgv_get_equiv_pl_def(WEFG_superelem_t *se, WEFG_vertex_t *v,
                      double *equiv_pl_def);

void *
efgv_new_sampler(WEFG_superelem_t *se, W_index_t pntidx, W_loc_t *loc);

void 
efgv_sampler_store(W_elem_t *e,
		   W_solution_history_t *h,
		   void *data,
		   double time);

void
efgv_write_tecplot(WEFG_vertex_t *v, WEFG_superelem_t *se, FILE *stream,
                   W_index_t *count);

W_node_t *
efgv_get_coupling_fen_conn (WEFG_superelem_t *se, WEFG_vertex_t *v);
W_node_t *
efgv_get_driven_fen_conn (WEFG_superelem_t *se, WEFG_vertex_t *v);
void
efgv_Newmark_update (WEFG_vertex_t *v, WEFG_superelem_t *se);

void
efgv_assemble_driven_fen_force (WEFG_vertex_t *v, WEFG_superelem_t *se);


/* HEXAHEDRA */

void
read_hexahedra(WEFG_superelem_t *t, W_domain_t *d);

void
hex_calc_eff_loads(WEFG_superelem_t *se, WEFG_hex_volume_t *h);

double
hex_total_mass(WEFG_superelem_t *se, WEFG_hex_volume_t *h);

void
hex_generate_efg_nodes(WEFG_superelem_t *se, WEFG_hex_volume_t *h);

void
hex_find_b_faces(WEFG_superelem_t *se, WEFG_hex_volume_t *h, int level);

void
hex_assemble_mass(WEFG_superelem_t *se, WEFG_hex_volume_t *h);

void
hex_assemble_restoring_forces(WEFG_superelem_t *se, W_mat_state_update_form_enum update_form, WEFG_hex_volume_t *h);

void
hex_write_hexvue_postproc_file (WEFG_superelem_t *se, WEFG_hex_volume_t *h,
                                ckit_iobuf_t IOB, char *mode);
void
hex_write_hexvue_XTRA_postproc_file (WEFG_superelem_t *se, WEFG_hex_volume_t *h,
                                ckit_iobuf_t IOB, char *mode);

void
hex_write_tecplot_postproc_file(WEFG_hex_volume_t *h, WEFG_superelem_t *se, 
                                ckit_iobuf_t IOB, W_index_t *count);

WEFG_hex_volume_t *
hex_loc_is_inside(WEFG_hex_volume_t *h, W_loc_t *loc, double eps,
                    W_loc_t *param_loc);

int
hex_is_interface_vol(WEFG_hex_volume_t *h);


void 
hex_construct_ramp(SFL_generic_ptr_t if_data,
                   double *R,
                   double D_R[3], double x, double y, double z);

int 
hex_elem_node_num(SFL_generic_ptr_t if_data,
                  SFL_efg_node_id_t efgp);

void
hex_fe_shape(SFL_generic_ptr_t if_data,
             int  node_num,
             double *n,
             double *nx, double *ny, double *nz,
             double x, double y, double z);

void
hex_modify_elem_grep(WEFG_superelem_t *se, double displ_scale, WEFG_hex_volume_t *h);

int
hex_averaged_stress(WEFG_superelem_t *se, WEFG_hex_volume_t *h,
                    W_mat_3d_stress_type_enum requested_out_type,
                    double stress[3][3]);

double
hex_deform_energy(WEFG_superelem_t *se, WEFG_hex_volume_t *h);




/* EFG NODES */

W_index_t
efgn_number_eqns(WEFG_node_t *en, W_index_t ne,
                 ListOf(WEFG_symm_plane_t *) symm_planes);

WEFG_node_t *
new_efg_node(WEFG_superelem_t *se, W_loc_t *loc, W_float_t support_size);

void
efgn_support_size (SFL_efg_node_id_t efgp, double *sx, double *sy, double *sz);

void
efgn_ref_coords(SFL_efg_node_id_t efgn, double *x, double *y, double *z);

void
efgn_prev_coords(WEFG_node_t *en, double *x, double *y, double *z);

#define EFGN_REF_COORDS(EN, x_, y_, z_)                                 \
  if        (EFGN_IS_TRUE_EFG(EN) || EFGN_IS_EFG_IF(EN)) {              \
    x_ = EN->ndata.ref_coords.x;                                    \
    y_ = EN->ndata.ref_coords.y;                                    \
    z_ = EN->ndata.ref_coords.z;                                    \
  } else {                                                              \
    WN_get_ref_coords (EN->fen, &x_, &y_, &z_);                    \
  }

#define EFGN_PREV_COORDS(EN, x_, y_, z_)                                \
  if        (EFGN_IS_TRUE_EFG(EN) || EFGN_IS_EFG_IF(EN)) {              \
    x_ = EN->ndata.ref_coords.x + EN->ndata.displ_t_minus_dt.x; \
    y_ = EN->ndata.ref_coords.y + EN->ndata.displ_t_minus_dt.y; \
    z_ = EN->ndata.ref_coords.z + EN->ndata.displ_t_minus_dt.z; \
  } else {                                                              \
    WN_get_prev_coords (EN->fen, &x_, &y_, &z_);                   \
  }

#define EFGN_PREV_DISPLS(EN, x_, y_, z_)                                \
  if        (EFGN_IS_TRUE_EFG(EN) || EFGN_IS_EFG_IF(EN)) {              \
    x_ = EN->ndata.displ_t_minus_dt.x; \
    y_ = EN->ndata.displ_t_minus_dt.y; \
    z_ = EN->ndata.displ_t_minus_dt.z; \
  } else {                                                              \
    WN_get_prev_displ (EN->fen, &x_, &y_, &z_);                   \
  }

void
efgn_curr_coords(WEFG_node_t *en, double *x, double *y, double *z);

#define EFGN_CURR_COORDS(EN, x_, y_, z_)                        \
  if        (EFGN_IS_TRUE_EFG(EN) || EFGN_IS_EFG_IF(EN)) {      \
    x_ = EN->ndata.ref_coords.x + EN->ndata.displ_t.x;  \
    y_ = EN->ndata.ref_coords.y + EN->ndata.displ_t.y;  \
    z_ = EN->ndata.ref_coords.z + EN->ndata.displ_t.z;  \
  } else {                                                      \
    WN_get_coords (EN->fen, &x_, &y_, &z_);                \
  }

#define EFGN_CURR_DISPLS(EN, x_, y_, z_)                        \
  if        (EFGN_IS_TRUE_EFG(EN) || EFGN_IS_EFG_IF(EN)) {      \
    x_ = EN->ndata.displ_t.x;  \
    y_ = EN->ndata.displ_t.y;  \
    z_ = EN->ndata.displ_t.z;  \
  } else {                                                      \
    WN_get_displ (EN->fen, &x_, &y_, &z_);                 \
  }

void
efgn_curr_vel(WEFG_node_t *en, double *x, double *y, double *z);

#define EFGN_CURR_VELOC(EN, x_, y_, z_)                         \
  if        (EFGN_IS_TRUE_EFG(EN) || EFGN_IS_EFG_IF(EN)) {      \
    x_ = EN->ndata.velocity_t.x;                            \
    y_ = EN->ndata.velocity_t.y;                            \
    z_ = EN->ndata.velocity_t.z;                            \
  } else {                                                      \
    WN_get_velocity (EN->fen, &x_, &y_, &z_);              \
  }

void
efgn_curr_accel(WEFG_node_t *en, double *x, double *y, double *z);

#define EFGN_CURR_ACCEL(EN, x_, y_, z_)                         \
  if        (EFGN_IS_TRUE_EFG(EN) || EFGN_IS_EFG_IF(EN)) {      \
    x_ = EN->ndata.acceleration_t.x;                        \
    y_ = EN->ndata.acceleration_t.y;                        \
    z_ = EN->ndata.acceleration_t.z;                        \
  } else  {                                                     \
    WN_get_acceleration (EN->fen, &x_, &y_, &z_);          \
  }

void
efgn_Warburton_update(WEFG_node_t *n, WEFG_superelem_t *se, W_float_t dt);

void
efgn_Newmark_update(WEFG_node_t *n, WEFG_superelem_t *se, W_float_t dt);

void
efgn_apply_init_cond(WEFG_node_t *en, W_vect_t *velocity, W_float_t time_step);

void
efgn_assemble_force(WEFG_node_t *en, WEFG_superelem_t *se, W_vect_t *f);

#if 1
#define EFGN_ASSEMBLE_FORCE(en, se, frc)                        \
  {                                                             \
    if (EFGN_IS_TRUE_EFG(en) || EFGN_IS_EFG_IF(en)) {           \
      register W_index_t *AEL_ne = en->ndata.eq_numbers;    \
      register W_vector_t v = se->effective_loads.v;            \
      if (*AEL_ne > 0)                                          \
        v[*AEL_ne] += frc.x;                                    \
      AEL_ne++;                                                 \
      if (*AEL_ne > 0)                                          \
        v[*AEL_ne] += frc.y;                                    \
      AEL_ne++;                                                 \
      if (*AEL_ne > 0)                                          \
        v[*AEL_ne] += frc.z;                           \
    } else if (EFGN_IS_FE_IF(en)) {                             \
      WN_assemble_force(en->fen, &frc);                    \
    }                                                           \
  }
#else
#define EFGN_ASSEMBLE_FORCE(en, se, frc) efgn_assemble_force(en, se, &frc)
#endif

void
efgn_assemble_mass(WEFG_node_t *en, WEFG_superelem_t *se, double M);

#define EFGN_IS_TRUE_EFG(EFGN) (EFGN->type == WEFG_TRUE_EFG_NODE)
#define EFGN_IS_EFG_IF(EFGN)   (EFGN->type == WEFG_EFG_IF_NODE)
#define EFGN_IS_FE_IF(EFGN)    (EFGN->type == WEFG_FE_IF_NODE)


void
efgn_change_to_efg_if_node(WEFG_node_t *en);

void
efgn_change_to_fe_if_node(WEFG_node_t *en, W_node_t *fen);

void
efgn_change_to_driving_node(WEFG_node_t *en, W_node_t *fen);

void
efgn_modify_grep(WEFG_superelem_t *se, double displ_scale, WEFG_node_t *n);

double
efgn_kinetic_energy(WEFG_node_t *en, WEFG_superelem_t *se, double dt);

void
efgn_assemble_damping(WEFG_node_t *en,
		      WEFG_superelem_t *se,
                      double damping_coeff,
		      double dt);

void
efgn_print_info(WEFG_node_t *en,
                FILE *fp,
                int indent,
                WEFG_superelem_t *se);



/* CONCENTRATED LOADS */

void
read_conc_loads(WEFG_superelem_t *t, W_domain_t *d);

void
conc_load_assemble(WEFG_superelem_t *se, WEFG_conc_load_t *cl);


/* DISTRIBUTED LOADS */

void
read_distrib_loads(WEFG_superelem_t *t, W_domain_t *d);

void
distrib_load_assemble(WEFG_superelem_t *se, WEFG_distrib_load_t *dl);


/* CONTACT LOADS */

void
read_contact_loads(WEFG_superelem_t *t, W_domain_t *d);

void
contact_load_assemble(WEFG_superelem_t *se, WEFG_contact_load_t *ctl);


/* BODY LOAD */

void
read_body_load(WEFG_superelem_t *se, W_domain_t *d);



/* CRACKS */


void
crack_find_b_faces(WEFG_superelem_t *se, WEFG_crack_surf_t *c);

void
read_crack(WEFG_superelem_t *t, W_domain_t *d);

void
efgcr_write_hexvue_postproc_file(WEFG_crack_surf_t *cs,
                                 ckit_iobuf_t IOB, char *mode);

#ifdef X_XFACE
void
efgcr_modify_crack_surf_grep (WEFG_superelem_t *se, /* EFG superelement */
                              double displ_scale,
                              WEFG_crack_surf_t *cs /* generic crack surface */
                              ) ;
#endif

void *
efgcr_new_sampler(WEFG_crack_surf_t *cs, W_index_t pntidx, W_loc_t *loc);

void 
efgcr_sampler_store(W_elem_t *e,
		    W_solution_history_t *h,
		    void *data,
		    double time);

int
efgcr_plane_generating_vector(WEFG_crack_surf_t *cs, 
                              double atx, double aty, double atz ,
                              double rcx, double rcy, double rcz ,
                              double *bx, double *by, double *bz );

int 
efgcr_wrap_around_point(WEFG_crack_surf_t *cs, 
                        double atx, double aty, double atz,
                        double rcx, double rcy, double rcz,
                        double bx, double by, double bz,
                        double *wapx, double *wapy, double *wapz);





/* GEOMETRIC OBJECTS */

WEFG_b_face_t *
new_b_face(WEFG_b_face_t *f);

W_bool_t
node_hidden_by_boundary (SFL_generic_ptr_t eval_for,
                         double seg_start_x,
                         double seg_start_y,
                         double seg_start_z ,
                         SFL_efg_node_id_t efgp,
                         double seg_end_x,
                         double seg_end_y,
                         double seg_end_z ,
                         double  eps);

W_bool_t
efggc_segment_intersection_with_face(G3D_segment_t *s,
                                     WEFG_b_face_t *f,
                                     W_bool_t line_intersections,
                                     W_loc_t *ip,
                                     double *lambda);

int 
efggc_use_cpp(SFL_generic_ptr_t eval_for, SFL_efg_node_id_t efgp);

int 
efggc_plane_generating_vector(SFL_generic_ptr_t eval_for,
                              double atx, double aty, double atz ,
                              double rcx, double rcy, double rcz ,
                              double *bx, double *by, double *bz );


int 
efggc_wrap_around_point(SFL_generic_ptr_t eval_for,
                        double atx, double aty, double atz,
                        double rcx, double rcy, double rcz,
                        double bx, double by, double bz,
                        double *wapx, double *wapy, double *wapz);

void
efggc_face_normal (WEFG_b_face_t *f, W_vect_t *n);


/* SYMMETRY PLANE */

void
read_symm_plane(WEFG_superelem_t *t, W_domain_t *d);


/* STRESS/STRAIN EVALUATION */

void
efgss_1st_PK_to_2nd_PK(void);

void
efgss_invert_3x3(W_float_t J[3][3], W_float_t inv_J[3][3]);


/* TIME STAMPS */

#define INIT_TIME_STAMP(stamp) (stamp) = -FLT_MAX
#define TOUCH_TIME_STAMP(stamp, se) { (stamp) = se->invalid_after_time; }

#define TIME_STAMP_IS_STALE(stamp, se) ((stamp) < se->invalid_after_time)


/* PRIVATE EFG SHAPE FUNCTION TABLE */

#define INIT_SFTAB(sftab)                       \
   {                                            \
     INIT_TIME_STAMP((sftab).time_stamp);       \
     (sftab).num_of_entries = 0;                \
     (sftab).tot_of_entries = 0;                \
     (sftab).rows           = NULL;             \
   }

#define STALE_SFTAB(sftab) INIT_TIME_STAMP((sftab).time_stamp)

#define SFTAB_IS_STALE(sftab, se) (TIME_STAMP_IS_STALE((sftab).time_stamp, se))

#define SFTAB_IS_EMPTY(sftab) ((sftab).tot_of_entries == 0)

/* NOTE:
   Requiring full check means that the rays of existing SF tables are
   checked for intersections with the *NEW* triangles.  If there are
   no intersections, the SF table is deemed undamaged.
 */
#define DAMAGED_FULL_CHECK               TRUE
#define DAMAGED_INSIDE_INV_REGION_ONLY   FALSE

#define SFTAB_DAMAGED_FULL(sftab, se, loc) \
   (efg_pt_is_inside_invalid_region(se, loc, &sftab, DAMAGED_FULL_CHECK))

#define SFTAB_DAMAGED_INSIDE_INV_REGION_ONLY(sftab, se, loc)           \
   (efg_pt_is_inside_invalid_region(se, loc, &sftab,                   \
                                    DAMAGED_INSIDE_INV_REGION_ONLY))

#define SFTAB_INVALID(sftab, se, loc)                            \
   (SFTAB_IS_EMPTY(sftab) || SFTAB_DAMAGED_FULL(sftab, se, loc))


#define TOUCH_SFTAB(sftab, se) { TOUCH_TIME_STAMP((sftab).time_stamp, se); }

#ifdef USE_ESFLIB
#   define USE_SFTAB_SEARCH_FUNC(se, sftab)				\
  {									\
    if (!ESFL3_set_search_structure(se->esflib, (char *)&(sftab))) {	\
      REPORT_FATAL_ERROR(se->domain->errh, (char *)ESFL3_explain_error(se->esflib));      \
    }									\
    if (!ESFL3_set_search_funcs(se->esflib,				\
				NULL,					\
                                efgsf_first,		        	\
                                efgsf_next)) {		        	\
      REPORT_FATAL_ERROR(se->domain->errh, (char *)ESFL3_explain_error(se->esflib));      \
    }									\
  }
#   define USE_DEFAULT_SEARCH_FUNC(se)                                       \
  {                                                                          \
    if (!ESFL3_set_search_structure (se->esflib, (char *)se->efg_node_ss)) { \
      REPORT_FATAL_ERROR (se->domain->errh, (char *)ESFL3_explain_error(se->esflib));          \
    }                                                                        \
    if (!ESFL3_set_search_funcs (se->esflib,                                 \
                                 (ESFLIB_find_overlaps_func)                 \
                                 lobs_3d_search_for_boxes_overlapping_pt,    \
                                 (ESFLIB_first_overlapping_func)             \
                                 lobs_3d_first_overlapping_object,           \
                                 (ESFLIB_next_overlapping_func)              \
                                 lobs_3d_next_overlapping_object)) {         \
      REPORT_FATAL_ERROR (se->domain->errh, (char *)ESFL3_explain_error(se->esflib));          \
    }                                                                        \
  }
#else
#   define USE_SFTAB_SEARCH_FUNC(se, sftab)             \
  {                                                     \
    if (!ESP_3_set_search_structure((char *)&(sftab))) {\
      REPORT_FATAL_ERROR(se->domain->errh, ESP_3_explain_error());        \
    }                                                   \
    if (!ESP_3_set_search_funcs(NULL,                   \
                                efgsf_first,            \
                                efgsf_next)) {          \
      REPORT_FATAL_ERROR(se->domain->errh, ESP_3_explain_error());        \
    }                                                   \
  }
#   define USE_DEFAULT_SEARCH_FUNC(se)                          \
  {                                                             \
    if (!ESP_3_set_search_structure((char *)se->efg_node_ss)) { \
      REPORT_FATAL_ERROR(se->domain->errh, ESP_3_explain_error());                \
    }                                                           \
    if (!ESP_3_set_search_funcs(SEARCH_FUNC_NAME,		\
                                FIRST_IN_RANGE_FUNC_NAME,    	\
                                NEXT_IN_RANGE_FUNC_NAME)) {     \
      REPORT_FATAL_ERROR(se->domain->errh, ESP_3_explain_error());                \
    }                                                           \
  }
#endif



#define SFTAB_NUM_OF_ENTRIES(sftab) sftab.num_of_entries
#define SFTAB_ROW_PHI(row)          (row).phi
#define SFTAB_ROW_DPHI_X(row)       (row).dphix
#define SFTAB_ROW_DPHI_Y(row)       (row).dphiy
#define SFTAB_ROW_DPHI_Z(row)       (row).dphiz
#define SFTAB_ROW_NODE(row)         (row).node
#define SFTAB_PHI(sftab, k)         SFTAB_ROW_PHI(sftab.rows[k])
#define SFTAB_DPHI_X(sftab, k)      SFTAB_ROW_DPHI_X(sftab.rows[k])
#define SFTAB_DPHI_Y(sftab, k)      SFTAB_ROW_DPHI_Y(sftab.rows[k])
#define SFTAB_DPHI_Z(sftab, k)      SFTAB_ROW_DPHI_Z(sftab.rows[k])
#define SFTAB_NODE(sftab, k)        SFTAB_ROW_NODE(sftab.rows[k])


W_bool_t
efgsf_create_sftab(WEFG_superelem_t *se,
                   WEFG_sf_tab_t *sftab,
                   SFL_sf_table_t sf,
                   W_loc_t *at, W_bool_t allow_draw_rays);

SFL_sf_table_t
efgsf_sf_table(WEFG_superelem_t *se);

void
efgsf_eval_curr_loc(WEFG_sf_tab_t *sftab,
                    double *ptx, double *pty, double *ptz);

void
efgsf_eval_curr_vel(WEFG_sf_tab_t *sftab,
                    double *x, double *y, double *z);

void
efgsf_eval_curr_acc(WEFG_sf_tab_t *sftab,
                    double *x, double *y, double *z);

SFL_efg_node_id_t 
efgsf_first (ESFLIB_search_struct_t SS);


SFL_efg_node_id_t 
efgsf_next (ESFLIB_search_struct_t SS);


void
efghr_update_history (W_elem_t *e, W_solution_history_t *h,
                   W_float_t time);
void *
efghr_new_history_client_data (W_elem_t *e,
                              W_index_t idx, W_history_writer_t *r,
                              char *input_line);

double
efg_closest_node_dist (WEFG_superelem_t *se, WEFG_node_t *en);

#endif /* WEFGPRIV_H */
