#ifndef WEFGUTIL_H
#   define WEFGUTIL_H


#ifdef IX
#  undef IX
#  define IX v_index
#endif

#define VECT_DO(START, END, AN_EXPR)                        \
   {                                                        \
      register long int v_index = (START), v_end = (END);   \
                                                            \
      for (; v_index <= v_end; v_index++) { AN_EXPR; }      \
   }               

#define EFGVDO(SE, AN_EXPR) VECT_DO(1, SE->num_of_equations, AN_EXPR)  

/* Note LENGTH+1 ... the vectors are index by 1, ..., LENGTH!!! */
#define MAKE_VECTOR(PTR, EL_T, LENGTH)          \
  {                                             \
    PTR = w_new_array(LENGTH+1, EL_T);          \
    CHECK_MALLOC(PTR);                          \
    VECT_DO(1, LENGTH, (PTR)[IX] = 0);          \
  }

#define MAKE_EFGV(NAME, LENGTH) \
   { MAKE_VECTOR(NAME.v, W_float_t, LENGTH); NAME.dim = LENGTH+1; }

#define REALLOC_EFGV(NAME, NEW_LENGTH)                                  \
   {                                                                    \
      NAME.v = w_realloc_array(NAME.v, NEW_LENGTH+1, W_float_t);        \
      CHECK_MALLOC(NAME.v);                                             \
      NAME.dim = NEW_LENGTH+1;                                          \
    }



/* To access the 3-dimensional array of integration points ip[n][n][n], */
/* use the following macros.  */

#define LOOP_I LOIPA_i
#define LOOP_J LOIPA_j
#define LOOP_K LOIPA_k

#define IPA_LOOP(NIP)                                           \
   {                                                            \
     register int LOIPA_i, LOIPA_j, LOIPA_k;                        \
     register int LOIPA_in2, LOIPA_jn;                              \
     register int LOIPA_nip = NIP;                                  \
     int LOIPA_nip2 = LOIPA_nip * LOIPA_nip;                    \
                                                                \
     for (LOIPA_i = 0; LOIPA_i < LOIPA_nip; LOIPA_i++) {        \
       LOIPA_in2 = LOIPA_i * LOIPA_nip2;                        \
       for (LOIPA_j = 0; LOIPA_j < LOIPA_nip; LOIPA_j++) {      \
         LOIPA_jn = LOIPA_j * LOIPA_nip;                        \
         for (LOIPA_k = 0; LOIPA_k < LOIPA_nip; LOIPA_k++) {

#define IPA_LOOP_END()                          \
         }                                      \
       }                                        \
     }                                          \
   }

#define IPA_ENTRY(A) \
  *((A) + LOIPA_in2 + LOIPA_jn + LOIPA_k)

#ifdef ENABLE_TWO_POINT_RULE
#   define INC_2PT_RULE(EXPR) EXPR
#   define FLAG_2PT_RULE(SE, H)                         \
   W_bool_t use_2pt_rule = ((SE)->enable_2pt_quadrature \
                            ? ((H)->quad_order == -2    \
                               ? TRUE                   \
                               : FALSE)                 \
                            : FALSE)
#   define CHECK_2PT_CASE(SE)                                           \
    ((!use_2pt_rule)                                                    \
     || (   (LOOP_I == 0 && LOOP_J == 0 && LOOP_K == 0)                 \
         || (LOOP_I == 1 && LOOP_J == 1 && LOOP_K == 1)))
    /* Sets a flag. */
#else
#   define INC_2PT_RULE(EXPR) 
#   define FLAG_2PT_RULE(SE, H) int dummy__ = 0
#endif
/* The following macro does nothing when the 2pt rule is not compiled in; */
/* othewise a run-time flag in the superelement is checked. */
#define USE_QUADR_PT(SE) INC_2PT_RULE(if (CHECK_2PT_CASE(SE)))
#define INCL_EXCL_2PT_CASE(exec_for_2pt_case, exec_otherwise)           \
   {                                                                    \
     INC_2PT_RULE(                                                      \
                  if (use_2pt_rule) {                                   \
                    exec_for_2pt_case;                                  \
                  } else                                                \
                  )                                                     \
       {                                                                \
         exec_otherwise;                                                \
       }                                                                \
   }


#define LOOP_2D_I LOIPA_2D_i
#define LOOP_2D_J LOIPA_2D_j

#define IPA_2D_LOOP(NIP)                                                  \
   {                                                                      \
     register int LOIPA_2D_i, LOIPA_2D_j;                                 \
     register int LOIPA_2D_in;                                            \
     register int LOIPA_2D_nip = NIP;                                     \
                                                                          \
     for (LOIPA_2D_i = 0; LOIPA_2D_i < LOIPA_2D_nip; LOIPA_2D_i++) {      \
       LOIPA_2D_in = LOIPA_2D_i * LOIPA_2D_nip;                           \
       for (LOIPA_2D_j = 0; LOIPA_2D_j < LOIPA_2D_nip; LOIPA_2D_j++) {   

#define IPA_2D_LOOP_END()                                                \
       }                                                                 \
     }                                                                   \
   }

#define IPA_ENTRY_2D(A) \
  *((A) + LOIPA_2D_in + LOIPA_2D_j)

/* Example:
   
int
ptrs_var_opt(double *a, double *A)
{
  IPA_LOOP(n)
                {
                  if (IPA_ENTRY(a) - IPA_ENTRY(A) != 0)
                    fprintf(stderr, "mismatch\n");
                }
                IPA_LOOP_END();
}


 */

#include "WEFGpriv.h"

/* Assemble force to the effective-loads array */
#define ASSEMBLE_TO_EFF_LOAD(SE, EFGP, FRC) EFGN_ASSEMBLE_FORCE(EFGP, SE, FRC)

/* Assemble to mass */
#define ASSEMBLE_TO_MASS(SE, EFGP, M) efgn_assemble_mass(EFGP, SE, M)

/* Get from the update vector */
#define GET_FROM_UPDATE_VECT(SE, NE) ((SE)->update_vector.v[NE])

#endif /* WEFGUTIL_H */
