#ifndef WFEFG_H
#   define WFEFG_H

#include "Wtypes.h"
#include "prap3.h" /* Point-in-Range Package */
/* NOTE:  The PRaP library expects W_loc_t to consist of three */
/* double-precision numbers.  */
#define PRAP_REF(n) ((double *)(n))

/* ESPACK */
#include "espack3.h" 

/* G3D */
#include "g3d.h" 

#define ListOf(type) LIST 


struct WFEFG_node_t;


typedef struct WFEFG_sf_tab_row_t {
  struct WEFG_node_t  *node;
  double               phi, dphix, dphiy, dphiz;
}              WEFG_sf_tab_row_t;


typedef struct WFEFG_sf_tab_t {
  double                time_stamp;
  int                   tot_of_entries;
  int                   num_of_entries;
  WEFG_sf_tab_row_t    *rows;
}              WFEFG_sf_tab_t;



typedef struct WFEFG_vertex_t {
  W_index_t                     id;
  W_loc_t                       loc;
  struct WFEFG_node_t          *efgn;
  WFEFG_sf_tab_t                sftab;
  W_matstate_t                  matstate;
}              WFEFG_vertex_t;





#define MAX_VERT_PER_FACE 4     /* Supported: triangles, quadrilaterals */
typedef struct WFEFG_b_face_t {
  int            num_vert;
  WEFG_vertex_t *v[MAX_VERT_PER_FACE];
  G3D_box_t      bounding_box;
}              WFEFG_b_face_t;



  

typedef struct WFEFG_node_data_t {
  W_coords_3d_t            ref_coords;
  W_coords_3d_t            coords_t;
  W_coords_3d_t            coords_t_minus_dt;
  W_velocity_3d_t          velocity_t;
  W_acceleration_3d_t      acceleration_t;
  W_index_t                num_of_eqns;
  W_index_vector_t         eq_numbers;
}              WFEFG_node_data_t;


typedef enum WFEFG_efg_node_type_t{
  WFEFG_TRUE_EFG_NODE = 0,
  WFEFG_EFG_IF_NODE   = 1,
  WFEFG_FE_IF_NODE    = 2
             }  WFEFG_efg_node_type_t;


typedef struct WFEFG_node_t {
  int                      id;
  WFEFG_efg_node_type_t    type;
  float                    d_m;
  union {
    WFEFG_node_data_t   efgn;
    W_node_t           *fen;
  }                        data;
#ifdef X_XFACE
  GraphicObj              *node_marker;
#endif
}              WFEFG_node_t;


#define XDir 0
#define YDir 1
#define ZDir 2

  
typedef struct WFEFG_integration_point_t {
  WEFG_sf_tab_t                 sftab;
  W_matstate_t                  matstate;
}              WFEFG_integration_point_t;




typedef struct WFEFG_quadr_points_t {
  W_gauss_integration_rule_t     *qr;
  WFEFG_integration_point_t     **efg_ip; 
}              WFEFG_quadr_points_t;



typedef struct WFEFG_sys_vector_t {
  W_index_t      dim;           /* It is 1-based!!! */
  W_vector_t     v;
}              WFEFG_sys_vector_t;
  

  
typedef struct WFEFG_conc_load_t {
  W_index_t              id;
  W_loc_t                loc;
  W_vect_t               amplitudes;
  W_index_t              time_func_num;
  WFEFG_sf_tab_t         sftab;
}              WFEFG_conc_load_t;


typedef struct WFEFG_body_load_t {
  W_bool_t            is_set;
  W_bool_t            is_time_independent;
  W_bool_t            is_constant;
  char               *body_load_expr[3];
  W_vect_t            const_body_load;
  WFEFG_sys_vector_t  time_independent_body_load;
}              WFEFG_body_load_t;



typedef enum    WFEFG_distrib_load_type_t {
  WFEFG_DISTRIB_LOAD_DEAD_PRESSURE = 0,
  WFEFG_DISTRIB_LOAD_PRESSURE      = 1, /* follower */
  WFEFG_DISTRIB_LOAD_DEAD_SHEAR    = 2,
  WFEFG_DISTRIB_LOAD_SHEAR         = 3  /* follower */
             }  WFEFG_distrib_load_type_t;

  
typedef struct WFEFG_distrib_load_t {
  W_index_t                   id;
  WFEFG_distrib_load_type_t   type;
  struct {
    W_loc_t              loc;
    W_float_t            x, y;
    W_float_t            traction;
  }                          vdata[4];
  W_index_t                  time_func_num;
  int                        suggested_qr; /* quadrature order */
  W_vect_t                   u, v, normal; /* local csys on the patch */
  WFEFG_quadr_points_t      *qp; /* quadrature points for the patch */
  W_vect_t                   dir_vect; /* only for shear loads -- */
                                       /* gives with the normal the plane */
                                       /* to which the shear is parallel; */
                                       /* the vector /u/ is then computed */
                                       /* such that the shear is parallel */
                                       /* with it */
  double                     time_stamp;
}              WFEFG_distrib_load_t;

  
typedef enum WFEFG_contact_load_type_t {
  WEFG_CONTACT_LOAD_UNILATERAL = 0,
  WEFG_CONTACT_LOAD_STICKY     = 1  /* acts both ways */
             }  WFEFG_contact_load_type_t;

  
typedef struct WFEFG_contact_load_t {
  W_index_t                  id;
  WFEFG_contact_load_type_t  type;
  double                     contact_press_scaling;
  struct {
    W_loc_t                loc;
    W_float_t              x, y;
  WFEFG_sf_tab_t           sftab;
  }                          vdata[4];
  int                        suggested_qr; /* quadrature order */
  W_vect_t                   u, v, normal; /* local csys on the patch */
  WFEFG_quadr_points_t      *qp; /* quadrature points for the patch */
  G3D_plane_t                contact_plane;
  double                     time_stamp;
}              WFEFG_contact_load_t;

  
typedef struct WFEFG_hex_volume_t {
  W_index_t                       id;
  WFEFG_vertex_t                 *vertex[8];
  WFEFG_quadr_points_t           *qp;
  float                           time_stamp;
  int                             quad_order;
}              WFEFG_hex_volume_t;


typedef struct WFEFG_if_data_t {
  WFEFG_hex_volume_t *h;
  double              xi, eta, theta;
}              WFEFG_if_data_t;


  
typedef void *W_ptr_t;



struct WFEFG_crack_surf_methods_t;

typedef void *WFEFG_crack_surf_data_t;

struct WFEFG_superelem_t;

typedef struct WFEFG_crack_surf_t {
  struct WFEFG_superelem_t            *se;
  int                                  id;
  WFEFG_crack_surf_data_t              data;
  struct WFEFG_crack_surf_methods_t   *methods;
}              WFEFG_crack_surf_t;

 
typedef enum WFEFG_pt_pos_wrt_crack_surf_enum {
  WFEFG_NEGATIVE_NORMAL     = -1,
  WFEFG_ON_SURF             =  0,
  WFEFG_POSITIVE_NORMAL     =  1,
  WFEFG_AT_FRONT            =  2,
  WFEFG_POS_WRT_SURF_UNSPEC =  1000
              }  WFEFG_pt_pos_wrt_crack_surf_enum;


typedef WFEFG_pt_pos_wrt_crack_surf_enum
(*WFEFG_node_pos_wrt_crack_surf_proc)(WFEFG_crack_surf_t *cs,
                                      WFEFG_node_t *efgn,
                                      double physical_tol);


typedef struct WFEFG_eval_for_packet_t {
  struct WFEFG_superelem_t                 *se;
  WFEFG_pt_pos_wrt_crack_surf_enum          pos_of_eval_pt_wrt_crack_surf;
}              WFEFG_eval_for_packet_t;


typedef ESPACK_bool_t 
      (*WFEFG_segment_intersects_crack_surf_proc)(WFEFG_eval_for_packet_t *efp,
                                                  G3D_segment_t *s,
                                                  WFEFG_node_t *efgn,
                                                  WFEFG_crack_surf_t *cs,
                                                  double  tol);

typedef WFEFG_pt_pos_wrt_crack_surf_enum
(*WFEFG_point_pos_wrt_crack_surf_proc)(WFEFG_crack_surf_t *cs,
                                      W_loc_t *pt, double physical_tol);

typedef void (*WFEFG_modify_crack_surf_grep_proc)(struct WFEFG_superelem_t *se,
                                                 WFEFG_crack_surf_t *cs);

typedef void (*WFEFG_propagate_proc)(WFEFG_crack_surf_t *cs);


typedef void (*WFEFG_write_hexvue_postproc_proc)(WFEFG_crack_surf_t *cs,
                                                FILE *stream);

typedef void (*WFEFG_update_stress_proc)(struct WFEFG_superelem_t *se,
					WFEFG_crack_surf_t *cs);


typedef void (*WFEFG_write_ve3d_loclist_proc)(WFEFG_crack_surf_t *cs,
                                             FILE *stream);

typedef void (*WFEFG_write_ve3d_polys_proc)(WFEFG_crack_surf_t *cs,
                                             FILE *stream);


typedef struct WFEFG_crack_surf_methods_t {
  WFEFG_segment_intersects_crack_surf_proc    segment_intersects_crack_surf;
  WFEFG_node_pos_wrt_crack_surf_proc          node_pos_wrt_crack_surf;
  WFEFG_propagate_proc                        propagate;
  WFEFG_modify_crack_surf_grep_proc           modify_crack_surf_grep;
  WFEFG_write_hexvue_postproc_proc            write_hexvue_postproc;
  WFEFG_write_ve3d_loclist_proc               write_ve3d_loclist;
  WFEFG_write_ve3d_polys_proc                 write_ve3d_polys;
  WFEFG_update_stress_proc                    update_stress;
}              WFEFG_crack_surf_methods_t;


typedef  WFEFG_pt_pos_wrt_crack_surf_enum
   (*WFEFG_on_which_face_of_crack_surf_proc)
      (WFEFG_crack_surf_t *cs,
       W_loc_t *pt_on_surf,
       W_vect_t *normal_to_surf,
       W_ptr_t data);


    


#include "Wmat3d.h"


typedef struct WFEFG_efg_sf_stat_t {
  double  sum_nneighbors;
  int     nevals;
  int     min_nneighbors;
  int     max_nneighbors;
  double  avg_nneighbors;
}              WFEFG_efg_sf_stat_t;



typedef struct WFEFG_region_tri_t {
  W_bool_t     is_valid;
  G3D_tri_t    tri;
  G3D_box_t    bbox;
  G3D_plane_t  plane;
}              WFEFG_region_tri_t;

typedef struct WFEFG_region_t {
  W_bool_t           is_initialized;
  G3D_box_t          bbox;
  WFEFG_region_tri_t  triangles[2]; /* quads are decomposed into triangles */
}              WFEFG_region_t;



#define MIN_NUM_NEIGHBORS  4
#define D_M_MULT_FACTOR    1.1
#define DEFAULT_QUAD_ORDER 2

typedef struct WFEFG_superelem_t {
  W_domain_t                    *domain;
  /* Generate EFG nodes at this spacing. Input. */
  double                         node_spacing; 
  /* Merging distance. Either input, or computed as d_m/10. */
  double                         merging_distance; 
  /* Shortest dist. between EFG nodes Computed. */
  double                         shortest_distance;
  /* The number of nodes per hex volume side.  Generates anisotropic */
  /* spacing of EFG nodes, in case the hex is not a cube.  Specified */
  /* node spacing takes precedence. */
  int                            num_nodes_per_hex_side;
  /* Enforced d_m for all EFG nodes.Input. */
  double                         d_m; 
  /* Largest d_m allowed. Input. */
  double                         max_d_m; 
  /* Suggested number of neighbors.  Input. */
  int                            num_neighbors_suggested;
  /* Smallest number of neighbors.  Input. */
  int                            num_neighbors_min;
  /* Quadrature order in ALL hex volumes */
  int                            quadrature_order;
# ifdef ENABLE_TWO_POINT_RULE
  /* Is the two-point quadrature enabled?  */
  W_bool_t                       enable_2pt_quadrature;
# endif
  /* List of all vertices in the EFG superelement */
  struct dict_dictionary        *verts;
  /* Search structure for vertices */
  char                          *verts_ss;
  /* Estimate of the number of vertices (used to set up the vertex search */
  /* structure). */
  int                            vertex_ss_bucket_size;
  /* List of all top-level hex volumes in the EFG superelement.  Note: */
  /* The top-level hex volumes contribute the boundary faces which can */
  /* affect the visibility computation in shape func evaluation.  They */
  /* can be subdivided further for the purpose of numerical integration. */
  /* Their offspring shows the boundary faces in the graphical output, */
  /* but do not list these faces for the purpose of visibility computation. */
  ListOf(WFEFG_hex_volume_t *)    hex_volumes;
  /* List of all the cracks */
  ListOf(WFEFG_crack_surf_t *)    cracks;
  /* List of boundary faces.  See comment concerning the hex volumes above. */
  ListOf(WFEFG_b_face_t *)        b_faces;
  /* Is the topology valid (for crack propagation purposes). */
  W_bool_t                       topology_valid;
  /* Is the superelement domain convex? */
  W_bool_t                       is_convex;
  /* Is the superelement domain convex with cracks? */
  W_bool_t                       is_convex_with_cracks;
  /* The nodes are to be generated independently of the cells? */
  W_bool_t                       generate_nodes_in_volume;
  /* ...inside this volume. */
  W_loc_t                        lower_corner, upper_corner;
  /* Generate nodes in the interior of the integration cells, not at their */
  /* boundaries and vertices */
  W_bool_t                       generate_nodes_inside_only;
  /* List of all EFG nodes. */
  struct dict_dictionary        *efgnodes;
  /* Search structure for EFG nodes. */
  char                          *efg_node_ss;
/* Estimate of the number of vertices (used to set up the efg node search */
  /* structure). */
  int                            efg_node_ss_bucket_size;
  /* List of concentrated loads. */
  ListOf(WFEFG_conc_load_t *)     conc_loads;
  /* List of distributed loads. */
  ListOf(WFEFG_distrib_load_t *)  distrib_loads;
  /* List of contact loads. */
  ListOf(WFEFG_contact_load_t *)  contact_loads;
  /* Body load (valid in the whole EFG domain) */
  WFEFG_body_load_t               body_load;
  /* Material id.  The same for all hex volumes constituting this */
  /* superelement */
  W_mat_3d_t                    *mat;
  /* Number of equations at the EFG nodes. */
  W_index_t                      num_of_equations;
  /* Mass matrix. */
  WFEFG_sys_vector_t              mass_matrix;
  /* Total mass of the EFG superelement */
  W_float_t                      total_mass;
  /* Effective loads */
  WFEFG_sys_vector_t              effective_loads;
  /* Update vector. */
  WFEFG_sys_vector_t              update_vector;
  /* External load work. */
  W_float_t                      ext_load_work;
  /* Search tolerance -- no two efg nodes or vertices are expected */
  /* to be closer than that */
  double                         tolerance;
  /* Compute stresses at the vertices and other output points */
  W_bool_t                       compute_stress_at_output_pts;
  /* Compute equiv. plastic def. at the vertices and other output points */
  W_bool_t                       compute_equiv_pl_def_at_output_pts;
  /* EFG shape functions statistics */
  WFEFG_efg_sf_stat_t             efg_sf_stat;
  /* Time stamp for shape function validity; all points at which */
  /* the time stamp is earlier need to recompute if they are */
  /* in the inclusion */
  double                         invalid_after_time;
  /* The bounding boxes define a region in which the shape functions */
  /* may become invalid after the time invalid_after_time */
  ListOf(WFEFG_region_t *)        invalid_regions;
  /* Search structure for integration points */
  char                          *ip_ss;
  /* Estimate of the integration points (used to set up the search */
  /* structure). */
  int                            ip_ss_bucket_size;
  /* Store the integration points in a search structure */
  W_bool_t                       store_ip_in_search_structure;
  /* Force recomputation of all shape functions */
  W_bool_t                       force_all_sf;
  /* Draw rays point-node for shape function tables. */
  W_bool_t                       draw_rays;
}              WFEFG_superelem_t;




/* Graphics */
#define EFG_SURFACE_LAYER      0
#define EFG_NODE_LAYER         1
#define EFG_LOAD_LAYER         2


void
fefg_assemble_restoring_forces(W_elem_t *el);
void
fefg_assemble_mass(W_elem_t *el);
W_float_t 
fefg_suggested_time_step(W_elem_t *el);
#ifdef X_XFACE
void
fefg_modify_elem_grep(W_elem_t *el);
#endif
W_float_t
fefg_total_mass(W_elem_t *el);
void 
fefg_calc_eff_loads(W_elem_t *el);
void 
fefg_solve_for(W_elem_t *el);
void 
fefg_update_Warburton(W_elem_t *el);
void 
fefg_update_Newmark(W_elem_t *el);
NODE
fefg_new_history_client_data(W_elem_t *e, W_index_t idx, W_history_writer_t *r,
                            char *input_line);
void
fefg_update_history(W_elem_t *e, W_solution_history_t *h,
                   W_float_t time, W_bool_t should_flush);
void
fefg_update_history_stress(W_elem_t *e, W_solution_history_t *h);
void
fefg_apply_initial_conditions(W_elem_t *e);
void
fefg_write_postproc_file(W_elem_t *el, FILE *stream, int type);
void
fefg_print_info(W_elem_t *el, FILE *fp);

W_bool_t 
fefg_reader(W_domain_t *d, FILE *stream);

W_float_t 
fefg_finalize_step(W_elem_t *el);

char *
fefg_name(W_elem_t *el);

W_bool_t
fefg_configure(W_elem_t *el, char *option);

#ifdef WFEFG_PRIVATE_HEADER

W_elem_methods_t fefg_methods = {
  fefg_name,
  fefg_assemble_restoring_forces,
  NULL,
  NULL,
  fefg_assemble_mass,
  fefg_total_mass,
  NULL,
  fefg_suggested_time_step,
  NULL,
  fefg_calc_eff_loads,
  fefg_solve_for,
  fefg_update_Warburton,
  fefg_update_Newmark,
  fefg_finalize_step,
  fefg_new_history_client_data,
  fefg_update_history,
  fefg_apply_initial_conditions,
  fefg_write_postproc_file,
#ifdef X_XFACE
  fefg_modify_elem_grep,
#endif
  fefg_print_info,
  fefg_configure
};
  
#else  /* !WFEFG_PRIVATE_HEADER */

extern W_elem_methods_t fefg_methods;
#   undef WFEFG_PRIVATE_HEADER

#endif /* !WFEFG_PRIVATE_HEADER */


#endif
