#ifndef WDOMAIN_H
#   define WDOMAIN_H

#include "Wtypes.h"
#include "Wutil.h"
#include "Woptions.h"


W_node_t *
WD_get_node_by_idx(W_domain_t *d,
                     W_index_t idx
                     );
W_bool_t
WD_add_node(W_domain_t *d,
            W_index_t  idx,
            W_node_t   *node
            );
W_node_t *
WD_get_next_node (W_domain_t *d,
                  W_node_t *curr_node
                  );
W_elem_t *
WD_get_next_element(W_domain_t *d,
                    W_elem_t *curr_elem
                    );
W_elem_t *
WD_get_elem_by_idx(W_domain_t *d,
                     W_index_t idx
                     );
W_bool_t
WD_add_element(W_domain_t     *d,
               W_index_t      idx,
               W_elem_t       *elem
               );

W_bool_t
WD_add_time_func(W_domain_t      *d,
                 W_index_t       idx,
                 W_time_func_t   *time_func
                );

W_material_t *
WD_get_material_by_idx(W_domain_t *d,
                         W_index_t idx
                         );
/*
   Add material to the domain.
 */
W_bool_t/* TRUE if successful; FALSE otherwise */
WD_add_material (W_domain_t     *d, /* domain */
                W_index_t      idx, /* index */
                W_material_t   *material /* material */
                );
W_bool_t /* TRUE if successful; FALSE otherwise */
WD_add_disp_bc (W_domain_t *d, /* add a node to the domain */
                W_index_t  idx, /* index */
                W_disp_bc_t   *disp_bc /* node */
                );
/*
   Get an displacement BC by index.
 */
W_disp_bc_t * /* element or NULL if there is no such element index */
WD_get_disp_bc_by_idx (W_domain_t *d, /* domain */
                    W_index_t idx /* index */
                       );
W_disp_bc_t * /* node or NULL if there are no more */
WD_get_next_disp_bc (W_domain_t *d, /* domain */
                     W_disp_bc_t *curr /* node or NULL if starting */
                     );
W_bool_t /* TRUE if successful; FALSE otherwise */
WD_add_cont_surf (W_domain_t *d, /* domain */
                  W_index_t  idx, /* index */
                  W_contact_surf_t   *conts /* contact surface */
                  );
W_contact_surf_t * /* node or NULL if there are no more */
WD_get_next_cont_surf (W_domain_t *d, /* domain */
                     W_contact_surf_t  *curr /*  NULL if starting */
                     );
W_contact_surf_t * /* time function or NULL if there is no such index */
WD_get_cont_surf_by_idx (W_domain_t *d, /* domain */
                          W_index_t idx /* index */
                          ) ;

/*
   Create a new domain by reading an input file.
 */
W_domain_t * /* domain */
WD_new_domain (struct W_options_t *o);

void
WD_integrate_expl (W_domain_t *d /* integrate */
              );
void
WD_integrate_impl (W_domain_t *d /* integrate */
                       );
W_bool_t
WD_add_load(W_domain_t *d,
            W_load_t   *load
            );

W_float_t
WD_compute_dt(W_domain_t *d);

void
WD_eff_loads_expl (W_domain_t *d);

void
WD_solve_expl (W_domain_t *d);

void
WD_updt_cfg_Warb_expl(W_domain_t *d);

void
WD_updt_cfg_Newm_expl(W_domain_t *d);

W_float_t
WD_time(W_domain_t *d);

W_float_t /* time */
WD_time_step (W_domain_t  *d /* domain */
              );

W_time_func_t *
WD_get_time_func_by_idx(W_domain_t *d, W_index_t idx);
W_time_func_t * /* node or NULL if there are no more */
WD_get_next_time_func (W_domain_t *d, /* domain */
                     W_time_func_t *curr /* node or NULL if starting */
                     );
W_float_t
WD_total_mass(W_domain_t *d);

void
WD_move_domain_by(W_domain_t *d, W_coords_3d_t *by);

#ifdef X_XFACE
void
WD_build_grep (W_domain_t *d /* domain */,
               W_bool_t immed_draw
               );
#endif

void
WD_assemble_into_load_vect(W_domain_t *d, W_index_t ne, W_float_t val);

W_float_t
WD_get_from_update_vect(W_domain_t *d, W_index_t ne);

W_float_t
WD_get_from_mass_matrix(W_domain_t *d, W_index_t ne);

void
WD_assemble_into_mass(W_domain_t *d, W_index_t ne, W_float_t val);


W_float_t
WD_current_deformation_energy(W_domain_t *d);

W_float_t
WD_current_kinetic_energy(W_domain_t *d);

W_float_t
WD_current_ext_load_work(W_domain_t *d);

void
WD_update_history(W_domain_t *d);

void
WD_register_elem_for_integration(W_domain_t *d, W_elem_t *e);

void
WD_apply_initial_conditions(W_domain_t *d);

void
WD_write_postfile(W_domain_t *d, double simulation_time, double label_time);

void
WD_initial_velocity(W_domain_t *d, W_loc_t *at, W_loc_t *v_initial);

void
WD_print_info(W_domain_t *d, FILE *stream);

void
WD_print_elem_info(W_domain_t *d,
                   W_index_t elid,
                   W_info_spec_t spec, 
                   FILE *stream);

void
WD_print_node_info(W_domain_t *d,
                   W_index_t nid,
                   W_info_spec_t spec, 
                   FILE *stream);

W_bool_t
WD_clean_up_before_exit(W_domain_t *d);

void
WD_flush_history(W_domain_t *d);


void
WD_impl_initialize(W_domain_t *d);

void
WD_impl_solve_for_accel(W_domain_t *d);

void
WD_impl_update_prediction(W_domain_t *d);

void
WD_impl_calc_forces(W_domain_t *d);

W_bool_t 
WD_impl_has_converged(W_domain_t *d);

W_float_t 
WD_finalize_expl(W_domain_t *d);

void
WD_impl_set_avd(W_domain_t *d, W_index_t ne, double a, double v);

W_bool_t
WD_probe(W_domain_t *d,
	 void *options,
	 W_loc_t *loc,
	 unsigned long spec,
	 W_probe_package_t *result
	 );

double
WD_mass_damping_coeff(W_domain_t *d);


double
WD_mass_damping_coeff_time_dep(W_domain_t *d);

LIST_OF (WCF_contact_face_t *)
WD_get_list_of_contact_faces (W_domain_t *d);

W_bool_t /* TRUE if successful; FALSE otherwise */
WD_add_iv_cond (W_domain_t *d, /* add a node to the domain */
                W_index_t  idx, /* index */
                W_initial_velocity_cond_p iv_cond /* node */
                );
/*
   Iterate displacement boundary conditions.
   Supply curr_node = NULL, if starting iteration.
 */
W_initial_velocity_cond_p /* node or NULL if there are no more */
WD_get_next_iv_cond (W_domain_t *d, /* domain */
                     W_initial_velocity_cond_p curr /* node or NULL if starting */
                     );

W_float_t /* total mass */
WD_avg_nodal_mass (W_domain_t *d /* domain */);

void
WD_resid_force_Newm_impl (
                          W_domain_t *d, /* domain */
                          W_newmark_assemble_residual_force_proc assemble_residual_force, /* callback into the Newmark integrator to assemble the residual force */
                          W_mat_state_update_form_enum update_form, /* update material
                                                                       permanently or
                                                                       temporarily? */
                          W_newmark_get_incr_proc newmark_get_incr, /* displacement
                                                                       increment access callback
                                                                    */
                          W_newmark_predictor_enum newmark_predictor, /* predictor to use */
                          double *extf_norm, /* as a side-effect, compute also
                                                the norm of the external forces
                                             */
                          double *inertf_norm, /* and the norm of the inertial forces */
                          double *resid_norm /* residual norm */
                          );



void
WD_updt_cfg_Newm_impl (W_domain_t *d, W_newmark_get_incr_proc newmark_get_incr,
                      W_newmark_predictor_enum newmark_predictor);

W_float_t
WD_finalize_Newm_impl (W_domain_t *d, W_newmark_incr_disp_proc incr_disp_callback);

int
WD_tot_of_equations (W_domain_t *d);

void
W_domain_set_response_observer_func (W_domain_t *d,
                                     W_response_observer_func response_observer_func);

void
WD_write_named_postfile (W_domain_t *d /* domain */,
                         char *out, 
                         int type
                   );

void
WD_set_interrupt_func (W_domain_t *d, W_check_interrupt_func ifunc);

void 
WD_eff_stiff_Newm_impl (W_domain_t *d, /* domain */
                        W_newmark_assemble_eff_stiff_proc assemble_eff_stiff_callback,
                        W_newmark_get_incr_proc newmark_get_incr, /* displacement
                                                                       increment access callback
                                                                    */
                          W_newmark_predictor_enum newmark_predictor, /* predictor to use */
                        W_float_t stiff_mult,
                        W_float_t mass_mult);

double /* value */
WD_curr_mass_damping_coeff (W_domain_t *d /* domain */
                            );

void
WD_solve_for_modr_expl (W_domain_t *d /* domain */);

void
WD_start_step_Newm_impl (W_domain_t *d /* domain */
                         );
    

void
W_domain_add_finalize_callback(W_domain_t *d,
				void *client,
				W_finalize_callback_func callback);

void
WD_summary(W_domain_t *d);

#endif /* !WDOMAIN_H */
