#ifndef WELEM_H
#   define WELEM_H

#include "Wtypes.h"
#include "Wutil.h"

#define ZERO_ELEM_VECT_DATA(VECT_DATA, LENGTH)   \
   {                                             \
      W_index_t i, l = (LENGTH);                 \
      VECT_DATA->local_vect_dim = l;             \
      for (i = 0; i < l; i++) {                  \
        VECT_DATA->localization_indices[i] = 0;  \
        VECT_DATA->data_vector[i] = 0;           \
      }                                          \
   }

W_elem_t *
WE_new_element(W_domain_t *d, 
               W_elem_methods_t *methods,
               void *data
#ifdef PVM_NEED_GHOST_ELEMENTS
               , W_bool_t is_ghost
#endif
               );


int
WE_elem_elcode (W_elem_t *el);

const char *
WE_elem_type_name (W_elem_t *el);

int
WE_elem_elcode (W_elem_t *el);

void
WE_assemble_restoring_forces (W_elem_t *el,
                              W_mat_state_update_form_enum update_form);

W_bool_t
W_element_reader (W_domain_t *d);

void
WE_assemble_mass (W_elem_t *el);

W_float_t
WE_elem_total_mass (W_elem_t *el);

W_float_t 
WE_suggested_time_step (W_elem_t *el);

void
WE_modify_elem_grep (W_elem_t *el, double displ_scale, W_bool_t immed_draw);

W_float_t
WE_deformation_energy (W_elem_t *el);

W_float_t /* kinetic energy or zero if method not specialized */
WE_kinetic_energy (W_elem_t *el /* element */,
                   W_mass_type_enum mass_type /* mass type */
                   );
void 
WE_calc_eff_loads (W_elem_t *el);

void 
WE_solve_for (W_elem_t *el);

void 
WE_update_config_Warburton_variant (W_elem_t *el);

void 
WE_update_config_Newmark_variant (W_elem_t *el);

void *
WE_new_history_client_data (W_elem_t *el, W_index_t idx,
                            W_history_writer_t *r,
                            char *input_line);

void
WE_update_history (W_elem_t *el,
                   W_solution_history_t *h,
                   W_float_t time);


void
WE_apply_initial_conditions (W_elem_t *e);

void
WE_write_postproc_file (W_elem_t *el,
                        ckit_iobuf_t IOB,
                        int type,
                        char *mode,
                        void *optdata);


void
WE_print_info (W_elem_t *el, W_info_spec_t spec, FILE *stream, int indent);

W_float_t
WE_finalize_step (W_elem_t *el);

W_bool_t
WE_configure (W_elem_t *el, char *option);

W_bool_t 
WE_probe (W_elem_t *el, void *options,
          W_loc_t *loc, unsigned long spec,
          W_probe_package_t *result);


void 
WE_initialize_readers (void);

void
WE_register_reader (int elcode, W_elem_reader_func reader);

int 
WE_num_of_faces (W_elem_t *el);

int /* number of nodes supplied in the array; if the array is not
     large enough, the desired dimension is returned. */
WE_get_face_vertex_nodes (W_elem_t *el, /* element */
                          int nface, /* face number */
                          int ndim, /* dimension of the below array; if the
                                       dimension is not big enough to hold
                                       all of the face nodes, the appropriate
                                       dimension is returned and no data
                                       are placed in the array.
                                    */
                          W_node_t *n[] /* the array to hold pointers
                                         to nodes on the face */
                          );
int
WE_get_contact_faces (W_elem_t *el, int nface, WCF_contact_face_t *cf[], int maxcf);

/*
   Assemble the residual nodal forces.
   Assemble the residual nodal forces by adding the contribution of
   this element.
 */
void
WE_assemble_residual_forces (W_elem_t *el,
                             W_mat_state_update_form_enum update_form,
                             W_mass_type_enum mass_type);

void
WE_assemble_eff_stiffness (W_elem_t *el/* element */,
                           W_newmark_assemble_eff_stiff_proc assemble_eff_stiff_callback, /* callback to use to actually assemble the submatrix into the system matrix; this has to be supplied by the domain */
                           W_float_t stiff_mult, /* multiply stiffness by this; may be zero, in which case it indicates that no stiffness should be included */
                           W_float_t mass_mult, /* multiply mass by this; may be zero, in which case no mass matrix should be included */
                           W_mass_type_enum mass_type /* type of the mass matrix (consistent, diagonal, ...) */
                           );

W_bool_t
WE_get_elem_prop (W_elem_t *el, const char *prop, void *value);

int
WE_get_num_nodes (W_elem_t *el);

W_node_t *
WE_get_node(W_elem_t *el, int j);

int 
WE_get_num_matstates(W_elem_t *el);

void *
WE_get_matstate(W_elem_t *el, int j);

int 
WE_get_material_id(W_elem_t *el);

#endif /* !WELEM_H */
