#ifndef WERRH_H
#define WERRH_H 

#define E_K_FATAL   'f'
#define E_K_ERROR   'e'
#define E_K_WARNING 'w'

void
W_g3d_error_callback(void);

struct W_error_handler_t *
WEH_new_handler (char *err_log_name);

int /* ignored */
WEH_error_msg (struct W_error_handler_t *eh,
             char *msg /* message */
             );
/*
   Start an error message.
   Print the initial part of an error message.
   This should be followed by WEH_error_msg() and WEH_error_end().
 */
int /* ignored */
WEH_error_start (struct W_error_handler_t *eh, /* error handler */
                 char *file, /* file name where error occurred */
                 int line, /* line in the file /file/ */
                 int kind /* severity of the error; default is E_K_FATAL */
                 );
int
WEH_error_end (struct W_error_handler_t *eh,
               int kind /* severity of the error (or warning) */
               );

int 
WEH_error (struct W_error_handler_t *eh,
           char *file, int line, int kind, char *explanation);

const char *
WEH_read_failure (struct W_error_handler_t *eh, int curr_line_num, char *input_line);

int
WEH_parse_error (struct W_error_handler_t *eh,
               char *file, /* file where error occurred */
               int line, /* line in the file */
               int tok_pos, /* token position */
               tokens_token_type expect /* expected token value */,
                 int curr_line_num
               );
const char *
WEH_unrecognized_input_error (struct W_error_handler_t *eh, char *file, int line,
                              int curr_line_num, char *input_line);


#define REPORT_WARNING_START(eh) \
   { WEH_error_start (eh, __FILE__, __LINE__, E_K_WARNING); }
#define REPORT_ERROR_START(domain) \
   { WEH_error_start (eh, __FILE__, __LINE__, E_K_ERROR); }
#define REPORT_FATAL_ERROR_START(domain) \
   { WEH_error_start (eh, __FILE__, __LINE__, E_K_FATAL); }

#define REPORT_WARNING_END(eh) \
   { WEH_error_end (eh, E_K_WARNING); }
#define REPORT_ERROR_END() \
   { WEH_error_end (eh, E_K_ERROR); }
#define REPORT_FATAL_ERROR_END() \
   { WEH_error_end (eh, E_K_FATAL); }

#define REPORT_WARNING(eh, w) {WEH_error(eh, __FILE__, __LINE__, E_K_WARNING, w);}

#define REPORT_ERROR(eh, err) {WEH_error(eh, __FILE__, __LINE__, E_K_ERROR, err);}

#define REPORT_FATAL_ERROR(eh, err) {WEH_error(eh, __FILE__, __LINE__, E_K_FATAL, err);}

#endif
