#ifndef WMAT2D_H
#   define WMAT2D_H

/* RCSHEADER "$Id: Wmat2d.h,v 1.2 2002/03/21 01:32:35 cvs Exp $" */

typedef void *W_mat_2d_state_t;

struct W_mat_2d_ip_ctx_t;

typedef W_mat_3d_kinem_var_type_enum W_mat_2d_kinem_var_type_enum;

typedef enum W_mat_2d_mat_const_enum {
  MAT_2D_YOUNG_MODULUS                  = 0,
  MAT_2D_POISSON                        = 1,
  MAT_2D_SHEAR_MODULUS                  = 2,
  MAT_2D_BULK_MODULUS                   = 3,
  MAT_2D_TENSILE_STRENGTH               = 4,
  MAT_2D_DYN_FRACTURE_TOUGHNESS         = 5,
  MAT_2D_DIL_WAVE_SPEED_2D              = 6,
  MAT_2D_DIL_WAVE_SPEED_PSTRAIN         = 7,
  MAT_2D_DIL_WAVE_SPEED_PSTRESS         = 8,
  MAT_2D_SHEAR_WAVE_SPEED               = 9,
  MAT_2D_RAYLEIGH_PSTRAIN_WAVE_SPEED    = 10,
  MAT_2D_RAYLEIGH_PSTRESS_WAVE_SPEED    = 11,
  MAT_2D_INITIAL_YIELD_STRESS           = 12
}            W_mat_2d_mat_const_enum;

typedef enum W_mat_2d_stiff_enum {
  MAT_2D_STIFF_LAGRANGIAN = 0,
  MAT_2D_STIFF_EULERIAN   = 1
}            W_mat_2d_stiff_enum;
  
typedef W_bool_t (*WM2D_get_kinem_var_proc)(struct W_mat_2d_ip_ctx_t *ctx,
                                            W_mat_2d_kinem_var_type_enum  type,
                                            double kinvar[3][3]);

typedef struct W_mat_2d_ip_ctx_t {
  double                     time;
  double                     time_step;
  void                      *ctx_data;
  WM2D_get_kinem_var_proc    get_kinem_var; 
}              W_mat_2d_ip_ctx_t; 

struct W_mat_2d_t;

typedef W_mat_2d_state_t (*WM2D_get_mat_state_proc)(struct W_mat_2d_t *m,
                                                    W_loc_t *at_point);

typedef W_bool_t 
   (*WM2D_get_mat_const_proc)(struct W_mat_2d_t *m,
                               W_mat_2d_mat_const_enum type,
                               double *c);

typedef W_bool_t 
    (*WM2D_update_stress_proc)(struct W_mat_2d_t           *m,
                               W_mat_2d_state_t             ms,
                               W_mat_state_update_form_enum update_form, 
                               W_mat_2d_ip_ctx_t           *ip_ctx,
                               W_mat_2d_stress_type_enum    stress_type,
                               double                       stress[3][3],
                               W_mat_2d_stress_type_enum   *out_stress_type
                               );


typedef W_bool_t 
    (*WM2D_get_mat_state_var_proc)(struct W_mat_2d_t           *m,
                                   W_mat_2d_state_t             ms,
                                   W_mat_2d_ip_ctx_t           *ip_ctx,
                                   W_mat_2d_mat_state_var_enum  var_type,
                                   double                      *var
                                   );

typedef W_bool_t (*WM2D_get_mat_stiff_func) (struct W_mat_2d_t *m,
                                             W_mat_2d_state_t ms,
                                             W_mat_2d_ip_ctx_t *ip_ctx,
                                             W_mat_2d_stiff_enum stiff_type,
                                             double C[6][6] 
                                             );

typedef struct WM2D_mat_state_pkg_t {
  W_mat_2d_state_t        ms;
  double                  weight;
}              WM2D_mat_state_pkg_t;

typedef W_mat_2d_state_t (*WM2D_interpolate_mat_state_proc)
                                (struct W_mat_2d_t *m,
				 W_loc_t *at_point,
				 int npkgs,
				 WM2D_mat_state_pkg_t *pkgs);
				 
typedef int (*WM2D_get_mcode_func) (struct W_mat_2d_t *m);

typedef struct W_mat_2d_methods_t {
  WM2D_get_mcode_func                     get_mcode;
  WM2D_get_mat_state_proc                 get_mat_state;
  WM2D_get_mat_const_proc                 get_mat_const;
  WM2D_update_stress_proc                 update_stress;
  WM2D_get_mat_state_var_proc             get_mat_state_var;
  WM2D_interpolate_mat_state_proc         interpolate_mat_state;
  WM2D_get_mat_stiff_func                 get_mat_stiff;
}              W_mat_2d_methods_t;

typedef struct W_mat_2d_t {
  W_float_t                       vol_mass;
  W_mat_2d_methods_t             *methods;
  void                           *data;
#ifdef X_XFACE
  EPixel                          color;
  EFringeTable                    fringes;
#endif
}              W_mat_2d_t;


typedef W_mat_2d_t * (*W_mat2d_reader_func) (W_domain_t *d);

void 
WM2D_initialize_readers (void);


/*
   Register a mat2d reader.
   To be able to use specific implementation
   of the general triaxial material (mat2d),
   its code needs to be linked to the
   executable, and its reader needs to be registered, so that the generic
   mat2d reader can look up the type-specific reader.
   */
void
WM2D_register_reader (int m2dcode /* unique integer among all elements */,
                      W_mat2d_reader_func reader /* the reader function */
                      );

W_mat_2d_t *
mat2d_reader (W_domain_t *d);

W_mat_2d_t *
WM2D_new_material(W_mat_2d_methods_t *methods,
                  void *data,
                  double vol_mass);

W_mat_2d_state_t 
WM2D_get_mat_state(W_mat_2d_t *m, W_loc_t *at_point);

W_mat_2d_state_t 
WM2D_interpolate_mat_state(W_mat_2d_t *m,
			   W_loc_t *at_point,
			   int npkgs,
			   WM2D_mat_state_pkg_t *pkgs);

W_bool_t 
WM2D_get_mat_const(W_mat_2d_t *m, W_mat_2d_mat_const_enum type, double *c);

W_bool_t 
WM2D_update_stress(W_mat_2d_t               *m,
                   W_mat_2d_state_t             ms,
                   W_mat_state_update_form_enum update_form,
                   W_mat_2d_ip_ctx_t           *ip_ctx,
                   W_mat_2d_stress_type_enum    stress_type,
                   double                       stress[3][3],
                   W_mat_2d_stress_type_enum   *out_stress_type
                   );

W_bool_t 
WM2D_get_mat_state_var(W_mat_2d_t                  *m,
                       W_mat_2d_state_t             ms,
                       W_mat_2d_ip_ctx_t           *ip_ctx,
                       W_mat_2d_mat_state_var_enum  var_type,
                       double                      *var
                       );

double 
WM2D_get_volume_mass(W_mat_2d_t *m);

W_bool_t
WM2D_convert_stress(double F[3][3],
                     W_mat_2d_stress_type_enum    in_stress_type,
                     double                       in_stress[3][3],
                     W_mat_2d_stress_type_enum    out_stress_type,
                     double                       out_stress[3][3]
                     );

/* The stress tensor must be symmetric; otherwise the results are */
/* unpredictable.  NO CHECK is performed. */
W_bool_t
WM2D_principal_stresses(double stress[3][3],
                        double principal_stress[3],
                        W_vect_t *n0, W_vect_t *n1, W_vect_t *n2
                        );


int
WM2D_get_mcode (W_mat_2d_t *m);

/*
  Compute material stiffness tensor.
  Compute material stiffness tensor in the reference configuration.
 */
W_bool_t 
WM2D_get_mat_stiff (W_mat_2d_t *m/* mat2d */,
                    W_mat_2d_state_t             ms,/* material state */
                    W_mat_2d_ip_ctx_t           *ip_ctx,/* integration point context;
                                                           used to generate kinematic
                                                           quantities which the material state
                                                           might need in order to compute
                                                           the requested stress */
                    W_mat_2d_stiff_enum stiff_type, /* type of material stiffness matrix */
                    double C[6][6] /* material stiffness tensor in matrix form */
                        );

#endif /* WMAT2D_H */
