#ifndef WMAT3D_H
#   define WMAT3D_H

#include "Wmat3dutil.h"
#include "Whistry.h"

/* RCSHEADER "$Id: Wmat3d.h,v 1.9 2005/02/14 02:47:23 cvs Exp $" */

struct W_mat_3d_state_t;
typedef struct W_mat_3d_state_t *W_mat_3d_state_p;

struct W_mat_3d_ip_ctx_t;

typedef enum W_mat_3d_kinem_var_type_enum {
  MAT_3D_DEF_GRAD                =  0,
  MAT_3D_REL_DEF_GRAD            =  1,
  MAT_3D_MIDSTEP_DEF_GRAD        =  2,
  MAT_3D_CAUCHY_GREEN            =  3,
  MAT_3D_RATE_OF_DEF             =  4,
  MAT_3D_SPIN                    =  5,
  MAT_3D_MIDSTEP_RATE_OF_DEF     =  6,
  MAT_3D_MIDSTEP_SPIN            =  7,
  MAT_3D_GREEN_LAGRANGE          =  8,
  MAT_3D_INV_DEF_GRAD            =  9,
  MAT_3D_POLAR_U                 = 10,
  MAT_3D_POLAR_V                 = 11,
  MAT_3D_POLAR_R                 = 12,
  MAT_3D_POLAR_REL_U             = 13,
  MAT_3D_POLAR_REL_V             = 14,
  MAT_3D_POLAR_REL_R             = 15,
  MAT_3D_LN_POLAR_REL_U          = 16,
  MAT_3D_PREV_DEF_GRAD           = 17,
  MAT_3D_DEF_GRAD_DOT            = 18,
  MAT_3D_INCR_DISPL_GRAD         = 19,  /* gradient of the displacement increment in step */
  MAT_3D_SYMM_DISPL_GRAD         = 20
}            W_mat_3d_kinem_var_type_enum;

typedef enum W_mat_3d_stress_type_enum {
  MAT_3D_ANY_STRESS     = -1,
  MAT_3D_CAUCHY         = 0,
  MAT_3D_1ST_PK         = 1,
  MAT_3D_2ND_PK         = 2
}            W_mat_3d_stress_type_enum;

typedef enum W_mat_3d_mat_state_var_enum {
  MAT_3D_EQUIV_PLASTIC_DEF         = 0,
  MAT_3D_DEFORM_ENERGY_DENSITY     = 1,
  MAT_3D_PRINCIPAL_STRETCH_MAX     = 2,
  MAT_3D_PRINCIPAL_STRETCH_MID     = 3,
  MAT_3D_PRINCIPAL_STRETCH_MIN     = 4
}            W_mat_3d_mat_state_var_enum;

typedef enum W_mat_3d_mat_const_enum {
  MAT_3D_YOUNG_MODULUS                        = 0,
  MAT_3D_POISSON                              = 1,
  MAT_3D_SHEAR_MODULUS                        = 2,
  MAT_3D_BULK_MODULUS                         = 3,
  MAT_3D_TENSILE_STRENGTH                     = 4,
  MAT_3D_DYN_FRACTURE_TOUGHNESS               = 5,
  MAT_3D_DIL_WAVE_SPEED_3D                    = 6,
  MAT_3D_DIL_WAVE_SPEED_PSTRAIN               = 7,
  MAT_3D_DIL_WAVE_SPEED_PSTRESS               = 8,
  MAT_3D_SHEAR_WAVE_SPEED                     = 9,
  MAT_3D_RAYLEIGH_PSTRAIN_WAVE_SPEED          = 10,
  MAT_3D_RAYLEIGH_PSTRESS_WAVE_SPEED          = 11,
  MAT_3D_INITIAL_YIELD_STRESS                 = 12,
  MAT_3D_SMA_MARTENSITE_INITIAL_YIELD_STRESS  = 13,
  MAT_3D_SMA_AUSTENITE_INITIAL_YIELD_STRESS   = 14
}            W_mat_3d_mat_const_enum;

typedef enum W_mat_3d_stiff_enum {
  MAT_3D_STIFF_LAGRANGIAN = 0,
  MAT_3D_STIFF_EULERIAN   = 1
}            W_mat_3d_stiff_enum;
  
typedef W_bool_t (*WM3D_get_kinem_var_proc)(struct W_mat_3d_ip_ctx_t *ctx,
                                            W_mat_3d_kinem_var_type_enum  type,
                                            double kinvar[3][3]);

typedef struct W_mat_3d_ip_ctx_t {
  double                     time;
  double                     time_step;
  void                      *ctx_data;
  WM3D_get_kinem_var_proc    get_kinem_var; 
}              W_mat_3d_ip_ctx_t; 

struct W_mat_3d_t;

typedef W_mat_3d_state_p (*WM3D_get_mat_state_proc)(struct W_mat_3d_t *m,
                                                    W_loc_t *at_point);

typedef W_bool_t 
   (*WM3D_get_mat_const_proc)(struct W_mat_3d_t *m,
                               W_mat_3d_mat_const_enum type,
                               double *c);

typedef W_bool_t 
    (*WM3D_update_stress_proc)(struct W_mat_3d_t           *m,
                               W_mat_3d_state_p             ms,
                               W_mat_state_update_form_enum update_form, 
                               W_mat_3d_ip_ctx_t           *ip_ctx,
                               W_mat_3d_stress_type_enum    stress_type,
                               double                       stress[3][3],
                               W_mat_3d_stress_type_enum   *out_stress_type
                               );


typedef W_bool_t 
    (*WM3D_get_mat_state_var_proc)(struct W_mat_3d_t           *m,
                                   W_mat_3d_state_p             ms,
                                   W_mat_3d_ip_ctx_t           *ip_ctx,
                                   W_mat_3d_mat_state_var_enum  var_type,
                                   double                      *var
                                   );

typedef W_bool_t 
(*WM3D_get_mat_state_named_var_proc)(struct W_mat_3d_t           *m,
                                     W_mat_3d_state_p             ms,
                                     W_mat_3d_ip_ctx_t           *ip_ctx,
                                     char *name, 
                                     double                      *var
                                     );

typedef W_bool_t (*WM3D_get_mat_stiff_func) (struct W_mat_3d_t *m,
                                             W_mat_3d_state_p ms,
                                             W_mat_3d_ip_ctx_t *ip_ctx,
                                             W_mat_3d_stiff_enum stiff_type,
                                             double C[6][6] 
                                             );

typedef struct WM3D_mat_state_pkg_t {
  W_mat_3d_state_p        ms;
  double                  weight;
}              WM3D_mat_state_pkg_t;

typedef W_mat_3d_state_p (*WM3D_interpolate_mat_state_proc)
                                (struct W_mat_3d_t *m,
				 W_loc_t *at_point,
				 int npkgs,
				 WM3D_mat_state_pkg_t *pkgs);
				 
typedef int (*WM3D_get_mcode_func) (struct W_mat_3d_t *m);

typedef void (*WM3D_update_history_proc) (struct W_mat_3d_t *m,
                                          W_mat_3d_state_p  ms,
                                          W_mat_3d_ip_ctx_t *ip_ctx,
                                          W_solution_history_t *h, 
                                          W_index_t pntidx,
                                          W_float_t time
                                          );
typedef W_bool_t (*WM3D_get_mat_stiff_9x9_func) (struct W_mat_3d_t *m,
                                             W_mat_3d_state_p ms,
                                             W_mat_3d_ip_ctx_t *ip_ctx,
                                             W_mat_3d_stiff_enum stiff_type,
                                             double C[9][9]
                                             );
typedef struct W_mat_3d_methods_t {
  WM3D_get_mcode_func                     get_mcode;
  WM3D_get_mat_state_proc                 get_mat_state;
  WM3D_get_mat_const_proc                 get_mat_const;
  WM3D_update_stress_proc                 update_stress;
  WM3D_get_mat_state_var_proc             get_mat_state_var;
  WM3D_interpolate_mat_state_proc         interpolate_mat_state;
  WM3D_get_mat_stiff_func                 get_mat_stiff;
  WM3D_get_mat_state_named_var_proc       get_mat_state_named_var;
  WM3D_update_history_proc                update_history;
  WM3D_get_mat_stiff_9x9_func             get_mat_stiff_9x9;  
}              W_mat_3d_methods_t;

typedef struct W_mat_3d_t {
  W_material_t			  *mat;
  W_float_t                        vol_mass;
  W_mat_3d_methods_t              *methods;
  void                            *data;
#ifdef X_XFACE
  EPixel                           color;
  EFringeTable                     fringes;
#endif
}              W_mat_3d_t;


typedef W_mat_3d_t * (*W_mat3d_reader_func) (W_domain_t *d);

void 
WM3D_initialize_readers (void);


/*
   Register a mat3d reader.
   To be able to use specific implementation
   of the general triaxial material (mat3d),
   its code needs to be linked to the
   executable, and its reader needs to be registered, so that the generic
   mat3d reader can look up the type-specific reader.
   */
void
WM3D_register_reader (int m3dcode /* unique integer among all elements */,
                      W_mat3d_reader_func reader /* the reader function */
                      );

W_mat_3d_t *
mat3d_reader (W_domain_t *d);

W_mat_3d_t *
WM3D_new_material(W_mat_3d_methods_t *methods,
                  void *data,
                  double vol_mass);

W_mat_3d_state_p 
WM3D_get_mat_state(W_mat_3d_t *m, W_loc_t *at_point);

W_mat_3d_state_p 
WM3D_interpolate_mat_state(W_mat_3d_t *m,
			   W_loc_t *at_point,
			   int npkgs,
			   WM3D_mat_state_pkg_t *pkgs);

W_bool_t 
WM3D_get_mat_const(W_mat_3d_t *m, W_mat_3d_mat_const_enum type, double *c);

W_bool_t 
WM3D_update_stress(W_mat_3d_t               *m,
                   W_mat_3d_state_p             ms,
                   W_mat_state_update_form_enum update_form,
                   W_mat_3d_ip_ctx_t           *ip_ctx,
                   W_mat_3d_stress_type_enum    stress_type,
                   double                       stress[3][3],
                   W_mat_3d_stress_type_enum   *out_stress_type
                   );

W_bool_t 
WM3D_get_mat_state_var(W_mat_3d_t                  *m,
                       W_mat_3d_state_p             ms,
                       W_mat_3d_ip_ctx_t           *ip_ctx,
                       W_mat_3d_mat_state_var_enum  var_type,
                       double                      *var
                       );
W_bool_t 
WM3D_get_mat_state_named_var(W_mat_3d_t                  *m,
                             W_mat_3d_state_p             ms,
                             W_mat_3d_ip_ctx_t           *ip_ctx,
                             char *name, 
                             double                      *var
                             );

double 
WM3D_get_volume_mass(W_mat_3d_t *m);

W_bool_t
WM3D_convert_stress(double F[3][3],
                     W_mat_3d_stress_type_enum    in_stress_type,
                     double                       in_stress[3][3],
                     W_mat_3d_stress_type_enum    out_stress_type,
                     double                       out_stress[3][3]
                     );

/* The stress tensor must be symmetric; otherwise the results are */
/* unpredictable.  NO CHECK is performed. */
W_bool_t
WM3D_principal_stresses(double stress[3][3],
                        double principal_stress[3],
                        W_vect_t *n0, W_vect_t *n1, W_vect_t *n2
                        );


int
WM3D_get_mcode (W_mat_3d_t *m);

/*
  Compute material stiffness tensor.
  Compute material stiffness tensor in the reference configuration.
 */
W_bool_t 
WM3D_get_mat_stiff (W_mat_3d_t *m/* mat3d */,
                    W_mat_3d_state_p             ms,/* material state */
                    W_mat_3d_ip_ctx_t           *ip_ctx,/* integration point context;
                                                           used to generate kinematic
                                                           quantities which the material state
                                                           might need in order to compute
                                                           the requested stress */
                    W_mat_3d_stiff_enum stiff_type, /* type of material stiffness matrix */
                    double C[6][6] /* material stiffness tensor in matrix form */
                        );
/*
  Compute material stiffness tensor.
  Compute material stiffness tensor in the reference configuration.
 */
W_bool_t
WM3D_get_mat_stiff_9x9 (W_mat_3d_t *m/* mat3d */,
                    W_mat_3d_state_p             ms,/* material state */
                    W_mat_3d_ip_ctx_t           *ip_ctx,/* integration point context;
                                                           used to generate kinematic
                                                           quantities which the material state
                                                           might need in order to compute
                                                           the requested stress */
                    W_mat_3d_stiff_enum stiff_type, /* type of material stiffness matrix */
		    double C[9][9]
                        );
void
WM3D_update_history (W_mat_3d_t *m,
                     W_mat_3d_state_p ms,
                     W_mat_3d_ip_ctx_t *ip_ctx,
                     W_solution_history_t *h, 
                     W_index_t pntidx,
                     W_float_t time
                     );


#endif /* WMAT3D_H */
