#ifndef WNEWMFEND_H
#   define WNEWMFEND_H

#include "Wtypes.h"

struct W_newmark_integrator_implem_t;

typedef void (*W_newmark_advance_implem_proc) (struct W_newmark_integrator_implem_t *ii);

typedef W_float_t (*W_newmark_time_implem_func) (struct W_newmark_integrator_implem_t *ii);

typedef W_float_t (*W_newmark_dt_implem_func) (struct W_newmark_integrator_implem_t *ii);

typedef void (*W_newmark_get_consts_implem_proc) (struct W_newmark_integrator_implem_t *ii, double *beta, double *gamma);

typedef void (*W_newmark_stat_implem_proc) (struct W_newmark_integrator_implem_t *ii,
                                            double *ass_rhs_time,
                                            double *ass_lhs_time,
                                            double *solve_time, 
                                            double *pred_time,
                                            double *updt_time
                                            );

typedef void (*W_newmark_update_history_implem_proc) (struct W_newmark_integrator_implem_t *ii,
                                                      W_history_writer_t *reporter);

typedef int (*W_newmark_done_implem_func) (struct W_newmark_integrator_implem_t *ii);

typedef struct W_newmark_integrator_methods_t {
W_newmark_advance_implem_proc        advance;
W_newmark_time_implem_func           time;
W_newmark_dt_implem_func             dt;
W_newmark_get_consts_implem_proc     get_consts;
W_newmark_stat_implem_proc           stat;
W_newmark_update_history_implem_proc update_history;
W_newmark_done_implem_func           done;
}              W_newmark_integrator_methods_t ;

/*
  Private to implementations.
 */
struct W_newmark_integrator_t * /* integrator or NULL if required methods not defined */
newmark_new_front_end (
                       struct W_newmark_integrator_implem_t *ii,
                       W_newmark_integrator_methods_t *methods
                       );
struct W_newmark_integrator_implem_t *
newmark_implem (struct W_newmark_integrator_t *i);

/*
  Public interface.
 */
void
W_newmark_advance (struct W_newmark_integrator_t *i /* integrator */);
W_float_t /* time */
W_newmark_get_time (struct W_newmark_integrator_t *i/* integrator */);
W_float_t
W_newmark_get_dt (struct W_newmark_integrator_t *i/* integrator */);
void
W_newmark_update_history (struct W_newmark_integrator_t *i, /* integrator */
                          W_history_writer_t *hw /* history writer */
                          );
int /* ignore */
W_newmark_done (struct W_newmark_integrator_t *i /* integrator */);
void
W_newmark_get_consts (struct W_newmark_integrator_t *i, /* integrator */
                      double *beta, /* Newmark beta */
                      double *gamma /* Newmark gamma */
                      );
void
W_newmark_get_stat (struct W_newmark_integrator_t *i,
                double *ass_rhs_time, /* CPU time to assemble rhs */
                double *ass_lhs_time,  /* CPU time to assemble lhs */
                double *solve_time,  /* CPU time to solve */
                double *start_time, /* CPU time to start step */
                double *updt_time /* CPU time to update configuration */
                );

#endif 
