#ifndef WNODE_H
#   define WNODE_H

#include "Wtypes.h"
#include "Wutil.h"

W_bool_t
W_node_reader(W_domain_t *d);

#define FREE_DOF  1
#define FIXED_DOF 0

void
WN_get_ref_coords (W_node_t *n, double *x, double *y, double *z);

void
WN_get_displ (W_node_t *n, double *x, double *y, double *z);

void
WN_get_coords (W_node_t *n, double *x, double *y, double *z);

void
WN_get_prev_displ (W_node_t *n, double *x, double *y, double *z);

void
WN_get_prev_coords (W_node_t *n, double *x, double *y, double *z);

void
WN_get_velocity (W_node_t *n, double *x, double *y, double *z);

void
WN_get_acceleration (W_node_t *n, double *x, double *y, double *z);

W_index_t
WN_number_your_equations(W_node_t *n, W_index_t eqn);

void
WN_update_config_Newmark_variant(W_node_t *n, W_domain_t *d, W_float_t dt);

void
WN_update_config_Warburton_variant(W_node_t *n, W_domain_t *d, W_float_t dt);

void
WN_setup_init_conds(W_node_t *n,
                    W_vect_t *velocity,
                    W_float_t dt
                    );
void
WN_set_acceleration (W_node_t *n, W_acceleration_3d_t *a, W_bool_t apply[3]);
void *
WN_new_history_client_data(W_node_t *n, W_index_t idx,
                           W_history_writer_t *r, char *input_line);

void
WN_update_history(W_node_t *n, W_solution_history_t *h,
                   W_float_t time);
/*
   Prepare for a force assembly.
   Zero the force vector.
 */
void
WN_prep_force_assembly (W_node_t *n);

void
WN_assemble_force (W_node_t *n, W_coords_3d_t *f);

void
WN_prep_mass_assembly (W_node_t *n);

void
WN_assemble_mass (W_node_t *n, W_float_t mass);


void
WN_transform_by_rot_matrix_t(W_node_t *n, W_triad_3d_t *in_t, W_triad_3d_t *out_t);

void
WN_assemble_moment(W_node_t *n, W_coords_3d_t *m);


void
WN_assemble_tensor_of_inertia(W_node_t *n, W_triad_3d_t *Im);

void
WN_resolve_rot_inertia(W_node_t *n);

void
WN_assemble_ang_vel_dep_load(W_node_t *n);

#ifdef X_XFACE
void
WN_modify_node_grep (W_node_t *n /* node */,
                     double displ_scale, W_bool_t draw_numbers, W_bool_t immed_draw) ;
#endif


W_float_t
WN_node_kinetic_energy(W_node_t *n);

void
WN_get_disp_increment(W_node_t *n,
                      W_coords_3d_t *transl_incr, W_coords_3d_t *rot_incr);

void
WN_print_info(W_node_t *n,
              W_domain_t *d,
              W_info_spec_t spec,
              FILE *stream, int indent);


#if defined(PVM_WASP_WORKER) || defined(PVM_WASP_MASTER)

void
WN_pack_interface_vector(W_node_t *n, W_vector_t to_pack, double packed[6]);

void
WN_unpack_interface_vector(W_node_t *n, double packed[6],
                           W_vector_t unpack_into);


#endif /* if defined(PVM_WASP_WORKER) || defined(PVM_WASP_MASTER) */

void
WN_impl_init(W_node_t *n, W_domain_t *d);

void
WN_impl_predict(W_node_t *n, W_domain_t *d, W_float_t dt);

void
WN_assemble_damping (W_node_t *n, /* node */
                     double c,  /* damping coefficient */
                     double dt /* time step */
                     );
void
WN_solve_for (W_node_t *n);
double
WN_nodal_mass (W_node_t *n);
void
WN_fix_node (W_node_t *n,
#if W_ROT_DOFS
             W_bool_t is_free[6]
#else
             W_bool_t is_free[3]
#endif             
) ;
int
WN_is_fixed (W_node_t *n,
#if W_ROT_DOFS
             W_bool_t is_fixed[6]
#else
             W_bool_t is_fixed[3]
#endif
             );
void
WN_write_tecplot_file (W_node_t *n, /* node */
                       FILE *stream, /* output stream */
                       int output_spec,
                       W_index_t *count /* index of current output point;
                                         this number is incremented inside
                                         and *count is therefore the number
                                         of the next TECPLOT point after
                                         this function completes
                                         */
                       );

int /* returns number of dofs per node */
WN_eqn_numbers_and_constraints (W_node_t *n,
#if W_ROT_DOFS
                                W_index_t ens[6],
                                W_bool_t constrained[6]
#else
                                W_index_t ens[3],
                                W_bool_t constrained[3]
#endif
                                );

void
WN_update_config_Newmark_implicit (W_node_t *n /* node */
                                   );

void
WN_get_force (W_node_t *n, double *x, double *y, double *z);

void
WN_set_force (W_node_t *n, W_vect_t *a, W_bool_t apply[3]);

void
WN_set_acceleration (W_node_t *n, W_acceleration_3d_t *a, W_bool_t apply[3]);


void
WN_set_velocity (W_node_t *n, W_vect_t *a, W_bool_t apply[3]);


void
WN_set_displ (W_node_t *n, W_vect_t *a, W_bool_t apply[3]);

void
WN_set_prev_displ (W_node_t *n, W_vect_t *a, W_bool_t apply[3]);

int
WN_is_fixed (W_node_t *n, W_bool_t is_fixed[6]);

double
WN_extf_norm_squared (W_node_t *n);

double
WN_inertf_norm_squared (W_node_t *n);

W_node_t * /* node or NULL if anything went wrong */
WN_new_node (W_domain_t *d, /* domain; the node is added to this domain */
             W_index_t id, /* node identifier; either positive integer, which means "I know what
                              the number of this node is", or a non-positive integer, which means "generate
                              an identifier for me, please".
                           */
             W_coords_3d_t *c /* reference coordinate of the node */
             );

#endif /* !WNODE_H */
