#ifndef WOCTREE_H
#   define WOCTREE_H

#ifdef X_XFACE
#  if !X_XFACE
#    undef X_XFACE
#  endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define WANT_VECTP_ALIASES
#include "vectP.h" 
#define loc_t CkitVector3   /* type defined in Ckit library */
#define vect_t loc_t   /* type defined in Ckit library */

#ifdef X_XFACE
#include "Esimple.h"
#endif

#include "g3d.h"
#include "g3d.h"
#include "g3dpoly.h"


#define new_instance(TYPE) ((TYPE *) malloc(sizeof(TYPE)))

#define new_array(NELT, TYPE) \
                      ((TYPE *) calloc((NELT), (size_t) sizeof(TYPE)))

#define realloc_array(PTR, NEW_SIZE, TYPE) \
   ((TYPE *) realloc(PTR, (size_t)(NEW_SIZE * sizeof(TYPE))))

#define free(PTR) free(PTR)

#define CHECK_MALLOC(ptr)                                               \
   {                                                                    \
     if (ptr == NULL) {                                                 \
       fprintf(stderr, "%s (%d): Failed malloc\n", __FILE__, __LINE__); \
       exit(EXIT_FAILURE);                                              \
     }                                                                  \
   }



struct octree_leaf_t;

struct octree_root_t;

typedef void *owner_ptr;

typedef void (*owner_get_vertex_loc_proc)(owner_ptr owner,
                                          int vertex, loc_t *loc);

typedef struct octree_root_t {
  owner_ptr                   owner;
  owner_get_vertex_loc_proc   get_vertex_loc;
  struct octree_leaf_t       *top_leaf;
  short int                   contains_crack;
}              octree_root_t;



typedef struct octree_leaf_t {
  octree_root_t          *root;
  loc_t                   param_center;
  double                  radius;
  struct octree_leaf_t  **leaves;
#ifdef X_XFACE
  GraphicObj             *g_rep;
#endif
  short int               level; /* top leaf is level 0 */
  short int               contains_crack;
}              octree_leaf_t;


octree_root_t *
octree_new_root(void *owner, owner_get_vertex_loc_proc get_vertex_loc);

void
octree_extend_crack(octree_root_t *root,
                    G3D_poly_t *poly,
                    double terminal_size);



#endif /* WOCTREE_H */

