#ifndef WPVM_H
#   define WPVM_H

#define WORKER_NAME       "wd_worker"
#define WORKER_GROUP      WORKER_NAME 
#define PVM_DATA_ENCODING PvmDataDefault
#define PVM_SPAWN_OPTS    (PvmTaskDefault)
#define PVM_SPAWN_WHERE   NULL

typedef struct W_pvm_status_t {
  int info;
  int bufid;
  int bytes;
  int msgtag;
  int tid;
}              W_pvm_status_t;

#ifdef WPVM_PRIVATE_HEADER
W_pvm_status_t          pvm_status;
#   undef WPVM_PRIVATE_HEADER
#else
extern W_pvm_status_t   pvm_status;
#endif

/* Wildcard */
#define ANY_TASK -1

/* TAGS */
typedef enum { ANY_TAG               = -1,
               STARTUP_ACK_TAG       =  1,
               TIME_STEP_REQUEST_TAG =  2,
               TIME_STEP_ANSWER_TAG  =  3,
               EFFLOAD_LOCAL_TAG     =  4,
               EFFLOAD_GLOBAL_TAG    =  5
             } W_pvm_msg_tag;

/* MACROS */
#define NUM_OF_WORKERS(master) (master)->num_of_workers
#define MASTER_TID(worker) (worker)->master_tid
#define INITSEND()                                                   \
  do {                                                               \
    if ((pvm_status.bufid = pvm_initsend(PVM_DATA_ENCODING)) < 0) {  \
      PVM_err_exit("pvm_initsend", __FILE__);                        \
    }                                                                \
  } while (0)
#define RECEIVE(tid, tag)                               \
  if ((pvm_status.bufid = pvm_recv(tid, tag)) < 0) {    \
    PVM_err_exit("pvm_recv", __FILE__);                 \
  }
#define SEND(tid, tag)                                 \
  if ((pvm_status.info = pvm_send(tid, tag)) != 0) {   \
    PVM_err_exit("pvm_send", __FILE__);                \
  }
#define BCAST(group, tag)                                       \
  if ((pvm_status.info = pvm_bcast(group, tag)) != 0) {         \
    PVM_err_exit("pvm_bcast", __FILE__);                        \
  }
#define BUFINFO()                                                       \
  if ((pvm_status.info                                                  \
       = pvm_bufinfo(pvm_status.bufid, &pvm_status.bytes,               \
                     &pvm_status.msgtag, &pvm_status.tid)) != 0) {      \
    PVM_err_exit("pvm_bufinfo", __FILE__);                              \
  }
#define BUFINFO_TID() pvm_status.tid
#define BUFINFO_TAG() pvm_status.msgtag
#define UNPACK(type, item) pvm_upk##type(item, 1, 1)
#define PACK(type, item)   pvm_pk##type(item, 1, 1)
#define UNPACK_ARRAY(type, array, nitem) pvm_upk##type(array, nitem, 1)
#define PACK_ARRAY(type, array, nitem)   pvm_pk##type(array, nitem, 1)
#define JOIN_GROUP(group)                                                   \
   if ((pvm_status.info = pvm_joingroup(group)) < 0) {                      \
     PVM_err_exit("pvm_joingroup", __FILE__);                               \
   }
#define BARRIER(group, ninstances)                                          \
   if ((pvm_status.info = pvm_barrier(group, ninstances)) < 0) {            \
     PVM_err_exit("pvm_barrier", __FILE__);                                 \
   }
#define MYTID()                                                              \
   (((pvm_status.info = pvm_mytid()) <= 0)                                   \
    ? PVM_err_exit("pvm_mytid", __FILE__)                                    \
    : pvm_status.info)                                          
#define PARENT()                                                             \
   (((pvm_status.info = pvm_parent()) <= 0)                                  \
    ? PVM_err_exit("pvm_parent", __FILE__)                                   \
    : pvm_status.info)                                          

W_bool_t
pvm_run_reader(W_domain_t *d, FILE *stream);

W_float_t
W_pvm_master_time_step(W_domain_t *d);

int
PVM_err_exit(char *pvm_routine, char *msg);

#endif
