#ifndef WTYPES_H
#   define WTYPES_H

/* RCSHEADER "$Id: Wtypes.h,v 1.8 2004/06/08 17:53:36 cvs Exp $" */

#ifdef __cplusplus
extern "C" {
#endif

#include "Wconfig.h"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>
#ifdef UNIX
#   include <sys/stat.h>
#   include <unistd.h>
#   include <float.h>
#   if HAVE_STRING_H
#      include <string.h>
#   else
#      if HAVE_STRINGS_H
#         include <strings.h>
#      endif
#   endif
#else
#   define caddr_t void *
#endif

/* Ckit */
#ifdef DMALLOC
#  include "dmalloc.h"
#endif

#include "newP.h" 
#include "ioP.h"  
#include "getoptP.h" 
#include "listP.h"   
#include "dictP.h"   
#include "hashP.h"   
#define VECMAC_VECTOR_TYPE 1
#include "vecmacP.h" 
#define WANT_VECTP_ALIASES
#include "vectP.h" 
#include "tokensP.h"
#include "expevalP.h"
#include "gaussP.h" 
#include "randP.h" 
#include "timeP.h"
#include "mathP.h"
/* G3D */
#include "g3d.h"
/* LOBS */
#include "lobs3d.h"
/* FLES */
#include "flestype.h"

#if defined (MPI_WASP) && MPI_WASP
# include <mpi.h>
#endif

#ifdef MAXFILENAMECHAR
#  undef MAXFILENAMECHAR
#endif
#define MAXFILENAMECHAR 1024
#if FILENAME_MAX > MAXFILENAMECHAR
#   undef MAXFILENAMECHAR
#   define MAXFILENAMECHAR FILENAME_MAX /* POSIX constant */
#endif

#define END_OF_BLOCK '&'

#define W_float_t double        /* set precision here */

#define W_loc_t             CkitVector3   /* type defined in Ckit library */
#define W_loc2d_t           CkitVector2
#define W_vect_t            CkitVector3 
#define W_vect2d_t          CkitVector2  
#define W_coords_3d_t       CkitVector3
#define W_velocity_3d_t     CkitVector3
#define W_acceleration_3d_t CkitVector3


typedef W_float_t *W_vector_t;

typedef int W_index_t;

typedef int W_bool_t;

typedef short int W_short_t;

typedef char W_short_bool_t;

typedef W_index_t *W_index_vector_t;

/*
  Direction (three for translation, three for rotation).
 */
typedef enum { W_DX = 0, W_DY = 1, W_DZ = 2, W_RX = 3, W_RY = 4, W_RZ = 5
             } W_direction_t;

/*
  Quaternion.
 */
typedef struct W_quaternion_t {
  W_float_t q0, q1, q2, q3;
}              W_quaternion_t;



/*
  Orthonormal unit vectors (Cartesian triad).
 */
typedef struct W_triad_3d_t {
  W_coords_3d_t      vect1;
  W_coords_3d_t      vect2;
  W_coords_3d_t      vect3;
}              W_triad_3d_t;


/*
  Node.
 */
typedef struct W_node_t {
  struct W_domain_t       *domain;
  W_index_t                id; /* unique integer */
  W_index_t                internal_id; /* sequential numbering of nodes */
  W_coords_3d_t            ref_coords; /* reference coordinates */
  W_coords_3d_t            displ_t; /* current displacement */
  W_coords_3d_t            displ_t_minus_dt; /* displacement at t-dt */
  W_velocity_3d_t          velocity_t; /* current velocity */
  W_acceleration_3d_t      acceleration_t; /* current acceleration */
#if W_ROT_DOFS
  W_triad_3d_t             nodal_triad_t; /* current nodal triad  */
  W_coords_3d_t            rot_incr_t; /* current rotation increment  */
  W_velocity_3d_t          ang_velocity_t; /* current angular velocity */
  W_acceleration_3d_t      ang_acceleration_t; /* current angular acceleration */
  W_triad_3d_t             mom_inertia_princ_dirs; /* orientation
                                                    of the tensor of inertia */
#endif
  W_coords_3d_t            force_t; /* current force */
#if W_ROT_DOFS
  W_coords_3d_t            moment_t; /* current moment */
#endif
  double                   mass; /* mass */
#if W_ROT_DOFS
  double                   mom_inertia[3]; /* principal values of the
                                                    tensor of inertia */
#endif
#ifdef X_XFACE
  GraphicObj               *node_marker;
# if W_ROT_DOFS
  GraphicObj               *node_t1;
  GraphicObj               *node_t2;
  GraphicObj               *node_t3;
# endif
#endif
#if W_ROT_DOFS
  W_short_bool_t           can_move_in_dir[6];
#else
  W_short_bool_t           can_move_in_dir[3]; /* Is node free to move
                                                in certain directions? */
#endif
}              W_node_t;


/*
  Contact face definition.
 */
typedef struct WCF_contact_face_t {
  int                           num_nodes;
  W_node_t                     *node[4];
}              WCF_contact_face_t;


struct W_elem_t;
struct W_load_t;

typedef enum W_mat_state_update_form_enum {
  W_MAT_STATE_UPDATE_FINAL = 0,
  W_MAT_STATE_UPDATE_TEMP  = 1
}            W_mat_state_update_form_enum;


typedef int (*WE_elem_elcode_func) (struct W_elem_t *el);

typedef char * (*WE_elem_type_name_func)(struct W_elem_t *el);

typedef void (*WE_assemble_restoring_forces_proc)(struct W_elem_t *el,
                                                  W_mat_state_update_form_enum update_form);


typedef void (*WE_assemble_load_forces_proc)(struct W_elem_t *el);


typedef void (*WE_assemble_mass_proc)(struct W_elem_t *el);

typedef W_float_t (*WE_suggested_time_step_proc)(struct W_elem_t *el);

struct W_history_writer_t;
struct W_solution_history_t;

typedef W_float_t (*WE_elem_total_mass_proc)(struct W_elem_t *el);


typedef void (*WE_modify_grep_proc)(struct W_elem_t *el, double displ_scale, W_bool_t immed_draw);


typedef W_float_t (*WE_deformation_energy_proc)(struct W_elem_t *el);

typedef void (*WE_calc_eff_loads_proc)(struct W_elem_t *el);

typedef void (*WE_solve_for_proc)(struct W_elem_t *el);

typedef void (*WE_update_config_Warburton_variant_proc)(struct W_elem_t *el);

typedef void (*WE_update_config_Newmark_variant_proc)(struct W_elem_t *el);

typedef void (*WE_update_history_proc)(struct W_elem_t *el,
                                       struct W_solution_history_t *h, W_float_t time);

typedef void *(*WE_new_history_client_data_proc)(struct W_elem_t *el,
                                      W_index_t idx,
                                      struct W_history_writer_t *r,
                                      char *input_line);

typedef void (*WE_apply_initial_conditions_proc)(struct W_elem_t *el);

typedef W_bool_t (*WE_get_elem_prop_proc) (struct W_elem_t *el, const char *prop,
                                           void *value);

#define HEXVUE_POSTPROC_FILE      1
#define VE3D_POSTPROC_FILE        2
#define TECPLOT_I_POSTPROC_FILE   3
#define HEXVUE_XTRA_POSTPROC_FILE 4

typedef void (*WE_write_postproc_file_proc)(struct W_elem_t *el,
                                            ckit_iobuf_t IOB,
                                            int type, /*hexvue, ve3d, tecplot*/
                                            char *mode, /* binary or ascii */
                                            void *optdata);

typedef int (*WE_num_of_faces_func) (struct W_elem_t *el);
typedef int (*WE_get_face_vertex_nodes_proc) (struct W_elem_t *el,
                                              int nface,
                                              int ndim, 
                                              W_node_t *n[]);

typedef int (*WE_get_num_nodes_func)(struct W_elem_t *el);
typedef W_node_t *(*WE_get_node_func)(struct W_elem_t *el,
					int j);
typedef int (*WE_get_num_matstates_func)(struct W_elem_t *el);
typedef void *(*WE_get_matstate_func)(struct W_elem_t *el,
                                        int j);
typedef int (*WE_get_material_id_func)(struct W_elem_t *el);

typedef enum W_info_spec_t {
  W_INFO_BRIEF         = 0,
  W_INFO_FULL          = 1
}            W_info_spec_t;



/*
  Mass matrix type.
 */
typedef enum W_mass_type_enum {
  W_MASS_DIAG = 0, /* diagonal */
  W_MASS_CONSISTENT = 1 /* consistent */
}            W_mass_type_enum;


#define CD_INTEGRATOR      0
#define NEWMARK_INTEGRATOR 1

struct W_cd_integrator_t;
struct W_newmark_integrator_t;

/*
  Newmark predictor type
 */
typedef enum W_newmark_predictor_enum {
  W_NEWMARK_PRED_UNKNOWN   = 0,
  W_NEWMARK_PRED_SAME_U    = 1,
  W_NEWMARK_PRED_SAME_V    = 2,
  W_NEWMARK_PRED_SAME_A    = 3,
  W_NEWMARK_PRED_ZERO_A    = 4,
  W_NEWMARK_PRED_ADAPTIVE  = 5,
  W_NEWMARK_PRED_PREV_INCR = 6,
  W_NEWMARK_PRED_DEFAULT   = 2
}            W_newmark_predictor_enum;

typedef void *W_newmark_client_t;

typedef int (*W_newmark_get_incr_proc) (struct W_newmark_integrator_t *i,
                                        int eqns[], /* for equation numbers */
                                        double values[], /* return values in this array */
                                        int dim /* dimension of preceding arrays */
                                        );

typedef void (*W_newmark_incr_disp_proc) (struct W_newmark_integrator_t *i,
                                          int nr, int *ri,
                                          double *values);

typedef void (*W_newmark_assemble_eff_stiff_proc)
     (struct W_newmark_integrator_t *i, /* Newmark integrator */
      fles_update_mode_enum update_mode, /* update mode;  */
      int nr, /* number of rows and columns in square matrix to be assembled */
      int *ri, /* array of equations numbers, which apply to both rows and columns
                  of the matrix to be assembled */
      double *values, /* square matrix to be assembled, passed
                         as a linear array of nr*nr values */
      W_bool_t *constrained /* array of nr elements; TRUE if degree-of-freedom
                               constrained, FALSE otherwise */
      );

typedef void (*W_newmark_assemble_eff_stiff_client_proc)
     (W_newmark_client_t client,
      W_newmark_assemble_eff_stiff_proc assemble_eff_stiff_callback,
      W_newmark_get_incr_proc get_incr,
      W_newmark_predictor_enum predictor,
      W_float_t stiff_mult,
      W_float_t mass_mult);

typedef void (*W_newmark_assemble_residual_force_proc)
     (struct W_newmark_integrator_t *i,
      fles_update_mode_enum update_mode, /* update mode */
      int nr,/* number of rows in vector to be assembled */
      int *ri,/* array of equations numbers */
      double *values /* vector to be assembled */
      );

typedef void (*W_newmark_assemble_residual_force_client_proc)
     (W_newmark_client_t client,
      W_newmark_assemble_residual_force_proc assemble_residual_force,
      W_mat_state_update_form_enum update_form,
      W_newmark_get_incr_proc get_incr,
      W_newmark_predictor_enum predictor,
      double *extf_norm,
      double *inertf_norm,
      double *resid_norm);

typedef W_float_t (*W_newmark_finalize_step_client_proc)
     (W_newmark_client_t client, W_newmark_incr_disp_proc incr_disp_callback);

typedef void (*W_newmark_start_step_client_proc)(W_newmark_client_t client);

typedef void (*W_newmark_update_config_client_proc)
     (W_newmark_client_t client,
      W_newmark_get_incr_proc newmark_get_incr,
      W_newmark_predictor_enum newmark_predictor);

typedef void (*WE_print_info_proc)(struct W_elem_t *el, W_info_spec_t spec,
                                   FILE *stream, int indent);

typedef W_float_t (*WE_finalize_step_proc)(struct W_elem_t *el);

typedef W_bool_t (*WE_configure_func)(struct W_elem_t *el, char *option);

typedef W_float_t (*WE_kinetic_energy_func)(struct W_elem_t *el, W_mass_type_enum mass_type);

struct W_probe_package_t;

typedef W_bool_t (*WE_probe_func)(struct W_elem_t *el, void *options,
				  W_loc_t *loc, unsigned long spec,
				  struct W_probe_package_t *result);

typedef int (*WE_get_contact_faces_func) (struct W_elem_t *el, int nface,
                                          WCF_contact_face_t *cf[], int maxcf);


typedef void (*WE_assemble_residual_forces_proc) (struct W_elem_t *el,
                                                  W_mat_state_update_form_enum update_form,
                                                  W_mass_type_enum mass_type);

typedef void (*WE_assemble_eff_stiffness_proc) (struct W_elem_t *el,
                                                W_newmark_assemble_eff_stiff_proc assemble_eff_stiff_callback,
                                                W_float_t stiff_mult,
                                                W_float_t mass_mult,
                                                W_mass_type_enum mass_type);

typedef void (*WE_update_connectivity_info_proc) (struct W_elem_t *el);
                                                  

/*
  Protocol of generic finite element.
 */
typedef struct W_elem_methods_t {
  WE_elem_elcode_func                       elem_elcode;
  WE_elem_type_name_func                    elem_type_name;
  WE_assemble_restoring_forces_proc         assemble_restoring_forces;
  WE_assemble_mass_proc                     assemble_mass;
  WE_elem_total_mass_proc                   elem_total_mass;
  WE_assemble_load_forces_proc              assemble_load_forces;
  WE_suggested_time_step_proc               suggested_time_step;
  WE_deformation_energy_proc                deformation_energy;
  WE_calc_eff_loads_proc                    calc_eff_loads;
  WE_solve_for_proc                         solve_for;
  WE_update_config_Warburton_variant_proc   update_config_Warburton_variant;
  WE_update_config_Newmark_variant_proc     update_config_Newmark_variant;
  WE_finalize_step_proc                     finalize_step;
  WE_new_history_client_data_proc           new_history_client_data;
  WE_update_history_proc                    update_history;
  WE_apply_initial_conditions_proc          apply_initial_conditions;
  WE_write_postproc_file_proc               write_postproc_file;
#ifdef X_XFACE
  WE_modify_grep_proc                       modify_grep;
#endif
  WE_print_info_proc                        print_info;
  WE_configure_func                         configure;
  WE_probe_func                             probe;
  WE_num_of_faces_func                      num_of_faces;
  WE_get_face_vertex_nodes_proc             get_face_vertex_nodes;
  WE_get_contact_faces_func                 get_contact_faces;
  WE_assemble_residual_forces_proc          assemble_residual_forces;
  WE_assemble_eff_stiffness_proc            assemble_eff_stiffness;
  WE_kinetic_energy_func                    kinetic_energy;
  WE_update_connectivity_info_proc          update_connectivity_info;
  WE_get_elem_prop_proc                     get_elem_prop;
  WE_get_num_nodes_func			    get_num_nodes;
  WE_get_node_func			    get_node;
  WE_get_num_matstates_func		    get_num_matstates;    
  WE_get_matstate_func			    get_matstate;
  WE_get_material_id_func		    get_material_id;    
}              W_elem_methods_t;
  

    


struct W_domain_t;

typedef void (*W_response_observer_func) (struct W_domain_t *d, int end_of_integration);

typedef W_bool_t (*W_elem_reader_func) (struct W_domain_t *);

/*
  Finite element.
 */
typedef struct W_elem_t {
  struct W_domain_t  *domain; /* belongs to domain */
  W_elem_methods_t   *methods; /* protocol */
  void               *data; /* type-specific data */
  W_float_t           deformation_energy; /* deformation energy */
#ifdef X_XFACE
  GraphicObj         *grep;
#endif
#ifdef PVM_NEED_GHOST_ELEMENTS
  W_bool_t            i_am_ghost;
#endif
}              W_elem_t;


typedef void *W_matstate_t;


struct W_material_t;


typedef W_bool_t (*WM_get_uniaxial_elast_const_proc)(struct W_material_t *m,
                                                     W_float_t *E,
                                                     W_float_t *nu);
     
     
typedef W_bool_t (*WM_get_uniaxial_stress_proc)(struct W_material_t *m,
                                                W_matstate_t ms,
                                                W_float_t strain,
                                                W_float_t strain_rate,
                                                W_float_t *stress
                                                );
typedef W_bool_t (*WM_get_pss_stress_proc)(struct W_material_t *m,
                                           W_matstate_t ms,
                                           W_float_t eps1,
                                           W_float_t eps2,
                                           W_float_t gamma,
                                           W_float_t eps1_rate,
                                           W_float_t eps2_rate,
                                           W_float_t gamma_rate,
                                           W_float_t *sigma1,
                                           W_float_t *sigma2,
                                           W_float_t *tau
                                           );

typedef W_bool_t (*WM_get_3d_stress_proc)(struct W_material_t *m,
                                          W_matstate_t ms,
                                          W_float_t eps[6],
                                          W_float_t eps_rate[6],
                                          W_float_t sigma[6]);

typedef W_matstate_t (*WM_get_mat_state_proc)(struct W_material_t *m);
   
typedef W_bool_t (*WM_is_in_plastic_state_proc)(struct W_material_t *m,
                                                W_matstate_t ms);
   
typedef W_float_t (*WM_get_uniaxial_def_energy_proc)(struct W_material_t *m,
                                                     W_matstate_t ms,
                                                     W_float_t strain);

typedef W_float_t (*WM_get_pss_def_energy_proc)(struct W_material_t *m,
                                                W_matstate_t ms,
                                                W_float_t eps1,
                                                W_float_t eps2,
                                                W_float_t gamma);

typedef W_float_t (*WM_get_3d_def_energy_proc)(struct W_material_t *m,
                                                W_matstate_t ms,
                                                W_float_t eps[6]);
/*
  Protocol of material.
 */
typedef struct W_material_methods_t {
  WM_get_uniaxial_elast_const_proc     get_uniaxial_elast_const;
  WM_get_uniaxial_stress_proc          get_uniaxial_stress;
  WM_get_uniaxial_def_energy_proc      get_uniaxial_def_energy;
  WM_get_pss_stress_proc               get_pss_stress;
  WM_get_pss_def_energy_proc           get_pss_def_energy;
  WM_get_3d_stress_proc                get_3d_stress;
  WM_get_3d_def_energy_proc            get_3d_def_energy;
  WM_get_mat_state_proc                get_mat_state;
  WM_is_in_plastic_state_proc          is_in_plastic_state;
}              W_material_methods_t;
  


typedef enum W_mcode_t {
  MCODE_UNIAX_LE      = 1, /* uniaxial linear elastic */
  MCODE_UNIAX_PEP     = 2, /* uniaxial perfectly elastic-plastic */
  MCODE_UNIAX_HEP     = 3, /* uniaxial hardening elastic-plastic */
  MCODE_BIAX_LE       = 4, /* biaxial linear isotropic elastic */
  MCODE_TRIAX_LE      = 5, /* tri-axial linear isotropic elastic */
  MCODE_GENERAL_TRIAX = 6, /* general triaxial material */
  MCODE_GENERAL_UNIAX = 7  /* general uniaxial material */
} W_mcode_t;

typedef enum W_pss_code_t {
  PSSCODE_PLANE_STRESS  = 1, 
  PSSCODE_PLANE_STRAIN  = 2
} W_pss_code_t;



/*
  Material.
 */
typedef struct W_material_t {
  W_mcode_t              mcode; /* material code */
  int                    id;
  W_float_t              vol_mass; /* volume mass */
  W_material_methods_t   *methods; /* protocol */
  void                  *data; /* type-specific data */
#ifdef X_XFACE
  EPixel                 color;
  EFringeTable           fringes;
#endif
}              W_material_t;

typedef enum { W_DISP_BC_NONE = -1, W_DISP_BC_A = 0, W_DISP_BC_V = 1, W_DISP_BC_D = 2 } W_disp_bc_enum;
typedef enum { W_DISPBC_TYPE_BOX = 0, W_DISPBC_TYPE_SURF = 1 } W_disp_bc_type_enum;
  
typedef struct W_disp_bc_surf_face_t {
  int     npts;
  W_loc_t pts[4];
} W_disp_bc_surf_face_t;

typedef struct W_disp_bc_surf_t {
  char                            *egf_surf_file;
  double                           dist;
  LIST_OF(W_disp_bc_surf_face_t *) fl;
} W_disp_bc_surf_t;

/*
  Displacement boundary condition.
 */
typedef struct W_disp_bc_t {
  int                     id; /* unique integer */
  struct W_domain_t      *domain; /* domain */
  W_disp_bc_type_enum     type;
  union {
    G3D_box_t                 box; /* applies to nodes in this box */
    W_disp_bc_surf_t          surf; /* applies to nodes near this surface */
  }                       data;
  LIST_OF (W_node_t *)    applies_to; /* cached list of nodes */
  char                   *prescribed[3]; /* expression for acceleration, velocity, or displacement */
  W_disp_bc_enum          kind[3]; /* what is being prescribed */
  struct {
    double on;
    double off;
  }                       applies_in_interval;
  W_bool_t                initialized;
}              W_disp_bc_t;


/* Type of time function. */
typedef enum { W_HARMONIC_TIME_FUNC = 1, W_TABULAR_TIME_FUNC = 2
             } W_tfcode_t;

/* Point in the tabular time function graph. */
typedef struct W_abscissa_ordinate_pair_t {
  W_float_t     abscissa;
  W_float_t     ordinate;
}              W_abscissa_ordinate_pair_t;

/* Tabular time function. */
typedef struct W_time_func_tabular_t {
  W_index_t                   num_of_pairs; /* number of pairs */
  W_float_t                   cached_time; /* cached time */
  W_float_t                   cached_value; /* cached time function value */
  W_abscissa_ordinate_pair_t  *pairs; /* time-value pairs */
}              W_time_func_tabular_t;

/* Harmonic time function. */
typedef struct W_time_func_harmonic_t {
  W_float_t        phase; /* phase */
  W_float_t        circular_frequency; /* circular frequency */
}              W_time_func_harmonic_t;

/* Time function. */
typedef struct W_time_func_t {
  W_index_t                    id;
  W_tfcode_t                   tfcode; /* code */
  union {
    W_time_func_harmonic_t  harmonic; 
    W_time_func_tabular_t   tabular;
  }                             data; /* type-specific data */
}              W_time_func_t;

/*
  Nodal load.
 */
typedef struct W_nodal_load_t {
  W_index_t         node_num; /* node number */
  W_index_t         direction; /* direction */
  W_float_t         amplitude; /* amplitude */
  W_index_t         time_func_num; /* time function */
#ifdef X_XFACE
  GraphicObj       *fg_rep;
  GraphicObj       *mg_rep;
#endif
}              W_nodal_load_t;

/*
  
 */
typedef struct W_node_group_node_t {
  W_node_t        *n;
#ifdef X_XFACE
  GraphicObj       *f_rep;
#endif
}              W_node_group_node_t;


typedef enum W_node_group_load_type_enum {
  W_NODE_GROUP_LOAD_BODY  = 0,
  W_NODE_GROUP_LOAD_FORCE = 1,
  W_NODE_GROUP_LOAD_TRAC  = 2
}            W_node_group_load_type_enum;

/*
    Load applied to a group of nodes.
 */
typedef struct W_node_group_load_t {
  W_node_group_load_type_enum       kind; /* kind of load */
  union {
    struct {
      W_index_t                         direction; /* direction */
      W_float_t                         amplitude; /* amplitude */
      W_index_t                         time_func_num; /* time function */
    } force;
    struct {
      char                             *accel[3]; /* expressions for accelerations */
      W_index_t                         time_func_num[3]; /* time function */
    } body;
    struct {
      W_index_t                         direction; /* direction */
      W_float_t                         amplitude; /* amplitude */
      W_index_t                         time_func_num; /* time function */
      LIST_OF (W_load_t *)              nll; /* list of nodal loads */
    } trac;
  } spec;
  G3D_box_t                         box; /* applies to nodes in the box */
  LIST_OF (W_node_group_node_t *)   nodes; /* cached nodes */
}              W_node_group_load_t;


typedef struct W_elem_load_t {
  W_index_t         elem_num;
  W_index_t         direction;
  W_float_t         amplitude;
  W_index_t         time_func_num;
}              W_elem_load_t;


typedef enum W_load_type_t {
  MOUSE_ANIM_LOAD = 0,
  NODAL_LOAD      = 1,
  ELEM_LOAD       = 2,
  NODE_GROUP_LOAD = 3
}                 W_load_type_t;

/*
  Load.
 */
typedef struct W_load_t {
  W_load_type_t     type; /* type of load */
  union {
    W_elem_load_t         elem;
    W_nodal_load_t        nodal;
    W_node_group_load_t   node_group;
  }                 data; /* type-specific data */
  struct W_domain_t *domain;
}              W_load_t;

typedef struct W_timer_t {
  double          num_calls;
  double          cummulative_time;
  ckit_timer_t   *timer;
}              W_timer_t;  


#if defined(PVM_WASP_WORKER) || defined(PVM_WASP_MASTER)

typedef struct W_interface_package_t {
  double      forces[6];
}              W_interface_package_t;


typedef struct W_pvm_partition_id_t {
  char    *split_input_file;
  char    *host_name;
}              W_pvm_partition_id_t;


typedef struct W_pvm_master_t {
  int                               mytid;
  unsigned long                     num_of_interface_nodes;
  W_interface_package_t            *interfaces;
  int                               num_of_workers;
  LIST_OF(W_pvm_partition_id_t *)   partitions;
  int                      *tids;
}              W_pvm_master_t;

typedef struct W_local_to_interface_map_t {
  unsigned long      local_num;
  unsigned long      interface_num;
}              W_local_to_interface_map_t;


typedef struct W_pvm_worker_t {
  int                          mytid;
  int                          master_tid;
  int                          group_inum;
  unsigned long                num_of_interface_nodes;
  W_local_to_interface_map_t  *map;
}              W_pvm_worker_t;

#endif /* PVM version */

#if defined (MPI_WASP) && MPI_WASP 

#if MPI_WASP_ELEM_CUT
typedef int W_mpi_rank_t;
/*
  Node/ partition pair
 */
typedef struct W_mpi_npmap_pair_t {
  W_node_t       *n; /* node (handle on processor /rank/) */
  W_mpi_rank_t    rank; /* partition number (rank in MPI parlance) */
}              W_mpi_npmap_pair_t;
#endif /* MPI_WASP_ELEM_CUT */

#if MPI_WASP_NODE_CUT
  /* this type needs to be changed if there is more than
     SHRT_MAX processors (32767 for a 16-bit short int)
   */
typedef short int W_mpi_rank_t;

/*
  Node/ partition pair
 */
typedef struct W_mpi_npmap_pair_t {
  W_node_t       *n; /* node (handle on processor /rank/) */
  W_mpi_rank_t    ranks_dim; /* number of processors to which node belongs */
  W_mpi_rank_t   *ranks; /* array of processor numbers to which node belongs
                            (rank in MPI parlance) if ranks_dim > 1;
                            otherwise ranks stores directly the rank
                            of the owning processor
                         */
}              W_mpi_npmap_pair_t;
#endif /* MPI_WASP_NODE_CUT */

/*
  Node/ partition map
 */
typedef struct W_mpi_npmap_t {
  W_index_t            num_pairs;
  W_mpi_npmap_pair_t  *pairs;
}              W_mpi_npmap_t;

/*
  Communication mapper
 */
typedef struct W_mpi_comm_buff_t {
  int             rank;         /* rank (partition number) */
  int             num_to_send;  /* number of nodes to send */
  int             num_to_recv;  /* number of nodes to receive */
  W_node_t      **to_send;
  W_node_t      **to_recv;
#if MPI_WASP_ELEM_CUT
  W_short_bool_t *send_flag; /* array for each global node;
                                TRUE means send data to the processor
                                whose comm buff this is
                                  */
#endif
  int             count; /* number of doubles in buffer */
  int             sbuf_dim;   /* send buffer dimension */
  double         *sbuf;  /* send-buffer */
  int             rbuf_dim;   /* receive buffer dimension */
  double         *rbuf;  /* receive-buffer */
}              W_mpi_comm_buff_t;

typedef struct W_mpi_process_t {
  struct W_domain_t *domain; /* domain on the processor */
  int                size; /* number of collaborating processes */
  int                rank; /* rank of process */
  MPI_Request       *recv_requests; /* array of MPI requests */
  W_mpi_npmap_t      node_part_map; /* node/partition map */
  W_mpi_comm_buff_t *comm_buffs; /* communication buffers */
#if TIMED_WASP
#endif
}              W_mpi_process_t;

typedef void 
(*W_mpi_pack_func) (W_mpi_process_t *p, /* IN: MPI process */
                    W_mpi_comm_buff_t *cb /* IN: communication buffer */
                    );

typedef void
(*W_mpi_unpack_func) (W_mpi_process_t *p, /* IN: MPI process */
                      W_mpi_comm_buff_t *cb /* IN: communication buffer */
                      );

#endif /* defined (MPI_WASP) && MPI_WASP  */


struct W_history_writer_t;


struct W_contact_surf_t;

typedef enum W_gap_penalty_type {
  W_PENALTY_OVERLAP      = 0,
  W_PENALTY_OVERLAP_RATE = 1,
  W_PENALTY_OVERLAP_COMB = 2
}            W_gap_penalty_type;

/*
  Penalty-based contact.
 */
typedef struct WPC_faceted_penalty_contact_t {
  struct W_contact_surf_t         *cont_surf; /* contact surface */
  lobs_3d_front_end                lobs; /* box search */
  int                              first_time;
  W_gap_penalty_type               force_type; /* type of contact force */
  double                           overlap_penalty; /* penalty for overlap */
  double                           overlap_rate_penalty; /* penalty
                                                          for overlap velocity */
  double                           avgnmass; /* average nodal mass */
  int                              subdiv_level; /* level of subdivision
                                                  of contact face;
                                                 if subdiv_level==1,
                                                 there are four contact
                                                 points per face;
                                                 if subdiv_level==2,
                                                 there are 4*4=16 contact
                                                 points per face, and
                                                 so on*/
}              WPC_faceted_penalty_contact_t;

typedef struct WPC_spheres_sphere_t {
  W_time_func_t                   *x_tf; /* drives evolution of radius */
  W_time_func_t                   *y_tf; /* drives evolution of radius */
  W_time_func_t                   *z_tf; /* drives evolution of radius */
#if defined(X_XFACE)
  char                            *egf_file;
  LIST_OF(GraphicObj *)            ogrepl;
  LIST_OF(GraphicObj *)            grepl;
#endif
  W_time_func_t                   *radius_tf; /* drives evolution of radius */
  W_vect_t                         center; /* cached: must be updated from the tf */
  double                           radius; /* cached: must be updated from the tf */
}              WPC_spheres_sphere_t;

/*
  Penalty-based contact. Specialized spherical surface.
 */
typedef struct WPC_spheres_penalty_contact_t {
  struct W_contact_surf_t         *cont_surf; /* contact surface */
  int                              first_time;
  W_gap_penalty_type               force_type; /* type of contact force */
  double                           overlap_penalty; /* penalty for overlap */
  double                           overlap_rate_penalty; /* penalty
                                                          for overlap velocity */
  double                           avgnmass; /* average nodal mass */
  LIST_OF(WPC_spheres_sphere_t *)  sl;
}              WPC_spheres_penalty_contact_t;

typedef enum W_contact_surf_type {
  W_FACETED_PENALTY_CONTACT_SURF = 0,
  W_SPHERES_PENALTY_CONTACT_SURF = 1
}            W_contact_surf_type;

/*
  Contact surface.
 */
typedef struct W_contact_surf_t {
  int                              id; /* unique integer */
  struct W_domain_t               *domain; /* domain */
  W_contact_surf_type              type; /* type */
  double                           search_eps; /* physical tolerance for searching */
  union {
    WPC_faceted_penalty_contact_t *faceted_penalty;
    WPC_spheres_penalty_contact_t *spheres_penalty;
  }                                data; /* type-specific data */
  LIST_OF (W_elem_t *)             celems; /* list of contact elements */
  W_bool_t                         all_rigid; /* are all contact faces rigid? */
}              W_contact_surf_t;


typedef struct W_initial_velocity_cond_t *W_initial_velocity_cond_p;

typedef struct W_modr_t *W_modr_p;

typedef int (*W_check_interrupt_func) (struct W_domain_t *d);

struct W_error_handler_t;
struct W_log_handler_t;

typedef struct W_dict_t {
  dict_dictionary d;
  dict_iterator   diter;
}              W_dict_t;

typedef void (*W_finalize_callback_func)(void *client);

typedef struct W_finalize_callback_t{
  void                      *client; /* 3 types of material? */
  W_finalize_callback_func  callback;
}              W_finalize_callback_t;

/*
  Domain.
 */
typedef struct W_domain_t {
  struct W_options_t                   *options;
  struct W_error_handler_t             *errh;
  struct W_log_handler_t               *logh;
  tokens_parser                         parser; /* input parser */
  char                                 *comment; /* comment read from input */
  struct W_history_writer_t            *history_writer; /* history writer */
  struct W_cd_integrator_t             *cd_integrator; /* central-difference
                                                     integrator */
  double                                stable_time_step;
  struct W_newmark_integrator_t        *newmark_integrator;
  W_newmark_get_incr_proc               newmark_get_incr;
  W_newmark_predictor_enum              newmark_predictor;
  W_dict_t                             *nodes; /* dictionary of nodes */
  W_dict_t                             *elements; /* dictionary of elements */
  LIST_OF(W_elem_t *)                   elems_reg_for_integrate; /* elements
                                                                  registered for
                                                                 integration
                                                                 of internal dofs */
  W_dict_t                             *materials; /* dictionary of materials */
  W_dict_t                             *time_funcs; /* dictionary of time functions
                                                     */
  W_dict_t                             *loads; /* dictionary of loads */
  W_dict_t                             *disp_bcs; /* dictionary of displacement
                                                   boundary conditions */
  W_dict_t                             *cont_surfs; /* dictionary of contact
                                                     surfaces */
  W_dict_t                             *iv_conds; /* dictionary of initial
                                                   conditions */
  W_float_t                             ext_load_work; /* work of external
                                                        loads */
  W_float_t                             total_mass; /* total mass of domain */
  W_float_t                             avg_nodal_mass; /* average nodal mass */
  double                                mass_damping_coeff; /* mass damping
                                                             coefficient */
  int                                   mass_damping_time_func; /* mass damping
                                                                 time function
                                                                */
  double                                curr_mass_damping_coeff; /* current mass
                                                                  damping
                                                                  coefficient */
#if defined(PVM_WASP_MASTER)
  W_pvm_master_t                       *pvm_master;
#endif
#if defined(PVM_WASP_WORKER)
  W_pvm_worker_t                       *pvm_worker;
#endif 
#if defined (MPI_WASP) && MPI_WASP
  W_mpi_process_t                       mpi_process;
#endif
#if TIMED_WASP
  W_timer_t                            *effl_timer;
  W_timer_t                            *solv_timer;
  W_timer_t                            *ucfg_timer;
  W_timer_t                            *fina_timer;
#if defined (MPI_WASP) && MPI_WASP
  W_timer_t                            *comm_timer;
#endif
#endif
  W_modr_p                              modr;
  struct expl_modr_data_t              *expl_modr_data;
  W_response_observer_func              response_observer_func;
  W_check_interrupt_func                check_interrupt;
  LIST_OF(W_finalize_callback_t *)      finalize_callbacks;
}              W_domain_t;

/*
  Type of reader function.
  Each reader is designed to read a block from the input stream.
  See User's guide for the notion of a block.
 */
typedef W_bool_t (*reader_proc)(W_domain_t *d);
     


/*
  Task type.  Only 3-D implemented at present.
 */
typedef enum { TASK_3D                  = 0,
               TASK_PLANE_STRESS_STRAIN = 1
             } W_task_type_t;


/*
  LIst of times for post-processing file output.
 */
typedef struct W_postfile_times_list_t {
  int     curr; /* current time */
  int     ntimes; /* total of times specified */
  double *times; /* list of times */
}              W_postfile_times_list_t;

typedef struct W_postfile_output_specs_t {
  int    nspecs;
  char **specs;
} W_postfile_output_specs_t;

/*
  Stress tensor type for output.
 */
typedef enum W_output_stress_type_enum {
  W_OUTPUT_STRESS_CAUCHY = 0, /* Cauchy */
  W_OUTPUT_STRESS_2ND_PK = 2 /* 2nd Piola-Kirchhoff */
}            W_output_stress_type_enum;


typedef struct W_probe_package_t {
  union {
    double scalar;
    double vector[3];
    double tensor[3][3];
  } value;
}              W_probe_package_t;


typedef enum W_probe_type_enum {
  W_PROBE_DISPL_MASK               = (1L<<0),
  W_PROBE_STRESS_MASK              = (1L<<1),
  W_PROBE_EQPLS_MASK               = (1L<<2),
  W_PROBE_VON_MISES_MASK           = (1L<<3),
  W_PROBE_MATERIAL_MASK            = (1L<<4),
  W_PROBE_DEF_ENERGY_DENSITY_MASK  = (1L<<5)
}            W_probe_type_enum;


#define W_POSTFILE_DISPL_MASK                  W_PROBE_DISPL_MASK
#define W_POSTFILE_STRESS_MASK                 W_PROBE_STRESS_MASK 
#define W_POSTFILE_EQPLS_MASK                  W_PROBE_EQPLS_MASK
#define W_POSTFILE_VON_MISES_MASK              W_PROBE_VON_MISES_MASK
#define W_POSTFILE_MATERIAL_MASK               W_PROBE_MATERIAL_MASK
#define W_POSTFILE_DEF_ENERGY_DENSITY_MASK    W_PROBE_DEF_ENERGY_DENSITY_MASK

/*
  Type of virtual extension function.
 */
typedef enum W_q_func_enum {
  W_QUADRATIC_Q_FUNC           = 0, /* C^0 quadratic */
  W_QUARTIC_Q_FUNC             = 1, /* C^2 quartic */
  W_LINEAR_Q_FUNC              = 2, /* C^0 linear */
  W_COMPOSITE_QUADRATIC_Q_FUNC = 3 /* C^1 composite quadratic */
}            W_q_func_enum;


/*
  Evaluate domain integrals for SIF determination
 */
typedef enum W_j_dom_eval_enum {
  W_PER_CELL_WF_PTS   = 0, /* weak-form (Gauss) points */
  W_PER_CELL_IPOL_PTS = 1 /* per-cell re-interpolation from Gauss points */
}            W_j_dom_eval_enum;

/*
  Evaluate stress intensity factors from
  interaction integral, J-integral, or ...
 */
typedef enum W_sifs_eval_from_enum {
  W_SIFS_FROM_II_INTEGRAL = 0,
  W_SIFS_FROM_J_INTEGRAL  = 1
}            W_sifs_eval_from_enum;


struct W_options_t;
  
/*
  History writer.
 */
typedef struct W_history_writer_t {
  W_domain_t       *domain; /* created for domain */
  char             *history_dir; /* history disk directory */
  int               hist_buff_length;
  dict_dictionary   var_hist_buff_d;
  LIST_OF(void *)   node_clients; /* list of node clients */
  LIST_OF(void *)   elem_clients; /* list of finite element clients  */
  W_bool_t          newmark_client;
  W_bool_t          domain_client;
}              W_history_writer_t;

/*
  Solution history record.
 */
typedef struct W_solution_history_t {
  W_history_writer_t            *writer; /* history writer */
  void                          *client_data; /* client data */
}              W_solution_history_t;

/*
  Gauss integration point for isoparametric elements.
 */
typedef struct W_gauss_integration_point_t {
  double  Loc;                  /* location */
  double  Wgt;                  /* weight */
}              W_gauss_integration_point_t;

/*
  GAuss integration rule for isoparametric elements.
 */
typedef struct W_gauss_integration_rule_t {
  int                           order; /* order of quadrature */
  W_gauss_integration_point_t  *ip; /* quadrature points, ie. ip is an array
                                       of type ip[order] */
}              W_gauss_integration_rule_t;

#ifdef __cplusplus
}
#endif


#endif /* !WTYPES_H */
