#ifndef WUTIL_H
#   define WUTIL_H

/* RCSHEADER "$Id: Wutil.h,v 1.2 2002/11/08 20:34:17 cvs Exp $" */

#include <string.h>

#include "Werrh.h"
#include "Wlogh.h"

const char *
W_get_input_line (W_domain_t *d);
W_bool_t
W_is_end_of_block (const char *s);

/* Macros for access to the parsed line */
#define TOKEN_DBL(ith) ((tokens_type_of_token(d->parser, ith) == TOKENS_NUMBER)    \
                        ? tokens_token_as_double(d->parser, ith)                   \
                        : (double)WEH_parse_error(d->errh, __FILE__, __LINE__,     \
                                                  ith, TOKENS_NUMBER, tokens_curr_line_num (d->parser)))
#define TOKEN_INT(ith) ((tokens_type_of_token(d->parser, ith) == TOKENS_NUMBER)    \
                        ? tokens_token_as_int(d->parser, ith)                      \
                        : (long)WEH_parse_error(d->errh, __FILE__, __LINE__,       \
                                                  ith, TOKENS_NUMBER, tokens_curr_line_num (d->parser)))
#define TOKEN_STR(ith) (tokens_token_as_string(d->parser, ith))
#define TOKEN_IS_NUMBER(ith) (tokens_type_of_token(d->parser, ith) == TOKENS_NUMBER)
#define TOKEN_IS_STRING(ith) (tokens_type_of_token(d->parser, ith) == TOKENS_STRING)
#define TOKENS_TOTAL() tokens_total_of_tokens(d->parser)
#define TOKENS_PARSE(buf)                                       \
  if (!tokens_parse_line(d->parser, (char *)buf))               \
    {REPORT_FATAL_ERROR(d->errh, "Failed parsing input");}
#define TOKENS_HAVE_KEYWORD(keyword) (tokens_have_keyword(d->parser, 1, keyword) > 0)

#define TOKENS_SET_BY_KEYWORD(keyw, required, if_present, if_not_present)       \
  {                                                                             \
    int _kpos_ = tokens_have_keyword(d->parser, 1, keyw);                       \
                                                                                \
    if (_kpos_ <= 0) {                                                          \
      int keyword_is_required = required;                                       \
      if (keyword_is_required) {                                                \
	sprintf(W_buf(), "Keyword \"%s\" required", keyw);                      \
	REPORT_FATAL_ERROR(d->errh, W_buf());                                   \
      }                                                                         \
      if_not_present                                                            \
    } else {                                                                    \
      if_present                                                                \
    }                                                                           \
  }
#define TOKEN_STR_NEXT(ijk) TOKEN_STR(_kpos_+ijk)
#define TOKEN_DBL_NEXT(ijk) TOKEN_DBL(_kpos_+ijk)
#define TOKEN_INT_NEXT(ijk) TOKEN_INT(_kpos_+ijk)
#define NOOP() { int _foo_ = 0; while(_foo_); }
  
#define BAD_INPUT() \
   { WEH_error(d->errh, __FILE__, __LINE__, E_K_FATAL, (char *) WEH_read_failure(d->errh, tokens_curr_line_num (d->parser),  (char *) input_line)); } 

#define UNRECOGNIZED_INPUT() {WEH_error(d->errh, __FILE__, __LINE__, E_K_FATAL, (char *) WEH_unrecognized_input_error (d->errh, __FILE__, __LINE__, tokens_curr_line_num (d->parser), (char *) input_line));}

#define CHECK_MALLOC(ptr) \
   { if (ptr == NULL) { REPORT_FATAL_ERROR (NULL, "failed malloc"); } }




/* Ordinary loop */
#define FOR(_Counter_, _Up_)                                      \
   for (_Counter_ = 0; _Counter_ < _Up_; _Counter_++)


#ifndef min
#   define min(A,B) (((A) < (B)) ? (A) : (B))
#endif
#ifndef max
#   define max(A,B) (((A) > (B)) ? (A) : (B))
#endif
#ifndef min_of_3
#   define min_of_3(A,B,C)           \
          (((A) < (B))?              \
           (((C) < (A))? (C): (A)):  \
           (((C) < (B))? (C): (B))   \
           )
#   define max_of_3(A,B,C)           \
          (((A) > (B))?              \
           (((C) > (A))? (C): (A)):  \
           (((C) > (B))? (C): (B))   \
           )
#endif


/* Efficient approximation of log(x) in the neighbourhood of x=1 */
#if 0
#define LN_FUNC(lnx, x)                                                   \
   { double _t_ = x;                                                      \
     lnx = ( -(1 - _t_)                                                   \
            * (1 + (1 - _t_) * (0.5 + 0.3333333333333333333 * (1 - _t_))) \
            );                                                            \
   }
#else
#define LN_FUNC(lnx, x) lnx = log(x)
#endif

/* Debugging malloc */
#ifdef DMALLOC
#  include "dmalloc.h"
#endif

#ifdef INQUISITIVE_MEM_ALLOC
caddr_t
logging_malloc(char *file, int line, char *type, int bytes);
caddr_t
logging_calloc(char *file, int line, char *type, int nelms, int bytes);
caddr_t
logging_realloc(char *file, int line, char *type, caddr_t ptr,
                int nelms, int bytes);

#define w_new(TYPE) \
  ((TYPE *) logging_malloc(__FILE__, __LINE__, #TYPE, sizeof(TYPE)))

#define w_new_array(NELT, TYPE)                             \
  ((TYPE *) logging_calloc(__FILE__, __LINE__, #TYPE,       \
                           (NELT), (size_t) sizeof(TYPE)))

#define w_realloc_array(PTR, NEW_SIZE, TYPE)                          \
   ((TYPE *) logging_realloc(__FILE__, __LINE__, #TYPE, (caddr_t)PTR, \
                             NEW_SIZE, sizeof(TYPE)))

#define w_free(PTR) free(PTR)
#else  /* INQUISITIVE_MEM_ALLOC  */
#define w_new(TYPE) ((TYPE *) malloc(sizeof(TYPE)))

#define w_new_array(NELT, TYPE) \
                      ((TYPE *) calloc((NELT), (size_t) sizeof(TYPE)))

#define w_realloc_array(PTR, NEW_SIZE, TYPE) \
   ((TYPE *) realloc(PTR, (size_t)(NEW_SIZE * sizeof(TYPE))))

#define w_free(PTR) free(PTR)
#endif  /* INQUISITIVE_MEM_ALLOC  */

#define DUPLICATE_STRING(dest, src)                             \
   if (((dest) = w_new_array(strlen(src)+1, char)) == NULL) {   \
      dest = NULL;                                              \
      CHECK_MALLOC(dest);                                       \
   }                                                            \
   strcpy((dest), (src))


double 
W_vector_3d_length(W_coords_3d_t *vec);

void 
W_cross_prod_3d(W_coords_3d_t *reslt_vec,
                W_coords_3d_t *Vec1, W_coords_3d_t *Vec2);

W_float_t
W_dot_prod_3d(W_coords_3d_t *Vec1, W_coords_3d_t *Vec2);

W_bool_t
W_normalize_vect_3d(W_coords_3d_t *vec);

void
W_rotate_triad_by_vector(W_triad_3d_t *t, W_coords_3d_t *v);

void
W_L_shift_triad(W_triad_3d_t *out, W_triad_3d_t *by, W_triad_3d_t *in);

void
W_transp_L_shift_triad(W_triad_3d_t *out, W_triad_3d_t *by, W_triad_3d_t *in);

void
W_R_shift_triad(W_triad_3d_t *out, W_triad_3d_t *in, W_triad_3d_t *by);

void
W_transp_R_shift_triad(W_triad_3d_t *out, W_triad_3d_t *in, W_triad_3d_t *by);

void
W_extract_rot_vector_from_triad(W_triad_3d_t *ROTM, W_coords_3d_t *ROTV);

void
W_extract_triad_from_rot_vector(W_coords_3d_t *rotv, W_triad_3d_t *rotm);

void
W_extract_quaternion_from_triad(W_triad_3d_t *rotm, W_quaternion_t *q);

void
W_extract_triad_from_quaternion(W_quaternion_t *q, W_triad_3d_t *rotm);

#ifdef DEBUG
#define DBG(l) l
#undef  NDEBUG
#else
#define DBG(l) 
#define NDEBUG                  /* define to get rid of assert */
#endif


#define DebugPV(arg1, arg2, arg3)   /* PV(arg1, arg2, arg3) */


#define RAW                     /* can also define as !!! */

#include <assert.h>


#ifdef MSDOS
#   define FILE_ACCESSIBLE(f) (access((f), 0) == 0)
#else
#   define FILE_ACCESSIBLE(f) (W_file_exists(f))
#endif


void 
W_get_date_and_time(char *return_string);
W_bool_t 
W_has_suffix(char *f, char *s);
W_bool_t 
W_file_exists(char *f);

W_float_t
W_eval_ordinate_at_abscissa(W_index_t num_of_entries,
                            W_abscissa_ordinate_pair_t *pairs,
                            W_float_t abscissa);


int
pv(char *name, W_float_t *v, int row, int col);
#define PV(v, row, col) pv(#v, (W_float_t *)v, row, col)


W_bool_t
W_get_times(double *real_time,
            double *usr_time,
            double *systm_time); /* Returns the real, user and system
                                      time for the process */

char *
W_buf(void);

char *
W_postfile_name(double time, char *basename, int type, W_bool_t number_seq);

#define W_INDENT(stream, indent) \
   { int k__; FOR(k__, indent) fprintf(stream, " "); }

int
W_check_interrupt (W_domain_t *d);

double
W_f_quartic_spline(double r);
double 
W_f_r_quartic_spline(double r);
double
W_f_quadratic(double r);
double 
W_f_r_quadratic(double r);
double
W_f_linear(double r);
double 
W_f_r_linear(double r);
double 
W_f_composite_quadratic(double x);
double 
W_f_r_composite_quadratic(double x);


int
W_write_hexvue_hex (ckit_iobuf_t IOB,
                    G3D_loc_t *p[8],
                    int num_of_values,
                    double *values[8],
                    char *mode
                    );

int
W_write_hexvue_tri3 (ckit_iobuf_t IOB,
                     G3D_loc_t *p[3],
                     int num_of_values,
                     double *values[3],
                     char *mode
                     );
int
W_write_hexvue_tet4 (ckit_iobuf_t IOB,
                     G3D_loc_t *p[4],
                     int num_of_values,
                     double *values[4],
                     char *mode
                     );
/*
   Compute the angle of fracture propagation.
   Compute the angle of fracture propagation using the maximum
   circumferential stress criterion.
   The coordinate system is as follows:
   e1 ... normal to the crack front, in crack plane
   e2 ... normal to both the crack front and the crack plane
   e3 ... tangent to the crack front

   K_I is required to be positive (crack is opening, not closing)
   K_II positive should yields negative angle, K_II negative yields
   a positive angle; in all cases the absolute value of theta is < PI/2.
 */
W_bool_t
W_compute_theta (double K_I, double K_II, double *theta);

/*
   Compute equivalent opening SIF.
 */
double
W_equiv_KI (double KI, /* opening SIF */
          double KII, /* shearing SIF */
          double theta /* angle in the forward/normal plane */
          ) ;

int
W_compute_Rayleigh_speed (double C_d, double C_s,
                          double percent_error,
                          double *C_R);

double
W_3D_dilatational_wave_speed (double E, double nu, double rho);

double
W_plstress_dilatational_wave_speed (double E, double nu, double rho);

double
W_plstrain_dilatational_wave_speed (double E, double nu, double rho);

double
W_shear_wave_speed (double E, double nu, double rho);

W_bool_t
W_read_and_discard_block (W_domain_t *d);

double
W_timer_get_time (W_timer_t *timer);

W_timer_t *
W_timer_new (void);

void
W_timer_restart (W_timer_t *timer);

void
W_timer_accumulate (W_timer_t *timer);

W_dict_t *
W_dict_new_dictionary (void);

/*
  Iterate objects in dictionary.
  Iterate objects in dictionary.  Pass current object as curr.
  W_dict_t dictionary uses dict_dictionary dict_iterator from Ckit.
 */
void * /* next object or NULL if at end */
W_dict_next_obj (W_dict_t *dict, /* dictionary */
                 void *curr /* current object or NULL if starting */
                 );

/*
  Find object in dictionary.
  Find object in WASP dictionary by index (positive integer).
 */
void * /* object or NULL if not found */
W_dict_find_by_idx (W_dict_t *dict, /* dictionary */
                    W_index_t idx /* index */
                    );

void
W_dict_add (W_dict_t *dict, /* dictionary */
            void *obj, /* object */
            W_index_t idx /* index */
            );
int
W_dict_count(W_dict_t *dict);

W_bool_t
W_polar_decomposition (double F[3][3], double R[3][3], double U[3][3]);

/*
  Principal directions and eigenvalues of a symmetric
  tensor.
 */
W_bool_t
W_eigendecomposition (      
               double t[3][3],/* IN symmetric tensor */
               double t_eig_vecs[3][3], /* OUT each row */
               /* is an eigenvector */
               double t_eig_vals[3] /* OUT principal stretches */
               ) ;

void
W_ln_U (double U[3][3], double lnU[3][3]);

#endif

