#ifndef WVSET_H
#   define WVSET_H

#include "Wtypes.h"
#include "Wutil.h"
#include "hashP.h"
#include <stdlib.h>

#define HASH_TABLE_DIM 10


#define VSET_NAMED_ITEM_TABLE 1
#define VSET_ARRAY_TABLE      2


typedef struct  W_vset_named_item_t {
  char            *name;
  int              type;
  union {
    double     d;
    char       *c;
  }                value;
}               W_vset_named_item_t;


typedef struct W_vset_named_item_table_t {
  int                          num_of_items;
  W_vset_named_item_t          *item;
}              W_vset_named_item_table_t;


typedef struct W_vset_array_table_t {
  int                          num_of_rows;
  int                          num_of_cols;
  W_float_t                    *item;
}              W_vset_array_table_t;



typedef struct W_vset_table_t {
  int     table_type;
  int     table_number;
  union {
    W_vset_named_item_table_t  named_item_table;
    W_vset_array_table_t       array_table;
  } data;
}              W_vset_table_t;

W_vset_table_t *
WVS_new_vset_named_item_table(void);


W_vset_table_t *
WVS_new_vset_array_table(int table_number, int n_rows, int n_cols);

int
WVS_get_vset_table_type(W_vset_table_t *t);

int
WVS_n_of_rows_in_array_table(W_vset_table_t *t);

int
WVS_n_of_cols_in_array_table(W_vset_table_t *t);

void
WVS_put_into_array_table(W_vset_table_t *t, int row, int col,
                         W_float_t item);

void
WVS_get_from_array_table(W_vset_table_t *t, int row, int col,
                         W_float_t *item_return);


W_vset_table_t *
WVS_get_table_by_number(int table_number, int table_type);

W_bool_t 
W_vset_table_reader(W_domain_t *d);



#endif /* !WVSET_H */
