
/* Pick the way in which the local frame is computed: either define */
/* FROM_POLAR_DECOMPOSITION or undefine it. */
#ifdef FROM_POLAR_DECOMPOSITION
#  if FROM_POLAR_DECOMPOSITION == 0
#    undef FROM_POLAR_DECOMPOSITION
#  endif
#endif


#undef HANTI

#include "Wconfig.h"


#define WASQBI3D_PRIVATE_HEADER
#include "Wasqbi3d.h"
#include "Wdomain.h"
#include "Wnode.h"
#include "Wmats.h"
#include "Wgaussp.h"
#include "Wxf.h"
#include "Wutil.h"
#include "Wcdi.h"
#include "g3d.h"
#include "g3dhex.h"

/* RCSHEADER "$Id: asqbi3d.c,v 1.1.1.1 2001/07/24 02:42:51 pkrysl Exp $" */

static int debug_print = 0;


typedef struct priv_elem_t {
  W_node_t         *node[8];
  W_float_t         volume; /* Element volume */
  W_float_t         lb1[3]; /* Local basis vectors in the reference config. */
  W_float_t         lb2[3];
  W_float_t         lb3[3];
  W_index_t         material;
#ifdef ASQBI_FULL_INTEGRATION
#  ifndef ASQBI_FULL_INTEGRATION_ORDER
#    define ASQBI_FULL_INTEGRATION_ORDER 2
#  endif
#  define theDim ASQBI_FULL_INTEGRATION_ORDER 
  W_matstate_t      matstate[theDim][theDim][theDim];
#  undef  theDim
#else
  W_matstate_t      matstate;
  W_float_t         hg_stress[3][4]; /* generalized hourglass stresses */
#endif
}              priv_elem_t;

#define GET_FROM_NODE(node, func, tovect)         \
   func (node, &tovect.x, &tovect.y, &tovect.z)


static int
elem_elcode (W_elem_t *el);
static char *
type_name(W_elem_t *el);
#ifdef ASQBI_FULL_INTEGRATION
static void
assemble_full_integration_restoring_forces(W_elem_t *el, W_mat_state_update_form_enum update_form);
#else
static void
assemble_restoring_forces(W_elem_t *el, W_mat_state_update_form_enum update_form);
#endif
static void
assemble_mass(W_elem_t *el);
static W_float_t 
suggested_time_step(W_elem_t *el);
#ifdef X_XFACE
static void
modify_elem_grep(W_elem_t *el, double displ_scale, W_bool_t immed_draw)
{
  WCRec pnts[8];
  W_coords_3d_t u[8], refc[8];
  W_material_t *m;
  priv_elem_t *t = (priv_elem_t *)el->data;
  int i;

  for (i = 0; i < 8; i++)  {
    GET_FROM_NODE (t->node[i], WN_get_coords, u[i]);    
    GET_FROM_NODE (t->node[i], WN_get_ref_coords, refc[i]); 
  }
  
  for (i = 0; i < 8; i++) {
    pnts[i].x = refc[i].x + displ_scale * (u[i].x);
    pnts[i].y = refc[i].y + displ_scale * (u[i].y);
    pnts[i].z = refc[i].z + displ_scale * (u[i].z);
  }

  m = WD_get_material_by_idx(el->domain, t->material);
  if (el->grep == NULL) {
    el->grep = CreateHexahedron(pnts);
    EASValsSetColor(WM_get_color(m));
    EGWithMaskChangeAttributes(COLOR_MASK, el->grep);
    EGAttachObject(el->grep, (EObjectP)el);
    EMNoDrawAddGraphicsToModel(ESIModel (), el->grep);
  } else {
    EGModifyGraphicsGeometry(el->grep, pnts);
  }
}
#endif

static W_float_t
total_mass(W_elem_t *el);
static void
write_postproc_file (W_elem_t *el, ckit_iobuf_t IOB, int type,
                     char *mode, void *optdata);



static W_elem_methods_t asqbi3d_methods = {
  elem_elcode,
  type_name,
#ifdef ASQBI_FULL_INTEGRATION
  assemble_full_integration_restoring_forces,
#else
  assemble_restoring_forces,
#endif
  assemble_mass,
  total_mass,
  NULL,
  suggested_time_step,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  write_postproc_file,
#ifdef X_XFACE
  modify_elem_grep,
#endif
  NULL,
  NULL
};


#define ALL_REF_COORDS    refc[0].x, refc[0].y, refc[0].z,   \
                          refc[1].x, refc[1].y, refc[1].z,   \
                          refc[2].x, refc[2].y, refc[2].z,   \
                          refc[3].x, refc[3].y, refc[3].z,   \
                          refc[4].x, refc[4].y, refc[4].z,   \
                          refc[5].x, refc[5].y, refc[5].z,   \
                          refc[6].x, refc[6].y, refc[6].z,   \
                          refc[7].x, refc[7].y, refc[7].z
#define ALL_CURR_COORDS   c[0].x, c[0].y, c[0].z,    \
                          c[1].x, c[1].y, c[1].z,    \
                          c[2].x, c[2].y, c[2].z,    \
                          c[3].x, c[3].y, c[3].z,    \
                          c[4].x, c[4].y, c[4].z,    \
                          c[5].x, c[5].y, c[5].z,    \
                          c[6].x, c[6].y, c[6].z,    \
                          c[7].x, c[7].y, c[7].z
#define ALL_REF_HAT_COORDS    Xhat[X][0], Xhat[Y][0], Xhat[Z][0],       \
                              Xhat[X][1], Xhat[Y][1], Xhat[Z][1],       \
                              Xhat[X][2], Xhat[Y][2], Xhat[Z][2],       \
                              Xhat[X][3], Xhat[Y][3], Xhat[Z][3],       \
                              Xhat[X][4], Xhat[Y][4], Xhat[Z][4],       \
                              Xhat[X][5], Xhat[Y][5], Xhat[Z][5],       \
                              Xhat[X][6], Xhat[Y][6], Xhat[Z][6],       \
                              Xhat[X][7], Xhat[Y][7], Xhat[Z][7]
#define ALL_CURR_HAT_COORDS   xhat[X][0], xhat[Y][0], xhat[Z][0],       \
                              xhat[X][1], xhat[Y][1], xhat[Z][1],       \
                              xhat[X][2], xhat[Y][2], xhat[Z][2],       \
                              xhat[X][3], xhat[Y][3], xhat[Z][3],       \
                              xhat[X][4], xhat[Y][4], xhat[Z][4],       \
                              xhat[X][5], xhat[Y][5], xhat[Z][5],       \
                              xhat[X][6], xhat[Y][6], xhat[Z][6],       \
                              xhat[X][7], xhat[Y][7], xhat[Z][7]


/* Hourglass-mode vectors */
static W_float_t h0[8] = { 1,  1, -1, -1, -1, -1,  1,  1};
static W_float_t h1[8] = { 1, -1, -1,  1, -1,  1,  1, -1};
static W_float_t h2[8] = { 1, -1,  1, -1,  1, -1,  1, -1};
static W_float_t h3[8] = {-1,  1, -1,  1,  1, -1,  1, -1};


/* Vectors Lambda */
static W_float_t LambdaX[8] = {-1,  1,  1, -1, -1,  1,  1, -1};
static W_float_t LambdaY[8] = {-1, -1,  1,  1, -1, -1,  1,  1};
static W_float_t LambdaZ[8] = {-1, -1, -1, -1,  1,  1,  1,  1};


static W_float_t xhat[3][8];
static W_float_t Xhat[3][8];

static W_float_t LX, LY, LZ;

#define LAMBDA_X(dir, x) (  Lambda##dir[0] * x[dir][0]  \
                          + Lambda##dir[1] * x[dir][1]  \
                          + Lambda##dir[2] * x[dir][2]  \
                          + Lambda##dir[3] * x[dir][3]  \
                          + Lambda##dir[4] * x[dir][4]  \
                          + Lambda##dir[5] * x[dir][5]  \
                          + Lambda##dir[6] * x[dir][6]  \
                          + Lambda##dir[7] * x[dir][7])
#define H_X(alpha, dir, x) (  h##alpha[0] * x[dir][0]  \
                            + h##alpha[1] * x[dir][1]  \
                            + h##alpha[2] * x[dir][2]  \
                            + h##alpha[3] * x[dir][3]  \
                            + h##alpha[4] * x[dir][4]  \
                            + h##alpha[5] * x[dir][5]  \
                            + h##alpha[6] * x[dir][6]  \
                            + h##alpha[7] * x[dir][7])

static W_float_t H[3][3];

static void
compute_local_frame(/* IN current coordinates */
                    W_float_t x1, W_float_t y1, W_float_t z1, 
                    W_float_t x2, W_float_t y2, W_float_t z2,
                    W_float_t x3, W_float_t y3, W_float_t z3,
                    W_float_t x4, W_float_t y4, W_float_t z4,
                    W_float_t x5, W_float_t y5, W_float_t z5, 
                    W_float_t x6, W_float_t y6, W_float_t z6,
                    W_float_t x7, W_float_t y7, W_float_t z7,
                    W_float_t x8, W_float_t y8, W_float_t z8,
                    /* IN reference coordinates */
                    W_float_t X1, W_float_t Y1, W_float_t Z1, 
                    W_float_t X2, W_float_t Y2, W_float_t Z2,
                    W_float_t X3, W_float_t Y3, W_float_t Z3,
                    W_float_t X4, W_float_t Y4, W_float_t Z4,
                    W_float_t X5, W_float_t Y5, W_float_t Z5, 
                    W_float_t X6, W_float_t Y6, W_float_t Z6,
                    W_float_t X7, W_float_t Y7, W_float_t Z7,
                    W_float_t X8, W_float_t Y8, W_float_t Z8,
                    /* IN reference frame */
                    W_float_t refb1[3], W_float_t refb2[3],W_float_t refb3[3], 
                    /* OUT current frame */
                    W_float_t b1[3], W_float_t b2[3],W_float_t b3[3],
                    W_float_t center0[3], W_float_t center[3]);


#ifdef HANTI
static int 
hanti(W_float_t *x, W_float_t *v,    /* IN local coordinates and velocities */
      W_float_t *nx, W_float_t *ny, W_float_t *nz, /* IN strain-displ. submat*/
      W_float_t delt,                /* IN time increment */
      W_float_t twomu, W_float_t c1, W_float_t c2, W_float_t c3,
      /* IN elastic constants above */
      W_float_t *fx, W_float_t *fy, W_float_t *fz, /* OUT hourgl. nodal forc*/
      W_float_t *hg /* OUT hourglass forces on element */
      );
static void
mean_value_B(W_float_t y[8], W_float_t z[8], W_float_t B[8]);
#endif


/* Reads asqbi3d element data from stream as long as there is no */
/* END_OF_BLOCK tag in the input line. Creates elements. */

W_bool_t 
asqbi3d_reader(W_domain_t *d)
{
  const char *input_line;
  long elem_num, i, nodeid;
  priv_elem_t *t;
#ifdef PVM_NEED_GHOST_ELEMENTS
  int is_ghost = FALSE;
#endif

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return TRUE;
    else {
      t = w_new(priv_elem_t);
      CHECK_MALLOC(t);
      TOKENS_PARSE(input_line);
      elem_num             = TOKEN_INT(1);
      FOR (i, 8) {
        nodeid = TOKEN_INT(2+i);
        t->node[i] = WD_get_node_by_idx (d, nodeid);
        if (t->node[i] == NULL) {
          sprintf (W_buf (), "No such node: %ld", nodeid);
          REPORT_FATAL_ERROR (d->errh, W_buf ());
        }
      }
      t->material          = TOKEN_INT(10);
      t->volume            = 0;
#ifdef ASQBI_FULL_INTEGRATION
      {
        int i, j, k;

        for (i = 0; i < ASQBI_FULL_INTEGRATION_ORDER; i++)
          for (j = 0; j < ASQBI_FULL_INTEGRATION_ORDER; j++) 
            for (k = 0; k < ASQBI_FULL_INTEGRATION_ORDER; k++) 
            t->matstate[i][j][k] = NULL;
      }
#else
      t->matstate          = NULL;
      for (i = 0; i < 3; i++)
        for (j = 0; j < 4; j++)
          t->hg_stress[i][j] = 0;
#endif
#ifdef PVM_NEED_GHOST_ELEMENTS
      is_ghost = FALSE;
      if (TOKENS_HAVE_KEYWORD("GHOST"))
        is_ghost = TRUE;
#endif      
      WD_add_element(d, elem_num,
                     WE_new_element(d, &asqbi3d_methods, t
#                                   ifdef PVM_NEED_GHOST_ELEMENTS
                                    , is_ghost
#                                   endif
                                    ));
    }
  }
  return FALSE;
}



static int
elem_elcode (W_elem_t *el)
{
  return ASQBI3D_ELCODE;
}


static char *
type_name(W_elem_t *el)
{
  static char buf[] = "asqbi3d";
  return buf;
}



/* Returns nodal forces by which the asqbi3d element acts on its nodal points. */

#ifndef ASQBI_FULL_INTEGRATION
#ifdef HANTI
static void
assemble_restoring_forces(W_elem_t *el, W_mat_state_update_form_enum update_form)
{
  W_coords_3d_t c[8], refc[8];
  W_coords_3d_t lf, gf;
  W_velocity_3d_t v[8];
  priv_elem_t *t = (priv_elem_t *)el->data;
  W_material_t *m;
  W_float_t eps_mean[6] = {0, 0, 0, 0, 0, 0};
  W_float_t eps_rate_mean[6] = {0, 0, 0, 0, 0, 0};
  W_float_t sigma_mean[6] = {0, 0, 0, 0, 0, 0};
  W_float_t jdetw;
  W_float_t center[3], center0[3], r[3], r0[3], dhat[3][8], vhat[3][8];
  W_float_t n_xi[8];
  W_float_t n_eta[8];
  W_float_t n_theta[8];
  W_float_t n_x[8], n_y[8], n_z[8];
  W_float_t n_x_mean[8]   = {0,0,0,0,0,0,0,0};
  W_float_t n_y_mean[8]   = {0,0,0,0,0,0,0,0};
  W_float_t n_z_mean[8]   = {0,0,0,0,0,0,0,0};
  W_float_t J[3][3], inv_J[3][3], F[3][3], R[3][3], U[3][3];
  W_float_t b1[3], b2[3], b3[3];
  double c1, c2, c3;
  W_float_t E, nu, mu, twomu, dt;
  W_gauss_integration_rule_t *gr;
  W_gauss_integration_point_t *ip;
  int i, j, k, p;
  W_triad_3d_t lb;
  W_float_t hfx[8], hfy[8], hfz[8];
#define X 0
#define Y 1
#define Z 2

  
  /* kinematic quantities */
  for (i = 0; i < 8; i++)  {
    GET_FROM_NODE (t->node[i], WN_get_coords, c[i]);    
    GET_FROM_NODE (t->node[i], WN_get_ref_coords, refc[i]); 
    GET_FROM_NODE (t->node[i], WN_get_velocity, v[i]);
  }
  
  m = WD_get_material_by_idx(el->domain, t->material);
  if (t->matstate == NULL) {
    /* General initialization */
    t->matstate = WM_get_mat_state(m);
    G3D_hex_local_basis(ALL_REF_COORDS, &lb.vect1, &lb.vect2, &lb.vect3);
    t->lb1[X] = lb.vect1.x; t->lb1[Y] = lb.vect1.y; t->lb1[Z] = lb.vect1.z;
    t->lb2[X] = lb.vect2.x; t->lb2[Y] = lb.vect2.y; t->lb2[Z] = lb.vect2.z;
    t->lb3[X] = lb.vect3.x; t->lb3[Y] = lb.vect3.y; t->lb3[Z] = lb.vect3.z;
  }

  compute_local_frame(ALL_REF_COORDS, ALL_CURR_COORDS,
                      t->lb1, t->lb2, t->lb3, b1, b2, b3, center0, center);
  
  for (i = 0; i < 8; i++) {
    r0[X]      = refc[i].x - center0[X];
    r0[Y]      = refc[i].y - center0[Y];
    r0[Z]      = refc[i].z - center0[Z];
    /* coords in the local reference system */
    Xhat[X][i] = r0[X] * t->lb1[X] + r0[Y] * t->lb1[Y] + r0[Z] * t->lb1[Z];
    Xhat[Y][i] = r0[X] * t->lb2[X] + r0[Y] * t->lb2[Y] + r0[Z] * t->lb2[Z];
    Xhat[Z][i] = r0[X] * t->lb3[X] + r0[Y] * t->lb3[Y] + r0[Z] * t->lb3[Z];
    r[X]       = c[i].x - center[X];
    r[Y]       = c[i].y - center[Y];
    r[Z]       = c[i].z - center[Z];
    xhat[X][i] = r[X] * b1[X] + r[Y] * b1[Y] + r[Z] * b1[Z];
    xhat[Y][i] = r[X] * b2[X] + r[Y] * b2[Y] + r[Z] * b2[Z];
    xhat[Z][i] = r[X] * b3[X] + r[Y] * b3[Y] + r[Z] * b3[Z];
    /* displacements in local coord system */
    dhat[X][i] = xhat[X][i] - Xhat[X][i];
    dhat[Y][i] = xhat[Y][i] - Xhat[Y][i];
    dhat[Z][i] = xhat[Z][i] - Xhat[Z][i];
    /* velocities in the local frame */
    vhat[X][i] = v[i].x * b1[X] + v[i].y * b1[Y] + v[i].z * b1[Z];
    vhat[Y][i] = v[i].x * b2[X] + v[i].y * b2[Y] + v[i].z * b2[Z];
    vhat[Z][i] = v[i].x * b3[X] + v[i].y * b3[Y] + v[i].z * b3[Z];
  }
/*  DebugPV(dhat, 3, 8); */
/*  DebugPV(vhat, 3, 8); */

  /* the B_mean matrix */
  mean_value_B(&xhat[Y][0], &xhat[Z][0], n_x_mean);
  mean_value_B(&xhat[Z][0], &xhat[X][0], n_y_mean);
  mean_value_B(&xhat[X][0], &xhat[Y][0], n_z_mean);

  /* Mean strains */
#define CEPS(N, INCR) \
    for (i = 0, eps_mean[N]=0; i < 8; i++) { eps_mean[N]+=INCR; }

  CEPS(0, n_x_mean[i] * dhat[X][i]);
  CEPS(1, n_y_mean[i] * dhat[Y][i]);
  CEPS(2, n_z_mean[i] * dhat[Z][i]);
  CEPS(3, n_y_mean[i] * dhat[X][i] + n_x_mean[i] * dhat[Y][i]);
  CEPS(4, n_z_mean[i] * dhat[X][i] + n_x_mean[i] * dhat[Z][i]);
  CEPS(5, n_z_mean[i] * dhat[Y][i] + n_y_mean[i] * dhat[Z][i]);

  /* Mean stresses */
  /* RAW not using update_form yet */
  WM_get_3d_stress(m, t->matstate, eps_mean, eps_rate_mean, sigma_mean);

  /* Hourglass forces */
  if (!WM_get_uniax_lin_el_consts(m, &E, &nu))
    REPORT_ERROR("No uniax_lin_el_consts");
  hanti((W_float_t *)xhat, (W_float_t *)vhat, n_x_mean, n_y_mean, n_z_mean,
        eid_cd_integrator_dt(el->domain->integrator),
        E / (1 + nu), 1 / (1 - nu), 2 * (1 + nu) / 3, nu / (1 - nu),
        hfx, hfy, hfz, (W_float_t *)t->hg_stress);
  
  /* Nodal forces */
#define COMPUTE_LOCAL_FORCE(N)                                          \
  {                                                                     \
    lf.x = (  t->volume * (  n_x_mean[N] * sigma_mean[0]                \
                           + n_y_mean[N] * sigma_mean[3]                \
                           + n_z_mean[N] * sigma_mean[4])               \
            + hfx[N]);                                                  \
    lf.y = (  t->volume * (  n_y_mean[N] * sigma_mean[1]                \
                           + n_x_mean[N] * sigma_mean[3]                \
                           + n_z_mean[N] * sigma_mean[5])               \
            + hfy[N]);                                                  \
    lf.z = (  t->volume * (  n_z_mean[N] * sigma_mean[2]                \
                           + n_x_mean[N] * sigma_mean[4]                \
                           + n_y_mean[N] * sigma_mean[5])               \
            + hfz[N]);                                                  \
  }

    /* Note the minus sign */
#define TRANSFORM_TO_GLOBAL()                                      \
  {                                                                \
    gf.x = -(lf.x * b1[X] + lf.y * b2[X] + lf.z * b3[X]);          \
    gf.y = -(lf.x * b1[Y] + lf.y * b2[Y] + lf.z * b3[Y]);          \
    gf.z = -(lf.x * b1[Z] + lf.y * b2[Z] + lf.z * b3[Z]);          \
  }

#define ASSEMBLE_GF(N)                          \
  {                                             \
    COMPUTE_LOCAL_FORCE(N);                     \
    TRANSFORM_TO_GLOBAL();                      \
    WN_assemble_force(n[N], &gf, el->domain);   \
  }

  ASSEMBLE_GF(0); ASSEMBLE_GF(1); ASSEMBLE_GF(2); ASSEMBLE_GF(3);
  ASSEMBLE_GF(4); ASSEMBLE_GF(5); ASSEMBLE_GF(6); ASSEMBLE_GF(7);
  
  /* Deformation energy */
  el->deformation_energy
    = 0.5 * t->volume
      *(  eps_mean[0] * sigma_mean[0] + eps_mean[1] * sigma_mean[1]
        + eps_mean[2] * sigma_mean[2] + eps_mean[3] * sigma_mean[3]
        + eps_mean[4] * sigma_mean[4] + eps_mean[5] * sigma_mean[5]);
}

#else  /* ifndef HANTI */
static void
assemble_restoring_forces(W_elem_t *el, W_mat_state_update_form_enum update_form)
{
  W_coords_3d_t *c[8], *refc[8];
  W_coords_3d_t lf, gf;
  W_velocity_3d_t *v[8];
  priv_elem_t *t = (priv_elem_t *)el->data;
  W_node_t *n[8];
  W_material_t *m;
  W_float_t eps_mean[6] = {0, 0, 0, 0, 0, 0};
  W_float_t eps_rate_mean[6] = {0, 0, 0, 0, 0, 0};
  W_float_t sigma_mean[6] = {0, 0, 0, 0, 0, 0};
  W_float_t jdetw;
  W_float_t center[3], center0[3], r[3], r0[3], dhat[3][8], vhat[3][8];
  W_float_t n_xi[8];
  W_float_t n_eta[8];
  W_float_t n_theta[8];
  W_float_t n_x[8], n_y[8], n_z[8];
  W_float_t n_x_mean[8]   = {0,0,0,0,0,0,0,0};
  W_float_t n_y_mean[8]   = {0,0,0,0,0,0,0,0};
  W_float_t n_z_mean[8]   = {0,0,0,0,0,0,0,0};
  W_float_t J[3][3], inv_J[3][3], F[3][3], R[3][3], U[3][3];
  W_float_t b1[3], b2[3], b3[3];
  W_float_t hg_strain_rate[3][4];
  W_float_t hg_stress_rate[3][4];
  double hax, hay, haz, c1, c2, c3;
  W_float_t E, nu, mu, twomu, dt;
  W_gauss_integration_rule_t *gr;
  W_gauss_integration_point_t *ip;
  int i, j, k, p;
  W_triad_3d_t lb;
  W_float_t gammahat[4][8];
#define X 0
#define Y 1
#define Z 2

  for (i = 0; i < 8; i++) 
    n[i]  = WD_get_node_by_idx(el->domain, t->node[i]); 
  
  /* kinematic quantities */
  for (i = 0; i < 8; i++)  {
    c[i]    = WN_get_coords(n[i]);
    refc[i] = WN_get_ref_coords(n[i]);
    v[i]    = WN_get_velocity(n[i]);
  }
  
  m = WD_get_material_by_idx(el->domain, t->material);
  if (t->matstate == NULL) {
    /* General initialization */
    t->matstate = WM_get_mat_state(m);
    G3D_hex_local_basis(ALL_REF_COORDS, &lb.vect1, &lb.vect2, &lb.vect3);
    t->lb1[X] = lb.vect1.x; t->lb1[Y] = lb.vect1.y; t->lb1[Z] = lb.vect1.z;
    t->lb2[X] = lb.vect2.x; t->lb2[Y] = lb.vect2.y; t->lb2[Z] = lb.vect2.z;
    t->lb3[X] = lb.vect3.x; t->lb3[Y] = lb.vect3.y; t->lb3[Z] = lb.vect3.z;
  }

  compute_local_frame(ALL_REF_COORDS, ALL_CURR_COORDS,
                      t->lb1, t->lb2, t->lb3, b1, b2, b3, center0, center);
  
  for (i = 0; i < 8; i++) {
    r0[X]      = refc[i]->x - center0[X];
    r0[Y]      = refc[i]->y - center0[Y];
    r0[Z]      = refc[i]->z - center0[Z];
    /* coords in the local reference system */
    Xhat[X][i] = r0[X] * t->lb1[X] + r0[Y] * t->lb1[Y] + r0[Z] * t->lb1[Z];
    Xhat[Y][i] = r0[X] * t->lb2[X] + r0[Y] * t->lb2[Y] + r0[Z] * t->lb2[Z];
    Xhat[Z][i] = r0[X] * t->lb3[X] + r0[Y] * t->lb3[Y] + r0[Z] * t->lb3[Z];
    r[X]       = c[i]->x - center[X];
    r[Y]       = c[i]->y - center[Y];
    r[Z]       = c[i]->z - center[Z];
    xhat[X][i] = r[X] * b1[X] + r[Y] * b1[Y] + r[Z] * b1[Z];
    xhat[Y][i] = r[X] * b2[X] + r[Y] * b2[Y] + r[Z] * b2[Z];
    xhat[Z][i] = r[X] * b3[X] + r[Y] * b3[Y] + r[Z] * b3[Z];
    /* displacements in local coord system */
    dhat[X][i] = xhat[X][i] - Xhat[X][i];
    dhat[Y][i] = xhat[Y][i] - Xhat[Y][i];
    dhat[Z][i] = xhat[Z][i] - Xhat[Z][i];
    /* velocities in the local frame */
    vhat[X][i] = v[i]->x * b1[X] + v[i]->y * b1[Y] + v[i]->z * b1[Z];
    vhat[Y][i] = v[i]->x * b2[X] + v[i]->y * b2[Y] + v[i]->z * b2[Z];
    vhat[Z][i] = v[i]->x * b3[X] + v[i]->y * b3[Y] + v[i]->z * b3[Z];
  }
/*  DebugPV(dhat, 3, 8); */
/*  DebugPV(vhat, 3, 8); */

  /* the B_mean matrix: first the N_i_mean (i=x,y,z) */
#define B_MEAN_INTEGRATION_ORDER 2
  gr = WG_get_integration_rule(B_MEAN_INTEGRATION_ORDER);
  ip = gr->ip;
  for (i = 0; i < B_MEAN_INTEGRATION_ORDER; i++) {
    for (j = 0; j < B_MEAN_INTEGRATION_ORDER; j++) {
      for (k = 0; k < B_MEAN_INTEGRATION_ORDER; k++) {
        G3D_hex_J(J, ip[i].Loc, ip[j].Loc, ip[k].Loc, ALL_REF_HAT_COORDS);
        G3D_hex_invJ(J, inv_J);
        G3D_hex_N_xi(n_xi, ip[i].Loc, ip[j].Loc, ip[k].Loc);
        G3D_hex_N_eta(n_eta, ip[i].Loc, ip[j].Loc, ip[k].Loc);
        G3D_hex_N_theta(n_theta, ip[i].Loc, ip[j].Loc, ip[k].Loc);
        G3D_hex_N_x(inv_J, n_xi, n_eta, n_theta, n_x);
        G3D_hex_N_y(inv_J, n_xi, n_eta, n_theta, n_y);
        G3D_hex_N_z(inv_J, n_xi, n_eta, n_theta, n_z);
        jdetw = ip[i].Wgt * ip[j].Wgt * ip[k].Wgt * G3D_hex_detJ(J);
        for (p = 0; p < 8; p++) {
          n_x_mean[p] += jdetw * n_x[p];
          n_y_mean[p] += jdetw * n_y[p];
          n_z_mean[p] += jdetw * n_z[p];
        }
      }
    }
  }
  jdetw = 1 / t->volume;
  for (i = 0; i < 8; i++) {
    n_x_mean[i] *= jdetw;
    n_y_mean[i] *= jdetw;
    n_z_mean[i] *= jdetw;
  }

  /* Mean strains */
#define CEPS(N, INCR) \
    for (i = 0, eps_mean[N]=0; i < 8; i++) { eps_mean[N]+=INCR; }

  CEPS(0, n_x_mean[i] * dhat[X][i]);
  CEPS(1, n_y_mean[i] * dhat[Y][i]);
  CEPS(2, n_z_mean[i] * dhat[Z][i]);
  CEPS(3, n_y_mean[i] * dhat[X][i] + n_x_mean[i] * dhat[Y][i]);
  CEPS(4, n_z_mean[i] * dhat[X][i] + n_x_mean[i] * dhat[Z][i]);
  CEPS(5, n_z_mean[i] * dhat[Y][i] + n_y_mean[i] * dhat[Z][i]);

  /* Mean stresses */
  /* RAW not using update_form yet */
  WM_get_3d_stress(m, t->matstate, eps_mean, eps_rate_mean, sigma_mean);

  /* Generalized hourglass strain rates */
  LX = LAMBDA_X(X, Xhat);
  LY = LAMBDA_X(Y, Xhat);
  LZ = LAMBDA_X(Z, Xhat);
  H[X][X] = 0.333333 * LY * LZ / LX;
  H[Y][Y] = 0.333333 * LX * LZ / LY;
  H[Z][Z] = 0.333333 * LX * LY / LZ;
  H[X][Y] = 0.333333 * LZ;
  H[X][Z] = 0.333333 * LY;
  H[Y][X] = 0.333333 * LZ;
  H[Y][Z] = 0.333333 * LX;
  H[Z][X] = 0.333333 * LY;
  H[Z][Y] = 0.333333 * LX;
  DebugPV(H, 3, 3); 

  hax = H_X(0, X, Xhat); hay = H_X(0, Y, Xhat); haz = H_X(0, Z, Xhat);
  for (j = 0; j < 8; j++) {
    gammahat[0][j]
      = 0.125*(h0[j] - (hax*n_x_mean[j] + hay*n_y_mean[j] + haz*n_z_mean[j]));
  }
  hax = H_X(1, X, Xhat); hay = H_X(1, Y, Xhat); haz = H_X(1, Z, Xhat);
  for (j = 0; j < 8; j++) {
    gammahat[1][j]
      = 0.125*(h1[j] - (hax*n_x_mean[j] + hay*n_y_mean[j] + haz*n_z_mean[j]));
  }
  hax = H_X(2, X, Xhat); hay = H_X(2, Y, Xhat); haz = H_X(2, Z, Xhat);
  for (j = 0; j < 8; j++) {
    gammahat[2][j]
      = 0.125*(h2[j] - (hax*n_x_mean[j] + hay*n_y_mean[j] + haz*n_z_mean[j]));
  }
  hax = H_X(3, X, Xhat); hay = H_X(3, Y, Xhat); haz = H_X(3, Z, Xhat);
  for (j = 0; j < 8; j++) {
    gammahat[3][j]
      = 0.125*(h3[j] - (hax*n_x_mean[j] + hay*n_y_mean[j] + haz*n_z_mean[j]));
  }

  DebugPV(gammahat, 4, 8); 
  
  for (i = 0; i < 4; i++) {
    hg_strain_rate[X][i] = 0;
    hg_strain_rate[Y][i] = 0;
    hg_strain_rate[Z][i] = 0;
    for (j = 0; j < 8; j++) {
      hg_strain_rate[X][i] += gammahat[i][j] * vhat[X][j];
      hg_strain_rate[Y][i] += gammahat[i][j] * vhat[Y][j];
      hg_strain_rate[Z][i] += gammahat[i][j] * vhat[Z][j];
    }
  }
  DebugPV(hg_strain_rate, 3, 4); 

  /* Generalized hourglass stress rates */
  if (!WM_get_uniax_lin_el_consts(m, &E, &nu))
    REPORT_ERROR("No uniax_lin_el_consts");
  mu = E / 2 / (1 + nu);   /* shear modulus */
  twomu = mu + mu;
  c1 = 1 / (1 - nu);
  c2 = 2 * (1 + nu) / 3;
  c3 = nu / (1 - nu);
  hg_stress_rate[X][X] = mu * (  (H[Y][Y] + H[Z][Z]) * hg_strain_rate[X][X]
                               + H[X][Y]             * hg_strain_rate[Y][Y]
                               + H[X][Z]             * hg_strain_rate[Z][Z]);
  hg_stress_rate[Y][Y] = mu * (  (H[X][X] + H[Z][Z]) * hg_strain_rate[Y][Y]
                               + H[Y][X]             * hg_strain_rate[X][X]
                               + H[Y][Z]             * hg_strain_rate[Z][Z]);
  hg_stress_rate[Z][Z] = mu * (  (H[X][X] + H[Y][Y]) * hg_strain_rate[Z][Z]
                               + H[Z][X]             * hg_strain_rate[X][X]
                               + H[Z][Y]             * hg_strain_rate[Y][Y]);
  hg_stress_rate[X][Y] =     twomu * (  c1 * H[X][X] * hg_strain_rate[X][Y]
                                      + c3 * H[X][Y] * hg_strain_rate[Y][X]);
  hg_stress_rate[X][Z] =     twomu * (  c1 * H[X][X] * hg_strain_rate[X][Z]
                                      + c3 * H[X][Z] * hg_strain_rate[Z][X]);
  hg_stress_rate[Y][X] =     twomu * (  c1 * H[Y][Y] * hg_strain_rate[Y][X]
                                      + c3 * H[Y][X] * hg_strain_rate[X][Y]);
  hg_stress_rate[Y][Z] =     twomu * (  c1 * H[Y][Y] * hg_strain_rate[Y][Z]
                                      + c3 * H[Y][Z] * hg_strain_rate[Z][Y]);
  hg_stress_rate[Z][X] =     twomu * (  c1 * H[Z][Z] * hg_strain_rate[Z][X]
                                      + c3 * H[Z][X] * hg_strain_rate[X][Z]);
  hg_stress_rate[Z][Y] =     twomu * (  c1 * H[Z][Z] * hg_strain_rate[Z][Y]
                                      + c3 * H[Z][Y] * hg_strain_rate[Y][Z]);
  hg_stress_rate[X][3] =    twomu     * c2 * H[X][X] * hg_strain_rate[X][3];
  hg_stress_rate[Y][3] =    twomu     * c2 * H[Y][Y] * hg_strain_rate[Y][3];
  hg_stress_rate[Z][3] =    twomu     * c2 * H[Z][Z] * hg_strain_rate[Z][3];
  DebugPV(hg_stress_rate, 3, 4); 
  
  /* Hourglass stresses */
  dt = eid_cd_integrator_dt(el->domain->integrator);
  for (i = 0; i < 4; i++) {
    t->hg_stress[X][i] += dt * hg_stress_rate[X][i];
    t->hg_stress[Y][i] += dt * hg_stress_rate[Y][i];
    t->hg_stress[Z][i] += dt * hg_stress_rate[Z][i];
  }
  DebugPV(t->hg_stress, 3, 4);
 
  /* Nodal forces */
#define COMPUTE_LOCAL_FORCE(N)                                          \
  {                                                                     \
    lf.x = (  t->volume * (  n_x_mean[N] * sigma_mean[0]                \
                           + n_y_mean[N] * sigma_mean[3]                \
                           + n_z_mean[N] * sigma_mean[4])               \
            + (  t->hg_stress[X][0] * gammahat[0][N]                    \
               + t->hg_stress[X][1] * gammahat[1][N]                    \
               + t->hg_stress[X][2] * gammahat[2][N]                    \
               + t->hg_stress[X][3] * gammahat[3][N]));                 \
    lf.y = (  t->volume * (  n_y_mean[N] * sigma_mean[1]                \
                           + n_x_mean[N] * sigma_mean[3]                \
                           + n_z_mean[N] * sigma_mean[5])               \
            + (  t->hg_stress[Y][0] * gammahat[0][N]                    \
               + t->hg_stress[Y][1] * gammahat[1][N]                    \
               + t->hg_stress[Y][2] * gammahat[2][N]                    \
               + t->hg_stress[Y][3] * gammahat[3][N]));                 \
    lf.z = (  t->volume * (  n_z_mean[N] * sigma_mean[2]                \
                           + n_x_mean[N] * sigma_mean[4]                \
                           + n_y_mean[N] * sigma_mean[5])               \
            + (  t->hg_stress[Z][0] * gammahat[0][N]                    \
               + t->hg_stress[Z][1] * gammahat[1][N]                    \
               + t->hg_stress[Z][2] * gammahat[2][N]                    \
               + t->hg_stress[Z][3] * gammahat[3][N]));                 \
  }

    /* Note the minus sign */
#define TRANSFORM_TO_GLOBAL()                                      \
  {                                                                \
    gf.x = -(lf.x * b1[X] + lf.y * b2[X] + lf.z * b3[X]);          \
    gf.y = -(lf.x * b1[Y] + lf.y * b2[Y] + lf.z * b3[Y]);          \
    gf.z = -(lf.x * b1[Z] + lf.y * b2[Z] + lf.z * b3[Z]);          \
  }

#define ASSEMBLE_GF(N)                          \
  {                                             \
    COMPUTE_LOCAL_FORCE(N);                     \
    TRANSFORM_TO_GLOBAL();                      \
    WN_assemble_force(n[N], &gf, el->domain);   \
  }

  ASSEMBLE_GF(0); ASSEMBLE_GF(1); ASSEMBLE_GF(2); ASSEMBLE_GF(3);
  ASSEMBLE_GF(4); ASSEMBLE_GF(5); ASSEMBLE_GF(6); ASSEMBLE_GF(7);
  
  /* Deformation energy */
  el->deformation_energy
    = 0.5 * t->volume
      *(  eps_mean[0] * sigma_mean[0] + eps_mean[1] * sigma_mean[1]
        + eps_mean[2] * sigma_mean[2] + eps_mean[3] * sigma_mean[3]
        + eps_mean[4] * sigma_mean[4] + eps_mean[5] * sigma_mean[5]);
}
#endif

#endif /* #ifndef ASQBI_FULL_INTEGRATION */


#ifdef HANTI

/* twomu = twice the shear mod. */
/* c1 = 1/(1-nu) */
/* c2 = 2*(1+nu)/3 */
/* c3 = nu/(1-nu) */

static int 
hanti(W_float_t *x, W_float_t *v,    /* IN local coordinates and velocities */
      W_float_t *nx, W_float_t *ny, W_float_t *nz, /* IN strain-displ. submat*/
      W_float_t delt,                /* IN time increment */
      W_float_t twomu, W_float_t c1, W_float_t c2, W_float_t c3,
      /* IN elastic constants above */
      W_float_t *fx, W_float_t *fy, W_float_t *fz, /* OUT hourgl. nodal forc*/
      W_float_t *hg /* OUT hourglass forces on element */
      )
{
  static W_float_t h[32]	/* was [8][4] */
    = { 1.,1.,-1.,-1.,
        -1.,-1.,1.,1.,
        1.,-1.,-1.,1.,
        -1.,1.,1.,-1.,
        1.,-1.,1.,-1.,
        1.,-1.,1.,-1.,
        -1.,1.,-1.,1.,
        1.,-1.,1.,-1.
      };
  W_float_t yeta, zxi, xpsi;
  W_float_t c, g[32]	/* was [8][4] */;
  int i, j;
  W_float_t q[12]	/* was [3][4] */;
  W_float_t hx[12]	/* was [3][4] */;
  W_float_t hxx, hxy, hyy, hyz, hzz, hxz;

  /* Parameter adjustments */
  hg -= 4;
  --nz;
  --ny;
  --nx;
  v -= 9;
  x -= 9;
  --fz;
  --fy;
  --fx;
  
  /* Function Body */
  for (i = 1; i <= 3; ++i) {
    hx[i - 1] = (  x[(i << 3) + 1] + x[(i << 3) + 2]
                 - x[(i << 3) + 3] - x[(i << 3) + 4]
                 - x[(i << 3) + 5] - x[(i << 3) + 6]
                 + x[(i << 3) + 7] + x[(i << 3) + 8]);
    hx[i + 2] = (  x[(i << 3) + 1] - x[(i << 3) + 2]
                 - x[(i << 3) + 3] + x[(i << 3) + 4]
                 - x[(i << 3) + 5] + x[(i << 3) + 6]
                 + x[(i << 3) + 7] - x[(i << 3) + 8]);
    hx[i + 5] = (  x[(i << 3) + 1] - x[(i << 3) + 2]
                 + x[(i << 3) + 3] - x[(i << 3) + 4]
                 + x[(i << 3) + 5] - x[(i << 3) + 6]
                 + x[(i << 3) + 7] - x[(i << 3) + 8]);
    hx[i + 8] = ( -x[(i << 3) + 1] + x[(i << 3) + 2]
                 - x[(i << 3) + 3] + x[(i << 3) + 4]
                 + x[(i << 3) + 5] - x[(i << 3) + 6]
                 + x[(i << 3) + 7] - x[(i << 3) + 8]);
  }
  xpsi = -x[9]  + x[10] + x[11] - x[12] - x[13] + x[14] + x[15] - x[16];
  yeta = -x[17] - x[18] + x[19] + x[20] - x[21] - x[22] + x[23] + x[24];
  zxi = -x[25] - x[26] - x[27] - x[28] + x[29] + x[30] + x[31] + x[32];
  hxx = yeta * zxi / (xpsi * 3);
  hyy = xpsi * zxi / (yeta * 3);
  hzz = xpsi * yeta / (zxi * 3);
  hyz = xpsi / 3;
  hxz = yeta / 3;
  hxy = zxi / 3;
  for (j = 1; j <= 4; ++j) {
    for (i = 1; i <= 8; ++i) {
      g[i + (j << 3) - 9]
        = (  h[i + (j << 3) - 9]
           - hx[j * 3 - 3] * nx[i]
           - hx[j * 3 - 2] * ny[i]
           - hx[j * 3 - 1] * nz[i]) * 0.125;
    }
    for (i = 1; i <= 3; ++i) {
      q[i + j * 3 - 4]
        = (  g[(j << 3) - 8] * v[(i << 3) + 1]
           + g[(j << 3) - 7] * v[(i << 3) + 2]
           + g[(j << 3) - 6] * v[(i << 3) + 3]
           + g[(j << 3) - 5] * v[(i << 3) + 4]
           + g[(j << 3) - 4] * v[(i << 3) + 5]
           + g[(j << 3) - 3] * v[(i << 3) + 6]
           + g[(j << 3) - 2] * v[(i << 3) + 7]
           + g[(j << 3) - 1] * v[(i << 3) + 8]);
    }
  }
  
  c = twomu * delt;
  hg[4]  += c * 0.5 * ((hyy + hzz) * q[0] + hxy * q[4] + hxz * q[8]);
  hg[8]  += c * 0.5 * ((hzz + hxx) * q[4] + hyz * q[8] + hxy * q[0]);
  hg[12] += c * 0.5 * ((hxx + hyy) * q[8] + hxz * q[0] + hyz * q[4]);
  hg[7]  += c * (c1 * hxx * q[3] + c3 * hxy * q[1]);
  hg[11] += c * (c1 * hyy * q[7] + c3 * hyz * q[5]);
  hg[6]  += c * (c1 * hzz * q[2] + c3 * hxz * q[6]);
  hg[10] += c * (c1 * hxx * q[6] + c3 * hxz * q[2]);
  hg[5]  += c * (c1 * hyy * q[1] + c3 * hxy * q[3]);
  hg[9]  += c * (c1 * hzz * q[5] + c3 * hyz * q[7]);
  hg[13] += c * c2 * hxx * q[9];
  hg[14] += c * c2 * hyy * q[10];
  hg[15] += c * c2 * hzz * q[11];

  for (i = 1; i <= 8; ++i) {
    fx[i] = (  hg[4] * g[i - 1]
             + hg[7] * g[i + 7]
             + hg[10] * g[i + 15]
             + hg[13] * g[i + 23]);
    fy[i] = (  hg[5] * g[i - 1]
             + hg[8] * g[i + 7]
             + hg[11] * g[i + 15]
             + hg[14] * g[i + 23]);
    fz[i] = (  hg[6] * g[i - 1]
             + hg[9] * g[i + 7]
             + hg[12] * g[i + 15]
             + hg[15] * g[i + 23]);
  }
  return 0;
} 


static void
mean_value_B(W_float_t y[8], W_float_t z[8], W_float_t B[8])
{
  register i;
  static int i2[8] = {1,2,3,0,7,4,5,6};
  static int i3[8] = {2,3,0,1,6,7,4,5};
  static int i4[8] = {3,0,1,2,5,6,7,4};
  static int i5[8] = {4,5,6,7,0,1,2,3};
  static int i6[8] = {5,6,7,4,3,0,1,2};
  static int i8[8] = {7,4,5,6,1,2,3,0};

  FOR(i, 8) {
    B[i] = (  y[i2[i]] * (z[i6[i]] - z[i3[i]] + z[i5[i]] - z[i4[i]])
            + y[i4[i]] * (z[i3[i]] - z[i8[i]] + z[i2[i]] - z[i5[i]])
            + y[i5[i]] * (z[i8[i]] - z[i6[i]] + z[i4[i]] - z[i2[i]])
            + y[i3[i]] * (z[i2[i]] - z[i4[i]])
            + y[i6[i]] * (z[i5[i]] - z[i2[i]])
            + y[i8[i]] * (z[i4[i]] - z[i5[i]])) / 12;
  }
}
 
#endif

static void
assemble_mass(W_elem_t *el)
{
  W_float_t m;
  priv_elem_t *t = (priv_elem_t *)el->data;
  int i;
  
  m = total_mass(el) / 8;

  for (i = 0; i < 8; i++)  
    WN_assemble_mass (t->node[i], m);
}




static W_float_t 
suggested_time_step(W_elem_t *el)
{
  priv_elem_t *t = (priv_elem_t *)el->data;
  W_coords_3d_t refc[8], d;
  W_float_t E, nu;
  W_float_t rho, speed_of_sound;
  W_material_t *m;
  int i;
  double shortest_side = FLT_MAX;

  for (i = 0; i < 8; i++) {
    GET_FROM_NODE (t->node[i], WN_get_ref_coords, refc[i]); 
  }

  /* RAW */
  m = WD_get_material_by_idx(el->domain, t->material);
  if (!WM_get_uniax_lin_el_consts(m, &E, &nu))
    REPORT_ERROR(el->domain->errh, "No uniax_lin_el_consts");
  rho      = WM_get_volume_mass(m);
  speed_of_sound = sqrt(E / (1 - 2*nu) / rho);
#define SUBTR(res, a, b) VECMAC_3_VEC (res, =, a, -, b)
#define COND_MIN(a, b) \
   { double the_B = b; if (the_B != 0) a = min(a, the_B); }
  SUBTR(d, refc[1], refc[0]);
  COND_MIN(shortest_side, W_vector_3d_length(&d));
  SUBTR(d, refc[2], refc[1]);
  COND_MIN(shortest_side, W_vector_3d_length(&d));
  SUBTR(d, refc[3], refc[2]);
  COND_MIN(shortest_side, W_vector_3d_length(&d));
  SUBTR(d, refc[0], refc[3]);
  COND_MIN(shortest_side, W_vector_3d_length(&d));
  SUBTR(d, refc[5], refc[4]);
  COND_MIN(shortest_side, W_vector_3d_length(&d));
  SUBTR(d, refc[6], refc[5]);
  COND_MIN(shortest_side, W_vector_3d_length(&d));
  SUBTR(d, refc[7], refc[6]);
  COND_MIN(shortest_side, W_vector_3d_length(&d));
  SUBTR(d, refc[4], refc[7]);
  COND_MIN(shortest_side, W_vector_3d_length(&d));
  SUBTR(d, refc[0], refc[4]);
  COND_MIN(shortest_side, W_vector_3d_length(&d));
  SUBTR(d, refc[1], refc[5]);
  COND_MIN(shortest_side, W_vector_3d_length(&d));
  SUBTR(d, refc[2], refc[6]);
  COND_MIN(shortest_side, W_vector_3d_length(&d));
  SUBTR(d, refc[3], refc[7]);
  COND_MIN(shortest_side, W_vector_3d_length(&d));
  
  return shortest_side / speed_of_sound;                 
}




static W_float_t
total_mass(W_elem_t *el)
{
  W_float_t volume_mass, J[3][3];
  W_coords_3d_t refc[8];
  priv_elem_t *t;
  W_material_t *m;
  W_gauss_integration_rule_t *gr;
  W_gauss_integration_point_t *ip;
  int i, j, k;
  double volume = 0;

  t   = (priv_elem_t *)el->data;

  for (i = 0; i < 8; i++)  {
    GET_FROM_NODE (t->node[i], WN_get_ref_coords, refc[i]);
  }
  
  m = WD_get_material_by_idx(el->domain, t->material);
  volume_mass = WM_get_volume_mass(m);

#define QR 2
  gr = WG_get_integration_rule(QR);
  ip = gr->ip;
  for (i = 0; i < QR; i++) {
    for (j = 0; j < QR; j++) {
      for (k = 0; k < QR; k++) {
        G3D_hex_J(J, ip[i].Loc, ip[j].Loc, ip[k].Loc, ALL_REF_COORDS);
        volume += ip[i].Wgt * ip[j].Wgt * ip[k].Wgt * G3D_hex_detJ(J);
      }
    }
  }

  t->volume = volume;
  
  return (volume_mass * t->volume);
}



#ifdef ASQBI_FULL_INTEGRATION
static void
assemble_full_integration_restoring_forces(W_elem_t *el,
                                           W_mat_state_update_form_enum update_form)
{
  W_coords_3d_t c[8], refc[8];
  W_coords_3d_t gf;
  W_velocity_3d_t v[8];
  priv_elem_t *t = (priv_elem_t *)el->data;
  W_material_t *m;
  W_float_t eps[6] = {0, 0, 0, 0, 0, 0};
  W_float_t eps_rate[6] = {0, 0, 0, 0, 0, 0};
  W_float_t sigma[6] = {0, 0, 0, 0, 0, 0};
  W_float_t jdetw;
  W_coords_3d_t lf[8]
    = {{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}};
  W_float_t center[3], center0[3], r[3], r0[3], dhat[3][8];
  W_float_t n_xi[8];
  W_float_t n_eta[8];
  W_float_t n_theta[8];
  W_float_t n_x[8]   = {0,0,0,0,0,0,0,0};
  W_float_t n_y[8]   = {0,0,0,0,0,0,0,0};
  W_float_t n_z[8]   = {0,0,0,0,0,0,0,0};
  W_float_t J[3][3], inv_J[3][3];
  W_float_t b1[3], b2[3], b3[3];
  W_gauss_integration_rule_t *gr;
  W_gauss_integration_point_t *ip;
  int i, j, k;
  W_triad_3d_t lb;

#define X 0
#define Y 1
#define Z 2

#define COMPUTE_EPS()                                                         \
  {                                                                           \
    eps[0] =         (  n_x[0] * dhat[X][0] + n_x[1] * dhat[X][1]             \
                      + n_x[2] * dhat[X][2] + n_x[3] * dhat[X][3]             \
                      + n_x[4] * dhat[X][4] + n_x[5] * dhat[X][5]             \
                      + n_x[6] * dhat[X][6] + n_x[7] * dhat[X][7]             \
                      );                                                      \
    eps[1] =         (  n_y[0] * dhat[Y][0] + n_y[1] * dhat[Y][1]             \
                      + n_y[2] * dhat[Y][2] + n_y[3] * dhat[Y][3]             \
                      + n_y[4] * dhat[Y][4] + n_y[5] * dhat[Y][5]             \
                      + n_y[6] * dhat[Y][6] + n_y[7] * dhat[Y][7]             \
                      );                                                      \
    eps[2] =         (  n_z[0] * dhat[Z][0] + n_z[1] * dhat[Z][1]             \
                      + n_z[2] * dhat[Z][2] + n_z[3] * dhat[Z][3]             \
                      + n_z[4] * dhat[Z][4] + n_z[5] * dhat[Z][5]             \
                      + n_z[6] * dhat[Z][6] + n_z[7] * dhat[Z][7]             \
                      );                                                      \
    eps[3] =         ( (  n_z[0] * dhat[Y][0] + n_z[1] * dhat[Y][1]           \
                        + n_z[2] * dhat[Y][2] + n_z[3] * dhat[Y][3]           \
                        + n_z[4] * dhat[Y][4] + n_z[5] * dhat[Y][5]           \
                        + n_z[6] * dhat[Y][6] + n_z[7] * dhat[Y][7]           \
                        )                                                     \
                      + (  n_y[0] * dhat[Z][0] + n_y[1] * dhat[Z][1]          \
                         + n_y[2] * dhat[Z][2] + n_y[3] * dhat[Z][3]          \
                         + n_y[4] * dhat[Z][4] + n_y[5] * dhat[Z][5]          \
                         + n_y[6] * dhat[Z][6] + n_y[7] * dhat[Z][7]          \
                         )                                                    \
                      );                                                      \
    eps[4] =         ( (  n_z[0] * dhat[X][0] + n_z[1] * dhat[X][1]           \
                        + n_z[2] * dhat[X][2] + n_z[3] * dhat[X][3]           \
                        + n_z[4] * dhat[X][4] + n_z[5] * dhat[X][5]           \
                        + n_z[6] * dhat[X][6] + n_z[7] * dhat[X][7]           \
                        )                                                     \
                      + (  n_x[0] * dhat[Z][0] + n_x[1] * dhat[Z][1]          \
                         + n_x[2] * dhat[Z][2] + n_x[3] * dhat[Z][3]          \
                         + n_x[4] * dhat[Z][4] + n_x[5] * dhat[Z][5]          \
                         + n_x[6] * dhat[Z][6] + n_x[7] * dhat[Z][7]          \
                         )                                                    \
                      );                                                      \
    eps[5] =         ( (  n_y[0] * dhat[X][0] + n_y[1] * dhat[X][1]           \
                        + n_y[2] * dhat[X][2] + n_y[3] * dhat[X][3]           \
                        + n_y[4] * dhat[X][4] + n_y[5] * dhat[X][5]           \
                        + n_y[6] * dhat[X][6] + n_y[7] * dhat[X][7]           \
                        )                                                     \
                      + (  n_x[0] * dhat[Y][0] + n_x[1] * dhat[Y][1]          \
                         + n_x[2] * dhat[Y][2] + n_x[3] * dhat[Y][3]          \
                         + n_x[4] * dhat[Y][4] + n_x[5] * dhat[Y][5]          \
                         + n_x[6] * dhat[Y][6] + n_x[7] * dhat[Y][7]          \
                         )                                                    \
                      );                                                      \
  }

#define INCR_FORCES()                                                     \
  {                                                                       \
    int theI;                                                             \
    for (theI = 0; theI < 8; theI++) {                                    \
      lf[theI].x -= jdetw * (  n_x[theI] * sigma[0]                       \
                             + n_z[theI] * sigma[4]                       \
                             + n_y[theI] * sigma[5]);                     \
      lf[theI].y -= jdetw * (  n_y[theI] * sigma[1]                       \
                             + n_z[theI] * sigma[3]                       \
                             + n_x[theI] * sigma[5]);                     \
      lf[theI].z -= jdetw * (  n_z[theI] * sigma[2]                       \
                             + n_y[theI] * sigma[3]                       \
                             + n_x[theI] * sigma[4]);                     \
    }                                                                     \
  }

  for (i = 0; i < 8; i++)  {
    GET_FROM_NODE (t->node[i], WN_get_coords, c[i]);    
    GET_FROM_NODE (t->node[i], WN_get_ref_coords, refc[i]); 
    GET_FROM_NODE (t->node[i], WN_get_velocity, v[i]);
  }
  
  m = WD_get_material_by_idx(el->domain, t->material);
  if (t->matstate[0][0][0] == NULL) { /* General initialization */
    for (i = 0; i < ASQBI_FULL_INTEGRATION_ORDER; i++)
      for (j = 0; j < ASQBI_FULL_INTEGRATION_ORDER; j++)
        for (k = 0; k < ASQBI_FULL_INTEGRATION_ORDER; k++) 
          t->matstate[i][j][k] = WM_get_mat_state(m);
    G3D_hex_local_basis(ALL_REF_COORDS, &lb.vect1, &lb.vect2, &lb.vect3);
    t->lb1[X] = lb.vect1.x; t->lb1[Y] = lb.vect1.y; t->lb1[Z] = lb.vect1.z;
    t->lb2[X] = lb.vect2.x; t->lb2[Y] = lb.vect2.y; t->lb2[Z] = lb.vect2.z;
    t->lb3[X] = lb.vect3.x; t->lb3[Y] = lb.vect3.y; t->lb3[Z] = lb.vect3.z;
  } /* General initialization */

  compute_local_frame(ALL_REF_COORDS, ALL_CURR_COORDS,
                      t->lb1, t->lb2, t->lb3, 
                      b1, b2, b3, center0, center);

  for (i = 0; i < 8; i++) {
    r0[X]      = refc[i].x - center0[X];
    r0[Y]      = refc[i].y - center0[Y];
    r0[Z]      = refc[i].z - center0[Z];
    /* coords in the local reference system */
    Xhat[X][i] = r0[X] * t->lb1[X] + r0[Y] * t->lb1[Y] + r0[Z] * t->lb1[Z];
    Xhat[Y][i] = r0[X] * t->lb2[X] + r0[Y] * t->lb2[Y] + r0[Z] * t->lb2[Z];
    Xhat[Z][i] = r0[X] * t->lb3[X] + r0[Y] * t->lb3[Y] + r0[Z] * t->lb3[Z];
    r[X]       = c[i].x - center[X];
    r[Y]       = c[i].y - center[Y];
    r[Z]       = c[i].z - center[Z];
    xhat[X][i] = r[X] * b1[X] + r[Y] * b1[Y] + r[Z] * b1[Z];
    xhat[Y][i] = r[X] * b2[X] + r[Y] * b2[Y] + r[Z] * b2[Z];
    xhat[Z][i] = r[X] * b3[X] + r[Y] * b3[Y] + r[Z] * b3[Z];
    /* displacements in local coord system */
    dhat[X][i] = xhat[X][i] - Xhat[X][i];
    dhat[Y][i] = xhat[Y][i] - Xhat[Y][i];
    dhat[Z][i] = xhat[Z][i] - Xhat[Z][i];
  }
/*  DebugPV(dhat, 3, 8); */
/*  DebugPV(vhat, 3, 8); */

  gr = WG_get_integration_rule(ASQBI_FULL_INTEGRATION_ORDER);
  ip = gr->ip;
  for (i = 0; i < ASQBI_FULL_INTEGRATION_ORDER; i++) {
    for (j = 0; j < ASQBI_FULL_INTEGRATION_ORDER; j++) {
      for (k = 0; k < ASQBI_FULL_INTEGRATION_ORDER; k++) {
        G3D_hex_J(J, ip[i].Loc, ip[j].Loc, ip[k].Loc, ALL_REF_HAT_COORDS);
        G3D_hex_invJ(J, inv_J);
        G3D_hex_N_xi(n_xi, ip[i].Loc, ip[j].Loc, ip[k].Loc);
        G3D_hex_N_eta(n_eta, ip[i].Loc, ip[j].Loc, ip[k].Loc);
        G3D_hex_N_theta(n_theta, ip[i].Loc, ip[j].Loc, ip[k].Loc);
        G3D_hex_N_x(inv_J, n_xi, n_eta, n_theta, n_x);
        G3D_hex_N_y(inv_J, n_xi, n_eta, n_theta, n_y);
        G3D_hex_N_z(inv_J, n_xi, n_eta, n_theta, n_z);
        jdetw = ip[i].Wgt * ip[j].Wgt * ip[k].Wgt * G3D_hex_detJ(J);
        COMPUTE_EPS();
        WM_get_3d_stress(m, t->matstate[i][j][k], eps, eps_rate, sigma);
        INCR_FORCES();
      }
    }
  }

  for (i = 0; i < 8; i++) {
    gf.x = lf[i].x * b1[X] + lf[i].y * b2[X] + lf[i].z * b3[X];          
    gf.y = lf[i].x * b1[Y] + lf[i].y * b2[Y] + lf[i].z * b3[Y];          
    gf.z = lf[i].x * b1[Z] + lf[i].y * b2[Z] + lf[i].z * b3[Z];
    WN_assemble_force (t->node[i], &gf);
  }
  
  /* Deformation energy */
  el->deformation_energy = 0;
  for (i = 0; i < 8; i++) {
    el->deformation_energy += lf[i].x * dhat[X][i];
    el->deformation_energy += lf[i].y * dhat[Y][i];
    el->deformation_energy += lf[i].z * dhat[Z][i];
  }
  el->deformation_energy *= 0.5;
}

#endif /* ifdef ASQBI_FULL_INTEGRATION */


static void
rotate_by(                /* IN rotation tensor */
                W_float_t R[3][3],
                                /* INOUT vector to rotate */
                W_float_t v[3]
                );


static void
rotate_by(                /* IN rotation tensor */
                W_float_t R[3][3],
                                /* INOUT vector to rotate */
                W_float_t v[3]
                )
{
  W_float_t result[3];
  
  result[0] = R[0][0] * v[0] + R[1][0] * v[1] + R[2][0] * v[2];
  result[1] = R[0][1] * v[0] + R[1][1] * v[1] + R[2][1] * v[2];
  result[2] = R[0][2] * v[0] + R[1][2] * v[1] + R[2][2] * v[2];
  v[0] = result[0];
  v[1] = result[1];
  v[2] = result[2];
}



static void
compute_local_frame(/* IN current coordinates */
                    W_float_t x1, W_float_t y1, W_float_t z1, 
                    W_float_t x2, W_float_t y2, W_float_t z2,
                    W_float_t x3, W_float_t y3, W_float_t z3,
                    W_float_t x4, W_float_t y4, W_float_t z4,
                    W_float_t x5, W_float_t y5, W_float_t z5, 
                    W_float_t x6, W_float_t y6, W_float_t z6,
                    W_float_t x7, W_float_t y7, W_float_t z7,
                    W_float_t x8, W_float_t y8, W_float_t z8,
                    /* IN reference coordinates */
                    W_float_t X1, W_float_t Y1, W_float_t Z1, 
                    W_float_t X2, W_float_t Y2, W_float_t Z2,
                    W_float_t X3, W_float_t Y3, W_float_t Z3,
                    W_float_t X4, W_float_t Y4, W_float_t Z4,
                    W_float_t X5, W_float_t Y5, W_float_t Z5, 
                    W_float_t X6, W_float_t Y6, W_float_t Z6,
                    W_float_t X7, W_float_t Y7, W_float_t Z7,
                    W_float_t X8, W_float_t Y8, W_float_t Z8,
                    /* IN reference frame */
                    W_float_t refb1[3], W_float_t refb2[3],W_float_t refb3[3], 
                    /* OUT current frame */
                    W_float_t b1[3], W_float_t b2[3], W_float_t b3[3],
                    W_float_t center0[3], W_float_t center[3])
{
#ifdef FROM_POLAR_DECOMPOSITION
  W_float_t F[3][3], R[3][3], U[3][3];
  
  G3D_hex_center(X1, Y1, Z1, 
                 X2, Y2, Z2,
                 X3, Y3, Z3,
                 X4, Y4, Z4,
                 X5, Y5, Z5, 
                 X6, Y6, Z6,
                 X7, Y7, Z7,
                 X8, Y8, Z8,
                 center0);
  G3D_hex_center(x1, y1, z1, 
                 x2, y2, z2,
                 x3, y3, z3,
                 x4, y4, z4,
                 x5, y5, z5, 
                 x6, y6, z6,
                 x7, y7, z7,
                 x8, y8, z8,
                 center);
  G3D_hex_deformation_gradient(0, 0, 0,
                               x1, y1, z1, x2, y2, z2, x3, y3, z3, x4, y4, z4,
                               x5, y5, z5, x6, y6, z6, x7, y7, z7, x8, y8, z8,
                               X1, Y1, Z1, X2, Y2, Z2, X3, Y3, Z3, X4, Y4, Z4,
                               X5, Y5, Z5, X6, Y6, Z6, X7, Y7, Z7, X8, Y8, Z8, 
                               F);
  W_polar_decomposition (F, R, U);
  b1[X] = refb1[X];    b1[Y] = refb1[Y];    b1[Z] = refb1[Z];
  b2[X] = refb2[X];    b2[Y] = refb2[Y];    b2[Z] = refb2[Z];
  b3[X] = refb3[X];    b3[Y] = refb3[Y];    b3[Z] = refb3[Z];
  rotate_by(R, b1);
  rotate_by(R, b2);
  rotate_by(R, b3);
#else  /* #ifndef FROM_POLAR_DECOMPOSITION */
  W_float_t n_xi[8], n_eta[8];
  W_coords_3d_t v1, v2, v3;
  
  G3D_hex_N_xi(n_xi, 0, 0, 0);
  v1.x = (  n_xi[0] * x1 + n_xi[1] * x2 + n_xi[2] * x3 + n_xi[3] * x4
          + n_xi[4] * x5 + n_xi[5] * x6 + n_xi[6] * x7 + n_xi[7] * x8);
  v1.y = (  n_xi[0] * y1 + n_xi[1] * y2 + n_xi[2] * y3 + n_xi[3] * y4
          + n_xi[4] * y5 + n_xi[5] * y6 + n_xi[6] * y7 + n_xi[7] * y8);
  v1.z = (  n_xi[0] * z1 + n_xi[1] * z2 + n_xi[2] * z3 + n_xi[3] * z4
          + n_xi[4] * z5 + n_xi[5] * z6 + n_xi[6] * z7 + n_xi[7] * z8);
  v3_normalize(&v1);
  G3D_hex_N_eta(n_eta, 0, 0, 0);
  v2.x = (  n_eta[0] * x1 + n_eta[1] * x2 + n_eta[2] * x3 + n_eta[3] * x4
          + n_eta[4] * x5 + n_eta[5] * x6 + n_eta[6] * x7 + n_eta[7] * x8);
  v2.y = (  n_eta[0] * y1 + n_eta[1] * y2 + n_eta[2] * y3 + n_eta[3] * y4
          + n_eta[4] * y5 + n_eta[5] * y6 + n_eta[6] * y7 + n_eta[7] * y8);
  v2.z = (  n_eta[0] * z1 + n_eta[1] * z2 + n_eta[2] * z3 + n_eta[3] * z4
          + n_eta[4] * z5 + n_eta[5] * z6 + n_eta[6] * z7 + n_eta[7] * z8);
  v3_cross(&v1, &v2, &v3);
  v3_normalize(&v3);
  v3_cross(&v3, &v1, &v2);
  
  b1[0] = v1.x;
  b1[1] = v1.y;
  b1[2] = v1.z;
  
  b2[0] = v2.x;
  b2[1] = v2.y;
  b2[2] = v2.z;

  b3[0] = v3.x;
  b3[1] = v3.y;
  b3[2] = v3.z;

  G3D_hex_center(X1, Y1, Z1, X2, Y2, Z2, X3, Y3, Z3, X4, Y4, Z4,
                 X5, Y5, Z5, X6, Y6, Z6, X7, Y7, Z7, X8, Y8, Z8,
                 center0);
  G3D_hex_center(x1, y1, z1, x2, y2, z2, x3, y3, z3, x4, y4, z4,
                 x5, y5, z5, x6, y6, z6, x7, y7, z7, x8, y8, z8,
                 center);
#endif
}


static void
write_postproc_file (W_elem_t *el, ckit_iobuf_t IOB, int type, char *mode,
                     void *optdata)
{
  priv_elem_t *t = (priv_elem_t *)el->data;
  W_coords_3d_t c[8], refc[8];
  int i;

  for (i = 0; i < 8; i++)  {
    GET_FROM_NODE (t->node[i], WN_get_coords, c[i]);    
    GET_FROM_NODE (t->node[i], WN_get_ref_coords, refc[i]); 
  }
  
  if (type == HEXVUE_POSTPROC_FILE) {
    /* RAW */
  } else if (type == TECPLOT_I_POSTPROC_FILE) {
    /* RAW */
  } /* type */
}
