#ifndef WBMAT_NNODES
# error WBMAT_NNODES not defined
#endif

/* RCSHEADER "$Id: bmat.c,v 1.4 2005/02/14 02:47:23 cvs Exp $" */

/*
  Computes the Total Lagrangian B matrix.
  The GL strain increments are
  _0 e_11, _0 e_22, _0 e_33, 2 *_0 e_12, 2 * _0 e_13, 2 * _0 e_23
 */
static void
b_l_mat (double n_X[WBMAT_NNODES],
         double n_Y[WBMAT_NNODES],
         double n_Z[WBMAT_NNODES],
         int nnodes,
         W_vect_t u_t[WBMAT_NNODES], /* current displacements
                                        u_x0, u_y0, ...., u_zN
                                      */
         double b_l[6][WBMAT_NNODES*3] /* the matrix is zeroed first, then
                                          its filled with linear and nonlinear
                                          terms
                                       */
         )
{
  int i, j, jx, jy, jz;
  double l[3][3] = {{0,0,0},{0,0,0},{0,0,0}};

  FOR (i, WBMAT_NNODES*3) {
    FOR (j, 6) { b_l[j][i] = 0; }
  }

  /* displacement independent strain-displacement matrix
     (_0^t B_L0 matrix from Bathe 1996, p.555) */
#undef B
#define B(I,J,V) { b_l[I][J] = V; }
  jx = 0;
  jy = 1;
  jz = 2;
  FOR (i, nnodes) {
    B (0, jx, n_X[i]); 
    B (1, jy, n_Y[i]);
    B (2, jz, n_Z[i]);
    B (3, jx, n_Y[i]); B (3, jy, n_X[i]);
    B (4, jy, n_Z[i]); B (4, jz, n_Y[i]);
    B (5, jx, n_Z[i]); B (5, jz, n_X[i]);
    jx += 3;
    jy += 3;
    jz += 3;
  }
  /* displacement-dependent strain-displacement matrix
     (_0^t B_L1 matrix, p. 556) */
  /* first the l matrix which collects the terms with displacements */
  FOR (i, nnodes) {
    l[0][0] += n_X[i] * u_t[i].x;
    l[1][0] += n_X[i] * u_t[i].y;
    l[2][0] += n_X[i] * u_t[i].z;

    l[0][1] += n_Y[i] * u_t[i].x;
    l[1][1] += n_Y[i] * u_t[i].y;
    l[2][1] += n_Y[i] * u_t[i].z;

    l[0][2] += n_Z[i] * u_t[i].x;
    l[1][2] += n_Z[i] * u_t[i].y;
    l[2][2] += n_Z[i] * u_t[i].z;
  }
  /* then the b matrix */
#undef B
#define B(I,J,V) { b_l[I][J] += V; }
  jx = 0;
  jy = 1;
  jz = 2;
  FOR (i, nnodes) {

    B(0,jx, l[0][0]*n_X[i]); B(0,jy, l[1][0]*n_X[i]); B(0,jz, l[2][0]*n_X[i]); 
    B(1,jx, l[0][1]*n_Y[i]); B(1,jy, l[1][1]*n_Y[i]); B(1,jz, l[2][1]*n_Y[i]); 
    B(2,jx, l[0][2]*n_Z[i]); B(2,jy, l[1][2]*n_Z[i]); B(2,jz, l[2][2]*n_Z[i]); 

    B(3,jx, (l[0][0]*n_Y[i] + l[0][1]*n_X[i])); 
    B(4,jx, (l[0][1]*n_Z[i] + l[0][2]*n_Y[i])); 
    B(5,jx, (l[0][2]*n_X[i] + l[0][0]*n_Z[i])); 

    B(3,jy, (l[1][0]*n_Y[i] + l[1][1]*n_X[i])); 
    B(4,jy, (l[1][1]*n_Z[i] + l[1][2]*n_Y[i])); 
    B(5,jy, (l[1][2]*n_X[i] + l[1][0]*n_Z[i])); 

    B(3,jz, (l[2][0]*n_Y[i] + l[2][1]*n_X[i])); 
    B(4,jz, (l[2][1]*n_Z[i] + l[2][2]*n_Y[i])); 
    B(5,jz, (l[2][2]*n_X[i] + l[2][0]*n_Z[i])); 
    
    jx += 3;
    jy += 3;
    jz += 3;
  }
#undef B
}

static void
b_nl_mat (double n_X[WBMAT_NNODES],
            double n_Y[WBMAT_NNODES],
            double n_Z[WBMAT_NNODES],
            int nnodes,
            double b_nl[9][WBMAT_NNODES*3]
            )
{
  int i, j, jx, jy, jz;
#undef B
#define B(I,J,V) { b_nl[I][J] = V; }

  FOR (i, WBMAT_NNODES*3) {
    FOR (j, 9) { b_nl[j][i] = 0; }
  }

  jx = 0;
  jy = 1;
  jz = 2;
  FOR (i, nnodes) {
    B (0,jx, n_X[i]);
    B (1,jx, n_Y[i]);
    B (2,jx, n_Z[i]);
    B (3,jy, n_X[i]);
    B (4,jy, n_Y[i]);
    B (5,jy, n_Z[i]);
    B (6,jz, n_X[i]);
    B (7,jz, n_Y[i]);
    B (8,jz, n_Z[i]);
    jx += 3;
    jy += 3;
    jz += 3;
  }
#undef B
}

static void
g_mat (double n_X[WBMAT_NNODES],
       double n_Y[WBMAT_NNODES],
       double n_Z[WBMAT_NNODES],
       int nnodes,
       W_vect_t u_t[WBMAT_NNODES], /* current displacements
                                        u_x0, u_y0, ...., u_zN
                                      */
       double g[9][WBMAT_NNODES*3]
      )
{
  int i, j, jx, jy, jz;
  double l[3][3] = {{0,0,0},{0,0,0},{0,0,0}};
#undef G
#define G(I,J,V) { g[I][J] = V; }
 
  FOR (i, WBMAT_NNODES*3) {
    FOR (j, 9) { g[j][i] = 0; }
  }
 
  jx = 0;
  jy = 1;
  jz = 2;

/* By Dr. Belytschko's FB formulation and Dr. Krysl's voigt notation */
/* Dr. Krysl's 'voigt notaion' */
/*   (0,0)->0; (1,1)->1; (2,2)->2; 
   (0,1)->3; (0,2)->4; 
   (1,2)->5; (1,0)->6; 
   (2,0)->7; (2,1)->8; */

  FOR (i, nnodes) {

/*  0 ~ 2 rows*/
    G (0,jx, n_X[i]);
    G (1,jy, n_Y[i]);
    G (2,jz, n_Z[i]);

/*  (1, 2), (1, 3) */
    G (3,jy, n_X[i]);
    G (4,jz, n_X[i]);

/*  (2, 3), (2, 1) */
    G (5,jz, n_Y[i]);
    G (6,jx, n_Y[i]);

/*  (3, 1), (3, 2) */
    G (7,jx, n_Z[i]);
    G (8,jy, n_Z[i]); 

    jx += 3;
    jy += 3;
    jz += 3;
  }
  /* displacement-dependent strain-displacement matrix
     (_0^t B_L1 matrix, p. 556) */
  /* first the l matrix which collects the terms with displacements */
  FOR (i, nnodes) {
    l[0][0] += n_X[i] * u_t[i].x;
    l[1][0] += n_X[i] * u_t[i].y;
    l[2][0] += n_X[i] * u_t[i].z;

    l[0][1] += n_Y[i] * u_t[i].x;
    l[1][1] += n_Y[i] * u_t[i].y;
    l[2][1] += n_Y[i] * u_t[i].z;

    l[0][2] += n_Z[i] * u_t[i].x;
    l[1][2] += n_Z[i] * u_t[i].y;
    l[2][2] += n_Z[i] * u_t[i].z;
  }
#undef G
  /* then the b matrix */
#define G(I,J,V) { g[I][J] += V; }
  jx = 0;
  jy = 1;
  jz = 2;
  FOR (i, nnodes) {

    G(0,jx, l[0][0]*n_X[i]); G(0,jy, l[1][0]*n_X[i]); G(0,jz, l[2][0]*n_X[i]); 
    G(1,jx, l[0][1]*n_Y[i]); G(1,jy, l[1][1]*n_Y[i]); G(1,jz, l[2][1]*n_Y[i]); 
    G(2,jx, l[0][2]*n_Z[i]); G(2,jy, l[1][2]*n_Z[i]); G(2,jz, l[2][2]*n_Z[i]); 

    G(3,jx, (l[0][1]*n_X[i])); 
    G(3,jy, (l[1][1]*n_X[i])); 
    G(3,jz, (l[2][1]*n_X[i])); 

    G(4,jx, (l[0][2]*n_X[i])); 
    G(4,jy, (l[1][2]*n_X[i])); 
    G(4,jz, (l[2][2]*n_X[i])); 

    G(5,jx, (l[0][2]*n_Y[i])); 
    G(5,jy, (l[1][2]*n_Y[i])); 
    G(5,jz, (l[2][2]*n_Y[i])); 

    G(6,jx, (l[0][0]*n_Y[i])); 
    G(6,jy, (l[1][0]*n_Y[i])); 
    G(6,jz, (l[2][0]*n_Y[i])); 

    G(7,jx, (l[0][0]*n_Z[i])); 
    G(7,jy, (l[1][0]*n_Z[i])); 
    G(7,jz, (l[2][0]*n_Z[i])); 

    G(8,jx, (l[0][1]*n_Z[i])); 
    G(8,jy, (l[1][1]*n_Z[i])); 
    G(8,jz, (l[2][1]*n_Z[i])); 

/* for 9x9 material moduli matrix
    G(3,jx, (l[0][0]*n_Y[i] + l[0][1]*n_X[i])); 
    G(4,jx, (l[0][1]*n_Z[i] + l[0][2]*n_Y[i])); 
    G(5,jx, (l[0][2]*n_X[i] + l[0][0]*n_Z[i])); 

    G(3,jy, (l[1][0]*n_Y[i] + l[1][1]*n_X[i])); 
    G(4,jy, (l[1][1]*n_Z[i] + l[1][2]*n_Y[i])); 
    G(5,jy, (l[1][2]*n_X[i] + l[1][0]*n_Z[i])); 

    G(3,jz, (l[2][0]*n_Y[i] + l[2][1]*n_X[i])); 
    G(4,jz, (l[2][1]*n_Z[i] + l[2][2]*n_Y[i])); 
    G(5,jz, (l[2][2]*n_X[i] + l[2][0]*n_Z[i])); 
*/
    
    jx += 3;
    jy += 3;
    jz += 3;
  }
#undef G
}

/*
  Computes the Eulerian B matrix; see J. Bonet, R. D. Wood,
  Nonlinear continuum mechanics for finite element analysis,
  Cambridge University Press, 1997.  Note however: the order
  of the strain increments is here changed.
   d_11,  d_22,  d_33, 2 * d_12, 2 * d_13, 2 * d_23
 */
static void
eulerian_b_mat (double n_x[WBMAT_NNODES],
                double n_y[WBMAT_NNODES],
                double n_z[WBMAT_NNODES],
                double b[6][WBMAT_NNODES*3] /* results is *added* to b */
         )
{
  int i, j, jx, jy, jz;

  FOR (i, WBMAT_NNODES*3) {
    FOR (j, 6) { b[j][i] = 0; }
  }

#undef B
#define B(I,J,V) { b[I][J] = V; }
  jx = 0;
  jy = 1;
  jz = 2;
  FOR (i, WBMAT_NNODES) {
    B (0, jx, n_x[i]); 
    B (1, jy, n_y[i]);
    B (2, jz, n_z[i]);
    B (3, jx, n_y[i]); B (3, jy, n_x[i]);
    B (4, jy, n_z[i]); B (4, jz, n_y[i]);
    B (5, jx, n_z[i]); B (5, jz, n_x[i]);
    jx += 3;
    jy += 3;
    jz += 3;
  }
#undef B
}

/*
  Computes the small-strain B matrix.
  The strain increments are
  _0 e_11, _0 e_22, _0 e_33, 2 *_0 e_12, 2 * _0 e_13, 2 * _0 e_23
 */
static void
b_ssl_mat (double n_X[WBMAT_NNODES],
           double n_Y[WBMAT_NNODES],
           double n_Z[WBMAT_NNODES],
           int nnodes,
           double b_l[6][WBMAT_NNODES*3] /* the matrix is zeroed first, then
                                            its filled 
                                       */
         )
{
  int i, j, jx, jy, jz;

  FOR (i, WBMAT_NNODES*3) {
    FOR (j, 6) { b_l[j][i] = 0; }
  }

#undef B
#define B(I,J,V) { b_l[I][J] = V; }
  jx = 0;
  jy = 1;
  jz = 2;
  FOR (i, nnodes) {
    B (0, jx, n_X[i]); 
    B (1, jy, n_Y[i]);
    B (2, jz, n_Z[i]);
    B (3, jx, n_Y[i]); B (3, jy, n_X[i]);
    B (4, jy, n_Z[i]); B (4, jz, n_Y[i]);
    B (5, jx, n_Z[i]); B (5, jz, n_X[i]);
    jx += 3;
    jy += 3;
    jz += 3;
  }
#undef B
}

/*
  Computes the small-strain B matrix.
  The strain increments are
  _0 e_11, _0 e_22, _0 e_33, 2 *_0 e_12, 2 * _0 e_13, 2 * _0 e_23
 */
static void
b_ssl_dil_mat (double n_X[WBMAT_NNODES],
               double n_Y[WBMAT_NNODES],
               double n_Z[WBMAT_NNODES],
               int nnodes,
               double b_l[6][WBMAT_NNODES*3] /* the matrix is zeroed first, then
                                                its filled 
                                             */
         )
{
  int i, j, jx, jy, jz;

  FOR (i, WBMAT_NNODES*3) {
    FOR (j, 6) { b_l[j][i] = 0; }
  }

#undef B
#define B(I,J,V) { b_l[I][J] = V; }
  jx = 0;
  jy = 1;
  jz = 2;
  FOR (i, nnodes) {
    B (0, jx, n_X[i]/3);     B (0, jy, n_Y[i]/3);      B (0, jz, n_Z[i]/3); 
    B (1, jx, n_X[i]/3);     B (1, jy, n_Y[i]/3);      B (1, jz, n_Z[i]/3); 
    B (2, jx, n_X[i]/3);     B (2, jy, n_Y[i]/3);      B (2, jz, n_Z[i]/3); 
    jx += 3;
    jy += 3;
    jz += 3;
  }
#undef B
}
