#include "Wconfig.h"
#include "Wtypes.h"
#include "Wutil.h"
#include "Wcdi.h"
#include "Wdomain.h"
#include "Wnode.h"
#include "Welem.h"
#include "Wxf.h"
#include "Wcontact.h"

/* RCSHEADER "$Id: contact.c,v 1.6 2004/04/27 20:10:10 cvs Exp $" */




W_contact_surf_t *
read_faceted_penalty_surf (W_domain_t *d); 
W_contact_surf_t *
read_spheres_penalty_surf (W_domain_t *d);



/*
  Contact surface reader.
 */
W_bool_t
W_contact_reader (W_domain_t *d)
{
  const char *input_line;
  W_contact_surf_t *cont_surf;

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line)) {
      return TRUE;
    } else {
      TOKENS_PARSE (input_line);
      if (TOKENS_HAVE_KEYWORD ("PENALTY") || TOKENS_HAVE_KEYWORD ("FACETED_PENALTY")) {
        if (!(cont_surf = read_faceted_penalty_surf (d))) {
          return FALSE;
        }
        WD_add_cont_surf (d, cont_surf->id, cont_surf);
      } else if (TOKENS_HAVE_KEYWORD ("SPHERES_PENALTY")) {
        if (!(cont_surf = read_spheres_penalty_surf (d))) {
          return FALSE;
        }
        WD_add_cont_surf (d, cont_surf->id, cont_surf);
      } else {
        REPORT_FATAL_ERROR (d->errh, "Unrecognized contact surface type");
      }
    }
  }

  return FALSE;                 /* something wrong; no end-of-block perhaps? */
}


/*
  Contact detection.
 */
void
WC_search (W_contact_surf_t *cont_surf /* contact surface */) 
{
  switch (cont_surf->type) {
  case W_FACETED_PENALTY_CONTACT_SURF:
    WPCF_search (cont_surf->data.faceted_penalty);
    break;
  case W_SPHERES_PENALTY_CONTACT_SURF:
    WPCS_search (cont_surf->data.spheres_penalty);
    break;
  default:
    break;
  }
}

/*
  Add contact element.
  Add contact element to contact surface.
  It is the caller's responsibility to make sure that
  the element being added is able to treat contact.
 */
void
WC_add_cont_elem (W_contact_surf_t *cont_surf, /* contact surface */
                  W_elem_t *e /* add this element */
                  )
{
  int i, num_nodes;
#define MAXNODES 4
  W_node_t *n[MAXNODES];
  W_loc_t loc;
  double avg_range;
  G3D_box_t box;

  /* add to list */
  if (!cont_surf->celems) cont_surf->celems = make_list ();
  add_to_tail (cont_surf->celems, e);
  
  /* update search epsilon */
  num_nodes = WE_get_face_vertex_nodes (e, 0, MAXNODES, n);
  if (num_nodes > MAXNODES) {
    REPORT_FATAL_ERROR (cont_surf->domain->errh, "Resize n[]");
  }
  FOR (i, num_nodes) {
    if (n[i]->can_move_in_dir[0] ||
        n[i]->can_move_in_dir[1] ||
        n[i]->can_move_in_dir[2]
        )
      cont_surf->all_rigid  = FALSE;
    WN_get_coords (n[i], &(loc.x), &(loc.y), &(loc.z));
    if (i == 0) { G3D_INIT_BOX (box, loc); }
    else        { G3D_UPDT_BOX (box, loc); }
  }
  avg_range = 0.3333 * (G3D_BOX_RANGE (box, x) +
                        G3D_BOX_RANGE (box, y) +
                        G3D_BOX_RANGE (box, z));
  cont_surf->search_eps = min (avg_range, cont_surf->search_eps);
}

