#include "Wconfig.h"
#include "Wxf.h"
#include "Wtypes.h"
#include "Wnode.h"
#include "Wtfuncs.h"
#include "Wdomain.h"
#include "Wutil.h"
#include "Wdispbc.h"

/* RCSHEADER "$Id: dispbc.c,v 1.10 2004/09/06 01:17:58 cvs Exp $" */

static void
make_node_group_list_box (W_disp_bc_t *bc)
{
  W_node_t *n;
  W_vect_t rc;

  if (!bc->applies_to) bc->applies_to = make_list ();
  n = NULL;
  while ((n = WD_get_next_node (bc->domain, n)) != NULL) {
    WN_get_ref_coords (n, &rc.x, &rc.y, &rc.z);
    if (G3D_PT_INSIDE_BOX (bc->data.box, rc)) {
      add_to_tail (bc->applies_to, n);
    } /* if */
  } /* while */
}

#ifdef X_XFACE
#define ELIXIR_FILE_HEADER  "# ELIXIR file (C) 1994 Petr Krysl; version "
static int
elixir_file_version(FILE *fp)
{
#define LINE_LEN  132
  char line[LINE_LEN];
  int version;
  
  fgets(line, LINE_LEN-1, fp);
  if (sscanf(line, ELIXIR_FILE_HEADER "%d", &version) != 1) {
    /* put back what you have read */
    rewind(fp);
    return 0;
  } else
    return version;
}

static lobs_3d_front_end
make_surf_lobs (W_disp_bc_t *bc)
{
  FILE *  fp = 0;
  lobs_3d_front_end lobs = lobs_3d_new_instance (NULL, 10); CHECK_MALLOC(lobs);
  fp = fopen(bc->data.surf.egf_surf_file, "r");
  if (fp == NULL) {
    sprintf (W_buf(), "Could not open \"%s\" file", bc->data.surf.egf_surf_file);
    REPORT_FATAL_ERROR(bc->domain->errh, W_buf());
  } else {
    int gpnum = 0;
    GraphicObj *p;
    bc->data.surf.fl=make_list(); CHECK_MALLOC(bc->data.surf.fl);
    EGSetFileVersion(elixir_file_version(fp));
    do {
      p = EGGraphicRetrieveFrom(fp);
      if (p != NULL) { gpnum++;
        if (EGGraphicType(p) == EG_TRIANGLE3D) {
          int np;
          WCRec pts[3];
          if (EGGetGraphicsGeometry(p, pts, &np)) {
            G3D_box_t box;
            int i;
            W_disp_bc_surf_face_t *f;
            f = w_new(W_disp_bc_surf_face_t); CHECK_MALLOC(f);
            f->npts = np;
            G3D_INIT_BOX(box, pts[0]);
            FOR (i, np) {
              f->pts[i] = pts[i];
              G3D_UPDT_BOX (box, pts[i]); 
            }
            lobs_3d_add_object (lobs, f,
                                box.lower.x, box.lower.y, box.lower.z,
                                box.upper.x, box.upper.y, box.upper.z);
            add_to_tail (bc->data.surf.fl, f);
          }
        } else if (EGGraphicType(p) == EG_QUAD3D) {
          int np;
          WCRec pts[4];
          if (EGGetGraphicsGeometry(p, pts, &np)) {
            G3D_box_t box;
            int i, j;
            W_disp_bc_surf_face_t *f;
            const int nti[2][3]={{0,1,2},{2,3,0}};
            for (j=0; j<2; j++) {
              f = w_new(W_disp_bc_surf_face_t); CHECK_MALLOC(f);
              f->npts = 3;
              G3D_INIT_BOX(box, pts[nti[j][0]]);
              FOR (i, 3) {
                f->pts[i] = pts[nti[j][i]];
                G3D_UPDT_BOX (box, pts[nti[j][i]]); 
              }
              lobs_3d_add_object (lobs, f,
                                  box.lower.x, box.lower.y, box.lower.z,
                                  box.upper.x, box.upper.y, box.upper.z);
              add_to_tail (bc->data.surf.fl, f);
            }
          }
        }
      }
      if (feof(fp))
        break;
    } while (p != NULL);
    fclose(fp);

  }
  return lobs;
}

static W_bool_t 
node_near_surf (lobs_3d_front_end lobs, double dist, W_loc_t *l)
{
  G3D_box_t box;
  W_disp_bc_surf_face_t *f;
  G3D_INIT_BOX(box, *l);
  G3D_INFLATE_BOX(box, dist);
  lobs_3d_search (lobs, 
                  box.lower.x, box.lower.y, box.lower.z,
                  box.upper.x, box.upper.y, box.upper.z);
  f = (W_disp_bc_surf_face_t *) lobs_3d_first_overlapping_object (lobs);
  while (f) {
    G3D_plane_t plane;
    double sdist;
    double uc[4], vc[4];
    G3D_vect_t *pp[4];
    G3D_plane_consts_3(&plane, &f->pts[0], &f->pts[1], &f->pts[2]);
    sdist = G3D_signed_dist_from_plane(&plane, l);
    if (fabs(sdist) < dist) {
      pp[0] = &f->pts[0];      pp[1] = &f->pts[1];      pp[2] = &f->pts[2];
      pp[3] = l;
      G3D_plane_project_points(&plane, pp, 4, uc, vc);
      if (G3D_is_inside_tri2d(uc[3], vc[3], uc[0], vc[0], uc[1], vc[1], uc[2], vc[2])) {
        return TRUE;
      }
    }
    f = (W_disp_bc_surf_face_t *) lobs_3d_next_overlapping_object (lobs);
  }
  return FALSE;
}

static void
make_node_group_list_surf (W_disp_bc_t *bc)
{
  W_node_t *n;
  W_vect_t rc;
  lobs_3d_front_end lobs = make_surf_lobs (bc);

  if (!bc->applies_to) bc->applies_to = make_list ();
  n = NULL;
  while ((n = WD_get_next_node (bc->domain, n)) != NULL) {
    WN_get_ref_coords (n, &rc.x, &rc.y, &rc.z); 
    if (node_near_surf(lobs, bc->data.surf.dist, &rc)) {
      add_to_tail (bc->applies_to, n);
    } /* if */
  } /* while */
  lobs_3d_free_instance (lobs);
}
#else /* no Elixir: no surf type */
static void
make_node_group_list_surf (W_disp_bc_t *bc)
{
  REPORT_FATAL_ERROR (bc->domain->errh,
                      "Surf type of disp bc available only with Elixir linked in: recompile with X_XFACE");
}
#endif


static void
make_node_group_list (W_disp_bc_t *bc)
{
  switch (bc->type) {
  case W_DISPBC_TYPE_BOX:
    make_node_group_list_box (bc);
    break;
  case W_DISPBC_TYPE_SURF:
    make_node_group_list_surf (bc);
    break;
  default:
    REPORT_FATAL_ERROR (bc->domain->errh, "Type not recognized");
    break;
  }
}

/*
  Displacement boundary condition reader.

  An arbitrary combination of prescribed displacements, velocities, or
  accelerations may be set, but for a given component only one of them
  may be set.  In other words, it is possible to set AX, VX, or DX, but
  not any combination of two or three of these specifiers.
  The prescribed displacement, velocity, or acceleration 
  applies to nodes in the specified box.  On/off time may also be given.
  (Condition is ignored before the `on' time and after the `off' time.)
  
  Example of input syntax:
"
"   Dispbc
"   1 VX "vx" VY "vy" VZ "vz" BOX lx ... uz INFLATE s
"   1 AX "ax" AY "ay" AZ "az" BOX lx ... uz INFLATE s
"   1 VX "vx" AY "ay" VZ "vz" BOX lx ... uz INFLATE s
"   1 DX "dx" VY "ay" AZ "vz" BOX lx ... uz INFLATE s
"   &
"
 */
W_bool_t /* returns TRUE if input OK */
W_disp_bc_reader (W_domain_t *d /* domain */)
{
  const char *input_line;
  W_disp_bc_t *bc;
  const char *s;
  double inflate = 0;
  int had_type_keyw = FALSE;

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return TRUE;
    else {
      had_type_keyw = FALSE;
      bc = w_new (W_disp_bc_t); CHECK_MALLOC (bc);
      TOKENS_PARSE(input_line);
      bc->id = TOKEN_INT (1);
      bc->domain     = d;
      bc->applies_to = make_list ();
      bc->prescribed[0] = NULL;
      bc->prescribed[1] = NULL;
      bc->prescribed[2] = NULL;
      bc->kind[0] = bc->kind[1] = bc->kind[2] = W_DISP_BC_NONE;
      {
        char kinds[] = "AVD";
        char dirs[]  = "XYZ";
        char keyw[3];
        int i, j;
        keyw[2] = '\0';
        FOR (i, 3) {
          keyw[1] = dirs[i];
          FOR (j, 3) {
            keyw[0] = kinds[j]; 
            TOKENS_SET_BY_KEYWORD (keyw, FALSE,
            {
              s = TOKEN_STR_NEXT (1);
              if (bc->prescribed[i]) {
                REPORT_WARNING (bc->domain->errh,
                                "Cannot apply more than one kind of boundary condition to a component: ignored");
              } else {
                DUPLICATE_STRING (bc->prescribed[i], s);
                bc->kind[i] = j;
              }
            }, {
            });
          }
        }
      }
      TOKENS_SET_BY_KEYWORD ("BOX", FALSE,
                             {
			       W_loc_t p;
			       p.x=TOKEN_DBL_NEXT (1);
                               p.y=TOKEN_DBL_NEXT (2);
                               p.z=TOKEN_DBL_NEXT (3);
			       G3D_INIT_BOX(bc->data.box,p);
                               p.x=TOKEN_DBL_NEXT (4);
                               p.y=TOKEN_DBL_NEXT (5);
                               p.z=TOKEN_DBL_NEXT (6);
			       G3D_UPDT_BOX(bc->data.box,p);
                               bc->type = W_DISPBC_TYPE_BOX;
                               had_type_keyw = TRUE;
                             }, {
                             });
      TOKENS_SET_BY_KEYWORD ("SURF", FALSE,
                             {
			       DUPLICATE_STRING(bc->data.surf.egf_surf_file, TOKEN_STR_NEXT(1));
                               bc->data.surf.dist = 0;
                               bc->data.surf.fl = 0;
                               bc->type = W_DISPBC_TYPE_SURF;
                               had_type_keyw = TRUE;
                             }, {
                             });
      if (!had_type_keyw) REPORT_FATAL_ERROR (d->errh, "Box or surf is required for displ BC's");
      TOKENS_SET_BY_KEYWORD ("INFLATE", FALSE,
                             {
                               inflate = TOKEN_DBL_NEXT (1);
                             }, {
                               inflate = 0;
                             });
      TOKENS_SET_BY_KEYWORD ("ON", FALSE,
                             {
                               if (tokens_type_of_token(d->parser, _kpos_) == TOKENS_NUMBER) {
                                 bc->applies_in_interval.on = TOKEN_DBL_NEXT (1);
                               } else {
                                 double f;
                                 int stat;
                                 const char *expr = (const char *) TOKEN_STR_NEXT (1);
                                 stat = expeval_eval ((char *) expr, &f);
                                 if (stat != EXPEVAL_E_OK) {
                                   char buf[1024];
                                   sprintf(buf, "%s", expeval_err_msgs[stat]);
                                   REPORT_FATAL_ERROR (bc->domain->errh, buf);
                                 }
                                 bc->applies_in_interval.on = f;
                               }
                             }, {
                               bc->applies_in_interval.on = 0;
                             });
      TOKENS_SET_BY_KEYWORD ("OFF", FALSE,
                             {
                               if (tokens_type_of_token(d->parser, _kpos_) == TOKENS_NUMBER) {
                                 bc->applies_in_interval.off = TOKEN_DBL_NEXT (1);
                               } else {
                                 double f;
                                 int stat;
                                 const char *expr = (const char *) TOKEN_STR_NEXT (1);
                                 stat = expeval_eval ((char *) expr, &f);
                                 if (stat != EXPEVAL_E_OK) {
                                   char buf[1024];
                                   sprintf(buf, "%s", expeval_err_msgs[stat]);
                                   REPORT_FATAL_ERROR (bc->domain->errh, buf);
                                 }
                                 bc->applies_in_interval.off = f;
                               }
                             }, {
                               bc->applies_in_interval.off = 0;
                             });
      if (bc->applies_in_interval.on > bc->applies_in_interval.off) {
        REPORT_FATAL_ERROR (d->errh, "OFF time precedes ON time");
      }
      bc->initialized = FALSE;
      switch (bc->type) {
      case W_DISPBC_TYPE_BOX:
        G3D_INFLATE_BOX (bc->data.box, inflate);
        break;
      case W_DISPBC_TYPE_SURF:
        bc->data.surf.dist = inflate;
        break;
      }
      WD_add_disp_bc (d, bc->id, bc); 
    }
  }
  return FALSE;
}

static void
beta_gamma (W_disp_bc_t *bc, double *beta, double *gamma)
{
  if (bc->domain->newmark_integrator) {
    W_newmark_get_consts (bc->domain->newmark_integrator, beta, gamma);
  } else if (bc->domain->cd_integrator) {
    *beta = 0; *gamma = 0.5;
  } else {
    *beta = 0; *gamma = 0.5;
  }
}

static void
eval_prescribed (W_disp_bc_t *bc, double dt,
                 W_vect_t *a, W_vect_t *v, W_vect_t *d,
                 W_vect_t *newa, W_vect_t *newv, W_vect_t *newd)
{
  int stat;
  double p;
  double beta, gamma;

  beta_gamma (bc, &beta, &gamma);

  VECMAC_SCALAR ((*newa), = , 0);
  VECMAC_SCALAR ((*newv), = , 0);
  VECMAC_SCALAR ((*newd), = , 0);

  if (bc->prescribed[0]) {
    stat = expeval_eval (bc->prescribed[0], &p);
    if (stat != EXPEVAL_E_OK) {
      char buf[1024];
      sprintf(buf, "%s", expeval_err_msgs[stat]);
      REPORT_WARNING(bc->domain->errh, buf);
      p = 0;
    } 
    switch (bc->kind[0]) {
    case W_DISP_BC_A:
      newa->x = p;
      newd->x = d->x + dt*v->x + (dt*dt)/2*((1-2*beta)*a->x+2*beta*newa->x);
      newv->x = v->x + dt*((1-gamma)*a->x+gamma*newa->x);
      break;
    case W_DISP_BC_V:
      newv->x = p;
      newa->x = (newv->x - v->x)/(dt*gamma) - (1-gamma)/gamma*a->x;
      newd->x = d->x + dt*v->x + (dt*dt)/2*((1-2*beta)*a->x+2*beta*newa->x);
      break;
    case W_DISP_BC_D:
      newd->x = p;
      if (beta == 0) REPORT_FATAL_ERROR (NULL, "beta == 0 for prescribed displacement");
      newa->x = (newd->x - d->x - dt*v->x - (dt*dt)/2*(1-2*beta)*a->x) / (dt*dt*beta);
      newv->x = v->x + dt*((1-gamma)*a->x+gamma*newa->x);
      break;
    }      
  }
 
  if (bc->prescribed[1]) {
    stat = expeval_eval (bc->prescribed[1], &p);
    if (stat != EXPEVAL_E_OK) {
      char buf[1024];
      sprintf(buf, "%s", expeval_err_msgs[stat]);
      REPORT_WARNING(bc->domain->errh, buf);
      p = 0;
    } 
    switch (bc->kind[1]) {
    case W_DISP_BC_A:
      newa->y = p;
      newd->y = d->y + dt*v->y + (dt*dt)/2*((1-2*beta)*a->y+2*beta*newa->y);
      newv->y = v->y + dt*((1-gamma)*a->y+gamma*newa->y);
      break;
    case W_DISP_BC_V:
      newv->y = p;
      newa->y = (newv->y - v->y)/(dt*gamma) - (1-gamma)/gamma*a->y;
      newd->y = d->y + dt*v->y + (dt*dt)/2*((1-2*beta)*a->y+2*beta*newa->y);
      break;
    case W_DISP_BC_D:
      newd->y = p;
      if (beta == 0) REPORT_FATAL_ERROR (NULL, "beta == 0 for prescribed displacement");
      newa->y = (newd->y - d->y - dt*v->y - (dt*dt)/2*(1-2*beta)*a->y) / (dt*dt*beta);
      newv->y = v->y + dt*((1-gamma)*a->y+gamma*newa->y);
      break;
    }      
  }
 
  if (bc->prescribed[2]) {
    stat = expeval_eval (bc->prescribed[2], &p);
    if (stat != EXPEVAL_E_OK) {
      char buf[1024];
      sprintf(buf, "%s", expeval_err_msgs[stat]);
      REPORT_WARNING(bc->domain->errh, buf);
      p = 0;
    } 
    switch (bc->kind[2]) {
    case W_DISP_BC_A:
      newa->z = p;
      newd->z = d->z + dt*v->z + (dt*dt)/2*((1-2*beta)*a->z+2*beta*newa->z);
      newv->z = v->z + dt*((1-gamma)*a->z+gamma*newa->z);
      break;
    case W_DISP_BC_V:
      newv->z = p;
      newa->z = (newv->z - v->z)/(dt*gamma) - (1-gamma)/gamma*a->z;
      newd->z = d->z + dt*v->z + (dt*dt)/2*((1-2*beta)*a->z+2*beta*newa->z);
      break;
    case W_DISP_BC_D:
      newd->z = p;
      if (beta == 0) REPORT_FATAL_ERROR (NULL, "beta == 0 for prescribed displacement");
      newa->z = (newd->z - d->z - dt*v->z - (dt*dt)/2*(1-2*beta)*a->z) / (dt*dt*beta);
      newv->z = v->z + dt*((1-gamma)*a->z+gamma*newa->z);
      break;
    }      
  }
 
}

/*
  Initialize boundary condition.
  Initialize essential boundary condition.
 */
void
WBC_init_disp_bc (W_disp_bc_t *bc /* boundary condition */
                  )
{
  W_node_t *n;
  W_vect_t rc, cc, cv;
  double t = WD_time (bc->domain);
  double dt = WD_time_step (bc->domain);

  if (bc->applies_in_interval.on == 2e+7){
     make_node_group_list (bc);
     return;
  }
  if (bc->applies_in_interval.on != bc->applies_in_interval.off) {
    if (bc->applies_in_interval.on > t) {
      return;                   /* this displ bc does not apply yet */
    }
  }
  
  make_node_group_list (bc);

  n = NULL;
  while ((n = (W_node_t *) get_list_next (bc->applies_to, n))) {
    W_bool_t is_fixed[W_NDOF_PER_NODE], is_free[W_NDOF_PER_NODE], apply[W_NDOF_PER_NODE];
    { int i; FOR (i, W_NDOF_PER_NODE) { is_fixed[i] = is_free[i] = apply[i] = FALSE; } }
    
    apply[0] = (bc->prescribed[0] != NULL);
    apply[1] = (bc->prescribed[1] != NULL);
    apply[2] = (bc->prescribed[2] != NULL);
    /* now need to merge the constraint associated with this dispbc with
       the current constraints of the node */
    {
      int i, ndof = WN_is_fixed (n, is_fixed);
      FOR (i, ndof) {
        is_fixed[i] = (is_fixed[i] || apply[i]);
        is_free[i] = ! is_fixed[i];
      }
    }
    WN_fix_node (n, is_free);
  }
  bc->initialized = TRUE;
}

/*
  Enforce essential boundary condition.
  Enforce essential boundary condition.
 */
void
WBC_enforce (W_disp_bc_t *bc /* displacement boundary condition */
             )
{
  W_node_t *n;
  W_vect_t rc = {0,0,0};
  W_vect_t cc = {0,0,0};
  W_vect_t cv = {0,0,0};
  W_acceleration_3d_t a = {0,0,0};
  W_velocity_3d_t v = {0,0,0};
  W_vect_t d = {0,0,0};
  W_acceleration_3d_t newa = {0,0,0};
  W_velocity_3d_t newv = {0,0,0};
  W_vect_t newd = {0,0,0};
  double t = WD_time (bc->domain);
  double dt = WD_time_step (bc->domain);
  double f;
  int stat;
  W_bool_t apply[3];

  if (!bc->initialized) WBC_init_disp_bc (bc);
  if (bc->applies_in_interval.on != bc->applies_in_interval.off) {
    if (bc->applies_in_interval.on > t) {
      return;                   /* this displ bc does not apply yet */
    }
    if (bc->applies_in_interval.off < t) {
      return;                   /* this displ bc does no longer applies */
    }
  }

  /*  if (!bc->initialized) WBC_init_disp_bc (bc); */

  apply[0] = (bc->prescribed[0] != NULL);
  apply[1] = (bc->prescribed[1] != NULL);
  apply[2] = (bc->prescribed[2] != NULL);
  
  n = NULL;
  while ((n = (W_node_t *) get_list_next (bc->applies_to, n))) {
    WN_get_ref_coords (n, &rc.x, &rc.y, &rc.z);
    WN_get_coords (n, &cc.x, &cc.y, &cc.z);
    WN_get_velocity (n, &cv.x, &cv.y, &cv.z);
    /* reference coordinates */
    expeval_set_value("xr", rc.x);
    expeval_set_value("yr", rc.y);
    expeval_set_value("zr", rc.z);
    /* current coordinate and velocity */
    expeval_set_value("x", cc.x);
    expeval_set_value("y", cc.y);
    expeval_set_value("z", cc.z);
    expeval_set_value("xdot", cv.x);
    expeval_set_value("ydot", cv.y);
    expeval_set_value("zdot", cv.z);
    /* time function values */
    {
      W_time_func_t *tf;
      char tfname[32];
      double tfv;
      tf = NULL;
      while (tf = WD_get_next_time_func (bc->domain, tf)) {
        sprintf (tfname, "tf_%d", (int) tf->id);
        tfv = WTF_func_val (tf, t);
        expeval_set_value (tfname, tfv);
      }
    }
    /* time and time step */
    expeval_set_value("t", t);
    expeval_set_value("dt", dt);

    WN_get_displ (n, &d.x, &d.y, &d.z);
    WN_get_velocity (n, &v.x, &v.y, &v.z);
    WN_get_acceleration (n, &a.x, &a.y, &a.z);

    /* compute the prescribed and derived vector components */
    eval_prescribed (bc, dt, &a, &v, &d, &newa, &newv, &newd);
    
    if (bc->domain->cd_integrator) {
      WN_set_force (n, &newa, apply);
    } else if (bc->domain->newmark_integrator) {
      WN_set_velocity (n, &newv, apply);
      WN_set_acceleration (n, &newa, apply);
      WN_set_displ (n, &newd, apply);
    } else {
      REPORT_FATAL_ERROR (bc->domain->errh, "Unknown integrator; cannot enforce BC");
    }
  }
}

W_node_t *
WBC_get_next_node (W_disp_bc_t *bc, W_node_t *n)
{
  if (!bc->initialized) WBC_init_disp_bc(bc);
  return (W_node_t *) get_list_next (bc->applies_to, n);
}


