#include "Wconfig.h"
#include "Wxf.h"
#include "Whistry.h"
#include "Welem.h"
#include "Wtfuncs.h"
#include "Wnode.h"
#include "Wloads.h"
#include "Wdispbc.h"
#include "Wivcond.h"
#include "Wmats.h"
#include "Wcontact.h"
#include "Wvset.h"
#include "Wpvm.h"
#include "Wmpi.h"
#include "Woptions.h"
#include "Wcdi.h"
#include "Wnewmarki.h"
#include "Wdomain.h"
#include "Wmodr.h"
#include "meschach.h"

/* RCSHEADER "$Id: domain.c,v 1.16 2005/02/18 19:56:16 cvs Exp $" */

typedef struct expl_modr_data_t
{
  MAT    *Phit_M_Phi_inv_Phit;
  double *modr_a;
}
expl_modr_data_t;

static void
update_ext_load_work(W_domain_t *d);

#define VL(E_nd, W_hat) VECT_LOOP(1, d->E_nd, W_hat)

#define theELEM an_elem
#define REG_ELEM_LOOP(dom, to_do)                                           \
   if (dom->elems_reg_for_integrate != NULL) {                              \
     W_elem_t *an_elem;                                                     \
                                                                            \
     loop_through_list(dom->elems_reg_for_integrate, an_elem, W_elem_t *) { \
       to_do;                                                               \
     }                                                                      \
   }


#define DICT_LOOP_START(dict, vartype, var)                                     \
  {                                                                             \
    vartype *var = NULL;                                                        \
    dict_iterator d_iter_ = dict_get_iterator ();                               \
    dict_reset_iterator (d_iter_);                                              \
    while ((var = (vartype *) dict_iter_objs (dict->d, d_iter_)) != NULL) {
#define DICT_LOOP_END                           \
    }                                           \
    dict_del_iterator (d_iter_);                \
  }

#define NODE_LOOP_START(domain, var) DICT_LOOP_START(domain->nodes, W_node_t, var)
#define NODE_LOOP_END                DICT_LOOP_END

#define ELEM_LOOP_START(domain, var) DICT_LOOP_START(domain->elements, W_elem_t, var)
#define ELEM_LOOP_END                DICT_LOOP_END

#define LOAD_LOOP_START(domain, var) DICT_LOOP_START(domain->loads, W_load_t, var)
#define LOAD_LOOP_END                DICT_LOOP_END

#define MATERIAL_LOOP_START(domain, var) DICT_LOOP_START(domain->materials, W_material_t, var)
#define MATERIAL_LOOP_END                DICT_LOOP_END

#define DISP_BC_LOOP_START(domain, var) DICT_LOOP_START(domain->disp_bcs, W_disp_bc_t, var)
#define DISP_BC_LOOP_END                DICT_LOOP_END

#define CONTACT_LOOP_START(domain, var) DICT_LOOP_START(domain->cont_surfs, W_contact_surf_t, var)
#define CONTACT_LOOP_END                DICT_LOOP_END

#define IV_COND_LOOP_START(domain, var) DICT_LOOP_START(domain->iv_conds, struct W_initial_velocity_cond_t, var)
#define IV_COND_LOOP_END                DICT_LOOP_END



/*
   Get a node by index.
 */
W_node_t * /* node */
WD_get_node_by_idx (W_domain_t *d, /* domain */
                    W_index_t idx /* index */
                   )
{
  return (W_node_t *) W_dict_find_by_idx (d->nodes, idx);
}


/*
   Add a node to the domain.
 */
W_bool_t /* TRUE if successful; FALSE otherwise */
WD_add_node (W_domain_t *d, /* add a node to the domain */
             W_index_t  idx, /* index */
             W_node_t   *node /* node */
            )
{
  W_node_t *n = WD_get_node_by_idx (d, idx);
  if (n) {
    sprintf (W_buf (), "Redefining node %d", (int) idx);
    REPORT_FATAL_ERROR (d->errh, W_buf ());
    return FALSE;
  } else {
    W_dict_add (d->nodes, node, idx);
    return TRUE;
  }
}

/*
   Iterate nodes.
   Supply curr_node = NULL, if starting iteration.
 */
W_node_t * /* node or NULL if there are no more */
WD_get_next_node (W_domain_t *d, /* domain */
                  W_node_t *curr /* node or NULL if starting */
                 )
{
  return (W_node_t *) W_dict_next_obj (d->nodes, curr);
}

/*
   Iterate elements.
   Supply curr_elem = NULL, if starting iteration.
 */
W_elem_t * /* element or NULL if there are no more */
WD_get_next_element (W_domain_t *d, /* domain */
                     W_elem_t *curr /* element or NULL if starting */
                    )
{
  return (W_elem_t *) W_dict_next_obj (d->elements, curr);
}


/*
   Add a new element to the domain.
 */
W_bool_t /* TRUE if successful; FALSE otherwise */
WD_add_element (W_domain_t     *d, /* domain */
                W_index_t      idx, /* index */
                W_elem_t       *elem /* element */
               )
{
  W_elem_t *e = WD_get_elem_by_idx (d, idx);
  if (e) {
    sprintf (W_buf (), "Redefining element %d", (int) idx);
    REPORT_FATAL_ERROR (d->errh, W_buf ());
    return FALSE;
  } else {
    W_dict_add (d->elements, elem, idx);
    return TRUE;
  }
}

/*
   Get an element by index.
 */
W_elem_t * /* element or NULL if there is no such element index */
WD_get_elem_by_idx (W_domain_t *d, /* domain */
                    W_index_t idx /* index */
                   )
{
  return W_dict_find_by_idx (d->elements, idx);
}


/*
   Get a material by index.
 */
W_material_t * /* material or NULL if there is no such material index */
WD_get_material_by_idx (W_domain_t *d, /* domain */
                        W_index_t idx /* index */
                       )
{
  return (W_material_t *) W_dict_find_by_idx (d->materials, idx);
}


/*
   Add material to the domain.
 */
W_bool_t/* TRUE if successful; FALSE otherwise */
WD_add_material (W_domain_t     *d, /* domain */
                 W_index_t      idx, /* index */
                 W_material_t   *material /* material */
                )
{
  W_material_t *n = WD_get_material_by_idx (d, idx);
  if (n) {
    REPORT_ERROR(d->errh, "Redefining material");
    return FALSE;
  } else {
    W_dict_add (d->materials, material, idx);
    return TRUE;
  }
}


/*
   Add a time function.
 */
W_bool_t/* TRUE if successful; FALSE otherwise */
WD_add_time_func (W_domain_t      *d, /* domain */
                  W_index_t       idx, /* index */
                  W_time_func_t   *time_func /* time function */
                 )
{
  W_time_func_t *tf = WD_get_time_func_by_idx (d, idx);
  if (tf) {
    REPORT_ERROR(d->errh, "Redefining time function");
    return FALSE;
  } else {
    W_dict_add (d->time_funcs, time_func, idx);
    return TRUE;
  }
}


/*
   Get time function by index.
 */
W_time_func_t * /* time function or NULL if there is no such index */
WD_get_time_func_by_idx (W_domain_t *d, /* domain */
                         W_index_t idx /* index */
                        )
{
  return (W_time_func_t *) W_dict_find_by_idx (d->time_funcs, idx);
}

/*
  Iterate displacement boundary conditions.
  Supply curr_node = NULL, if starting iteration.
*/
W_time_func_t * /* node or NULL if there are no more */
WD_get_next_time_func (W_domain_t *d, /* domain */
                       W_time_func_t *curr /* node or NULL if starting */
                      )
{
  return (W_time_func_t *) W_dict_next_obj (d->time_funcs, curr);
}

/*
   Add external load.
 */
W_bool_t/* TRUE if successful; FALSE otherwise */
WD_add_load (W_domain_t *d, /* domain */
             W_load_t   *load /* load */
            )
{
  int idx = dict_count (d->loads->d);
  W_dict_add (d->loads, load, idx++);
  return TRUE;
}


/*
   Add a displacement BC to the domain.
 */
W_bool_t /* TRUE if successful; FALSE otherwise */
WD_add_disp_bc (W_domain_t *d, /* add a node to the domain */
                W_index_t  idx, /* index */
                W_disp_bc_t   *disp_bc /* node */
               )
{
  W_disp_bc_t *bc = dict_find_obj_with_int_name (d->disp_bcs->d, idx);
  if (bc) {
    REPORT_ERROR(d->errh, "Redefining displacement BC");
    return FALSE;
  } else {
    W_dict_add (d->disp_bcs, disp_bc, idx);
    return TRUE;
  }
}

/*
   Iterate displacement boundary conditions.
   Supply curr_node = NULL, if starting iteration.
 */
W_disp_bc_t * /* node or NULL if there are no more */
WD_get_next_disp_bc (W_domain_t *d, /* domain */
                     W_disp_bc_t *curr /* node or NULL if starting */
                    )
{
  return (W_disp_bc_t *) W_dict_next_obj (d->disp_bcs, curr);
}

/*
   Get an displacement BC by index.
 */
W_disp_bc_t * /* element or NULL if there is no such element index */
WD_get_disp_bc_by_idx (W_domain_t *d, /* domain */
                    W_index_t idx /* index */
                   )
{
  return W_dict_find_by_idx (d->disp_bcs, idx);
}


/*
   Add a displacement BC to the domain.
 */
W_bool_t /* TRUE if successful; FALSE otherwise */
WD_add_iv_cond (W_domain_t *d, /* add a node to the domain */
                W_index_t  idx, /* index */
                W_initial_velocity_cond_p iv_cond /* node */
               )
{
  W_initial_velocity_cond_p ivc = dict_find_obj_with_int_name (d->iv_conds->d, idx);
  if (ivc) {
    REPORT_ERROR(d->errh, "Redefining");
    return FALSE;
  } else {
    W_dict_add (d->iv_conds, iv_cond, idx);
    return TRUE;
  }
}

/*
   Iterate displacement boundary conditions.
   Supply curr_node = NULL, if starting iteration.
 */
W_initial_velocity_cond_p /* node or NULL if there are no more */
WD_get_next_iv_cond (W_domain_t *d, /* domain */
                     W_initial_velocity_cond_p curr /* node or NULL if starting */
                    )
{
  return (W_initial_velocity_cond_p) W_dict_next_obj (d->iv_conds, curr);
}



/*
   Add a contact surface to the domain.
 */
W_bool_t /* TRUE if successful; FALSE otherwise */
WD_add_cont_surf (W_domain_t *d, /* domain */
                  W_index_t  idx, /* index */
                  W_contact_surf_t   *conts /* contact surface */
                 )
{
  W_contact_surf_t *cs = WD_get_cont_surf_by_idx (d, idx);
  if (cs) {
    REPORT_ERROR(d->errh, "Redefining contact surface");
    return FALSE;
  } else {
    W_dict_add (d->cont_surfs, conts, idx);
    return TRUE;
  }
}

/*
   Iterate displacement boundary conditions.
   Supply curr = NULL, if starting iteration.
 */
W_contact_surf_t * /* node or NULL if there are no more */
WD_get_next_cont_surf (W_domain_t *d, /* domain */
                       W_contact_surf_t  *curr /*  NULL if starting */
                      )
{
  return (W_contact_surf_t *) W_dict_next_obj (d->cont_surfs, curr);
}

/*
   Get contact surface by index.
 */
W_contact_surf_t * /* time function or NULL if there is no such index */
WD_get_cont_surf_by_idx (W_domain_t *d, /* domain */
                         W_index_t idx /* index */
                        )
{
  return (W_contact_surf_t *) W_dict_find_by_idx (d->cont_surfs, idx);
}


static struct
{
  char          reader_tag;
  reader_proc   reader;
}
readers[] = {
              { 'E',  W_element_reader         },
              { 'N',  W_node_reader            },
              { 'L',  W_load_reader            },
              { 'M',  W_material_reader        },
              { 'T',  W_time_func_reader       },
              { 'D',  W_disp_bc_reader         },
              { 'C',  W_contact_reader         },
              { 'O',  W_options_reader         },
#if defined(PVM_WASP_WORKER) || defined(PVM_WASP_MASTER)
              { 'P',  W_pvm_run_reader
              },
#endif
#if defined (MPI_WASP) && MPI_WASP
              { 'P',  W_mpi_run_reader         },
#endif
              { 'H',  W_history_clients_reader },
              { 'S',  W_vset_table_reader      },
              { 'I',  W_init_cond_reader       },
              { 'R',  W_modr_reader            },
              { '\0', W_read_and_discard_block }
            };

static reader_proc current_reader = NULL;
static int current_reader_tag_num = 0;

static reader_proc
get_reader_proc (const char *input_line);


static W_bool_t
read_problem_data (W_domain_t *d)
{
  char buffer[MAXFILENAMECHAR];

  while ((current_reader = get_reader_proc (W_get_input_line (d))) != NULL) {
    if (!(*current_reader)(d)) {
      sprintf(buffer, "Failed in data reader: reader tag = %c\n",
              readers[current_reader_tag_num].reader_tag);
      REPORT_FATAL_ERROR(d->errh, buffer);
      return FALSE;
    }
  }
  return TRUE;
}

static int
get_tag_num (const char *input_line)
{
  char first_non_blank = '\0';
  int i, n;
  for (i = 0, n = strlen (input_line); i < n; i++) {
    if (!isspace (input_line[i])) {
      first_non_blank = input_line[i];
      break;
    }
  }
  i = 0;
  while (readers[i].reader_tag != '\0') {
    if (readers[i].reader_tag == first_non_blank) {
      current_reader_tag_num = i;
      return i;
    }
    i++;
  }
  /* at this point we're at the last entry in the table of readers;
     that means no reader has been defined for this block, warn the user
     and skip it (by returning index of W_read_and_discard_block())
   */
  sprintf (W_buf (), "Skipping unrecognized block \"%s\"", input_line);
  REPORT_WARNING (NULL, W_buf ());
  return i;
}



static reader_proc
get_reader_proc (const char *input_line)
{
  if (input_line == NULL)
    return NULL;
  else
    return readers[get_tag_num(input_line)].reader;
}



/*
   Create a new domain by reading an input file.
 */
W_domain_t * /* domain */
WD_new_domain (struct W_options_t *o)
{
  W_domain_t *d;
  const char *input_line;

  d = w_new (W_domain_t);
  CHECK_MALLOC (d);

  d->options = o;
  d->errh    = WEH_new_handler (WOPT_OPT (d->options, error_log_name));
  d->logh    = W_log_new_handler (WOPT_OPT (d->options, log_name));

  d->parser  = tokens_new_parser ();

  if (!tokens_open_file (d->parser, WOPT_OPT (d->options, in_file_name)))
  {
    REPORT_FATAL_ERROR (d->errh, "Could not open input file");
  }

  if (WOPT_OPT (d->options, verbose))
  {
    W_log_update (d->logh, "Input file:");
    sprintf (W_buf (), " %s", WOPT_OPT (d->options, in_file_name));
    W_log_update (d->logh, W_buf ());
  }

  if (WOPT_OPT (d->options, expeval_init) != NULL)
  {
    int stat;
    double f;
    stat = expeval_eval (WOPT_OPT (d->options, expeval_init), &f);
    if (stat != EXPEVAL_E_OK) {
      sprintf (W_buf (), "%s: \"%s\"", expeval_err_msgs[stat],
               WOPT_OPT (d->options, expeval_init));
      REPORT_FATAL_ERROR (d->errh, W_buf ());
    }
  }

  input_line = W_get_input_line (d);
  if (input_line != NULL)
    DUPLICATE_STRING (d->comment, input_line);
  if (WOPT_OPT (d->options, verbose))
  {
    W_log_update (d->logh, "Problem description:");
    sprintf (W_buf (), " %s", d->comment);
    W_log_update (d->logh, W_buf ());
  }

  input_line = W_get_input_line (d);
  if (input_line == NULL) /* Just skip this line;
                                 the total used to be used in older versions,
                                 but one line still has to be supplied at this point
                                 (one could use 'ignore' or supply a zero)
                              */
    BAD_INPUT ();


  d->cd_integrator      = NULL;
  d->newmark_integrator = NULL;
  d->newmark_get_incr   = NULL;
  d->newmark_predictor  = W_NEWMARK_PRED_UNKNOWN;

  d->avg_nodal_mass = 0;
  d->total_mass     = 0;

  d->history_writer = NULL;
  if (!WOPT_OPT (d->options, no_output))
    d->history_writer = WHR_new_history_writer (d);

  d->nodes = W_dict_new_dictionary ();
  CHECK_MALLOC (d->nodes);

  d->elements = W_dict_new_dictionary ();
  CHECK_MALLOC (d->elements);

  d->elems_reg_for_integrate = NULL; /* No elements require to */
  /* be called when the domain */
  /* is integrated in time */

  d->materials = W_dict_new_dictionary ();
  CHECK_MALLOC (d->materials);

  d->time_funcs = W_dict_new_dictionary ();
  CHECK_MALLOC (d->time_funcs);

  d->loads = W_dict_new_dictionary ();
  CHECK_MALLOC (d->loads);

  d->disp_bcs = W_dict_new_dictionary ();
  CHECK_MALLOC (d->disp_bcs);

  d->iv_conds = W_dict_new_dictionary ();
  CHECK_MALLOC (d->iv_conds);

  d->cont_surfs = W_dict_new_dictionary ();
  CHECK_MALLOC (d->cont_surfs);

  d->ext_load_work   = 0;

#if defined (MPI_WASP) && MPI_WASP

  d->mpi_process.domain = d;
  d->mpi_process.rank   = -1;
  d->mpi_process.size   = 0;
#endif

  d->modr = NULL;
  d->expl_modr_data = w_new (expl_modr_data_t);
  CHECK_MALLOC (d->expl_modr_data);
  d->expl_modr_data->Phit_M_Phi_inv_Phit = NULL;
  d->expl_modr_data->modr_a = NULL;

  /*
     d->finalize_callbacks must be made in front of 
     read_mat_3d_cryspl_fcc(), which calls 
     add_to_tail(d->finalize_callbacks, solve_for_hardening) 
   */

  d->finalize_callbacks = make_list();

  if (!read_problem_data (d))
  {
    REPORT_FATAL_ERROR (d->errh, "Data not read");
  }

  tokens_close_curr_file (d->parser);

  d->mass_damping_coeff      = WOPT_OPT (d->options, rayleigh_mass_damping_coeff);
  d->curr_mass_damping_coeff = d->mass_damping_coeff;
  d->mass_damping_time_func  = WOPT_OPT (d->options, mass_damping_time_func);

#if defined (MPI_WASP) && MPI_WASP

  if (!WOPT_OPT (d->options, serial))
    W_mpi_build_comm_buffs (&d->mpi_process);
#endif

  d->response_observer_func = NULL;

#if TIMED_WASP

  d->effl_timer = W_timer_new ();
  d->solv_timer = W_timer_new ();
  d->ucfg_timer = W_timer_new ();
  d->fina_timer = W_timer_new ();
#if defined (MPI_WASP) && MPI_WASP

  d->comm_timer = W_timer_new ();
#endif
#endif

  d->check_interrupt = NULL;

  return d;
}

void
W_domain_add_finalize_callback(W_domain_t *d,
                               void  *client,
                               W_finalize_callback_func callback)
{
  W_finalize_callback_t *c = w_new( W_finalize_callback_t);
  c->client = client;
  c->callback = callback;
  add_to_tail(d->finalize_callbacks, c);
}

void
W_domain_set_response_observer_func (W_domain_t *d,
                                     W_response_observer_func response_observer_func)
{
  d->response_observer_func = response_observer_func;
}

/*
   Register element for integration.
   Needs to be done only if the element has internal degrees of freedom.
   If an element registers, it needs to define the functions related
   to the central difference algorithm.
 */
void
WD_register_elem_for_integration (W_domain_t *d, /* domain */
                                  W_elem_t *e /* element */
                                 )
{
  if (d->elems_reg_for_integrate == NULL)
    d->elems_reg_for_integrate = make_list();

  add_to_tail(d->elems_reg_for_integrate, e);
}



static void
make_modr_transf (W_domain_t *d)
{
  int i, j, k;
  MAT *Phit_M_Phi = NULL, *Phi = NULL, *Phit_M_Phi_inv = NULL;
  int M, ndof, dofpn;
  double *Phi_row;
  double *massmat, nmass;
#if W_ROT_DOFS

  W_index_t ens[6];
  W_bool_t constrained[6];
#else

  W_index_t ens[3];
  W_bool_t constrained[3];
#endif

  double mean;

  M = W_modr_get_nmodes (d->modr);
  Phit_M_Phi = m_get (M, M);
  CHECK_MALLOC (Phit_M_Phi);
  ndof = WD_tot_of_equations (d);
  Phi_row = w_new_array (M, double);
  CHECK_MALLOC (Phi_row);
  massmat = w_new_array (ndof, double);
  CHECK_MALLOC (massmat);

  Phi = m_get (ndof, M);
  CHECK_MALLOC (Phi);

  FOR (k, ndof) {
    massmat[k] = 0;
    W_modr_dof_modes (d->modr, k, Phi_row, &mean);
    FOR (i, M) {
      Phi->me[k][i] = Phi_row[i];
    }
  }

  NODE_LOOP_START (d, n) {
    nmass = WN_nodal_mass (n);
    dofpn = WN_eqn_numbers_and_constraints (n, ens, constrained);
    FOR (k, dofpn) {
      massmat[ens[k]] = nmass;
    }
  }
  NODE_LOOP_END;

  FOR (i, M) {
    FOR (j, M) {
      Phit_M_Phi->me[i][j] = 0;
      FOR (k, ndof) {
        Phit_M_Phi->me[i][j] += massmat[k] * Phi->me[k][i] * Phi->me[k][j];
      }
    }
  }

  Phit_M_Phi_inv = m_get (Phit_M_Phi->m, Phit_M_Phi->n);
  Phit_M_Phi_inv = m_inverse (Phit_M_Phi, Phit_M_Phi_inv);

  d->expl_modr_data->Phit_M_Phi_inv_Phit
  = mmtr_mlt (Phit_M_Phi_inv, Phi, NULL);

  d->expl_modr_data->modr_a = w_new_array (M, double);
  CHECK_MALLOC (d->expl_modr_data->modr_a);

  m_free (Phit_M_Phi_inv);
  m_free (Phit_M_Phi);
  w_free (massmat);
  m_free (Phi);
}

/*
  Solve for the primary unknowns (accelerations) in terms of the 
  modes.
 */
void
WD_solve_for_modr_expl (W_domain_t *d /* domain */)
{
  int M, k, i, j, dofpn;
#if W_ROT_DOFS

  W_index_t ens[6];
  W_bool_t constrained[6];
#else

  W_index_t ens[3];
  W_bool_t constrained[3];
#endif

  W_vect_t accel;
  double *a;
  double mean, f[3], *Phi_row;
  W_bool_t apply[] = { TRUE, TRUE, TRUE, TRUE, TRUE, TRUE };
  MAT *Phit_M_Phi_inv_Phit = NULL;
#if TIMED_WASP

  W_timer_restart (d->solv_timer);
#endif


  M = W_modr_get_nmodes (d->modr);

  Phi_row = w_new_array (M, double);
  CHECK_MALLOC (Phi_row);

  if (d->expl_modr_data->Phit_M_Phi_inv_Phit == NULL)
    make_modr_transf (d);
  Phit_M_Phi_inv_Phit = d->expl_modr_data->Phit_M_Phi_inv_Phit;

  /* Mode unknowns */
  a = d->expl_modr_data->modr_a;
  FOR (k, M) {
    a[k] = 0;
  }

  NODE_LOOP_START (d, n) {
    WN_get_force (n, &f[0], &f[1], &f[2]);
    dofpn = WN_eqn_numbers_and_constraints (n, ens, constrained);
    FOR (i, M) {
      FOR (k, dofpn) {
        j = ens[k];
        a[i] += Phit_M_Phi_inv_Phit->me[i][j] * f[k];
      }
    }
  }
  NODE_LOOP_END;

  NODE_LOOP_START (d, n) {
    dofpn = WN_eqn_numbers_and_constraints (n, ens, constrained);

    W_modr_dof_modes (d->modr, ens[0], Phi_row, &mean);
    accel.x = 0;
    FOR (i, M) {
      accel.x += Phi_row[i] * a[i];
    }

    W_modr_dof_modes (d->modr, ens[1], Phi_row, &mean);
    accel.y = 0;
    FOR (i, M) {
      accel.y += Phi_row[i] * a[i];
    }

    W_modr_dof_modes (d->modr, ens[2], Phi_row, &mean);
    accel.z = 0;
    FOR (i, M) {
      accel.z += Phi_row[i] * a[i];
    }

    WN_set_force (n, &accel, apply); /* force stores the new accelerations
                                          computed in the regular solve
                                         */
  }
  NODE_LOOP_END;

  DISP_BC_LOOP_START (d, bc) {
    WBC_enforce (bc);
  }
  DISP_BC_LOOP_END;

  w_free (Phi_row);

#if TIMED_WASP

  W_timer_accumulate (d->solv_timer);
#endif
}

static void
assemble_mass (W_domain_t *d /* domain */)
{
  int nnodes;
  double nmass;

  d->avg_nodal_mass = 0;

  NODE_LOOP_START (d, n) {
    WN_prep_mass_assembly (n);
  }
  NODE_LOOP_END;

  ELEM_LOOP_START (d, e) {
    WE_assemble_mass(e);
  }
  ELEM_LOOP_END;

#if W_ROT_DOFS

  NODE_LOOP_START (d, n) {
    WN_resolve_rot_inertia (n);
  }
  NODE_LOOP_END;
#endif

#if defined (MPI_WASP) && MPI_WASP_NODE_CUT

  if (!WOPT_OPT (d->options, serial))
    W_mpi_exchange_masses (d);
#endif

  nnodes = 0;
  NODE_LOOP_START (d, n) {
    nmass = WN_nodal_mass (n);
    d->avg_nodal_mass += nmass;
    if (nmass <= 0 &&
        (n->can_move_in_dir[0] ||
         n->can_move_in_dir[1] ||
         n->can_move_in_dir[2])) {
      W_bool_t free[] = {FALSE,FALSE,FALSE,FALSE,FALSE,FALSE};
      sprintf (W_buf (), "Zero or negative mass %g associated with node %d", nmass, (int)n->id);
      REPORT_WARNING (d->errh, W_buf ());
      WN_assemble_mass (n, 2*fabs(nmass));
    }
    nnodes++;
  }
  NODE_LOOP_END;
  d->avg_nodal_mass = d->avg_nodal_mass / nnodes;
  d->total_mass     = WD_total_mass (d);
  sprintf (W_buf (), "Total mass = %g", d->total_mass);
  if (WOPT_OPT (d->options, verbose)) {
    W_log_update (d->logh, W_buf());
  }
}



/*
   Integrate the equations of motion for the domain.
 */
void
WD_integrate_expl (W_domain_t *d /* integrate */
                  )
{
  W_float_t start_time = 0;
  double real_time, usr_time, systm_time;
  W_postfile_times_list_t OL;
  W_bool_t should_flush
#ifdef X_XFACE
  = (WOPT_OPT (d->options, X_xface)                      /* flush for X interface */
     || WOPT_OPT (d->options, time_between_outputs) == 0 /* for forced output each step */
    );
#else

  = 0;
#endif

  OL = WOPT_OPT (d->options, postfile_times_list);

#if defined (MPI_WASP) && MPI_WASP
  /* Wait until all processes have been started up and arrive at the barrier */
  if (!options.serial)
    MPI_Barrier (MPI_COMM_WORLD);
#endif

  if (d->cd_integrator == NULL) {
    if (d->modr && W_modr_is_in_use_mode (d->modr)) {
      if (WOPT_OPT (d->options, integrator_detail) == WARBURTON_CD_VARIANT) {
        d->cd_integrator =
          W_cd_new_Warburton_integrator
          (d, 0.0, WD_compute_dt(d),
           (W_cdi_calc_eff_loads_proc)WD_eff_loads_expl,
           (W_cdi_solve_for_proc)WD_solve_for_modr_expl,
           (W_cdi_update_config_proc)WD_updt_cfg_Warb_expl,
           (W_cdi_finalize_step_proc)WD_finalize_expl,
           WOPT_OPT(d->options, cdi_options));
      } else if (WOPT_OPT (d->options, integrator_detail) == NEWMARK_CD_VARIANT ||
                 WOPT_OPT (d->options, integrator_detail) == 0 /* default */
                ) {
        d->cd_integrator =
          W_cd_new_Newmark_integrator
          (d, 0.0, WD_compute_dt(d),
           (W_cdi_calc_eff_loads_proc)WD_eff_loads_expl,
           (W_cdi_solve_for_proc)WD_solve_for_modr_expl,
           (W_cdi_update_config_proc)WD_updt_cfg_Newm_expl,
           (W_cdi_finalize_step_proc)WD_finalize_expl,
           WOPT_OPT(d->options, cdi_options));
      } else {
        REPORT_FATAL_ERROR(d->errh, "Unknown central difference integrator");
      }
    } else {
      if (WOPT_OPT (d->options, integrator_detail) == WARBURTON_CD_VARIANT) {
        d->cd_integrator =
          W_cd_new_Warburton_integrator
          (d, 0.0, WD_compute_dt(d),
           (W_cdi_calc_eff_loads_proc)WD_eff_loads_expl,
           (W_cdi_solve_for_proc)WD_solve_expl,
           (W_cdi_update_config_proc)WD_updt_cfg_Warb_expl,
           (W_cdi_finalize_step_proc)WD_finalize_expl, WOPT_OPT(d->options, cdi_options));
      } else if (WOPT_OPT (d->options, integrator_detail) == NEWMARK_CD_VARIANT ||
                 WOPT_OPT (d->options, integrator_detail) == 0 /* default */
                ) {
        d->cd_integrator =
          W_cd_new_Newmark_integrator
          (d, 0.0, WD_compute_dt(d),
           (W_cdi_calc_eff_loads_proc)WD_eff_loads_expl,
           (W_cdi_solve_for_proc)WD_solve_expl,
           (W_cdi_update_config_proc)WD_updt_cfg_Newm_expl,
           (W_cdi_finalize_step_proc)WD_finalize_expl, WOPT_OPT(d->options, cdi_options));
      } else {
        REPORT_FATAL_ERROR(d->errh, "Unknown central difference integrator");
      }
    }
#   if !defined(PVM_WASP_MASTER)
    assemble_mass (d);
    WD_apply_initial_conditions (d);
#   endif

  }

  WD_flush_history(d);

  start_time = WD_time(d);
  if (WOPT_OPT (d->options, running_patch_test))
    WOPT_SET_OPT (d->options, target_time, 1.9 * W_cd_get_dt(d->cd_integrator));

advance:
  do {
    if (d->check_interrupt)
      d->check_interrupt (d);
    if (WD_time(d) - start_time >= WOPT_OPT (d->options, time_between_outputs)) {
      start_time = WD_time(d);

      if (!WOPT_OPT (d->options, no_output)) {
        if (WOPT_OPT (d->options, verbose)) {
          W_get_times (&real_time, &usr_time, &systm_time);
          sprintf(W_buf(), "Time=%g (r=%g; u=%g; s=%g)",
                  WD_time(d), real_time, usr_time, systm_time);
          W_log_update(d->logh, W_buf());
        }
        WHR_update_history(d->history_writer, should_flush);
      }
#ifdef X_XFACE
      if (WOPT_OPT (d->options, X_xface)) {
        WD_build_grep (d, FALSE);
        sprintf(W_buf(), "Time = %g", WD_time(d));
        WXF_show_graphics (WOPT_OPT (d->options, interactive_during_outputs),
                           W_buf(), WD_time (d));
      }
#endif

    }
    if (WOPT_OPT (d->options, write_postfile_file_now)) {
      W_bool_t nseq = WOPT_OPT (d->options, postfile_number_seq);
      WOPT_SET_OPT (d->options, postfile_number_seq, FALSE); /* force labelling w/ time, not sequential number */
      WD_write_postfile(d, WD_time(d), WD_time(d));
      WOPT_SET_OPT (d->options, write_postfile_file_now, FALSE);
      WOPT_SET_OPT (d->options, postfile_number_seq, nseq);
    } else {
      if (   (!WOPT_OPT (d->options, no_output))
             && (   (OL.ntimes > 0) && (OL.curr < OL.ntimes))) {
        if ((OL.ntimes > 0) && (WD_time(d) >= OL.times[OL.curr])) {
          WD_write_postfile(d, WD_time(d), OL.times[OL.curr]);
          OL.curr++;
        }
      }
    }
    W_cd_advance (d->cd_integrator);
    if (d->response_observer_func)
      (d->response_observer_func) (d, FALSE);
  } while (WD_time(d) < WOPT_OPT (d->options, target_time));

#ifdef X_XFACE

  if (WOPT_OPT (d->options, X_xface)) {
    W_get_times(&real_time, &usr_time, &systm_time);
    sprintf(W_buf(), "Time reached = %g (r=%g; u=%g; s=%g)",
            WD_time(d), real_time, usr_time, systm_time);
    WXF_loop (TRUE, W_buf());
    if (WD_time (d) < WOPT_OPT (d->options, target_time))
      goto advance;
  }
#endif

  if (d->response_observer_func)
    (d->response_observer_func) (d, TRUE);

  WD_clean_up_before_exit(d);

  if (WOPT_OPT (d->options, verbose)) {
    sprintf(W_buf(), "Reached target time = %g", WD_time(d));
    W_log_update(d->logh, W_buf());
    W_get_times(&real_time, &usr_time, &systm_time);
    sprintf(W_buf(), "   Used time [sec]: real %g, user %g, sys %g",
            real_time, usr_time, systm_time);
    W_log_update(d->logh, W_buf());
  }

#ifdef NEED_STAT_MALLOC
  W_stat_malloc_statistics(stderr);
#endif
}



/*
   Get the current time. 
 */
W_float_t /* time */
WD_time (W_domain_t  *d /* domain */
        )
{
  assert(d != NULL);
  if      (d->cd_integrator)
    return W_cd_get_time (d->cd_integrator);
  else if (d->newmark_integrator)
    return W_newmark_get_time (d->newmark_integrator);
  else
    return 0;
}


/*
   Get the current time step. 
 */
W_float_t /* time */
WD_time_step (W_domain_t  *d /* domain */
             )
{
  if      (d->cd_integrator)
    return W_cd_get_dt (d->cd_integrator);
  else if (d->newmark_integrator)
    return W_newmark_get_dt (d->newmark_integrator);
  else
    return 0;
}




#if !(defined(PVM_WASP_WORKER) || defined(PVM_WASP_MASTER)) /* serial code */
/*
   Compute the time step.
 */
W_float_t /* time step */
WD_compute_dt (W_domain_t *d /* domain */)
{
  double dt = FLT_MAX;

  if (WOPT_OPT (d->options, time_step) == 0) {
    ELEM_LOOP_START (d, e) {
      dt = min(dt, WE_suggested_time_step(e));
    }
    ELEM_LOOP_END;
  } else {
    dt = WOPT_OPT (d->options, time_step);
  }

#if defined (MPI_WASP) && MPI_WASP
  {
    double sdt = dt;
    MPI_Allreduce (&sdt, &dt, 1, MPI_DOUBLE, MPI_MIN, MPI_COMM_WORLD);
  }
#endif

  d->stable_time_step = dt;

  if (WOPT_OPT (d->options, time_step) == 0) { /* Time step not directly prescribed */
    /* Reduction/stretching factor should be applied to the stable time step */
    int tfid = WOPT_OPT (d->options, time_step_reduction_func);
    if (tfid != 0) { /* use the time function */
      double R = 0;
      W_time_func_t *tf = WD_get_time_func_by_idx(d, tfid);
      if (tf == NULL) {
        sprintf (W_buf(), "Time function %d for the time step reduction factor not found", tfid);
        REPORT_FATAL_ERROR (d->errh, W_buf());
      }
      R = WTF_func_val (tf, WD_time(d));
      dt *= R;
    } else { /* use the constant factor */
      dt *= WOPT_OPT (d->options, time_step_reduction);
    }
  }

  if (WOPT_OPT (d->options, verbose)) {
    sprintf(W_buf(), "Time step = %g", (double)dt);
    W_log_update(d->logh, W_buf());
  }
  return dt;
}

#endif /* serial code for WD_compute_dt */

/*
  Current mass-proportional damping coefficient.
  Evaluate current mass-proportional damping coefficient.
 */
double /* value */
WD_curr_mass_damping_coeff (W_domain_t *d /* domain */
                           )
{
  d->curr_mass_damping_coeff = d->mass_damping_coeff;
  if (d->mass_damping_coeff != 0) {
    if (d->mass_damping_time_func > 0) {
      W_time_func_t *tf
      = WD_get_time_func_by_idx (d, d->mass_damping_time_func);
      d->curr_mass_damping_coeff *= WTF_func_val (tf, WD_time(d));
    }
  }
  return d->curr_mass_damping_coeff;
}


#if !(defined(PVM_WASP_WORKER) || defined(PVM_WASP_MASTER))
/* serial or MPI code */
/*
   Compute the effective nodal loads.
 */
void
WD_eff_loads_expl (W_domain_t *d /* domain */)
{
  double curr_mass_damping_coeff = WD_curr_mass_damping_coeff (d);
#if TIMED_WASP

  W_timer_restart (d->effl_timer);
#endif

  NODE_LOOP_START (d, n) {
    WN_prep_force_assembly (n);
  }
  NODE_LOOP_END;

  LOAD_LOOP_START (d, l) {
    WL_assemble_load (l, d);
  }
  LOAD_LOOP_END;

  CONTACT_LOOP_START (d, c) {
    WC_search (c);
  }
  CONTACT_LOOP_END;

  REG_ELEM_LOOP(d, WE_calc_eff_loads(theELEM));

  ELEM_LOOP_START (d, e) {
    WE_assemble_restoring_forces (e, W_MAT_STATE_UPDATE_FINAL);
  }
  ELEM_LOOP_END;


#if W_ROT_DOFS

  NODE_LOOP_START (d, n) {
    WN_assemble_ang_vel_dep_load (n);
  }
  NODE_LOOP_END;
#endif

  if (d->mass_damping_coeff != 0) {
    W_float_t dt = W_cd_get_dt(d->cd_integrator);
    NODE_LOOP_START (d, n) {
      WN_assemble_damping (n, d->mass_damping_coeff, dt);
    }
    NODE_LOOP_END;
  }

#if TIMED_WASP
  W_timer_accumulate (d->effl_timer);
#endif

#if defined (MPI_WASP) && MPI_WASP && MPI_WASP_NODE_CUT

  if (!WOPT_OPT (d->options, serial))
    W_mpi_exchange_forces (d);
#endif

}
#endif /* serial code for WD_eff_loads_expl */

#if !(defined(PVM_WASP_MASTER))
/* serial or worker's code, or MPI code */
/*
   Solve for the primary position unknowns.
 */
void
WD_solve_expl (W_domain_t *d /* domain */)
{
#if TIMED_WASP
  W_timer_restart (d->solv_timer);
#endif

  NODE_LOOP_START (d, n) {
    WN_solve_for (n);
  }
  NODE_LOOP_END;

  REG_ELEM_LOOP(d, WE_solve_for (theELEM));

  DISP_BC_LOOP_START (d, bc) {
    WBC_enforce (bc);
  }
  DISP_BC_LOOP_END;

#if TIMED_WASP

  W_timer_accumulate (d->solv_timer);
#endif

#if defined (MPI_WASP) && MPI_WASP && MPI_WASP_ELEM_CUT

  if (!WOPT_OPT (d->options, serial))
    W_mpi_exchange_accels (d);
#endif
}
#endif /* serial code for WD_solve_expl */


#if !defined(PVM_WASP_MASTER) /* serial or worker's code */
/*
   Update the configuration for the Warburton's variant of the
   central differences.
 */
void
WD_updt_cfg_Warb_expl (W_domain_t *d /* domain */)
{
  W_float_t dt = W_cd_get_dt(d->cd_integrator);
#if TIMED_WASP

  W_timer_restart (d->ucfg_timer);
#endif

  NODE_LOOP_START (d, n) {
    WN_update_config_Warburton_variant (n, d, dt);
  }
  NODE_LOOP_END;
  REG_ELEM_LOOP(d, WE_update_config_Warburton_variant(theELEM));
  update_ext_load_work(d);

#if TIMED_WASP

  W_timer_accumulate (d->ucfg_timer);
#endif
}
#endif /* serial or worker's code */

#if !defined(PVM_WASP_MASTER) /* serial or worker's code */
/*
   Update the configuration for the Newmark's variant of the
   central differences.
 */
void
WD_updt_cfg_Newm_expl (W_domain_t *d /* domain */)
{
  W_float_t dt = W_cd_get_dt(d->cd_integrator);
#if TIMED_WASP

  W_timer_restart (d->ucfg_timer);
#endif

  NODE_LOOP_START (d, n) {
    WN_update_config_Newmark_variant (n, d, dt);
  }
  NODE_LOOP_END;
  REG_ELEM_LOOP(d, WE_update_config_Newmark_variant(theELEM));
  update_ext_load_work(d);

#if TIMED_WASP

  W_timer_accumulate (d->ucfg_timer);
#endif
}
#endif /* serial or worker's code */


#ifdef X_XFACE

static int
redraw_view (NODE ptr, NODE elem)
{
  EVFastRedraw ((EView *)elem);
  return 1;
}
/*
   Build graphic representation of domain.
 */
void
WD_build_grep (W_domain_t *d /* domain */,
               W_bool_t immed_draw
              )
{
  ELEM_LOOP_START (d, e) {
    WE_modify_elem_grep (e, WOPT_OPT (d->options, displ_scale), immed_draw);
  }
  ELEM_LOOP_END;
  NODE_LOOP_START (d, n) {
    WN_modify_node_grep (n, WOPT_OPT (d->options, displ_scale),
                         WOPT_OPT (d->options, draw_numbers), immed_draw);
  }
  NODE_LOOP_END;
  LOAD_LOOP_START (d, l) {
    WL_modify_load_grep (l, d, immed_draw);
  }
  LOAD_LOOP_END;

  if (!immed_draw)
    EMDispatchToDependentViews (ESIModel (), redraw_view, NULL);
}

#endif

/*
   Compute the total mass of the domain.
 */
W_float_t /* total mass */
WD_total_mass (W_domain_t *d /* domain */)
{
  assert(d != NULL);

  if (d->total_mass == 0) {
    ELEM_LOOP_START (d, e) {
      d->total_mass += WE_elem_total_mass(e);
    }
    ELEM_LOOP_END;
  }
  return d->total_mass;
}



/*
   Compute the average mass of the nodes.
 */
W_float_t /* total mass */
WD_avg_nodal_mass (W_domain_t *d /* domain */)
{
  return d->avg_nodal_mass;
}






/*
   Compute current kinetic energy.
 */
W_float_t /* energy */
WD_current_kinetic_energy (W_domain_t *d /* domain */)
{
  W_float_t ke = 0;

  if (WOPT_OPT (d->options, mass_type) == W_MASS_DIAG ||
      d->cd_integrator != NULL) { /* explicit or implicit with diagonal mass */
    NODE_LOOP_START (d, n) {
      ke += WN_node_kinetic_energy (n);
    }
    NODE_LOOP_END;
  } else { /* implicit with non-diagonal mass */
    ELEM_LOOP_START (d, e) {
      ke += WE_kinetic_energy (e, WOPT_OPT (d->options, mass_type)); /* RAW: should warn if zero is returned? */
    }
    ELEM_LOOP_END;
  }
  REG_ELEM_LOOP(d, ke += WE_kinetic_energy (theELEM, WOPT_OPT (d->options, mass_type)));
  return ke;
}



/*
   Compute current work of deformation.
 */
W_float_t /* value of work of deformation */
WD_current_deformation_energy (W_domain_t *d /* domain */)
{
  W_float_t en = 0;

  ELEM_LOOP_START (d, e) {
    en += WE_deformation_energy(e);
  }
  ELEM_LOOP_END;
  return en;
}


static void
update_ext_load_work(W_domain_t *d /* domain */)
{
  W_float_t dew = 0;

  LOAD_LOOP_START (d, l) {
    dew += WL_incr_ext_load_work (l, d);
  }
  LOAD_LOOP_END;
  d->ext_load_work += dew;
}


/*
   Compute value of current work of external loads.
 */
W_float_t
WD_current_ext_load_work (W_domain_t *d /* domain */)
{
  return (d->ext_load_work);
}

static void
modr_proj_coeffs (W_domain_t *d, int how_many, double *a,
                  void (*accessor) (W_node_t *n, double *x, double *y, double *z)
                 )
{
  W_modr_p modr = d->modr;
  int j, k, M = W_modr_get_nmodes (modr);
  double mean, *modes;
  W_node_t *n;
#if W_ROT_DOFS

  W_index_t ens[6];
  W_bool_t constrained[6];
#else

  W_index_t ens[3];
  W_bool_t constrained[3];
#endif

  double fr[3];

  modes = w_new_array (M, double);
  CHECK_MALLOC (modes);

  FOR (k, how_many) {
    a[k] = 0;
  }

  NODE_LOOP_START (d, n) {
    (*accessor) (n, &fr[0], &fr[1], &fr[2]);
    WN_eqn_numbers_and_constraints (n, ens, constrained);
    FOR (j, 3) {
      W_modr_dof_modes (modr, ens[j], modes, &mean);
      FOR (k, how_many) {
        a[k] += fr[j] * modes[k];
      }
    }
  }
  NODE_LOOP_END;

  w_free (modes);
}

static void
record_modr_mode_participation (W_domain_t *d)
{
  W_modr_p modr = d->modr;
  int j, M = W_modr_get_nmodes (modr);
  char name[128];
  double *coeffs, sum;

  coeffs = w_new_array (M, double);
  CHECK_MALLOC (coeffs);
  /* project onto modes */
  modr_proj_coeffs (d, M, coeffs, WN_get_displ);
  /* normalize coefficients */
  sum = 0;
  FOR (j, M) {
    sum += coeffs[j]*coeffs[j];
  }
  ;
  sum = sqrt (sum);
  /* save history */
  FOR (j, M) {
    sprintf (name, "nDpart-m%d", j);
    WHR_store (d->history_writer, name, WD_time (d), coeffs[j] / sum);
  }
  /* project onto modes */
  modr_proj_coeffs (d, M, coeffs, WN_get_velocity);
  /* normalize coefficients */
  sum = 0;
  FOR (j, M) {
    sum += coeffs[j]*coeffs[j];
  }
  ;
  sum = sqrt (sum);
  /* save history */
  FOR (j, M) {
    sprintf (name, "nVpart-m%d", j);
    WHR_store (d->history_writer, name, WD_time (d), coeffs[j] / sum);
  }
  w_free (coeffs);
}

/*
   Update the history records.
 */
void
WD_update_history (W_domain_t *d /* domain */
                  )
{
  W_float_t ke, de, ew;
  double time = WD_time (d);

  ke = WD_current_kinetic_energy(d);
  de = WD_current_deformation_energy(d);
  ew = WD_current_ext_load_work(d);

  WHR_store (d->history_writer, "KinetE", time, ke);
  WHR_store (d->history_writer, "DeformE", time, de);
  WHR_store (d->history_writer, "ExtLWork", time, ew);
  WHR_store (d->history_writer, "TotalE", time, (ke + de));

#if 0

  if (d->modr && W_modr_is_in_use_mode (d->modr))
    record_modr_mode_participation (d);
#endif
}

static void
proj_initial_velocity (W_domain_t *d)
{
  double *a;
  W_modr_p modr = d->modr;
  int j, k, M = W_modr_get_nmodes (modr);
  double iv[3], mean, *dof_modes;
#if W_ROT_DOFS

  W_index_t ens[6];
  W_bool_t constrained[6];
#else

  W_index_t ens[3];
  W_bool_t constrained[3];
#endif

  dof_modes = w_new_array (M, double);
  CHECK_MALLOC (dof_modes);
  a         = w_new_array (M, double);
  CHECK_MALLOC (a);

  modr_proj_coeffs (d, M, a, WN_get_velocity);

  NODE_LOOP_START (d, n) {
    WN_eqn_numbers_and_constraints (n, ens, constrained);
    FOR (j, 3) {
      W_modr_dof_modes (modr, ens[j], dof_modes, &mean);
      iv[j] = 0;
      FOR (k, M) {
        iv[j] += a[k] * dof_modes[k];
      }
    }
    n->velocity_t.x = iv[0];
    n->velocity_t.y = iv[1];
    n->velocity_t.z = iv[2];
  }
  NODE_LOOP_END;

  w_free (a);
  w_free (dof_modes);
}

/*
   Apply initial conditions to the entities constituting domain.
 */
void
WD_apply_initial_conditions (W_domain_t *d /* domain */)
{
  W_vect_t rc, v, a, pd;
  W_float_t dt = 0;
  W_bool_t is_fixed[6]={0,0,0,0,0,0};
  W_bool_t apply[6];

  if (!(d->cd_integrator || d->newmark_integrator))
    REPORT_ERROR(d->errh, "no integrator");

  dt = WD_time_step (d);

  NODE_LOOP_START (d, n) {
    WN_get_ref_coords (n, &rc.x, &rc.y, &rc.z);
    WD_initial_velocity (d, &rc, &v);
    WN_is_fixed (n, is_fixed);
    {
      int i;
      FOR (i, 6) apply[i] = ! is_fixed[i];
    }
    if (is_fixed[0])
      v.x = 0;
    if (is_fixed[1])
      v.y = 0;
    if (is_fixed[2])
      v.z = 0;
    WN_get_acceleration (n, &a.x, &a.y, &a.z);
    VECMAC_3_VEC (pd, =, -dt * v, +, dt*dt * a);
    WN_set_velocity (n, &v, apply);
    WN_set_prev_displ (n, &pd, apply);
  }
  NODE_LOOP_END;

  if (d->modr && W_modr_is_in_use_mode (d->modr))
    proj_initial_velocity (d);

  ELEM_LOOP_START (d, e) {
    WE_apply_initial_conditions (e);
  }
  ELEM_LOOP_END;

  DISP_BC_LOOP_START (d, bc) {
    WBC_init_disp_bc (bc);
  }
  DISP_BC_LOOP_END;
}


/*
   Compute initial velocity at a given point.
 */
void
WD_initial_velocity (W_domain_t *d /* domain */,
                     W_loc_t *at, /* location */
                     W_loc_t *v_initial /* returns the velocity */
                    )
{
  W_velocity_3d_t zerov = {0,0,0}, v;
  *v_initial = zerov;

  IV_COND_LOOP_START (d, ivc) {
    if (WIVC_eval (ivc, at, &v)) {
      *v_initial = v;
    }
  }
  IV_COND_LOOP_END;
}

static void
write_hexvue_file (W_domain_t *d /* domain */, ckit_iobuf_t IOB)
{
  char buf[132];
  int n_output_values = 0;

  W_get_date_and_time (buf);
  ckit_io_put_line (IOB, "! HEXVUE file (C) 1995 Petr Krysl.");
  sprintf (W_buf (), "!     Generated by WASP on %s.", buf);
  ckit_io_put_line (IOB, W_buf ());
  sprintf (W_buf (), "%s; Time: %g (file %s, %s)",
           d->comment, WD_time(d), WOPT_OPT (d->options, in_file_name), buf);
  ckit_io_put_line (IOB, W_buf ());
  if (WOPT_OPT (d->options, postfile_output_spec_deprec) & W_POSTFILE_DISPL_MASK)
    n_output_values += 3;
  if (WOPT_OPT (d->options, postfile_output_spec_deprec) & W_POSTFILE_STRESS_MASK)
    n_output_values += 6;
  if (WOPT_OPT (d->options, postfile_output_spec_deprec) & W_POSTFILE_EQPLS_MASK)
    n_output_values += 1;
  if (WOPT_OPT (d->options, postfile_output_spec_deprec) & W_POSTFILE_VON_MISES_MASK)
    n_output_values += 1;
  if (WOPT_OPT (d->options, postfile_output_spec_deprec) & W_POSTFILE_MATERIAL_MASK)
    n_output_values += 1;
  if (WOPT_OPT (d->options, postfile_output_spec_deprec) & W_POSTFILE_DEF_ENERGY_DENSITY_MASK)
    n_output_values += 1;

  sprintf (W_buf (), "%d ! # of values per vertex", n_output_values);
  ckit_io_put_line (IOB, W_buf ());

  if (WOPT_OPT (d->options, postfile_output_spec_deprec) & W_POSTFILE_DISPL_MASK) {
    ckit_io_put_line (IOB, "\"Displacement U\"");
    ckit_io_put_line (IOB, "\"Displacement V\"");
    ckit_io_put_line (IOB, "\"Displacement W\"");
  }

  if (WOPT_OPT (d->options, postfile_output_spec_deprec) & W_POSTFILE_STRESS_MASK) {
    if        (WOPT_OPT (d->options, output_stress_type) == W_OUTPUT_STRESS_CAUCHY) {
      ckit_io_put_line (IOB, "\"Cauchy stress sigma_11\"");
      ckit_io_put_line (IOB, "\"Cauchy stress sigma_12\"");
      ckit_io_put_line (IOB, "\"Cauchy stress sigma_13\"");
      ckit_io_put_line (IOB, "\"Cauchy stress sigma_22\"");
      ckit_io_put_line (IOB, "\"Cauchy stress sigma_23\"");
      ckit_io_put_line (IOB, "\"Cauchy stress sigma_33\"");
    } else if (WOPT_OPT (d->options, output_stress_type) == W_OUTPUT_STRESS_2ND_PK) {
      ckit_io_put_line (IOB, "\"2nd PK stress S_11\"");
      ckit_io_put_line (IOB, "\"2nd PK stress S_12\"");
      ckit_io_put_line (IOB, "\"2nd PK stress S_13\"");
      ckit_io_put_line (IOB, "\"2nd PK stress S_22\"");
      ckit_io_put_line (IOB, "\"2nd PK stress S_23\"");
      ckit_io_put_line (IOB, "\"2nd PK stress S_33\"");
    }
  }

  if (WOPT_OPT (d->options, postfile_output_spec_deprec) & W_POSTFILE_EQPLS_MASK) {
    ckit_io_put_line (IOB, "\"Equivalent plastic strain\"");
  }

  if (WOPT_OPT (d->options, postfile_output_spec_deprec) & W_POSTFILE_VON_MISES_MASK) {
    ckit_io_put_line (IOB, "\"Von Mises stress\"");
  }

  if (WOPT_OPT (d->options, postfile_output_spec_deprec) & W_POSTFILE_MATERIAL_MASK) {
    ckit_io_put_line (IOB, "\"Material ID\"");
  }

  if (WOPT_OPT (d->options, postfile_output_spec_deprec) & W_POSTFILE_DEF_ENERGY_DENSITY_MASK) {
    ckit_io_put_line (IOB, "\"Deformation energy density\"");
  }

  ckit_io_put_line (IOB, "! END OF HEADER");

  ELEM_LOOP_START (d, e) {
    WE_write_postproc_file (e, IOB,
                            HEXVUE_POSTPROC_FILE,
                            WOPT_OPT (d->options, hexvue_format), NULL);
  }
  ELEM_LOOP_END;
  ckit_io_put_line (IOB, "END ! HEXVUE file.");
}



static void
write_hexvue_XTRA_file (W_domain_t *d /* domain */, ckit_iobuf_t IOB)
{
  char buf[132];
  int i, n_output_values = 0;
  W_postfile_output_specs_t *os = &d->options->postfile_output_specs_list;

  W_get_date_and_time (buf);
  ckit_io_put_line (IOB, "! HEXVUE file (C) 1995-2004 Petr Krysl.");
  sprintf (W_buf (), "!     Generated by WASP on %s.", buf);
  ckit_io_put_line (IOB, W_buf ());
  sprintf (W_buf (), "%s; Time: %g (file %s, %s)",
           d->comment, WD_time(d), WOPT_OPT (d->options, in_file_name), buf);
  ckit_io_put_line (IOB, W_buf ());
  n_output_values = os->nspecs;

  sprintf (W_buf (), "%d ! # of values per vertex", n_output_values);
  ckit_io_put_line (IOB, W_buf ());

  for (i = 0; i < n_output_values; i++) {
    char buf[1024];
    sprintf (buf, "\"%s\"", os->specs[i]);
    ckit_io_put_line (IOB, buf);
  }

  ckit_io_put_line (IOB, "! END OF HEADER");

  ELEM_LOOP_START (d, e) {
    WE_write_postproc_file (e, IOB,
                            HEXVUE_XTRA_POSTPROC_FILE,
                            WOPT_OPT (d->options, hexvue_format), WOPT_OPT(d->options,hexvue_xtra_opts));
  }
  ELEM_LOOP_END;
  ckit_io_put_line (IOB, "END ! HEXVUE file.");
}


static void
write_ve3d_file (W_domain_t *d /* domain */, ckit_iobuf_t IOB)
{
  char buf[132];

  W_get_date_and_time(buf);

  ckit_io_put_line (IOB, "! VE3D file (C) 1996 Petr Krysl.");
  sprintf (W_buf (), "!    Generated by WASP (%s).\n", buf);
  ckit_io_put_line (IOB, W_buf ());
  sprintf (W_buf (), "BODY \"%s\"\n\n", WOPT_OPT (d->options, in_file_name));
  ckit_io_put_line (IOB, W_buf ());

#if 0                           /* RAW */

  ELEM_LOOP_START (d, e) {
    if (WE_elem_elcode (e) == ELCODE_EFG_SUPERELEM) {
      sprintf (W_buf (), "REGION \"%s\"",
               WE_elem_name(e);
               ckit_io_put_line (IOB, W_buf ());
               WE_write_postproc_file (e, IOB,
                                       VE3D_POSTPROC_FILE, "ascii", NULL);
               ckit_io_put_line (IOB, "END_REGION");
             }
           }
         }
         ELEM_LOOP_END;
#endif

sprintf (W_buf (), "END_BODY \"%s\"\n\n", WOPT_OPT (d->options, in_file_name));
ckit_io_put_line (IOB, W_buf ());
}


static void
write_tecplot_file (W_domain_t *d /* domain */, ckit_iobuf_t IOB)
{
  /* This had been deprecated: I don't use Tecplot anymore. */
}


/*
   Write a post-processing file.
   
 */
void
WD_write_named_postfile (W_domain_t *d /* domain */,
                         char *out,
                         int type
                        )
{
  FILE *stream;
  ckit_iobuf_t IOB;

  stream = ckit_io_open_stream (out, "w");
  if (stream == NULL)
    REPORT_ERROR (d->errh, (char *)ckit_io_error_explanation ());
  IOB = ckit_io_new_buffer (stream);
  if (IOB == NULL)
    REPORT_ERROR (d->errh, (char *)ckit_io_error_explanation ());

  switch (type) {
  case HEXVUE_POSTPROC_FILE:
    write_hexvue_file(d, IOB);
    break;
  case HEXVUE_XTRA_POSTPROC_FILE:
    write_hexvue_XTRA_file(d, IOB);
    break;
  case VE3D_POSTPROC_FILE:
    write_ve3d_file(d, IOB);
    break;
  case TECPLOT_I_POSTPROC_FILE:
    write_tecplot_file(d, IOB);
    break;
  default:
    break;
  }
  ckit_io_delete_buffer (IOB);
  fclose (stream);
}

/*
  Write post-processing file.
  
 */
void
WD_write_postfile (W_domain_t *d /* domain */,
                   double simulation_time,
                   double label_with_time /* label with time (might be different
                                                         from the problem time) */
                  )
{
  FILE *stream;
  ckit_iobuf_t IOB;
  double t = simulation_time;
  if (WOPT_OPT(d->options, postfile_user_labels))
    t = label_with_time;
  {
    char *out = W_postfile_name (t,
                                 WOPT_OPT (d->options, in_file_basename),
                                 WOPT_OPT (d->options, postfile_type),
                                 WOPT_OPT (d->options, postfile_number_seq));
    int type = WOPT_OPT (d->options, postfile_type);
    WD_write_named_postfile (d, out, type);
  }
}





/*
   Print info for a domain.
 */
void
WD_print_info (W_domain_t *d /* domain */,
               FILE *stream /* print to stream */
              )
{
  fprintf(stream, "Domain\n");
  fprintf(stream, "End Domain\n");
}


/*
   Print info for a given element.
 */
void
WD_print_elem_info(W_domain_t *d /* domain */,
                   W_index_t elid, /* element index */
                   W_info_spec_t spec, /* specification */
                   FILE *stream /* output stream */
                  )
{
  W_elem_t *el;

  el = WD_get_elem_by_idx(d, elid);
  if (el == NULL) {
    fprintf(stream, "No such element: %ld\n", elid);
  } else {
    WE_print_info(el, spec, stream, 0);
  }
}

/*
   Print info for a given node.
 */
void
WD_print_node_info (W_domain_t *d /* domain */,
                    W_index_t nid, /* node index */
                    W_info_spec_t spec, /* specification */
                    FILE *stream) /* output stream */
{
  W_node_t *n;

  n = WD_get_node_by_idx(d, nid);
  if (n == NULL) {
    fprintf(stream, "No such node: %ld\n", nid)
    ;
  } else
  {
    fprintf(stream, "Node %ld\n", nid);
    WN_print_info(n, d, spec, stream, 0);
  }
}


/*
   Clean up before the exit from the program.
 */
W_bool_t /* TRUE if successful; FALSE otherwise */
WD_clean_up_before_exit (W_domain_t *d /* domain */)
{
  WD_flush_history(d);
  return TRUE;
}


/*
   Flush history buffers.
 */
void
WD_flush_history (W_domain_t *d /* domain */)
{
  /* Flush history records to disk to clean up */
  if (!WOPT_OPT (d->options, no_output)) {
    WHR_update_history (d->history_writer, TRUE);
  }
}



/*
   Finalize the current time step.
 This function can: (i) set a new time step to use, (ii) give a chance 
 to registered elements to do some action (e.g. propagate a crack). 
 */
W_float_t /* new time step */
WD_finalize_expl (W_domain_t *d /* domain */)
{
  W_float_t dt = W_cd_get_dt(d->cd_integrator);
  W_float_t newdt = dt;
  W_float_t sdt = 0;
  W_float_t new_time_step = 0;
#if TIMED_WASP

  W_timer_restart (d->fina_timer);
#endif

  if (!list_empty(d->finalize_callbacks)) {
    W_finalize_callback_t *cd;
    loop_through_list(d->finalize_callbacks, cd, W_finalize_callback_t *) {
      cd->callback(cd->client);
    }
  }

  REG_ELEM_LOOP(d, sdt = WE_finalize_step(theELEM));
  if (sdt > 0)
    newdt = min (newdt, sdt);

  { /* Reduction/stretching factor should be applied to the stable time step */
    double R = 0;
    int tfid = WOPT_OPT (d->options, time_step_reduction_func);
    if (tfid != 0) { /* use the time function */
      W_time_func_t *tf = WD_get_time_func_by_idx(d, tfid);
      if (tf == NULL) {
        sprintf (W_buf(), "Time function %d for the time step reduction factor not found", tfid);
        REPORT_FATAL_ERROR (d->errh, W_buf());
      }
      R = WTF_func_val (tf, WD_time(d));
    } else { /* use the constant factor */
      R = WOPT_OPT (d->options, time_step_reduction);
    }
    new_time_step = d->stable_time_step * R;
  }

  if (new_time_step > 0)
    newdt = min (newdt, new_time_step);

#if TIMED_WASP

  W_timer_accumulate (d->fina_timer);
#endif

  if (d->modr)
    W_modr_accumulate (d->modr);

  return newdt; /* We are ignoring the suggestions from the elements */
  /* for the time being.  Needs to be implemented. */
}


/*
   Probe some field.
 */
W_bool_t
WD_probe (W_domain_t *d, /* domain */
          void *options, /* options */
          W_loc_t *loc, /* at location */
          unsigned long spec, /* specification */
          W_probe_package_t *result /* returns result */
         )
{
  ELEM_LOOP_START (d, e) {
    if (WE_probe (e, options, loc, spec, result)) {
      return TRUE;
    }
  }
  ELEM_LOOP_END;
  return FALSE;
}

/*
   Get mass damping coefficient.
  */
double /* coefficient */
WD_mass_damping_coeff (W_domain_t *d /* domain */
                      )
{
  return d->mass_damping_coeff;
}



/*
   Get ratio of current and the input mass damping coefficient.
 */
double /* ratio */
WD_mass_damping_coeff_time_dep (W_domain_t *d /* domain */)
{
  return (d->mass_damping_coeff != 0
          ? d->curr_mass_damping_coeff / d->mass_damping_coeff
          : 0);
}

#define MAXFEFREFNODES 9
typedef struct fef_ref_t {
  W_elem_t *e;
  int       nface;
  int       nn;
  W_node_t *n[MAXFEFREFNODES];
  int       ref;
}
fef_ref_t;


static LIST_OF (fef_ref_t *)
find_frl (W_node_t *n, dict_dictionary nfrd)
{
  return dict_find_obj_with_handle_name (nfrd, n);
}


static W_node_t *
min_id_fvn (W_node_t *fvn[MAXFEFREFNODES], int nfvn)
{
  int mini = 0, minid = INT_MAX, i;

  FOR (i, nfvn) {
    if (fvn[i]->id < minid) {
      minid = fvn[i]->id;
      mini  = i;
    }
  }
  return fvn[mini];
}

static fef_ref_t *
find_fr (LIST frl, int nfvn, W_node_t *fvn[MAXFEFREFNODES])
{
  int had = 0;
  int i, j;
  fef_ref_t *fr;

  loop_through_list (frl, fr, fef_ref_t *) {
    if (fr->nn == nfvn) {
      had = 0;
      FOR (i, nfvn) {
        FOR (j, nfvn) {
          if (fvn[i] == fr->n[j]) {
            had++;
            break;
          }
        }
      }
      if (had == nfvn)
        return fr;
    }
  }
  return NULL;
}

/*
   Get list of contact faces.
 */
static LIST_OF (fef_ref_t *)
get_list_of_bdry_faces (W_domain_t *d)
{
  dict_dictionary nfrld;
  int i, j, nf, nfvn;
  LIST_OF (fef_ref_t *) frefl;
  W_node_t *fvn[MAXFEFREFNODES], *n;
  LIST_OF (fef_ref_t *) frl, tmpl;
  fef_ref_t *fr;

  /* accumulate FE faces */

  nfrld = dict_create_dictionary ();

  tmpl = make_list ();

  ELEM_LOOP_START (d, e) {
    nf = WE_num_of_faces (e);
    FOR (i, nf) {
      nfvn = WE_get_face_vertex_nodes (e, i, MAXFEFREFNODES, fvn);
      if (nfvn > MAXFEFREFNODES) {
        REPORT_FATAL_ERROR (d->errh,
                            "Can handle only 3- or 4-noded boundary faces");
      }
      n    = min_id_fvn (fvn, nfvn);
      frl  = find_frl (n, nfrld);
      if (!frl) {
        frl = make_list ();
        dict_add_obj_handle_name (nfrld, frl, n);
      }
      fr = find_fr (frl, nfvn, fvn);
      if (!fr) {
        fr = w_new (fef_ref_t);
        CHECK_MALLOC (fr);
        fr->e     = e;
        fr->nface = i;
        fr->nn    = nfvn;
        FOR (j, fr->nn) {
          fr->n[j] = fvn[j];
        }
        fr->ref   = 0;
        add_to_tail (frl, fr);
        add_to_tail (tmpl, fr);
      }
      fr->ref++;
    }
  }
  ELEM_LOOP_END;

  /* create face list by inspecting the accumulated normals */

  frefl = make_list ();
  fr     = NULL;
  while ((fr = (fef_ref_t *) get_list_next (tmpl, fr))) {
    if (fr->ref == 1) {
      add_to_tail (frefl, fr);
    }
  }

  /* cleanup */

  free_list (tmpl, NOT_NODES);
  {
    dict_iterator diter;
    diter = dict_get_iterator ();
    dict_reset_iterator (diter);
    while ((frl = (LIST) dict_iter_objs (nfrld, diter))) {
      free_list (frl, NOT_NODES);
    }
    dict_del_iterator (diter);
  }
  dict_destroy_dictionary (nfrld);

  return frefl;
}

/*
   Get list of contact faces.
 */
LIST_OF (WCF_contact_face_t *) /* list of contact faces */
WD_get_list_of_contact_faces (W_domain_t *d /* domain */
                             )
{
  LIST_OF (WCF_contact_face_t *) cfaces;
  LIST_OF (fef_ref_t *) frl;
  int i, nf;
  fef_ref_t *fr;
#define MAX_CF_PER_FEF 16

  WCF_contact_face_t *cf[MAX_CF_PER_FEF];

  /* find boundary faces */

  frl = get_list_of_bdry_faces (d);

  /* create contact face list  */

  cfaces = make_list ();
  fr     = NULL;
  while ((fr = (fef_ref_t *) get_list_next (frl, fr))) {
    if (fr->ref == 1) {
      nf = WE_get_contact_faces (fr->e, fr->nface, cf, MAX_CF_PER_FEF);
      if (nf > MAX_CF_PER_FEF)
        REPORT_FATAL_ERROR (d->errh, "too many contact faces");
      FOR (i, nf) {
        add_to_tail (cfaces, cf[i]);
      }
    }
  }

  /* cleanup */

  free_list (frl, AND_NODES);

  return cfaces;
}

static void
assemble_mass_from_nodal_masses (W_domain_t *d,
                                 W_newmark_assemble_eff_stiff_proc assemble_eff_stiff_callback,
                                 double stiff_mult, double mass_mult
                                )
{
  W_node_t *n;
  W_index_t idcs[3];
  W_bool_t constrained[3];
  double massmatrix[3][3]={{0,0,0},{0,0,0},{0,0,0}};
  NODE_LOOP_START (d, n) {
    massmatrix[0][0] = massmatrix[1][1] = massmatrix[2][2] = mass_mult * WN_nodal_mass (n);
    WN_eqn_numbers_and_constraints (n, &idcs[0], &constrained[0]);
    (*assemble_eff_stiff_callback)  (d->newmark_integrator, FLES_ADD,
                                     3, idcs, (double *) massmatrix, constrained);
  }
  NODE_LOOP_END;
}

/*
  Assemble effective stiffness.
  Assemble effective stiffness for the Newmark integrator.
  It is the scaled mass matrix and the constitutive and geometrical stiffness (and load stiffness
  if applicable).
 */
void
WD_eff_stiff_Newm_impl (W_domain_t *d, /* domain */
                        W_newmark_assemble_eff_stiff_proc assemble_eff_stiff_callback, /* callback into the integrator to assemble the element matrices */
                        W_newmark_get_incr_proc newmark_get_incr, /* displacement
                                                                                               increment access callback
                                                                                            */
                        W_newmark_predictor_enum newmark_predictor, /* predictor to use */
                        W_float_t stiff_mult, /* stiffness matrix multiplier; should be
                                                                     disregarded if < 0; the domain is in that
                                                                    case free to use the default Newmark
                                                                    stiffness multiplier (including stiffness-proportional
                                                                    Rayleigh damping). */
                        W_float_t mass_mult /* mass matrix multiplier; should be
                                                                     disregarded if < 0; the domain is in that
                                                                    case free to use the default Newmark
                                                                    mass multiplier (including mass-proportional
                                                                    Rayleigh damping). */
                       )
{
  double curr_mass_damping_coeff = WD_curr_mass_damping_coeff (d);
  double beta, gamma, ic00, ic01, dt = WD_time_step (d);
  double sm = stiff_mult, mm = mass_mult;

  W_newmark_get_consts (d->newmark_integrator, &beta, &gamma);
  ic00 = 1.0 / dt / dt / beta;
  ic01 = gamma / dt / beta;
  if (sm < 0)
    sm = 1.0;
  if (mm < 0)
    mm = ic00 + curr_mass_damping_coeff * ic01;

  /* configure domain for this call */
  d->newmark_get_incr  = newmark_get_incr;
  d->newmark_predictor = newmark_predictor;

  if (sm == 0 && mm != 0 &&
      (WOPT_OPT (d->options, mass_type) == W_MASS_DIAG)) {
    assemble_mass_from_nodal_masses (d, assemble_eff_stiff_callback, sm, mm);
    goto exit;
  }

  ELEM_LOOP_START (d, e) {
    WE_assemble_eff_stiffness (e, assemble_eff_stiff_callback, sm, mm,
                               WOPT_OPT (d->options, mass_type));
  }
  ELEM_LOOP_END;

  /* undo the configuration */
exit:
  d->newmark_get_incr = NULL;
  d->newmark_predictor = W_NEWMARK_PRED_UNKNOWN;
}

/*
  Assemble residual.
  Assemble residual forces for the Newmark integrator.
  The integrator supplies a callback to call from within
  node-data access routines to compute current displacement,
  velocity, and acceleration which corresponds to the predicted
  configuration variables and the current displacement
  increment within the time step.
  The callback into the Newmark integrator must return
  the current value of the displacement increment.
 */
void
WD_resid_force_Newm_impl (
  W_domain_t *d, /* domain */
  W_newmark_assemble_residual_force_proc assemble_residual_force, /* callback into the Newmark integrator to assemble the residual force */
  W_mat_state_update_form_enum update_form, /* update material
                                                                         permanently or
                                                                         temporarily? */
  W_newmark_get_incr_proc newmark_get_incr, /* displacement
                                                                         increment access callback
                                                                      */
  W_newmark_predictor_enum newmark_predictor, /* predictor to use */
  double *extf_norm, /* as a side-effect, compute also
                                                  the norm of the external forces
                                               */
  double *inertf_norm, /* and the norm of the inertial forces */
  double *resid_norm /* residual norm */
)
{
  double f[3];
  double displ[3]; /* current displacement */
#if W_ROT_DOFS

  W_index_t ens[6];
  W_bool_t constrained[6];
#else

  W_index_t ens[3];
  W_bool_t constrained[3];
#endif

  int iens[3];
  double curr_mass_damping_coeff = WD_curr_mass_damping_coeff (d);

  /* configure domain for this call */
  d->newmark_get_incr  = newmark_get_incr;
  d->newmark_predictor = newmark_predictor;

  *extf_norm   = 0;
  *inertf_norm = 0;
  *resid_norm  = 0;

  NODE_LOOP_START (d, n) {
    WN_prep_force_assembly (n);
  }
  NODE_LOOP_END;

  LOAD_LOOP_START (d, l) {
    WL_assemble_load (l, d);
  }
  LOAD_LOOP_END;

  NODE_LOOP_START (d, n) {
    *extf_norm += WN_extf_norm_squared (n);
    *inertf_norm += WN_inertf_norm_squared (n);
  }
  NODE_LOOP_END;
  *extf_norm   = sqrt (*extf_norm);
  *inertf_norm = sqrt (*inertf_norm);

  CONTACT_LOOP_START (d, c) {
    WC_search (c);
  }
  CONTACT_LOOP_END;

  ELEM_LOOP_START (d, e) {
    WE_assemble_residual_forces (e, update_form, WOPT_OPT (d->options, mass_type));
  }
  ELEM_LOOP_END;

  if (curr_mass_damping_coeff != 0) {
    W_float_t dt = WD_time_step (d);
    NODE_LOOP_START (d, n) {
      WN_assemble_damping (n, curr_mass_damping_coeff, dt);
    }
    NODE_LOOP_END;
  }

  NODE_LOOP_START (d, n) {
    WN_eqn_numbers_and_constraints (n, ens, constrained);
    WN_get_force (n, &f[0], &f[1], &f[2]);
    WN_get_displ (n, &displ[0], &displ[1], &displ[2]);
    if (!constrained[0])
      *resid_norm += (f[0] * f[0]);
    else
      f[0] = 0;
    if (!constrained[1])
      *resid_norm += (f[1] * f[1]);
    else
      f[1] = 0;
    if (!constrained[2])
      *resid_norm += (f[2] * f[2]);
    else
      f[2] = 0;
    /*
        fprintf (stderr, "%s = [ ", "displ");
        fprintf (stderr, "%g ", displ[0]);
        fprintf (stderr, "%g ", displ[1]);
        fprintf (stderr, "%g ", displ[2]);
        fprintf (stderr, "]\n");
     
        fprintf (stderr, "%s = [ ", "resid_f");
        fprintf (stderr, "%g ", f[0]);
        fprintf (stderr, "%g ", f[1]);
        fprintf (stderr, "%g ", f[2]);
        fprintf (stderr, "]\n");
    */
    iens[0] = ens[0];
    iens[1] = ens[1];
    iens[2] = ens[2];
    (*assemble_residual_force) (d->newmark_integrator, FLES_ADD, 3, iens, f);
  }
  NODE_LOOP_END;
  *resid_norm = sqrt (*resid_norm);

#ifdef X_XFACE                  /* RAW for debugging */

  if (WOPT_OPT (d->options, X_xface) &&
      WOPT_OPT (d->options, interactive_during_outputs)) {
    WD_build_grep (d, FALSE);
    WXF_show_graphics (TRUE, "Iteration", WD_time (d));
  }
#endif

  /* undo the configuration */
  d->newmark_get_incr = NULL;
  d->newmark_predictor = W_NEWMARK_PRED_UNKNOWN;
}

/*
  Start implicit Newmark step.
 */
void
WD_start_step_Newm_impl (W_domain_t *d /* domain */
                        )
{
  /* undo configuration of the callbacks */
  d->newmark_get_incr = NULL;
  d->newmark_predictor = W_NEWMARK_PRED_UNKNOWN;
}

/*
  Implicit Newmark configuration update.
  Update configuration for implicit Newmark integrator.
  The actual configuration variables (displacements, velocities, accelerations)
  are obtained by the node-data access functions, which `know' how to interpret
  the Newmark equations, and the final values in the step are then
  copied to the node by using WN_set_xxx().
 */
void
WD_updt_cfg_Newm_impl (W_domain_t *d, /* domain */
                       W_newmark_get_incr_proc newmark_get_incr, /*
                                                                                        callback into the Newmark
                                                                                        integrator to obtain the displacement
                                                                                        increment in the current step
                                                                                      */
                       W_newmark_predictor_enum newmark_predictor /* predictor to use */
                      )
{
  W_vect_t u, v, a;
  W_bool_t apply[6] = {0,0,0,0,0,0};
  W_bool_t fixed[6]
    = {0,0,0,0,0,0};


  /* first set the previous displacement */
  d->newmark_get_incr = NULL;
  d->newmark_predictor = W_NEWMARK_PRED_UNKNOWN;

  /* enforce essential boundary conditions */
  DISP_BC_LOOP_START (d, bc) {
    WBC_enforce (bc);
  }
  DISP_BC_LOOP_END;

  NODE_LOOP_START (d, n) {
    WN_is_fixed (n, fixed)
      ;
    {
      int j;
      FOR (j, 6) {
        apply[j] = ! fixed[j]
                     ;
      }
    }
    WN_get_displ (n, &u.x, &u.y, &u.z);
    WN_set_prev_displ (n, &u, apply); /* restart value */
  }
  NODE_LOOP_END;

  /* configure domain for integrator; access the new, converged values
     by calling into the integrator   */
  d->newmark_get_incr = newmark_get_incr;
  d->newmark_predictor = newmark_predictor;
  NODE_LOOP_START (d, n) {
    WN_is_fixed (n, fixed)
      ;
    {
      int j;
      FOR (j, 6) {
        apply[j] = ! fixed[j]
                     ;
      }
    }

    WN_get_displ (n, &u.x, &u.y, &u.z);
    WN_get_velocity (n, &v.x, &v.y, &v.z);
    WN_get_acceleration (n, &a.x, &a.y, &a.z);

    WN_set_displ (n, &u, apply);
    WN_set_velocity (n, &v, apply);
    WN_set_acceleration (n, &a, apply);
  }
  NODE_LOOP_END;

  /* undo configuration */
  d->newmark_get_incr = NULL;
  d->newmark_predictor = W_NEWMARK_PRED_UNKNOWN;
}


W_float_t
WD_finalize_Newm_impl (W_domain_t *d, W_newmark_incr_disp_proc incr_disp_callback)
{
  W_coords_3d_t transl_incr;
  W_coords_3d_t rot_incr;
#if W_ROT_DOFS

  W_index_t ens[6];
  W_bool_t constrained[6];
#else

  W_index_t ens[3];
  W_bool_t constrained[3];
  int iens[3];
  double u[3];
#endif

  if (d->modr)
    W_modr_accumulate (d->modr);

  /* compute external-load work */
  update_ext_load_work (d);

  /* tell the integrator what the increment of displacement was */
  if (incr_disp_callback) {
    NODE_LOOP_START (d, n) {
      WN_get_disp_increment (n, &transl_incr, &rot_incr);
      WN_eqn_numbers_and_constraints (n, ens, constrained);
      iens[0] = ens[0];
      iens[1] = ens[1];
      iens[2] = ens[2]; /* RAW */
      u[0] = transl_incr.x;
      u[1] = transl_incr.y;
      u[2] = transl_incr.z;
      (*incr_disp_callback) (d->newmark_integrator, 3, iens, u);
    }
    NODE_LOOP_END;
  }

  if (!list_empty(d->finalize_callbacks)) {
    W_finalize_callback_t *cd;

    loop_through_list(d->finalize_callbacks, cd, W_finalize_callback_t *) {
      cd->callback(cd->client);
    }
  }

  {
    int tfid = WOPT_OPT (d->options, time_step_reduction_func);
    if (tfid != 0) {
      double R = 0;
      W_time_func_t *tf = WD_get_time_func_by_idx(d, tfid);
      if (tf == NULL) {
        sprintf (W_buf(), "Time function %d for the time step reduction factor not found", tfid);
        REPORT_FATAL_ERROR (d->errh, W_buf());
      }
      R = WTF_func_val (tf, WD_time(d));
      return (d->stable_time_step * R);
    } else { /* use the previous time step */
      return WD_time_step (d);
    }
  }
}

/*
  Get number of equations.
  Get number of equations of motion the domain nodes possess.
 */
int
WD_tot_of_equations (W_domain_t *d /* domain */
                    )
{
#if W_ROT_DOFS
  W_index_t ens[6];
  W_bool_t constrained[6];
#else

  W_index_t ens[3];
  W_bool_t constrained[3];
#endif

  W_index_t maxeqn;

  maxeqn = 0;
  NODE_LOOP_START (d, n) {
    WN_eqn_numbers_and_constraints (n, ens, constrained);
    maxeqn = max (maxeqn, ens[0]);
    maxeqn = max (maxeqn, ens[1]);
    maxeqn = max (maxeqn, ens[2]);
  }
  NODE_LOOP_END;
  return (maxeqn + 1);
}

/*
  Implicit Newmark integration.
  Integrate the equations of motion for the domain.  Use implicit (Newmark)
  integration.
 */
void
WD_integrate_impl (W_domain_t *d /* integrate */
                  )
{
  W_float_t start_time = 0;
  double real_time, usr_time, systm_time;
  W_postfile_times_list_t OL;
  W_bool_t should_flush
#ifdef X_XFACE
  = (WOPT_OPT (d->options, X_xface)                      /* flush for X interface */
     || WOPT_OPT (d->options, time_between_outputs) == 0 /* for forced output each step */
    );
#else

  = 0;
#endif

  OL = WOPT_OPT (d->options, postfile_times_list);

  if (d->newmark_integrator == NULL) {

    if (d->modr && W_modr_is_in_use_mode (d->modr)) {
      d->newmark_integrator =
        W_newmark_new_modr_integrator
        ((W_newmark_client_t) d, d, 0.0, WD_compute_dt (d),
         0, 0, /* Newmark constants take default values */
         (fles_tot_of_equations_callback) WD_tot_of_equations,
         (W_newmark_assemble_eff_stiff_client_proc) WD_eff_stiff_Newm_impl,
         (W_newmark_assemble_residual_force_client_proc) WD_resid_force_Newm_impl,
         (W_newmark_start_step_client_proc) WD_start_step_Newm_impl,
         (W_newmark_update_config_client_proc) WD_updt_cfg_Newm_impl,
         (W_newmark_finalize_step_client_proc) WD_finalize_Newm_impl,
         WOPT_OPT (d->options, newmarki_options)
        );
      CHECK_MALLOC (d->newmark_integrator);
    } else { /* default: full FE system integration */
      d->newmark_integrator =
        W_newmark_new_dflt_integrator
        ((W_newmark_client_t) d, d, 0.0, WD_compute_dt (d),
         0, 0, /* Newmark constants take default values */
         (fles_tot_of_equations_callback) WD_tot_of_equations,
         (W_newmark_assemble_eff_stiff_client_proc) WD_eff_stiff_Newm_impl,
         (W_newmark_assemble_residual_force_client_proc) WD_resid_force_Newm_impl,
         (W_newmark_start_step_client_proc) WD_start_step_Newm_impl,
         (W_newmark_update_config_client_proc) WD_updt_cfg_Newm_impl,
         (W_newmark_finalize_step_client_proc) WD_finalize_Newm_impl,
         WOPT_OPT (d->options, newmarki_options)
        );
      CHECK_MALLOC (d->newmark_integrator);
    }
  }
  assemble_mass (d);
  WD_apply_initial_conditions (d);

  WD_flush_history (d);

  start_time = WD_time (d);

advance:
  do {
    if (d->check_interrupt)
      d->check_interrupt (d);
    if (WD_time(d) - start_time >= WOPT_OPT (d->options, time_between_outputs)) {
      start_time = WD_time(d);

      if (!WOPT_OPT (d->options, no_output)) {
        if (WOPT_OPT (d->options, verbose)) {
          W_get_times (&real_time, &usr_time, &systm_time);
          sprintf(W_buf(), "Time=%g (r=%g; u=%g; s=%g)",
                  WD_time(d), real_time, usr_time, systm_time);
          W_log_update(d->logh, W_buf());
        }
        WHR_update_history(d->history_writer, should_flush);
      }
#ifdef X_XFACE
      if (WOPT_OPT (d->options, X_xface)) {
        WD_build_grep (d, FALSE);
        sprintf(W_buf(), "Time = %g", WD_time(d));
        WXF_show_graphics (WOPT_OPT (d->options, interactive_during_outputs),
                           W_buf(), WD_time (d));
      }
#endif

    }
    if (WOPT_OPT (d->options, write_postfile_file_now)) {
      W_bool_t nseq = WOPT_OPT (d->options, postfile_number_seq);
      WOPT_SET_OPT (d->options, postfile_number_seq, FALSE); /* force labelling w/ time, not sequential number */
      WD_write_postfile(d, WD_time(d), WD_time(d));
      WOPT_SET_OPT (d->options, write_postfile_file_now, FALSE);
      WOPT_SET_OPT (d->options, postfile_number_seq, nseq);
    } else {
      if (   (!WOPT_OPT (d->options, no_output))
             && (   (OL.ntimes > 0) && (OL.curr < OL.ntimes))) {
        if ((OL.ntimes > 0) && (WD_time(d) >= OL.times[OL.curr])) {
          WD_write_postfile(d, WD_time(d), OL.times[OL.curr]);
          OL.curr++;
        }
      }
    }
    if (d->response_observer_func)
      (d->response_observer_func) (d, FALSE);
    
    W_newmark_advance (d->newmark_integrator);
  } while (WD_time (d) < WOPT_OPT (d->options, target_time));

#ifdef X_XFACE

  if (WOPT_OPT (d->options, X_xface)) {
    W_get_times(&real_time, &usr_time, &systm_time);
    sprintf(W_buf(), "Time reached = %g (r=%g; u=%g; s=%g)",
            WD_time(d), real_time, usr_time, systm_time);
    WXF_loop (TRUE, W_buf());
    if (WD_time (d) < WOPT_OPT (d->options, target_time))
      goto advance;
  }
#endif

  if (d->response_observer_func)
    (d->response_observer_func) (d, TRUE);

  WD_clean_up_before_exit(d);

  if (WOPT_OPT (d->options, verbose)) {
    sprintf(W_buf(), "Reached target time = %g", WD_time(d));
    W_log_update(d->logh, W_buf());
    W_get_times(&real_time, &usr_time, &systm_time);
    sprintf(W_buf(), "   Used time [sec]: real %g, user %g, sys %g",
            real_time, usr_time, systm_time);
    W_log_update(d->logh, W_buf());
  }

#ifdef NEED_STAT_MALLOC
  W_stat_malloc_statistics(stderr);
#endif
#undef OL
}


/*
  Set interrupt function.
  Set the function the domain should use to check for interrupts.
 */
void
WD_set_interrupt_func (W_domain_t *d, /* domain */
                       W_check_interrupt_func ifunc /* pointer to interrupt function */
                      )
{
  d->check_interrupt = ifunc;
}

void
WD_summary(W_domain_t *d)
{
  sprintf (W_buf (), "%d nodes, %d elements", W_dict_count(d->nodes), W_dict_count(d->elements));
    W_log_update (d->logh, W_buf ());
}
