#include "Wconfig.h" 
 
#ifdef EFG_WASP

#include "Wdomain.h"
#include "Welem.h"
#include "Wnode.h"
#include "Wmat3d.h"
#include "Wgaussp.h"
#include "Wxf.h"
#include "Wutil.h"
#include "Wcdi.h" 
#define WEFG_PRIVATE_HEADER
#include "WEFG.h" 
#include "WEFGutil.h"
#include "WEFGgc.h"
#include "WEFGpriv.h"
#include "WEFGss.h"
#include "WEFGcr.h"
#include "WEFGpub.h"
#include "Wcontface.h"

/* RCSHEADER "$Id: efg.c,v 1.7 2004/09/06 01:17:58 cvs Exp $" */

static void
set_up_efg_superelem (WEFG_superelem_t *se);
static void
find_b_faces (WEFG_superelem_t *se);

static void
generate_nodes_in_hexvols (WEFG_superelem_t *se) ;

static void
generate_nodes_inside_volume (WEFG_superelem_t *se);
static void
number_eqns (WEFG_superelem_t *se); 
static void
update_system (WEFG_superelem_t *se);
static void
recompute_mass (WEFG_superelem_t *se);
static void
apply_init_cond (WEFG_superelem_t *se);
static int
read_support_size_region (WEFG_superelem_t *se /* EFG superelement */, W_domain_t *d);


#define ALMOST_INFINITY FLT_MAX
#define TO_KEEP 3

#ifdef USE_ESFLIB
static void 
W_esflib_error_handler (ESFLIB_lib_t lib);
static int
W_esflib_find_overlaps (ESFLIB_search_struct_t SS,
                        double x, double y, double z ) 
{
  return lobs_3d_search_for_boxes_overlapping_pt ((lobs_3d_front_end) SS,
                                                   x, y, z);
} 
#endif


#define LIST_LOOP_START(list, var, type) \
        for (var = (type) get_list_next (list, NULL);    \
             var;                               \
             var = (type) get_list_next (list, var)) {
#define LIST_LOOP_END() }         


/*
   Create a superelement structure.
   Initialize to default values.  Most of the setup is finished
   in set_up_efg_superelem ().
 */
static WEFG_superelem_t *
new_efg_superelem (W_domain_t *d, int basis_order) 
{
  WEFG_superelem_t *se;
#ifndef USE_ESFLIB
  static int first = TRUE;
#endif
  W_loc_t origin =  {0,0,0} ;
  

  se = w_new (WEFG_superelem_t);
  CHECK_MALLOC (se);

  se->domain                                   = d;
  se->node_spacing                             = 0;
  se->merging_distance                         = 100 * MACHEPS;
  se->shortest_distance                        = FLT_MAX;
  se->num_nodes_per_hex_side                   = 0;
  
  se->compute_support_size                     = FALSE;
  se->support_size                             = 0;
  se->support_size_region_list                 = NULL;
  se->max_support_size                         = 0;
  se->support_size_multiplier                  = 1;

  se->sflib_prec_tol                           = 1.0e4 * MACHEPS;
  se->num_neighbors_min                        = MIN_NUM_NEIGHBORS;
  se->num_neighbors_suggested                  = 20;
  se->quadrature_order                         = DEFAULT_QUAD_ORDER;
# ifdef ENABLE_TWO_POINT_RULE
  se->enable_2pt_quadrature                    = FALSE;
# endif
  se->verts                                    = dict_create_dictionary ();
  se->vertex_ss_bucket_size                    = 100;
  se->verts_ss
    = pirat_new_pirat (se->vertex_ss_bucket_size, 1000 * MACHEPS);
  se->hex_volumes                              = make_list ();
  se->hex_ss                                   = lobs_3d_new_instance (NULL, 10);
  se->cracks                                   = make_list ();
  se->b_faces                                  = make_list ();
  se->is_convex                                = TRUE;
  se->is_convex_with_cracks                    = FALSE;
  se->generate_nodes_inside_only               = FALSE;
  se->generate_nodes_at_vertices               = FALSE;
  se->generate_nodes_in_volume                 = FALSE;
  se->lower_corner = se->upper_corner = origin;
  se->efgnodes                                 = dict_create_dictionary ();
  se->efg_node_ss_bucket_size                  = 15;
  se->efg_node_ss                              = lobs_3d_new_instance (NULL, se->efg_node_ss_bucket_size);
  se->conc_loads                               = make_list ();
  se->distrib_loads                            = make_list ();
  se->contact_loads                            = make_list ();
  se->body_load.is_set                         = FALSE;
  se->body_load.is_time_independent            = FALSE;
  se->body_load.is_constant                    = FALSE;
  se->body_load.body_load_expr[0]              =
  se->body_load.body_load_expr[1]              =
  se->body_load.body_load_expr[2]              = NULL;
  se->body_load.time_independent_body_load.dim = 0;
  se->body_load.time_independent_body_load.v   = NULL;
  se->mat                                      = NULL;
  
  se->num_of_equations                         = 0;
  
  se->mass_matrix.dim                          = 0;
  se->mass_matrix.v                            = NULL;

  se->internal_mass_only                       = FALSE;
  se->time_independent_mass                    = FALSE;
  
  se->total_mass                               = 0;
  
  se->effective_loads.dim                      = 0;
  se->effective_loads.v                        = NULL;
  
  se->update_vector.dim                        = 0;
  se->update_vector.v                          = NULL;
  
  se->ext_load_work                            = 0;

  se->tolerance                                = 0.0001;

  se->symm_planes                              = NULL;

  se->use_wrap_around                          = FALSE;

  se->compute_stress_at_output_pts
    = ((WOPT_OPT (se->domain->options, postfile_output_spec_deprec) & W_POSTFILE_STRESS_MASK) 
       || (WOPT_OPT (se->domain->options, postfile_output_spec_deprec) & W_POSTFILE_EQPLS_MASK) 
       || (WOPT_OPT (se->domain->options, postfile_output_spec_deprec) & W_POSTFILE_VON_MISES_MASK) 
       );
  se->compute_equiv_pl_def_at_output_pts
    = (WOPT_OPT (se->domain->options, postfile_output_spec_deprec) & W_POSTFILE_EQPLS_MASK);

  INIT_EFG_SF_STAT (se);

  se->invalid_after_time                        = 0;
  se->invalid_regions.triangles_are_reliable    = TRUE;
  se->invalid_regions.regions                   = NULL;

  se->ip_ss_bucket_size                         = 100;
  se->ip_ss =
    pirat_new_pirat (se->ip_ss_bucket_size, 1000 * MACHEPS);
  se->store_ip_in_search_structure              = TRUE;
  se->force_all_sf                              = FALSE;
  se->draw_rays                                 = FALSE;
  se->mass_damping_coeff               = WOPT_OPT (d->options, rayleigh_mass_damping_coeff);
  se->general_constraint[0] =
  se->general_constraint[1] =
  se->general_constraint[2] = FREE_DOF;

  se->scramble_diagonals    = FALSE;

  se->weight_func_type      = WEFG_SMOOTH_QUADRATIC_SPHERICAL_WEIGHT_FUNC;

  se->basis_order           = basis_order;

#ifdef USE_ESFLIB
  se->A_solver              = ESFLIB_LU_DECOMP;
#endif
  
  se->coupling_fen_conn = dict_create_dictionary ();
  se->driven_fen_conn = dict_create_dictionary ();

#ifdef USE_ESFLIB
  se->esflib
    = ESFL3_init (W_esflib_error_handler,
                  NULL,
 (ESFLIB_find_overlaps_func) 
                     lobs_3d_search_for_boxes_overlapping_pt,
 (ESFLIB_first_overlapping_func) 
                     lobs_3d_first_overlapping_object,
 (ESFLIB_next_overlapping_func) 
                     lobs_3d_next_overlapping_object,
                  efgn_ref_coords,
 (WOPT_OPT (se->domain->options, optimization_level) == 0
                   ? ESFL3_poly_basis
                   : NULL),
 (WOPT_OPT (se->domain->options, optimization_level) == 0
                   ? ESFL3_D_poly_basis
                   : NULL),
                  (se->basis_order == WEFG_COMPLETE_LINEAR
                   ? ESFL3_NUM_LINEAR_BASIS_TERMS
                   : ESFL3_NUM_QUADRATIC_BASIS_TERMS) ,
                  ESFL3_qs_w,
                  ESFL3_qs_D1,
                  efgn_support_size,
                  hex_construct_ramp,  
                  hex_elem_node_num,   
                  hex_fe_shape,        
                  0, 
                  node_hidden_by_boundary
                  );
  if (se->esflib == NULL) {
    REPORT_FATAL_ERROR (d->errh, (char *) ESFL3_explain_error (se->esflib));
    return NULL;
  } 
  if (WOPT_OPT (se->domain->options, optimization_level) == 0) 
    ESFL3_set_opt_level (se->esflib, ESFLIB_OPT_NONE);
  else
    ESFL3_set_opt_level (se->esflib, ESFLIB_OPT_FULL);

  ESFL3_choose_A_solver (se->esflib, se->A_solver);
  
#else
  if (first) {
    if (!ESP_3_init (NULL,
                    1.0,
                    SFL_search_cast SEARCH_FUNC_NAME,
                    FIRST_IN_RANGE_FUNC_NAME,	
                    NEXT_IN_RANGE_FUNC_NAME,
                    efgn_ref_coords,
 (options.optimization_level == 0 
                     ? ESP_3_poly_basis
                     : NULL),
 (options.optimization_level == 0
                     ? ESP_3_D_poly_basis
                     : NULL),
                    ESP_3_NUM_LINEAR_BASIS_TERMS,
                    ESP_3_qs_w,
                    ESP_3_qs_D1,
                    efgn_support_size,
                    hex_construct_ramp,  
                    hex_elem_node_num,   
                    hex_fe_shape,        
                    0, 
                    node_hidden_by_boundary
                    )) {
      REPORT_FATAL_ERROR (d->errh, ESP_3_explain_error ());
      return NULL;
    } 
    if (options.optimization_level == 0) 
      ESP_3_set_opt_level (ESPACK_OPT_NONE);
    else
      ESP_3_set_opt_level (ESPACK_OPT_FULL);
    first = FALSE;
  }  /* if (first) */
#endif
  return se;
} 



static int
num_of_faces (W_elem_t *el)
{
  WEFG_superelem_t *se = (WEFG_superelem_t *)el->data;
  return list_length(se->b_faces);
}

static int
get_face_vertex_nodes (W_elem_t *el, int nface, int ndim, W_node_t *n[])
{
  WEFG_superelem_t *se = (WEFG_superelem_t *)el->data;
  int i;

  if (ndim < 4)
    return 0;

  { WEFG_b_face_t *bf = get_list_pos (se->b_faces, nface+1);
    if (bf) {
      FOR (i, bf->num_vert) {
        n[i] = bf->v[i]->efgn->fen;
      }
      return bf->num_vert;
    } else
      return 0;
  }
}

static int
get_contact_faces (W_elem_t *el, int nface,
                   WCF_contact_face_t *cf[], int maxcf)
{
  WEFG_superelem_t *se = (WEFG_superelem_t *)el->data;
  int i;

  if (maxcf < 1) {
    FOR (i, maxcf) {
      cf[i] = NULL;
    }
    return 1;
  }


  { WEFG_b_face_t *bf = get_list_pos (se->b_faces, nface+1);
    if (bf) {
      if (bf->num_vert == 3) 
        cf[0] =    WCF_new_contact_face (bf->v[0]->efgn->fen, bf->v[1]->efgn->fen, bf->v[2]->efgn->fen, 0);
      else if (bf->num_vert == 4)
        cf[0] =    WCF_new_contact_face (bf->v[0]->efgn->fen, bf->v[1]->efgn->fen, bf->v[2]->efgn->fen, bf->v[3]->efgn->fen);
      else {
        cf[0] = 0;
        return 0;
      }
    } else
      cf[0] = 0;
  }

  return 1;
}




static void
set_up_symm_planes (WEFG_superelem_t *se);

static void
check_for_subsplitted_iface_hvs (WEFG_superelem_t *se);

static void
generate_nodes_in_hexvols (WEFG_superelem_t *se) 
{
  WEFG_hex_volume_t *h;
  
  LIST_LOOP_START (se->hex_volumes, h, WEFG_hex_volume_t *) {
    hex_generate_efg_nodes (se, h);
  }  LIST_LOOP_END ();
} 

/*
   Generate nodes at vertices.
   Generate nodes at vertices.
 */
static void
generate_nodes_at_vertices (WEFG_superelem_t *se) 
{
    WEFG_vertex_t *v;
    v = NULL;
    while ((v = (WEFG_vertex_t *) dict_next_obj (se->verts, v)) ) {
      efgv_generate_efgn (v, se);
    } 
} 


/*
   Set up settings.
   Set up the settings for the superlement once it has been read.
   Configure ESFLib.  Find boundary faces.  Set up symmetry planes.
   Number equations.  Update equation-of-motion systems.
   If cracks defined, intersect hex integration volumes by cracks.
 */
static void
set_up_efg_superelem (WEFG_superelem_t *se) 
{
  /* Need to establish:
 (i) Boundary faces
 (ii) Generate EFG nodes
 (iii) Number equations
 (iv) Generate the list of coupling hex volumes
   */
# ifdef ENABLE_TWO_POINT_RULE
  if (se->enable_2pt_quadrature) 
    se->quadrature_order = 2;
# endif
  
  if (se->verts_ss == NULL) {
    se->verts_ss = pirat_new_pirat (se->vertex_ss_bucket_size, 1000 * MACHEPS);
  } 

  if (se->efg_node_ss == NULL) {
    se->efg_node_ss = lobs_3d_new_instance (NULL, se->efg_node_ss_bucket_size);
  } 

#ifdef USE_ESFLIB
  ESFL3_set_search_structure (se->esflib, (char *) se->efg_node_ss);
  ESFL3_set_min_num_neighbors (se->esflib, se->num_neighbors_min);
#ifdef USE_WRAP_AROUND
  if (se->use_wrap_around) {
    REPORT_FATAL_ERROR ("Wrap-around not available");
  } 
#endif
  if (se->weight_func_type
             == WEFG_QUADRATIC_TENSOR_PROD_WEIGHT_FUNC) {
    ESFL3_set_weight_funcs (se->esflib, ESFL3_qtp_w, ESFL3_qtp_D1);
  }  else if (se->weight_func_type
             == WEFG_SMOOTH_QUADRATIC_TENSOR_PROD_WEIGHT_FUNC) {
    ESFL3_set_weight_funcs (se->esflib, ESFL3_sqtp_w, ESFL3_sqtp_D1);
  }  else if (se->weight_func_type
             == WEFG_QUARTIC_SPLINE_SPHERICAL_WEIGHT_FUNC) {
    ESFL3_set_weight_funcs (se->esflib, ESFL3_qs_w, ESFL3_qs_D1);
  }  else if (se->weight_func_type
             == WEFG_QUADRATIC_SPHERICAL_WEIGHT_FUNC) {
    ESFL3_set_weight_funcs (se->esflib, ESFL3_q_w, ESFL3_q_D1);
  }  else if (se->weight_func_type
             == WEFG_GAUSSIAN_SPHERICAL_WEIGHT_FUNC) {
    ESFL3_gh_init (4.0);
    ESFL3_set_weight_funcs (se->esflib, ESFL3_gh_w, ESFL3_gh_D1);
  }  else if (se->weight_func_type
             == WEFG_SMOOTH_QUADRATIC_SPHERICAL_WEIGHT_FUNC) {
    ESFL3_set_weight_funcs (se->esflib, ESFL3_sq_w, ESFL3_sq_D1);
  }  else  {
    REPORT_FATAL_ERROR (se->domain->errh, "Unknown weight function type");
  } 
  ESFL3_set_prec_tol (se->esflib, se->sflib_prec_tol);
#else
  ESP_3_set_search_structure ((char *) se->efg_node_ss);
  ESP_3_set_range_size (se->max_support_size);
  ESP_3_set_min_num_neighbors (se->num_neighbors_min);
#ifdef USE_WRAP_AROUND
  if (se->use_wrap_around) {
    if (!ESP_3_enable_wrap_around_CPP ()) {
      REPORT_FATAL_ERROR ("Wrap-around not available");
    } 
    if (!ESP_3_set_wrap_around_CPP_funcs (efggc_use_cpp,
                                         efggc_plane_generating_vector,
                                         efggc_wrap_around_point)) {
      REPORT_FATAL_ERROR ("Wrap-around not available");
    } 
  } 
#endif
  if (se->weight_func_type
             == WEFG_QUADRATIC_TENSOR_PROD_WEIGHT_FUNC) {
    if (se->use_wrap_around) {
      REPORT_FATAL_ERROR ("Cannot mix tensor-product weight + wrap around");
    } 
    ESP_3_set_weight_funcs (ESP_3_qtp_w, ESP_3_qtp_D1);
  }  else if (se->weight_func_type
             == WEFG_SMOOTH_QUADRATIC_TENSOR_PROD_WEIGHT_FUNC) {
    if (se->use_wrap_around) {
      REPORT_FATAL_ERROR ("Cannot mix tensor-product weight + wrap around");
    } 
    ESP_3_set_weight_funcs (ESP_3_sqtp_w, ESP_3_sqtp_D1);
  }  else if (se->weight_func_type
             == WEFG_QUARTIC_SPLINE_SPHERICAL_WEIGHT_FUNC) {
    ESP_3_set_weight_funcs (ESP_3_qs_w, ESP_3_qs_D1);
  }  else if (se->weight_func_type
             == WEFG_QUADRATIC_SPHERICAL_WEIGHT_FUNC) {
    ESP_3_set_weight_funcs (ESP_3_q_w, ESP_3_q_D1);
  }  else if (se->weight_func_type
             == WEFG_GAUSSIAN_SPHERICAL_WEIGHT_FUNC) {
    ESP_3_gh_init (4.0);
    ESP_3_set_weight_funcs (ESP_3_gh_w, ESP_3_gh_D1);
  }  else if (se->weight_func_type
             == WEFG_SMOOTH_QUADRATIC_SPHERICAL_WEIGHT_FUNC) {
    ESP_3_set_weight_funcs (ESP_3_sq_w, ESP_3_sq_D1);
  }  else  {
    REPORT_FATAL_ERROR ("Unknown weight function type");
  } 
#endif

  find_b_faces (se);

  set_up_symm_planes (se);

  if (se->generate_nodes_in_volume) 
    generate_nodes_inside_volume (se);
  else if (se->generate_nodes_at_vertices)
    generate_nodes_at_vertices (se);
  else
    generate_nodes_in_hexvols (se);

  check_for_subsplitted_iface_hvs (se);

  number_eqns (se);

  update_system (se);

  if (WOPT_OPT (se->domain->options, j_dom_eval) == W_PER_CELL_IPOL_PTS) {
    WEFG_crack_surf_t *cs;
    if (se->cracks) {
      LIST_LOOP_START (se->cracks, cs, WEFG_crack_surf_t *) {
        WEFG_hex_volume_t *h;
        LIST_LOOP_START (se->hex_volumes, h, WEFG_hex_volume_t *) {
          if (!h->is_intersected_by_crack_surf) 
            h->is_intersected_by_crack_surf
              = efgcr_hex_intersected_by_crack_surf (cs, h);
        }  LIST_LOOP_END ();
      }  LIST_LOOP_END ();
    } 
  } 
} 



/*
  Has an interface hex volume been split?
   Has an interface hex volume been sub-split?  Disallowed!
 */
static void
check_for_subsplitted_iface_hvs (WEFG_superelem_t *se) 
{
  WEFG_hex_volume_t *h;
  
  LIST_LOOP_START (se->hex_volumes, h, WEFG_hex_volume_t *) {
    if (hex_is_interface_vol (h) && h->offspring) {
      REPORT_FATAL_ERROR (se->domain->errh, "EFG/FE hex volume CANNOT be sub-split");
    } 
  }  LIST_LOOP_END ();
} 

/*
   Set up symmetry planes.
 */
static void
set_up_symm_planes (WEFG_superelem_t *se) 
{
  WEFG_symm_plane_t *sp;
#define SHIFT(p, by) {p.x += by.x;    p.y += by.y;    p.z += by.z;} 

  if (se->symm_planes == NULL) return;

  LIST_LOOP_START (se->symm_planes, sp, WEFG_symm_plane_t *) {
    if (sp->tolerance <= 0) sp->tolerance = se->tolerance;
     {
      W_loc_t px =  {1, 0, 0} , py =  {0, 1, 0} , pz =  {0, 0, 1} ;
      SHIFT (px, sp->center);
      SHIFT (py, sp->center);
      SHIFT (pz, sp->center);
      if (sp->direction_of_normal == 'X') {
        G3D_plane_consts_3 (&sp->plane, &sp->center, &py, &pz);
      }  else if (sp->direction_of_normal == 'Y') {
        G3D_plane_consts_3 (&sp->plane, &px, &sp->center, &pz);
      }  else if (sp->direction_of_normal == 'Z') {
        G3D_plane_consts_3 (&sp->plane, &px, &py, &sp->center);
      }  else  {
        REPORT_FATAL_ERROR (se->domain->errh, "bad symm plane normal");
      } 
    } 
  }  LIST_LOOP_END ();
} 

static void 
init_body_load (WEFG_superelem_t *se);


/*
   Apply initial conditions.
   Apply initial conditions to all EFG superlement nodes.
   Initialize body load.
 */
static void
apply_init_cond (WEFG_superelem_t *se) 
{
  WEFG_node_t *en;
  W_vect_t v;
  W_float_t time_step = WD_time_step (se->domain);
  W_loc_t rl;
  double x, y, z;

  en = NULL;
  while ((en = (WEFG_node_t *) dict_next_obj (se->efgnodes, en)) ) {
    efgn_ref_coords (en, &x, &y, &z);
    rl.x = x; rl.y = y; rl.z = z;
    WD_initial_velocity (se->domain, &rl, &v);
    efgn_apply_init_cond (en, &v, time_step);
  } 
  if (se->body_load.is_set && se->body_load.is_time_independent) {
    init_body_load (se);
  } 
} 

/*
   Find boundary faces.
   Find boundary faces by looking at faces of the hex integration volumes
   which are not shared by two integration volumes.

   Limitations:

   If adjacent integration volumes are split unequally it is incorrectly deduced
   that the faces of the subsplit volumes are boundary faces eventhough
   the faces of the unsplit volumes are shared.
 */
static void
find_b_faces (WEFG_superelem_t *se) 
{
  WEFG_hex_volume_t *h;
  
  if (se->hex_volumes) {
    LIST_LOOP_START (se->hex_volumes, h, WEFG_hex_volume_t *) {
      hex_find_b_faces (se, h, 0);
    }  LIST_LOOP_END ();
  } 
#if 0
  fprintf (stderr, "b faces = %d\n", list_length (se->b_faces));
#endif
} 


/*
   Generate nodes inside a given box.
   Generate nodes inside a given rectangular box.
   The nodes are generated in a uniform cubic lattice.
   If there are any FE/EFG interface cells, also the nodes
   at the vertices of these cells are generated.

   Limitations:

   The function does not check whether a node is inside any integration volume.
 */
static void
generate_nodes_inside_volume (WEFG_superelem_t *se) 
{
  double x, y, z, xs, ys, zs, dx=0, dy=0, dz=0, xe, ye, ze;
  W_loc_t loc;
  G3D_box_t range;
# define SUPRNG(at) \
   {                                             \
    G3D_INIT_BOX (range, at);                    \
    G3D_INFLATE_BOX (range, drange);             \
  } 
  double drange = se->merging_distance;

  xs = se->lower_corner.x;
  ys = se->lower_corner.y;
  zs = se->lower_corner.z;
  if (se->node_spacing > 0) {
    dx = dy = dz = se->node_spacing;
    if (se->compute_support_size) {
      se->shortest_distance 
        = sqrt (3) * se->node_spacing * se->support_size_multiplier;
      efg_set_support_size (se, se->shortest_distance);
    } 
  }  else  {
    REPORT_FATAL_ERROR (se->domain->errh, "Spacing not specified with GENERATE_NODES_IN_VOLUME");
  }  
  xe = se->upper_corner.x + dx/10;
  ye = se->upper_corner.y + dy/10;
  ze = se->upper_corner.z + dz/10;
  for (x = xs; x < xe; x += dx) {
    for (y = ys; y < ye; y += dy) {
      for (z = zs; z < ze; z += dz) {
        loc.x = x; loc.y = y; loc.z = z;
        SUPRNG (loc); 
        if (!efg_node_in_range (se, &range)) 
          efg_add_node (se, &loc);
        /* RAW should check whether the node is inside some cell; */
        /* if not, don't generate it */
      } 
    } 
  } 

   { /* Now generate nodes at the vertices of interface cells */
    WEFG_hex_volume_t *h;
    int i;
    WEFG_node_t *exn;

    LIST_LOOP_START (se->hex_volumes, h, WEFG_hex_volume_t *) {
      if (hex_is_interface_vol (h)) {
        FOR (i, 8) {
          if (h->vertex[i]->efgn == NULL) {
            SUPRNG ((h->vertex[i]->loc));
            exn = efg_node_in_range (se, &range);
            if (exn == NULL) 
              h->vertex[i]->efgn
                = new_efg_node (se, &h->vertex[i]->loc,
                                efg_support_size_at (se,
                                                     h->vertex[i]->loc.x,
                                                     h->vertex[i]->loc.y,
                                                     h->vertex[i]->loc.z));
            else
              h->vertex[i]->efgn = exn;
            efgn_change_to_efg_if_node (h->vertex[i]->efgn);
          }  else if (!EFGN_IS_FE_IF (h->vertex[i]->efgn)) {
            efgn_change_to_efg_if_node (h->vertex[i]->efgn);
          } 
          efg_attach_node (se, h->vertex[i]->efgn);
        } 
      } 
    }  LIST_LOOP_END ();
  } 
} 
 







/*
   Number equations of motion.
   Number equations of motion by looping over all EFG nodes.
 */
static void
number_eqns (WEFG_superelem_t *se) 
{
  WEFG_node_t *en;

  se->num_of_equations = 0;
  en = NULL;
  while ((en = (WEFG_node_t *) dict_next_obj (se->efgnodes, en)) ) {
    se->num_of_equations = efgn_number_eqns (en, se->num_of_equations,
                                            se->symm_planes);
  } 
  if (WOPT_OPT (se->domain->options, verbose)) {
    sprintf (W_buf (), "EFG superelement: Total of internal unknowns = %ld",
            se->num_of_equations);
    W_log_update (se->domain->logh, W_buf ());
  } 
}  



/*
   Update the system of equations.
   Update the system of equations of motion.
   Re-allocate where necessary.  Recompute mass matrix.
 */
static void
update_system (WEFG_superelem_t *se) 
{
  if (se->mass_matrix.v == NULL) {
    MAKE_EFGV (se->mass_matrix, se->num_of_equations);
  }  else if (se->mass_matrix.dim <= se->num_of_equations) {
    recompute_mass (se);
  } 
  if (se->effective_loads.v == NULL) {
    MAKE_EFGV (se->effective_loads, se->num_of_equations);
  }  else if (se->effective_loads.dim <= se->num_of_equations) {
    REALLOC_EFGV (se->effective_loads, se->num_of_equations);
  } 
  if (se->update_vector.v == NULL) {
    MAKE_EFGV (se->update_vector, se->num_of_equations);
  }  else if (se->update_vector.dim <= se->num_of_equations) {
    REALLOC_EFGV (se->update_vector, se->num_of_equations);
  } 
  if (se->body_load.is_set && se->body_load.is_time_independent) {
    MAKE_EFGV (se->body_load.time_independent_body_load, se->num_of_equations);
  } 
} 


/*
   Initialize body load.
 */
static void 
init_body_load (WEFG_superelem_t *se) 
{
  WEFG_hex_volume_t *h;
  int stat;
  double f;

  expeval_set_value ("x", FLT_MAX);
  expeval_set_value ("y", FLT_MAX);
  expeval_set_value ("z", FLT_MAX);
  expeval_set_value ("t", 0);
  if (se->body_load.is_constant) {
    stat = expeval_eval (se->body_load.body_load_expr[0], &f);
    if (stat != EXPEVAL_E_OK) {
      sprintf (W_buf (), "%s", expeval_err_msgs[stat]);
      REPORT_ERROR (se->domain->errh, W_buf ());
    } 
    se->body_load.const_body_load.x = f;
    stat = expeval_eval (se->body_load.body_load_expr[1], &f);
    if (stat != EXPEVAL_E_OK) {
      sprintf (W_buf (), "%s", expeval_err_msgs[stat]);
      REPORT_ERROR (se->domain->errh, W_buf ());
    } 
    se->body_load.const_body_load.y = f;
    stat = expeval_eval (se->body_load.body_load_expr[2], &f);
    if (stat != EXPEVAL_E_OK) {
      sprintf (W_buf (), "%s", expeval_err_msgs[stat]);
      REPORT_ERROR (se->domain->errh, W_buf ());
    } 
    se->body_load.const_body_load.z = f;
  } 

  EFGVDO (se, se->effective_loads.v[IX] = 0);
  LIST_LOOP_START (se->hex_volumes, h, WEFG_hex_volume_t *) {
    hex_calc_eff_loads (se, h);
  }  LIST_LOOP_END ();
  EFGVDO (se,
         se->body_load.time_independent_body_load.v[IX]
         = se->effective_loads.v[IX]);
} 

/*
   Compute total mass.
   Compute total mass of superelement by summing over all integration volumes.
 */
static W_float_t
get_total_mass (WEFG_superelem_t *se) 
{
  double tm = 0;
  WEFG_hex_volume_t *h;

  LIST_LOOP_START (se->hex_volumes, h, WEFG_hex_volume_t *) {
    tm += hex_total_mass (se, h);
  }  LIST_LOOP_END ();
  return tm;
} 


/*
   Recompute mass matrix.
 */
static void
recompute_mass (WEFG_superelem_t *se) 
{
  WEFG_hex_volume_t *h;
  W_vector_t m = se->mass_matrix.v;

#ifdef USE_ESFLIB
  ESFL3_set_search_structure (se->esflib, (char *) se->efg_node_ss);
#else
  ESP_3_set_search_structure ((char *) se->efg_node_ss);
#endif

  EFGVDO (se, m[IX] = 0);
  LIST_LOOP_START (se->hex_volumes, h, WEFG_hex_volume_t *) {
    hex_assemble_mass (se, h);
  }  LIST_LOOP_END ();
  
  se->total_mass = get_total_mass (se);
} 



/*
   Find node by integer index.
 */
WEFG_vertex_t * /* node */
efg_get_vertex (WEFG_superelem_t *se /* EFG superelement */,
                W_index_t n /* unique integer */
                ) 
{
  WEFG_vertex_t *v;

  v = dict_find_obj_with_int_name (se->verts, n);
  return v;
} 


static void
compute_bbox (WEFG_b_face_t *f);


/*
   Compute bounding box of a boundary face.
 */
static void
compute_bbox (WEFG_b_face_t *f) 
{
  int i;
  double v;
#define FPT(n) f->v[n]->loc
#define ASS(dir, v) \
   {                                                                     \
    f->bounding_box.upper.dir = max (f->bounding_box.upper.dir, v);      \
    f->bounding_box.lower.dir = min (f->bounding_box.lower.dir, v);      \
  } 
  f->bounding_box.upper.x = f->bounding_box.lower.x = FPT (0) .x;
  f->bounding_box.upper.y = f->bounding_box.lower.y = FPT (0) .y;
  f->bounding_box.upper.z = f->bounding_box.lower.z = FPT (0) .z;
  for (i = 1; i < f->num_vert; i++) {
    v = FPT (i) .x; ASS (x, v);
    v = FPT (i) .y; ASS (y, v);
    v = FPT (i) .z; ASS (z, v);
  } 
} 

/*
   Attach boundary face.
   Attach boundary face to a superelement.
 */
void
attach_b_face (WEFG_superelem_t *se /* EFG superelement */, WEFG_b_face_t *f /* boundary face */
               ) 
{
  compute_bbox (f);
  add_to_tail (se->b_faces, f);
} 

/*
   Attach crack surface.
   Attach crack surface to a superelement.
 */
void
attach_crack_surf (WEFG_superelem_t *se /* EFG superelement */, WEFG_crack_surf_t *cs) 
{
  add_to_tail (se->cracks, cs);
} 




static void
add_en_to_ss (WEFG_superelem_t *se /* EFG superelement */, WEFG_node_t *en);

/*
   Add a node.
   Add a node to the node search structure.
   The LOBS (library for overlapping box searching) is used.
 */
static void
add_en_to_ss (WEFG_superelem_t *se /* EFG superelement */,
              WEFG_node_t *en /* node */
              ) 
{
  G3D_box_t box;
  if (EFGN_IS_TRUE_EFG (en) || EFGN_IS_EFG_IF (en)) {
    G3D_INIT_BOX (box, en->ndata.ref_coords);
  }  else if (EFGN_IS_FE_IF (en)) {
    W_loc_t loc;
    WN_get_ref_coords (en->fen, &loc.x, &loc.y, &loc.z);
    G3D_INIT_BOX (box, loc);
  }  else  {
    REPORT_FATAL_ERROR (se->domain->errh, "bad EFG node type");
  }  
  G3D_INFLATE_BOX (box, en->support_size);
  lobs_3d_add_object (se->efg_node_ss, en,
                      box.lower.x, box.lower.y, box.lower.z,
                      box.upper.x, box.upper.y, box.upper.z);
} 



/*
   Any node inside a box?
   Is a there any EFG node inside a given box.
   The node search structure is used to find out.
 */
WEFG_node_t *
efg_node_in_range (WEFG_superelem_t *se /* EFG superelement */,
                   G3D_box_t *range /* box */
                   ) 
{
  WEFG_node_t *n, *nn = NULL;
  G3D_loc_t pt;

  if ((n = (WEFG_node_t *) dict_next_obj (se->efgnodes, NULL)) == NULL) 
    return NULL;

  lobs_3d_search (se->efg_node_ss,
                  range->lower.x, range->lower.y, range->lower.z,
                  range->upper.x, range->upper.y, range->upper.z);
  
  n = (WEFG_node_t *) lobs_3d_first_overlapping_object (se->efg_node_ss);
  while (n) {
    efgn_ref_coords (n, &pt.x, &pt.y, &pt.z);
    if (G3D_PT_INSIDE_BOX (*range, pt)) {
      if (nn) {
        REPORT_WARNING (se->domain->errh, "two or more nodes at a location!?");
      } 
      nn = n;
    } 
    n = (WEFG_node_t *) lobs_3d_next_overlapping_object (se->efg_node_ss);
  } 

  return nn;
} 

/*
   How far is the closest node?
   The node search structure is used to find out.
 */
 double
efg_closest_node_dist (WEFG_superelem_t *se, WEFG_node_t *en) 
{
  WEFG_node_t *n, *nn = NULL;
  G3D_loc_t enpt, pt, d;
  G3D_box_t range;
  double mindist = FLT_MAX;

  efgn_ref_coords (en, &enpt.x, &enpt.y, &enpt.z);
  G3D_INIT_BOX(range, enpt);
  G3D_INFLATE_BOX(range, en->support_size);
  
  lobs_3d_search (se->efg_node_ss,
                  range.lower.x, range.lower.y, range.lower.z,
                  range.upper.x, range.upper.y, range.upper.z);
  
  n = (WEFG_node_t *) lobs_3d_first_overlapping_object (se->efg_node_ss);
  while (n) {
    if (n != en) { /* not self */
      efgn_ref_coords (n, &pt.x, &pt.y, &pt.z);
      VECMAC_3_VEC(d, =, enpt, -, pt);
      if (VECMAC_NORM(d) < mindist) {
        nn = n;
        mindist = min(mindist, VECMAC_NORM(d));
      }
    }
    n = (WEFG_node_t *) lobs_3d_next_overlapping_object (se->efg_node_ss);
  } 
  
  return mindist;
} 


/*
   Add a node.
   Add a node to the superelement.
 */
WEFG_node_t * /* created and attached node */
efg_add_node (WEFG_superelem_t *se /* EFG superelement */,
              W_loc_t *loc /* location */
              ) 
{
  WEFG_node_t *en;
  en = new_efg_node (se, loc,
                     efg_support_size_at (se, loc->x, loc->y, loc->z));
  dict_add_obj_int_name (se->efgnodes, en, en->id); 
  add_en_to_ss (se, en);
  return en;
} 

/*
   Attach a node.
   Attach a node to the superelement.
 */
void
efg_attach_node (WEFG_superelem_t *se /* EFG superelement */,
                 WEFG_node_t *en /* node */
                 ) 
{
  WEFG_node_t *n;

  n = (WEFG_node_t *) dict_find_obj_with_int_name (se->efgnodes, en->id);
  if (n == NULL) {
    dict_add_obj_int_name (se->efgnodes, en, en->id); 
    add_en_to_ss (se, en);
  } 
} 

/*
   Get next node.
   Get next node from the superelement.
 */
WEFG_node_t * /* node or NULL to indicate there is no more nodes */
efg_next_node (WEFG_superelem_t *se /* EFG superelement */,
               WEFG_node_t *en /* node; pass NULL to get the first */) 
{
  return (WEFG_node_t *) dict_next_obj (se->efgnodes, en);
} 

/*
   Return element code.
 */
static int /* code; unique integer among the element types */
elem_elcode (W_elem_t *el /* generic element */
             ) 
{
  return EFG_SUPERELEM_ELCODE;
} 

/* Return element type name
 */
static char *
type_name (W_elem_t *el /* generic element */) 
{
  static char buf[] = "efg_superelement";
  return buf;
} 


/*
   Assemble the restoring forces.
   Assemble the restoring forces and update the material state
   at the same time.
   Loop over all integration volumes, update the material state
   for all integration points, and compute the forces through which
   the element acts on its nodes.  Assemble those forces.
   Update mechanical state for all vertices of integration volumes,
   and for all crack surface vertices.
 */
static void
assemble_restoring_forces (W_elem_t *el,  /* generic element */
                           W_mat_state_update_form_enum update_form
                           ) 
{
  WEFG_superelem_t *se = (WEFG_superelem_t *) el->data;
  WEFG_hex_volume_t *h;
  WEFG_crack_surf_t *cs;

#ifdef USE_ESFLIB
  ESFL3_set_search_structure (se->esflib, (char *) se->efg_node_ss);
#else
  ESP_3_set_search_structure ((char *) se->efg_node_ss);
#endif
  
  /* Assemble the restoring forces to both the internal degrees of freedom,
     and to the coupling FE nodes.
   */
  LIST_LOOP_START (se->hex_volumes, h, WEFG_hex_volume_t *) {
    hex_assemble_restoring_forces (se, update_form, h);
  }  LIST_LOOP_END ();
  /*
    If required, update the stress state at the vertices of the hexahedra,
    and also at all vertices at the output points
   */
  if (se->compute_stress_at_output_pts) {
    WEFG_vertex_t *v;
    v = NULL;
    while ((v = (WEFG_vertex_t *) dict_next_obj (se->verts, v)) ) {
      efgv_update_stress (se, v);
    } 
  } 
  /*
    If there are cracks, update the stress state at the crack-face
    vertices.
   */
  if (se->cracks) {
    LIST_LOOP_START (se->cracks, cs, WEFG_crack_surf_t *) {
      efgcr_update_stress (se, cs); 
    }  LIST_LOOP_END ();
  } 
} 

/*
   Assemble the mass.
   Assemble the mass of the element to its nodes.
 */
static void
assemble_mass (W_elem_t *el /* generic element */) 
{
  WEFG_superelem_t *se = (WEFG_superelem_t *) el->data;
  recompute_mass (se);
} 

static double
smallest_distance_between_nodes (WEFG_superelem_t *se)
{
  double mindist = FLT_MAX;
  WEFG_node_t *efgn = 0;
  while ((efgn = (WEFG_node_t *) dict_next_obj (se->efgnodes, efgn)) ) {
    mindist = min( mindist, efg_closest_node_dist(se, efgn) );
  }
  return mindist;
}

/*
   Compute the suggested time step.
   The smallest time step recommended by the superelement
   is computed from the dilatational wave speed and the smallest distance
   between the nodes of this element.
 */
static W_float_t 
suggested_time_step (W_elem_t *el /* generic element */) 
{
  WEFG_superelem_t *se = (WEFG_superelem_t *) el->data;
  W_float_t E, nu;
  W_float_t rho, dilatation_wave_speed;
  double hmin = FLT_MAX;

#ifdef USE_ESFLIB
  ESFL3_set_search_structure (se->esflib, (char *) se->efg_node_ss);
#else
  ESP_3_set_search_structure ((char *) se->efg_node_ss);
#endif
  
  if (!WM3D_get_mat_const (se->mat, MAT_3D_YOUNG_MODULUS, &E)) 
    REPORT_FATAL_ERROR (se->domain->errh, "No Young's modulus");
  if (!WM3D_get_mat_const (se->mat, MAT_3D_POISSON, &nu)) 
    REPORT_FATAL_ERROR (se->domain->errh, "No Poisson constant");
  rho      = WM3D_get_volume_mass (se->mat);
  /* Speed of spherical wave -- highest eigenfrequency is associated */
  /* to uniform expansion */
  dilatation_wave_speed = W_3D_dilatational_wave_speed (E, nu, rho);
  
  hmin = smallest_distance_between_nodes (se);
  
  return (hmin / dilatation_wave_speed);
} 


#ifdef X_XFACE
/*
   Modify the graphic representation of this element.
 */
static void
modify_elem_grep (W_elem_t *el, /* generic element */
                  double displ_scale, 
                  W_bool_t immed_draw /* ignored */
                  ) 
{
  WEFG_superelem_t *se = (WEFG_superelem_t *) el->data;
  WEFG_hex_volume_t *h;
  WEFG_node_t *n;
  WEFG_crack_surf_t *cs;
  
  LIST_LOOP_START (se->hex_volumes, h, WEFG_hex_volume_t *) {
    hex_modify_elem_grep (se, displ_scale, h);
  }  LIST_LOOP_END ();
  n = NULL;
  while ((n = (WEFG_node_t *) dict_next_obj (se->efgnodes, n)) ) {
    efgn_modify_grep (se, displ_scale, n);
  } 
  if (se->cracks) {
    LIST_LOOP_START (se->cracks, cs, WEFG_crack_surf_t *) {
      efgcr_modify_crack_surf_grep (se, displ_scale, cs); 
    }  LIST_LOOP_END ();
  } 
} 
#endif




/*
   Compute the total mass.
   Compute the total mass of the element
 */
static W_float_t
total_mass (W_elem_t *el /* generic element */
            ) 
{
  WEFG_superelem_t *se = (WEFG_superelem_t *) el->data;
  return get_total_mass (se);
} 


/*
   Compute the effective nodal loads.
   Compute the effective nodal loads by which the element loadings (body loads,
   concentrated forces, surface loads, penalty contact loads, damping) act
   on the nodes of the element.
 */
static void 
calc_eff_loads (W_elem_t *el /* generic element */
                ) 
{
  WEFG_superelem_t *se = (WEFG_superelem_t *) el->data;
  WEFG_distrib_load_t *dl;
  WEFG_contact_load_t *ctl;
  WEFG_conc_load_t *cl;
  WEFG_hex_volume_t *h;

  assert (se->effective_loads.v);

  if (se->body_load.is_set && se->body_load.is_time_independent) {
    EFGVDO (se,  { se->effective_loads.v[IX]
                   = se->body_load.time_independent_body_load.v[IX];
               } );
  }  else  {
    EFGVDO (se,  { se->effective_loads.v[IX] = 0; } );
  } 

  /* Body load */
  if (se->body_load.is_set && (!se->body_load.is_time_independent)) {
    LIST_LOOP_START (se->hex_volumes, h, WEFG_hex_volume_t *) {
      hex_calc_eff_loads (se, h);
    }  LIST_LOOP_END ();
  } 
  
  /* Concentrated loads */
  LIST_LOOP_START (se->conc_loads, cl, WEFG_conc_load_t *) {
    conc_load_assemble (se, cl);
  }  LIST_LOOP_END ();
  
  /* Distributed loads */
  LIST_LOOP_START (se->distrib_loads, dl, WEFG_distrib_load_t *) {
    distrib_load_assemble (se, dl);
  }  LIST_LOOP_END ();

  /* Contact loads */
  LIST_LOOP_START (se->contact_loads, ctl, WEFG_contact_load_t *) {
    contact_load_assemble (se, ctl);
  }  LIST_LOOP_END ();

  /* Damping */
  if (se->mass_damping_coeff != 0) {
    WEFG_node_t *n;
    W_float_t dt = WD_time_step (se->domain);
    double damping_coeff
      = se->mass_damping_coeff * WD_mass_damping_coeff_time_dep (se->domain);
    n = NULL;
    while ((n = (WEFG_node_t *) dict_next_obj (se->efgnodes, n)) ) {
      efgn_assemble_damping (n, se, damping_coeff, dt);
    } 
  } 

} 



/*
   Solve for the update vector.
   Solve for the update vector from the equations of motion for the internal degrees
   of freedom.
 */
static void 
solve_for (W_elem_t *el /* generic element */) 
{
  WEFG_superelem_t *se = (WEFG_superelem_t *) el->data;

  if (dict_count (se->driven_fen_conn) > 0) {
    WEFG_vertex_t *v = 0;
    while (v = dict_next_obj (se->verts, v)) {
      if (v->efgn && ((EFGN_IS_TRUE_EFG(v->efgn) || EFGN_IS_EFG_IF(v->efgn))
                      && (v->efgn->fen != 0)))
        efgv_assemble_driven_fen_force (v, se);
    }
  }

  W_vector_t u = se->update_vector.v;
  W_vector_t l = se->effective_loads.v;
  W_vector_t m = se->mass_matrix.v;

  EFGVDO (se, u[IX] = l[IX] / m[IX]);
} 

/*
   Update the configuration.
   Update the configuration of the element nodes for the internal degrees of freedom.
   Warburton variant of the central differences.
 */
static void 
update_Warburton (W_elem_t *el /* generic element */) 
{
  WEFG_superelem_t *se = (WEFG_superelem_t *) el->data;
  WEFG_node_t *n;
  W_float_t dt = WD_time_step  (se->domain);

  n = NULL;
  while ((n = (WEFG_node_t *) dict_next_obj (se->efgnodes, n)) ) {
    efgn_Warburton_update (n, se, dt);
  } 
} 

/*
   Update the configuration .
   Update the configuration of the element nodes for the internal degrees of freedom.
   Newmark variant of the central differences.
 */
static void 
update_Newmark (W_elem_t *el /* generic element */) 
{
  WEFG_superelem_t *se = (WEFG_superelem_t *) el->data;
  WEFG_node_t *n;
  W_float_t dt = WD_time_step  (se->domain);

  n = NULL;
  while ((n = (WEFG_node_t *) dict_next_obj (se->efgnodes, n)) ) {
    efgn_Newmark_update (n, se, dt);
  }

  if (dict_count(se->driven_fen_conn) > 0) {
    WEFG_vertex_t *v;
    v = NULL;
    while ((v = (WEFG_vertex_t *) dict_next_obj (se->verts, v)) ) {
      efgv_Newmark_update (v, se);
    } 
  } 
} 

/*
   Any vertex in a box?
   Is any vertex present in the given rectangular box?
 */
WEFG_vertex_t * /* vertex or NULL */
efg_vertex_in_range (WEFG_superelem_t *se /* EFG superelement */,
                     G3D_box_t *range /* rectangular box */
                     ) 
{
  WEFG_vertex_t *v;
  
  pirat_search_range (se->verts_ss, range, INT_MAX);
  
  v = (WEFG_vertex_t *) pirat_first_in_range (se->verts_ss);
  if (v) {
    if (pirat_next_in_range (se->verts_ss)) {
      REPORT_WARNING (se->domain->errh, "two or more vertices at a location!?");
    } 
  } 
  return v;
} 

/*
   Report warning if something bad happened during searching
   for something.
 */
void
bad_search (char *ss, G3D_box_t *range) 
{
  REPORT_WARNING (NULL, "bad search");
} 


/*
   Attach vertex to a superelement.
 */
void
efg_attach_vertex (WEFG_superelem_t *se /* EFG superelement */,
                   WEFG_vertex_t *v /* vertex */
                   ) 
{
  dict_add_obj_int_name (se->verts, v, v->id);
  pirat_add_pt (se->verts_ss, v, &v->loc);
} 



 /* apply initial conditions.
    apply initial conditions on all EFG nodes of the superelement */
static void
apply_initial_conditions (W_elem_t *el /* generic element */
                          ) 
{
  WEFG_superelem_t *se = (WEFG_superelem_t *) el->data;
  apply_init_cond (se);
} 


/*
   Write a post-processing file.
   
 */
static void
write_postproc_file (W_elem_t *el /* generic element */,
                    ckit_iobuf_t IOB, /* I/O buffer */
                    int type, /* type of output */
                    char *mode, /* mode of output */
                    void *optdata /* optional data interpreted by the
                                     element
                                   */
                     ) 
{
  WEFG_superelem_t *se = (WEFG_superelem_t *) el->data;
  WEFG_hex_volume_t *h;
  WEFG_crack_surf_t *cs;

  if (type == HEXVUE_POSTPROC_FILE) {
    LIST_LOOP_START (se->hex_volumes, h, WEFG_hex_volume_t *) {
      hex_write_hexvue_postproc_file (se, h, IOB, mode);
    }  LIST_LOOP_END ();
    if (se->cracks) {
      LIST_LOOP_START (se->cracks, cs, WEFG_crack_surf_t *) {
        efgcr_write_hexvue_postproc_file (cs, IOB, mode);
      }  LIST_LOOP_END ();
    } 
  } else if (type == HEXVUE_XTRA_POSTPROC_FILE) {
    LIST_LOOP_START (se->hex_volumes, h, WEFG_hex_volume_t *) {
      hex_write_hexvue_XTRA_postproc_file (se, h, IOB, mode);
    }  LIST_LOOP_END ();
    /*
    if (se->cracks) {
      LIST_LOOP_START (se->cracks, cs, WEFG_crack_surf_t *) {
        efgcr_write_hexvue_postproc_file (cs, IOB, mode);
      }  LIST_LOOP_END ();
      }
    */
  }  else if (type == VE3D_POSTPROC_FILE) {
    efg_write_ve3d (el, IOB);
  }  else if (type == TECPLOT_I_POSTPROC_FILE) {
    efg_write_tecplot (el, IOB, optdata);
  } 
} 


/* attach a symmetry plane.  attach a symmetry plane specification to a superelement */
void
attach_symm_plane (WEFG_superelem_t *se /* EFG superelement */,
                   WEFG_symm_plane_t *sp /* symmetry plane */
                   ) 
{
  if (se->symm_planes == NULL) {
    se->symm_planes = make_list ();
  } 
  add_to_tail (se->symm_planes, sp);
} 


/*
   Find integration volume enclosing point.
   Find the integration volume enclosing a given point.
 */
WEFG_hex_volume_t *             /* the enclosing hex or NULL */
efg_find_enclosing_hex
 (
 WEFG_superelem_t *se /* EFG superelement */, 
 W_loc_t *loc,         /* location */
 double eps,           /* tolerance in the parametric <-1;1>^3 space */
 W_loc_t *param_loc    /* returns the parametric location inside the hex if the function returned non-NULL value */
 ) 
{
  WEFG_hex_volume_t *h;
  WEFG_hex_volume_t *found_h = NULL;
  G3D_box_t locbox;

  G3D_INIT_BOX (locbox, (*loc));
  G3D_INFLATE_BOX (locbox, se->max_support_size / 10);
  
  lobs_3d_search (se->hex_ss,
                  locbox.lower.x, locbox.lower.y, locbox.lower.z,
                  locbox.upper.x, locbox.upper.y, locbox.upper.z);
  
  for (h = (WEFG_hex_volume_t *) lobs_3d_first_overlapping_object (se->hex_ss);
       h;
       h = (WEFG_hex_volume_t *) lobs_3d_next_overlapping_object (se->hex_ss) 
       ) {
    if ((found_h = hex_loc_is_inside (h, loc, eps, param_loc)) ) 
      return found_h;
  } 
  return NULL;
} 



static int
count_efgnodes (WEFG_superelem_t *se);
static int
count_vertices (WEFG_superelem_t *se);

/*
   Print element info.
 */
static void
print_info (W_elem_t *el, /* generic element */
            W_info_spec_t spec, /* ignored */
            FILE *fp, /* output stream */
            int indent /* initial indentation */
            ) 
{
  WEFG_superelem_t *se = (WEFG_superelem_t *) el->data;

  indent += 3;
  W_INDENT (fp, indent);
  fprintf (fp, "General constraints x: %s, y: %s, z: %s\n",
          se->general_constraint[0] == FREE_DOF
          ? "FREE"
          : "FIXED",
          se->general_constraint[1] == FREE_DOF
          ? "FREE"
          : "FIXED",
          se->general_constraint[2] == FREE_DOF
          ? "FREE"
          : "FIXED");
  W_INDENT (fp, indent);
  fprintf (fp, "Prescribed node spacing = %g\n", se->node_spacing);
  W_INDENT (fp, indent);
  fprintf (fp, "Prescribed merging distance = %g\n", se->merging_distance);
  W_INDENT (fp, indent);
  fprintf (fp, "Prescribed # nodes per hex side = %d\n",
          se->num_nodes_per_hex_side);
  W_INDENT (fp, indent);
  fprintf (fp, "Support sizes are %s\n",
          se->compute_support_size
          ? "COMPUTED"
          : "READ FROM INPUT");
  W_INDENT (fp, indent);
  fprintf (fp, "Prescribed support size multiplier = %g\n",
          se->support_size_multiplier);
  W_INDENT (fp, indent);
  fprintf (fp, "Prescribed support size = %g\n", se->support_size);
  W_INDENT (fp, indent);
  fprintf (fp, "Prescribed max support size = %g\n", se->max_support_size);
  W_INDENT (fp, indent);
  fprintf (fp, "Prescribed min # neighbors = %d\n", se->num_neighbors_min);
  W_INDENT (fp, indent);
  fprintf (fp, "Prescribed suggested # neighbors = %d\n", 
          se->num_neighbors_suggested);
  W_INDENT (fp, indent);
  fprintf (fp, "Prescribed quadrature order = %d\n", se->quadrature_order);
# ifdef ENABLE_TWO_POINT_RULE
  W_INDENT (fp, indent);
  fprintf (fp, "2-point quadrature is %s\n",
          se->enable_2pt_quadrature ? "ENABLED" : "DISABLED");
# endif
  W_INDENT (fp, indent);
  fprintf (fp, "Prescribed vertex search-structure bucket size = %d\n", 
          se->vertex_ss_bucket_size);
  W_INDENT (fp, indent);
  fprintf (fp, "# of hex volumes    = %d\n", list_length (se->hex_volumes));
  W_INDENT (fp, indent);
  fprintf (fp, "# of cracks         = %d\n", list_length (se->cracks));
  W_INDENT (fp, indent);
  fprintf (fp, "# of boundary faces = %d\n", list_length (se->b_faces));
  W_INDENT (fp, indent);
  fprintf (fp, "Element region is   = %s\n",
 (se->is_convex
           ? "CONVEX"
           : (se->is_convex_with_cracks
              ? "CONVEX WITH CRACKS"
              : "NON-CONVEX")) );
  W_INDENT (fp, indent);
  fprintf (fp, "Nodes are generated %s\n",
 (se->generate_nodes_inside_only
           ? "INSIDE ONLY"
           : "without restrictions"));
  fprintf (fp, "Nodes are generated %s\n",
 (se->generate_nodes_at_vertices
           ? "AT VERTICES"
           : ": not requested at the vertices"));
  if (se->generate_nodes_in_volume) {
    W_INDENT (fp, indent);
    fprintf (fp, "Nodes are generated in a volume \n");
    W_INDENT (fp, indent);
    fprintf (fp, "    box =  { {%g, %g, %g} ,  {%g, %g, %g} } \n",
            se->lower_corner.x, se->lower_corner.y, se->lower_corner.z,
            se->upper_corner.x, se->upper_corner.y, se->upper_corner.z);
  } 
  W_INDENT (fp, indent);
  fprintf (fp, "Prescribed node search-structure bucket size %d\n", 
          se->efg_node_ss_bucket_size);
  W_INDENT (fp, indent);
  fprintf (fp, "Body load is %s\n",
          se->body_load.is_set
          ? "SET"
          : "NOT SET");
  if (se->body_load.is_set) {
    W_INDENT (fp, indent);
    fprintf (fp, "Body load is %s\n",
          se->body_load.is_time_independent
          ? "TIME-DEPENDENT"
          : "TIME-INDEPENDENT");
    W_INDENT (fp, indent);
    fprintf (fp, "Body load is %s\n",
            se->body_load.is_constant
            ? "CONSTANT IN SPACE"
            : "VARIABLE IN SPACE");
    W_INDENT (fp, indent);
    fprintf (fp, "Body load expressions\n");
    W_INDENT (fp, indent);
    fprintf (fp, "   x: %s\n", se->body_load.body_load_expr[0]);
    W_INDENT (fp, indent);
    fprintf (fp, "   y: %s\n", se->body_load.body_load_expr[1]);
    W_INDENT (fp, indent);
    fprintf (fp, "   z: %s\n", se->body_load.body_load_expr[2]);
  } 
  W_INDENT (fp, indent);
  fprintf (fp, "Material type = %d\n", WM3D_get_mcode (se->mat));
  W_INDENT (fp, indent);
  fprintf (fp, "Mass-proportional damping coefficient = %g\n",
          se->mass_damping_coeff);
  W_INDENT (fp, indent);
  fprintf (fp, "Shortest distance between nodes = %g\n", se->shortest_distance);
  W_INDENT (fp, indent);
  fprintf (fp, "# of vertices           = %d\n", count_vertices (se));
  W_INDENT (fp, indent);
  fprintf (fp, "# of nodes              = %d\n", count_efgnodes (se));
  W_INDENT (fp, indent);
  fprintf (fp, "# of equations          = %ld\n", se->num_of_equations);
  W_INDENT (fp, indent);
  fprintf (fp, "# of concentrated loads = %d\n",
          list_length (se->conc_loads)); 
  W_INDENT (fp, indent);
  fprintf (fp, "# of distributed loads  = %d\n",
          list_length (se->distrib_loads)); 
  W_INDENT (fp, indent);
  fprintf (fp, "# of contact loads      = %d\n",
          list_length (se->contact_loads));
  
  W_INDENT (fp, indent);
  fprintf (fp, "Total mass = %g\n", se->total_mass);

   {
    double mass = 0;
    WEFG_node_t *en /* node */;
    
    en = NULL;
    while ((en = (WEFG_node_t *) dict_next_obj (se->efgnodes, en)) ) {
      if (EFGN_IS_TRUE_EFG (en) || EFGN_IS_EFG_IF (en)) {
        if (en->ndata.eq_numbers[W_DX] > 0) 
          mass += se->mass_matrix.v[en->ndata.eq_numbers[W_DX]];
        else if (en->ndata.eq_numbers[W_DY] > 0) 
          mass += se->mass_matrix.v[en->ndata.eq_numbers[W_DY]];
        else if (en->ndata.eq_numbers[W_DZ] > 0) 
          mass += se->mass_matrix.v[en->ndata.eq_numbers[W_DZ]];
      } 
    } 
    W_INDENT (fp, indent);
    fprintf (fp, "Mass of EFG nodes %g\n", mass);
  } 

  
  W_INDENT (fp, indent);
  fprintf (fp, "Tolerance = %g\n", se->tolerance);

  W_INDENT (fp, indent);
  fprintf (fp, "Stress is %s computed at vertices\n",
          se->compute_stress_at_output_pts
          ? ""
          : "NOT");

  W_INDENT (fp, indent);
  fprintf (fp, "Equiv. plastic strain is %s computed\n",
          se->compute_equiv_pl_def_at_output_pts
          ? ""
          : "NOT");
  W_INDENT (fp, indent);
  fprintf (fp, "Integration points are %s stored in a search structure\n",
          se->store_ip_in_search_structure
          ? ""
          : "NOT");


  W_INDENT (fp, indent);
#ifdef USE_ESFLIB
  fprintf (fp, "Using ESFLIB to construct shape functions\n");
#else
  fprintf (fp, "Using ESPACK to construct shape functions\n");
#endif
  W_INDENT (fp, indent);
  fprintf (fp, "Shape functions constructed as %s\n",
          se->use_wrap_around ? "C^0 using diffraction" : "C^-1");
  W_INDENT (fp, indent);
  fprintf (fp, "Basis order %s\n",
 (se->basis_order
           == WEFG_COMPLETE_LINEAR
           ? "LINEAR"
  : "QUADRATIC"));
  W_INDENT (fp, indent);
  fprintf (fp, "Weight function %s\n",
 (se->weight_func_type
           == WEFG_QUADRATIC_TENSOR_PROD_WEIGHT_FUNC
           ? "QUADRATIC TENSOR-PRODUCT"
           :
 (se->weight_func_type
           == WEFG_SMOOTH_QUADRATIC_TENSOR_PROD_WEIGHT_FUNC
           ? "QUADRATIC SMOOTH (C^1) TENSOR-PRODUCT"
           :
 (se->weight_func_type
           == WEFG_QUARTIC_SPLINE_SPHERICAL_WEIGHT_FUNC
           ? "QUARTIC SPHERICAL"
           :
 (se->weight_func_type
           == WEFG_GAUSSIAN_SPHERICAL_WEIGHT_FUNC
           ? "GAUSSIAN_SPHERICAL"
           :
 (se->weight_func_type
           == WEFG_QUADRATIC_SPHERICAL_WEIGHT_FUNC
           ? "QUADRATIC SPHERICAL"
           :
 (se->weight_func_type
           == WEFG_SMOOTH_QUADRATIC_SPHERICAL_WEIGHT_FUNC
           ? "SMOOTH QUADRATIC SPHERICAL"
           :
          "unknown!?"
           )) )) )) 
          );
  W_INDENT (fp, indent);
  fprintf (fp, "Shape functions invalid after time %g\n",
          se->invalid_after_time);
  EVAL_EFG_SF_STAT (se);  
  W_INDENT (fp, indent);
  fprintf (fp, "Shape func statistics:  # neighbors min=%d, max=%d, avg=%d\n",
 (int) se->efg_sf_stat.min_nneighbors,
 (int) se->efg_sf_stat.max_nneighbors,
 (int) se->efg_sf_stat.avg_nneighbors);

  W_INDENT (fp, indent);
  fprintf (fp, "EFG nodes\n");
   {
    WEFG_node_t *en /* node */;
    
    en = NULL;
    while ((en = (WEFG_node_t *) dict_next_obj (se->efgnodes, en)) ) {
      efgn_print_info (en, fp, indent+3, se);
    } 
  } 
} 

/* count nodes. count all nodes associated with the superelement */
static int
count_efgnodes (WEFG_superelem_t *se) 
{
  WEFG_node_t *en /* node */;
  int count = 0;
  
  en = NULL;
  while ((en = (WEFG_node_t *) dict_next_obj (se->efgnodes, en)) ) {
    count++;
  } 

  return count;
} 

/* count vertices. count all vertices associated with the superelement */
static int
count_vertices (WEFG_superelem_t *se) 
{
  WEFG_vertex_t *v;
  int count = 0;
  
  v = NULL;
  while ((v = (WEFG_vertex_t *) dict_next_obj (se->verts, v)) ) {
    count++;
  } 

  return count;
} 

 

static void
clear_invalid_regions (WEFG_superelem_t *se);


/*
   Finalize the time step.
   Clear regions of invalidated connectivities.
   Propagate any cracks present.  If any crack changed shape, recompute the mass matrix.
 */
static W_float_t /* suggested time step or zero if none suggested */
finalize_step (W_elem_t *el /* generic element */) 
{
  WEFG_superelem_t *se = (WEFG_superelem_t *) el->data;
  WEFG_crack_surf_t *cs;
  W_bool_t any_changed_shape = FALSE;
  double t;

  if (se->invalid_regions.regions) {
    clear_invalid_regions (se);
  } 
  if (se->cracks) {
    LIST_LOOP_START (se->cracks, cs, WEFG_crack_surf_t *) {
      cs->changed_shape = FALSE;
      efgcr_propagate (cs);
      any_changed_shape = (any_changed_shape || cs->changed_shape);
      if (WOPT_OPT (se->domain->options, j_dom_eval)
          == W_PER_CELL_IPOL_PTS && cs->changed_shape) {
        WEFG_hex_volume_t *h;
        LIST_LOOP_START (se->hex_volumes, h, WEFG_hex_volume_t *) {
          if (!h->is_intersected_by_crack_surf) 
            h->is_intersected_by_crack_surf
              = efgcr_hex_intersected_by_crack_surf (cs, h);
        }  LIST_LOOP_END ();
      }  /* if (options.j_dom_eval ... */
    }  LIST_LOOP_END ();
    t  = WD_time (se->domain);
    if (any_changed_shape
        /*  && se->invalid_after_time > t+dt WHY THIS CONDITION? */
        ) {
      se->internal_mass_only = TRUE;
      if (!se->time_independent_mass) 
        recompute_mass (se);
      se->internal_mass_only = FALSE;
      se->invalid_after_time = t;
    }  
  } 

  return 0; /* ignore this suggestion of the new time step */
} 

/* clear all regions. clear all regions of invalidated connectivities, ie. undefine
   all of the regions
 */
static void
clear_invalid_regions (WEFG_superelem_t *se) 
{
  WEFG_region_t *r;

  LIST_LOOP_START (se->invalid_regions.regions, r, WEFG_region_t *) {
    r->is_initialized = FALSE;
  }  LIST_LOOP_END ();
  se->invalid_regions.triangles_are_reliable = TRUE;
} 


/* add a region.  add a region of invalidated connectivities */
void
efg_add_invalid_region (WEFG_superelem_t *se /* EFG superelement */,
                        int ntriangles, /* number of triangles in the below array */
                        G3D_tri_t *triangles /* triangles in the region */
                        ) 
{
  WEFG_region_t *r;
  int i;
  double s;
  double t, dt;
#define SET_UP_BBOX(b, ntriangles, triangles) \
     {                                                   \
      G3D_INIT_BOX (b, triangles[0].loc[0]);            \
      switch (ntriangles) {                             \
      case 1:                                           \
        G3D_UPDT_BOX (b,  triangles[0].loc[1]);          \
        G3D_UPDT_BOX (b,  triangles[0].loc[2]);          \
        break;                                          \
      case 2:                                           \
        G3D_UPDT_BOX (b,  triangles[0].loc[1]);          \
        G3D_UPDT_BOX (b,  triangles[0].loc[2]);          \
        G3D_UPDT_BOX (b,  triangles[1].loc[0]);          \
        G3D_UPDT_BOX (b,  triangles[1].loc[1]);          \
        G3D_UPDT_BOX (b,  triangles[1].loc[2]);          \
        break;                                          \
      }                                                  \
    } 

  if (se->invalid_regions.regions) {
    LIST_LOOP_START (se->invalid_regions.regions, r, WEFG_region_t *) {
      if (!r->is_initialized) {
        goto have_free_region;
      } 
    }  LIST_LOOP_END ();
  }  else  {
    se->invalid_regions.regions = make_list ();
  } 

  r = w_new (WEFG_region_t); CHECK_MALLOC (r);
  add_to_tail (se->invalid_regions.regions, r);
  
 have_free_region:
  
  r->is_initialized      = TRUE;
  SET_UP_BBOX (r->bbox, ntriangles, triangles);
  G3D_INFLATE_BOX (r->bbox, se->max_support_size);
  /* can handle only one or two triangles */
  if (ntriangles > 2) {
    ntriangles = 2;
    REPORT_WARNING (se->domain->errh, "Too many triangles in efg_add_invalid_region () ");
  } 
  /* initialize triangles stored with the region */
  r->triangles[0].is_valid = r->triangles[1].is_valid = FALSE;
  FOR (i, ntriangles) {
    r->triangles[i].is_valid = TRUE;
    r->triangles[i].tri      = triangles[i];
    /* set up the bounding box of the triangle */
    G3D_INIT_BOX (r->triangles[i].bbox, triangles[i].loc[0]);
    G3D_UPDT_BOX (r->triangles[i].bbox, triangles[i].loc[1]);
    G3D_UPDT_BOX (r->triangles[i].bbox, triangles[i].loc[2]);
    s = (r->triangles[i].bbox.upper.x - r->triangles[i].bbox.lower.x +
         r->triangles[i].bbox.upper.y - r->triangles[i].bbox.lower.y +
         r->triangles[i].bbox.upper.z - r->triangles[i].bbox.lower.z);
    s *= 0.001;
    G3D_INFLATE_BOX (r->triangles[i].bbox, s); /* make it larger to go sure */
    /* and cache the polygon plane */
    G3D_plane_consts_3 (&r->triangles[i].plane,
                        &triangles[i].loc[0],
                        &triangles[i].loc[1],
                        &triangles[i].loc[2]);
  } 
  
  dt = WD_time_step (se->domain);
  t  = WD_time (se->domain) - 0.01 * dt;
  se->invalid_after_time = t;
  return;
} 


static int
any_seg_severed (W_loc_t *pt, WEFG_sf_tab_t *sftab,
                 WEFG_region_tri_t *triangles);

/*
   Check if point affected by region of invalidated connectivities.
   Check if the given point is inside any region of invalidated connectivities.
   If full check is desired,
   compute intersections of the rays
   connecting the point with the nodes with triangles in
   the regions of invalidated connectivities; otherwise presence of the location
   within a region of invalidated connectivities is sufficient.

   If the superelement has been configured to force recomputation of *ALL*
   connectivities, TRUE is returned immediately.
 */
W_bool_t
efg_pt_is_inside_invalid_region (WEFG_superelem_t *se /* EFG superelement */,
                                 W_loc_t *pt, /* location */
                                 WEFG_sf_tab_t *sftab, /* shape function table
                                                          at the location
                                                        */
                                 W_bool_t full_check /* perform full check if TRUE
                                                      */
                                 ) 
{
  WEFG_region_t *r;

  if (se->force_all_sf) {
    return TRUE;
  } 
 
  if (se->invalid_regions.regions) {
    LIST_LOOP_START (se->invalid_regions.regions, r, WEFG_region_t *) {
      if (r->is_initialized) {
        if (G3D_PT_INSIDE_BOX (r->bbox, (*pt)) ) {
          if (full_check && se->invalid_regions.triangles_are_reliable) {
            if (any_seg_severed (pt, sftab, r->triangles)) 
              return TRUE;
          }  else  {
            return TRUE;
          } 
        } 
      }  
    }  LIST_LOOP_END ();
  }  
  return FALSE;
#undef INSIDE_BBOX
} 


/*
   Indicate reliability of triangles in regions.
   Set flag indicating reliability of the triangles in regions of invalidated
   connectivities.
 */
void
efg_invalid_region_triangles_reliable (WEFG_superelem_t *se /* EFG superelement */,
                                       W_bool_t YES_OR_NO /* if TRUE, triangles
                                                             inside the regions
                                                             can be relied upon.
                                                           */
                                       ) 
{
  se->invalid_regions.triangles_are_reliable = YES_OR_NO;
} 


/*
   Any ray severed in region?
   Is any ray in the shape function table severed by any triangle in the given
   region of invalidated
   connectivities?
 */
static int /* TRUE if any ray severed; FALSE otherwise */
any_seg_severed (W_loc_t *pt, /* point at which the shape function table is evaluated */
                 WEFG_sf_tab_t *sftab, /* shape function table */
                 WEFG_region_tri_t *triangles /* triangles in the region */
                 ) 
{
  int i, k, n = sftab->num_of_entries;
  G3D_loc_t a, b;
  G3D_vect_t ab;
  G3D_loc_t s;
  G3D_bool_t parallel;
  double lambda;
  double eps = 0;
  WEFG_node_t *efgn;
#define INSIDE_BBOX(bbox, pt) \
 ((pt.x >= (bbox) .lower.x && pt.x <= (bbox) .upper.x) \
    || (pt.y >= (bbox) .lower.y && pt.y <= (bbox) .upper.y) \
    || (pt.z >= (bbox) .lower.z && pt.z <= (bbox) .upper.z)) 
    
  VECMAC_2_VEC (a, =, (*pt));
  
  FOR (i, n) {
    efgn = SFTAB_NODE ((*sftab), i);
    EFGN_CURR_COORDS (efgn, b.x, b.y, b.z);
    VECMAC_3_VEC (ab, =, b, -, a);
    FOR (k, 2) {
      if (triangles[k].is_valid) {
        G3D_intersect_plane_and_line (&s, &lambda, &parallel, &a, &ab,
                                      &triangles[k].plane, 0, FLT_MAX);
        if (INSIDE_BBOX (triangles[k].bbox, s)) {
          if (G3D_intersect_tri_and_seg (&s, &lambda, &a, &ab,
                                         &triangles[k].tri, eps)) 
            return TRUE;
        } 
      } 
    } 
  } 
  return FALSE;
} 


/*
   Find intersection of segment with boundary face.
   Find intersection of given segment with a boundary face.
 */
WEFG_b_face_t * /* boundary face or NULL */
efg_find_intersection_with_bface (WEFG_superelem_t *se /* EFG superelement */,
                                  G3D_segment_t *s, /* input segment */
                                  double lower_lambda, /* lower limit of acceptable
                                                          range of the segment parameterization
                                                          (in the range 0<= l <=1)
                                                        */
                                  double upper_lambda, /* upper limit of acceptable
                                                          range of the segment parameterization
                                                          (in the range 0<= l <=1)
                                                        */
                                  W_loc_t *intersection, /* output intersection point */
                                  WEFG_which_b_face_enum which /* which face should
                                                                  be picked if there is
                                                                  more than one intersection?
                                                                */
                                  ) 
{
  WEFG_b_face_t *f;
  double lambda;
  double d, dist = FLT_MAX;
  W_loc_t xs, xsection =  {0,0,0} ;
  WEFG_b_face_t *have_xs = NULL;

  /* RAW speed up by using a bounding box tree for the
     boundary faces
   */
  LIST_LOOP_START (se->b_faces, f, WEFG_b_face_t *) {
    if (efggc_segment_intersection_with_face (s, f, TRUE, &xs, &lambda)) {
      if (lambda >= lower_lambda && lambda <= upper_lambda) {
        if (which == WEFG_ANY_B_FACE) {
          return f;
        }  else if (which == WEFG_B_FACE_CLOSEST_TO_START) {
          d = v3_distance (&s->start, &xs);
          if (d < dist) {
            dist = d;
            xsection = xs;
            have_xs = f;
          } 
        }  else if (which == WEFG_B_FACE_CLOSEST_TO_END) {
          d = v3_distance (&s->end, &xs);
          if (d < dist) {
            dist = d;
            xsection = xs;
            have_xs = f;
          } 
        } 
      } 
    }  /* if (efggc_segment_intersection_with_face... */
  }  LIST_LOOP_END ();
  *intersection = xsection;
  return have_xs;
} 



/*
   Add an integration point.
   Add an integration point to a search structure.
 */
void
efg_add_ip_to_search_structure (WEFG_superelem_t          *se, /* superelement */
                                W_loc_t                   *ip_loc, /* location */
                                WEFG_hex_volume_t         *enclosing_hex, /* integration volume in which this integration point is located */
                                WEFG_integration_point_t  *ip /* integration point */
                                ) 
{
  if (se->store_ip_in_search_structure) {
    /* RAW need to store somewhere the pointers on the location
       records; otherwise this could potentially constitute a memory leak. */
    WEFG_ip_ref_package_t *ippkg;
    
    ippkg                = w_new (WEFG_ip_ref_package_t); CHECK_MALLOC (ippkg);
    ippkg->ip            = ip;
    ippkg->loc           = *ip_loc;
    ippkg->enclosing_hex = enclosing_hex;
    pirat_add_pt (se->ip_ss, ippkg, &ippkg->loc);
  } 
} 


/*
   enable storage of integration points.
   Set a flag to enable storage of integration points in a search structure.
 */
void
efg_enable_ip_storage_to_search_structure (WEFG_superelem_t *se) 
{
  se->store_ip_in_search_structure = TRUE;
} 



/*
   Compute average stress.
   Compute average stress tensor in a given box.
   This is done for searching for all quadrature points in
   a given box, and summing the stress tensor at those points.
 */
W_bool_t /* TRUE if OK; FALSE otherwise */
efg_average_stress_in_box (WEFG_superelem_t *se /* EFG superelement */,
                           G3D_box_t *box, /* rectangular region */
                           W_mat_3d_stress_type_enum requested_out_type,
                           double stress[3][3] /* output stress */
                           ) 
{
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  double P[3][3];
  W_mat_3d_stress_type_enum out_type;
  int nip = 0;
  double one_over_num_of_ip;
  WEFG_ip_ref_package_t *ippkg;

  MAT3DUTIL_ZERO_3X3 (stress);
  
  ip_ctx_data.se       = se;
  ip_ctx.time          = WD_time (se->domain);
  ip_ctx.time_step     = WD_time_step (se->domain);
  ip_ctx.ctx_data      = &ip_ctx_data;
  ip_ctx.get_kinem_var = efgss_get_kinem_var;

  pirat_search_range (se->ip_ss, box, INT_MAX);
  
  ippkg = (WEFG_ip_ref_package_t *) pirat_first_in_range (se->ip_ss);
  while (ippkg) {
    ip_ctx_data.sftab    = ippkg->ip->sftab;
    EFGSS_AT_NEW_IP ();
    if (!WM3D_update_stress (se->mat, ippkg->ip->matstate, W_MAT_STATE_UPDATE_TEMP, 
			 &ip_ctx, requested_out_type, P,
			 &out_type)) 
      REPORT_ERROR (se->domain->errh, "Failed to compute stress");
    if (out_type != requested_out_type) {
      REPORT_ERROR (se->domain->errh, "types do not match");
      return FALSE;
    } 
    MAT3DUTIL_INCR_A_BY_B_3X3 (stress, P);
    nip++;
    ippkg = (WEFG_ip_ref_package_t *) pirat_next_in_range (se->ip_ss);
  } 

  if (nip == 0) /* No integration point inside the volume */
    return FALSE;
  
  one_over_num_of_ip = 1.0 / nip;
  MAT3DUTIL_SCALE_3X3 (stress, one_over_num_of_ip);
  return TRUE;
} 



/*
   Interpolate material state.
 Interpolate material state at a given point.
*/
W_mat_3d_state_p /* new material state */
efg_interpolate_mat_state (WEFG_superelem_t *se /* EFG superelement */,
                           W_loc_t *at_point,
                           double initial_box_size /* hint of the size of the box
                                                      in which quadrature points
                                                      are interpolated from
                                                    */
                           ) 
{
  int nip = 0;
  WEFG_ip_ref_package_t *ippkg;
  G3D_box_t range;
  double bs = initial_box_size;
#define MAX_PKGS 100
  WM3D_mat_state_pkg_t mspkgs[MAX_PKGS];
  int num_of_restarts = 0;
  double w;

restart:
  G3D_INIT_BOX (range, *at_point);
  G3D_INFLATE_BOX (range, bs);
  if (! pirat_search_range (se->ip_ss, &range, INT_MAX)) {
    REPORT_ERROR (se->domain->errh, "Search failed");
    return NULL; 
  } 
  
  ippkg = (WEFG_ip_ref_package_t *) pirat_first_in_range (se->ip_ss);
  nip   = 0;
  while (ippkg) {
    if (! (nip < MAX_PKGS)) 
      break;
    if (ippkg->ip->matstate) {
      mspkgs[nip].ms = ippkg->ip->matstate;
      w = v3_distance (&ippkg->loc, at_point);
      mspkgs[nip].weight = (w != 0? 1/w : FLT_MAX);
      nip++;
    } 
    ippkg = (WEFG_ip_ref_package_t *) pirat_next_in_range (se->ip_ss);
  } 

  if (nip == 0) {/* No integration point inside the volume */
    num_of_restarts++;
    bs *= 1.5;
    if (num_of_restarts > 5) {
      REPORT_ERROR (se->domain->errh, "No integration points available for interpolation");
      return NULL;
    } 
    goto restart;
  } 
    
  return WM3D_interpolate_mat_state (se->mat, at_point, nip, mspkgs);
} 




/*
   Configure superelement.
   Configure superelement by passing command line.
   The superelement parses the command line and if one of the options
   is recognized, configures itself accordingly.
 */
static W_bool_t
configure (W_elem_t *el /* generic element */, char *option) 
{
  WEFG_superelem_t *se = (WEFG_superelem_t *) el->data;
  W_domain_t *d = el->domain;
  
  TOKENS_PARSE (option);
  if (TOKENS_TOTAL () < 1) 
    return FALSE;
  
  if (TOKENS_HAVE_KEYWORD ("FORCE_ALL_SF")) {
    if (TOKENS_HAVE_KEYWORD ("FALSE")) 
      se->force_all_sf = FALSE;
    else
      se->force_all_sf = TRUE;
    return TRUE;
  }  else if (TOKENS_HAVE_KEYWORD ("DRAW_RAYS")) {
    if (TOKENS_HAVE_KEYWORD ("FALSE")) 
      se->draw_rays = FALSE;
    else
      se->draw_rays = TRUE;
    return TRUE;
  }  else if (TOKENS_HAVE_KEYWORD ("RECOMPUTE_SFS")) {
    se->invalid_after_time = FLT_MAX;
    se->force_all_sf       = TRUE;
    REPORT_WARNING (se->domain->errh, "Recomputing all shape functions");
    return TRUE;
  }  else  {
    REPORT_WARNING (se->domain->errh, "Unknown configuration option");
    return FALSE;
  } 
} 



#ifdef USE_ESFLIB
static void 
W_esflib_error_handler (ESFLIB_lib_t lib) 
{
  /* do nothing; relegate all error processing to return code handling */
} 

/*
   handle error in ESFLIB.
   Handle an error which occurred in the ESFLIB library.
 */
int
efg_error_esflib (ESFLIB_lib_t lib, /* ESFLib library */
                  char *file, /* source file */
                  int line, /* line in source file */
                  double x, /* x-coordinate of point */
                  double y, /* y-coordinate of point */
                  double z, /* z-coordinate of point */
                  int kind /* severity of error in terms of WASP understanding
                              (fatal, ordinary error, warning)
                            */
                  ) 
{
  ESFLIB_error_enum_t en = ESFL3_status (lib);
  
  sprintf (W_buf (), "At [%g;%g;%g]\n%s\n", x, y, z,
           ESFL3_explain_error (lib));
  switch (en) {
  case ESFLIB_OK:
    /* why are we here at all? */
    break;
  default:
    if (en == ESFLIB_BAD_PREC) {
      if (ESFL3_prec_loss (lib) > 1.e-6) {
        WEH_error (NULL, file, line, kind, W_buf ());
      }  else  {
        WEH_error (NULL, file, line, E_K_WARNING, W_buf ());
      } 
    }  else  {
      WEH_error (NULL, file, line, kind, W_buf ());
    } 
    break;
  } 
  return 0;
} 
#else
int
efg_error_espack (char *file, int line, double x, double y, double z) 
{
  ESPACK_error_enum_t en =ESP_3_status ();
  
  sprintf (W_buf (), "At [%g;%g;%g]\n%s\n", x, y, z, ESP_3_explain_error ());
  switch (en) {
  case ESPACK_OK:
    /* why are we here at all? */
    break;
  default:
    if (en == ESPACK_BAD_PREC) {
      if (ESP_3_prec_loss () > 1.e-6) {
        WEH_error (file, line, E_K_ERROR, W_buf ());
      }  else  {
        WEH_error (file, line, E_K_WARNING, W_buf ());
      } 
    }  else  {
      WEH_error (file, line, E_K_ERROR, W_buf ());
    } 
    break;
  } 
  return 0;
} 
#endif

/*
Compute kinetic energy.
   Compute the kinetic energy associated with the internal degrees of freedom.
 */
static double
kinetic_energy (W_elem_t *el    /* generic element */,
                W_mass_type_enum mass_type /* mass type; ignored: EFG superelement recognizes diagonal mass only */
                ) 
{
  WEFG_superelem_t *se = (WEFG_superelem_t *) el->data;
  double ke = 0;
  WEFG_node_t *en /* node */;
  W_float_t time_step = WD_time_step (se->domain);
  
  en = NULL;
  while ((en = (WEFG_node_t *) dict_next_obj (se->efgnodes, en)) ) {
    ke += efgn_kinetic_energy (en, se, time_step);
  } 
  return ke;
}  



/*
   Compute deformation work.
   Compute the deformation work done by the superelement by looping over
   the integration volumes.
 */
static double
deformation_energy (W_elem_t *el /* generic element */) 
{
  WEFG_superelem_t *se = (WEFG_superelem_t *) el->data;
  double de = 0;
  WEFG_hex_volume_t *h;
  
  LIST_LOOP_START (se->hex_volumes, h, WEFG_hex_volume_t *) {
    de += hex_deform_energy (se, h);
  }  LIST_LOOP_END ();
  
  return de;
} 


static void 
displ_at (WEFG_superelem_t *se /* EFG superelement */,
          WEFG_hex_volume_t *h,
          W_loc_t *param_loc,
          W_loc_t *loc,
          double displ[3]);


/*
   Probe at point
   Probe mechanical state at a given point.
 */
static W_bool_t /* TRUE if ok; FALSE otherwise */
probe (W_elem_t *el, /* finite element (generic element
                        whose instantiation is the superelement) */
       void *options, /* options specified as a string with blank-separated tokens */
       W_loc_t *loc, /* location */
       unsigned long spec, /* specification */
       W_probe_package_t *result /* result is returned in this */
       ) 
{
  WEFG_superelem_t *se = (WEFG_superelem_t *) el->data;
  WEFG_hex_volume_t *h;
  W_loc_t param_loc;
  int interpolate = FALSE;
  WEFG_field_package_t field_pkg;
  
  h = efg_find_enclosing_hex (se, loc, 0.001, &param_loc);
  if (h == NULL) return FALSE;

  if (options) {
    if (strcmp (options, "interpolate") == 0) {
      interpolate = TRUE;
    } 
  } 

  switch (spec) {
  case W_PROBE_DISPL_MASK:
    if (interpolate) {
      field_pkg.value.vector[0] =
        field_pkg.value.vector[1] =
          field_pkg.value.vector[2] = 0;
      if (!efg_interpolate_at (se, loc, NULL, efg_update_displ_pkg,
			      &field_pkg)) {
	return FALSE;
      } 
      result->value.vector[0] = field_pkg.value.vector[0];
      result->value.vector[1] = field_pkg.value.vector[1];
      result->value.vector[2] = field_pkg.value.vector[2];
    }  else  {
      displ_at (se, h, &param_loc, loc, result->value.vector);
    } 
    break;
  case W_PROBE_STRESS_MASK:
    break;
  case W_PROBE_EQPLS_MASK:
    break;
  case W_PROBE_VON_MISES_MASK:
    break;
  default:
    return FALSE;
  } 
  return TRUE;
} 

/*
   Evaluate displacement.
   Evaluate displacement at a particular point inside the superelement.

   Assumptions:

   It is assumed the point is *inside* the superelement.
 */
static void 
displ_at (WEFG_superelem_t *se /* EFG superelement */,
          WEFG_hex_volume_t *h,
          W_loc_t *param_loc,
          W_loc_t *loc, /* location */
          double displ[3] /* displacement */
          ) 
{
  WEFG_if_data_t if_data;
  SFL_sf_table_t sf;
  WEFG_sf_tab_t sftab;
  WEFG_eval_for_packet_t efp;
  double  x, y, z;

  efp.se = se;
  efp.pt_wrt_crack_surf_classification = WEFG_POS_WRT_SURF_UNSPEC;

  sf = efgsf_sf_table (se);

  if (hex_is_interface_vol (h)) {
    if_data.h     = h;
    if_data.xi    = param_loc->x;
    if_data.eta   = param_loc->y;
    if_data.theta = param_loc->z;
    if (!SFL_shape_func_at (sf, loc->x, loc->y, loc->z,
			     &efp, (SFL_generic_ptr_t) &if_data)) 
      SFL_ERROR (sf, loc->x, loc->y, loc->z, E_K_WARNING);
  }  else  {
    if (!SFL_shape_func_at (sf, loc->x, loc->y, loc->z, &efp, NULL)) 
      SFL_ERROR (sf, loc->x, loc->y, loc->z, E_K_WARNING);
  } 
  INIT_SFTAB (sftab);
  efgsf_create_sftab (se, &sftab, sf, NULL, FALSE);
  efgsf_eval_curr_loc (&sftab, &x, &y, &z);
  displ[0] = x - loc->x;
  displ[1] = y - loc->y;
  displ[2] = z - loc->z;
} 


  
/*
   Update displacement package.
   Used for interpolation-based probe.
 */
W_bool_t 
efg_update_displ_pkg (void *option,
                      WEFG_superelem_t *se /* EFG superelement */,
                      WEFG_ip_ref_package_t *ippkg,
                      double weight,
                      WEFG_field_package_t *pkg) 
{
  double cx = 0, cy = 0, cz = 0;

  efgsf_eval_curr_loc (&ippkg->ip->sftab, &cx, &cy, &cz);

  pkg->value.vector[0] += weight * (cx - ippkg->loc.x);
  pkg->value.vector[1] += weight * (cy - ippkg->loc.y);
  pkg->value.vector[2] += weight * (cz - ippkg->loc.z);

  return TRUE;
} 



/*
   Update displacement gradient package.
   Used for interpolation-based probe.
 */
W_bool_t 
efg_update_displ_grad_pkg (void *options,
                           WEFG_superelem_t *se /* EFG superelement */,
                           WEFG_ip_ref_package_t *ippkg,
                           double weight,
                           WEFG_field_package_t *pkg) 
{
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  double F[3][3];
  
  ip_ctx_data.se       = se;
  ip_ctx.time          = WD_time (se->domain);
  ip_ctx.time_step     = WD_time_step (se->domain);
  ip_ctx.ctx_data      = &ip_ctx_data;
  ip_ctx.get_kinem_var = efgss_get_kinem_var;
  ip_ctx_data.sftab    = ippkg->ip->sftab;
  
  if (! (*ip_ctx.get_kinem_var) (&ip_ctx, MAT_3D_DEF_GRAD, F)) 
    return FALSE;
  
  pkg->value.tensor[0][0] += weight * (F[0][0] - 1);
  pkg->value.tensor[0][1] += weight * F[0][1];
  pkg->value.tensor[0][2] += weight * F[0][2];

  pkg->value.tensor[1][0] += weight * F[1][0];
  pkg->value.tensor[1][1] += weight * (F[1][1] - 1);
  pkg->value.tensor[1][2] += weight * F[1][2];

  pkg->value.tensor[2][0] += weight * F[2][0];
  pkg->value.tensor[2][1] += weight * F[2][1];
  pkg->value.tensor[2][2] += weight * (F[2][2] - 1);

  return TRUE;
} 



/*
   Update stress package.
   Used for interpolation-based probe.
 */
W_bool_t 
efg_update_stress_pkg (void *options,
                       WEFG_superelem_t *se /* EFG superelement */,
                       WEFG_ip_ref_package_t *ippkg,
                       double weight,
                       WEFG_field_package_t *pkg) 
{
  W_mat_3d_stress_type_enum requested_out_type =
 (W_mat_3d_stress_type_enum) options;
  W_mat_3d_stress_type_enum out_type;
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  double s[3][3];
  
  ip_ctx_data.se       = se;
  ip_ctx.time          = WD_time (se->domain);
  ip_ctx.time_step     = WD_time_step (se->domain);
  ip_ctx.ctx_data      = &ip_ctx_data;
  ip_ctx.get_kinem_var = efgss_get_kinem_var;
  ip_ctx_data.sftab    = ippkg->ip->sftab;
  
  if (!WM3D_update_stress (se->mat, ippkg->ip->matstate, W_MAT_STATE_UPDATE_TEMP,
                        &ip_ctx, requested_out_type, s,
                        &out_type)) {
    REPORT_ERROR (se->domain->errh, "Failed to compute stress");
    return FALSE;
  } 
  if (out_type != requested_out_type) {
    REPORT_ERROR (se->domain->errh, "types do not match");
    return FALSE;
  } 

  pkg->value.tensor[0][0] += weight * s[0][0];
  pkg->value.tensor[0][1] += weight * s[0][1];
  pkg->value.tensor[0][2] += weight * s[0][2];

  pkg->value.tensor[1][0] += weight * s[1][0];
  pkg->value.tensor[1][1] += weight * s[1][1];
  pkg->value.tensor[1][2] += weight * s[1][2];

  pkg->value.tensor[2][0] += weight * s[2][0];
  pkg->value.tensor[2][1] += weight * s[2][1];
  pkg->value.tensor[2][2] += weight * s[2][2];

  return TRUE;
} 




/*
   Update deformation energy package.
   Used for interpolation-based probe.
 */
W_bool_t 
efg_update_deform_energy_pkg (void *options,
                              WEFG_superelem_t *se /* EFG superelement */,
                              WEFG_ip_ref_package_t *ippkg,
                              double weight,
                              WEFG_field_package_t *pkg) 
{
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  double de;

  ip_ctx_data.se       = se;
  ip_ctx.time          = WD_time (se->domain);
  ip_ctx.time_step     = WD_time_step (se->domain);
  ip_ctx.ctx_data      = &ip_ctx_data;
  ip_ctx.get_kinem_var = efgss_get_kinem_var;
  ip_ctx_data.sftab    = ippkg->ip->sftab;

  if (!WM3D_get_mat_state_var (se->mat, ippkg->ip->matstate,
                               &ip_ctx, MAT_3D_DEFORM_ENERGY_DENSITY,
                               &de)) 
    REPORT_FATAL_ERROR (se->domain->errh, "Failed to retrieve deformation energy density");

  pkg->value.scalar += weight * de;

  return TRUE;
} 


/*
   Create a new crack tip sampler.
 Create a new crack tip sampler of mechanical quantities.
*/
void *
efg_new_crack_tip_sampler (WEFG_superelem_t *se /* EFG superelement */,
                           W_index_t pntidx, /* unique integer by which
                                                is this point referred to
                                              */
                           W_loc_t *loc /* location; needs to be on the
                                           crack front of at least one
                                           of the cracks
                                         */
                           ) 
{
  WEFG_crack_surf_t *cs;
  void *s;

  if (se->cracks) {
    LIST_LOOP_START (se->cracks, cs, WEFG_crack_surf_t *) {
      s = efgcr_new_sampler (cs, pntidx, loc);
      if (s) { 
	return s;
      } 
    }  LIST_LOOP_END ();
  }  
  return NULL;
}  


/*
   Evaluate support size.
   Evaluate support size at a given point.
   Support size is either given by the default support size associated
   with the superelement, or if the point falls into one of the regions
   where support size is specified, the support size is taken to be
   the smallest of support sizes as specified by all regions overlapping
   given point.
   
 */
double
efg_support_size_at (WEFG_superelem_t *se /* EFG superelement */,
                     double x, /* x-coord */
                     double y, /* y-coord */
                     double z /* z-coord */
                     ) 
{
  double ss  = 0;
  W_loc_t pt;

  if (se->compute_support_size) 
    return se->support_size;
  
  if (se->support_size_region_list == NULL) {
    ss = se->support_size;
  }  else  {
    WEFG_support_size_region_t *r;
    pt.x = x;
    pt.y = y;
    pt.z = z;
    LIST_LOOP_START (se->support_size_region_list, r,
                      WEFG_support_size_region_t *) {
      if (G3D_PT_INSIDE_BOX (r->box, pt)) {
        if (ss == 0) ss = FLT_MAX;
        ss = min (ss, r->support_size);
      } 
    }  LIST_LOOP_END ();
    if (ss == 0) {
      ss = se->support_size;    /* not in any of the regions */
    } 
  } 

  if (ss <= 0) {
    REPORT_FATAL_ERROR (se->domain->errh, "Zero support size");
  } 

  return ss; 
} 
	
 
/*
   Read rectangular region.
   Read rectangular region with specification of support size
 */
static int
read_support_size_region (WEFG_superelem_t *se /* EFG superelement */,
                          W_domain_t *d
                          ) 
{
  WEFG_support_size_region_t *r;
 
  /* SUPPORT_SIZE_REGION  lowx lowy lowz hix hiy hiz support_size */
  r = w_new (WEFG_support_size_region_t);
  CHECK_MALLOC (r);
  r->support_size = TOKEN_DBL (2);
  r->box.lower.x  = TOKEN_DBL (3);
  r->box.lower.y  = TOKEN_DBL (4);
  r->box.lower.z  = TOKEN_DBL (5);
  r->box.upper.x  = TOKEN_DBL (6);
  r->box.upper.y  = TOKEN_DBL (7);
  r->box.upper.z  = TOKEN_DBL (8);
  if (se->support_size_region_list == NULL) 
    se->support_size_region_list = make_list ();
  add_to_tail (se->support_size_region_list, r);

  return TRUE;
} 




/*
   Set default support size.
 */
void
efg_set_support_size (WEFG_superelem_t *se /* EFG superelement */,
                      double support_size /* support size; interpretation depends
                                               on the type of the weight function chosen
                                               for ESFLib */
                      ) 
{
  if (se->max_support_size < support_size) {
    efg_set_max_support_size (se, support_size);
  } 
  se->support_size = support_size;
} 


/*
   Set max support size
 */
void
efg_set_max_support_size (WEFG_superelem_t *se /* EFG superelement */,
                          double support_size /* support size; interpretation depends
                                               on the type of the weight function chosen
                                               for ESFLib
                                               */
                          ) 
{
  se->max_support_size = max (se->max_support_size, support_size);
#ifndef USE_ESFLIB
  ESP_3_set_range_size (se->max_support_size);
#endif
} 



/*
   Attach integration volume.
   Attach integration volume to the superelement
 */
void
efg_attach_hex (WEFG_superelem_t *se /* EFG superelement */,
                WEFG_hex_volume_t *h /* integration volume */
                ) 
{
  G3D_box_t box;
  G3D_INIT_BOX (box, h->vertex[0]->loc);
  G3D_UPDT_BOX (box, h->vertex[1]->loc);
  G3D_UPDT_BOX (box, h->vertex[2]->loc);
  G3D_UPDT_BOX (box, h->vertex[3]->loc);
  G3D_UPDT_BOX (box, h->vertex[4]->loc);
  G3D_UPDT_BOX (box, h->vertex[5]->loc);
  G3D_UPDT_BOX (box, h->vertex[6]->loc);
  G3D_UPDT_BOX (box, h->vertex[7]->loc);
  lobs_3d_add_object (se->hex_ss, h,
                      box.lower.x, box.lower.y, box.lower.z,
                      box.upper.x, box.upper.y, box.upper.z);
  add_to_tail (se->hex_volumes, h);
} 

static W_elem_methods_t efg_methods =  {
  elem_elcode,
  type_name,
  assemble_restoring_forces,
  assemble_mass,
  total_mass,
  NULL,
  suggested_time_step,
  deformation_energy,
  calc_eff_loads,
  solve_for,
  update_Warburton,
  update_Newmark,
  finalize_step,
  efghr_new_history_client_data,
  efghr_update_history,
  apply_initial_conditions,
  write_postproc_file,
#ifdef X_XFACE
  modify_elem_grep,
#endif
  print_info,
  configure,
  probe,
  num_of_faces,
  get_face_vertex_nodes,
  get_contact_faces,
  NULL,
  NULL,
  kinetic_energy,
  NULL,
  NULL
} ;

/*
   Read material.
   Material id is read from the input stream.
   The material is check if it is of type `general triaxial'.
 */
static void
read_mat3d (WEFG_superelem_t *se /* EFG superelement */, W_domain_t *d) 
{
  W_index_t mid;
  W_material_t *mat;
  
  mid = TOKEN_INT (2);
  mat = WD_get_material_by_idx (d, mid);
  if (mat == NULL) {
    sprintf (W_buf (), "No such material: %ld", mid);
    REPORT_ERROR (d->errh, W_buf ());
  } 
  if (mat->mcode != MCODE_GENERAL_TRIAX) {
    sprintf (W_buf (), "Material %d should be of code %d", (int) mid,
 (int) MCODE_GENERAL_TRIAX);
    REPORT_ERROR (d->errh, W_buf ());
  } 
  se->mat = (W_mat_3d_t *) mat->data;
} 


#define SHOULDHAVEHAD(subcommand, prereq_subcommand) \
   {                                                             \
    if (!had_##prereq_subcommand) {                             \
      sprintf (W_buf (),                                        \
               "%s is a prerequisite of %s",                    \
               #prereq_subcommand, #subcommand);                \
      REPORT_FATAL_ERROR (d->errh, W_buf ());                            \
    }                                                            \
  } 
#define SHOULDNOTHAVEHAD(subcommand, prereq_subcommand) \
   {                                                             \
    if (had_##prereq_subcommand) {                              \
      sprintf (W_buf (),                                        \
               "%s should precede %s",                          \
               #subcommand, #prereq_subcommand);                \
      REPORT_FATAL_ERROR (d->errh, W_buf ());                            \
    }                                                            \
  } 
#define HAD(subcommand) had_##subcommand = TRUE

#define PREAMBLE_STUFF() {                                                    \
                            SHOULDNOTHAVEHAD (preamble subcommands, VERT);    \
                            SHOULDNOTHAVEHAD (preamble subcommands, HEXVOL);  \
                            SHOULDNOTHAVEHAD (preamble subcommands, CONNECT); \
                            SHOULDNOTHAVEHAD (preamble subcommands, CRACK);   \
                          } 

/*
   Read superelement data.
   Reads superelement data from the input stream.
   The reader is invoked from within the generic element reader.
   See the User's manual for details on syntax.
   Once the read is finished, the superelement is set up and
   registered for integration of the internal degrees of freedom.
 */
W_bool_t 
efg_reader (W_domain_t *d) 
{
  const char *input_line;
  long elem_num = 0;
  WEFG_superelem_t *se;
#ifdef PVM_NEED_GHOST_ELEMENTS
  int is_ghost = FALSE;
#endif
  int got_all = FALSE;
  W_elem_t *e;
  int had_VERT = FALSE;
  int had_HEXVOL = FALSE;
  int had_CONNECT = FALSE;
  int had_MATERIAL = FALSE;
  int had_CRACK = FALSE;
  
  /* See the file WASP/src/input_template for a writeup of the superlement */
  /* input data structure. */
  
  se = 0;
  
  while ((input_line = W_get_input_line (d)) ) {
    TOKENS_PARSE (input_line);
    if (input_line[0] == END_OF_BLOCK) 
      return got_all;
    else  {
      if (TOKENS_HAVE_KEYWORD ("EFGSUPERELEM")) {
        int order = 1;
        elem_num = TOKEN_INT (2);
        if (TOKENS_HAVE_KEYWORD ("QUADRATIC")) order = 2;
        se = new_efg_superelem (d, order);
      }  else if (TOKENS_HAVE_KEYWORD ("VERT")) {
        SHOULDNOTHAVEHAD (VERT, CONNECT);
        SHOULDNOTHAVEHAD (VERT, HEXVOL);
        SHOULDHAVEHAD (VERT, MATERIAL);
        read_vertices (se, d);
        HAD (VERT);
      }  else if (TOKENS_HAVE_KEYWORD ("HEXVOL")) {
        SHOULDHAVEHAD (HEXVOL, VERT);
        SHOULDHAVEHAD (HEXVOL, CONNECT);
        SHOULDHAVEHAD (HEXVOL, MATERIAL);
        read_hexahedra (se, d);
        HAD (HEXVOL);
      }  else if (TOKENS_HAVE_KEYWORD ("MATERIAL")) {
        read_mat3d (se, d);
        HAD (MATERIAL);
      }  else if (TOKENS_HAVE_KEYWORD ("SPACING")) {
        PREAMBLE_STUFF ();
        se->node_spacing = TOKEN_DBL (2);
      }  else if (TOKENS_HAVE_KEYWORD ("NODES_PER_SIDE")) {
        PREAMBLE_STUFF ();
        se->num_nodes_per_hex_side = TOKEN_INT (2);
      }  else if (TOKENS_HAVE_KEYWORD ("COMPUTE_SUPPORT_SIZE")) {
        PREAMBLE_STUFF ();
        se->compute_support_size = TRUE;
      }  else if (TOKENS_HAVE_KEYWORD ("MAX_D_M") ||
                  TOKENS_HAVE_KEYWORD ("MAX_SUPPORT_SIZE")) {
        PREAMBLE_STUFF ();
        se->max_support_size = TOKEN_DBL (2);
      }  else if (TOKENS_HAVE_KEYWORD ("D_M") ||
                  TOKENS_HAVE_KEYWORD ("SUPPORT_SIZE")) {
        PREAMBLE_STUFF ();
        se->support_size = TOKEN_DBL (2);
        if (se->merging_distance <= 0) 
          se->merging_distance = se->support_size / 100;
      }  else if (TOKENS_HAVE_KEYWORD ("SUPPORT_SIZE_MULTIPLIER")) {
        PREAMBLE_STUFF ();
        se->support_size_multiplier = TOKEN_DBL (2);
      }  else if (TOKENS_HAVE_KEYWORD ("SUPPORT_SIZE_REGION")) {
        PREAMBLE_STUFF ();
        read_support_size_region (se, d);
      }  else if (TOKENS_HAVE_KEYWORD ("SYMMPLANE")) {
        PREAMBLE_STUFF ();
        read_symm_plane (se, d);
      }  else if (TOKENS_HAVE_KEYWORD ("BODYL")) {
        read_body_load (se, d);
      }  else if (TOKENS_HAVE_KEYWORD ("SFLIB_PREC_TOL")) {
        PREAMBLE_STUFF ();
        se->sflib_prec_tol = TOKEN_DBL (2);
      }  else if (TOKENS_HAVE_KEYWORD ("A_SOLVER")) {
        PREAMBLE_STUFF ();
        if (TOKENS_HAVE_KEYWORD ("LU")) 
          se->A_solver = ESFLIB_LU_DECOMP;
        else if (TOKENS_HAVE_KEYWORD ("QR")) 
          se->A_solver = ESFLIB_QR_DECOMP;
        else if (TOKENS_HAVE_KEYWORD ("CH")) 
          se->A_solver = ESFLIB_CH_DECOMP;
        else
          REPORT_FATAL_ERROR (d->errh, "Unknown solver");
      }  else if (TOKENS_HAVE_KEYWORD ("MIN_NEIGHBORS")) {
        PREAMBLE_STUFF ();
        se->num_neighbors_min = TOKEN_INT (2);
      }  else if (TOKENS_HAVE_KEYWORD ("SUGGESTED_NEIGHBORS")) {
        PREAMBLE_STUFF ();
        se->num_neighbors_suggested = TOKEN_INT (2);
      }  else if (TOKENS_HAVE_KEYWORD ("USE_WRAP_AROUND")) {
        PREAMBLE_STUFF ();
#ifdef USE_WRAP_AROUND
        se->use_wrap_around = TRUE;
#else
        REPORT_FATAL_ERROR (d->errh, "Wrap around not compiled in");
#endif
      }  else if (TOKENS_HAVE_KEYWORD ("GENERAL_CONSTRAINT")) {
        PREAMBLE_STUFF ();
        se->general_constraint[0] = TOKEN_INT (2);
        se->general_constraint[1] = TOKEN_INT (3);
        se->general_constraint[2] = TOKEN_INT (4);
      }  else if (TOKENS_HAVE_KEYWORD ("VSS_BUCKET")) {
        PREAMBLE_STUFF ();
        se->vertex_ss_bucket_size = TOKEN_INT (2);
        if (se->verts_ss == NULL) {
          se->verts_ss
            = pirat_new_pirat (se->vertex_ss_bucket_size, 1000 * MACHEPS);
        } 
      }  else if (TOKENS_HAVE_KEYWORD ("EFGN_BUCKET")) {
        PREAMBLE_STUFF ();
        se->efg_node_ss_bucket_size = TOKEN_INT (2);
        if (se->efg_node_ss == NULL) {
          se->efg_node_ss
            = lobs_3d_new_instance (NULL, se->efg_node_ss_bucket_size);
        } 
      }  else if (TOKENS_HAVE_KEYWORD ("QUADRATURE") 
#       ifdef ENABLE_TWO_POINT_RULE
                 && (!se->enable_2pt_quadrature) 
#       endif
                 ) {
        PREAMBLE_STUFF ();
        se->quadrature_order = TOKEN_INT (2);
#     ifdef ENABLE_TWO_POINT_RULE
      }  else if (TOKENS_HAVE_KEYWORD ("ENABLE_2PT_QUADRATURE")) {
        PREAMBLE_STUFF ();
        se->enable_2pt_quadrature = TRUE;
        se->quadrature_order = 2;
#     endif
      }  else if (TOKENS_HAVE_KEYWORD ("TOLERANCE")) {
        PREAMBLE_STUFF ();
        se->tolerance = TOKEN_DBL (2);
      }  else if (TOKENS_HAVE_KEYWORD ("TIME_INDEPENDENT_MASS")) {
        PREAMBLE_STUFF ();
        se->time_independent_mass = TRUE;
      }  else if (TOKENS_HAVE_KEYWORD ("MERGING_DISTANCE")) {
        PREAMBLE_STUFF ();
        se->merging_distance = TOKEN_DBL (2);
      }  else if (TOKENS_HAVE_KEYWORD ("CONVEX")) {
        PREAMBLE_STUFF ();
        se->is_convex             = TRUE;
        se->is_convex_with_cracks = FALSE;
#       ifdef USE_ESFLIB
        if (!ESFLIB_set_node_hidden_by_bdry_func (se->esflib, NULL)) /*disable*/
          REPORT_ERROR (d->errh, (char *) ESFL3_explain_error (se->esflib));
#       else
	if (!ESP_3_set_seg_intersects_bdry_func (NULL)) /* disable */
          REPORT_ERROR (d->errh, ESP_3_explain_error ());
#       endif
      }  else if (   TOKENS_HAVE_KEYWORD ("NON_CONVEX") 
                 || TOKENS_HAVE_KEYWORD ("CONCAVE")) {
        PREAMBLE_STUFF ();
        se->is_convex             = FALSE;
        se->is_convex_with_cracks = FALSE;
      }  else if (TOKENS_HAVE_KEYWORD ("CONVEX_WITH_CRACKS")) {
        PREAMBLE_STUFF ();
        se->is_convex_with_cracks = TRUE;
        se->is_convex             = FALSE;
      }  else if (TOKENS_HAVE_KEYWORD ("WEIGHT_FUNC_TYPE")) {
        PREAMBLE_STUFF ();
        if (TOKENS_HAVE_KEYWORD ("QUADRATIC_TENSOR_PROD")) {
          se->weight_func_type = WEFG_QUADRATIC_TENSOR_PROD_WEIGHT_FUNC;
        }  else if (TOKENS_HAVE_KEYWORD ("SMOOTH_QUADRATIC_TENSOR_PROD")) {
          se->weight_func_type = WEFG_SMOOTH_QUADRATIC_TENSOR_PROD_WEIGHT_FUNC;
        }  else if (TOKENS_HAVE_KEYWORD ("QUARTIC_SPLINE_SPHERICAL")) {
          se->weight_func_type = WEFG_QUARTIC_SPLINE_SPHERICAL_WEIGHT_FUNC;
        }  else if (TOKENS_HAVE_KEYWORD ("QUADRATIC_SPHERICAL")) {
          se->weight_func_type = WEFG_QUADRATIC_SPHERICAL_WEIGHT_FUNC;
        }  else if (TOKENS_HAVE_KEYWORD ("SMOOTH_QUADRATIC_SPHERICAL")) {
          se->weight_func_type = WEFG_SMOOTH_QUADRATIC_SPHERICAL_WEIGHT_FUNC;
        }  else if (TOKENS_HAVE_KEYWORD ("GAUSSIAN_SPHERICAL")) {
          se->weight_func_type = WEFG_GAUSSIAN_SPHERICAL_WEIGHT_FUNC;
        }  else  {
          REPORT_FATAL_ERROR (d->errh, "Unknown weight function type");
        } 
      }  else if (TOKENS_HAVE_KEYWORD ("GENERATE_NODES_INSIDE_ONLY")) {
        PREAMBLE_STUFF ();
        se->generate_nodes_inside_only = TRUE;
      }  else if (TOKENS_HAVE_KEYWORD ("SCRAMBLE_DIAGONALS") ||
                 TOKENS_HAVE_KEYWORD ("RANDOMIZE_DIAGONAL_ORIENTATION")) {
        PREAMBLE_STUFF ();
        se->scramble_diagonals = TRUE;
      }  else if (TOKENS_HAVE_KEYWORD ("GENERATE_NODES_IN_VOLUME")) {
        PREAMBLE_STUFF ();
        se->generate_nodes_in_volume = TRUE;
        se->lower_corner.x = TOKEN_DBL (2);
        se->lower_corner.y = TOKEN_DBL (3);
        se->lower_corner.z = TOKEN_DBL (4);
        se->upper_corner.x = TOKEN_DBL (5);
        se->upper_corner.y = TOKEN_DBL (6);
        se->upper_corner.z = TOKEN_DBL (7);
      }  else if (TOKENS_HAVE_KEYWORD ("GENERATE_NODES_AT_VERTICES")) {
        PREAMBLE_STUFF ();
        se->generate_nodes_at_vertices = TRUE;
      }  else if (TOKENS_HAVE_KEYWORD ("RAYLEIGH_MASS_DAMPING")) {
        PREAMBLE_STUFF ();
	se->mass_damping_coeff = TOKEN_DBL (2);
      }  else if (TOKENS_HAVE_KEYWORD ("CONNECT")) {
        SHOULDNOTHAVEHAD (CONNECT, HEXVOL);
        SHOULDHAVEHAD (CONNECT, VERT);
        read_connectivities (se, d);
        HAD (CONNECT);
      }  else if (TOKENS_HAVE_KEYWORD ("CONCL")) {
        read_conc_loads (se, d);
      }  else if (TOKENS_HAVE_KEYWORD ("DISTRIBL")) {
        read_distrib_loads (se, d);
      }  else if (TOKENS_HAVE_KEYWORD ("CONTACTL")) {
        read_contact_loads (se, d);
      }  else if (TOKENS_HAVE_KEYWORD ("CRACK")) {
        SHOULDHAVEHAD (CRACK, VERT);
        SHOULDHAVEHAD (CRACK, CONNECT);
        SHOULDHAVEHAD (CRACK, HEXVOL);
        SHOULDHAVEHAD (CRACK, MATERIAL);
        read_crack (se, d);
        HAD (CRACK); 
      }  else if (   TOKENS_HAVE_KEYWORD ("END")) {
        WD_add_element (d, elem_num,
                        (e = WE_new_element (d, &efg_methods, se
#                                   ifdef PVM_NEED_GHOST_ELEMENTS
                                             , FALSE
#                                   endif
                                             )) );
        set_up_efg_superelem (se);
        WD_register_elem_for_integration (d, e);
        got_all = TRUE;
      }  else  {
        UNRECOGNIZED_INPUT ();
        return FALSE;
      } 
    } 
  } 
  
  if (se->max_support_size <= 0) 
    REPORT_FATAL_ERROR (d->errh, "Must set max support_size");
  return got_all;
} 




#endif /* EFG_WASP  */
