#include "Wconfig.h"

#ifdef EFG_WASP

#include "WEFG.h"
#include "WEFGsf.h"
#include "WEFGpriv.h"
#include "WEFGutil.h"
#include "Wgaussp.h"
#include "Wxf.h"
#include "Wutil.h"
#include "Wcdi.h"
#include "Wdomain.h"
#include "Wtfuncs.h"
#include "Wnode.h"

/* RCSHEADER "$Id: efgcl.c,v 1.1.1.1 2001/07/24 02:42:51 pkrysl Exp $" */


                                /* read concentrated load.
                                   read a concentrated load from
                                   the input stream.
                                 */
void
read_conc_loads(WEFG_superelem_t *se, W_domain_t *d)
{
  const char *input_line;
  WEFG_conc_load_t *c;

  while ((input_line = W_get_input_line(d)) != NULL) {
    TOKENS_PARSE(input_line);
    if (TOKENS_HAVE_KEYWORD("END")) {
      if (TOKENS_HAVE_KEYWORD("CONCL"))
        return;
      else
        REPORT_FATAL_ERROR(se->domain->errh, "CONCL does not end by END CONCL");
    } else {
      c = w_new(WEFG_conc_load_t);
      CHECK_MALLOC(c);
      c->id                   = TOKEN_INT(1);
      c->loc.x                = TOKEN_DBL(2);
      c->loc.y                = TOKEN_DBL(3);
      c->loc.z                = TOKEN_DBL(4);
      c->amplitudes.x         = TOKEN_DBL(5);
      c->amplitudes.y         = TOKEN_DBL(6);
      c->amplitudes.z         = TOKEN_DBL(7);
      c->time_func_num        = TOKEN_INT(8);
      INIT_SFTAB(c->sftab); 
      add_to_tail(se->conc_loads, c);
    }
  }
  REPORT_FATAL_ERROR(se->domain->errh, "CONCL does not end by END CONCL");
}

static void
check_eh(WEFG_hex_volume_t *eh, W_loc_t *at, W_loc_t *ploc);

#define CHECK_EH(eh, at, ploc) check_eh(eh, at, ploc)

static void
check_eh(WEFG_hex_volume_t *eh, W_loc_t *at, W_loc_t *ploc)
{
  if (eh == NULL) {
    sprintf(W_buf(), "No enclosing hex at [%g;%g;%g]", at->x, at->y, at->z);
    REPORT_FATAL_ERROR(NULL, W_buf());
  }
}


/*
  Assemble concentrated load.
 */
void
conc_load_assemble(WEFG_superelem_t *se, /* EFG superelement */
                   WEFG_conc_load_t *cl /* concentrated load */
                   )
{
  int K, ne;
  W_vect_t f, frc;
  double mult;
  W_time_func_t *tf = WD_get_time_func_by_idx(se->domain, cl->time_func_num);
  WEFG_hex_volume_t *eh;
  W_loc_t ploc;
  WEFG_if_data_t if_data;
  double phi;
  WEFG_node_t *efgp;
  WEFG_eval_for_packet_t efp;

  efp.se = se;
  efp.pt_wrt_crack_surf_classification = WEFG_POS_WRT_SURF_UNSPEC;
  efp.crack_surf = NULL;
  
  if (SFTAB_IS_STALE(cl->sftab, se)) {
    if (SFTAB_INVALID (cl->sftab, se, &cl->loc)) {
      SFL_sf_table_t sf;
    
      sf = efgsf_sf_table(se);
      eh = efg_find_enclosing_hex (se, &cl->loc, 0.0001, &ploc);

      if (!SFTAB_IS_EMPTY(cl->sftab)) {
        USE_SFTAB_SEARCH_FUNC(se, cl->sftab);
      }
    
      CHECK_EH(eh, &cl->loc, &ploc);
      if (hex_is_interface_vol(eh)) {
        if_data.h     = eh;
        if_data.xi    = ploc.x;
        if_data.eta   = ploc.y; 
        if_data.theta = ploc.z;
        if (!SFL_shape_func_at(sf, cl->loc.x, cl->loc.y, cl->loc.z,
                               &efp, (SFL_generic_ptr_t)&if_data)) {
          SFL_ERROR(sf, cl->loc.x, cl->loc.y, cl->loc.z, E_K_WARNING);
          ESFL3_set_num_of_entries (sf, 0);
        }
      } else {
        if (!SFL_shape_func_at(sf, cl->loc.x, cl->loc.y, cl->loc.z,
                                 &efp, NULL)) {
          SFL_ERROR(sf, cl->loc.x, cl->loc.y, cl->loc.z, E_K_WARNING);
          ESFL3_set_num_of_entries (sf, 0);
        }
      }
      efgsf_create_sftab (se, &cl->sftab, sf, NULL, FALSE);
      USE_DEFAULT_SEARCH_FUNC(se);
      UPDATE_EFG_SF_STAT(se, SFTAB_NUM_OF_ENTRIES(cl->sftab));
    }
    TOUCH_SFTAB(cl->sftab, se);
  }
  
  f = cl->amplitudes;
  mult = WTF_func_val(tf, WD_time(se->domain));
  f.x *= mult;
  f.y *= mult;
  f.z *= mult;
  ne = SFTAB_NUM_OF_ENTRIES(cl->sftab);
  for (K = 0; K < ne; K++) {
    phi = SFTAB_PHI(cl->sftab, K);
    frc.x = f.x * phi;
    frc.y = f.y * phi;
    frc.z = f.z * phi;
    efgp = SFTAB_NODE(cl->sftab, K);
    ASSEMBLE_TO_EFF_LOAD(se, efgp, frc); 
  }
}


#endif /* EFG_WASP  */
