
#include "Wconfig.h"

#ifdef EFG_WASP

#include "Wtypes.h"
#include "Wgaussp.h"
#include "Wxf.h"
#include "Wutil.h"
#include "Wcdi.h"
#include "Wdomain.h"
#include "WEFG.h"
#include "WEFGsf.h"
#include "WEFGpriv.h"
#include "WEFGutil.h"
#include "Wnode.h"

/* RCSHEADER "$Id: efgctl.c,v 1.1.1.1 2001/07/24 02:42:51 pkrysl Exp $" */

static void
compute_local_csys(WEFG_contact_load_t *ctl);


#define subt_vec(ab, a, b)    \
   (ab)->x = (a)->x - (b)->x, \
   (ab)->y = (a)->y - (b)->y, \
   (ab)->z = (a)->z - (b)->z
#define avg_vec(ab, a, b)             \
   (ab)->x = 0.5 * ((a)->x + (b)->x), \
   (ab)->y = 0.5 * ((a)->y + (b)->y), \
   (ab)->z = 0.5 * ((a)->z + (b)->z)
#define cross_p(ab, a, b)                      \
  (ab)->x = ((a)->y*(b)->z) - ((a)->z*(b)->y), \
  (ab)->y = ((a)->z*(b)->x) - ((a)->x*(b)->z), \
  (ab)->z = ((a)->x*(b)->y) - ((a)->y*(b)->x)
#define dot_p(a, b) ((a)->x*(b)->x + (a)->y*(b)->y + (a)->z*(b)->z)
#define lin_comb(r, p, pq, l)                   \
  (r)->x = (p)->x + (l) * (pq)->x,              \
  (r)->y = (p)->y + (l) * (pq)->y,              \
  (r)->z = (p)->z + (l) * (pq)->z

/* REad contact load specification.
   REad contact load specification from the input stream.
   See the User's guide for details.x
 */
void
read_contact_loads(WEFG_superelem_t *se, /* EFG superelement */
                   W_domain_t *d
                   )
{
  const char                *input_line;
  WEFG_contact_load_t       *ctl;
  int                        i, kp;
  int                        id;
  WEFG_contact_load_type_t   type = WEFG_CONTACT_LOAD_UNILATERAL;
  double                     contact_press_scaling;
  W_loc_t                    loc[4];
  int                        suggested_qr;
  int                        split;
  G3D_plane_t                contact_plane;
  double                     xi, eta, dxi, deta, n[4];
#define NL(num, DIR)                                                          \
   { ctl->vdata[num].loc.DIR                                                  \
       = (n[0]*loc[0].DIR+n[1]*loc[1].DIR+n[2]*loc[2].DIR+n[3]*loc[3].DIR); }
#define N_LOC(num) { NL(num, x);  NL(num, y); NL(num, z); }

  while ((input_line = W_get_input_line(d)) != NULL) {
    TOKENS_PARSE(input_line);
    if (TOKENS_HAVE_KEYWORD("END")) {
      if (TOKENS_HAVE_KEYWORD("CONTACTL"))
        return;
      else
        REPORT_FATAL_ERROR(se->domain->errh, "CONTACTL does not end by END CONTACTL");
    } else {
      id = TOKEN_INT(1);
      if        (strcmp(TOKEN_STR(2), "UNILATERAL") == 0) {
        type = WEFG_CONTACT_LOAD_UNILATERAL;
      } else if (strcmp(TOKEN_STR(2), "BILATERAL") == 0) {
        type = WEFG_CONTACT_LOAD_STICKY;
      } else if (strcmp(TOKEN_STR(2), "STICKY") == 0) {
        REPORT_FATAL_ERROR(se->domain->errh, "Unknown contact load type");
      } else {
        REPORT_FATAL_ERROR(se->domain->errh, "Unknown contact load type");
      }
      if ((kp = tokens_have_keyword(d->parser, 1, "QR")) > 0)
        suggested_qr = TOKEN_INT(kp+1);
      else
        suggested_qr = 6;
      contact_press_scaling = 1;
      if ((kp = tokens_have_keyword(d->parser, 1, "CONTACT_PRESS_SCALING")) > 0)
        contact_press_scaling = TOKEN_DBL(kp+1);
      split = 1;
      if ((kp = tokens_have_keyword(d->parser, 1, "SPLIT")) > 0)
        split = TOKEN_INT(kp+1);
      if (split <= 0) {
        BAD_INPUT();
      }

      FOR(i, 4) {
        if ((input_line = W_get_input_line(d)) == NULL)
          BAD_INPUT();
        TOKENS_PARSE(input_line);
        loc[i].x = TOKEN_DBL(1);
        loc[i].y = TOKEN_DBL(2);
        loc[i].z = TOKEN_DBL(3);
      }

      if ((input_line = W_get_input_line(d)) == NULL)
        BAD_INPUT();
      TOKENS_PARSE(input_line);
      if (TOKENS_HAVE_KEYWORD("BY_PATCH")) {
        /* NOTE:  The contact plane has a normal oriented opposite to */
        /* the normal of the patch */
        G3D_plane_consts_4(&contact_plane, &loc[3], &loc[2], &loc[1], &loc[0]);
      } else if (TOKENS_HAVE_KEYWORD("BY_POINTS")) {
        W_loc_t p0, p1, p2;
        p0.x = TOKEN_DBL(2);
        p0.y = TOKEN_DBL(3);
        p0.z = TOKEN_DBL(4);
        p1.x = TOKEN_DBL(5);
        p1.y = TOKEN_DBL(6);
        p1.z = TOKEN_DBL(7);
        p2.x = TOKEN_DBL(8);
        p2.y = TOKEN_DBL(9);
        p2.z = TOKEN_DBL(10);
        G3D_plane_consts_3(&contact_plane, &p0, &p1, &p2);
      } else {
        REPORT_FATAL_ERROR(se->domain->errh, "Not yet implemented");
      }

      dxi = deta = 2.0 / split;
      for (xi = -1; xi < 0.99999; xi += dxi) { 
        for (eta = -1; eta < 0.99999; eta += deta) {
          ctl = w_new(WEFG_contact_load_t);
          CHECK_MALLOC(ctl); 
          ctl->id                    = id;
          ctl->type                  = type;
          ctl->contact_press_scaling = contact_press_scaling;
          ctl->suggested_qr          = suggested_qr;
          ctl->qp                    = NULL; 
          G3D_q4_N(n, xi, eta);          N_LOC(0); 
          G3D_q4_N(n, xi+dxi, eta);      N_LOC(1); 
          G3D_q4_N(n, xi+dxi, eta+deta); N_LOC(2); 
          G3D_q4_N(n, xi, eta+deta);     N_LOC(3);
          INIT_SFTAB(ctl->vdata[0].sftab);
          INIT_SFTAB(ctl->vdata[1].sftab);
          INIT_SFTAB(ctl->vdata[2].sftab);
          INIT_SFTAB(ctl->vdata[3].sftab);
          compute_local_csys(ctl);
          ctl->contact_plane         = contact_plane;
          add_to_tail(se->contact_loads, ctl);
        }
      }
      INIT_TIME_STAMP(ctl->time_stamp);
    }
  }
  REPORT_FATAL_ERROR(se->domain->errh, "CONTACTL does not end by END CONTACTL");
#undef N_LOC
#undef NL
}


static void
check_eh(WEFG_hex_volume_t *eh, W_loc_t *at, W_loc_t *ploc);

#define CHECK_EH(eh, at, ploc) check_eh(eh, at, ploc)

static void
check_eh(WEFG_hex_volume_t *eh, W_loc_t *at, W_loc_t *ploc)
{
  if (eh == NULL) {
    sprintf(W_buf(), "No enclosing hex at [%g;%g;%g]", at->x, at->y, at->z);
    REPORT_FATAL_ERROR(NULL, W_buf());
  }
}


/*
   Recompute quadrature points.
   Re-evaluate shape function tables for the quadrature-point array
   of the integration volume.

   Create the quadrature points if there were none; recompute them
   if they need updating.
 */
static void
recompute_ip_array(WEFG_superelem_t *se, /* EFG superelement */
                   WEFG_contact_load_t *ctl /* contact load  */
                   )
{
  WEFG_quadr_points_t *qp = ctl->qp;
  W_gauss_integration_rule_t *qr = qp->qr;
  W_gauss_integration_point_t *gp = qr->ip;
  int qo = qr->order, tot = qo * qo;
  WEFG_integration_point_t *ip;
  W_loc_t at;
  double n[4];
  WEFG_hex_volume_t *eh;
  W_loc_t ploc;
  WEFG_if_data_t if_data;
  int i;
#define CAST_IF_DATA (SFL_generic_ptr_t)&if_data
#define N_LOC(v, DIR, n)                          \
   v.DIR = (  n[0] * ctl->vdata[0].loc.DIR      \
            + n[1] * ctl->vdata[1].loc.DIR      \
            + n[2] * ctl->vdata[2].loc.DIR      \
            + n[3] * ctl->vdata[3].loc.DIR)
  WEFG_eval_for_packet_t efp;

  efp.se = se;
  efp.pt_wrt_crack_surf_classification = WEFG_POS_WRT_SURF_UNSPEC;
  efp.crack_surf = NULL;
  
  FOR (i, 4) {
    if (SFTAB_IS_STALE(ctl->vdata[i].sftab, se)) {
      if (SFTAB_INVALID (ctl->vdata[i].sftab, se, &ctl->vdata[i].loc)) {
        SFL_sf_table_t sf;
        sf = efgsf_sf_table(se);
        at = ctl->vdata[i].loc;
        eh = efg_find_enclosing_hex (se, &at, 0.0001, &ploc);
        CHECK_EH(eh, &at, &ploc);
        if (hex_is_interface_vol(eh)) {
          if_data.h     = eh;
          if_data.xi    = ploc.x;
          if_data.eta   = ploc.y;
          if_data.theta = ploc.z;
          if (!SFL_shape_func_at(sf, at.x, at.y, at.z,
                                   &efp, CAST_IF_DATA)) {
            SFL_ERROR(sf, at.x, at.y, at.z, E_K_WARNING);
            ESFL3_set_num_of_entries (sf, 0);
          }
        } else {
          if (!SFL_shape_func_at(sf, at.x, at.y, at.z,
                                   &efp, NULL)) {
            SFL_ERROR(sf, at.x, at.y, at.z, E_K_WARNING);
            ESFL3_set_num_of_entries (sf, 0);
          }
        }
        efgsf_create_sftab (se, &ctl->vdata[i].sftab, sf, NULL, FALSE);
        UPDATE_EFG_SF_STAT(se, SFTAB_NUM_OF_ENTRIES(ctl->vdata[i].sftab));
      } /* if (SFTAB_INVALID( */
      TOUCH_SFTAB(ctl->vdata[i].sftab, se);
    }
  }
  
  /* If the quadrature points have not been created yet */
  if (qp->efg_ip == NULL) {
    SFL_sf_table_t sf;
    sf = efgsf_sf_table(se);
    qp->efg_ip = w_new_array(tot, WEFG_integration_point_t *);
    CHECK_MALLOC(qp->efg_ip);
    IPA_2D_LOOP(qo)
      {
        ip = w_new(WEFG_integration_point_t);
        CHECK_MALLOC(ip); ip->matstate = NULL;
        G3D_q4_N(n, gp[LOOP_2D_I].Loc, gp[LOOP_2D_J].Loc);
        N_LOC(at, x, n);
        N_LOC(at, y, n);
        N_LOC(at, z, n);
        INIT_SFTAB(ip->sftab);
        eh = efg_find_enclosing_hex (se, &at, 0.0001, &ploc);
        CHECK_EH(eh, &at, &ploc);
        if (hex_is_interface_vol(eh)) {
          if_data.h     = eh;
          if_data.xi    = ploc.x;
          if_data.eta   = ploc.y;
          if_data.theta = ploc.z;
          if (!SFL_shape_func_at(sf, at.x, at.y, at.z,
                                   &efp, CAST_IF_DATA)) {
            SFL_ERROR(sf, at.x, at.y, at.z, E_K_WARNING);
            ESFL3_set_num_of_entries (sf, 0);
          }
        } else {
          if (!SFL_shape_func_at(sf, at.x, at.y, at.z,
                                   &efp, NULL)) {
            SFL_ERROR(sf, at.x, at.y, at.z, E_K_WARNING);
            ESFL3_set_num_of_entries (sf, 0);
          }
        }
        efgsf_create_sftab (se, &ip->sftab, sf, NULL, FALSE);
        UPDATE_EFG_SF_STAT(se, SFTAB_NUM_OF_ENTRIES(ip->sftab));
        IPA_ENTRY_2D(qp->efg_ip) = ip;
      }
      IPA_2D_LOOP_END();
  } else {
    SFL_sf_table_t sf;
    sf = efgsf_sf_table(se);
    IPA_2D_LOOP(qo)
      {
        ip = IPA_ENTRY_2D(qp->efg_ip);
        G3D_q4_N(n, gp[LOOP_2D_I].Loc, gp[LOOP_2D_J].Loc);
        N_LOC(at, x, n);
        N_LOC(at, y, n);
        N_LOC(at, z, n);
        if (SFTAB_INVALID (ip->sftab, se, &at)) {
          if (!SFTAB_IS_EMPTY(ip->sftab)) {
            USE_SFTAB_SEARCH_FUNC(se, ip->sftab);
          }
          eh = efg_find_enclosing_hex (se, &at, 0.0001, &ploc);
          CHECK_EH(eh, &at, &ploc);
          if (hex_is_interface_vol(eh)) {
            if_data.h     = eh;
            if_data.xi    = ploc.x;
            if_data.eta   = ploc.y;
            if_data.theta = ploc.z;
            if (!SFL_shape_func_at(sf, at.x, at.y, at.z,
                                     &efp, CAST_IF_DATA)) {
              SFL_ERROR(sf, at.x, at.y, at.z, E_K_WARNING);
            ESFL3_set_num_of_entries (sf, 0);
          }
          } else {
            if (!SFL_shape_func_at(sf, at.x, at.y, at.z,
                                     &efp, NULL)) {
              SFL_ERROR(sf, at.x, at.y, at.z, E_K_WARNING);
            ESFL3_set_num_of_entries (sf, 0);
          }
          }
          efgsf_create_sftab (se, &ip->sftab, sf, NULL, FALSE);
          USE_DEFAULT_SEARCH_FUNC(se);
          UPDATE_EFG_SF_STAT(se,
                             SFTAB_NUM_OF_ENTRIES(ip->sftab));
        } /* if (SFTAB_INVALID(ip->sftab, se, &at)) */
        TOUCH_SFTAB(ip->sftab, se);
      }
      IPA_2D_LOOP_END();
}
}




/* compute local coordinate system.
    compute local coordinate system on the quadrilateral patch
    which defines potential contact area.
 */
static void
compute_local_csys(WEFG_contact_load_t *ctl /* contact load */
                   )
{
  W_vect_t p01, p12, p23, p30, f, s;

  avg_vec(&p01, &ctl->vdata[0].loc, &ctl->vdata[1].loc);
  avg_vec(&p23, &ctl->vdata[2].loc, &ctl->vdata[3].loc);
  avg_vec(&p12, &ctl->vdata[1].loc, &ctl->vdata[2].loc);
  avg_vec(&p30, &ctl->vdata[0].loc, &ctl->vdata[3].loc);
  subt_vec(&f, &p12, &p30);
  subt_vec(&s, &p23, &p01);

  cross_p(&ctl->normal, &f, &s);    
  v3_normalize(&ctl->normal);

  v3_normalize(&f);
  ctl->u = f;

  cross_p(&ctl->v, &ctl->normal, &ctl->u);

  ctl->vdata[0].x = 0;
  ctl->vdata[0].y = 0;
  subt_vec(&f, &ctl->vdata[0].loc, &ctl->vdata[1].loc);
  ctl->vdata[1].x = dot_p(&f, &ctl->u);
  ctl->vdata[1].y = dot_p(&f, &ctl->v);
  subt_vec(&f, &ctl->vdata[0].loc, &ctl->vdata[2].loc);
  ctl->vdata[2].x = dot_p(&f, &ctl->u);
  ctl->vdata[2].y = dot_p(&f, &ctl->v);
  subt_vec(&f, &ctl->vdata[0].loc, &ctl->vdata[3].loc);
  ctl->vdata[3].x = dot_p(&f, &ctl->u);
  ctl->vdata[3].y = dot_p(&f, &ctl->v);
}




/*
  Assemble contact tractions.
  Assemble contact tractions for a penalty-based contact patch.
  Evaluate shape function tables at quadrature points, and compute
  current locations of vertices.  Evaluate contact tractions from signed
  distances of vertices wrt contact plane.
 */
void
contact_load_assemble(WEFG_superelem_t *se, /* EFG superelement */
                      WEFG_contact_load_t *ctl /* contact load */
                      )
{
  int k, ne;
  W_vect_t f, F;
  W_gauss_integration_point_t *gp;
  WEFG_integration_point_t **ip, *cip;
  double jdetw, press;
  W_float_t n[4], J[2][2];
  double x0, y0, x1, y1, x2, y2, x3, y3;
  double p[4];
  double d[4];
  double bulkmod;
  WEFG_node_t *efgp;
  W_vect_t Normal = ctl->contact_plane.n;
  double phi;
  int i;
  double x, y, z;

  if (ctl->qp == NULL) {
    ctl->qp = w_new(WEFG_quadr_points_t);
    CHECK_MALLOC(ctl->qp);
    ctl->qp->qr     = WG_get_integration_rule(ctl->suggested_qr);
    ctl->qp->efg_ip = NULL;
    recompute_ip_array(se, ctl);
    TOUCH_TIME_STAMP(ctl->time_stamp, se);
  } else {
    if (TIME_STAMP_IS_STALE(ctl->time_stamp, se))
      recompute_ip_array(se, ctl);
    TOUCH_TIME_STAMP(ctl->time_stamp, se);
  }

  /* evaluate current locations */
  FOR(i, 4) {
    efgsf_eval_curr_loc(&ctl->vdata[i].sftab, &x, &y, &z);
    ctl->vdata[i].loc.x = x;
    ctl->vdata[i].loc.y = y;
    ctl->vdata[i].loc.z = z;
  }
  compute_local_csys(ctl);
  x0 = ctl->vdata[0].x, y0 = ctl->vdata[0].y;
  x1 = ctl->vdata[1].x, y1 = ctl->vdata[1].y;
  x2 = ctl->vdata[2].x, y2 = ctl->vdata[2].y;
  x3 = ctl->vdata[3].x, y3 = ctl->vdata[3].y;
  
  /* compute the tractions from the signed distances */
  FOR(i, 4) {
    d[i] = G3D_signed_dist_from_plane(&ctl->contact_plane, &ctl->vdata[i].loc);
  }
  if (ctl->type == WEFG_CONTACT_LOAD_UNILATERAL
      && (d[0] >= 0 && d[1] >= 0 && d[2] >= 0 && d[3] >= 0))
    return; /* no contact */
  if (!WM3D_get_mat_const(se->mat, MAT_3D_BULK_MODULUS, &bulkmod))
    REPORT_FATAL_ERROR(se->domain->errh, "No bulk modulus");
  FOR(i, 4) {
    if (ctl->type == WEFG_CONTACT_LOAD_UNILATERAL && d[i] >= 0)
      p[i] = 0;
    else
      p[i] = (- ctl->contact_press_scaling * bulkmod * d[i]);
  }
  
  /* Now integrate the traction */
  ip = ctl->qp->efg_ip;
  gp = ctl->qp->qr->ip;
  IPA_2D_LOOP(ctl->qp->qr->order)
    {
      G3D_q4_J(J,
               gp[LOOP_2D_I].Loc, gp[LOOP_2D_J].Loc,
               x0, y0, x1, y1, x2, y2, x3, y3);
      jdetw = (  gp[LOOP_2D_I].Wgt * gp[LOOP_2D_J].Wgt
                 * G3D_q4_detJ(J));
      if (jdetw == 0)
        REPORT_FATAL_ERROR(se->domain->errh, "bad load area");
      G3D_q4_N(n, gp[LOOP_2D_I].Loc, gp[LOOP_2D_J].Loc);
      press = jdetw*(n[0]*p[0]+n[1]*p[1]+n[2]*p[2]+n[3]*p[3]);
      cip   = IPA_ENTRY_2D(ip);
      f.x = press * Normal.x;
      f.y = press * Normal.y;
      f.z = press * Normal.z;
      ne    = SFTAB_NUM_OF_ENTRIES(cip->sftab);
      FOR(k, ne) {
        phi = SFTAB_PHI(cip->sftab, k);
        F.x = f.x * phi;
        F.y = f.y * phi;
        F.z = f.z * phi;
        efgp = SFTAB_NODE(cip->sftab, k);
        ASSEMBLE_TO_EFF_LOAD(se, efgp, F);
      }
    } 
    IPA_2D_LOOP_END();
}


#endif /* EFG_WASP  */



