
#include "Wconfig.h"

#ifdef EFG_WASP

#include "WEFG.h"
#include "WEFGsf.h"
#include "WEFGpriv.h"
#include "WEFGutil.h"
#include "Wgaussp.h"
#include "Wpss.h"
#include "Wxf.h"
#include "Wutil.h"
#include "Wcdi.h"
#include "Wdomain.h"
#include "Wtfuncs.h"
#include "Wnode.h"

/* RCSHEADER "$Id: efgdl.c,v 1.1.1.1 2001/07/24 02:42:51 pkrysl Exp $" */

static void
compute_local_csys(WEFG_distrib_load_t *dl);


#define subt_vec(ab, a, b)    \
   (ab)->x = (a)->x - (b)->x, \
   (ab)->y = (a)->y - (b)->y, \
   (ab)->z = (a)->z - (b)->z
#define avg_vec(ab, a, b)             \
   (ab)->x = 0.5 * ((a)->x + (b)->x), \
   (ab)->y = 0.5 * ((a)->y + (b)->y), \
   (ab)->z = 0.5 * ((a)->z + (b)->z)
#define cross_p(ab, a, b)                      \
  (ab)->x = ((a)->y*(b)->z) - ((a)->z*(b)->y), \
  (ab)->y = ((a)->z*(b)->x) - ((a)->x*(b)->z), \
  (ab)->z = ((a)->x*(b)->y) - ((a)->y*(b)->x)
#define dot_p(a, b) ((a)->x*(b)->x + (a)->y*(b)->y + (a)->z*(b)->z)
#define lin_comb(r, p, pq, l)                   \
  (r)->x = (p)->x + (l) * (pq)->x,              \
  (r)->y = (p)->y + (l) * (pq)->y,              \
  (r)->z = (p)->z + (l) * (pq)->z


static void
make_distrib_loads (WEFG_superelem_t *se,
                    int split,
                    int id,
                    WEFG_distrib_load_type_t type,
                    int time_func_num,
                    W_vect_t *dir_vect,
                    int suggested_qr,
                    double traction[4],
                    W_loc_t locs[4]
                    )
{
  int nsubdiv = max (split, 1);
  double dxi, deta, xi, eta;
  int i, j;
  double n[4];
  WEFG_distrib_load_t *dl;
#define CHILD_ID_INC 1000000  
#define IPOL(num)                               \
  {                                             \
    dl->vdata[num].loc.x                        \
      =   n[0] * locs[0].x                      \
        + n[1] * locs[1].x                      \
        + n[2] * locs[2].x                      \
        + n[3] * locs[3].x;                     \
    dl->vdata[num].loc.y                        \
      =   n[0] * locs[0].y                      \
        + n[1] * locs[1].y                      \
        + n[2] * locs[2].y                      \
        + n[3] * locs[3].y;                     \
    dl->vdata[num].loc.z                        \
      =   n[0] * locs[0].z                      \
        + n[1] * locs[1].z                      \
        + n[2] * locs[2].z                      \
        + n[3] * locs[3].z;                     \
    dl->vdata[num].traction                     \
      =   n[0] * traction[0]                    \
        + n[1] * traction[1]                    \
        + n[2] * traction[2]                    \
        + n[3] * traction[3];                   \
  }

  dxi  = 2.0 / nsubdiv;
  deta = dxi;

  xi = -1;
  FOR (i, nsubdiv) {
    eta = -1;
    FOR (j, nsubdiv) {
      dl = w_new(WEFG_distrib_load_t);
      CHECK_MALLOC(dl);
      dl->id = id + (nsubdiv == 1 ? 0 : CHILD_ID_INC);
      dl->type = type;
      dl->time_func_num = time_func_num;
      dl->suggested_qr = suggested_qr;
      dl->dir_vect = *dir_vect;
      G3D_q4_N (n, xi, eta);
      IPOL (0);
      G3D_q4_N (n, xi+dxi, eta);
      IPOL (1);
      G3D_q4_N (n, xi+dxi, eta+deta);
      IPOL (2);
      G3D_q4_N (n, xi, eta+deta);
      IPOL (3);
      compute_local_csys(dl);
      dl->qp = NULL;
      INIT_TIME_STAMP(dl->time_stamp);
      add_to_tail (se->distrib_loads, dl);
      eta += deta;
    }
    xi += dxi;
  }
}

/*
   Read a distributed load.
   Read a distributed load (traction).
   See User's guide for syntax and other details.
 */
void
read_distrib_loads(WEFG_superelem_t *se, /* EFG superelement */
                   W_domain_t *d /* input stream */
                   )
{
  const char *input_line;
  int kp, i;
  int id;
  WEFG_distrib_load_type_t type;
  int time_func_num;
  W_loc_t dir_vect;
  int suggested_qr;
  W_loc_t locs[4];
  double traction[4];
  int split;
  
  while ((input_line = W_get_input_line (d)) != NULL) {
    TOKENS_PARSE (input_line);
    if (TOKENS_HAVE_KEYWORD ("END")) {
      if (TOKENS_HAVE_KEYWORD ("DISTRIBL"))
        return;
      else
        REPORT_FATAL_ERROR (se->domain->errh, "DISTRIBL does not end by END DISTRIBL");
    } else {
      id            = TOKEN_INT (1);
      type          = WEFG_DISTRIB_LOAD_DEAD_PRESSURE;
      if        (tokens_token_and_keyword_equiv (TOKEN_STR(2),
                                                 "DEAD_PRESSURE")) {
        type = WEFG_DISTRIB_LOAD_DEAD_PRESSURE;
      } else if (tokens_token_and_keyword_equiv (TOKEN_STR(2),
                                                 "PRESSURE")) {
        type = WEFG_DISTRIB_LOAD_PRESSURE;
        REPORT_FATAL_ERROR (se->domain->errh, "WEFG_DISTRIB_LOAD_PRESSURE not implemented");
      } else if (tokens_token_and_keyword_equiv (TOKEN_STR(2),
                                                 "DEAD_SHEAR")) {
        type = WEFG_DISTRIB_LOAD_DEAD_SHEAR;
      } else if (tokens_token_and_keyword_equiv (TOKEN_STR(2),
                                                 "SHEAR")) {
        type = WEFG_DISTRIB_LOAD_SHEAR;
        REPORT_FATAL_ERROR (se->domain->errh, "WEFG_DISTRIB_LOAD_SHEAR not implemented");
      } else {
        REPORT_FATAL_ERROR (se->domain->errh, "Unknown distributed load type");
      }
      time_func_num = TOKEN_INT (3);
      dir_vect.x = dir_vect.y = dir_vect.z = 0;
      if (   type == WEFG_DISTRIB_LOAD_DEAD_SHEAR
          || type == WEFG_DISTRIB_LOAD_SHEAR) {
        dir_vect.x = TOKEN_DBL (4);
        dir_vect.y = TOKEN_DBL (5);
        dir_vect.z = TOKEN_DBL (6);
      }
      if ((kp = tokens_have_keyword (d->parser, 1, "SPLIT")) > 0)
        split = TOKEN_INT(kp+1);
      else
        split = 1;
      if ((kp = tokens_have_keyword (d->parser, 1, "QR")) > 0)
        suggested_qr = TOKEN_INT (kp+1);
      else
        suggested_qr = 6;
      FOR(i, 4) {
        if ((input_line = W_get_input_line (d)) == NULL) {
          BAD_INPUT ();
        }
        TOKENS_PARSE (input_line);
        locs[i].x   = TOKEN_DBL (1);
        locs[i].y   = TOKEN_DBL (2);
        locs[i].z   = TOKEN_DBL (3);
        traction[i] = TOKEN_DBL (4);
      } 
      make_distrib_loads (se, split, id, type, time_func_num, &dir_vect,
                          suggested_qr, traction, locs);
    }
  }
  REPORT_FATAL_ERROR(se->domain->errh, "DISTRIBL does not end by END DISTRIBL");
}




static void
check_eh(WEFG_hex_volume_t *eh, W_loc_t *at, W_loc_t *ploc);

#define CHECK_EH(eh, at, ploc) check_eh(eh, at, ploc)

static void
check_eh(WEFG_hex_volume_t *eh, W_loc_t *at, W_loc_t *ploc)
{
  if (eh == NULL) {
    sprintf(W_buf(), "No enclosing hex at [%g;%g;%g]", at->x, at->y, at->z);
    REPORT_FATAL_ERROR(NULL, W_buf());
  }
}

#define N_LOC(v, DIR, n)                          \
   v.DIR = (  n[0] * dl->vdata[0].loc.DIR      \
            + n[1] * dl->vdata[1].loc.DIR      \
            + n[2] * dl->vdata[2].loc.DIR      \
            + n[3] * dl->vdata[3].loc.DIR)



static void
recompute_ip_array(WEFG_superelem_t *se, WEFG_distrib_load_t *dl)
{
  WEFG_quadr_points_t *qp = dl->qp;
  W_gauss_integration_rule_t *qr = qp->qr;
  W_gauss_integration_point_t *gp = qr->ip;
  int qo = qr->order, tot = qo * qo;
  WEFG_integration_point_t *ip;
  W_loc_t at;
  W_float_t n[4];
  WEFG_hex_volume_t *eh;
  W_loc_t ploc;
  WEFG_if_data_t if_data;
  SFL_sf_table_t sf;
#define CAST_IF_DATA (SFL_generic_ptr_t)&if_data
  WEFG_eval_for_packet_t efp;

  efp.se = se;
  efp.pt_wrt_crack_surf_classification = WEFG_POS_WRT_SURF_UNSPEC;
  efp.crack_surf = NULL;
  
  sf = efgsf_sf_table(se);

  /* If the quadrature points have not been created yet */
  if (qp->efg_ip == NULL) {
    qp->efg_ip = w_new_array(tot, WEFG_integration_point_t *);
    CHECK_MALLOC(qp->efg_ip);
    IPA_2D_LOOP(qo)
      {
        ip = w_new(WEFG_integration_point_t);
        CHECK_MALLOC(ip); ip->matstate = NULL;
        pss_N_matrix(n, gp[LOOP_2D_I].Loc, gp[LOOP_2D_J].Loc);
        N_LOC(at, x, n);
        N_LOC(at, y, n);
        N_LOC(at, z, n);
        INIT_SFTAB(ip->sftab);
        eh = efg_find_enclosing_hex (se, &at, 0.0001, &ploc);
        CHECK_EH(eh, &at, &ploc);
        if (hex_is_interface_vol(eh)) {
          if_data.h     = eh;
          if_data.xi    = ploc.x;
          if_data.eta   = ploc.y;
          if_data.theta = ploc.z;
          if (!SFL_shape_func_at(sf, at.x, at.y, at.z,
                                   &efp, CAST_IF_DATA)) {
            SFL_ERROR (sf, at.x, at.y, at.z, E_K_WARNING);
            ESFL3_set_num_of_entries (sf, 0);
          }
        } else {
          if (!SFL_shape_func_at(sf, at.x, at.y, at.z,
                                   &efp, NULL)) {
            SFL_ERROR(sf, at.x, at.y, at.z, E_K_WARNING);
            ESFL3_set_num_of_entries (sf, 0);
          }
        }
        efgsf_create_sftab (se, &ip->sftab, sf, NULL, FALSE);
        UPDATE_EFG_SF_STAT(se, SFTAB_NUM_OF_ENTRIES(ip->sftab));
        IPA_ENTRY_2D(qp->efg_ip) = ip;
      }
      IPA_2D_LOOP_END();
  } else {
    IPA_2D_LOOP(qo)
      {
        ip = IPA_ENTRY_2D(qp->efg_ip);
        pss_N_matrix(n, gp[LOOP_2D_I].Loc, gp[LOOP_2D_J].Loc);
        N_LOC(at, x, n);
        N_LOC(at, y, n); 
        N_LOC(at, z, n);
        if (SFTAB_INVALID (ip->sftab, se, &at)) {
          eh = efg_find_enclosing_hex (se, &at, 0.0001, &ploc);
          CHECK_EH(eh, &at, &ploc);
          USE_SFTAB_SEARCH_FUNC (se, ip->sftab);
          if (hex_is_interface_vol(eh)) {
            if_data.h     = eh;
            if_data.xi    = ploc.x;
            if_data.eta   = ploc.y;
            if_data.theta = ploc.z;
            if (!SFL_shape_func_at(sf, at.x, at.y, at.z,
                                     &efp, CAST_IF_DATA)) {
              SFL_ERROR(sf, at.x, at.y, at.z, E_K_WARNING);
              ESFL3_set_num_of_entries (sf, 0);
            }
          } else {
            if (!SFL_shape_func_at(sf, at.x, at.y, at.z,
                                     &efp, NULL)) {
              SFL_ERROR(sf, at.x, at.y, at.z, E_K_WARNING);
              ESFL3_set_num_of_entries (sf, 0);
            }
          }
          efgsf_create_sftab (se, &ip->sftab, sf, NULL, FALSE);
          USE_DEFAULT_SEARCH_FUNC(se);
          UPDATE_EFG_SF_STAT(se,
                             SFTAB_NUM_OF_ENTRIES(ip->sftab));
        } /* if (SFTAB_INVALID( */
        TOUCH_SFTAB(ip->sftab, se);
      }
      IPA_2D_LOOP_END();
  }
}




/*
   Compute the local coordinate system.
   Compute the local coordinate system for the traction quadrilateral patch.
 */
static void
compute_local_csys(WEFG_distrib_load_t *dl /* distributed load */
                   )
{
  W_vect_t p01, p12, p23, p30, f, s;

  avg_vec(&p01, &dl->vdata[0].loc, &dl->vdata[1].loc);
  avg_vec(&p23, &dl->vdata[2].loc, &dl->vdata[3].loc);
  avg_vec(&p12, &dl->vdata[1].loc, &dl->vdata[2].loc);
  avg_vec(&p30, &dl->vdata[0].loc, &dl->vdata[3].loc);
  subt_vec(&f, &p12, &p30);
  subt_vec(&s, &p23, &p01);
  cross_p(&dl->normal, &f, &s);    
  v3_normalize(&dl->normal);
  if (   dl->type == WEFG_DISTRIB_LOAD_DEAD_SHEAR
      || dl->type == WEFG_DISTRIB_LOAD_SHEAR)
    { /* shear loads -- they act along /u/ */
    cross_p(&s, &dl->normal, &dl->dir_vect);
    cross_p(&dl->u, &s, &dl->normal);
    v3_normalize(&dl->u);
  } else { /* pressure loads */
    v3_normalize(&f);
    dl->u = f;
  }
  cross_p(&dl->v, &dl->normal, &dl->u);
  dl->vdata[0].x = 0;
  dl->vdata[0].y = 0;
  subt_vec(&f, &dl->vdata[0].loc, &dl->vdata[1].loc);
  dl->vdata[1].x = dot_p(&f, &dl->u);
  dl->vdata[1].y = dot_p(&f, &dl->v);
  subt_vec(&f, &dl->vdata[0].loc, &dl->vdata[2].loc);
  dl->vdata[2].x = dot_p(&f, &dl->u);
  dl->vdata[2].y = dot_p(&f, &dl->v);
  subt_vec(&f, &dl->vdata[0].loc, &dl->vdata[3].loc);
  dl->vdata[3].x = dot_p(&f, &dl->u);
  dl->vdata[3].y = dot_p(&f, &dl->v);
}



/*
  Assemble distributed load.
  Assemble tractions acting on a quadrilateral patch.
 */
void
distrib_load_assemble(WEFG_superelem_t *se, /* EFG superelement */
                      WEFG_distrib_load_t *dl /* distributed load patch */
                      )
{
  int k, ne;
  W_vect_t f, frc;
  double mult;
  W_time_func_t *tf = WD_get_time_func_by_idx(se->domain, dl->time_func_num);
  W_gauss_integration_point_t *gp;
  WEFG_integration_point_t **ip, *cip;
  double jdetw, p;
  W_float_t n[4], J[2][2];
  double x0 = dl->vdata[0].x, y0 = dl->vdata[0].y;
  double x1 = dl->vdata[1].x, y1 = dl->vdata[1].y;
  double x2 = dl->vdata[2].x, y2 = dl->vdata[2].y;
  double x3 = dl->vdata[3].x, y3 = dl->vdata[3].y;
  double p0 = dl->vdata[0].traction;
  double p1 = dl->vdata[1].traction;
  double p2 = dl->vdata[2].traction;
  double p3 = dl->vdata[3].traction;
  WEFG_node_t *efgp;
  W_vect_t Normal = dl->normal;
  W_vect_t U = dl->u;
  double phi;

  if (dl->qp == NULL) {
    dl->qp = w_new(WEFG_quadr_points_t);
    CHECK_MALLOC(dl->qp);
    dl->qp->qr     = WG_get_integration_rule(dl->suggested_qr);
    dl->qp->efg_ip = NULL;
    recompute_ip_array(se, dl);
    TOUCH_TIME_STAMP(dl->time_stamp, se);
  } else {
    if (TIME_STAMP_IS_STALE(dl->time_stamp, se))
      recompute_ip_array(se, dl);
    TOUCH_TIME_STAMP(dl->time_stamp, se);
  }
  
  /* Now integrate the traction */
  ip = dl->qp->efg_ip;
  gp = dl->qp->qr->ip;
  mult = WTF_func_val(tf, WD_time(se->domain));
  p0 *= mult;  p1 *= mult;  p2 *= mult;  p3 *= mult;
  IPA_2D_LOOP(dl->qp->qr->order)
    {
      pss_jacobian_matrix(J,
                          gp[LOOP_2D_I].Loc, gp[LOOP_2D_J].Loc,
                          x0, y0, x1, y1, x2, y2, x3, y3);
      jdetw = (  gp[LOOP_2D_I].Wgt * gp[LOOP_2D_J].Wgt
                 * pss_jacobian(J));
      if (jdetw == 0)
        REPORT_FATAL_ERROR(se->domain->errh, "bad load area");
      pss_N_matrix(n, gp[LOOP_2D_I].Loc, gp[LOOP_2D_J].Loc);
      p = jdetw * (n[0]*p0 + n[1]*p1 + n[2]*p2 + n[3]*p3);
      cip  = IPA_ENTRY_2D(ip);
      switch (dl->type) {
      case WEFG_DISTRIB_LOAD_DEAD_PRESSURE:
        f.x = p * Normal.x;
        f.y = p * Normal.y;
        f.z = p * Normal.z;
        break;
      case WEFG_DISTRIB_LOAD_DEAD_SHEAR:
        f.x = p * U.x;
        f.y = p * U.y;
        f.z = p * U.z;
        break;
      default:
        REPORT_FATAL_ERROR(se->domain->errh, "unimplemented load");
        break;
      }
      ne = SFTAB_NUM_OF_ENTRIES(cip->sftab);
      FOR(k, ne) {
        phi = SFTAB_PHI(cip->sftab, k);
        frc.x = f.x * phi;
        frc.y = f.y * phi;
        frc.z = f.z * phi;
        efgp = SFTAB_NODE(cip->sftab, k);
        ASSEMBLE_TO_EFF_LOAD(se, efgp, frc);
      } 
    }
    IPA_2D_LOOP_END();
}


#endif /* EFG_WASP  */



