#include "Wconfig.h"

#ifdef EFG_WASP

#include "Wtypes.h"
#include "Wmat3d.h"
#include "WEFG.h"
#include "WEFGsf.h"
#include "Wdomain.h"
#include "Whistry.h"
#include "Wgaussp.h"
#include "Wxf.h"
#include "Wutil.h"
#include "Wcdi.h"
#include "WEFGutil.h"
#include "WEFGpriv.h"
#include "WEFGss.h"

/* RCSHEADER "$Id: efghr.c,v 1.1.1.1 2001/07/24 02:42:51 pkrysl Exp $" */

typedef void (*store_proc)(W_elem_t *e,
			   W_solution_history_t *h,
			   void *data,
			   double time);

/* history client data */
typedef struct efg_history_client_data_t {
  WEFG_superelem_t   *se; /* EFG superelement */
  W_index_t          pntidx; /* unique integer  */
  W_loc_t            loc; /* location */
  void              *data; /* data */
  store_proc         store; /* `store' procedure */
}              efg_history_client_data_t;





/*
  Create a history client.
  Create a history client for the EFG superelement.
  It can be either an arbitrary material point, or it can be
  a crack front point (point moving with the crack front).
 */
void * /* client or NULL (not enough memory) */
efghr_new_history_client_data (W_elem_t *e, /* generic element (specialized
                                             to the EFG superelement) */
                               W_index_t idx, /* unique integer  */
                               W_history_writer_t *r, /* history writer */
                               char *input_line /* input line (to allow
                                                 for additional parsing of
                                                the line contents by this
                                                function) */
                               )
{
  WEFG_superelem_t *se = (WEFG_superelem_t *)e->data;
  efg_history_client_data_t *cd;
  W_domain_t *d = e->domain;
  
  cd = w_new(efg_history_client_data_t);
  CHECK_MALLOC(cd);
 
  TOKENS_PARSE(input_line);
  cd->se     = se;
  cd->pntidx = TOKEN_INT(3);
  cd->loc.x  = TOKEN_DBL(4);
  cd->loc.y  = TOKEN_DBL(5);
  cd->loc.z  = TOKEN_DBL(6);
  if (TOKENS_HAVE_KEYWORD("CRACK_TIP")) {
    cd->data  = efg_new_crack_tip_sampler (se, cd->pntidx, &cd->loc);
    cd->store = efgcr_sampler_store;
  } else {
    cd->data  = efgv_new_sampler(se, cd->pntidx, &cd->loc);
    cd->store = efgv_sampler_store;
  }

  return (NODE)cd;
}



/*
  Update history.
  Update history record.
 */
void
efghr_update_history (W_elem_t *e, /* generic element */
                      W_solution_history_t *h, /* solution history */
                     W_float_t time /* current time */
                      )
{
  efg_history_client_data_t *cd =
    (efg_history_client_data_t *)h->client_data;

  cd->store (e, h, cd->data, time);
}


#endif /* EFG_WASP  */
