#include "Wconfig.h"

#ifdef EFG_WASP

#include "Wtypes.h"
#include "Wdomain.h"
#include "Wmat3d.h"
#include "Wgaussp.h"
#include "Wxf.h"
#include "Wutil.h"
#include "Wcdi.h"
#include "WEFG.h"
#include "WEFGsf.h"
#include "WEFGutil.h"
#include "WEFGpriv.h"
#include "WEFGss.h"
#include "Wnode.h"
#include "g3d.h"
#include "g3dhex.h"


#undef I

/* RCSHEADER "$Id: efghv.c,v 1.7 2004/09/06 01:17:58 cvs Exp $" */

static int debug_print = 0;
#undef DebugPV
#define DebugPV(arg1, arg2, arg3)   if (debug_print) PV(arg1, arg2, arg3) 

/* The hexahedron is given by its vertices. */
/* Vertex # is associated to the intrinsic coordinates */
/* =================================================== */
/*        0                        -1 -1 -1 */
/*        1                         1 -1 -1 */
/*        2                         1  1 -1 */
/*        3                        -1  1 -1 */
/*        ================================= */
/*        4                        -1 -1  1 */
/*        5                         1 -1  1 */
/*        6                         1  1  1 */
/*        7                        -1  1  1 */
/* */
/* The faces are: */
/* face #      vertices (counterclockwise when seen from outside) */
static W_index_t face_vertex_indices[6 /* face # */][4 /* vertex index */] = {
/* 0 */ { 0, 3, 2, 1 },
/* 1 */ { 0, 1, 5, 4 },
/* 2 */ { 1, 2, 6, 5 },
/* 3 */ { 2, 3, 7, 6 },
/* 4 */ { 3, 0, 4, 7 },
/* 5 */ { 4, 5, 6, 7 }
      };

#define SET_FACE(face_num, into, hex)                                   \
   {                                                                    \
     (into)->num_vert = 4;                                              \
     (into)->v[0] = hex->vertex[face_vertex_indices[face_num][0]];      \
     (into)->v[1] = hex->vertex[face_vertex_indices[face_num][1]];      \
     (into)->v[2] = hex->vertex[face_vertex_indices[face_num][2]];      \
     (into)->v[3] = hex->vertex[face_vertex_indices[face_num][3]];      \
   }


#define N_LOC(h, v, DIR, n)                     \
   v.DIR = (  n[0] * h->vertex[0]->loc.DIR      \
            + n[1] * h->vertex[1]->loc.DIR      \
            + n[2] * h->vertex[2]->loc.DIR      \
            + n[3] * h->vertex[3]->loc.DIR      \
            + n[4] * h->vertex[4]->loc.DIR      \
            + n[5] * h->vertex[5]->loc.DIR      \
            + n[6] * h->vertex[6]->loc.DIR      \
            + n[7] * h->vertex[7]->loc.DIR)


#define ALL_REF_COORDS                                                  \
  h->vertex[0]->loc.x, h->vertex[0]->loc.y, h->vertex[0]->loc.z,        \
  h->vertex[1]->loc.x, h->vertex[1]->loc.y, h->vertex[1]->loc.z,        \
  h->vertex[2]->loc.x, h->vertex[2]->loc.y, h->vertex[2]->loc.z,        \
  h->vertex[3]->loc.x, h->vertex[3]->loc.y, h->vertex[3]->loc.z,        \
  h->vertex[4]->loc.x, h->vertex[4]->loc.y, h->vertex[4]->loc.z,        \
  h->vertex[5]->loc.x, h->vertex[5]->loc.y, h->vertex[5]->loc.z,        \
  h->vertex[6]->loc.x, h->vertex[6]->loc.y, h->vertex[6]->loc.z,        \
  h->vertex[7]->loc.x, h->vertex[7]->loc.y, h->vertex[7]->loc.z


static int
shared_face(WEFG_b_face_t *f, WEFG_hex_volume_t *with_h);
static void
renew_quadrature_array(WEFG_superelem_t *se,
                       WEFG_hex_volume_t *h);
static void
split_hex(WEFG_superelem_t *se, WEFG_hex_volume_t *hp,
          int nsubdiv, double subdiv_size);


#define ALL_GP_LOCS gp[LOOP_I].Loc, gp[LOOP_J].Loc, gp[LOOP_K].Loc
#define GP_WGT(gp) (gp[LOOP_I].Wgt * gp[LOOP_J].Wgt * gp[LOOP_K].Wgt)

static void
scramble_diagonal(WEFG_hex_volume_t *h);

/*
   Read hexahedral integration volumes from the input stream.
 */
void
read_hexahedra (WEFG_superelem_t *se, /* EFG superelement */
                W_domain_t *d /* input stream */
                )
{
  const char *input_line;
  WEFG_hex_volume_t *h;
  int nsubdiv = 0, kp, i, f;
  double subdiv_size;
  W_bool_t is_interface;

  while ((input_line = W_get_input_line(d)) != NULL) {
    TOKENS_PARSE(input_line);
    if (TOKENS_HAVE_KEYWORD("END")) {
      if (TOKENS_HAVE_KEYWORD("HEXVOL"))
        return;
      else
        REPORT_FATAL_ERROR(se->domain->errh, "HEXVOL does not end by END HEXVOL");
    } else {
      h = w_new(WEFG_hex_volume_t);
      CHECK_MALLOC(h);
      h->id = TOKEN_INT(1);
      is_interface = FALSE;
      if ((TOKENS_TOTAL() == 6) || (TOKENS_TOTAL() == 5)) {
        int idc[8]={0,1,2,2,3,3,3,3};
        WEFG_vertex_t *v[4];
        FOR(i, 4) {
          v[i] = efg_get_vertex (se, TOKEN_INT(i+2));
          if (v[i] == NULL)
            REPORT_FATAL_ERROR(se->domain->errh, "Bad vertex reference");
          if (efgv_get_coupling_fen_conn (se, v[i]) != NULL)
            is_interface = TRUE;
        }
        FOR(i, 8) { h->vertex[i] = v[idc[i]]; }
      } else if ((TOKENS_TOTAL() >= 9)) {
        FOR(i, 8) {
          h->vertex[i] = efg_get_vertex (se, TOKEN_INT(i+2));
          if (h->vertex[i] == NULL)
            REPORT_FATAL_ERROR(se->domain->errh, "Bad vertex reference");
          if (efgv_get_coupling_fen_conn (se, h->vertex[i]) != NULL)
            is_interface = TRUE;
        }
      } else {
        REPORT_FATAL_ERROR(se->domain->errh, "Bad hexvol connectivity: need either 8 (hex) or 4 (tet -> hex)");
      }        
      if (se->scramble_diagonals) {
        scramble_diagonal(h);
      }
      FOR (i, 8) 
        attach_hex_vol (h->vertex[i], h);
      FOR (f, 6) {
        h->face_on_boundary[f] = TRUE; /* Initially assume that all faces */
        /* ARE on the boundary; later check */
        /* whether the face is shared by */
        /* another volume, and if it is */
        /* then mark it as not being */
        /* on the boundary. */
#ifdef X_XFACE
        h->face_grep[f] = NULL;
#endif
      }
#     ifdef ENABLE_TWO_POINT_RULE
      /*  0 signals that the order is to be taken from the superelement;
         -2 signals that the 2pt diagonal rule is to be used
         */
      h->quad_order = (se->enable_2pt_quadrature ? -2 : 0);
#     else
      h->quad_order = 0;
#     endif
      if ((kp = tokens_have_keyword(d->parser, 1, "QO")) > 0) {
        h->quad_order = TOKEN_INT(kp+1);
      } 
      h->qp = NULL;
      h->volume = 0;
      if ((kp = tokens_have_keyword(d->parser, 1, "SPLIT")) > 0) {
        int kp_;
        nsubdiv = 0; subdiv_size = 0;
        if ((kp_ = tokens_have_keyword(d->parser, 1, "SIZE")) > 0) {
          subdiv_size = TOKEN_DBL(kp_+1);
        } else {
          nsubdiv = TOKEN_INT(kp+1);
        }
        if (!is_interface)
          split_hex (se, h, nsubdiv, subdiv_size);
      } else {
        h->offspring = NULL;
      }
      INIT_TIME_STAMP(h->time_stamp);
      
      efg_attach_hex (se, h);
    }
  }
  REPORT_FATAL_ERROR (se->domain->errh, "HEXVOL does not end by END HEXVOL");
}


#define MAKE_VERTEX_AT(N, xi, eta, tht)                                     \
   {                                                                        \
     G3D_hex_N(n, xi, eta, tht);                                            \
     N_LOC(hp, loc, x, n); N_LOC(hp, loc, y, n); N_LOC(hp, loc, z, n);      \
     G3D_INIT_BOX(range, loc);                                              \
     G3D_INFLATE_BOX(range, EPS);                                           \
     v = efg_vertex_in_range(se, &range);                                   \
     if (v == NULL) {                                                       \
       v = efgv_new_vertex(0, &loc);                                        \
       efg_attach_vertex(se, v);                                            \
     }                                                                      \
     ho->vertex[N] = v;                                                     \
   }

 

static int
one_of_four(void);
static int
flip(void);


static int
one_of_four(void)
{
  
  if (flip())
    if (flip()) return 1;
    else        return 2;
  else
    if (flip()) return 3;
    else        return 4;
}


static int
flip(void)
{
  long rn;
  int two = 2;

  rn = (long) ((double) LONG_MAX * rand_0_to_1());

  return (((rn / two) * 2) == rn);
}

/*
   Randomize the orientation of the `main' diagonal.
   Randomize the orientation of the `main' diagonal of the integration
   volume.  (The `main' diagonal connects vertices number 0 and 6.)
 */
static void
scramble_diagonal(WEFG_hex_volume_t *h /* integration volume */
                  )
{
  int which = one_of_four();
#define FOUR_SHIFT(start)                         \
  {                                               \
    WEFG_vertex_t *t = h->vertex[start];          \
    h->vertex[start  ] = h->vertex[start+1];      \
    h->vertex[start+1] = h->vertex[start+2];      \
    h->vertex[start+2] = h->vertex[start+3];      \
    h->vertex[start+3] = t;                       \
  }
#define SHIFT_ONCE() { FOUR_SHIFT(0); FOUR_SHIFT(4); }

  switch (which) {
  case 1:
    /* do nothing */
    break;
  case 2:
    /* rotate once */
    SHIFT_ONCE();
    break;
  case 3:
    /* rotate twice */
    SHIFT_ONCE();
    SHIFT_ONCE();
    break;
  case 4:
    /* rotate thrice */
    SHIFT_ONCE();
    SHIFT_ONCE();
    SHIFT_ONCE();
    break;
  default:
    break;
  }
}

/*
   Is the integration volume at FE/EFG interface?
   Is the integration volume at FE/EFG interface?
   If any of its vertices has an EFG node attached which
   is of type `FE interface', the integration volume is
   of interface type.
 */
int
hex_is_interface_vol(WEFG_hex_volume_t *h /* integration volume */
                     )
{
  return 
    (   (h->vertex[0]->efgn != NULL && EFGN_IS_FE_IF(h->vertex[0]->efgn))
     || (h->vertex[1]->efgn != NULL && EFGN_IS_FE_IF(h->vertex[1]->efgn))
     || (h->vertex[2]->efgn != NULL && EFGN_IS_FE_IF(h->vertex[2]->efgn))
     || (h->vertex[3]->efgn != NULL && EFGN_IS_FE_IF(h->vertex[3]->efgn))
     || (h->vertex[4]->efgn != NULL && EFGN_IS_FE_IF(h->vertex[4]->efgn))
     || (h->vertex[5]->efgn != NULL && EFGN_IS_FE_IF(h->vertex[5]->efgn))
     || (h->vertex[6]->efgn != NULL && EFGN_IS_FE_IF(h->vertex[6]->efgn))
     || (h->vertex[7]->efgn != NULL && EFGN_IS_FE_IF(h->vertex[7]->efgn))
     );
}

#define CBND(hbbox, hex, which, func, N)                         \
  {                                                              \
    hbbox.which.x = func(hbbox.which.x, hex->vertex[N]->loc.x);  \
    hbbox.which.y = func(hbbox.which.y, hex->vertex[N]->loc.y);  \
    hbbox.which.z = func(hbbox.which.z, hex->vertex[N]->loc.z);  \
  }
#define DEFINE_HBBOX(hex, hbbox)                                     \
  {                                                                  \
    hbbox.lower.x = hbbox.upper.x = hex->vertex[0]->loc.x;           \
    hbbox.lower.y = hbbox.upper.y = hex->vertex[0]->loc.y;           \
    hbbox.lower.z = hbbox.upper.z = hex->vertex[0]->loc.z;           \
    CBND(hbbox, hex, lower, min, 1);                                 \
    CBND(hbbox, hex, lower, min, 2);                                 \
    CBND(hbbox, hex, lower, min, 3);                                 \
    CBND(hbbox, hex, lower, min, 4);                                 \
    CBND(hbbox, hex, lower, min, 5);                                 \
    CBND(hbbox, hex, lower, min, 6);                                 \
    CBND(hbbox, hex, lower, min, 7);                                 \
    CBND(hbbox, hex, upper, max, 1);                                 \
    CBND(hbbox, hex, upper, max, 2);                                 \
    CBND(hbbox, hex, upper, max, 3);                                 \
    CBND(hbbox, hex, upper, max, 4);                                 \
    CBND(hbbox, hex, upper, max, 5);                                 \
    CBND(hbbox, hex, upper, max, 6);                                 \
    CBND(hbbox, hex, upper, max, 7);                                 \
  }

/*
   Is point inside integration volume?
   Is a given point inside a hexahedral integration volume?
 */
WEFG_hex_volume_t *    /* the enclosing hex */
hex_loc_is_inside (WEFG_hex_volume_t *h, /* hex */
                   W_loc_t *loc, /* location in the reference coordinates */
                   double eps, /* tolerance in the parametric <-1;1>^3 space */
                   W_loc_t *param_loc /* returns the parametric location inside the hex if the function returned non-NULL value */
                   )
{
#define IN_BI_UNIT(val) ((val >= -1-eps) && (val <= 1+eps))
  G3D_hex_t hex;
#define ASSP1(num, dir) hex.loc[num].dir = h->vertex[num]->loc.dir
#define ASP(num) { ASSP1(num, x); ASSP1(num, y); ASSP1(num, z); }
#define INIT_H {ASP(0);ASP(1);ASP(2);ASP(3);ASP(4);ASP(5);ASP(6);ASP(7);}

  if (h->offspring != NULL) {
    WEFG_hex_volume_t *found_h, *ho;
    loop_through_list(h->offspring, ho, WEFG_hex_volume_t *) {
      if ((found_h = hex_loc_is_inside (ho, loc, eps, param_loc)) != NULL)
        return found_h;
    }
    return NULL;
  } else { /* No offspring */
    INIT_H;
    if (G3D_hex_inv_map(&hex, loc, eps, param_loc)) {
      if (   IN_BI_UNIT(param_loc->x)
          && IN_BI_UNIT(param_loc->y)
          && IN_BI_UNIT(param_loc->z))
        return h;
      else
        return NULL;
    } else {
      REPORT_WARNING(NULL, "failed to evaluate parametric coordinates");
    }
    return NULL;
  }
}

/*
   Split integration volume.
   Split integration volume into sub-volumes.
   The sub-volumes are listed as offspring by the integration volume.
 */
static void
split_hex(WEFG_superelem_t *se, /* EFG superelement */
          WEFG_hex_volume_t *hp, /* integration volume to subdivide */
          int nsubdiv, /* subdivision level; used only if
                        subdiv_size==0 */
          double subdiv_size /* subdivision size >0 or zero if
                              the above subdivision level should be
                              used instead of size */
          )
{
  WEFG_hex_volume_t *ho;
  WEFG_hex_volume_t *h = hp;
  double xi, eta, tht, dxi, deta, dtht;
  int i, j, k;
  W_loc_t loc;
  W_float_t n[8];
  G3D_box_t range;
  double EPS = se->tolerance;
  WEFG_vertex_t *v;
  int nsubdivx, nsubdivy, nsubdivz;
#define PARENT_MULT 1000000
  static W_index_t offspring_id = 0;
  double hex_length_x;
  double hex_length_y;
  double hex_length_z;
  W_triad_3d_t lb;
  W_vect_t vect;

  hp->offspring = make_list();

  if (subdiv_size <= 0) {
    nsubdivx = nsubdivy = nsubdivz = nsubdiv;
  } else {

    G3D_hex_local_basis(ALL_REF_COORDS, &lb.vect1, &lb.vect2, &lb.vect3);
  
#   define DIFF(f, s) (h->vertex[f]->loc.DIR - h->vertex[s]->loc.DIR)

#   define SUM (DIFF(1,0) + DIFF(2,3) + DIFF(5,4) + DIFF(6,7))
#   define DIR x
    vect.x = 0.25 * (SUM);
#   undef  DIR
#   define DIR y
    vect.y = 0.25 * (SUM);
#   undef  DIR
#   define DIR z
    vect.z = 0.25 * (SUM);
#   undef  DIR
#   undef  SUM
    hex_length_x = fabs(VECMAC_DOT(vect, lb.vect1));

#   define SUM (DIFF(2,1) + DIFF(3,0) + DIFF(6,5) + DIFF(7,4))
#   define DIR x
    vect.x = 0.25 * (SUM);
#   undef  DIR
#   define DIR y
    vect.y = 0.25 * (SUM);
#   undef  DIR
#   define DIR z
    vect.z = 0.25 * (SUM);
#   undef  DIR
#   undef  SUM
    hex_length_y = fabs(VECMAC_DOT(vect, lb.vect2));
  
#   define SUM DIFF(4,0) + DIFF(5,1) + DIFF(6,2) + DIFF(7,3)
#   define DIR x
    vect.x = 0.25 * (SUM);
#   undef  DIR
#   define DIR y
    vect.y = 0.25 * (SUM);
#   undef  DIR
#   define DIR z
    vect.z = 0.25 * (SUM);
#   undef  DIR
#   undef  SUM
    hex_length_z = fabs(VECMAC_DOT(vect, lb.vect3));
#   define COMPUTE_N_SUBDIV(dir)                                   \
     if (hex_length_##dir < subdiv_size) {                         \
       nsubdiv##dir = 1;                                           \
     } else {                                                      \
       nsubdiv##dir = (int)(hex_length_##dir / subdiv_size);       \
     }
    COMPUTE_N_SUBDIV(x);
    COMPUTE_N_SUBDIV(y);
    COMPUTE_N_SUBDIV(z);
  }

  dxi = (double)2 / (double)nsubdivx;
  deta = (double)2 / (double)nsubdivy;
  dtht = (double)2 / (double)nsubdivz;

  xi = -1;
  FOR(i, nsubdivx) {
    eta = -1;
    FOR(j, nsubdivy) {
      tht = -1;
      FOR(k, nsubdivz) {
        ho = w_new(WEFG_hex_volume_t);
        CHECK_MALLOC(ho);
        ho->id = -(PARENT_MULT * hp->id + ++offspring_id);

        MAKE_VERTEX_AT(0, xi     , eta     , tht     );
        MAKE_VERTEX_AT(1, xi+dxi, eta     , tht     );
        MAKE_VERTEX_AT(2, xi+dxi, eta+deta, tht     );
        MAKE_VERTEX_AT(3, xi     , eta+deta, tht     );
        MAKE_VERTEX_AT(4, xi     , eta     , tht+dtht);
        MAKE_VERTEX_AT(5, xi+dxi, eta     , tht+dtht);
        MAKE_VERTEX_AT(6, xi+dxi, eta+deta, tht+dtht);
        MAKE_VERTEX_AT(7, xi     , eta+deta, tht+dtht);

        attach_hex_vol(ho->vertex[0], ho);
        attach_hex_vol(ho->vertex[1], ho);
        attach_hex_vol(ho->vertex[2], ho);
        attach_hex_vol(ho->vertex[3], ho);
        attach_hex_vol(ho->vertex[4], ho);
        attach_hex_vol(ho->vertex[5], ho);
        attach_hex_vol(ho->vertex[6], ho);
        attach_hex_vol(ho->vertex[7], ho);
        {
          int I;
          FOR(I, 6) {
            ho->face_on_boundary[I] = TRUE; 
#ifdef X_XFACE
            ho->face_grep[I] = NULL;
#endif
          }
        }
        ho->quad_order = hp->quad_order;
        ho->qp         = NULL;
        ho->volume     = 0;
        ho->offspring  = NULL;
        add_to_tail(hp->offspring, ho);
        tht += dtht;
      } /* FOR(k, nsubdiv) */
      eta += deta;
    } /* FOR(j, nsubdiv)  */
    xi += dxi;
  } /* FOR(i, nsubdiv) { */
#undef DIFF
#undef SUM
}

/*
   Compute mass of integration volume.
 */
double /* mass */
hex_total_mass(WEFG_superelem_t *se, /* EFG superelement */
               WEFG_hex_volume_t *h /* integration volume */
               )
{
  W_float_t rho, J[3][3];
  W_gauss_integration_rule_t *gr;
  W_gauss_integration_point_t *ip;
  double volume = 0;
  FLAG_2PT_RULE(se, h);

  rho = WM3D_get_volume_mass(se->mat);

#define QR 2
  gr = WG_get_integration_rule(QR); 
  ip = gr->ip;
  IPA_LOOP(QR)
    {
      USE_QUADR_PT(se) {
        G3D_hex_J(J, ip[LOOP_I].Loc, ip[LOOP_J].Loc, ip[LOOP_K].Loc,
                  ALL_REF_COORDS);
        INCL_EXCL_2PT_CASE(
                           { volume += 4 * GP_WGT(ip) * G3D_hex_detJ(J); },
                           { volume += GP_WGT(ip) * G3D_hex_detJ(J); }
                           );
      }
    } 
    IPA_LOOP_END();

  h->volume = volume;
  
  return (rho * h->volume);
}
  

/*
   CAlculate effective loads.
   Evaluate effective loads by summing the contribution
   of body loads.
 */
void
hex_calc_eff_loads(WEFG_superelem_t *se, /* EFG superelement */
                   WEFG_hex_volume_t *h /* integration volume */
                   )
{
  int qo;
  WEFG_integration_point_t **ip, *cip;
  W_gauss_integration_point_t *gp;
  W_float_t J[3][3];
  int i, ne;
  double rho;
  WEFG_hex_volume_t *ho;
  double phi, fact, n[8];
  WEFG_node_t *efgp;
  W_vect_t cbf, f, at;
  double t = WD_time(se->domain);
  FLAG_2PT_RULE(se, h);
#define COMP_AT()                                               \
  {                                                             \
    G3D_hex_N(n, ALL_GP_LOCS);                                  \
    N_LOC(h, at, x, n); N_LOC(h, at, y, n); N_LOC(h, at, z, n); \
  }
#define EVAL_BL()                                               \
  {                                                             \
    int stat;                                                   \
    double v;                                                   \
                                                                \
    COMP_AT();                                                  \
    expeval_set_value("x", at.x);                               \
    expeval_set_value("y", at.y);                               \
    expeval_set_value("z", at.z);                               \
                                                                \
    stat = expeval_eval(se->body_load.body_load_expr[0], &v);   \
    if (stat != EXPEVAL_E_OK) {                                 \
      sprintf(W_buf(), "%s", expeval_err_msgs[stat]);           \
      REPORT_ERROR(se->domain->errh, W_buf());                                    \
    }                                                           \
    cbf.x = v;                                                  \
    stat = expeval_eval(se->body_load.body_load_expr[1], &v);   \
    if (stat != EXPEVAL_E_OK) {                                 \
      sprintf(W_buf(), "%s", expeval_err_msgs[stat]);           \
      REPORT_ERROR(se->domain->errh, W_buf());                                    \
    }                                                           \
    cbf.y = v;                                                  \
    stat = expeval_eval(se->body_load.body_load_expr[2], &v);   \
    if (stat != EXPEVAL_E_OK) {                                 \
      sprintf(W_buf(), "%s", expeval_err_msgs[stat]);           \
      REPORT_ERROR(se->domain->errh, W_buf());                                    \
    }                                                           \
    cbf.z = v;                                                  \
  }

  if (h->offspring != NULL) {
    loop_through_list(h->offspring, ho, WEFG_hex_volume_t *) {
      hex_calc_eff_loads(se, ho);
    }
  } else { /* No offspring -- do the computation */
    renew_quadrature_array(se, h);
    qo = h->qp->qr->order;
    
    rho = WM3D_get_volume_mass(se->mat);

    expeval_set_value("t", t);
    if (se->body_load.is_constant) {
      cbf = se->body_load.const_body_load;
    }

    ip = h->qp->efg_ip;
    gp = h->qp->qr->ip;
    IPA_LOOP(qo)
      {
        USE_QUADR_PT(se) {
          if (!se->body_load.is_constant) { EVAL_BL(); }
          G3D_hex_J(J, ALL_GP_LOCS, ALL_REF_COORDS);
          INCL_EXCL_2PT_CASE({ fact
                                 = 4 * rho * (GP_WGT(gp) * G3D_hex_detJ(J)); },
                             { fact = rho * (GP_WGT(gp) * G3D_hex_detJ(J)); }
                             );
          cip = IPA_ENTRY(ip);
          ne  = SFTAB_NUM_OF_ENTRIES(cip->sftab);
          FOR(i, ne) { /* Row-sum lumping */
            phi = SFTAB_PHI(cip->sftab, i);
            f.x = cbf.x * fact * phi;
            f.y = cbf.y * fact * phi;
            f.z = cbf.z * fact * phi;
            efgp = SFTAB_NODE(cip->sftab, i);
            ASSEMBLE_TO_EFF_LOAD(se, efgp, f);
          }
        }
      }
      IPA_LOOP_END();
  } /* No offspring -- do the computation */
}


/*
   Generate nodes.
   Generate EFG nodes inside the integration volume.
   The nodes are generated either according to the node spacing
   or according to the number of nodes per side length of the integration
   volume.

   If the integration volume is at an interface, the nodes are also generated
   at the vertices of the volume.

   As a side-effect the shortest distance between nodes is updated
   and cached in the superelement.
 */
void
hex_generate_efg_nodes(WEFG_superelem_t *se, /* EFG superelement */
                       WEFG_hex_volume_t *h /* integration volume */
                       )
{
  double hex_length_x;
  double hex_length_y;
  double hex_length_z;
  W_triad_3d_t lb;
  W_vect_t v;
  W_float_t n[8];
  W_float_t xi, eta, tht, dxi, deta, dtht, xis, etas, thts;
  int n_x, n_y, n_z, i;
  G3D_box_t range;
  double drange = se->merging_distance;
  W_node_t *fen;
  W_bool_t is_interface;
#define SUPRNG(at)                              \
  {                                             \
    G3D_INIT_BOX(range, at);                    \
    G3D_INFLATE_BOX(range, drange);             \
  }

  G3D_hex_local_basis(ALL_REF_COORDS, &lb.vect1, &lb.vect2, &lb.vect3);
  
#undef DIFF
#define DIFF(f, s, DIR) (h->vertex[f]->loc.DIR - h->vertex[s]->loc.DIR)

#define SUM(DIR) DIFF(1,0,DIR) + DIFF(2,3,DIR) + DIFF(5,4,DIR) + DIFF(6,7,DIR)
  v.x = 0.25 * (SUM(x));
  v.y = 0.25 * (SUM(y));
  v.z = 0.25 * (SUM(z));
  hex_length_x = fabs(VECMAC_DOT(v, lb.vect1));

#undef  SUM
#define SUM(DIR) DIFF(2,1,DIR) + DIFF(3,0,DIR) + DIFF(6,5,DIR) + DIFF(7,4,DIR)
  v.x = 0.25 * (SUM(x));
  v.y = 0.25 * (SUM(y));
  v.z = 0.25 * (SUM(z));
  hex_length_y = fabs(VECMAC_DOT(v, lb.vect2));
  
#undef  SUM
#define SUM(DIR) DIFF(4,0,DIR) + DIFF(5,1,DIR) + DIFF(6,2,DIR) + DIFF(7,3,DIR)
  v.x = 0.25 * (SUM(x));
  v.y = 0.25 * (SUM(y));
  v.z = 0.25 * (SUM(z));
  hex_length_z = fabs(VECMAC_DOT(v, lb.vect3));

#define COMPUTE_N_(dir)                                         \
  if (hex_length_##dir < se->node_spacing) {                    \
    n_##dir = (se->num_nodes_per_hex_side > 0                   \
           ? se->num_nodes_per_hex_side                         \
           : 1);                                                \
  } else {                                                      \
    n_##dir = (int)(hex_length_##dir / se->node_spacing);       \
    if (se->num_nodes_per_hex_side > 0) {                       \
      if (n_##dir < se->num_nodes_per_hex_side)                 \
        n_##dir = se->num_nodes_per_hex_side;                   \
    }                                                           \
  }

  if (se->node_spacing > 0) {
    COMPUTE_N_(x);
    COMPUTE_N_(y);
    COMPUTE_N_(z);
  } else {
    if (se->num_nodes_per_hex_side > 0) {
      n_x = n_y = n_z = se->num_nodes_per_hex_side;
    } else {
      n_x = n_y = n_z = 1;
    }
  }

  if (se->compute_support_size) { 
    if (se->node_spacing > 0) {
      efg_set_support_size (se, sqrt(3) * se->node_spacing *
                            se->support_size_multiplier);
    } else {
      efg_set_support_size (se,
                            sqrt((hex_length_x / n_x) * (hex_length_x / n_x) +
                                 (hex_length_y / n_y) * (hex_length_y / n_y) +
                                 (hex_length_z / n_z) * (hex_length_z / n_z))
                            * se->support_size_multiplier);
    }
  }

  se->shortest_distance = min(se->shortest_distance, (hex_length_x / n_x));
  se->shortest_distance = min(se->shortest_distance, (hex_length_y / n_y));
  se->shortest_distance = min(se->shortest_distance, (hex_length_z / n_z));

  dxi   = 2.0 / n_x;
  deta  = 2.0 / n_y;
  dtht  = 2.0 / n_z;
  xis   = (se->generate_nodes_inside_only ? -1 + dxi/2  : -1);
  etas  = (se->generate_nodes_inside_only ? -1 + deta/2 : -1);
  thts  = (se->generate_nodes_inside_only ? -1 + dtht/2 : -1);
  for (xi = xis; xi < 1.001; xi += dxi) {
    for (eta = etas; eta < 1.001; eta += deta) {
      for (tht = thts; tht < 1.001; tht += dtht) {
        G3D_hex_N(n, xi, eta, tht);
        N_LOC(h, v, x, n);
        N_LOC(h, v, y, n);
        N_LOC(h, v, z, n);
        if (se->generate_nodes_inside_only) {
          efg_add_node(se, &v);
        } else {
          WEFG_node_t *en;
          SUPRNG(v);
          en = efg_node_in_range (se, &range);
          if (en == NULL)
            efg_add_node (se, &v);
        }
      } /* for (tht */
    } /* for (eta */
  } /* for (xi */
  
  efg_set_max_support_size (se, sqrt(hex_length_x * hex_length_x +
                                     hex_length_y * hex_length_y +
                                     hex_length_z * hex_length_z));
  
  /* Check connectivities; if the hex is at the interface superelement/FE
     generate and adjust types of EFG nodes
   */
  is_interface = FALSE;
  FOR (i, 8) {
    if ((fen = efgv_get_coupling_fen_conn (se, h->vertex[i])) != NULL) {
      is_interface = TRUE;
      if (h->vertex[i]->efgn == NULL) {
        SUPRNG(h->vertex[i]->loc);
        h->vertex[i]->efgn = efg_node_in_range (se, &range);
        if (h->vertex[i]->efgn == NULL) {
          h->vertex[i]->efgn
            = new_efg_node (se, &h->vertex[i]->loc,
                            efg_support_size_at(se,
                                                h->vertex[i]->loc.x,
                                                h->vertex[i]->loc.y,
                                                h->vertex[i]->loc.z));
        }
      }
      efgn_change_to_fe_if_node (h->vertex[i]->efgn, fen);
      efg_attach_node (se, h->vertex[i]->efgn);
    } else if ((fen = efgv_get_driven_fen_conn (se, h->vertex[i])) != NULL) {
      /* driven FE node connection: the FE node is driven by the motion of the vertex */
      if (h->vertex[i]->efgn == NULL) {
        SUPRNG(h->vertex[i]->loc);
        h->vertex[i]->efgn = efg_node_in_range (se, &range);
        if (h->vertex[i]->efgn == NULL) {
          h->vertex[i]->efgn
            = new_efg_node (se, &h->vertex[i]->loc,
                            efg_support_size_at(se,
                                                h->vertex[i]->loc.x,
                                                h->vertex[i]->loc.y,
                                                h->vertex[i]->loc.z));
        }
        efgn_change_to_driving_node (h->vertex[i]->efgn, fen);
      }
      efg_attach_node (se, h->vertex[i]->efgn);
    }
  }
  
  /* If any of the vertices is associated with an EFG node of type */
  /* FE interface node, the hex volume is interfacial */
  if (is_interface) {
    int i;
    WEFG_node_t *exn;

    FOR (i, 8) {
      if (h->vertex[i]->efgn == NULL) {
        SUPRNG((h->vertex[i]->loc));
        exn = efg_node_in_range(se, &range);
        if (exn == NULL)
          h->vertex[i]->efgn 
            = new_efg_node (se, &h->vertex[i]->loc,
                            efg_support_size_at(se,
                                                h->vertex[i]->loc.x,
                                                h->vertex[i]->loc.y,
                                                h->vertex[i]->loc.z));
        else {
          h->vertex[i]->efgn = exn;
        }
        efgn_change_to_efg_if_node (h->vertex[i]->efgn);
      } else if (!EFGN_IS_FE_IF(h->vertex[i]->efgn)) {
        efgn_change_to_efg_if_node (h->vertex[i]->efgn);
      }
      efg_attach_node (se, h->vertex[i]->efgn);
    } /* FOR(i, 8) */
  } /* if (hex_is_interface_vol */
#undef DIFF
#undef SUM
}

static int
num_unique_vertices (WEFG_b_face_t *f)
{
  int i, j;
  int num_uniq = 4;
  FOR(i, 4) {
    for (j = i+1; j<4; j++) {
      if (f->v[i] == f->v[j]) num_uniq--;
    }
  }
  return (num_uniq < 1 ? 1 : num_uniq);
}

/*
   Find boundary faces.
   Find which faces of integration volume are `boundary' faces.
   Any face which is not shared by two adjacent volumes is considered
   to be a boundary face.  Faces of sub-volumes are not considered.
 */
void 
hex_find_b_faces(WEFG_superelem_t *se, /* EFG superelement */
                 WEFG_hex_volume_t *h, /* integration volume */
                 int level /* level of sub-volume (top-level is zero) */
                 )
{
  WEFG_b_face_t f;
  WEFG_hex_volume_t *nh;
  int i, j, shared;
  WEFG_hex_volume_t *ho;
  int next_level = level+1;

  if (h->offspring != NULL) {
    loop_through_list(h->offspring, ho, WEFG_hex_volume_t *) {
      hex_find_b_faces(se, ho, next_level);
    }
  }
  FOR(i, 8) {
    nh = next_incident_hex_vol(h->vertex[i], NULL);
    while (nh != NULL) {
      if (nh != h) { /* The same hexahedron must not be used */
        FOR(j, 6) {
          if (h->face_on_boundary[j]) {
            SET_FACE(j, &f, h);
            if ((shared = shared_face(&f, nh)) >= 0) {
              h->face_on_boundary[j]       = FALSE;
              nh->face_on_boundary[shared] = FALSE;
              goto next_vertex;
            }
          }
        }
      }
      nh = next_incident_hex_vol(h->vertex[i], nh);
    }
  next_vertex:
    ;
  }
  if (level == 0) {   /* Attach only for the parent */
    FOR(j, 6) {
      if ((h->face_on_boundary[j])) {
        SET_FACE(j, &f, h);
        if (num_unique_vertices(&f) > 2) /* only triangles or quadrangles */
          attach_b_face(se, copy_b_face(&f));
      }
    }
  }
}



/* are faces shared?
   Are faces shared by two integration volumes?
 */
static int
faces_shared(WEFG_b_face_t *f1, WEFG_b_face_t *f2)
{
  int i, found = 0;
  WEFG_vertex_t *v;

  FOR(i, 4) {
    v = f1->v[i];
    if        (v == f2->v[0]) {
      found++; 
    } else if (v == f2->v[1]) {
      found++;
    } else if (v == f2->v[2]) {
      found++;
    } else if (v == f2->v[3]) {
      found++;
    }
  }
  return (found == 4);
}


/*
   Is face shared?
   Is the boundary face shared by two integration volumes?
 */
static int
shared_face(WEFG_b_face_t *f, /* boundary face */
            WEFG_hex_volume_t *with_h /* with this integration volume? */
            )
{
  WEFG_b_face_t af;
  int i;

  if (num_unique_vertices(f) < 3) return -1;

  FOR(i, 6) {
    if (with_h->face_on_boundary[i]) {
      SET_FACE(i, &af, with_h);
      if (num_unique_vertices(&af) > 2)
        if (faces_shared(&af, f))
          return i;
    }
  }
  return -1;
}



static void
recompute_ip_array(WEFG_superelem_t *se, WEFG_hex_volume_t *h);


/*
   Recompute quadrature points.
   Re-evaluate shape function tables for the quadrature-point array
   of the integration volume.
 */
static void
renew_quadrature_array(WEFG_superelem_t *se, /* EFG superelement */
                       WEFG_hex_volume_t *h /* integration volume */
                       )
{
  W_bool_t need_to_recompute = FALSE;
  WEFG_quadr_points_t *qp = h->qp;
  int qo;

  if (qp == NULL) {
#     ifdef ENABLE_TWO_POINT_RULE
    if (se->enable_2pt_quadrature) {
      if (h->quad_order == -2)
        qo = 2;
      else
        qo = (h->quad_order > 0 ? h->quad_order : se->quadrature_order);
    } else {
      qo = (h->quad_order > 0 ? h->quad_order : se->quadrature_order);
    } 
#   else
    qo = (h->quad_order > 0 ? h->quad_order : se->quadrature_order);
#   endif

    qp = w_new(WEFG_quadr_points_t);
    CHECK_MALLOC(qp);

    qp->qr            = WG_get_integration_rule(qo);
    qp->efg_ip        = NULL;
    h->qp             = qp;
    need_to_recompute = TRUE;
  } else {
    if (TIME_STAMP_IS_STALE(h->time_stamp, se))
      need_to_recompute = TRUE;
  }
  if (need_to_recompute)
    recompute_ip_array(se, h);
}


/*
   Recompute quadrature points.
   Re-evaluate shape function tables for the quadrature-point array
   of the integration volume.

   Create the quadrature points if there were none; recompute them
   if they need updating.
 */
static void
recompute_ip_array(WEFG_superelem_t *se, /* EFG superelement */
                   WEFG_hex_volume_t *h /* integration volume */
                   )
{
  WEFG_quadr_points_t *qp = h->qp;
  W_gauss_integration_rule_t *qr = qp->qr;
  W_gauss_integration_point_t *gp = qr->ip;
  int qo = qr->order, tot = qo * qo *qo;
  WEFG_integration_point_t *ip;
  W_loc_t at;
  W_float_t n[8];
  W_bool_t is_interface_vol;
  SFL_sf_table_t sf;
#define CAST_IF_DATA (SFL_generic_ptr_t)&if_data
  WEFG_eval_for_packet_t efp;
  FLAG_2PT_RULE(se, h);

  efp.se                               = se;
  efp.pt_wrt_crack_surf_classification = WEFG_POS_WRT_SURF_UNSPEC;
  efp.crack_surf                       = NULL;
  
  sf = efgsf_sf_table(se);

  /* If the quadrature points have not been created yet */
  if (qp->efg_ip == NULL) {
    qp->efg_ip = w_new_array(tot, WEFG_integration_point_t *);
    CHECK_MALLOC(qp->efg_ip);
    is_interface_vol = hex_is_interface_vol(h);
    IPA_LOOP(qo)
      {
        USE_QUADR_PT(se) {
          ip = w_new(WEFG_integration_point_t);
          CHECK_MALLOC(ip); ip->matstate = NULL;
          G3D_hex_N(n, ALL_GP_LOCS);
          N_LOC(h, at, x, n);   
          N_LOC(h, at, y, n);
          N_LOC(h, at, z, n);
          INIT_SFTAB(ip->sftab);
          if (is_interface_vol) { 
            WEFG_if_data_t if_data;
            if_data.h     = h;
            if_data.xi    = gp[LOOP_I].Loc;
            if_data.eta   = gp[LOOP_J].Loc;
            if_data.theta = gp[LOOP_K].Loc;
            if (!SFL_shape_func_at(sf, at.x, at.y, at.z,
                                     &efp, CAST_IF_DATA)) {
              SFL_ERROR(sf, at.x, at.y, at.z, E_K_WARNING);
              ESFL3_set_num_of_entries (sf, 0);
            }
          } else {
            if (!SFL_shape_func_at(sf, at.x, at.y, at.z,
                                     &efp, NULL)) {
              SFL_ERROR(sf, at.x, at.y, at.z, E_K_WARNING);
              ESFL3_set_num_of_entries (sf, 0);
            }
          }
          efgsf_create_sftab (se, &ip->sftab, sf, &at, TRUE);
          UPDATE_EFG_SF_STAT(se, SFTAB_NUM_OF_ENTRIES(ip->sftab));
          ip->matstate = WM3D_get_mat_state(se->mat, &at);
          IPA_ENTRY(qp->efg_ip) = ip;
          efg_add_ip_to_search_structure(efp.se, &at, h, ip);
        }
      }
      IPA_LOOP_END();
  } else {
    is_interface_vol = hex_is_interface_vol(h);
    IPA_LOOP(qo)
      {
        USE_QUADR_PT(se) {
          ip = IPA_ENTRY(qp->efg_ip);
          G3D_hex_N(n, ALL_GP_LOCS);
          N_LOC(h, at, x, n);   
          N_LOC(h, at, y, n);
          N_LOC(h, at, z, n);
          if (SFTAB_INVALID (ip->sftab, se, &at)) {
            USE_SFTAB_SEARCH_FUNC(se, ip->sftab);
            if (is_interface_vol) { 
              WEFG_if_data_t if_data;
              if_data.h     = h;
              if_data.xi    = gp[LOOP_I].Loc;
              if_data.eta   = gp[LOOP_J].Loc;
              if_data.theta = gp[LOOP_K].Loc;
              if (!SFL_shape_func_at (sf, at.x, at.y, at.z,
                                      &efp, CAST_IF_DATA)) {
                SFL_ERROR (sf, at.x, at.y, at.z, E_K_WARNING);
                ESFL3_set_num_of_entries (sf, 0);
              }
            } else {
              if (!SFL_shape_func_at (sf, at.x, at.y, at.z,
                                      &efp, NULL)) {
                SFL_ERROR (sf, at.x, at.y, at.z, E_K_WARNING);
                ESFL3_set_num_of_entries (sf, 0);
              }
            }
            efgsf_create_sftab (se, &ip->sftab, sf, &at, TRUE);
            USE_DEFAULT_SEARCH_FUNC (se);
            UPDATE_EFG_SF_STAT (se, SFTAB_NUM_OF_ENTRIES(ip->sftab));
          } /* if (SFTAB_INVALID(ip->sftab, se, &at)) */
          TOUCH_SFTAB(ip->sftab, se);
        } 
      }
      IPA_LOOP_END();
  }
  TOUCH_TIME_STAMP(h->time_stamp, se);
}
  
  
  
  
#ifdef X_XFACE
static void
modify_face_grep(WEFG_superelem_t *se, double displ_scale, WEFG_hex_volume_t *h, int face_num);

void
hex_modify_elem_grep(WEFG_superelem_t *se, double displ_scale, WEFG_hex_volume_t *h)
{
  int j;
  WEFG_hex_volume_t *ho;

  if (h->offspring != NULL) {
    loop_through_list(h->offspring, ho, WEFG_hex_volume_t *) {
      hex_modify_elem_grep(se, displ_scale, ho);
    }
  } else { /* No offspring */
    FOR(j, 6) {
      if (h->face_on_boundary[j]) {
        modify_face_grep(se, displ_scale, h, j);
      }
    }
  } /* No offspring */
}


#define CURR_LOC(v, loc) vertex_current_loc(se, v, loc)
#define ADJ_LOC(dest, src, vert)                                            \
  {                                                                         \
    (dest).x = vert->loc.x + displ_scale * ((src).x - vert->loc.x); \
    (dest).y = vert->loc.y + displ_scale * ((src).y - vert->loc.y); \
    (dest).z = vert->loc.z + displ_scale * ((src).z - vert->loc.z); \
  }

static EPixel efg_face_color = 0;

static void
modify_face_grep(WEFG_superelem_t *se, double displ_scale, WEFG_hex_volume_t *h, int face_num)
{ 
  WCRec pnts[4];
  WEFG_b_face_t f;
  W_loc_t loc; 

  SET_FACE(face_num, &f, h);

  CURR_LOC(f.v[0], &loc);
  ADJ_LOC(pnts[0], loc, f.v[0]);

  CURR_LOC(f.v[1], &loc);
  ADJ_LOC(pnts[1], loc, f.v[1]);

  CURR_LOC(f.v[2], &loc);
  ADJ_LOC(pnts[2], loc, f.v[2]);

  CURR_LOC(f.v[3], &loc);
  ADJ_LOC(pnts[3], loc, f.v[3]);

  if (h->face_grep[face_num] == NULL) {
    BOOLEAN suc;
    int layer;
    
    h->face_grep[face_num] = CreateQuad3D(pnts);
    efg_face_color = ColorGetPixelFromString("yellow", &suc);
    EASValsSetColor(efg_face_color/* WM_get_color(m) */);
    EASValsSetEdgeColor( ColorGetPixelFromString("black", &suc) );
    EASValsSetEdgeFlag( TRUE );
    layer = EASValsGetLayer();
    EASValsSetLayer(EFG_SURFACE_LAYER);
    EGWithMaskChangeAttributes(COLOR_MASK | EDGE_COLOR_MASK | EDGE_FLAG_MASK | 
                               LAYER_MASK,
                               h->face_grep[face_num]);
    EMNoDrawAddGraphicsToModel(ESIModel (), h->face_grep[face_num]);
    EASValsSetLayer(layer);
  } else {
    EGModifyGraphicsGeometry(h->face_grep[face_num], pnts);
  }
}

#endif /* X_XFACE */



/*
   Assemble mass.
   Assemble mass contribution of this integration volume
   to its nodes (ie. all nodes listed in the shape function tables
   of its quadrature points).
 */
void
hex_assemble_mass(WEFG_superelem_t *se, /* EFG superelement */
                  WEFG_hex_volume_t *h /* integration volume */
                  )
{
  int qo;
  WEFG_integration_point_t **ip, *cip;
  W_gauss_integration_point_t *gp;
  W_float_t J[3][3];
  W_float_t jdetw;
  int i, n;
  double M, rho;
  WEFG_hex_volume_t *ho;
  FLAG_2PT_RULE(se, h);
  WEFG_node_t *efgp;

  if (h->offspring != NULL) {
    loop_through_list(h->offspring, ho, WEFG_hex_volume_t *) {
      hex_assemble_mass(se, ho);
    }
  } else { /* No offspring */
    renew_quadrature_array(se, h);
    qo = h->qp->qr->order;
    
    rho = WM3D_get_volume_mass(se->mat);

    ip = h->qp->efg_ip;
    gp = h->qp->qr->ip;
    IPA_LOOP(qo)
      {
        USE_QUADR_PT(se) {
          G3D_hex_J(J, ALL_GP_LOCS, ALL_REF_COORDS);
          INCL_EXCL_2PT_CASE( { jdetw = 4 * GP_WGT(gp) * G3D_hex_detJ(J); },
                              { jdetw = GP_WGT(gp) * G3D_hex_detJ(J); }
                              );
          cip = IPA_ENTRY(ip);
          n   = SFTAB_NUM_OF_ENTRIES(cip->sftab);
          FOR(i, n) { /* Row-sum lumping */
            M    = rho * jdetw * SFTAB_PHI(cip->sftab, i);
            efgp = SFTAB_NODE(cip->sftab, i);
            ASSEMBLE_TO_MASS(se, efgp, M);
          }
        } 
      }
      IPA_LOOP_END();
  } /* No offspring -- do the computation */
}




#ifdef PATCH_TEST_SPECIAL
#define DUMP_STRESS(se, hex, gplocs, SPK)               \
  {                                                     \
    if (   running_patch_test                   \
        && WD_time(se->domain) > 0) {                   \
      dump_stress_at(hex, gplocs, SPK);                 \
    }                                                   \
  }
static void
dump_stress_at(WEFG_hex_volume_t *hex,
               double xi, double eta, double theta,
               W_float_t SPK[3][3]);
static void
dump_stress_at(WEFG_hex_volume_t *hex,
               double xi, double eta, double tht,
               W_float_t SPK[3][3])
{
  W_float_t n[8];
  W_vect_t v;
  
  G3D_hex_N(n, xi, eta, tht);
  N_LOC(hex, v, x, n);
  N_LOC(hex, v, y, n);
  N_LOC(hex, v, z, n);
  fprintf(stderr, "at %g %g %g   \n", v.x, v.y, v.z);
  fprintf(stderr, "SPK %g %g %g     %g %g     %g\n",
          SPK[0][0], SPK[0][1], SPK[0][2], SPK[1][1], SPK[1][2], SPK[2][2]);
}
#else
#define DUMP_STRESS(se, hex, gplocs, SPK)
#endif



/*
   Assemble restoring forces.
   Assemble restoring forces, ie. forces through which the material in
   the integration volume acts on its nodes.
 */
void
hex_assemble_restoring_forces(WEFG_superelem_t *se, /* EFG superelement */
                              W_mat_state_update_form_enum update_form,
                              WEFG_hex_volume_t *h /* integration volume */
                              )
{
  int qo;
  WEFG_integration_point_t **ip, *cip;
  W_gauss_integration_point_t *gp;
  W_float_t J[3][3];
  W_float_t jdetw;
  int k, n;
  W_vect_t f;
  WEFG_node_t *efgp;
  WEFG_hex_volume_t *ho;
  double dphix, dphiy, dphiz;
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  double P[3][3];
  W_mat_3d_stress_type_enum out_type;
  WEFG_sf_tab_row_t *sftrow;
  FLAG_2PT_RULE(se, h);
#define COMP_F()                                            \
  {                                                         \
    f.x = -(P[0][0]*dphix + P[0][1]*dphiy + P[0][2]*dphiz); \
    f.y = -(P[1][0]*dphix + P[1][1]*dphiy + P[1][2]*dphiz); \
    f.z = -(P[2][0]*dphix + P[2][1]*dphiy + P[2][2]*dphiz); \
  }

  ip_ctx_data.se       = se;
  ip_ctx.time          = WD_time(se->domain);
  ip_ctx.time_step     = WD_time_step(se->domain);
  ip_ctx.ctx_data      = &ip_ctx_data;
  ip_ctx.get_kinem_var = efgss_get_kinem_var;

  if (h->offspring != NULL) {
    loop_through_list(h->offspring, ho, WEFG_hex_volume_t *) {
      hex_assemble_restoring_forces(se, update_form, ho);
    }
  } else { /* No offspring */
    renew_quadrature_array(se, h);
    qo = h->qp->qr->order;
    
    ip = h->qp->efg_ip;
    gp = h->qp->qr->ip;
    IPA_LOOP(qo)
      {
        USE_QUADR_PT(se) {
          G3D_hex_J(J, ALL_GP_LOCS, ALL_REF_COORDS);
          cip               = IPA_ENTRY(ip);
          ip_ctx_data.sftab = cip->sftab;
          EFGSS_AT_NEW_IP();
          if (!WM3D_update_stress(se->mat, cip->matstate, update_form,
                                  &ip_ctx, MAT_3D_1ST_PK, P,
                                  &out_type))
            REPORT_ERROR(se->domain->errh, "Failed to compute stress");
          INCL_EXCL_2PT_CASE( { jdetw = 4 * GP_WGT(gp) * G3D_hex_detJ(J); },
                              { jdetw = GP_WGT(gp) * G3D_hex_detJ(J); }
                              );
          n = SFTAB_NUM_OF_ENTRIES(cip->sftab);
          FOR(k, n) {
            sftrow = &cip->sftab.rows[k];
            dphix = jdetw * SFTAB_ROW_DPHI_X((*sftrow));
            dphiy = jdetw * SFTAB_ROW_DPHI_Y((*sftrow));
            dphiz = jdetw * SFTAB_ROW_DPHI_Z((*sftrow)); 
            COMP_F(); 
            efgp = SFTAB_ROW_NODE((*sftrow));
            ASSEMBLE_TO_EFF_LOAD(se, efgp, f);
          }
        } 
      } 
      IPA_LOOP_END();
  } /* no offspring */
}
  
  
  
  
/*
   Write a HEXVUE post-processing file.
   Write a HEXVUE post-processing file.
 */
void
hex_write_hexvue_postproc_file(WEFG_superelem_t *se, /* EFG superelement */
                               WEFG_hex_volume_t *h, /* integration volume */
                               ckit_iobuf_t IOB, /* I/O buffer */
                               char *mode /* mode (ascii, binary, compressed) */
                               )
{
  WEFG_hex_volume_t *ho;
  W_loc_t cl;
  double S[3][3];
  double equiv_pl_def;
  int reqlen;
  G3D_loc_t *pts[8];
  static int buflen        = 0;
  static double *values[8] = {NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL};
  W_output_stress_type_enum output_stress_type
    = WOPT_OPT (se->domain->options, output_stress_type);

  if (h->offspring != NULL) {
    loop_through_list(h->offspring, ho, WEFG_hex_volume_t *) {
      hex_write_hexvue_postproc_file(se, ho, IOB, mode);
    }
  } else { /* No offspring */
    int i, ival;
    
    reqlen = 0;
    if (WOPT_OPT (se->domain->options, postfile_output_spec_deprec) & W_POSTFILE_DISPL_MASK) 
      reqlen += 3;
    if (WOPT_OPT (se->domain->options, postfile_output_spec_deprec) & W_POSTFILE_STRESS_MASK)
      reqlen += 6;
    if (WOPT_OPT (se->domain->options, postfile_output_spec_deprec) & W_POSTFILE_EQPLS_MASK)
      reqlen++;
    if (WOPT_OPT (se->domain->options, postfile_output_spec_deprec) & W_POSTFILE_VON_MISES_MASK)
      reqlen++;
    
    if (buflen < reqlen) {
      FOR (i, 8) {
        values[i] = NEWP_REALLOC_OBJ_ARRAY (values[i], reqlen, double);
      }
      buflen = reqlen;
    }
    
    FOR(i, 8) {
      pts[i] = &h->vertex[i]->loc;
      ival = 0;
      if (WOPT_OPT (se->domain->options, postfile_output_spec_deprec) & W_POSTFILE_DISPL_MASK) {
        vertex_current_loc(se, h->vertex[i], &cl);
        values[i][ival] = cl.x - h->vertex[i]->loc.x; ival++;
        values[i][ival] = cl.y - h->vertex[i]->loc.y; ival++;
        values[i][ival] = cl.z - h->vertex[i]->loc.z; ival++;
      }
      if (WOPT_OPT (se->domain->options, postfile_output_spec_deprec) & W_POSTFILE_STRESS_MASK) {
        efgv_get_stress(se, h->vertex[i], output_stress_type, S);
        values[i][ival] = S[0][0]; ival++;
        values[i][ival] = S[0][1]; ival++;
        values[i][ival] = S[0][2]; ival++;
        values[i][ival] = S[1][1]; ival++;
        values[i][ival] = S[1][2]; ival++;
        values[i][ival] = S[2][2]; ival++;
      }
      if (WOPT_OPT (se->domain->options, postfile_output_spec_deprec) & W_POSTFILE_EQPLS_MASK) {
        efgv_get_equiv_pl_def(se, h->vertex[i], &equiv_pl_def);
        values[i][ival] = equiv_pl_def; ival++;
      }
      if (WOPT_OPT (se->domain->options, postfile_output_spec_deprec) & W_POSTFILE_VON_MISES_MASK) {
        double means;
        efgv_get_stress(se, h->vertex[i], W_OUTPUT_STRESS_CAUCHY, S);
        means = (S[0][0] + S[1][1] + S[2][2]) / 3;
        S[0][0] -= means; S[1][1] -= means; S[2][2] -= means;
        values[i][ival] = 
          sqrt(1.5
               * (  S[0][0]*S[0][0] + S[0][1]*S[0][1] + S[0][2]*S[0][2]
                  + S[1][0]*S[1][0] + S[1][1]*S[1][1] + S[1][2]*S[1][2]
                  + S[2][0]*S[2][0] + S[2][1]*S[2][1] + S[2][2]*S[2][2]
                  )
               ); ival++;
      }
      if (ival > reqlen) REPORT_FATAL_ERROR (se->domain->errh, "wrong");
    } /* FOR */
    W_write_hexvue_hex (IOB, pts, reqlen, values, mode);
  }
}



/*
   Write a HEXVUE post-processing file.
   Write a HEXVUE post-processing file.
 */
void
hex_write_hexvue_XTRA_postproc_file(WEFG_superelem_t *se, /* EFG superelement */
                                    WEFG_hex_volume_t *h, /* integration volume */
                                    ckit_iobuf_t IOB, /* I/O buffer */
                                    char *mode /* mode (ascii, binary, compressed) */
                                    )
{
  WEFG_hex_volume_t *ho;

  if (h->offspring != NULL) {
    loop_through_list(h->offspring, ho, WEFG_hex_volume_t *) {
      hex_write_hexvue_XTRA_postproc_file(se, ho, IOB, mode);
    }
  } else { /* No offspring */
    W_postfile_output_specs_t *os = &se->domain->options->postfile_output_specs_list;
    W_loc_t cl;
    double equiv_pl_def = 0, von_mises = 0;
    int reqlen;
    G3D_loc_t *pts[8], loc[8];
    int i;
    static double *values[8] = {NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL};
    W_output_stress_type_enum output_stress_type
      = WOPT_OPT (se->domain->options, output_stress_type);
    W_mat_3d_stress_type_enum outtype = (output_stress_type == W_OUTPUT_STRESS_CAUCHY
                                         ? MAT_3D_CAUCHY
                                         : (output_stress_type == W_OUTPUT_STRESS_2ND_PK
                                            ? MAT_3D_2ND_PK
                                            : MAT_3D_CAUCHY)
                                         );
    
    double S[3][3] = {{0,0,0},{0,0,0},{0,0,0}};
    int ival;
    
    reqlen = os->nspecs;
    
    FOR(i, 8) {
      double v = 0;
      loc[i] = h->vertex[i]->loc;
      pts[i] = &loc[i];
      values[i] = NEWP_REALLOC_OBJ_ARRAY (values[i], reqlen, double);
      efgv_get_stress(se, h->vertex[i], W_OUTPUT_STRESS_CAUCHY, S);
      efgv_get_equiv_pl_def(se, h->vertex[i], &equiv_pl_def);
      { /* von Mises */
        double means = (S[0][0] + S[1][1] + S[2][2]) / 3;
        double s00 = S[0][0] - means;
        double s11 = S[1][1] - means;
        double s22 = S[2][2] - means;
        
        von_mises
          = sqrt(1.5 * (s00*s00 + S[0][1]*S[0][1] + S[0][2]*S[0][2] + S[1][0]*S[1][0] + s11*s11
                        + S[1][2]*S[1][2] + S[2][0]*S[2][0] + S[2][1]*S[2][1] + s22*s22));
      }
      FOR (ival, reqlen) {
        if        (tokens_token_and_keyword_equiv (os->specs[ival], "displ_x")) {
          vertex_current_loc(se, h->vertex[i], &cl);
          v = cl.x - h->vertex[i]->loc.x;
        } else if (tokens_token_and_keyword_equiv (os->specs[ival], "displ_y")) { 
          vertex_current_loc(se, h->vertex[i], &cl);
          v = cl.y - h->vertex[i]->loc.y;
        } else if (tokens_token_and_keyword_equiv (os->specs[ival], "displ_z")) { 
          vertex_current_loc(se, h->vertex[i], &cl);
          v = cl.z - h->vertex[i]->loc.z;
        } else { /* quadrature-point data */
          if (tokens_token_and_keyword_equiv (os->specs[ival], "s_xx")) { 
            v = S[0][0];
          } else if (tokens_token_and_keyword_equiv (os->specs[ival], "s_xy")) { 
            v = S[0][1];
          } else if (tokens_token_and_keyword_equiv (os->specs[ival], "s_xz")) { 
            v = S[0][2];
          } else if (tokens_token_and_keyword_equiv (os->specs[ival], "s_yy")) { 
            v = S[1][1];
          } else if (tokens_token_and_keyword_equiv (os->specs[ival], "s_yz")) { 
            v = S[1][2];
          } else if (tokens_token_and_keyword_equiv (os->specs[ival], "s_zz")) { 
            v = S[2][2];
          } else if (tokens_token_and_keyword_equiv (os->specs[ival], "vonm")) { 
            v = von_mises;
          } else if (tokens_token_and_keyword_equiv (os->specs[ival], "press")) { 
            v = (S[0][0] + S[1][1] + S[2][2]) / 3;
          } else if (tokens_token_and_keyword_equiv (os->specs[ival], "eqpls")) { 
            v = equiv_pl_def;
          } else { /* try to get the named variable from the material state */
            v = efgv_get_named_var (se, h->vertex[i], os->specs[ival]);
          }
        }
        (values[i])[ival] = v;
      }
    } /* FOR(i, 8) */
    W_write_hexvue_hex (IOB, pts, reqlen, values, mode);
  }
} 



static W_float_t J[3][3], inv_J[3][3];
static W_float_t N[8], N_xi[8], N_eta[8], N_theta[8], N_x[8], N_y[8], N_z[8];
/*
   Construct ramp.
   Compute the ramp function.
   
   shared data:

   Caching the matrices N, ..., N_z to be used later by hex_fe_shape().
 */
void 
hex_construct_ramp(SFL_generic_ptr_t if_data, /* interface package data */
                   double *R, /* output: value of the ramp function */
                   double D_R[3], /* output: derivatives of the ramp
                                   function
                                   */
                   double x, /* point at which evaluating the ramp */
                   double y, /* point at which evaluating the ramp */
                   double z /* point at which evaluating the ramp */
                   )
{
  WEFG_if_data_t *id = (WEFG_if_data_t *)if_data;
  WEFG_hex_volume_t *h = id->h;
  double xi = id->xi, eta = id->eta, tht = id->theta;

  int i;

  G3D_hex_N(N, xi, eta, tht);
  G3D_hex_J(J, xi, eta, tht, ALL_REF_COORDS);
  G3D_hex_invJ(J, inv_J);
  G3D_hex_N_xi(   N_xi,    xi, eta, tht);
  G3D_hex_N_eta(  N_eta,   xi, eta, tht);
  G3D_hex_N_theta(N_theta, xi, eta, tht);
  G3D_hex_N_x(inv_J, N_xi, N_eta, N_theta, N_x);
  G3D_hex_N_y(inv_J, N_xi, N_eta, N_theta, N_y);
  G3D_hex_N_z(inv_J, N_xi, N_eta, N_theta, N_z);
  *R = D_R[XDir] = D_R[YDir] = D_R[ZDir] = 0;
  FOR(i, 8) {
    if (h->vertex[i]->efgn != NULL) {
      if (EFGN_IS_EFG_IF(h->vertex[i]->efgn)) {
        *R        += N[i];
        D_R[XDir] += N_x[i];
        D_R[YDir] += N_y[i];
        D_R[ZDir] += N_z[i];
      }
    }
  }
}

/*
   Evaluate the FE shape function
   at the node given by /node_num/.

   shared data:

   Reusing the matrices computed by hex_construct_ramp().
 */
void
hex_fe_shape(SFL_generic_ptr_t if_data,
             int  node_num,
             double *n,
             double *nx, double *ny, double *nz,
             double x, double y, double z)
{
  /* NOTE:  Reusing the matrices computed by hex_construct_ramp() */
  *n  = N[node_num];
  *nx = N_x[node_num];
  *ny = N_y[node_num];
  *nz = N_z[node_num];
}

/*
   Which node is it?
   Which node in the interface integration volume
   is the EFG node on input?
 */
int 
hex_elem_node_num(SFL_generic_ptr_t if_data, /* interface package */
                  SFL_efg_node_id_t efgp /* EFG node */
                  )
{
  WEFG_if_data_t *id = (WEFG_if_data_t *)if_data;
  WEFG_hex_volume_t *h = id->h;
  WEFG_node_t *en = (WEFG_node_t *)efgp;
#define HVEFGN(HV, NUM) (HV->vertex[NUM]->efgn)
#define IFRET(NUM) if (en == HVEFGN(h, NUM)) return NUM;

  IFRET(0);
  IFRET(1);
  IFRET(2);
  IFRET(3);
  IFRET(4);
  IFRET(5);
  IFRET(6);
  IFRET(7);

  return -1;
}




/*
   Compute averaged stress.
   Compute averaged stress in an integration volume.

   caveat:

   The averaging technique is extremely primitive.  The stresses
   at the quadrature points are simply summed and then divided by the
   number of quadrature points.
 */
int
hex_averaged_stress(WEFG_superelem_t *se, /* EFG superelement */
                    WEFG_hex_volume_t *h, /* integration volume */
                    W_mat_3d_stress_type_enum requested_out_type, /* type of stress */
                    double stress[3][3] /* output: averaged stress */
                    )
{
  int qo;
  WEFG_integration_point_t **ip, *cip;
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  double P[3][3];
  W_mat_3d_stress_type_enum out_type;
  int nip = 0;
  double one_over_num_of_ip;
  FLAG_2PT_RULE(se, h);

  ip_ctx_data.se       = se;
  ip_ctx.time          = WD_time (se->domain);
  ip_ctx.time_step     = WD_time_step (se->domain);
  ip_ctx.ctx_data      = &ip_ctx_data;
  ip_ctx.get_kinem_var = efgss_get_kinem_var;

  renew_quadrature_array(se, h);
  qo = h->qp->qr->order;

  MAT3DUTIL_ZERO_3X3(stress);
  
  ip  = h->qp->efg_ip;
  nip = 0;
  IPA_LOOP(qo)
    {
      USE_QUADR_PT(se) {
        cip               = IPA_ENTRY(ip);
        ip_ctx_data.sftab = cip->sftab;
        EFGSS_AT_NEW_IP();
        if (!WM3D_update_stress(se->mat, cip->matstate, W_MAT_STATE_UPDATE_TEMP,
                             &ip_ctx, requested_out_type, P,
                             &out_type))
          REPORT_ERROR(se->domain->errh, "Failed to compute stress");
        if (out_type != requested_out_type) {
          REPORT_ERROR(se->domain->errh, "types do not match");
          return 0;
        }
        MAT3DUTIL_INCR_A_BY_B_3X3(stress, P);
        nip++;
      }
    } 
    IPA_LOOP_END();
    one_over_num_of_ip = 1.0 / nip;
    MAT3DUTIL_SCALE_3X3(stress, one_over_num_of_ip);
    return 1;
}


/*
   Evaluate deformation work (energy).
   Evaluate deformation work (energy) within an integration volume.
 */
double
hex_deform_energy(WEFG_superelem_t *se, /* EFG superelement */
                  WEFG_hex_volume_t *h /* integration volume */
                  )
{
  int qo;
  WEFG_integration_point_t **ip, *cip;
  W_gauss_integration_point_t *gp;
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  double J[3][3];
  FLAG_2PT_RULE(se, h);
  double ide, de = 0;
  double jdetw;

  if (h->offspring != NULL) {
    WEFG_hex_volume_t *ho;
    loop_through_list(h->offspring, ho, WEFG_hex_volume_t *) {
      de += hex_deform_energy(se, ho);
    }
  } else { /* No offspring */
    
    ip_ctx_data.se       = se;
    ip_ctx.time          = WD_time(se->domain);
    ip_ctx.time_step     = WD_time_step(se->domain);
    ip_ctx.ctx_data      = &ip_ctx_data;
    ip_ctx.get_kinem_var = efgss_get_kinem_var;
    
    renew_quadrature_array(se, h);
    qo = h->qp->qr->order;
    
    ip  = h->qp->efg_ip;
    gp  = h->qp->qr->ip;
    IPA_LOOP(qo) {
      USE_QUADR_PT(se) {
        cip               = IPA_ENTRY(ip);
        ip_ctx_data.sftab = cip->sftab;
        EFGSS_AT_NEW_IP();
        if (!WM3D_get_mat_state_var(se->mat, cip->matstate,
                                    &ip_ctx, MAT_3D_DEFORM_ENERGY_DENSITY,
                                    &ide))
          REPORT_FATAL_ERROR(se->domain->errh, "Failed to retrieve deformation energy density");
        G3D_hex_J(J, ALL_GP_LOCS, ALL_REF_COORDS);
        INCL_EXCL_2PT_CASE( { jdetw = 4 * GP_WGT(gp) * G3D_hex_detJ(J); },
                            { jdetw = GP_WGT(gp) * G3D_hex_detJ(J); }
                            );
        de += ide * jdetw;
      } 
    } IPA_LOOP_END();
  } /* no offspring */

  return de;
}



/*
   Write TECPLOT post-processing file.
 */
void
hex_write_tecplot_postproc_file(WEFG_hex_volume_t *h, WEFG_superelem_t *se, 
                                ckit_iobuf_t IOB, W_index_t *count)
{
  WEFG_hex_volume_t *ho;
  double S[3][3];
  double equiv_pl_def = 0;
  double vonmises = 0;
  int qo;
  WEFG_integration_point_t **ip, *cip;
  W_gauss_integration_point_t *gp;
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  FLAG_2PT_RULE(se, h);
  double n[8];
  G3D_vect_t rloc, cloc;

  if (h->offspring != NULL) {
    loop_through_list(h->offspring, ho, WEFG_hex_volume_t *) {
      hex_write_tecplot_postproc_file(ho, se, IOB, count);
    }
  } else { /* No offspring */

    renew_quadrature_array(se, h);
    qo = h->qp->qr->order;
    
    ip_ctx_data.se       = se;
    ip_ctx.time          = WD_time (se->domain);
    ip_ctx.time_step     = WD_time_step (se->domain);
    ip_ctx.ctx_data      = &ip_ctx_data;
    ip_ctx.get_kinem_var = efgss_get_kinem_var;

    ip = h->qp->efg_ip;
    gp = h->qp->qr->ip;
    IPA_LOOP(qo) {
      USE_QUADR_PT(se) {
        cip               = IPA_ENTRY(ip);
        ip_ctx_data.sftab = cip->sftab;
        EFGSS_AT_NEW_IP();
        G3D_hex_N(n, ALL_GP_LOCS); 
        N_LOC(h, rloc, x, n); N_LOC(h, rloc, y, n); N_LOC(h, rloc, z, n);
        sprintf (W_buf (), "%g %g %g ", rloc.x, rloc.y, rloc.z);
        ckit_io_put_line (IOB, W_buf ());
        if (WOPT_OPT (se->domain->options, postfile_output_spec_deprec) & W_POSTFILE_DISPL_MASK) {
          efgsf_eval_curr_loc(&cip->sftab, &cloc.x, &cloc.y, &cloc.z);
          sprintf (W_buf (), "%g %g %g ",
                   cloc.x - rloc.x, cloc.y - rloc.y, cloc.z - rloc.z);
          ckit_io_put_line (IOB, W_buf ());
        }
        if (WOPT_OPT (se->domain->options, postfile_output_spec_deprec) & W_POSTFILE_STRESS_MASK ||
            WOPT_OPT (se->domain->options, postfile_output_spec_deprec) & W_POSTFILE_VON_MISES_MASK) {
          W_mat_3d_stress_type_enum out_type;
          double means;
          W_mat_3d_stress_type_enum st
            = (WOPT_OPT (se->domain->options, output_stress_type) == W_OUTPUT_STRESS_CAUCHY ?
               MAT_3D_CAUCHY
               : MAT_3D_2ND_PK);
          if (!WM3D_update_stress(se->mat, cip->matstate, W_MAT_STATE_UPDATE_TEMP, &ip_ctx, st, S, &out_type))
            REPORT_ERROR(se->domain->errh, "Failed to compute stress");
          if (out_type != st) {
            REPORT_ERROR(se->domain->errh, "Stress types do not match");
          }
          sprintf (W_buf (), "%g %g %g %g %g %g",
                   S[0][0], S[0][1], S[0][2], S[1][1], S[1][2], S[2][2]);
          ckit_io_put_line (IOB, W_buf ());
          means = (S[0][0]
                   + S[1][1]
                   + S[2][2]) / 3;
          S[0][0] -= means;
          S[1][1] -= means;
          S[2][2] -= means;
          vonmises
            = sqrt(1.5
                   * (  S[0][0]*S[0][0] + S[0][1]*S[0][1] + S[0][2]*S[0][2]
                      + S[1][0]*S[1][0] + S[1][1]*S[1][1] + S[1][2]*S[1][2]
                      + S[2][0]*S[2][0] + S[2][1]*S[2][1] + S[2][2]*S[2][2]));
        }
        if (WOPT_OPT (se->domain->options, postfile_output_spec_deprec) & W_POSTFILE_EQPLS_MASK) {
          if (!WM3D_get_mat_state_var(se->mat, cip->matstate,
                                      &ip_ctx, MAT_3D_EQUIV_PLASTIC_DEF,
                                      &equiv_pl_def))
            REPORT_FATAL_ERROR(se->domain->errh, "Failed to retrieve equiv plastic def");
          sprintf (W_buf (), "%g", equiv_pl_def);
          ckit_io_put_line (IOB, W_buf ());
        }
        if (WOPT_OPT (se->domain->options, postfile_output_spec_deprec) & W_POSTFILE_VON_MISES_MASK) {
          sprintf (W_buf (), "%g", vonmises);
          ckit_io_put_line (IOB, W_buf ());
        }
        ckit_io_put_line (IOB, "");
        (*count)++;
      } 
    } 
    IPA_LOOP_END();
    
  }
}

  
#endif /* EFG_WASP  */
