#include "Wconfig.h" 
 
#ifdef EFG_WASP

#include "Wtypes.h"
#include "Wdomain.h"
#include "Welem.h"
#include "Wnode.h"
#include "Wmat3d.h"
#include "Wgaussp.h"
#include "Wxf.h"
#include "Wutil.h"
#include "Wcdi.h" 
#include "WEFG.h" 
#include "WEFGutil.h"
#include "WEFGgc.h"
#include "WEFGpriv.h"
#include "WEFGss.h"
#include "WEFGcr.h"
 
/* RCSHEADER "$Id: efgip.c,v 1.1.1.1 2001/07/24 02:42:51 pkrysl Exp $" */

static int tot_pkgs = 0;
static int num_pkgs = 0;
static WEFG_ip_ipol_package_t *pkgs = NULL;
static WEFG_ip_ipol_package_t *oldpkgs = NULL;


#define CLEAR_PKGS()				\
   {						\
     int _i;					\
     FOR(_i, tot_pkgs) {			\
       pkgs[_i].ip_ref = NULL;			\
       pkgs[_i].weight = 0;			\
     }						\
   }
       
#define USE_FUNC_TO_REALLOCATE 0
#if USE_FUNC_TO_REALLOCATE
#define REALLOC_PKGS_BUF(ns)  realloc_pkgs_buf(ns)
#else
#define REALLOC_PKGS_BUF(ns)						\
   {									\
     int new_size = ns;							\
     									\
     if (pkgs == NULL) {						\
       tot_pkgs = max(new_size, 3);					\
       pkgs     = w_new_array(tot_pkgs, WEFG_ip_ipol_package_t);	\
       CHECK_MALLOC(pkgs);						\
     } else {								\
       if (new_size >= tot_pkgs) {					\
	 int _i;							\
	 								\
	 tot_pkgs = new_size + 3;					\
	 oldpkgs  = pkgs;						\
	 pkgs     = w_new_array(tot_pkgs, WEFG_ip_ipol_package_t);	\
	 CHECK_MALLOC(pkgs);						\
	 for (_i = 0; _i < num_pkgs; _i++) {				\
	   pkgs[_i].ip_ref = oldpkgs[_i].ip_ref;			\
	   pkgs[_i].weight = oldpkgs[_i].weight;			\
	 }								\
	 for (_i = num_pkgs; _i < tot_pkgs; _i++) {			\
	   pkgs[_i].ip_ref = NULL;					\
	   pkgs[_i].weight = 0;						\
	 }								\
	 w_free(oldpkgs);						\
       }								\
     }									\
   }
#endif


#if USE_FUNC_TO_REALLOCATE
static void
realloc_pkgs_buf(int ns)
{
  int new_size = ns;

  if (pkgs == NULL) {
    tot_pkgs = max(new_size, 3);
    pkgs     = w_new_array(tot_pkgs, WEFG_ip_ipol_package_t);
    CHECK_MALLOC(pkgs);
  } else {
    if (new_size >= tot_pkgs) {
      int _i;
      
      tot_pkgs = new_size + 3;
      oldpkgs  = pkgs;
      pkgs     = w_new_array(tot_pkgs, WEFG_ip_ipol_package_t);
      CHECK_MALLOC(pkgs);
      for (_i = 0; _i < num_pkgs; _i++) {
	pkgs[_i].ip_ref = oldpkgs[_i].ip_ref;
	pkgs[_i].weight = oldpkgs[_i].weight;
      }
      for (_i = num_pkgs; _i < tot_pkgs; _i++) {
	pkgs[_i].ip_ref = NULL;
	pkgs[_i].weight = 0;
      }
      w_free(oldpkgs);
    }
  }
}
#endif


static W_bool_t
efg_set_up_interpolation(WEFG_superelem_t *se,
			 G3D_box_t *in_box,
			 double hsize
			 );


static W_bool_t
efg_set_up_interpolation(WEFG_superelem_t *se,
			 G3D_box_t *in_box,
			 double hsize
			 )
{
  WEFG_ip_ref_package_t *ippkg;
  G3D_box_t range;

  G3D_INIT_BOX (range, in_box->lower);
  G3D_UPDT_BOX (range, in_box->upper);
  G3D_INFLATE_BOX (range, hsize);
  if (! pirat_search_range (se->ip_ss, &range, INT_MAX)) {
    REPORT_ERROR (se->domain->errh, "Search failed");
    return FALSE; 
  }

  num_pkgs = 0;
  REALLOC_PKGS_BUF(num_pkgs);
  CLEAR_PKGS();
  
  ippkg = (WEFG_ip_ref_package_t *) pirat_first_in_range (se->ip_ss);
  while (ippkg != NULL) {
    pkgs[num_pkgs].ip_ref = ippkg;
    pkgs[num_pkgs].weight = 0;
    num_pkgs++;
    REALLOC_PKGS_BUF(num_pkgs);
    ippkg = (WEFG_ip_ref_package_t *) pirat_next_in_range (se->ip_ss);
  }

  if (num_pkgs <= 0) {
    REPORT_WARNING(se->domain->errh, "Number of integration points too low for interpolation");
    return FALSE;
  }

  return TRUE;
}



static W_bool_t 
intersects_crack_surf(WEFG_superelem_t *se,
		      W_loc_t *evalpt, W_loc_t *pt);


static WEFG_eval_for_packet_t efp;
   

static W_bool_t
efg_interpolation_funcs_at(WEFG_superelem_t *se,
			   W_loc_t *at
			   )
{
  int i;
  W_vect_t cv;
  double d2, d4;
  double c = se->max_support_size * se->max_support_size;
  double ws, iws;
  G3D_box_t box;
  double hsize = se->max_support_size / 4;

  box.lower = box.upper = *at;
  if (!efg_set_up_interpolation(se, &box, hsize))
    return FALSE;

  efp.se = se;
  efp.pt_wrt_crack_surf_classification = WEFG_POS_WRT_SURF_UNSPEC;
  efp.crack_surf = NULL;
  
  FOR(i, num_pkgs) {
    if (pkgs[i].ip_ref == NULL) {
      REPORT_ERROR(se->domain->errh, "NULL among the ip references");
    }
    if (intersects_crack_surf(se, at, &pkgs[i].ip_ref->loc)) {
      pkgs[i].weight = 0;
    } else {
      VECMAC_3_VEC(cv, =, pkgs[i].ip_ref->loc, -, (*at));
      d2 = VECMAC_DOT(cv, cv);
      d4 = d2 * d2;
      pkgs[i].weight = 1 / (d4 + c);
    }
  }
  ws = 0;
  FOR(i, num_pkgs) {
    ws += pkgs[i].weight;
  }
  if (ws == 0)
    return FALSE;
  
  iws = 1/ws;
  FOR(i, num_pkgs) {
    pkgs[i].weight *= iws;
  }
  return TRUE;
}



static W_bool_t 
intersects_crack_surf(WEFG_superelem_t *se,
		      W_loc_t *evalpt, W_loc_t *pt)
{
  WEFG_crack_surf_t *cs;
  G3D_segment_t s;

  s.start = *evalpt;
  s.end   = *pt;
  
  loop_through_list(se->cracks, cs, WEFG_crack_surf_t *) {
    if (efgcr_segment_intersects_crack_surf(&efp, &s, NULL, cs, 0))
      return TRUE;
  }
  return FALSE;
}


W_bool_t
efg_interpolate_at(WEFG_superelem_t *se,
		   W_loc_t *at,
		   void *options,
		   WEFG_update_field_pkg_proc update_pkg,
		   WEFG_field_package_t *result
	) 
{
  int i;
  
  if (!efg_interpolation_funcs_at(se, at))
    return FALSE;

  FOR(i, num_pkgs) {
    (*update_pkg)(options, se, pkgs[i].ip_ref, pkgs[i].weight, result);
  } 

  return TRUE;
}


#endif /* EFG_WASP  */
