#include "Wconfig.h"

#ifdef EFG_WASP

#include "Wtypes.h"
#include "Wdomain.h"
#include "Wmat3d.h"
#include "Wmat3dutil.h"
#include "Wgaussp.h"
#include "Wxf.h"
#include "Wutil.h"
#include "Wcdi.h"
#include "WEFG.h"
#include "WEFGsf.h"
#include "WEFGutil.h"
#include "WEFGpriv.h"
#include "WEFGss.h"
#include "Wnode.h"
#include "g3d.h"
#include "g3dhex.h"

/* RCSHEADER "$Id: efgj.c,v 1.4 2003/11/25 16:55:47 cvs Exp $" */

static void
hex_j_contrib(WEFG_hex_volume_t *h,
	      WEFG_superelem_t *se);
static void
hex_j_contrib_2x2x2(WEFG_hex_volume_t *h,
		    WEFG_superelem_t *se);
static void
determine_bounding_box(void);
static void
q_func(void);






typedef struct j_integral_opts_t {
  WEFG_superelem_t               *se;
  W_loc_t                         at;
  W_vect_t                        forward;
  W_vect_t                        tangent;
  W_vect_t                        normal;
  double                          hsize_Xbar_plus;	/* forward */
  double                          hsize_Xbar_minus;	/* forward */
  double                          hsize_Ybar_plus;	/* tangent */
  double                          hsize_Ybar_minus;	/* tangent */
  double                          hsize_Zbar_plus;	/* normal */
  double                          hsize_Zbar_minus;	/* normal */
  G3D_box_t                       bounding_box;
  double                          transf_mat[3][3];
  int                             quad_order;
  W_loc_t                         currpt;
  W_loc_t                         currpt_bar;
  double                          q;
  double                          q_der[3]; 
  double                          rho;
  WEFG_tensor_prod_comp_func      f;
  WEFG_tensor_prod_comp_der_func  f_r;
  int                             qo_for_2x2x2_ipol;
  double                          j_xbar_static;
  double                          j_xbar_dynamic;
  double                          j_zbar_static;
  double                          j_zbar_dynamic;
}              j_integral_opts_t;




static j_integral_opts_t jopts;


void
efgj_j_volume(WEFG_superelem_t *se,
              W_loc_t *at_point,
              W_vect_t *forward,
              W_vect_t *tangent,
              W_vect_t *normal,
              double hsize_forward_plus,
              double hsize_forward_minus,
              double hsize_tangent_plus,
              double hsize_tangent_minus,
              double hsize_normal_plus,
              double hsize_normal_minus,
              WEFG_tensor_prod_comp_func func,
              WEFG_tensor_prod_comp_der_func func_der,
              double *J_forward_static,
              double *J_forward_dynamic,
              double *J_normal_static,
              double *J_normal_dynamic
              )
{
  WEFG_hex_volume_t *h;
  W_loc_t origin = {0,0,0};

  /* the options */
  jopts.se               = se;
  jopts.at               = *at_point;
  jopts.forward          = *forward;
  jopts.tangent          = *tangent;
  jopts.normal           = *normal;
  jopts.hsize_Xbar_plus  = hsize_forward_plus;
  jopts.hsize_Ybar_plus  = hsize_tangent_plus;
  jopts.hsize_Zbar_plus  = hsize_normal_plus;
  jopts.hsize_Xbar_minus = hsize_forward_minus;
  jopts.hsize_Ybar_minus = hsize_tangent_minus;
  jopts.hsize_Zbar_minus = hsize_normal_minus;
  jopts.quad_order       = 3;
  jopts.transf_mat[0][0] = jopts.forward.x;
  jopts.transf_mat[1][0] = jopts.forward.y;
  jopts.transf_mat[2][0] = jopts.forward.z;
  jopts.transf_mat[0][1] = jopts.tangent.x;
  jopts.transf_mat[1][1] = jopts.tangent.y;
  jopts.transf_mat[2][1] = jopts.tangent.z;
  jopts.transf_mat[0][2] = jopts.normal.x;
  jopts.transf_mat[1][2] = jopts.normal.y;
  jopts.transf_mat[2][2] = jopts.normal.z;
  jopts.j_xbar_static    = 0;
  jopts.j_xbar_dynamic   = 0;
  jopts.j_zbar_static    = 0;
  jopts.j_zbar_dynamic   = 0;
  jopts.currpt_bar       = origin;
  jopts.rho              = WM3D_get_volume_mass(jopts.se->mat);
  jopts.qo_for_2x2x2_ipol= 6;
  jopts.f                = func;
  jopts.f_r              = func_der;
  
  determine_bounding_box();
  
  /* evaluate the domain volume integral */
  if (WOPT_OPT (jopts.se->domain->options, j_dom_eval) == W_PER_CELL_IPOL_PTS) {
    loop_through_list(se->hex_volumes, h, WEFG_hex_volume_t *) {
      hex_j_contrib_2x2x2(h, se);
    }
  } else { /* default */
    loop_through_list(se->hex_volumes, h, WEFG_hex_volume_t *) {
      hex_j_contrib(h, se);
    }
  }

  *J_forward_static   = jopts.j_xbar_static;
  *J_forward_dynamic  = jopts.j_xbar_dynamic;
  *J_normal_static    = jopts.j_zbar_static;
  *J_normal_dynamic   = jopts.j_zbar_dynamic;
}

#define ALL_REF_COORDS                                                  \
  h->vertex[0]->loc.x, h->vertex[0]->loc.y, h->vertex[0]->loc.z,        \
  h->vertex[1]->loc.x, h->vertex[1]->loc.y, h->vertex[1]->loc.z,        \
  h->vertex[2]->loc.x, h->vertex[2]->loc.y, h->vertex[2]->loc.z,        \
  h->vertex[3]->loc.x, h->vertex[3]->loc.y, h->vertex[3]->loc.z,        \
  h->vertex[4]->loc.x, h->vertex[4]->loc.y, h->vertex[4]->loc.z,        \
  h->vertex[5]->loc.x, h->vertex[5]->loc.y, h->vertex[5]->loc.z,        \
  h->vertex[6]->loc.x, h->vertex[6]->loc.y, h->vertex[6]->loc.z,        \
  h->vertex[7]->loc.x, h->vertex[7]->loc.y, h->vertex[7]->loc.z

#define ALL_GAUSS_LOCS ip[LOOP_I].Loc, ip[LOOP_J].Loc, ip[LOOP_K].Loc

#define GP_WGT(gp) (gp[LOOP_I].Wgt * gp[LOOP_J].Wgt * gp[LOOP_K].Wgt)

#define N_LOC(h, v, DIR, n)                     \
   v.DIR = (  n[0] * h->vertex[0]->loc.DIR      \
            + n[1] * h->vertex[1]->loc.DIR      \
            + n[2] * h->vertex[2]->loc.DIR      \
            + n[3] * h->vertex[3]->loc.DIR      \
            + n[4] * h->vertex[4]->loc.DIR      \
            + n[5] * h->vertex[5]->loc.DIR      \
            + n[6] * h->vertex[6]->loc.DIR      \
            + n[7] * h->vertex[7]->loc.DIR)




static void
determine_bounding_box(void)
{
  W_loc_t pt;
  W_loc_t Xbar, Ybar, Zbar;

  G3D_INIT_BOX(jopts.bounding_box, jopts.at);

#define INITV(SIGNF, SIGNT, SIGNN, WHEREF, WHERET, WHEREN)                 \
  {                                                                        \
    VECMAC_3_SCALAR(Xbar, =,                                               \
                    (SIGNF jopts.hsize_Xbar_##WHEREF * jopts.forward.x),  \
                    (SIGNF jopts.hsize_Xbar_##WHEREF * jopts.forward.y),  \
                    (SIGNF jopts.hsize_Xbar_##WHEREF * jopts.forward.z)); \
                                                                           \
    VECMAC_3_SCALAR(Ybar, =,                                               \
                    (SIGNT jopts.hsize_Ybar_##WHERET * jopts.tangent.x),  \
                    (SIGNT jopts.hsize_Ybar_##WHERET * jopts.tangent.y),  \
                    (SIGNT jopts.hsize_Ybar_##WHERET * jopts.tangent.z)); \
                                                                           \
    VECMAC_3_SCALAR(Zbar, =,                                               \
                    (SIGNN jopts.hsize_Zbar_##WHEREN * jopts.normal.x),   \
                    (SIGNN jopts.hsize_Zbar_##WHEREN * jopts.normal.y),   \
                    (SIGNN jopts.hsize_Zbar_##WHEREN * jopts.normal.z));  \
    VECMAC_3_SCALAR(pt, =,                                                 \
		  jopts.at.x + Xbar.x + Ybar.x + Zbar.x,                   \
		  jopts.at.y + Xbar.y + Ybar.y + Zbar.y,                   \
		  jopts.at.z + Xbar.z + Ybar.z + Zbar.z);                  \
  }
  
  INITV(-, -, -, minus, minus, minus);
  G3D_UPDT_BOX(jopts.bounding_box, pt);
  INITV(+, -, -, plus, minus, minus);
  G3D_UPDT_BOX(jopts.bounding_box, pt);
  INITV(+, +, -, plus, plus, minus);
  G3D_UPDT_BOX(jopts.bounding_box, pt);
  INITV(-, +, -, minus, plus, minus);
  G3D_UPDT_BOX(jopts.bounding_box, pt);
 
  INITV(-, -, +, minus, minus, plus);
  G3D_UPDT_BOX(jopts.bounding_box, pt);
  INITV(+, -, +, plus, minus, plus);
  G3D_UPDT_BOX(jopts.bounding_box, pt);
  INITV(+, +, +, plus, plus, plus);
  G3D_UPDT_BOX(jopts.bounding_box, pt);
  INITV(-, +, +, minus, plus, plus);
  G3D_UPDT_BOX(jopts.bounding_box, pt);
}


  
  
  
  


#define CBND(hbbox, hex, which, func, N)                       \
  {                                                              \
    hbbox.which.x = func(hbbox.which.x, hex->vertex[N]->loc.x);  \
    hbbox.which.y = func(hbbox.which.y, hex->vertex[N]->loc.y);  \
    hbbox.which.z = func(hbbox.which.z, hex->vertex[N]->loc.z);  \
  }
#define DEFINE_HBBOX(hex, hbbox)                                 \
  {                                                                  \
    hbbox.lower.x = hbbox.upper.x = hex->vertex[0]->loc.x;           \
    hbbox.lower.y = hbbox.upper.y = hex->vertex[0]->loc.y;           \
    hbbox.lower.z = hbbox.upper.z = hex->vertex[0]->loc.z;           \
    CBND(hbbox, hex, lower, min, 1);                                 \
    CBND(hbbox, hex, lower, min, 2);                                 \
    CBND(hbbox, hex, lower, min, 3);                                 \
    CBND(hbbox, hex, lower, min, 4);                                 \
    CBND(hbbox, hex, lower, min, 5);                                 \
    CBND(hbbox, hex, lower, min, 6);                                 \
    CBND(hbbox, hex, lower, min, 7);                                 \
    CBND(hbbox, hex, upper, max, 1);                                 \
    CBND(hbbox, hex, upper, max, 2);                                 \
    CBND(hbbox, hex, upper, max, 3);                                 \
    CBND(hbbox, hex, upper, max, 4);                                 \
    CBND(hbbox, hex, upper, max, 5);                                 \
    CBND(hbbox, hex, upper, max, 6);                                 \
    CBND(hbbox, hex, upper, max, 7);                                 \
  }


#define BOXES_OVERLAP(r1, r2) 				     	\
   (   (r1.lower.x <= r2.upper.x) && (r1.upper.x >= r2.lower.x)	\
    && (r1.lower.y <= r2.upper.y) && (r1.upper.y >= r2.lower.y)	\
    && (r1.lower.z <= r2.upper.z) && (r1.upper.z >= r2.lower.z))


#define TRANSF_V_3(VP_, A_, V_)                                         \
 {                                                                      \
   VP_.x  = (A_[0][0]*V_.x + A_[1][0]*V_.y + A_[2][0]*V_.z);            \
   VP_.y  = (A_[0][1]*V_.x + A_[1][1]*V_.y + A_[2][1]*V_.z);            \
   VP_.z  = (A_[0][2]*V_.x + A_[1][2]*V_.y + A_[2][2]*V_.z);            \
 }


static void
eval_j_contrib(WEFG_integration_point_t *cip,
               W_mat_3d_ip_ctx_t *ip_ctx,
               double *j_xbar_statcontrib, double *j_xbar_dyncontrib,
               double *j_zbar_statcontrib, double *j_zbar_dyncontrib);



static void
hex_j_contrib(WEFG_hex_volume_t *h,
	      WEFG_superelem_t *se)
{
  WEFG_integration_point_t **ip, *cip;
  W_gauss_integration_point_t *gp;
  double jdetw, J[3][3];
  double Nmat[8];
  double j_xbar_statcontrib, j_xbar_dyncontrib;
  double j_zbar_statcontrib, j_zbar_dyncontrib;
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  int qo;
  G3D_box_t hbox;
  WEFG_hex_volume_t *ho;
  FLAG_2PT_RULE(se, h);
#define CURRPT()				        \
  {					        	\
    N_LOC(h, jopts.currpt, x, Nmat);			\
    N_LOC(h, jopts.currpt, y, Nmat);			\
    N_LOC(h, jopts.currpt, z, Nmat);			\
  }
#undef ALL_GAUSS_LOCS 
#define ALL_GAUSS_LOCS gp[LOOP_I].Loc, gp[LOOP_J].Loc, gp[LOOP_K].Loc
#undef GP_WGT
#define GP_WGT(gp) (gp[LOOP_I].Wgt * gp[LOOP_J].Wgt * gp[LOOP_K].Wgt)
 
  DEFINE_HBBOX(h, hbox);
  if (!BOXES_OVERLAP(hbox, jopts.bounding_box))
    return;

  if (h->offspring != NULL) {
    loop_through_list(h->offspring, ho, WEFG_hex_volume_t *) {
      hex_j_contrib(ho, se);
    }
  } else { /* No offspring */

    ip_ctx_data.se       = se;
    ip_ctx.time          = WD_time(se->domain);
    ip_ctx.time_step     = WD_time_step (se->domain);
    ip_ctx.ctx_data      = &ip_ctx_data;
    ip_ctx.get_kinem_var = efgss_get_kinem_var;
    
    qo = h->qp->qr->order;
    ip = h->qp->efg_ip;
    gp = h->qp->qr->ip;
    IPA_LOOP(qo) {
      USE_QUADR_PT(se) {
	G3D_hex_N(Nmat, ALL_GAUSS_LOCS);  
	G3D_hex_J(J, ALL_GAUSS_LOCS, ALL_REF_COORDS);
	cip               = IPA_ENTRY(ip);
	ip_ctx_data.sftab = cip->sftab;
	EFGSS_AT_NEW_IP();
	INCL_EXCL_2PT_CASE( { jdetw = 4 * GP_WGT(gp) * G3D_hex_detJ(J); },
			    { jdetw = GP_WGT(gp) * G3D_hex_detJ(J); }
			    );
	CURRPT();
        if (G3D_PT_INSIDE_BOX(jopts.bounding_box, jopts.currpt)) {
          eval_j_contrib(cip, &ip_ctx,
                         &j_xbar_statcontrib, &j_xbar_dyncontrib,
                         &j_zbar_statcontrib, &j_zbar_dyncontrib);
          jopts.j_xbar_static   += j_xbar_statcontrib * jdetw;
          jopts.j_xbar_dynamic  += j_xbar_dyncontrib * jdetw;
          jopts.j_zbar_static   += j_zbar_statcontrib * jdetw;
          jopts.j_zbar_dynamic  += j_zbar_dyncontrib * jdetw;
        } /* if (G3D_PT_INSIDE_BOX */
      }
    } IPA_LOOP_END();

  } /* No offspring */
}




static void
eval_j_contrib(WEFG_integration_point_t *cip,
               W_mat_3d_ip_ctx_t *ip_ctx,
               double *j_xbar_statcontrib, double *j_xbar_dyncontrib,
               double *j_zbar_statcontrib, double *j_zbar_dyncontrib)
{
  W_loc_t d;
  double displ_grad[3][3];
  double displ_grad_dot[3][3];
  double firstPK[3][3];
  double tmp[3][3];
  W_vect_t vec, vel, acc;
  int i, j;
  double W, K;
  W_mat_3d_stress_type_enum out_type;

  VECMAC_3_VEC(d, =, jopts.currpt, -, jopts.at);
  jopts.currpt_bar.x = VECMAC_DOT(jopts.forward, d);
  jopts.currpt_bar.y = VECMAC_DOT(jopts.tangent, d);
  jopts.currpt_bar.z = VECMAC_DOT(jopts.normal,  d);

  q_func();

  *j_xbar_statcontrib = *j_xbar_dyncontrib =
    *j_zbar_statcontrib = *j_zbar_dyncontrib = 0;
  
  if (jopts.q == 0)
    return; /* evaluation point is outside of q support */

  /* Get displacement gradient at point */
  if (!(ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, displ_grad)) {
    REPORT_ERROR(jopts.se->domain->errh, "No displacement gradient");
    return;
  }
  displ_grad[0][0] -= 1;
  displ_grad[1][1] -= 1;
  displ_grad[2][2] -= 1;
  
  /* Get F_dot (material time derivative of the
     displacement gradient at point */
  if (!(ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD_DOT, displ_grad_dot)) {
    REPORT_ERROR(jopts.se->domain->errh, "No displacement gradient");
    return;
  }
  
  /* Get 1st Piola-Kirchhoff stress at point */
  if (!WM3D_update_stress(jopts.se->mat, cip->matstate, W_MAT_STATE_UPDATE_TEMP,
		       ip_ctx, MAT_3D_1ST_PK, firstPK,
		       &out_type)) {
    REPORT_ERROR(jopts.se->domain->errh, "Failed to compute stress");
    return;
  }

  /* Deformation energy (strain energy) at point */
  if (!WM3D_get_mat_state_var(jopts.se->mat, cip->matstate,
			      ip_ctx, MAT_3D_DEFORM_ENERGY_DENSITY, &W)) {
    REPORT_ERROR(jopts.se->domain->errh, "Failed to retrieve deformation energy");
    return;
  }

  /* Kinetic energy */
  efgsf_eval_curr_vel(&cip->sftab, &vec.x, &vec.y, &vec.z);
  TRANSF_V_3(vel, jopts.transf_mat, vec);
  efgsf_eval_curr_acc(&cip->sftab, &vec.x, &vec.y, &vec.z);
  TRANSF_V_3(acc, jopts.transf_mat, vec);
  K = 0.5 * jopts.rho * VECMAC_DOT(vel, vel);

  /* Transform the data to the barred coordinate system */
  MAT3DUTIL_MULT_ATB_3X3(tmp, /* = */ jopts.transf_mat, /* * */ displ_grad);
  MAT3DUTIL_MULT_AB_3X3(displ_grad, /* = */ tmp, /* * */ jopts.transf_mat);

  MAT3DUTIL_MULT_ATB_3X3(tmp, /* = */ jopts.transf_mat, /* * */ firstPK);
  MAT3DUTIL_MULT_AB_3X3(firstPK, /* = */ tmp, /* * */ jopts.transf_mat);

  MAT3DUTIL_MULT_ATB_3X3(tmp, /* = */jopts.transf_mat, /* * */displ_grad_dot);
  MAT3DUTIL_MULT_AB_3X3(displ_grad_dot, /* = */tmp, /* * */jopts.transf_mat);

  /*
    Evaluate the product P_ij * u_i,X_k * q_k,X_j  
    where for the mode I opening k=0 (no summation on k)
   */
  FOR(i, 3) {
    FOR(j, 3) {
      *j_xbar_statcontrib +=
        firstPK[i][j] * displ_grad[i][0] * jopts.q_der[j];
      *j_zbar_statcontrib +=
        firstPK[i][j] * displ_grad[i][2] * jopts.q_der[j];
    }
  }

  /* The strain energy:  -W * q_k,X_k */
  *j_xbar_statcontrib += (-W) * jopts.q_der[0];
  *j_zbar_statcontrib += (-W) * jopts.q_der[2];

  /* The kinetic energy: -K * q_k,X_k  */
  *j_xbar_dyncontrib = (-K) * jopts.q_der[0];
  *j_xbar_dyncontrib += jopts.rho * (acc.x * displ_grad[0][0] +
                                     acc.y * displ_grad[1][0] +
                                     acc.z * displ_grad[2][0]) * jopts.q;
  *j_xbar_dyncontrib -= (jopts.rho * (vel.x * displ_grad_dot[0][0] +
                                      vel.y * displ_grad_dot[1][0] +
                                      vel.z * displ_grad_dot[2][0]) * jopts.q);
  
  *j_zbar_dyncontrib = (-K) * jopts.q_der[2];
  *j_zbar_dyncontrib += jopts.rho * (acc.x * displ_grad[0][2] +
                                     acc.y * displ_grad[1][2] +
                                     acc.z * displ_grad[2][2]) * jopts.q;
  *j_zbar_dyncontrib -= (jopts.rho * (vel.x * displ_grad_dot[0][2] +
                                      vel.y * displ_grad_dot[1][2] +
                                      vel.z * displ_grad_dot[2][2]) * jopts.q);
  
  return;
}









static void
q_func(void)
{
  double xs = (jopts.currpt_bar.x > 0
               ? jopts.hsize_Xbar_plus
               : jopts.hsize_Xbar_minus);
  double ys = (jopts.currpt_bar.y > 0
               ? jopts.hsize_Ybar_plus
               : jopts.hsize_Ybar_minus);
  double zs = (jopts.currpt_bar.z > 0
               ? jopts.hsize_Zbar_plus
               : jopts.hsize_Zbar_minus);
  double xbar = jopts.currpt_bar.x / xs;
  double ybar = jopts.currpt_bar.y / ys;
  double zbar = jopts.currpt_bar.z / zs;
  
  /* Virtual displacement along the forward vector */
  jopts.q = (jopts.f(xbar) * jopts.f(ybar) * jopts.f(zbar));

  /* Derivatives */
  jopts.q_der[0] = (jopts.f_r(xbar) / xs * jopts.f(ybar) * jopts.f(zbar));
  jopts.q_der[1] = (jopts.f(xbar) * jopts.f_r(ybar) / ys * jopts.f(zbar));
  jopts.q_der[2] = (jopts.f(xbar) * jopts.f(ybar) * jopts.f_r(zbar) / zs);
}

  
  
 






static void
eval_j_contrib_2x2x2(double xi, double eta, double theta,
                     double *j_xbar_statcontrib, double *j_xbar_dyncontrib,
                     double *j_zbar_statcontrib, double *j_zbar_dyncontrib);

static void
fill_ip_2x2x2_pkg(double xi, double eta, double theta,
		  WEFG_integration_point_t *cip, W_mat_3d_ip_ctx_t *ip_ctx);
 

typedef struct ipol_2x2x2_pkg_t {
  double          displ_grad[3][3];
  double          displ_grad_dot[3][3];
  W_vect_t        vel, acc;
  double          firstPK[3][3];
  double          W;
  double          K;
}              ipol_2x2x2_pkg_t;


static ipol_2x2x2_pkg_t ipol_2x2x2_pkgs[8];


static void
hex_j_contrib_2x2x2(WEFG_hex_volume_t *h,
		    WEFG_superelem_t *se)
{
  WEFG_integration_point_t **ip, *cip;
  W_gauss_integration_point_t *gp;
  double jdetw, J[3][3];
  double Nmat[8];
  double j_xbar_statcontrib, j_xbar_dyncontrib;
  double j_zbar_statcontrib, j_zbar_dyncontrib;
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  int qo;
  G3D_box_t hbox;
  WEFG_hex_volume_t *ho;
  FLAG_2PT_RULE(se, h);
  W_gauss_integration_rule_t *gr;
#define CURRPT()				        \
  {					        	\
    N_LOC(h, jopts.currpt, x, Nmat);			\
    N_LOC(h, jopts.currpt, y, Nmat);			\
    N_LOC(h, jopts.currpt, z, Nmat);			\
  }
#undef ALL_GAUSS_LOCS 
#define ALL_GAUSS_LOCS gp[LOOP_I].Loc, gp[LOOP_J].Loc, gp[LOOP_K].Loc
#undef GP_WGT
#define GP_WGT(gp) (gp[LOOP_I].Wgt * gp[LOOP_J].Wgt * gp[LOOP_K].Wgt)
 
  DEFINE_HBBOX(h, hbox);
  if (!BOXES_OVERLAP(hbox, jopts.bounding_box))
    return;

  if (h->offspring != NULL) {
    loop_through_list(h->offspring, ho, WEFG_hex_volume_t *) {
      hex_j_contrib_2x2x2(ho, se);
    }
  } else { /* No offspring */

    /* Construct the interpolation */
    ip_ctx_data.se       = se;
    ip_ctx.time          = WD_time (se->domain);
    ip_ctx.time_step     = WD_time_step (se->domain);
    ip_ctx.ctx_data      = &ip_ctx_data;
    ip_ctx.get_kinem_var = efgss_get_kinem_var;

    qo = h->qp->qr->order;
    if (qo != 2) {
      REPORT_FATAL_ERROR(jopts.se->domain->errh, "Interpolated field not applicable");
    }
    ip = h->qp->efg_ip;
    gp = h->qp->qr->ip;
    INC_2PT_RULE(if (use_2pt_rule) {
      REPORT_FATAL_ERROR(jopts.se->domain->errh, "Interpolated field not applicable");
    });
    IPA_LOOP(qo) {
      G3D_hex_N(Nmat, ALL_GAUSS_LOCS);  
      cip               = IPA_ENTRY(ip);
      ip_ctx_data.sftab = cip->sftab;
      EFGSS_AT_NEW_IP();
      fill_ip_2x2x2_pkg(ALL_GAUSS_LOCS, cip, &ip_ctx);
    } IPA_LOOP_END();

    /* Integrate the interpolated field */
    gr       = WG_get_integration_rule(jopts.qo_for_2x2x2_ipol);
    gp       = gr->ip;
    IPA_LOOP(jopts.qo_for_2x2x2_ipol) {
      G3D_hex_N(Nmat, ALL_GAUSS_LOCS);  
      G3D_hex_J(J, ALL_GAUSS_LOCS, ALL_REF_COORDS);
      jdetw = GP_WGT(gp) * G3D_hex_detJ(J);
      CURRPT();
      eval_j_contrib_2x2x2(ALL_GAUSS_LOCS, 
                           &j_xbar_statcontrib, &j_xbar_dyncontrib,
                           &j_zbar_statcontrib, &j_zbar_dyncontrib);
      jopts.j_xbar_static  += j_xbar_statcontrib * jdetw;
      jopts.j_xbar_dynamic += j_xbar_dyncontrib * jdetw;
      jopts.j_zbar_static  += j_zbar_statcontrib * jdetw;
      jopts.j_zbar_dynamic += j_zbar_dyncontrib * jdetw;
    } IPA_LOOP_END();
  } /* No offspring */
  
  return;
}



static void
fill_ip_2x2x2_pkg(double xi, double eta, double theta,
		  WEFG_integration_point_t *cip, W_mat_3d_ip_ctx_t *ip_ctx)
{
  double displ_grad[3][3];
  double displ_grad_dot[3][3];
  double firstPK[3][3];
  double tmp[3][3];
  W_vect_t vel, acc, vec;
  double W, K;
  W_mat_3d_stress_type_enum out_type;
  int entry = -1;
  
  if        (xi < 0 && eta < 0 && theta < 0) {entry = 0;
  } else if (xi > 0 && eta < 0 && theta < 0) {entry = 1;
  } else if (xi > 0 && eta > 0 && theta < 0) {entry = 2;
  } else if (xi < 0 && eta > 0 && theta < 0) {entry = 3;
  } else if (xi < 0 && eta < 0 && theta > 0) {entry = 4;
  } else if (xi > 0 && eta < 0 && theta > 0) {entry = 5;
  } else if (xi > 0 && eta > 0 && theta > 0) {entry = 6;
  } else if (xi < 0 && eta > 0 && theta > 0) {entry = 7;
  }

  /* Get displacement gradient at point */
  if (!(ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, displ_grad)) {
    REPORT_ERROR(jopts.se->domain->errh, "No displacement gradient");
    return;
  }
  displ_grad[0][0] -= 1;
  displ_grad[1][1] -= 1;
  displ_grad[2][2] -= 1;
  
  /* Get F_dot (material time derivative of the
     displacement gradient at point */
  if (!(ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD_DOT, displ_grad_dot)) {
    REPORT_ERROR(jopts.se->domain->errh, "No displacement gradient");
    return;
  }
  
  /* Get 1st Piola-Kirchhoff stress at point */
  if (!WM3D_update_stress(jopts.se->mat, cip->matstate, W_MAT_STATE_UPDATE_TEMP,
		       ip_ctx, MAT_3D_1ST_PK, firstPK,
		       &out_type)) {
    REPORT_ERROR(jopts.se->domain->errh, "Failed to compute stress");
    return;
  }

  /* Deformation energy (strain energy) at point */
  if (!WM3D_get_mat_state_var(jopts.se->mat, cip->matstate,
			      ip_ctx, MAT_3D_DEFORM_ENERGY_DENSITY,
			      &W)) {
    REPORT_ERROR(jopts.se->domain->errh, "Failed to retrieve deformation energy");
    return;
  }

  /* Kinetic energy */
  efgsf_eval_curr_vel(&cip->sftab, &vec.x, &vec.y, &vec.z);
  TRANSF_V_3(vel, jopts.transf_mat, vec);
  efgsf_eval_curr_acc(&cip->sftab, &vec.x, &vec.y, &vec.z);
  TRANSF_V_3(acc, jopts.transf_mat, vec);
  K = 0.5 * jopts.rho * VECMAC_DOT(vel, vel);

  /* Transform the data to the barred coordinate system */
  MAT3DUTIL_MULT_ATB_3X3(tmp, /* = */ jopts.transf_mat, /* * */ displ_grad);
  MAT3DUTIL_MULT_AB_3X3(displ_grad, /* = */ tmp, /* * */ jopts.transf_mat);

  MAT3DUTIL_MULT_ATB_3X3(tmp, /* = */ jopts.transf_mat, /* * */ firstPK);
  MAT3DUTIL_MULT_AB_3X3(firstPK, /* = */ tmp, /* * */ jopts.transf_mat);

  MAT3DUTIL_MULT_ATB_3X3(tmp, /* = */jopts.transf_mat, /* * */displ_grad_dot);
  MAT3DUTIL_MULT_AB_3X3(displ_grad_dot, /* = */tmp, /* * */jopts.transf_mat);

  /* Store into the package */
  MAT3DUTIL_COPY_3X3(ipol_2x2x2_pkgs[entry].displ_grad, displ_grad);
  MAT3DUTIL_COPY_3X3(ipol_2x2x2_pkgs[entry].displ_grad_dot, displ_grad_dot);
  MAT3DUTIL_COPY_3X3(ipol_2x2x2_pkgs[entry].firstPK, firstPK);
  ipol_2x2x2_pkgs[entry].vel = vel;
  ipol_2x2x2_pkgs[entry].acc = acc;
  ipol_2x2x2_pkgs[entry].W = W;
  ipol_2x2x2_pkgs[entry].K = K;
}



static void
eval_j_contrib_2x2x2(double xi, double eta, double theta,
                     double *j_xbar_statcontrib, double *j_xbar_dyncontrib,
                     double *j_zbar_statcontrib, double *j_zbar_dyncontrib)
{
  W_loc_t d;
  double displ_grad[3][3];
  double displ_grad_dot[3][3];
  double firstPK[3][3];
#define SQRT_3 1.73205080756887729352
  double modxi    = SQRT_3 * xi;
  double modeta   = SQRT_3 * eta;
  double modtheta = SQRT_3 * theta;
  double Nmat[8];
  double W, K;
  W_vect_t vel, acc;
  int i, j;

  VECMAC_3_VEC(d, =, jopts.currpt, -, jopts.at);
  jopts.currpt_bar.x = VECMAC_DOT(jopts.forward, d);
  jopts.currpt_bar.y = VECMAC_DOT(jopts.tangent, d);
  jopts.currpt_bar.z = VECMAC_DOT(jopts.normal,  d);

  q_func(); 

  *j_xbar_statcontrib = *j_xbar_dyncontrib =
    *j_zbar_statcontrib = *j_zbar_dyncontrib = 0;
  
  if (jopts.q == 0)
    return; /* evaluation point is outside of q support */

  /* Interpolate */
  G3D_hex_N(Nmat, modxi, modeta, modtheta);
#define COMP_T1(T, I, J)				\
  { T[I][J] = (						\
	       Nmat[0] * ipol_2x2x2_pkgs[0].T[I][J] +	\
	       Nmat[1] * ipol_2x2x2_pkgs[1].T[I][J] +	\
	       Nmat[2] * ipol_2x2x2_pkgs[2].T[I][J] +	\
	       Nmat[3] * ipol_2x2x2_pkgs[3].T[I][J] +	\
	       Nmat[4] * ipol_2x2x2_pkgs[4].T[I][J] +	\
	       Nmat[5] * ipol_2x2x2_pkgs[5].T[I][J] +	\
	       Nmat[6] * ipol_2x2x2_pkgs[6].T[I][J] +	\
	       Nmat[7] * ipol_2x2x2_pkgs[7].T[I][J]	\
	       );					\
  }
#define COMP_T(T)						\
  {								\
    COMP_T1(T, 0, 0); COMP_T1(T, 0, 1); COMP_T1(T, 0, 2);	\
    COMP_T1(T, 1, 0); COMP_T1(T, 1, 1); COMP_T1(T, 1, 2);	\
    COMP_T1(T, 2, 0); COMP_T1(T, 2, 1); COMP_T1(T, 2, 2);	\
  }
#define COMP_S(S)				\
  { S = (					\
	 Nmat[0] * ipol_2x2x2_pkgs[0].S +	\
	 Nmat[1] * ipol_2x2x2_pkgs[1].S +	\
	 Nmat[2] * ipol_2x2x2_pkgs[2].S +	\
	 Nmat[3] * ipol_2x2x2_pkgs[3].S +	\
	 Nmat[4] * ipol_2x2x2_pkgs[4].S +	\
	 Nmat[5] * ipol_2x2x2_pkgs[5].S +	\
	 Nmat[6] * ipol_2x2x2_pkgs[6].S +	\
	 Nmat[7] * ipol_2x2x2_pkgs[7].S		\
	 );					\
  }
#define COMP_V1(V, DIR)                         \
  { V.DIR = (                                   \
           Nmat[0] * ipol_2x2x2_pkgs[0].V.DIR + \
           Nmat[1] * ipol_2x2x2_pkgs[1].V.DIR + \
           Nmat[2] * ipol_2x2x2_pkgs[2].V.DIR + \
           Nmat[3] * ipol_2x2x2_pkgs[3].V.DIR + \
           Nmat[4] * ipol_2x2x2_pkgs[4].V.DIR + \
           Nmat[5] * ipol_2x2x2_pkgs[5].V.DIR + \
           Nmat[6] * ipol_2x2x2_pkgs[6].V.DIR + \
           Nmat[7] * ipol_2x2x2_pkgs[7].V.DIR   \
           );                                   \
  }
#define COMP_V(V) { COMP_V1(V, x); COMP_V1(V, y); COMP_V1(V, z); }

  COMP_T(displ_grad);
  COMP_T(displ_grad_dot);
  COMP_T(firstPK);
  COMP_V(vel);
  COMP_V(acc);
  COMP_S(W);
  COMP_S(K);
  
  /*
    Evaluate the product P_ij * u_i,X_k * q_k,X_j  
    where for the mode I opening k=0 (no summation on k)
   */
  FOR(i, 3) {
    FOR(j, 3) {
      *j_xbar_statcontrib +=
        firstPK[i][j] * displ_grad[i][0] * jopts.q_der[j];
      *j_zbar_statcontrib +=
        firstPK[i][j] * displ_grad[i][2] * jopts.q_der[j];
    }
  }

  /* The strain energy:  -W * q_k,X_k */
  *j_xbar_statcontrib += (-W) * jopts.q_der[0];
  *j_zbar_statcontrib += (-W) * jopts.q_der[2];

  /* The kinetic energy: -K * q_k,X_k  */
  *j_xbar_dyncontrib = (-K) * jopts.q_der[0];
  *j_xbar_dyncontrib +=    jopts.rho * (acc.x * displ_grad[0][0] +
                                        acc.y * displ_grad[1][0] +
                                        acc.z * displ_grad[2][0]) * jopts.q;
  *j_xbar_dyncontrib += -(jopts.rho *
                          (vel.x * displ_grad_dot[0][0] +
                           vel.y * displ_grad_dot[1][0] +
                           vel.z * displ_grad_dot[2][0]) * jopts.q);
  
  *j_zbar_dyncontrib = (-K) * jopts.q_der[2];
  *j_zbar_dyncontrib += jopts.rho * (acc.x * displ_grad[0][2] +
                                     acc.y * displ_grad[1][2] +
                                     acc.z * displ_grad[2][2]) * jopts.q;
  *j_zbar_dyncontrib += -(jopts.rho *
                          (vel.x * displ_grad_dot[0][2] +
                           vel.y * displ_grad_dot[1][2] +
                           vel.z * displ_grad_dot[2][2]) * jopts.q);
  
  return;
}

 

#endif /* EFG_WASP  */
