#include "Wconfig.h"

#ifdef EFG_WASP

#include "Wtypes.h"
#include "WEFG.h"
#include "WEFGsf.h"
#include "Wdomain.h"
#include "Wmats.h"
#include "Wgaussp.h"
#include "Wxf.h"
#include "Wutil.h"
#include "Wcdi.h"
#include "WEFGutil.h"
#include "WEFGpriv.h"



void
read_symm_plane(WEFG_superelem_t *se, W_domain_t *d)
{
  const char *input_line;
  WEFG_symm_plane_t *sp;

  /* Syntax: */
  /* SYMMPLANE dir cx cy cz [tol] */
  /* dir = x|y|z|X|Y|Z */
  /* cx,cy,cz = center coordinates */
  /* tol = tolerance.  If the distance of a node is less than /tol/, */
  /*       the node is considered to be located on the symmetry plane. */
  
  while ((input_line = W_get_input_line(d)) != NULL) {
    TOKENS_PARSE(input_line);
    if (TOKENS_HAVE_KEYWORD("END")) {
      if (TOKENS_HAVE_KEYWORD("SYMMPLANE"))
        return;
      else
        REPORT_FATAL_ERROR(se->domain->errh, "SYMMPLANE does not end by END SYMMPLANE");
    } else {
      sp = w_new(WEFG_symm_plane_t);
      CHECK_MALLOC(sp);
      sp->direction_of_normal = toupper(TOKEN_STR(2)[0]);
      if (!(   sp->direction_of_normal == 'X'
            || sp->direction_of_normal == 'Y'
            || sp->direction_of_normal == 'Z')) {
        REPORT_FATAL_ERROR(se->domain->errh, "Bad normal of symmetry plane");
      }
      sp->center.x = TOKEN_DBL(3);
      sp->center.y = TOKEN_DBL(4);
      sp->center.z = TOKEN_DBL(5);
      if (TOKENS_TOTAL() == 6) {
        sp->tolerance = TOKEN_DBL(6);
      } else {
        sp->tolerance = 0; /* In this case se->tolerance applies */
      }
      attach_symm_plane(se, sp);
    }
  }
}



#endif /* EFG_WASP  */
