#include "Wconfig.h"

#ifdef EFG_WASP

#if 0

#include "WEFGwf.h"
#include "Wdomain.h"
#include "Welem.h"
#include "Wnode.h"
#include "Wmats.h"
#include "Wgaussp.h"
#include "Wxf.h"
#include "Wutil.h"
#include "EID.h"


/* This is the quartic spline weight function.  Definition: */
/* w(r) = 1 - 6r^2 + 8r^3 - 3r^4, with r = sqrt(dksi^2 + deta^2)/dist_max */

double                              /* OUT weight function */
wf_qs_w(
        double dksi,
        double deta,
        double dtheta, /* IN dksi = ksi - ksiI, deta = eta - etaI, */
                       /* dtheta = theta - thetaI */
        double dist_max /* IN max distance */
        )
{
  double dist = sqrt(dksi * dksi + deta * deta + dtheta * dtheta);
  double tmp = (dist + dist) / dist_max;
  /* Quartic spline function */
  if (dist >= dist_max)
    return 0;
  else {
    tmp = (1 + ((-1.5 + (1 - 0.1875 * tmp) * tmp) * tmp) * tmp);
    return tmp > 0? tmp: 0;     /* catch round-off errors */
  }
}


void                         
wf_qs_D1(
         double dksi,     /* IN dksi = ksi - ksiI */
         double deta,     /* IN deta = eta - etaI */
         double dtheta,   /* IN dtheta = theta - thetaI */
         double dist_max, /* IN max distance */
         double der[3]    /* OUT der[0] = dwdksi, */
                          /*     der[1] = dwdeta  */
                          /*     der[2] = dwdtheta  */
         )
{
  double dist = sqrt(dksi * dksi + deta * deta + dtheta * dtheta);
  double oodm = 1.0 / dist_max;
  double mu   = (dist + dist) * oodm;
  double muex = -1.0 + mu - 0.25 * mu * mu;
  double todm2 = 12.0 * oodm * oodm;
  
  /* Quartic spline function derivative */
  if (dist >= dist_max) {
    der[0] = der[1] = der[2] = 0;
  } else {
    der[0]  = /* dwdksi   */ todm2 * dksi   * muex;
    der[1]  = /* dwdksi   */ todm2 * deta   * muex;
    der[2]  = /* dwdtheta */ todm2 * dtheta * muex;
  }
}

#endif /* if 0 */
#endif /* EFG_WASP */
