#include "Wconfig.h"
#include "Wxf.h"
#include "Welem.h"
#include "Whistry.h" 

/* RCSHEADER "$Id: elem.c,v 1.13 2005/02/14 02:47:23 cvs Exp $" */

typedef struct reader_intermediary_t {
  W_elem_reader_func  reader;
}              reader_intermediary_t;

static dict_dictionary reader_intermediary_table = NULL;


static W_elem_reader_func
find_reader (int elcode)
{
  reader_intermediary_t *ri = NULL;
  
  ri = dict_find_obj_with_int_name (reader_intermediary_table, elcode);
  if (ri == NULL) {
    sprintf (W_buf (), "Reader for element type %d not registered",
             elcode);
    REPORT_ERROR (NULL, W_buf ());
  }
  return ri->reader;
}


static void 
initialize_readers_for_builtins (void)
{
  
#define W_DCL_READER(reader) W_elem_reader_func reader
#define W_REGISTER_READER(elcode, reader) WE_register_reader (elcode, reader)
#include "Wtruss.h"
#include "Wtnbeam.h"
#include "Wqmem4.h"
#include "Wasqbi3d.h"
#include "Whex.h"

#include "Wahex.h"
#include "Wahex20.h"

#include "WEFGpub.h"
#include "Wtet10.h"
#include "Wftet10.h"
#include "Wfhex.h"
#include "Wpcont3.h"
#include "Wtet4.h"
#include "Wtet4amc.h"
#include "Wntruss.h"
#include "Wcmass.h"
#include "Wtmem3.h"
#include "Wt3press.h"
#include "Wq4press.h"
#include "Wq8press.h"
#include "Whex20.h"
#include "Whex20aq.h"
#include "Weulhex.h"
#include "Wqmem4c.h"
  /* #include "Whex20unsymm.h" */
#include "Whexwfbar.h"
#include "Wpcontpt.h"
#include "Wspring.h"
#include "Whexssbbar.h"
   W_REGISTER_READER (TRUSS_ELCODE, truss_reader);
# if W_ROT_DOFS
   W_REGISTER_READER (TNBEAM_ELCODE, tnbeam_reader);
# endif
   W_REGISTER_READER (QMEM4_ELCODE, qmem4_reader);
   W_REGISTER_READER (ASQBI3D_ELCODE, asqbi3d_reader);
#ifdef EFG_WASP
   W_REGISTER_READER (EFG_SUPERELEM_ELCODE, efg_reader);
#endif
   W_REGISTER_READER (HEX_ELCODE, hex_reader);

   W_REGISTER_READER (AHEX_ELCODE, ahex_reader);

   W_REGISTER_READER (HEX20_ELCODE, hex20_reader);
   W_REGISTER_READER (AHEX20_ELCODE, ahex20_reader);
   /*   W_REGISTER_READER (HEX20UNSYM_ELCODE, hex20unsymm_reader); */
   W_REGISTER_READER (HEX20AQ_ELCODE, hex20aq_reader);
   W_REGISTER_READER (TET10_ELCODE, tet10_reader);
   W_REGISTER_READER (FTET10_ELCODE, ftet10_reader);
   W_REGISTER_READER (FHEX_ELCODE, fhex_reader);
   W_REGISTER_READER (PCONT3_ELCODE, pcont3_reader);
   W_REGISTER_READER (TET4_ELCODE, tet4_reader);
   W_REGISTER_READER (NTRUSS_ELCODE, ntruss_reader);
   W_REGISTER_READER (CMASS_ELCODE, cmass_reader);
   W_REGISTER_READER (TMEM3_ELCODE, tmem3_reader);
   W_REGISTER_READER (T3PRESS_ELCODE, t3press_reader);
   W_REGISTER_READER (Q4PRESS_ELCODE, q4press_reader);
   W_REGISTER_READER (Q8PRESS_ELCODE, q8press_reader);
   W_REGISTER_READER (EULHEX_ELCODE, eulhex_reader);
   /*   W_REGISTER_READER (HEXWFBAR_ELCODE, hexwfbar_reader); */
   W_REGISTER_READER (QMEM4C_ELCODE, qmem4c_reader);
   W_REGISTER_READER (PCONTPT_ELCODE, pcontpt_reader);
   W_REGISTER_READER (SPRING_ELCODE, spring_reader);
   W_REGISTER_READER (TET4AMC_ELCODE, tet4amc_reader);
   W_REGISTER_READER (HEXSSBBAR_ELCODE, hexssbbar_reader);
#undef  W_DCL_READER
#undef  W_REGISTER_READER
}

/*
  Element reader.
  Generic element reader, which delegates the actual task
  of reading to the reader registered for the element type.
  If no reader was registered for the type, FALSE is returned.
 */
W_bool_t                        /* TRUE if successful */
W_element_reader (W_domain_t *d /* domain */)
{
  static int readers_initialized = FALSE;
  const char *input_line;
  long elcode;
  W_elem_reader_func reader = NULL;
  
  if (!readers_initialized) {
    initialize_readers_for_builtins ();
    readers_initialized = TRUE;
  }
  
  input_line = W_get_input_line(d);
  
  if (sscanf(input_line, "%ld", &elcode) != 1)
    BAD_INPUT();

  reader = find_reader (elcode);
  if (reader != NULL) {
    return (*reader) (d);
  } else {
    return FALSE;
  }
}


/*
   Register an element reader.
   To be able to use an element, its code needs to be linked to the
   executable, and its reader needs to be registered, so that the generic
   element reader can look up the type-specific reader.
 */
void
WE_register_reader (int elcode /* unique integer among all elements */,
                    W_elem_reader_func reader /* the reader function */
                    )
{
  reader_intermediary_t *ri;
  
  if (reader_intermediary_table == NULL) {
    reader_intermediary_table = dict_create_dictionary ();
  }
  ri = (reader_intermediary_t *) dict_find_obj_with_int_name (reader_intermediary_table, elcode);
  if (ri != NULL) {
    sprintf (W_buf (), "Reader for element type %d already registered",
             elcode);
    REPORT_ERROR (NULL, W_buf ());
  }
  ri = w_new (reader_intermediary_t); CHECK_MALLOC (ri);
  ri->reader = reader;
  dict_add_obj_int_name (reader_intermediary_table, ri, elcode);
}


/*
   Create a new element.
   Create a new element of appropriate type.  The type specific
   information is supplied in "methods" and "data".
 */
W_elem_t *                      		/* new element */
WE_new_element (W_domain_t *d, 			/* domain */
                W_elem_methods_t *methods,	/* type-specific methods */
                void *data 			/* type-specific data */
#ifdef PVM_NEED_GHOST_ELEMENTS
                , W_bool_t is_ghost 		/* is the element a ghost? */
#endif
                )
{
  W_elem_t *el;

  el = w_new(W_elem_t);
  CHECK_MALLOC (el);
  
  el->domain             = d;
  el->data               = data;
  el->methods            = methods;
  el->deformation_energy = 0;
#ifdef X_XFACE
  el->grep               = NULL;
#endif
#ifdef PVM_NEED_GHOST_ELEMENTS
  el->i_am_ghost         = is_ghost;
#endif
  return el;
}


/* return the element code
   The element code is a unique type identifier.  It is set in the
   element header file, and the element reader is registered under this
   code.  If the element does not define a type-specific method,
   the invalid code -13131313 is returned.
 */
int /* element code */
WE_elem_elcode (W_elem_t *el /* element */
                )
{
  if (el->methods->elem_elcode != NULL)
    return (*el->methods->elem_elcode) (el);
  else
    return -13131313; /* invalid code */
}



/* element type as a string
   The element type is returned as a string.
 */
const char * /* the element type as a read-only string */
WE_elem_type_name (W_elem_t *el /* element */)
{
  if (el->methods->elem_type_name != NULL)
    return (*el->methods->elem_type_name) (el);
  else
    return "unknown element type";
}




/*
   Assemble the restoring nodal forces.
   Assemble the restoring nodal forces by adding the contribution of
   this element.
 */
void
WE_assemble_restoring_forces (W_elem_t *el,
                              W_mat_state_update_form_enum update_form)
{
#ifdef PVM_NEED_GHOST_ELEMENTS
  if (!el->i_am_ghost) {
#endif
    if (el->methods->assemble_restoring_forces != NULL)
      (*el->methods->assemble_restoring_forces) (el, update_form);
#ifdef PVM_NEED_GHOST_ELEMENTS
  }
#endif
}
 






/*
   Write a post-processing file.
   Write a post-processing data for this element.
 */
void
WE_write_postproc_file (W_elem_t *el, /* element */
                        ckit_iobuf_t IOB, /* I/O buffer */
                        int type, /* type of the file */
                        char *mode, /* mode of the file */
                        void *optdata /* optional data interpreted
                                         by the element
                                       */
                        )
{
  if (el->methods->write_postproc_file != NULL)
    (*el->methods->write_postproc_file) (el, IOB, type, mode, optdata);
}

/* Assemble the nodal masses.
   Assemble the nodal masses by adding the contributions of this element.
 */
void
WE_assemble_mass (W_elem_t *el/* element */)
{
  if (el->methods->assemble_mass != NULL)
    (*el->methods->assemble_mass) (el);
}


/*
   Compute the suggested limit on the stable time step.
 */
W_float_t /* limit on the stable time step as given by the properties
             of this element
           */
WE_suggested_time_step (W_elem_t *el/* element */)
{
#ifdef PVM_NEED_GHOST_ELEMENTS
  if (!el->i_am_ghost) {
#endif
    if (el->methods->suggested_time_step != NULL)
      return (*el->methods->suggested_time_step) (el);
    else
      return FLT_MAX; /* no restriction on time step  */
#ifdef PVM_NEED_GHOST_ELEMENTS
  } else
    return FLT_MAX;
#endif
}

/*
   Compute the energy of deformation.
   Compute the energy of deformation accumulated in this element.
 */
W_float_t /* deformation energy */
WE_deformation_energy (W_elem_t *el/* element */)
{
#ifdef PVM_NEED_GHOST_ELEMENTS
  if (!el->i_am_ghost) {
#endif
    if (el->methods->deformation_energy != NULL)
      return (*el->methods->deformation_energy) (el);
    else
      return el->deformation_energy;
#ifdef PVM_NEED_GHOST_ELEMENTS
  } else
    return 0;
#endif
}

/*
   Compute the kinetic energy.
   Compute the kinetic energy associated with the mass of this element.
   This method needs to be defined for implicit integration, or if
   the element has internal degrees of freedom.
   If the element type does not define the method, zero (0) is returned.
 */
W_float_t /* kinetic energy or zero if method not specialized */
WE_kinetic_energy (W_elem_t *el, /* element */
                   W_mass_type_enum mass_type /* mass type */
                   )
{
#ifdef PVM_NEED_GHOST_ELEMENTS
  if (!el->i_am_ghost) {
#endif
    if (el->methods->kinetic_energy != NULL)
      return (*el->methods->kinetic_energy) (el, mass_type);
    else
      return 0;
#ifdef PVM_NEED_GHOST_ELEMENTS
  } else
    return 0;
#endif
}


/*
   Compute the total mass of the element.
 */
W_float_t /* the total mass */
WE_elem_total_mass (W_elem_t *el/* element */)
{
  if (el->methods->elem_total_mass != NULL)
    return (*el->methods->elem_total_mass) (el);
  else
    return 0.0;
}


#ifdef X_XFACE
/*
   Modify the element graphic representation.
 */
void
WE_modify_elem_grep (W_elem_t *el/* element */, double displ_scale, W_bool_t immed_draw)
{
#ifdef PVM_NEED_GHOST_ELEMENTS
  if (!el->i_am_ghost) {
#endif
    if (el->methods->modify_grep != NULL)
      (*el->methods->modify_grep) (el, displ_scale, immed_draw);
#ifdef PVM_NEED_GHOST_ELEMENTS
  }
#endif
}

#endif /* #ifdef X_XFACE */

/*
   Apply initial conditions.
   Applies to elements with internal degrees-of-freedom only.
   Apply initial conditions for internal degrees-of-freedom for this element.
 */
void
WE_apply_initial_conditions (W_elem_t *el/* element */)
{
  if (el->methods->apply_initial_conditions != NULL)
    (*el->methods->apply_initial_conditions) (el);
}


/*
   Calculate the effective loads.
   Applies to elements with internal degrees-of-freedom only.
   Calculate the effective loads for the internal degrees-of-freedom
   of this element.
 */
void 
WE_calc_eff_loads (W_elem_t *el/* element */)
{
  if (el->methods->calc_eff_loads!= NULL)
    (*el->methods->calc_eff_loads) (el);
}


/*
   Solve for the primary kinematic quantities.
   Applies to elements with internal degrees-of-freedom only.
   Solve for the primary kinematic quantities
   for the internal degrees-of-freedom
   of this element.
 */
void 
WE_solve_for (W_elem_t *el/* element */)
{
  if (el->methods->solve_for!= NULL)
    (*el->methods->solve_for) (el);
}


/*
   Update the configuration.
   Applies to elements with internal degrees-of-freedom only.
   Update the configuration
   for the internal degrees-of-freedom
   of this element.  Warburton's variant of the central differences.
 */
void 
WE_update_config_Warburton_variant (W_elem_t *el/* element */)
{
  if (el->methods->update_config_Warburton_variant!= NULL)
    (*el->methods->update_config_Warburton_variant) (el);
}

/*
   Update the configuration.
   Applies to elements with internal degrees-of-freedom only.
   Update the configuration
   for the internal degrees-of-freedom
   of this element.  Newmark's variant of the central differences.
 */
void 
WE_update_config_Newmark_variant (W_elem_t *el/* element */)
{
  if (el->methods->update_config_Newmark_variant!= NULL)
    (*el->methods->update_config_Newmark_variant) (el);
}


/*
   Update the history.
   Applies to elements with internal degrees-of-freedom only.
   Update the history clients
   of this element.  
 */
void
WE_update_history (W_elem_t *el/* element */,
                   W_solution_history_t *h,
                   W_float_t time)
{
  if (el->methods->update_history != NULL)
    (*el->methods->update_history) (el, h, time);
}


/*
   Finalize the current step in time.
   Applies to elements with internal degrees-of-freedom only.
   Finalize the current step in time, ie., do whatever is necessary
   to conclude the step (for example, propagate a crack).
 */
W_float_t
WE_finalize_step (W_elem_t *el/* element */)
{
  if (el->methods->finalize_step != NULL)
    return (*el->methods->finalize_step) (el);
  else
    return 0;
}

/*
   Create a new element history client.
 */
void *
WE_new_history_client_data (W_elem_t *el/* element */,
                            W_index_t idx, /* id of the client */
                            W_history_writer_t *r, /* history reporter */
                            char *input_line /* line from the
                                                input file on which
                                                the client was defined
                                                */
                            )
{
  if (el->methods->new_history_client_data != NULL)
    return (*el->methods->new_history_client_data) (el, idx, r, input_line);
  else
    return NULL;
}





/*
   Print element info.
 */
void
WE_print_info (W_elem_t *el/* element */,
               W_info_spec_t spec, /* specification */
               FILE *fp, /* to file */
               int indent /* optional indent by white spaces */
               )
{
  W_INDENT(fp, indent);
  fprintf(fp, "Element type \"%s\"\n", WE_elem_type_name(el));
  indent += 3;
#ifdef PVM_NEED_GHOST_ELEMENTS
  if (el->i_am_ghost) {
    W_INDENT(fp, indent);
    fprintf(fp, "Element is a ghost\n");
  }
#endif
  if (el->methods->print_info != NULL) {
    (*el->methods->print_info) (el, spec, fp, indent);
  }
}





/*
   Configure an element.
   Configure an element by passing to it a string with the options and
   values.  The interpretation of the string is up to the particular
   element.
 */
W_bool_t /* signals success or failure of the configuration attempt */
WE_configure (W_elem_t *el/* element */,
              char *option /* option and value in string form */
              )
{
  if (el->methods->configure != NULL)
    return (*el->methods->configure) (el, option);
  else
    return FALSE;
}

/*
   Probe the given field inside the element.
 */
W_bool_t 
WE_probe (W_elem_t *el/* element */,
          void *options, /* options */
          W_loc_t *loc, /* at this point */
          unsigned long spec, /* what to inspect? */
          W_probe_package_t *result /* return the result in this packet */
          )
{
  if (el->methods->probe != NULL)
    return (*el->methods->probe) (el, options, loc, spec, result);
  else
    return FALSE;
}
 

/*
   Get the number of topological faces.
   Get the number of topological faces defined by the element.
 */
int 
WE_num_of_faces (W_elem_t *el)
{
  if (el->methods->num_of_faces)
    return (*el->methods->num_of_faces) (el);
  else
    return 0;
}

/*
   Get the nodes at the vertices of face.
   Get the nodes at the vertices of the topological face
   number nface.  Note that only topologically triangular
   or quadratilaterla faces are allowed.
 */
int /* number of nodes supplied in the array; if the array is not
     large enough, the desired dimension is returned. */
WE_get_face_vertex_nodes (W_elem_t *el, /* element */
                          int nface, /* face number */
                          int ndim, /* dimension of the below array; if the
                                       dimension is not big enough to hold
                                       all of the face nodes, the appropriate
                                       dimension is returned and no data
                                       are placed in the array.
                                    */
                          W_node_t *n[] /* the array to hold pointers
                                         to nodes on the face */
                          )
{
  if (el->methods->get_face_vertex_nodes)
    return (*el->methods->get_face_vertex_nodes) (el, nface, ndim, n);
  else
    return 0;
}


/*
   Get the contact faces.
   Get the contact faces on the topological face nface of the
   element.  If the array cf is too small to hold all of the
   contact faces, the function returns the true number of
   contact faces, eventhough not all of them are defined
   in the array.   Therefore check the returned value, if it is
   larger than maxcf, none of the contact faces in cf is valid.
   Reallocate cf, and call the function again.
 */
int /* number of contact faces pointers defined in cf[] or the actual
       number of contact faces the element wants to define if maxcf is
       not big enough
     */
WE_get_contact_faces (W_elem_t *el, int nface, WCF_contact_face_t *cf[], int maxcf)
{
  if (el->methods->get_contact_faces)
    return (*el->methods->get_contact_faces) (el, nface, cf, maxcf);
  else
    return 0;
}



/*
   Assemble the residual nodal forces.
   Assemble the residual nodal forces by adding the contribution of
   this element.
 */
void
WE_assemble_residual_forces (W_elem_t *el/* element */,
                             W_mat_state_update_form_enum update_form,
                             W_mass_type_enum mass_type)
{
  if (el->methods->assemble_residual_forces != NULL)
    (*el->methods->assemble_residual_forces) (el, update_form, mass_type);
}
 


/*
   Assemble effective stiffness.
   Assemble the effective stiffness by adding the contribution of
   this element.
 */
void
WE_assemble_eff_stiffness (W_elem_t *el/* element */,
                           W_newmark_assemble_eff_stiff_proc assemble_eff_stiff_callback, /* callback to use to actually assemble the submatrix into the system matrix; this has to be supplied by the domain */
                           W_float_t stiff_mult, /* multiply stiffness by this; may be zero, in which case it indicates that no stiffness should be included */
                           W_float_t mass_mult, /* multiply mass by this; may be zero, in which case no mass matrix should be included */
                           W_mass_type_enum mass_type /* type of the mass matrix (consistent, diagonal, ...) */
                           )
{
  if (el->methods->assemble_eff_stiffness != NULL)
    (*el->methods->assemble_eff_stiffness) (el, assemble_eff_stiff_callback,
                                            stiff_mult, mass_mult, mass_type);
}
 

W_bool_t
WE_get_elem_prop (W_elem_t *el, const char *prop, void *value)
{
  if (el->methods->get_elem_prop)
    return (*el->methods->get_elem_prop) (el, prop, value);
  else
    return FALSE;
}

int
WE_get_num_nodes (W_elem_t *el)
{
  if (el->methods->get_num_nodes)
     return (*el->methods->get_num_nodes)(el);
  else 
     return 0;
}

W_node_t *
WE_get_node(W_elem_t *el, int j)
{
  if (el->methods->get_node)
     return (*el->methods->get_node)(el, j);
  else
     return 0;
}

int 
WE_get_num_matstates(W_elem_t *el)
{
  if (el->methods->get_num_matstates)
     return (*el->methods->get_num_matstates)(el);
  else 
     return 0;
}

void *
WE_get_matstate(W_elem_t *el, int j)
{
  if (el->methods->get_matstate)
     return (*el->methods->get_matstate)(el, j);
  else
     return 0;
}

int 
WE_get_material_id(W_elem_t *el)
{
  if (el->methods->get_material_id)
     return (*el->methods->get_material_id)(el);
  else 
     return 0;
}
