#include "Wconfig.h"
#include "Wxf.h"
#include "Welem.h"

/*
     In order to be able to link an element to the WASP package, it needs
     to register a procedure which is able to
       (i)  read the element data from the input file, and
       (ii) add the element(s) to the domain.
     The element reader is of the following structure:

"  W_bool_t 
"  xxxxxx_reader(W_domain_t *d, FILE *stream)
"  {
"    const char *input_line;
"    ... some other declarations here ...
"      
"    while ((input_line = W_get_input_line(stream)) != NULL) {
"      if (input_line[0] == END_OF_BLOCK)
"        return TRUE; / * ok; end of element xxxxx block * /
"      else {
"        (i)   create the element struct
"        (ii)  read the data from the input file
"        (iii) add the element to the WASP domain:
"          WD_add_element(d, elem_num,
"                       WE_new_element(d, &xxxxx_methods, t
"  #                    ifdef PVM_NEED_GHOST_ELEMENTS
"                                      , is_ghost
"  #                    endif
"                       ));
"      }
"    }
"    return FALSE; / * block not finished properly by '&' * /
"  }
"
       The element is registered when the WE_initialize_readers()
       procedure is invoked during the initialization of the WASP
       package.  To register an element reader, declare it and invoke
       the macro W_REGISTER_READER(elcode, reader) in the file
       "element.readers".  This file is included in the code of
       the WE_initialize_readers() procedure.

       In order to be a good companion, the element needs to comply
       with a certain protocol.  The protocol is carried out by calling
       functions, pointer to which are stored in the element method table.
       (Note that the method table is shared by all elements of the same type,
       but one could easily implement a private table for each element,
       for example to implement special behaviors for certain select
       element individuals.)

       The element method table prototype is given in the include file
       Wtypes.h.

     */
void 
WE_initialize_readers (void)
{
  
#define W_DCL_READER(reader) W_elem_reader_func reader
#define W_REGISTER_READER(elcode, reader) WE_register_reader (elcode, reader)
#include "element.readers" 
#undef  W_DCL_READER
#undef  W_REGISTER_READER
}
