#ifndef WBMAT_NNODES
# error WBMAT_NNODES not defined
#endif

/* RCSHEADER "$Id: elstiff.c,v 1.4 2002/11/08 20:34:17 cvs Exp $" */

static void
mirror (double K[WBMAT_NNODES*3][WBMAT_NNODES*3])
{
  int i, j;
  const int kdim = WBMAT_NNODES*3;
 /* copy the upper triangle to the lower one */
  for (i = 1; i < kdim; i++) {
    for (j = 0; j < i; j++) {
      K[i][j] = K[j][i];
    }
  }
}

static void
accum (double K[WBMAT_NNODES*3][WBMAT_NNODES*3], double toadd[WBMAT_NNODES*3][WBMAT_NNODES*3])
{
  int i, j;
  const int kdim = WBMAT_NNODES*3;
  FOR(i,kdim) {
    FOR(j,kdim) {
      K[i][j] += toadd[i][j];
    }
  }
}

static void
add_to_diag (double M[WBMAT_NNODES*3][WBMAT_NNODES*3], double value)
{
  int i;
  const int kdim = WBMAT_NNODES*3;

  FOR (i, kdim) {
    M[i][i] += value;
  }
}

static void
add_rhontn_upper_triangle (double M[WBMAT_NNODES*3][WBMAT_NNODES*3],
                           double nmat[WBMAT_NNODES], double fact)
{
  int hi, hj, i, j;
  double ninj;

  FOR (hi, WBMAT_NNODES) {
    for (hj = hi; hj < WBMAT_NNODES; hj++) {
      ninj = fact * nmat[hi] * nmat[hj];
      j = hj * 3;
      i = hi * 3;
      M[i++][j++] += ninj;
      M[i++][j++] += ninj;
      M[i++][j++] += ninj;
    }
  }
}

static void
add_rhontn (double M[WBMAT_NNODES*3][WBMAT_NNODES*3],
                           double nmat[WBMAT_NNODES], double fact)
{
  int hi, hj, i, j;
  double ninj;
 
  FOR (hi, WBMAT_NNODES) {
    FOR (hj, WBMAT_NNODES) {
      ninj = fact * nmat[hi] * nmat[hj];
      j = hj * 3;
      i = hi * 3;
      M[i++][j++] += ninj;
      M[i++][j++] += ninj;
      M[i++][j++] += ninj;
    }
  }
}
static void
add_bltcbl_upper_triangle (double K[WBMAT_NNODES*3][WBMAT_NNODES*3],
                           double b_l[6][WBMAT_NNODES*3], double C[6][6], double fact)
{
  int i, j, k, m;
  const int kdim = WBMAT_NNODES*3;
  double c, blkic;
  /* compute the upper triangle */
  FOR (k, 6) {
    FOR (m, 6) {
      c = C[k][m];
      if (c != 0) {
        c *= fact;
        FOR (i, kdim) {
          blkic = c * b_l[k][i];
          for (j = i; j < kdim; j++) {
            K[i][j] += blkic * b_l[m][j];
          }
        }
      }
    }
  }
}

static void
add_bltcbl (double K[WBMAT_NNODES*3][WBMAT_NNODES*3],
            double b_l[6][WBMAT_NNODES*3], double C[6][6], double fact)
{
  int i, j, k, m;
  const int kdim = WBMAT_NNODES*3;
  double c, blkic;
  /* compute the upper triangle */
  FOR (k, 6) {
    FOR (m, 6) {
      c = C[k][m];
      if (c != 0) {
        c *= fact;
        FOR (i, kdim) {
          blkic = c * b_l[k][i];
          FOR (j, kdim) {
            K[i][j] += blkic * b_l[m][j];
          }
        }
      }
    }
  }
}


static void
add_bnltsbnl_upper_triangle (double K[WBMAT_NNODES*3][WBMAT_NNODES*3],
                             double b_nl[9][WBMAT_NNODES*3],
                             double S[3][3], double fact)

{
  int i, j, k, m, ik, im;
  const int kdim = WBMAT_NNODES*3;
  const int si[] = {0,1,2,0,1,2,0,1,2};
  double kij, sb;
  
  /* compute the upper triangle */
  FOR (i, kdim) {
    for (j = i; j < kdim; j++) {
      kij = 0;
      for (ik =0, k = 0; k < 3; k++, ik++) {
        sb = 0;
        for (im = 0, m = 0; m < 3; m++, im++) {
          sb += S[ik][im] * b_nl[m][j];
        }
        kij += b_nl[k][i] * sb;
      }
      for (ik =0, k = 3; k < 6; k++, ik++) {
        sb = 0;
        for (im = 0, m = 3; m < 6; m++, im++) {
          sb += S[ik][im] * b_nl[m][j];
        }
        kij += b_nl[k][i] * sb;
      }
      for (ik =0, k = 6; k < 9; k++, ik++) {
        sb = 0;
        for (im = 0, m = 6; m < 9; m++, im++) {
          sb += S[ik][im] * b_nl[m][j];
        }
        kij += b_nl[k][i] * sb;
      }
      K[i][j] += fact * kij;
    }
  }
}

static void
add_bnltsbnl (double K[WBMAT_NNODES*3][WBMAT_NNODES*3],
                             double b_nl[9][WBMAT_NNODES*3],
                             double S[3][3], double fact)
 
{
  int i, j, k, m, ik, im;
  const int kdim = WBMAT_NNODES*3;
  const int si[] = {0,1,2,0,1,2,0,1,2};
  double kij, sb;
 
  /* compute the upper triangle */
  FOR (i, kdim) {
    FOR (j, kdim) {
      kij = 0;
      for (ik =0, k = 0; k < 3; k++, ik++) {
        sb = 0;
        for (im = 0, m = 0; m < 3; m++, im++) {
          sb += S[ik][im] * b_nl[m][j];
        }
        kij += b_nl[k][i] * sb;
      }
      for (ik =0, k = 3; k < 6; k++, ik++) {
        sb = 0;
        for (im = 0, m = 3; m < 6; m++, im++) {
          sb += S[ik][im] * b_nl[m][j];
        }
        kij += b_nl[k][i] * sb;
      }
      for (ik =0, k = 6; k < 9; k++, ik++) {
        sb = 0;
        for (im = 0, m = 6; m < 9; m++, im++) {
          sb += S[ik][im] * b_nl[m][j];
        }
        kij += b_nl[k][i] * sb;
      }
      K[i][j] += fact * kij;
    }
  }
}

static void
add_gtcg (double K[WBMAT_NNODES*3][WBMAT_NNODES*3],
            double g[9][WBMAT_NNODES*3], double C[9][9], double fact)
{
  int i, j, k, m;
  const int kdim = WBMAT_NNODES*3;
  double c, gkic;
  FOR (k, 9) {
    FOR (m, 9) {
      c = C[k][m];
      if (c != 0) {
        c *= fact;
        FOR (i, kdim) {
          gkic = c * g[k][i];
          FOR (j, kdim) {
            K[i][j] += gkic * g[m][j];
          }
        }
      }
    }
  }
}

static void
add_gtcg_upper_triangle (double K[WBMAT_NNODES*3][WBMAT_NNODES*3],
                           double g[9][WBMAT_NNODES*3], double C[9][9], double fact)
{
  int i, j, k, m;
  const int kdim = WBMAT_NNODES*3;
  double c, gkic;
  /* compute the upper triangle */
  FOR (k, 9) {
    FOR (m, 9) {
      c = C[k][m];
      if (c != 0) {
        c *= fact;
        FOR (i, kdim) {
          gkic = c * g[k][i];
          for (j = i; j < kdim; j++) {
            K[i][j] += gkic * g[m][j];
          }
        }
      }
    }
  }
}

static void
add_eulerian_istress_upper_triangle (
   double K[WBMAT_NNODES*3][WBMAT_NNODES*3],
   double n_x[WBMAT_NNODES],
   double n_y[WBMAT_NNODES],
   double n_z[WBMAT_NNODES],
   double cauchy[3][3], double fact
   )
{
  int a, b, k, l, i, j;
  double *nder[3];
  double nsn;
  nder[0] = n_x;
  nder[1] = n_y;
  nder[2] = n_z;
  FOR (a, WBMAT_NNODES) {
    for (b = a; b < WBMAT_NNODES; b++) {
      nsn = 0;
      FOR (k, 3) {
        FOR (l, 3) {
          nsn += nder[k][a] * cauchy[k][l] * nder[l][b];
        }
      }
      i = 3 * a;
      j = 3 * b;
      K[i][j] += fact * nsn; i++; j++;
      K[i][j] += fact * nsn; i++; j++;
      K[i][j] += fact * nsn;
    }
  }
}
