#include "Wconfig.h"
#include "Wtypes.h"
#include "Wdomain.h"
#include "Woptions.h"
#include "Wnode.h"
#include "Wmats.h"
#include "Wgaussp.h"
#include "Wxf.h"
#include "Wutil.h"
#include "Whistry.h"
#include "Wcdi.h"
#include "g3d.h"
#include "Wmat3d.h"
#include "Wnewmarki.h"
#include "Wcdi.h"
#include "Wcontface.h"
#include "Weulhex.h"

/* RCSHEADER "$Id: eulhex.c,v 1.3 2002/03/21 01:32:35 cvs Exp $" */

/*
  This file implements an Eulerian variant of the hexahedron from hex.c.
  Eulerian means that all weak forms are evaluated in the current
  configuration.  That does not mean `updated Lagrangian', but really
  the latest configuration, as produced for instance by each Newton
  iteration.
 */

#undef WBMAT_NNODES
#define WBMAT_NNODES 8

#include "bmat.c"
#include "elstiff.c"
 
#define GET_FROM_NODE(node, func, tovect)         \
   func (node, &tovect.x, &tovect.y, &tovect.z)

#define CHECK_2PT_CASE(H, I, J, K)                    \
    ((!use_2pt_rule)                                  \
     || (   (I == 0 && J == 0 && K == 0)              \
         || (I == 1 && J == 1 && K == 1)))
#define USE_QUADR_PT(H, I, J, K) if (CHECK_2PT_CASE(H, I, J, K))
#define ALL_GP_LOCS(i,j,k) gp[i].Loc, gp[j].Loc, gp[k].Loc
#define GP_WGT(gp,I,J,K) (gp[I].Wgt * gp[J].Wgt * gp[K].Wgt)

static int msidcs[2][2][2] = {
  {{0,7}, {2,3}},
  {{4,5}, {6,1}}
};

static int invmsidcs[8][3] = {
/* invmsidcs[0] = */ {0, 0, 0},
/* invmsidcs[1] = */ {1, 1, 1},
/* invmsidcs[2] = */ {0, 1, 0},
/* invmsidcs[3] = */ {0, 1, 1},
/* invmsidcs[4] = */ {1, 0, 0},
/* invmsidcs[5] = */ {1, 0, 1},
/* invmsidcs[6] = */ {1, 1, 0},
/* invmsidcs[7] = */ {0, 0, 1},
};

typedef struct priv_elem_t {
  W_node_t          *node[WBMAT_NNODES];
  W_material_t      *material;
  W_mat_3d_state_p  *matst;
  W_bool_t           use_2pt_rule;
  double             mass;
}              priv_elem_t;

static int
elem_elcode (W_elem_t *el)
{
  return EULHEX_ELCODE;
}

static char *
type_name (W_elem_t *el)
{
  static char buf[] = "eulhex";
  return buf;
}

static int 
num_of_faces (W_elem_t *el)
{
  return 6;
}

static int 
get_face_vertex_nodes (W_elem_t *el, int nface, int ndim, W_node_t *n[])
{
  priv_elem_t *h = (priv_elem_t *)el->data;
  int i;

  if (ndim >= 4) {
    FOR (i, 4) {
      n[i] = h->node[g3dhex_vertex_on_face[nface][i]];
    }
  } 
  return 4;
}

static int
get_contact_faces (W_elem_t *el, int nface,
                   WCF_contact_face_t *cf[], int maxcf)
{
  priv_elem_t *h = (priv_elem_t *)el->data;
  int i;
  
  if (maxcf < 1) {
    FOR (i, maxcf) { cf[i] = NULL; }
    return 1;
  }
  
  cf[0] = WCF_new_contact_face (h->node[g3dhex_vertex_on_face[nface][0]],
                                h->node[g3dhex_vertex_on_face[nface][1]],
                                h->node[g3dhex_vertex_on_face[nface][2]],
                                h->node[g3dhex_vertex_on_face[nface][3]]);
  
  return 1;
}

static W_gauss_integration_rule_t *gr = NULL;
static W_gauss_integration_point_t *gp = NULL;

typedef struct shpf_t {
  double n[WBMAT_NNODES];
  double n_x[WBMAT_NNODES];
  double n_y[WBMAT_NNODES];
  double n_z[WBMAT_NNODES];
  double J[3][3];
  double invJ[3][3];
  double detJ;
}              shpf_t;

typedef struct ip_ctx_data_t {
  W_domain_t  *domain;
  priv_elem_t *h;
  shpf_t       sf;    /* shape function table */
}              ip_ctx_data_t;


#define VALID(which) which##_is_valid

#define DCL_TENSOR(T)                           \
   static double T[3][3];                       \
   static W_bool_t VALID(T) = FALSE
 
DCL_TENSOR(F);
DCL_TENSOR(prev_F);
DCL_TENSOR(midstep_F);
DCL_TENSOR(inv_F);
DCL_TENSOR(R);
DCL_TENSOR(U);
DCL_TENSOR(rel_F);
DCL_TENSOR(inv_rel_F);
DCL_TENSOR(rel_R);
DCL_TENSOR(rel_U);
DCL_TENSOR(C);
DCL_TENSOR(E);
DCL_TENSOR(D);
DCL_TENSOR(L);
DCL_TENSOR(ln_rel_U);
DCL_TENSOR(midstep_D);
DCL_TENSOR(midstep_W);


static void
at_new_ip (void)
{
#define INVALIDATE(which) VALID(which) = FALSE
  
  INVALIDATE(F);
  INVALIDATE(prev_F);
  INVALIDATE(midstep_F);
  INVALIDATE(inv_F);
  INVALIDATE(R);
  INVALIDATE(U);
  INVALIDATE(rel_F);
  INVALIDATE(inv_rel_F);
  INVALIDATE(rel_R);
  INVALIDATE(rel_U);
  INVALIDATE(C);
  INVALIDATE(E);
  INVALIDATE(D);
  INVALIDATE(L);
  INVALIDATE(ln_rel_U);
  INVALIDATE(midstep_D);
  INVALIDATE(midstep_W);
}

static void
compute_inv_def_grad (ip_ctx_data_t *ipctx)
{
  priv_elem_t *h = ipctx->h;
  double *n_x = ipctx->sf.n_x;
  double *n_y = ipctx->sf.n_y;
  double *n_z = ipctx->sf.n_z;
  int k;
  double X, Y, Z;
  register double Nx, Ny, Nz;

  /*  Evaluate dX/dx (inverse of deformation gradient) */
  k = 0;
  WN_get_ref_coords (h->node[k], &X, &Y, &Z);
  Nx  = n_x[k];
  Ny  = n_y[k];
  Nz  = n_z[k];
  inv_F[0][0] = Nx * X; inv_F[0][1] = Ny * X; inv_F[0][2] = Nz * X;
  inv_F[1][0] = Nx * Y; inv_F[1][1] = Ny * Y; inv_F[1][2] = Nz * Y;
  inv_F[2][0] = Nx * Z; inv_F[2][1] = Ny * Z; inv_F[2][2] = Nz * Z;
  for (k = 1; k < WBMAT_NNODES; k++) {
    WN_get_ref_coords (h->node[k], &X, &Y, &Z);
    Nx  = n_x[k];
    Ny  = n_y[k];
    Nz  = n_z[k];
    inv_F[0][0] += Nx * X; inv_F[0][1] += Ny * X; inv_F[0][2] += Nz * X;
    inv_F[1][0] += Nx * Y; inv_F[1][1] += Ny * Y; inv_F[1][2] += Nz * Y;
    inv_F[2][0] += Nx * Z; inv_F[2][1] += Ny * Z; inv_F[2][2] += Nz * Z;
  }

  VALID(inv_F) = TRUE;
}

static void
compute_rel_def_grad (ip_ctx_data_t *ipctx)
{
  priv_elem_t *h = ipctx->h;
  double *n_x = ipctx->sf.n_x;
  double *n_y = ipctx->sf.n_y;
  double *n_z = ipctx->sf.n_z;
  int k;
  double x, y, z;
  double ux, uy, uz;
  double pux, puy, puz;
  register double Nx, Ny, Nz;

  k = 0;
  WN_get_prev_displ (h->node[k], &pux, &puy, &puz);
  WN_get_displ (h->node[k], &ux, &uy, &uz);
  x = pux - ux; y = puy - uy; z = puz - uz;
  Nx  = n_x[k]; Ny  = n_y[k]; Nz  = n_z[k];
  inv_rel_F[0][0] = Nx * x; inv_rel_F[0][1] = Ny * x; inv_rel_F[0][2] = Nz * x;
  inv_rel_F[1][0] = Nx * y; inv_rel_F[1][1] = Ny * y; inv_rel_F[1][2] = Nz * y;
  inv_rel_F[2][0] = Nx * z; inv_rel_F[2][1] = Ny * z; inv_rel_F[2][2] = Nz * z;
  for (k = 1; k < WBMAT_NNODES; k++) {
    WN_get_prev_displ (h->node[k], &pux, &puy, &puz);
    WN_get_displ (h->node[k], &ux, &uy, &uz);
    x = pux - ux; y = puy - uy; z = puz - uz;
    Nx  = n_x[k]; Ny  = n_y[k]; Nz  = n_z[k];
    inv_rel_F[0][0] += Nx * x; inv_rel_F[0][1] += Ny * x; inv_rel_F[0][2] += Nz * x;
    inv_rel_F[1][0] += Nx * y; inv_rel_F[1][1] += Ny * y; inv_rel_F[1][2] += Nz * y;
    inv_rel_F[2][0] += Nx * z; inv_rel_F[2][1] += Ny * z; inv_rel_F[2][2] += Nz * z;
  }
  inv_rel_F[0][0] += 1; inv_rel_F[1][1] += 1; inv_rel_F[2][2] += 1;

  if (!W_mat3dutil_invert_3x3 (inv_rel_F, rel_F)) {
    sprintf(W_buf(), "Failed to invert inv def grad");
    REPORT_FATAL_ERROR(NULL, W_buf());
  }

  VALID(inv_rel_F) = TRUE;
  VALID(rel_F) = TRUE;
}

static void
compute_def_grad (ip_ctx_data_t *ipctx)
{
  if (!VALID(inv_F)) compute_inv_def_grad(ipctx);

  if (!W_mat3dutil_invert_3x3(inv_F, F)) {
    sprintf(W_buf(), "Failed to invert inv def grad");
    REPORT_FATAL_ERROR(NULL, W_buf());
  }

  VALID(F) = TRUE;
}

static void
compute_prev_def_grad (ip_ctx_data_t *ipctx)
{
  if (!VALID(F)) compute_def_grad(ipctx);
  if (!VALID(inv_rel_F)) compute_rel_def_grad(ipctx);

  MAT3DUTIL_MULT_AB_3X3 (prev_F, inv_rel_F, F);

  VALID(prev_F) = TRUE;
}

static void
compute_polar_decomp (ip_ctx_data_t *ipctx)
{
  if (!VALID(F)) compute_def_grad (ipctx);
   
  if (!W_polar_decomposition (F, R, U)) {
    REPORT_FATAL_ERROR (NULL, "Failed in W_polar_decomposition()");
  }
  VALID(R) = TRUE;
  VALID(U) = TRUE;
}

static void
compute_rel_polar_decomp (ip_ctx_data_t *ipctx)
{
  if (!VALID(rel_F)) compute_rel_def_grad(ipctx);

  if (!W_polar_decomposition (rel_F, rel_R, rel_U)) {
    REPORT_FATAL_ERROR (NULL, "Failed in W_polar_decomposition()");
  }
  VALID(rel_R) = TRUE;
  VALID(rel_U) = TRUE;
}

static void
compute_midstep_def_grad (ip_ctx_data_t *ipctx)
{
  priv_elem_t *h = ipctx->h;
  
  if (!VALID (F)) compute_def_grad (ipctx);
  if (!VALID (prev_F)) compute_prev_def_grad (ipctx);
   
  MAT3DUTIL_COMB_AB_3X3 (midstep_F, 0.5, F, 0.5, prev_F);

  VALID(midstep_F) = TRUE;
}

static void
compute_ln_rel_U (ip_ctx_data_t *ipctx)
{
  if (!VALID(rel_U)) compute_rel_polar_decomp(ipctx);
  
  W_ln_U (rel_U, ln_rel_U);
  
  VALID(ln_rel_U) = TRUE;
}

static void
compute_vel_grad (ip_ctx_data_t *ipctx)
{
  priv_elem_t *h = ipctx->h;
  double *n_x = ipctx->sf.n_x;
  double *n_y = ipctx->sf.n_y;
  double *n_z = ipctx->sf.n_z;
  int k;
  double x, y, z;
  register double Nx, Ny, Nz;

  /* this is the spatial gradient of velocity */
  k = 0;
  WN_get_velocity (h->node[k], &x, &y, &z);
  Nx  = n_x[k];
  Ny  = n_y[k];
  Nz  = n_z[k];
  L[0][0] = Nx * x; L[0][1] = Ny * x; L[0][2] = Nz * x;
  L[1][0] = Nx * y; L[1][1] = Ny * y; L[1][2] = Nz * y;
  L[2][0] = Nx * z; L[2][1] = Ny * z; L[2][2] = Nz * z;
  for (k = 1; k < WBMAT_NNODES; k++) {
    WN_get_velocity (h->node[k], &x, &y, &z);
    Nx  = n_x[k];
    Ny  = n_y[k];
    Nz  = n_z[k];
    L[0][0] += Nx * x; L[0][1] += Ny * x; L[0][2] += Nz * x;
    L[1][0] += Nx * y; L[1][1] += Ny * y; L[1][2] += Nz * y;
    L[2][0] += Nx * z; L[2][1] += Ny * z; L[2][2] += Nz * z;
  }

  VALID(L) = TRUE;
}

static void
compute_rate_of_def (ip_ctx_data_t *ipctx)
{
  int i, j;
  
  if (!VALID(L)) compute_vel_grad (ipctx);

  FOR(i, 3)
    FOR(j, 3) 
      D[i][j] = 0.5 * (L[i][j] + L[j][i]);

  VALID(D) = TRUE;
}

static void
compute_cauchy_green (ip_ctx_data_t *ipctx)
{
  if (!VALID(F)) compute_def_grad (ipctx);
  
  MAT3DUTIL_MULT_ATB_3X3(C, F, F);

  VALID(C) = TRUE;
}

static void
compute_green_lagrange (ip_ctx_data_t *ipctx)
{
  if (!VALID(C)) compute_cauchy_green (ipctx);

  /* Green-Lagrange strain tensor */
  E[0][0] = 0.5 * (C[0][0] - 1);
  E[0][1] = 0.5 * C[0][1];
  E[0][2] = 0.5 * C[0][2];
  E[1][0] = 0.5 * C[1][0];
  E[1][1] = 0.5 * (C[1][1] - 1);
  E[1][2] = 0.5 * C[1][2];
  E[2][0] = 0.5 * C[2][0];
  E[2][1] = 0.5 * C[2][1];
  E[2][2] = 0.5 * (C[2][2] - 1);
  
  VALID(E) = TRUE;
}

static void
compute_midstep_rate_of_def (ip_ctx_data_t *ipctx, double time_step)
{
  int i, j;
  double m[3][3] = { {1, 0, 0}, {0, 1, 0}, {0, 0, 1} };
  double M[3][3] = { {-1, 0, 0}, {0, -1, 0}, {0, 0, -1} };
  double s[3][3], inv_s[3][3], q[3][3], r[3][3];
  double inv_dt = 1 / time_step;
  
  if (!VALID(rel_F)) compute_rel_def_grad(ipctx);

  MAT3DUTIL_ADD_AB_3X3 (s, m, rel_F);
  MAT3DUTIL_SCALE_3X3 (s, 0.5);
  if (!W_mat3dutil_invert_3x3(s, inv_s)) {
    sprintf(W_buf(), "Failed to invert def grad");
    REPORT_FATAL_ERROR(NULL, W_buf());
  }
  MAT3DUTIL_ADD_AB_3X3 (q, rel_F, M);
  MAT3DUTIL_MULT_AB_3X3 (r, inv_s, q);

  MAT3DUTIL_SYMM  (midstep_D, r);
  MAT3DUTIL_ASYMM (midstep_W, r);

  MAT3DUTIL_SCALE_3X3(midstep_D, inv_dt);
  MAT3DUTIL_SCALE_3X3(midstep_W, inv_dt);
  
  VALID(midstep_D) = TRUE;
  VALID(midstep_W) = TRUE;
}

static W_bool_t 
get_kinem_var(W_mat_3d_ip_ctx_t *ctx, 
              W_mat_3d_kinem_var_type_enum  type,
              double kinvar[3][3])
{
  ip_ctx_data_t *ipctx = (ip_ctx_data_t *)ctx->ctx_data;

  switch (type) {
  case MAT_3D_DEF_GRAD:
    if (!VALID(F)) compute_def_grad(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, F);
    break;
  case MAT_3D_REL_DEF_GRAD:
    if (!VALID(rel_F)) compute_rel_def_grad(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, rel_F);
    break;
  case MAT_3D_PREV_DEF_GRAD:
    if (!VALID(prev_F)) compute_prev_def_grad(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, prev_F);
    break;
  case MAT_3D_MIDSTEP_DEF_GRAD:
    if (!VALID(midstep_F)) compute_midstep_def_grad(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, midstep_F);
    break;
  case MAT_3D_CAUCHY_GREEN:
    if (!VALID(C)) compute_cauchy_green(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, C);
    break;
  case MAT_3D_RATE_OF_DEF:
    if (!VALID(D)) compute_rate_of_def(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, D);
    break;
  case MAT_3D_GREEN_LAGRANGE:
    if (!VALID(E)) compute_green_lagrange(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, E);
    break;
  case MAT_3D_POLAR_R:
    if (!VALID(R)) compute_polar_decomp(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, R);
    break;
  case MAT_3D_POLAR_U:
    if (!VALID(U)) compute_polar_decomp(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, U);
    break;
  case MAT_3D_POLAR_REL_R:
    if (!VALID(rel_R)) compute_rel_polar_decomp(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, rel_R);
    break;
  case MAT_3D_POLAR_REL_U:
    if (!VALID(rel_U)) compute_rel_polar_decomp(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, rel_U);
    break;
  case MAT_3D_LN_POLAR_REL_U:
    if (!VALID(ln_rel_U)) compute_ln_rel_U(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, ln_rel_U);
    break;
  case MAT_3D_MIDSTEP_RATE_OF_DEF:
    if (!VALID(midstep_D))
      compute_midstep_rate_of_def(ipctx, ctx->time_step);
    MAT3DUTIL_COPY_3X3(kinvar, midstep_D);
    break;
  case MAT_3D_MIDSTEP_SPIN:
    if (!VALID(midstep_W))
      compute_midstep_rate_of_def(ipctx, ctx->time_step);
    MAT3DUTIL_COPY_3X3(kinvar, midstep_W);
    break;
  default:
    {
      static W_bool_t complained = FALSE;
      if (!complained) {
        sprintf (W_buf(), "Unknown kinematic tensor type: %d",     (int)type);
        REPORT_WARNING (ipctx->domain->errh, W_buf());
        complained = TRUE;
      }
    }
    return FALSE; /* don't have this thing */
  }

  return TRUE;
}


#define XC(V, I) V[I].x
#define YC(V, I) V[I].y
#define ZC(V, I) V[I].z
#define XYZC(V, I) XC(V, I), YC(V, I), ZC(V, I)
#define ALLC(V) XYZC (V, 0), XYZC (V, 1), XYZC (V, 2), XYZC (V, 3), \
                XYZC (V, 4), XYZC (V, 5), XYZC (V, 6), XYZC (V, 7)

static void
eval_sf (shpf_t *sf, double xi, double eta, double theta, W_vect_t x[8])
{ 
  G3D_hex_interp_opt (sf->n, sf->n_x, sf->n_y, sf->n_z,  
                      &sf->detJ, sf->J, sf->invJ, 
                      xi, eta, theta, ALLC(x)); 
}

/*
  Returns nodal forces by which the  element acts on its nodal points.
*/
static void
assemble_restoring_forces (W_elem_t *el,
                           W_mat_state_update_form_enum update_form)
{
  priv_elem_t *h = (priv_elem_t *)el->data;
  W_mat_3d_t *m = (W_mat_3d_t *)h->material->data;
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  double cauchy[3][3];
  W_mat_3d_stress_type_enum out_type;
  W_vect_t f;
  double jdetw, dde;
  W_bool_t use_2pt_rule = h->use_2pt_rule;
  double Nx, Ny, Nz;
  int i, j, k, n;
  W_vect_t x[WBMAT_NNODES];

  FOR (i, WBMAT_NNODES) { GET_FROM_NODE (h->node[i], WN_get_coords, x[i]); }

  ip_ctx_data.domain   = el->domain;    
  ip_ctx_data.h        = h;
  ip_ctx.time          = WD_time (el->domain);
  ip_ctx.time_step     = WD_time_step (el->domain);
  ip_ctx.ctx_data      = &ip_ctx_data;
  ip_ctx.get_kinem_var = get_kinem_var;

  el->deformation_energy = 0;
  gp = gr->ip;
  FOR (i, 2) {
    FOR (j, 2) {
      FOR (k, 2) {
        USE_QUADR_PT (h, i, j, k) {
          eval_sf (&ip_ctx_data.sf, ALL_GP_LOCS(i,j,k), x);
          at_new_ip ();
          compute_def_grad (&ip_ctx_data);
          if (!WM3D_update_stress (m, h->matst[msidcs[i][j][k]], update_form,
                                   &ip_ctx, MAT_3D_CAUCHY, cauchy, &out_type))
            REPORT_ERROR(el->domain->errh, "Failed to compute stress");
          if (out_type != MAT_3D_CAUCHY) {
            REPORT_ERROR(el->domain->errh, "Failed to compute Cauchy stress");
          }
          if (!WM3D_get_mat_state_var (m, h->matst[msidcs[i][j][k]],
                                       &ip_ctx, MAT_3D_DEFORM_ENERGY_DENSITY, &dde))
            REPORT_ERROR(el->domain->errh, "Failed to retrieve work of deformation");
          dde /= MAT3DUTIL_DET_3X3 (F); /* dde is per reference volume; adjust for it */
          jdetw = GP_WGT(gp,i,j,k) * ip_ctx_data.sf.detJ;
          if (use_2pt_rule) {jdetw *= 4;}
          el->deformation_energy += dde * jdetw;
          FOR (n, WBMAT_NNODES) {
            Nx = jdetw * ip_ctx_data.sf.n_x[n];
            Ny = jdetw * ip_ctx_data.sf.n_y[n];
            Nz = jdetw * ip_ctx_data.sf.n_z[n];
            f.x = -(Nx*cauchy[0][0]+Ny*cauchy[0][1]+Nz*cauchy[0][2]);
            f.y = -(Nx*cauchy[1][0]+Ny*cauchy[1][1]+Nz*cauchy[1][2]);
            f.z = -(Nx*cauchy[2][0]+Ny*cauchy[2][1]+Nz*cauchy[2][2]);
            WN_assemble_force (h->node[n], &f);
          }
        } /* USE_QUADR_PT */
      } /* FOR(k, 2) */
    } /* FOR(j, 2) */
  } /* FOR(i, 2) */
}

static void
assemble_inertial_forces_consistent (W_elem_t *el)
{
  priv_elem_t *h = (priv_elem_t *)el->data;
  W_mat_3d_t *m = (W_mat_3d_t *)h->material->data;
  W_vect_t f[WBMAT_NNODES], a[WBMAT_NNODES];
  double Nmat[WBMAT_NNODES];
  double jdetw, J[3][3], rho;
  W_bool_t use_2pt_rule = h->use_2pt_rule;
  int i, j, k, n, l;
  W_vect_t X[WBMAT_NNODES];

  FOR (i, WBMAT_NNODES) {
    GET_FROM_NODE (h->node[i], WN_get_acceleration, a[i]);
    GET_FROM_NODE (h->node[i], WN_get_ref_coords, X[i]);
    VECMAC_SCALAR (f[i], =, 0);
  }

  rho = WM3D_get_volume_mass(m);

  gp = gr->ip;
  FOR(i, 2) {
    FOR(j, 2) {
      FOR(k, 2) {
        USE_QUADR_PT(h, i, j, k) {
          G3D_hex_N (Nmat, ALL_GP_LOCS(i,j,k));
          G3D_hex_J (J, ALL_GP_LOCS(i,j,k), ALLC(X));
          jdetw = -rho * GP_WGT(gp,i,j,k) * G3D_hex_detJ (J);
          if (use_2pt_rule) {jdetw *= 4;}
          FOR(n, WBMAT_NNODES) {
            FOR (l, WBMAT_NNODES) {
              VECMAC_2_VEC (f[n], +=, jdetw * Nmat[n] * Nmat[l] * a[l]);
            }
          }
        } /* USE_QUADR_PT */
      } /* FOR(k, 2) */
    } /* FOR(j, 2) */
  } /* FOR(i, 2) */
  FOR (i, WBMAT_NNODES) {
    WN_assemble_force (h->node[i], &f[i]);
  }
}


static W_float_t
compute_total_mass (priv_elem_t *h)
{
  W_mat_3d_t *m = (W_mat_3d_t *)h->material->data;
  W_float_t volume_mass, J[3][3];
  int i, j, k;
  double volume = 0;
  W_vect_t X[WBMAT_NNODES];

  FOR (i, WBMAT_NNODES) { GET_FROM_NODE (h->node[i], WN_get_ref_coords, X[i]); }

  volume_mass = WM3D_get_volume_mass(m);
  
  gp = gr->ip;
  for (i = 0; i < 2; i++) {
    for (j = 0; j < 2; j++) {
      for (k = 0; k < 2; k++) {
        G3D_hex_J(J, ALL_GP_LOCS(i,j,k), ALLC(X));
        volume += GP_WGT(gp,i,j,k) * G3D_hex_detJ(J);
      }
    }
  }

  return (volume_mass * volume);
}

static W_float_t
total_mass (W_elem_t *el)
{
  priv_elem_t *h = (priv_elem_t *)el->data;
  return h->mass;
}

static void
assemble_inertial_forces_diag (W_elem_t *el)
{
  priv_elem_t *h = (priv_elem_t *)el->data;
  W_vect_t f, a;
  int i;
  double m = h->mass / WBMAT_NNODES;

  FOR (i, WBMAT_NNODES) {
    GET_FROM_NODE (h->node[i], WN_get_acceleration, a);
    VECMAC_2_VEC (f, =, -m * a);
    WN_assemble_force (h->node[i], &f);
  }
}

static void
assemble_mass(W_elem_t *el)
{
  W_float_t m;
  priv_elem_t *h = (priv_elem_t *)el->data;
  int i;
  
  m = h->mass / WBMAT_NNODES;

  for (i = 0; i < WBMAT_NNODES; i++)  
    WN_assemble_mass (h->node[i], m);
}


static W_float_t 
suggested_time_step(W_elem_t *el)
{
  priv_elem_t *h = (priv_elem_t *)el->data;
  W_mat_3d_t *m = (W_mat_3d_t *)h->material->data;
  W_coords_3d_t d;
  W_float_t E, nu;
  W_float_t rho, speed_of_sound;
  double shortest_side = FLT_MAX;

  if (!WM3D_get_mat_const(m, MAT_3D_YOUNG_MODULUS, &E))
    REPORT_FATAL_ERROR(el->domain->errh, "No Young's modulus");
  if (!WM3D_get_mat_const(m, MAT_3D_POISSON, &nu))
    REPORT_FATAL_ERROR(el->domain->errh, "No Poisson constant");
  
  rho      = WM3D_get_volume_mass(m);
  speed_of_sound = sqrt(E / (1 - 2*nu) / rho);
#define SUBTR(res, a, b) \
   {res.x = (a).x - (b).x; res.y = (a).y - (b).y; res.z = (a).z - (b).z;}
#define COND_MIN(a, b) \
   { double the_B = b; if (the_B != 0) a = min(a, the_B); }
  SUBTR(d, h->node[1]->ref_coords, h->node[0]->ref_coords);
  COND_MIN(shortest_side, W_vector_3d_length(&d));
  SUBTR(d, h->node[2]->ref_coords, h->node[1]->ref_coords);
  COND_MIN(shortest_side, W_vector_3d_length(&d));
  SUBTR(d, h->node[3]->ref_coords, h->node[2]->ref_coords);
  COND_MIN(shortest_side, W_vector_3d_length(&d));
  SUBTR(d, h->node[0]->ref_coords, h->node[3]->ref_coords);
  COND_MIN(shortest_side, W_vector_3d_length(&d));
  SUBTR(d, h->node[5]->ref_coords, h->node[4]->ref_coords);
  COND_MIN(shortest_side, W_vector_3d_length(&d));
  SUBTR(d, h->node[6]->ref_coords, h->node[5]->ref_coords);
  COND_MIN(shortest_side, W_vector_3d_length(&d));
  SUBTR(d, h->node[7]->ref_coords, h->node[6]->ref_coords);
  COND_MIN(shortest_side, W_vector_3d_length(&d));
  SUBTR(d, h->node[4]->ref_coords, h->node[7]->ref_coords);
  COND_MIN(shortest_side, W_vector_3d_length(&d));
  SUBTR(d, h->node[0]->ref_coords, h->node[4]->ref_coords);
  COND_MIN(shortest_side, W_vector_3d_length(&d));
  SUBTR(d, h->node[1]->ref_coords, h->node[5]->ref_coords);
  COND_MIN(shortest_side, W_vector_3d_length(&d));
  SUBTR(d, h->node[2]->ref_coords, h->node[6]->ref_coords);
  COND_MIN(shortest_side, W_vector_3d_length(&d));
  SUBTR(d, h->node[3]->ref_coords, h->node[7]->ref_coords);
  COND_MIN(shortest_side, W_vector_3d_length(&d));
  
  return shortest_side / speed_of_sound;                 
}


#ifdef X_XFACE
static void
modify_elem_grep(W_elem_t *el, double displ_scale, W_bool_t immed_draw)
{
  priv_elem_t *h = (priv_elem_t *)el->data;
  int i;
  W_vect_t refc, u;
  WCRec pnts[WBMAT_NNODES];

  for (i = 0; i < WBMAT_NNODES; i++) {
    GET_FROM_NODE (h->node[i], WN_get_ref_coords, refc);
    GET_FROM_NODE (h->node[i], WN_get_displ, u);
    pnts[i].x = refc.x + displ_scale * u.x;
    pnts[i].y = refc.y + displ_scale * u.y;
    pnts[i].z = refc.z + displ_scale * u.z;
  }

  if (el->grep == NULL) {
    el->grep = CreateHexahedron(pnts);
    EASValsSetColor(WM_get_color(h->material));
    EASValsSetLayer (WX_ELEM_LAYER);
    EGWithMaskChangeAttributes(COLOR_MASK | LAYER_MASK, el->grep);
    EMNoDrawAddGraphicsToModel(ESIModel (), el->grep);
  } else {
    EGModifyGraphicsGeometry(el->grep, pnts);
  }
}
#endif

static int
flip(void)
{
  long rn;
  int two = 2;

  rn = (long) ((double) LONG_MAX * rand_0_to_1());

  return (((rn / two) * 2) == rn);
}

static int
one_of_four(void)
{
  
  if (flip())
    if (flip()) return 1;
    else        return 2;
  else
    if (flip()) return 3;
    else        return 4;
}

static void
scramble_diagonal(priv_elem_t *h)
{
  int which = one_of_four();
#define FOUR_SHIFT(start)                       \
  {                                             \
    W_node_t *t = h->node[start];               \
    h->node[start  ] = h->node[start+1];        \
    h->node[start+1] = h->node[start+2];        \
    h->node[start+2] = h->node[start+3];        \
    h->node[start+3] = t;                       \
  }
#define SHIFT_ONCE() { FOUR_SHIFT(0); FOUR_SHIFT(4); }

  switch (which) {
  case 1:
    /* do nothing */
    break;
  case 2:
    /* rotate once */
    SHIFT_ONCE();
    break;
  case 3:
    /* rotate twice */
    SHIFT_ONCE();
    SHIFT_ONCE();
    break;
  case 4:
    /* rotate thrice */
    SHIFT_ONCE();
    SHIFT_ONCE();
    SHIFT_ONCE();
    break;
  default:
    break;
  }
}

typedef struct ipol_2x2x2_pkg_t {
  double          stress[3][3];
  double          equiv_pl_strain;
  double          von_mises;
}              ipol_2x2x2_pkg_t;

static ipol_2x2x2_pkg_t ipol_2x2x2_pkgs[WBMAT_NNODES];

static void
fill_ip_2x2x2_pkg(double xi, double eta, double theta,
                  W_mat_3d_state_p matstate,
                  W_mat_3d_ip_ctx_t *ip_ctx,
                  W_mat_3d_t *mat,
                  W_mat_3d_stress_type_enum stress_type
                  )
{
  double stress[3][3];
  double equiv_pl_strain, von_mises;
  W_mat_3d_stress_type_enum out_type;
  int entry = -1;
  
  if        (xi < 0 && eta < 0 && theta < 0) {entry = 0;
  } else if (xi > 0 && eta < 0 && theta < 0) {entry = 1;
  } else if (xi > 0 && eta > 0 && theta < 0) {entry = 2;
  } else if (xi < 0 && eta > 0 && theta < 0) {entry = 3;
  } else if (xi < 0 && eta < 0 && theta > 0) {entry = 4;
  } else if (xi > 0 && eta < 0 && theta > 0) {entry = 5;
  } else if (xi > 0 && eta > 0 && theta > 0) {entry = 6;
  } else if (xi < 0 && eta > 0 && theta > 0) {entry = 7;
  }

  /* Get 1st Piola-Kirchhoff stress at point */
  if (!WM3D_update_stress (mat, matstate, W_MAT_STATE_UPDATE_TEMP,
                        ip_ctx, stress_type, stress,
                        &out_type)) {
    REPORT_ERROR (NULL, "Failed to compute stress");
    return;
  }
  if (out_type != stress_type)
    REPORT_WARNING (NULL, "output stress type mismatch");

  if (!WM3D_get_mat_state_var (mat, matstate,
                               ip_ctx, MAT_3D_EQUIV_PLASTIC_DEF,
                               &equiv_pl_strain))
    REPORT_FATAL_ERROR(NULL, "Failed to retrieve equiv plastic def");

  {
    double means = (stress[0][0] + stress[1][1] + stress[2][2]) / 3;
    double s00 = stress[0][0] - means;
    double s11 = stress[1][1] - means;
    double s22 = stress[2][2] - means;
    
    von_mises
      = sqrt(1.5
             * (  s00*s00
                + stress[0][1]*stress[0][1]
                + stress[0][2]*stress[0][2]
                + stress[1][0]*stress[1][0]
                + s11*s11
                + stress[1][2]*stress[1][2]
                + stress[2][0]*stress[2][0]
                + stress[2][1]*stress[2][1]
                + s22*s22
                )
             );
  }
  
  /* Store into the package */
  MAT3DUTIL_COPY_3X3(ipol_2x2x2_pkgs[entry].stress, stress);
  ipol_2x2x2_pkgs[entry].equiv_pl_strain = equiv_pl_strain;
  ipol_2x2x2_pkgs[entry].von_mises       = von_mises;
}


static void
eval_at_extrapolated_pt (double xi, double eta, double theta,
                         double stress[3][3], double *eps,
                         double *vm)
{
#define SQRT_3 1.73205080756887729352
  double modxi    = SQRT_3 * xi;
  double modeta   = SQRT_3 * eta;
  double modtheta = SQRT_3 * theta;
  double Nmat[WBMAT_NNODES];
  double equiv_pl_strain, von_mises;

  /* Interpolate */
  G3D_hex_N(Nmat, modxi, modeta, modtheta);

#define COMP_T1(T, I, J)				\
  { T[I][J] = (						\
	       Nmat[0] * ipol_2x2x2_pkgs[0].T[I][J] +	\
	       Nmat[1] * ipol_2x2x2_pkgs[1].T[I][J] +	\
	       Nmat[2] * ipol_2x2x2_pkgs[2].T[I][J] +	\
	       Nmat[3] * ipol_2x2x2_pkgs[3].T[I][J] +	\
	       Nmat[4] * ipol_2x2x2_pkgs[4].T[I][J] +	\
	       Nmat[5] * ipol_2x2x2_pkgs[5].T[I][J] +	\
	       Nmat[6] * ipol_2x2x2_pkgs[6].T[I][J] +	\
	       Nmat[7] * ipol_2x2x2_pkgs[7].T[I][J]	\
	       );					\
  }
#define COMP_T(T)						\
  {								\
    COMP_T1(T, 0, 0); COMP_T1(T, 0, 1); COMP_T1(T, 0, 2);	\
    COMP_T1(T, 1, 0); COMP_T1(T, 1, 1); COMP_T1(T, 1, 2);	\
    COMP_T1(T, 2, 0); COMP_T1(T, 2, 1); COMP_T1(T, 2, 2);	\
  }
#define COMP_S(S)				\
  { S = (					\
	 Nmat[0] * ipol_2x2x2_pkgs[0].S +	\
	 Nmat[1] * ipol_2x2x2_pkgs[1].S +	\
	 Nmat[2] * ipol_2x2x2_pkgs[2].S +	\
	 Nmat[3] * ipol_2x2x2_pkgs[3].S +	\
	 Nmat[4] * ipol_2x2x2_pkgs[4].S +	\
	 Nmat[5] * ipol_2x2x2_pkgs[5].S +	\
	 Nmat[6] * ipol_2x2x2_pkgs[6].S +	\
	 Nmat[7] * ipol_2x2x2_pkgs[7].S		\
	 );					\
  }
#define COMP_V1(V, DIR)                         \
  { V.DIR = (                                   \
           Nmat[0] * ipol_2x2x2_pkgs[0].V.DIR + \
           Nmat[1] * ipol_2x2x2_pkgs[1].V.DIR + \
           Nmat[2] * ipol_2x2x2_pkgs[2].V.DIR + \
           Nmat[3] * ipol_2x2x2_pkgs[3].V.DIR + \
           Nmat[4] * ipol_2x2x2_pkgs[4].V.DIR + \
           Nmat[5] * ipol_2x2x2_pkgs[5].V.DIR + \
           Nmat[6] * ipol_2x2x2_pkgs[6].V.DIR + \
           Nmat[7] * ipol_2x2x2_pkgs[7].V.DIR   \
           );                                   \
  }
#define COMP_V(V) { COMP_V1(V, x); COMP_V1(V, y); COMP_V1(V, z); }

  COMP_T(stress);
  COMP_S(equiv_pl_strain); *eps = equiv_pl_strain;
  COMP_S(von_mises); *vm = von_mises;

  return;
}

static void
clear_2x2x2_pkgs (void)
{
  int i;

  FOR (i, WBMAT_NNODES) {
    MAT3DUTIL_ZERO_3X3 (ipol_2x2x2_pkgs[i].stress);
    ipol_2x2x2_pkgs[i].equiv_pl_strain = 0;
    ipol_2x2x2_pkgs[i].von_mises       = 0;
  }
}

static void
accum_unused_2x2x2_pkgs (double xi, double eta, double theta)
{
  int i;
  
  FOR (i, WBMAT_NNODES) {
    if (i != 0 && i != 6) {
      MAT3DUTIL_INCR_A_BY_B_3X3 (ipol_2x2x2_pkgs[i].stress,
                                 ipol_2x2x2_pkgs[0].stress);
      MAT3DUTIL_INCR_A_BY_B_3X3 (ipol_2x2x2_pkgs[i].stress,
                                 ipol_2x2x2_pkgs[6].stress);
      ipol_2x2x2_pkgs[i].von_mises += ipol_2x2x2_pkgs[0].von_mises;
      ipol_2x2x2_pkgs[i].von_mises += ipol_2x2x2_pkgs[1].von_mises;
      ipol_2x2x2_pkgs[i].equiv_pl_strain += ipol_2x2x2_pkgs[0].equiv_pl_strain;
      ipol_2x2x2_pkgs[i].equiv_pl_strain += ipol_2x2x2_pkgs[1].equiv_pl_strain;
    }
  }
}

static void
finish_unused_2x2x2_pkgs (void)
{
  int i;

  FOR (i, WBMAT_NNODES) {
    if (i != 0 && i != 6) {
      MAT3DUTIL_SCALE_3X3 (ipol_2x2x2_pkgs[i].stress, 0.5);
      ipol_2x2x2_pkgs[i].equiv_pl_strain *= 0.5;
      ipol_2x2x2_pkgs[i].von_mises *= 0.5;
    }
  }
}

static void
write_postproc_file (W_elem_t *el, ckit_iobuf_t IOB,
                     int type, char *mode, void *optdata)
{
  priv_elem_t *h = (priv_elem_t *)el->data;
  W_bool_t use_2pt_rule = h->use_2pt_rule;
  int i, j, k;
  W_vect_t refc[WBMAT_NNODES], u[WBMAT_NNODES];
  W_mat_3d_t *m = (W_mat_3d_t *)h->material->data;
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  int pt, ival;
  int reqlen;
  G3D_loc_t *pts[WBMAT_NNODES];
  static int buflen        = 0;
  static double *values[WBMAT_NNODES] = {NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL};
  W_vect_t x[WBMAT_NNODES];
  W_mat_3d_stress_type_enum outtype;
  W_output_stress_type_enum output_stress_type
    = WOPT_OPT (el->domain->options, output_stress_type);
  int output_spec
    = WOPT_OPT (el->domain->options, postfile_output_spec_deprec);

  outtype = (output_stress_type == W_OUTPUT_STRESS_CAUCHY
             ? MAT_3D_CAUCHY
             : (output_stress_type == W_OUTPUT_STRESS_2ND_PK
                ? MAT_3D_2ND_PK
                : MAT_3D_CAUCHY)
             );
             

  if (type == HEXVUE_POSTPROC_FILE) {
    double S[3][3] = {{0,0,0},{0,0,0},{0,0,0}};
    double equiv_pl_def = 0, von_mises = 0;

    reqlen = 0;
    if (output_spec & W_POSTFILE_DISPL_MASK) 
      reqlen += 3;
    if (output_spec & W_POSTFILE_STRESS_MASK)
      reqlen += 6;
    if (output_spec & W_POSTFILE_EQPLS_MASK)
      reqlen++;
    if (output_spec & W_POSTFILE_VON_MISES_MASK)
      reqlen++;
    if (output_spec & W_POSTFILE_MATERIAL_MASK)
      reqlen++;
    
    if (buflen < reqlen) {
      FOR (i, WBMAT_NNODES) {
        values[i] = NEWP_REALLOC_OBJ_ARRAY (values[i], reqlen, double);
      }
      buflen = reqlen;
    }
  
    ip_ctx_data.domain   = el->domain;    
    ip_ctx_data.h        = h;
    ip_ctx.time          = WD_time (el->domain);
    ip_ctx.time_step     = WD_time_step (el->domain);
    ip_ctx.ctx_data      = &ip_ctx_data;
    ip_ctx.get_kinem_var = get_kinem_var;

    FOR (i, WBMAT_NNODES) { GET_FROM_NODE (h->node[i], WN_get_coords, x[i]); }

    clear_2x2x2_pkgs ();
    gp = gr->ip;
    FOR(i, 2) {
      FOR(j, 2) {
        FOR(k, 2) {
          USE_QUADR_PT(h, i, j, k) {
            eval_sf (&ip_ctx_data.sf, ALL_GP_LOCS(i,j,k), x);
            at_new_ip();
            fill_ip_2x2x2_pkg (ALL_GP_LOCS(i,j,k),
                               h->matst[msidcs[i][j][k]],   
                               &ip_ctx, m, outtype);
          } else {
            accum_unused_2x2x2_pkgs (ALL_GP_LOCS(i,j,k));
          }
        }
      }
    }
    if (use_2pt_rule)
      finish_unused_2x2x2_pkgs ();

    FOR (i, WBMAT_NNODES) {
      GET_FROM_NODE (h->node[i], WN_get_ref_coords, refc[i]);
      GET_FROM_NODE (h->node[i], WN_get_displ, u[i]);
    }

    pt = 0;
    FOR(i, 2) {
      FOR(j, 2) {
        FOR(k, 2) {
          pts[pt]   = &refc[pt];
          ival = 0;
          if (output_spec & W_POSTFILE_DISPL_MASK) {
            values[pt][ival] = u[pt].x; ival++;
            values[pt][ival] = u[pt].y; ival++;
            values[pt][ival] = u[pt].z; ival++;
          }

          eval_at_extrapolated_pt (g3dhex_vertex_param_coord[pt][0],
                                   g3dhex_vertex_param_coord[pt][1],
                                   g3dhex_vertex_param_coord[pt][2],
                                   S, &equiv_pl_def, &von_mises);
          
          if (output_spec & W_POSTFILE_STRESS_MASK) {
            values[pt][ival] = S[0][0]; ival++;
            values[pt][ival] = S[0][1]; ival++;
            values[pt][ival] = S[0][2]; ival++;
            values[pt][ival] = S[1][1]; ival++;
            values[pt][ival] = S[1][2]; ival++;
            values[pt][ival] = S[2][2]; ival++;
          }
          if (output_spec & W_POSTFILE_EQPLS_MASK) {
            values[pt][ival] = equiv_pl_def; ival++;
          }
          if (output_spec & W_POSTFILE_VON_MISES_MASK) {
            values[pt][ival] = von_mises; ival++;
          }
          if (output_spec & W_POSTFILE_MATERIAL_MASK) {
            values[pt][ival] = WM_get_id (h->material); ival++;
          }
          if (ival > reqlen) REPORT_FATAL_ERROR (el->domain->errh, "wrong");
          pt++;
        }
      }
    }
    W_write_hexvue_hex (IOB, pts, reqlen, values, mode);
  } else if (type == TECPLOT_I_POSTPROC_FILE) {
    /* RAW */
  } /* type */
}

static void
assemble_eff_stiffness (W_elem_t *el,
                        W_newmark_assemble_eff_stiff_proc assemble_eff_stiff_callback,
                        W_float_t stiff_mult,
                        W_float_t mass_mult,
                        W_mass_type_enum mass_type
                        )
{
  priv_elem_t *h = (priv_elem_t *)el->data;
  W_mat_3d_t *m = (W_mat_3d_t *)h->material->data;
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  double cauchy[3][3];
  W_mat_3d_stress_type_enum out_type;
  double jdetw;
  W_bool_t use_2pt_rule = h->use_2pt_rule;
  double Nx, Ny, Nz;
  int i, j, k, n;
  double rho;
  double C[6][6];
  double K[WBMAT_NNODES*3][WBMAT_NNODES*3];
  W_index_t idcs[WBMAT_NNODES*3];
  W_bool_t constrained[WBMAT_NNODES*3];
  double b[6][WBMAT_NNODES*3];
  W_vect_t x[WBMAT_NNODES];
  
  ip_ctx_data.domain   = el->domain;    
  ip_ctx_data.h        = h;
  ip_ctx.time          = WD_time (el->domain);
  ip_ctx.time_step     = WD_time_step (el->domain);
  ip_ctx.ctx_data      = &ip_ctx_data;
  ip_ctx.get_kinem_var = get_kinem_var;

  rho = WM3D_get_volume_mass (m);

  FOR (i, WBMAT_NNODES*3) {FOR (j, WBMAT_NNODES*3) { K[i][j] = 0; }}
  
  FOR (i, WBMAT_NNODES) { GET_FROM_NODE (h->node[i], WN_get_coords, x[i]); }

  gp = gr->ip;
  FOR(i, 2) {
    FOR(j, 2) {
      FOR(k, 2) {
        USE_QUADR_PT(h, i, j, k) {
          eval_sf (&ip_ctx_data.sf, ALL_GP_LOCS(i,j,k), x);
          at_new_ip();
          jdetw = GP_WGT(gp,i,j,k) * ip_ctx_data.sf.detJ;
          if (use_2pt_rule) {jdetw *= 4;}
          
          eulerian_b_mat (ip_ctx_data.sf.n_x, ip_ctx_data.sf.n_y, ip_ctx_data.sf.n_z, b);
          if (!WM3D_get_mat_stiff (m, h->matst[msidcs[i][j][k]], &ip_ctx,
                                   MAT_3D_STIFF_EULERIAN, C))
            REPORT_FATAL_ERROR (el->domain->errh, "No eulerian material tangent");
          add_bltcbl_upper_triangle (K, b, C, stiff_mult * jdetw); 
          if (!WM3D_update_stress (m, h->matst[msidcs[i][j][k]], W_MAT_STATE_UPDATE_TEMP,
                                   &ip_ctx, MAT_3D_CAUCHY, cauchy, &out_type))
            REPORT_ERROR(el->domain->errh, "Failed to compute stress");
          add_eulerian_istress_upper_triangle (K, ip_ctx_data.sf.n_x, ip_ctx_data.sf.n_y, ip_ctx_data.sf.n_z,
                                               cauchy, stiff_mult * jdetw);
          switch (mass_type) {
          case W_MASS_CONSISTENT:
            add_rhontn_upper_triangle (K, ip_ctx_data.sf.n, mass_mult * rho * jdetw); 
            break;
          default:
            /* VOID: see the switch below */
            break;
          }
        } /* USE_QUADR_PT */
      } /* FOR(k, 2) */
    } /* FOR(j, 2) */
  } /* FOR(i, 2) */

  switch (mass_type) {
  case W_MASS_CONSISTENT:
    /* VOID: handled in the quadrature loop above */
    break;
  case W_MASS_DIAG:
  default:
    add_to_diag (K, mass_mult * h->mass / WBMAT_NNODES);
    break;
  }
  
  mirror (K);
  
  WN_eqn_numbers_and_constraints (h->node[0], &idcs[0], &constrained[0]);
  WN_eqn_numbers_and_constraints (h->node[1], &idcs[3], &constrained[3]);
  WN_eqn_numbers_and_constraints (h->node[2], &idcs[6], &constrained[6]);
  WN_eqn_numbers_and_constraints (h->node[3], &idcs[9], &constrained[9]);
  WN_eqn_numbers_and_constraints (h->node[4], &idcs[12], &constrained[12]);
  WN_eqn_numbers_and_constraints (h->node[5], &idcs[15], &constrained[15]);
  WN_eqn_numbers_and_constraints (h->node[6], &idcs[18], &constrained[18]);
  WN_eqn_numbers_and_constraints (h->node[7], &idcs[21], &constrained[21]);

  (*assemble_eff_stiff_callback)  (el->domain->newmark_integrator, FLES_ADD,
                                   WBMAT_NNODES*3, idcs, (double *) K, constrained);
}

static void
assemble_residual_forces (W_elem_t *el,
                          W_mat_state_update_form_enum update_form,
                          W_mass_type_enum mass_type)
{
  assemble_restoring_forces (el, update_form);
  switch (mass_type) {
  case W_MASS_DIAG:
    assemble_inertial_forces_diag (el);
    break;
  case W_MASS_CONSISTENT:
  default:
    assemble_inertial_forces_consistent (el);
    break;
  }
}


static int
which_qp (priv_elem_t *h, W_loc_t *loc)
{
  int i, j, k, m;
  double n[8];
  int closest = -1;
  double mindist, dist[8]
    = {FLT_MAX,FLT_MAX,FLT_MAX,FLT_MAX,FLT_MAX,FLT_MAX,FLT_MAX,FLT_MAX};
  W_loc_t l;

  gp = gr->ip;
  for (i = 0; i < 2; i++) {
    for (j = 0; j < 2; j++) {
      for (k = 0; k < 2; k++) {
        if (h->matst[msidcs[i][j][k]]) {
          G3D_hex_N (n, ALL_GP_LOCS (i,j,k));
          VECMAC_SCALAR (l, =, 0);
          FOR (m, 8) {
            VECMAC_2_VEC (l, +=, n[m] * h->node[m]->ref_coords);
          }
          VECMAC_2_VEC (l, -=, (*loc));
          dist[msidcs[i][j][k]] = VECMAC_NORM (l);
        }
      }
    }
  }
  closest = -1;
  mindist = FLT_MAX;
  FOR (i, 8) {
    if (dist[i] < FLT_MAX && dist[i] < mindist) {
      closest = i;
      mindist = dist[i];
    }
  }
  return closest;
}

/* history client data */
typedef struct history_client_data_t {
  W_elem_t          *el; /* FE */
  W_index_t          pntidx; /* unique integer  */
  int                which_qp;  /* which quadrature point */
}              history_client_data_t;

static void *
new_history_client_data (W_elem_t *e,
                         W_index_t idx, /* unique integer  */
                         W_history_writer_t *r, /* history writer */
                         char *input_line /* input line (to allow
                                             for additional parsing of
                                             the line contents by this
                                             function) */
                         )
{
  priv_elem_t *h = (priv_elem_t *)e->data;
  history_client_data_t *cd;
  W_loc_t loc;
  W_domain_t *d = e->domain;
  
  cd = w_new (history_client_data_t);
  CHECK_MALLOC(cd);
 
  TOKENS_PARSE(input_line);
  cd->el     = e;
  cd->pntidx = TOKEN_INT(3);
  loc.x  = TOKEN_DBL(4);
  loc.y  = TOKEN_DBL(5);
  loc.z  = TOKEN_DBL(6);
  cd->which_qp = which_qp (h, &loc);
  if (cd->which_qp < 0 && cd->which_qp >= 8)
    REPORT_FATAL_ERROR (d->errh, "Wrong quadrature point");

  return (NODE)cd;
}

static void
calc_stress (W_elem_t *el,
            int which_qp,
            W_output_stress_type_enum stress_type, /* stress type */
            double stress[3][3] /* output: stress */
            )
{
  priv_elem_t *h = (priv_elem_t *)el->data;
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  W_mat_3d_stress_type_enum out_type;
  W_mat_3d_t *m = (W_mat_3d_t *)h->material->data;
  int i;
  W_vect_t x[WBMAT_NNODES];

  ip_ctx_data.domain   = el->domain;    
  ip_ctx_data.h        = h;
  ip_ctx.time          = WD_time (el->domain);
  ip_ctx.time_step     = WD_time_step (el->domain);
  ip_ctx.ctx_data      = &ip_ctx_data;
  ip_ctx.get_kinem_var = get_kinem_var;
  
  FOR (i, WBMAT_NNODES) { GET_FROM_NODE (h->node[i], WN_get_coords, x[i]); }

  eval_sf (&ip_ctx_data.sf,
           ALL_GP_LOCS(invmsidcs[which_qp][0],
                       invmsidcs[which_qp][1],
                       invmsidcs[which_qp][2]), x);
  at_new_ip();
  if      (stress_type == W_OUTPUT_STRESS_CAUCHY) {
    if (!WM3D_update_stress (m, h->matst[which_qp], W_MAT_STATE_UPDATE_TEMP,
                         &ip_ctx, MAT_3D_CAUCHY, stress, &out_type))
      REPORT_FATAL_ERROR(el->domain->errh, "Failed to compute stress");
    if (out_type != MAT_3D_CAUCHY)
      REPORT_FATAL_ERROR(el->domain->errh, "Stress type not available");
  } else if (stress_type == W_OUTPUT_STRESS_2ND_PK) {
    if (!WM3D_update_stress (m, h->matst[which_qp], W_MAT_STATE_UPDATE_TEMP,
                         &ip_ctx, MAT_3D_2ND_PK, stress, &out_type))
      REPORT_FATAL_ERROR(el->domain->errh, "Failed to compute stress");
    if (out_type != MAT_3D_2ND_PK)
      REPORT_FATAL_ERROR(el->domain->errh, "Stress type not available");
  }
}



static void
get_equiv_pl_def (W_elem_t *el,
                  int which_qp,
                  double *equiv_pl_def
                  )
{
  priv_elem_t *h = (priv_elem_t *)el->data;
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  int i;
  W_mat_3d_t *m = (W_mat_3d_t *)h->material->data;
  W_vect_t x[WBMAT_NNODES];

  ip_ctx_data.domain   = el->domain;    
  ip_ctx_data.h        = h;
  ip_ctx.time          = WD_time (el->domain);
  ip_ctx.time_step     = WD_time_step (el->domain);
  ip_ctx.ctx_data      = &ip_ctx_data;
  ip_ctx.get_kinem_var = get_kinem_var;
  
  FOR (i, WBMAT_NNODES) { GET_FROM_NODE (h->node[i], WN_get_coords, x[i]); }

  eval_sf (&ip_ctx_data.sf,
           ALL_GP_LOCS(invmsidcs[which_qp][0],
                       invmsidcs[which_qp][1],
                       invmsidcs[which_qp][2]), x);
  at_new_ip();
  if (!WM3D_get_mat_state_var(m, h->matst[which_qp],
                              &ip_ctx, MAT_3D_EQUIV_PLASTIC_DEF, equiv_pl_def))
    REPORT_FATAL_ERROR(el->domain->errh, "Failed to retrieve equiv plastic def");
}


static void
update_history (W_elem_t *e, /* generic element */
                W_solution_history_t *h, /* solution history */
                W_float_t time
                )
{
  history_client_data_t *cd =
    (history_client_data_t *)h->client_data;
  double S[3][3], pressure, equiv_pl_def;
#define STORE(CD, HW, NAME, TIME, VAL)          \
  {                                             \
    char buf[128];                              \
    sprintf (buf, "p%d%s", CD->pntidx, NAME);   \
    WHR_store (HW, buf, TIME, VAL);             \
  }

  calc_stress (e, cd->which_qp,
              WOPT_OPT (h->writer->domain->options, output_stress_type),
              S);
  pressure  = (S[0][0] + S[1][1] + S[2][2]) / 3;
  if        (WOPT_OPT (h->writer->domain->options, output_stress_type) == W_OUTPUT_STRESS_CAUCHY) {
    STORE (cd, h->writer, "sxx", time, S[0][0]);
    STORE (cd, h->writer, "sxy", time, S[0][1]);
    STORE (cd, h->writer, "sxz", time, S[0][2]);
    STORE (cd, h->writer, "syy", time, S[1][1]);
    STORE (cd, h->writer, "syz", time, S[1][2]);
    STORE (cd, h->writer, "szz", time, S[2][2]);
    STORE (cd, h->writer, "press", time, pressure);
  } else if (WOPT_OPT (h->writer->domain->options, output_stress_type) == W_OUTPUT_STRESS_2ND_PK) {
    STORE (cd, h->writer, "Sxx", time, S[0][0]);
    STORE (cd, h->writer, "Sxy", time, S[0][1]);
    STORE (cd, h->writer, "Sxz", time, S[0][2]);
    STORE (cd, h->writer, "Syy", time, S[1][1]);
    STORE (cd, h->writer, "Syz", time, S[1][2]);
    STORE (cd, h->writer, "Szz", time, S[2][2]);
    STORE (cd, h->writer, "Press", time, pressure);
  } else {
    REPORT_ERROR(h->writer->domain->errh, "Unknown stress type for output");
  }
  get_equiv_pl_def (e, cd->which_qp, &equiv_pl_def);
  STORE (cd, h->writer, "EPlS", time, equiv_pl_def);
}


static W_float_t
kinetic_energy_consistent (W_elem_t *el)
{
  priv_elem_t *h = (priv_elem_t *)el->data;
  W_mat_3d_t *m = (W_mat_3d_t *)h->material->data;
  W_vect_t v[WBMAT_NNODES];
  double Nmat[WBMAT_NNODES];
  double jdetw, J[3][3], rho;
  W_bool_t use_2pt_rule = h->use_2pt_rule;
  int i, j, k, n, l;
  double dke, ke = 0;
  double vdotv[WBMAT_NNODES][WBMAT_NNODES];
  W_vect_t X[WBMAT_NNODES];

  FOR (i, WBMAT_NNODES) {
    GET_FROM_NODE (h->node[i], WN_get_velocity, v[i]);
    GET_FROM_NODE (h->node[i], WN_get_ref_coords, X[i]);
  }

  /* note that we fill only the upper triangle */
  FOR (i, WBMAT_NNODES) {
    FOR (j, WBMAT_NNODES) {
      vdotv[i][j] = VECMAC_DOT (v[i], v[j]);
    }
  }

  rho = WM3D_get_volume_mass(m);

  gp = gr->ip;
  FOR(i, 2) {
    FOR(j, 2) {
      FOR(k, 2) {
        USE_QUADR_PT(h, i, j, k) {
          G3D_hex_N (Nmat, ALL_GP_LOCS(i,j,k));
          G3D_hex_J (J, ALL_GP_LOCS(i,j,k), ALLC(X));
          jdetw = GP_WGT(gp,i,j,k) * G3D_hex_detJ (J);
          if (use_2pt_rule) {jdetw *= 4;}
          FOR(n, WBMAT_NNODES) {
            FOR (l, WBMAT_NNODES) {
              dke = jdetw * Nmat[n] * Nmat[l] * vdotv[n][l];
              ke  += dke;
            }
          }
        } /* USE_QUADR_PT */
      } /* FOR(k, 2) */
    } /* FOR(j, 2) */
  } /* FOR(i, 2) */

  return (rho * ke / 2);
}

static W_float_t
kinetic_energy_diag (W_elem_t *el)
{
  priv_elem_t *e = (priv_elem_t *)el->data;
  W_vect_t v;
  int i;
  double ke = 0;
  double m = e->mass / WBMAT_NNODES;

  FOR (i, WBMAT_NNODES) {
    GET_FROM_NODE (e->node[i], WN_get_velocity, v);
    ke += VECMAC_DOT (v, v);
  }
  return (m * ke) / 2;
}

static W_float_t
kinetic_energy (W_elem_t *el, W_mass_type_enum mass_type)
{
  switch (mass_type) {
  case W_MASS_CONSISTENT:
    return kinetic_energy_consistent (el);
  default:
    return kinetic_energy_diag (el);
  }
}

static W_elem_methods_t methods = {
  elem_elcode,
  type_name,
  assemble_restoring_forces,
  assemble_mass,
  total_mass,
  NULL,
  suggested_time_step,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  new_history_client_data,
  update_history,
  NULL,
  write_postproc_file,
#ifdef X_XFACE
  modify_elem_grep,
#endif
  NULL,
  NULL,
  NULL,
  num_of_faces,
  get_face_vertex_nodes,
  get_contact_faces,
  assemble_residual_forces,
  assemble_eff_stiffness,
  kinetic_energy
};

/*
  Element reader.
  Reads element data for the element type
  from stream as long as there is no
  END_OF_BLOCK tag in the input line. Creates elements.
 */
W_bool_t /* TRUE if successful */
eulhex_reader (W_domain_t *d /* domain */)
{
  const char *input_line;
  long elem_num;
  priv_elem_t *h;
#ifdef PVM_NEED_GHOST_ELEMENTS
  int is_ghost = FALSE;
#endif
  W_index_t nodeid, matid;
  int i, j, k;
  W_material_t *mat;
  W_mat_3d_t *m;
  W_bool_t use_2pt_rule = FALSE;
  double nmat[WBMAT_NNODES];
#define LOC(h, nmat, loc)                                               \
  {                                                                     \
    int m_;                                                             \
    VECMAC_SCALAR (loc, =, 0);                                          \
    FOR (m_, WBMAT_NNODES) {                                            \
      VECMAC_2_VEC (loc, +=, nmat[m_] * h->node[m_]->ref_coords);       \
    }                                                                   \
  }

  if (gr == NULL) {
    gr = WG_get_integration_rule(2);
  }

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line)) {
      return TRUE;
    } else {
      TOKENS_PARSE(input_line);
      if (TOKENS_HAVE_KEYWORD("USE_2PT_RULE")) {
        use_2pt_rule = TRUE;
      } else if (TOKENS_HAVE_KEYWORD("DO_NOT_USE_2PT_RULE")) {
        use_2pt_rule = FALSE;
      } else { /* ordinary line with element data */
        h = w_new(priv_elem_t);
        CHECK_MALLOC(h);
        elem_num             = TOKEN_INT(1);
        FOR(i, WBMAT_NNODES) {
          nodeid = TOKEN_INT(2+i);
          h->node[i] = WD_get_node_by_idx(d, nodeid);
          if (h->node[i] == NULL) {
            sprintf(W_buf(), "No such node: %ld", nodeid);
            REPORT_ERROR(d->errh, W_buf());
          }
        }
        matid = TOKEN_INT(10);
        mat = WD_get_material_by_idx(d, matid);
        if (mat == NULL) {
          sprintf(W_buf(), "No such material: %ld", matid);
          REPORT_ERROR(d->errh, W_buf());
        }
        h->material     = mat;
        m = (W_mat_3d_t *)h->material->data;
        h->use_2pt_rule = use_2pt_rule;
        h->matst        = w_new_array (h->use_2pt_rule ? 2 : 9, W_mat_3d_state_p);
        CHECK_MALLOC (h->matst);
        gp = gr->ip;
        if (h->use_2pt_rule) {
          W_loc_t loc={0,0,0};
          scramble_diagonal(h);
          i = j = k = 0;
          G3D_hex_N(nmat, ALL_GP_LOCS(i,j,k));
          LOC(h, nmat, loc);
          h->matst[msidcs[i][j][k]] = WM3D_get_mat_state(m, &loc);
          i = j = k = 1;
          G3D_hex_N(nmat, ALL_GP_LOCS(i,j,k));
          LOC(h, nmat, loc);
          h->matst[msidcs[i][j][k]] = WM3D_get_mat_state(m, &loc);
        } else {
          FOR(i, 2) {
            FOR(j, 2) {
              FOR(k, 2) {
                W_loc_t loc;
                G3D_hex_N(nmat, ALL_GP_LOCS(i,j,k));
                LOC(h, nmat, loc);
                h->matst[msidcs[i][j][k]] = WM3D_get_mat_state(m, &loc);
              }
            }
          }
        }
        h->mass = compute_total_mass (h);
#ifdef PVM_NEED_GHOST_ELEMENTS
        is_ghost = FALSE;
        if (TOKENS_HAVE_KEYWORD("GHOST"))
          is_ghost = TRUE;
#endif      
        WD_add_element(d, elem_num,
                       WE_new_element(d, &methods, h
#                                   ifdef PVM_NEED_GHOST_ELEMENTS
                                      , is_ghost
#                                   endif
                                      ));
      } /* ordinary data line */
    } /* closing else */
  }
  return FALSE;
}

