#include "Wconfig.h"


#include "Wtypes.h"
#include "Wftet10.h"
#include "Wdomain.h"
#include "Wnode.h"
#include "Wmats.h"
#include "Wgaussp.h"
#include "Wxf.h"
#include "Wutil.h"
#include "Wcdi.h"
#include "g3d.h"
#include "Wmat3d.h"
#include "Wcontface.h"

/* RCSHEADER "$Id: ftet10.c,v 1.3 2002/03/21 01:32:35 cvs Exp $" */

#define GET_FROM_NODE(node, func, tovect) \
   func (node, &tovect.x, &tovect.y, &tovect.z) 


typedef struct tet10_shape_func_t     {
  double n[10];
  double n_x[10];
  double n_y[10];
  double n_z[10];
  double J[3][3];
  double invJ[3][3];
  double detJ;
}                  tet10_shape_func_t;

typedef struct gp_data_t     {
  W_mat_3d_state_p   matst;
  double            *n_x;
  double            *n_y;
  double            *n_z;
  double             detJ;
}                  gp_data_t;

typedef struct priv_elem_t     {
  W_node_t          *node[10];
  W_material_t      *material;
  gp_data_t          gpdata[4];
}                  priv_elem_t;


typedef struct gp_coord_t     {
  double r, s, t, u;
}                  gp_coord_t;

static gp_coord_t gps[4] =     {
      { 0.13819660,0.13819660,0.13819660,0.58541020 }    ,
      { 0.58541020,0.13819660,0.13819660,0.13819660 }    ,
      { 0.13819660,0.58541020,0.13819660,0.13819660 }    ,
      { 0.13819660,0.13819660,0.58541020,0.13819660 }    ,
}    ;
/* The following factor includes W_l=1/4 and c=1/6;
   see Hughes, The finite element method, 1987. */
#define GPW 0.041666666666666666667 /* = (1.0/4/6) */


static int
elem_elcode (W_elem_t *el);
static char *
type_name (W_elem_t *el);
static void
assemble_restoring_forces (W_elem_t *el, W_mat_state_update_form_enum update_form);
static void
assemble_mass (W_elem_t *el);
static W_float_t 
suggested_time_step (W_elem_t *el);


#ifdef X_XFACE
static void
modify_elem_grep (W_elem_t *el, double displ_scale, W_bool_t immed_draw) 
{
  priv_elem_t *tet = (priv_elem_t *) el->data;
  int i;
  W_vect_t refc, u;
  WCRec pnts[10];

  for (i = 0; i < 10; i++) {
    GET_FROM_NODE (tet->node[i], WN_get_ref_coords, refc);
    GET_FROM_NODE (tet->node[i], WN_get_displ, u);
    pnts[i].x = refc.x + displ_scale * u.x;
    pnts[i].y = refc.y + displ_scale * u.y;
    pnts[i].z = refc.z + displ_scale * u.z;
  }    

  if (el->grep == NULL) {
    el->grep = CreateTet10 (pnts);
    EASValsSetColor (WM_get_color (tet->material));
    EGWithMaskChangeAttributes (COLOR_MASK, el->grep);
    EMNoDrawAddGraphicsToModel (ESIModel (), el->grep);
  }     else     {
    EGModifyGraphicsGeometry (el->grep, pnts);
  }    
}    
#endif

static W_float_t
total_mass (W_elem_t *el);
static void
write_postproc_file (W_elem_t *el, ckit_iobuf_t IOB, int type,
                     char *mode, void *optdata);



static void
fill_ip_pkg (double r, double s, double t, double u,
             int entry,
             G3D_loc_t *iploc,
             W_mat_3d_state_p matstate,
             W_mat_3d_ip_ctx_t *ip_ctx,
             W_mat_3d_t *mat,
             W_mat_3d_stress_type_enum stress_type
             );
static void
eval_at_extrapolated_pt (G3D_loc_t *pt,
                         double stress[3][3], double *eps,
                         double *vm);

static void
clear_pkgs (void);

static int 
num_of_faces (W_elem_t *el) 
{
  return 4;
}    

static int
get_face_vertex_nodes (W_elem_t *el, int nface, int ndim, W_node_t *n[]) 
{
  priv_elem_t *tet10 = (priv_elem_t *) el->data;
  int i;

  if (ndim >= 6) {
    FOR (i, 6) {
      n[i] = tet10->node[g3dtet10_vertex_on_face[nface][i]];
    }    
  }
  return 6;
}    

static int
get_contact_faces (W_elem_t *el, int nface, WCF_contact_face_t *cf[], int maxcf) 
{
  priv_elem_t *tet10 = (priv_elem_t *) el->data;
  int i;
  
  if (maxcf < 4) {
    FOR (i, maxcf) { cf[i] = NULL; }    
    return 4;
  }    
  
#define DEFCF(cfn)                                                      \
      {                                                                 \
    cf[cfn] =                                                           \
       WCF_new_contact_face (                                           \
           tet10->node[g3dtet10_subtri_verts[nface][cfn][0]],           \
           tet10->node[g3dtet10_subtri_verts[nface][cfn][1]],           \
           tet10->node[g3dtet10_subtri_verts[nface][cfn][2]], NULL);    \
  }    

  DEFCF (0);
  DEFCF (1);
  DEFCF (2);
  DEFCF (3);

  return 4;
}    

static W_elem_methods_t tet10_methods =     {
  elem_elcode,
  type_name,
  assemble_restoring_forces,
  assemble_mass,
  total_mass,
  NULL,
  suggested_time_step,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  write_postproc_file,
#ifdef X_XFACE
  modify_elem_grep,
#endif
  NULL,
  NULL,
  NULL,
  num_of_faces,
  get_face_vertex_nodes,
  get_contact_faces
}    ;
  




typedef struct ip_ctx_data_t     {
  priv_elem_t                 *tet10;
  double                  *n_x, *n_y, *n_z;
}                  ip_ctx_data_t;

static void
at_new_ip (void);
static W_bool_t 
get_kinem_var (W_mat_3d_ip_ctx_t *ctx, 
              W_mat_3d_kinem_var_type_enum  type,
              double kinvar[3][3]);


#define ALL_GP_LOCS(i) gps[i].r, gps[i].s, gps[i].t, gps[i].u
#define ALL_REF_COORDS(H) \
 H->node[0]->ref_coords.x, H->node[0]->ref_coords.y, H->node[0]->ref_coords.z,\
 H->node[1]->ref_coords.x, H->node[1]->ref_coords.y, H->node[1]->ref_coords.z,\
 H->node[2]->ref_coords.x, H->node[2]->ref_coords.y, H->node[2]->ref_coords.z,\
 H->node[3]->ref_coords.x, H->node[3]->ref_coords.y, H->node[3]->ref_coords.z,\
 H->node[4]->ref_coords.x, H->node[4]->ref_coords.y, H->node[4]->ref_coords.z,\
 H->node[5]->ref_coords.x, H->node[5]->ref_coords.y, H->node[5]->ref_coords.z,\
 H->node[6]->ref_coords.x, H->node[6]->ref_coords.y, H->node[6]->ref_coords.z,\
 H->node[7]->ref_coords.x, H->node[7]->ref_coords.y, H->node[7]->ref_coords.z,\
 H->node[8]->ref_coords.x, H->node[8]->ref_coords.y, H->node[8]->ref_coords.z,\
 H->node[9]->ref_coords.x, H->node[9]->ref_coords.y, H->node[9]->ref_coords.z


static tet10_shape_func_t tet10_shape_func;
  
#define IPOL10(n, v0, v1, v2, v3, v4, v5, v6, v7, v8, v9) \
 (  n[0] * (v0) \
         + n[1] * (v1) \
         + n[2] * (v2) \
         + n[3] * (v3) \
         + n[4] * (v4) \
         + n[5] * (v5) \
         + n[6] * (v6) \
         + n[7] * (v7) \
         + n[8] * (v8) \
         + n[9] * (v9)) 

#define N_LOC(h, v, DIR, n) \
   v.DIR = (IPOL10 (n,                               \
                    h->node[0]->ref_coords.DIR,      \
                    h->node[1]->ref_coords.DIR,      \
                    h->node[2]->ref_coords.DIR,      \
                    h->node[3]->ref_coords.DIR,      \
                    h->node[4]->ref_coords.DIR,      \
                    h->node[5]->ref_coords.DIR,      \
                    h->node[6]->ref_coords.DIR,      \
                    h->node[7]->ref_coords.DIR,      \
                    h->node[8]->ref_coords.DIR,      \
                    h->node[9]->ref_coords.DIR)) 

#define LOC(h, gp) \
      {                                             \
    N_LOC (h, gp, x, tet10_shape_func.n);          \
    N_LOC (h, gp, y, tet10_shape_func.n);          \
    N_LOC (h, gp, z, tet10_shape_func.n);          \
  }    

/*
  Element reader.
  Reads element data for the pcont3 element type
  from stream as long as there is no
  END_OF_BLOCK tag in the input line. Creates elements.
 */
W_bool_t                        /* TRUE if successful */
ftet10_reader (W_domain_t *d /* domain */) 
{
  const char *input_line;
  long elem_num; 
  priv_elem_t *tet10;
#ifdef PVM_NEED_GHOST_ELEMENTS
  int is_ghost = FALSE;
#endif
  W_index_t nodeid, matid;
  int i, k;
  W_material_t *mat;
  W_mat_3d_t *m;

  while ((input_line = W_get_input_line (d)) ) {
    if (W_is_end_of_block (input_line)) {
      return TRUE;
    }     else     {
      TOKENS_PARSE (input_line);
      tet10 = w_new (priv_elem_t);
      CHECK_MALLOC (tet10);
      elem_num = TOKEN_INT (1);
      FOR (i, 10) {
        nodeid = TOKEN_INT (2+i);
        tet10->node[i] = WD_get_node_by_idx (d, nodeid);
        if (tet10->node[i] == NULL) {
          sprintf (W_buf (), "No such node: %ld", nodeid);
          REPORT_ERROR (d->errh, W_buf ());
        }    
      }    
      matid = TOKEN_INT (12);
      mat = WD_get_material_by_idx (d, matid);
      if (mat == NULL) {
        sprintf (W_buf (), "No such material: %ld", matid);
        REPORT_ERROR (d->errh, W_buf ());
      }    
      tet10->material     = mat;
      m = (W_mat_3d_t *) tet10->material->data;
      FOR (k, 4) {
        W_loc_t loc;
        G3D_tet10_N (tet10_shape_func.n, ALL_GP_LOCS (k));
        LOC (tet10, loc);
        tet10->gpdata[k].matst = WM3D_get_mat_state (m, &loc);
        tet10->gpdata[k].n_x = tet10->gpdata[k].n_y = tet10->gpdata[k].n_z = NULL;
        tet10->gpdata[k].detJ = 0;
      }    
#ifdef PVM_NEED_GHOST_ELEMENTS
        is_ghost = FALSE;
        if (TOKENS_HAVE_KEYWORD ("GHOST")) 
          is_ghost = TRUE;
#endif      
        WD_add_element (d, elem_num,
                       WE_new_element (d, &tet10_methods, tet10
#                                   ifdef PVM_NEED_GHOST_ELEMENTS
                                      , is_ghost
#                                   endif
                                      ));
    }     /* closing else */
  }    
  return FALSE;
}    



static int
elem_elcode (W_elem_t *el) 
{
  return FTET10_ELCODE;
}    


static char *
type_name (W_elem_t *el) 
{
  static char buf[] = "ftet10";
  return buf;
}    








/* Returns nodal forces by which the  element acts on its nodal points. */

static void
assemble_restoring_forces (W_elem_t *el, W_mat_state_update_form_enum update_form) 
{
  priv_elem_t *tet10 = (priv_elem_t *) el->data;
  W_mat_3d_t *m = (W_mat_3d_t *) tet10->material->data;
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  double P[3][3];
  W_mat_3d_stress_type_enum out_type;
  W_vect_t f;
  double jdetw, ded;
  double dphix, dphiy, dphiz;
  int k, n;
#define EVAL_SF(k) \
      {                                                                                             \
    if (!tet10->gpdata[k].n_x) {                                                                \
      int _counter_i;                                                                           \
      G3D_tet10_interp_opt (tet10_shape_func.n,                                                 \
                            tet10_shape_func.n_x,                                               \
                            tet10_shape_func.n_y,                                               \
                            tet10_shape_func.n_z,                                               \
                            &tet10_shape_func.detJ,                                             \
                            tet10_shape_func.J,                                                 \
                            tet10_shape_func.invJ,                                              \
                            ALL_GP_LOCS (k),                                                     \
                            ALL_REF_COORDS (tet10));                                             \
      tet10->gpdata[k].n_x = w_new_array (10, double); CHECK_MALLOC (tet10->gpdata[k].n_x);     \
      tet10->gpdata[k].n_y = w_new_array (10, double); CHECK_MALLOC (tet10->gpdata[k].n_y);     \
      tet10->gpdata[k].n_z = w_new_array (10, double); CHECK_MALLOC (tet10->gpdata[k].n_z);     \
      FOR (_counter_i, 10) {                                                                    \
        tet10->gpdata[k].n_x[_counter_i] = tet10_shape_func.n_x[_counter_i];                    \
        tet10->gpdata[k].n_y[_counter_i] = tet10_shape_func.n_y[_counter_i];                    \
        tet10->gpdata[k].n_z[_counter_i] = tet10_shape_func.n_z[_counter_i];                    \
      }                                                                                             \
      tet10->gpdata[k].detJ = tet10_shape_func.detJ;                                            \
    }                                                                                               \
  }    
#define COMP_F() \
      {                                                         \
    f.x = - (P[0][0]*dphix + P[0][1]*dphiy + P[0][2]*dphiz); \
    f.y = - (P[1][0]*dphix + P[1][1]*dphiy + P[1][2]*dphiz); \
    f.z = - (P[2][0]*dphix + P[2][1]*dphiy + P[2][2]*dphiz); \
  }    
  
  ip_ctx_data.tet10    = tet10;
  ip_ctx.time          = WD_time (el->domain);
  ip_ctx.time_step     = WD_time_step (el->domain);
  ip_ctx.ctx_data      = &ip_ctx_data;
  ip_ctx.get_kinem_var = get_kinem_var;

  el->deformation_energy = 0;
  FOR (k, 4) {
    EVAL_SF (k);
    ip_ctx_data.n_x = tet10->gpdata[k].n_x;
    ip_ctx_data.n_y = tet10->gpdata[k].n_y;
    ip_ctx_data.n_z = tet10->gpdata[k].n_z;
    at_new_ip ();
    if (!WM3D_update_stress (m, tet10->gpdata[k].matst, update_form,
                            &ip_ctx, MAT_3D_1ST_PK, P, &out_type)) 
      REPORT_ERROR (el->domain->errh, "Failed to compute stress");
    if (!WM3D_get_mat_state_var (m, tet10->gpdata[k].matst,
                                 &ip_ctx, MAT_3D_DEFORM_ENERGY_DENSITY,
                                 &ded))
      REPORT_ERROR(el->domain->errh, "Failed to retrieve work of deformation");
    jdetw = GPW * tet10->gpdata[k].detJ;
    el->deformation_energy += ded * jdetw;
    FOR (n, 10) {
      dphix = jdetw * tet10->gpdata[k].n_x[n];
      dphiy = jdetw * tet10->gpdata[k].n_y[n];
      dphiz = jdetw * tet10->gpdata[k].n_z[n];
      COMP_F ();
      WN_assemble_force (tet10->node[n], &f);
    }    
  }     /* FOR (k, 2) */
#undef  EVAL_SF
}    

static void
assemble_mass (W_elem_t *el) 
{
  priv_elem_t *tet10 = (priv_elem_t *) el->data;
  int i;
  double tm = total_mass (el);

  for (i = 0; i < 4; i++) 
    WN_assemble_mass (tet10->node[i], tm / 32.0);
  for (i = 4; i < 10; i++) 
    WN_assemble_mass (tet10->node[i], 7 * tm / 48.0);
}    


static W_float_t 
suggested_time_step (W_elem_t *el) 
{
  priv_elem_t *tet10 = (priv_elem_t *) el->data;
  W_vect_t refc;
  G3D_box_t b;
  int k;
  double E, nu, rho, xs, ys, zs, mins, c_d;
  W_mat_3d_t *m = (W_mat_3d_t *) tet10->material->data;

  GET_FROM_NODE (tet10->node[0], WN_get_ref_coords, refc);
  G3D_INIT_BOX (b, refc);
  for (k = 1; k < 10; k++) {
    GET_FROM_NODE (tet10->node[k], WN_get_ref_coords, refc);
    G3D_UPDT_BOX (b, refc);
  }    
  xs = G3D_BOX_RANGE (b, x);
  ys = G3D_BOX_RANGE (b, y);
  zs = G3D_BOX_RANGE (b, z);
  mins = xs; mins = min (mins, ys); mins = min (mins, zs);

  if (!WM3D_get_mat_const (m, MAT_3D_YOUNG_MODULUS, &E)) 
    REPORT_FATAL_ERROR (el->domain->errh, "No Young's modulus");
  if (!WM3D_get_mat_const (m, MAT_3D_POISSON, &nu)) 
    REPORT_FATAL_ERROR (el->domain->errh, "No Poisson constant");
  rho = WM3D_get_volume_mass (m);
  c_d = W_3D_dilatational_wave_speed (E, nu, rho);
  return mins / c_d / 30;
}    


static W_float_t
total_mass (W_elem_t *el) 
{
  priv_elem_t *tet10 = (priv_elem_t *) el->data;
  W_mat_3d_t *m = (W_mat_3d_t *) tet10->material->data;
  W_float_t volume_mass, J[3][3];
  int k;
  double volume = 0;

  volume_mass = WM3D_get_volume_mass (m);
  
  FOR (k, 4) {
    G3D_tet10_J (J, ALL_GP_LOCS (k), ALL_REF_COORDS (tet10));
    volume += GPW * G3D_tet10_detJ (J);
  }    

  return (volume_mass * volume);
}    








static void
write_postproc_file (W_elem_t *el, ckit_iobuf_t IOB,
                     int type, char *mode, void *optdata) 
{
  priv_elem_t *tet10 = (priv_elem_t *) el->data;
  int i, j, k;
  W_vect_t refc[15], u[15];
  W_mat_3d_t *m = (W_mat_3d_t *) tet10->material->data;
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  int ival;
  int reqlen;
  G3D_loc_t *pts[15], iploc;
  static int buflen        = 0;
  static double *values[15] =     { NULL,NULL,NULL,NULL,NULL,
                                NULL,NULL,NULL,NULL,NULL,
                                NULL,NULL,NULL,NULL,NULL
                              }    ;
  double *hvals[8] =     { NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL }    ;
  G3D_loc_t *hpts[8];
  int hexi[4][8] =     {
        {10,5,2,6,14,12,9,13}    ,
        {4,1,5,10,11,8,12,14}    ,
        {0,4,10,6,7,11,14,13}    ,
        {7,11,14,13,3,8,12,9}    
  }    ;

#define EVAL_SF(k) \
      {                                               \
    G3D_tet10_interp_opt (tet10_shape_func.n,     \
                       tet10_shape_func.n_x,      \
                       tet10_shape_func.n_y,      \
                       tet10_shape_func.n_z,      \
                       &tet10_shape_func.detJ,    \
                       tet10_shape_func.J,        \
                       tet10_shape_func.invJ,     \
                       ALL_GP_LOCS (k),            \
                       ALL_REF_COORDS (tet10));    \
  }    
  W_mat_3d_stress_type_enum outtype;
  W_output_stress_type_enum output_stress_type
    = WOPT_OPT (el->domain->options, output_stress_type);
  int postfile_output_spec
    = WOPT_OPT (el->domain->options, postfile_output_spec_deprec);

  outtype = (output_stress_type == W_OUTPUT_STRESS_CAUCHY
             ? MAT_3D_CAUCHY
             : (output_stress_type == W_OUTPUT_STRESS_2ND_PK
                ? MAT_3D_2ND_PK
                : MAT_3D_CAUCHY)
             );
             
  if (type == HEXVUE_POSTPROC_FILE) {
    double S[3][3] =     {    {0,0,0}    ,    {0,0,0}    ,    {0,0,0}    }    ;
    double equiv_pl_def = 0, von_mises = 0;

    reqlen = 0;
    if (postfile_output_spec & W_POSTFILE_DISPL_MASK) 
      reqlen += 3;
    if (postfile_output_spec & W_POSTFILE_STRESS_MASK) 
      reqlen += 6;
    if (postfile_output_spec & W_POSTFILE_EQPLS_MASK) 
      reqlen++;
    if (postfile_output_spec & W_POSTFILE_VON_MISES_MASK) 
      reqlen++;
    
    if (buflen < reqlen) {
      FOR (i, 15) {
        values[i] = NEWP_REALLOC_OBJ_ARRAY (values[i], reqlen, double);
      }    
      buflen = reqlen;
    }    
  
    ip_ctx_data.tet10    = tet10;
    ip_ctx.time          = WD_time (el->domain);
    ip_ctx.time_step     = WD_time_step (el->domain);
    ip_ctx.ctx_data      = &ip_ctx_data;
    ip_ctx.get_kinem_var = get_kinem_var;

    clear_pkgs ();
    FOR (k, 4) {
      EVAL_SF (k);
      at_new_ip ();
      ip_ctx_data.n_x = tet10->gpdata[k].n_x;
      ip_ctx_data.n_y = tet10->gpdata[k].n_y;
      ip_ctx_data.n_z = tet10->gpdata[k].n_z;
      G3D_tet10_N (tet10_shape_func.n, ALL_GP_LOCS (k));
      LOC (tet10, iploc);
      fill_ip_pkg (ALL_GP_LOCS (k), k, &iploc,
                   tet10->gpdata[k].matst, &ip_ctx, m, outtype);
    }    

    FOR (i, 10) {
      GET_FROM_NODE (tet10->node[i], WN_get_ref_coords, refc[i]);
      GET_FROM_NODE (tet10->node[i], WN_get_displ, u[i]);
    }    
    for (i = 10; i < 15; i++) {
      G3D_tet10_N (tet10_shape_func.n,
                   g3dtet10_vertex_param_coords[i][0],
                   g3dtet10_vertex_param_coords[i][1],
                   g3dtet10_vertex_param_coords[i][2],
                   g3dtet10_vertex_param_coords[i][3]);
      LOC (tet10, refc[i]);
      u[i].x = IPOL10 (tet10_shape_func.n,
                       u[0].x, u[1].x, u[2].x, u[3].x, u[4].x, 
                       u[5].x, u[6].x, u[7].x, u[8].x, u[9].x); 
      u[i].y = IPOL10 (tet10_shape_func.n,
                       u[0].y, u[1].y, u[2].y, u[3].y, u[4].y, 
                       u[5].y, u[6].y, u[7].y, u[8].y, u[9].y); 
      u[i].z = IPOL10 (tet10_shape_func.n,
                       u[0].z, u[1].z, u[2].z, u[3].z, u[4].z, 
                       u[5].z, u[6].z, u[7].z, u[8].z, u[9].z); 
    }    

    FOR (k, 15) {
      pts[k] = &refc[k];
      ival   = 0;
      if (postfile_output_spec & W_POSTFILE_DISPL_MASK) {
        values[k][ival] = u[k].x; ival++;
        values[k][ival] = u[k].y; ival++;
        values[k][ival] = u[k].z; ival++;
      }    
      
      eval_at_extrapolated_pt (pts[k], S, &equiv_pl_def, &von_mises);
      
      if (postfile_output_spec & W_POSTFILE_STRESS_MASK) {
        values[k][ival] = S[0][0]; ival++;
        values[k][ival] = S[0][1]; ival++;
        values[k][ival] = S[0][2]; ival++;
        values[k][ival] = S[1][1]; ival++;
        values[k][ival] = S[1][2]; ival++;
        values[k][ival] = S[2][2]; ival++;
      }    
      if (postfile_output_spec & W_POSTFILE_EQPLS_MASK) {
        values[k][ival] = equiv_pl_def; ival++;
      }    
      if (postfile_output_spec & W_POSTFILE_VON_MISES_MASK) {
        values[k][ival] = von_mises; ival++;
      }    
      if (ival > reqlen) REPORT_FATAL_ERROR (el->domain->errh, "wrong");
    }    
    FOR (i,  4) {
      FOR (j, 8) {
        hvals[j] = values[hexi[i][j]];
        hpts[j]  = pts[hexi[i][j]];
      }    
      W_write_hexvue_hex (IOB, hpts, reqlen, hvals, mode);
    }    
  }     else if (type == TECPLOT_I_POSTPROC_FILE) {
    /* RAW */
  }     /* type */
#undef  EVAL_SF
}    



#define VALID(which) which##_is_valid

#define DCL_TENSOR(T) \
   static double T[3][3];                       \
   static W_bool_t VALID (T) = FALSE
 
DCL_TENSOR (F);
DCL_TENSOR (prev_F);
DCL_TENSOR (inv_F);
DCL_TENSOR (rel_F);
DCL_TENSOR (midstep_F);
DCL_TENSOR (C);
DCL_TENSOR (E);
DCL_TENSOR (D);
DCL_TENSOR (midstep_D);
DCL_TENSOR (midstep_W);
DCL_TENSOR (F_dot);
DCL_TENSOR (L);
DCL_TENSOR (R);
DCL_TENSOR (U);
DCL_TENSOR (rel_R);
DCL_TENSOR (rel_U);
DCL_TENSOR (ln_rel_U);


static void
at_new_ip (void) 
{
#define INVALIDATE(which) VALID (which) = FALSE
  
  INVALIDATE (F);
  INVALIDATE (inv_F);
  INVALIDATE (prev_F);
  INVALIDATE (rel_F);
  INVALIDATE (midstep_F);
  INVALIDATE (C);
  INVALIDATE (E);
  INVALIDATE (D);
  INVALIDATE (midstep_D);
  INVALIDATE (midstep_W);
  INVALIDATE (F_dot);
  INVALIDATE (L);
  INVALIDATE (R);
  INVALIDATE (U);
  INVALIDATE (rel_R);
  INVALIDATE (rel_U);
  INVALIDATE (ln_rel_U);
}    


static void
compute_def_grad (ip_ctx_data_t *ipctx);
static void
compute_cauchy_green (ip_ctx_data_t *ipctx);
static void
compute_green_lagrange (ip_ctx_data_t *ipctx);
static void
compute_rel_def_grad (ip_ctx_data_t *ipctx);
static void
compute_midstep_def_grad (ip_ctx_data_t *ipctx);
static void
compute_rate_of_def (ip_ctx_data_t *ipctx);
static void
compute_midstep_rate_of_def (ip_ctx_data_t *ipctx, double time_step);
static void
compute_def_grad_dot (ip_ctx_data_t *ipctx);
static void
compute_vel_grad (ip_ctx_data_t *ipctx);
static void
compute_inv_def_grad (ip_ctx_data_t *ipctx);
static void
compute_polar_decomp (ip_ctx_data_t *ipctx);
static void
compute_rel_polar_decomp (ip_ctx_data_t *ipctx);
static void
compute_ln_rel_U (ip_ctx_data_t *ipctx);



static W_bool_t 
get_kinem_var (W_mat_3d_ip_ctx_t *ctx, 
              W_mat_3d_kinem_var_type_enum  type,
              double kinvar[3][3]) 
{
  ip_ctx_data_t *ipctx = (ip_ctx_data_t *) ctx->ctx_data;

  switch (type) {
  case MAT_3D_DEF_GRAD:
    if (!VALID (F)) compute_def_grad (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, F);
    break;
  case MAT_3D_DEF_GRAD_DOT:
    if (!VALID (F)) compute_def_grad_dot (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, F_dot);
    break;
  case MAT_3D_REL_DEF_GRAD:
    if (!VALID (rel_F)) compute_rel_def_grad (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, rel_F);
    break;
  case MAT_3D_PREV_DEF_GRAD:
    if (!VALID (prev_F)) compute_rel_def_grad (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, prev_F);
    break;
  case MAT_3D_MIDSTEP_DEF_GRAD:
    if (!VALID (midstep_F)) compute_midstep_def_grad (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, midstep_F);
    break;
  case MAT_3D_CAUCHY_GREEN:
    if (!VALID (C)) compute_cauchy_green (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, C);
    break;
  case MAT_3D_RATE_OF_DEF:
    if (!VALID (D)) compute_rate_of_def (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, D);
    break;
  case MAT_3D_MIDSTEP_RATE_OF_DEF:
    if (!VALID (midstep_D)) 
      compute_midstep_rate_of_def (ipctx, ctx->time_step);
    MAT3DUTIL_COPY_3X3 (kinvar, midstep_D);
    break;
  case MAT_3D_MIDSTEP_SPIN:
    if (!VALID (midstep_W)) 
      compute_midstep_rate_of_def (ipctx, ctx->time_step);
    MAT3DUTIL_COPY_3X3 (kinvar, midstep_W);
    break;
  case MAT_3D_GREEN_LAGRANGE:
    if (!VALID (E)) compute_green_lagrange (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, E);
    break;
  case MAT_3D_POLAR_R:
    if (!VALID (R)) compute_polar_decomp (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, R);
    break;
  case MAT_3D_POLAR_U:
    if (!VALID (U)) compute_polar_decomp (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, U);
    break;
  case MAT_3D_POLAR_REL_R:
    if (!VALID (rel_R)) compute_rel_polar_decomp (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, rel_R);
    break;
  case MAT_3D_POLAR_REL_U:
    if (!VALID (rel_U)) compute_rel_polar_decomp (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, rel_U);
    break;
  case MAT_3D_LN_POLAR_REL_U:
    if (!VALID (ln_rel_U)) compute_ln_rel_U (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, ln_rel_U);
    break;
  default:
    REPORT_FATAL_ERROR (NULL, "Unknown kinematic tensor type");
    break;
  }    

  return TRUE;
}    




static void
compute_rel_def_grad (ip_ctx_data_t *ipctx) 
{
  priv_elem_t *tet10 = ipctx->tet10;
  double *n_x = ipctx->n_x;
  double *n_y = ipctx->n_y;
  double *n_z = ipctx->n_z;
  int k;
  double f00 = 0, f01 = 0, f02 = 0;
  double f10 = 0, f11 = 0, f12 = 0;
  double f20 = 0, f21 = 0, f22 = 0;
  double x, y, z;
  double inv_prev_F[3][3];
  double det_prev_F;
  register double dphix, dphiy, dphiz;
  W_node_t *n;

  if (!VALID (F)) compute_def_grad (ipctx);
   
  FOR (k, 10) {
    n = tet10->node[k];
    WN_get_prev_coords (n, &x, &y, &z);
    dphix  = n_x[k];
    dphiy  = n_y[k];
    dphiz  = n_z[k];
    f00 += dphix * x;
    f01 += dphiy * x;
    f02 += dphiz * x;
    f10 += dphix * y;
    f11 += dphiy * y;
    f12 += dphiz * y;
    f20 += dphix * z;
    f21 += dphiy * z;
    f22 += dphiz * z;
  }    

  prev_F[0][0] = f00;
  prev_F[0][1] = f01;
  prev_F[0][2] = f02;
  
  prev_F[1][0] = f10;
  prev_F[1][1] = f11;
  prev_F[1][2] = f12;
  
  prev_F[2][0] = f20;
  prev_F[2][1] = f21;
  prev_F[2][2] = f22;

  det_prev_F = MAT3DUTIL_DET_3X3 (prev_F);
  if (det_prev_F <= 0) {
    sprintf (W_buf (), "Failed to invert def grad");
    REPORT_FATAL_ERROR (NULL, W_buf ());
  }    

  MAT3DUTIL_INVERT_3X3 (det_prev_F, prev_F, inv_prev_F);

  MAT3DUTIL_MULT_AB_3X3 (rel_F, F, inv_prev_F);

  VALID (rel_F) = TRUE;
  VALID (prev_F) = TRUE;
}    




static void
compute_midstep_def_grad (ip_ctx_data_t *ipctx) 
{
  double m[3][3] =     {     {1, 0, 0}    ,     {0, 1, 0}    ,     {0, 0, 1}     }    , s[3][3], inv_s[3][3];
  
  if (!VALID (rel_F)) compute_rel_def_grad (ipctx);

  MAT3DUTIL_ADD_AB_3X3 (s, m, rel_F);
  MAT3DUTIL_SCALE_3X3 (s, 0.5);
  if (!W_mat3dutil_invert_3x3 (s, inv_s)) {
    sprintf (W_buf (), "Failed to invert");
    REPORT_FATAL_ERROR (NULL, W_buf ());
  }    
  MAT3DUTIL_MULT_AB_3X3 (midstep_F, inv_s, rel_F);

  VALID (midstep_F) = TRUE;
}    


static void
compute_vel_grad (ip_ctx_data_t *ipctx) 
{
  if (!VALID (F)) compute_def_grad (ipctx);
  if (!VALID (F_dot)) compute_def_grad_dot (ipctx);
  if (!VALID (inv_F)) compute_inv_def_grad (ipctx);

  MAT3DUTIL_MULT_AB_3X3 (L, F_dot, inv_F);
  
  VALID (L) = TRUE;
}    


static void
compute_rate_of_def (ip_ctx_data_t *ipctx) 
{
  int i, j;
  
  if (!VALID (L)) compute_vel_grad (ipctx);

  FOR (i, 3) 
    FOR (j, 3) 
      D[i][j] = 0.5 * (L[i][j] + L[j][i]);

  VALID (D) = TRUE;
}    



static void
compute_def_grad (ip_ctx_data_t *ipctx) 
{
  priv_elem_t *tet10 = ipctx->tet10;
  double *n_x = ipctx->n_x;
  double *n_y = ipctx->n_y;
  double *n_z = ipctx->n_z;
  int k;
  double f00 = 0, f01 = 0, f02 = 0;
  double f10 = 0, f11 = 0, f12 = 0;
  double f20 = 0, f21 = 0, f22 = 0;
  W_node_t *n;
  double x, y, z;
  register double dphix, dphiy, dphiz;

  /* Deformation gradient */
  FOR (k, 10) {
    n = tet10->node[k];
    WN_get_coords (n, &x, &y, &z);
    dphix  = n_x[k];
    dphiy  = n_y[k];
    dphiz  = n_z[k];
    f00 += dphix * x;
    f01 += dphiy * x;
    f02 += dphiz * x;
    f10 += dphix * y;
    f11 += dphiy * y;
    f12 += dphiz * y;
    f20 += dphix * z;
    f21 += dphiy * z;
    f22 += dphiz * z;
  }    

  F[0][0] = f00;
  F[0][1] = f01;
  F[0][2] = f02;
  
  F[1][0] = f10;
  F[1][1] = f11;
  F[1][2] = f12;
  
  F[2][0] = f20;
  F[2][1] = f21;
  F[2][2] = f22;

  VALID (F) = TRUE;
}    



static void
compute_def_grad_dot (ip_ctx_data_t *ipctx) 
{
  priv_elem_t *tet10 = ipctx->tet10;
  double *n_x = ipctx->n_x;
  double *n_y = ipctx->n_y;
  double *n_z = ipctx->n_z;
  int k;
  double f00 = 0, f01 = 0, f02 = 0;
  double f10 = 0, f11 = 0, f12 = 0;
  double f20 = 0, f21 = 0, f22 = 0;
  W_node_t *n;
  double x, y, z;
  register double dphix, dphiy, dphiz;

  /* Deformation gradient */
  FOR (k, 10) {
    n = tet10->node[k];
    x = n->velocity_t.x;
    y = n->velocity_t.y;
    z = n->velocity_t.z;
    dphix  = n_x[k];
    dphiy  = n_y[k];
    dphiz  = n_z[k];
    f00 += dphix * x;
    f01 += dphiy * x;
    f02 += dphiz * x;
    f10 += dphix * y;
    f11 += dphiy * y;
    f12 += dphiz * y;
    f20 += dphix * z;
    f21 += dphiy * z;
    f22 += dphiz * z;
  }    

  F_dot[0][0] = f00;
  F_dot[0][1] = f01;
  F_dot[0][2] = f02;
  
  F_dot[1][0] = f10;
  F_dot[1][1] = f11;
  F_dot[1][2] = f12;
  
  F_dot[2][0] = f20;
  F_dot[2][1] = f21;
  F_dot[2][2] = f22;

  VALID (F_dot) = TRUE;
}    




static void
compute_green_lagrange (ip_ctx_data_t *ipctx) 
{
  if (!VALID (C)) compute_cauchy_green (ipctx);

  /* Green-Lagrange strain tensor */
  E[0][0] = 0.5 * (C[0][0] - 1);
  E[0][1] = 0.5 * C[0][1];
  E[0][2] = 0.5 * C[0][2];
  E[1][0] = 0.5 * C[1][0];
  E[1][1] = 0.5 * (C[1][1] - 1);
  E[1][2] = 0.5 * C[1][2];
  E[2][0] = 0.5 * C[2][0];
  E[2][1] = 0.5 * C[2][1];
  E[2][2] = 0.5 * (C[2][2] - 1);
  
  VALID (E) = TRUE;
}    



static void
compute_cauchy_green (ip_ctx_data_t *ipctx) 
{
  if (!VALID (F)) compute_def_grad (ipctx);
  
  /* Cauchy-Green deformation tensor */
  MAT3DUTIL_MULT_ATB_3X3 (C, F, F);

  VALID (C) = TRUE;
}    


static void
compute_inv_def_grad (ip_ctx_data_t *ipctx) 
{
  if (!VALID (F)) compute_def_grad (ipctx);

  if (!W_mat3dutil_invert_3x3 (F, inv_F)) {
    sprintf (W_buf (), "Failed to invert def grad");
    REPORT_FATAL_ERROR (NULL, W_buf ());
  }    

  VALID (inv_F) = TRUE;
}    


static void
compute_midstep_rate_of_def (ip_ctx_data_t *ipctx, double time_step) 
{
  int i, j;
  double m[3][3] =     {     {1, 0, 0}    ,     {0, 1, 0}    ,     {0, 0, 1}     }    ;
  double M[3][3] =     {     {-1, 0, 0}    ,     {0, -1, 0}    ,     {0, 0, -1}     }    ;
  double s[3][3], inv_s[3][3], q[3][3], r[3][3];
  double inv_dt = 1 / time_step;
  
  if (!VALID (rel_F)) compute_rel_def_grad (ipctx);

  MAT3DUTIL_ADD_AB_3X3 (s, m, rel_F);
  MAT3DUTIL_SCALE_3X3 (s, 0.5);
  if (!W_mat3dutil_invert_3x3 (s, inv_s)) {
    sprintf (W_buf (), "Failed to invert def grad");
    REPORT_FATAL_ERROR (NULL, W_buf ());
  }    
  MAT3DUTIL_ADD_AB_3X3 (q, rel_F, M);
  MAT3DUTIL_MULT_AB_3X3 (r, inv_s, q);

  FOR (i, 3) 
    FOR (j, 3) {
      midstep_D[i][j] = 0.5 * (r[i][j] + r[j][i]);
      midstep_W[i][j] = 0.5 * (r[i][j] - r[j][i]);
    }    
  MAT3DUTIL_SCALE_3X3 (midstep_D, inv_dt);
  MAT3DUTIL_SCALE_3X3 (midstep_W, inv_dt);
  
  VALID (midstep_D) = TRUE;
  VALID (midstep_W) = TRUE;
}    




static void
compute_polar_decomp (ip_ctx_data_t *ipctx) 
{
  if (!VALID (F)) compute_def_grad (ipctx);
  
  if (!W_polar_decomposition (F, R, U)) {
    REPORT_FATAL_ERROR (NULL, "Failed in W_polar_decomposition()");
  }
  VALID (R) = TRUE;
  VALID (U) = TRUE;
}    


static void
compute_rel_polar_decomp (ip_ctx_data_t *ipctx) 
{
  if (!VALID (rel_F)) compute_rel_def_grad (ipctx);
  
  if (!W_polar_decomposition (rel_F, rel_R, rel_U)) {
    REPORT_FATAL_ERROR (NULL, "Failed in W_polar_decomposition()");
  }
  VALID (rel_R) = TRUE;
  VALID (rel_U) = TRUE;
}    



static void
compute_ln_rel_U (ip_ctx_data_t *ipctx) 
{
  double lneval[3];

  if (!VALID (rel_U)) compute_rel_polar_decomp (ipctx);
  
  W_ln_U (rel_U, ln_rel_U);

  VALID (ln_rel_U) = TRUE;
}    


 

typedef struct ipol_pkg_t     {
  G3D_loc_t       iploc;
  double          stress[3][3];
  double          equiv_pl_strain;
  double          von_mises;
}                  ipol_pkg_t;


static ipol_pkg_t ipol_pkgs[4];


static void
fill_ip_pkg (double r, double s, double t, double u,
             int entry,
             G3D_loc_t *iploc,
             W_mat_3d_state_p matstate,
             W_mat_3d_ip_ctx_t *ip_ctx,
             W_mat_3d_t *mat,
             W_mat_3d_stress_type_enum stress_type
             ) 
{
  double stress[3][3];
  double equiv_pl_strain, von_mises;
  W_mat_3d_stress_type_enum out_type;

  /* Get 1st Piola-Kirchhoff stress at point */
  if (!WM3D_update_stress (mat, matstate, W_MAT_STATE_UPDATE_TEMP,
                        ip_ctx, stress_type, stress,
                        &out_type)) {
    REPORT_ERROR (NULL, "Failed to compute stress");
    return;
  }    
  if (out_type != stress_type) 
    REPORT_WARNING (NULL, "output stress type mismatch");

  if (!WM3D_get_mat_state_var (mat, matstate,
                               ip_ctx, MAT_3D_EQUIV_PLASTIC_DEF,
                               &equiv_pl_strain)) 
    REPORT_FATAL_ERROR (NULL, "Failed to retrieve equiv plastic def");

      {
    double means = (stress[0][0] + stress[1][1] + stress[2][2]) / 3;
    double s00 = stress[0][0] - means;
    double s11 = stress[1][1] - means;
    double s22 = stress[2][2] - means;
    
    von_mises
      = sqrt (1.5
             * (  s00*s00
                + stress[0][1]*stress[0][1]
                + stress[0][2]*stress[0][2]
                + stress[1][0]*stress[1][0]
                + s11*s11
                + stress[1][2]*stress[1][2]
                + stress[2][0]*stress[2][0]
                + stress[2][1]*stress[2][1]
                + s22*s22
                ) 
             );
  }    
  
  /* Store into the package */
  ipol_pkgs[entry].iploc           = *iploc;
  MAT3DUTIL_COPY_3X3 (ipol_pkgs[entry].stress, stress);
  ipol_pkgs[entry].equiv_pl_strain = equiv_pl_strain;
  ipol_pkgs[entry].von_mises       = von_mises;
}    


static void
eval_at_extrapolated_pt (G3D_loc_t *pt,
                         double stress[3][3], double *eps,
                         double *vm) 
{
  double equiv_pl_strain, von_mises;
  G3D_tet_t tet;
  G3D_loc4d_t param_loc;

  tet.loc[0] = ipol_pkgs[0].iploc;
  tet.loc[1] = ipol_pkgs[1].iploc;
  tet.loc[2] = ipol_pkgs[2].iploc;
  tet.loc[3] = ipol_pkgs[3].iploc;

  if (!G3D_tet_inv_map (&tet, pt, &param_loc)) 
    REPORT_WARNING (NULL, "inv map failed");

  /* Interpolate */
#define COMP_T1(T, I, J) 				\
      { T[I][J] = (						\
	       param_loc.x * ipol_pkgs[0].T[I][J] +	\
	       param_loc.y * ipol_pkgs[1].T[I][J] +	\
	       param_loc.z * ipol_pkgs[2].T[I][J] +	\
	       param_loc.w * ipol_pkgs[3].T[I][J]       \
	       );					\
  }    
#define COMP_T(T) 						\
      {								\
    COMP_T1 (T, 0, 0); COMP_T1 (T, 0, 1); COMP_T1 (T, 0, 2);	\
    COMP_T1 (T, 1, 0); COMP_T1 (T, 1, 1); COMP_T1 (T, 1, 2);	\
    COMP_T1 (T, 2, 0); COMP_T1 (T, 2, 1); COMP_T1 (T, 2, 2);	\
  }    
#define COMP_S(S) 				\
      { S = (					\
	 param_loc.x * ipol_pkgs[0].S +	\
	 param_loc.y * ipol_pkgs[1].S +	\
	 param_loc.z * ipol_pkgs[2].S +	\
	 param_loc.w * ipol_pkgs[3].S 	\
	 );					\
  }    
#define COMP_V1(V, DIR) \
      { V.DIR = (                                   \
           param_loc.x * ipol_pkgs[0].V.DIR + \
           param_loc.y * ipol_pkgs[1].V.DIR + \
           param_loc.z * ipol_pkgs[2].V.DIR + \
           param_loc.w * ipol_pkgs[3].V.DIR   \
           );                                   \
  }    
#define COMP_V(V) { COMP_V1 (V, x); COMP_V1 (V, y); COMP_V1 (V, z); }    

  COMP_T (stress);
  COMP_S (equiv_pl_strain); *eps = equiv_pl_strain;
  COMP_S (von_mises); *vm = von_mises;

  return;
}    


static void
clear_pkgs (void) 
{
  int i;
  G3D_loc_t zero =     {0,0,0}    ;

  FOR (i, 4) {
    ipol_pkgs[i].iploc           = zero;
    MAT3DUTIL_ZERO_3X3 (ipol_pkgs[i].stress);
    ipol_pkgs[i].equiv_pl_strain = 0;
    ipol_pkgs[i].von_mises       = 0;
  }    
}    


