#ifndef HARDENING_H
#   define HARDENING_H

#include "fles.h" 

struct mat3dcryspl_elem_t;
 
typedef void (*lhs_func)(struct mat3dcryspl_elem_t *el);
 
typedef void (*rhs_func)(struct mat3dcryspl_elem_t *el);
typedef void (*update_ys_func)(struct mat3dcryspl_elem_t *el);
typedef double (*get_rho_func)(struct mat3dcryspl_elem_t *el, void *matst, int i);
typedef void (*write_hexvue_XTRA_file_func)(struct mat3dcryspl_elem_t *el, ckit_iobuf_t IOB);
 
typedef struct mat3dcryspl_elem_methods_t {
  lhs_func                      assemble_lhs;
  rhs_func                      assemble_rhs;
  update_ys_func                update_ys;
  get_rho_func			get_rho;
  write_hexvue_XTRA_file_func   write_hexvue_XTRA_file;
}   mat3dcryspl_elem_methods_t;
 
 
typedef struct mat3dcryspl_elem_t {
  W_mat_3d_t                      *m;
  mat3dcryspl_elem_methods_t      *methods;
  void				  *data; /* holds priv_elem_t info */
}   mat3dcryspl_elem_t;
 
typedef double rhos_t[N_SLIP_SYSTEMS];
typedef W_bool_t constrained_t[N_SLIP_SYSTEMS];
 
typedef struct hardening_solver_t {
  W_mat_3d_t	       *m;
  fles_solver_ctx_t    fs;
  fles_solver_ctx_t    fs_dmeschach;
  dict_dictionary      eld;
  dict_dictionary      eqnd; /* hold the nodes */
  dict_dictionary      bdry_nodes; 
  dict_dictionary      matstd;
  rhos_t               *rhost; /* pointer to an array or storage 
			         containing dislocation densities 
				 of all the slip systems at all 
				 the nodes of the domain, as 
 				 "rhost[num_nodes_of_domain][N_SLIP_SYSTEMS]",
    			        */

  rhos_t              *rhost_dt; 
  int 		       curr_sys; /* starts from 0 to (N_SLIP_SYSTEMS-1)*/ 
  /*  W_bool_t	     (*constrained)[N_SLIP_SYSTEMS];*/
  constrained_t	     *constrained;
  rhos_t	       *constraints;
}       hardening_solver_t;

typedef mat3dcryspl_elem_t *(*make_mat3dcryspl_elem_func)(W_elem_t *el, 
		                                          hardening_solver_t *hsi);
static mat3dcryspl_elem_t *make_hex20(W_elem_t *el,
                                      hardening_solver_t *hsi);
static hardening_solver_t *make_hardening_solver(W_mat_3d_t *m);

#endif
