#include "hardening.h"
#include "Wgaussp.h"   

#undef WBMAT_NNODES
#define WBMAT_NNODES 20

#define GET_FROM_NODE(node, func, tovect)         \
   func (node, &tovect.x, &tovect.y, &tovect.z)

static int msidcs[2][2][2] = {
  {{0,7}, {2,3}},
  {{4,5}, {6,1}}
};

typedef struct priv_elem_t {
  mat3dcryspl_elem_t *e;
  W_node_t           *node[WBMAT_NNODES];
  W_index_t          eqns[WBMAT_NNODES]; /* holds the eqn # of the nodes
					    in the hardening element */
  W_mat_3d_state_p   matst[8];     /* array of material states */
}              priv_elem_t;

static W_gauss_integration_rule_t *gr = NULL;
static W_gauss_integration_rule_t *gr3 = NULL; /* 3x3x3 integration rule */

static W_gauss_integration_point_t *gp = NULL;

typedef struct shpf_hard_t {
  double n[WBMAT_NNODES];
  double n_x[WBMAT_NNODES];
  double n_y[WBMAT_NNODES];
  double n_z[WBMAT_NNODES];
  double J[3][3];
  double invJ[3][3];
  double detJ;
}              shpf_hard_t;

typedef struct ip_ctx_data_t {
  W_domain_t  *domain;
  priv_elem_t *h;
  shpf_hard_t      *sf;    /* shape function table */
}              ip_ctx_data_t;

#define CHECK_2PT_CASE(H, I, J, K)                    \
    ((!use_2pt_rule)                                  \
     || (   (I == 0 && J == 0 && K == 0)              \
         || (I == 1 && J == 1 && K == 1)))
#define USE_QUADR_PT(H, I, J, K) if (CHECK_2PT_CASE(H, I, J, K))

#define ALL_GP_LOCS(i,j,k) gp[i].Loc, gp[j].Loc, gp[k].Loc
#define GP_WGT(gp,I,J,K) (gp[I].Wgt * gp[J].Wgt * gp[K].Wgt)
#define ALL_REF_COORDS(H)                                                               \
 H->node[0]->ref_coords.x, H->node[0]->ref_coords.y, H->node[0]->ref_coords.z,          \
 H->node[1]->ref_coords.x, H->node[1]->ref_coords.y, H->node[1]->ref_coords.z,          \
 H->node[2]->ref_coords.x, H->node[2]->ref_coords.y, H->node[2]->ref_coords.z,          \
 H->node[3]->ref_coords.x, H->node[3]->ref_coords.y, H->node[3]->ref_coords.z,          \
 H->node[4]->ref_coords.x, H->node[4]->ref_coords.y, H->node[4]->ref_coords.z,          \
 H->node[5]->ref_coords.x, H->node[5]->ref_coords.y, H->node[5]->ref_coords.z,          \
 H->node[6]->ref_coords.x, H->node[6]->ref_coords.y, H->node[6]->ref_coords.z,          \
 H->node[7]->ref_coords.x, H->node[7]->ref_coords.y, H->node[7]->ref_coords.z,          \
 H->node[8]->ref_coords.x, H->node[8]->ref_coords.y, H->node[8]->ref_coords.z,          \
 H->node[9]->ref_coords.x, H->node[9]->ref_coords.y, H->node[9]->ref_coords.z,          \
 H->node[10]->ref_coords.x, H->node[10]->ref_coords.y, H->node[10]->ref_coords.z,       \
 H->node[11]->ref_coords.x, H->node[11]->ref_coords.y, H->node[11]->ref_coords.z,       \
 H->node[12]->ref_coords.x, H->node[12]->ref_coords.y, H->node[12]->ref_coords.z,       \
 H->node[13]->ref_coords.x, H->node[13]->ref_coords.y, H->node[13]->ref_coords.z,       \
 H->node[14]->ref_coords.x, H->node[14]->ref_coords.y, H->node[14]->ref_coords.z,       \
 H->node[15]->ref_coords.x, H->node[15]->ref_coords.y, H->node[15]->ref_coords.z,       \
 H->node[16]->ref_coords.x, H->node[16]->ref_coords.y, H->node[16]->ref_coords.z,       \
 H->node[17]->ref_coords.x, H->node[17]->ref_coords.y, H->node[17]->ref_coords.z,       \
 H->node[18]->ref_coords.x, H->node[18]->ref_coords.y, H->node[18]->ref_coords.z,       \
 H->node[19]->ref_coords.x, H->node[19]->ref_coords.y, H->node[19]->ref_coords.z

/* evaluate the shape functions at the integration pts */
#define EVAL_SF                                 \
  {                                             \
    G3D_hex20_interp_opt(shpf_hex20.n,        \
                       shpf_hex20.n_x,      \
                       shpf_hex20.n_y,      \
                       shpf_hex20.n_z,      \
                       &shpf_hex20.detJ,    \
                       shpf_hex20.J,        \
                       shpf_hex20.invJ,     \
                       ALL_GP_LOCS(i,j,k),      \
                       ALL_REF_COORDS(h));      \
  }


static shpf_hard_t shpf_hex20;


static void
b_mat (double n_X[WBMAT_NNODES],
       double n_Y[WBMAT_NNODES],
       double n_Z[WBMAT_NNODES],
       double b[3][WBMAT_NNODES]
       )
{
  int i, j, jx, jy, jz;

  FOR (i, WBMAT_NNODES) {
    FOR (j, 3) { b[j][i] = 0; }
  }

#undef B
#define B(I,J,V) { b[I][J] = V; }
  jx = 0;
  jy = 1;
  jz = 2;
  FOR (i, WBMAT_NNODES){
    B (jx, i, n_X[i]); 
    B (jy, i, n_Y[i]);
    B (jz, i, n_Z[i]);
  }
}

/*
static void
h_mat (double n[WBMAT_NNODES],
       double sf[WBMAT_NNODES]
       )
{
  int i;

  FOR (i, WBMAT_NNODES) { sf[i] = 0; }

#undef SF
#define SF(I,V) { sf[I]= V; }
  FOR (i, WBMAT_NNODES){
    SF (i, n[i]); 
  }
#undef SF
}
*/


static void
add_btdb(double K[WBMAT_NNODES][WBMAT_NNODES],
         double b[3][WBMAT_NNODES], double D[3][3], double fact)
{
  int i, j, k, m;
  const int kdim = WBMAT_NNODES;
  double d, bkic;
  FOR (k, 3) {
    FOR (m, 3) {
      d = D[k][m];
      if (d != 0) {
        d *= fact;
        FOR (i, kdim) {
          bkic = d * b[k][i];
          FOR (j, kdim) {
            K[i][j] += bkic * b[m][j];
          } /* for j */
        } /* for i */
      } /* if */
    } /* for m */ 
  } /* for k */
}

static void
add_hht(double K[WBMAT_NNODES][WBMAT_NNODES],
        double sf[WBMAT_NNODES], double fact)
{
  int i, j;
  const int kdim = WBMAT_NNODES;
  double hkic;
  FOR (i, kdim) {
    hkic = fact * sf[i];
    FOR (j ,kdim) {
      K[i][j] += hkic * sf[j];
    }
  }
}

static void
add_st(double 	sf[WBMAT_NNODES], /* shape functions at the quadrature pts*/
       int    	sys, /* sys is the current slip system # */
       double 	dgammadt, /* shear strain on the current slip system of the integration pt*/
       priv_elem_t  *h,
       rhos_t 	*rhost, /* dislocation densities of previous time step */
       double 	J,
       double 	K,
       double 	jdetw,
       double 	s_t[WBMAT_NNODES]  
      )
{ 
  int i,j, id;
  double rhost_sum = 0;
  double rhost_qp[N_SLIP_SYSTEMS];
  double drhosdt;

  FOR (i, N_SLIP_SYSTEMS) { rhost_qp[i] = 0;}
  FOR (i, N_SLIP_SYSTEMS){ 
      FOR (j, WBMAT_NNODES) { 
          id = h->eqns[j]; /* id is the eqn# of the node, which starts from 0 */
          rhost_qp[i] += rhost[id][i] * sf[j]; 
      }
  }

  FOR (i, N_SLIP_SYSTEMS){ rhost_sum += rhost_qp[i]; }
  
  FOR (i, N_SLIP_SYSTEMS) { rhost_qp[i] = 0;}
  FOR (i, N_SLIP_SYSTEMS){ 
      FOR (j, WBMAT_NNODES) { 
          id = h->eqns[j]; /* id is the eqn# of the node, which starts from 0 */
          rhost_qp[i] += rhost[id][i] * sf[j]; 
      }
  }

  FOR (i, N_SLIP_SYSTEMS){ rhost_sum += rhost_qp[i]; }
  
  /* evolution of rhos, see the equation(25) given in L.J. Sluys & Y.Estrin's paper) */

  drhosdt = dgammadt * (J* sqrt( rhost_sum - rhost_qp[sys]) - K * rhost_qp[sys]);

  FOR (i, WBMAT_NNODES){
    s_t[i] += sf[i] * drhosdt * jdetw;
  }
}


/** 
 **  Calculate the diffusivity matrix D
 **/

static void
transform_diffusivity(mat_t *u/* mat3d */,
	              double d,
  	              double D[3][3]
	              )
{
  int i, j;
  double temp[3][3];
  double T[3][3];
  double a[3], b[3], c[3]; 
  double la, lb, lc; 
  
  a[0] = u->a.x;
  a[1] = u->a.y;
  a[2] = u->a.z;
  b[0] = u->b.x;
  b[1] = u->b.y;
  b[2] = u->b.z;
  c[0] = u->c.x;
  c[1] = u->c.y;
  c[2] = u->c.z;

  /* Normalize the vectors */
  la = sqrt(MAT3DUTIL_DOT_V(a,a));
  lb = sqrt(MAT3DUTIL_DOT_V(b,b));
  lc = sqrt(MAT3DUTIL_DOT_V(c,c));

  a[0] = a[0]/la;
  a[1] = a[1]/la;
  a[2] = a[2]/la;

  b[0] = b[0]/lb;
  b[1] = b[1]/lb;
  b[2] = b[2]/lb;

  c[0] = c[0]/lc;
  c[1] = c[1]/lc;
  c[2] = c[2]/lc;

  /* Transformation tensor from the lattice coordinate to the global coordinate */
  T[0][0] = a[0];
  T[1][0] = b[0];
  T[2][0] = c[0];
  T[0][1] = a[1];
  T[1][1] = b[1];
  T[2][1] = c[1];
  T[0][2] = a[2];
  T[1][2] = b[2];
  T[2][2] = c[2];

  MAT3DUTIL_ZERO_3X3(D);
  MAT3DUTIL_ZERO_3X3(temp);
  temp[0][0] = d;

  /* Do the transformation */
  MAT3DUTIL_MULT_ATBA_3X3(D, T, temp);
}

static double 
get_diffusivity(matstate_t *ms,
                int         curr_sys
                )
{
  double p; /* probability */
  double d; /* diffusitivity on the current quadrature pt, 
               and current slip system */
  double tau = fabs(ms->tau[curr_sys]);
  double g   = ms->g[curr_sys];
  
  double exponent;

  if (tau < g)
  {
     exponent = -360.0 * (1 - tau/g);
     p = exp(exponent);
  }

  else p = 1.0;

  d = 2 * p; /* unit mm*mm/s ? */
  return d;
}

static void
get_flux (double flux[3],
	  double b[3][WBMAT_NNODES],
          priv_elem_t  *h,
	  hardening_solver_t *hsi) 
{
  int i, j, sys;
  sys = hsi->curr_sys;
  FOR(i, 3) {
     flux[i] = 0.0;
     FOR(j, WBMAT_NNODES) {
	flux[i] += b[i][j] * hsi->rhost[h->eqns[j]][sys];
     }
  }
}

/* at integration points */
static double
get_dislocation (hardening_solver_t *hsi,
	         priv_elem_t        *h,
                 double              shpf[WBMAT_NNODES])
{
   int sys = hsi->curr_sys, n, id;
   double component, sum = 0.0;

   double rhost_qp = 0.0;
   FOR (n, WBMAT_NNODES) {
     id = h->eqns[n];  
     component = hsi->rhost[id][sys] * shpf[n];
     rhost_qp += component; 
     if (hsi->rhost[id][sys] != 0){
        sum += shpf[n];
     }
   } 
   return rhost_qp;       
}

static void
assemble_lhs (mat3dcryspl_elem_t *el)
{
  priv_elem_t *h = (priv_elem_t *)el->data;
  W_mat_3d_t *m = el->m;
  mat_t *u = (mat_t *)m->data;
  matstate_t *ms;
  int curr_sys = u->hsi->curr_sys;
  /* curr_sys should start from zero to (N_SLIP_SYSTEMS-1) */

  double jdetw;
  int i, j, k, n;

  int theta = 1; /* fully implicit */
  double K[WBMAT_NNODES][WBMAT_NNODES];
  double M[WBMAT_NNODES][WBMAT_NNODES];
  double b[3][WBMAT_NNODES];
  double D[3][3];
  W_bool_t constrained[WBMAT_NNODES];
  int nodeid[WBMAT_NNODES];
  double dt, d;
  
  double flux[3];
  double rhost_qp; 

  if (gr == NULL) {
    gr = WG_get_integration_rule(2);
  }

  if (gr3 == NULL) {
    gr3 = WG_get_integration_rule(3);
  }

  dt = WD_time_step(u->d);

  FOR (i, WBMAT_NNODES) {FOR (j, WBMAT_NNODES) { K[i][j] = 0; }}
  FOR (i, WBMAT_NNODES) {FOR (j, WBMAT_NNODES) { M[i][j] = 0; }}
  FOR (i, WBMAT_NNODES) { nodeid[i] = h->node[i]->id; }

  /** K -- stiffness matrix **/

  gp = gr->ip; /* 2x2x2 integration rule */
  FOR(i, 2) {
    FOR(j, 2) {
      FOR(k, 2) {

          EVAL_SF;
          jdetw = GP_WGT(gp,i,j,k) * shpf_hex20.detJ;

          b_mat (shpf_hex20.n_x, shpf_hex20.n_y, shpf_hex20.n_z, b);

	  d = 2.0;
	  transform_diffusivity(u, d, D);

          add_btdb(K, b, D, jdetw); /*Contribution from K*/

      } /* FOR(k, 2) */
    } /* FOR(j, 2) */
  } /* FOR(i, 2) */

  /** M -- mass matrix **/

  gp = gr3->ip; /* 3x3x3 integration rule */
  FOR(i, 3) {
    FOR(j, 3) {
      FOR(k, 3) {

          EVAL_SF;
          jdetw = GP_WGT(gp,i,j,k) * shpf_hex20.detJ;

          add_hht(M, shpf_hex20.n, jdetw/dt/theta); /*Contribution from M*/

      } /* FOR(k, 3) */
    } /* FOR(j, 3) */
  } /* FOR(i, 3) */


  /** add M to K **/

  FOR (i, WBMAT_NNODES) {
      FOR (j, WBMAT_NNODES) {
	  K[i][j] += M[i][j];
      }
  }

  /** apply the zero EBC 
      to the nodes on the boundary surfaces**/

  FOR (i, WBMAT_NNODES) {
     constrained[i] = u->hsi->constrained[h->eqns[i]][curr_sys];
  }

 assemble_lhs_callback(u->hsi, FLES_ADD, WBMAT_NNODES, h->eqns, (double *)K, constrained);
}

#define DBGPg(tit, n, what)  {                                          \
    int I_I_;                                                           \
    fprintf (stderr, "%s of qpt# %d = [ ", tit, n);                     \
    FOR (I_I_, N_SLIP_SYSTEMS) { fprintf (stderr, "%g ", what[I_I_]); } \
    fprintf (stderr, "]\n");                                            \
  }

/* PRINT SLIP SYSTEM 1,2,7,8 ONLY */
#define DBGP1278(tit,what)  {                                               \
    fprintf (stderr, "%s = [ ", tit);                                       \
    fprintf (stderr, "%g %g, %g %g ", what[0], what[1], what[6], what[7]);  \
    fprintf (stderr, "]\n");                                                \
  }

#define DBGPH(tit, n, what)  {                                          \
    int I_I_, J_J_;                                                     \
    fprintf (stderr, "%s of qpt# %d = [ ", tit, n);                     \
    FOR (I_I_, N_SLIP_SYSTEMS) { 					\
      FOR (J_J_, N_SLIP_SYSTEMS) { 					\
	fprintf (stderr, "%g ", what[I_I_][J_J_]);  			\
      }									\
       fprintf (stderr, ";\n");                         		\
    } 									\
    fprintf (stderr, "]\n");                                            \
  }

static void
apply_constraints( int nr,
                   double *values, 
                   W_bool_t *constrained)
{
  int j, k, l;
  W_bool_t need_to_rewrite = FALSE; /* do I need to modify the element matrix
                                       because of some constraints?
                                     */
 
  FOR (j, nr) { if (constrained[j]) { need_to_rewrite = TRUE; break; } }
 
  if (need_to_rewrite) {
    l = 0;
    FOR (j, nr) {
      FOR (k, nr) {
        if (j != k && (constrained[j] || constrained[k])) {
          values[l] = 0;
        }
        l++;
      }
    }
  }
}

static void
assemble_rhs(mat3dcryspl_elem_t *el)
{
  priv_elem_t *h = (priv_elem_t *)el->data; 
  W_mat_3d_t *m = el->m;
  mat_t *u = (mat_t *)m->data;
  matstate_t *ms;

  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  double jdetw;
  double K = u->K;
  double J = u->J;
  int i, j, k, id;
  int theta = 1; /* fully implicit */

  double s_t[WBMAT_NNODES];
  double f_t[WBMAT_NNODES];
  double dgammadt; /* the shear strain rate on the current slip system */
  double M[WBMAT_NNODES][WBMAT_NNODES];
  double dt, d;
  W_bool_t constrained[WBMAT_NNODES];
  int nodeid[WBMAT_NNODES];
  
  ip_ctx_data.domain   = u->d;
  ip_ctx_data.h        = h;
  ip_ctx.time          = WD_time (u->d);
  ip_ctx.time_step     = WD_time_step (u->d);
  ip_ctx_data.sf       = &shpf_hex20;

  if (gr == NULL) {
    gr = WG_get_integration_rule(2);
  }
  if (gr3 == NULL) {
    gr3 = WG_get_integration_rule(3);
  }

  dt = ip_ctx.time_step;

  FOR (i, WBMAT_NNODES) {FOR (j, WBMAT_NNODES) { M[i][j] = 0; }}
  FOR (i, WBMAT_NNODES) { s_t[i] = 0; }
  FOR (i, WBMAT_NNODES) { f_t[i] = 0; }
  FOR (i, WBMAT_NNODES) { nodeid[i] = h->node[i]->id; }

  gp = gr->ip;
  FOR(i, 2) {
    FOR(j, 2) {
      FOR(k, 2) {

          EVAL_SF;
          jdetw = GP_WGT(gp,i,j,k) * shpf_hex20.detJ;

          ms = (matstate_t *)h->matst[msidcs[i][j][k]];

          /* dgammadt = fabs( ms->dgamma[u->hsi->curr_sys] )/dt;	 */
	  dgammadt = 0.0;

	  add_st(shpf_hex20.n, u->hsi->curr_sys, dgammadt, h, u->hsi->rhost, J, K, jdetw, s_t);

      } /* FOR(k, 2) */
    } /* FOR(j, 2) */
  } /* FOR(i, 2) */

  gp = gr3->ip;
  FOR(i, 3) {
    FOR(j, 3) {
      FOR(k, 3) {

          EVAL_SF;
          jdetw = GP_WGT(gp,i,j,k) * shpf_hex20.detJ;

          add_hht(M, shpf_hex20.n, jdetw/dt/theta); /* get the M matrix */

      } /* FOR(k, 3) */
    } /* FOR(j, 3) */
  } /* FOR(i, 3) */

  /** apply the zero EBC 
      to the nodes on the boundary surfaces**/
  FOR (i, WBMAT_NNODES) {
     constrained[i] = u->hsi->constrained[h->eqns[i]][u->hsi->curr_sys];
  }

  apply_constraints(WBMAT_NNODES, (double *)M, constrained);

  FOR(i, WBMAT_NNODES){
    f_t[i] = s_t[i];
    FOR(j, WBMAT_NNODES){
      id = h->eqns[j];
      f_t[i] += M[i][j]*u->hsi->rhost[id][u->hsi->curr_sys]; /* + M[i][j]*drhosdt[j]*(1-theta)/theta;*/
      /* here theta == 1 at the moment*/
    }
  }

  assemble_rhs_callback(u->hsi,FLES_ADD, WBMAT_NNODES, h->eqns, f_t);
}

/**
 **    update the yield stress and 
 **    the reference shear-strain rate 
 **    at the quadrature pts 
 **/

static void
update_ys(mat3dcryspl_elem_t *el)
{
  int i, j, k, sys, n, id; 
  int nqp = 0;
  priv_elem_t *h = (priv_elem_t *)el->data;
  W_mat_3d_t *m = el->m;
  mat_t *u = (mat_t *)m->data;
  matstate_t *ms = NULL;

  double rhost_qp[N_SLIP_SYSTEMS];

  double factor = (u->const_a) * (u->G) * (u->mb);
  double const_k = u->const_k;
  double J = u->J;
  double K = u->K;

  double rhos_sum = 0.0;
  double component;
  double g, under_sqrt;
 
  if (gr == NULL) {
    gr = WG_get_integration_rule(2);
  }

   /* go over the integration points*/
   gp = gr->ip;
   FOR(i, 2) {
      FOR(j, 2) {
         FOR(k, 2) {
            EVAL_SF;

	    /* the following two loops calculate:
	      (1) the dislocation densities
	          of all slip systems at the quadrature pts  
	      (2) the sum of dislocation densities 
	          of all slip systems at each quadrature pt
             */

	    rhos_sum = 0.0;	
	    FOR(sys, N_SLIP_SYSTEMS) { 	

               rhost_qp[sys] = 0.0;

	       /* get the dislocation density at the quadrature pts */	
               FOR (n, WBMAT_NNODES) {
                  id = h->eqns[n];  
		  component = u->hsi->rhost_dt[id][sys] * shpf_hex20.n[n];
                  rhost_qp[sys] += component; 
               } 

               rhos_sum += rhost_qp[sys];

            } /* FOR(sys, N_SLIP_SYSTEMS) */


	    /* set the yield stress, reference shear-strain rate */

	    ms = (matstate_t *)h->matst[msidcs[i][j][k]];

	    FOR(sys, N_SLIP_SYSTEMS) { 
              under_sqrt = rhos_sum; /* - rhost_qp[sys]; 	*/
	       g = factor * sqrt(under_sqrt);
               ms->g[sys] = g; 
               /*	       ms->gamma_0[sys] = const_k * rhost_qp[sys];*/
            }
	   

	/* calculate the hardening moduli */
            /* FOR( sysi, N_SLIP_SYSTEMS ) {
	         FOR( sysj, N_SLIP_SYSTEMS ) {
		    if( sysi == sysj ) h1[sysi][sysj] = 0; 
		    else{
			h1[sysi][sysj] = 0.5 * factor * ( J * sqrt( rhos_sum - rhost_qp[sysj] ) - K * rhost_qp[sysj] );
			h1[sysi][sysj] /= sqrt( rhos_sum - rhost_qp[sysi] );
		    }
		}
               }*/
	 
         nqp++;

         } /* FOR(k, 2) */
      } /* FOR(j, 2) */
   } /* FOR(i, 2) */

   /* done */
}


/**
 ** get dislocation density on the intergration pts which holds *matst 
 **/

static double  
get_rho(mat3dcryspl_elem_t *e,
        void               *matst,
        int                sys)
{
   mat_t *u = (mat_t *)e->m->data;
   priv_elem_t *h = (priv_elem_t *)e->data;
   matstate_t *ms = NULL;
 
   double rhost_qp = 0;
   int id;
 
   int i, j, k, n;
 
   if (gr == NULL) {
     gr = WG_get_integration_rule(2);
   }
 
   /* go over the integration points*/
   gp = gr->ip;
   FOR(i, 2) {
      FOR(j, 2) {
         FOR(k, 2) {
            ms = (matstate_t *)h->matst[msidcs[i][j][k]];
            if( ms == (matstate_t *)matst )
            {
               EVAL_SF;
               /* get the dislocation density at the quadrature pts */
               FOR (n, WBMAT_NNODES) {
                  id = h->eqns[n];
                  rhost_qp += u->hsi->rhost_dt[id][sys] * shpf_hex20.n[n];
               }
            }
 
        } /* for i */
      } /* for j */
   } /* for k */
 
  return rhost_qp;
 
}


static void
write_hexvue_XTRA_file (mat3dcryspl_elem_t *e, ckit_iobuf_t IOB)
{
  priv_elem_t *h = (priv_elem_t *)e->data;
  mat_t       *u = (mat_t *)(e->m->data);
  hardening_solver_t *hsi = u->hsi;

  int         i, j, k, id;
  G3D_loc_t   *refcptr[8];
  W_vect_t    refc[8];
  double      *values[8] = {NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL};
  double      rhos[8][N_SLIP_SYSTEMS];

  /*fill the rhos array with hsi->rhost_dt at 8 corners */
  FOR (i, 8) {
      id = h->eqns[i];
      FOR (j, N_SLIP_SYSTEMS) { rhos[i][j] = hsi->rhost_dt[id][j]; }

      values[i] = rhos[i];
  }
    
    FOR (i, 8) {
      GET_FROM_NODE (h->node[i], WN_get_ref_coords, refc[i]);
    }
   
   FOR (i, 8) { refcptr[i] = &refc[i]; } 

/*
    FOR (ival, reqlen) {
      if        (tokens_token_and_keyword_equiv (os->specs[ival], "displ_x")) {
        FOR (j, WBMAT_NNODES) { epol_pkg[j].vals[ival] = u[j].x; }
      } else if (tokens_token_and_keyword_equiv (os->specs[ival], "displ_y")) { 
        FOR (j, WBMAT_NNODES) { epol_pkg[j].vals[ival] = u[j].y; }
      } else if (tokens_token_and_keyword_equiv (os->specs[ival], "displ_z")) { 
        FOR (j, WBMAT_NNODES) { epol_pkg[j].vals[ival] = u[j].z; }
      } else if (tokens_token_and_keyword_equiv (os->specs[ival], "vel_x")) {
        FOR (j, WBMAT_NNODES) { epol_pkg[j].vals[ival] = v[j].x; }
      } else if (tokens_token_and_keyword_equiv (os->specs[ival], "vel_y")) { 
        FOR (j, WBMAT_NNODES) { epol_pkg[j].vals[ival] = v[j].y; }
      } else if (tokens_token_and_keyword_equiv (os->specs[ival], "vel_z")) { 
        FOR (j, WBMAT_NNODES) { epol_pkg[j].vals[ival] = v[j].z; }
      }
    }
*/    

  W_write_hexvue_hex (IOB, refcptr, N_SLIP_SYSTEMS, values, "ascii");
}

static mat3dcryspl_elem_methods_t hex20_methods = {
  assemble_lhs,
  assemble_rhs,
  update_ys,
  get_rho,
  write_hexvue_XTRA_file 
};


/**
 ** make hardening element 
 ** like hex20_reader
 **/

static mat3dcryspl_elem_t *
make_hex20(W_elem_t *el,
           hardening_solver_t *hsi)
{ 
  mat3dcryspl_elem_t *e; 
  priv_elem_t *pe;
  int i;
  int num_nodes = WE_get_num_nodes(el);
  int num_matsts = WE_get_num_matstates(el);

  pe = w_new(priv_elem_t);
  CHECK_MALLOC(pe);
  e = w_new(mat3dcryspl_elem_t);
  CHECK_MALLOC(e);

  e->m = hsi->m;
  e->methods = &hex20_methods;
  e->data = pe;
  pe->e = e;

  FOR(i, num_nodes){
    pe->node[i] = WE_get_node(el,i);
    pe->eqns[i] = find_eqn(hsi, pe->node[i]);
  }

  FOR(i, num_matsts){
    pe->matst[i] = WE_get_matstate(el,i);
    add_matst(hsi, e, pe->matst[i]);
  }
  return e;
 
}

#undef ALL_GP_LOCS
#undef GP_WGT
#undef ALL_REF_COORDS
#undef N_LOC
#undef LOC
#undef B
#undef SF
#undef WBMAT_NNODES 
#undef MAXNODES
#undef ELEM_LOOP_START
#undef ELEM_LOOP_END
#undef DICT_LOOP_START
#undef DICT_LOOP_END
#undef EVAL_SF
#undef GET_FROM_NODE
