#include "Wconfig.h"
#include "Wtypes.h"
#include "Wdomain.h"
#include "Woptions.h"
#include "Wnode.h"
#include "Wmats.h"
#include "Wxf.h"
#include "Wutil.h"
#include "Whistry.h"
#include "Wcdi.h"
#include "g3d.h"
#include "Wmat3d.h"
#include "Wnewmarki.h"
#include "Wcdi.h"
#include "Wcontface.h"
#include "Whex20aq.h"
#include "Wgaussp.h"

/* RCSHEADER "$Id: hex20aq.c,v 1.5 2002/03/21 01:32:35 cvs Exp $" */

#undef WBMAT_NNODES
#define WBMAT_NNODES 20

#include "bmat.c"
#include "elstiff.c"
 
#define GET_FROM_NODE(node, func, tovect)         \
   func (node, &tovect.x, &tovect.y, &tovect.z)

#define FOR_ALL_ELEM_IN_BLOCK elem_t *e = NULL; while ((e = (elem_t *) get_list_next (eb->elist, e)) != 0) 

typedef struct integration_point_t {
  double  Loc[3];               /* location */
  double  Wgt;                  /* weight */
}              integration_point_t;

typedef struct elem_t {
  W_node_t          *node[WBMAT_NNODES];
  W_mat_3d_state_p   matst[9];
  double             mass;
  int                serial_num;
# ifdef X_XFACE
  GraphicObj        *grep;
# endif
}              elem_t;

/* history client data */
typedef struct history_client_data_t {
  W_elem_t          *el; /* FE */
  W_index_t          pntidx; /* unique integer  */
  elem_t            *h;
  int                which_qp;  /* which quadrature point */
}              history_client_data_t;

typedef struct elem_block_t {
  W_elem_t               *el;
  LIST_OF (elem_t *)      elist;
  W_material_t           *material;
  double                  block_mass;
  double                  W0;
  double                  a;
  double                  Wa;
  integration_point_t     ip[9];
}              elem_block_t;

static void
make_integration_points (elem_block_t *eb)
{
  /* This assumes the 9-point integration rule for quadratic hex solids
     from E. L. Wilson's book, Three dimensional static and dynamic analysis
     of structures, Computers and Structures, Inc., 2nd Edition, 1998,
     p. 5-15.
     The ninth integration point is always in the center of the parametric
     coordinates.
     If W0 = 0, the rule reduces to 2x2x2 gaussian integration,
     for W0 = 16/3, the eight integration points are at the corners, and
     Wa = 1/3.
   */
  eb->Wa = 1.0 - eb->W0 / 8;
  eb->a  = sqrt (1.0 / 3.0 / eb->Wa);
#define INITIP(which, A0, A1, A2, W)            \
  {                                             \
    eb->ip[which].Loc[0] = A0;                  \
    eb->ip[which].Loc[1] = A1;                  \
    eb->ip[which].Loc[2] = A2;                  \
    eb->ip[which].Wgt    = W;                   \
  }
  INITIP (0, -eb->a, -eb->a, -eb->a, eb->Wa);
  INITIP (1,  eb->a, -eb->a, -eb->a, eb->Wa);
  INITIP (2,  eb->a,  eb->a, -eb->a, eb->Wa);
  INITIP (3, -eb->a,  eb->a, -eb->a, eb->Wa);
  INITIP (4, -eb->a, -eb->a,  eb->a, eb->Wa);
  INITIP (5,  eb->a, -eb->a,  eb->a, eb->Wa);
  INITIP (6,  eb->a,  eb->a,  eb->a, eb->Wa);
  INITIP (7, -eb->a,  eb->a,  eb->a, eb->Wa);
  INITIP (8,      0,      0,      0, eb->W0);
}

static int 
num_of_faces (W_elem_t *el)
{
  elem_block_t *eb = (elem_block_t *)el->data;
  return 6 * list_length (eb->elist);
}

static elem_t *
find_in_elist (elem_block_t *eb, int serial_num)
{
  FOR_ALL_ELEM_IN_BLOCK {
    if (e->serial_num == serial_num) return e;
  }
  return NULL;
}

static int 
get_face_vertex_nodes (W_elem_t *el, int nface, int ndim, W_node_t *n[]) 
{
  elem_block_t *eb = (elem_block_t *)el->data;
  elem_t *h = NULL;
  int serial_num = nface / 6;
  int nf = nface % 6;
  int i;
  h = find_in_elist (eb, serial_num);
  if (h) {
    if (ndim >= 8) {
      FOR (i, 8) {
        n[i] = h->node[g3dhex20_vertex_on_face[nf][i]];
      } 
    }
    return 8;
  } else
    return 0;
} 


typedef struct shpf_t {
  double n[WBMAT_NNODES];
  double n_x[WBMAT_NNODES];
  double n_y[WBMAT_NNODES];
  double n_z[WBMAT_NNODES];
  double J[3][3];
  double invJ[3][3];
  double detJ;
}              shpf_t;

typedef struct ip_ctx_data_t {
  W_domain_t  *domain;
  elem_t      *h;
  shpf_t      *sf;    /* shape function table */
}              ip_ctx_data_t;

static void
at_new_ip(void);
static W_bool_t 
get_kinem_var(W_mat_3d_ip_ctx_t *ctx, 
              W_mat_3d_kinem_var_type_enum  type,
              double kinvar[3][3]);


#define ALL_GP_LOCS(i) ip[i].Loc[0], ip[i].Loc[1], ip[i].Loc[2]
#define GP_WGT(I) (ip[I].Wgt)
#define ALL_REF_COORDS(H)                                                               \
 H->node[0]->ref_coords.x, H->node[0]->ref_coords.y, H->node[0]->ref_coords.z,          \
 H->node[1]->ref_coords.x, H->node[1]->ref_coords.y, H->node[1]->ref_coords.z,          \
 H->node[2]->ref_coords.x, H->node[2]->ref_coords.y, H->node[2]->ref_coords.z,          \
 H->node[3]->ref_coords.x, H->node[3]->ref_coords.y, H->node[3]->ref_coords.z,          \
 H->node[4]->ref_coords.x, H->node[4]->ref_coords.y, H->node[4]->ref_coords.z,          \
 H->node[5]->ref_coords.x, H->node[5]->ref_coords.y, H->node[5]->ref_coords.z,          \
 H->node[6]->ref_coords.x, H->node[6]->ref_coords.y, H->node[6]->ref_coords.z,          \
 H->node[7]->ref_coords.x, H->node[7]->ref_coords.y, H->node[7]->ref_coords.z,          \
 H->node[8]->ref_coords.x, H->node[8]->ref_coords.y, H->node[8]->ref_coords.z,          \
 H->node[9]->ref_coords.x, H->node[9]->ref_coords.y, H->node[9]->ref_coords.z,          \
 H->node[10]->ref_coords.x, H->node[10]->ref_coords.y, H->node[10]->ref_coords.z,       \
 H->node[11]->ref_coords.x, H->node[11]->ref_coords.y, H->node[11]->ref_coords.z,       \
 H->node[12]->ref_coords.x, H->node[12]->ref_coords.y, H->node[12]->ref_coords.z,       \
 H->node[13]->ref_coords.x, H->node[13]->ref_coords.y, H->node[13]->ref_coords.z,       \
 H->node[14]->ref_coords.x, H->node[14]->ref_coords.y, H->node[14]->ref_coords.z,       \
 H->node[15]->ref_coords.x, H->node[15]->ref_coords.y, H->node[15]->ref_coords.z,       \
 H->node[16]->ref_coords.x, H->node[16]->ref_coords.y, H->node[16]->ref_coords.z,       \
 H->node[17]->ref_coords.x, H->node[17]->ref_coords.y, H->node[17]->ref_coords.z,       \
 H->node[18]->ref_coords.x, H->node[18]->ref_coords.y, H->node[18]->ref_coords.z,       \
 H->node[19]->ref_coords.x, H->node[19]->ref_coords.y, H->node[19]->ref_coords.z

static shpf_t shpf;
  
#define N_LOC(h, v, DIR, n)                          \
   v.DIR = (  n[0] * h->node[0]->ref_coords.DIR      \
            + n[1] * h->node[1]->ref_coords.DIR      \
            + n[2] * h->node[2]->ref_coords.DIR      \
            + n[3] * h->node[3]->ref_coords.DIR      \
            + n[4] * h->node[4]->ref_coords.DIR      \
            + n[5] * h->node[5]->ref_coords.DIR      \
            + n[6] * h->node[6]->ref_coords.DIR      \
            + n[7] * h->node[7]->ref_coords.DIR      \
            + n[8] * h->node[8]->ref_coords.DIR      \
            + n[9] * h->node[9]->ref_coords.DIR      \
            + n[10] * h->node[10]->ref_coords.DIR      \
            + n[11] * h->node[11]->ref_coords.DIR      \
            + n[12] * h->node[12]->ref_coords.DIR      \
            + n[13] * h->node[13]->ref_coords.DIR      \
            + n[14] * h->node[14]->ref_coords.DIR      \
            + n[15] * h->node[15]->ref_coords.DIR      \
            + n[16] * h->node[16]->ref_coords.DIR      \
            + n[17] * h->node[17]->ref_coords.DIR      \
            + n[18] * h->node[18]->ref_coords.DIR      \
            + n[19] * h->node[19]->ref_coords.DIR      \
              )

#define LOC(h, gp)                              \
  {                                             \
    N_LOC(h, gp, x, shpf.n);          \
    N_LOC(h, gp, y, shpf.n);          \
    N_LOC(h, gp, z, shpf.n);          \
  }


static int
elem_elcode (W_elem_t *el)
{
  return HEX20AQ_ELCODE;
}


static char *
type_name(W_elem_t *el)
{
  static char buf[] = "hex20aq";
  return buf;
}

/* Returns nodal forces by which the  element acts on its nodal points. */
static void
el_assemble_restoring_forces (elem_t *h, elem_block_t *eb, 
                              W_mat_state_update_form_enum update_form)
{
  W_mat_3d_t *m = (W_mat_3d_t *)eb->material->data;
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  double P[3][3];
  W_mat_3d_stress_type_enum out_type;
  W_vect_t f;
  double jdetw, dde, de;
  double dphix, dphiy, dphiz;
  integration_point_t *ip = eb->ip;
  int i, n;
#define EVAL_SF(i)                                              \
    G3D_hex20_interp_opt (shpf.n, shpf.n_x, shpf.n_y, shpf.n_z, \
                          &shpf.detJ, shpf.J, shpf.invJ,        \
                          ALL_GP_LOCS(i), ALL_REF_COORDS(h))
#define COMP_F()                                            \
  {                                                         \
    f.x = -(P[0][0]*dphix + P[0][1]*dphiy + P[0][2]*dphiz); \
    f.y = -(P[1][0]*dphix + P[1][1]*dphiy + P[1][2]*dphiz); \
    f.z = -(P[2][0]*dphix + P[2][1]*dphiy + P[2][2]*dphiz); \
  }
  
  ip_ctx_data.domain   = eb->el->domain;
  ip_ctx_data.h        = h;
  ip_ctx.time          = WD_time (eb->el->domain);
  ip_ctx.time_step     = WD_time_step (eb->el->domain);
  ip_ctx.ctx_data      = &ip_ctx_data;
  ip_ctx.get_kinem_var = get_kinem_var;
  ip_ctx_data.sf       = &shpf;

  de = 0;

  FOR (i, 9) {
    EVAL_SF(i);
    at_new_ip();
    if (!WM3D_update_stress (m, h->matst[i], update_form,
                             &ip_ctx, MAT_3D_1ST_PK, P, &out_type))
      REPORT_ERROR (eb->el->domain->errh, "Failed to compute stress");
    if (!WM3D_get_mat_state_var (m, h->matst[i],
                                 &ip_ctx, MAT_3D_DEFORM_ENERGY_DENSITY, &dde))
      REPORT_ERROR (eb->el->domain->errh, "Failed to get work of deformation");
    jdetw = GP_WGT(i) * shpf.detJ;
    de += dde * jdetw;
    FOR (n, WBMAT_NNODES) {
      dphix = jdetw * shpf.n_x[n];
      dphiy = jdetw * shpf.n_y[n];
      dphiz = jdetw * shpf.n_z[n];
      COMP_F();
      WN_assemble_force(h->node[n], &f);
    }
  }

  eb->el->deformation_energy += de;
#undef  EVAL_SF
}

static void
assemble_restoring_forces (W_elem_t *el,
                           W_mat_state_update_form_enum update_form)
{
  elem_block_t *eb = (elem_block_t *)el->data;
  eb->el->deformation_energy = 0;
  {
    FOR_ALL_ELEM_IN_BLOCK {
      el_assemble_restoring_forces (e, eb, update_form);
    }
  }
}

static void
el_assemble_inertial_forces_consistent (elem_t *h, elem_block_t *eb)
{
  W_mat_3d_t *m = (W_mat_3d_t *)eb->material->data;
  W_vect_t f[WBMAT_NNODES], a[WBMAT_NNODES];
  double Nmat[WBMAT_NNODES];
  double jdetw, J[3][3], rho;
  int i, n, l;
  integration_point_t *ip = eb->ip;

  FOR (i, WBMAT_NNODES) {
    GET_FROM_NODE (h->node[i], WN_get_acceleration, a[i]);
    VECMAC_SCALAR (f[i], =, 0);
  }

  rho = WM3D_get_volume_mass (m);

  FOR (i, 9) {
    G3D_hex20_N (Nmat, ALL_GP_LOCS(i));
    G3D_hex20_J (J, ALL_GP_LOCS(i), ALL_REF_COORDS(h));
    jdetw = -rho * GP_WGT(i) * G3D_hex20_detJ (J);
    FOR(n, WBMAT_NNODES) {
      FOR (l, WBMAT_NNODES) {
        VECMAC_2_VEC (f[n], +=, jdetw * Nmat[n] * Nmat[l] * a[l]);
      }
    }
  } 

  FOR (i, WBMAT_NNODES) {
    WN_assemble_force (h->node[i], &f[i]);
  }
#undef  EVAL_SF
}

static void
assemble_inertial_forces_consistent (W_elem_t *el)
{
  elem_block_t *eb = (elem_block_t *)el->data;
  FOR_ALL_ELEM_IN_BLOCK {
    el_assemble_inertial_forces_consistent (e, eb);
  }
}

static W_float_t
el_compute_total_mass (elem_t *h, elem_block_t *eb)
{
  W_mat_3d_t *m = (W_mat_3d_t *)eb->material->data;
  W_float_t volume_mass, J[3][3];
  int i;
  integration_point_t *ip = eb->ip;
  double volume = 0;

  volume_mass = WM3D_get_volume_mass(m);
  
  FOR (i, 9) {
    G3D_hex20_J (J, ALL_GP_LOCS(i), ALL_REF_COORDS(h));
    volume += GP_WGT(i) * G3D_hex20_detJ (J);
  }

  return (h->mass = (volume_mass * volume));
}

static void
compute_total_mass (elem_block_t *eb)
{
  double tm = 0;
  {
    FOR_ALL_ELEM_IN_BLOCK {
      tm += el_compute_total_mass (e, eb);
    }
  }
  eb->block_mass = tm;
}

static W_float_t
total_mass (W_elem_t *el)
{
  elem_block_t *eb = (elem_block_t *)el->data;
  return eb->block_mass;
}

static void
el_assemble_inertial_forces_diag (elem_t *h, elem_block_t *eb)
{
  W_mat_3d_t *m = (W_mat_3d_t *)eb->material->data;
  W_vect_t a;
  double Nmat[WBMAT_NNODES];
  double jdetw, J[3][3], rho;
  int i, n;
  integration_point_t *ip = eb->ip;
  double diagmass[WBMAT_NNODES], diagmasssum = 0, mr, ms;

  rho = WM3D_get_volume_mass(m);

  FOR (i, WBMAT_NNODES) {diagmass[i] = 0;}

  FOR (i, 9) {
    G3D_hex20_N (Nmat, ALL_GP_LOCS(i));
    G3D_hex20_J (J, ALL_GP_LOCS(i), ALL_REF_COORDS(h));
    jdetw = rho * GP_WGT(i) * G3D_hex20_detJ (J);
    FOR(n, WBMAT_NNODES) {
      diagmass[n] += jdetw * Nmat[n] * Nmat[n];
    }
  }

  FOR (i, WBMAT_NNODES) { diagmasssum += diagmass[i]; }
  mr = (h->mass / diagmasssum);
  FOR (i, WBMAT_NNODES) {
    GET_FROM_NODE (h->node[i], WN_get_acceleration, a);
    ms = diagmass[i] * mr;
    VECMAC_SCALAR (a, *=, -ms);
    WN_assemble_force (h->node[i], &a);
  }
}

static void
assemble_inertial_forces_diag (W_elem_t *el)
{
  elem_block_t *eb = (elem_block_t *)el->data;
  FOR_ALL_ELEM_IN_BLOCK {
    el_assemble_inertial_forces_diag (e, eb);
  }
}

static void
el_assemble_mass (elem_t *h, elem_block_t *eb)
{
  W_mat_3d_t *m = (W_mat_3d_t *)eb->material->data;
  double Nmat[WBMAT_NNODES];
  double jdetw, J[3][3], rho;
  int i, n;
  integration_point_t *ip = eb->ip;
  double diagmass[WBMAT_NNODES], diagmasssum = 0;

  FOR (i, WBMAT_NNODES) { diagmass[i] = 0; }

  rho = WM3D_get_volume_mass(m);

  FOR (i, 9) {
    G3D_hex20_N (Nmat, ALL_GP_LOCS(i));
    G3D_hex20_J (J, ALL_GP_LOCS(i), ALL_REF_COORDS(h));
    jdetw = rho * GP_WGT(i) * G3D_hex20_detJ (J);
    FOR(n, WBMAT_NNODES) {
      diagmass[n] += jdetw * Nmat[n] * Nmat[n];
    }
  } 

  FOR (i, WBMAT_NNODES) { diagmasssum += diagmass[i]; }
  FOR (i, WBMAT_NNODES) {
    WN_assemble_mass(h->node[i], diagmass[i] * (h->mass / diagmasssum));
  }
}

static void
assemble_mass (W_elem_t *el)
{
  elem_block_t *eb = (elem_block_t *)el->data;
  FOR_ALL_ELEM_IN_BLOCK {
    el_assemble_mass (e, eb);
  }
}

static W_float_t 
el_suggested_time_step (elem_t *h, elem_block_t *eb)
{
  W_mat_3d_t *m = (W_mat_3d_t *)eb->material->data;
  W_coords_3d_t d;
  W_float_t E, nu;
  W_float_t rho, speed_of_sound;
  double shortest_side = FLT_MAX;

  if (!WM3D_get_mat_const(m, MAT_3D_YOUNG_MODULUS, &E))
    REPORT_FATAL_ERROR(eb->el->domain->errh, "No Young's modulus");
  if (!WM3D_get_mat_const(m, MAT_3D_POISSON, &nu))
    REPORT_FATAL_ERROR(eb->el->domain->errh, "No Poisson constant");
  
  rho      = WM3D_get_volume_mass(m);
  speed_of_sound = sqrt(E / (1 - 2*nu) / rho);
#define SUBTR(res, a, b) \
   {res.x = (a).x - (b).x; res.y = (a).y - (b).y; res.z = (a).z - (b).z;}
#define COND_MIN(a, b) \
   { double the_B = b; if (the_B != 0) a = min(a, the_B); }
  SUBTR(d, h->node[1]->ref_coords, h->node[0]->ref_coords);
  COND_MIN(shortest_side, W_vector_3d_length(&d));
  SUBTR(d, h->node[2]->ref_coords, h->node[1]->ref_coords);
  COND_MIN(shortest_side, W_vector_3d_length(&d));
  SUBTR(d, h->node[3]->ref_coords, h->node[2]->ref_coords);
  COND_MIN(shortest_side, W_vector_3d_length(&d));
  SUBTR(d, h->node[0]->ref_coords, h->node[3]->ref_coords);
  COND_MIN(shortest_side, W_vector_3d_length(&d));
  SUBTR(d, h->node[5]->ref_coords, h->node[4]->ref_coords);
  COND_MIN(shortest_side, W_vector_3d_length(&d));
  SUBTR(d, h->node[6]->ref_coords, h->node[5]->ref_coords);
  COND_MIN(shortest_side, W_vector_3d_length(&d));
  SUBTR(d, h->node[7]->ref_coords, h->node[6]->ref_coords);
  COND_MIN(shortest_side, W_vector_3d_length(&d));
  SUBTR(d, h->node[4]->ref_coords, h->node[7]->ref_coords);
  COND_MIN(shortest_side, W_vector_3d_length(&d));
  SUBTR(d, h->node[0]->ref_coords, h->node[4]->ref_coords);
  COND_MIN(shortest_side, W_vector_3d_length(&d));
  SUBTR(d, h->node[1]->ref_coords, h->node[5]->ref_coords);
  COND_MIN(shortest_side, W_vector_3d_length(&d));
  SUBTR(d, h->node[2]->ref_coords, h->node[6]->ref_coords);
  COND_MIN(shortest_side, W_vector_3d_length(&d));
  SUBTR(d, h->node[3]->ref_coords, h->node[7]->ref_coords);
  COND_MIN(shortest_side, W_vector_3d_length(&d));
  
  return shortest_side / speed_of_sound / 2;                 
}

static W_float_t 
suggested_time_step (W_elem_t *el)
{
  double sdt = FLT_MAX;
  elem_block_t *eb = (elem_block_t *)el->data;
  FOR_ALL_ELEM_IN_BLOCK {
    sdt = min (sdt, el_suggested_time_step (e, eb));
  }
  return sdt;
}

#ifdef X_XFACE
static void
el_modify_elem_grep (elem_t *h, elem_block_t *eb,
                     double displ_scale, W_bool_t immed_draw)
{
  int i;
  W_vect_t refc, u;
  WCRec pnts[WBMAT_NNODES];

  for (i = 0; i < 8; i++) {     /* note: we're using only the corner nodes here */
    GET_FROM_NODE (h->node[i], WN_get_ref_coords, refc);
    GET_FROM_NODE (h->node[i], WN_get_displ, u);
    pnts[i].x = refc.x + displ_scale * u.x;
    pnts[i].y = refc.y + displ_scale * u.y;
    pnts[i].z = refc.z + displ_scale * u.z;
  }

  if (h->grep == NULL) {
    h->grep = CreateHexahedron(pnts);
    EASValsSetColor(WM_get_color(eb->material));
    EASValsSetLayer (WX_ELEM_LAYER);
    EGWithMaskChangeAttributes(COLOR_MASK | LAYER_MASK, h->grep);
    EMNoDrawAddGraphicsToModel(ESIModel (), h->grep);
  } else {
    EGModifyGraphicsGeometry(h->grep, pnts);
  }
}

static void
modify_elem_grep(W_elem_t *el, double displ_scale, W_bool_t immed_draw)
{
  elem_block_t *eb = (elem_block_t *)el->data;
  FOR_ALL_ELEM_IN_BLOCK {
    el_modify_elem_grep (e, eb, displ_scale, immed_draw);
  }
}
#endif

#define VALID(which) which##_is_valid

#define DCL_TENSOR(T)                           \
   static double T[3][3];                       \
   static W_bool_t VALID(T) = FALSE
 
DCL_TENSOR(F);
DCL_TENSOR(prev_F);
DCL_TENSOR(inv_F);
DCL_TENSOR(rel_F);
DCL_TENSOR(midstep_F);
DCL_TENSOR(C);
DCL_TENSOR(E);
DCL_TENSOR(D);
DCL_TENSOR(midstep_D);
DCL_TENSOR(midstep_W);
DCL_TENSOR(F_dot);
DCL_TENSOR(L);
DCL_TENSOR(R);
DCL_TENSOR(U);
DCL_TENSOR(rel_R);
DCL_TENSOR(rel_U);
DCL_TENSOR(ln_rel_U);


static void
at_new_ip(void)
{
#define INVALIDATE(which) VALID(which) = FALSE
  
  INVALIDATE(F);
  INVALIDATE(inv_F);
  INVALIDATE(prev_F);
  INVALIDATE(rel_F);
  INVALIDATE(midstep_F);
  INVALIDATE(C);
  INVALIDATE(E);
  INVALIDATE(D);
  INVALIDATE(midstep_D);
  INVALIDATE(midstep_W);
  INVALIDATE(F_dot);
  INVALIDATE(L);
  INVALIDATE(R);
  INVALIDATE(U);
  INVALIDATE(rel_R);
  INVALIDATE(rel_U);
  INVALIDATE(ln_rel_U);
}


static void
compute_def_grad(ip_ctx_data_t *ipctx);
static void
compute_cauchy_green(ip_ctx_data_t *ipctx);
static void
compute_green_lagrange(ip_ctx_data_t *ipctx);
static void
compute_rel_def_grad(ip_ctx_data_t *ipctx);
static void
compute_midstep_def_grad(ip_ctx_data_t *ipctx);
static void
compute_rate_of_def(ip_ctx_data_t *ipctx);
static void
compute_midstep_rate_of_def(ip_ctx_data_t *ipctx, double time_step);
static void
compute_def_grad_dot(ip_ctx_data_t *ipctx);
static void
compute_vel_grad(ip_ctx_data_t *ipctx);
static void
compute_inv_def_grad(ip_ctx_data_t *ipctx);
static void
compute_polar_decomp(ip_ctx_data_t *ipctx);
static void
compute_rel_polar_decomp(ip_ctx_data_t *ipctx);
static void
compute_ln_rel_U(ip_ctx_data_t *ipctx);



static W_bool_t 
get_kinem_var(W_mat_3d_ip_ctx_t *ctx, 
              W_mat_3d_kinem_var_type_enum  type,
              double kinvar[3][3])
{
  ip_ctx_data_t *ipctx = (ip_ctx_data_t *)ctx->ctx_data;

  switch (type) {
  case MAT_3D_DEF_GRAD:
    if (!VALID(F)) compute_def_grad(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, F);
    break;
  case MAT_3D_DEF_GRAD_DOT:
    if (!VALID(F)) compute_def_grad_dot(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, F_dot);
    break;
  case MAT_3D_REL_DEF_GRAD:
    if (!VALID(rel_F)) compute_rel_def_grad(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, rel_F);
    break;
  case MAT_3D_PREV_DEF_GRAD:
    if (!VALID(prev_F)) compute_rel_def_grad(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, prev_F);
    break;
  case MAT_3D_MIDSTEP_DEF_GRAD:
    if (!VALID(midstep_F)) compute_midstep_def_grad(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, midstep_F);
    break;
  case MAT_3D_CAUCHY_GREEN:
    if (!VALID(C)) compute_cauchy_green(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, C);
    break;
  case MAT_3D_RATE_OF_DEF:
    if (!VALID(D)) compute_rate_of_def(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, D);
    break;
  case MAT_3D_MIDSTEP_RATE_OF_DEF:
    if (!VALID(midstep_D))
      compute_midstep_rate_of_def(ipctx, ctx->time_step);
    MAT3DUTIL_COPY_3X3(kinvar, midstep_D);
    break;
  case MAT_3D_MIDSTEP_SPIN:
    if (!VALID(midstep_W))
      compute_midstep_rate_of_def(ipctx, ctx->time_step);
    MAT3DUTIL_COPY_3X3(kinvar, midstep_W);
    break;
  case MAT_3D_GREEN_LAGRANGE:
    if (!VALID(E)) compute_green_lagrange(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, E);
    break;
  case MAT_3D_POLAR_R:
    if (!VALID(R)) compute_polar_decomp(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, R);
    break;
  case MAT_3D_POLAR_U:
    if (!VALID(U)) compute_polar_decomp(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, U);
    break;
  case MAT_3D_POLAR_REL_R:
    if (!VALID(rel_R)) compute_rel_polar_decomp(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, rel_R);
    break;
  case MAT_3D_POLAR_REL_U:
    if (!VALID(rel_U)) compute_rel_polar_decomp(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, rel_U);
    break;
  case MAT_3D_LN_POLAR_REL_U:
    if (!VALID(ln_rel_U)) compute_ln_rel_U(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, ln_rel_U);
    break;
  default:
    {
      static W_bool_t complained = FALSE;
      if (!complained) {
        sprintf (W_buf(), "Unknown kinematic tensor type: %d",     (int)type);
        REPORT_WARNING (ipctx->domain->errh, W_buf());
        complained = TRUE;
      }
    }
    return FALSE;
  }

  return TRUE;
}




static void
compute_rel_def_grad (ip_ctx_data_t *ipctx)
{
  elem_t *h = ipctx->h;
  shpf_t *sftab = ipctx->sf;
  int k;
  double f00 = 0, f01 = 0, f02 = 0;
  double f10 = 0, f11 = 0, f12 = 0;
  double f20 = 0, f21 = 0, f22 = 0;
  double x, y, z;
  double inv_prev_F[3][3];
  double det_prev_F;
  register double dphix, dphiy, dphiz;
  W_node_t *n;

  if (!VALID(F)) compute_def_grad(ipctx);
   
  FOR(k, WBMAT_NNODES) {
    n = h->node[k];
    WN_get_prev_coords (n, &x, &y, &z);
    dphix  = sftab->n_x[k];
    dphiy  = sftab->n_y[k];
    dphiz  = sftab->n_z[k];
    f00 += dphix * x;
    f01 += dphiy * x;
    f02 += dphiz * x;
    f10 += dphix * y;
    f11 += dphiy * y;
    f12 += dphiz * y;
    f20 += dphix * z;
    f21 += dphiy * z;
    f22 += dphiz * z;
  }

  prev_F[0][0] = f00;
  prev_F[0][1] = f01;
  prev_F[0][2] = f02;
  
  prev_F[1][0] = f10;
  prev_F[1][1] = f11;
  prev_F[1][2] = f12;
  
  prev_F[2][0] = f20;
  prev_F[2][1] = f21;
  prev_F[2][2] = f22;

  det_prev_F = MAT3DUTIL_DET_3X3(prev_F);
  if (det_prev_F <= 0) {
    sprintf(W_buf(), "Failed to invert def grad");
    REPORT_FATAL_ERROR(NULL, W_buf());
  }

  MAT3DUTIL_INVERT_3X3(det_prev_F, prev_F, inv_prev_F);

  MAT3DUTIL_MULT_AB_3X3(rel_F, F, inv_prev_F);

  VALID(rel_F) = TRUE;
  VALID(prev_F) = TRUE;
}




static void
compute_midstep_def_grad(ip_ctx_data_t *ipctx)
{
  if (!VALID(F)) compute_def_grad(ipctx);
  if (!VALID(prev_F)) compute_rel_def_grad (ipctx);
   
  MAT3DUTIL_COMB_AB_3X3 (midstep_F, 0.5, F, 0.5, prev_F);

  VALID(midstep_F) = TRUE;
}


static void
compute_vel_grad(ip_ctx_data_t *ipctx)
{
  if (!VALID(F)) compute_def_grad(ipctx);
  if (!VALID(F_dot)) compute_def_grad_dot(ipctx);
  if (!VALID(inv_F)) compute_inv_def_grad(ipctx);

  MAT3DUTIL_MULT_AB_3X3(L, F_dot, inv_F);
  
  VALID(L) = TRUE;
}


static void
compute_rate_of_def(ip_ctx_data_t *ipctx)
{
  int i, j;
  
  if (!VALID(L)) compute_vel_grad(ipctx);

  FOR(i, 3)
    FOR(j, 3) 
      D[i][j] = 0.5 * (L[i][j] + L[j][i]);

  VALID(D) = TRUE;
}




static void
compute_def_grad(ip_ctx_data_t *ipctx)
{
  elem_t *h = ipctx->h;
  shpf_t *sftab = ipctx->sf;
  int k;
  double f00 = 0, f01 = 0, f02 = 0;
  double f10 = 0, f11 = 0, f12 = 0;
  double f20 = 0, f21 = 0, f22 = 0;
  W_node_t *n;
  double x, y, z;
  register double dphix, dphiy, dphiz;

  /* Deformation gradient */
  FOR(k, WBMAT_NNODES) {
    n = h->node[k];
    WN_get_displ (n, &x, &y, &z);
    dphix  = sftab->n_x[k];
    dphiy  = sftab->n_y[k];
    dphiz  = sftab->n_z[k];
    f00 += dphix * x;
    f01 += dphiy * x;
    f02 += dphiz * x;
    f10 += dphix * y;
    f11 += dphiy * y;
    f12 += dphiz * y;
    f20 += dphix * z;
    f21 += dphiy * z;
    f22 += dphiz * z;
  }

  F[0][0] = f00 + 1;
  F[0][1] = f01;
  F[0][2] = f02;
  
  F[1][0] = f10;
  F[1][1] = f11 + 1;
  F[1][2] = f12;
  
  F[2][0] = f20;
  F[2][1] = f21;
  F[2][2] = f22 + 1;

  VALID(F) = TRUE;
}



static void
compute_def_grad_dot(ip_ctx_data_t *ipctx)
{
  elem_t *h = ipctx->h;
  shpf_t *sftab = ipctx->sf;
  int k;
  double f00 = 0, f01 = 0, f02 = 0;
  double f10 = 0, f11 = 0, f12 = 0;
  double f20 = 0, f21 = 0, f22 = 0;
  W_node_t *n;
  double x, y, z;
  register double dphix, dphiy, dphiz;

  /* Deformation gradient */
  FOR(k, WBMAT_NNODES) {
    n = h->node[k];
    x = n->velocity_t.x;
    y = n->velocity_t.y;
    z = n->velocity_t.z;
    dphix  = sftab->n_x[k];
    dphiy  = sftab->n_y[k];
    dphiz  = sftab->n_z[k];
    f00 += dphix * x;
    f01 += dphiy * x;
    f02 += dphiz * x;
    f10 += dphix * y;
    f11 += dphiy * y;
    f12 += dphiz * y;
    f20 += dphix * z;
    f21 += dphiy * z;
    f22 += dphiz * z;
  }

  F_dot[0][0] = f00;
  F_dot[0][1] = f01;
  F_dot[0][2] = f02;
  
  F_dot[1][0] = f10;
  F_dot[1][1] = f11;
  F_dot[1][2] = f12;
  
  F_dot[2][0] = f20;
  F_dot[2][1] = f21;
  F_dot[2][2] = f22;

  VALID(F_dot) = TRUE;
}




static void
compute_green_lagrange(ip_ctx_data_t *ipctx)
{
  if (!VALID(C)) compute_cauchy_green(ipctx);

  /* Green-Lagrange strain tensor */
  E[0][0] = 0.5 * (C[0][0] - 1);
  E[0][1] = 0.5 * C[0][1];
  E[0][2] = 0.5 * C[0][2];
  E[1][0] = 0.5 * C[1][0];
  E[1][1] = 0.5 * (C[1][1] - 1);
  E[1][2] = 0.5 * C[1][2];
  E[2][0] = 0.5 * C[2][0];
  E[2][1] = 0.5 * C[2][1];
  E[2][2] = 0.5 * (C[2][2] - 1);
  
  VALID(E) = TRUE;
}



static void
compute_cauchy_green(ip_ctx_data_t *ipctx)
{
  if (!VALID(F)) compute_def_grad(ipctx);
  
  /* Cauchy-Green deformation tensor */
  MAT3DUTIL_MULT_ATB_3X3(C, F, F);

  VALID(C) = TRUE;
}


static void
compute_inv_def_grad(ip_ctx_data_t *ipctx)
{
  if (!VALID(F)) compute_def_grad(ipctx);

  if (!W_mat3dutil_invert_3x3(F, inv_F)) {
    sprintf(W_buf(), "Failed to invert def grad");
    REPORT_FATAL_ERROR(NULL, W_buf());
  }

  VALID(inv_F) = TRUE;
}


static void
compute_midstep_rate_of_def (ip_ctx_data_t *ipctx, double time_step)
{
  int i, j;
  double m[3][3] = { {1, 0, 0}, {0, 1, 0}, {0, 0, 1} };
  double M[3][3] = { {-1, 0, 0}, {0, -1, 0}, {0, 0, -1} };
  double s[3][3], inv_s[3][3], q[3][3], r[3][3];
  double inv_dt = 1 / time_step;
  
  if (!VALID(rel_F)) compute_rel_def_grad(ipctx);

  MAT3DUTIL_ADD_AB_3X3(s, m, rel_F);
  MAT3DUTIL_SCALE_3X3(s, 0.5);
  if (!W_mat3dutil_invert_3x3(s, inv_s)) {
    sprintf(W_buf(), "Failed to invert def grad");
    REPORT_FATAL_ERROR(NULL, W_buf());
  }
  MAT3DUTIL_ADD_AB_3X3(q, rel_F, M);
  MAT3DUTIL_MULT_AB_3X3(r, inv_s, q);

  FOR(i, 3)
    FOR(j, 3) {
      midstep_D[i][j] = 0.5 * (r[i][j] + r[j][i]);
      midstep_W[i][j] = 0.5 * (r[i][j] - r[j][i]);
    }
  MAT3DUTIL_SCALE_3X3(midstep_D, inv_dt);
  MAT3DUTIL_SCALE_3X3(midstep_W, inv_dt);
  
  VALID(midstep_D) = TRUE;
  VALID(midstep_W) = TRUE;
}




static void
compute_polar_decomp(ip_ctx_data_t *ipctx)
{
  if (!VALID(F)) compute_def_grad(ipctx);
   
  if (!W_polar_decomposition (F, R, U)) {
    REPORT_FATAL_ERROR (NULL, "Failed in W_polar_decomposition()");
  }
  VALID(R) = TRUE;
  VALID(U) = TRUE;
}


static void
compute_rel_polar_decomp(ip_ctx_data_t *ipctx)
{
  if (!VALID(rel_F)) compute_rel_def_grad(ipctx);

  if (!W_polar_decomposition (rel_F, rel_R, rel_U)) {
    REPORT_FATAL_ERROR (NULL, "Failed in W_polar_decomposition()");
  }
  VALID(rel_R) = TRUE;
  VALID(rel_U) = TRUE;
}



static void
compute_ln_rel_U(ip_ctx_data_t *ipctx)
{
  if (!VALID(rel_U)) compute_rel_polar_decomp(ipctx);
  
  W_ln_U (rel_U, ln_rel_U);
  
  VALID(ln_rel_U) = TRUE;
}

typedef struct ipol_pkg_t {
  double          stress[3][3];
  double          equiv_pl_strain;
  double          von_mises;
}              ipol_pkg_t;


static ipol_pkg_t ipol_pkgs[8];


static void
fill_ip_pkg (double xi, double eta, double theta,
             int entry,
             W_mat_3d_state_p matstate,
             W_mat_3d_ip_ctx_t *ip_ctx,
             W_mat_3d_t *mat,
             W_mat_3d_stress_type_enum stress_type
             )
{
  double stress[3][3];
  double equiv_pl_strain, von_mises;
  W_mat_3d_stress_type_enum out_type;

  /* Get 1st Piola-Kirchhoff stress at point */
  if (!WM3D_update_stress (mat, matstate, W_MAT_STATE_UPDATE_TEMP,
                        ip_ctx, stress_type, stress,
                        &out_type)) {
    REPORT_ERROR (NULL, "Failed to compute stress");
    return;
  }
  if (out_type != stress_type)
    REPORT_WARNING (NULL, "output stress type mismatch");

  if (!WM3D_get_mat_state_var (mat, matstate,
                               ip_ctx, MAT_3D_EQUIV_PLASTIC_DEF,
                               &equiv_pl_strain))
    REPORT_FATAL_ERROR(NULL, "Failed to retrieve equiv plastic def");

  {
    double means = (stress[0][0] + stress[1][1] + stress[2][2]) / 3;
    double s00 = stress[0][0] - means;
    double s11 = stress[1][1] - means;
    double s22 = stress[2][2] - means;
    
    von_mises
      = sqrt(1.5
             * (  s00*s00
                + stress[0][1]*stress[0][1]
                + stress[0][2]*stress[0][2]
                + stress[1][0]*stress[1][0]
                + s11*s11
                + stress[1][2]*stress[1][2]
                + stress[2][0]*stress[2][0]
                + stress[2][1]*stress[2][1]
                + s22*s22
                )
             );
  }
  
  /* Store into the package */
  MAT3DUTIL_COPY_3X3(ipol_pkgs[entry].stress, stress);
  ipol_pkgs[entry].equiv_pl_strain = equiv_pl_strain;
  ipol_pkgs[entry].von_mises       = von_mises;
}


static void
eval_at_extrapolated_pt (double a,
                         double xi, double eta, double theta,
                         double stress[3][3], double *eps,
                         double *vm)
{
  double inv_a = 1.0 / a;
  double modxi    = inv_a * xi;
  double modeta   = inv_a * eta;
  double modtheta = inv_a * theta;
  double Nmat[8];
  double equiv_pl_strain, von_mises;

  /* Interpolate */
  G3D_hex_N (Nmat, modxi, modeta, modtheta);

#define COMP_T1(T, I, J)				\
  { T[I][J] = (						\
	       Nmat[0] * ipol_pkgs[0].T[I][J] +	\
	       Nmat[1] * ipol_pkgs[1].T[I][J] +	\
	       Nmat[2] * ipol_pkgs[2].T[I][J] +	\
	       Nmat[3] * ipol_pkgs[3].T[I][J] +	\
	       Nmat[4] * ipol_pkgs[4].T[I][J] +	\
	       Nmat[5] * ipol_pkgs[5].T[I][J] +	\
	       Nmat[6] * ipol_pkgs[6].T[I][J] +	\
	       Nmat[7] * ipol_pkgs[7].T[I][J]	\
	       );					\
  }
#define COMP_T(T)						\
  {								\
    COMP_T1(T, 0, 0); COMP_T1(T, 0, 1); COMP_T1(T, 0, 2);	\
    COMP_T1(T, 1, 0); COMP_T1(T, 1, 1); COMP_T1(T, 1, 2);	\
    COMP_T1(T, 2, 0); COMP_T1(T, 2, 1); COMP_T1(T, 2, 2);	\
  }
#define COMP_S(S)				\
  { S = (					\
	 Nmat[0] * ipol_pkgs[0].S +	\
	 Nmat[1] * ipol_pkgs[1].S +	\
	 Nmat[2] * ipol_pkgs[2].S +	\
	 Nmat[3] * ipol_pkgs[3].S +	\
	 Nmat[4] * ipol_pkgs[4].S +	\
	 Nmat[5] * ipol_pkgs[5].S +	\
	 Nmat[6] * ipol_pkgs[6].S +	\
	 Nmat[7] * ipol_pkgs[7].S		\
	 );					\
  }
#define COMP_V1(V, DIR)                         \
  { V.DIR = (                                   \
           Nmat[0] * ipol_pkgs[0].V.DIR + \
           Nmat[1] * ipol_pkgs[1].V.DIR + \
           Nmat[2] * ipol_pkgs[2].V.DIR + \
           Nmat[3] * ipol_pkgs[3].V.DIR + \
           Nmat[4] * ipol_pkgs[4].V.DIR + \
           Nmat[5] * ipol_pkgs[5].V.DIR + \
           Nmat[6] * ipol_pkgs[6].V.DIR + \
           Nmat[7] * ipol_pkgs[7].V.DIR   \
           );                                   \
  }
#define COMP_V(V) { COMP_V1(V, x); COMP_V1(V, y); COMP_V1(V, z); }

  COMP_T(stress);
  COMP_S(equiv_pl_strain); *eps = equiv_pl_strain;
  COMP_S(von_mises); *vm = von_mises;

  return;
}

static void
clear_ip_pkgs (void)
{
  int i;

  FOR (i, 8) {
    MAT3DUTIL_ZERO_3X3 (ipol_pkgs[i].stress);
    ipol_pkgs[i].equiv_pl_strain = 0;
    ipol_pkgs[i].von_mises       = 0;
  }
}

static void
el_write_hexvue_XTRA_file (elem_t *h, elem_block_t *eb,
                           ckit_iobuf_t IOB,
                           int type, char *mode, void *optdata)
{
  int i, j, k;
  W_vect_t refc[WBMAT_NNODES+1], u[WBMAT_NNODES+1];
  W_mat_3d_t *m = (W_mat_3d_t *)eb->material->data;
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  integration_point_t *ip = eb->ip;
  int pt, ival;
  int reqlen;
  G3D_loc_t *pts[WBMAT_NNODES];
  static int buflen        = 0;
  static double *values[WBMAT_NNODES] = {NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL};
  W_postfile_output_specs_t *os = &eb->el->domain->options->postfile_output_specs_list;
  W_mat_3d_stress_type_enum outtype;
  W_mat_3d_stress_type_enum actual_outtype;
  W_output_stress_type_enum output_stress_type
    = WOPT_OPT (eb->el->domain->options, output_stress_type);
  typedef struct my_epol_2x2x2_pkg_t {
    double                    *vals;
  }              my_epol_2x2x2_pkg_t;
  /* for the 20 nodes plus the center */
  my_epol_2x2x2_pkg_t epol_pkg[WBMAT_NNODES+1]
    ={{0},{0},{0},{0},{0},{0},{0},{0},{0},{0},{0},{0},{0},{0},{0},{0},{0},{0},{0},{0},{0}};
  W_gauss_integration_point_t *gp = NULL;
    static int msidcs[2][2][2] = {
      {{0,1}, {2,3}},
      {{4,5}, {6,7}}
    };
#define EVAL_SF(i)                                              \
  G3D_hex20_interp_opt (shpf.n, shpf.n_x, shpf.n_y, shpf.n_z,   \
                        &shpf.detJ, shpf.J, shpf.invJ,          \
                        ALL_GP_LOCS(i), ALL_REF_COORDS(h))

  outtype = (output_stress_type == W_OUTPUT_STRESS_CAUCHY
             ? MAT_3D_CAUCHY
             : (output_stress_type == W_OUTPUT_STRESS_2ND_PK
                ? MAT_3D_2ND_PK
                : MAT_3D_CAUCHY)
             );
             
  {
    double S[3][3] = {{0,0,0},{0,0,0},{0,0,0}};
    double equiv_pl_def = 0, von_mises = 0;
    W_gauss_integration_rule_t *gr = WG_get_integration_rule (2);
#define SQRT_3 1.73205080756887729352
    
    reqlen = os->nspecs;
    
    if (buflen < reqlen) {
      FOR (i, WBMAT_NNODES) {
        values[i] = NEWP_REALLOC_OBJ_ARRAY (values[i], reqlen, double);
      }
      buflen = reqlen;
    }
    
    FOR (i, WBMAT_NNODES+1) {
      epol_pkg[i].vals = w_new_array (reqlen, double);
      FOR (ival, reqlen) { epol_pkg[i].vals[ival] = 0; }
    }
    
    ip_ctx_data.domain   = eb->el->domain;    
    ip_ctx_data.h        = h;
    ip_ctx.time          = WD_time (eb->el->domain);
    ip_ctx.time_step     = WD_time_step (eb->el->domain);
    ip_ctx.ctx_data      = &ip_ctx_data;
    ip_ctx.get_kinem_var = get_kinem_var;
    ip_ctx_data.sf       = &shpf;
    
    FOR (i, WBMAT_NNODES) {
      GET_FROM_NODE (h->node[i], WN_get_ref_coords, refc[i]);
      GET_FROM_NODE (h->node[i], WN_get_displ, u[i]);
    }
    G3D_hex20_interp_opt (shpf.n, shpf.n_x, shpf.n_y, shpf.n_z,   
                          &shpf.detJ, shpf.J, shpf.invJ,
                          ALL_GP_LOCS(8), ALL_REF_COORDS(h));
    VECMAC_SCALAR (refc[WBMAT_NNODES], =, 0);
    VECMAC_SCALAR (u[WBMAT_NNODES], =, 0);
    FOR (i, WBMAT_NNODES) {
      VECMAC_2_VEC (refc[WBMAT_NNODES], +=, shpf.n[i] * refc[i]);
      VECMAC_2_VEC (u[WBMAT_NNODES], +=, shpf.n[i] * u[i]);
    }
    
    FOR (ival, reqlen) {
      if        (tokens_token_and_keyword_equiv (os->specs[ival], "displ_x")) {
        FOR (j, WBMAT_NNODES+1) { epol_pkg[j].vals[ival] = u[j].x; }
      } else if (tokens_token_and_keyword_equiv (os->specs[ival], "displ_y")) { 
        FOR (j, WBMAT_NNODES+1) { epol_pkg[j].vals[ival] = u[j].y; }
      } else if (tokens_token_and_keyword_equiv (os->specs[ival], "displ_z")) { 
        FOR (j, WBMAT_NNODES+1) { epol_pkg[j].vals[ival] = u[j].z; }
      }
    }
    
    gp = gr->ip;
    FOR(i, 8) { /* the ninth integration point is ignored here */
      EVAL_SF(i);
      at_new_ip();
      if (!WM3D_update_stress (m, h->matst[i], W_MAT_STATE_UPDATE_TEMP,
                            &ip_ctx, outtype, S,
                            &actual_outtype)) {
        REPORT_ERROR (NULL, "Failed to compute stress");
        return;
      }
      if (actual_outtype != outtype)
        REPORT_WARNING (NULL, "output stress type mismatch");
      
      if (!WM3D_get_mat_state_var (m, h->matst[i],
                                   &ip_ctx, MAT_3D_EQUIV_PLASTIC_DEF,
                                   &equiv_pl_def))
        REPORT_FATAL_ERROR(NULL, "Failed to retrieve equiv plastic def");
      
      { /* von Mises */
        double means = (S[0][0] + S[1][1] + S[2][2]) / 3;
        double s00 = S[0][0] - means;
        double s11 = S[1][1] - means;
        double s22 = S[2][2] - means;
        
        von_mises
          = sqrt(1.5 * (s00*s00 + S[0][1]*S[0][1] + S[0][2]*S[0][2] + S[1][0]*S[1][0] + s11*s11
                        + S[1][2]*S[1][2] + S[2][0]*S[2][0] + S[2][1]*S[2][1] + s22*s22));
      }
      FOR (ival, reqlen) {
        int L;
        double v = 0;
        if        (tokens_token_and_keyword_equiv (os->specs[ival], "s_xx")) { 
          v = S[0][0];
        } else if (tokens_token_and_keyword_equiv (os->specs[ival], "s_xy")) { 
          v = S[0][1];
        } else if (tokens_token_and_keyword_equiv (os->specs[ival], "s_xz")) { 
          v = S[0][2];
        } else if (tokens_token_and_keyword_equiv (os->specs[ival], "s_yy")) { 
          v = S[1][1];
        } else if (tokens_token_and_keyword_equiv (os->specs[ival], "s_yz")) { 
          v = S[1][2];
        } else if (tokens_token_and_keyword_equiv (os->specs[ival], "s_zz")) { 
          v = S[2][2];
        } else if (tokens_token_and_keyword_equiv (os->specs[ival], "vonm")) { 
          v = von_mises;
        } else if (tokens_token_and_keyword_equiv (os->specs[ival], "eqpls")) { 
          v = equiv_pl_def;
        } else { /* try to get the named variable from the material state */
          double val = 0;
          if (WM3D_get_mat_state_named_var (m, h->matst[i], 
                                            &ip_ctx, os->specs[ival],
                                            &val)) {
            v = val;
          } else
            v = 0;
        }
        FOR (L, 8) {
          double modxi    = g3dhex_vertex_param_coord[L][0] / eb->a;
          double modeta   = g3dhex_vertex_param_coord[L][1] / eb->a;
          double modtheta = g3dhex_vertex_param_coord[L][2] / eb->a;
          double Nmat[WBMAT_NNODES];
          G3D_hex_N (Nmat, modxi, modeta, modtheta);
          epol_pkg[L].vals[ival] += Nmat[i] * v;
        }
      }
    } /* FOR (i,9) */
    
    FOR(i, 2) {
      FOR(j, 2) {
        FOR(k, 2) {
          pt = msidcs[i][j][k];
          pts[pt]   = &refc[pt];
          FOR (ival, reqlen) {
            values[pt][ival] = epol_pkg[pt].vals[ival]; 
          }
        }
      }
    }
    W_write_hexvue_hex (IOB, pts, reqlen, values, mode);
    
    FOR (i, WBMAT_NNODES+1) {
      w_free (epol_pkg[i].vals);
    }
#undef EVAL_SF
  }
}

static void
el_write_hexvue_file (elem_t *h, elem_block_t *eb,
                        ckit_iobuf_t IOB,
                        int type, char *mode, void *optdata)
{
  int i, j, k;
  W_vect_t refc[WBMAT_NNODES], u[WBMAT_NNODES];
  W_mat_3d_t *m = (W_mat_3d_t *)eb->material->data;
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  integration_point_t *ip = eb->ip;
  int pt, ival;
  int reqlen;
  G3D_loc_t *pts[WBMAT_NNODES];
  static int buflen        = 0;
  static double *values[WBMAT_NNODES] = {NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL};
  W_mat_3d_stress_type_enum outtype;
  W_output_stress_type_enum output_stress_type
    = WOPT_OPT (eb->el->domain->options, output_stress_type);
  int output_spec
    = WOPT_OPT (eb->el->domain->options, postfile_output_spec_deprec);

#define EVAL_SF(i)                                              \
  G3D_hex20_interp_opt (shpf.n, shpf.n_x, shpf.n_y, shpf.n_z,   \
                        &shpf.detJ, shpf.J, shpf.invJ,          \
                        ALL_GP_LOCS(i), ALL_REF_COORDS(h))

  outtype = (output_stress_type == W_OUTPUT_STRESS_CAUCHY
             ? MAT_3D_CAUCHY
             : (output_stress_type == W_OUTPUT_STRESS_2ND_PK
                ? MAT_3D_2ND_PK
                : MAT_3D_CAUCHY)
             );
             

  {
    double S[3][3] = {{0,0,0},{0,0,0},{0,0,0}};
    double equiv_pl_def = 0, von_mises = 0;

    reqlen = 0;
    if (output_spec & W_POSTFILE_DISPL_MASK) 
      reqlen += 3;
    if (output_spec & W_POSTFILE_STRESS_MASK)
      reqlen += 6;
    if (output_spec & W_POSTFILE_EQPLS_MASK)
      reqlen++;
    if (output_spec & W_POSTFILE_VON_MISES_MASK)
      reqlen++;
    if (output_spec & W_POSTFILE_MATERIAL_MASK)
      reqlen++;
    
    if (buflen < reqlen) {
      FOR (i, WBMAT_NNODES) {
        values[i] = NEWP_REALLOC_OBJ_ARRAY (values[i], reqlen, double);
      }
      buflen = reqlen;
    }
  
    ip_ctx_data.domain   = eb->el->domain;
    ip_ctx_data.h        = h;
    ip_ctx.time          = WD_time (eb->el->domain);
    ip_ctx.time_step     = WD_time_step (eb->el->domain);
    ip_ctx.ctx_data      = &ip_ctx_data;
    ip_ctx.get_kinem_var = get_kinem_var;
    ip_ctx_data.sf       = &shpf;

    clear_ip_pkgs ();
    FOR (i, 8) {
      EVAL_SF(i);
      at_new_ip();
      fill_ip_pkg (ALL_GP_LOCS(i), i,
                   h->matst[i], &ip_ctx, m, outtype);
    }

    FOR (i, WBMAT_NNODES) {
      GET_FROM_NODE (h->node[i], WN_get_ref_coords, refc[i]);
      GET_FROM_NODE (h->node[i], WN_get_displ, u[i]);
    }

    pt = 0;
    FOR (i, 2) {
      FOR (j, 2) {
        FOR (k, 2) {
          pts[pt]   = &refc[pt];
          ival = 0;
          if (output_spec & W_POSTFILE_DISPL_MASK) {
            values[pt][ival] = u[pt].x; ival++;
            values[pt][ival] = u[pt].y; ival++;
            values[pt][ival] = u[pt].z; ival++;
          }

          eval_at_extrapolated_pt (eb->a,
                                   g3dhex_vertex_param_coord[pt][0],
                                   g3dhex_vertex_param_coord[pt][1],
                                   g3dhex_vertex_param_coord[pt][2],
                                   S, &equiv_pl_def, &von_mises);
          
          if (output_spec & W_POSTFILE_STRESS_MASK) {
            values[pt][ival] = S[0][0]; ival++;
            values[pt][ival] = S[0][1]; ival++;
            values[pt][ival] = S[0][2]; ival++;
            values[pt][ival] = S[1][1]; ival++;
            values[pt][ival] = S[1][2]; ival++;
            values[pt][ival] = S[2][2]; ival++;
          }
          if (output_spec & W_POSTFILE_EQPLS_MASK) {
            values[pt][ival] = equiv_pl_def; ival++;
          }
          if (output_spec & W_POSTFILE_VON_MISES_MASK) {
            values[pt][ival] = von_mises; ival++;
          }
          if (output_spec & W_POSTFILE_MATERIAL_MASK) {
            values[pt][ival] = WM_get_id (eb->material); ival++;
          }
          if (ival > reqlen) REPORT_FATAL_ERROR (eb->el->domain->errh, "wrong");
          pt++;
        }
      }
    }
    W_write_hexvue_hex (IOB, pts, reqlen, values, mode);
  } /* type */
#undef  EVAL_SF
}

static void
el_write_postproc_file (elem_t *e, elem_block_t *eb,
                        ckit_iobuf_t IOB,
                        int type, char *mode, void *optdata)
{
  if (type == HEXVUE_POSTPROC_FILE) {
    el_write_hexvue_file (e, eb, IOB, type, mode, optdata);
  } else if (type == HEXVUE_XTRA_POSTPROC_FILE) {
    el_write_hexvue_XTRA_file (e, eb, IOB, type, mode, optdata);
  }
}

static void
write_postproc_file (W_elem_t *el, ckit_iobuf_t IOB,
                     int type, char *mode, void *optdata)
{
  elem_block_t *eb = (elem_block_t *)el->data;
  FOR_ALL_ELEM_IN_BLOCK {
    el_write_postproc_file (e, eb, IOB, type, mode, optdata);
  }
}

static void
el_assemble_eff_stiffness (elem_t *h, elem_block_t *eb,
                           W_newmark_assemble_eff_stiff_proc assemble_eff_stiff_callback,
                           W_float_t stiff_mult,
                           W_float_t mass_mult,
                           W_mass_type_enum mass_type
                        )
{
  W_mat_3d_t *m = (W_mat_3d_t *)eb->material->data;
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  double S[3][3];
  W_mat_3d_stress_type_enum out_type;
  W_vect_t u[WBMAT_NNODES];
  double jdetw;
  int i, j, k, n;
  double rho;
  double C[6][6];
  double K[WBMAT_NNODES*3][WBMAT_NNODES*3];
  W_index_t idcs[WBMAT_NNODES*3];
  W_bool_t constrained[WBMAT_NNODES*3];
  double b_l[6][WBMAT_NNODES*3];
  double b_nl[9][WBMAT_NNODES*3];
  integration_point_t *ip = eb->ip;
#define EVAL_SF(i)                                              \
    G3D_hex20_interp_opt(shpf.n, shpf.n_x, shpf.n_y, shpf.n_z,  \
                         &shpf.detJ, shpf.J, shpf.invJ,         \
                         ALL_GP_LOCS(i), ALL_REF_COORDS(h))
  
  ip_ctx_data.domain   = eb->el->domain;
  ip_ctx_data.h        = h;
  ip_ctx.time          = WD_time (eb->el->domain);
  ip_ctx.time_step     = WD_time_step (eb->el->domain);
  ip_ctx.ctx_data      = &ip_ctx_data;
  ip_ctx.get_kinem_var = get_kinem_var;
  ip_ctx_data.sf       = &shpf;

  FOR (i, WBMAT_NNODES) { GET_FROM_NODE (h->node[i], WN_get_displ, u[i]); }

  rho = WM3D_get_volume_mass (m);

  FOR (i, WBMAT_NNODES*3) {FOR (j, WBMAT_NNODES*3) { K[i][j] = 0; }}
  
  FOR (i, 9) {
    EVAL_SF(i);
    at_new_ip();
    jdetw = GP_WGT(i) * shpf.detJ;
    b_l_mat  (shpf.n_x, shpf.n_y, shpf.n_z, WBMAT_NNODES, u, b_l);
    if (!WM3D_get_mat_stiff (m, h->matst[i], &ip_ctx,
                             MAT_3D_STIFF_LAGRANGIAN, C))
      REPORT_FATAL_ERROR (eb->el->domain->errh, "No material tangent");
    add_bltcbl_upper_triangle (K, b_l, C, stiff_mult * jdetw); 
    if (!WM3D_update_stress (m, h->matst[i], W_MAT_STATE_UPDATE_TEMP,
                          &ip_ctx, MAT_3D_2ND_PK, S, &out_type))
      REPORT_ERROR(eb->el->domain->errh, "Failed to compute stress");
    b_nl_mat (shpf.n_x, shpf.n_y, shpf.n_z, WBMAT_NNODES, b_nl);
    add_bnltsbnl_upper_triangle (K, b_nl, S, stiff_mult * jdetw);
    switch (mass_type) {
    case W_MASS_CONSISTENT:
      add_rhontn_upper_triangle (K, shpf.n, mass_mult * rho * jdetw); 
      break;
    default:
      /* VOID: see the switch below */
      break;
    }
  }

  switch (mass_type) {
  case W_MASS_CONSISTENT:
    /* VOID: handled in the quadrature loop above */
    break;
  case W_MASS_DIAG:
  default:
    add_to_diag (K, mass_mult * h->mass / WBMAT_NNODES);
    break;
  }
  
  mirror (K);
  
  k = 0;
  FOR (i, WBMAT_NNODES) {
    WN_eqn_numbers_and_constraints (h->node[i], &idcs[k], &constrained[k]);
    k += 3;
  }

  (*assemble_eff_stiff_callback)  (eb->el->domain->newmark_integrator,
                                   FLES_ADD,
                                   WBMAT_NNODES*3, idcs,
                                   (double *) K, constrained);
#undef  EVAL_SF
}

static void
assemble_eff_stiffness (W_elem_t *el,
                        W_newmark_assemble_eff_stiff_proc assemble_eff_stiff_callback,
                        W_float_t stiff_mult,
                        W_float_t mass_mult,
                        W_mass_type_enum mass_type
                        )
{
  elem_block_t *eb = (elem_block_t *)el->data;
  FOR_ALL_ELEM_IN_BLOCK {
    el_assemble_eff_stiffness (e, eb, assemble_eff_stiff_callback,
                               stiff_mult, mass_mult, mass_type);
  }
}

static void
el_assemble_residual_forces (elem_t *h, elem_block_t *eb,
                          W_mat_state_update_form_enum update_form,
                          W_mass_type_enum mass_type)
{
  el_assemble_restoring_forces (h, eb, update_form);
  switch (mass_type) {
  case W_MASS_DIAG:
    el_assemble_inertial_forces_diag (h, eb);
    break;
  case W_MASS_CONSISTENT:
  default:
    el_assemble_inertial_forces_consistent (h, eb);
    break;
  }
}

static void
assemble_residual_forces (W_elem_t *el,
                          W_mat_state_update_form_enum update_form,
                          W_mass_type_enum mass_type)
{
  assemble_restoring_forces (el, update_form);
  {
    elem_block_t *eb = (elem_block_t *)el->data;
    FOR_ALL_ELEM_IN_BLOCK {
      switch (mass_type) {
      case W_MASS_DIAG:
        el_assemble_inertial_forces_diag (e, eb);
        break;
      case W_MASS_CONSISTENT:
      default:
        el_assemble_inertial_forces_consistent (e, eb);
        break;
      }
    }
  }
}

static void
el_which_qp (elem_t *h, elem_block_t *eb, W_loc_t *loc,
             int *whichqp, double *mindist)
{
  int i, m;
  double n[WBMAT_NNODES];
  integration_point_t *ip = eb->ip;
  double dist[9] = {
    FLT_MAX,FLT_MAX,FLT_MAX,FLT_MAX,FLT_MAX,FLT_MAX,FLT_MAX,FLT_MAX,FLT_MAX
  };
  W_loc_t l;

  FOR (i, 9) {
    if (h->matst[i]) {
      G3D_hex_N (n, ALL_GP_LOCS (i));
      VECMAC_SCALAR (l, =, 0);
      FOR (m, 8) {
        VECMAC_2_VEC (l, +=, n[m] * h->node[m]->ref_coords);
      }
      VECMAC_2_VEC (l, -=, (*loc));
      dist[i] = VECMAC_NORM (l);
    }
  }

  *whichqp = -1;
  *mindist = FLT_MAX;
  FOR (i, 9) {
    if (dist[i] < FLT_MAX && dist[i] < *mindist) {
      *whichqp = i;
      *mindist = dist[i];
    }
  }
}

static void
sel_elem_and_qp (elem_block_t *eb, W_loc_t *loc, elem_t **h, int *qp)
{
  elem_t *chosen_h = NULL;
  int chosen_qp = -1;
  int w;
  double md;
  double global_mindist = FLT_MAX;
  FOR_ALL_ELEM_IN_BLOCK {
    el_which_qp (e, eb, loc, &w, &md);
    if (md < global_mindist) {
      chosen_h = e;
      chosen_qp = w;
      global_mindist = md;
    }
  }
  *h  = chosen_h;
  *qp = chosen_qp;
}

static void *
new_history_client_data (W_elem_t *el,
                         W_index_t idx, /* unique integer  */
                         W_history_writer_t *r, /* history writer */
                         char *input_line /* input line (to allow
                                             for additional parsing of
                                             the line contents by this
                                             function) */
                         )
{
  elem_block_t *eb = (elem_block_t *)el->data;
  history_client_data_t *cd;
  W_loc_t loc;
  W_domain_t *d = el->domain;
  
  cd = w_new (history_client_data_t);
  CHECK_MALLOC(cd);
 
  TOKENS_PARSE(input_line);
  cd->el     = el;
  cd->pntidx = TOKEN_INT(3);
  loc.x  = TOKEN_DBL(4);
  loc.y  = TOKEN_DBL(5);
  loc.z  = TOKEN_DBL(6);
  sel_elem_and_qp (eb, &loc, &cd->h, &cd->which_qp);
  if (cd->h == NULL || (cd->which_qp < 0 || cd->which_qp >= 9))
    REPORT_FATAL_ERROR (d->errh, "Wrong quadrature point");

  return (NODE)cd;
}

static void
calc_stress (elem_t *h, elem_block_t *eb,
            int which_qp,
            W_output_stress_type_enum stress_type, /* stress type */
            double stress[3][3] /* output: stress */
            )
{
  W_domain_t *d = eb->el->domain;
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  W_mat_3d_stress_type_enum out_type;
  W_mat_3d_t *m = (W_mat_3d_t *)eb->material->data;
  integration_point_t *ip = eb->ip;
#undef EVAL_SF
#define EVAL_SF(i) \
  {                                             \
    G3D_hex20_interp_opt(shpf.n,        \
                       shpf.n_x,      \
                       shpf.n_y,      \
                       shpf.n_z,      \
                       &shpf.detJ,    \
                       shpf.J,        \
                       shpf.invJ,     \
                       ALL_GP_LOCS(which_qp),      \
                       ALL_REF_COORDS(h));      \
  }

  ip_ctx_data.domain   = eb->el->domain;
  ip_ctx_data.h        = h;
  ip_ctx.time          = WD_time (eb->el->domain);
  ip_ctx.time_step     = WD_time_step (eb->el->domain);
  ip_ctx.ctx_data      = &ip_ctx_data;
  ip_ctx.get_kinem_var = get_kinem_var;
  ip_ctx_data.sf       = &shpf;
  
  EVAL_SF(which_qp);
  at_new_ip();
  if      (stress_type == W_OUTPUT_STRESS_CAUCHY) {
    if (!WM3D_update_stress (m, h->matst[which_qp], W_MAT_STATE_UPDATE_TEMP,
                         &ip_ctx, MAT_3D_CAUCHY, stress, &out_type))
      REPORT_FATAL_ERROR(d->errh, "Failed to compute stress");
    if (out_type != MAT_3D_CAUCHY)
      REPORT_FATAL_ERROR(d->errh, "Stress type not available");
  } else if (stress_type == W_OUTPUT_STRESS_2ND_PK) {
    if (!WM3D_update_stress (m, h->matst[which_qp], W_MAT_STATE_UPDATE_TEMP,
                         &ip_ctx, MAT_3D_2ND_PK, stress, &out_type))
      REPORT_FATAL_ERROR(d->errh, "Failed to compute stress");
    if (out_type != MAT_3D_2ND_PK)
      REPORT_FATAL_ERROR(d->errh, "Stress type not available");
  }
}

static void
get_equiv_pl_def (elem_t *h, elem_block_t *eb,
                  int which_qp,
                  double *equiv_pl_def
                  )
{
  W_domain_t *d = eb->el->domain;
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  W_mat_3d_t *m = (W_mat_3d_t *)eb->material->data;
  integration_point_t *ip = eb->ip;
#undef EVAL_SF
#define EVAL_SF(i) \
  {                                             \
    G3D_hex20_interp_opt(shpf.n,        \
                       shpf.n_x,      \
                       shpf.n_y,      \
                       shpf.n_z,      \
                       &shpf.detJ,    \
                       shpf.J,        \
                       shpf.invJ,     \
                       ALL_GP_LOCS(which_qp),      \
                       ALL_REF_COORDS(h));      \
  }

  ip_ctx_data.domain   = eb->el->domain;
  ip_ctx_data.h        = h;
  ip_ctx.time          = WD_time (eb->el->domain);
  ip_ctx.time_step     = WD_time_step (eb->el->domain);
  ip_ctx.ctx_data      = &ip_ctx_data;
  ip_ctx.get_kinem_var = get_kinem_var;
  ip_ctx_data.sf       = &shpf;
  
  EVAL_SF(which_qp);
  at_new_ip();
  if (!WM3D_get_mat_state_var(m, h->matst[which_qp],
                              &ip_ctx, MAT_3D_EQUIV_PLASTIC_DEF, equiv_pl_def))
    REPORT_FATAL_ERROR(d->errh, "Failed to retrieve equiv plastic def");
}


static void
update_history_at_material (elem_t *el, elem_block_t *eb,
                            int which_qp,
                            W_solution_history_t *h,
                            W_index_t pntidx
                            )
{
  W_domain_t *d = eb->el->domain;
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  W_mat_3d_stress_type_enum out_type;
  W_mat_3d_t *m = (W_mat_3d_t *)eb->material->data;
  integration_point_t *ip = eb->ip;
#undef EVAL_SF
#define EVAL_SF(i) \
  {                                             \
    G3D_hex20_interp_opt(shpf.n,        \
                       shpf.n_x,      \
                       shpf.n_y,      \
                       shpf.n_z,      \
                       &shpf.detJ,    \
                       shpf.J,        \
                       shpf.invJ,     \
                       ALL_GP_LOCS(which_qp),      \
                       ALL_REF_COORDS(h));      \
  }

  ip_ctx_data.domain   = eb->el->domain;
  ip_ctx_data.h        = el;
  ip_ctx.time          = WD_time (eb->el->domain);
  ip_ctx.time_step     = WD_time_step (eb->el->domain);
  ip_ctx.ctx_data      = &ip_ctx_data;
  ip_ctx.get_kinem_var = get_kinem_var;
  ip_ctx_data.sf       = &shpf;
  
  WM3D_update_history (m, el->matst[which_qp], &ip_ctx, h, pntidx, ip_ctx.time);
}

static void
update_history (W_elem_t *el, /* generic element */
                W_solution_history_t *h, /* solution history */
                W_float_t time
                )
{
  elem_block_t *eb = (elem_block_t *)el->data;
  history_client_data_t *cd =
    (history_client_data_t *)h->client_data;
  double S[3][3], pressure, equiv_pl_def;
#define STORE(CD, HW, NAME, TIME, VAL)          \
  {                                             \
    char buf[128];                              \
    sprintf (buf, "p%d%s", CD->pntidx, NAME);   \
    WHR_store (HW, buf, TIME, VAL);             \
  }

  calc_stress (cd->h, eb, cd->which_qp,
              WOPT_OPT (h->writer->domain->options, output_stress_type),
              S);
  pressure  = (S[0][0] + S[1][1] + S[2][2]) / 3;
  if        (WOPT_OPT (h->writer->domain->options, output_stress_type) == W_OUTPUT_STRESS_CAUCHY) {
    STORE (cd, h->writer, "sxx", time, S[0][0]);
    STORE (cd, h->writer, "sxy", time, S[0][1]);
    STORE (cd, h->writer, "sxz", time, S[0][2]);
    STORE (cd, h->writer, "syy", time, S[1][1]);
    STORE (cd, h->writer, "syz", time, S[1][2]);
    STORE (cd, h->writer, "szz", time, S[2][2]);
    STORE (cd, h->writer, "press", time, pressure);
  } else if (WOPT_OPT (h->writer->domain->options, output_stress_type) == W_OUTPUT_STRESS_2ND_PK) {
    STORE (cd, h->writer, "Sxx", time, S[0][0]);
    STORE (cd, h->writer, "Sxy", time, S[0][1]);
    STORE (cd, h->writer, "Sxz", time, S[0][2]);
    STORE (cd, h->writer, "Syy", time, S[1][1]);
    STORE (cd, h->writer, "Syz", time, S[1][2]);
    STORE (cd, h->writer, "Szz", time, S[2][2]);
    STORE (cd, h->writer, "Press", time, pressure);
  } else {
    REPORT_ERROR(h->writer->domain->errh, "Unknown stress type for output");
  }
  get_equiv_pl_def (cd->h, eb, cd->which_qp, &equiv_pl_def);
  STORE (cd, h->writer, "EPlS", time, equiv_pl_def);
  update_history_at_material (cd->h, eb, cd->which_qp, h, cd->pntidx);
}


static W_float_t
el_kinetic_energy_consistent (elem_t *h, elem_block_t *eb)
{
  W_mat_3d_t *m = (W_mat_3d_t *)eb->material->data;
  W_vect_t v[WBMAT_NNODES];
  double Nmat[WBMAT_NNODES];
  double jdetw, J[3][3], rho;
  int i, n, l;
  double dke, ke = 0;
  double vdotv[WBMAT_NNODES][WBMAT_NNODES];
  integration_point_t *ip = eb->ip;

  FOR (i, WBMAT_NNODES) {
    GET_FROM_NODE (h->node[i], WN_get_velocity, v[i]);
  }

  /* note that we fill only the upper triangle */
  FOR (l, WBMAT_NNODES) {
    FOR (n, WBMAT_NNODES) {
      vdotv[l][n] = VECMAC_DOT (v[l], v[n]);
    }
  }

  rho = WM3D_get_volume_mass(m);

  FOR (i, 9) {
    G3D_hex20_N (Nmat, ALL_GP_LOCS(i));
    G3D_hex20_J (J, ALL_GP_LOCS(i), ALL_REF_COORDS(h));
    jdetw = GP_WGT(i) * G3D_hex20_detJ (J);
    FOR(n, WBMAT_NNODES) {
      FOR (l, WBMAT_NNODES) {
        dke = jdetw * Nmat[n] * Nmat[l] * vdotv[n][l];
        ke  += dke;
      }
    }
  }

  return (rho * ke / 2);
#undef  EVAL_SF
}

static W_float_t
el_kinetic_energy_diag (elem_t *h, elem_block_t *eb)
{
  W_mat_3d_t *m = (W_mat_3d_t *)eb->material->data;
  W_vect_t v[WBMAT_NNODES];
  double Nmat[WBMAT_NNODES];
  double jdetw, J[3][3], rho;
  int i, j;
  double dke, ke = 0;
  double massmat[WBMAT_NNODES];
  integration_point_t *ip = eb->ip;
  
  FOR (i, WBMAT_NNODES) {
    GET_FROM_NODE (h->node[i], WN_get_velocity, v[i]);
    massmat[i] = 0;
  }
  
  rho = WM3D_get_volume_mass(m);
  
  FOR (i, 9) {
    G3D_hex20_N (Nmat, ALL_GP_LOCS(i));
    G3D_hex20_J (J, ALL_GP_LOCS(i), ALL_REF_COORDS(h));
    jdetw = rho * GP_WGT(i) * G3D_hex20_detJ (J);
    FOR (j, WBMAT_NNODES) {
      massmat[j] += jdetw * Nmat[j];
    }
  }
  
  FOR (i, WBMAT_NNODES) {
    dke = massmat[i] * VECMAC_DOT (v[i],v[i]);
    ke  += dke;
  }

  return (ke / 2);
#undef  EVAL_SF 
}

static W_float_t
kinetic_energy (W_elem_t *el, W_mass_type_enum mass_type)
{
  elem_block_t *eb = (elem_block_t *)el->data;
  double ke = 0;
  FOR_ALL_ELEM_IN_BLOCK {
    switch (mass_type) {
    case W_MASS_CONSISTENT:
      ke += el_kinetic_energy_consistent (e, eb);
    default:
      ke += el_kinetic_energy_diag (e, eb);
    }
  }
  return ke;
}

static W_elem_methods_t methods = {
  elem_elcode,
  type_name,
  assemble_restoring_forces,
  assemble_mass,
  total_mass,
  NULL,
  suggested_time_step,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  new_history_client_data,
  update_history,
  NULL,
  write_postproc_file,
#ifdef X_XFACE
  modify_elem_grep,
#endif
  NULL,
  NULL,
  NULL,
  num_of_faces,
  get_face_vertex_nodes,
  NULL,
  assemble_residual_forces,
  assemble_eff_stiffness,
  kinetic_energy
};

/*
  Element reader.
  Reads element data for the hex20aq element block type
  from stream as long as there is no
  END_OF_BLOCK tag in the input line.
  Creates hex volumes, and the block element.

  Note:

  Hex20aq is an element block, which means that while it pretends to
  be a single element, it is actually a block (collection) of hex20 elements
  with 9-point quadrature, of the same material, and all using the same
  quadrature.  The block dispatches all protocol requests to its hex volumes.
  
 */
W_bool_t /* TRUE if successful */
hex20aq_reader (W_domain_t *d /* domain */)
{
  const char *input_line;
  long elem_num = 0;
  elem_block_t *eb;
  elem_t *h;
#ifdef PVM_NEED_GHOST_ELEMENTS
  int is_ghost = FALSE;
#endif
  W_index_t nodeid, matid;
  int i, j, k;
  W_material_t *mat;
  W_mat_3d_t *m;
  int had_material = FALSE;
  int had_w0 = FALSE;
  int had_elem_num = FALSE;

  eb = w_new(elem_block_t); CHECK_MALLOC(eb);
  eb->el = NULL;
  eb->elist = make_list (); CHECK_MALLOC (eb->elist);
  eb->material = NULL;
  eb->W0 = 16.0 / 3.0;
  make_integration_points (eb);
  
  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line)) {
#     ifdef PVM_NEED_GHOST_ELEMENTS
      is_ghost = FALSE;
      if (TOKENS_HAVE_KEYWORD("GHOST"))
        is_ghost = TRUE;
#     endif      
      compute_total_mass (eb);
      eb->el = WE_new_element (d, &methods, eb
#     ifdef PVM_NEED_GHOST_ELEMENTS
                               , is_ghost
#     endif
                               );
      if (!had_elem_num) elem_num = 1;
      WD_add_element(d, elem_num, eb->el);
      return TRUE;
    } else {
      TOKENS_PARSE (input_line);
      if        (TOKENS_HAVE_KEYWORD("MATERIAL")) {
        if (had_material) {
          REPORT_ERROR(d->errh, "Material may be specified only once");
        }
        matid = TOKEN_INT (2);
        mat = WD_get_material_by_idx (d, matid);
        if (mat == NULL) {
          sprintf(W_buf(), "No such material: %ld", matid);
          REPORT_ERROR(d->errh, W_buf());
        }
        eb->material = mat;
        had_material = TRUE;
      } else if (TOKENS_HAVE_KEYWORD("CENTER_WEIGHT")) {
        if (had_w0) {
          REPORT_ERROR(d->errh, "Center weight may be specified only once");
        }
        eb->W0 = TOKEN_DBL (2);
        if (eb->W0 > 16.0 / 3.0) {
          REPORT_ERROR(d->errh, "Center weight must be <= 16/3");
        }
        make_integration_points (eb);
        had_w0 = TRUE;
      } else if (TOKENS_HAVE_KEYWORD("ELEM_NUM")) {
        if (had_elem_num) {
          REPORT_ERROR(d->errh, "Element number may be specified only once");
        }
        elem_num = TOKEN_INT (2);
        had_elem_num = TRUE;
      } else {
        h = w_new (elem_t); CHECK_MALLOC (h);
        FOR (i, WBMAT_NNODES) {
          nodeid = TOKEN_INT(2+i);
          h->node[i] = WD_get_node_by_idx (d, nodeid);
          if (h->node[i] == NULL) {
            sprintf(W_buf(), "No such node: %ld", nodeid);
            REPORT_ERROR (d->errh, W_buf());
          }
        }
        if (!had_material) {
          REPORT_ERROR(d->errh,
                       "Material needs to be specified prior to element connectivity");
        }
        m = (W_mat_3d_t *)eb->material->data;
        {
          integration_point_t *ip = eb->ip;
          FOR(i, 9) {
            W_loc_t loc;
            G3D_hex20_N(shpf.n, ALL_GP_LOCS(i));
            LOC(h, loc);
            h->matst[i] = WM3D_get_mat_state (m, &loc);
          }
        }
        h->serial_num = list_length (eb->elist);
        add_to_tail (eb->elist, h);
      }
    } /* closing else */
  }
  return FALSE;
}

