#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define DIM 3
#define DIM2 (DIM*DIM)
#define REPEAT 1000000

int
arrays_fixed(double a[DIM][DIM][DIM], double A[DIM][DIM][DIM]);
int
arrays_var(double a[DIM][DIM][DIM], double A[DIM][DIM][DIM]);
int
ptrs_fixed(double *a, double *A);
int
ptrs_var(double *a, double *A);
int
ptrs_var_opt(double *a, double *A);

static int n, n2;

int
main()
{
  double a[DIM][DIM][DIM];
  int i, j, k;
  
  for (i = 0; i < DIM; i++)
    for (j = 0; j < DIM; j++)
      for (k = 0; k < DIM; k++)
        a[i][j][k] = rand();
  
#if 0
#  if 0
     fprintf(stderr, "arrays_fixed\n");
     for (i = 0; i < REPEAT; i++)
       arrays_fixed(a, a);
#  else
     fprintf(stderr, "arrays_var\n");
     n = DIM;
     n2 = DIM * DIM;
     for (i = 0; i < REPEAT; i++)
       arrays_var(a, a);
#  endif
#else 
#  if 0
     fprintf(stderr, "ptrs_fixed\n");
     for (i = 0; i < REPEAT; i++)
       ptrs_fixed((double *)a, (double *)a);
#  else
#    if 0
          fprintf(stderr, "ptrs_var\n");
          n = DIM;
          n2 = DIM * DIM;
          for (i = 0; i < REPEAT; i++)
            ptrs_var((double *)a, (double *)a);
#     else
          fprintf(stderr, "ptrs_var_opt\n");
          n = DIM;
          n2 = DIM * DIM;
          for (i = 0; i < REPEAT; i++)
            ptrs_var_opt((double *)a, (double *)a);

#     endif
#  endif
#endif
}


#define ACCESS(A, i, j, k) (*(A + i * DIM2 + j * DIM + k)) 
#if 0
/* M 2 A 3 */
#define Access(A, i, j, k) (*(A + i * n2 + j * n + k)) 
#else
/* M 2 A 3 */
#define Access(A, i, j, k) (*(A + ((i * n + j) * n) + k)) 
#endif

#define AccessOpt(A, in2, jn, k) (*(A + in2 + jn + k)) 


/* This requires 7.8 sec */
int
arrays_fixed(double a[DIM][DIM][DIM], double A[DIM][DIM][DIM])
{
  int i, j, k;
  
  for (i = 0; i < DIM; i++)
    for (j = 0; j < DIM; j++)
      for (k = 0; k < DIM; k++) {
        if (a[i][j][k] - A[i][j][k] != 0)
          fprintf(stderr, "mismatch\n");
      }
}


/* This requires 9.5 sec */
int
arrays_var(double a[DIM][DIM][DIM], double A[DIM][DIM][DIM])
{
  int i, j, k;
  
  for (i = 0; i < n; i++)
    for (j = 0; j < n; j++)
      for (k = 0; k < n; k++) {
        if (a[i][j][k] - A[i][j][k] != 0)
          fprintf(stderr, "mismatch\n");
      }
}


/* This requires 7.8 sec */
int
ptrs_fixed(double *a, double *A)
{
  int i, j, k;
  
  for (i = 0; i < DIM; i++)
    for (j = 0; j < DIM; j++)
      for (k = 0; k < DIM; k++) {
        if (ACCESS(a, i, j, k) - ACCESS(A, i, j, k) != 0)
          fprintf(stderr, "mismatch\n");
      }
}


/* This requires 19.8 sec */
int
ptrs_var(double *a, double *A)
{
  int i, j, k;
  
  for (i = 0; i < n; i++)
    for (j = 0; j < n; j++)
      for (k = 0; k < n; k++) {
        if (Access(a, i, j, k) - Access(A, i, j, k) != 0)
          fprintf(stderr, "mismatch\n");
      }
}

#define LOOP_OVER_IPA(NIP, ANEPXR)                     \
   {                                                   \
     register LOIPA_i, LOIPA_j, LOIPA_k;               \
     register LOIPA_in2, LOIPA_jn;                     \
     int LOIPA_NIP2 = NIP * NIP;                       \
                                                       \
     for (LOIPA_i = 0; LOIPA_i < NIP; LOIPA_i++) {     \
       LOIPA_in2 = LOIPA_i * LOIPA_NIP2;               \
       for (LOIPA_j = 0; LOIPA_j < NIP; LOIPA_j++) {   \
         LOIPA_jn = LOIPA_j * NIP;                     \
         for (LOIPA_k = 0; LOIPA_k < NIP; LOIPA_k++) { \
           ANEPXR;                                     \
         }                                             \
       }                                               \
     }                                                 \
   }

#define IPA_ENTRY(A) \
  (*(A + LOIPA_in2 + LOIPA_jn + LOIPA_k)) 


/* This requires 11.8 sec */
int
ptrs_var_opt(double *a, double *A)
{
  LOOP_OVER_IPA(n,
                {
                  if (IPA_ENTRY(a) - IPA_ENTRY(A) != 0)
                    fprintf(stderr, "mismatch\n");
                }
                );
}



