

#include "Wconfig.h"
#include "Wxf.h"
#include "IMPLi.h"
#include "Wutil.h"


static void
impl_advance(impl_integrator_t *i);


impl_integrator_t *
impl_new_integrator
    (void                        *domain,
     W_float_t                   t0,
     W_float_t                   dt,
     impl_initialize_proc        initialize,
     impl_solve_for_accel_proc   solve_for_accel,
     impl_update_prediction_proc update_prediction,
     impl_calc_forces_proc       calc_forces,
     impl_has_converged_proc     has_converged
     )
{  
  impl_integrator_t *i;
  
  /* Create integrator */
  i = w_new(impl_integrator_t);
  CHECK_MALLOC(i);

  /* Domain */
  i->domain            = domain;
  
  /* Select time step dt */
  i->t                 = t0;
  i->dt                = dt;
  
  /* Initialize procedures */
  i->initialize        = initialize;
  i->solve_for_accel   = solve_for_accel;
  i->update_prediction = update_prediction;
  i->calc_forces       = calc_forces;
  i->has_converged     = has_converged;

  i->advance           = impl_advance;
  
  return i;
}




void
impl_advance_integrator(impl_integrator_t *i)
{
  assert(i != NULL);
  assert(i->advance != NULL);
  (*i->advance)(i);
}



static void
impl_advance(impl_integrator_t *i)
{
  i->converged = FALSE;
  
  do {

    i->calc_forces(i->domain);    /* Calculate nodal forces */

    i->solve_for_accel(i->domain); /* Solve for accelerations */

    i->converged = i->has_converged(i->domain); /* Converged? */
    
    i->update_prediction(i->domain); /* Update predicted veloc and displ */

  } while (!i->converged);
  
  i->t += i->dt;
}




W_float_t
impl_integrator_time(impl_integrator_t *i)
{
  /* If not yet converged, the integrator pretends the time had already */
  /* been advanced */
  return (i->converged ? i->t : i->t + i->dt);
}



W_float_t
impl_integrator_dt(impl_integrator_t *i)
{
  return i->dt;
}
