#include "Wconfig.h"
#include "Wxf.h"
#include "Wdomain.h"
#include "Wnode.h"
#include "Wivcond.h"
#include "Wvset.h"
#include "Wpvm.h"
#include "Wcdi.h"

/* RCSHEADER "$Id: ivcond.c,v 1.1.1.1 2001/07/24 02:42:51 pkrysl Exp $" */

typedef struct W_initial_velocity_cond_t {
  int                    id;
  W_domain_t            *domain;
  LIST_OF (G3D_box_t *)  boxl;
  char                  *vel_expr[3];
}              W_initial_velocity_cond_t;



static W_bool_t
iv (W_initial_velocity_cond_p iv_cond, W_coords_3d_t *at, W_loc_t *v_initial)
{
  int stat;
  double f;
  
  expeval_set_value ("x", at->x);
  expeval_set_value ("y", at->y);
  expeval_set_value ("z", at->z);

  if (iv_cond->vel_expr[0]) {
    stat = expeval_eval(iv_cond->vel_expr[0], &f);
    if (stat != EXPEVAL_E_OK) {
      char buf[1024];
      sprintf(buf, "%s", expeval_err_msgs[stat]);
      REPORT_WARNING(iv_cond->domain->errh, buf);
    }
    v_initial->x = f;
  } else {
    v_initial->x = 0;
  }
  if (iv_cond->vel_expr[1]) {
    stat = expeval_eval(iv_cond->vel_expr[1], &f);
    if (stat != EXPEVAL_E_OK) {
      char buf[1024];
      sprintf(buf, "%s", expeval_err_msgs[stat]);
      REPORT_WARNING(iv_cond->domain->errh, buf);
    }
    v_initial->y = f;
  } else {
    v_initial->y = 0;
  }
  if (iv_cond->vel_expr[2]) {
    stat = expeval_eval(iv_cond->vel_expr[2], &f);
    if (stat != EXPEVAL_E_OK) {
      char buf[1024];
      sprintf(buf, "%s", expeval_err_msgs[stat]);
      REPORT_WARNING(iv_cond->domain->errh, buf);
    }
    v_initial->z = f;
  } else {
    v_initial->z = 0;
  }
  return TRUE;
}

/*
  Evaluate initial velocity at a point.
  Evaluate initial velocity at a given point.
  If no initial condition is defined at the point, FALSE is returned
  and the data is not valid.
 */
W_bool_t /* TRUE if successful */
WIVC_eval (W_initial_velocity_cond_p iv_cond, /* initial condition */
           W_coords_3d_t *at, /* at point */
           W_loc_t *v_initial /* output vector */
           )
{
  if (iv_cond->boxl) {
    G3D_box_t *box;
    box = NULL;
    while ((box = (G3D_box_t *) get_list_next (iv_cond->boxl, box))) {
      if (G3D_PT_INSIDE_BOX ((*box), (*at))) {
        return iv (iv_cond, at, v_initial);
      }
    }
  } else {
    return iv (iv_cond, at, v_initial);
  }
  return FALSE;
}
 

/*
  Initial condition reader.
  Reads data for the initial conditions.
 */
W_bool_t                        /* TRUE if successful */
W_init_cond_reader (W_domain_t *d/* domain */)
{
  const char *input_line;
  W_initial_velocity_cond_t *ivc = NULL;
  W_coords_3d_t x;

  while ((input_line = W_get_input_line(d)) != NULL) {
    TOKENS_PARSE(input_line);
    if (W_is_end_of_block (input_line))
      return TRUE;
    else {
      TOKENS_PARSE (input_line);
      if        (TOKENS_HAVE_KEYWORD ("ID")) {
        ivc = w_new (W_initial_velocity_cond_t); CHECK_MALLOC (ivc);
        ivc->id          = TOKEN_INT (2);
        ivc->domain      = d;
        ivc->boxl        = NULL;
        ivc->vel_expr[0] = NULL;
        ivc->vel_expr[1] = NULL;
        ivc->vel_expr[2] = NULL;
      } else if (TOKENS_HAVE_KEYWORD ("END")) {
        WD_add_iv_cond (d, ivc->id, ivc);
        ivc = NULL;
      } else if (TOKENS_HAVE_KEYWORD ("BOX")) {
        G3D_box_t *box;
        if (!ivc)
          REPORT_FATAL_ERROR (d->errh, "No initial velocity being defined");
        box = w_new (G3D_box_t); CHECK_MALLOC (box);
        x.x = TOKEN_DBL (2);
        x.y = TOKEN_DBL (3);
        x.z = TOKEN_DBL (4);
        G3D_INIT_BOX ((*box), x);
        x.x = TOKEN_DBL (5);
        x.y = TOKEN_DBL (6);
        x.z = TOKEN_DBL (7);
        G3D_UPDT_BOX ((*box), x);
        if (!ivc->boxl) ivc->boxl = make_list (); CHECK_MALLOC (ivc->boxl);
        add_to_tail (ivc->boxl, box);
      } else if (TOKENS_HAVE_KEYWORD ("IV")) {
        const char *s;
        if (!ivc)
          REPORT_FATAL_ERROR (d->errh, "No initial velocity being defined");
        s = TOKEN_STR (2);
        DUPLICATE_STRING (ivc->vel_expr[0], s);
        s = TOKEN_STR (3);
        DUPLICATE_STRING (ivc->vel_expr[1], s);
        s = TOKEN_STR (4);
        DUPLICATE_STRING (ivc->vel_expr[2], s);
      } else {
        sprintf (W_buf (), "Input not recognized: \"%s\"", input_line);
        REPORT_FATAL_ERROR (d->errh, W_buf ());
      }
    }
  }
  return FALSE;
}
