#include "Wconfig.h"
#include "Wxf.h"
#include "Wloads.h"
#include "Wdomain.h"
#include "Wnode.h"
#include "Welem.h"
#include "Wtfuncs.h"

#include <ctype.h>

/* RCSHEADER "$Id: loads.c,v 1.2 2002/03/04 17:37:13 cvs Exp $" */

#define SMALL_NUM 0.01
#define LARGE_NUM 100

/*
   Create a new load.
 */
static W_load_t * /* new load */
new_load (void)
{
  W_load_t *l;

  l = w_new(W_load_t);
  CHECK_MALLOC(l);
  
  return l;
}

static void
make_node_group_list (W_load_t *l, W_domain_t *d)
{
  W_node_group_node_t *gn;
  W_node_t *n;
  W_vect_t rc;

  l->data.node_group.nodes = make_list ();
  n = NULL;
  while ((n = WD_get_next_node (d, n)) != NULL) {
    WN_get_ref_coords (n, &rc.x, &rc.y, &rc.z);
    if (G3D_PT_INSIDE_BOX (l->data.node_group.box, rc)) {
      gn = w_new (W_node_group_node_t);
      CHECK_MALLOC (gn);
      gn->n     = n;
#ifdef X_XFACE
      gn->f_rep = NULL;
#endif
      add_to_tail (l->data.node_group.nodes, gn);
    } /* if */
  } /* while */
}

#ifdef X_XFACE

static void
assemble_mouse_anim_load(W_domain_t *d, W_node_t *n)
{ 
  WCRec wcp;
  W_coords_3d_t c, V, F;
  W_float_t tm, f, L, cos_ax, cos_ay, cos_az;

  EVLastPointerWCPos(&wcp);
  WN_get_coords (n, &c.x, &c.y, &c.z);
  V.x = wcp.x - c.x;
  V.y = wcp.y - c.y;
  V.z = wcp.z - c.z;
  L   = W_vector_3d_length (&V);
  cos_ax = V.x / L;  cos_ay = V.y / L;  cos_az = V.z / L;
  tm  = WD_total_mass(d);
  if (L != 0) {
    f = LARGE_NUM * tm * (1/SMALL_NUM - 1/(L+SMALL_NUM));
  } else
    f = 0;
  F.x = f * cos_ax;   F.y = f * cos_ay;  F.z = f * cos_az;
  WN_assemble_force (n, &F);
}

#endif

static W_bool_t 
all_nodes_on_face_in_list (dict_dictionary nd, W_node_t *n[4])
{
  int i;
  FOR (i, 4) {
    if (n[i]) {
      if (dict_find_obj_with_handle_name (nd, n[i]) == NULL) {
        return FALSE;
      }
    }
  }
  return TRUE;
}

typedef struct nld_t {
  W_node_t         *n;
  W_float_t         area; /* area around node */
}              nld_t;

static nld_t *
new_nld (W_node_t *n)
{
  nld_t *nld = w_new (nld_t);
  CHECK_MALLOC (nld);
  nld->n = n;
  nld->area = 0;
  return nld;
}

static   dict_dictionary 
l_to_d (LIST l)
{
  dict_dictionary d = dict_create_dictionary ();
  W_node_group_node_t *ng;
  nld_t *nld;
  ng = NULL;
  while (ng = (W_node_group_node_t *) get_list_next (l, ng)) {
    nld = new_nld (ng->n);
    dict_add_obj_handle_name (d, nld, ng->n);
  }
  return d;
}

static void
trac_on_face (int nn, W_node_t *n[], dict_dictionary nd)
{
  int i;
  double a;
  W_vect_t txi, teta, normal={0,0,0};
  nld_t **nld;
  W_vect_t *c;
  
  nld = w_new_array (nn, nld_t *); CHECK_MALLOC (nld);
  c   = w_new_array (nn, W_vect_t); CHECK_MALLOC (c);

  FOR (i, nn) {
    WN_get_ref_coords (n[i], &c[i].x, &c[i].y, &c[i].z);
    nld[i] = dict_find_obj_with_handle_name (nd, n[i]);
    if (!nld[i]) REPORT_FATAL_ERROR (NULL, "Not found");
  }

  if        (nn == 4) { /* 4-noded quadrilateral */
    double n_xi[4], n_eta[4];
    int j;
    G3D_q4_N_xi (n_xi, 0, 0);
    G3D_q4_N_eta (n_eta, 0, 0);
    VECMAC_SCALAR (txi, =, 0);
    VECMAC_SCALAR (teta, =, 0);
    FOR (j, 4) {
      VECMAC_2_VEC ( txi, +=,  n_xi[j] * c[j]);
      VECMAC_2_VEC (teta, +=, n_eta[j] * c[j]);
    }
    VECMAC_CROSS (normal, =, txi, x, teta);
    a = VECMAC_NORM (normal);
    FOR (j, 4) {
      nld[j]->area += a / 4;
    }
  } else if (nn == 3) { /* triangular face */
    VECMAC_3_VEC (txi, =, c[1], -, c[0]);
    VECMAC_3_VEC (teta, =, c[2], -, c[0]);
    VECMAC_CROSS (normal, =, txi, x, teta);
    a = VECMAC_NORM (normal) / 2;
    FOR (i, 3) {
      nld[i]->area += a / 3;
    }
  } else if (nn == 8) { /* 8-noded quadrilateral */
    double n_xi[8], n_eta[8], nmat[8];
    int i, j;
    GAUSS_2d_integration_rule_t *gr = gauss_rule_2d (2);
    FOR (i, gr->npoints) {
      G3D_q8_N     ( nmat, gr->ip[i].xi, gr->ip[i].eta);
      G3D_q8_N_xi  ( n_xi, gr->ip[i].xi, gr->ip[i].eta);
      G3D_q8_N_eta (n_eta, gr->ip[i].xi, gr->ip[i].eta);
      VECMAC_SCALAR (txi, =, 0);
      VECMAC_SCALAR (teta, =, 0);
      FOR (j, 8) {
        VECMAC_2_VEC ( txi, +=,  n_xi[j] * c[j]);
        VECMAC_2_VEC (teta, +=, n_eta[j] * c[j]);
      }
      VECMAC_CROSS (normal, =, txi, x, teta); 
      a = gr->ip[i].Wgt * VECMAC_NORM (normal);
      FOR (j, 8) {
        nld[j]->area += nmat[j] * a;
      }
    }
    gauss_free_2d (gr);
  }

  /* cleanup */
  w_free (nld);
  w_free (c);
}

static void
convert_to_nodal_loads (W_load_t *l, W_domain_t *d)
{
  W_elem_t *e;
  int i, f, nf;
  int nv, ndim = 4;
  W_node_t **n;
  dict_dictionary nd = l_to_d (l->data.node_group.nodes);
  
  n = w_new_array (ndim, W_node_t *); CHECK_MALLOC (n);
  
  e = NULL;
  while (e = WD_get_next_element (d, e)) {
    nf = WE_num_of_faces (e);
    FOR (f, nf) {
      FOR (i, 4) { n[i] = NULL; }
    retry:
      nv = WE_get_face_vertex_nodes (e, f, ndim, n);
      if (nv > ndim) {
        ndim = nv;
        n = w_realloc_array (n, ndim, W_node_t *); CHECK_MALLOC (n);
        goto retry;
      }
      if (all_nodes_on_face_in_list (nd, n)) {
        trac_on_face (nv, n, nd);
      }
    }
  }

  /* convert */
  { /* cleanup */
    nld_t *nld;
    W_load_t *nl;
    double total_force = 0;
    dict_iterator di = dict_get_iterator ();
    dict_reset_iterator (di);
    while (nld = (nld_t *) dict_iter_objs (nd, di)) {
      nl = new_load ();
      nl->type                     = NODAL_LOAD;
      nl->domain                   = d;
      nl->data.nodal.node_num      = nld->n->id;
      nl->data.nodal.direction     = l->data.node_group.spec.trac.direction;
      nl->data.nodal.amplitude     = nld->area * l->data.node_group.spec.trac.amplitude;
      total_force += nl->data.nodal.amplitude;
      nl->data.nodal.time_func_num = l->data.node_group.spec.trac.time_func_num;
#     ifdef X_XFACE
      nl->data.nodal.fg_rep        = NULL;
      nl->data.nodal.mg_rep        = NULL;
#     endif
      WD_add_load (d, nl); 
    }
    dict_del_iterator (di);
  }

  { /* cleanup */
    nld_t *nld;
    dict_iterator di = dict_get_iterator ();
    dict_reset_iterator (di);
    while (nld = (nld_t *) dict_iter_objs (nd, di)) {
      w_free (nld);
    }
    dict_del_iterator (di);
  }
  dict_destroy_dictionary (nd);
}
  
/*
   Reader of load data.
 */
W_bool_t
W_load_reader (W_domain_t *d)
{
  const char *input_line;
  W_load_t *l;
  long n_num, lcode, tf_num[3], dir;
  double amplitude;
  char sdir[2];

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return TRUE;
    else {
      if (sscanf(input_line, "%ld", &lcode) != 1)
        BAD_INPUT();
      switch (lcode) {
      case MOUSE_ANIM_LOAD:
        if ((input_line = W_get_input_line(d)) == NULL)
          BAD_INPUT();
        if (sscanf(input_line, "%ld", &n_num) != 1) {
          BAD_INPUT();
        } else {
          l = new_load ();
          l->type                     = MOUSE_ANIM_LOAD;
          l->data.nodal.node_num      = n_num;
          l->domain                   = d;
          WD_add_load(d, l); 
        }
        break;
      case NODAL_LOAD:
        if ((input_line = W_get_input_line(d)) == NULL)
          BAD_INPUT();
        
        if (sscanf(input_line, "%ld %ld %lf %ld",
                   &n_num, &dir, &amplitude, &tf_num[0]) != 4) {
          if (sscanf(input_line, "%ld %s %lf %ld",
                     &n_num, sdir, &amplitude, &tf_num[0]) != 4) {
            BAD_INPUT();
          } else {
            dir = W_DX;
            if      (toupper(sdir[0]) == 'D')
              if      (toupper(sdir[1]) == 'X') dir = W_DX;
              else if (toupper(sdir[1]) == 'Y') dir = W_DY;
              else if (toupper(sdir[1]) == 'Z') dir = W_DZ;
              else  { BAD_INPUT(); }
            else if (toupper(sdir[0]) == 'R')
              if      (toupper(sdir[1]) == 'X') dir = W_RX;
              else if (toupper(sdir[1]) == 'Y') dir = W_RY;
              else if (toupper(sdir[1]) == 'Z') dir = W_RZ;
              else  { BAD_INPUT(); }
            else
              { BAD_INPUT(); }
          }
        }
        l = new_load();
        l->type                     = NODAL_LOAD;
        l->domain                   = d;
        l->data.nodal.node_num      = n_num;
        
        l->data.nodal.direction     = dir;
        l->data.nodal.amplitude     = amplitude;
        l->data.nodal.time_func_num = tf_num[0];
        if (WD_get_time_func_by_idx (d, tf_num[0]) == NULL) {
          sprintf (W_buf (), "Time function #%d not found", (int) tf_num[0]);
          REPORT_FATAL_ERROR (d->errh, W_buf ());
        }
#       ifdef X_XFACE
        l->data.nodal.fg_rep        = NULL;
        l->data.nodal.mg_rep        = NULL;
#       endif
        WD_add_load(d, l); 
        break;
      case ELEM_LOAD:
        break;
      case NODE_GROUP_LOAD:
        {
          G3D_box_t box;
          double inflate;
          W_loc_t x;
          W_node_group_load_type_enum kind = W_NODE_GROUP_LOAD_FORCE;
          char *accel[3] = {NULL,NULL,NULL};
          
          if ((input_line = W_get_input_line(d)) == NULL)
            BAD_INPUT();
          TOKENS_PARSE (input_line);

          if (TOKEN_IS_STRING (1)) {
            const char *s;
            s = TOKEN_STR (1);
            if        (tokens_token_and_keyword_equiv ("BODY", s)) {
              kind = W_NODE_GROUP_LOAD_BODY;
            } else if (tokens_token_and_keyword_equiv ("TRAC", s)) {
              kind = W_NODE_GROUP_LOAD_TRAC;
            }
          }

          if (kind == W_NODE_GROUP_LOAD_FORCE) {
            if (TOKEN_IS_STRING (1)) {
              const char *s;
              s = TOKEN_STR (1);
              if        (tokens_token_and_keyword_equiv ("DX", s)) {
                dir = W_DX;
              } else if (tokens_token_and_keyword_equiv ("DY", s)) {
                dir = W_DY;
              } else if (tokens_token_and_keyword_equiv ("DZ", s)) {
                dir = W_DZ;
              }
            } else {
              dir = TOKEN_INT (1);
            }
            amplitude = TOKEN_DBL (2);
            tf_num[0] = 
            tf_num[1] = 
            tf_num[2] = TOKEN_INT (3);
          } else if (kind == W_NODE_GROUP_LOAD_BODY) {
            const char *s;
            s = TOKEN_STR (2);
            DUPLICATE_STRING (accel[0], s);
            s = TOKEN_STR (3);
            DUPLICATE_STRING (accel[1], s);
            s = TOKEN_STR (4);
            DUPLICATE_STRING (accel[2], s);
            tf_num[0] = tf_num[1] = tf_num[2] = TOKEN_INT (5);
            if (TOKENS_TOTAL () == 7) {
              tf_num[1] = TOKEN_INT (6);
              tf_num[2] = TOKEN_INT (7);
            }
          } else if (kind == W_NODE_GROUP_LOAD_TRAC) {
            if (TOKEN_IS_STRING (2)) {
              const char *s;
              s = TOKEN_STR (2);
              if        (tokens_token_and_keyword_equiv ("DX", s)) {
                dir = W_DX;
              } else if (tokens_token_and_keyword_equiv ("DY", s)) {
                dir = W_DY;
              } else if (tokens_token_and_keyword_equiv ("DZ", s)) {
                dir = W_DZ;
              }
            } else {
              dir = TOKEN_INT (2);
            }
            amplitude = TOKEN_DBL (3);
            tf_num[0] = 
            tf_num[1] = 
            tf_num[2] = TOKEN_INT (4);
          }
          { /* are time functions defined? */
            int m;
            FOR (m, 3) {
              if (WD_get_time_func_by_idx (d, tf_num[m]) == NULL) {
                sprintf (W_buf (), "Time function #%d not found", (int) tf_num[m]);
                REPORT_FATAL_ERROR (d->errh, W_buf ());
              }
            }
          }
            
          if ((input_line = W_get_input_line(d)) == NULL)
            BAD_INPUT();
          TOKENS_PARSE (input_line);

          x.x = TOKEN_DBL (1);
          x.y = TOKEN_DBL (2);
          x.z = TOKEN_DBL (3);
          G3D_INIT_BOX (box, x);
          x.x = TOKEN_DBL (4);
          x.y = TOKEN_DBL (5);
          x.z = TOKEN_DBL (6);
          G3D_UPDT_BOX (box, x);

          TOKENS_SET_BY_KEYWORD ("INFLATE",
                                 FALSE,
                                 {
                                   inflate = TOKEN_DBL(_kpos_+1);
                                 },
                                 {
                                   inflate = 0;
                                 });
          G3D_INFLATE_BOX (box, inflate);
          
          l = new_load ();
          l->type = NODE_GROUP_LOAD;
          l->domain = d;
          l->data.node_group.kind = kind;
          switch (l->data.node_group.kind) {
          case W_NODE_GROUP_LOAD_FORCE:
            l->data.node_group.spec.force.direction  = dir;
            l->data.node_group.spec.force.amplitude  = amplitude;
            l->data.node_group.spec.force.time_func_num = tf_num[0];
            break;
          case W_NODE_GROUP_LOAD_TRAC:
            l->data.node_group.spec.trac.direction  = dir;
            l->data.node_group.spec.trac.amplitude  = amplitude;
            l->data.node_group.spec.trac.time_func_num = tf_num[0];
            break;
          case W_NODE_GROUP_LOAD_BODY:
            l->data.node_group.spec.body.accel[0]  = accel[0];
            l->data.node_group.spec.body.accel[1]  = accel[1];
            l->data.node_group.spec.body.accel[2]  = accel[2];
            {
              int m;
              FOR (m, 3) 
                l->data.node_group.spec.body.time_func_num[m] = tf_num[m];
            }
            break;
          }
          l->data.node_group.box                   = box;
          make_node_group_list (l, d);
          WD_add_load (d, l); 
          if (l->data.node_group.kind == W_NODE_GROUP_LOAD_TRAC)
            convert_to_nodal_loads (l, d);
        }
        break;
      default: 
        return FALSE;
      }
    }
  }
  return FALSE;
}

static void
eval_bodyl (W_node_group_load_t *ngl, W_node_t *n, W_vect_t *ampl)
{
  int stat;
  double f;
  W_loc_t x, X, u;

  /* current coordinates */
  WN_get_ref_coords (n, &X.x, &X.y, &X.z);
  WN_get_displ (n, &u.x, &u.y, &u.z);
  VECMAC_3_VEC (x, =, X, +, u);
  expeval_set_value ("x", x.x);
  expeval_set_value ("y", x.y);
  expeval_set_value ("z", x.z);
  /* current velocity */
  WN_get_velocity (n, &u.x, &u.y, &u.z);
  expeval_set_value ("xdot", u.x);
  expeval_set_value ("ydot", u.y);
  expeval_set_value ("zdot", u.z);
    
  f = 0;
  if (ngl->spec.body.accel[0]) {
    stat = expeval_eval (ngl->spec.body.accel[0], &f);
    if (stat != EXPEVAL_E_OK) {
      sprintf (W_buf (), "%s: \"%s\"", expeval_err_msgs[stat], ngl->spec.body.accel[0]);
      REPORT_ERROR (NULL, W_buf ());
    } 
  }
  ampl->x = f;

  f = 0;
  if (ngl->spec.body.accel[1]) {
    stat = expeval_eval (ngl->spec.body.accel[1], &f);
    if (stat != EXPEVAL_E_OK) {
      sprintf (W_buf (), "%s: \"%s\"", expeval_err_msgs[stat], ngl->spec.body.accel[1]);
      REPORT_ERROR (NULL, W_buf ());
    } 
  }
  ampl->y = f;

  f = 0;
  if (ngl->spec.body.accel[2]) {
    stat = expeval_eval (ngl->spec.body.accel[2], &f);
    if (stat != EXPEVAL_E_OK) {
      sprintf (W_buf (), "%s: \"%s\"", expeval_err_msgs[stat], ngl->spec.body.accel[2]);
      REPORT_ERROR (NULL, W_buf ());
    } 
  }
  ampl->z = f;
}


#ifdef X_XFACE
static EPixel force_color = 0x0, moment_color = 0x0;

static void
modify_vect_rep (W_node_t *n,
                 W_coords_3d_t *f, int is_moment,
                 double displ_scale,
                 GraphicObj **grep)
{
  BOOLEAN suc;
  WCRec pnts[2];
  EPixel color;
  W_loc_t x, X, u;
  int np;
  
  if ((VECMAC_NORM ((*f)) == 0)) { /* zero length vector: do not draw */
    if (*grep != NULL) {        /* and if it was added to the graphic model, remove it */
      EMDeleteGraphics (ESIModel (), *grep);
      *grep = NULL;
    }
    return;
  }

  if (force_color == 0x0) {
    force_color = ColorGetPixelFromString("yellow", &suc);
    moment_color = ColorGetPixelFromString("magenta", &suc);
  }
  WN_get_ref_coords (n, &X.x, &X.y, &X.z);
  WN_get_displ (n, &u.x, &u.y, &u.z);
  VECMAC_3_VEC (x, =, X, +, displ_scale * u);
  if (is_moment) {
    color = moment_color;
  } else {
    color = force_color;
  }
  VECMAC_2_VEC (pnts[0], =, x);
  VECMAC_2_VEC (pnts[1], =, (*f));
  if (*grep == NULL) {
    EMNoDrawAddGraphicsToModel(ESIModel (), (*grep = CreateVector3D (pnts)));
    EASValsSetColor(color);
    EGWithMaskChangeAttributes(COLOR_MASK, *grep);
  }
  EGModifyGraphicsGeometry(*grep, pnts);
}


static void
modify_node_group_load_body_grep (
                                   W_load_t *l,
                                   W_domain_t *d
                                   )
{
  W_node_group_load_t *ngl = &l->data.node_group;
  W_time_func_t *tf[3];
  double tfv[3];
  W_node_group_node_t *gn;
  W_coords_3d_t F = {0, 0, 0};
  double force_scale = WOPT_OPT (d->options, force_scale);
  double displ_scale = WOPT_OPT (d->options, displ_scale);
  int i;

  FOR (i, 3) {
    tf[i]  = WD_get_time_func_by_idx(d, ngl->spec.body.time_func_num[i]);
    tfv[i] = WTF_func_val(tf[i], WD_time(d));
  }
  loop_through_list (ngl->nodes, gn, W_node_group_node_t *) {
    if (tfv[i] != 0) {
      eval_bodyl (&l->data.node_group, gn->n, &F);
      F.x *= gn->n->mass * force_scale * tfv[0];
      F.y *= gn->n->mass * force_scale * tfv[1];
      F.z *= gn->n->mass * force_scale * tfv[2];
    } 
    modify_vect_rep (gn->n, &F, FALSE, displ_scale, &gn->f_rep);
  }
}

static void
modify_node_group_load_force_grep (
                                   W_load_t *l,
                                   W_domain_t *d
                                   )
{
  W_node_group_load_t *ngl = &l->data.node_group;
  W_time_func_t *tf[3];
  double tfv[3];
  double amplitude = ngl->spec.force.amplitude;
  int dir = ngl->spec.force.direction ;
  W_node_group_node_t *gn;
  W_coords_3d_t F = {0, 0, 0};
  double displ_scale = WOPT_OPT (d->options, displ_scale);
  double force_scale = WOPT_OPT (d->options, force_scale);
  int i;

  FOR (i, 3) {
    tf[i]  = WD_get_time_func_by_idx(d, ngl->spec.force.time_func_num);
    tfv[i] = WTF_func_val(tf[i], WD_time(d));
  }
  F.x = F.y = F.z = 0;
  if      (dir == W_DX) F.x = force_scale * amplitude * tfv[0];
  else if (dir == W_DY) F.y = force_scale * amplitude * tfv[1];
  else if (dir == W_DZ) F.z = force_scale * amplitude * tfv[2];
  loop_through_list (ngl->nodes, gn, W_node_group_node_t *) {
    modify_vect_rep (gn->n, &F, FALSE, displ_scale, &gn->f_rep);
  }
#if W_ROT_DOFS
  F.x = F.y = F.z = 0;
  if      (dir == W_RX) F.x = force_scale * amplitude * tfv[0];
  else if (dir == W_RY) F.y = force_scale * amplitude * tfv[1];
  else if (dir == W_RZ) F.z = force_scale * amplitude * tfv[2];
  loop_through_list (ngl->nodes, gn, W_node_group_node_t *) {
    modify_vect_rep (gn->n, &F, TRUE, displ_scale, &gn->f_rep);
  }
#endif
}

static void
modify_node_group_load_grep (
                             W_load_t *l,
                             W_domain_t *d
                             )
{
  switch (l->data.node_group.kind) {
  case W_NODE_GROUP_LOAD_FORCE:
    modify_node_group_load_force_grep (l, d);
    break;
  case W_NODE_GROUP_LOAD_BODY:
    modify_node_group_load_body_grep (l, d);
    break;
  }
}

static void
modify_nodal_load_grep(W_load_t *l, W_domain_t *d);

static void
modify_nodal_load_grep (W_load_t *l, W_domain_t *d)
{
  W_coords_3d_t F = {0, 0, 0};
  W_coords_3d_t M = {0, 0, 0};
  W_float_t val;
  W_nodal_load_t *nl = &l->data.nodal;
  W_time_func_t *tf = WD_get_time_func_by_idx(d, nl->time_func_num);
  W_node_t *n = WD_get_node_by_idx(d, nl->node_num);
  double displ_scale = WOPT_OPT (d->options, displ_scale);
  double force_scale = WOPT_OPT (d->options, force_scale);
  
  val = nl->amplitude * WTF_func_val (tf, WD_time(d));
  if      (nl->direction == W_DX) F.x = force_scale * val;
  else if (nl->direction == W_DY) F.y = force_scale * val;
  else if (nl->direction == W_DZ) F.z = force_scale * val;
  modify_vect_rep (n, &F, FALSE, displ_scale, &nl->fg_rep);

  if      (nl->direction == W_RX) M.x = force_scale * val;
  else if (nl->direction == W_RY) M.y = force_scale * val;
  else if (nl->direction == W_RZ) M.z = force_scale * val;
  modify_vect_rep (n, &M, TRUE, displ_scale, &nl->mg_rep);
}
#endif


static void
assemble_nodal_force_or_moment (W_load_t *l, W_domain_t *d)
{
  W_coords_3d_t F = {0, 0, 0};
#if W_ROT_DOFS
  W_coords_3d_t M = {0, 0, 0};
#endif
  W_float_t val;
  W_nodal_load_t *nl = &l->data.nodal;
  W_time_func_t *tf = WD_get_time_func_by_idx(d, nl->time_func_num);
  W_node_t *n = WD_get_node_by_idx(d, nl->node_num);
  
  val = nl->amplitude * WTF_func_val(tf, WD_time(d));
  if      (nl->direction == W_DX) F.x = val;
  else if (nl->direction == W_DY) F.y = val;
  else if (nl->direction == W_DZ) F.z = val;
  if (val != 0) {
    WN_assemble_force (n, &F);
  }
#if W_ROT_DOFS
  if      (nl->direction == W_RX) M.x = val;
  else if (nl->direction == W_RY) M.y = val;
  else if (nl->direction == W_RZ) M.z = val;
  if (val != 0) {
    WN_assemble_moment (n, &M);
  }
#endif
}


static void
assemble_node_group_force_load (W_load_t *l, W_domain_t *d)
{
  W_coords_3d_t F = {0, 0, 0};
#if W_ROT_DOFS
  W_coords_3d_t M = {0, 0, 0};
#endif
  W_node_group_load_t *ngl = &l->data.node_group;
  W_time_func_t *tf = WD_get_time_func_by_idx (d, ngl->spec.force.time_func_num);
  double tfv = WTF_func_val(tf, WD_time(d));
  double amplitude = ngl->spec.force.amplitude;
  double val = amplitude * tfv;
  int dir = ngl->spec.force.direction ;
  W_node_group_node_t *gn;

  if (val != 0) {
    F.x = F.y = F.z = 0;
    if      (dir == W_DX) F.x = val;
    else if (dir == W_DY) F.y = val;
    else if (dir == W_DZ) F.z = val;
    loop_through_list (ngl->nodes, gn, W_node_group_node_t *) {
      WN_assemble_force (gn->n, &F);
    }
#if W_ROT_DOFS
    M.x = M.y = M.z = 0;
    if      (dir == W_RX) M.x = val;
    else if (dir == W_RY) M.y = val;
    else if (dir == W_RZ) M.z = val;
    loop_through_list (ngl->nodes, gn, W_node_group_node_t *) {
      WN_assemble_moment (gn->n, &M);
    }
#endif
  }
}

static void
assemble_node_group_body_load (W_load_t *l, W_domain_t *d)
{
  W_coords_3d_t F = {0, 0, 0};
  W_node_group_load_t *ngl = &l->data.node_group;
  W_time_func_t *tf[3];
  double tfv[3];
  W_node_group_node_t *gn;
  int i;

  FOR (i, 3) {
    tf[i]  = WD_get_time_func_by_idx(d, ngl->spec.body.time_func_num[i]);
    tfv[i] = WTF_func_val(tf[i], WD_time(d));
  }
  if (tfv[0] != 0 || tfv[1] != 0 || tfv[2] != 0) {
    loop_through_list (ngl->nodes, gn, W_node_group_node_t *) {
      eval_bodyl (&l->data.node_group, gn->n, &F);
      if (F.x != 0 || F.y != 0 || F.z != 0) {
        F.x *= gn->n->mass * tfv[0];
        F.y *= gn->n->mass * tfv[1];
        F.z *= gn->n->mass * tfv[2];
        WN_assemble_force (gn->n, &F);
      }
    }
  }
}

static void
assemble_node_group_load (W_load_t *l, W_domain_t *d)
{
  switch (l->data.node_group.kind) {
  case W_NODE_GROUP_LOAD_FORCE:
    assemble_node_group_force_load (l, d);
    break;
  case W_NODE_GROUP_LOAD_BODY:
    assemble_node_group_body_load (l, d);
    break;
  case W_NODE_GROUP_LOAD_TRAC:
    /* VOID traction nodal group had been converted to nodal loads */
    break;
  }
}

/*
   Assemble current value of external load.
 */
void
WL_assemble_load (W_load_t *l, /* load */
                  W_domain_t *d /* domain */
                  )
{
  switch (l->type) {
#ifdef X_XFACE
  case MOUSE_ANIM_LOAD:
    {
      W_nodal_load_t *nl = &l->data.nodal;

      if (WOPT_OPT (d->options, X_xface)) 
        assemble_mouse_anim_load(d, WD_get_node_by_idx(d, nl->node_num));
    }
    break;
#endif
  case NODAL_LOAD:
    assemble_nodal_force_or_moment (l, d);
    break;
  case NODE_GROUP_LOAD:
    assemble_node_group_load (l, d);
    break;
  case ELEM_LOAD:
    break;
  default:
    /* VOID */
    break;
  }
}



static W_float_t
incr_ext_nodal_load_work (W_nodal_load_t *nl, W_domain_t *d)
{
  W_float_t prev_val = 0, val = 0, comb_val, dew = 0;
  W_time_func_t *tf = WD_get_time_func_by_idx(d, nl->time_func_num);
  W_coords_3d_t incr_disp, incr_rot;
  W_node_t *n = WD_get_node_by_idx(d, nl->node_num);
  
  val      = nl->amplitude * WTF_func_val(tf, WD_time(d));
  prev_val = nl->amplitude * WTF_func_val(tf, WD_time(d) - WD_time_step(d));
  comb_val = (val + prev_val) / 2;

  WN_get_disp_increment (n, &incr_disp, &incr_rot);
  if      (nl->direction == W_DX) dew += comb_val * incr_disp.x;
  else if (nl->direction == W_DY) dew += comb_val * incr_disp.y;
  else if (nl->direction == W_DZ) dew += comb_val * incr_disp.z;
  if      (nl->direction == W_RX) dew += comb_val * incr_rot.x;
  else if (nl->direction == W_RY) dew += comb_val * incr_rot.y;
  else if (nl->direction == W_RZ) dew += comb_val * incr_rot.z;

  return dew;
}


static W_float_t
incr_ext_load_force_node_group (W_node_group_load_t *ngl, W_domain_t *d)
{
  W_float_t dew = 0;
  W_time_func_t *tf[3];
  W_coords_3d_t incr_disp, incr_rot;
  double tfv[3];
  double amplitude = ngl->spec.force.amplitude;
  int dir = ngl->spec.force.direction ;
  W_node_group_node_t *gn;
  int i;

  FOR (i, 3) {
    tf[i]  = WD_get_time_func_by_idx(d, ngl->spec.force.time_func_num);
    tfv[i] = WTF_func_val(tf[i], WD_time(d));
  }

  loop_through_list (ngl->nodes, gn, W_node_group_node_t *) {
    WN_get_disp_increment (gn->n, &incr_disp, &incr_rot);
    if      (dir == W_DX) dew += amplitude * tfv[0] * incr_disp.x;
    else if (dir == W_DY) dew += amplitude * tfv[1] * incr_disp.y;
    else if (dir == W_DZ) dew += amplitude * tfv[2] * incr_disp.z;
    if      (dir == W_RX) dew += amplitude * tfv[0] * incr_rot.x;
    else if (dir == W_RY) dew += amplitude * tfv[1] * incr_rot.y;
    else if (dir == W_RZ) dew += amplitude * tfv[2] * incr_rot.z;
  }

  return dew;
}

static W_float_t
incr_ext_load_body_node_group (W_node_group_load_t *ngl, W_domain_t *d)
{
  W_float_t dew = 0;
  W_coords_3d_t F = {0, 0, 0};
  W_coords_3d_t incr_disp, incr_rot;
  W_time_func_t *tf[3];
  double tfv[3];
  W_node_group_node_t *gn;
  int i;
  
  FOR (i, 3) {
    tf[i]  = WD_get_time_func_by_idx(d, ngl->spec.body.time_func_num[i]);
    tfv[i] = WTF_func_val(tf[i], WD_time(d));
  }
  if (tfv[0] != 0 || tfv[1] != 0 || tfv[2] != 0) {
    loop_through_list (ngl->nodes, gn, W_node_group_node_t *) {
      WN_get_disp_increment (gn->n, &incr_disp, &incr_rot);
      eval_bodyl (ngl, gn->n, &F);
      if (F.x != 0 || F.y != 0 || F.z != 0) {
        F.x *= gn->n->mass * tfv[0];
        F.y *= gn->n->mass * tfv[1];
        F.z *= gn->n->mass * tfv[2];
        dew += VECMAC_DOT (incr_disp, F);
      }
    }
  }

  return dew;
}

static W_float_t
incr_ext_node_group_load_work (W_node_group_load_t *ngl, W_domain_t *d)
{
  W_float_t dew = 0;

  switch (ngl->kind) {
  case W_NODE_GROUP_LOAD_FORCE:
    dew = incr_ext_load_force_node_group (ngl, d);
    break;
  case W_NODE_GROUP_LOAD_TRAC:
    dew = 0; /* traction group is converted to nodal loads */
    break;
  case W_NODE_GROUP_LOAD_BODY:
    dew = incr_ext_load_body_node_group (ngl, d);
    break;
  }
  return dew;
}

/*
   Compute increment of work of external load.
 */
W_float_t /* increment of work in the time step */
WL_incr_ext_load_work (W_load_t *l, /* load */
                       W_domain_t *d /* domain */
                       ) 
{
  switch (l->type) {
  case NODAL_LOAD:
    return incr_ext_nodal_load_work (&l->data.nodal, d);
  case NODE_GROUP_LOAD:
    return incr_ext_node_group_load_work (&l->data.node_group, d);
  case ELEM_LOAD:
    return 0; RAW
  default:
    return 0;
  }
}


#ifdef X_XFACE
/*
   Modify load graphic representation.
 */
void
WL_modify_load_grep (W_load_t *l, /* load */
                       W_domain_t *d /* domain */ ,
                     W_bool_t immed_draw
                       ) 
{
  switch (l->type) {
  case MOUSE_ANIM_LOAD:
    /* VOID */
    break;
  case NODAL_LOAD:
    modify_nodal_load_grep (l, d);
    break;
  case NODE_GROUP_LOAD:
    modify_node_group_load_grep (l, d);
    break;
  case ELEM_LOAD:
    /* RAW -- need to update */
    break;
  }
}
#endif
