#include "Wconfig.h"


#if defined(PVM_WASP_MASTER)
#include "Wxf.h"
#include "Wdomain.h"
#include "Whistry.h"
#include "Welem.h"
#include "Wtfuncs.h"
#include "Wnode.h"
#include "Wloads.h"
#include "Wmats.h"
#include "Wvset.h"
#include "Wpvm.h"
#include "EID.h"

/* RCSHEADER "$Id: master.c,v 1.1.1.1 2001/07/24 02:42:51 pkrysl Exp $" */

#define VL(E_nd, W_hat) VECT_LOOP(1, d->E_nd, W_hat)

W_float_t
WD_compute_dt(W_domain_t *d)
{
  int w;
  double dt = FLT_MAX, sent_dt;

  if (options.time_step == 0) { /* compute it */
    INITSEND();
    BCAST(WORKER_GROUP, TIME_STEP_REQUEST_TAG);
    for (w = 1; w <= NUM_OF_WORKERS(d->pvm_master); w++) {
      RECEIVE(d->pvm_master->tids[w], TIME_STEP_ANSWER_TAG);
      UNPACK(double, &sent_dt);
      dt = min(dt, sent_dt);
    }
    dt *= options.time_step_reduction;
  } else 
    dt = options.time_step;

  INITSEND();
  PACK(double, &dt);
  BCAST(WORKER_GROUP, TIME_STEP_ANSWER_TAG);

  if (options.verbose) {
    sprintf(W_buf(), "Time step = %g", (double)dt);
    W_update_log(W_buf());
  }
  return dt;
}


static void
send_interface_loads(W_domain_t *d);
static void
receive_interface_loads(W_domain_t *d);
static void
zero_interface_loads(W_domain_t *d);

void
WD_calc_eff_loads(W_domain_t *d)
{
  receive_interface_loads(d);
  send_interface_loads(d);
}


static W_interface_package_t *interface_loads = NULL;

static void
zero_interface_loads(W_domain_t *d)
{
  W_pvm_master_t *m = d->pvm_master;
  int i;
  int n = m->num_of_interface_nodes;
  
  if (interface_loads == NULL) {
    interface_loads = w_new_array(n+1, W_interface_package_t);
    CHECK_MALLOC(interface_loads);
  }
  for (i = 1; i <= n; i++) {
    interface_loads[i].forces[0] = 0; 
    interface_loads[i].forces[1] = 0; 
    interface_loads[i].forces[2] = 0; 
    interface_loads[i].forces[3] = 0; 
    interface_loads[i].forces[4] = 0; 
    interface_loads[i].forces[5] = 0;
  }
}

static void
receive_interface_loads(W_domain_t *d)
{
  unsigned long nin, in;
  int i;
  W_pvm_master_t *m = d->pvm_master;
  double forces[6];
  W_float_t *pf;
  int nrecv = 0;

  zero_interface_loads(d);

  while (nrecv++ < NUM_OF_WORKERS(m)) {
    RECEIVE(ANY_TASK, EFFLOAD_LOCAL_TAG);
    UNPACK(ulong, &nin);
    for (i = 1; i <= nin; i++) {
      UNPACK(ulong, &in);
      UNPACK_ARRAY(double, forces, 6);
      interface_loads[in].forces[0] += forces[0];
      interface_loads[in].forces[1] += forces[1];
      interface_loads[in].forces[2] += forces[2];
      interface_loads[in].forces[3] += forces[3];
      interface_loads[in].forces[4] += forces[4];
      interface_loads[in].forces[5] += forces[5];
    }
  }
}

static void
send_interface_loads(W_domain_t *d)
{
  W_pvm_master_t *m = d->pvm_master;
  unsigned long nin = m->num_of_interface_nodes;
  unsigned long i;
  
  INITSEND();
  PACK(ulong, &nin);
  for (i = 1; i <= nin; i++) {
    PACK_ARRAY(double, interface_loads[i].forces, 6);
  }
  for (i = 1; i <= NUM_OF_WORKERS(m); i++) {
    SEND(m->tids[i], EFFLOAD_GLOBAL_TAG);
  }
}

void
WD_update_config_Newmark_variant(W_domain_t *d)
{
  /* VOID -- the work is done in workers */
}

void
WD_update_config_Warburton_variant(W_domain_t *d)
{
  /* VOID -- the work is done in workers */
}
  
void
WD_solve_for(W_domain_t *d)
{
  /* VOID -- the work is done in workers */
}

#endif /* if defined(PVM_WASP_MASTER) */
