
#include "Wconfig.h"
#include "Wmat001.h"
#include "Wdomain.h"
#include "Wmats.h"
#include "Wutil.h"

/* RCSHEADER "$Id: mat001.c,v 1.1.1.1 2001/07/24 02:42:51 pkrysl Exp $" */

typedef struct mat_t {
  W_float_t    E;
  W_float_t    nu;
}              mat_t;



static W_bool_t
get_uniaxial_elast_const(W_material_t *m,
                         W_float_t *E,
                         W_float_t *nu);
static W_matstate_t 
get_mat_state(W_material_t *m);
static W_bool_t
get_uniaxial_stress(W_material_t *m,
                    W_matstate_t ms,
                    W_float_t strain,
                    W_float_t strain_rate,
                    W_float_t *stress
                    );
static W_float_t
get_uniaxial_def_energy(W_material_t *m, W_matstate_t ms,
                        W_float_t strain);


static W_material_methods_t methods = {
  get_uniaxial_elast_const,
  get_uniaxial_stress,
  get_uniaxial_def_energy,
  NULL,
  NULL,
  NULL,
  NULL,
  get_mat_state,
  NULL
};
  

/* Reads uniaxial linear elastic material data from stream */
/* as long as there is no */
/* END_OF_BLOCK tag in the input line. Creates materials. */

W_bool_t 
mat001_reader(W_domain_t *d)
{
  const char *input_line;
  long mat_num;
  double E, nu, vol_mass;
  mat_t *u;

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return TRUE;
    else {
      if (sscanf(input_line, "%ld %lf %lf %lf",
                 &mat_num, &E, &nu, &vol_mass) != 4) {
        BAD_INPUT();
      } else {
        u = w_new(mat_t);
        CHECK_MALLOC(u);
        u->E        = E; 
        u->nu       = nu;
        WD_add_material(d, mat_num, 
                        WM_new_material (MCODE_UNIAX_LE, mat_num,
                                         &methods, u, vol_mass));
      }
    }
  }
  return FALSE;
}


static W_bool_t
get_uniaxial_elast_const(W_material_t *m, W_float_t *E, W_float_t *nu)
{
  mat_t *u = (mat_t *)m->data;

  *E  = u->E;
  *nu = u->nu;
  return TRUE;
}



static int dummy_mat001_state = 0;

static W_matstate_t 
get_mat_state(W_material_t *m)
{
  return (W_matstate_t)&dummy_mat001_state; /* to avoid returning NULL */
}



static W_bool_t
get_uniaxial_stress(W_material_t *m,
                    W_matstate_t ms,
                    W_float_t strain,
                    W_float_t strain_rate,
                    W_float_t *stress
                    )
{
  mat_t *u = (mat_t *)m->data;
  
  *stress = u->E * strain;
  return TRUE;
}



static W_float_t
get_uniaxial_def_energy(W_material_t *m, W_matstate_t ms,
                        W_float_t strain)
{
  mat_t *u = (mat_t *)m->data;
  
  return 0.5 * u->E * strain * strain;
}



