
#include "Wconfig.h"
#include "Wmat006.h"
#include "Wdomain.h"
#include "Wmats.h"
#include "Wmat3d.h"

/* RCSHEADER "$Id: mat006.c,v 1.2 2002/11/08 20:34:17 cvs Exp $" */

static W_bool_t
get_uniaxial_elast_const(W_material_t *m, W_float_t *E, W_float_t *nu);


static W_material_methods_t methods = {
  get_uniaxial_elast_const,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL
};
  

/*
  Reads general triaxial material data from stream 
  as long as there is no END_OF_BLOCK tag in the input line.
  Creates materials.  The reading of the actual material data
  is delegated to mat3d_reader ().
*/
W_bool_t 
mat006_reader(W_domain_t *d)
{
  const char *input_line;
  long mat_num;
  W_mat_3d_t *mat3d;
  W_material_t *mat = NULL;

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return TRUE;
    else {
      if (sscanf(input_line, "%ld", &mat_num) != 1) {
        BAD_INPUT();
      } else {
        mat3d = mat3d_reader (d);
        mat = WM_new_material(MCODE_GENERAL_TRIAX, mat_num,
                                        &methods, mat3d,
                                        WM3D_get_volume_mass(mat3d));
        mat3d->mat = mat;
        WD_add_material(d, mat_num, mat);
      }
    }
  }
  return FALSE;
}


static W_bool_t
get_uniaxial_elast_const(W_material_t *m, W_float_t *E, W_float_t *nu)
{
  double theE, thenu;

  if (!WM3D_get_mat_const((W_mat_3d_t *)m->data, MAT_3D_YOUNG_MODULUS, &theE))
    return FALSE;
  if (!WM3D_get_mat_const((W_mat_3d_t *)m->data, MAT_3D_POISSON, &thenu))
    return FALSE;
  *E  = theE;
  *nu = thenu;
  return TRUE;
}
