
#include "Wconfig.h"
#include "Wmat007.h"
#include "Wdomain.h"
#include "Wmats.h"
#include "Wmat1d.h"

/* RCSHEADER "$Id: mat007.c,v 1.1.1.1 2001/07/24 02:42:51 pkrysl Exp $" */

static W_bool_t
get_uniaxial_elast_const(W_material_t *m, W_float_t *E, W_float_t *nu);


static W_material_methods_t methods = {
  get_uniaxial_elast_const,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL
};
  

/*
  Reads general uniaxial material data from stream 
  as long as there is no END_OF_BLOCK tag in the input line.
  Creates materials.  The reading of the actual material data
  is delegated to mat1d_reader ().
*/
W_bool_t                        /* TRUE if successful */
mat007_reader (W_domain_t *d    /* domain */
               )
{
  const char *input_line;
  long mat_num;
  W_mat_1d_t *mat1d;

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return TRUE;
    else {
      if (sscanf(input_line, "%ld", &mat_num) != 1) {
        BAD_INPUT();
      } else {
        mat1d = mat1d_reader (d);
        if (!mat1d) REPORT_FATAL_ERROR (d->errh, "Failed to create material");
        WD_add_material(d, mat_num, 
                        WM_new_material (MCODE_GENERAL_UNIAX, mat_num,
                                         &methods, mat1d,
                                         WM1D_get_volume_mass (mat1d)));
      }
    }
  }
  return FALSE;
}


static W_bool_t
get_uniaxial_elast_const (W_material_t *m, W_float_t *E, W_float_t *nu)
{
  double theE, thenu;

  if (!WM1D_get_mat_const((W_mat_1d_t *)m->data, MAT_1D_YOUNG_MODULUS, &theE))
    return FALSE;
  if (!WM1D_get_mat_const((W_mat_1d_t *)m->data, MAT_1D_POISSON, &thenu))
    return FALSE;
  *E  = theE;
  *nu = thenu;
  return TRUE;
}
