#include "Wconfig.h"
#include "Wtypes.h"
#include "Wutil.h"
#include "Wxf.h"
#include "Wmat1d.h"

/*
   Register mat1d material readers.
     In order to be able to link a general triaxial
     material (mat1d) to the WASP package, it needs
     to register a procedure which is able to
     read the material data from the input file.
     The mat1d reader is of the following structure:

"   W_mat_1d_t *
"   mat1d_reader_xxxxx  (FILE *stream)
"   {
"     ... some local variable declarations ...
"   
"     while ((input_line = W_get_input_line(stream)) != NULL) {
"       if (input_line[0] == END_OF_BLOCK)
"         return NULL;
"       else {
"         ... read the material data, set up the material by calling
"             WM1D_new_material(), and return it ...
"         return newmat1d;
"       }
"     }
"     return NULL; ... some kind of failure ...
"   }

       The mat1d reader is registered when the WM1D_initialize_readers()
       procedure is invoked during the initialization of the WASP
       package.  To register a mat1d reader, declare it and invoke
       the macro W_REGISTER_READER(m3ccode, reader) in the file
       "mat1d.readers".  This file is included in the code of
       the WM1D_initialize_readers() procedure.

       In order to be a good companion, the mat1d needs to comply
       with a certain protocol.  The protocol is carried out by calling
       functions, pointer to which are stored in the mat1d method table.
       (Note that the method table is shared by all mat1d's of the same type.)
     */
void 
WM1D_initialize_readers (void)
{
  
#define W_DCL_READER(reader) W_mat1d_reader_func reader
#define W_REGISTER_READER(m1dcode, reader) WM1D_register_reader (m1dcode, reader)
#include "mat1d.readers" 
#undef  W_DCL_READER
#undef  W_REGISTER_READER
}
