
#include "Wtypes.h"
#include "Wutil.h"
#include "Wxf.h"
#include "Wmat3d.h"
#include "meschach.h"
 
/* RCSHEADER "$Id: mat3d.c,v 1.16 2004/09/06 01:17:58 cvs Exp $" */

typedef struct reader_intermediary_t {
  W_mat3d_reader_func reader;
}              reader_intermediary_t;

static dict_dictionary reader_intermediary_table = NULL;


static W_mat3d_reader_func
find_reader (int m3dcode)
{
  reader_intermediary_t *ri = NULL;
  
  ri = dict_find_obj_with_int_name (reader_intermediary_table, m3dcode);
  if (ri == NULL) {
    sprintf (W_buf (), "Reader for mat3d type %d not registered",
             m3dcode);
    REPORT_ERROR (NULL, W_buf ());
  }
  return ri->reader;
}

static void 
initialize_readers_for_builtins (void)
{
#define W_DCL_READER(reader) W_mat3d_reader_func reader
#define W_REGISTER_READER(m3dcode, reader) WM3D_register_reader (m3dcode, reader)
#include "Wmat3difr.h"
#include "Wmat3drel.h"
#include "Wmat3dstvk.h"
#include "Wmat3dpf.h"
#include "Wmat3dhalq.h"
#include "Wmat3dneohk.h"
#include "Wmat3dunrot.h"
#include "Wmat3dssj2ep.h"
#include "Wmat3durd.h"
#include "Wmat3dred2d.h"
#include "Wmat3dis.h"
#include "Wmat3dcryspl.h"
#include "Wmat3dcrysplfcc.h"
#include "Wmat3dcrysplfccps.h"
#include "Wmat3dckink.h"
#include "Wmat3dckinks.h"
#include "Wmat3dsmass1.h"
  /*#include "Wmat3dbond.h"*/
#include "Wmat3dviscel.h"
#include "Wmat3dfung.h"
#include "Wmat3dqfung.h"
#include "Wmat3dmodstvk.h"
   W_REGISTER_READER (MAT_3D_STV_K, read_mat_3d_stv_k);
   W_REGISTER_READER (MAT_3D_PINSKY_ETAL_EL, read_mat_3d_rate_el);
   W_REGISTER_READER (MAT_3D_RATE_INST_FINAL_ROT, read_mat_3d_ifr);
   W_REGISTER_READER (MAT_3D_PERF_FLUID, read_mat_3d_perf_fluid);
   W_REGISTER_READER (MAT_3D_RATE_HALLQUIST, read_mat_3d_halq);
   W_REGISTER_READER (MAT_3D_NEOHOOKEAN, read_mat_3d_neohk);
   W_REGISTER_READER (MAT_3D_UNROTATED, read_mat_3d_unrot);
   W_REGISTER_READER (MAT_3D_SSJ2EP, read_mat_3d_ssj2ep);
   W_REGISTER_READER (MAT_3D_URD, read_mat_3d_urd);
   W_REGISTER_READER (MAT_3D_RED2D, read_mat_3d_red2d);
   W_REGISTER_READER (MAT_3D_IS, read_mat_3d_is);
   W_REGISTER_READER (MAT_3D_CRYSPL, read_mat_3d_cryspl);
   W_REGISTER_READER (MAT_3D_CRYSPL_FCC, read_mat_3d_cryspl_fcc);
   W_REGISTER_READER (MAT_3D_CRYSPL_FCC_PS, read_mat_3d_cryspl_fcc_ps);
   W_REGISTER_READER (MAT_3D_CKINK, read_mat_3d_ckink);
   W_REGISTER_READER (MAT_3D_CKINKS, read_mat_3d_ckinks);
   W_REGISTER_READER (MAT_3D_SMASS1, read_mat_3d_smass1);
   /*   W_REGISTER_READER (MAT_3D_BOND, read_mat_3d_bond); */
   W_REGISTER_READER (MAT_3D_VISCEL, read_mat_3d_viscel);
   W_REGISTER_READER (MAT_3D_FUNG, read_mat_3d_fung);
   W_REGISTER_READER (MAT_3D_QFUNG, read_mat_3d_qfung);
   W_REGISTER_READER (MAT_3D_MOD_STV_K, read_mat_3d_mod_stvk);
#undef  W_DCL_READER
#undef  W_REGISTER_READER
}

/*
  Top-level 3d material reader.
   The generic mat3d reader.  The reader for the particular
   material in the input is found from among the registered
   mat3d readers.  If no reader for the material type was
   registered, NULL is returned; otherwise
 */
W_mat_3d_t * /* If no reader for the material type was
   registered, NULL is returned; otherwise the result of the reader to which
   the reading was delegated is returned. */
mat3d_reader (W_domain_t *d     /* domain */
              )
{
  static int readers_initialized = FALSE;
  const char *input_line;
  long m3dcode;
  W_mat3d_reader_func reader = NULL;
  
  if (!readers_initialized) {
    initialize_readers_for_builtins ();
    readers_initialized = TRUE;
  }

  input_line = W_get_input_line(d);
  TOKENS_PARSE(input_line);
  m3dcode = TOKEN_INT(1);

  reader = find_reader (m3dcode);
  if (reader != NULL) {
    return (*reader) (d);
  } else {
    return NULL;
  }
}


/*
   Register a mat3d reader.
   To be able to use specific implementation
   of the general triaxial material (mat3d),
   its code needs to be linked to the
   executable, and its reader needs to be registered, so that the generic
   mat3d reader can look up the type-specific reader.
   */
void
WM3D_register_reader (int m3dcode /* unique integer among all elements */,
                      W_mat3d_reader_func reader /* the reader function */
                      )
{
  reader_intermediary_t *ri;
  
  if (reader_intermediary_table == NULL) {
    reader_intermediary_table = dict_create_dictionary ();
  }
  ri = (reader_intermediary_t *) dict_find_obj_with_int_name (reader_intermediary_table, m3dcode);
  if (ri != NULL) {
    sprintf (W_buf (), "Reader for element type %d already registered",
             m3dcode);
    REPORT_ERROR (NULL, W_buf ());
  }
  ri = w_new (reader_intermediary_t); CHECK_MALLOC (ri);
  ri->reader = reader;
  dict_add_obj_int_name (reader_intermediary_table, ri, m3dcode);
}
  
/*
   Create a new 3-D material instance.
   The type specific information is supplied in "methods" and "data".
 */
W_mat_3d_t * /* material */
WM3D_new_material(
                  W_mat_3d_methods_t *methods, /* table of type-specific
                                                  functions
                                                */
                  void *data, /* mat3d type-specific data */
                  double vol_mass /* volume mass density */
                  )
{
  W_mat_3d_t *m;

  m = w_new(W_mat_3d_t); CHECK_MALLOC(m);

  m->vol_mass = vol_mass;
  m->methods  = methods;
  m->data     = data;
#ifdef X_XFACE
  m->color    = 0x0; 
  m->fringes  = NULL;
#endif
  return m;
}


/*
   Get volume mass density.
 */
double /* volume mass density */
WM3D_get_volume_mass (W_mat_3d_t *m/* mat3d */)
{
  return (m->vol_mass);
}


/*
   Get material type code.
 */
int /* code or -131313 signalling an invalid material or
       some kind of failure */
WM3D_get_mcode (W_mat_3d_t *m/* mat3d */)
{
  if (m->methods->get_mcode) {
    return (*m->methods->get_mcode) (m);
  } else {
    return -131313;
  }
}
 


/*
   Get material constant.
 */
W_bool_t 
WM3D_get_mat_const (W_mat_3d_t *m/* mat3d */,
                    W_mat_3d_mat_const_enum type, /* type of constant */
                    double *c /* value */
                    )
{
  if (m->methods->get_mat_const) {
    return (*m->methods->get_mat_const)(m, type, c);
  } else {
    *c = 0;
    return FALSE; 
  }
} 


/*
   Update stress in a material state and return it.
 */
W_bool_t /* TRUE if successful; FALSE otherwise, in which case an error
            message is generated
          */
WM3D_update_stress (W_mat_3d_t                  *m/* mat3d */,
                    W_mat_3d_state_p             ms, /* material state */
                    W_mat_state_update_form_enum update_form,
                    W_mat_3d_ip_ctx_t           *ip_ctx,/* integration point context; used to generate kinematic quantities which the material state
                                                           might need in order to compute the requested stress */
                    W_mat_3d_stress_type_enum    stress_type,/* type of stress tensor */
                    double                       stress[3][3],/* stress tensor */
                    W_mat_3d_stress_type_enum   *out_stress_type/* type of actually returned stress tensor; ideally, output type == requested type, but
                                                                   sometimes the material state fails to compute the requested tensor for lack of information
                                                                   */
                    )
{
  if (m->methods->update_stress) {
    return (*m->methods->update_stress)(m, ms, update_form, ip_ctx,
                                        stress_type, stress, out_stress_type);
  } else {
    return FALSE;
  }
}


/*
   Get material state variable.
 */
W_bool_t /* TRUE if successful; FALSE otherwise */
WM3D_get_mat_state_var (W_mat_3d_t           *m/* mat3d */,
                        W_mat_3d_state_p             ms, /* material state */
                        W_mat_3d_ip_ctx_t           *ip_ctx, /* integration point context; used to generate kinematic quantities which the material state
might need in order to compute the requested variable
                                                              */
                        W_mat_3d_mat_state_var_enum  var_type, /* variable type */
                        double                      *var /* variable value is returned in this argument  */
                        )
{
  if (m->methods->get_mat_state_var) {
    return (*m->methods->get_mat_state_var)(m, ms, ip_ctx,
                                            var_type, var);
  } else {
    return FALSE;
  }
}


/*
   Get material state variable.
 */
W_bool_t /* TRUE if successful; FALSE otherwise */
WM3D_get_mat_state_named_var (W_mat_3d_t *m/* mat3d */,
                              W_mat_3d_state_p ms, /* material state */
                              W_mat_3d_ip_ctx_t *ip_ctx, /* integration point context; used to generate kinematic quantities which the material state
                                                                      might need in order to compute the requested variable
                                                                   */
                              char *name, /* variable name */
                              double *var /* variable value is returned in this argument  */
                        )
{
  if (m->methods->get_mat_state_named_var) {
    return (*m->methods->get_mat_state_named_var)(m, ms, ip_ctx, name, var);
  } else {
    return FALSE;
  }
}


/*
   Get a new material state at a given point.
   */
W_mat_3d_state_p /* material state */
WM3D_get_mat_state (W_mat_3d_t *m/* mat3d */,
                    W_loc_t *at_point /* point */
                    )
{
  if (m->methods->get_mat_state)
    return m->methods->get_mat_state(m, at_point);
  else
    return NULL;
}


/*
   Interpolate material state at a given point from data at nearby
   points at which the material state is known.
   */
W_mat_3d_state_p /* generated material state */
WM3D_interpolate_mat_state (W_mat_3d_t *m/* mat3d */,
                            W_loc_t *at_point, /* point */
                            int npkgs, /* number of material state packages supplied below */
                            WM3D_mat_state_pkg_t *pkgs /* material state packages */
                            )
{
  if (m->methods->interpolate_mat_state)
    return m->methods->interpolate_mat_state(m, at_point, npkgs, pkgs);
  else
    return NULL;
}

/*
  Compute material stiffness tensor.
  Compute material stiffness tensor in the reference configuration.
 */
W_bool_t 
WM3D_get_mat_stiff (W_mat_3d_t *m/* mat3d */,
                    W_mat_3d_state_p ms,/* material state */
                    W_mat_3d_ip_ctx_t *ip_ctx,/* integration point context;
                                                           used to generate kinematic
                                                           quantities which the material state
                                                           might need in order to compute
                                                           the requested stress */
                    W_mat_3d_stiff_enum stiff_type, /* type of material stiffness matrix */
                    double C[6][6] /* material stiffness tensor in matrix form */
                    )
{
  if (m->methods->get_mat_stiff) {
    return (*m->methods->get_mat_stiff) (m, ms, ip_ctx, stiff_type, C);
  } else {
    return FALSE;
  }
}

W_bool_t
WM3D_get_mat_stiff_9x9 (W_mat_3d_t *m/* mat3d */,
                    W_mat_3d_state_p ms,/* material state */
                    W_mat_3d_ip_ctx_t *ip_ctx,/* integration point context;
                                                           used to generate kinematic
                                                           quantities which the material state
                                                           might need in order to compute
                                                           the requested stress */
                    W_mat_3d_stiff_enum stiff_type, /* type of material stiffness matrix */
		    double C[9][9]
                    )
{
  if (m->methods->get_mat_stiff_9x9) {
    return (*m->methods->get_mat_stiff_9x9) (m, ms, ip_ctx, stiff_type, C);
  } else {
    return FALSE;
  }
}

/*
   Convert stress tensor by using the deformation gradient.
   The stress tensor can be converted to/from any of these
   types: Cauchy, 1st and 2nd Piola-Kirchhoff.
   */
W_bool_t
WM3D_convert_stress (double F[3][3], /* deformation gradient */
                     W_mat_3d_stress_type_enum    in_stress_type, /* input stress tensor type */
                     double                       in_stress[3][3], /* input stress tensor */
                     W_mat_3d_stress_type_enum    out_stress_type, /* output stress tensor type */
                     double                       out_stress[3][3] /* output stress tensor */
                     )
{
  double F_cof[3][3], inv_F[3][3], P[3][3];
  double J;

  if (in_stress_type == out_stress_type) {
    MAT3DUTIL_COPY_3X3(out_stress, in_stress);
    return TRUE;
  }

  switch (in_stress_type) {
  case MAT_3D_CAUCHY:
    switch (out_stress_type) {
    case MAT_3D_1ST_PK:
      MAT3DUTIL_COFACTOR_3X3(F, F_cof);
      MAT3DUTIL_MULT_ABT_3X3(out_stress, in_stress, F_cof);
      break;
    case MAT_3D_2ND_PK:
      J = MAT3DUTIL_DET_3X3(F);
      if (!W_mat3dutil_invert_3x3(F, inv_F))
        REPORT_FATAL_ERROR(NULL, "Failed inversion of F");
      MAT3DUTIL_MULT_ABT_3X3(P, in_stress, inv_F);
      MAT3DUTIL_MULT_AB_3X3(out_stress, inv_F, P);
      MAT3DUTIL_SCALE_3X3(out_stress, J);
      break;
    default:
      REPORT_FATAL_ERROR(NULL, "Unknown output stress type");
      break;
    }
    break;
  default:
    REPORT_FATAL_ERROR(NULL, "Unknown input stress type");
    break;
  }
    
  return TRUE;
}


/*
   Compute the principal stresses.
   Compute the principal stresses and principal stress directions
   for the given stress tensor.

   The stress tensor must be symmetric; otherwise the results are 
   unpredictable.  NO CHECK is performed.
   */
W_bool_t
WM3D_principal_stresses (double stress[3][3], /* symmetric stress tensor */
                         double principal_stress[3], /* principal stresses */
                         W_vect_t *n0, W_vect_t *n1, W_vect_t *n2 /* principal directions */
                         )
{
  double eig_vecs[3][3], eig_vals[3];
  W_bool_t result = TRUE;
  result = W_eigendecomposition (stress, eig_vecs, eig_vals);
  if (result) {
    principal_stress[0] = eig_vals[0];
    principal_stress[1] = eig_vals[1];
    principal_stress[2] = eig_vals[2];
    n0->x = eig_vecs[0][0];
    n0->y = eig_vecs[0][1];
    n0->z = eig_vecs[0][2];
    n1->x = eig_vecs[1][0];
    n1->y = eig_vecs[1][1];
    n1->z = eig_vecs[1][2];
    n2->x = eig_vecs[2][0];
    n2->y = eig_vecs[2][1];
    n2->z = eig_vecs[2][2];
  }
  return result;
}

void
WM3D_update_history (W_mat_3d_t *m,
                     W_mat_3d_state_p ms,
                     W_mat_3d_ip_ctx_t *ip_ctx,
                     W_solution_history_t *h, 
                     W_index_t pntidx,
                     W_float_t time
                     )
{
  if (m->methods->update_history) {
    (*m->methods->update_history) (m, ms, ip_ctx, h, pntidx, time);
  }
}
  
