#include "Wconfig.h"
#include "Wtypes.h"
#include "Wutil.h"
#include "Wxf.h"
#include "Wmat3d.h"

/*
   Register mat3d material readers.
     In order to be able to link a general triaxial
     material (mat3d) to the WASP package, it needs
     to register a procedure which is able to
     read the material data from the input file.
     The mat3d reader is of the following structure:

"   W_mat_3d_t *
"   mat3d_reader_xxxxx  (FILE *stream)
"   {
"     ... some local variable declarations ...
"   
"     while ((input_line = W_get_input_line(stream)) != NULL) {
"       if (input_line[0] == END_OF_BLOCK)
"         return NULL;
"       else {
"         ... read the material data, set up the material by calling
"             WM3D_new_material(), and return it ...
"         return newmat3d;
"       }
"     }
"     return NULL; ... some kind of failure ...
"   }

       The mat3d reader is registered when the WM3D_initialize_readers()
       procedure is invoked during the initialization of the WASP
       package.  To register a mat3d reader, declare it and invoke
       the macro W_REGISTER_READER(m3ccode, reader) in the file
       "mat3d.readers".  This file is included in the code of
       the WM3D_initialize_readers() procedure.

       In order to be a good companion, the mat3d needs to comply
       with a certain protocol.  The protocol is carried out by calling
       functions, pointer to which are stored in the mat3d method table.
       (Note that the method table is shared by all mat3d's of the same type.)
     */
void 
WM3D_initialize_readers (void)
{
  
#define W_DCL_READER(reader) W_mat3d_reader_func reader
#define W_REGISTER_READER(m3dcode, reader) WM3D_register_reader (m3dcode, reader)
#include "mat3d.readers" 
#undef  W_DCL_READER
#undef  W_REGISTER_READER
}
