 
#include "Wtypes.h"
#include "Wutil.h"
#include "Wxf.h"
#include "Wmat3d.h"
#include "Wmat3dbond.h"
#include "Wutil.h"

/* RCSHEADER "$Id: mat3dbond.c,v 1.1 2004/04/02 19:39:59 cvs Exp $" */

static int
get_mcode (W_mat_3d_t *m)
{
  return MAT_3D_BOND;
}


typedef struct bond_mat_t {
  double E, nu;
  double cutoff;
  double lambda, G, bulk_mod;
  double dil_wave_speed_3d;
  double dil_wave_speed_pstrain;
  double dil_wave_speed_pstress;
  double shear_wave_speed;
  double rayleigh_pstrain_wave_speed;
  double rayleigh_pstress_wave_speed;
  double tensile_strength;
  double dyn_fracture_toughness;
}              bond_mat_t;


typedef struct matstate_t {
  double   deform_energy_density;
}              matstate_t;


static W_bool_t
get_mat_const (W_mat_3d_t *m, W_mat_3d_mat_const_enum type, double *c)
{
  bond_mat_t *u = (bond_mat_t *)m->data;

  switch (type) {
  case MAT_3D_YOUNG_MODULUS:
    *c = u->E;
    break;
  case MAT_3D_POISSON:
    *c = u->nu;
    break;
  case MAT_3D_SHEAR_MODULUS:
    *c = u->G;
    break;
  case MAT_3D_BULK_MODULUS:
    *c = u->bulk_mod;
    break;
  case MAT_3D_TENSILE_STRENGTH:
    *c = u->tensile_strength; 
    break;
  case MAT_3D_DYN_FRACTURE_TOUGHNESS:
    *c = u->dyn_fracture_toughness; 
    break;
  case MAT_3D_DIL_WAVE_SPEED_3D:
    *c = u->dil_wave_speed_3d;
    break;
  case MAT_3D_DIL_WAVE_SPEED_PSTRAIN:
    *c = u->dil_wave_speed_pstrain;
    break;
  case MAT_3D_DIL_WAVE_SPEED_PSTRESS:
    *c = u->dil_wave_speed_pstress;
    break;
  case MAT_3D_SHEAR_WAVE_SPEED:
    *c = u->shear_wave_speed;
    break;
  case MAT_3D_RAYLEIGH_PSTRAIN_WAVE_SPEED:
    *c = u->rayleigh_pstrain_wave_speed;
    break;
  case MAT_3D_RAYLEIGH_PSTRESS_WAVE_SPEED:
    *c = u->rayleigh_pstress_wave_speed;
    break;
  default:
    *c = 0;
    return FALSE;
  }
  return TRUE;
}

  
#if 0
#define DUMP(S) {                                               \
  printf("%g\t %g\t %g\n", S[0][0], S[0][1], S[0][2]);          \
  printf("%g\t %g\t %g\n", S[1][0], S[1][1], S[1][2]);          \
  printf("%g\t %g\t %g\n", S[2][0], S[2][1], S[2][2]); }

#define HAVENAN(S) (   isnan(S[0][0]) || isnan(S[0][1]) || isnan(S[0][2])  \
                    || isnan(S[1][0]) || isnan(S[1][1]) || isnan(S[1][2])  \
                    || isnan(S[2][0]) || isnan(S[2][1]) || isnan(S[2][2]))
#endif

#define DBGP_VCT(tit,what, n1)  {                                        \
    int I_I_;                                                           \
    fprintf (stderr, "%s = [ ", tit);                                   \
    FOR (I_I_, n1) { fprintf (stderr, "%g ", what[I_I_]); }	\
    fprintf (stderr, "]\n");                                            \
  }

#define DBGP_T(tit,what, n1, n2)  {                                        \
    int I_I_, J_J_;                                     \
    fprintf (stderr, "%s = [ ", tit);                    \
    FOR (I_I_, n1) {                        \
       FOR (J_J_, n2) {                     \
         fprintf (stderr, "%g ", what[I_I_][J_J_]);     \
       }                                                \
       fprintf (stderr, ";\n");                          \
    }                                                   \
    fprintf (stderr, "]\n");                             \
  }

/* Function updates the stress state, and returns the stress (with proper */
/* conversion into the requested type).  Note that the function get_stress() */
/* is identical with this function since no stress state is saved for this */
/* type of elastic material. */
static W_bool_t 
update_stress (W_mat_3d_t                  *m,
               W_mat_3d_state_p             ms,
               W_mat_state_update_form_enum update_form,
               W_mat_3d_ip_ctx_t           *ip_ctx,
               W_mat_3d_stress_type_enum    stress_type,
               double                       stress[3][3],
               W_mat_3d_stress_type_enum   *out_stress_type
               )
{
  double S[3][3], P[3][3], sigma[3][3], E[3][3], F[3][3];
  bond_mat_t *u = (bond_mat_t *)m->data;
  double lambda = u->lambda, twomu = 2 * u->G;
  double trE;
  double invdetF;
  int i, j, k;
  W_bool_t result = TRUE;
  matstate_t *matst = (matstate_t *)ms;
  matstate_t input_matst;

  double S_eigvct[3][3], S_eigval[3];
  int c1, c2; /* counters */

  if (!ip_ctx->get_kinem_var) return FALSE;

  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_GREEN_LAGRANGE, E))
    return FALSE;
#if 0
  {
    double F[3][3];
  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F))
    return FALSE;
    { int i, j;
    fprintf (stderr, "F=[\n");
    FOR (i, 3) {
      FOR (j, 3) {
        fprintf (stderr, "%g ", F[i][j]);
      }
    }
    fprintf (stderr, "]\n");
    }
  }
#endif
#if 0
  {
    double C[3][3];
  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_CAUCHY_GREEN, C))
    return FALSE;
    { int i, j;
    fprintf (stderr, "Ce=[\n");
    FOR (i, 3) {
      FOR (j, 3) {
        fprintf (stderr, "%g ", C[i][j]);
      }
    }
    fprintf (stderr, "]\n");
    }
  }
#endif

  /* save the material state if the update is temporary */
  if (update_form == W_MAT_STATE_UPDATE_TEMP) input_matst = *matst;
  
  trE = E[0][0] + E[1][1] + E[2][2];
  S[0][0] = lambda * trE + twomu * E[0][0];
  S[0][1] =                twomu * E[0][1];
  S[0][2] =                twomu * E[0][2];
  S[1][0] =                twomu * E[1][0];
  S[1][1] = lambda * trE + twomu * E[1][1];
  S[1][2] =                twomu * E[1][2];
  S[2][0] =                twomu * E[2][0];
  S[2][1] =                twomu * E[2][1];
  S[2][2] = lambda * trE + twomu * E[2][2];

  /* calculate the principal stresses */
  W_eigendecomposition(S, S_eigvct, S_eigval);

  FOR(i, 3) {
    if (S_eigval[i] >= u->cutoff) 
    {
     /** DBGP_T("upd:S",S, 3,3);
      DBGP_VCT("upd:S_eigval",S_eigval, 3); ***/
      for (c1 = 0; c1 < 3; c1++) {
          for (c2 = 0; c2 < 3; c2++) {
              S[c1][c2] = 0;
          }
      }
    }
  }

#if 0
    { int i, j;
    fprintf (stderr, "E=[\n");
    FOR (i, 3) {
      FOR (j, 3) {
        fprintf (stderr, "%g ", E[i][j]);
      }
    }
    fprintf (stderr, "]\n");
    }
    { int i, j;
    fprintf (stderr, "S=[\n");
    FOR (i, 3) {
      FOR (j, 3) {
        fprintf (stderr, "%g ", S[i][j]);
      }
    }
    fprintf (stderr, "]\n");
    }
#endif

  matst->deform_energy_density = 0.5 * (MAT3DUTIL_DBL_PROD_3X3(S, E));

  *out_stress_type = stress_type;
  switch (stress_type) {
  case MAT_3D_CAUCHY:
    if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F)) {
      result = FALSE; goto exit;
    }
    FOR(i, 3) {
      FOR(j, 3) {
        P[i][j] = 0; /* /F/ * /S/ */ 
        FOR(k, 3) P[i][j] += F[i][k] * S[k][j];
      }
    }
    invdetF = 1 / MAT3DUTIL_DET_3X3(F);
    FOR(i, 3) {
      FOR(j, 3) {
        sigma[i][j] = 0; /* 1/detF /P/ * /F/^T */ 
        FOR(k, 3) sigma[i][j] += invdetF * P[i][k] * F[j][k];
      }
    }
    MAT3DUTIL_COPY_3X3(stress, sigma);
    break;
  case MAT_3D_1ST_PK:
    if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F)) {
      result = FALSE; goto exit;
    }
    FOR(i, 3) {
      FOR(j, 3) {
        P[i][j] = 0; /* /F/ * /S/ */ 
        FOR(k, 3) P[i][j] += F[i][k] * S[k][j];
      }
    }
    MAT3DUTIL_COPY_3X3(stress, P);
    break;
  case MAT_3D_2ND_PK:
    MAT3DUTIL_COPY_3X3(stress, S);
    break;
  default:
    MAT3DUTIL_COPY_3X3(stress, S);
    *out_stress_type = MAT_3D_2ND_PK;
    break;
  }
    
 exit:
  if (update_form == W_MAT_STATE_UPDATE_TEMP) *matst = input_matst;
  return result;
}

static W_bool_t 
get_stress (W_mat_3d_t                  *m,
            W_mat_3d_state_p             ms,
            W_mat_3d_ip_ctx_t           *ip_ctx,
            W_mat_3d_stress_type_enum    stress_type,
            double                       stress[3][3],
            W_mat_3d_stress_type_enum   *out_stress_type
            )
{
  double S[3][3], P[3][3], sigma[3][3], E[3][3], F[3][3];
  bond_mat_t *u = (bond_mat_t *)m->data;
  double lambda = u->lambda, twomu = 2 * u->G;
  double trE;
  double invdetF;
  int i, j, k;
  W_bool_t result = TRUE;
  matstate_t *matst = (matstate_t *)ms;

  double S_eigvct[3][3], S_eigval[3];
  int c1, c2; /* counters */

  if (!ip_ctx->get_kinem_var) return FALSE;

  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_GREEN_LAGRANGE, E))
    return FALSE;

  trE = E[0][0] + E[1][1] + E[2][2];
  S[0][0] = lambda * trE + twomu * E[0][0];
  S[0][1] =                twomu * E[0][1];
  S[0][2] =                twomu * E[0][2];
  S[1][0] =                twomu * E[1][0];
  S[1][1] = lambda * trE + twomu * E[1][1];
  S[1][2] =                twomu * E[1][2];
  S[2][0] =                twomu * E[2][0];
  S[2][1] =                twomu * E[2][1];
  S[2][2] = lambda * trE + twomu * E[2][2];

  /* calculate the principal stresses */
  W_eigendecomposition(S, S_eigvct, S_eigval);

  FOR(i, 3) {
    if (S_eigval[i] >= u->cutoff) 
    {
      /*****DBGP_T("get_stress:S",S, 3,3);
      DBGP_VCT("get_stress:S_eigval",S_eigval, 3);******/
      for (c1 = 0; c1 < 3; c1++) {
          for (c2 = 0; c2 < 3; c2++) {
              S[c1][c2] = 0;
          }
      }
    }
  }

  matst->deform_energy_density = 0.5 * (MAT3DUTIL_DBL_PROD_3X3(S, E));

  *out_stress_type = stress_type;
  switch (stress_type) {
  case MAT_3D_CAUCHY:
    if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F)) {
      result = FALSE; goto exit;
    }
    FOR(i, 3) {
      FOR(j, 3) {
        P[i][j] = 0; /* /F/ * /S/ */ 
        FOR(k, 3) P[i][j] += F[i][k] * S[k][j];
      }
    }
    invdetF = 1 / MAT3DUTIL_DET_3X3(F);
    FOR(i, 3) {
      FOR(j, 3) {
        sigma[i][j] = 0; /* 1/detF /P/ * /F/^T */ 
        FOR(k, 3) sigma[i][j] += invdetF * P[i][k] * F[j][k];
      }
    }
    MAT3DUTIL_COPY_3X3(stress, sigma);
    break;
  case MAT_3D_1ST_PK:
    if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F)) {
      result = FALSE; goto exit;
    }
    FOR(i, 3) {
      FOR(j, 3) {
        P[i][j] = 0; /* /F/ * /S/ */ 
        FOR(k, 3) P[i][j] += F[i][k] * S[k][j];
      }
    }
    MAT3DUTIL_COPY_3X3(stress, P);
    break;
  case MAT_3D_2ND_PK:
    MAT3DUTIL_COPY_3X3(stress, S);
    break;
  default:
    MAT3DUTIL_COPY_3X3(stress, S);
    *out_stress_type = MAT_3D_2ND_PK;
    break;
  }
    
 exit:
  return result;
}

static W_mat_3d_state_p 
get_mat_state (W_mat_3d_t *m, W_loc_t *at_point)
{
  matstate_t *ms;

  ms = w_new(matstate_t); CHECK_MALLOC(ms);

  ms->deform_energy_density = 0;

  return (W_mat_3d_state_p)ms;
}

static W_bool_t 
get_mat_state_var(W_mat_3d_t           *m,
                       W_mat_3d_state_p             ms,
                       W_mat_3d_ip_ctx_t           *ip_ctx,
                       W_mat_3d_mat_state_var_enum  var_type,
                       double                      *var
                       )
{
  matstate_t *matst = (matstate_t *)ms;

  switch(var_type) {
  case MAT_3D_EQUIV_PLASTIC_DEF:
    *var = 0; /* elastic material */
    return TRUE;
  case MAT_3D_DEFORM_ENERGY_DENSITY:
    *var = matst->deform_energy_density;
    return TRUE;  
  default:
    return FALSE;
  }
}




static W_mat_3d_state_p 
interpolate_mat_state(W_mat_3d_t *m,
		      W_loc_t *at_point,
		      int npkgs,
		      WM3D_mat_state_pkg_t *pkgs)
{
  matstate_t *ms, *pms;
  int k;
  double weight_sum = 0;
#define WGHT(K) (pkgs[k].weight / weight_sum)

  FOR(k, npkgs) {weight_sum += pkgs[k].weight;}

  ms = (matstate_t *) get_mat_state(m, at_point);

  FOR(k, npkgs) {
    pms = (matstate_t *)pkgs[k].ms;
    ms->deform_energy_density += pms->deform_energy_density * WGHT(k);
  }

  return (W_mat_3d_state_p)ms;
#undef WGHT
}


static W_bool_t
get_mat_stiff (W_mat_3d_t *m/* mat3d */,
               W_mat_3d_state_p ms,/* material state */
               W_mat_3d_ip_ctx_t *ip_ctx,/* integration point context;
                                            used to generate kinematic
                                            quantities which the material state
                                            might need in order to compute
                                            the requested stress */
               W_mat_3d_stiff_enum stiff_type, /* material stiffness type */
               double C_mat[6][6] /* material stiffness tensor in matrix form */
                   )
{
  bond_mat_t *u = (bond_mat_t *)m->data;
  double lambda = u->lambda, mu = u->G;
  if (stiff_type == MAT_3D_STIFF_LAGRANGIAN) {
#undef C
#define C(I,J,V) { C_mat[I][J] = V; }
    C(0,0,lambda + 2*mu); C(0,1,lambda);        C(0,2,lambda);        C(0,3,0);  C(0,4,0);  C(0,5,0);
    C(1,0,lambda);        C(1,1,lambda + 2*mu); C(1,2,lambda);        C(1,3,0);  C(1,4,0);  C(1,5,0);
    C(2,0,lambda);        C(2,1,lambda);        C(2,2,lambda + 2*mu); C(2,3,0);  C(2,4,0);  C(2,5,0);
    C(3,0,0);             C(3,1,0);             C(3,2,0);             C(3,3,mu); C(3,4,0);  C(3,5,0); 
    C(4,0,0);             C(4,1,0);             C(4,2,0);             C(4,3,0);  C(4,4,mu); C(4,5,0); 
    C(5,0,0);             C(5,1,0);             C(5,2,0);             C(5,3,0);  C(5,4,0);  C(5,5,mu); 
#undef C
#if 0
    { int i, j;
    fprintf (stderr, "C=[\n");
    FOR (i, 6) {
      FOR (j, 6) {
        fprintf (stderr, "%g ", C_mat[i][j]);
      }
    }
    fprintf (stderr, "]\n");
    }
#endif
    return TRUE;
  } else
    return FALSE;
}
	
static W_mat_3d_methods_t mat3dbond_methods = {
  get_mcode,
  get_mat_state,
  get_mat_const, 
  update_stress,
  get_mat_state_var,
  interpolate_mat_state,
  get_mat_stiff,
  NULL,
  NULL
};

W_mat_3d_t *
new_mat_3d_bond  (double E, double nu, 
                   double tensile_strength, 
                   double cutoff,
                   double dyn_fracture_toughness,
                   double rho)
{
  bond_mat_t *u;
  
  u = w_new(bond_mat_t); CHECK_MALLOC(u);

  u->E                      = E;
  u->nu                     = nu;
  u->cutoff                 = cutoff;
  u->tensile_strength       = tensile_strength;
  u->dyn_fracture_toughness = dyn_fracture_toughness;
  
  u->lambda           = u->E * u->nu / (1 + u->nu) / (1 - 2*(u->nu));
  u->G                = u->E / 2 / (1 + u->nu);
  u->bulk_mod         = u->E / 3 / (1 - 2*u->nu);
  u->dil_wave_speed_3d
    = W_3D_dilatational_wave_speed (u->E, u->nu, rho);
  u->dil_wave_speed_pstrain
    = W_plstrain_dilatational_wave_speed (u->E, u->nu, rho);
  u->dil_wave_speed_pstress
    = W_plstress_dilatational_wave_speed (u->E, u->nu, rho);
  u->shear_wave_speed
    = W_shear_wave_speed (u->E, u->nu, rho);
  if (!W_compute_Rayleigh_speed (u->dil_wave_speed_pstress,
                                 u->shear_wave_speed,
                                 0.01,
                                 &u->rayleigh_pstress_wave_speed))
    REPORT_FATAL_ERROR (NULL, "Rayleigh plane stress speed does not exist");
  if (!W_compute_Rayleigh_speed (u->dil_wave_speed_pstrain,
                                 u->shear_wave_speed,
                                 0.01,
                                 &u->rayleigh_pstrain_wave_speed))
    REPORT_FATAL_ERROR (NULL, "Rayleigh plane strain speed does not exist");
  return WM3D_new_material (&mat3dbond_methods, (void *)u, rho);
}

W_mat_3d_t *
read_mat_3d_bond (W_domain_t *d)
{
  const char *input_line;
  double E = 0, nu = 0, tensile_strength = 0, dyn_fracture_toughness = 0, rho = 0;
  double cutoff = 0;
 
  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return NULL;
    else {
      TOKENS_PARSE(input_line);
      TOKENS_SET_BY_KEYWORD("E", TRUE,
			    { E = TOKEN_DBL_NEXT(1); },
			    { NOOP(); }
			    );
      TOKENS_SET_BY_KEYWORD("NU", TRUE,
			    { nu = TOKEN_DBL_NEXT(1); },
			    { NOOP(); }
			    );
      if  (nu >= 0.5 || nu < 0)
        REPORT_FATAL_ERROR  (d->errh, "Bad nu  (should be 0 <= nu < 0.5)");
      TOKENS_SET_BY_KEYWORD("CUTOFF", FALSE,
			    { cutoff = TOKEN_DBL_NEXT(1); },
			    { cutoff = 1; }
			    );
      TOKENS_SET_BY_KEYWORD("TENSILE_STRENGTH", FALSE,
			    { tensile_strength = TOKEN_DBL_NEXT(1); },
			    { tensile_strength = E / 100; }
			    );
      TOKENS_SET_BY_KEYWORD("DYN_FRACTURE_TOUGHNESS", FALSE,
			    { dyn_fracture_toughness = TOKEN_DBL_NEXT(1); },
			    { dyn_fracture_toughness = E / 1000; }
			    );
      TOKENS_SET_BY_KEYWORD("RHO", TRUE,
			    { rho = TOKEN_DBL_NEXT(1); },
			    { NOOP(); }
			    );
      return new_mat_3d_bond (E, nu, tensile_strength, cutoff, dyn_fracture_toughness, rho);
    }
  }
  return NULL;
}

