 
#include "Wtypes.h"
#include "Wutil.h"
#include "Wxf.h"
#include "Wmat3d.h"
#include "Wmat3dqfung.h"

/* RCSHEADER "$Id: mat3dqfung.c,v 1.1 2004/09/06 01:17:58 cvs Exp $" */

static int
get_mcode (W_mat_3d_t *m)
{
  return MAT_3D_QFUNG;
}


typedef struct mat_t {
  double fE, nu, fc;
}              mat_t;


typedef struct matstate_t {
  double   deform_energy_density;
}              matstate_t;


  
#if 0
#define DUMP(S) {                                               \
  printf("%g\t %g\t %g\n", S[0][0], S[0][1], S[0][2]);          \
  printf("%g\t %g\t %g\n", S[1][0], S[1][1], S[1][2]);          \
  printf("%g\t %g\t %g\n", S[2][0], S[2][1], S[2][2]); }

#define HAVENAN(S) (   isnan(S[0][0]) || isnan(S[0][1]) || isnan(S[0][2])  \
                    || isnan(S[1][0]) || isnan(S[1][1]) || isnan(S[1][2])  \
                    || isnan(S[2][0]) || isnan(S[2][1]) || isnan(S[2][2]))
#endif

/* Function updates the stress state, and returns the stress (with proper */
/* conversion into the requested type).  Note that the function get_stress() */
/* is identical with this function since no stress state is saved for this */
/* type of elastic material. */
static W_bool_t 
update_stress (W_mat_3d_t                  *m,
               W_mat_3d_state_p             ms,
               W_mat_state_update_form_enum update_form,
               W_mat_3d_ip_ctx_t           *ip_ctx,
               W_mat_3d_stress_type_enum    stress_type,
               double                       stress[3][3],
               W_mat_3d_stress_type_enum   *out_stress_type
               )
{
  double S[3][3], E[3][3], F[3][3], C[3][3], invC[3][3];
  mat_t *u = (mat_t *)m->data;
  double lambda = u->fE * u->nu / (1 + u->nu) / (1 - 2*(u->nu));
  double twomu = u->fE /  (1 + u->nu);
  double B = u->fE * u->fc / 3 / (1 - 2*u->nu);
  double J, detC3, Q, pressure;
  double trE, Qfact;
  W_bool_t result = TRUE;
  matstate_t *matst = (matstate_t *)ms;
  matstate_t input_matst;
  const double n = 2;
  
  if (!ip_ctx->get_kinem_var) return FALSE;

  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F))
    return FALSE;
  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_CAUCHY_GREEN, C))
    return FALSE;

  /* save the material state if the update is temporary */
  if (update_form == W_MAT_STATE_UPDATE_TEMP) input_matst = *matst;

  J = MAT3DUTIL_DET_3X3(F);
  detC3 = pow(J, 3./2);
  E[0][0] = 0.5 *  (C[0][0]/detC3 - 1);
  E[0][1] = 0.5 * C[0][1]/detC3;
  E[0][2] = 0.5 * C[0][2]/detC3;
  E[1][0] = 0.5 * C[1][0]/detC3;
  E[1][1] = 0.5 *  (C[1][1]/detC3 - 1);
  E[1][2] = 0.5 * C[1][2]/detC3;
  E[2][0] = 0.5 * C[2][0]/detC3;
  E[2][1] = 0.5 * C[2][1]/detC3;
  E[2][2] = 0.5 *  (C[2][2]/detC3 - 1);
  
  trE =   MAT3DUTIL_TRACE_3X3(E); 
  S[0][0] = lambda * trE + twomu * E[0][0];
  S[0][1] =                twomu * E[0][1];
  S[0][2] =                twomu * E[0][2];
  S[1][0] =                twomu * E[1][0];
  S[1][1] = lambda * trE + twomu * E[1][1];
  S[1][2] =                twomu * E[1][2];
  S[2][0] =                twomu * E[2][0];
  S[2][1] =                twomu * E[2][1];
  S[2][2] = lambda * trE + twomu * E[2][2];

  Q = 0.5 * (MAT3DUTIL_DBL_PROD_3X3(S, E));

  Qfact = (2/n) * pow(Q, (2 - n)/n);
  lambda *= u->fc * Qfact;
  twomu  *= u->fc * Qfact;
  if (!W_mat3dutil_invert_3x3 (C, invC)) {
    result = FALSE; goto exit;
  }
  pressure =  (- (J - 1)) * B;
  S[0][0] = lambda * trE + twomu * E[0][0] - pressure * invC[0][0];
  S[0][1] =                twomu * E[0][1] - pressure * invC[0][1];
  S[0][2] =                twomu * E[0][2] - pressure * invC[0][2];
  S[1][0] =                twomu * E[1][0] - pressure * invC[1][0];
  S[1][1] = lambda * trE + twomu * E[1][1] - pressure * invC[1][1];
  S[1][2] =                twomu * E[1][2] - pressure * invC[1][2];
  S[2][0] =                twomu * E[2][0] - pressure * invC[2][0];
  S[2][1] =                twomu * E[2][1] - pressure * invC[2][1];
  S[2][2] = lambda * trE + twomu * E[2][2] - pressure * invC[2][2];

  matst->deform_energy_density = 0.5 * (MAT3DUTIL_DBL_PROD_3X3(S, E));
  /*  fprintf(stderr, "J=%g\n", MAT3DUTIL_DET_3X3(F)); */
  
  *out_stress_type = stress_type;
  switch (stress_type) {
  case MAT_3D_CAUCHY:
    {
      double P[3][3], sigma[3][3], invdetF;
      int i, j, k;
      
      if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F)) {
        result = FALSE; goto exit;
      }
      FOR(i, 3) {
        FOR(j, 3) {
          P[i][j] = 0; /* /F/ * /S/ */ 
          FOR(k, 3) P[i][j] += F[i][k] * S[k][j];
        }
      }
      invdetF = 1 / MAT3DUTIL_DET_3X3(F);
      FOR(i, 3) {
        FOR(j, 3) {
          sigma[i][j] = 0; /* 1/detF /P/ * /F/^T */ 
          FOR(k, 3) sigma[i][j] += invdetF * P[i][k] * F[j][k];
        }
      }
      MAT3DUTIL_COPY_3X3(stress, sigma);
    }
    break;
  case MAT_3D_1ST_PK:
    {
      double P[3][3];
      int i, j, k;
      if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F)) {
        result = FALSE; goto exit;
      }
      FOR(i, 3) {
        FOR(j, 3) {
          P[i][j] = 0; /* /F/ * /S/ */ 
          FOR(k, 3) P[i][j] += F[i][k] * S[k][j];
        }
      }
      MAT3DUTIL_COPY_3X3(stress, P);
    }
    break;
  case MAT_3D_2ND_PK:
    MAT3DUTIL_COPY_3X3(stress, S);
    break;
  default:
    MAT3DUTIL_COPY_3X3(stress, S);
    *out_stress_type = MAT_3D_2ND_PK;
    break;
  }
    
 exit:
  if (update_form == W_MAT_STATE_UPDATE_TEMP) *matst = input_matst;
  return result;
}

static W_bool_t
get_mat_const (W_mat_3d_t *m, W_mat_3d_mat_const_enum type, double *c)
{
  mat_t *u = (mat_t *)m->data;

  switch (type) {
  case MAT_3D_YOUNG_MODULUS:
    *c = u->fE * u->fc;
    break;
  case MAT_3D_POISSON:
    *c = u->nu;
    break;
  case MAT_3D_SHEAR_MODULUS:
    *c = u->fE * u->fc / 2 / (1 + u->nu);
    break;
  case MAT_3D_BULK_MODULUS:
    *c = u->fE * u->fc / 3 / (1 - 2*u->nu);
    break;
  default:
    *c = 0;
    return FALSE;
  }
  return TRUE;
}

static W_bool_t 
get_mat_state_named_var (W_mat_3d_t *m,
                         W_mat_3d_state_p ms,
                         W_mat_3d_ip_ctx_t *ip_ctx,
                         char *name,
                         double *var
                         )
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *matst = (matstate_t *)ms;

  if        (tokens_token_and_keyword_equiv (name, "ln_U_1") ||
             tokens_token_and_keyword_equiv (name, "ln_U_2") ||
             tokens_token_and_keyword_equiv (name, "ln_U_3") ||
             tokens_token_and_keyword_equiv (name, "min_ln_U_x") ||
             tokens_token_and_keyword_equiv (name, "min_ln_U_y") ||
             tokens_token_and_keyword_equiv (name, "min_ln_U_z") ||
             tokens_token_and_keyword_equiv (name, "mid_ln_U_x") ||
             tokens_token_and_keyword_equiv (name, "mid_ln_U_y") ||
             tokens_token_and_keyword_equiv (name, "mid_ln_U_z") ||
             tokens_token_and_keyword_equiv (name, "max_ln_U_x") ||
             tokens_token_and_keyword_equiv (name, "max_ln_U_y") ||
             tokens_token_and_keyword_equiv (name, "max_ln_U_z") ||
             tokens_token_and_keyword_equiv (name, "max_ln_U") 
             ) {
    double F[3][3], R[3][3], U[3][3], ln_U[3][3];
    double evals[3], evecs[3][3], lnevals[3];
    *var = 0;
    if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F)) {
      return FALSE;
    }
    if (W_polar_decomposition (F, R, U)) {
      int imin, imid, imax;
      double logevals[3];
      W_eigendecomposition (U, evecs, evals);
      logevals[0] = log(evals[0]);
      logevals[1] = log(evals[1]);
      logevals[2] = log(evals[2]);
      if (fabs(logevals[0]) > fabs(logevals[1])) {
        if (fabs(logevals[1]) > fabs(logevals[2])) {
          imax = 0; imid = 1; imin = 2; 
        } else {
          if (fabs(logevals[0]) > fabs(logevals[2])) {
            imax = 0; imid = 2; imin = 1; 
          } else {
            imax = 2; imid = 0; imin = 1; 
          }          
        }
      } else {
        if (fabs(logevals[0]) > fabs(logevals[2])) {
          imax = 1; imid = 0; imin = 2; 
        } else {
          if (fabs(logevals[1]) > fabs(logevals[2])) {
            imax = 1; imid = 2; imin = 0; 
          } else {
            imax = 2; imid = 1; imin = 0; 
          }          
        }
      }

      if (tokens_token_and_keyword_equiv (name, "ln_U_1")) *var = log(evals[0]);
      if (tokens_token_and_keyword_equiv (name, "ln_U_2")) *var = log(evals[1]);
      if (tokens_token_and_keyword_equiv (name, "ln_U_3")) *var = log(evals[2]);
      if (tokens_token_and_keyword_equiv (name, "max_ln_U")) {
        *var = fabs(log(evals[0]));
        if (fabs(log(evals[1])) > *var) *var = fabs(log(evals[1]));
        if (fabs(log(evals[2])) > *var) *var = fabs(log(evals[2]));
      }
      if (tokens_token_and_keyword_equiv (name, "min_ln_U_x")) *var = evecs[imin][0];
      if (tokens_token_and_keyword_equiv (name, "min_ln_U_y")) *var = evecs[imin][1];
      if (tokens_token_and_keyword_equiv (name, "min_ln_U_z")) *var = evecs[imin][2];
      if (tokens_token_and_keyword_equiv (name, "mid_ln_U_x")) *var = evecs[imid][0];
      if (tokens_token_and_keyword_equiv (name, "mid_ln_U_y")) *var = evecs[imid][1];
      if (tokens_token_and_keyword_equiv (name, "mid_ln_U_z")) *var = evecs[imid][2];
      if (tokens_token_and_keyword_equiv (name, "max_ln_U_x")) *var = evecs[imax][0];
      if (tokens_token_and_keyword_equiv (name, "max_ln_U_y")) *var = evecs[imax][1];
      if (tokens_token_and_keyword_equiv (name, "max_ln_U_z")) *var = evecs[imax][2];
      goto exit;
    }
  } else if (tokens_token_and_keyword_equiv (name, "J")) {
    double F[3][3];
    if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F)) {
      return FALSE;
    }
    *var = MAT3DUTIL_DET_3X3(F);
  } else {
    *var = 0;
    return FALSE;
  }
 exit:
  return TRUE;
}

static W_bool_t 
get_stress (W_mat_3d_t                  *m,
            W_mat_3d_state_p             ms,
            W_mat_3d_ip_ctx_t           *ip_ctx,
            W_mat_3d_stress_type_enum    stress_type,
            double                       stress[3][3],
            W_mat_3d_stress_type_enum   *out_stress_type
            )
{
  return update_stress (m, ms, W_MAT_STATE_UPDATE_TEMP, ip_ctx,
                        stress_type, stress, out_stress_type);
}

static W_mat_3d_state_p 
get_mat_state (W_mat_3d_t *m, W_loc_t *at_point)
{
  matstate_t *ms;

  ms = w_new(matstate_t); CHECK_MALLOC(ms);

  ms->deform_energy_density = 0;

  return (W_mat_3d_state_p)ms;
}

static W_bool_t 
get_mat_state_var(W_mat_3d_t           *m,
                       W_mat_3d_state_p             ms,
                       W_mat_3d_ip_ctx_t           *ip_ctx,
                       W_mat_3d_mat_state_var_enum  var_type,
                       double                      *var
                       )
{
  matstate_t *matst = (matstate_t *)ms;

  switch(var_type) {
  case MAT_3D_EQUIV_PLASTIC_DEF:
    *var = 0; /* elastic material */
    return TRUE;
  case MAT_3D_DEFORM_ENERGY_DENSITY:
    *var = matst->deform_energy_density;
    return TRUE;  
  default:
    return FALSE;
  }
}




static W_mat_3d_state_p 
interpolate_mat_state(W_mat_3d_t *m,
		      W_loc_t *at_point,
		      int npkgs,
		      WM3D_mat_state_pkg_t *pkgs)
{
  matstate_t *ms, *pms;
  int k;
  double weight_sum = 0;
#define WGHT(K) (pkgs[k].weight / weight_sum)

  FOR(k, npkgs) {weight_sum += pkgs[k].weight;}

  ms = (matstate_t *) get_mat_state(m, at_point);

  FOR(k, npkgs) {
    pms = (matstate_t *)pkgs[k].ms;
    ms->deform_energy_density += pms->deform_energy_density * WGHT(k);
  }

  return (W_mat_3d_state_p)ms;
#undef WGHT
}


static W_bool_t
get_mat_stiff (W_mat_3d_t *m/* mat3d */,
               W_mat_3d_state_p ms,/* material state */
               W_mat_3d_ip_ctx_t *ip_ctx,/* integration point context;
                                            used to generate kinematic
                                            quantities which the material state
                                            might need in order to compute
                                            the requested stress */
               W_mat_3d_stiff_enum stiff_type, /* material stiffness type */
               double C_mat[6][6] /* material stiffness tensor in matrix form */
                   )
{
  mat_t *u = (mat_t *)m->data;
    /* RAW this is *not* correct: need to properly modify for the Fung pseudoelasticity */
    return FALSE;
}
	
static W_mat_3d_methods_t mat3dstvk_methods = {
  get_mcode,
  get_mat_state,
  get_mat_const, 
  update_stress,
  get_mat_state_var,
  interpolate_mat_state,
  get_mat_stiff,
  get_mat_state_named_var,
  NULL
};

static W_mat_3d_t *
new_mat  (double fE, double nu, double fc, 
                   double tensile_strength, double dyn_fracture_toughness,
                   double rho)
{
  mat_t *u;
  double E = fE * fc;
  
  u = w_new(mat_t); CHECK_MALLOC(u);

  u->fE                     = fE;
  u->nu                     = nu;
  u->fc                     = fc;
  return WM3D_new_material (&mat3dstvk_methods, (void *)u, rho);
}

W_mat_3d_t *
read_mat_3d_qfung (W_domain_t *d)
{
  const char *input_line;
  double E = 0, nu = 0, c = 0, tensile_strength = 0, dyn_fracture_toughness = 0, rho = 0;
 
  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return NULL;
    else {
      TOKENS_PARSE(input_line);
      TOKENS_SET_BY_KEYWORD("E", TRUE,
			    { E = TOKEN_DBL_NEXT(1); },
			    { NOOP(); }
			    );
      TOKENS_SET_BY_KEYWORD("NU", TRUE,
			    { nu = TOKEN_DBL_NEXT(1); },
			    { NOOP(); }
			    );
      if  (nu >= 0.5 || nu < 0)
        REPORT_FATAL_ERROR  (d->errh, "Bad nu  (should be 0 <= nu < 0.5)");
      TOKENS_SET_BY_KEYWORD("C", TRUE,
			    { c = TOKEN_DBL_NEXT(1); },
			    { NOOP(); }
			    );
      TOKENS_SET_BY_KEYWORD("TENSILE_STRENGTH", FALSE,
			    { tensile_strength = TOKEN_DBL_NEXT(1); },
			    { tensile_strength = E / 100; }
			    );
      TOKENS_SET_BY_KEYWORD("DYN_FRACTURE_TOUGHNESS", FALSE,
			    { dyn_fracture_toughness = TOKEN_DBL_NEXT(1); },
			    { dyn_fracture_toughness = E / 1000; }
			    );
      TOKENS_SET_BY_KEYWORD("RHO", TRUE,
			    { rho = TOKEN_DBL_NEXT(1); },
			    { NOOP(); }
			    );
      return new_mat (E, nu, c, tensile_strength, dyn_fracture_toughness, rho);
    }
  }
  return NULL;
}

