
#include "Wtypes.h"
#include "Wutil.h"
#include "Wxf.h"
#include "Wmat3d.h"
#include "Wmat3dutil.h"




/*
  Invert a 3x3 matrix.
  Invert a 3x3 matrix if non-singular; otherwise report error.
 */
W_bool_t /* TRUE if result valid (matrix was non-singular) */
W_mat3dutil_invert_3x3(double J[3][3], /* input matrix */
                       double inv_J[3][3] /* output matrix */
                       )
{
  double tmp;

  assert(J != NULL);
  assert(inv_J != NULL);
  
  tmp  = MAT3DUTIL_DET_3X3(J);
  if (tmp == 0) return FALSE;

  tmp = 1 / tmp;

  inv_J[0][0] =   tmp * (J[1][1]*J[2][2] - J[1][2]*J[2][1]);
  inv_J[0][1] = - tmp * (J[0][1]*J[2][2] - J[0][2]*J[2][1]);
  inv_J[0][2] =   tmp * (J[0][1]*J[1][2] - J[0][2]*J[1][1]);
    
  inv_J[1][0] = - tmp * (J[1][0]*J[2][2] - J[1][2]*J[2][0]);
  inv_J[1][1] =   tmp * (J[0][0]*J[2][2] - J[0][2]*J[2][0]);
  inv_J[1][2] = - tmp * (J[0][0]*J[1][2] - J[0][2]*J[1][0]);
    
  inv_J[2][0] =   tmp * (J[1][0]*J[2][1] - J[1][1]*J[2][0]);
  inv_J[2][1] = - tmp * (J[0][0]*J[2][1] - J[0][1]*J[2][0]);
  inv_J[2][2] =   tmp * (J[0][0]*J[1][1] - J[0][1]*J[1][0]);

  return TRUE;
}

