#include "Wconfig.h"
#include "Wxf.h"
#include "Wmats.h"
#include "Wmat001.h"
#include "Wmat002.h"
#include "Wmat003.h"
#include "Wmat004.h"
#include "Wmat005.h"
#include "Wmat006.h"
#include "Wmat007.h"

/* RCSHEADER "$Id: mats.c,v 1.1.1.1 2001/07/24 02:42:51 pkrysl Exp $" */

/*
   ##########################################################################
   THE CODE IN THIS FILE NEEDS TO BE CLEANED UP ALONG THE LINES ALREADY
   IMPLEMENTED FOR A GENERAL 3-D MATERIAL (MAT3D).  DON'T HOLD YOUR BREATH,
   THOUGH.  I SIMPLY DO NOT HAVE THE TIME, SOMEHOW.
   ##########################################################################
 */


#define SIGMA_MAX(_E_) (0.0001 * _E_)

#ifdef X_XFACE
#define NCOLORS 9
static char *color_table[NCOLORS] = {
  "white", "blue", "green", "red", "yellow", "magenta", "cyan", "orange", "grey"
};
static int suc = 0;
#endif


/*
   Create a new material.
 */
W_material_t * /* new material */
WM_new_material (W_mcode_t mcode, /* material code */
                 int mat_id,
                 W_material_methods_t *methods, /* type-specific methods */
                 void *data, /* type-specific data */
                 double rho /* volume mass */
                 )
{
  W_material_t *m/* material */;

  m = w_new(W_material_t);
  CHECK_MALLOC(m);

  m->mcode   = mcode;
  m->id      = mat_id;
  m->data    = data;
  m->methods = methods;
  m->vol_mass = rho;
#ifdef X_XFACE
  m->color = 0x0;
  m->fringes = NULL;
#endif
  return m;
}

/*
  Reader of materials.
 */
W_bool_t /* TRUE if successful; FALSE otherwise. */
W_material_reader (W_domain_t *d /* domain */
                   )
{
  const char *input_line;
  long mcode;

  input_line = W_get_input_line(d);
  
  if (sscanf(input_line, "%ld", &mcode) != 1)
    BAD_INPUT();

  switch (mcode) {
  case MCODE_UNIAX_LE:
    return mat001_reader(d); 
  case MCODE_UNIAX_PEP:
    return mat002_reader(d); 
  case MCODE_UNIAX_HEP:
    return mat003_reader(d); 
  case MCODE_BIAX_LE:
    return mat004_reader(d); 
  case MCODE_TRIAX_LE:
    return mat005_reader(d); 
  case MCODE_GENERAL_TRIAX:
    return mat006_reader(d); 
  case MCODE_GENERAL_UNIAX:
    return mat007_reader (d); 
  default:
    fprintf(stderr, "Unknown mcode\n");
    return FALSE;
  }
}


/*
   Get the volume mass of material.
 */
W_float_t /* value */
WM_get_volume_mass (W_material_t *m/* material */)
{
  return m->vol_mass;
}

#ifdef X_XFACE

EPixel
WM_get_color(W_material_t *m/* material */)
{
  if (m->color == 0x0)
    m->color = ColorGetPixelFromString (color_table[m->id % NCOLORS], &suc);
  return m->color;
}


EFringeTable
WM_get_fringe_table(W_material_t *m/* material */)
{
  W_float_t E, nu;
  double minv, maxv;
  
  if (m->fringes == NULL) {
    m->fringes = ColorCreateFringeTable();
    if (!WM_get_uniax_lin_el_consts(m, &E, &nu)) {
      minv = -1; maxv = 1;
    } else {
      minv = -SIGMA_MAX(E); maxv = SIGMA_MAX(E);
    }
    ColorSetupFringeTableByMinMax(m->fringes, minv, maxv);
  }
  return m->fringes;
}

#endif


W_bool_t
WM_get_uniax_stress(W_material_t *m/* material */,
                    W_matstate_t ms,
                    W_float_t strain,
                    W_float_t strain_rate,
                    W_float_t *stress
                    )
{
  if (m->methods->get_uniaxial_stress != NULL)
    return m->methods->get_uniaxial_stress(m, ms, strain, strain_rate, stress);
  else
    return FALSE;
}


/* Computes the total stresses for plane stress/strain material law */
W_bool_t
WM_get_pss_stress(W_material_t *m/* material */,
                  W_matstate_t ms,
                  W_float_t eps1, W_float_t eps2, W_float_t gamma,
                  W_float_t eps1_rate,
                  W_float_t eps2_rate,
                  W_float_t gamma_rate,
                  W_float_t *sigma1,  W_float_t *sigma2,
                  W_float_t *tau
                  )
{
  if (m->methods->get_pss_stress != NULL)
    return m->methods->get_pss_stress(m, ms, eps1, eps2, gamma,
                                      eps1_rate, eps2_rate, gamma_rate,
                                      sigma1,  sigma2, tau);
  else
    return FALSE;
}



W_bool_t
WM_get_uniax_lin_el_consts(W_material_t *m/* material */,
                           W_float_t *E,
                           W_float_t *nu
                           )
{
  if (m->methods->get_uniaxial_elast_const != NULL)
    return m->methods->get_uniaxial_elast_const(m, E, nu);
  else
    return FALSE;
}



W_float_t
WM_get_uniax_def_energy(W_material_t *m/* material */, W_matstate_t ms, W_float_t strain)
{
  if (m->methods->get_uniaxial_def_energy != NULL)
    return m->methods->get_uniaxial_def_energy(m, ms, strain);
  else
    return 0;
}


W_float_t
WM_get_pss_def_energy(W_material_t *m/* material */, W_matstate_t ms,
                      W_float_t eps1, W_float_t eps2, W_float_t gamma)
{
  if (m->methods->get_pss_def_energy != NULL)
    return m->methods->get_pss_def_energy(m, ms, eps1, eps2, gamma);
  else
    return 0;
}


W_matstate_t 
WM_get_mat_state(W_material_t *m/* material */)
{
  if (m->methods->get_mat_state != NULL)
    return m->methods->get_mat_state(m);
  else
    return NULL;
}

 

W_bool_t
WM_is_in_plastic_state(W_material_t *m/* material */,
                       W_matstate_t ms)
{
  if (m->methods->is_in_plastic_state != NULL)
    return m->methods->is_in_plastic_state(m, ms);
  else
    return FALSE;
}


W_float_t 
WM_get_3d_def_energy (W_material_t *m/* material */,
                      W_matstate_t ms, W_float_t eps[6])
{
  if (m->methods->get_3d_def_energy != NULL)
    return m->methods->get_3d_def_energy(m, ms, eps);
  else
    return 0;
}

W_bool_t 
WM_get_3d_stress (W_material_t *m/* material */,
                  W_matstate_t ms,
                  W_float_t eps[6],
                  W_float_t eps_rate[6],
                  W_float_t sigma[6]
                  )
{
  if (m->methods->get_3d_stress != NULL)
    return m->methods->get_3d_stress (m, ms, eps, eps_rate, sigma);
  else
    return FALSE;
}

int
WM_get_id (W_material_t *m)
{
  return m->id;
}
