#include "Wconfig.h"
#include "Wtypes.h"
#include "Wxf.h"




static LIST nodes = NULL;
static LIST elements = NULL;

typedef struct node_t {
  double x, y, z;
  int    n;
} node_t;

static int total_of_nodes;

static char el_props[132];

typedef struct elem_t {
  int n1, n2;
} elem_t;


int 
dump_elems(void);
int
dump_nodes(void);
int
add_node(double x, double y, double z);
int 
add_elem(int n1, int n2);


int 
main(int argc, char **argv)
{
  FILE *stream;
  char buffer[MAXFILENAMECHAR];
  double x1, y1, z1, x2, y2, z2;

  if (argc == 1) {
    fprintf(stderr, "No input file!?\n");
    exit(0);
  }

  fprintf(stderr, "El props?");
  fgets(el_props, 131, stdin);
  
  if ((stream = fopen(argv[1], "r")) == NULL) {
    fprintf(stderr, "Failed open input file %s!?\n", argv[1]);
    exit(0);
  }

  nodes = make_list();
  elements = make_list();

  while (fgets(buffer, MAXFILENAMECHAR-1, stream) != NULL) {
    if (fgets(buffer, MAXFILENAMECHAR-1, stream) != NULL) {
      sscanf(buffer, "LINE3D [%lf,%lf,%lf] [%lf,%lf,%lf]",
                           &x1, &y1, &z1, &x2, &y2, &z2);
      add_elem(add_node(x1, y1, z1), add_node(x2, y2, z2));
    } else {
      fprintf(stderr, "Failed read input file %s!?\n", argv[1]);
      exit(0);
    }
  }

  dump_nodes();
  dump_elems();

  fclose(stream);
  return 1;
}
  


int 
dump_elems(void)
{
  elem_t *el = NULL;
  int n = 0;

  el = (elem_t *)get_list_next(elements, el);
  while (el != NULL) {
    printf("%d %d %d %d %s\n",
           ++n, el->n1, el->n2, total_of_nodes+1, el_props);
    el = (elem_t *)get_list_next(elements, el);
  }
  return 1;
}


int
dump_nodes(void)
{
  node_t *nr = NULL;

  nr = (node_t *)get_list_next(nodes, nr);
  while (nr != NULL) {
    printf("%d %f %f %f 1 1 1 1 1 1\n", nr->n, nr->x, nr->y, nr->z);
    nr = (node_t *)get_list_next(nodes, nr);
  }
  return 1;
}


#define EPS 0.0001

#define CMP_XYZ(X1, Y1, Z1, X2, Y2, Z2) (   (fabs((X2)-(X1)) < EPS )  \
                                         && (fabs((Y2)-(Y1)) < EPS )  \
                                         && (fabs((Z2)-(Z1)) < EPS ))

int
add_node(double x, double y, double z)
{
  int n = 0;
  node_t *nr = NULL;

  nr = (node_t *)get_list_next(nodes, nr);
  while (nr != NULL) {
    n++;
    if (CMP_XYZ(x, y, z, nr->x, nr->y, nr->z))
      return nr->n;
    nr = (node_t *)get_list_next(nodes, nr);
  }
  nr = (node_t *)make_node(sizeof(node_t));
  nr->x = x, nr->y = y, nr->z = z, nr->n = ++n;
  total_of_nodes = nr->n;
  add_to_tail(nodes, nr);
  return n;
}



int 
add_elem(int n1, int n2)
{
  elem_t *el;
  
  el = (elem_t *)make_node(sizeof(elem_t));
  el->n1 = n1, el->n2 = n2;
  add_to_tail(elements, el);
  return 1;
}


