#include "Esimple.h"
#define VECMAC_VECTOR_TYPE 1
#include "vecmacP.h"
#include "vectP.h"


/*
 cc -o mkquads mkquads.c -I../../Ckit/include -I../../Elixir/include ../../Ckit/lib/`archdir`/libckit.a ../../Elixir/lib/`archdir`/libesi.a ../../Elixir/lib/`archdir`/libelixir.a -lm
*/

typedef struct { double x, y, z; } v_t;
static v_t normal = {0,0,1};

static LIST the_nodes = NULL;
static LIST the_elements = NULL;

typedef struct node_t {
  double x, y, z;
  int    n;
} node_t;

typedef struct elem_t {
  node_t *n1, *n2, *n3, *n4;
} elem_t;

static int 
dump_elems (void)
{
  elem_t *el = NULL;
  int n = 0;

  printf ("ELEM\n");
  printf ("23 ! qmem4c\n");
  el = (elem_t *)get_list_next(the_elements, el);
  while (el != NULL) {
    printf("%d %d %d %d %d 1 1.0\n", ++n,
           el->n1->n, el->n2->n, el->n3->n, el->n4->n);
    el = (elem_t *)get_list_next(the_elements, el);
  }
  printf ("&\n");
  return 1;
}

static int
dump_nodes (void)
{
  node_t *nr = NULL;

  nr = (node_t *)get_list_next(the_nodes, nr);
  printf ("NODES\n");
  while (nr != NULL) {
    printf("%d %g %g %g\n", nr->n, nr->x, nr->y, nr->z);
    nr = (node_t *)get_list_next(the_nodes, nr);
  }
  printf ("&\n");
  return 1;
}

#define EPS 0.0001

#define CMP_XYZ(X1, Y1, Z1, X2, Y2, Z2) (   (fabs((X2)-(X1)) < EPS )  \
                                         && (fabs((Y2)-(Y1)) < EPS )  \
                                         && (fabs((Z2)-(Z1)) < EPS ))

static node_t *
add_node (double x, double y, double z)
{
  node_t *nr = NULL;
  int n = 0;

  nr = (node_t *)get_list_next(the_nodes, nr);
  while (nr != NULL) {
    n++;
    if (CMP_XYZ(x, y, z, nr->x, nr->y, nr->z))
      return nr;
    nr = (node_t *)get_list_next(the_nodes, nr);
  }
  nr = (node_t *)make_node(sizeof(node_t));
  nr->x = x, nr->y = y, nr->z = z, nr->n = ++n;
  add_to_tail(the_nodes, nr);
  return nr;
}

static void
reorient (elem_t *el)
{
#define SWAP(i,j) { node_t *n = el->n##i; el->n##i = el->n##j; el->n##j = n; }
  CkitVector3 na={0,0,0}, nb={0,0,0}, va, vb;
  double vdot;
  VECMAC_3_VEC (va, =, (*el->n2), -, (*el->n1));
  VECMAC_3_VEC (vb, =, (*el->n3), -, (*el->n1));
  VECMAC_CROSS (na, =, va, x, vb); 
  VECMAC_3_VEC (va, =, (*el->n3), -, (*el->n1));
  VECMAC_3_VEC (vb, =, (*el->n4), -, (*el->n1));
  VECMAC_CROSS (nb, =, va, x, vb); 
  vdot = VECMAC_DOT (na, nb);
  if (vdot < 0) {
    SWAP (2,4);
  }
  vdot = VECMAC_DOT (na, normal) + VECMAC_DOT (nb, normal);
  if (vdot < 0) {
    SWAP (1,2); SWAP (3,4);
  }
}

static int 
add_elem (node_t *n1, node_t *n2, node_t *n3, node_t *n4)
{
  elem_t *el;
  
  el = (elem_t *)make_node(sizeof(elem_t));
  el->n1 = n1, el->n2 = n2, el->n3 = n3, el->n4 = n4;
  reorient (el);
  add_to_tail(the_elements, el);
  return 1;
}

int 
main (int argc, char **argv)
{
  FILE *stream;
  char buffer[MAXFILENAMECHAR];
  double x1, y1, z1, x2, y2, z2, x3, y3, z3, x4, y4, z4;

  if (argc == 1) {
    fprintf(stderr, "No input file!?\n");
    fprintf(stderr, "Usage: %s input_file\n", argv[0]);
    exit(0);
  }
  
  if ((stream = fopen(argv[1], "r")) == NULL) {
    fprintf(stderr, "Failed open input file %s!?\n", argv[1]);
    exit(0);
  }

  if (argc == 5) {
    
  }

  the_nodes    = make_list ();
  the_elements = make_list ();

  while (fgets (buffer, MAXFILENAMECHAR-1, stream) != NULL) {
    if (strncmp (buffer, "QUAD", 4) == 0) {
      int status
        = sscanf(buffer,
                 "QUAD [%lf,%lf,%lf] [%lf,%lf,%lf] [%lf,%lf,%lf] [%lf,%lf,%lf]",
                 &x1, &y1, &z1, &x2, &y2, &z2, &x3, &y3, &z3, &x4, &y4, &z4);
      if (status != 12) {
        fprintf(stderr, "Failed scan line \"%s\"!?\n", buffer);
      }
      add_elem (add_node (x1, y1, z1), add_node (x2, y2, z2),
                add_node (x3, y3, z3), add_node (x4, y4, z4));
    }
  }

  dump_nodes ();
  dump_elems ();

  fclose (stream);
  return 1;
}
