#include "Esimple.h"
#define VECMAC_VECTOR_TYPE 1
#include "vecmacP.h"
#include "vectP.h"


static LIST fe_nodes = NULL;
static LIST fe_elements = NULL;

typedef struct node_t {
  double x, y, z;
  int    n;
} node_t;

typedef struct elem_t {
  int n1, n2, layer;
} elem_t;



static int 
dump_elems(void);
static int
dump_nodes(void);
static int
add_node(double x, double y, double z);
static int 
add_elem(int n1, int n2, int layer);

static void
dump_prelude (void)
{
  printf ("<<< replace by problem description >>>\n");
  printf ("! num_of_nodes num_of_elements num_of_materials num_of_loads num_of_time_funcs\n");
  printf ("%d %d 1 0 1\n", list_length (fe_nodes), list_length (fe_elements));
  printf ("M \n");
  printf ("! uniaxial elastic/perfectly plastic\n");
  printf ("   1\n");
  printf ("! material data \n");
  printf ("   1  2.1e9  0.15  7.85 200000 -200000\n");
  printf ("&\n");
  printf ("!  ======================================================= \n");
  printf ("TIME_FUNCTIONS \n");
  printf ("2  ! tabular \n");
  printf ("! func#   #pairs \n");
  printf ("1        3\n");
  printf ("! time     val \n");
  printf ("! Heaviside step function:\n");
  printf ("0        1\n");
  printf ("1        1\n");
  printf ("10000000 1\n");
  printf ("!  END TIME_FUNCTIONS \n");
  printf ("&\n");
  printf ("!  ======================================================\n");
}

int 
main(int argc, char **argv)
{
  FILE *stream;
  char buffer[MAXFILENAMECHAR];
  double x1, y1, z1, x2, y2, z2;
  int layer;

  if (argc == 1) {
    fprintf(stderr, "No input file!?\n");
    exit(0);
  }
  
  if ((stream = fopen(argv[1], "r")) == NULL) {
    fprintf(stderr, "Failed open input file %s!?\n", argv[1]);
    exit(0);
  }

  fe_nodes = make_list();
  fe_elements = make_list();

  layer = 0;
  while (fgets (buffer, MAXFILENAMECHAR-1, stream) != NULL) {
    if (strncmp (buffer, "GID", 3) == 0) {
      int timestamp, serial, status
        = sscanf(buffer,
                 "GID time stamp: %d, serial: %d, layer %d",
                 &timestamp, &serial, &layer);
      if (status != 3) {
        fprintf(stderr, "Failed read input file %s!?\n", argv[1]);
        fprintf(stderr, "Got: %s\n", buffer);
      }
    }
    if (strncmp (buffer, "LINE3D", 6) == 0) {
      int status
        = sscanf(buffer,
                 "LINE3D [%lf,%lf,%lf] [%lf,%lf,%lf]",
                 &x1, &y1, &z1, &x2, &y2, &z2);
      if (status != 6) {
        fprintf(stderr, "Failed read input file %s!?\n", argv[1]);
        fprintf(stderr, "Got: %s\n", buffer);
      }
      add_elem (add_node(x1, y1, z1), add_node(x2, y2, z2), layer);
      layer = 0;
    }
  }
  
  dump_prelude ();
  dump_nodes ();
  dump_elems ();

  fclose(stream);
  return 1;
}
  


static int 
dump_elems (void)
{
  elem_t *el = NULL;
  int n = 0;

  printf ("ELEM\n");
  printf ("1 ! truss\n");
  el = (elem_t *) get_list_next (fe_elements, el);
  while (el != NULL) {
    printf("%d %d %d   1 csa%d\n", ++n, el->n1, el->n2, el->layer);
    el = (elem_t *) get_list_next (fe_elements, el);
  }
  printf ("&\n");
  return 1;
}


static int
dump_nodes(void)
{
  node_t *nr = NULL;

  printf ("NODES\n");
  nr = (node_t *)get_list_next(fe_nodes, nr);
  while (nr != NULL) {
    printf("%d %f %f %f\n", nr->n, nr->x, nr->y, nr->z);
    nr = (node_t *)get_list_next(fe_nodes, nr);
  }
  printf ("&\n");
  return 1;
}


#define EPS 0.0001

#define CMP_XYZ(X1, Y1, Z1, X2, Y2, Z2) (   (fabs((X2)-(X1)) < EPS )  \
                                         && (fabs((Y2)-(Y1)) < EPS )  \
                                         && (fabs((Z2)-(Z1)) < EPS ))

static int
add_node(double x, double y, double z)
{
  int n = 0;
  node_t *nr = NULL;

  nr = (node_t *)get_list_next(fe_nodes, nr);
  while (nr != NULL) {
    n++;
    if (CMP_XYZ(x, y, z, nr->x, nr->y, nr->z))
      return nr->n;
    nr = (node_t *)get_list_next(fe_nodes, nr);
  }
  nr = (node_t *)make_node(sizeof(node_t));
  nr->x = x, nr->y = y, nr->z = z, nr->n = ++n;
  add_to_tail(fe_nodes, nr);
  return n;
}



static int 
add_elem (int n1, int n2, int layer)
{
  elem_t *el;
  
  el = (elem_t *)make_node(sizeof(elem_t));
  el->n1 = n1, el->n2 = n2; el->layer = layer;
  add_to_tail(fe_elements, el);
  return 1;
}


