#! /bin/ksh -f

#
# This is a script for use with WD.
# The run-time arguments are set up so as to run WD in model-reduction collect mode.
#

R=1
T=1
O=undefined
dt=undefined
nsamples=200
conv_ratio=0.001
conv_res_mag=0
max_iter=25
pass_on_max=f
report_iter=f
modnewton=f
initial_stiff=f
recomp_keff_after=1
method=snapshots
dont_subtract_mean=f
nz_per_row=undefined
cutoff_ev=undefined
explicit=f
conv_in="equil"
precond=lu
ksp=preonly
b=20
suffix=""
echocmdonly=f
nles="newton"
wd=wd
silent=f

function usage {
cat <<EOF
Usage: $(basename $0) wd_input [ options ]
Options:
 -R #                       (default $R)
 -T #                       (default $T)
 -O #                       (default $(echo "$T/100" | expeval))
 -dt #                      (default $(echo "$T/$nsamples" | expeval))
 -nsamples #                (default $(echo "$nsamples" | expeval))
 -conv_ratio #              (default $conv_ratio)
 -conv_res_mag #            (default $conv_res_mag)
 -method {direct|snapshots} (default $method)
 -dont_subtract_mean        (default $dont_subtract_mean)
 -max_iter #                (default $max_iter)
 -pass_on_max               (default $pass_on_max)
 -conv_in                   (default $conv_in; equil | displ)
 -report_iter               (default $report_iter)
 -nles #                    (default $nles)
 -modnewton                 (default $modnewton)
 -initial_stiff             (default $initial_stiff)
 -recomp_keff_after n       (default $recomp_keff_after)
 -nz_per_row #              (default 0.01)
 -explicit                  (default $explicit)
 -precond method            (default $precond)
 -ksp method                (default $ksp)
 -cutoff_ev #               (default $cutoff_ev)
 -b #                       (default $b)
 -suffix s                  (default "$suffix")
 -echocmdonly               (default: run the command)
 -wd prog                   (default: $wd)
 -silent                    (default: $silent)
EOF
}

if [ $# -lt 1 ] ; then
    usage; exit 1;
fi
file=$1
if [ $# -gt 1 ] ; then 
   shift;
fi

args=""
while [ $# -ge 1 ] 
do
   a=$1; shift
   case $a in
    -help)
        usage; exit 0; 
        ;;
    -b)
        if [ $# -ge 1 ] ; then
            b=$1; shift
        else
            usage; exit 1;
        fi
        ;;
    -R)
        if [ $# -ge 1 ] ; then
            R=$1; shift
        else
            usage; exit 1;
        fi
        ;;
    -wd)
        if [ $# -ge 1 ] ; then
            wd=$1; shift
        else
            usage; exit 1;
        fi
        ;;
    -O)
        if [ $# -ge 1 ] ; then
            O=$1; shift
        else
            usage; exit 1;
        fi
        ;;
    -T)
        if [ $# -ge 1 ] ; then
            T=$1; shift
        else
            usage; exit 1;
        fi
        ;;
    -suffix)
        if [ $# -ge 1 ] ; then
            suffix=$1; shift
        else
            usage; exit 1;
        fi
        ;;
    -dt)
        if [ $# -ge 1 ] ; then
            dt=$1; shift
        else
            usage; exit 1;
        fi
        ;;
    -nsamples)
        if [ $# -ge 1 ] ; then
            nsamples=$1; shift
        else
            usage; exit 1;
        fi
        ;;
    -conv_ratio)
        if [ $# -ge 1 ] ; then
            conv_ratio=$1; shift
        else
            usage; exit 1;
        fi
        ;;
    -conv_in)
        if [ $# -ge 1 ] ; then
            conv_in=$1; shift
        else
            usage; exit 1;
        fi
        ;;
    -conv_res_mag)
        if [ $# -ge 1 ] ; then
            conv_res_mag=$1; shift
        else
            usage; exit 1;
        fi
        ;;
    -max_iter)
        if [ $# -ge 1 ] ; then
            max_iter=$1; shift
        else
            usage; exit 1;
        fi
        ;;
    -pass_on_max)
        pass_on_max=t;
        ;;
    -nz_per_row)
        if [ $# -ge 1 ] ; then
            nz_per_row=$1; shift
        else
            usage; exit 1;
        fi
        ;;
    -report_iter)
        report_iter=t
        ;;
    -nles)
        if [ $# -ge 1 ] ; then
            nles=$1; shift
        else
            usage; exit 1;
        fi
        ;;
    -modnewton)
        modnewton=t
        ;;
    -initial_stiff)
        initial_stiff=t
        ;;
    -recomp_keff_after)
        if [ $# -ge 1 ] ; then
            recomp_keff_after=$1; shift
        else
            usage; exit 1;
        fi
        ;;
    -echocmdonly)
        echocmdonly=t
        ;;
    -cutoff_ev)
        if [ $# -ge 1 ] ; then
            cutoff_ev=$1; shift
        else
            usage; exit 1;
        fi
        ;;
     -method)
        if [ $# -ge 1 ] ; then
            method=$1; shift
        else
            usage; exit 1;
        fi
        ;;
     -dont_subtract_mean)
        dont_subtract_mean=t
        ;;
    -precond)
        if [ $# -ge 1 ] ; then
            precond=$1; shift
        else
            usage; exit 1;
        fi
        ;;
    -ksp)
        if [ $# -ge 1 ] ; then
            ksp=$1; shift
        else
            usage; exit 1;
        fi
        ;;
    -explicit)
        explicit=t;
        ;;
    -silent)
        silent=t;
        ;;
    *)
        ta="$args $a"
        args="$ta"
        ;;
    esac
done

if [ x"$O" = xundefined ] ; then
   O=$(echo "$T/100" |expeval)
fi

if [ $dt = undefined ] ; then
   dt=$(echo "$T/$nsamples" | expeval)
fi

if [ x"$nz_per_row" = xundefined ] ; then
   nz_per_row=0.01
fi

if [ $report_iter = t ] ; then 
    report_iter=report_iter
else
    report_iter=
fi

if [ $pass_on_max = t ] ; then
    iter_limit="pass_on_max max_iter $max_iter"
else
    iter_limit="max_iter $max_iter"
fi

if [ $modnewton = t ] ; then 
    newton_conf="modnewton recomp_keff_after $recomp_keff_after"
elif [ $initial_stiff = t ] ; then 
    newton_conf="initial_stiff recomp_keff_after $recomp_keff_after"
else
    newton_conf=
fi

if [ $dont_subtract_mean = t ] ; then 
    subtract_mean="dont_subtract_mean"
else
    subtract_mean=""
fi


if [ x"$suffix" = x ] ; then
    suff=""
else
    suff="-$suffix"
fi

if [ $cutoff_ev = undefined ] ; then 
    cutoff_ev=
else
    c="$cutoff_ev"
    cutoff_ev="cutoff_ev $c"
fi

if [ $conv_in = displ ] ; then
    conv_in=conv_in_displ
else
    conv_in=conv_in_equil
fi

if [ "$conv_res_mag" != "0" ] ; then
    conv_res_mag_opt="conv_res_mag $conv_res_mag"
else
    conv_res_mag_opt=
fi


if [ $silent = t ] ; then
  verbose=
else
  verbose=-verbose
fi

if [ $explicit = t ] ; then

command="\
$wd $verbose -! -input $file -integrator 0 -R $R -W $file-full-expl-R$R$suff \
-T $T -b $b -O $O \
-modr_options \"collect method $method $subtract_mean dt $dt $cutoff_ev\" \
$args"

else

command="\
$wd $verbose -! -input $file -integrator 1 -R $R -W $file-full-R$R$suff \
-T $T -b $b -O $O \
-modr_options \"collect method $method dt $dt $subtract_mean $cutoff_ev\" \
-newmarki_options \"$nles conv_ratio $conv_ratio $conv_res_mag_opt solver spetsc nz_per_row $nz_per_row $iter_limit $conv_in $report_iter $newton_conf\" \
-pc_type $precond -ksp_type $ksp \
$args"

fi
if [ $silent != t ] ; then echo "$command"; fi
if [ $echocmdonly != t ] ; then eval $command; fi

