#! /bin/ksh -f

#
# This is a script for use with WD.
# The run-time arguments are set up so as to run WD in model-reduction use mode.
#

method=snapshots
nmodes_list="1"
report_iter=f
nles=newton
modnewton=f
initial_stiff=f
recomp_keff_after=1
modenl=""
R_list="200"
conv_ratio_list="0.001"
T=1
O=undefined
suffix=""
explicit=f
adaptive_modr=f
nitadapt=0
max_iter=25
pass_on_max=f
conv_in="equil"
b=20
modr_file=""
wd=wd
silent=f

function usage {
cat <<EOF
Usage: $(basename $0) wd_input [ options ]
Options:
 -T #                       (default $T)
 -O #                       (default $(echo "$T/100" | bc -l))
 -nmodes_list L             (default $nmodes_list)
 -nmodes n                  (default $nmodes_list)
 -modenl L                  (default "$modenl")
 -R_list L                  (default $R_list)
 -R r                       (default $R_list)
 -conv_ratio_list L         (default $conv_ratio_list)
 -conv_ratio r              (default $conv_ratio_list)
 -method m                  (default $method)
 -report_iter               (default $report_iter)
 -nles #                    (default $nles)
 -modnewton                 (default $modnewton)
 -initial_stiff             (default $initial_stiff)
 -recomp_keff_after n       (default $recomp_keff_after)
 -conv_in                   (default $conv_in; equil | displ)
 -explicit                  (default $explicit)
 -modr_file s               (default $modr_file)
 -adaptive_modr             (default $adaptive_modr)
 -nitadapt n                (default $nitadapt)
 -max_iter n                (default $max_iter)
 -pass_on_max               (default $pass_on_max)
 -suffix s                  (default "$suffix")
 -b n                       (default $b)
 -wd prog                   (default $wd)
 -silent                    (default $silent)
EOF
exit 1
}


if [ $# -lt 1 ] ; then
    usage; exit 0;
fi
file=$1; 
if [ $# -gt 1 ] ; then 
   shift;
fi

args=""
while [ $# -ge 1 ] 
do
   a=$1; shift
   case $a in
    -help)
        usage; exit 0; 
        ;;
    -wd)
        if [ $# -ge 1 ] ; then
            wd=$1; shift
        else
            usage; exit 1;
        fi
        ;;
    -b)
        if [ $# -ge 1 ] ; then
            b=$1; shift
        else
            usage; exit 1;
        fi
        ;;
    -T)
        if [ $# -ge 1 ] ; then
            T=$1; shift
        else
            usage; exit 1;
        fi
        ;;
    -O)
        if [ $# -ge 1 ] ; then
            O=$1; shift
        else
            usage; exit 1;
        fi
        ;;
    -modr_file)
        if [ $# -ge 1 ] ; then
            modr_file=$1; shift
        else
            usage; exit 1;
        fi
        ;;
    -report_iter)
        report_iter=t
        ;;
    -modnewton)
        modnewton=t
        ;;
    -initial_stiff)
        initial_stiff=t
        ;;
    -conv_in)
        if [ $# -ge 1 ] ; then
            conv_in=$1; shift
        else
            usage; exit 1;
        fi
        ;;
    -nles)
        if [ $# -ge 1 ] ; then
            nles=$1; shift
        else
            usage; exit 1;
        fi
        ;;
    -recomp_keff_after)
        if [ $# -ge 1 ] ; then
            recomp_keff_after=$1; shift
        else
            usage; exit 1;
        fi
        ;;
    -explicit)
        explicit=t;
        ;;
    -adaptive_modr)
        adaptive_modr=t;
        ;;
    -pass_on_max)
        pass_on_max=t;
        ;;
    -max_iter)
        if [ $# -ge 1 ] ; then
            max_iter="$1"; shift
        else
            usage; exit 1;
        fi
        ;;
    -nitadapt)
        if [ $# -ge 1 ] ; then
            nitadapt="$1"; shift
        else
            usage; exit 1;
        fi
        ;;
    -nmodes_list)
        if [ $# -ge 1 ] ; then
            nmodes_list="$1"; shift
        else
            usage; exit 1;
        fi
        ;;
    -nmodes)
        if [ $# -ge 1 ] ; then
            nmodes_list="$1"; shift
        else
            usage; exit 1;
        fi
        ;;
     -modenl)
        if [ $# -ge 1 ] ; then
            modenl="$1"; shift
        else
            usage; exit 1;
        fi
        ;;
    -R_list)
        if [ $# -ge 1 ] ; then
            R_list="$1"; shift
        else
            usage; exit 1;
        fi
        ;;
    -R)
        if [ $# -ge 1 ] ; then
            R_list="$1"; shift
        else
            usage; exit 1;
        fi
        ;;
    -conv_ratio_list)
        if [ $# -ge 1 ] ; then
            conv_ratio_list="$1"; shift
        else
            usage; exit 1;
        fi
        ;;
    -conv_ratio)
        if [ $# -ge 1 ] ; then
            conv_ratio_list="$1"; shift
        else
            usage; exit 1;
        fi
        ;;
    -method)
        if [ $# -ge 1 ] ; then
            method=$1; shift
        else
            usage; exit 1;
        fi
        ;;
    -suffix)
        if [ $# -ge 1 ] ; then
            suffix=$1; shift
        else
            usage; exit 1;
        fi
        ;;
    -silent)
        silent=t
        ;;
    *)
        ta="$args $a"
        args="$ta"
        ;;
    esac
done

if [ x"$O" = xundefined ] ; then
   O=$(echo "$T/100" |bc -l)
fi

if [ x"$suffix" = x ] ; then
    suff=""
else
    suff="-$suffix"
fi

if [ $report_iter = t ] ; then 
    report_iter=report_iter
else
    report_iter=
fi

if [ $modnewton = t ] ; then 
    newton_conf="modnewton recomp_keff_after $recomp_keff_after"
elif [ $initial_stiff = t ] ; then 
    newton_conf="initial_stiff recomp_keff_after $recomp_keff_after"
else
    newton_conf=
fi

if [ $explicit = t ] ; then
    integrator=0
else
    integrator=1
fi

if [ $adaptive_modr = t ] ; then
    adaptive_modr="adaptive_modr"
else
    adaptive_modr=
fi

if [ $pass_on_max = t ] ; then
    iter_limit="pass_on_max max_iter $max_iter"
else
    iter_limit="max_iter $max_iter"
fi

if [ $conv_in = displ ] ; then
    conv_in=conv_in_displ
else
    conv_in=conv_in_equil
fi

if [ "$modr_file" != "" ] ; then
    p="$modr_file"
    modr_file="file $p"
fi


if [ $silent = t ] ; then
  verbose=
else
  verbose=-verbose
fi

for R in $(echo $R_list)
do
   if [ x"$modenl" != x ] ; then
       for conv_ratio in $(echo $conv_ratio_list)
       do
            command="$wd $verbose -! -input $file -integrator $integrator \
-T $T -b $b -O $O \
-R $R -W $file-red-R$R-c$conv_ratio-M-$method$suff \
-modr_options \"use $modr_file modenl $modenl method $method\" \
-newmarki_options \"modr $nles solver dmeschach conv_ratio $conv_ratio $iter_limit $report_iter $conv_in $newton_conf $adaptive_modr nitadapt $nitadapt\" $args"
            echo "$command"
            eval $command
            if [ $? -ne 0 ] ; then
                echo "*** Error"
            fi
       done
   else
     for nmodes in $(echo $nmodes_list)
     do
       for conv_ratio in $(echo $conv_ratio_list)
       do
            command="$wd $verbose -! -input $file -integrator $integrator \
-T $T -b $b -O $O \
-R $R -W $file-red-R$R-c$conv_ratio-M$nmodes-$method$suff \
-modr_options \"use $modr_file nmodes $nmodes method $method\" \
-newmarki_options \"modr $nles solver dmeschach conv_ratio $conv_ratio $iter_limit $report_iter $conv_in $newton_conf $adaptive_modr nitadapt $nitadapt\" $args"
            if [ $silent != t ] ; then echo "$command"; fi
            eval $command
            if [ $? -ne 0 ] ; then
                echo "*** Error"
            fi
       done
     done
  fi
done
