#include "Wconfig.h"
#include "Wxf.h"
#include "Wnewmarki.h"
#include "Woptions.h"
#include "Wnode.h"
#include "Whistry.h"
#include "Wutil.h"
#include "Wnewmfend.h"
#include "Wnewmmodr.h"
#include "Wnewmdflt.h"
#include "Wmodr.h"
#include "meschach.h"

/* RCSHEADER "$Id: newmmodr.c,v 1.2 2001/08/16 22:07:10 pkrysl Exp $" */

#define MODRP W_modr_p 

typedef struct history_client_data_t {
  double                            resid_work;
  double                            resid_work_incr;
  double                            resid_work_bound;
  double                            resid_work_incr_bound;
}              history_client_data_t;

typedef struct W_newmark_integrator_implem_t {
  /* client and client callbacks */
  W_newmark_client_t                            client;
  fles_tot_of_equations_callback                tot_of_equations;
  W_newmark_finalize_step_client_proc           finalize_step;
  W_newmark_assemble_eff_stiff_client_proc      assemble_eff_stiff;
  W_newmark_assemble_residual_force_client_proc residual_force;
  W_newmark_start_step_client_proc              start_step;
  W_newmark_update_config_client_proc           update_config;
  /* variables */
  struct W_newmark_integrator_t *delegate_i;
  W_domain_t                    *domain;
  int                            num_eqns; /* number of equations (# of dofs in the full FE model) */
  int                            nmodes; /* number of modes */
  double                        *dof_modes; 
  double                        *Delta_p; /* increment of modal coefficients */
  int                           *idcs;
  W_bool_t                      *constrained;
  MAT                           *Phi; 
  MAT                           *Phi_T_K_Phi; 
  MAT                           *K_e; 
  MAT                           *aux; 
  double                        *ser_Phi_T_K_Phi;
  double                        *Phi_T_R; /* tranformed R */
  double                        *Delta_u; /* full increment of displacement */
  double                        *R; /* full residual */
  double                        *prev_R; /* full residual */
  double                         resid_work;
  double                         resid_work_incr;
  double                         resid_work_bound;
  double                         resid_work_incr_bound;
  /* history */
  history_client_data_t      hist_buf;
}              W_newmark_integrator_implem_t;

static void
save_hist (W_newmark_integrator_implem_t *i)
{
  i->hist_buf.resid_work                   = i->resid_work;
  i->hist_buf.resid_work_bound             = i->resid_work_bound;
  i->hist_buf.resid_work_incr              = i->resid_work_incr;
  i->hist_buf.resid_work_incr_bound        = i->resid_work_incr_bound;
}


static double
vdot (double *a, double *b, int n)
{
  int i;
  double r = 0;
  FOR (i, n) { r += a[i] * b[i]; }
  return r;
}

static void
vsub (double *c, double *a, double *b, int n)
{
  int i;
  FOR (i, n) { c[i] = a[i] - b[i]; }
}

static void
vadd (double *c, double *a, double *b, int n)
{
  int i;
  FOR (i, n) { c[i] = a[i] + b[i]; }
}

static void
vcomb (double *c, double alpha, double *a, double beta, double *b, int n)
{
  int i;
  FOR (i, n) { c[i] = alpha * a[i] + beta * b[i]; }
}

static void
vcopy (double *dest, double *src, int n)
{
  int i;
  FOR (i, n) { dest[i] = src[i]; }
}

static void
vzero (double *c, int n)
{
  int i;
  FOR (i, n) { c[i] = 0; }
}

static double
vnorm (double *fr, int n)
{
  return sqrt (vdot (fr, fr, n));
}

static double
ventrymaxabs (double *c, int n)
{
  int i;
  double m = 0;
  FOR (i, n) { m = max (fabs (c[i]), m); }
  return m;
}

#define  VPRINT(array, n)                               \
{                                                       \
  int _i_;                                              \
  printf ("%s, l%d: %s\n", __FILE__, __LINE__, #array); \
  FOR (_i_, n) {                                        \
    printf ("%g ", array[_i_]);                         \
  }                                                     \
  printf ("\n");                                        \
}

static void
advance (W_newmark_integrator_implem_t *ii)
{
  W_newmark_advance (ii->delegate_i);
}

static W_float_t
get_time (W_newmark_integrator_implem_t *ii)
{
  return W_newmark_get_time (ii->delegate_i);
}

static W_float_t
get_dt (W_newmark_integrator_implem_t *ii)
{
  return W_newmark_get_dt (ii->delegate_i);
}

static void
get_consts (W_newmark_integrator_implem_t *ii, double *beta, double *gamma)
{
  W_newmark_get_consts (ii->delegate_i, beta, gamma);
}

static void
get_stat (W_newmark_integrator_implem_t *ii,
                double *ass_rhs_time, 
                double *ass_lhs_time, 
                double *solve_time,  
                double *start_time, 
                double *updt_time
                )
{
  W_newmark_get_stat (ii->delegate_i, ass_rhs_time, ass_lhs_time, solve_time, start_time, updt_time);
}

static void
update_history (W_newmark_integrator_implem_t *ii, W_history_writer_t *hw)
{
  history_client_data_t *cd = &ii->hist_buf;
  double time = W_newmark_get_time (ii->delegate_i);

  W_newmark_update_history (ii->delegate_i, hw);

  WHR_store (hw, "newmmodr-resid_work", time, cd->resid_work);
  WHR_store (hw, "newmmodr-resid_work_bound", time, cd->resid_work_bound);
  WHR_store (hw, "newmmodr-resid_work_incr", time, cd->resid_work_incr);
  WHR_store (hw, "newmmodr-resid_work_incr_bound", time, cd->resid_work_incr_bound);
  if (cd->resid_work_incr_bound != 0) {
    WHR_store (hw, "newmmodr-resid_work_cos", time,
               fabs (cd->resid_work_incr) / cd->resid_work_incr_bound);
  }
}

static int 
done (W_newmark_integrator_implem_t *ii)
{
  return W_newmark_done (ii->delegate_i);
}

static W_newmark_integrator_methods_t methods = {
  advance,
  get_time,
  get_dt,
  get_consts,
  get_stat,
  update_history,
  done
};

static int
modr_tot_of_equations (void *iip)
{
  W_newmark_integrator_implem_t *ii = (W_newmark_integrator_implem_t *)iip;
  return ii->nmodes;
}

static void
build_Phi (W_newmark_integrator_implem_t *ii, int nr, int *ri)
{
  int j, k, nmodes = ii->nmodes;
  double mean;
  double *dof_modes = ii->dof_modes;
  MODRP modr = ii->domain->modr;
  MAT *Phi;
  
  Phi = ii->Phi = m_resize (ii->Phi, nr, nmodes);
  
  FOR (j, nr) {
    W_modr_dof_modes (modr, ri[j], dof_modes, &mean);
    FOR (k, nmodes) {
      Phi->me[j][k] = dof_modes[k];
    }
  }
}

static void
deserialize (double *v, MAT *m, int n)
{
  int i, j, k;
  k = 0;
  FOR (i, n) {
    FOR (j, n) {
      m->me[i][j] = v[k];
      k++;
    }
  }
}

static void
serialize (MAT *m, double *v, int n)
{
  int i, j, k;
  k = 0;
  FOR (i, n) {
    FOR (j, n) {
      v[k] = m->me[i][j];
      k++;
    }
  }
}

static void
modr_assemble_lhs_callback (struct W_newmark_integrator_t *i, /* Newmark integrator */
                            fles_update_mode_enum update_mode, /* update mode;  */
                            int nr, /* number of rows and columns in square matrix to be assembled */
                            int *ri, /* array of equations numbers, which apply to both rows and columns
                                        of the matrix to be assembled */
                            double *values, /* square matrix to be assembled, passed
                                               as a linear array of nr*nr values */
                            W_bool_t *constrained /* array of nr elements; TRUE if degree-of-freedom
                                                     constrained, FALSE otherwise */
                            )
{
  W_newmark_integrator_implem_t *ii = newmark_implem (i);

  /* build the transform matrix Phi */
  build_Phi (ii, nr, ri);
  /* transform */
  m_resize (ii->K_e, nr, nr);
  deserialize (values, ii->K_e, nr);
  mtrm_mlt (ii->Phi, ii->K_e, ii->aux);
  m_mlt (ii->aux, ii->Phi, ii->K_e);
  /* assemble the transformed matrix */
  m_add (ii->K_e, ii->Phi_T_K_Phi, ii->Phi_T_K_Phi);
}

static int 
modr_get_incr (struct W_newmark_integrator_t *i,
               int eqns[], /* for equation numbers */
               double values[], /* return values in this array */
               int dim /* dimension of preceding arrays */
               )
{
  W_newmark_integrator_implem_t *ii = newmark_implem (i);
  if (dim > 1) {
    double *Delta_u = ii->Delta_u;
    int j;
    FOR (j, dim) {
      values[j] = Delta_u[eqns[j]];
    }
    return dim;
  } else {
    values[0] = ii->Delta_u[eqns[0]];
    return 1;
  }
}

static void
modr_assemble_eff_stiff (W_newmark_client_t iip,
                         W_newmark_assemble_eff_stiff_proc assemble_eff_stiff_callback,
                         W_newmark_get_incr_proc get_incr,
                         W_newmark_predictor_enum predictor,
                         W_float_t stiff_mult,
                         W_float_t mass_mult)
{
  W_newmark_integrator_implem_t *ii = (W_newmark_integrator_implem_t *)iip;
  m_zero (ii->Phi_T_K_Phi);
  ii->assemble_eff_stiff (ii->client, modr_assemble_lhs_callback,
                          modr_get_incr, predictor, stiff_mult, mass_mult);
  serialize (ii->Phi_T_K_Phi, ii->ser_Phi_T_K_Phi, ii->nmodes);
  (*assemble_eff_stiff_callback) (ii->delegate_i, FLES_ADD,
                                  ii->nmodes, ii->idcs,
                                  ii->ser_Phi_T_K_Phi, ii->constrained);
}

static void
modr_assemble_rhs_callback (struct W_newmark_integrator_t *i, /* Newmark integrator */
                            fles_update_mode_enum update_mode, /* update mode */
                            int nr,/* number of rows in vector to be assembled */
                            int *ri,/* array of equations numbers */
                            double *values /* vector to be assembled */
                            )
{
  W_newmark_integrator_implem_t *ii = newmark_implem (i);
  double *R = ii->R;
  int k;

  FOR (k, nr) { 
    R[ri[k]] += values[k];
  }
}

static void
untransf_Delta_u (W_newmark_integrator_implem_t *ii, W_newmark_get_incr_proc get_incr)
{
  int j, num_eqns = ii->num_eqns, nmodes = ii->nmodes;
  MODRP modr = ii->domain->modr;
  double mean;
  double *Delta_u = ii->Delta_u;
  double *Delta_p = ii->Delta_p;
  double *dof_modes = ii->dof_modes;
  (*get_incr) (ii->delegate_i, ii->idcs, Delta_p, nmodes);
  vzero (Delta_u, num_eqns);
  FOR (j, num_eqns) {
    W_modr_dof_modes (modr, j, dof_modes, &mean);
    Delta_u[j] += vdot (dof_modes, Delta_p, nmodes);
  }
}

static void
transf_R (W_newmark_integrator_implem_t *ii)
{
  int j, k, num_eqns = ii->num_eqns, nmodes = ii->nmodes;
  MODRP modr = ii->domain->modr;
  double mean;
  double *R = ii->R;
  double *Phi_T_R = ii->Phi_T_R;
  double *dof_modes = ii->dof_modes;

  vzero (Phi_T_R, nmodes);
  FOR (j, num_eqns) {
    W_modr_dof_modes (modr, j, dof_modes, &mean);
    FOR (k, nmodes) {
      Phi_T_R[k] += R[j] * dof_modes[k];
    }
  }
}

static void
modr_residual_force (W_newmark_client_t iip,
                     W_newmark_assemble_residual_force_proc assemble_residual_force,
                     W_mat_state_update_form_enum update_form,
                     W_newmark_get_incr_proc get_incr,
                     W_newmark_predictor_enum predictor,
                     double *extf_norm,
                     double *inertf_norm,
                     double *resid_norm)
{
  W_newmark_integrator_implem_t *ii = (W_newmark_integrator_implem_t *)iip;
  untransf_Delta_u (ii, get_incr);
  vzero (ii->R, ii->num_eqns);
  ii->residual_force (ii->client, modr_assemble_rhs_callback, update_form, modr_get_incr,
                      predictor, extf_norm, inertf_norm, resid_norm);
  transf_R (ii);
  (*assemble_residual_force) (ii->delegate_i, FLES_ADD,
                              ii->nmodes, ii->idcs, ii->Phi_T_R);
}

static void
modr_start_step (W_newmark_client_t iip)
{
  W_newmark_integrator_implem_t *ii = (W_newmark_integrator_implem_t *)iip;
  ii->start_step (ii->client);
}

static void
modr_give_me_disp_incr (struct W_newmark_integrator_t *i, int nr, int *ri, double *du)
{
  W_newmark_integrator_implem_t *ii = newmark_implem (i);
  int j, k;
  double *fr = ii->R;
  double *pfr = ii->prev_R;
  double du_norm = 0;
  double fr_norm = 0;
  FOR (j, nr) {
    k = ri[j];
    du_norm += du[j] * du[j];
    fr_norm += ((fr[k] + pfr[k]) / 2) * ((fr[k] + pfr[k]) / 2);
    ii->resid_work_incr += (fr[k] + pfr[k]) / 2 * du[j];
  }
  ii->resid_work_incr_bound += sqrt (du_norm * fr_norm);
}

static W_float_t
modr_finalize_step (W_newmark_client_t iip, W_newmark_incr_disp_proc incr_disp_callback)
{
  W_newmark_integrator_implem_t *ii = (W_newmark_integrator_implem_t *)iip;
  double newdt = 0;

  ii->resid_work_incr = 0;
  ii->resid_work_incr_bound = 0;
  if (ii->finalize_step) newdt = ii->finalize_step (ii->client, modr_give_me_disp_incr);
  else                   newdt = 0;

  ii->resid_work += ii->resid_work_incr;
  ii->resid_work_bound += ii->resid_work_incr_bound;

  vcopy (ii->prev_R, ii->R, ii->num_eqns);
  
  save_hist (ii);

  return newdt;
}

static void
modr_update_config (W_newmark_client_t iip,
                    W_newmark_get_incr_proc get_incr,
                    W_newmark_predictor_enum predictor)
{
  W_newmark_integrator_implem_t *ii = (W_newmark_integrator_implem_t *)iip;
  untransf_Delta_u (ii, get_incr);
  ii->update_config (ii->client, modr_get_incr, predictor);
}

static void
build (W_newmark_integrator_implem_t *ii)
{
  int j, nmodes, num_eqns;
  
  num_eqns = ii->num_eqns = ii->tot_of_equations (ii->client);
  nmodes = ii->nmodes   = W_modr_get_nmodes (ii->domain->modr);

  ii->dof_modes = w_new_array (nmodes, double);
  CHECK_MALLOC (ii->dof_modes);
  vzero (ii->dof_modes, nmodes);

  ii->Delta_p = w_new_array (nmodes, double);
  CHECK_MALLOC (ii->Delta_p);
  vzero (ii->Delta_p, nmodes);

  ii->idcs = w_new_array (nmodes, int);
  CHECK_MALLOC (ii->idcs);
  FOR (j, nmodes) { ii->idcs[j] = j; }

  ii->constrained = w_new_array (nmodes, W_bool_t);
  CHECK_MALLOC (ii->constrained);
  FOR (j, nmodes) { ii->constrained[j] = FALSE; }
  
  CATCH(E_MEM,
        { 
          ii->Phi = m_get (24, nmodes);
        }, {
          REPORT_FATAL_ERROR (ii->domain->errh, "Not enough memory");
        });
  CATCH(E_MEM,
        { 
          ii->K_e = m_get (30, 30);
        }, {
          REPORT_FATAL_ERROR (ii->domain->errh, "Not enough memory");
        });
  CATCH(E_MEM,
        {
          ii->Phi_T_K_Phi = m_get (nmodes, nmodes);
        }, {
          REPORT_FATAL_ERROR (ii->domain->errh, "Not enough memory");
        });
  CATCH(E_MEM,
        {
          ii->aux = m_get (nmodes, nmodes);
        }, {
          REPORT_FATAL_ERROR (ii->domain->errh, "Not enough memory");
        });

  ii->ser_Phi_T_K_Phi = w_new_array (nmodes*nmodes, double);
  CHECK_MALLOC (ii->ser_Phi_T_K_Phi);

  ii->R = w_new_array (num_eqns, double);
  CHECK_MALLOC (ii->R);

  ii->prev_R = w_new_array (num_eqns, double);
  CHECK_MALLOC (ii->prev_R);
  vzero (ii->prev_R, num_eqns);

  ii->Phi_T_R = w_new_array (nmodes, double);
  CHECK_MALLOC (ii->Phi_T_R);

  ii->Delta_u = w_new_array (num_eqns, double);
  CHECK_MALLOC (ii->Delta_u);

  ii->resid_work_incr_bound = ii->resid_work_incr = ii->resid_work_bound = ii->resid_work = 0;
}

/* ######################################################################### */
/*              Newmark public interface functions follow:                  */
/* ######################################################################### */

/*
   New Newmark integrator.
   Create a new Newmark integrator implementation which represents the
   model reduction layer superimposed on top of the default Newmark integrator.
 */
struct W_newmark_integrator_t *
W_newmark_new_modr_integrator (W_newmark_client_t client, /* client (may be, but needn't be, the domain) */
                             W_domain_t *domain, /* domain */
                             W_float_t               t0, /* initial time */
                             W_float_t               dt, /* time step */
                             W_float_t               beta, /* Newmark beta */
                             W_float_t               gamma, /* Newmark gamma */
                             fles_tot_of_equations_callback
                             tot_of_equations_callback, /* total-of-equations callback */
                             W_newmark_assemble_eff_stiff_client_proc
                             assemble_eff_stiff, /* procedure to assemble
                                                    effective stiffness */
                             W_newmark_assemble_residual_force_client_proc
                             residual_force, /* procedure to assemble
                                                residual force */
                             W_newmark_start_step_client_proc
                             start_step, /* procedure to start a new step */
                             W_newmark_update_config_client_proc
                             update_config, /* procedure to update configuration */
                             W_newmark_finalize_step_client_proc
                             finalize_step, /* procedure to finalize step */
                             char *optstring /* options string; specify option+value pairs */
                             )
{  
  W_newmark_integrator_implem_t *ii;
    
  /* Create integrator */
  ii = w_new  (W_newmark_integrator_implem_t); CHECK_MALLOC(ii);

  /* Client, domain, front-end */
  ii->client = client;
  /* my client supplied these procedures */
  ii->tot_of_equations    = tot_of_equations_callback;
  ii->assemble_eff_stiff  = assemble_eff_stiff;
  ii->update_config       = update_config;
  ii->start_step          = start_step;
  ii->residual_force      = residual_force;
  ii->finalize_step       = finalize_step;
  
  ii->domain = domain;

  build (ii);

  ii->delegate_i = W_newmark_new_dflt_integrator (ii, domain, t0, dt, beta, gamma, 
                                                  modr_tot_of_equations,
                                                  modr_assemble_eff_stiff, 
                                                  modr_residual_force, 
                                                  modr_start_step, 
                                                  modr_update_config, 
                                                  modr_finalize_step, 
                                                  optstring
                                                  );

  save_hist (ii);

  return newmark_new_front_end (ii, &methods);
}
