#include "Wconfig.h"
#include "Wxf.h"
#include "Wtypes.h"
#include "Wnode.h"
#include "Wdomain.h"
#include "Wutil.h"
#include "Whistry.h"
#include "Wnewmarki.h"
#include "Wcdi.h"
#include "meschach.h"

/* RCSHEADER "$Id: node.c,v 1.5 2004/09/06 01:17:58 cvs Exp $" */

#define COPY_TRIAD(theDEST, theSRC)       \
 {                                        \
   (theDEST).vect1.x = (theSRC).vect1.x;  \
   (theDEST).vect1.y = (theSRC).vect1.y;  \
   (theDEST).vect1.z = (theSRC).vect1.z;  \
   (theDEST).vect2.x = (theSRC).vect2.x;  \
   (theDEST).vect2.y = (theSRC).vect2.y;  \
   (theDEST).vect2.z = (theSRC).vect2.z;  \
   (theDEST).vect3.x = (theSRC).vect3.x;  \
   (theDEST).vect3.y = (theSRC).vect3.y;  \
   (theDEST).vect3.z = (theSRC).vect3.z;  \
 }

#define COPY_VECTOR(theDEST, theSRC)      \
  {                                       \
    (theDEST).x += (theSRC).x;            \
    (theDEST).y += (theSRC).y;            \
    (theDEST).z += (theSRC).z;            \
  }

#define ADD_VECTOR(theSUM, theINCR)       \
  {                                       \
    (theSUM).x += (theINCR).x;            \
    (theSUM).y += (theINCR).y;            \
    (theSUM).z += (theINCR).z;            \
  }

#define SUBTRACT_VECTOR(theSUM, theINCR)  \
  {                                       \
    (theSUM).x -= (theINCR).x;            \
    (theSUM).y -= (theINCR).y;            \
    (theSUM).z -= (theINCR).z;            \
  }

#define MULT_VECTOR(theVECTOR, theFACTOR)    \
 {                                           \
    (theVECTOR).x *= (theFACTOR);            \
    (theVECTOR).y *= (theFACTOR);            \
    (theVECTOR).z *= (theFACTOR);            \
  }

#define ZERO_VECTOR(theVECTOR) (theVECTOR).x=(theVECTOR).y=(theVECTOR).z= 0


#define MAKE_SPATIAL(theMATERIAL, theSPATIAL)                                \
  {                                                                          \
    (theSPATIAL).x = (  (theMATERIAL).x * n->mom_inertia_princ_dirs.vect1.x  \
                      + (theMATERIAL).y * n->mom_inertia_princ_dirs.vect2.x  \
                      + (theMATERIAL).z * n->mom_inertia_princ_dirs.vect3.x);\
    (theSPATIAL).y = (  (theMATERIAL).x * n->mom_inertia_princ_dirs.vect1.y  \
                      + (theMATERIAL).y * n->mom_inertia_princ_dirs.vect2.y  \
                      + (theMATERIAL).z * n->mom_inertia_princ_dirs.vect3.y);\
    (theSPATIAL).z = (  (theMATERIAL).x * n->mom_inertia_princ_dirs.vect1.z  \
                      + (theMATERIAL).y * n->mom_inertia_princ_dirs.vect2.z  \
                      + (theMATERIAL).z * n->mom_inertia_princ_dirs.vect3.z);\
  }
   
 
/*
   Predict acceleration of node.
   Predict acceleration of node at the end of the time step for the
   Newmark implicit algorithm.
   */
static void
pred_acceleration (W_node_t *n, 
                   double *x, double *y, double *z)
{
  if (n->domain->newmark_integrator) {
    switch (n->domain->newmark_predictor) {
    case W_NEWMARK_PRED_SAME_U:
      {
        double beta, gamma, dt, c0, c1;
        W_newmark_get_consts (n->domain->newmark_integrator, &beta, &gamma);
        dt = WD_time_step (n->domain);
        c0 = - 1 / (beta * dt);
        c1 = (1 - 1 / (2 * beta));
        *x = c0 * n->velocity_t.x + c1 * n->acceleration_t.x;
        *y = c0 * n->velocity_t.y + c1 * n->acceleration_t.y;
        *z = c0 * n->velocity_t.z + c1 * n->acceleration_t.z;
      }
      break;
    case W_NEWMARK_PRED_SAME_A:
      *x = n->acceleration_t.x;
      *y = n->acceleration_t.y;
      *z = n->acceleration_t.z;
      break;
    case W_NEWMARK_PRED_ZERO_A:
      *x = 0;
      *y = 0;
      *z = 0;
      break;
    case W_NEWMARK_PRED_SAME_V:
      {
        double beta, gamma, c;
        W_newmark_get_consts (n->domain->newmark_integrator, &beta, &gamma);
        c = - (1 - gamma) / gamma;
        *x = c * n->acceleration_t.x;
        *y = c * n->acceleration_t.y;
        *z = c * n->acceleration_t.z;
      }
      break;
    case W_NEWMARK_PRED_PREV_INCR:
      {
        double beta, gamma, dt, c0, c1, c2;
        W_newmark_get_consts (n->domain->newmark_integrator, &beta, &gamma);
        dt = WD_time_step (n->domain);
        c0 = 1 / (beta * dt * dt);
        c1 = - 1 / (beta * dt);
        c2 = (1 - 1 / (2 * beta));
        *x = c0 * (n->displ_t.x - n->displ_t_minus_dt.x) + c1 * n->velocity_t.x + c2 * n->acceleration_t.x;
        *y = c0 * (n->displ_t.y - n->displ_t_minus_dt.y) + c1 * n->velocity_t.y + c2 * n->acceleration_t.y;
        *z = c0 * (n->displ_t.z - n->displ_t_minus_dt.z) + c1 * n->velocity_t.z + c2 * n->acceleration_t.z;
      }
      break;
    default:
      /* break the computation to signal error; should not get here in the first place */
      *x = FLT_MAX;
      *y = FLT_MAX;
      *z = FLT_MAX;
      break;
    }
  }
}

static void
pred_displ (W_node_t *n,
            double *x, double *y, double *z
            )
{
  double beta, gamma, dt;
  W_vect_t a, a0t;
  W_newmark_get_consts (n->domain->newmark_integrator, &beta, &gamma);
  dt = WD_time_step (n->domain);
  pred_acceleration (n, &a0t.x, &a0t.y, &a0t.z);
  VECMAC_3_VEC (a, =, (1 - 2*beta) * n->acceleration_t, +, (2*beta) * a0t);
  *x = n->displ_t.x + dt * n->velocity_t.x + (dt * dt / 2) * a.x;
  *y = n->displ_t.y + dt * n->velocity_t.y + (dt * dt / 2) * a.y;
  *z = n->displ_t.z + dt * n->velocity_t.z + (dt * dt / 2) * a.z;
}

/*
   Get reference position of node.
 */
void
WN_get_ref_coords (W_node_t *n, /* node */
                   double *x, /* x-coordinate */
                   double *y, /* y-coordinate */ 
                   double *z /* z-coordinate */
                   )
{
  *x = n->ref_coords.x;
  *y = n->ref_coords.y;
  *z = n->ref_coords.z;
}

/*
   Get current displacements of node.
 */
void
WN_get_displ (W_node_t *n, /* node */
              double *x, /* x-displacement */
              double *y, /* y-displacement */
              double *z /* z-displacement */
              )
{
  if (n->domain->newmark_integrator && n->domain->newmark_get_incr) {
    W_index_t ens[W_NDOF_PER_NODE];
    W_bool_t constrained[W_NDOF_PER_NODE];
    double delta_u[W_NDOF_PER_NODE];
    W_vect_t u0t;
    WN_eqn_numbers_and_constraints (n, ens, constrained);
    n->domain->newmark_get_incr (n->domain->newmark_integrator,
                                 ens, delta_u, W_NDOF_PER_NODE);
    pred_displ (n, &u0t.x, &u0t.y, &u0t.z);
    *x = u0t.x + delta_u[0];
    *y = u0t.y + delta_u[1];
    *z = u0t.z + delta_u[2];
  } else {
    *x = n->displ_t.x;
    *y = n->displ_t.y;
    *z = n->displ_t.z;
  }
}

/*
   Get previous displacement of node (ie. at time t_(N-1))
 */
void
WN_get_prev_displ (W_node_t *n, /* node */
                   double *x, /* x-displacement */
                   double *y, /* y-displacement */
                   double *z /* z-displacement */
                   )
{
  if (n->domain->newmark_integrator && n->domain->newmark_get_incr) {
    *x = n->displ_t.x;
    *y = n->displ_t.y;
    *z = n->displ_t.z;
  } else {
    *x = n->displ_t_minus_dt.x;
    *y = n->displ_t_minus_dt.y;
    *z = n->displ_t_minus_dt.z;
  }
}

/*
   Get increment of displacements in the current time step.
 */
void
WN_get_disp_increment (W_node_t *n, /* node */
                       W_coords_3d_t *transl_incr,/* translational increment */
                       W_coords_3d_t *rot_incr /* rotary increment */
                       )
{
  if (n->domain->newmark_integrator && n->domain->newmark_get_incr) {
    W_vect_t uit;
    WN_get_displ (n, &uit.x, &uit.y, &uit.z);
    VECMAC_3_VEC ((*transl_incr), =, uit, -, n->displ_t);
    /* RAW: increment for rotational dofs */
    VECMAC_SCALAR ((*rot_incr), =, 0);/* RAW */
  } else {
    VECMAC_3_VEC ((*transl_incr), =, n->displ_t, -, n->displ_t_minus_dt);
    
#if W_ROT_DOFS
    VECMAC_2_VEC ((*rot_incr), =, n->rot_incr_t);
#endif
  }
}

/*
   Get current coordinates of node.
 */
void
WN_get_coords (W_node_t *n, /* node */
               double *x, /* x-coordinate */
               double *y, /* y-coordinate */
               double *z /* z-coordinate */
               )
{
  double ux, uy, uz;
  WN_get_displ (n, &ux, &uy, &uz);
  *x = n->ref_coords.x + ux;
  *y = n->ref_coords.y + uy;
  *z = n->ref_coords.z + uz;
}


/*
   Get previous coordinates of node.
 */
void
WN_get_prev_coords (W_node_t *n, /* node */
                    double *x, /* x-coordinate */
                    double *y, /* y-coordinate */
                    double *z /* z-coordinate */
                    )
{
  W_vect_t prev_u;
  WN_get_prev_displ (n, &prev_u.x, &prev_u.y, &prev_u.z);
  *x = n->ref_coords.x + prev_u.x;
  *y = n->ref_coords.y + prev_u.y;
  *z = n->ref_coords.z + prev_u.z;
}



/*
   Get current velocity of node.
 */
void
WN_get_velocity (W_node_t *n, /* node */
                 double *x, /* x-velocity */
                 double *y, /* y-velocity */
                 double *z /* z-velocity */
                 )
{
  if (n->domain->newmark_integrator && n->domain->newmark_get_incr) {
    double beta, gamma, dt, ax, ay, az;
    W_newmark_get_consts (n->domain->newmark_integrator, &beta, &gamma);
    dt = WD_time_step (n->domain);
    WN_get_acceleration (n, &ax, &ay, &az);
    *x = n->velocity_t.x + dt * ((1 - gamma) * n->acceleration_t.x + gamma * ax);
    *y = n->velocity_t.y + dt * ((1 - gamma) * n->acceleration_t.y + gamma * ay);
    *z = n->velocity_t.z + dt * ((1 - gamma) * n->acceleration_t.z + gamma * az);
  } else {
    *x = n->velocity_t.x;
    *y = n->velocity_t.y;
    *z = n->velocity_t.z;
  }
}

/*
   Get current acceleration of node.
   This version uses the formula 
     a(i)_t+dt = 1/(beta*dt^2){u(i)_t+dt - u_t} - 1/(beta*dt) v_t + (1-1/(2*beta))a_t
   */
void
WN_get_acceleration (W_node_t *n, /* node */
                     double *x, /* x-acceleration */
                     double *y, /* y-acceleration */
                     double *z /* z-acceleration */
                 )
{
  if (n->domain->newmark_integrator && n->domain->newmark_get_incr) {
    double beta, gamma, dt, c0, c1, c2;
    W_vect_t uit;
    W_newmark_get_consts (n->domain->newmark_integrator, &beta, &gamma);
    dt = WD_time_step (n->domain);
    c0 = (1 / beta / dt / dt);
    c1 = - 1 / (beta * dt);
    c2 = (1 - 1 / (2 * beta));
    WN_get_displ (n, &uit.x, &uit.y, &uit.z);
    *x = c0 * (uit.x - n->displ_t.x) + c1 * n->velocity_t.x + c2 * n->acceleration_t.x;
    *y = c0 * (uit.y - n->displ_t.y) + c1 * n->velocity_t.y + c2 * n->acceleration_t.y;
    *z = c0 * (uit.z - n->displ_t.z) + c1 * n->velocity_t.z + c2 * n->acceleration_t.z;
  } else {
    *x = n->acceleration_t.x;
    *y = n->acceleration_t.y;
    *z = n->acceleration_t.z;
  }
}

/*
  New node.
  Create a new node.  The node identifier may be input as a positive
  integer, or as a negative number.  In the latter case, the node identifier
  is generated internally.  Be warned though, that could lead to conflicts
  because another request to create a node might specify
  the same identifier as was generated here for a negative /id/.

  The default internal identifier is also generated here.
  
  The default initial displacement, velocity, and acceleration
  are a zero vector.

  The node is added to the domain on input.
 */
W_node_t * /* node or NULL if anything went wrong */
WN_new_node (W_domain_t *d, /* domain; the node is added to this domain */
             W_index_t id, /* node identifier; either positive integer, which means "I know what
                              the number of this node is", or a non-positive integer, which means "generate
                              an identifier for me, please".
                           */
             W_coords_3d_t *c /* reference coordinate of the node */
             )
{
  W_node_t *n;
  W_loc_t zerov = {0,0,0};
  static W_index_t default_internal_id = 0;

  n = w_new(W_node_t);
  CHECK_MALLOC(n);

  n->domain         = d;
  /* Note: internal identifiers start with zero, and are supposed to be contiguous
     non-negative integers.  Their purpose is to generate equation numbers for linear
     equation solvers.
  */
  n->internal_id      = default_internal_id++;
  n->id               = (id >= 1 ? id : n->internal_id+1);
  n->ref_coords       = *c;
  n->displ_t          = zerov;
  n->displ_t_minus_dt = zerov;
  n->velocity_t       = zerov;
  n->acceleration_t   = zerov;

#if W_ROT_DOFS
  n->nodal_triad_t.vect1.x = 1;
  n->nodal_triad_t.vect1.y = n->nodal_triad_t.vect1.z = 0; 
  n->nodal_triad_t.vect2.y = 1;
  n->nodal_triad_t.vect2.x = n->nodal_triad_t.vect2.z = 0; 
  n->nodal_triad_t.vect3.z = 1;
  n->nodal_triad_t.vect3.x = n->nodal_triad_t.vect3.y = 0;

  n->ang_velocity_t = zerov;

  n->ang_acceleration_t.x= n->ang_acceleration_t.y= n->ang_acceleration_t.z= 0;

  n->rot_incr_t = zerov;
  
  n->mom_inertia_princ_dirs.vect1.x = n->mom_inertia_princ_dirs.vect1.y =
    n->mom_inertia_princ_dirs.vect1.z = 0;
  n->mom_inertia_princ_dirs.vect2.x = n->mom_inertia_princ_dirs.vect2.y =
    n->mom_inertia_princ_dirs.vect2.z = 0;
  n->mom_inertia_princ_dirs.vect3.x = n->mom_inertia_princ_dirs.vect3.y =
    n->mom_inertia_princ_dirs.vect3.z = 0;
#endif

  n->can_move_in_dir[W_DX] = n->can_move_in_dir[W_DY] = n->can_move_in_dir[W_DZ] = TRUE;
#if W_ROT_DOFS
  n->can_move_in_dir[W_RX] = n->can_move_in_dir[W_RY] = n->can_move_in_dir[W_RZ] = TRUE;
#endif

  n->force_t = zerov;

#if W_ROT_DOFS
  n->moment_t = zerov;
#endif

#ifdef X_XFACE
  n->node_marker = NULL;
# if W_ROT_DOFS
  n->node_t1 = n->node_t2 = n->node_t3 = NULL;
# endif
#endif

  n->mass = 0;
#if W_ROT_DOFS
  n->mom_inertia[0] = n->mom_inertia[1] = n->mom_inertia[2] = 0;
#endif
  
  WD_add_node (d, n->id, n); 
  return n;
}


/*
   Node reader.
   Used by domain to read node data from the input file.
 */
W_bool_t
W_node_reader (W_domain_t *d)
{
  const char *input_line;
  W_node_t *n;
  long n_num;
  W_coords_3d_t c = {0, 0, 0};
 
  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return TRUE;
    else {
      TOKENS_PARSE (input_line);
      n_num = TOKEN_INT (1);
      c.x   = TOKEN_DBL (2);
      c.y   = TOKEN_DBL (3);
      c.z   = TOKEN_DBL (4);
      n = WN_new_node (d, n_num, &c);
      if (TOKENS_TOTAL() >= 7) {
        W_bool_t free[] = {TRUE,TRUE,TRUE,TRUE,TRUE,TRUE};
        free[0] = TOKEN_INT (5);
        free[1] = TOKEN_INT (6);
        free[2] = TOKEN_INT (7);
        if (TOKENS_TOTAL() >= 10) {
          free[3] = TOKEN_INT (8);
          free[4] = TOKEN_INT (9);
          free[5] = TOKEN_INT (10);
        }
        WN_fix_node (n, free);
      }
    }
  }
  return FALSE;
}


/*
   Assemble force.
 */
void
WN_assemble_force (W_node_t *n, /* node */
                   W_coords_3d_t *f
                   )
{
    n->force_t.x += f->x;
    n->force_t.y += f->y;
    n->force_t.z += f->z;
}


#if W_ROT_DOFS
/*
   Assemble moment.
 */
void
WN_assemble_moment (W_node_t *n, /* node */
                    W_coords_3d_t *m
                    )
{
  W_coords_3d_t tm;
  W_float_t p;

  /* Pull back */
  tm.x = (  m->x * n->nodal_triad_t.vect1.x
          + m->y * n->nodal_triad_t.vect1.y
          + m->z * n->nodal_triad_t.vect1.z);
  tm.y = (  m->x * n->nodal_triad_t.vect2.x
          + m->y * n->nodal_triad_t.vect2.y
          + m->z * n->nodal_triad_t.vect2.z);
  tm.z = (  m->x * n->nodal_triad_t.vect3.x
          + m->y * n->nodal_triad_t.vect3.y
          + m->z * n->nodal_triad_t.vect3.z);
  /* Transform into principal directions of the nodal moment of inertia */
  /* Assemble */
  if (n->can_move_in_dir[W_RX]) {
    p = (  tm.x * n->mom_inertia_princ_dirs.vect1.x
         + tm.y * n->mom_inertia_princ_dirs.vect1.y
         + tm.z * n->mom_inertia_princ_dirs.vect1.z);
    n->moment_t.x += p;
  }
  if (n->can_move_in_dir[W_RY]) {
    p = (  tm.x * n->mom_inertia_princ_dirs.vect2.x
         + tm.y * n->mom_inertia_princ_dirs.vect2.y
         + tm.z * n->mom_inertia_princ_dirs.vect2.z);
    n->moment_t.y += p;
  }
  if (n->can_move_in_dir[W_RZ]) {
    p = (  tm.x * n->mom_inertia_princ_dirs.vect3.x
         + tm.y * n->mom_inertia_princ_dirs.vect3.y
         + tm.z * n->mom_inertia_princ_dirs.vect3.z);
    n->moment_t.z += p;
  }
}
#endif


#if W_ROT_DOFS
/*
   Assemble angular velocity dependent load.
   Applicable to problems with rotational degrees of freedom only.
 */
void
WN_assemble_ang_vel_dep_load (W_node_t *n
                              )
{
  W_float_t v = 0;

  if (n->can_move_in_dir[W_RX]) {
    W_float_t Izz = 0, Iyy = 0;

    if (n->can_move_in_dir[W_RY]) Izz = n->mom_inertia[0];
    if (n->can_move_in_dir[W_RZ]) Iyy = n->mom_inertia[1];
    v = - (Izz - Iyy) * n->ang_velocity_t.y * n->ang_velocity_t.z;
    n->moment_t.x += v;
  }
  if (n->can_move_in_dir[W_RY]) {
    W_float_t Izz = 0, Ixx = 0;

    if (n->can_move_in_dir[W_RZ]) Izz = n->mom_inertia[2];
    if (n->can_move_in_dir[W_RX]) Ixx = n->mom_inertia[0];
    v = - (Ixx - Izz) * n->ang_velocity_t.x * n->ang_velocity_t.z;
    n->moment_t.y += v;
  }
  if (n->can_move_in_dir[W_RZ]) {
    W_float_t Iyy = 0, Ixx = 0;

    if (n->can_move_in_dir[W_RY]) Iyy = n->mom_inertia[1];
    if (n->can_move_in_dir[W_RX]) Ixx = n->mom_inertia[0];
    v = - (Iyy - Ixx) * n->ang_velocity_t.x * n->ang_velocity_t.y;
    n->moment_t.z += v;
  }
}
#endif



#if W_ROT_DOFS
static void
update_triad_Warburton_variant(W_node_t *n, W_domain_t *d, W_float_t dt);
static void
update_triad_Newmark_variant(W_node_t *n, W_domain_t *d, W_float_t dt);
#endif

static void
update_cva_Warburton_variant(W_node_t *n, W_domain_t *d, W_float_t dt);
static void
update_cva_Newmark_variant(W_node_t *n, W_domain_t *d, W_float_t dt);


/*
   Update configuration for the Warburton's variant of the central
   difference algorithm.
 */
void
WN_update_config_Warburton_variant (W_node_t *n, /* node */
                                    W_domain_t *d, /* domain */
                                    W_float_t dt /* time step */
                                    )
{
  if (n->can_move_in_dir[W_DX] ||
      n->can_move_in_dir[W_DY] ||
      n->can_move_in_dir[W_DZ]) {
    update_cva_Warburton_variant (n, d, dt);
  }
#if W_ROT_DOFS
  if (n->can_move_in_dir[W_RX] ||
      n->can_move_in_dir[W_RY] ||
      n->can_move_in_dir[W_RZ]) {
    update_triad_Warburton_variant (n, d, dt);
  }
#endif
}

/*
   Update configuration for the Newmark's variant of the central
   difference algorithm.
 */
void
WN_update_config_Newmark_variant (W_node_t *n, /* node */
                                    W_domain_t *d, /* domain */
                                    W_float_t dt /* time step */
                                    )
{
  update_cva_Newmark_variant (n, d, dt);
#if W_ROT_DOFS
  if (n->can_move_in_dir[W_RX] ||
      n->can_move_in_dir[W_RY] ||
      n->can_move_in_dir[W_RZ]) {
    update_triad_Newmark_variant (n, d, dt);
  }
#endif
}


/*
   Create a new nodal history client.
 */
void *
WN_new_history_client_data (W_node_t *n, /* node number */
                            W_index_t idx, /* id of the client */
                            W_history_writer_t *r, /* history reporter */
                            char *input_line /* input line as read
                                              from the input file */
                            )
{
  return (void *)n;
}


/*
   Update history for a node.
 */
void
WN_update_history (W_node_t *n, /* node */
                   W_solution_history_t *h, /* history surrogate */
                   W_float_t time
                  )
{
  W_node_t *cd = (W_node_t *) h->client_data;
  W_vect_t rv;
#define STORE(CD, SUFFIX, data)                 \
  {                                             \
    char name[64];                              \
    sprintf (name, "N%d%s", CD->id, SUFFIX);    \
    WHR_store (h->writer, name, time, data);    \
  }

  if (cd != n) REPORT_FATAL_ERROR (h->writer->domain->errh, "Wrong client");
  
  STORE (cd, "TDx", n->displ_t.x);
  STORE (cd, "TDy", n->displ_t.y);
  STORE (cd, "TDz", n->displ_t.z);
  STORE (cd, "TVx", n->velocity_t.x);
  STORE (cd, "TVy", n->velocity_t.y);
  STORE (cd, "TVz", n->velocity_t.z);
  STORE (cd, "TAx", n->acceleration_t.x);
  STORE (cd, "TAy", n->acceleration_t.y);
  STORE (cd, "TAz", n->acceleration_t.z);

#if W_ROT_DOFS
  STORE (cd, "RVx", 
         (  n->ang_velocity_t.x * n->mom_inertia_princ_dirs.vect1.x
            + n->ang_velocity_t.y * n->mom_inertia_princ_dirs.vect2.x
            + n->ang_velocity_t.z * n->mom_inertia_princ_dirs.vect3.x));
  STORE (cd, "RVy", 
         (  n->ang_velocity_t.x * n->mom_inertia_princ_dirs.vect1.y
            + n->ang_velocity_t.y * n->mom_inertia_princ_dirs.vect2.y
            + n->ang_velocity_t.z * n->mom_inertia_princ_dirs.vect3.y));
  STORE (cd, "RVz", 
         (  n->ang_velocity_t.x * n->mom_inertia_princ_dirs.vect1.z
            + n->ang_velocity_t.y * n->mom_inertia_princ_dirs.vect2.z
            + n->ang_velocity_t.z * n->mom_inertia_princ_dirs.vect3.z));
  
  STORE (cd, "RAx", 
         (  n->ang_acceleration_t.x * n->mom_inertia_princ_dirs.vect1.x
            + n->ang_acceleration_t.y * n->mom_inertia_princ_dirs.vect2.x
            + n->ang_acceleration_t.z * n->mom_inertia_princ_dirs.vect3.x));
  STORE (cd, "RAy", 
         (  n->ang_acceleration_t.x * n->mom_inertia_princ_dirs.vect1.y
            + n->ang_acceleration_t.y * n->mom_inertia_princ_dirs.vect2.y
            + n->ang_acceleration_t.z * n->mom_inertia_princ_dirs.vect3.y));
  STORE (cd, "RAz", 
         (  n->ang_acceleration_t.x * n->mom_inertia_princ_dirs.vect1.z
            + n->ang_acceleration_t.y * n->mom_inertia_princ_dirs.vect2.z
            + n->ang_acceleration_t.z * n->mom_inertia_princ_dirs.vect3.z));

  W_extract_rot_vector_from_triad (&n->nodal_triad_t, &rv);
  STORE (cd, "RRx", rv.x);
  STORE (cd, "RRy", rv.y);
  STORE (cd, "RRz", rv.z);
#endif
}


/*
   Prepare mass assembly.
 */
void
WN_prep_mass_assembly (W_node_t *n) { n->mass = 0; }

/*
   Assemble mass to node.
 */
void
WN_assemble_mass (W_node_t *n, /* node */
                  W_float_t mass
                  )
{
  n->mass += mass;
}



static void
update_cva_Warburton_variant (W_node_t *n, W_domain_t *d, W_float_t dt)
{
  W_float_t dt2 = dt * dt, U, A, u_t_minus_dt;

  if (n->can_move_in_dir[W_DX]) {
    A = n->force_t.x;
    U = (-n->displ_t_minus_dt.x + 2 * n->displ_t.x + dt2 * A);
    u_t_minus_dt           = n->displ_t_minus_dt.x;
    n->displ_t_minus_dt.x  = n->displ_t.x;
    n->displ_t.x           = U;
    n->acceleration_t.x    = A;
    n->velocity_t.x        = (n->displ_t.x - u_t_minus_dt) / dt * 0.5;
  } 
  if (n->can_move_in_dir[W_DY]) {
    A = n->force_t.y;
    U = (-n->displ_t_minus_dt.y + 2 * n->displ_t.y + dt2 * A);
    u_t_minus_dt           = n->displ_t_minus_dt.y;
    n->displ_t_minus_dt.y  = n->displ_t.y;
    n->displ_t.y           = U;
    n->acceleration_t.y    = A;
    n->velocity_t.y        = (n->displ_t.y - u_t_minus_dt) / dt * 0.5;
  }
  if (n->can_move_in_dir[W_DZ]) {
    A = n->force_t.z;
    U = (-n->displ_t_minus_dt.z + 2 * n->displ_t.z + dt2 * A);
    u_t_minus_dt           = n->displ_t_minus_dt.z;
    n->displ_t_minus_dt.z  = n->displ_t.z;
    n->displ_t.z           = U;
    n->acceleration_t.z    = A;
    n->velocity_t.z        = (n->displ_t.z - u_t_minus_dt) / dt * 0.5;
  }
}


#if W_ROT_DOFS
static void
update_triad_Warburton_variant (W_node_t *n, W_domain_t *d, W_float_t dt)
{
  W_float_t dt2 = dt * dt;
  W_acceleration_3d_t A = {0, 0, 0};
  W_velocity_3d_t v;
  W_coords_3d_t RI = {0, 0, 0}, ri = {0, 0, 0};
  W_triad_3d_t nt, rim;

  if (n->can_move_in_dir[W_RX]) A.x = n->moment_t.x;
  if (n->can_move_in_dir[W_RY]) A.y = n->moment_t.y;
  if (n->can_move_in_dir[W_RZ]) A.z = n->moment_t.z;
 
  /* Rotation increment in material setting */
  RI.x = dt2 * A.x;
  RI.y = dt2 * A.y;
  RI.z = dt2 * A.z;
  ri.x = (  RI.x * n->mom_inertia_princ_dirs.vect1.x
          + RI.y * n->mom_inertia_princ_dirs.vect2.x
          + RI.z * n->mom_inertia_princ_dirs.vect3.x);
  ri.y = (  RI.x * n->mom_inertia_princ_dirs.vect1.y
          + RI.y * n->mom_inertia_princ_dirs.vect2.y
          + RI.z * n->mom_inertia_princ_dirs.vect3.y);
  ri.z = (  RI.x * n->mom_inertia_princ_dirs.vect1.z
          + RI.y * n->mom_inertia_princ_dirs.vect2.z
          + RI.z * n->mom_inertia_princ_dirs.vect3.z);

  /* Rotate by the spatial counterpart */
  ri.x += n->rot_incr_t.x;
  ri.y += n->rot_incr_t.y;
  ri.z += n->rot_incr_t.z;
  W_extract_triad_from_rot_vector(&ri, &rim);
  COPY_TRIAD(nt, n->nodal_triad_t);
  W_R_shift_triad(&n->nodal_triad_t, &nt, &rim);
  
  /* Update spatial rotation increment, material ang. vel. & accel. */
  v.x =  (ri.x + n->rot_incr_t.x) * 0.5 / dt;
  v.y =  (ri.y + n->rot_incr_t.y) * 0.5 / dt;
  v.z =  (ri.z + n->rot_incr_t.z) * 0.5 / dt;
  n->rot_incr_t.x = ri.x;
  n->rot_incr_t.y = ri.y;
  n->rot_incr_t.z = ri.z;
  n->ang_velocity_t.x = (  v.x * n->mom_inertia_princ_dirs.vect1.x
                         + v.y * n->mom_inertia_princ_dirs.vect1.y
                         + v.z * n->mom_inertia_princ_dirs.vect1.z);
  n->ang_velocity_t.y = (  v.x * n->mom_inertia_princ_dirs.vect2.x
                         + v.y * n->mom_inertia_princ_dirs.vect2.y
                         + v.z * n->mom_inertia_princ_dirs.vect2.z);
  n->ang_velocity_t.z = (  v.x * n->mom_inertia_princ_dirs.vect3.x
                         + v.y * n->mom_inertia_princ_dirs.vect3.y
                         + v.z * n->mom_inertia_princ_dirs.vect3.z);
  n->ang_acceleration_t.x = A.x;
  n->ang_acceleration_t.y = A.y;
  n->ang_acceleration_t.z = A.z;
}
#endif


static void
update_cva_Newmark_variant (W_node_t *n, W_domain_t *d, W_float_t dt)
{
  double gamma = W_cd_get_gamma(d->cd_integrator);
  W_float_t dt2 = dt * dt, A;

  A = n->force_t.x;
  n->velocity_t.x        += dt * ((1 - gamma) * n->acceleration_t.x + gamma * A);
  n->displ_t_minus_dt.x  = n->displ_t.x;
  n->displ_t.x           += dt * n->velocity_t.x + 0.5 * dt2 * A;
  n->acceleration_t.x    = A;
  
  A = n->force_t.y;
  n->velocity_t.y        += dt * ((1 - gamma) * n->acceleration_t.y + gamma * A);
  n->displ_t_minus_dt.y  = n->displ_t.y;
  n->displ_t.y           += dt * n->velocity_t.y + 0.5 * dt2 * A;
  n->acceleration_t.y    = A;
  
  A = n->force_t.z;
  n->velocity_t.z        += dt * ((1 - gamma) * n->acceleration_t.z + gamma * A);
  n->displ_t_minus_dt.z  = n->displ_t.z;
  n->displ_t.z           += dt * n->velocity_t.z + 0.5 * dt2 * A;
  n->acceleration_t.z    = A;
}



#if W_ROT_DOFS
static void
update_triad_Newmark_variant (W_node_t *n, W_domain_t *d, W_float_t dt)
{
  W_float_t dt2 = dt * dt;
  W_acceleration_3d_t A = {0, 0, 0};
  W_coords_3d_t RI = {0, 0, 0};
  W_triad_3d_t nt, rim;

  if (n->can_move_in_dir[W_RX]) A.x = n->moment_t.x;
  if (n->can_move_in_dir[W_RY]) A.y = n->moment_t.y;
  if (n->can_move_in_dir[W_RZ]) A.z = n->moment_t.z;
 
  /* Update material velocity */
  n->ang_velocity_t.x += dt * 0.5 * (n->ang_acceleration_t.x + A.x);
  n->ang_velocity_t.y += dt * 0.5 * (n->ang_acceleration_t.y + A.y);
  n->ang_velocity_t.z += dt * 0.5 * (n->ang_acceleration_t.z + A.z);
    
  /* Rotation increment in material setting */
  RI.x = dt * n->ang_velocity_t.x + 0.5 * dt2 * A.x;
  RI.y = dt * n->ang_velocity_t.y + 0.5 * dt2 * A.y;
  RI.z = dt * n->ang_velocity_t.z + 0.5 * dt2 * A.z;
  n->rot_incr_t.x = (  RI.x * n->mom_inertia_princ_dirs.vect1.x
                     + RI.y * n->mom_inertia_princ_dirs.vect2.x
                     + RI.z * n->mom_inertia_princ_dirs.vect3.x);
  n->rot_incr_t.y = (  RI.x * n->mom_inertia_princ_dirs.vect1.y
                     + RI.y * n->mom_inertia_princ_dirs.vect2.y
                     + RI.z * n->mom_inertia_princ_dirs.vect3.y);
  n->rot_incr_t.z = (  RI.x * n->mom_inertia_princ_dirs.vect1.z
                     + RI.y * n->mom_inertia_princ_dirs.vect2.z
                     + RI.z * n->mom_inertia_princ_dirs.vect3.z);

  /* Rotate by the spatial counterpart */
  W_extract_triad_from_rot_vector(&n->rot_incr_t, &rim);
  COPY_TRIAD(nt, n->nodal_triad_t);
  W_R_shift_triad(&n->nodal_triad_t, &nt, &rim);
  
  /* Update material ang. accel. */
  n->ang_acceleration_t.x = A.x;
  n->ang_acceleration_t.y = A.y;
  n->ang_acceleration_t.z = A.z;
}
#endif


#if W_ROT_DOFS
/*
   Transform triad by nodal rotation matrix.
 */
void
WN_transform_by_rot_matrix_t (W_node_t *n, /* node */
                              W_triad_3d_t *in, /* input triad */
                              W_triad_3d_t *out /* output triad */
                              )
{
  W_L_shift_triad (out, &(n->nodal_triad_t), in);
}
#endif


#if W_ROT_DOFS
/*
   Assemble tensor of inertia to node.
 */
void
WN_assemble_tensor_of_inertia (W_node_t *n, /* node */
                               W_triad_3d_t *I_global
                               )
{
  /* Save temporarily in the triad: call later on WN_resolve_rot_inertia. */
  n->mom_inertia_princ_dirs.vect1.x += I_global->vect1.x;
  n->mom_inertia_princ_dirs.vect1.y += I_global->vect1.y;
  n->mom_inertia_princ_dirs.vect1.z += I_global->vect1.z;
  n->mom_inertia_princ_dirs.vect2.x += I_global->vect2.x;
  n->mom_inertia_princ_dirs.vect2.y += I_global->vect2.y;
  n->mom_inertia_princ_dirs.vect2.z += I_global->vect2.z;
  n->mom_inertia_princ_dirs.vect3.x += I_global->vect3.x;
  n->mom_inertia_princ_dirs.vect3.y += I_global->vect3.y;
  n->mom_inertia_princ_dirs.vect3.z += I_global->vect3.z;
} 
#endif


#if W_ROT_DOFS
/*
   Resolve rotary inertia of node by computing the principal directions
   and the principal values    of the nodal tensor of inertia.
 */
void
WN_resolve_rot_inertia (W_node_t *n
                        )
{
  W_index_t n_rot_freedoms = 0;
  W_float_t Iaa, Iab, Ibb, alpha, I1, I2, ca, sa, s2a;

  if (n->can_move_in_dir[W_RX]) n_rot_freedoms++;
  if (n->can_move_in_dir[W_RY]) n_rot_freedoms++;
  if (n->can_move_in_dir[W_RZ]) n_rot_freedoms++;

  if (n_rot_freedoms == 3) { /* Completely free node */
    MAT *I = NULL;
    VEC *evals = NULL;
    MAT *evecs = NULL;
    I = m_get(3, 3);
    evals = v_get(3);
    evecs = m_get(3, 3);
    I->me[0][0] = n->mom_inertia_princ_dirs.vect1.x;
    I->me[1][0] = n->mom_inertia_princ_dirs.vect1.y;
    I->me[2][0] = n->mom_inertia_princ_dirs.vect1.z;
    I->me[0][1] = n->mom_inertia_princ_dirs.vect2.x;
    I->me[1][1] = n->mom_inertia_princ_dirs.vect2.y;
    I->me[2][1] = n->mom_inertia_princ_dirs.vect2.z;
    I->me[0][2] = n->mom_inertia_princ_dirs.vect3.x;
    I->me[1][2] = n->mom_inertia_princ_dirs.vect3.y;
    I->me[2][2] = n->mom_inertia_princ_dirs.vect3.z;
    evals = symmeig(I, evecs, evals); /* Eigenproblem solution */
    n->mom_inertia[0] = evals->ve[0];
    n->mom_inertia[1] = evals->ve[1];
    n->mom_inertia[2] = evals->ve[2];
    n->mom_inertia_princ_dirs.vect1.x = evecs->me[0][0];
    n->mom_inertia_princ_dirs.vect1.y = evecs->me[1][0];
    n->mom_inertia_princ_dirs.vect1.z = evecs->me[2][0];
    n->mom_inertia_princ_dirs.vect2.x = evecs->me[0][1];
    n->mom_inertia_princ_dirs.vect2.y = evecs->me[1][1];
    n->mom_inertia_princ_dirs.vect2.z = evecs->me[2][1];
    n->mom_inertia_princ_dirs.vect3.x = evecs->me[0][2];
    n->mom_inertia_princ_dirs.vect3.y = evecs->me[1][2];
    n->mom_inertia_princ_dirs.vect3.z = evecs->me[2][2];
    m_free (I);
    m_free (evecs);
    v_free (evals);
  } else if (n_rot_freedoms == 2) { /* Free about two axes */
    if        (!n->can_move_in_dir[W_RX]) { /* Fixed about X axis */
      Iaa   = n->mom_inertia_princ_dirs.vect2.y;
      Ibb   = n->mom_inertia_princ_dirs.vect3.z;
      Iab   = n->mom_inertia_princ_dirs.vect2.z;
      alpha = atan2((2*Iab), (Ibb-Iaa)) / 2;
      ca    = cos(alpha); sa = sin(alpha); s2a = sin(2*alpha);
      I1    = Iaa * ca * ca - Iab * s2a + Ibb * sa * sa;
      I2    = Ibb * ca * ca + Iab * s2a + Iaa * sa * sa;
      Iab   = (Iaa - Ibb) * s2a / 2 + Iab * cos(2*alpha);
      n->mom_inertia[1] = I1;
      n->mom_inertia[2] = I2;
      n->mom_inertia_princ_dirs.vect1.x =   1;
      n->mom_inertia_princ_dirs.vect1.y =   0;
      n->mom_inertia_princ_dirs.vect1.z =   0;
      n->mom_inertia_princ_dirs.vect2.x =   0;
      n->mom_inertia_princ_dirs.vect2.y =  ca;
      n->mom_inertia_princ_dirs.vect2.z =  sa;
      n->mom_inertia_princ_dirs.vect3.x =   0;
      n->mom_inertia_princ_dirs.vect3.y = -sa;
      n->mom_inertia_princ_dirs.vect3.z =  ca;
    } else if (!n->can_move_in_dir[W_RY]) { /* Fixed about Y axis */
      Iaa   = n->mom_inertia_princ_dirs.vect3.z;
      Ibb   = n->mom_inertia_princ_dirs.vect1.x;
      Iab   = n->mom_inertia_princ_dirs.vect1.z;
      alpha = atan2((2*Iab), (Ibb-Iaa)) / 2;
      ca    = cos(alpha); sa = sin(alpha); s2a = sin(2*alpha);
      I1    = Iaa * ca * ca - Iab * s2a + Ibb * sa * sa;
      I2    = Ibb * ca * ca + Iab * s2a + Iaa * sa * sa;
      Iab   = (Iaa - Ibb) * s2a / 2 + Iab * cos(2*alpha);
      n->mom_inertia[2] = I1;
      n->mom_inertia[0] = I2;
      n->mom_inertia_princ_dirs.vect1.x =  ca;
      n->mom_inertia_princ_dirs.vect1.y =   0;
      n->mom_inertia_princ_dirs.vect1.z = -sa;
      n->mom_inertia_princ_dirs.vect2.x =   0;
      n->mom_inertia_princ_dirs.vect2.y =   1;
      n->mom_inertia_princ_dirs.vect2.z =   0;
      n->mom_inertia_princ_dirs.vect3.x =  sa;
      n->mom_inertia_princ_dirs.vect3.y =   0;
      n->mom_inertia_princ_dirs.vect3.z =  ca;
    } else if (!n->can_move_in_dir[W_RZ]) { /* Fixed about Z axis */
      Iaa   = n->mom_inertia_princ_dirs.vect1.x;
      Ibb   = n->mom_inertia_princ_dirs.vect2.y;
      Iab   = n->mom_inertia_princ_dirs.vect1.y;
      alpha = atan2((2*Iab), (Ibb-Iaa)) / 2;
      ca    = cos(alpha); sa = sin(alpha); s2a = sin(2*alpha);
      I1    = Iaa * ca * ca - Iab * s2a + Ibb * sa * sa;
      I2    = Ibb * ca * ca + Iab * s2a + Iaa * sa * sa;
      Iab   = (Iaa - Ibb) * s2a / 2 + Iab * cos(2*alpha);
      n->mom_inertia[0] = I1;
      n->mom_inertia[1] = I2;
      n->mom_inertia_princ_dirs.vect1.x =  ca;
      n->mom_inertia_princ_dirs.vect1.y =  sa;
      n->mom_inertia_princ_dirs.vect1.z =   0;
      n->mom_inertia_princ_dirs.vect2.x = -sa;
      n->mom_inertia_princ_dirs.vect2.y =  ca;
      n->mom_inertia_princ_dirs.vect2.z =   0;
      n->mom_inertia_princ_dirs.vect3.x =   0;
      n->mom_inertia_princ_dirs.vect3.y =   0;
      n->mom_inertia_princ_dirs.vect3.z =   1;
    }
  } else if (n_rot_freedoms == 1) { /* Free about only one axis */
    if (n->can_move_in_dir[W_RX]) n->mom_inertia[0] = n->mom_inertia_princ_dirs.vect1.x;
    if (n->can_move_in_dir[W_RY]) n->mom_inertia[1] = n->mom_inertia_princ_dirs.vect2.y;
    if (n->can_move_in_dir[W_RZ]) n->mom_inertia[2] = n->mom_inertia_princ_dirs.vect3.z;
    n->mom_inertia_princ_dirs.vect1.x = 1; /* Coincides with global c.s. */
    n->mom_inertia_princ_dirs.vect1.y = 0;
    n->mom_inertia_princ_dirs.vect1.z = 0;
    n->mom_inertia_princ_dirs.vect2.x = 0;
    n->mom_inertia_princ_dirs.vect2.y = 1;
    n->mom_inertia_princ_dirs.vect2.z = 0;
    n->mom_inertia_princ_dirs.vect3.x = 0;
    n->mom_inertia_princ_dirs.vect3.y = 0;
    n->mom_inertia_princ_dirs.vect3.z = 1;
  } else { /* Completely fixed node */
  /* VOID */
  }
}
#endif


#ifdef X_XFACE

static WCRec pnts[2] = {{0, 0, 0}, {0, 0, 0}};
static W_bool_t first = TRUE;
static EPixel ntriad_color = 0;
static EPixel essbc_color = 0;
static EPixel number_color = 0;
static GraphicObj *number_label = NULL;

static void
create_node_grep (W_node_t *n)
{
  int layer;

  pnts[0].x = n->ref_coords.x + n->displ_t.x;
  pnts[0].y = n->ref_coords.y + n->displ_t.y;
  pnts[0].z = n->ref_coords.z + n->displ_t.z;
  n->node_marker = CreateMarker3D(&pnts[0]);
  EMNoDrawAddGraphicsToModel(ESIModel(), n->node_marker);
  EASValsSetColor(ntriad_color);
  EASValsSetMType(CIRCLE_MARKER);
  EASValsSetMSize(8);
  layer = EASValsGetLayer();
  EASValsSetLayer(WX_NODE_LAYER);
  EGWithMaskChangeAttributes(COLOR_MASK |
                             MTYPE_MASK |
                             MSIZE_MASK |
                             LAYER_MASK,
                             n->node_marker);
  EASValsSetLayer(layer);
}

static void
edit_node_grep (W_node_t *n)
{
  W_bool_t is_fixed[W_NDOF_PER_NODE];
  W_bool_t with_ess_bc = FALSE;

  { int i, ndof;
  ndof = WN_is_fixed (n, is_fixed);
  FOR (i, ndof) { if (is_fixed[i]) { with_ess_bc = TRUE; break; } }
  }

  if (with_ess_bc) {
    EASValsSetColor(essbc_color);
    EASValsSetMType(SQUARE_MARKER);
  }  else {
    EASValsSetColor(ntriad_color);
    EASValsSetMType(CIRCLE_MARKER);
  }
  EGWithMaskChangeAttributes(COLOR_MASK | MTYPE_MASK, n->node_marker);
}

static void
first_time(void)
{
  int layer;

  ntriad_color = ColorGetPixelFromString("cyan", &first);
  essbc_color = ColorGetPixelFromString("orange", &first);
  number_color = ColorGetPixelFromString("white", &first);
  
  number_label = CreateAnnText3D(&pnts[0], "");
  EASValsSetColor(number_color);
  layer = EASValsGetLayer();
  EASValsSetLayer(WX_ANNOT_LAYER);
  EGWithMaskChangeAttributes(COLOR_MASK |
                             LAYER_MASK,
                             number_label);
  EASValsSetLayer(layer);
  first = FALSE;
}

#if W_ROT_DOFS
static void
make_node_triad (W_node_t *n)
{
  W_triad_3d_t *t = &n->nodal_triad_t;
  
  pnts[0].x = n->ref_coords.x + n->displ_t.x;
  pnts[0].y = n->ref_coords.y + n->displ_t.y;
  pnts[0].z = n->ref_coords.z + n->displ_t.z;
  pnts[1].x = pnts[0].x + options.triad_leg_length * t->vect1.x;
  pnts[1].y = pnts[0].y + options.triad_leg_length * t->vect1.y;
  pnts[1].z = pnts[0].z + options.triad_leg_length * t->vect1.z;
  EMNoDrawAddGraphicsToModel(ESIModel (), (n->node_t1 = CreateLine3D(pnts)));
  
  pnts[1].x = pnts[0].x + options.triad_leg_length * t->vect2.x;
  pnts[1].y = pnts[0].y + options.triad_leg_length * t->vect2.y;
  pnts[1].z = pnts[0].z + options.triad_leg_length * t->vect2.z;
  EMNoDrawAddGraphicsToModel(ESIModel (), (n->node_t2 = CreateLine3D(pnts)));
  
  pnts[1].x = pnts[0].x + options.triad_leg_length * t->vect3.x;
  pnts[1].y = pnts[0].y + options.triad_leg_length * t->vect3.y;
  pnts[1].z = pnts[0].z + options.triad_leg_length * t->vect3.z;
  EMNoDrawAddGraphicsToModel(ESIModel (), (n->node_t3 = CreateLine3D(pnts)));
  
  EASValsSetColor(ntriad_color);
  EGWithMaskChangeAttributes(COLOR_MASK, n->node_t1);
  EGWithMaskChangeAttributes(COLOR_MASK, n->node_t2);
  EGWithMaskChangeAttributes(COLOR_MASK, n->node_t3);
}
#endif


/*
   Modify node graphic representation.
 */
void
WN_modify_node_grep (W_node_t *n /* node */,
                     double displ_scale,
                     W_bool_t draw_numbers,
                     W_bool_t immed_draw
                     ) 
{
#if W_ROT_DOFS
  W_triad_3d_t *t = &n->nodal_triad_t;
#endif

  if (first) first_time();
  
#if W_ROT_DOFS
  if (options.triad_leg_length > 0) {
    if (   n->node_t1 == NULL
        || n->node_t2 == NULL
        || n->node_t3 == NULL)
      make_node_triad(n);
    else {
      pnts[0].x = n->ref_coords.x + n->displ_t.x;
      pnts[0].y = n->ref_coords.y + n->displ_t.y;
      pnts[0].z = n->ref_coords.z + n->displ_t.z;

      pnts[1].x = pnts[0].x + options.triad_leg_length * t->vect1.x;
      pnts[1].y = pnts[0].y + options.triad_leg_length * t->vect1.y;
      pnts[1].z = pnts[0].z + options.triad_leg_length * t->vect1.z;
      EGModifyGraphicsGeometry(n->node_t1, pnts);
      
      pnts[1].x = pnts[0].x + options.triad_leg_length * t->vect2.x;
      pnts[1].y = pnts[0].y + options.triad_leg_length * t->vect2.y;
      pnts[1].z = pnts[0].z + options.triad_leg_length * t->vect2.z;
      EGModifyGraphicsGeometry(n->node_t2, pnts);
      
      pnts[1].x = pnts[0].x + options.triad_leg_length * t->vect3.x;
      pnts[1].y = pnts[0].y + options.triad_leg_length * t->vect3.y;
      pnts[1].z = pnts[0].z + options.triad_leg_length * t->vect3.z;
      EGModifyGraphicsGeometry(n->node_t3, pnts);
    }
  }
#endif

  if (n->node_marker == NULL) {
    create_node_grep(n);
  } else {
    pnts[0].x = (n->ref_coords.x + displ_scale * (n->displ_t.x));
    pnts[0].y = (n->ref_coords.y + displ_scale * (n->displ_t.y));
    pnts[0].z = (n->ref_coords.z + displ_scale * (n->displ_t.z));
    EGModifyGraphicsGeometry(n->node_marker, &pnts[0]);
    edit_node_grep (n);
    if (draw_numbers && immed_draw) {
      char buf[32];
      pnts[0].x = n->ref_coords.x + n->displ_t.x;
      pnts[0].y = n->ref_coords.y + n->displ_t.y;
      pnts[0].z = n->ref_coords.z + n->displ_t.z;
      EGModifyGraphicsGeometry(number_label, pnts);
      sprintf(buf, " %ld", n->id);
      EASValsSetText(buf);
      EGWithMaskChangeAttributes(TEXT_STRING_MASK,
                                 number_label);        
      EMDrawGraphics(ESIModel (), number_label);
    }
  }
}


#endif


/*
   Get node kinetic energy.
 */
W_float_t /* energy */
WN_node_kinetic_energy (W_node_t *n
                        )
{
  W_float_t e = 0, v;

  if (n->can_move_in_dir[W_DX]) {
    v = n->velocity_t.x;
    e += v * v * n->mass;
  }
  if (n->can_move_in_dir[W_DY])  {
    v = n->velocity_t.y;
    e += v * v * n->mass;
  }
  if (n->can_move_in_dir[W_DZ]) {
    v = n->velocity_t.z;
    e += v * v * n->mass;
  }
#if W_ROT_DOFS
  if (n->can_move_in_dir[W_RX]) {
    v = n->ang_velocity_t.x;
    e += v * v * n->mom_inertia[0];
  }
  if (n->can_move_in_dir[W_RY])  {
    v = n->ang_velocity_t.y;
    e += v * v * n->mom_inertia[1];
  }
  if (n->can_move_in_dir[W_RZ]) {
    v = n->ang_velocity_t.z;
    e += v * v * n->mom_inertia[2];
  } 
#endif

  return e / 2;
}





#if defined(PVM_WASP_WORKER) || defined(PVM_WASP_MASTER)

void
WN_pack_interface_vector (W_node_t *n, /* node */
                          W_vector_t to_pack,/* pack from this vector */
                          double packed[6] /* packed data */
                          )
{
  W_index_t ne;
  
  if ((ne = n->eq_numbers[W_DX]) > 0) packed[0] = to_pack[ne];
  else                                packed[0] = 0;
  if ((ne = n->eq_numbers[W_DY]) > 0) packed[1] = to_pack[ne];
  else                                packed[1] = 0;
  if ((ne = n->eq_numbers[W_DZ]) > 0) packed[2] = to_pack[ne];
  else                                packed[2] = 0;
  if ((ne = n->eq_numbers[W_RX]) > 0) packed[3] = to_pack[ne];
  else                                packed[3] = 0;
  if ((ne = n->eq_numbers[W_RY]) > 0) packed[4] = to_pack[ne];
  else                                packed[4] = 0;
  if ((ne = n->eq_numbers[W_RZ]) > 0) packed[5] = to_pack[ne];
  else                                packed[5] = 0;
}

/*
   Unpack interface package.
 */
void
WN_unpack_interface_vector(W_node_t *n, /* node */
                           double packed[6], /* packed data */
                           W_vector_t unpack_into /* unpack into this vector */
                           )
{
  W_index_t ne;
  
  if ((ne = n->eq_numbers[W_DX]) > 0) unpack_into[ne] = packed[0];
  if ((ne = n->eq_numbers[W_DY]) > 0) unpack_into[ne] = packed[1];
  if ((ne = n->eq_numbers[W_DZ]) > 0) unpack_into[ne] = packed[2];
  if ((ne = n->eq_numbers[W_RX]) > 0) unpack_into[ne] = packed[3];
  if ((ne = n->eq_numbers[W_RY]) > 0) unpack_into[ne] = packed[4];
  if ((ne = n->eq_numbers[W_RZ]) > 0) unpack_into[ne] = packed[5];
}

#endif /* if defined(PVM_WASP_WORKER) || defined(PVM_WASP_MASTER) */



#ifdef WASP_IMPLICIT
/*
   Implicit initialization of motion quantities.
 */
void
WN_impl_init (W_node_t *n, /* node */
              W_domain_t *d /* domain */
              )
{
  W_index_t ne;

  if ((ne = n->eq_numbers[W_DX]) > 0) {
    WD_impl_set_avd (d, ne,
                    n->acceleration_t.x,
                    n->velocity_t.x);
  } 
  if ((ne = n->eq_numbers[W_DY]) > 0) {
    WD_impl_set_avd (d, ne,
                    n->acceleration_t.y,
                    n->velocity_t.y);
  }
  if ((ne = n->eq_numbers[W_DZ]) > 0) {
    WD_impl_set_avd (d, ne,
                    n->acceleration_t.z,
                    n->velocity_t.z);
  }
}
#endif


#ifdef WASP_IMPLICIT
/*
   Implicit Newmark predictor.
   Not yet implemented.
 */
void
WN_impl_predict (W_node_t *n, /* node */
                 W_domain_t *d, /* domain */
                 W_float_t dt /* time step */
                 )
{
  W_index_t ne;
  double a, v, dthalf = dt / 2;

  if ((ne = n->eq_numbers[W_DX]) > 0) {
    a                   = n->acceleration_t.x;
    n->acceleration_t.x = WD_get_from_update_vect(d, ne);
    v                   = n->velocity_t.x;
    n->velocity_t.x     = v + dthalf * (n->acceleration_t.x + a);
    n->displ_t.x       = (  n->displ_t_minus_dt.x
                           + dthalf * (n->velocity_t.x + v)); 
  } 
  if ((ne = n->eq_numbers[W_DY]) > 0) {
    a                   = n->acceleration_t.y;
    n->acceleration_t.y = WD_get_from_update_vect(d, ne);
    v                   = n->velocity_t.y;
    n->velocity_t.y     = v + dthalf * (n->acceleration_t.y + a);
    n->displ_t.y       = (  n->displ_t_minus_dt.y
                           + dthalf * (n->velocity_t.y + v));
  }
  if ((ne = n->eq_numbers[W_DZ]) > 0) {
    a                   = n->acceleration_t.z;
    n->acceleration_t.z = WD_get_from_update_vect(d, ne);
    v                   = n->velocity_t.z;
    n->velocity_t.z     = v + dthalf * (n->acceleration_t.z + a);
    n->displ_t.z       = (  n->displ_t_minus_dt.z
                           + dthalf * (n->velocity_t.z + v));
  }
}
#endif /* WASP_IMPLICIT */


/*
   Print info for node.
 */
void
WN_print_info (W_node_t *n, /* node */
              W_domain_t *d, /* domain */
              W_info_spec_t spec, /* specification */
              FILE *stream, /* output stream */
               int indent /* optional indent in terms of spaces */
               )
{
  W_INDENT(stream, indent);
  fprintf(stream, "Node # %ld\n", n->id);
  indent += 3;
#define PRINT_VECTOR(name, member)                      \
  {                                                     \
    W_INDENT(stream, indent);                           \
    fprintf(stream,  name " = {%g, %g, %g}\n",          \
            n->member.x, n->member.y, n->member.z);     \
  }
  PRINT_VECTOR("Ref. location       ", ref_coords);
  PRINT_VECTOR("Displacement        ", displ_t);
  PRINT_VECTOR("Prev. displacement  ", displ_t_minus_dt);
  PRINT_VECTOR("Velocity            ", velocity_t);
  PRINT_VECTOR("Acceleration        ", acceleration_t);
#if W_ROT_DOFS
  PRINT_VECTOR("Triad vector #1     ", nodal_triad_t.vect1);
  PRINT_VECTOR("Triad vector #2     ", nodal_triad_t.vect2);
  PRINT_VECTOR("Triad vector #3     ", nodal_triad_t.vect3);
  PRINT_VECTOR("Angular velocity    ", ang_velocity_t);
  PRINT_VECTOR("Angular acceleration", ang_acceleration_t);
  PRINT_VECTOR("Mom. of inertia princ dir #1",
               mom_inertia_princ_dirs.vect1);
  PRINT_VECTOR("Mom. of inertia princ dir #2",
               mom_inertia_princ_dirs.vect2);
  PRINT_VECTOR("Mom. of inertia princ dir #3",
               mom_inertia_princ_dirs.vect3);
#endif
  W_INDENT(stream, indent);
  fprintf(stream, "Associated mass = %g\n", n->mass);
}


/*
   Solve for accelerations.
 */
void
WN_solve_for (W_node_t *n)
{
  double invm;
  if (n->mass > 0) {
    invm = 1 / n->mass;
    n->force_t.x *= invm;
    n->force_t.y *= invm;
    n->force_t.z *= invm;
  } else {
    n->force_t.x = n->force_t.y = n->force_t.z = 0;
  }
#if W_ROT_DOFS
  {
    if (n->mom_inertia[0] > 0) {
      n->moment_t.x /= n->mom_inertia[0];
    } else {
      n->moment_t.x = 0;
    }
    if (n->mom_inertia[1] > 0) {
      n->moment_t.y /= n->mom_inertia[1];
    } else {
      n->moment_t.y = 0;
    }
    if (n->mom_inertia[2] > 0) {
      n->moment_t.z /= n->mom_inertia[2];
    } else {
      n->moment_t.z = 0;
    }
  }
#endif
}

void
WN_set_force (W_node_t *n, W_vect_t *a, W_bool_t apply[3])
{
  if (apply[0]) n->force_t.x = a->x;
  if (apply[1]) n->force_t.y = a->y;
  if (apply[2]) n->force_t.z = a->z;
}

void
WN_set_acceleration (W_node_t *n, W_acceleration_3d_t *a, W_bool_t apply[3])
{
  if (apply[0]) n->acceleration_t.x = a->x;
  if (apply[1]) n->acceleration_t.y = a->y;
  if (apply[2]) n->acceleration_t.z = a->z;
}

void
WN_set_velocity (W_node_t *n, W_vect_t *a, W_bool_t apply[3])
{
  if (apply[0]) n->velocity_t.x = a->x;
  if (apply[1]) n->velocity_t.y = a->y;
  if (apply[2]) n->velocity_t.z = a->z;
}

void
WN_set_displ (W_node_t *n, W_vect_t *a, W_bool_t apply[3])
{
  if (apply[0]) n->displ_t.x = a->x;
  if (apply[1]) n->displ_t.y = a->y;
  if (apply[2]) n->displ_t.z = a->z;
}

void
WN_set_prev_displ (W_node_t *n, W_vect_t *a, W_bool_t apply[3])
{
  if (apply[0]) n->displ_t_minus_dt.x = a->x;
  if (apply[1]) n->displ_t_minus_dt.y = a->y;
  if (apply[2]) n->displ_t_minus_dt.z = a->z;
}

/*
   Get nodal mass.
 */
double
WN_nodal_mass (W_node_t *n) { return n->mass; }

/*
   Fix node.
 */
void
WN_fix_node (W_node_t *n, /* node */
             W_bool_t is_free[W_NDOF_PER_NODE] /* array of boolean flags;
                                                degree of freedom  is not
                                                constrained if (is_free[i]==TRUE) */
) 
{
  n->can_move_in_dir[0] = is_free[0];
  n->can_move_in_dir[1] = is_free[1];
  n->can_move_in_dir[2] = is_free[2];
  
#if W_ROT_DOFS
  n->can_move_in_dir[3] = is_free[3];
  n->can_move_in_dir[4] = is_free[4];
  n->can_move_in_dir[5] = is_free[5];
#endif
}

/*
  Is the node fixed?
  Is the node constrained by some essential boundary condition?
 */
int
WN_is_fixed (W_node_t *n, /* node */
             W_bool_t is_fixed[W_NDOF_PER_NODE] /* boolean flags for each degree of freedom */
             )
{
  is_fixed[0] = ! n->can_move_in_dir[0];
  is_fixed[1] = ! n->can_move_in_dir[1];
  is_fixed[2] = ! n->can_move_in_dir[2];
#if W_ROT_DOFS
  is_fixed[3] = ! n->can_move_in_dir[3];
  is_fixed[4] = ! n->can_move_in_dir[4];
  is_fixed[5] = ! n->can_move_in_dir[5];
#endif
  return W_NDOF_PER_NODE;
}

/*
   Assemble damping force for node.
 */
void
WN_assemble_damping (W_node_t *n, /* node */
                     double c,  /* damping coefficient */
                     double dt /* time step */
                     ) 
{
  if (n->can_move_in_dir[W_DX]) { 
    n->force_t.x -= c * n->mass * n->velocity_t.x;
  }                                              
  if (n->can_move_in_dir[W_DY]) { 
    n->force_t.y -= c * n->mass * n->velocity_t.y;
  }                                              
  if (n->can_move_in_dir[W_DZ]) { 
    n->force_t.z -= c * n->mass * n->velocity_t.z;
  }                                              
} 


/*
   Prepare for a force assembly.
   Zero the force vector.
 */
void
WN_prep_force_assembly (W_node_t *n)
{
  n->force_t.x = n->force_t.y = n->force_t.z = 0;
#if W_ROT_DOFS
  n->moment_t.x = n->moment_t.y = n->moment_t.z = 0;
#endif
}

void
WN_get_force (W_node_t *n, double *x, double *y, double *z
              )
{
  *x = n->force_t.x;
  *y = n->force_t.y;
  *z = n->force_t.z;
}


/*
   Write TECPLOT file for node.
   The data that get written are determined by the global
   options.postfile_output_spec variable. 
 */
void
WN_write_tecplot_file (W_node_t *n, /* node */
                       FILE *stream, /* output stream */
                       int output_spec,
                       W_index_t *count /* index of current output point;
                                         this number is incremented inside
                                         and *count is therefore the number
                                         of the next TECPLOT point after
                                         this function completes
                                         */
                       )
{
  fprintf(stream, "%ld ", *count);
  if (output_spec & W_POSTFILE_DISPL_MASK) {
    fprintf(stream, "%g ", n->displ_t.x);
    fprintf(stream, "%g ", n->displ_t.y);
    fprintf(stream, "%g ", n->displ_t.z);
  }
  if (output_spec & W_POSTFILE_STRESS_MASK ||
      output_spec & W_POSTFILE_VON_MISES_MASK) {
    double S[3][3] = {{0,0,0},{0,0,0},{0,0,0}};
    fprintf(stream, "%g ", S[0][0]);
    fprintf(stream, "%g ", S[0][1]);
    fprintf(stream, "%g ", S[0][2]);
    fprintf(stream, "%g ", S[1][1]);
    fprintf(stream, "%g ", S[1][2]);
    fprintf(stream, "%g ", S[2][2]);
  }
  if (output_spec & W_POSTFILE_EQPLS_MASK) {
    fprintf(stream, "%g ", 0.0);
  }
  if (output_spec & W_POSTFILE_VON_MISES_MASK) {
    fprintf(stream, "%g ", 0.0);
  }
  fprintf(stream, "\n");
  (*count)++;
}

/*
  Get equation numbers associated with the node degrees of freedom.
 */
int /* returns number of dofs per node */
WN_eqn_numbers_and_constraints (W_node_t *n,
                                W_index_t ens[W_NDOF_PER_NODE],
                                W_bool_t constrained[W_NDOF_PER_NODE]
                                )
{
  W_index_t base;
  int i;
  base = n->internal_id * W_NDOF_PER_NODE;
  FOR (i, W_NDOF_PER_NODE) {
    ens[i]         = base + i;
    constrained[i] = ! n->can_move_in_dir[i];
  }
  return W_NDOF_PER_NODE;
}

double
WN_extf_norm_squared (W_node_t *n)
{
  double f, efn = 0;
  f = n->force_t.x;
  efn += f*f;
  f = n->force_t.y;
  efn += f*f;
  f = n->force_t.z;
  efn += f*f;
#if W_ROT_DOFS
  f = n->moment_t.x;
  efn += f*f;
  f = n->moment_t.y;
  efn += f*f;
  f = n->moment_t.z;
  efn += f*f;
#endif
  return efn;
}

double
WN_inertf_norm_squared (W_node_t *n)
{
  double f, efn = 0;
  f = n->acceleration_t.x * n->mass;
  efn += f*f;
  f = n->acceleration_t.y * n->mass;
  efn += f*f;
  f = n->acceleration_t.z * n->mass;
  efn += f*f;
#if W_ROT_DOFS
  f = n->ang_acceleration_t.x * n->mom_inertia[0];
  efn += f*f;
  f = n->ang_acceleration_t.y * n->mom_inertia[1];
  efn += f*f;
  f = n->ang_acceleration_t.z * n->mom_inertia[2];
  efn += f*f;
#endif
  return efn;
}
