#include "Wconfig.h"
#include "getoptP.h"
#include "Wxf.h"
#include "Wtypes.h"
#include "Wcdtypes.h"
#include "Wutil.h"
#include "Wpvm.h"
#include "Whistry.h"
#include "Wmodr.h"
#include "Wnewmarki.h"
#include "Wcdi.h"
#if defined (MPI_WASP) && MPI_WASP /* parallel execution with MPI */
# include "Wmpi.h"
#endif
#include "Woptions.h"

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include <sys/stat.h> 

/* RCSHEADER "$Id: options.c,v 1.7 2005/02/16 15:06:28 cvs Exp $" */

#define WOPT_DFN_ACCESS_OPT_VALUE_FUNC(o, type, optname)        \
   WOPT_DCL_ACCESS_OPT_VALUE_FUNC(o, type, optname)             \
   {                                                            \
     return o->optname;                                         \
   }
#undef WOPT_LIST_ITEM
#define WOPT_LIST_ITEM(o,type,optname,value) WOPT_DFN_ACCESS_OPT_VALUE_FUNC(o, type, optname)
WOPT_LIST

#define WOPT_DFN_SET_OPT_VALUE_FUNC(o, type, optname)           \
   WOPT_DCL_SET_OPT_VALUE_FUNC(o, type, optname)                \
   {                                                            \
     o->optname = arg;                                          \
   }
#undef WOPT_LIST_ITEM
#define WOPT_LIST_ITEM(o,type,optname,value) WOPT_DFN_SET_OPT_VALUE_FUNC(o, type, optname)
WOPT_LIST



static void
mark_option_as_interactive (W_options_t *o, char *option)
{
  if (o->cl_options_found == NULL) {
    o->cl_options_found_table_dim = 1;
    o->cl_options_found = w_new_array (o->cl_options_found_table_dim, char *);
  } else {
    o->cl_options_found_table_dim++;
    o->cl_options_found = w_realloc_array (o->cl_options_found,
                                           o->cl_options_found_table_dim, char *);
  }
  CHECK_MALLOC (o->cl_options_found);

  DUPLICATE_STRING (o->cl_options_found[o->cl_options_found_idx], option);
  o->cl_options_found_idx++;
}


static void
create_postfile_times_list (W_options_t *o)
{
  int nt, i, t;
  int kp;
  static tokens_parser parser = NULL;
  W_postfile_times_list_t *tl = &o->postfile_times_list;
  char *list_string = o->postfile_times;

  if (!parser) parser = tokens_new_parser ();
  
  tokens_parse_line (parser, list_string);
  if ((kp = tokens_have_keyword(parser, 1, "SEQUENCE")) > 0) {
    double from = 0, to = o->target_time, step = 0, t;
    if (tokens_total_of_tokens (parser) >= kp+1)
      if (tokens_type_of_token (parser, kp+1)  == TOKENS_NUMBER)
        from = tokens_token_as_double (parser, kp+1);
    if (tokens_total_of_tokens (parser) >= kp+2)
      if (tokens_type_of_token (parser, kp+2) == TOKENS_NUMBER)
        to = tokens_token_as_double (parser, kp+2);
    if (tokens_total_of_tokens (parser) >= kp+3)
      if (tokens_type_of_token (parser, kp+3) == TOKENS_NUMBER)
        step = tokens_token_as_double (parser, kp+3);
    if (to < from) { double tmp = from; from = to; to = tmp; }
    if (step <= 0)
      step = (to - from) / 10;
    tl->ntimes = (int)((to - from) / step) + 1;
    tl->curr = 0;
    tl->times = w_new_array (tl->ntimes, double);
      CHECK_MALLOC (tl->times);
    t = from;
    for(i = 0; i < tl->ntimes; i++) {
      tl->times[i] = t;
      t += step;
    }
  } else {
    nt = tokens_total_of_tokens (parser);
    tl->curr = tl->ntimes = 0;
    for(i = 1; i <= nt; i++) {
      if (tokens_type_of_token (parser, i)  == TOKENS_NUMBER)
        tl->ntimes++;
      else
        REPORT_FATAL_ERROR (NULL, "Bad times list");
    }
    if (tl->ntimes > 0) {
      tl->times = w_new_array(tl->ntimes, double);
      CHECK_MALLOC (tl->times);
      t = 0;
      for(i = 1; i <= nt; i++) {
        if (tokens_type_of_token (parser, i) == TOKENS_NUMBER) {
          tl->times[t++] = tokens_token_as_double (parser, i);
        }
      }
    }
  }
}

static void
create_postfile_output_specs (W_options_t *o)
{
  int i;
  static tokens_parser parser = NULL;
  W_postfile_output_specs_t *os = &o->postfile_output_specs_list;
  char *list_string = o->postfile_output_specs;

  if (!parser) parser = tokens_new_parser ();
  
  tokens_parse_line (parser, list_string);
  os->nspecs = tokens_total_of_tokens (parser);
  os->specs = w_new_array (os->nspecs, char *);
  
  for (i = 1; i <= os->nspecs; i++) {
    const char *s = tokens_token_as_string (parser, i);
    DUPLICATE_STRING (os->specs[i-1], s);
  }
}

static void
strip_path (char *dest, char *src)
{
  char *p;
  
  p = strrchr(src, '/');
  if (p == NULL) sprintf(dest, "%s", src);
  else {
    sprintf(dest, "%s", ++p);
  }
}

static void
strip_suffix (char *dest, char *src, char *suffix)
{
  if (!ckit_io_file_has_suffix (src, suffix)) {
    sprintf (dest, "%s", src);
    return;
  } else {
    int i, n;
    n = strlen (src) - strlen (suffix);
    for (i = 0; i < n; i++) dest[i] = src[i];
    dest[n] = '\0';
  }
}

#undef SET_BOOL
#define SET_BOOL(name)                                                  \
  {                                                                     \
    if (o->report_opts) { fprintf (stderr, "Set: %s\n", #name); }  \
      o->name = TRUE;                                              \
  }
#undef SET_WITH_ARG_NUM
#define SET_WITH_ARG_NUM(name, format)                                  \
  {                                                                     \
    if (o->report_opts) {                                          \
      if        (strcmp (#format, "int") == 0) {                        \
        int _i_ = (int)o->name;                                    \
        fprintf (stderr, "Set: %s = %d\n", #name, _i_);                 \
      } else if (strcmp (#format, "double") == 0 ||                     \
                 strcmp (#format, "float") == 0) {                      \
        double  _d_ = (double)o->name;                             \
        fprintf (stderr, "Set: %s = %g\n", #name, _d_);                 \
      }                                                                 \
    }                                                                   \
  }
#undef SET_WITH_ARG_STR
#define SET_WITH_ARG_STR(name)                                          \
  {                                                                     \
    if (o->report_opts) {                                          \
      fprintf (stderr, "Set: %s = %s\n", #name, o->name);          \
    }                                                                   \
  }

#undef  GET_OPT_BOOL
#define GET_OPT_BOOL(argc, argv, opt, todo_when_set, usage)             \
  {                                                                     \
    int how_many_occurrences =                                          \
      ckit_opt_is_set (argc, argv, opt, usage, 1);                      \
    if (how_many_occurrences == 1) {                                    \
      todo_when_set;                                                    \
      mark_option_as_interactive (o, opt);                              \
    } else if (how_many_occurrences > 1) {                              \
      sprintf (W_buf (), "More than one occurrence of option %s", opt); \
      REPORT_FATAL_ERROR (NULL, W_buf ());                              \
    }                                                                   \
  }
#define GET_OPT_BOOL_(opt, optvar, usage) \
   GET_OPT_BOOL(argc, argv, opt, SET_BOOL(optvar), ": bool; " usage)

#define BAD_ARG(opt, arg_desc)                                          \
  {                                                                     \
    fprintf(stderr, "Bad argument.  Usage: %s arg, where arg = "        \
            arg_desc "\n", opt);                                        \
    exit (EXIT_FAILURE);                                                \
  }
#undef  GET_OPT_ARG
#define GET_OPT_ARG(argc, argv, opt, type, var, todo_when_set, arg_desc)        \
  {                                                                             \
    int how_many_occurrences =                                                  \
      ckit_opt_is_set (argc, argv, opt,                                         \
                       "arg: arg = " arg_desc, 1);                              \
    if (how_many_occurrences == 1) {                                            \
      if (ckit_opt_has_arg ()) {                                                \
        if (!ckit_opt_##type##_val (&o->var))                                   \
          BAD_ARG(opt, arg_desc);                                               \
      } else {                                                                  \
        BAD_ARG(opt, arg_desc);                                                 \
      }                                                                         \
      todo_when_set;                                                            \
      {                                                                         \
        mark_option_as_interactive (o, opt);                                    \
      }                                                                         \
    } else if (how_many_occurrences > 1) {                                      \
      sprintf (W_buf (), "More than one occurrence of option %s", opt);         \
      REPORT_FATAL_ERROR (NULL, W_buf ());                                      \
    }                                                                           \
  }
#define GET_OPT_ARG_INT_(opt, optvar, usage)    \
   GET_OPT_ARG(argc, argv, opt, int, optvar,   \
               SET_WITH_ARG_NUM(optvar, int), usage)
#define GET_OPT_ARG_DBL_(opt, optvar, usage)            \
   GET_OPT_ARG(argc, argv, opt, double, optvar,        \
               SET_WITH_ARG_NUM(optvar, double), usage)
#define GET_OPT_ARG_STR_(opt, optvar, usage)          \
   GET_OPT_ARG(argc, argv, opt, char, optvar,        \
               SET_WITH_ARG_STR(optvar), usage)

#if defined (MPI_WASP) && MPI_WASP
/* generate name of input file for a partition;
 the template is: input:  xxx(suffix) 
                  output: xxx-nnnn(suffix)
 where xxx = base input file name, and
       nnnn = zero-padded number of partition

 examples:
  
 twoel.in -> twoel-0000.in
 twoel    -> twoel-0000
*/
static void
gen_partition_file_name (char *dest, char *src, char *suffix)
{
  int rank;
  
  MPI_Comm_rank (MPI_COMM_WORLD, &rank);

  if (ckit_io_file_has_suffix (src, suffix)) {
    char *name_wo_suffix;
    name_wo_suffix = w_new_array (strlen (src) + 1 + strlen (suffix), char);
    CHECK_MALLOC (name_wo_suffix);
    sprintf (name_wo_suffix, "%s", src);
    name_wo_suffix[strlen (src) - strlen (suffix)] = '\0';
    sprintf (dest, "%s-%04d%s", name_wo_suffix, rank, suffix);
    w_free (name_wo_suffix);
  } else {
    sprintf (dest, "%s-%04d", src, rank);
  }
}
#endif

/*
  Create new options record.
 */
W_options_t * /* non-NULL if successful */
WO_new_options (void)
{
  W_options_t *o;
  W_postfile_times_list_t default_times_list = {
    0,
    0,
    NULL
  };
  W_postfile_output_specs_t default_specs_list = {
    0,
    NULL
  };
  
  o = w_new (W_options_t); CHECK_MALLOC (o);
  
  /* Init private fields */
  o->cl_options_found_table_dim = 0;
  o->cl_options_found_idx       = 0;
  o->cl_options_found           = NULL;

  /* Init option fields */
#define WOPT_INIT_VALUE(optname, value) o->optname = value
#undef WOPT_LIST_ITEM
#define WOPT_LIST_ITEM(o,type,optname,value) WOPT_INIT_VALUE(optname, value)
  WOPT_LIST

  return o;
}

static char *
expand_string (char *str, char *to_append)
{
  int n = strlen(str) + strlen(" ") + strlen(to_append) + 1;
  char *new_string = w_new_array(n, char);
  sprintf(new_string, "%s %s", str, to_append);
  w_free (str);
  return new_string;
}

/*
  Get option values from command-line arguments.
 */
W_bool_t /* TRUE if successful */
WO_get_options (W_options_t *o, /* options */
                int argc, /* number of arguments */
                char **argv /* array of arguments */
                ) 
{
  int help = FALSE;
  char buf[MAXFILENAMECHAR];

  o->argc = argc;
  o->argv = argv;
  o->argcp = &o->argc;
  o->argvp = &o->argv;
  
  GET_OPT_BOOL_("-report_opts", report_opts,
                "prints options as they are being set");
  
  GET_OPT_BOOL_("-help", help,
                "print help on command line options");
  
#if defined (MPI_WASP) && MPI_WASP
  GET_OPT_BOOL_("-serial", serial,
                "run as a serial program (default is parallel)");
#endif
  
  GET_OPT_ARG_STR_("-i", in_file_name, "input file");
  if (o->in_file_name == NULL)
    GET_OPT_ARG_STR_("-input", in_file_name, "input file");
  if (o->in_file_name == NULL && !o->help) {
    fprintf (stderr, "The input file name must be supplied\n");
    exit (EXIT_FAILURE);
  }
#if defined (MPI_WASP) && MPI_WASP
  if (o->in_file_name && (!o->serial)) {
    gen_partition_file_name (buf, o->in_file_name, ".in");
    DUPLICATE_STRING (o->in_file_name, buf);
  }
#endif
  if (o->in_file_name != NULL) {
    strip_suffix (buf, o->in_file_name, ".in");
    sprintf (o->in_file_name, "%s", buf);
    strip_path (buf, o->in_file_name);
    DUPLICATE_STRING (o->in_file_basename, buf);
  }
  
  GET_OPT_ARG_STR_("-W", hist_file_name,
                   "history file name (default name identical "
                   "with in_file_basename)");
  if (o->hist_file_name != NULL) {
    strip_path (buf, o->hist_file_name);
    w_free (o->hist_file_name);
    DUPLICATE_STRING (o->hist_file_name, buf);
  } else if (o->in_file_basename) {
    DUPLICATE_STRING (o->hist_file_name, o->in_file_basename);
  }

  GET_OPT_ARG_STR_("-E", error_log_name,
                   "error log name (default name identical "
                   "with in_file_name.err)");
  if (o->error_log_name == NULL && o->in_file_basename) {
    sprintf (buf, "%s.err", o->in_file_basename);
    DUPLICATE_STRING (o->error_log_name, buf);
  } 
  
  GET_OPT_ARG_STR_("-L", log_name,
                   "log name (default is `not given' = log to stderr)");
#if defined (MPI_WASP) && MPI_WASP
  /* if not given, log file name has to be deduced for parallel execution */
  if (o->log_name == NULL && (!o->serial) && o->in_file_basename) {
    sprintf (buf, "%s.log", o->in_file_basename);
    DUPLICATE_STRING (o->log_name, buf);
  } 
#endif

  GET_OPT_ARG_STR_("-expeval_init", expeval_init,
                   "expeval initialization expression");
    
  GET_OPT_ARG_INT_("-integrator", integrator,
                   "integrator (default 0=central difference algorithm; 1=implicit Newmark ");
  
  GET_OPT_ARG_INT_("-integrator_detail", integrator_detail,
                   "for -integrator 0: 0=default, Warburton=1, Newmark=2;\n"
                   "for -integrator 1: 0=default, same_u=1, same_v=2, same_a=3, zero_a=4");
  
#if defined(PVM_WASP_MASTER) 
  o->no_output = TRUE;
#else
  o->no_output = FALSE;
#endif
  GET_OPT_BOOL_("-n", no_output,
                "no output (default for PVM master)");
  
  GET_OPT_ARG_STR_("-o", out_file_name,
                   "out_file (default file in_file_name.out)");
  
#ifdef X_XFACE
  GET_OPT_BOOL_("-X", X_xface,
                "use X Window System graphical interface (default is don't)");
  
  GET_OPT_ARG_DBL_("-x", displ_scale,
                   "displacement scale for plotting (default is 1)");
  if (o->displ_scale <= 0) o->displ_scale = 1;
  
  GET_OPT_ARG_DBL_("-f", force_scale,
                   "force scale for plotting (default is 1)");
  if (o->force_scale <= 0) o->force_scale = 1;
  
  GET_OPT_ARG_STR_("-elixir_cmds", elixir_cmd_file,
                   "Elixir command file (no default file)");

  GET_OPT_BOOL_("-gather_frames", gather_frames,
                "gather frame images for movies (default is don't)");
  
  GET_OPT_ARG_STR_("-frame_format", frame_format,
                   "frame image format (default JPG)");

  GET_OPT_ARG_DBL_("-frame_dt", frame_dt,
                   "time between frames (default is the same as -O)");
  
#endif

  GET_OPT_BOOL_("-v", verbose,
                "verbose (produce log, default is don't)");
  if (!WOPT_OPT (o, verbose)) {
    GET_OPT_BOOL_("-verbose", verbose,
                  "verbose (produce log, default is don't)");
  }
  
  GET_OPT_ARG_DBL_("-T", target_time,
                   "target time (default is 1.0)");
  
  GET_OPT_ARG_DBL_("-O", time_between_outputs,
                   "time between outputs (default is 0)");
  
  GET_OPT_ARG_DBL_("-S", time_step,
                   "time step (default is compute from element data)");
  
  GET_OPT_ARG_DBL_("-R", time_step_reduction,
                   "time step reduction factor (default is 1: no reduction)");

  GET_OPT_ARG_INT_("-R_func", time_step_reduction_func,
                   "time step reduction factor function (default is 0: no function is specified)");
  
  GET_OPT_BOOL_("-e", color_encoded_forces,
                "color encode internal forces (default is don't)");
  
  GET_OPT_ARG_STR_("-modr_options", modr_options,
                   "Model reduction options [use | collect][method direct|snapshots][nmodes #][nm #][max_nm #][modenl #][dt #][up_to_t #][from_t #][cutoff_ev #]");

  GET_OPT_ARG_STR_("-newmarki_options", newmarki_options,
                   "options for the Newmark integrator [solver spetsc|smeschach|dmeschach][nz_per_row #][modnewton][report_iter][pass_on_max][max_iter #][conv_ratio #][recomp_keff_after #][gamma #][modr][adaptive_modr][nitadapt #]");
  
  GET_OPT_ARG_STR_("-cdi_options", cdi_options,
                   "options for the explicit centered difference integrator [gamma #]");
  
  GET_OPT_BOOL_("-!", careless,
                "careless: overwrites history w/o warning (default is don't)");
  
  GET_OPT_BOOL_("-I", interactive_during_outputs,
                "GUI-interactive during outputs (default is not)");
  
#ifdef X_XFACE
  GET_OPT_BOOL_("-t", track_nodes,
                "track position of nodes by not erasing (default is don't)");
  
  GET_OPT_ARG_DBL_("-N", triad_leg_length,
                   "length of node triad leg (default is 0; triads not drawn)");
#endif
  GET_OPT_ARG_INT_("-d", task_type,
                   "task type: dimension (default 3D=0, plane s/s=1)");
  if (o->task_type != TASK_PLANE_STRESS_STRAIN &&
      o->task_type != TASK_3D) {
    o->task_type = TASK_3D;
    REPORT_WARNING (NULL, "Unknown task; resetting to 3D");
  }
  
#ifdef PVM_WASP_STUFF
  GET_OPT_BOOL_("-D", debug_pvm,
                "debug PVM version (default is don't)");
#endif
  
  GET_OPT_BOOL_("-P", running_patch_test,
                "run a patch test (a single time step) (default is don't)");
  
  GET_OPT_ARG_INT_("-p", postfile_type,
                   "post-processing file type "
                   "(default 1=HEXVUE, 2=VE3D, 3=TECPLOT, 4=HEXVUE w/ extra capabilities)");
  if (o->postfile_type != HEXVUE_POSTPROC_FILE &&
      o->postfile_type != VE3D_POSTPROC_FILE   &&
      o->postfile_type != TECPLOT_I_POSTPROC_FILE &&
      o->postfile_type != HEXVUE_XTRA_POSTPROC_FILE
      ) {
    REPORT_WARNING (NULL, "Unknown post-processing file type; resetting to HEXVUE");
    o->postfile_type = HEXVUE_POSTPROC_FILE;
  }
  
  GET_OPT_ARG_INT_("-post_file_type", postfile_type,
                   "post-processing file type "
                   "(default 1=HEXVUE, 2=VE3D, 3=TECPLOT, 4=HEXVUE w/ extra capabilities)");
  if (o->postfile_type != HEXVUE_POSTPROC_FILE &&
      o->postfile_type != VE3D_POSTPROC_FILE   &&
      o->postfile_type != TECPLOT_I_POSTPROC_FILE &&
      o->postfile_type != HEXVUE_XTRA_POSTPROC_FILE
      ) {
    REPORT_WARNING (NULL, "Unknown post-processing file type; resetting to HEXVUE");
    o->postfile_type = HEXVUE_POSTPROC_FILE;
  }
  
  GET_OPT_ARG_STR_("-hexvue_format", hexvue_format,
                   "HEXVUE format (ascii, binary, default compressed)");
  if (! (
         (strcmp (o->hexvue_format, "compressed") == 0) ||
         (strcmp (o->hexvue_format, "binary") == 0) ||
         (strcmp (o->hexvue_format, "ascii") == 0)
         ))
    o->hexvue_format = "compressed";
  
  GET_OPT_ARG_STR_("-output_specs", postfile_output_specs,
                   "list of output quantities");
  if (o->postfile_output_specs != NULL) {
    create_postfile_output_specs (o);
  }
  
  { /* This is to catch usage of the -h option (which is deprecated) */
    int ul = o->postfile_output_spec_deprec;
    GET_OPT_ARG_INT_("-h", postfile_output_spec_deprec,
                     "post-processing options (add "
                     "1=displacements (default), "
                     "2=stresses, "
                     "4=equiv.pl.strain, "
                     "8=von Mises stress, "
                     "16=material id)");
    o->postfile_output_spec_deprec |= ul;
  } 
  
  GET_OPT_ARG_STR_("-H", postfile_times,
                   "produce post-processing file at times in list arg");
  if (o->postfile_times != NULL) {
    create_postfile_times_list (o);
  }
  
  GET_OPT_BOOL_("-number_seq", postfile_number_seq,
                "number post-processing files sequentially (default is mark by time)");
  
  GET_OPT_BOOL_("-user_labels", postfile_user_labels,
                "label post-processing files with user-specified times (default is simulation time)");
  
  GET_OPT_ARG_INT_("-+", optimization_level,
                   "optimize speed (default 0=no, >0=yes)");
  
  GET_OPT_ARG_INT_("-b", hist_buff_length,
                   "history-buffer length");
  if (o->hist_buff_length > MAXHISTBUFFLNGTH) {
    REPORT_WARNING(NULL, "Too large history buffer");
    o->hist_buff_length = MAXHISTBUFFLNGTH;
  }
  
  GET_OPT_ARG_INT_("-s", output_stress_type,
                   "stress output (default 0=Cauchy, 2=2nd PK)");
  if (!(   o->output_stress_type == W_OUTPUT_STRESS_CAUCHY
        || o->output_stress_type == W_OUTPUT_STRESS_2ND_PK
        )) {
    REPORT_WARNING (NULL, "unknown stress type; resetting to Cauchy");
    o->output_stress_type = W_OUTPUT_STRESS_CAUCHY;
  }
  
  GET_OPT_ARG_INT_("-q", q_func_type,
                   "J-int domain q-function type (default "
                   "0=quadratic, "
                   "1=quartic, "
                   "2=linear)");
  if (o->q_func_type != W_QUADRATIC_Q_FUNC &&
      o->q_func_type != W_QUARTIC_Q_FUNC &&
      o->q_func_type != W_LINEAR_Q_FUNC &&
      o->q_func_type != W_COMPOSITE_QUADRATIC_Q_FUNC
      ) {
    REPORT_WARNING (NULL, "unknown q-func type; resetting to composite quadratic");
    o->q_func_type = W_COMPOSITE_QUADRATIC_Q_FUNC;
  }
  
  GET_OPT_ARG_INT_("-J", j_dom_eval,
                   "J-int evaluation ("
                   "0=per cell weak-form Gauss pts, "
                   "default 1=per cell Gauss pt interpolation)");
  if (o->j_dom_eval != W_PER_CELL_WF_PTS &&
      o->j_dom_eval != W_PER_CELL_IPOL_PTS) {
    REPORT_WARNING (NULL, "unknown J-dom eval type; resetting to per cell");
    o->j_dom_eval = W_PER_CELL_WF_PTS;
  }
  
  GET_OPT_ARG_INT_("-K", sifs_from,
                   "Compute SIF's from ("
                   "default 0=interaction integrals, "
                   "1=J-integrals (works for KI only!))");
  if (o->sifs_from != W_SIFS_FROM_II_INTEGRAL &&
      o->sifs_from != W_SIFS_FROM_J_INTEGRAL) {
    REPORT_WARNING(NULL, "unknown SIF eval type; resetting to interaction integral");
    o->sifs_from = W_SIFS_FROM_II_INTEGRAL;
  }
  
  GET_OPT_ARG_DBL_("-M", rayleigh_mass_damping_coeff,
                   "Rayleigh mass-proportional damping coeff "
                   "(default 0=no damping");
  
  GET_OPT_ARG_INT_("-m", mass_damping_time_func,
                   "Rayleigh mass-prop. damping time function "
                   "(default 0=no time dependency");
  
  GET_OPT_ARG_INT_("-mass_type", mass_type,
                   "mass type (default 1=consistent, 0=diagonal)");
  
  GET_OPT_ARG_STR_("-hexvue_xtra_opts", hexvue_xtra_opts,
                   "HEXVUE XTRA options (\"matid #\")");

  if (o->postfile_output_specs_list.nspecs == 0) { /* not set by -output_specs */
    unsigned int ul = o->postfile_output_spec_deprec;
    o->postfile_output_specs = w_new_array(1, char);
    if (ul & W_POSTFILE_DISPL_MASK) {
      o->postfile_output_specs = expand_string (o->postfile_output_specs, "displ_x");
      o->postfile_output_specs = expand_string (o->postfile_output_specs, "displ_y");
      o->postfile_output_specs = expand_string (o->postfile_output_specs, "displ_z");
    }
    if (ul & W_POSTFILE_STRESS_MASK) {
      if (o->output_stress_type == W_OUTPUT_STRESS_2ND_PK) {
        o->postfile_output_specs = expand_string (o->postfile_output_specs, "S_xx");
        o->postfile_output_specs = expand_string (o->postfile_output_specs, "S_xy");
        o->postfile_output_specs = expand_string (o->postfile_output_specs, "S_xz");
        o->postfile_output_specs = expand_string (o->postfile_output_specs, "S_yy");
        o->postfile_output_specs = expand_string (o->postfile_output_specs, "S_yz");
        o->postfile_output_specs = expand_string (o->postfile_output_specs, "S_zz");
      } else {
        o->postfile_output_specs = expand_string (o->postfile_output_specs, "s_xx");
        o->postfile_output_specs = expand_string (o->postfile_output_specs, "s_xy");
        o->postfile_output_specs = expand_string (o->postfile_output_specs, "s_xz");
        o->postfile_output_specs = expand_string (o->postfile_output_specs, "s_yy");
        o->postfile_output_specs = expand_string (o->postfile_output_specs, "s_yz");
        o->postfile_output_specs = expand_string (o->postfile_output_specs, "s_zz");
      } 
    }
    if (ul & W_POSTFILE_EQPLS_MASK)
      o->postfile_output_specs = expand_string (o->postfile_output_specs, "eqpls");
    if (ul & W_POSTFILE_VON_MISES_MASK)
      o->postfile_output_specs = expand_string (o->postfile_output_specs, "vonm");
    if (ul & W_POSTFILE_MATERIAL_MASK)
      o->postfile_output_specs = expand_string (o->postfile_output_specs, "mat");
    if (ul & W_POSTFILE_DEF_ENERGY_DENSITY_MASK)
      o->postfile_output_specs = expand_string (o->postfile_output_specs, "defed");
    create_postfile_output_specs (o);
  }
  return TRUE;
}

static W_bool_t
option_marked_interactive (W_options_t *o, char *option)
{
  int i;

  for (i = 0; i < o->cl_options_found_idx; i++)
    if (strcmp (o->cl_options_found[i], option) == 0)
      return TRUE;
  return FALSE;
}

/*
  Options block reader.
  Read options block for given domain.
 */
W_bool_t /* TRUE if success; FALSE otherwise  */
W_options_reader (W_domain_t *d /* domain */
                  )
{
  const char *input_line;
  char option[256];
  W_options_t *o = d->options;
#define OPTCASE(opt) (strcmp (option, #opt) == 0)

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line)) {
      return TRUE;
    } else {
      
      TOKENS_PARSE(input_line);
      if (!TOKEN_IS_STRING(1)) {
        sprintf(W_buf(), "Bad option %s", tokens_token_as_string(d->parser, 1));
        REPORT_FATAL_ERROR(NULL, W_buf());
      } else {
        const char *s;
        s = tokens_token_as_string(d->parser, 1);
        if (s[0] != '-')
          sprintf (option, "-%s", s);
        else
          sprintf (option, "%s", s);
      }
       
      if (!option_marked_interactive (o, option)) {
        /* cannot override command line */
        if        (OPTCASE (-T)) {
          o->target_time = TOKEN_DBL(2);
        } else if (OPTCASE (-O)) {
          o->time_between_outputs = TOKEN_DBL(2);
        } else if (OPTCASE (-S)) {
          o->time_step = TOKEN_DBL(2);
        } else if (OPTCASE (-expeval_init)) {
          int stat;
          double f;
          o->expeval_init = (char *) TOKEN_STR(2);
          stat = expeval_eval (WOPT_OPT (d->options, expeval_init), &f);
          if (stat != EXPEVAL_E_OK) {
            sprintf (W_buf (), "%s: \"%s\"", expeval_err_msgs[stat],
                     WOPT_OPT (d->options, expeval_init));
            REPORT_FATAL_ERROR (d->errh, W_buf ());
          }
        } else if (OPTCASE (-R)) {  
          o->time_step_reduction = TOKEN_DBL(2);
        } else if (OPTCASE (-R_func)) {  
          o->time_step_reduction_func = TOKEN_INT(2);
        } else if (OPTCASE (-d)) {  
          o->task_type = (W_task_type_t) TOKEN_INT(2);
        } else if (OPTCASE (-b)) {  
          o->hist_buff_length = TOKEN_INT(2);
        } else if (OPTCASE (-h)) {   
          if (o->postfile_output_specs_list.nspecs == 0) {
            unsigned int ul = TOKEN_INT(2);
            o->postfile_output_spec_deprec |= ul;
            o->postfile_output_specs = w_new_array(1, char);
            if (ul & W_POSTFILE_DISPL_MASK) {
              o->postfile_output_specs = expand_string (o->postfile_output_specs, "displ_x");
              o->postfile_output_specs = expand_string (o->postfile_output_specs, "displ_y");
              o->postfile_output_specs = expand_string (o->postfile_output_specs, "displ_z");
            }
            if (ul & W_POSTFILE_STRESS_MASK) {
              if (o->output_stress_type == W_OUTPUT_STRESS_2ND_PK) {
                o->postfile_output_specs = expand_string (o->postfile_output_specs, "S_xx");
                o->postfile_output_specs = expand_string (o->postfile_output_specs, "S_xy");
                o->postfile_output_specs = expand_string (o->postfile_output_specs, "S_xz");
                o->postfile_output_specs = expand_string (o->postfile_output_specs, "S_yy");
                o->postfile_output_specs = expand_string (o->postfile_output_specs, "S_yz");
                o->postfile_output_specs = expand_string (o->postfile_output_specs, "S_zz");
              } else {
                o->postfile_output_specs = expand_string (o->postfile_output_specs, "s_xx");
                o->postfile_output_specs = expand_string (o->postfile_output_specs, "s_xy");
                o->postfile_output_specs = expand_string (o->postfile_output_specs, "s_xz");
                o->postfile_output_specs = expand_string (o->postfile_output_specs, "s_yy");
                o->postfile_output_specs = expand_string (o->postfile_output_specs, "s_yz");
                o->postfile_output_specs = expand_string (o->postfile_output_specs, "s_zz");
              } 
            }
            if (ul & W_POSTFILE_EQPLS_MASK)
              o->postfile_output_specs = expand_string (o->postfile_output_specs, "eqpls");
            if (ul & W_POSTFILE_VON_MISES_MASK)
              o->postfile_output_specs = expand_string (o->postfile_output_specs, "vonm");
            if (ul & W_POSTFILE_MATERIAL_MASK)
              o->postfile_output_specs = expand_string (o->postfile_output_specs, "mat");
            if (ul & W_POSTFILE_DEF_ENERGY_DENSITY_MASK)
              o->postfile_output_specs = expand_string (o->postfile_output_specs, "defed");
            create_postfile_output_specs (o);
          }
        } else if (OPTCASE (-hexvue_format)) {
          DUPLICATE_STRING(o->hexvue_format, TOKEN_STR(2));
          if (! (
                 (strcmp (o->hexvue_format, "compressed") == 0) ||
                 (strcmp (o->hexvue_format, "binary") == 0) ||
                 (strcmp (o->hexvue_format, "ascii") == 0)
                 ))
            o->hexvue_format = "compressed";
        } else if (OPTCASE (-hexvue_xtra_opts)) {
          DUPLICATE_STRING(o->hexvue_xtra_opts, TOKEN_STR(2));
        } else if (OPTCASE (-output_specs)) {
          DUPLICATE_STRING (o->postfile_output_specs, TOKEN_STR(2));
          create_postfile_output_specs (o);
        } else if (OPTCASE (-number_seq)) {
          o->postfile_number_seq = TRUE;
#ifdef X_XFACE
        } else if (OPTCASE (-x)) {   
          o->displ_scale = TOKEN_DBL(2);
        } else if (OPTCASE (-f)) {   
          o->force_scale = TOKEN_DBL(2);
        } else if (OPTCASE (-N)) {   
          o->triad_leg_length = TOKEN_DBL(2);
        } else if (OPTCASE (-elixir_cmds)) { 
          DUPLICATE_STRING (o->elixir_cmd_file, TOKEN_STR(2));
          WXF_process_elixir_cmds (o->elixir_cmd_file);
#endif
        } else if (OPTCASE (-M)) {   
          o->rayleigh_mass_damping_coeff = TOKEN_DBL(2); 
        } else if (OPTCASE (-m)) {    
          o->mass_damping_time_func = TOKEN_INT(2); 
        } else if (OPTCASE (-mass_type)) {    
          o->mass_type = (W_mass_type_enum) TOKEN_INT(2);
          if (!(o->mass_type == 0 || o->mass_type == 1)) {
            sprintf (W_buf (), "Unknown mass type: %d", o->mass_type);
            REPORT_FATAL_ERROR (NULL, W_buf());
          }
        } else if (OPTCASE (-q)) {   
          o->q_func_type = (W_q_func_enum)  TOKEN_INT(2);
        } else if (OPTCASE (-J)) {   
          o->j_dom_eval = (W_j_dom_eval_enum) TOKEN_INT(2);
          if (o->j_dom_eval != W_PER_CELL_WF_PTS &&
              o->j_dom_eval != W_PER_CELL_IPOL_PTS)
            o->j_dom_eval = W_PER_CELL_WF_PTS;
          break;
        } else if (OPTCASE (-W)) { 
          {
            char buf[MAXFILENAMECHAR];
            
            DUPLICATE_STRING (o->hist_file_name, TOKEN_STR(2));
            strip_path(buf, o->hist_file_name);
            DUPLICATE_STRING (o->hist_file_name, buf);
          }
        } else if (OPTCASE (-H)) { 
          {
            DUPLICATE_STRING (o->postfile_times, TOKEN_STR(2));
            create_postfile_times_list (o);
          }
        } else if (OPTCASE (-post_file_type)) {
          o->postfile_type = TOKEN_INT(2);
          if (o->postfile_type != HEXVUE_POSTPROC_FILE &&
              o->postfile_type != VE3D_POSTPROC_FILE   &&
              o->postfile_type != TECPLOT_I_POSTPROC_FILE &&
              o->postfile_type != HEXVUE_XTRA_POSTPROC_FILE
              ) {
            REPORT_WARNING (NULL, "Unknown post-processing file type; resetting to HEXVUE");
            o->postfile_type = HEXVUE_POSTPROC_FILE;
          }
        } else {
          sprintf (W_buf (), "Unknown option: %s", option);
          REPORT_WARNING (NULL, W_buf ());
        }
      } /*if (!option_marked_interactive(option_letter))  */
    }
  } /* while */

  return FALSE;
}


