#include "Wconfig.h"

#include "Wtypes.h"
#include "Wpcontpt.h"
#include "Wdomain.h"
#include "Wnode.h"
#include "Wmats.h"
#include "Wgaussp.h"
#include "Wxf.h"
#include "Wutil.h"
#include "Wcdi.h"
#include "g3d.h"
#include "Wmat3d.h"
#include "Wcontact.h"
#include "Wtfuncs.h"

/* RCSHEADER "$Id: pcontpt.c,v 1.9 2004/09/06 01:17:58 cvs Exp $" */

/* for computing with geometric tolerances  */
#define MACHEPS 1e-16

typedef struct elem_t {
  W_contact_surf_t             *cs;
  W_dict_t                     * node_dict;
  }           elem_t;


static int
elem_elcode (W_elem_t *el)
{
  return PCONTPT_ELCODE;
}


static char *
type_name (W_elem_t *el)
{
  static char buf[] = "pcontpt";
  return buf;
}

#define GET_COORD(n, vec) WN_get_coords (n, &(vec.x), &(vec.y), &(vec.z))
#define GET_PREV_COORD(n, vec) WN_get_prev_coords (n, &(vec.x), &(vec.y), &(vec.z))
#define GET_VELOC(n, vec) WN_get_velocity (n, &(vec.x), &(vec.y), &(vec.z))


static void
assemble_restoring_forces (W_elem_t *el, W_mat_state_update_form_enum update_form)
{
  elem_t *pe = (elem_t *) el->data;
  WPC_spheres_penalty_contact_t *pc = pe->cs->data.spheres_penalty;
  WPC_spheres_sphere_t *s = 0;
  
  while (s = (WPC_spheres_sphere_t *) get_list_next (pc->sl, s)) {
    W_node_t *n = 0;
    double rtfv = WTF_func_val(s->radius_tf, WD_time(pc->cont_surf->domain));
    double xtfv = WTF_func_val(s->x_tf, WD_time(pc->cont_surf->domain));
    double ytfv = WTF_func_val(s->y_tf, WD_time(pc->cont_surf->domain));
    double ztfv = WTF_func_val(s->z_tf, WD_time(pc->cont_surf->domain));
    s->radius = rtfv;
    if (s->radius < 0) {
      REPORT_WARNING(pc->cont_surf->domain->errh, "Radius < 0");
      s->radius = 0;
    }
    s->center.x = xtfv;
    s->center.y = ytfv;
    s->center.z = ztfv;
    while (n = (W_node_t *)  W_dict_next_obj (pe->node_dict, n)) {
      W_vect_t d;
      W_loc_t c;
      double dist, overlap=0, overlap_rate=0;
      GET_COORD(n, c);
      VECMAC_3_VEC(d, =, c, -, s->center);
      dist = VECMAC_NORM(d);
      if (dist < s->radius) {
        W_vect_t of={0,0,0};
        W_vect_t orf={0,0,0};
        W_vect_t f={0,0,0};
        if (pc->force_type == W_PENALTY_OVERLAP_RATE ||
            pc->force_type == W_PENALTY_OVERLAP_COMB) {
          W_velocity_3d_t v;
          GET_VELOC(n, v);
          overlap_rate = (-1)* VECMAC_DOT(v, d)/dist;
          VECMAC_2_VEC (orf, =, (overlap_rate / dist * pc->overlap_rate_penalty) * d);
        }
        
        if (pc->force_type == W_PENALTY_OVERLAP ||
            pc->force_type == W_PENALTY_OVERLAP_COMB) {
          overlap=s->radius-dist;
          VECMAC_2_VEC (of, =, (overlap / dist * pc->overlap_penalty) * d);
        }
        
        VECMAC_3_VEC (f, =, of, +, orf);
        VECMAC_2_VEC (f, =, (n->mass/pc->avgnmass)*f); /* scale force to account for different masses */
        WN_assemble_force (n, &f);
      }
    }
  }
}

static int 
num_of_faces (W_elem_t *el)
{
  return 0;
}


static void
assemble_residual_forces (W_elem_t *el,
                          W_mat_state_update_form_enum update_form,
                          W_mass_type_enum mass_type)
{
  assemble_restoring_forces (el, update_form);
}

#ifdef X_XFACE
#define ELIXIR_FILE_HEADER  "# ELIXIR file (C) 1994 Petr Krysl; version "
static int
elixir_file_version(FILE *fp)
{
#define LINE_LEN  132
  char line[LINE_LEN];
  int version;
  
  fgets(line, LINE_LEN-1, fp);
  if (sscanf(line, ELIXIR_FILE_HEADER "%d", &version) != 1) {
    /* put back what you have read */
    rewind(fp);
    return 0;
  } else
    return version;
}

static void
modify_elem_grep(W_elem_t *el, double displ_scale, W_bool_t immed_draw)
{
  elem_t *pe = (elem_t *)el->data;
  WPC_spheres_penalty_contact_t *pc = pe->cs->data.spheres_penalty;
  WPC_spheres_sphere_t *s =0;
  
  while (s = (WPC_spheres_sphere_t *) get_list_next (pc->sl, s)) {
    double rtfv = WTF_func_val(s->radius_tf, WD_time(pc->cont_surf->domain));
    double xtfv = WTF_func_val(s->x_tf, WD_time(pc->cont_surf->domain));
    double ytfv = WTF_func_val(s->y_tf, WD_time(pc->cont_surf->domain));
    double ztfv = WTF_func_val(s->z_tf, WD_time(pc->cont_surf->domain));
    s->radius = rtfv;
    if (s->radius < 0) {
      REPORT_WARNING(pc->cont_surf->domain->errh, "Radius < 0");
      s->radius = 0;
    }
    s->center.x = xtfv;
    s->center.y = ytfv;
    s->center.z = ztfv;
    if (s->ogrepl == NULL) {
      if (s->egf_file) {
        FILE *  fp = 0;
        OnDiskSetFileName(s->egf_file);
        fp = OnDiskOpenAGEFile("r");
        s->grepl = 0;
        s->ogrepl = make_list();
        if (fp == NULL) {
          REPORT_WARNING(pc->cont_surf->domain->errh, "Could not open sphere grep file");
        } else {
          GraphicObj *p;
          EGSetFileVersion(elixir_file_version(fp));
          do {
            p = EGGraphicRetrieveFrom(fp);
            if (p != NULL) {
              if (EGGraphicType(p) == EG_CTLBLOCK) {
              } else {
                add_to_tail(s->ogrepl, p);
              }
            }
            if (feof(fp))
              break;
          } while (p != NULL);
        }
      }
    }
    if (s->ogrepl) {
      if (s->grepl == NULL) {
        GraphicObj *og=0;
        s->grepl=make_list();
        while (og=(GraphicObj*)get_list_next(s->ogrepl, og)) {
          GraphicObj *g=EGDeepCopyGraphics(og);
          /* EASValsSetLayer (WX_BCS_LAYER); */
          EGWithMaskChangeAttributes(LAYER_MASK, g);
          EMNoDrawAddGraphicsToModel(ESIModel (), g);
          add_to_tail(s->grepl, g);
        }
      }
    }
    if (s->grepl) {
      GraphicObj *g=0;
      GraphicObj *og=0;
      while (g=(GraphicObj*)get_list_next(s->grepl, g)) {
        WCRec points[1000];
        int npoints=1000;
        og = (GraphicObj*)get_list_next(s->ogrepl, og);
        if (EGGetGraphicsGeometry(og, points, &npoints)) {
          int i;
          FOR(i, npoints) {
            VECMAC_2_VEC(points[i],=,s->radius*points[i]);
            VECMAC_2_VEC(points[i],+=,s->center);
          }
          EGModifyGraphicsGeometry(g, points);
        }
      }
    }
  }
}
#endif

static W_elem_methods_t elem_methods = {
  elem_elcode,
  type_name,
  assemble_restoring_forces,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
#ifdef X_XFACE
  modify_elem_grep,
#endif
  NULL,
  NULL,
  NULL,
  num_of_faces,
  NULL,
  NULL,
  assemble_residual_forces,
  NULL
};

/*
  Element reader.
  Reads element data for the pcontpt element type
  from stream as long as there is no
  END_OF_BLOCK tag in the input line. Creates elements.
 */
W_bool_t /* TRUE if successful */
pcontpt_reader (W_domain_t *d /* domain */)
{
  const char *input_line;
  long elem_num;
  W_node_t *n[3];
  W_elem_t *el;
  int inx;
  static W_elem_methods_t pcont3_methods;
  int max_elem_num;
#ifdef PVM_NEED_GHOST_ELEMENTS
  int is_ghost = FALSE;
#endif
  int i;
  W_index_t nodeid, cont_surf_id;
  W_contact_surf_t *cs = NULL;

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line)) {
      return TRUE;
    } else {
      TOKENS_PARSE(input_line);
      if (TOKENS_HAVE_KEYWORD ("NATURAL_FACES")) {
        elem_t *pe = 0;
        LIST_OF (WCF_contact_face_t *) cfaces = 0;
        cont_surf_id = TOKEN_INT (2);
        cs = WD_get_cont_surf_by_idx (d, cont_surf_id);
        if (cs == NULL) {
          sprintf(W_buf(), "No such contact surface: %ld", cont_surf_id);
          REPORT_ERROR(d->errh, W_buf());
        }
        if (cs->type != W_SPHERES_PENALTY_CONTACT_SURF) {
          sprintf(W_buf(), "Element must belong to a spherical penalty contact");
          REPORT_FATAL_ERROR (d->errh, W_buf ());
        }
        pe = w_new(elem_t);
        CHECK_MALLOC(pe);
        pe->cs = cs;
        pe->node_dict =  W_dict_new_dictionary ();
        CHECK_MALLOC(pe->node_dict);
        cfaces = WD_get_list_of_contact_faces (d);
        if (list_length (cfaces) > 0) {
          WCF_contact_face_t *cf;
          max_elem_num = cont_surf_id*11000000;
          while (WD_get_elem_by_idx (d, max_elem_num++)) /* VOID */;
          cf = NULL;
          while ((cf = (WCF_contact_face_t *) get_list_next (cfaces, cf))) {
            for (inx=0; inx< cf->num_nodes; inx++) {
	      W_dict_add(pe->node_dict, cf->node[inx], cf->node[inx]->id); 
            }
          }
        }
 
         WD_add_element (d, max_elem_num,
                        el = WE_new_element (d, &elem_methods, pe
#                                   ifdef PVM_NEED_GHOST_ELEMENTS
                                             , is_ghost
#                                   endif
                                             ));
        free_list (cfaces, AND_NODES);
      } else {
        REPORT_FATAL_ERROR(d->errh, "No contact points defined");
      }
    } /* closing else */ 
  }
  return FALSE;
}
