#include "Wconfig.h"
#include "Wtypes.h"
#include "Wutil.h"
#include "Wcdi.h"
#include "Wdomain.h"
#include "Wnode.h"
#include "Welem.h"
#include "Wxf.h"
#include "Wcontact.h"

/* RCSHEADER "$Id: pencont.c,v 1.4 2004/04/27 20:10:11 cvs Exp $" */

static void
build_search_structure (WPC_faceted_penalty_contact_t *pc)
{
  W_elem_t *el;
#define MAXNODES 4
  W_node_t *n[MAXNODES];
  int i, num_nodes;
  W_coords_3d_t loc[MAXNODES];
  G3D_box_t box;

  if (pc->cont_surf->all_rigid && pc->lobs) return;

  if (pc->lobs)  pc->lobs = lobs_3d_recycle_instance (pc->lobs);
  if (!pc->lobs) pc->lobs = lobs_3d_new_instance (NULL, 10);
   
  el = NULL;
  while ((el = (W_elem_t *) get_list_next (pc->cont_surf->celems, el))) {
    num_nodes = WE_get_face_vertex_nodes (el, 0, MAXNODES, n);
    if (num_nodes > MAXNODES) {
      REPORT_FATAL_ERROR (pc->cont_surf->domain->errh,
                          "Can handle only 3- or 4-noded contact faces");
    }
    FOR (i, num_nodes) {
      WN_get_coords (n[i], &(loc[i].x), &(loc[i].y), &(loc[i].z));
      if (i == 0) { G3D_INIT_BOX (box, loc[0]); }
      else        { G3D_UPDT_BOX (box, loc[i]); }
    }
    FOR (i, num_nodes) {
      WN_get_prev_coords (n[i], &(loc[i].x), &(loc[i].y), &(loc[i].z));
      G3D_UPDT_BOX (box, loc[i]); 
    }
#if 0                           /* don't know if this is needed */
    G3D_INFLATE_BOX (box, pc->cont_surf->search_eps);
#endif
    lobs_3d_add_object (pc->lobs, el->data,
                        box.lower.x, box.lower.y, box.lower.z,
                        box.upper.x, box.upper.y, box.upper.z);
  }
}

  
void
WPCF_search (WPC_faceted_penalty_contact_t *pc)
{
  build_search_structure (pc);

  if (pc->avgnmass <= 0)
    pc->avgnmass = WD_avg_nodal_mass (pc->cont_surf->domain);
}

W_contact_surf_t *
read_faceted_penalty_surf (W_domain_t *d)
{
  const char *input_line;
  W_contact_surf_t *cont_surf = NULL;

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line)) {
      return NULL;
    } else {
      TOKENS_PARSE (input_line);
      if (TOKENS_HAVE_KEYWORD ("ID")) {
        cont_surf = w_new (W_contact_surf_t);
        CHECK_MALLOC (cont_surf);
        cont_surf->id                       = TOKEN_INT (2);
        cont_surf->domain                   = d;
        cont_surf->type                     = W_FACETED_PENALTY_CONTACT_SURF;
        cont_surf->celems                   = make_list ();
        cont_surf->search_eps               = FLT_MAX;
        cont_surf->all_rigid                = TRUE;

        cont_surf->data.faceted_penalty             = w_new (WPC_faceted_penalty_contact_t);
        CHECK_MALLOC (cont_surf->data.faceted_penalty);
        cont_surf->data.faceted_penalty->cont_surf            = cont_surf;
        cont_surf->data.faceted_penalty->lobs                 = NULL;
        cont_surf->data.faceted_penalty->first_time           = TRUE;
        cont_surf->data.faceted_penalty->force_type           = W_PENALTY_OVERLAP;
        cont_surf->data.faceted_penalty->overlap_penalty      = 1;
        cont_surf->data.faceted_penalty->overlap_rate_penalty = 1;
        cont_surf->data.faceted_penalty->avgnmass             = 0;
        cont_surf->data.faceted_penalty->subdiv_level         = 1;
      } else if (TOKENS_HAVE_KEYWORD ("SUBDIV_LEVEL")) {
        cont_surf->data.faceted_penalty->subdiv_level = TOKEN_INT (2);
      } else if (TOKENS_HAVE_KEYWORD ("FORCE_TYPE")) {
        if        (TOKENS_HAVE_KEYWORD ("RATE")) {
          cont_surf->data.faceted_penalty->force_type = W_PENALTY_OVERLAP_RATE;
        } else if (TOKENS_HAVE_KEYWORD ("COMB")) {
          cont_surf->data.faceted_penalty->force_type = W_PENALTY_OVERLAP_COMB;
        } else {
          cont_surf->data.faceted_penalty->force_type = W_PENALTY_OVERLAP;
        }
      } else if (TOKENS_HAVE_KEYWORD ("OVERLAP_PENALTY")) {
        cont_surf->data.faceted_penalty->overlap_penalty = TOKEN_DBL (2);
      } else if (TOKENS_HAVE_KEYWORD ("OVERLAP_RATE_PENALTY")) {
        cont_surf->data.faceted_penalty->overlap_rate_penalty = TOKEN_DBL (2);
      } else if (TOKENS_HAVE_KEYWORD ("END")) {
        return cont_surf;
      }
    }
  }

 

  return cont_surf;
}
