
#include "Wconfig.h"
#include "Wxf.h"
#include "Wtypes.h"
#include "Wutil.h"
#include "Wmpi.h"
#include "Wmat3dutil.h"

#include <stdio.h>
#include <stdlib.h>


/* RCSHEADER "$Id: polar.c,v 1.1.1.1 2001/07/24 02:42:51 pkrysl Exp $" */

#ifndef min
#   define min(A,B) (((A) < (B)) ? (A) : (B))
#endif
#ifndef max
#   define max(A,B) (((A) > (B)) ? (A) : (B))
#endif

#undef sign
#define sign(mag, sgn) (fabs(mag) * (sgn > 0? +1: -1))

#ifndef TRUE
#  define FALSE 0
#  define TRUE (!FALSE)
#endif

/*
    This function computes the eigenvalues of a 3x3        
    symmetric positive definite matrix in stress vector storage.
*/
static void
eigenvalues (double K[6], double LAMDA[3]) 
{
  double M[6], kbari, kbarj, kbar, ki, kj, mi, mj,
    scale, alpha, gamma, x, xsign, rad, swap,
    krat, sqtol, thresh;
  int  sweep, iexp;
  const int maxsweep = 15;
  const double  jactol = 1.0e-08, SMALL = 0.0001;
#define k(i) K[i-1]
#define lamda(i) LAMDA[i-1]
#define m(i) M[i-1]

  /* initialize */ 
  sweep = 0;
  m(1) = 1.;
  m(2) = 1.;
  m(3) = 1.;
  lamda(1) = k(1);
  lamda(2) = k(2);
  lamda(3) = k(3);
    
  /* Max and min terms on the diagonal of K */
  kj = k(1);
  kj = min (k(2),kj);
  kj = min (k(3),kj);
  ki = k(1);
  ki = max (k(2),ki);
  ki = max (k(3),ki);
  mj = 1.;
  mi = 1.;

  /* Scale to avoid problems with overflow and underflow. */
  iexp = ((log10(kj)+log10(ki)+ log10(mj)+log10(mi)) / 4.);
  scale = 1. /  (pow (10., iexp));
  m(1) = m(1) * scale;
  m(2) = m(2) * scale;
  m(3) = m(3) * scale;
  k(1) = k(1) * scale;
  k(4) = k(4) * scale;
  k(2) = k(2) * scale;
  k(6) = k(6) * scale;
  k(5) = k(5) * scale;
  k(3) = k(3) * scale;
  
 new_sweep:
  
  sweep = sweep + 1;
  thresh = pow (SMALL,  sweep);
  sqtol = jactol * jactol;
  if(thresh < sqtol) thresh = sqtol;
  /* Sweep loop; work on lower triangle only  (i > j)
    Rows are done from top to bottom;
    Columns are done from left to right. */
  
  /* row 2 and column 1. */
  krat = (k(4)*k(4))/(k(2)*k(1));
  if  (krat >= thresh) {
    /* Compute the rotatiom matrix:  an identity 
       matrix with alpha at position (2,1) and
       gamma at position (1,2). */ 
    kbari = -m(2)*k(4);
    kbarj = -m(1)*k(4);
    kbar = k(2)*m(1)-k(1)*m(2);
    rad = (kbar*kbar/4.)+kbari*kbarj;
    xsign = 1.;
    x = kbar/2.+sign(xsign,kbar)* sqrt(rad);
    if ((fabs(x) < jactol*fabs(kbarj)) || (fabs(x) < jactol*fabs(kbari))) {
      alpha = 0.;
      gamma = -k(4)/k(2);
    } else {
      alpha = kbarj/x;
      gamma = -kbari/x;
    }
    /* row 3, column 2
       row 3, column 1 */
    ki = k(5);
    kj = k(6);
    k(5) = ki+gamma*kj;
    k(6) = kj+alpha*ki;
    /* term (2,1) and diagonal terms (2,2) and (1,1). */
    kj = k(1);
    mj = m(1);
    ki = k(2);
    mi = m(2);
    k(1) = kj+alpha*alpha*ki+ 2.*alpha*k(4);
    m(1) = mj+alpha*alpha*mi;
    k(2) = ki+gamma*gamma*kj+ 2.*gamma*k(4);
    m(2) = mi+gamma*gamma*mj;
    k(4) = 0.;
  }
  
  /* row 3 and column 1. */ 
  krat = (k(6)*k(6))/(k(3)*k(1));
  if  (krat >= thresh) {
    /* Compute the rotatiom matrix:  an identity 
       matrix with alpha at position (3,1) and
       gamma at position (1,3). */ 
    kbari = -m(3)*k(6);
    kbarj = -m(1)*k(6);
    kbar = k(3)*m(1)-k(1)*m(3);
    rad = (kbar*kbar/4.)+kbari*kbarj;
    xsign = 1.;
    if (kbar <= 0.) xsign = -xsign;
    x = kbar/2.+xsign*sqrt(rad);
    if ((fabs(x) < jactol*fabs(kbarj)) || (fabs(x) < jactol*fabs(kbari))) {
      alpha = 0.;
      gamma = -k(6)/k(3);
    } else {
      alpha = kbarj/x;
      gamma = -kbari/x;
    }
    /* row 3, column 2
      row 2, column 1 */ 
    ki = k(5);
    kj = k(4);
    k(5) = ki+gamma*kj;
    k(4) = kj+alpha*ki;
    /* term (3,1) and diagonal terms (3,3) and (1,1). */ 
    kj = k(1);
    mj = m(1);
    ki = k(3);
    mi = m(3);
    k(1) = kj+alpha*alpha*ki+ 2.*alpha*k(6);
    m(1) = mj+alpha*alpha*mi;
    k(3) = ki+gamma*gamma*kj+ 2.*gamma*k(6);
    m(3) = mi+gamma*gamma*mj;
    k(6) = 0.;
  }
  
  /* row 3 and column 2.       */
  krat = (k(5)*k(5))/(k(3)*k(2));
  if  (krat >= thresh) {
    /* compute the rotatiom matrix:  an identity 
      matrix with alpha at position (3,2) and
      gamma at position (2,3). */ 
    kbari = -m(3)*k(5);
    kbarj = -m(2)*k(5);
    kbar = k(3)*m(2)-k(2)*m(3);
    rad = (kbar*kbar/4.)+kbari*kbarj;
    xsign = 1.;
    if (kbar <= 0.) xsign = -xsign;
    x = kbar/2.+xsign*sqrt(rad);
    if ((fabs(x) < jactol*fabs(kbarj)) || (fabs(x) < jactol*fabs(kbari))) {
      alpha = 0.;
      gamma = -k(5)/k(3);
    } else {
      alpha = kbarj/x;
      gamma = -kbari/x;
    }
    /* row 3, column 1
      row 2, column 1 */ 
    ki = k(6);
    kj = k(4);
    k(6) = ki+gamma*kj;
    k(4) = kj+alpha*ki;
    /* term (3,2) and diagonal terms (3,3) and (2,2). */ 
    kj = k(2);
    mj = m(2);
    ki = k(3);
    mi = m(3);
    k(2) = kj+alpha*alpha*ki+ 2.*alpha*k(5);
    m(2) = mj+alpha*alpha*mi;
    k(3) = ki+gamma*gamma*kj+ 2.*gamma*k(5);
    m(3) = mi+gamma*gamma*mj;
    k(5) = 0.;
    
  }

  /* Sweep done; update eigenvalue vector, lamda */             
  lamda(1) = k(1)/m(1);
  lamda(2) = k(2)/m(2);
  lamda(3) = k(3)/m(3);
  /* Check off-diagonal elements for convergence */
  if ((k(4)*k(4)/(k(2)*k(1)) < sqtol &&
       k(6)*k(6)/(k(3)*k(1)) < sqtol &&
       k(5)*k(5)/(k(3)*k(2)) < sqtol)) goto exit;

  if (sweep < maxsweep) goto new_sweep;
  
 exit:
  
  /* reorder the eigenvalues */
  if(lamda(2) < lamda(1)) {
    swap = lamda(1); lamda(1) = lamda(2); lamda(2) = swap;
  }
  if(lamda(3) < lamda(1)) {
    swap = lamda(1); lamda(1) = lamda(3); lamda(3) = swap;
  }
  if(lamda(3) < lamda(2)) {
    swap = lamda(2); lamda(2) = lamda(3); lamda(3) = swap;
  }
}


/*
  This function computes the invariants of the right     
  stretch tensor, the metric tensor, and its square.
  The C-G tensor and its square are in the upper triangular storage.
*/
static void
cauchygreen (double F[3][3], double C[6], double C2[6])
{
#define c(i) C[i-1]
#define f(i,j) F[i-1][j-1]     
#define c2(i) C2[i-1]
  /* C-G tensor */
  c(1) = f(1,1)*f(1,1)+f(2,1)*f(2,1)+f(3,1)*f(3,1); /* c(1,1) */
  c(2) = f(1,1)*f(1,2)+f(2,1)*f(2,2)+f(3,1)*f(3,2); /* c(1,2) */
  c(3) = f(1,2)*f(1,2)+f(2,2)*f(2,2)+f(3,2)*f(3,2); /* c(2,2) */
  c(4) = f(1,1)*f(1,3)+f(2,1)*f(2,3)+f(3,1)*f(3,3); /* c(1,3) */
  c(5) = f(1,2)*f(1,3)+f(2,2)*f(2,3)+f(3,2)*f(3,3); /* c(2,3) */
  c(6) = f(1,3)*f(1,3)+f(2,3)*f(2,3)+f(3,3)*f(3,3); /* c(3,3) */
  /* Compute the square of the C-G tensor. */
  c2(1) = c(1)*c(1)+c(2)*c(2)+c(4)*c(4);
  c2(2) = c(1)*c(2)+c(2)*c(3)+c(4)*c(5);
  c2(3) = c(2)*c(2)+c(3)*c(3)+c(5)*c(5);
  c2(4) = c(1)*c(4)+c(2)*c(5)+c(4)*c(6);
  c2(5) = c(2)*c(4)+c(3)*c(5)+c(5)*c(6);
  c2(6) = c(4)*c(4)+c(5)*c(5)+c(6)*c(6);
#undef c
#undef f
#undef c2
}

/*
  This function computes the invariants of the right     
  stretch tensor, the metric tensor, and its square.       
*/
static void
right_stretch_invariants (double F[3][3], double C[6], double C2[6],
                          double *iu, double *iiu, double *iiiu)
{
  double CT[6], EV[3];
  /* C and C2 are in symmetric upper triangular form. */
#define c(i) C[i-1]
#define f(i,j) F[i-1][j-1]     
#define c2(i) C2[i-1]
#define ev(i) EV[i-1]
#define ct(i) CT[i-1]

  /* C-G tensor */
  c(1) = f(1,1)*f(1,1)+f(2,1)*f(2,1)+f(3,1)*f(3,1); /* c(1,1) */
  c(2) = f(1,1)*f(1,2)+f(2,1)*f(2,2)+f(3,1)*f(3,2); /* c(1,2) */
  c(3) = f(1,2)*f(1,2)+f(2,2)*f(2,2)+f(3,2)*f(3,2); /* c(2,2) */
  c(4) = f(1,1)*f(1,3)+f(2,1)*f(2,3)+f(3,1)*f(3,3); /* c(1,3) */
  c(5) = f(1,2)*f(1,3)+f(2,2)*f(2,3)+f(3,2)*f(3,3); /* c(2,3) */
  c(6) = f(1,3)*f(1,3)+f(2,3)*f(2,3)+f(3,3)*f(3,3); /* c(3,3) */
  /* Compute the square of the C-G tensor. */
  c2(1) = c(1)*c(1)+c(2)*c(2)+c(4)*c(4);
  c2(2) = c(1)*c(2)+c(2)*c(3)+c(4)*c(5);
  c2(3) = c(2)*c(2)+c(3)*c(3)+c(5)*c(5);
  c2(4) = c(1)*c(4)+c(2)*c(5)+c(4)*c(6);
  c2(5) = c(2)*c(4)+c(3)*c(5)+c(5)*c(6);
  c2(6) = c(4)*c(4)+c(5)*c(5)+c(6)*c(6);
  /* Copy the metric tensor to stress vector;
     eigenvalues() relies on it. */           
  ct(1) = c(1);
  ct(2) = c(3);
  ct(3) = c(6);
  ct(4) = c(2);
  ct(5) = c(5);
  ct(6) = c(4);
  /* compute the principal values of the
     metric tensor. */
  eigenvalues (CT, EV);
  ev(1) = sqrt(ev(1));
  ev(2) = sqrt(ev(2));
  ev(3) = sqrt(ev(3));
  /* compute the invariants of the right
    stretch tensor. */
  *iu = ev(1)+ev(2)+ev(3);
  *iiu = ev(1)*ev(2)+ev(2)*ev(3)+ev(1)*ev(3);
  *iiiu = ev(1)*ev(2)*ev(3);
  /* cleanup */
#undef c
#undef f
#undef c2
#undef ev
#undef ct
}

#if 0
/*
  This algorithm is implemented from Simo, Hughes, p.242.
 */
static void
princ_stretches (double C[3][3], /* IN: Cauchy-Green tensor */
                 double lambda[3] /* OUT: principal stretches of C */
                 )
{
  const double tol3 = 1e-16;
  double I1, I2, I3;
  double C2[3][3];
  double x[3], b, c, i1, i2, i3, D;
  int A;
  I1 = MAT3DUTIL_TRACE_3X3 (C);
  MAT3DUTIL_MULT_AB_3X3 (C2, C, C);
  I2 = (I1*I1 - MAT3DUTIL_TRACE_3X3 (C2)) / 2;
  I3 = MAT3DUTIL_DET_3X3 (C);
  FOR (A, 3) {
    b  = I2 - I1*I1 / 3;
    c  = -(2*I1*I1*I1)/27 + I1*I2/3 - I3;
    if (b < tol3) {
      x[A] = -pow(c, 1.0/3);
    } else {
      double m, n, t;
      m = 2 * sqrt (-b / 3);
      n = (3 * c) / (m * b);
      t = atan2 (sqrt (1 - n*n), n) / 3;
      x[A] = m * cos (t + 2 * (A - 1) * M_PI / 3);
    }
    lambda[A] = x[A] + I1/3;
  }
  i1 = lambda[0] + lambda[1] + lambda[2];
  i2 = lambda[0]*lambda[1] + lambda[0]*lambda[2] + lambda[1]*lambda[2];
  i3 = lambda[0]*lambda[1]*lambda[2];
  D  = i1*i2 - i3;
  /* U    = 1/D * [ -C^2 + (i1^2 - i2)*C + i1*i3*1 ] */
  /* U^-1 = 1/i3 * [ C - i1*U + i2*1 ] */
}
#endif

/*
    this subroutine computes the polar decompostion of the   
    deformation gradient into the right stretch tensor.      
    the computations are for a gauss point of a q3disop      
    or l3disop element.                                      
*/
W_bool_t
W_polar_decomposition (double F[3][3], /* deformation gradient */
                       double R[3][3], /* output rotation tensor */
                       double U[3][3] /* output deformation tensor U */
                       )
{
  double inv_U[3][3], C[6], C2[6], iu, iiu, iiiu;
  double a1, b1, c1;
#define c(i) C[i-1]
#define c2(i) C2[i-1]

  /* the Cauchy-Green tensor and its square. */
  cauchygreen (F, C, C2);
  /* the invariants of the right stretch tensor */
  right_stretch_invariants (F, C, C2, &iu, &iiu, &iiiu);
  /* the right stretch tensor. */
  a1 = 1.0/(iu*iiu-iiiu);
  b1 = iu*iiiu;
  c1 = iu*iu-iiu;
  U[0][0] = a1 * (b1 + c1*c(1) - c2(1));
  U[0][1] = 
  U[1][0] = a1 * (     c1*c(2) - c2(2));
  U[1][1] = a1 * (b1 + c1*c(3) - c2(3));
  U[0][2] =
  U[2][0] = a1 * (     c1*c(4) - c2(4));
  U[1][2] =
  U[2][1] = a1 * (     c1*c(5) - c2(5));
  U[2][2] = a1 * (b1 + c1*c(6) - c2(6));

  if (!W_mat3dutil_invert_3x3 (U, inv_U)) return FALSE;
  MAT3DUTIL_MULT_AB_3X3 (R, F, inv_U);

  return TRUE;
#undef c
#undef c2
}

/*
  Evaluates the eigenvalues and principal directions of a symmetric matrix 
  using the Jacobi iteration.  Adapted from numerical recipes.
*/
static W_bool_t
do_jacobi (double *mat, /* IN: symmetric 3x3 matrix.
                           WARNING: it is overwritten inside this routine */
           double *pdir, /* OUT: 3x3 matrix, eigenvectors in columns */
           double *eval /* eigenvalues */
           )
{
  /* System generated locals */
  double d__1, d__2, d__3, d__4;
  /* Local variables */
  double c, g, h;
  int i, j;
  double s, t, theta, hd, od;
  int ip, iq, ir, is;
  double tau, sum;
  const double tol = 1e-16;
  

  /*     Initialise pdir to the identity */
  pdir -= 4;
  --eval;
  mat -= 4;
  
  /* Function Body */
  for (i = 1; i <= 3; ++i) {
    for (j = 1; j <= 3; ++j) {
      pdir[i + j * 3] = 0.;
    }
    pdir[i + i * 3] = 1.;
    eval[i] = mat[i + i * 3];
  }
  
  /*     Starts sweeping. */
  for (is = 1; is <= 50; ++is) {
    sum = 0.;
    for (ip = 1; ip <= 2; ++ip) {
      for (iq = ip + 1; iq <= 3; ++iq) {
        sum += (d__1 = mat[ip + iq * 3], fabs(d__1));
      }
    }
    
    /* If the sum of off-diagonal terms is zero evaluates the */
    /* evalches and returns */
    
    if (sum < tol) {
      return TRUE;
    }
    
    /*     Performs the sweep in three rotations. One per off diagonal term */
    
    for (ip = 1; ip <= 2; ++ip) {
      for (iq = ip + 1; iq <= 3; ++iq) {
        od = (d__1 = mat[ip + iq * 3], fabs(d__1)) * 100.;
        if (od + (d__1 = eval[ip], fabs(d__1)) != (d__2 = eval[ip], fabs(d__2)) &&
            od + (d__3 = eval[iq], fabs(d__3)) != (d__4 = eval[iq], fabs(d__4))
            ) {
          hd = eval[iq] - eval[ip];
          /* Evaluates the rotation angle */
          if (fabs(hd) + od == fabs(hd)) {
            t = mat[ip + iq * 3] / hd;
          } else {
            theta = hd * .5 / mat[ip + iq * 3];
            d__1 = theta;
            t = 1. / (fabs(theta) + sqrt(d__1 * d__1 + 1.));
            if (theta < 0.) {t = -t;}
          }
          /* Re-evaluates the diagonal terms */
          d__1 = t;
          c = 1. / sqrt(d__1 * d__1 + 1.);
          s = t * c;
          tau = s / (c + 1.);
          h = t * mat[ip + iq * 3];
          eval[ip] -= h;
          eval[iq] += h;
         /* Re-evaluates the remaining off-diagonal terms */
          ir = 6 - ip - iq;
          g = mat[min(ir,ip) + max(ir,ip) * 3];
          h = mat[min(ir,iq) + max(ir,iq) * 3];
          mat[min(ir,ip) + max(ir,ip) * 3] = g - s * (h + g * tau);
          mat[min(ir,iq) + max(ir,iq) * 3] = h + s * (g - h * tau);
          /* Rotates the eigenvectors */
          for (ir = 1; ir <= 3; ++ir) {
            g = pdir[ir + ip * 3];
            h = pdir[ir + iq * 3];
            pdir[ir + ip * 3] = g - s * (h + g * tau);
            pdir[ir + iq * 3] = h + s * (g - h * tau);
          }
        }
        mat[ip + iq * 3] = 0.;
      }
    }
  }
  return FALSE;
} 
     
/*
  Principal directions and eigenvalues of a symmetric
  tensor.  Uses an implementation of the Jacobi rotation.
 */
W_bool_t
W_eigendecomposition (      
                      double t[3][3],/* IN symmetric tensor */
                      double eig_vecs[3][3], /* OUT each row */
                      /* is an eigenvector */
                      double eig_vals[3] /* OUT principal stretches */
                      )
{
  double privt[3][3];
  MAT3DUTIL_COPY_3X3 (privt, t);
  return do_jacobi ((double *) privt, (double *)eig_vecs, (double *)eig_vals);
}

/*
  Compute the logarithmic strain tensor.
  Given the U tensor, compute its logarithm.
  The eigendecomposition is needed to perform
  this task.
 */
void
W_ln_U (double U[3][3], /* input tensor */
        double lnU[3][3] /* output tensor */
        )
{
  double evals[3];
  double evecs[3][3];
  double lnevals[3];
 
  W_eigendecomposition (U, evecs, evals);

  LN_FUNC (lnevals[0], /* = log */ evals[0]);
  LN_FUNC (lnevals[1], /* = log */ evals[1]);
  LN_FUNC (lnevals[2], /* = log */ evals[2]);
  lnU[0][0]
    = (lnevals[0] * evecs[0][0] * evecs[0][0] +
       lnevals[1] * evecs[1][0] * evecs[1][0] +
       lnevals[2] * evecs[2][0] * evecs[2][0]);
  lnU[0][1] = lnU[1][0]
    = (lnevals[0] * evecs[0][0] * evecs[0][1] +
       lnevals[1] * evecs[1][0] * evecs[1][1] +
       lnevals[2] * evecs[2][0] * evecs[2][1]);
  lnU[0][2] = lnU[2][0]
    = (lnevals[0] * evecs[0][0] * evecs[0][2] +
       lnevals[1] * evecs[1][0] * evecs[1][2] +
       lnevals[2] * evecs[2][0] * evecs[2][2]);
  lnU[1][1]
    = (lnevals[0] * evecs[0][1] * evecs[0][1] +
       lnevals[1] * evecs[1][1] * evecs[1][1] +
       lnevals[2] * evecs[2][1] * evecs[2][1]);
  lnU[1][2] = lnU[2][1]
    = (lnevals[0] * evecs[0][1] * evecs[0][2] +
       lnevals[1] * evecs[1][1] * evecs[1][2] +
       lnevals[2] * evecs[2][1] * evecs[2][2]);
  lnU[2][2]
    = (lnevals[0] * evecs[0][2] * evecs[0][2] +
       lnevals[1] * evecs[1][2] * evecs[1][2] +
       lnevals[2] * evecs[2][2] * evecs[2][2]);
}


#if 0
#include <randP.h>
int
test_polar_decomp (void)
{
  double C[3][3], Cev[3][3], Cee[3];
  double C1[3][3], Cev1[3][3], Cee1[3];
  int i, j, k;
#define PR(M)                                   \
  {                                             \
    int I_, J_;                                 \
    printf ("%s = [\n", #M);                    \
    FOR (I_, 3) {                               \
      FOR (J_, 3) {                             \
        printf (" %g; ", M[I_][J_]);            \
      }                                         \
      printf ("\n");                            \
    }                                           \
    printf ("]\n");                             \
  }


  FOR (k, 5){
    printf ("\n");
    FOR (i, 3) {
      FOR (j, 3) {
        C[i][j] = rand_0_to_1 ();
      }
    }
    MAT3DUTIL_SYMM (C, C);
    MAT3DUTIL_COPY_3X3 (C1, C);
    PR(C);
    PR(C1);
    W_eigendecomp (C, Cev, Cee);
    W_eigendecomposition (C1, Cev1, Cee1);
    printf ("ee=[%g %g %g]\n", Cee[0], Cee[1], Cee[2]);
    printf ("ee1=[%g %g %g]\n", Cee1[0], Cee1[1], Cee1[2]);
    PR (Cev);
    PR (Cev1);
  }
  return 0;
}
#endif


