#include "Wconfig.h"
#include "Wtypes.h"
#include "Wpss.h"

/* RCSHEADER "$Id: pss.c,v 1.1.1.1 2001/07/24 02:42:51 pkrysl Exp $" */

/* Returns the [3x8] strain-displacement matrix {B} of the receiver, eva-
   luated at a Gauss point. */
void
pss_compute_B_matrix(W_float_t b[3][8],
                     W_float_t eta, W_float_t xi,
                     W_float_t inv_jacobian[2][2],
                     W_float_t inv_jacobian_at_0_0[2][2])

{
  double xiX, xiY, etaX, etaY;
  double n1Xi, n2Xi, n3Xi, n4Xi, n1Eta, n2Eta, n3Eta, n4Eta;
  
  xiX = inv_jacobian[0][0];
  xiY = inv_jacobian[1][0];
  etaX = inv_jacobian[0][1];
  etaY = inv_jacobian[1][1];
  
  /* partial derivatives of the shape functions N_i, with respect to xi
     and eta : */
  n1Xi = (-1. + eta) * 0.25;
  n2Xi = ( 1. - eta) * 0.25;
  n3Xi = ( 1. + eta) * 0.25;
  n4Xi = (-1. - eta) * 0.25;
  n1Eta = (-1. + xi) * 0.25;
  n2Eta = (-1. - xi) * 0.25;
  n3Eta = ( 1. + xi) * 0.25;
  n4Eta = ( 1. - xi) * 0.25;
  
  /* B matrix  -  3 rows : epsilon-X, epsilon-Y, gamma-XY: */
  b[0][0] = n1Xi*xiX + n1Eta*etaX;
  b[0][1] = 0;
  b[0][2] = n2Xi*xiX + n2Eta*etaX;
  b[0][3] = 0;
  b[0][4] = n3Xi*xiX + n3Eta*etaX;
  b[0][5] = 0;
  b[0][6] = n4Xi*xiX + n4Eta*etaX;
  b[0][7] = 0;
  b[1][0] = 0;
  b[1][1] = n1Xi*xiY + n1Eta*etaY;
  b[1][2] = 0;
  b[1][3] = n2Xi*xiY + n2Eta*etaY;
  b[1][4] = 0;
  b[1][5] = n3Xi*xiY + n3Eta*etaY;
  b[1][6] = 0;
  b[1][7] = n4Xi*xiY + n4Eta*etaY;
  /* reduced integration 
     partial derivatives of the shape functions N_i, with respect to xi
     and eta at eta = xi = 0.0 :
     partial derivatives of xi and eta, with respect to x and y: */
  
  xiX = inv_jacobian_at_0_0[0][0];
  xiY = inv_jacobian_at_0_0[0][1];
  etaX = inv_jacobian_at_0_0[1][0];
  etaY = inv_jacobian_at_0_0[1][1];
  
  eta = 0;
  xi = 0;
  n1Xi = (-1. + eta) * 0.25;
  n2Xi = ( 1. - eta) * 0.25;
  n3Xi = ( 1. + eta) * 0.25;
  n4Xi = (-1. - eta) * 0.25;
  n1Eta = (-1. + xi) * 0.25;
  n2Eta = (-1. - xi) * 0.25;
  n3Eta = ( 1. + xi) * 0.25;
  n4Eta = ( 1. - xi) * 0.25;
  /* end of reduced integration */
  
  b[2][0]  = n1Xi*xiY + n1Eta*etaY;
  b[2][1]  = n1Xi*xiX + n1Eta*etaX;
  b[2][2]  = n2Xi*xiY + n2Eta*etaY;
  b[2][3]  = n2Xi*xiX + n2Eta*etaX;
  b[2][4]  = n3Xi*xiY + n3Eta*etaY;
  b[2][5]  = n3Xi*xiX + n3Eta*etaX;
  b[2][6]  = n4Xi*xiY + n4Eta*etaY;
  b[2][7]  = n4Xi*xiX + n4Eta*etaX;
}



/* Returns the displacement interpolation matrix {N} of the receiver,
   evaluated at a Gauss point eta, xi. */
void
pss_compute_N_matrix(W_float_t n[2][8], W_float_t eta, W_float_t xi)
{
  double       n1, n2, n3, n4;

  n1  = (1. - xi) * (1. - eta) * 0.25;
  n2  = (1. + xi) * (1. - eta) * 0.25;
  n3  = (1. + xi) * (1. + eta) * 0.25;
  n4  = (1. - xi) * (1. + eta) * 0.25;
  
  n[0][0] = n1;
  n[0][1] =  0;
  n[0][2] = n2;
  n[0][3] =  0;
  n[0][4] = n3;
  n[0][5] =  0;
  n[0][6] = n4;
  n[0][7] =  0;
  n[1][0] =  0;
  n[1][1] = n1;
  n[1][2] =  0;
  n[1][3] = n2;
  n[1][4] =  0;
  n[1][5] = n3;
  n[1][6] =  0;
  n[1][7] = n4;
}




/* Returns the displacement interpolation matrix {N} of the receiver,
   evaluated at a Gauss point eta, xi. */
void
pss_N_matrix(W_float_t n[4], W_float_t eta, W_float_t xi)
{
  n[0]  = (1. - xi) * (1. - eta) * 0.25;
  n[1]  = (1. + xi) * (1. - eta) * 0.25;
  n[2]  = (1. + xi) * (1. + eta) * 0.25;
  n[3]  = (1. - xi) * (1. + eta) * 0.25;
}




/* Returns the derivative of the submatrix {N1} of the displacement */
/* interpolation matrix {N} by xi, evaluated at a Gauss point eta, xi. */
void
pss_compute_N_xi_matrix(W_float_t n_xi[4], W_float_t eta, W_float_t xi)
{
  double       n1, n2;

  n1  =   (1. - eta) * 0.25;
  n2  =   (1. + eta) * 0.25;
  
  n_xi[0] = -n1;
  n_xi[1] =  n1;
  n_xi[2] =  n2;
  n_xi[3] = -n2;
}


/* Returns the derivative of the submatrix {N2} of the displacement */
/* interpolation matrix {N} by eta, evaluated at a Gauss point eta, xi. */
void
pss_compute_N_eta_matrix(W_float_t n_eta[4], W_float_t eta, W_float_t xi)
{
  double       n1, n2;

  n1  =   (1. - xi) * 0.25;
  n2  =   (1. + xi) * 0.25;
  
  n_eta[0]  = -n1;
  n_eta[1]  = -n2;
  n_eta[2]  =  n2;
  n_eta[3]  =  n1;
}



/* Computes the jacobian matrix  J(x,y) */
void
pss_jacobian_matrix(W_float_t J[2][2], W_float_t eta, W_float_t xi,
                    W_float_t x1, W_float_t y1, 
                    W_float_t x2, W_float_t y2,
                    W_float_t x3, W_float_t y3,
                    W_float_t x4, W_float_t y4
                    ) 
{
  W_float_t derivx[4] = {0, 0, 0, 0};
  W_float_t derivy[4] = {0, 0, 0, 0};

  derivx[0] = -0.25*(1-eta);
  derivx[1] =  0.25*(1-eta);
  derivx[2] =  0.25*(1+eta);
  derivx[3] = -0.25*(1+eta);

  derivy[0] = -0.25*(1-xi);
  derivy[1] = -0.25*(1+xi);
  derivy[2] =  0.25*(1+xi);
  derivy[3] =  0.25*(1-xi);
  
  J[0][0]   = (derivx[0]*x1 + derivx[1]*x2 +
               derivx[2]*x3 + derivx[3]*x4);
  J[0][1]   = (derivx[0]*y1 + derivx[1]*y2 +
               derivx[2]*y3 + derivx[3]*y4);
  J[1][0]   = (derivy[0]*x1 + derivy[1]*x2 +
               derivy[2]*x3 + derivy[3]*x4);
  J[1][1]   = (derivy[0]*y1 + derivy[1]*y2 +
               derivy[2]*y3 + derivy[3]*y4);
}


W_float_t
pss_jacobian(W_float_t J[2][2])
{
  return J[0][0] * J[1][1] - J[0][1] * J[1][0];
}



void
pss_invert_jacobian_matrix(W_float_t J[2][2], W_float_t inv_J[2][2])
{
  double det;

  assert(J != NULL);
  assert(inv_J != NULL);
  
  det  = J[0][0] * J[1][1] - J[0][1] * J[1][0];
  det  = 1 / det;
  inv_J[0][0] =  J[1][1] * det;
  inv_J[0][1] = -J[0][1] * det;
  inv_J[1][0] = -J[1][0] * det;
  inv_J[1][1] =  J[0][0] * det;
}
