#include "Wconfig.h"
#include "Wxf.h"
#include "Wdomain.h"
#define WPVM_PRIVATE_HEADER
#include "Wpvm.h"

/* RCSHEADER "$Id: pvmrun.c,v 1.1.1.1 2001/07/24 02:42:52 pkrysl Exp $" */

#if defined(PVM_WASP_WORKER) || defined(PVM_WASP_MASTER)

int
PVM_err_exit(char *pvm_routine, char *msg)
{
  char buffer[512];
  
  sprintf(buffer, "%s -- info = ", pvm_routine);
  switch (pvm_status.info) {
  case PvmOk:
    strcat(buffer, "PvmOk\n");
    break;
  case PvmBadParam:
    strcat(buffer, "PvmBadParam\n");
    break;
  case PvmMismatch:
    strcat(buffer, "PvmMismatch\n");
    break;
  case PvmOverflow:
    strcat(buffer, "PvmOverflow\n");
    break;
  case PvmNoData:
    strcat(buffer, "PvmNoData\n");
    break;
  case PvmNoHost:
    strcat(buffer, "PvmNoHost\n");
    break;
  case PvmNoFile:
    strcat(buffer, "PvmNoFile\n");
    break;
  case PvmNoMem:
    strcat(buffer, "PvmNoMem\n");
    break;
  case PvmBadMsg:
    strcat(buffer, "PvmBadMsg\n");
    break;
  case PvmSysErr:
    strcat(buffer, "PvmSysErr\n");
    break;
  case PvmNoBuf:
    strcat(buffer, "PvmNoBuf\n");
    break;
  case PvmNoSuchBuf:
    strcat(buffer, "PvmNoSuchBuf\n");
    break;
  case PvmNullGroup:
    strcat(buffer, "PvmNullGroup\n");
    break;
  case PvmDupGroup:
    strcat(buffer, "PvmDupGroup\n");
    break;
  case PvmNoGroup:
    strcat(buffer, "PvmNoGroup\n");
    break;
  case PvmNotInGroup:
    strcat(buffer, "PvmNotInGroup\n");
    break;
  case PvmNoInst:
    strcat(buffer, "PvmNoInst\n");
    break;
  case PvmHostFail:
    strcat(buffer, "PvmHostFail\n");
    break;
  case PvmNoParent:
    strcat(buffer, "PvmNoParent\n");
    break;
  case PvmNotImpl:
    strcat(buffer, "PvmNotImpl\n");
    break;
  case PvmDSysErr:
    strcat(buffer, "PvmDSysErr\n");
    break;
  case PvmBadVersion:
    strcat(buffer, "PvmBadVersion\n");
    break;
  case PvmOutOfRes:
    strcat(buffer, "PvmOutOfRes\n");
    break;
  case PvmDupHost:
    strcat(buffer, "PvmDupHost\n");
    break;
  case PvmCantStart:
    strcat(buffer, "PvmCantStart\n");
    break;
  case PvmAlready:
    strcat(buffer, "PvmAlready\n");
    break;
  case PvmNoTask:
    strcat(buffer, "PvmNoTask\n");
    break;
  case PvmNoEntry:
    strcat(buffer, "PvmNoEntry\n");
    break;
  case PvmDupEntry:
    strcat(buffer, "PvmDupEntry\n");
    break;
  default:
    strcat(buffer, "unknown textual form\n");
    break;
  }
  strcat(buffer, msg);
  pvm_perror(buffer);
  return 0;
}

/*  
! MASTER PVM SETUP
PVM
! tot-num-of-interface-nodes  
137
! split files
    "file1" 
    "file2" 
    "file3" 
! end-of-data
&
*/


#if defined(PVM_WASP_MASTER)

static void
spawn(W_domain_t *d);
static void
construct_full_name(char *dest, char *src);

W_bool_t
pvm_run_reader(W_domain_t *d, FILE *stream)
{
  const char *input_line;
  long n;
  W_pvm_master_t *m;
  char name[256];
  char full_name[256];
  W_pvm_partition_id_t *ptid;
  int np;

  input_line = W_get_input_line(stream);

  if (sscanf(input_line, "%ld", &n) != 1)
    BAD_INPUT();
  m = w_new(W_pvm_master_t);
  CHECK_MALLOC(m);
  m->mytid = 0;
  m->num_of_interface_nodes = n;
  m->interfaces
    = w_new_array(m->num_of_interface_nodes+1, W_interface_package_t);
  CHECK_MALLOC(m->interfaces);

  m->partitions = make_list();
  
  while ((input_line = W_get_input_line(stream)) != NULL) {
    if (input_line[0] == END_OF_BLOCK)
      break;
    else {
      TOKENS_PARSE(input_line);
      sprintf(name, "%s", TOKEN_STR(1));
      construct_full_name(full_name, name);
      ptid = w_new(W_pvm_partition_id_t);
      CHECK_MALLOC(ptid);
      DUPLICATE_STRING(ptid->split_input_file, full_name);
      ptid->host_name = NULL;
      if (TOKENS_HAVE_KEYWORD("HOST")) {
        sprintf(name, "%s", TOKEN_STR(2));
        DUPLICATE_STRING(ptid->host_name, name);
      }
      add_to_tail(m->partitions, ptid);
    }
  }

  m->num_of_workers = list_length(m->partitions);
  if (m->num_of_workers == 0) REPORT_FATAL_ERROR("No worker");
  m->tids = NULL; /* no workers yet */

  d->pvm_master = m;
  
  spawn(d);

  /* Wait until all workers have been started up and arrive at the barrier */
  for (np = 1; np <= m->num_of_workers; np++) {
    INITSEND();
    PACK(int, &m->tids[np]);
    SEND(m->tids[np], STARTUP_ACK_TAG);
    RECEIVE(m->tids[np], STARTUP_ACK_TAG);
  }
  
  return TRUE;
}

static void
construct_full_name(char *dest, char *src)
{
  char *p;

  p = strrchr(src, '/');
  if (p != NULL)
    sprintf(dest, "%s", src); /* worker file name is absolute or relative; */
                              /* not just a file name */
  else {
    sprintf(dest, "%s", options.in_file_name);
    p = strrchr(dest, '/');
    if (p != NULL) {            /* it gives absolute or relative path */
      ++p;                      /* move beyond the slash */
      sprintf(p, "%s", src);
    } else {
      sprintf(dest, "%s", src); /* master name is plain file name */
    }
  }
}

static void
modify_i_opt(char **wd_args, char *fn);

/* Replaces the master's input file name by the worker's file name */
static void
modify_i_opt(char **wd_args, char *fn)
{
  while (*wd_args != NULL) {
    if ((*wd_args)[0] == '-' && (*wd_args)[1] == 'i') {
      ++wd_args;
      w_free(*wd_args);
      DUPLICATE_STRING(*wd_args, fn);
      return;
    }
    ++wd_args;
  }
  assert(0);
}

static void
spawn(W_domain_t *d)
{
  W_pvm_master_t *m = d->pvm_master;
  int i, numt, nt;
  int pvm_opts = PVM_SPAWN_OPTS;
  W_pvm_partition_id_t *ptid;
  char *where = PVM_SPAWN_WHERE;

  if (options.debug_pvm)
    pvm_opts += PvmTaskDebug;

  assert(m != NULL);
  
  m->mytid = MYTID();

  m->tids = w_new_array(m->num_of_workers+1, int);
  CHECK_MALLOC(m->tids);
  nt = 0;
  ptid = (W_pvm_partition_id_t *)get_list_next(m->partitions, NULL);
  while (ptid != NULL) {
    modify_i_opt(wd_args, ptid->split_input_file);
    nt++;
    if (ptid->host_name != NULL) {
      pvm_opts = PvmTaskHost; where = ptid->host_name;
    } else  {
      pvm_opts = PVM_SPAWN_OPTS; where = NULL;
    }
    numt = pvm_spawn(WORKER_NAME,
                     wd_args,
                     pvm_opts,
                     where,
                     1 /* worker */,
                     m->tids+nt);
    if (numt != 1) pvm_perror(__FILE__);
    ptid = (W_pvm_partition_id_t *)get_list_next(m->partitions, ptid);
  }
}

#endif /* defined(PVM_WASP_MASTER) */

/* 
! WORKER PVM SETUP
PVM
! num-of-interface-nodes
 12
!      node  interface-num
         1    23
         2    4
         8    23
! end-of-data
&
*/
#if defined(PVM_WASP_WORKER)
W_bool_t
pvm_run_reader(W_domain_t *d, FILE *stream)
{
  const char *input_line;
  long n, l, i;
  int np, tid;
  W_pvm_worker_t *w;

  input_line = W_get_input_line(stream);

  if (sscanf(input_line, "%ld", &n) != 1)
    BAD_INPUT();
  w = w_new(W_pvm_worker_t);
  CHECK_MALLOC(w);
  w->num_of_interface_nodes = n;
  w->map
    = w_new_array(w->num_of_interface_nodes+1, W_local_to_interface_map_t);
  CHECK_MALLOC(w->map);

  n = 0;
  while ((input_line = W_get_input_line(stream)) != NULL) {
    if (input_line[0] == END_OF_BLOCK)
      break;
    else {
      if (sscanf(input_line, "%ld %ld", &l, &i) != 2) {
        BAD_INPUT();
      } else {
        n++;
        if (n > w->num_of_interface_nodes)
          REPORT_FATAL_ERROR("Too many local-interface pairs");
        w->map[n].local_num     = l;
        w->map[n].interface_num = i;
      }
    }
  }

  d->pvm_worker = w;
  
  w->mytid = MYTID();

  w->master_tid = PARENT(); /* No parent yet */

  JOIN_GROUP(WORKER_GROUP);


  RECEIVE(MASTER_TID(w), STARTUP_ACK_TAG);
  UNPACK(int, &tid);
  INITSEND();
  SEND(MASTER_TID(w), STARTUP_ACK_TAG);
  
  return TRUE;
}
#endif /* defined(PVM_WASP_WORKER) */

#endif /* defined(PVM_WASP_WORKER) || defined(PVM_WASP_MASTER) */
