#include <stdio.h>
#include <stdlib.h>
#include <math.h>


static double
D_of_C (double C_d, double C_s, double C)
{
  double beta1, beta2;
  beta1 = sqrt (1 - C*C / (C_d*C_d));
  beta2 = sqrt (1 - C*C / (C_s*C_s));
  return 4 * beta1 * beta2 - (1 - beta2 * beta2) * (1 - beta2 * beta2);
}

static int
W_compute_Rayleigh_speed (double C_d, double C_s,
                          double percent_error,
                          double *C_R)
{
  double hi     = C_s;
  double lo     = 0;
  double cr     = C_s;
  double v, vhi, vlo;

  vhi = D_of_C (C_d, C_s, hi);
  vlo = D_of_C (C_d, C_s, lo);
  if (vhi * vlo > 0) {
    *C_R = -999999;
    return 0;
  }
  
  while ((hi - lo) / 2 > percent_error * cr) {
    cr = (hi + lo) / 2;
    v  = D_of_C (C_d, C_s, cr);
    if (v > 0) {
      if (vhi > 0)
        hi = cr;
      else
        lo = cr;
    } else {
      if (vlo < 0)
        lo = cr;
      else
        hi = cr;
    }
    vhi    = D_of_C (C_d, C_s, hi);
    vlo    = D_of_C (C_d, C_s, lo);
  }
  *C_R = (hi+lo)/2;
  return 1;
}

int
main (int argc, char **argv)
{
  double c_d, c_s, c_R;
  char *ptr;
  
  if (argc != 3) {
    fprintf (stderr, "Error: Expecting arguments %s c_d c_s\n", argv[0]);
    exit (1);
  }
  c_d = strtod (argv[1], &ptr);
  if (c_d <= 0) {
    fprintf (stderr, "Error: c_d = %g; expected value >=0\n", c_d);
    exit (1);
  }
  c_s = strtod (argv[2], &ptr);
  if (c_s <= 0) {
    fprintf (stderr, "Error: c_s = %g; expected value >=0\n", c_s);
    exit (1);
  }
  W_compute_Rayleigh_speed (c_d, c_s, 0.001, &c_R);
  fprintf (stdout, "Results: c_d = %g, c_s = %g, c_R = %g\n", c_d, c_s, c_R);
}
