#include "Wconfig.h"
#include "Wxf.h"
#include "Wdomain.h"
#include "Welem.h"
#include "Wutil.h"
#include "Wrun.h"
#include "Wpod.h"
#include <stdio.h>
#include <unistd.h>
#include "Wdriver.h"


#ifdef CUSTOM_SIGNAL_HANDLER
#include <signal.h>
void (*old_signal_handler)(int sig);

#define FAILED_SIGNAL()                                                 \
  {                                                                     \
    fprintf(stderr, "Failed to install signal handler (%s, %d)\n",      \
            __FILE__, __LINE__);                                        \
    exit(EXIT_FAILURE);                                                 \
  }
#define SET_HANDLER(sig, handler)                               \
  { if ((old_signal_handler = signal(sig, handler)) == SIG_ERR) \
      FAILED_SIGNAL();                                          \
  }


 /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */


static int
menu(void);


typedef void (*action_proc)(void);

typedef struct menu_choice_t {
  char            option;
  int             only_with_X;
  char           *descr;
  action_proc     action;
}              menu_choice_t;

static void
stop_choice(void);
static void
stop_choice(void)
{
  if (wasp_domain) WD_clean_up_before_exit(wasp_domain);
  exit(EXIT_SUCCESS);
}


#ifdef X_XFACE
static void
xiface_interact_choice(void);
static void
xiface_interact_choice(void)
{
  sprintf(W_buf(),
	  "Stopped by an interrupt.  Press ^P inside the view to continue.\n");
  xiface_loop(TRUE, W_buf());
}
#endif


static void
abort_choice(void);
static void
abort_choice(void)
{
  exit(EXIT_FAILURE);
}


static void
hexvue_choice(void);
static void
hexvue_choice(void)
{
  if (wasp_domain == NULL)
    return;
  
  options.write_postfile_file_now = TRUE;
  options.postfile_type = HEXVUE_POSTPROC_FILE;
}


static void
ve3d_choice(void);
static void
ve3d_choice(void)
{
  if (wasp_domain == NULL)
    return;
  
  options.write_postfile_file_now = TRUE;
  options.postfile_type = VE3D_POSTPROC_FILE;
}



static void
tecplot_choice(void);
static void
tecplot_choice(void)
{
  if (wasp_domain == NULL)
    return;
  
  options.write_postfile_file_now = TRUE;
  options.postfile_type = TECPLOT_I_POSTPROC_FILE;
}


static void
flushhist_choice(void);
static void
flushhist_choice(void)
{
  if (wasp_domain) WD_flush_history(wasp_domain);
}


static void
configure_choice(void);
static void
configure_choice(void)
{
  W_elem_t *el;
  W_index_t id;
  W_domain_t *d = wasp_domain;

#define MXC 1024
  char line[MXC];
#define GETINP()                                \
  {                                             \
    gets(line);                                 \
  }
  
  if (wasp_domain == NULL)
    return;

  fprintf(stderr, "Input element number <"); fflush(stderr);
  fflush(stdin);
  GETINP();
  if (tokens_parse_line(d->parser, line) < 1) {
    REPORT_WARNING("Invalid input");  fflush(stderr);
    return;
  }
  
  if (!TOKEN_IS_NUMBER(1)) {
    REPORT_WARNING("Expecting number"); fflush(stderr);
    return;
  } 
  
  id = TOKEN_INT(1);
  el = WD_get_elem_by_idx(wasp_domain, id);
  if (el == NULL) {
    REPORT_WARNING("No such element"); fflush(stderr);
    return;
  }

getopt:
  fprintf(stderr, "Input configuration option (<CR> to continue)<");
  fflush(stderr);
  GETINP();
  if (tokens_parse_line(d->parser, line) < 1) {
    return;
  }
  if (!WE_configure(el, line)) {
    REPORT_WARNING("Configuration failed"); fflush(stderr);
  }
  goto getopt;
}


static void
elinfo_choice(void);
static void
elinfo_choice(void)
{
  W_index_t id;
  W_domain_t *d = wasp_domain;
#define MXC 1024
  char line[MXC];
#define GETINP()                                \
  {                                             \
    gets(line);                                 \
  }
  
  if (wasp_domain == NULL)
    return;

  fprintf(stderr, "Input element number <"); fflush(stderr);
  fflush(stdin);
  GETINP();
  if (tokens_parse_line(d->parser, line) < 1) {
    REPORT_WARNING("Invalid input");  fflush(stderr);
    return;
  }
  
  if (!TOKEN_IS_NUMBER(1)) {
    REPORT_WARNING("Expecting number"); fflush(stderr);
    return;
  } 
  
  id = TOKEN_INT(1);
  {
    char buf[L_tmpnam] = "./tmp";
    FILE *fp;

    tmpnam(buf);
    fp = fopen(buf, "w");
    fprintf(stderr, "Writing element info to %s\n", buf);
    WD_print_elem_info(wasp_domain, id, W_INFO_FULL, fp);
    fclose(fp);
    fprintf(stderr, "Done writing %s\n", buf);
  }
}



static void
set_target_time_choice(void);
static void
set_target_time_choice(void)
{
  W_domain_t *d = wasp_domain;
#define MXC 1024
  char line[MXC];
#define GETINP()                                \
  {                                             \
    gets(line);                                 \
  }
  
  if (wasp_domain == NULL)
    return;

  fprintf(stderr, "Input target time <"); fflush(stderr);
  fflush(stdin);
  GETINP();
  if (tokens_parse_line(d->parser, line) < 1) {
    REPORT_WARNING("Invalid input");  fflush(stderr);
    return;
  }
  
  if (!TOKEN_IS_NUMBER(1)) {
    REPORT_WARNING("Expecting number"); fflush(stderr);
    return;
  } 
  
  options.target_time = TOKEN_DBL(1);

  return;
}



static menu_choice_t choices[] = {
  { 's',  FALSE, "stop (with cleanup)",             stop_choice },
  { 'a',  FALSE, "abort (no cleanup)",              abort_choice },
  { 'c',  FALSE, "continue",                        NULL },
  { 'T',  FALSE, "set target time",                 set_target_time_choice },
  { 'h',  FALSE, "write a HEXVUE file",             hexvue_choice },
  { 'v',  FALSE, "write a VE3D file  ",             ve3d_choice },
  { 't',  FALSE, "write a TECPLOT file",            tecplot_choice },
  { 'f',  FALSE, "flush history files",             flushhist_choice },
#ifdef X_XFACE
  { 'X',   TRUE, "enter X interface",               xiface_interact_choice },
#endif
  { 'e',  FALSE, "configure element",               configure_choice },
  { 'i',  FALSE, "element info",                    elinfo_choice },
  { '\0', FALSE, NULL,                              NULL }
};



static int received_an_interrupt = FALSE;

void 
wasp_signal_handler(int sig)
{
#define DO_GLUE(A, B)  A##B
#define GLUE(A, B)     DO_GLUE(A, B)
#define DO_STRNGZ(S)   #S
#define STRNGZ(S)      DO_STRNGZ(S) 
#define SIGNLNM(S)     GLUE(STRNGZ(SIG),STRNGZ(S))
#define PRERR(NAME)    fprintf(stderr,                                  \
                               "wasp_signal_handler: signal %s\n",      \
                               SIGNLNM(NAME))
#define CASE(NAME)                                                      \
   case GLUE(SIG,NAME):                                                 \
     PRERR(NAME);                                                       \
     SET_HANDLER(SIGINT, SIG_DFL);                                      \
     if (isatty (2)) {                                                  \
       fprintf(stderr, "      ^c will interrupt the program\n");        \
       fflush(stderr);                                                  \
     } else {                                                           \
       REPORT_FATAL_ERROR ("Caught signal " SIGNLNM(NAME));             \
     }                                                                  \
     break

  switch (sig) {
  CASE(FPE);
  CASE(SEGV);
  CASE(BUS);
  CASE(ILL);
  case SIGINT:
    {
      received_an_interrupt = TRUE;
      SET_HANDLER(SIGINT, &wasp_signal_handler);
    }
    break;
  default:
    fprintf(stderr, "wasp_signal_handler: signal %d\n", sig);
    fflush(stderr);
    exit(EXIT_FAILURE);
    break;
  }

  return;
}


int
W_check_interrupt(void)
{
  if (!received_an_interrupt) {
    return TRUE;
  } else {
    int c;
    signal(SIGINT, SIG_IGN);
    c = menu();
    if (choices[c].action) choices[c].action();
    SET_HANDLER(SIGINT, &wasp_signal_handler);
    received_an_interrupt = FALSE;
    return TRUE;
  }
}


static int
menu(void)
{
  int i, c;
  double realt, usert, syst;
  
  W_get_times(&realt, &usert, &syst);
  while (1) {
    fprintf(stderr, "Interrupt (problem time %g, user %g, sys %g).\n",
            wasp_domain ? WD_time(wasp_domain) : 0, usert, syst);
    i = 0;
    while (choices[i].option != '\0') {
      if ((!choices[i].only_with_X)
#         ifdef X_XFACE
	  || options.X_xface
#         endif
	  )
	fprintf(stderr, "   %c   %s\n", choices[i].option, choices[i].descr);
      i++;
    }
    fprintf(stderr,   "         Your choice? ");
    while (1) {
      c = fgetc(stdin);
      i = 0;
      while (choices[i].option != '\0') {
        if (choices[i].option == c
	    && ((!choices[i].only_with_X)
#               ifdef X_XFACE
		|| options.X_xface
#               endif
		))
          return i;
        i++;
      }
    }
  }
}

#else
int
W_check_interrupt(void)
{
  return TRUE;
}
#endif /* CUSTOM_SIGNAL_HANDLER */


int
run_wasp(int *argcp, char ***argvp)
{
  { /* this is to initialize the real time */
    double real_time, usr_time, systm_time;
    W_get_times(&real_time, &usr_time, &systm_time);
  }
  
  if ((wasp_domain = WD_new_domain (options.in_file_name, argcp, argvp)) == NULL) {
    REPORT_FATAL_ERROR("Failed create domain.");
  }
#ifdef X_XFACE
  if (options.X_xface) {
    WD_build_grep (wasp_domain, FALSE);
    sprintf (W_buf (), "layer on %d", WX_NODE_LAYER);
    ESIHandleCmd (W_buf ());
    sprintf (W_buf (), "layer on %d", WX_ELEM_LAYER);
    ESIHandleCmd (W_buf ());
    sprintf (W_buf (), "layer off %d", WX_NATURAL_CF_LAYER);
    ESIHandleCmd (W_buf ());
    sprintf (W_buf (), "layer off %d", WX_RIGID_CF_LAYER);
    ESIHandleCmd (W_buf ());
    sprintf (W_buf (), "layer off %d", WX_NATURAL_CE_LAYER);
    ESIHandleCmd (W_buf ());
    sprintf (W_buf (), "layer off %d", WX_RIGID_CE_LAYER);
    ESIHandleCmd (W_buf ());
    sprintf (W_buf (), "layer off %d", WX_ANNOT_LAYER);
    ESIHandleCmd (W_buf ());
    xiface_loop (TRUE, "Ready to run central difference integrator: ");
  }
#endif

  wasp_custom_setup (argcp, argvp, wasp_domain);
  
#ifdef CUSTOM_SIGNAL_HANDLER
  SET_HANDLER(SIGINT, &wasp_signal_handler);
#endif
  if (options.verbose) {
    sprintf (W_buf (), "Started: %s", ckit_get_date_and_time ());
    W_update_log (W_buf ());
  }
  if (options.integrator == CD_INTEGRATOR)
    WD_integrate (wasp_domain);
  else if (options.integrator == NEWMARK_INTEGRATOR)
    WD_integrate_implicit (wasp_domain);
  else {
    REPORT_FATAL_ERROR ("Unknown integrator");
  }

  if (wasp_domain->pod && W_pod_is_in_collect_mode (wasp_domain->pod)) {
    W_pod_compute_modes (wasp_domain->pod);
#   ifdef X_XFACE
    if (options.X_xface)
      xiface_loop (1, "Inspect POD modes");
    else 
#   endif
      {
        FILE *fp;
        sprintf (W_buf (), "%s.pod", options.in_file_basename);
        fp = fopen (W_buf (), "w");
        if (fp) {
          W_pod_write_pod (wasp_domain->pod, fp);
          fclose (fp);
        } 
      }
  }
  
  if (options.verbose) {
    sprintf (W_buf (), "Finished: %s", ckit_get_date_and_time ());
    W_update_log (W_buf ());
  }

  return 1;
}
