#! /bin/ksh -f

R=1
T=1
O=undefined
dt=0
conv_ratio=0.001
max_iter=25
report_iter=f
method=direct
nz_per_row=undefined
explicit=f
precond=lu
ksp=preonly
pod=f


function usage {
cat <<EOF
Usage: $(basename $0) wd_input [ options ]
Options:
 -R #                       (default $R)
 -T #                       (default $T)
 -O #                       (default $(echo "$T/100" | bc -l))
 -dt #                      (default $(echo "$T/200" | bc -l))
 -conv_ratio #              (default $conv_ratio)
 -method {direct|snapshots} (default $method)
 -max_iter #                (default $max_iter)
 -report_iter               (default $report_iter)
 -nz_per_row #              (default 0.01)
 -explicit                  (default $explicit)
 -precond method            (default $precond)
 -ksp method                (default $ksp)
 -pod                       (default $pod)
EOF
}

if [ $# -lt 1 ] ; then
    usage; exit 1;
fi
file=$1

args=""
while [ $# -ge 1 ] 
do
   a=$1; shift
   case $a in
    -R)
        if [ $# -ge 1 ] ; then
            R=$1; shift
        else
            usage; exit 1;
        fi
        ;;
    -O)
        if [ $# -ge 1 ] ; then
            O=$1; shift
        else
            usage; exit 1;
        fi
        ;;
    -T)
        if [ $# -ge 1 ] ; then
            T=$1; shift
        else
            usage; exit 1;
        fi
        ;;
    -dt)
        if [ $# -ge 1 ] ; then
            dt=$1; shift
        else
            usage; exit 1;
        fi
        ;;
    -conv_ratio)
        if [ $# -ge 1 ] ; then
            conv_ratio=$1; shift
        else
            usage; exit 1;
        fi
        ;;
    -max_iter)
        if [ $# -ge 1 ] ; then
            max_iter=$1; shift
        else
            usage; exit 1;
        fi
        ;;
    -method)
        if [ $# -ge 1 ] ; then
            method=$1; shift
        else
            usage; exit 1;
        fi
        ;;
    -precond)
        if [ $# -ge 1 ] ; then
            precond=$1; shift
        else
            usage; exit 1;
        fi
        ;;
    -ksp)
        if [ $# -ge 1 ] ; then
            ksp=$1; shift
        else
            usage; exit 1;
        fi
        ;;
    -explicit)
        explicit=t;
        ;;
    -pod)
        pod=t;
        ;;
    -report_iter)
        report_iter=t;
        ;;
    *)
        ta="$args $a"
        args="$ta"
        ;;
    esac
done

if [ x"$O" = xundefined ] ; then
   O=$(echo "$T/100" |bc -l)
fi

if [ $dt -eq 0 ] ; then
   dt=$(echo "$T/200" | bc -l)
fi

if [ x"$nz_per_row" = xundefined ] ; then
   nz_per_row=0.01
fi

if [ $pod = t ] ; then
    pod_options="-pod_options \"collect method $method dt $dt\""
else
    pod_options=
fi

if [ $report_iter = t ] ; then
    report_iter=report_iter
else
    report_iter=
fi

if [ $explicit = t ] ; then

command="\
wd -v -! -input $file -integrator 0 -R $R -W $file-full-expl-R$R \
-T $T -b 20 -O $O \
$pod_options \
$args"

else

command="\
wd -v -! -input $file -integrator 1 -R $R -W $file-full-R$R \
-T $T -b 20 -O $O \
$pod_options \
-newmarki_options \"conv_ratio $conv_ratio solver spetsc nz_per_row $nz_per_row max_iter $max_iter $report_iter\" \
-pc_type $precond -ksp_type $ksp \
$args"

fi
echo "$command"
eval $command
