#include "Wconfig.h"
#include "Wtypes.h"
#include "Wutil.h"
#include "Wcdi.h"
#include "Wdomain.h"
#include "Wnode.h"
#include "Welem.h"
#include "Wxf.h"
#include "Wcontact.h"
#include "Wtfuncs.h"


/* RCSHEADER "$Id: spencont.c,v 1.3 2004/04/27 20:10:11 cvs Exp $" */


  
void
WPCS_search (WPC_spheres_penalty_contact_t *pc)
{
  {
    if (pc->avgnmass <= 0)
      pc->avgnmass = WD_avg_nodal_mass (pc->cont_surf->domain);
  }

  if (pc->sl) {
    WPC_spheres_sphere_t *s = 0;
    while (s = (WPC_spheres_sphere_t *) get_list_next (pc->sl, s)) {
      double rtfv = WTF_func_val(s->radius_tf, WD_time(pc->cont_surf->domain));
      double xtfv = WTF_func_val(s->x_tf, WD_time(pc->cont_surf->domain));
      double ytfv = WTF_func_val(s->y_tf, WD_time(pc->cont_surf->domain));
      double ztfv = WTF_func_val(s->z_tf, WD_time(pc->cont_surf->domain));
      s->radius = rtfv;
      if (s->radius < 0) {
        REPORT_WARNING(pc->cont_surf->domain->errh, "Radius < 0");
        s->radius = 0;
      }
      s->center.x = xtfv;
      s->center.y = ytfv;
      s->center.z = ztfv;
    }
  }
}

W_contact_surf_t *
read_spheres_penalty_surf (W_domain_t *d)
{
  const char *input_line;
  W_contact_surf_t *cont_surf = NULL;

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line)) {
      return NULL;
    } else {
      TOKENS_PARSE (input_line);
      if (TOKENS_HAVE_KEYWORD ("ID")) {
        cont_surf = w_new (W_contact_surf_t);
        CHECK_MALLOC (cont_surf);
        cont_surf->id                       = TOKEN_INT (2);
        cont_surf->domain                   = d;
        cont_surf->type                     = W_SPHERES_PENALTY_CONTACT_SURF;
        cont_surf->celems                   = make_list ();
        cont_surf->search_eps               = FLT_MAX;
        cont_surf->all_rigid                = TRUE;

        cont_surf->data.spheres_penalty             = w_new (WPC_spheres_penalty_contact_t);
        CHECK_MALLOC (cont_surf->data.spheres_penalty);
        cont_surf->data.spheres_penalty->cont_surf            = cont_surf;
        cont_surf->data.spheres_penalty->first_time           = TRUE;
        cont_surf->data.spheres_penalty->force_type           = W_PENALTY_OVERLAP;
        cont_surf->data.spheres_penalty->overlap_penalty      = 1;
        cont_surf->data.spheres_penalty->overlap_rate_penalty = 1;
        cont_surf->data.spheres_penalty->avgnmass             = 0;
        cont_surf->data.spheres_penalty->sl                   = make_list();
      } else if (TOKENS_HAVE_KEYWORD ("FORCE_TYPE")) {
        if        (TOKENS_HAVE_KEYWORD ("RATE")) {
          cont_surf->data.spheres_penalty->force_type = W_PENALTY_OVERLAP_RATE;
        } else if (TOKENS_HAVE_KEYWORD ("COMB")) {
          cont_surf->data.spheres_penalty->force_type = W_PENALTY_OVERLAP_COMB;
        } else {
          cont_surf->data.spheres_penalty->force_type = W_PENALTY_OVERLAP;
        }
      } else if (TOKENS_HAVE_KEYWORD ("OVERLAP_PENALTY")) {
        cont_surf->data.spheres_penalty->overlap_penalty = TOKEN_DBL (2);
      } else if (TOKENS_HAVE_KEYWORD ("OVERLAP_RATE_PENALTY")) {
        cont_surf->data.spheres_penalty->overlap_rate_penalty = TOKEN_DBL (2);
      } else if (TOKENS_HAVE_KEYWORD ("SPHERE")) {
        W_index_t ridx, xidx, yidx, zidx;
        WPC_spheres_sphere_t *s = w_new(WPC_spheres_sphere_t);
        CHECK_MALLOC(s);
        xidx         = TOKEN_INT(2);
        yidx         = TOKEN_INT(3);
        zidx         = TOKEN_INT(4);
        ridx         = TOKEN_INT(5);
        s->radius_tf       = WD_get_time_func_by_idx (d, ridx);
        if (s->radius_tf == NULL) REPORT_FATAL_ERROR(d->errh, "No such time function");
        s->x_tf       = WD_get_time_func_by_idx (d, xidx);
        if (s->x_tf == NULL) REPORT_FATAL_ERROR(d->errh, "No such time function for center");
        s->y_tf       = WD_get_time_func_by_idx (d, yidx);
        if (s->y_tf == NULL) REPORT_FATAL_ERROR(d->errh, "No such time function for center");
        s->z_tf       = WD_get_time_func_by_idx (d, zidx);
        if (s->z_tf == NULL) REPORT_FATAL_ERROR(d->errh, "No such time function for center");

#if defined(X_XFACE)
        s->egf_file = 0;
        s->grepl = 0;
        s->ogrepl = 0;
        if (TOKENS_TOTAL() >= 6) {
          DUPLICATE_STRING(s->egf_file, TOKEN_STR(6));
        }
#endif
        if (cont_surf == NULL) REPORT_FATAL_ERROR(d->errh, "Need to define contact surf first");
        add_to_tail(cont_surf->data.spheres_penalty->sl, s);
      } else if (TOKENS_HAVE_KEYWORD ("END")) {
        return cont_surf;
      }
    }
  }

 
  return cont_surf;
}
