#include "Wconfig.h"
#include "Wxf.h"
#include "Wtypes.h"
#include "Wutil.h"
#include "Welem.h"
#include "Wdomain.h"
#include "Wnode.h"
#include "Wmats.h"
#include "Wnewmarki.h"
#include "Wcdi.h"
#include "Wdispbc.h"
#include "Wspring.h"
#include "Whistry.h"
#include "meschach.h"

/* RCSHEADER "$Id: spring.c,v 1.2 2004/09/06 01:17:58 cvs Exp $" */

/*
  The type of a truss element.  
 */
typedef struct priv_elem_t {
  W_node_t         *node; /* node ID's */
  W_loc_t           attachment_pt;
  W_float_t         Kspring;		  /* spring constant */
}              priv_elem_t;


#define GET_FROM_NODE(node, func, tovect)         \
   func (node, &tovect.x, &tovect.y, &tovect.z)


/*
  Declarations of element methods.
 */
static int
elem_elcode (W_elem_t *el);
static char *
type_name(W_elem_t *el);
static void
assemble_restoring_forces(W_elem_t *el, W_mat_state_update_form_enum update_form);
static void
assemble_mass(W_elem_t *el);
static W_float_t 
suggested_time_step(W_elem_t *el);
#ifdef X_XFACE
static EPixel
get_color()
{
  BOOLEAN suc;
  return ColorGetPixelFromString ("hotpink", &suc);
}
/*
  Modify the graphic representation of the element (a 3D line).
 */
static void
modify_elem_grep(W_elem_t *el, double displ_scale, W_bool_t immed_draw)
{
  WCRec pnts[2];
  W_coords_3d_t refc, u;
  W_material_t *m;
  priv_elem_t *t = (priv_elem_t *)el->data;

  GET_FROM_NODE (t->node, WN_get_ref_coords, refc);
  GET_FROM_NODE (t->node, WN_get_displ, u);

  VECMAC_3_VEC (pnts[0], =, t->attachment_pt, +, 0 * u);
  VECMAC_3_VEC (pnts[1], =, refc, +, displ_scale * u);

  if (el->grep == NULL) {
    GraphicObj *nm = 0;
    el->grep = CreateLine3D(pnts);
    EASValsSetLayer (WX_ELEM_LAYER);
    EASValsSetColor(get_color());
    EASValsSetLineWidth(6);
    EGWithMaskChangeAttributes(COLOR_MASK | LAYER_MASK, el->grep);
    EMNoDrawAddGraphicsToModel(ESIModel (), el->grep);
    nm = CreateMarker3D(&pnts[0]);
    EMNoDrawAddGraphicsToModel(ESIModel(), nm);
    EASValsSetColor(get_color());
    EASValsSetMType(CIRCLE_MARKER);
    EASValsSetMSize(2);
    EASValsSetLayer(WX_ELEM_LAYER);
    EGWithMaskChangeAttributes(COLOR_MASK |
                               MTYPE_MASK |
                               MSIZE_MASK |
                               LAYER_MASK,
                               nm);
  } else {
    EGModifyGraphicsByHandle(el->grep, 1, &pnts[0]);
    EGModifyGraphicsByHandle(el->grep, 2, &pnts[1]);
  }
}

#endif


static W_float_t
total_mass(W_elem_t *el);


static int
elem_elcode (W_elem_t *el)
{
  return SPRING_ELCODE;
}


/*
  Returns a unmodifiable string with the element name.
 */
static char *
type_name(W_elem_t *el)
{
  static char buf[] = "spring";
  return buf;
}




/*
  Assembles nodal forces by which the truss element acts on its nodes.
  All effects (restoring and damping forces) must be included.
 */
static void
assemble_restoring_forces (W_elem_t *el, W_mat_state_update_form_enum update_form)
{
  W_float_t nforce;
  W_float_t Nd0, Nd1;
  W_coords_3d_t c[2], force;
  W_loc_t d;
  priv_elem_t *t = (priv_elem_t *)el->data;

  /* kinematic quantities */
  GET_FROM_NODE (t->node, WN_get_ref_coords, c[0]);
  VECMAC_3_VEC (d, =, c[0], -, t->attachment_pt);
  Nd0 = VECMAC_NORM (d);
  GET_FROM_NODE (t->node, WN_get_coords, c[1]);
  VECMAC_3_VEC (d, =, c[1], -, t->attachment_pt);
  Nd1 = VECMAC_NORM (d);
  if (Nd1 > 0 && (Nd1 > Nd0)) {
    nforce = - t->Kspring * (Nd1 - Nd0) / Nd1;
    VECMAC_3_SCALAR (force, =, nforce * d.x, nforce * d.y, nforce * d.z);
    WN_assemble_force (t->node, &force);
    /* RAW: should be    el->deformation_energy = (nforce*nforce/t->Kspring/2); */
    el->deformation_energy = 0;
  }
}


/*
  Compute the total mass of the element.
 */
static W_float_t
total_mass(W_elem_t *el)
{
  return 0;
}


/*
  Assemble mass to the nodes.
 */
static void
assemble_mass (W_elem_t *el)
{
  /* VOID */
}


/*
  Compute the suggested time step.
 */
static W_float_t 
suggested_time_step(W_elem_t *el)
{
  return FLT_MAX;
}



static void
assemble_eff_stiffness (W_elem_t *el,
                        W_newmark_assemble_eff_stiff_proc assemble_eff_stiff_callback,
                        W_float_t stiff_mult,
                        W_float_t mass_mult,
                        W_mass_type_enum mass_type
                        )
{
  priv_elem_t *t = (priv_elem_t *)el->data;
  double tk[9]={0,0,0,0,0,0,0,0,0};
  int idcs[3], constrained[3];
  
  tk[0]=tk[4]=tk[8]=t->Kspring;

  WN_eqn_numbers_and_constraints (t->node, idcs, constrained);
  (*assemble_eff_stiff_callback) (el->domain->newmark_integrator, FLES_ADD, 3, idcs, tk, constrained);
}




/*
  Assembles nodal forces by which the truss element acts on its nodes.
  All effects (restoring and damping forces) must be included.
 */
static void
assemble_residual_forces (W_elem_t *el, W_mat_state_update_form_enum update_form,
                          W_mass_type_enum mass_type)
{
  assemble_restoring_forces (el, update_form);
}



static W_bool_t
get_elem_prop (W_elem_t *el, const char *prop, void *value)
{
    return FALSE;
}



/*
  Element method dispatch table.
 */
static W_elem_methods_t methods = {
  elem_elcode,
  type_name,
  assemble_restoring_forces,
  assemble_mass,
  total_mass,
  NULL,
  suggested_time_step,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
#ifdef X_XFACE
  modify_elem_grep,
#endif
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  assemble_residual_forces,
  assemble_eff_stiffness,
  NULL,
  NULL,
  NULL
};

/*
  Reads truss element data from stream as long as there is no 
  END_OF_BLOCK tag in the input line.  Creates elements. 
  */
W_bool_t 
spring_reader(W_domain_t *d)
{
  const char *input_line;
  long elem_num, nodeid;
  priv_elem_t *t;
#ifdef PVM_NEED_GHOST_ELEMENTS
  int is_ghost = FALSE;
#endif
  W_material_t *m;

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return TRUE;
    else {
      TOKENS_PARSE(input_line);
      if (TOKENS_HAVE_KEYWORD("DISPBC")) {
        int dispbcid = 0;
        double K = 0;
        W_bool_t attachment_given = FALSE;
        W_loc_t at;
        W_node_t *n;
        W_disp_bc_t *bc = 0;
        TOKENS_SET_BY_KEYWORD ("DISPBC", TRUE,
                               {
                                 dispbcid = TOKEN_INT_NEXT (1);
                               }, {
                               });
        bc = WD_get_disp_bc_by_idx (d, dispbcid);
        if (bc == 0) {
          sprintf (W_buf (), "No such displacement BC: %ld", dispbcid);
          REPORT_FATAL_ERROR (d->errh, W_buf ());
        }
        TOKENS_SET_BY_KEYWORD ("K", TRUE,
                               {
                                 K = TOKEN_DBL_NEXT (1);
                               }, {
                               });
        TOKENS_SET_BY_KEYWORD ("ATTACHMENT", FALSE,
                               {
                                 attachment_given = TRUE;
                                 at.x = TOKEN_DBL_NEXT (1);
                                 at.y = TOKEN_DBL_NEXT (2);
                                 at.z = TOKEN_DBL_NEXT (3);
                               }, {
                               });
        elem_num = 13000000 * dispbcid;
        n=0;
        while (n = WBC_get_next_node (bc, n)) {
          elem_num++;
          t = w_new(priv_elem_t);
          CHECK_MALLOC(t);
          t->node = n;
          t->Kspring = K;
          if (attachment_given) t->attachment_pt = at;
          else         {        GET_FROM_NODE (t->node, WN_get_ref_coords, t->attachment_pt); }
#ifdef PVM_NEED_GHOST_ELEMENTS
          is_ghost = FALSE;
          if (TOKENS_HAVE_KEYWORD("GHOST"))
            is_ghost = TRUE;
#endif
          WD_add_element(d, elem_num,
                         WE_new_element(d, &methods, t
#                    ifdef PVM_NEED_GHOST_ELEMENTS
                                        , is_ghost
#                    endif
                                        ));
        }
      } else {
        t = w_new(priv_elem_t);
        CHECK_MALLOC(t);
        elem_num    = TOKEN_INT(1);
        nodeid      = TOKEN_INT(2);
        t->node     = WD_get_node_by_idx (d, nodeid);
        if (t->node == NULL) {
          sprintf (W_buf (), "No such node: %ld", nodeid);
          REPORT_FATAL_ERROR (d->errh, W_buf ());
        }
        t->Kspring = TOKEN_INT(3);
#ifdef PVM_NEED_GHOST_ELEMENTS
        is_ghost = FALSE;
        if (TOKENS_HAVE_KEYWORD("GHOST"))
          is_ghost = TRUE;
#endif
        WD_add_element(d, elem_num,
                       WE_new_element(d, &methods, t
#                    ifdef PVM_NEED_GHOST_ELEMENTS
                                      , is_ghost
#                    endif
                                      ));
      }
    }
  }
  return FALSE;
}
