#include "listP.h"
#include "tokensP.h"
#include "getoptP.h"



static int nnodes, nedges, ntriangles, ntetras;
static int etype, edegree;
static tokens_parser PARSER = NULL;

#define GETLINE()                                                               \
  {                                                                             \
    if (!tokens_next_line (PARSER)) {                                           \
      fprintf (stderr, "Error in %s (l.%d)\n", __FILE__, __LINE__);             \
      fprintf (stderr, "Input: \"%s\", failed to get next line at l.%d\n",      \
               "-", tokens_curr_line_num (PARSER));                            \
      return FALSE;                                                             \
    }                                                                           \
  }

#define EXPECT_TOT(tot)                                                 \
  {                                                                     \
    if (tokens_total_of_tokens (PARSER) < tot) {                        \
      fprintf (stderr, "Error in %s (l.%d)\n", __FILE__, __LINE__);     \
      fprintf (stderr, "Expected %d tokens, got \"%s\"\n",              \
               tot, tokens_curr_line_contents (PARSER));                \
      return FALSE;                                                     \
    }                                                                   \
  }



int
main (int argc, char **argv)
{
  int i, id;
  double x, y, z;
  int mvids[10];
  int regid;
  
  if (argc != 2) {
    fprintf(stderr, "Usage: %s t3d_file_name\n", argv[0]);
    exit (1);
  }
  
  if (!PARSER) PARSER = tokens_new_parser ();

  if (!tokens_open_file (PARSER, argv[1])) {
    fprintf(stderr, "%s: t3d file \"%s\" could not be opened\n", argv[0], argv[1]);
    exit (1);
  }

  /* nnodes nedges ntriangles ntetras */
  GETLINE ();
  EXPECT_TOT (4);
  nnodes     = tokens_token_as_int (PARSER, 1);
  nedges     = tokens_token_as_int (PARSER, 2);
  ntriangles = tokens_token_as_int (PARSER, 3);
  ntetras    = tokens_token_as_int (PARSER, 4);

  /* etype edegree */
  GETLINE ();
  EXPECT_TOT (2);
  etype   = tokens_token_as_int (PARSER, 1);
  edegree = tokens_token_as_int (PARSER, 2);

  /* nodes */
  printf ("Nodes\n");
  for (i = 0; i < nnodes; i++) {
    GETLINE ();
    EXPECT_TOT (4);
    id    = tokens_token_as_int (PARSER, 1);
    x     = tokens_token_as_double (PARSER, 2);
    y     = tokens_token_as_double (PARSER, 3);
    z     = tokens_token_as_double (PARSER, 4);
    printf ("%d %g %g %g\n", id, x, y, z);
  }
  printf ("&\n");

  /* skip edges */
  for (i = 0; i < nedges; i++) {
    GETLINE ();
  }
  
  /* skip triangles */
  for (i = 0; i < ntriangles; i++) {
    GETLINE ();
  }
  
  /* tetras */
  printf ("Element\n");
  printf ("%d ! %d-noded tetrahedral\n", (edegree == 2? 8: 12), (edegree == 2? 10: 4));
  for (i = 0; i < ntetras; i++) {
    GETLINE ();
    if (edegree == 2) {
      EXPECT_TOT (13);
      id       = tokens_token_as_int (PARSER, 1);
      mvids[0] = tokens_token_as_int (PARSER, 2);
      mvids[1] = tokens_token_as_int (PARSER, 3);
      mvids[2] = tokens_token_as_int (PARSER, 4);
      mvids[3] = tokens_token_as_int (PARSER, 5);
      mvids[4] = tokens_token_as_int (PARSER, 6);
      mvids[5] = tokens_token_as_int (PARSER, 7);
      mvids[6] = tokens_token_as_int (PARSER, 8);
      mvids[7] = tokens_token_as_int (PARSER, 9);
      mvids[8] = tokens_token_as_int (PARSER, 10);
      mvids[9] = tokens_token_as_int (PARSER, 11);
      regid    = tokens_token_as_int (PARSER, 13);
      printf ("%d %d %d %d %d %d %d %d %d %d %d %d\n", id,
              mvids[0], mvids[1], mvids[2], mvids[3], mvids[4],
              mvids[5], mvids[6], mvids[7], mvids[8], mvids[9],
              regid);
    } else {
      EXPECT_TOT (7);
      id       = tokens_token_as_int (PARSER, 1);
      mvids[0] = tokens_token_as_int (PARSER, 2);
      mvids[1] = tokens_token_as_int (PARSER, 3);
      mvids[2] = tokens_token_as_int (PARSER, 4);
      mvids[3] = tokens_token_as_int (PARSER, 5);
      regid    = tokens_token_as_int (PARSER, 6);
      printf ("%d %d %d %d %d %d\n", id,
              mvids[0], mvids[1], mvids[2], mvids[3], 
              regid);
    }
  }
  printf ("&\n");
  
  tokens_close_curr_file (PARSER);

  return TRUE;
}
