
#include "Woctree.h"


static void
set_up_octree(void);


#define FOR(_Counter_, _Up_)                                      \
   for (_Counter_ = 0; _Counter_ < _Up_; _Counter_++)

#ifndef X_XFACE
static void run_test(void);
#endif

int 
main(int argc, char *argv[])
{
#ifdef X_XFACE
  ESIBuildInterface(ESI_GRAPHIC_EDITOR_MASK, argc, argv);

  EMSetAssocFringeTable(ESIModel(), ColorCreateFringeTable());
  ColorSetupFringeTableByMinMax(
            EMGetAssocFringeTable(ESIModel()), -1, 1);
         
  ESIPopup(); 

  ESIHandleCmd("COLOR YELLOW");
  ESIHandleCmd("LAYER ACTIVE 0");
  ESIHandleCmd("FILL OFF");
  ESIHandleCmd("NEWFRAME WIRE");
  ESIHandleCmd("load \"QUAD.egf\"");
  
  set_up_octree();

  ESIPopupAndRun();
#else
  set_up_octree(); run_test();
#endif

  return 1;
}



#ifdef X_XFACE
static void
extend_crack_CB(Widget w, XtPointer ptr, XtPointer call_data);

static double tsize = 0.05;

static void
set_tsize(void);

static void
set_tsize(void)
{
  tsize = TypeInGetTokenAsDouble(2);
}


/* new commands */
static top_command   table[] = {
  {"TSIZE",  5, set_tsize},
};


void
ESICustomize(Widget parent_pane)
{
  Widget w;
  Arg args[3];
  int argn;

  TypeInAugmentCommandTable(table, XtNumber(table));
  
  argn = 0;
  w = ESIAddButton("extend_crack", "Extend crack", commandWidgetClass,
                   parent_pane, args, argn, extend_crack_CB, NULL);
}



static octree_root_t *root = NULL;
static WCRec hexpts[8] = {
  {-1, -1, -1},
  { 1.2, -1, -1},
  { 1.3,  1, -1},
  {-1,  1, -1.02},
  {-1, -1.1,  1},
  { 0.8, -1,  1},
  { 1,  1,  1},
  {-0.9,  1,  1.3}
};

static void 
get_vertex_loc(owner_ptr owner, int vertex, loc_t *loc);

static void 
get_vertex_loc(owner_ptr owner, int vertex, loc_t *loc)
{
  WCRec *p = (WCRec *)owner;

  loc->x = p[vertex].x;
  loc->y = p[vertex].y;
  loc->z = p[vertex].z;
}

static void
set_up_octree(void)
{
  GraphicObj *h;

  h = CreateHexahedron(hexpts);
  EGWithMaskChangeAttributes(COLOR_MASK | FILL_MASK, h);
  EMNoDrawAddGraphicsToModel(ESIModel(), h);
  
  root = octree_new_root(hexpts, get_vertex_loc);
}


static void 
do_extend_crack(EView *v_p, caddr_t data_p);

static void
extend_crack_CB(Widget w, XtPointer ptr, XtPointer call_data)
{
  EVSetSelectActionFunction(do_extend_crack, NULL, NULL);
  EMInstallHandler(ESIModel(), EVSelectGraphicsHandler,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL);
  ESIHandleCmd("COLOR RED");
  ERptStatusMessage(1, 1, "Select quadrangle or triangle to define crack increment");
}


static void
get_loc(G3D_vertex_t v, double *x, double *y, double *z);

static void 
do_extend_crack(EView *v_p, caddr_t data_p)
{
  GraphicObj *g = NULL;
  int i;
  WCRec pts[8];
  int np;
  G3D_poly_t *poly;
  G3D_vertex_t vertices[4];
  
  if (EVItemsInSelectionList() <= 0) {
    ERptErrMessage(1, 2, "No objects selected.", ERROR_GRADE);
    goto exit;
  }

  while ((g = EVGetFromSelectionList()) != NULL) {
    if        (   EGGraphicType(g) == EG_TRIANGLE3D
               || EGGraphicType(g) == EG_QUAD3D) {
      if (!EGGetGraphicsGeometry(g, pts, &np))
        ERptStatusMessage(1, 1, "EGGetGraphicsGeometry failed");
      FOR(i, np) {
        vertices[i] = (G3D_vertex_t)&pts[i];
      }
      poly = G3D_new_poly(np, vertices, get_loc);
      if (poly == NULL) {
        ERptStatusMessage(1, 1, "Failed G3D_new_poly()");
        goto exit;
      }
      octree_extend_crack(root, poly, tsize);
    } 
  }

 exit:
  EVUninstallHandler(v_p);
  ERptStatusMessage(1, 1, "Done");
}


static void
get_loc(G3D_vertex_t v, double *x, double *y, double *z)
{
  WCRec *wc = (WCRec *)v;
  *x = wc->x;
  *y = wc->y;
  *z = wc->z;
}

#else  /* X_XFACE */

static octree_root_t *root = NULL;
static loc_t hexpts[8] = {
  {-1, -1, -1},
  { 1.2, -1, -1},
  { 1.3,  1, -1},
  {-1,  1, -1.02},
  {-1, -1.1,  1},
  { 0.8, -1,  1},
  { 1,  1,  1},
  {-0.9,  1,  1.3}
};

static void 
get_vertex_loc(owner_ptr owner, int vertex, loc_t *loc);

static void 
get_vertex_loc(owner_ptr owner, int vertex, loc_t *loc)
{
  loc_t *p = (loc_t *)owner;

  loc->x = p[vertex].x;
  loc->y = p[vertex].y;
  loc->z = p[vertex].z;
}

static void
set_up_octree(void)
{
  root = octree_new_root(hexpts, get_vertex_loc);
}

static void
get_loc(G3D_vertex_t v, double *x, double *y, double *z);

static void
run_test(void)
{
  G3D_poly_t *poly;
  G3D_vertex_t vertices[4];
  loc_t pts[4] = {
    {0.200060,-0.157474,-0.285233},
    {-0.235633,0.357304,-0.007709},
    {-0.152366,0.076296,0.304505},
    {0.223600,-0.389363,0.107926}
  };
  int i;
  
  FOR(i, 4) {
    vertices[i] = (G3D_vertex_t)&pts[i];
  }
  poly = G3D_new_poly(4, vertices, get_loc);
  octree_extend_crack(root, poly, 0.05);
}


static void
get_loc(G3D_vertex_t v, double *x, double *y, double *z)
{
  loc_t *wc = (loc_t *)v;
  *x = wc->x;
  *y = wc->y;
  *z = wc->z;
}


#endif /* X_XFACE */


