#include "Wconfig.h"
#include "Wtypes.h"
#include "Wtet10.h"
#include "Wdomain.h"
#include "Woptions.h"
#include "Wnode.h"
#include "Wmats.h"
#include "Wgaussp.h"
#include "Wxf.h"
#include "Wutil.h"
#include "Wcdi.h"
#include "g3d.h"
#include "Wmat3d.h"
#include "Whistry.h"
#include "Wcontface.h"

/* RCSHEADER "$Id: tet10.c,v 1.10 2004/06/08 04:32:01 cvs Exp $" */

#undef WBMAT_NNODES
#define WBMAT_NNODES 10

/* history client data */
typedef struct history_client_data_t {
  W_elem_t          *el; /* FE */
  W_index_t          pntidx; /* unique integer  */
  int                which_qp;  /* which quadrature point */
}              history_client_data_t;


 
#define GET_FROM_NODE(node, func, tovect)         \
   func (node, &tovect.x, &tovect.y, &tovect.z)


typedef struct priv_elem_t {
  W_node_t          *node[WBMAT_NNODES];
  W_material_t      *material;
  W_mat_3d_state_p  *matst;
}              priv_elem_t;


typedef struct gp_coord_t {
  double r, s, t, u;
}              gp_coord_t;

static gp_coord_t gps[4] = {
  { 0.13819660,0.13819660,0.13819660,0.58541020 },
  { 0.58541020,0.13819660,0.13819660,0.13819660 },
  { 0.13819660,0.58541020,0.13819660,0.13819660 },
  { 0.13819660,0.13819660,0.58541020,0.13819660 },
};
/* The following factor includes W_l=1/4 and c=1/6;
   see Hughes, The finite element method, 1987. */
#define GPW 0.041666666666666666667 /* = (1.0/4/6) */

typedef struct shpf_t {
  double n[WBMAT_NNODES];
  double n_x[WBMAT_NNODES];
  double n_y[WBMAT_NNODES];
  double n_z[WBMAT_NNODES];
  double J[3][3];
  double invJ[3][3];
  double detJ;
}              shpf_t;


typedef struct ip_ctx_data_t {
  W_domain_t *domain;
  priv_elem_t *pe;
  shpf_t      *sf;    /* shape function table */
}              ip_ctx_data_t;



#define ALL_GP_LOCS(i) gps[i].r, gps[i].s, gps[i].t, gps[i].u
#define ALL_REF_COORDS(H)                                                     \
 H->node[0]->ref_coords.x, H->node[0]->ref_coords.y, H->node[0]->ref_coords.z,\
 H->node[1]->ref_coords.x, H->node[1]->ref_coords.y, H->node[1]->ref_coords.z,\
 H->node[2]->ref_coords.x, H->node[2]->ref_coords.y, H->node[2]->ref_coords.z,\
 H->node[3]->ref_coords.x, H->node[3]->ref_coords.y, H->node[3]->ref_coords.z,\
 H->node[4]->ref_coords.x, H->node[4]->ref_coords.y, H->node[4]->ref_coords.z,\
 H->node[5]->ref_coords.x, H->node[5]->ref_coords.y, H->node[5]->ref_coords.z,\
 H->node[6]->ref_coords.x, H->node[6]->ref_coords.y, H->node[6]->ref_coords.z,\
 H->node[7]->ref_coords.x, H->node[7]->ref_coords.y, H->node[7]->ref_coords.z,\
 H->node[8]->ref_coords.x, H->node[8]->ref_coords.y, H->node[8]->ref_coords.z,\
 H->node[9]->ref_coords.x, H->node[9]->ref_coords.y, H->node[9]->ref_coords.z


static shpf_t sf;
  
#define IPOL10(n, v0, v1, v2, v3, v4, v5, v6, v7, v8, v9) \
        (  n[0] * (v0)                                 \
         + n[1] * (v1)                                 \
         + n[2] * (v2)                                 \
         + n[3] * (v3)                                 \
         + n[4] * (v4)                                 \
         + n[5] * (v5)                                 \
         + n[6] * (v6)                                 \
         + n[7] * (v7)                                 \
         + n[8] * (v8)                                 \
         + n[9] * (v9))

#define N_LOC(h, v, DIR, n)                          \
   v.DIR = (IPOL10 (n,                               \
                    h->node[0]->ref_coords.DIR,      \
                    h->node[1]->ref_coords.DIR,      \
                    h->node[2]->ref_coords.DIR,      \
                    h->node[3]->ref_coords.DIR,      \
                    h->node[4]->ref_coords.DIR,      \
                    h->node[5]->ref_coords.DIR,      \
                    h->node[6]->ref_coords.DIR,      \
                    h->node[7]->ref_coords.DIR,      \
                    h->node[8]->ref_coords.DIR,      \
                    h->node[9]->ref_coords.DIR))

#define LOC(h, gp)                              \
  {                                             \
    N_LOC(h, gp, x, sf.n);          \
    N_LOC(h, gp, y, sf.n);          \
    N_LOC(h, gp, z, sf.n);          \
  }

static void
assemble_restoring_forces(W_elem_t *el, W_mat_state_update_form_enum update_form);
static void
assemble_mass(W_elem_t *el);
static W_float_t 
suggested_time_step(W_elem_t *el);

#ifdef X_XFACE
static void
modify_elem_grep(W_elem_t *el, double displ_scale, W_bool_t immed_draw)
{
  priv_elem_t *tet = (priv_elem_t *)el->data;
  int i;
  W_vect_t refc, u;
  WCRec pnts[WBMAT_NNODES];

  for (i = 0; i < WBMAT_NNODES; i++) {
    GET_FROM_NODE (tet->node[i], WN_get_ref_coords, refc);
    GET_FROM_NODE (tet->node[i], WN_get_displ, u);
    pnts[i].x = refc.x + displ_scale * u.x;
    pnts[i].y = refc.y + displ_scale * u.y;
    pnts[i].z = refc.z + displ_scale * u.z;
  }

  if (el->grep == NULL) {
    el->grep = CreateTet10 (pnts);
    EASValsSetColor (WM_get_color (tet->material));
    EGWithMaskChangeAttributes (COLOR_MASK, el->grep);
    EGAttachObject (el->grep, (EObjectP)el);
    EMNoDrawAddGraphicsToModel (ESIModel (), el->grep);
  } else {
    EGModifyGraphicsGeometry (el->grep, pnts);
  }
}
#endif

static int
elem_elcode (W_elem_t *el)
{
  return TET10_ELCODE;
}


static char *
type_name (W_elem_t *el)
{
  static char buf[] = "tet10";
  return buf;
}


static W_float_t
total_mass(W_elem_t *el);
static void
write_postproc_file (W_elem_t *el, ckit_iobuf_t IOB, int type,
                     char *mode, void *optdata);



static void
fill_ip_pkg (double r, double s, double t, double u,
             int entry,
             G3D_loc_t *iploc,
             W_mat_3d_state_p matstate,
             W_mat_3d_ip_ctx_t *ip_ctx,
             W_mat_3d_t *mat,
             W_mat_3d_stress_type_enum stress_type
             );
static void
eval_at_extrapolated_pt (G3D_loc_t *pt,
                         double stress[3][3], double *eps,
                         double *vm);

static void
clear_pkgs (void);

static int 
num_of_faces (W_elem_t *el)
{
  return 4;
}

static int
get_face_vertex_nodes (W_elem_t *el, int nface, int ndim, W_node_t *n[])
{
  priv_elem_t *pe = (priv_elem_t *)el->data;
  int i;

  if (ndim >= 6) {
    FOR (i, 6) {
      n[i] = pe->node[g3dtet10_vertex_on_face[nface][i]];
    }    
  }
  return 6;
}

static int
get_contact_faces (W_elem_t *el, int nface, WCF_contact_face_t *cf[], int maxcf)
{
  priv_elem_t *pe = (priv_elem_t *)el->data;
  int i;
  
  if (maxcf < 4) {
    FOR (i, maxcf) { cf[i] = NULL; }
    return 4;
  }
  
#define DEFCF(cfn)                                                      \
  {                                                                     \
    cf[cfn] =                                                           \
       WCF_new_contact_face (                                           \
           pe->node[g3dtet10_subtri_verts[nface][cfn][0]],           \
           pe->node[g3dtet10_subtri_verts[nface][cfn][1]],           \
           pe->node[g3dtet10_subtri_verts[nface][cfn][2]], NULL);    \
  }

  DEFCF (0);
  DEFCF (1);
  DEFCF (2);
  DEFCF (3);

  return 4;
}


static int
which_qp (priv_elem_t *pe, W_loc_t *loc)
{
  int k, m;
  double n[WBMAT_NNODES];
  int closest = -1;
  double mindist, dist[4] = {FLT_MAX,FLT_MAX,FLT_MAX,FLT_MAX};
  W_loc_t l;
 
  for (k = 0; k < 4; k++) {
    G3D_tet10_N (n,
                 g3dtet10_vertex_param_coords[k][0],
                 g3dtet10_vertex_param_coords[k][1],
                 g3dtet10_vertex_param_coords[k][2],
                 g3dtet10_vertex_param_coords[k][3]);
    VECMAC_SCALAR (l, =, 0);
    FOR (m, WBMAT_NNODES) {
      VECMAC_2_VEC (l, +=, n[m] * pe->node[m]->ref_coords);
    }
    VECMAC_2_VEC (l, -=, (*loc));
    dist[k] = VECMAC_NORM (l);
  }

  closest = -1;
  mindist = FLT_MAX;
  FOR (k, 4) {
    if (dist[k] < FLT_MAX && dist[k] < mindist) {
      closest = k;
      mindist = dist[k];
    }
  }
  return closest;
}

static void
at_new_ip(void);
static W_bool_t 
get_kinem_var(W_mat_3d_ip_ctx_t *ctx, 
              W_mat_3d_kinem_var_type_enum  type,
              double kinvar[3][3]);

static void *
new_history_client_data (W_elem_t *e,
                         W_index_t idx, /* unique integer  */
                         W_history_writer_t *r, /* history writer */
                         char *input_line /* input line (to allow
                                             for additional parsing of
                                             the line contents by this
                                             function) */
                         )
{
  priv_elem_t *pe = (priv_elem_t *)e->data;
  history_client_data_t *cd;
  W_loc_t loc;
  W_domain_t *d = e->domain;
  
  cd = w_new (history_client_data_t);
  CHECK_MALLOC(cd);
 
  TOKENS_PARSE(input_line);
  cd->el     = e;
  cd->pntidx = TOKEN_INT(3);
  loc.x  = TOKEN_DBL(4);
  loc.y  = TOKEN_DBL(5);
  loc.z  = TOKEN_DBL(6);
  cd->which_qp = which_qp (pe, &loc);
  if (cd->which_qp < 0 && cd->which_qp >= 4)
    REPORT_FATAL_ERROR (d->errh, "Wrong quadrature point");

  return (NODE)cd;
}

static void
calc_stress (W_elem_t *el,
            int which_qp,
            W_output_stress_type_enum stress_type, /* stress type */
            double stress[3][3] /* output: stress */
            )
{
  priv_elem_t *pe = (priv_elem_t *)el->data;
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  W_mat_3d_stress_type_enum out_type;
  W_mat_3d_t *m = (W_mat_3d_t *)pe->material->data;
#undef EVAL_SF
#define EVAL_SF(k)                                      \
  {                                                     \
    G3D_tet10_interp_opt (sf.n,           \
                          sf.n_x,         \
                          sf.n_y,         \
                          sf.n_z,         \
                          &sf.detJ,       \
                          sf.J,           \
                          sf.invJ,        \
                          ALL_GP_LOCS(k),               \
                          ALL_REF_COORDS(pe));       \
  }

  ip_ctx_data.domain   = el->domain;
  ip_ctx_data.pe    = pe;
  ip_ctx.time          = WD_time (el->domain);
  ip_ctx.time_step     = WD_time_step (el->domain);
  ip_ctx.ctx_data      = &ip_ctx_data;
  ip_ctx.get_kinem_var = get_kinem_var;
  ip_ctx_data.sf       = &sf;
  
  EVAL_SF (which_qp);
  at_new_ip();
  if      (stress_type == W_OUTPUT_STRESS_CAUCHY) {
    if (!WM3D_update_stress (m, pe->matst[which_qp], W_MAT_STATE_UPDATE_TEMP,
                         &ip_ctx, MAT_3D_CAUCHY, stress, &out_type))
      REPORT_FATAL_ERROR(el->domain->errh, "Failed to compute stress");
    if (out_type != MAT_3D_CAUCHY)
      REPORT_FATAL_ERROR(el->domain->errh, "Stress type not available");
  } else if (stress_type == W_OUTPUT_STRESS_2ND_PK) {
    if (!WM3D_update_stress (m, pe->matst[which_qp], W_MAT_STATE_UPDATE_TEMP,
                         &ip_ctx, MAT_3D_2ND_PK, stress, &out_type))
      REPORT_FATAL_ERROR(el->domain->errh, "Failed to compute stress");
    if (out_type != MAT_3D_2ND_PK)
      REPORT_FATAL_ERROR(el->domain->errh, "Stress type not available");
  }
}



static void
get_equiv_pl_def (W_elem_t *el,
                  int which_qp,
                  double *equiv_pl_def
                  )
{
  priv_elem_t *pe = (priv_elem_t *)el->data;
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  W_mat_3d_t *m = (W_mat_3d_t *)pe->material->data;
#undef EVAL_SF
#define EVAL_SF(k)                                      \
  {                                                     \
    G3D_tet10_interp_opt (sf.n,           \
                          sf.n_x,         \
                          sf.n_y,         \
                          sf.n_z,         \
                          &sf.detJ,       \
                          sf.J,           \
                          sf.invJ,        \
                          ALL_GP_LOCS(k),               \
                          ALL_REF_COORDS(pe));       \
  }

  ip_ctx_data.domain   = el->domain;
  ip_ctx_data.pe    = pe;
  ip_ctx.time          = WD_time (el->domain);
  ip_ctx.time_step     = WD_time_step (el->domain);
  ip_ctx.ctx_data      = &ip_ctx_data;
  ip_ctx.get_kinem_var = get_kinem_var;
  ip_ctx_data.sf       = &sf;
  
  EVAL_SF (which_qp);
  at_new_ip();
  if (!WM3D_get_mat_state_var(m, pe->matst[which_qp],
                              &ip_ctx, MAT_3D_EQUIV_PLASTIC_DEF, equiv_pl_def))
    REPORT_FATAL_ERROR(el->domain->errh, "Failed to retrieve equiv plastic def");
}


static void
update_history (W_elem_t *e, /* generic element */
                W_solution_history_t *h, /* solution history */
                W_float_t time /* current time */
                )
{
  history_client_data_t *cd =
    (history_client_data_t *)h->client_data;
  double S[3][3], pressure, equiv_pl_def;
#define STORE(CD, HW, NAME, TIME, VAL)          \
  {                                             \
    char buf[128];                              \
    sprintf (buf, "p%d%s", CD->pntidx, NAME);   \
    WHR_store (HW, buf, TIME, VAL);             \
  }

  calc_stress (e, cd->which_qp,
              WOPT_OPT (h->writer->domain->options, output_stress_type),
              S);
  pressure  = (S[0][0] + S[1][1] + S[2][2]) / 3;
  if        (WOPT_OPT (h->writer->domain->options, output_stress_type) == W_OUTPUT_STRESS_CAUCHY) {
    STORE (cd, h->writer, "sxx", time, S[0][0]);
    STORE (cd, h->writer, "sxy", time, S[0][1]);
    STORE (cd, h->writer, "sxz", time, S[0][2]);
    STORE (cd, h->writer, "syy", time, S[1][1]);
    STORE (cd, h->writer, "syz", time, S[1][2]);
    STORE (cd, h->writer, "szz", time, S[2][2]);
    STORE (cd, h->writer, "press", time, pressure);
  } else if (WOPT_OPT (h->writer->domain->options, output_stress_type) == W_OUTPUT_STRESS_2ND_PK) {
    STORE (cd, h->writer, "Sxx", time, S[0][0]);
    STORE (cd, h->writer, "Sxy", time, S[0][1]);
    STORE (cd, h->writer, "Sxz", time, S[0][2]);
    STORE (cd, h->writer, "Syy", time, S[1][1]);
    STORE (cd, h->writer, "Syz", time, S[1][2]);
    STORE (cd, h->writer, "Szz", time, S[2][2]);
    STORE (cd, h->writer, "Press", time, pressure);
  } else {
    REPORT_ERROR(h->writer->domain->errh, "Unknown stress type for output");
  }
  get_equiv_pl_def (e, cd->which_qp, &equiv_pl_def);
  STORE (cd, h->writer, "EPlS", time, equiv_pl_def);
}









/* Returns nodal forces by which the  element acts on its nodal points. */

static void
assemble_restoring_forces (W_elem_t *el, W_mat_state_update_form_enum update_form)
{
  priv_elem_t *pe = (priv_elem_t *)el->data;
  W_mat_3d_t *m = (W_mat_3d_t *)pe->material->data;
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  double P[3][3];
  W_mat_3d_stress_type_enum out_type;
  W_vect_t f;
  double jdetw, dde;
  double dphix, dphiy, dphiz;
  int k, n;
#undef EVAL_SF
#define EVAL_SF(k)                                      \
  {                                                     \
    G3D_tet10_interp_opt (sf.n, sf.n_x, sf.n_y, sf.n_z, \
                          &sf.detJ, sf.J, sf.invJ,      \
                          ALL_GP_LOCS(k),               \
                          ALL_REF_COORDS(pe));          \
  }
#define COMP_F()                                            \
  {                                                         \
    f.x = -(P[0][0]*dphix + P[0][1]*dphiy + P[0][2]*dphiz); \
    f.y = -(P[1][0]*dphix + P[1][1]*dphiy + P[1][2]*dphiz); \
    f.z = -(P[2][0]*dphix + P[2][1]*dphiy + P[2][2]*dphiz); \
  }
  
  ip_ctx_data.domain   = el->domain;
  ip_ctx_data.pe       = pe;
  ip_ctx.time          = WD_time (el->domain);
  ip_ctx.time_step     = WD_time_step (el->domain);
  ip_ctx.ctx_data      = &ip_ctx_data;
  ip_ctx.get_kinem_var = get_kinem_var;
  ip_ctx_data.sf       = &sf;

  el->deformation_energy = 0;
  FOR(k, 4) {
    EVAL_SF (k);
    at_new_ip();
    if (!WM3D_update_stress(m, pe->matst[k], update_form,
                            &ip_ctx, MAT_3D_1ST_PK, P, &out_type))
      REPORT_ERROR(el->domain->errh, "Failed to compute stress");
    if (!WM3D_get_mat_state_var (m, pe->matst[k],
                                 &ip_ctx, MAT_3D_DEFORM_ENERGY_DENSITY,
                                 &dde))
      REPORT_ERROR(el->domain->errh, "Failed to retrieve work of deformation");
    jdetw = GPW * sf.detJ;
    el->deformation_energy += dde * jdetw;
    FOR(n, WBMAT_NNODES) {
      dphix = jdetw * sf.n_x[n];
      dphiy = jdetw * sf.n_y[n];
      dphiz = jdetw * sf.n_z[n];
      COMP_F();
      WN_assemble_force(pe->node[n], &f);
    }
  } /* FOR(k, 2) */
#undef  EVAL_SF
}

static void
assemble_mass (W_elem_t *el)
{
  priv_elem_t *pe = (priv_elem_t *)el->data;
  int i;
  double tm = total_mass(el);

  for (i = 0; i < 4; i++)  
    WN_assemble_mass (pe->node[i], tm / 32.0);
  for (i = 4; i < WBMAT_NNODES; i++)  
    WN_assemble_mass (pe->node[i], 7 * tm / 48.0);
}


static W_float_t 
suggested_time_step_orig (W_elem_t *el)
{
  priv_elem_t *pe = (priv_elem_t *)el->data;
  W_vect_t refc;
  G3D_box_t b;
  int k;
  double E, nu, rho, xs, ys, zs, mins, c_d;
  W_mat_3d_t *m = (W_mat_3d_t *)pe->material->data;

  GET_FROM_NODE (pe->node[0], WN_get_ref_coords, refc);
  G3D_INIT_BOX (b, refc);
  for (k = 1; k < WBMAT_NNODES; k++) {
    GET_FROM_NODE (pe->node[k], WN_get_ref_coords, refc);
    G3D_UPDT_BOX (b, refc);
  }
  xs = G3D_BOX_RANGE (b, x);
  ys = G3D_BOX_RANGE (b, y);
  zs = G3D_BOX_RANGE (b, z);
  mins = xs; mins = min (mins, ys); mins = min (mins, zs);

  if (!WM3D_get_mat_const(m, MAT_3D_YOUNG_MODULUS, &E))
    REPORT_FATAL_ERROR(el->domain->errh, "No Young's modulus");
  if (!WM3D_get_mat_const(m, MAT_3D_POISSON, &nu))
    REPORT_FATAL_ERROR(el->domain->errh, "No Poisson constant");
  rho = WM3D_get_volume_mass (m);
  c_d = W_3D_dilatational_wave_speed (E, nu, rho);
  return mins / c_d;
}


static W_float_t 
suggested_time_step (W_elem_t *el)
{
  priv_elem_t *pe = (priv_elem_t *)el->data;
  W_vect_t v, p1, p2;
  int j, k;
  double E, nu, rho, c_d;
  W_mat_3d_t *m = (W_mat_3d_t *)pe->material->data;
  double d, mind = FLT_MAX;
  
  if (!WM3D_get_mat_const(m, MAT_3D_YOUNG_MODULUS, &E))
    REPORT_FATAL_ERROR(el->domain->errh, "No Young's modulus");
  if (!WM3D_get_mat_const(m, MAT_3D_POISSON, &nu))
    REPORT_FATAL_ERROR(el->domain->errh, "No Poisson constant");
  rho = WM3D_get_volume_mass (m);
  c_d = W_3D_dilatational_wave_speed (E, nu, rho);
 
  FOR(j, WBMAT_NNODES) {
    GET_FROM_NODE (pe->node[j], WN_get_ref_coords, p1);
    for (k = j+1; k < WBMAT_NNODES; k++) {
      GET_FROM_NODE (pe->node[k], WN_get_ref_coords, p2);
      VECMAC_3_VEC(v, =, p2, -, p1);
      d = VECMAC_NORM(v);
      mind = min(d, mind);
    }
  }
  
  return mind / c_d;
}

static W_float_t
total_mass(W_elem_t *el)
{
  priv_elem_t *pe = (priv_elem_t *)el->data;
  W_mat_3d_t *m = (W_mat_3d_t *)pe->material->data;
  W_float_t volume_mass, J[3][3], invJ[3][3];
  int k;
  double volume = 0, detJ;

  volume_mass = WM3D_get_volume_mass(m);

  FOR (k, 4) {
    G3D_tet10_J (J, ALL_GP_LOCS (k), ALL_REF_COORDS (pe));
    detJ    = G3D_tet10_detJ(J);
    if (detJ <= 0) {
      sprintf (W_buf (), "Tet10 %d %d %d %d %d %d %d %d %d %d: Non-positive Jacobian",
               pe->node[0]->id,
               pe->node[1]->id,
               pe->node[2]->id,
               pe->node[3]->id,
               pe->node[4]->id,
               pe->node[5]->id,
               pe->node[6]->id,
               pe->node[7]->id,
               pe->node[8]->id,
               pe->node[9]->id
               );
      REPORT_FATAL_ERROR (el->domain->errh, W_buf());
    }
    volume += GPW * detJ;
  }

  return (volume_mass * volume);
}

typedef struct ipol_pkg_t {
  G3D_loc_t       iploc;
  double          stress[3][3];
  double          equiv_pl_strain;
  double          von_mises;
}              ipol_pkg_t;


static ipol_pkg_t ipol_pkgs[4];



static void
clear_pkgs (void)
{
  int i;
  G3D_loc_t zero = {0,0,0};

  FOR (i, 4) {
    ipol_pkgs[i].iploc           = zero;
    MAT3DUTIL_ZERO_3X3 (ipol_pkgs[i].stress);
    ipol_pkgs[i].equiv_pl_strain = 0;
    ipol_pkgs[i].von_mises       = 0;
  }
}


static void
write_hexvue_file (W_elem_t *el, ckit_iobuf_t IOB,
                     int type, char *mode, void *optdata)
{
  priv_elem_t *pe = (priv_elem_t *)el->data;
  int i, j, k;
  W_vect_t refc[15], u[15];
  W_mat_3d_t *m = (W_mat_3d_t *)pe->material->data;
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  int ival;
  int reqlen;
  G3D_loc_t *pts[15], iploc;
  static int buflen        = 0;
  static double *values[15] = { NULL,NULL,NULL,NULL,NULL,
                                NULL,NULL,NULL,NULL,NULL,
                                NULL,NULL,NULL,NULL,NULL
                              };
  double *hvals[8] = { NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL };
  G3D_loc_t *hpts[8];
  int hexi[4][8] = {
    {10,5,2,6,14,12,9,13},
    {4,1,5,10,11,8,12,14},
    {0,4,10,6,7,11,14,13},
    {7,11,14,13,3,8,12,9}
  };

#define EVAL_SF(k)                                \
  {                                               \
    G3D_tet10_interp_opt (sf.n,     \
                       sf.n_x,      \
                       sf.n_y,      \
                       sf.n_z,      \
                       &sf.detJ,    \
                       sf.J,        \
                       sf.invJ,     \
                       ALL_GP_LOCS(k),            \
                       ALL_REF_COORDS(pe));    \
  }

#if 0
#define HAVENODEID(tet, n)                                      \
  ((pe->node[0]->id == n)||(pe->node[1]->id == n)||       \
   (pe->node[2]->id == n)||(pe->node[3]->id == n))
  if (   HAVENODEID (pe, 43)
      && HAVENODEID (pe, 46)
      && HAVENODEID (pe, 47)
      && HAVENODEID (pe, 48))
    fprintf (stderr, "43 46 47 48\n");
#endif
  W_mat_3d_stress_type_enum outtype;
  W_output_stress_type_enum output_stress_type
    = WOPT_OPT (el->domain->options, output_stress_type);
  int postfile_output_spec
    = WOPT_OPT (el->domain->options, postfile_output_spec_deprec);

  outtype = (output_stress_type == W_OUTPUT_STRESS_CAUCHY
             ? MAT_3D_CAUCHY
             : (output_stress_type == W_OUTPUT_STRESS_2ND_PK
                ? MAT_3D_2ND_PK
                : MAT_3D_CAUCHY)
             );
             

  if (type == HEXVUE_POSTPROC_FILE) {
    double S[3][3] = {{0,0,0},{0,0,0},{0,0,0}};
    double equiv_pl_def = 0, von_mises = 0;

    reqlen = 0;
    if (postfile_output_spec & W_POSTFILE_DISPL_MASK) 
      reqlen += 3;
    if (postfile_output_spec & W_POSTFILE_STRESS_MASK)
      reqlen += 6;
    if (postfile_output_spec & W_POSTFILE_EQPLS_MASK)
      reqlen++;
    if (postfile_output_spec & W_POSTFILE_VON_MISES_MASK)
      reqlen++;
    
    if (buflen < reqlen) {
      FOR (i, 15) {
        values[i] = NEWP_REALLOC_OBJ_ARRAY (values[i], reqlen, double);
      }
      buflen = reqlen;
    }
  
    FOR (i, WBMAT_NNODES) {
      GET_FROM_NODE (pe->node[i], WN_get_ref_coords, refc[i]);
      GET_FROM_NODE (pe->node[i], WN_get_displ, u[i]);
    }

    ip_ctx_data.domain   = el->domain;
    ip_ctx_data.pe    = pe;
    ip_ctx.time          = WD_time (el->domain);
    ip_ctx.time_step     = WD_time_step (el->domain);
    ip_ctx.ctx_data      = &ip_ctx_data;
    ip_ctx.get_kinem_var = get_kinem_var;
    ip_ctx_data.sf       = &sf;

    clear_pkgs ();
    FOR(k, 4) {
      EVAL_SF (k);
      at_new_ip();
      LOC (pe, iploc);
      fill_ip_pkg (ALL_GP_LOCS(k), k, &iploc,
                   pe->matst[k], &ip_ctx, m, outtype);
    }

    for (i = 10; i < 15; i++) {
      G3D_tet10_N (sf.n,
                   g3dtet10_vertex_param_coords[i][0],
                   g3dtet10_vertex_param_coords[i][1],
                   g3dtet10_vertex_param_coords[i][2],
                   g3dtet10_vertex_param_coords[i][3]);
      LOC (pe, refc[i]);
      u[i].x = IPOL10 (sf.n,
                       u[0].x, u[1].x, u[2].x, u[3].x, u[4].x, 
                       u[5].x, u[6].x, u[7].x, u[8].x, u[9].x); 
      u[i].y = IPOL10 (sf.n,
                       u[0].y, u[1].y, u[2].y, u[3].y, u[4].y, 
                       u[5].y, u[6].y, u[7].y, u[8].y, u[9].y); 
      u[i].z = IPOL10 (sf.n,
                       u[0].z, u[1].z, u[2].z, u[3].z, u[4].z, 
                       u[5].z, u[6].z, u[7].z, u[8].z, u[9].z); 
    }

    FOR (k, 15) {
      pts[k] = &refc[k];
      ival   = 0;
      if (postfile_output_spec & W_POSTFILE_DISPL_MASK) {
        values[k][ival] = u[k].x; ival++;
        values[k][ival] = u[k].y; ival++;
        values[k][ival] = u[k].z; ival++;
      }
      
      eval_at_extrapolated_pt (pts[k], S, &equiv_pl_def, &von_mises);
      
      if (postfile_output_spec & W_POSTFILE_STRESS_MASK) {
        values[k][ival] = S[0][0]; ival++;
        values[k][ival] = S[0][1]; ival++;
        values[k][ival] = S[0][2]; ival++;
        values[k][ival] = S[1][1]; ival++;
        values[k][ival] = S[1][2]; ival++;
        values[k][ival] = S[2][2]; ival++;
      }
      if (postfile_output_spec & W_POSTFILE_EQPLS_MASK) {
        values[k][ival] = equiv_pl_def; ival++;
      }
      if (postfile_output_spec & W_POSTFILE_VON_MISES_MASK) {
        values[k][ival] = von_mises; ival++;
      }
      if (ival > reqlen) REPORT_FATAL_ERROR (el->domain->errh, "wrong");
    }

    FOR (i,  4) {
      FOR (j, 8) {
        hvals[j] = values[hexi[i][j]];
        hpts[j]  = pts[hexi[i][j]];
      }
      W_write_hexvue_hex (IOB, hpts, reqlen, hvals, mode);
    }
  } else if (type == TECPLOT_I_POSTPROC_FILE) {
    /* RAW */
  } /* type */
#undef  EVAL_SF
}

static void
write_hexvue_XTRA_file (W_elem_t *el, ckit_iobuf_t IOB,
                        int type, char *mode, void *optdata)
{
  priv_elem_t *pe = (priv_elem_t *)el->data;
  int hi, i, j, k, ival;
  W_vect_t refc[15], u[15];
  W_mat_3d_t *m = (W_mat_3d_t *)pe->material->data;
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  int reqlen;
  int buflen        = 0;
  W_mat_3d_stress_type_enum outtype;
  W_output_stress_type_enum output_stress_type
    = WOPT_OPT (el->domain->options, output_stress_type);
  W_postfile_output_specs_t *os = &el->domain->options->postfile_output_specs_list;
  double *hvals[8] = { NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL };
  G3D_loc_t *hpts[8];
  int hexi[4][8] = {
    {10,5,2,6,14,12,9,13},
    {4,1,5,10,11,8,12,14},
    {0,4,10,6,7,11,14,13},
    {7,11,14,13,3,8,12,9}
  };

#define EVAL_SF(k)                                \
  {                                               \
    G3D_tet10_interp_opt (sf.n,     \
                       sf.n_x,      \
                       sf.n_y,      \
                       sf.n_z,      \
                       &sf.detJ,    \
                       sf.J,        \
                       sf.invJ,     \
                       ALL_GP_LOCS(k),            \
                       ALL_REF_COORDS(pe));    \
  }

  outtype = (output_stress_type == W_OUTPUT_STRESS_CAUCHY
             ? MAT_3D_CAUCHY
             : (output_stress_type == W_OUTPUT_STRESS_2ND_PK
                ? MAT_3D_2ND_PK
                : MAT_3D_CAUCHY)
             );
             
  {
    reqlen = os->nspecs;
    
    if (buflen < reqlen) {
      FOR (i, 8) {
        hvals[i] = NEWP_REALLOC_OBJ_ARRAY (hvals[i], reqlen, double);
      }
      buflen = reqlen;
    }
  
    ip_ctx_data.domain   = el->domain;
    ip_ctx_data.pe       = pe;
    ip_ctx.time          = WD_time (el->domain);
    ip_ctx.time_step     = WD_time_step (el->domain);
    ip_ctx.ctx_data      = &ip_ctx_data;
    ip_ctx.get_kinem_var = get_kinem_var;
    ip_ctx_data.sf       = &sf;

    FOR (i, 10) {
      GET_FROM_NODE (pe->node[i], WN_get_ref_coords, refc[i]);
      GET_FROM_NODE (pe->node[i], WN_get_displ, u[i]);
    }

    for (i = 10; i < 15; i++) {
      G3D_tet10_N (sf.n,
                   g3dtet10_vertex_param_coords[i][0],
                   g3dtet10_vertex_param_coords[i][1],
                   g3dtet10_vertex_param_coords[i][2],
                   g3dtet10_vertex_param_coords[i][3]);
      LOC (pe, refc[i]);
      u[i].x = IPOL10 (sf.n,
                       u[0].x, u[1].x, u[2].x, u[3].x, u[4].x, 
                       u[5].x, u[6].x, u[7].x, u[8].x, u[9].x); 
      u[i].y = IPOL10 (sf.n,
                       u[0].y, u[1].y, u[2].y, u[3].y, u[4].y, 
                       u[5].y, u[6].y, u[7].y, u[8].y, u[9].y); 
      u[i].z = IPOL10 (sf.n,
                       u[0].z, u[1].z, u[2].z, u[3].z, u[4].z, 
                       u[5].z, u[6].z, u[7].z, u[8].z, u[9].z); 
    }

    FOR (hi, 4) {
      FOR (k, 8) {
        hpts[k]=&refc[hexi[hi][k]];
        FOR (ival, reqlen) {
          if        (tokens_token_and_keyword_equiv (os->specs[ival], "displ_x")) {
              hvals [k][ival]   = u[hexi[hi][k]].x; 
          } else if (tokens_token_and_keyword_equiv (os->specs[ival], "displ_y")) { 
              hvals[k][ival]=u[hexi[hi][k]].y;
          } else if (tokens_token_and_keyword_equiv (os->specs[ival], "displ_z")) { 
              hvals[k][ival]=u[hexi[hi][k]].z;
          } else { /* quadrature-point data */
            EVAL_SF (hi);
            at_new_ip();
            {
              double stress[3][3];
              double equiv_pl_strain=0, von_mises=0, press=0;
              W_mat_3d_stress_type_enum actual_outtype;
              
              if (!WM3D_update_stress (m, pe->matst[hi], W_MAT_STATE_UPDATE_TEMP,
                                       &ip_ctx, outtype, stress,
                                       &actual_outtype)) {
                REPORT_ERROR (NULL, "Failed to compute stress");
                return;
              }
              if (actual_outtype != outtype)
                REPORT_WARNING (NULL, "output stress type mismatch");
              
              if (!WM3D_get_mat_state_var (m, pe->matst[hi],
                                           &ip_ctx, MAT_3D_EQUIV_PLASTIC_DEF,
                                           &equiv_pl_strain))
                REPORT_FATAL_ERROR(NULL, "Failed to retrieve equiv plastic def");
              
              press = (stress[0][0] + stress[1][1] + stress[2][2]) / 3;
              {
                double s00 = stress[0][0] - press;
                double s11 = stress[1][1] - press;
                double s22 = stress[2][2] - press;
                
                von_mises
                  = sqrt(1.5
                         * (  s00*s00
                              + stress[0][1]*stress[0][1]
                              + stress[0][2]*stress[0][2]
                              + stress[1][0]*stress[1][0]
                              + s11*s11
                              + stress[1][2]*stress[1][2]
                              + stress[2][0]*stress[2][0]
                              + stress[2][1]*stress[2][1]
                              + s22*s22
                              )
                         );
              }
              
              if (tokens_token_and_keyword_equiv (os->specs[ival], "s_xx")) { 
                hvals[k][ival] = stress[0][0];
              } else if (tokens_token_and_keyword_equiv (os->specs[ival], "s_xy")) { 
                hvals[k][ival] = stress[0][1];
              } else if (tokens_token_and_keyword_equiv (os->specs[ival], "s_xz")) { 
                hvals[k][ival] = stress[0][2];
              } else if (tokens_token_and_keyword_equiv (os->specs[ival], "s_yy")) { 
                hvals[k][ival] = stress[1][1];
              } else if (tokens_token_and_keyword_equiv (os->specs[ival], "s_yz")) { 
                hvals[k][ival] = stress[1][2];
              } else if (tokens_token_and_keyword_equiv (os->specs[ival], "s_zz")) { 
                hvals[k][ival] = stress[2][2];
              } else if (tokens_token_and_keyword_equiv (os->specs[ival], "vonm")) { 
                hvals[k][ival] = von_mises;
              } else if (tokens_token_and_keyword_equiv (os->specs[ival], "press")) { 
                hvals[k][ival] = press;
              } else if (tokens_token_and_keyword_equiv (os->specs[ival], "eqpls")) { 
                hvals[k][ival] = equiv_pl_strain;
              } else { /* try to get the named variable from the material state */
                double val = 0;
                if (WM3D_get_mat_state_named_var (m, pe->matst[hi],
                                                  &ip_ctx, os->specs[ival],
                                                  &val)) {
                  hvals[k][ival] = val;
                } else
                  hvals[k][ival] = 0;
              }
            }
          }
        }
      } 
      W_write_hexvue_hex (IOB, hpts, reqlen, hvals, mode);
    }
  }
  FOR (i, 8) {
    w_free (hvals[i]);
  }
#undef EVAL_SF
}

static double
tet_volume (priv_elem_t *pe)
{
  double x0 = pe->node[0]->ref_coords.x;
  double y0 = pe->node[0]->ref_coords.y;
  double z0 = pe->node[0]->ref_coords.z;
  double x1 = pe->node[1]->ref_coords.x;
  double y1 = pe->node[1]->ref_coords.y;
  double z1 = pe->node[1]->ref_coords.z;
  double x2 = pe->node[2]->ref_coords.x;
  double y2 = pe->node[2]->ref_coords.y;
  double z2 = pe->node[2]->ref_coords.z;
  double x3 = pe->node[3]->ref_coords.x;
  double y3 = pe->node[3]->ref_coords.y;
  double z3 = pe->node[3]->ref_coords.z;

  double a = x1-x0;
  double b = y1-y0;
  double c = z1-z0;
  double d = x2-x0;
  double e = y2-y0;
  double f = z2-z0;
  double g = x3-x0;
  double h = y3-y0;
  double i = z3-z0;
  
  double VOL = a*e*i - a*h*f - b*d*i + b*g*f + c*d*h - c*g*e;
  
  return  VOL / 6;
} 
  
static double
tet_rms_edge_len (priv_elem_t *pe)
{
  double l_rms_2 = 0, l_2;
  CkitPoint3 v;
  int i;

  FOR (i, 6) {
    VECMAC_3_VEC (v, =,
                  pe->node[g3dtet_vertex_on_edge[i][0]]->ref_coords, -,
                  pe->node[g3dtet_vertex_on_edge[i][1]]->ref_coords);
    l_2     = VECMAC_DOT (v, v);
    l_rms_2 += l_2;
  }
  return sqrt (l_rms_2 / 6);
}

static double
tet_gamma_quality (priv_elem_t *pe)
{
  double vol = tet_volume (pe);
  double l_rms = tet_rms_edge_len (pe);
  return (vol / (l_rms * l_rms * l_rms) * 8.47967); 
}


#define VALID(which) which##_is_valid

#define DCL_TENSOR(T)                           \
   static double T[3][3];                       \
   static W_bool_t VALID(T) = FALSE
 
DCL_TENSOR(F);
DCL_TENSOR(prev_F);
DCL_TENSOR(inv_F);
DCL_TENSOR(rel_F);
DCL_TENSOR(midstep_F);
DCL_TENSOR(C);
DCL_TENSOR(E);
DCL_TENSOR(D);
DCL_TENSOR(midstep_D);
DCL_TENSOR(midstep_W);
DCL_TENSOR(F_dot);
DCL_TENSOR(L);
DCL_TENSOR(R);
DCL_TENSOR(U);
DCL_TENSOR(rel_R);
DCL_TENSOR(rel_U);
DCL_TENSOR(ln_rel_U);

static void
at_new_ip(void)
{
#define INVALIDATE(which) VALID(which) = FALSE
  
  INVALIDATE(F);
  INVALIDATE(inv_F);
  INVALIDATE(prev_F);
  INVALIDATE(rel_F);
  INVALIDATE(midstep_F);
  INVALIDATE(C);
  INVALIDATE(E);
  INVALIDATE(D);
  INVALIDATE(midstep_D);
  INVALIDATE(midstep_W);
  INVALIDATE(F_dot);
  INVALIDATE(L);
  INVALIDATE(R);
  INVALIDATE(U);
  INVALIDATE(rel_R);
  INVALIDATE(rel_U);
  INVALIDATE(ln_rel_U);
}

static void
compute_def_grad(ip_ctx_data_t *ipctx)
{
  priv_elem_t *pe = ipctx->pe;
  shpf_t *sftab = ipctx->sf;
  int k;
  double f00 = 0, f01 = 0, f02 = 0;
  double f10 = 0, f11 = 0, f12 = 0;
  double f20 = 0, f21 = 0, f22 = 0;
  W_node_t *n;
  double x, y, z;
  register double dphix, dphiy, dphiz;

  /* Deformation gradient */
  FOR(k, WBMAT_NNODES) {
    n = pe->node[k];
    WN_get_coords (n, &x, &y, &z);
    dphix  = sftab->n_x[k];
    dphiy  = sftab->n_y[k];
    dphiz  = sftab->n_z[k];
    f00 += dphix * x;
    f01 += dphiy * x;
    f02 += dphiz * x;
    f10 += dphix * y;
    f11 += dphiy * y;
    f12 += dphiz * y;
    f20 += dphix * z;
    f21 += dphiy * z;
    f22 += dphiz * z;
  }

  F[0][0] = f00;
  F[0][1] = f01;
  F[0][2] = f02;
  
  F[1][0] = f10;
  F[1][1] = f11;
  F[1][2] = f12;
  
  F[2][0] = f20;
  F[2][1] = f21;
  F[2][2] = f22;

  VALID(F) = TRUE;
}

static void
compute_rel_def_grad(ip_ctx_data_t *ipctx)
{
  priv_elem_t *pe = ipctx->pe;
  shpf_t *sftab = ipctx->sf;
  int k;
  double f00 = 0, f01 = 0, f02 = 0;
  double f10 = 0, f11 = 0, f12 = 0;
  double f20 = 0, f21 = 0, f22 = 0;
  double x, y, z;
  double inv_prev_F[3][3];
  double det_prev_F;
  register double dphix, dphiy, dphiz;
  W_node_t *n;

  if (!VALID(F)) compute_def_grad(ipctx);
   
  FOR (k, WBMAT_NNODES) {
    n = pe->node[k];
    WN_get_prev_coords (n, &x, &y, &z);
    dphix  = sftab->n_x[k];
    dphiy  = sftab->n_y[k];
    dphiz  = sftab->n_z[k];
    f00 += dphix * x;
    f01 += dphiy * x;
    f02 += dphiz * x;
    f10 += dphix * y;
    f11 += dphiy * y;
    f12 += dphiz * y;
    f20 += dphix * z;
    f21 += dphiy * z;
    f22 += dphiz * z;
  }

  prev_F[0][0] = f00;
  prev_F[0][1] = f01;
  prev_F[0][2] = f02;
  
  prev_F[1][0] = f10;
  prev_F[1][1] = f11;
  prev_F[1][2] = f12;
  
  prev_F[2][0] = f20;
  prev_F[2][1] = f21;
  prev_F[2][2] = f22;

  det_prev_F = MAT3DUTIL_DET_3X3(prev_F);
  if (det_prev_F <= 0) {
    char buf[1024];
    sprintf (buf, "Failed to invert def grad; el qual=%g", tet_gamma_quality(pe));
    REPORT_FATAL_ERROR(NULL, buf);
  }

  MAT3DUTIL_INVERT_3X3(det_prev_F, prev_F, inv_prev_F);

  MAT3DUTIL_MULT_AB_3X3(rel_F, F, inv_prev_F);

  VALID(rel_F) = TRUE;
  VALID(prev_F) = TRUE;
}

static void
compute_midstep_def_grad (ip_ctx_data_t *ipctx)
{
  if (!VALID (F)) compute_def_grad (ipctx);
  if (!VALID (prev_F)) compute_rel_def_grad (ipctx);
   
  MAT3DUTIL_COMB_AB_3X3 (midstep_F, 0.5, F, 0.5, prev_F);

  VALID(midstep_F) = TRUE;
}

static void
compute_def_grad_dot(ip_ctx_data_t *ipctx)
{
  priv_elem_t *pe = ipctx->pe;
  shpf_t *sftab = ipctx->sf;
  int k;
  double f00 = 0, f01 = 0, f02 = 0;
  double f10 = 0, f11 = 0, f12 = 0;
  double f20 = 0, f21 = 0, f22 = 0;
  W_node_t *n;
  double x, y, z;
  register double dphix, dphiy, dphiz;

  /* Deformation gradient */
  FOR(k, WBMAT_NNODES) {
    n = pe->node[k];
    x = n->velocity_t.x;
    y = n->velocity_t.y;
    z = n->velocity_t.z;
    dphix  = sftab->n_x[k];
    dphiy  = sftab->n_y[k];
    dphiz  = sftab->n_z[k];
    f00 += dphix * x;
    f01 += dphiy * x;
    f02 += dphiz * x;
    f10 += dphix * y;
    f11 += dphiy * y;
    f12 += dphiz * y;
    f20 += dphix * z;
    f21 += dphiy * z;
    f22 += dphiz * z;
  }

  F_dot[0][0] = f00;
  F_dot[0][1] = f01;
  F_dot[0][2] = f02;
  
  F_dot[1][0] = f10;
  F_dot[1][1] = f11;
  F_dot[1][2] = f12;
  
  F_dot[2][0] = f20;
  F_dot[2][1] = f21;
  F_dot[2][2] = f22;

  VALID(F_dot) = TRUE;
}

static void
compute_inv_def_grad(ip_ctx_data_t *ipctx)
{
  if (!VALID(F)) compute_def_grad(ipctx);

  if (!W_mat3dutil_invert_3x3(F, inv_F)) {
    sprintf(W_buf(), "Failed to invert def grad");
    REPORT_FATAL_ERROR(NULL, W_buf());
  }

  VALID(inv_F) = TRUE;
}

static void
compute_vel_grad(ip_ctx_data_t *ipctx)
{
  if (!VALID(F)) compute_def_grad(ipctx);
  if (!VALID(F_dot)) compute_def_grad_dot(ipctx);
  if (!VALID(inv_F)) compute_inv_def_grad(ipctx);

  MAT3DUTIL_MULT_AB_3X3(L, F_dot, inv_F);
  
  VALID(L) = TRUE;
}

static void
compute_rate_of_def(ip_ctx_data_t *ipctx)
{
  int i, j;
  
  if (!VALID(L)) compute_vel_grad(ipctx);

  FOR(i, 3)
    FOR(j, 3) 
      D[i][j] = 0.5 * (L[i][j] + L[j][i]);

  VALID(D) = TRUE;
}

static void
compute_cauchy_green(ip_ctx_data_t *ipctx)
{
  if (!VALID(F)) compute_def_grad(ipctx);
  
  /* Cauchy-Green deformation tensor */
  MAT3DUTIL_MULT_ATB_3X3(C, F, F);

  VALID(C) = TRUE;
}

static void
compute_green_lagrange(ip_ctx_data_t *ipctx)
{
  if (!VALID(C)) compute_cauchy_green(ipctx);

  /* Green-Lagrange strain tensor */
  E[0][0] = 0.5 * (C[0][0] - 1);
  E[0][1] = 0.5 * C[0][1];
  E[0][2] = 0.5 * C[0][2];
  E[1][0] = 0.5 * C[1][0];
  E[1][1] = 0.5 * (C[1][1] - 1);
  E[1][2] = 0.5 * C[1][2];
  E[2][0] = 0.5 * C[2][0];
  E[2][1] = 0.5 * C[2][1];
  E[2][2] = 0.5 * (C[2][2] - 1);
  
  VALID(E) = TRUE;
}

static void
compute_midstep_rate_of_def (ip_ctx_data_t *ipctx, double time_step)
{
  int i, j;
  double m[3][3] = { {1, 0, 0}, {0, 1, 0}, {0, 0, 1} };
  double M[3][3] = { {-1, 0, 0}, {0, -1, 0}, {0, 0, -1} };
  double s[3][3], inv_s[3][3], q[3][3], r[3][3];
  double inv_dt = 1 / time_step;
  
  if (!VALID(rel_F)) compute_rel_def_grad(ipctx);

  MAT3DUTIL_ADD_AB_3X3(s, m, rel_F);
  MAT3DUTIL_SCALE_3X3(s, 0.5);
  if (!W_mat3dutil_invert_3x3(s, inv_s)) {
    sprintf(W_buf(), "Failed to invert def grad");
    REPORT_FATAL_ERROR(NULL, W_buf());
  }
  MAT3DUTIL_ADD_AB_3X3(q, rel_F, M);
  MAT3DUTIL_MULT_AB_3X3(r, inv_s, q);

  FOR(i, 3)
    FOR(j, 3) {
      midstep_D[i][j] = 0.5 * (r[i][j] + r[j][i]);
      midstep_W[i][j] = 0.5 * (r[i][j] - r[j][i]);
    }
  MAT3DUTIL_SCALE_3X3(midstep_D, inv_dt);
  MAT3DUTIL_SCALE_3X3(midstep_W, inv_dt);
  
  VALID(midstep_D) = TRUE;
  VALID(midstep_W) = TRUE;
}

static void
compute_polar_decomp(ip_ctx_data_t *ipctx)
{
  if (!VALID(F)) compute_def_grad(ipctx);
  
  if (!W_polar_decomposition (F, R, U)) {
    REPORT_FATAL_ERROR (NULL, "Failed in W_polar_decomposition()");
  }
  VALID(R) = TRUE;
  VALID(U) = TRUE;
}

static void
compute_rel_polar_decomp(ip_ctx_data_t *ipctx)
{
  if (!VALID(rel_F)) compute_rel_def_grad(ipctx);
  
  if (!W_polar_decomposition (rel_F, rel_R, rel_U)) {
    REPORT_FATAL_ERROR (NULL, "Failed in W_polar_decomposition()");
  }
  VALID(rel_R) = TRUE;
  VALID(rel_U) = TRUE;
}

static void
compute_ln_rel_U(ip_ctx_data_t *ipctx)
{
  double lneval[3];

  if (!VALID(rel_U)) compute_rel_polar_decomp(ipctx);
  
  W_ln_U (rel_U, ln_rel_U);
  
  VALID(ln_rel_U) = TRUE;
}

static W_bool_t 
get_kinem_var(W_mat_3d_ip_ctx_t *ctx, 
              W_mat_3d_kinem_var_type_enum  type,
              double kinvar[3][3])
{
  ip_ctx_data_t *ipctx = (ip_ctx_data_t *)ctx->ctx_data;

  switch (type) {
  case MAT_3D_DEF_GRAD:
    if (!VALID(F)) compute_def_grad(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, F);
    break;
  case MAT_3D_DEF_GRAD_DOT:
    if (!VALID(F)) compute_def_grad_dot(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, F_dot);
    break;
  case MAT_3D_REL_DEF_GRAD:
    if (!VALID(rel_F)) compute_rel_def_grad(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, rel_F);
    break;
  case MAT_3D_PREV_DEF_GRAD:
    if (!VALID(prev_F)) compute_rel_def_grad(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, prev_F);
    break;
  case MAT_3D_MIDSTEP_DEF_GRAD:
    if (!VALID(midstep_F)) compute_midstep_def_grad(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, midstep_F);
    break;
  case MAT_3D_CAUCHY_GREEN:
    if (!VALID(C)) compute_cauchy_green(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, C);
    break;
  case MAT_3D_RATE_OF_DEF:
    if (!VALID(D)) compute_rate_of_def(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, D);
    break;
  case MAT_3D_MIDSTEP_RATE_OF_DEF:
    if (!VALID(midstep_D))
      compute_midstep_rate_of_def(ipctx, ctx->time_step);
    MAT3DUTIL_COPY_3X3(kinvar, midstep_D);
    break;
  case MAT_3D_MIDSTEP_SPIN:
    if (!VALID(midstep_W))
      compute_midstep_rate_of_def(ipctx, ctx->time_step);
    MAT3DUTIL_COPY_3X3(kinvar, midstep_W);
    break;
  case MAT_3D_GREEN_LAGRANGE:
    if (!VALID(E)) compute_green_lagrange(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, E);
    break;
  case MAT_3D_POLAR_R:
    if (!VALID(R)) compute_polar_decomp(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, R);
    break;
  case MAT_3D_POLAR_U:
    if (!VALID(U)) compute_polar_decomp(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, U);
    break;
  case MAT_3D_POLAR_REL_R:
    if (!VALID(rel_R)) compute_rel_polar_decomp(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, rel_R);
    break;
  case MAT_3D_POLAR_REL_U:
    if (!VALID(rel_U)) compute_rel_polar_decomp(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, rel_U);
    break;
  case MAT_3D_LN_POLAR_REL_U:
    if (!VALID(ln_rel_U)) compute_ln_rel_U(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, ln_rel_U);
    break;
  default:
    {
      static W_bool_t complained = FALSE;
      if (!complained) {
        sprintf (W_buf(), "Unknown kinematic tensor type: %d",     (int)type);
        REPORT_WARNING (ipctx->domain->errh, W_buf());
        complained = TRUE;
      }
    }
    return FALSE;
  }

  return TRUE;
}

static void
fill_ip_pkg (double r, double s, double t, double u,
             int entry,
             G3D_loc_t *iploc,
             W_mat_3d_state_p matstate,
             W_mat_3d_ip_ctx_t *ip_ctx,
             W_mat_3d_t *mat,
             W_mat_3d_stress_type_enum stress_type
             )
{
  double stress[3][3];
  double equiv_pl_strain, von_mises;
  W_mat_3d_stress_type_enum out_type;

  /* Get 1st Piola-Kirchhoff stress at point */
  if (!WM3D_update_stress (mat, matstate, W_MAT_STATE_UPDATE_TEMP,
                        ip_ctx, stress_type, stress,
                        &out_type)) {
    REPORT_ERROR (NULL, "Failed to compute stress");
    return;
  }
  if (out_type != stress_type)
    REPORT_WARNING (NULL, "output stress type mismatch");

  if (!WM3D_get_mat_state_var (mat, matstate,
                               ip_ctx, MAT_3D_EQUIV_PLASTIC_DEF,
                               &equiv_pl_strain))
    REPORT_FATAL_ERROR(NULL, "Failed to retrieve equiv plastic def");

  {
    double means = (stress[0][0] + stress[1][1] + stress[2][2]) / 3;
    double s00 = stress[0][0] - means;
    double s11 = stress[1][1] - means;
    double s22 = stress[2][2] - means;
    
    von_mises
      = sqrt(1.5
             * (  s00*s00
                + stress[0][1]*stress[0][1]
                + stress[0][2]*stress[0][2]
                + stress[1][0]*stress[1][0]
                + s11*s11
                + stress[1][2]*stress[1][2]
                + stress[2][0]*stress[2][0]
                + stress[2][1]*stress[2][1]
                + s22*s22
                )
             );
  }
  
  /* Store into the package */
  /* fprintf(stderr, "%g %g %g %d\n", r, s, t, entry); */
  ipol_pkgs[entry].iploc           = *iploc;
    /* fprintf(stderr, "%g %g %g\n", ipol_pkgs[entry].iploc.x, ipol_pkgs[entry].iploc.y, ipol_pkgs[entry].iploc.z); */
  MAT3DUTIL_COPY_3X3(ipol_pkgs[entry].stress, stress);
  ipol_pkgs[entry].equiv_pl_strain = equiv_pl_strain;
  ipol_pkgs[entry].von_mises       = von_mises;
}


static void
eval_at_extrapolated_pt (G3D_loc_t *pt,
                         double stress[3][3], double *eps,
                         double *vm)
{
  double equiv_pl_strain, von_mises;
  G3D_tet_t tet;
  G3D_loc4d_t param_loc;

  tet.loc[0] = ipol_pkgs[0].iploc;
  tet.loc[1] = ipol_pkgs[1].iploc;
  tet.loc[2] = ipol_pkgs[2].iploc;
  tet.loc[3] = ipol_pkgs[3].iploc;

  if (!G3D_tet_inv_map(&tet, pt, &param_loc)) {
    REPORT_WARNING (NULL, "inv map failed");
    /*
    fprintf(stderr, "%g %g %g\n", tet.loc[0].x, tet.loc[0].y, tet.loc[0].z);
    fprintf(stderr, "%g %g %g\n", tet.loc[1].x, tet.loc[1].y, tet.loc[1].z);
    fprintf(stderr, "%g %g %g\n", tet.loc[2].x, tet.loc[2].y, tet.loc[2].z);
    fprintf(stderr, "%g %g %g\n", tet.loc[3].x, tet.loc[3].y, tet.loc[3].z);
    abort();
    */
  }

  /* Interpolate */
#define COMP_T1(T, I, J)				\
  { T[I][J] = (						\
	       param_loc.x * ipol_pkgs[0].T[I][J] +	\
	       param_loc.y * ipol_pkgs[1].T[I][J] +	\
	       param_loc.z * ipol_pkgs[2].T[I][J] +	\
	       param_loc.w * ipol_pkgs[3].T[I][J]       \
	       );					\
  }
#define COMP_T(T)						\
  {								\
    COMP_T1(T, 0, 0); COMP_T1(T, 0, 1); COMP_T1(T, 0, 2);	\
    COMP_T1(T, 1, 0); COMP_T1(T, 1, 1); COMP_T1(T, 1, 2);	\
    COMP_T1(T, 2, 0); COMP_T1(T, 2, 1); COMP_T1(T, 2, 2);	\
  }
#define COMP_S(S)				\
  { S = (					\
	 param_loc.x * ipol_pkgs[0].S +	\
	 param_loc.y * ipol_pkgs[1].S +	\
	 param_loc.z * ipol_pkgs[2].S +	\
	 param_loc.w * ipol_pkgs[3].S 	\
	 );					\
  }
#define COMP_V1(V, DIR)                         \
  { V.DIR = (                                   \
           param_loc.x * ipol_pkgs[0].V.DIR + \
           param_loc.y * ipol_pkgs[1].V.DIR + \
           param_loc.z * ipol_pkgs[2].V.DIR + \
           param_loc.w * ipol_pkgs[3].V.DIR   \
           );                                   \
  }
#define COMP_V(V) { COMP_V1(V, x); COMP_V1(V, y); COMP_V1(V, z); }

  COMP_T(stress);
  COMP_S(equiv_pl_strain); *eps = equiv_pl_strain;
  COMP_S(von_mises); *vm = von_mises;

  return;
}

static void
write_postproc_file (W_elem_t *el, ckit_iobuf_t IOB,
                     int type, char *mode, void *optdata)
{
  if (type == HEXVUE_POSTPROC_FILE) {
    write_hexvue_file (el, IOB, type, mode, optdata);
  } else if (type == HEXVUE_XTRA_POSTPROC_FILE) {
    write_hexvue_XTRA_file (el, IOB, type, mode, optdata);
  }
}



static W_elem_methods_t methods = {
  elem_elcode,
  type_name,
  assemble_restoring_forces,
  assemble_mass,
  total_mass,
  NULL,
  suggested_time_step,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  new_history_client_data,
  update_history,
  NULL,
  write_postproc_file,
#ifdef X_XFACE
  modify_elem_grep,
#endif
  NULL,
  NULL,
  NULL,
  num_of_faces,
  get_face_vertex_nodes,
  get_contact_faces
};
  



static W_bool_t
acceptable_tet (priv_elem_t *pe, double tol)
{
  return (tet_gamma_quality (pe) > tol);
}

/*
  Element reader.
  Reads element data for the pcont3 element type
  from stream as long as there is no
  END_OF_BLOCK tag in the input line. Creates elements.
 */
W_bool_t                        /* TRUE if successful */
tet10_reader (W_domain_t *d /* domain */)
{
  const char *input_line;
  long elem_num; 
  priv_elem_t *pe;
#ifdef PVM_NEED_GHOST_ELEMENTS
  int is_ghost = FALSE;
#endif
  W_index_t nodeid, matid;
  int i, k;
  W_material_t *mat;
  W_mat_3d_t *m;
  double quality_tol = 0;

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line)) {
      return TRUE;
    } else {
      TOKENS_PARSE(input_line);
      if (TOKENS_HAVE_KEYWORD("QUALITY_TOL")) {
        TOKENS_SET_BY_KEYWORD("QUALITY_TOL", 0, {quality_tol = TOKEN_DBL_NEXT(1);}, { quality_tol=0; });
      } else {
        pe = w_new(priv_elem_t);
        CHECK_MALLOC(pe);
        elem_num = TOKEN_INT(1);
        FOR(i, WBMAT_NNODES) {
          nodeid = TOKEN_INT(2+i);
          pe->node[i] = WD_get_node_by_idx(d, nodeid);
          if (pe->node[i] == NULL) {
            sprintf(W_buf(), "No such node: %ld", nodeid);
            REPORT_ERROR(d->errh, W_buf());
          }
        }
        matid = TOKEN_INT(12);
        mat = WD_get_material_by_idx(d, matid);
        if (mat == NULL) {
          sprintf(W_buf(), "No such material: %ld", matid);
          REPORT_ERROR(d->errh, W_buf());
        }
        pe->material     = mat;
        m = (W_mat_3d_t *)pe->material->data;
        pe->matst        = w_new_array (4, W_mat_3d_state_p);
        CHECK_MALLOC (pe->matst);
        FOR(k, 4) {
          W_loc_t loc;
          G3D_tet10_N (sf.n, ALL_GP_LOCS(k));
          LOC(pe, loc);
          pe->matst[k] = WM3D_get_mat_state(m, &loc);
        }
#ifdef PVM_NEED_GHOST_ELEMENTS
        is_ghost = FALSE;
        if (TOKENS_HAVE_KEYWORD("GHOST"))
          is_ghost = TRUE;
#endif      
        if ((quality_tol == 0) || (quality_tol > 0 && acceptable_tet (pe, quality_tol))) 
          WD_add_element(d, elem_num,
                         WE_new_element(d, &methods, pe
#                                   ifdef PVM_NEED_GHOST_ELEMENTS
                                        , is_ghost
#                                   endif
                                        ));
        else {
          char buf[1024];
          sprintf(buf, "Ignoring el. %d: qual=%g < quality_tol=%g", elem_num, tet_gamma_quality(pe), quality_tol);
          REPORT_WARNING(d->errh, buf);
        }
      }
    } /* closing else */
    }
    return FALSE;
  }
  
  
