#include "Wconfig.h"
#include "Wtypes.h"
#include "Wtet4.h"
#include "Wdomain.h"
#include "Wnode.h"
#include "Wmats.h"
#include "Wgaussp.h"
#include "Wxf.h"
#include "Wutil.h"
#include "Wcdi.h"
#include "g3d.h"
#include "Wmat3d.h"
#include "Wnewmarki.h"
#include "Wcontface.h"

/* RCSHEADER "$Id: tet4.c,v 1.8 2004/09/06 01:17:58 cvs Exp $" */

void
G3D_tet_N_deriv (double n_x[4],
                 double n_y[4],
                 double n_z[4],
                 double x1, double y1, double z1,
                 double x2, double y2, double z2,
                 double x3, double y3, double z3,
                 double x4, double y4, double z4
                );
#define GET_FROM_NODE(node, func, tovect)         \
   func (node, &tovect.x, &tovect.y, &tovect.z)
#define GP_WGT() 1.0
#define ALL_REF_COORDS(H)                                                     \
 H->node[0]->ref_coords.x, H->node[0]->ref_coords.y, H->node[0]->ref_coords.z,\
 H->node[1]->ref_coords.x, H->node[1]->ref_coords.y, H->node[1]->ref_coords.z,\
 H->node[2]->ref_coords.x, H->node[2]->ref_coords.y, H->node[2]->ref_coords.z,\
 H->node[3]->ref_coords.x, H->node[3]->ref_coords.y, H->node[3]->ref_coords.z

#define CENTROID(loc, t)                             \
  {                                                     \
    VECMAC_2_VEC (loc,  =, t->node[0]->ref_coords);  \
    VECMAC_2_VEC (loc, +=, t->node[1]->ref_coords);  \
    VECMAC_2_VEC (loc, +=, t->node[2]->ref_coords);  \
    VECMAC_2_VEC (loc, +=, t->node[3]->ref_coords);  \
    VECMAC_SCALAR (loc, *=, 0.25);                      \
  }

#undef WBMAT_NNODES
#define WBMAT_NNODES 4

#include "bmat.c"
#include "elstiff.c"

typedef struct priv_elem_t
{
  W_node_t          *node[WBMAT_NNODES];
  W_material_t      *material;
  W_mat_3d_state_p   matst;
}              priv_elem_t;


static int
elem_elcode (W_elem_t *el);
static char *
type_name(W_elem_t *el);
static void
assemble_restoring_forces(W_elem_t *el, W_mat_state_update_form_enum update_form);
static void
assemble_mass(W_elem_t *el);
static W_float_t
suggested_time_step(W_elem_t *el);

#ifdef X_XFACE
static void
modify_elem_grep(W_elem_t *el, double displ_scale, W_bool_t immed_draw)
{
  priv_elem_t *tet = (priv_elem_t *)el->data;
  int i;
  W_vect_t refc, u;
  WCRec pnts[4];

  for (i = 0; i < 4; i++) {
    GET_FROM_NODE (tet->node[i], WN_get_ref_coords, refc);
    GET_FROM_NODE (tet->node[i], WN_get_displ, u);
    pnts[i].x = refc.x + displ_scale * u.x;
    pnts[i].y = refc.y + displ_scale * u.y;
    pnts[i].z = refc.z + displ_scale * u.z;
  }

  if (el->grep == NULL) {
    el->grep = CreateTetra (pnts);
    EASValsSetColor (WM_get_color (tet->material));
    EASValsSetFillStyle (FILL_SOLID);
    EASValsSetLayer (WX_ELEM_LAYER);
    EGWithMaskChangeAttributes (COLOR_MASK | LAYER_MASK | FILL_MASK, el->grep);
    EMNoDrawAddGraphicsToModel (ESIModel (), el->grep);
  } else {
    EGModifyGraphicsGeometry (el->grep, pnts);
  }
}
#endif

static W_float_t
total_mass(W_elem_t *el);
static void
write_postproc_file (W_elem_t *el, ckit_iobuf_t IOB, int type,
                     char *mode, void *optdata);




static int
num_of_faces (W_elem_t *el)
{
  return 4;
}

static int
get_face_vertex_nodes (W_elem_t *el, int nface, int ndim, W_node_t *n[])
{
  priv_elem_t *t = (priv_elem_t *)el->data;
  int i;

  if (ndim < 3)
    return 3;

  FOR (i, 3) {
    n[i] = t->node[g3dtet_vertex_on_face[nface][i]];
  }
  return 3;
}

static int
get_contact_faces (W_elem_t *el, int nface,
                   WCF_contact_face_t *cf[], int maxcf)
{
  priv_elem_t *t = (priv_elem_t *)el->data;
  int i;

  if (maxcf < 1) {
    FOR (i, maxcf) {
      cf[i] = NULL;
    }
    return 1;
  }

  cf[0] =
    WCF_new_contact_face (
      t->node[g3dtet_vertex_on_face[nface][0]],
      t->node[g3dtet_vertex_on_face[nface][1]],
      t->node[g3dtet_vertex_on_face[nface][2]],
      NULL);

  return 1;
}

static void
at_new_ip(void);
static W_bool_t
get_kinem_var(W_mat_3d_ip_ctx_t *ctx,
              W_mat_3d_kinem_var_type_enum  type,
              double kinvar[3][3]);


typedef struct shape_func_t
{
  double n_x[4];
  double n_y[4];
  double n_z[4];
  double volume;
}
shape_func_t;

static shape_func_t shpf;

typedef struct ip_ctx_data_t
{
  priv_elem_t          *t;
  shape_func_t         *sf;    /* shape function table */
}
ip_ctx_data_t;

static double 
curr_volume (W_elem_t *el)
{
  priv_elem_t *pe = (priv_elem_t *)el->data;
  W_vect_t v, c[4];
  int j, k;
  
  FOR(j, 4) {
    GET_FROM_NODE (pe->node[j], WN_get_coords, c[j]);
  }
  
  return G3D_tet_volume (c[0].x, c[0].y, c[0].z,
                         c[1].x, c[1].y, c[1].z,
                         c[2].x, c[2].y, c[2].z,
                         c[3].x, c[3].y, c[3].z
                         );
}


static void
assemble_eff_stiffness (W_elem_t *el,
                        W_newmark_assemble_eff_stiff_proc assemble_eff_stiff_callback,
                        W_float_t stiff_mult,
                        W_float_t mass_mult,
                        W_mass_type_enum mass_type
                       )
{
  priv_elem_t *t = (priv_elem_t *)el->data;
  W_mat_3d_t *m = (W_mat_3d_t *)t->material->data;
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  double S[3][3];
  W_mat_3d_stress_type_enum out_type;
  W_vect_t u[WBMAT_NNODES];
  double jdetw;
  double dphix, dphiy, dphiz;
  int i, j, k, n;
  double rho;
  double C[6][6];
  double K[WBMAT_NNODES*3][WBMAT_NNODES*3];
  W_index_t idcs[WBMAT_NNODES*3];
  W_bool_t constrained[WBMAT_NNODES*3];
  double b_l[6][WBMAT_NNODES*3];
  double b_nl[9][WBMAT_NNODES*3];
  double nmat[4] = {0.5,0.5,0.5,0.5};

  ip_ctx_data.t        = t;
  ip_ctx.time          = WD_time (el->domain);
  ip_ctx.time_step     = WD_time_step (el->domain);
  ip_ctx.ctx_data      = &ip_ctx_data;
  ip_ctx.get_kinem_var = get_kinem_var;
  ip_ctx_data.sf       = &shpf;

  FOR (i, WBMAT_NNODES) {
    GET_FROM_NODE (t->node[i], WN_get_displ, u[i]);
  }

  rho = WM3D_get_volume_mass (m);

  FOR (i, WBMAT_NNODES*3) {
    FOR (j, WBMAT_NNODES*3) {
      K[i][j] = 0;
    }
  }

  G3D_tet_N_deriv (shpf.n_x, shpf.n_y, shpf.n_z, ALL_REF_COORDS(t));
  shpf.volume = G3D_tet_volume (ALL_REF_COORDS(t));
  at_new_ip();
  jdetw = shpf.volume;
  b_l_mat  (shpf.n_x, shpf.n_y, shpf.n_z, WBMAT_NNODES, u, b_l);
  if (!WM3D_get_mat_stiff (m, t->matst, &ip_ctx,
                           MAT_3D_STIFF_LAGRANGIAN, C))
    REPORT_FATAL_ERROR (el->domain->errh, "No material tangent");
  add_bltcbl_upper_triangle (K, b_l, C, stiff_mult * jdetw);
  if (!WM3D_update_stress (m, t->matst, W_MAT_STATE_UPDATE_TEMP,
                           &ip_ctx, MAT_3D_2ND_PK, S, &out_type))
    REPORT_ERROR(el->domain->errh, "Failed to compute stress");
  b_nl_mat (shpf.n_x, shpf.n_y, shpf.n_z, WBMAT_NNODES, b_nl);
  add_bnltsbnl_upper_triangle (K, b_nl, S, stiff_mult * jdetw);
  add_to_diag (K, mass_mult * rho * jdetw / WBMAT_NNODES); /* diagonal mass only is supported */

  mirror (K);

  WN_eqn_numbers_and_constraints (t->node[0], &idcs[0], &constrained[0]);
  WN_eqn_numbers_and_constraints (t->node[1], &idcs[3], &constrained[3]);
  WN_eqn_numbers_and_constraints (t->node[2], &idcs[6], &constrained[6]);
  WN_eqn_numbers_and_constraints (t->node[3], &idcs[9], &constrained[9]);

  (*assemble_eff_stiff_callback)  (el->domain->newmark_integrator, FLES_ADD,
                                   WBMAT_NNODES*3, idcs, (double *) K, constrained);
}

static void
assemble_inertial_forces_diag (W_elem_t *el)
{
  priv_elem_t *t = (priv_elem_t *)el->data;
  W_vect_t f, a;
  int i;
  double m = total_mass (el) / WBMAT_NNODES;

  FOR (i, WBMAT_NNODES) {
    GET_FROM_NODE (t->node[i], WN_get_acceleration, a);
    VECMAC_2_VEC (f, =, -m * a);
    WN_assemble_force (t->node[i], &f);
  }
}

static void
assemble_residual_forces (W_elem_t *el,
                          W_mat_state_update_form_enum update_form,
                          W_mass_type_enum mass_type)
{
  assemble_restoring_forces (el, update_form);
  assemble_inertial_forces_diag (el);
}

static int
elem_elcode (W_elem_t *el)
{
  return TET4_ELCODE;
}


static char *
type_name (W_elem_t *el)
{
  static char buf[] = "tet4";
  return buf;
}

/* Returns nodal forces by which the  element acts on its nodal points. */

static void
assemble_restoring_forces (W_elem_t *el, W_mat_state_update_form_enum update_form)
{
  priv_elem_t *t = (priv_elem_t *)el->data;
  W_mat_3d_t *m = (W_mat_3d_t *)t->material->data;
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  double P[3][3];
  W_mat_3d_stress_type_enum out_type;
  W_vect_t f;
  double jdetw, ded;
  double dphix, dphiy, dphiz;
  int k, n;
#define COMP_F()                                            \
  {                                                         \
    f.x = -(P[0][0]*dphix + P[0][1]*dphiy + P[0][2]*dphiz); \
    f.y = -(P[1][0]*dphix + P[1][1]*dphiy + P[1][2]*dphiz); \
    f.z = -(P[2][0]*dphix + P[2][1]*dphiy + P[2][2]*dphiz); \
  }

  G3D_tet_N_deriv (shpf.n_x, shpf.n_y, shpf.n_z, ALL_REF_COORDS(t));
  shpf.volume = G3D_tet_volume (ALL_REF_COORDS(t));
  at_new_ip();

  ip_ctx_data.t        = t;
  ip_ctx.time          = WD_time (el->domain);
  ip_ctx.time_step     = WD_time_step (el->domain);
  ip_ctx.ctx_data      = &ip_ctx_data;
  ip_ctx.get_kinem_var = get_kinem_var;
  ip_ctx_data.sf       = &shpf;

  if (!WM3D_update_stress(m, t->matst, update_form,
                          &ip_ctx, MAT_3D_1ST_PK, P, &out_type))
    REPORT_ERROR(el->domain->errh, "Failed to compute stress");
  if (!WM3D_get_mat_state_var (m, t->matst,
                               &ip_ctx, MAT_3D_DEFORM_ENERGY_DENSITY, &ded))
    REPORT_ERROR(el->domain->errh, "Failed to retrieve work of deformation");
  jdetw = shpf.volume;
  el->deformation_energy = ded * jdetw;
  FOR(n, 4) {
    dphix = jdetw * shpf.n_x[n];
    dphiy = jdetw * shpf.n_y[n];
    dphiz = jdetw * shpf.n_z[n];
    COMP_F();
    WN_assemble_force(t->node[n], &f);
  }
}

static void
assemble_mass (W_elem_t *el)
{
  priv_elem_t *t = (priv_elem_t *)el->data;
  int i;
  double tm = total_mass (el);

  for (i = 0; i < 4; i++)
    WN_assemble_mass (t->node[i], tm / 4.0);
}


static W_float_t 
suggested_time_step (W_elem_t *el)
{
  priv_elem_t *pe = (priv_elem_t *)el->data;
  W_vect_t v, c[4];
  int j, k;
  double E, nu, rho, c_d;
  W_mat_3d_t *m = (W_mat_3d_t *)pe->material->data;
  double d, mind = FLT_MAX;
  
  if (!WM3D_get_mat_const(m, MAT_3D_YOUNG_MODULUS, &E))
    REPORT_FATAL_ERROR(el->domain->errh, "No Young's modulus");
  if (!WM3D_get_mat_const(m, MAT_3D_POISSON, &nu))
    REPORT_FATAL_ERROR(el->domain->errh, "No Poisson constant");
  rho = WM3D_get_volume_mass (m);
  c_d = W_3D_dilatational_wave_speed (E, nu, rho);
 
  FOR(j, 4) {
    GET_FROM_NODE (pe->node[j], WN_get_coords, c[j]);
  }
  
  FOR(j, 4) {
    for (k = j+1; k < 4; k++) {
      VECMAC_3_VEC(v, =, c[k], -, c[j]);
      d = VECMAC_NORM(v);
      mind = min(d, mind);
    }
  }
  
  if (G3D_tet_volume (c[0].x, c[0].y, c[0].z,
                      c[1].x, c[1].y, c[1].z,
                      c[2].x, c[2].y, c[2].z,
                      c[3].x, c[3].y, c[3].z
                      ) <= 0) {
    sprintf (W_buf (), "Zero volume tetrahedron {%d,%d,%d,%d}",
             pe->node[0]->id, pe->node[1]->id, pe->node[2]->id, pe->node[3]->id);
    REPORT_WARNING (el->domain->errh, W_buf ());
  }

  return (mind / c_d);
}


static W_float_t
total_mass(W_elem_t *el)
{
  priv_elem_t *t = (priv_elem_t *)el->data;
  W_mat_3d_t *m = (W_mat_3d_t *)t->material->data;

  double volume = G3D_tet_volume (ALL_REF_COORDS (t));
  double volume_mass = WM3D_get_volume_mass(m);

  return (volume_mass * volume);
}





static void
write_hexvue_XTRA_file (W_elem_t *el, ckit_iobuf_t IOB,
                        int type, char *mode, void *optdata)
{
  priv_elem_t *pe = (priv_elem_t *)el->data;
  int  i, j, k, ival=0;
  W_vect_t refc[4], u[4];
  W_mat_3d_t *m = (W_mat_3d_t *)pe->material->data;
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  int reqlen;
  W_mat_3d_stress_type_enum outtype;
  W_output_stress_type_enum output_stress_type
  = WOPT_OPT (el->domain->options, output_stress_type);
  W_postfile_output_specs_t *os = &el->domain->options->postfile_output_specs_list;
  double *hvals[8] = { NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL };
  G3D_loc_t *hpts[8];
  int idc[8]={0,1,2,2,3,3,3,3};

  outtype = (output_stress_type == W_OUTPUT_STRESS_CAUCHY
             ? MAT_3D_CAUCHY
             : (output_stress_type == W_OUTPUT_STRESS_2ND_PK
                ? MAT_3D_2ND_PK
                : MAT_3D_CAUCHY)
            );

  {
    reqlen = os->nspecs;

      FOR (i, 8) {
        hvals[i] = NEWP_REALLOC_OBJ_ARRAY (hvals[i], reqlen, double);
      }

    ip_ctx_data.t        = pe;
    ip_ctx.time          = WD_time (el->domain);
    ip_ctx.time_step     = WD_time_step (el->domain);
    ip_ctx.ctx_data      = &ip_ctx_data;
    ip_ctx.get_kinem_var = get_kinem_var;
    ip_ctx_data.sf       = &shpf;

    FOR (i, 4) {
      GET_FROM_NODE (pe->node[i], WN_get_ref_coords, refc[i]);
      GET_FROM_NODE (pe->node[i], WN_get_displ, u[i]);
    }

    {
      FOR (k, 8) {
        hpts[k]=&refc[idc[k]];
        FOR (ival, reqlen) {
          if        (tokens_token_and_keyword_equiv (os->specs[ival], "displ_x")) {
            hvals[k][ival]=u[idc[k]].x;
          } else if (tokens_token_and_keyword_equiv (os->specs[ival], "displ_y")) {
            hvals[k][ival]=u[idc[k]].y;
          } else if (tokens_token_and_keyword_equiv (os->specs[ival], "displ_z")) {
            hvals[k][ival]=u[idc[k]].z;
          } else { /* quadrature-point data */
            G3D_tet_N_deriv (shpf.n_x, shpf.n_y, shpf.n_z, ALL_REF_COORDS(pe));
            at_new_ip();
            {
              double stress[3][3];
              double equiv_pl_strain=0, von_mises=0, press=0;
              W_mat_3d_stress_type_enum actual_outtype;

              if (!WM3D_update_stress (m, pe->matst, W_MAT_STATE_UPDATE_TEMP,
                                       &ip_ctx, outtype, stress,
                                       &actual_outtype)) {
                REPORT_ERROR (NULL, "Failed to compute stress");
                return;
              }
              if (actual_outtype != outtype)
                REPORT_WARNING (NULL, "output stress type mismatch");

              if (!WM3D_get_mat_state_var (m, pe->matst,
                                           &ip_ctx, MAT_3D_EQUIV_PLASTIC_DEF,
                                           &equiv_pl_strain))
                REPORT_FATAL_ERROR(NULL, "Failed to retrieve equiv plastic def");

              press = (stress[0][0] + stress[1][1] + stress[2][2]) / 3;
              {
                double s00 = stress[0][0] - press;
                double s11 = stress[1][1] - press;
                double s22 = stress[2][2] - press;

                von_mises
                = sqrt(1.5
                       * (  s00*s00
                            + stress[0][1]*stress[0][1]
                            + stress[0][2]*stress[0][2]
                            + stress[1][0]*stress[1][0]
                            + s11*s11
                            + stress[1][2]*stress[1][2]
                            + stress[2][0]*stress[2][0]
                            + stress[2][1]*stress[2][1]
                            + s22*s22
                         )
                      );
              }

              if (tokens_token_and_keyword_equiv (os->specs[ival], "s_xx")) {
                hvals[k][ival] = stress[0][0];
              } else if (tokens_token_and_keyword_equiv (os->specs[ival], "s_xy")) {
                hvals[k][ival] = stress[0][1];
              } else if (tokens_token_and_keyword_equiv (os->specs[ival], "s_xz")) {
                hvals[k][ival] = stress[0][2];
              } else if (tokens_token_and_keyword_equiv (os->specs[ival], "s_yy")) {
                hvals[k][ival] = stress[1][1];
              } else if (tokens_token_and_keyword_equiv (os->specs[ival], "s_yz")) {
                hvals[k][ival] = stress[1][2];
              } else if (tokens_token_and_keyword_equiv (os->specs[ival], "s_zz")) {
                hvals[k][ival] = stress[2][2];
              } else if (tokens_token_and_keyword_equiv (os->specs[ival], "vonm")) {
                hvals[k][ival] = von_mises;
              } else if (tokens_token_and_keyword_equiv (os->specs[ival], "press")) {
                hvals[k][ival] = press;
              } else if (tokens_token_and_keyword_equiv (os->specs[ival], "eqpls")) {
                hvals[k][ival] = equiv_pl_strain;
              } else { /* try to get the named variable from the material state */
                double val = 0;
                if (WM3D_get_mat_state_named_var (m, pe->matst,
                                                  &ip_ctx, os->specs[ival],
                                                  &val)) {
                  hvals[k][ival] = val;
                } else
                  hvals[k][ival] = 0;
              }
            }
          }
        }
      }
      W_write_hexvue_hex (IOB, hpts, reqlen, hvals, mode);
    }
  }
  FOR (i, 8) {
    w_free (hvals[i]);
  }
#undef EVAL_SF
}


static void
write_hexvue_file (W_elem_t *el, ckit_iobuf_t IOB,
                   int type, char *mode, void *optdata)
{
  priv_elem_t *t = (priv_elem_t *)el->data;
  W_vect_t refc[4], u[4];
  int i, ival, pt, reqlen;
  G3D_loc_t *pts[8];
  static int buflen        = 0;
  static double *values[8] = {NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL};
  int output_spec
  = WOPT_OPT (el->domain->options, postfile_output_spec_deprec);
  int whichpt[4] = {0,1,2,4};

  if (type == HEXVUE_POSTPROC_FILE) {
    double nx = 0, ny = 0, nxy = 0;
    double equiv_pl_def = 0, von_mises = 0;

    reqlen = 0;
    if (output_spec & W_POSTFILE_DISPL_MASK)
      reqlen += 3;
    if (output_spec & W_POSTFILE_STRESS_MASK)
      reqlen += 6;
    if (output_spec & W_POSTFILE_EQPLS_MASK)
      reqlen++;
    if (output_spec & W_POSTFILE_VON_MISES_MASK)
      reqlen++;
    if (output_spec & W_POSTFILE_MATERIAL_MASK)
      reqlen++;

    if (buflen < reqlen) {
      values[0] = NEWP_REALLOC_OBJ_ARRAY (values[0], reqlen, double);
      values[1] = NEWP_REALLOC_OBJ_ARRAY (values[1], reqlen, double);
      values[2] = NEWP_REALLOC_OBJ_ARRAY (values[2], reqlen, double);
      values[3] = values[2];
      values[4] = NEWP_REALLOC_OBJ_ARRAY (values[4], reqlen, double);
      values[5] = values[4];
      values[6] = values[4];
      values[7] = values[4];
      buflen = reqlen;
    }

    FOR (i, 4) {
      GET_FROM_NODE (t->node[i], WN_get_ref_coords, refc[i]);
      GET_FROM_NODE (t->node[i], WN_get_displ, u[i]);
    }

    FOR(i, 4) {
      pt = whichpt[i];
      pts[pt] = &refc[i];
      ival = 0;
      if (output_spec & W_POSTFILE_DISPL_MASK) {
        values[pt][ival] = u[i].x;
        ival++;
        values[pt][ival] = u[i].y;
        ival++;
        values[pt][ival] = u[i].z;
        ival++;
      }

      /* RAW evaluate stress etc */

      if (output_spec & W_POSTFILE_STRESS_MASK) {
        values[pt][ival] = nx;
        ival++;
        values[pt][ival] = ny;
        ival++;
        values[pt][ival] = nxy;
        ival++;
        values[pt][ival] = 0;
        ival++;
        values[pt][ival] = 0;
        ival++;
        values[pt][ival] = 0;
        ival++;
      }
      if (output_spec & W_POSTFILE_EQPLS_MASK) {
        values[pt][ival] = equiv_pl_def;
        ival++;
      }
      if (output_spec & W_POSTFILE_VON_MISES_MASK) {
        values[pt][ival] = von_mises;
        ival++;
      }
      if (output_spec & W_POSTFILE_MATERIAL_MASK) {
        values[pt][ival] = WM_get_id (t->material);
        ival++;
      }
      if (ival > reqlen)
        REPORT_FATAL_ERROR (el->domain->errh, "wrong");
    }
    pts[3] = pts[2];
    pts[5] = pts[4];
    pts[6] = pts[4];
    pts[7] = pts[4];
    W_write_hexvue_hex (IOB, pts, reqlen, values, mode);
  } /* type */
}


static void
write_postproc_file (W_elem_t *el, ckit_iobuf_t IOB,
                     int type, char *mode, void *optdata)
{
  if (type == HEXVUE_POSTPROC_FILE) {
    write_hexvue_file (el, IOB, type, mode, optdata);
  } else if (type == HEXVUE_XTRA_POSTPROC_FILE) {
    write_hexvue_XTRA_file (el, IOB, type, mode, optdata);
  }
}


#define VALID(which) which##_is_valid

#define DCL_TENSOR(T)                           \
   static double T[3][3];                       \
   static W_bool_t VALID(T) = FALSE

DCL_TENSOR(F);
DCL_TENSOR(prev_F);
DCL_TENSOR(inv_F);
DCL_TENSOR(rel_F);
DCL_TENSOR(midstep_F);
DCL_TENSOR(C);
DCL_TENSOR(E);
DCL_TENSOR(D);
DCL_TENSOR(midstep_D);
DCL_TENSOR(midstep_W);
DCL_TENSOR(F_dot);
DCL_TENSOR(L);
DCL_TENSOR(R);
DCL_TENSOR(U);
DCL_TENSOR(rel_R);
DCL_TENSOR(rel_U);
DCL_TENSOR(ln_rel_U);

static void
at_new_ip(void)
{
#define INVALIDATE(which) VALID(which) = FALSE

  INVALIDATE(F);
  INVALIDATE(inv_F);
  INVALIDATE(prev_F);
  INVALIDATE(rel_F);
  INVALIDATE(midstep_F);
  INVALIDATE(C);
  INVALIDATE(E);
  INVALIDATE(D);
  INVALIDATE(midstep_D);
  INVALIDATE(midstep_W);
  INVALIDATE(F_dot);
  INVALIDATE(L);
  INVALIDATE(R);
  INVALIDATE(U);
  INVALIDATE(rel_R);
  INVALIDATE(rel_U);
  INVALIDATE(ln_rel_U);
}


static void
compute_def_grad(ip_ctx_data_t *ipctx);
static void
compute_cauchy_green(ip_ctx_data_t *ipctx);
static void
compute_green_lagrange(ip_ctx_data_t *ipctx);
static void
compute_rel_def_grad(ip_ctx_data_t *ipctx);
static void
compute_midstep_def_grad(ip_ctx_data_t *ipctx);
static void
compute_rate_of_def(ip_ctx_data_t *ipctx);
static void
compute_midstep_rate_of_def(ip_ctx_data_t *ipctx, double time_step);
static void
compute_def_grad_dot(ip_ctx_data_t *ipctx);
static void
compute_vel_grad(ip_ctx_data_t *ipctx);
static void
compute_inv_def_grad(ip_ctx_data_t *ipctx);
static void
compute_polar_decomp(ip_ctx_data_t *ipctx);
static void
compute_rel_polar_decomp(ip_ctx_data_t *ipctx);
static void
compute_ln_rel_U(ip_ctx_data_t *ipctx);



static W_bool_t
get_kinem_var(W_mat_3d_ip_ctx_t *ctx,
              W_mat_3d_kinem_var_type_enum  type,
              double kinvar[3][3])
{
  ip_ctx_data_t *ipctx = (ip_ctx_data_t *)ctx->ctx_data;

  switch (type) {
  case MAT_3D_DEF_GRAD:
    if (!VALID(F))
      compute_def_grad(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, F);
    break;
  case MAT_3D_DEF_GRAD_DOT:
    if (!VALID(F))
      compute_def_grad_dot(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, F_dot);
    break;
  case MAT_3D_REL_DEF_GRAD:
    if (!VALID(rel_F))
      compute_rel_def_grad(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, rel_F);
    break;
  case MAT_3D_PREV_DEF_GRAD:
    if (!VALID(prev_F))
      compute_rel_def_grad(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, prev_F);
    break;
  case MAT_3D_MIDSTEP_DEF_GRAD:
    if (!VALID(midstep_F))
      compute_midstep_def_grad(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, midstep_F);
    break;
  case MAT_3D_CAUCHY_GREEN:
    if (!VALID(C))
      compute_cauchy_green(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, C);
    break;
  case MAT_3D_RATE_OF_DEF:
    if (!VALID(D))
      compute_rate_of_def(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, D);
    break;
  case MAT_3D_MIDSTEP_RATE_OF_DEF:
    if (!VALID(midstep_D))
      compute_midstep_rate_of_def(ipctx, ctx->time_step);
    MAT3DUTIL_COPY_3X3(kinvar, midstep_D);
    break;
  case MAT_3D_MIDSTEP_SPIN:
    if (!VALID(midstep_W))
      compute_midstep_rate_of_def(ipctx, ctx->time_step);
    MAT3DUTIL_COPY_3X3(kinvar, midstep_W);
    break;
  case MAT_3D_GREEN_LAGRANGE:
    if (!VALID(E))
      compute_green_lagrange(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, E);
    break;
  case MAT_3D_POLAR_R:
    if (!VALID(R))
      compute_polar_decomp(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, R);
    break;
  case MAT_3D_POLAR_U:
    if (!VALID(U))
      compute_polar_decomp(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, U);
    break;
  case MAT_3D_POLAR_REL_R:
    if (!VALID(rel_R))
      compute_rel_polar_decomp(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, rel_R);
    break;
  case MAT_3D_POLAR_REL_U:
    if (!VALID(rel_U))
      compute_rel_polar_decomp(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, rel_U);
    break;
  case MAT_3D_LN_POLAR_REL_U:
    if (!VALID(ln_rel_U))
      compute_ln_rel_U(ipctx);
    MAT3DUTIL_COPY_3X3(kinvar, ln_rel_U);
    break;
  default: {
      static W_bool_t complained = FALSE;
      if (!complained) {
        sprintf (W_buf(), "Unknown kinematic tensor type: %d",     (int)type);
        REPORT_WARNING (0, W_buf());
        complained = TRUE;
      }
    }
    return FALSE;
  }

  return TRUE;
}




static void
compute_rel_def_grad(ip_ctx_data_t *ipctx)
{
  priv_elem_t *t = ipctx->t;
  shape_func_t *sftab = ipctx->sf;
  int k;
  double f00 = 0, f01 = 0, f02 = 0;
  double f10 = 0, f11 = 0, f12 = 0;
  double f20 = 0, f21 = 0, f22 = 0;
  double x, y, z;
  double inv_prev_F[3][3];
  double det_prev_F;
  register double dphix, dphiy, dphiz;
  W_node_t *n;

  if (!VALID(F))
    compute_def_grad(ipctx);

  FOR (k, 4) {
    n = t->node[k];
    WN_get_prev_coords (n, &x, &y, &z);
    dphix  = sftab->n_x[k];
    dphiy  = sftab->n_y[k];
    dphiz  = sftab->n_z[k];
    f00 += dphix * x;
    f01 += dphiy * x;
    f02 += dphiz * x;
    f10 += dphix * y;
    f11 += dphiy * y;
    f12 += dphiz * y;
    f20 += dphix * z;
    f21 += dphiy * z;
    f22 += dphiz * z;
  }

  prev_F[0][0] = f00;
  prev_F[0][1] = f01;
  prev_F[0][2] = f02;

  prev_F[1][0] = f10;
  prev_F[1][1] = f11;
  prev_F[1][2] = f12;

  prev_F[2][0] = f20;
  prev_F[2][1] = f21;
  prev_F[2][2] = f22;

  det_prev_F = MAT3DUTIL_DET_3X3(prev_F);
  if (det_prev_F <= 0) {
    sprintf(W_buf(), "Failed to invert def grad");
    REPORT_FATAL_ERROR(NULL, W_buf());
  }

  MAT3DUTIL_INVERT_3X3(det_prev_F, prev_F, inv_prev_F);

  MAT3DUTIL_MULT_AB_3X3(rel_F, F, inv_prev_F);

  VALID(rel_F) = TRUE;
  VALID(prev_F) = TRUE;
}




static void
compute_midstep_def_grad(ip_ctx_data_t *ipctx)
{
  double m[3][3] = { {1, 0, 0}, {0, 1, 0}, {0, 0, 1} }, s[3][3], inv_s[3][3];

  if (!VALID(rel_F))
    compute_rel_def_grad(ipctx);

  MAT3DUTIL_ADD_AB_3X3(s, m, rel_F);
  MAT3DUTIL_SCALE_3X3(s, 0.5);
  if (!W_mat3dutil_invert_3x3(s, inv_s)) {
    sprintf(W_buf(), "Failed to invert");
    REPORT_FATAL_ERROR(NULL, W_buf());
  }
  MAT3DUTIL_MULT_AB_3X3(midstep_F, inv_s, rel_F);

  VALID(midstep_F) = TRUE;
}


static void
compute_vel_grad(ip_ctx_data_t *ipctx)
{
  if (!VALID(F))
    compute_def_grad(ipctx);
  if (!VALID(F_dot))
    compute_def_grad_dot(ipctx);
  if (!VALID(inv_F))
    compute_inv_def_grad(ipctx);

  MAT3DUTIL_MULT_AB_3X3(L, F_dot, inv_F);

  VALID(L) = TRUE;
}


static void
compute_rate_of_def(ip_ctx_data_t *ipctx)
{
  int i, j;

  if (!VALID(L))
    compute_vel_grad(ipctx);

  FOR(i, 3)
  FOR(j, 3)
  D[i][j] = 0.5 * (L[i][j] + L[j][i]);

  VALID(D) = TRUE;
}



static void
compute_def_grad(ip_ctx_data_t *ipctx)
{
  priv_elem_t *t = ipctx->t;
  shape_func_t *sftab = ipctx->sf;
  int k;
  double f00 = 0, f01 = 0, f02 = 0;
  double f10 = 0, f11 = 0, f12 = 0;
  double f20 = 0, f21 = 0, f22 = 0;
  W_node_t *n;
  double x, y, z;
  register double dphix, dphiy, dphiz;

  /* Deformation gradient */
  FOR(k, 4) {
    n = t->node[k];
    WN_get_coords (n, &x, &y, &z);
    dphix  = sftab->n_x[k];
    dphiy  = sftab->n_y[k];
    dphiz  = sftab->n_z[k];
    f00 += dphix * x;
    f01 += dphiy * x;
    f02 += dphiz * x;
    f10 += dphix * y;
    f11 += dphiy * y;
    f12 += dphiz * y;
    f20 += dphix * z;
    f21 += dphiy * z;
    f22 += dphiz * z;
  }

  F[0][0] = f00;
  F[0][1] = f01;
  F[0][2] = f02;

  F[1][0] = f10;
  F[1][1] = f11;
  F[1][2] = f12;

  F[2][0] = f20;
  F[2][1] = f21;
  F[2][2] = f22;

  VALID(F) = TRUE;
}



static void
compute_def_grad_dot(ip_ctx_data_t *ipctx)
{
  priv_elem_t *t = ipctx->t;
  shape_func_t *sftab = ipctx->sf;
  int k;
  double f00 = 0, f01 = 0, f02 = 0;
  double f10 = 0, f11 = 0, f12 = 0;
  double f20 = 0, f21 = 0, f22 = 0;
  W_node_t *n;
  double x, y, z;
  register double dphix, dphiy, dphiz;

  /* Deformation gradient */
  FOR(k, 4) {
    n = t->node[k];
    x = n->velocity_t.x;
    y = n->velocity_t.y;
    z = n->velocity_t.z;
    dphix  = sftab->n_x[k];
    dphiy  = sftab->n_y[k];
    dphiz  = sftab->n_z[k];
    f00 += dphix * x;
    f01 += dphiy * x;
    f02 += dphiz * x;
    f10 += dphix * y;
    f11 += dphiy * y;
    f12 += dphiz * y;
    f20 += dphix * z;
    f21 += dphiy * z;
    f22 += dphiz * z;
  }

  F_dot[0][0] = f00;
  F_dot[0][1] = f01;
  F_dot[0][2] = f02;

  F_dot[1][0] = f10;
  F_dot[1][1] = f11;
  F_dot[1][2] = f12;

  F_dot[2][0] = f20;
  F_dot[2][1] = f21;
  F_dot[2][2] = f22;

  VALID(F_dot) = TRUE;
}




static void
compute_green_lagrange(ip_ctx_data_t *ipctx)
{
  if (!VALID(C))
    compute_cauchy_green(ipctx);

  /* Green-Lagrange strain tensor */
  E[0][0] = 0.5 * (C[0][0] - 1);
  E[0][1] = 0.5 * C[0][1];
  E[0][2] = 0.5 * C[0][2];
  E[1][0] = 0.5 * C[1][0];
  E[1][1] = 0.5 * (C[1][1] - 1);
  E[1][2] = 0.5 * C[1][2];
  E[2][0] = 0.5 * C[2][0];
  E[2][1] = 0.5 * C[2][1];
  E[2][2] = 0.5 * (C[2][2] - 1);

  VALID(E) = TRUE;
}



static void
compute_cauchy_green(ip_ctx_data_t *ipctx)
{
  if (!VALID(F))
    compute_def_grad(ipctx);

  /* Cauchy-Green deformation tensor */
  MAT3DUTIL_MULT_ATB_3X3(C, F, F);

  VALID(C) = TRUE;
}


static void
compute_inv_def_grad(ip_ctx_data_t *ipctx)
{
  if (!VALID(F))
    compute_def_grad(ipctx);

  if (!W_mat3dutil_invert_3x3(F, inv_F)) {
    sprintf(W_buf(), "Failed to invert def grad");
    REPORT_FATAL_ERROR(NULL, W_buf());
  }

  VALID(inv_F) = TRUE;
}


static void
compute_midstep_rate_of_def (ip_ctx_data_t *ipctx, double time_step)
{
  int i, j;
  double m[3][3] = { {1, 0, 0}, {0, 1, 0}, {0, 0, 1} };
  double M[3][3] = { {-1, 0, 0}, {0, -1, 0}, {0, 0, -1} };
  double s[3][3], inv_s[3][3], q[3][3], r[3][3];
  double inv_dt = 1 / time_step;

  if (!VALID(rel_F))
    compute_rel_def_grad(ipctx);

  MAT3DUTIL_ADD_AB_3X3(s, m, rel_F);
  MAT3DUTIL_SCALE_3X3(s, 0.5);
  if (!W_mat3dutil_invert_3x3(s, inv_s)) {
    sprintf(W_buf(), "Failed to invert def grad");
    REPORT_FATAL_ERROR(NULL, W_buf());
  }
  MAT3DUTIL_ADD_AB_3X3(q, rel_F, M);
  MAT3DUTIL_MULT_AB_3X3(r, inv_s, q);

  FOR(i, 3)
  FOR(j, 3) {
    midstep_D[i][j] = 0.5 * (r[i][j] + r[j][i]);
    midstep_W[i][j] = 0.5 * (r[i][j] - r[j][i]);
  }
  MAT3DUTIL_SCALE_3X3(midstep_D, inv_dt);
  MAT3DUTIL_SCALE_3X3(midstep_W, inv_dt);

  VALID(midstep_D) = TRUE;
  VALID(midstep_W) = TRUE;
}




static void
compute_polar_decomp(ip_ctx_data_t *ipctx)
{
  if (!VALID(F))
    compute_def_grad(ipctx);

  if (!W_polar_decomposition (F, R, U)) {
    REPORT_FATAL_ERROR (NULL, "Failed in W_polar_decomposition()");
  }
  VALID(R) = TRUE;
  VALID(U) = TRUE;
}


static void
compute_rel_polar_decomp(ip_ctx_data_t *ipctx)
{
  if (!VALID(rel_F))
    compute_rel_def_grad(ipctx);

  if (!W_polar_decomposition (rel_F, rel_R, rel_U)) {
    REPORT_FATAL_ERROR (NULL, "Failed in W_polar_decomposition()");
  }
  VALID(rel_R) = TRUE;
  VALID(rel_U) = TRUE;
}



static void
compute_ln_rel_U(ip_ctx_data_t *ipctx)
{
  double lneval[3];

  if (!VALID(rel_U))
    compute_rel_polar_decomp(ipctx);

  W_ln_U (rel_U, ln_rel_U);

  VALID(ln_rel_U) = TRUE;
}

static W_float_t
kinetic_energy_diag (W_elem_t *el)
{
  priv_elem_t *e = (priv_elem_t *)el->data;
  W_vect_t v;
  int i;
  double ke = 0;
  double m = total_mass (el) / WBMAT_NNODES;

  FOR (i, WBMAT_NNODES) {
    GET_FROM_NODE (e->node[i], WN_get_velocity, v);
    ke += VECMAC_DOT (v, v);
  }
  return (m * ke) / 2;
}

static W_float_t
kinetic_energy (W_elem_t *el, W_mass_type_enum mass_type)
{
  return kinetic_energy_diag (el); /* diagonal mass only */
}

static W_elem_methods_t methods = {
                                    elem_elcode,
                                    type_name,
                                    assemble_restoring_forces,
                                    assemble_mass,
                                    total_mass,
                                    NULL,
                                    suggested_time_step,
                                    NULL,
                                    NULL,
                                    NULL,
                                    NULL,
                                    NULL,
                                    NULL,
                                    NULL,
                                    NULL,
                                    NULL,
                                    write_postproc_file,
#ifdef X_XFACE
                                    modify_elem_grep,
#endif
                                    NULL,
                                    NULL,
                                    NULL,
                                    num_of_faces,
                                    get_face_vertex_nodes,
                                    get_contact_faces,
                                    assemble_residual_forces,
                                    assemble_eff_stiffness,
                                    kinetic_energy
                                  };


/*
  Element reader.
  Reads element data for the tet4 element type
  from stream as long as there is no
  END_OF_BLOCK tag in the input line. Creates elements.
 */
W_bool_t
tet4_reader (W_domain_t *d /* domain */)
{
  const char *input_line;
  long elem_num;
  priv_elem_t *t;
#ifdef PVM_NEED_GHOST_ELEMENTS

  int is_ghost = FALSE;
#endif

  W_index_t nodeid, matid;
  int i, k;
  W_material_t *mat;
  W_mat_3d_t *m;
  W_loc_t loc;

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line)) {
      return TRUE;
    } else {
      TOKENS_PARSE(input_line);
      t = w_new(priv_elem_t);
      CHECK_MALLOC(t);
      elem_num = TOKEN_INT(1);
      FOR(i, 4) {
        nodeid = TOKEN_INT(2+i);
        t->node[i] = WD_get_node_by_idx(d, nodeid);
        if (t->node[i] == NULL) {
          sprintf(W_buf(), "No such node: %ld", nodeid);
          REPORT_ERROR(d->errh, W_buf());
        }
      }
      matid = TOKEN_INT(6);
      mat = WD_get_material_by_idx(d, matid);
      if (mat == NULL) {
        sprintf(W_buf(), "No such material: %ld", matid);
        REPORT_ERROR(d->errh, W_buf());
      }
      t->material     = mat;
      m = (W_mat_3d_t *)t->material->data;
      CENTROID (loc, t);
      t->matst        = WM3D_get_mat_state(m, &loc);
#ifdef PVM_NEED_GHOST_ELEMENTS

      is_ghost = FALSE;
      if (TOKENS_HAVE_KEYWORD("GHOST"))
        is_ghost = TRUE;
#endif

      WD_add_element(d, elem_num,
                     WE_new_element(d, &methods, t
#                                   ifdef PVM_NEED_GHOST_ELEMENTS
                                    , is_ghost
#                                   endif
                                   ));
    } /* closing else */
  }
  return FALSE;
}
