#include "Wconfig.h"
#include "Wxf.h"
#include "Wtfuncs.h"
#include "Wdomain.h"

/* RCSHEADER "$Id: tfuncs.c,v 1.1.1.1 2001/07/24 02:42:52 pkrysl Exp $" */

static W_float_t
value_of_tabular_tfunc(W_time_func_t *tf,
                           W_float_t time) 
{
  W_float_t v;
  W_time_func_tabular_t *tft = &tf->data.tabular;

  if (tf->data.tabular.cached_time == time)
    return tf->data.tabular.cached_value;

  v = W_eval_ordinate_at_abscissa(tf->data.tabular.num_of_pairs,
                                  tft->pairs, time); 

  tf->data.tabular.cached_value = v;
  tf->data.tabular.cached_time  = time;

  return v;
}


/*
   Evaluate time function for the given time.
 */
W_float_t /* value of the time function */
WTF_func_val (W_time_func_t *tf, /* time function */
             W_float_t time /* time */
              )
{
  switch (tf->tfcode) {
  case W_HARMONIC_TIME_FUNC:
    return sin (tf->data.harmonic.circular_frequency * time + tf->data.harmonic.phase);
  case W_TABULAR_TIME_FUNC:
    return value_of_tabular_tfunc (tf, time);
  default:
    REPORT_FATAL_ERROR(NULL, "Unknown time function type");
    break;
  }
  return 0;
}

/*
   Create a new time function.
 */
static W_time_func_t * /* new time function */
new_time_func(void)
{
  W_time_func_t *tf;
  
  tf = w_new(W_time_func_t);
  CHECK_MALLOC(tf);

  return tf;
}


static void
read_tabular_tfunc(W_domain_t *d)
{
  const char *input_line;
  W_time_func_t *tf;
  long tf_num, num_pairs;
  double a, o;
  int i;
  
  if ((input_line = W_get_input_line(d)) == NULL)
    BAD_INPUT();
  if (sscanf(input_line, "%ld %ld", &tf_num, &num_pairs) != 2) {
    BAD_INPUT();
  } else {
    tf = new_time_func();
    tf->id = tf_num;
    tf->tfcode = W_TABULAR_TIME_FUNC;
    if (num_pairs <= 0)
      BAD_INPUT();
    tf->data.tabular.num_of_pairs = num_pairs;
    tf->data.tabular.cached_time  = -1;
    tf->data.tabular.cached_value = -1;
    tf->data.tabular.pairs =
      w_new_array(num_pairs, W_abscissa_ordinate_pair_t);
    CHECK_MALLOC(tf->data.tabular.pairs);
    
    for (i = 0; i < num_pairs; i++) {
      if ((input_line = W_get_input_line(d)) == NULL)
        BAD_INPUT();
      if (sscanf(input_line, "%lf %lf", &a, &o) != 2) {
        BAD_INPUT();
      } else {
        if (i > 0)
          if (tf->data.tabular.pairs[i-1].abscissa > a)
            BAD_INPUT();
        tf->data.tabular.pairs[i].abscissa = a;
        tf->data.tabular.pairs[i].ordinate = o;
      }
    }
    WD_add_time_func (d, tf->id, tf);
  }
}


/*
   Reader of time functions.
 */
W_bool_t
W_time_func_reader (W_domain_t *d)
{
  const char *input_line;
  W_time_func_t *tf;
  long tf_num, tfcode;
  double phase, circ_freq;

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return TRUE;
    else {
      if (sscanf(input_line, "%ld", &tfcode) != 1)
        BAD_INPUT();
      switch (tfcode) {
      case W_HARMONIC_TIME_FUNC:
        if ((input_line = W_get_input_line(d)) == NULL)
          BAD_INPUT();
        if (sscanf(input_line, "%ld %lf %lf",
                 &tf_num, &circ_freq, &phase) != 3) {
          BAD_INPUT();
        } else {
          tf = new_time_func();
          tf->id = tf_num; 
          tf->tfcode = W_HARMONIC_TIME_FUNC;
          tf->data.harmonic.circular_frequency = circ_freq;
          tf->data.harmonic.phase              = phase;
          WD_add_time_func(d, tf->id, tf); 
        }
        break;
      case W_TABULAR_TIME_FUNC:
        read_tabular_tfunc (d);
        break;
      default:
        return FALSE;
      }
    }
  }
  return FALSE;
}





