
#include "Esimple.h" /* The only one Elixir include needed */
#include "WEFG.h" /* The only one Elixir include needed */





int 
main(int argc, char **argv)
{
  ESIBuildInterface(ESI_GRAPHIC_EDITOR_MASK, argc, argv); /* create GUI */
  ESIPopup(); /* Just to make sure the main control panel is popped up */
              /* before the view. */
  ESIHandleCmd("COLOR WHITE");  /* Change active color */
  ESIHandleCmd("NEWFRAME WIRE");     /* Open new view */
  ESIPopupAndRun(); /* Now let us see what we can do with it. */
  return 1;
}



static void 
CheckXsect(Widget w, XtPointer text_ptr, XtPointer call_data);


void
ESICustomize(Widget parent_pane)
{
  Arg args[3];
  int argn;
  Widget w;

  argn = 0;
  w = ESIAddButton("check_xsect", "Check intersection",
                   commandWidgetClass,
                   parent_pane, args, argn, CheckXsect, NULL);
}


static void 
do_check_xsect(EView *v_p, caddr_t data_p);

static void
CheckXsect(Widget w, XtPointer ptr, XtPointer call)
{
  EVSetSelectActionFunction(do_check_xsect, NULL, NULL);
  EMInstallHandler(ESIModel(), EVSelectGraphicsHandler,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL,
		      NULL, NULL);
  ERptStatusMessage(1, 1, "Select line and polygon");
}


int
check_isect(WEFG_segment_t *s, WEFG_b_face_t *f);


static void 
do_check_xsect(EView *v_p, caddr_t data_p)
{
  GraphicObj *gp;
  int i;
  WCRec pts[8];
  int np;
  int had_poly = 0;
  int had_line = 0;
  WEFG_segment_t s;
  WEFG_b_face_t f;
  WEFG_vertex_t v[4];
  static int planar = 0; 

  f.v[0] = &v[0];
  f.v[1] = &v[1];
  f.v[2] = &v[2];
  f.v[3] = &v[3];

  f.planar = planar = !planar;
  
  if (EVItemsInSelectionList() <= 0) {
    ERptErrMessage(1, 2, "No objects selected.", ERROR_GRADE);
    goto exit;
  }

  while ((gp = EVGetFromSelectionList()) != NULL) {
    if        (EGGraphicType(gp) == EG_LINE3D) {
      had_line = 1;
      if (!EGGetGraphicsGeometry(gp, pts, &np)) 
        ERptStatusMessage(1, 1, "EGGetGraphicsGeometry failed");
      s.start.x = pts[0].x;
      s.start.y = pts[0].y;
      s.start.z = pts[0].z;
      s.end.x   = pts[1].x;
      s.end.y   = pts[1].y;
      s.end.z   = pts[1].z;
    }
    if        (EGGraphicType(gp) == EG_TRIANGLE3D) {
      had_poly = 1;
      if (!EGGetGraphicsGeometry(gp, pts, &np)) 
        ERptStatusMessage(1, 1, "EGGetGraphicsGeometry failed");
      f.num_vert = 3;
      f.v[0]->loc.x = pts[0].x;
      f.v[0]->loc.y = pts[0].y;
      f.v[0]->loc.z = pts[0].z;
      f.v[1]->loc.x = pts[1].x;
      f.v[1]->loc.y = pts[1].y;
      f.v[1]->loc.z = pts[1].z;
      f.v[2]->loc.x = pts[2].x;
      f.v[2]->loc.y = pts[2].y;
      f.v[2]->loc.z = pts[2].z;
    }
    if        (EGGraphicType(gp) == EG_QUAD3D) {
      had_poly = 1;
      if (!EGGetGraphicsGeometry(gp, pts, &np)) 
        ERptStatusMessage(1, 1, "EGGetGraphicsGeometry failed");
      f.num_vert = 4;
      f.v[0]->loc.x = pts[0].x;
      f.v[0]->loc.y = pts[0].y;
      f.v[0]->loc.z = pts[0].z;
      f.v[1]->loc.x = pts[1].x;
      f.v[1]->loc.y = pts[1].y;
      f.v[1]->loc.z = pts[1].z;
      f.v[2]->loc.x = pts[2].x;
      f.v[2]->loc.y = pts[2].y;
      f.v[2]->loc.z = pts[2].z;
      f.v[3]->loc.x = pts[3].x;
      f.v[3]->loc.y = pts[3].y;
      f.v[3]->loc.z = pts[3].z;
    }
  }

  if ((!had_line) || (!had_poly)) {
    ERptErrMessage(1, 2, "No line or poly in selection.", ERROR_GRADE);
    goto exit;
  } else {
    char buf[132];

    if (check_isect(&s, &f))
      sprintf(buf, "INTERSECT");
    else
      sprintf(buf, "NO INTERSECTION");
    ESIPopupInfo(buf);
  }

 exit:
  EVUninstallHandler(v_p);
  ERptStatusMessage(1, 1, "Done");
}
