#include "Wconfig.h"
#include "Wxf.h"
#include "Wdomain.h"
#include "Whistry.h"
#include "Welem.h"
#include "Wtfuncs.h"
#include "Wnode.h"
#include "Wloads.h"
#include "Wdispbc.h"
#include "Wivcond.h"
#include "Wmats.h"
#include "Wcontact.h"
#include "Wvset.h"
#include "Wpvm.h"
#include "Wmpi.h"
#include "Wcdi.h"
#include "Wnewmarki.h"
#include "Wpod.h"


static double max_dyn_displ_of_tip          = 0;
static double static_displ_of_tip           = 0;
static double settling_time                 = 0;
static double total_weight                  = 0;
static W_index_t tip_node                   = 1;
static double settled_deform_energy_ratio   = 0.001;

static int num_samples                      = 0;
static int tot_samples                      = 100;
static double deform_energy_accumulator     = 0;
static double max_deform_energy             = 0;
static double min_deform_energy             = FLT_MAX;

static void 
update_truss_response (W_domain_t *d)
{
  double deform_energy = 0, deform_energy_average = 0, kinetic_energy = 0;
  double x, y, z;
  W_node_t *n;

  n = WD_get_node_by_idx (d, tip_node);
  if (!n) REPORT_FATAL_ERROR ("Node not found");
  WN_get_displ (n, &x, &y, &z);
  if (fabs (z) > fabs (max_dyn_displ_of_tip))
    max_dyn_displ_of_tip = z;
  static_displ_of_tip  = z;

  deform_energy  = WD_current_deformation_energy (d);
  deform_energy_accumulator += deform_energy; num_samples++;
  min_deform_energy = min (min_deform_energy, deform_energy);
  max_deform_energy = max (max_deform_energy, deform_energy);
  kinetic_energy = WD_current_kinetic_energy (d);

  if (num_samples > tot_samples) {
    deform_energy_average = deform_energy_accumulator / num_samples;
    if (fabs (max_deform_energy - min_deform_energy) / deform_energy_average < settled_deform_energy_ratio &&
        kinetic_energy / deform_energy_average < settled_deform_energy_ratio) {
      options.target_time = WD_time (d); /* force the integrator to stop in the next step */
    } else {
      num_samples = 0;
      deform_energy_accumulator = 0;
      min_deform_energy = FLT_MAX;
      max_deform_energy = 0;
    }
  }
}

static void
print_truss_response (W_domain_t *d)
{
  total_weight = WD_total_mass (d);
  settling_time = WD_time(d);
  printf ("# max_dyn_displ_of_tip  static_displ_of_tip   settling_time   total_weight\n");
  printf ("%g %g %g %g\n", max_dyn_displ_of_tip, static_displ_of_tip, settling_time, total_weight);
}


static void 
collect_truss_response (W_domain_t *d, int final_call)
{
  if (final_call) print_truss_response (d);
  else            update_truss_response (d);
}

int
wasp_custom_setup (int *argcp, char ***argvp, W_domain_t *d)
{
  W_domain_set_response_observer_func (d, collect_truss_response);
  return TRUE;
}

