
#include "Wconfig.h"
#include "Wxf.h"
#include "Wtypes.h"
#include "Wutil.h"
#include "Wmpi.h"

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <timeP.h>

/* RCSHEADER "$Id: util.c,v 1.3 2002/11/08 20:34:17 cvs Exp $" */

const char *
W_get_input_line (W_domain_t *d)
{
  if (tokens_next_line (d->parser) == 0) {
    return NULL;
  } else {
    return tokens_curr_line_contents (d->parser);
  }
}

/*
   Compute length of vector.
 */
double 
W_vector_3d_length (W_coords_3d_t *vec)
{
  double sum;
  
  sum = vec->x * vec->x;
  sum += vec->y * vec->y;
  sum += vec->z * vec->z;
  
  return sqrt(sum); 
} 



/*
   Compute vector cross product.
 */
void
W_cross_prod_3d (W_coords_3d_t *reslt_vec,
                 W_coords_3d_t *Vec1,
                 W_coords_3d_t *Vec2
                 )
{
  reslt_vec->x = Vec1->y * Vec2->z - Vec1->z * Vec2->y;
  reslt_vec->y = Vec1->z * Vec2->x - Vec1->x * Vec2->z;
  reslt_vec->z = Vec1->x * Vec2->y - Vec1->y * Vec2->x; 
}



/*
   Compute dot product of vectors.
 */
W_float_t
W_dot_prod_3d (W_coords_3d_t *Vec1, W_coords_3d_t *Vec2)
{
  return (Vec1->x * Vec2->x + Vec1->y * Vec2->y + Vec1->z * Vec2->z);
}


/*
   Normalize vector.
 */
W_bool_t /* FALSE if zero vector */
W_normalize_vect_3d (W_coords_3d_t *vec)
{
  double tmp, length;

  length = 0;
  tmp = vec->x; tmp *= tmp;
  length += tmp;
  tmp = vec->y; tmp *= tmp;
  length += tmp;
  tmp = vec->z; tmp *= tmp;
  length += tmp;
  if (length < (double)0) {
    return FALSE;
  } else if (length == (double)0) {
    vec->x = vec->y = vec->z = 0;
  } else {
    length = sqrt(length);
    vec->x /= length;
    vec->y /= length;
    vec->z /= length;
  }
  return TRUE;
}


/*
   Rotate triad by the rotation vector.
 */
void
W_rotate_triad_by_vector (W_triad_3d_t *t, W_coords_3d_t *v)
{
  double ad, cosad, sadad, cad2, rb1, rb2;
  W_coords_3d_t rxb1, rxb2;
  
  ad = sqrt(v->x*v->x + v->y*v->y + v->z*v->z);
  if (ad > 0) {
    cosad = cos(ad);
    sadad = sin(ad)/ad;
    cad2  = (1.- cosad)/ad/ad;
    rb1   = W_dot_prod_3d(v, &(t->vect1));
    W_cross_prod_3d(&rxb1, v, &(t->vect1));
    rb2   = W_dot_prod_3d(v, &(t->vect2));
    W_cross_prod_3d(&rxb2, v, &(t->vect2));
    t->vect1.x = (t->vect1.x * cosad + v->x * rb1 * cad2 + rxb1.x * sadad);
    t->vect2.x = (t->vect2.x * cosad + v->x * rb2 * cad2 + rxb2.x * sadad);
    t->vect1.y = (t->vect1.y * cosad + v->y * rb1 * cad2 + rxb1.y * sadad);
    t->vect2.y = (t->vect2.y * cosad + v->y * rb2 * cad2 + rxb2.y * sadad);
    t->vect1.z = (t->vect1.z * cosad + v->z * rb1 * cad2 + rxb1.z * sadad);
    t->vect2.z = (t->vect2.z * cosad + v->z * rb2 * cad2 + rxb2.z * sadad);
    W_cross_prod_3d(&(t->vect3), &(t->vect1), &(t->vect2));
  }
}



/*
   Left-shift triad by another triad (triad == orthonormal matrix).
 */
void
W_L_shift_triad (W_triad_3d_t *out, W_triad_3d_t *by, W_triad_3d_t *in)
{
  assert(out != NULL && by != NULL && in != NULL);
  
  /* OUT = BY * IN, i.e. OUT_ij = Sum_k BY_ik * IN_kj, */
  /* i.e. out->vectj.i = Sum_k by->vectk.i * in->vectj.k */
  
  out->vect1.x = (  by->vect1.x * in->vect1.x
                  + by->vect2.x * in->vect1.y
                  + by->vect3.x * in->vect1.z);
  out->vect1.y = (  by->vect1.y * in->vect1.x
                  + by->vect2.y * in->vect1.y
                  + by->vect3.y * in->vect1.z);
  out->vect1.z = (  by->vect1.z * in->vect1.x
                  + by->vect2.z * in->vect1.y
                  + by->vect3.z * in->vect1.z);

  out->vect2.x = (  by->vect1.x * in->vect2.x
                  + by->vect2.x * in->vect2.y
                  + by->vect3.x * in->vect2.z);
  out->vect2.y = (  by->vect1.y * in->vect2.x
                  + by->vect2.y * in->vect2.y
                  + by->vect3.y * in->vect2.z);
  out->vect2.z = (  by->vect1.z * in->vect2.x
                  + by->vect2.z * in->vect2.y
                  + by->vect3.z * in->vect2.z);

  out->vect3.x = (  by->vect1.x * in->vect3.x
                  + by->vect2.x * in->vect3.y
                  + by->vect3.x * in->vect3.z);
  out->vect3.y = (  by->vect1.y * in->vect3.x
                  + by->vect2.y * in->vect3.y
                  + by->vect3.y * in->vect3.z);
  out->vect3.z = (  by->vect1.z * in->vect3.x
                  + by->vect2.z * in->vect3.y
                  + by->vect3.z * in->vect3.z);
}


/*
   Left-shift triad by another, transposed triad (triad == orthonormal matrix).
 */
void
W_transp_L_shift_triad (W_triad_3d_t *out, W_triad_3d_t *by, W_triad_3d_t *in)
{
  assert(out != NULL && by != NULL && in != NULL);
  
  /* OUT = transp(BY) * IN, i.e. OUT_ij = Sum_k BY_ki * IN_kj, */
  /* i.e. out->vectj.i = Sum_k by->vecti.k * in->vectj.k */
  
  out->vect1.x = (  by->vect1.x * in->vect1.x
                  + by->vect1.y * in->vect1.y
                  + by->vect1.z * in->vect1.z);
  out->vect1.y = (  by->vect2.x * in->vect1.x
                  + by->vect2.y * in->vect1.y
                  + by->vect2.z * in->vect1.z);
  out->vect1.z = (  by->vect3.x * in->vect1.x
                  + by->vect3.y * in->vect1.y
                  + by->vect3.z * in->vect1.z);

  out->vect2.x = (  by->vect1.x * in->vect2.x
                  + by->vect1.y * in->vect2.y
                  + by->vect1.z * in->vect2.z);
  out->vect2.y = (  by->vect2.x * in->vect2.x
                  + by->vect2.y * in->vect2.y
                  + by->vect2.z * in->vect2.z);
  out->vect2.z = (  by->vect3.x * in->vect2.x
                  + by->vect3.y * in->vect2.y
                  + by->vect3.z * in->vect2.z);

  out->vect3.x = (  by->vect1.x * in->vect3.x
                  + by->vect1.y * in->vect3.y
                  + by->vect1.z * in->vect3.z);
  out->vect3.y = (  by->vect2.x * in->vect3.x
                  + by->vect2.y * in->vect3.y
                  + by->vect2.z * in->vect3.z);
  out->vect3.z = (  by->vect3.x * in->vect3.x
                  + by->vect3.y * in->vect3.y
                  + by->vect3.z * in->vect3.z);
}


/*
   Right-shift triad by another triad (triad == orthonormal matrix).
 */
void
W_R_shift_triad (W_triad_3d_t *out, W_triad_3d_t *in, W_triad_3d_t *by)
{
  assert(out != NULL && by != NULL && in != NULL);
  
  /* OUT = IN * BY, i.e. OUT_ij = Sum_k IN_ik * BY_kj, */
  /* i.e. out->vectj.i = Sum_k in->vectk.i * by->vectj.k */
  
  out->vect1.x = (  in->vect1.x * by->vect1.x
                  + in->vect2.x * by->vect1.y
                  + in->vect3.x * by->vect1.z);
  out->vect1.y = (  in->vect1.y * by->vect1.x
                  + in->vect2.y * by->vect1.y
                  + in->vect3.y * by->vect1.z);
  out->vect1.z = (  in->vect1.z * by->vect1.x
                  + in->vect2.z * by->vect1.y
                  + in->vect3.z * by->vect1.z);

  out->vect2.x = (  in->vect1.x * by->vect2.x
                  + in->vect2.x * by->vect2.y
                  + in->vect3.x * by->vect2.z);
  out->vect2.y = (  in->vect1.y * by->vect2.x
                  + in->vect2.y * by->vect2.y
                  + in->vect3.y * by->vect2.z);
  out->vect2.z = (  in->vect1.z * by->vect2.x
                  + in->vect2.z * by->vect2.y
                  + in->vect3.z * by->vect2.z);

  out->vect3.x = (  in->vect1.x * by->vect3.x
                  + in->vect2.x * by->vect3.y
                  + in->vect3.x * by->vect3.z);
  out->vect3.y = (  in->vect1.y * by->vect3.x
                  + in->vect2.y * by->vect3.y
                  + in->vect3.y * by->vect3.z);
  out->vect3.z = (  in->vect1.z * by->vect3.x
                  + in->vect2.z * by->vect3.y
                  + in->vect3.z * by->vect3.z);
}


/*
Right-shift triad by another, transposed triad (triad == orthonormal matrix).
 */

void
W_transp_R_shift_triad(W_triad_3d_t *out, W_triad_3d_t *in, W_triad_3d_t *by)
{
  assert(out != NULL && by != NULL && in != NULL);
  
  /* OUT = IN * transp(BY), i.e. OUT_ij = Sum_k IN_ik * BY_jk, */
  /* i.e. out->vectj.i = Sum_k in->vectk.i * by->vectk.j */
  
  out->vect1.x = (  in->vect1.x * by->vect1.x
                  + in->vect2.x * by->vect2.x
                  + in->vect3.x * by->vect3.x);
  out->vect1.y = (  in->vect1.y * by->vect1.x
                  + in->vect2.y * by->vect2.x
                  + in->vect3.y * by->vect3.x);
  out->vect1.z = (  in->vect1.z * by->vect1.x
                  + in->vect2.z * by->vect2.x
                  + in->vect3.z * by->vect3.x);

  out->vect2.x = (  in->vect1.x * by->vect1.y
                  + in->vect2.x * by->vect2.y
                  + in->vect3.x * by->vect3.y);
  out->vect2.y = (  in->vect1.y * by->vect1.y
                  + in->vect2.y * by->vect2.y
                  + in->vect3.y * by->vect3.y);
  out->vect2.z = (  in->vect1.z * by->vect1.y
                  + in->vect2.z * by->vect2.y
                  + in->vect3.z * by->vect3.y);

  out->vect3.x = (  in->vect1.x * by->vect1.z
                  + in->vect2.x * by->vect2.z
                  + in->vect3.x * by->vect3.z);
  out->vect3.y = (  in->vect1.y * by->vect1.z
                  + in->vect2.y * by->vect2.z
                  + in->vect3.y * by->vect3.z);
  out->vect3.z = (  in->vect1.z * by->vect1.z
                  + in->vect2.z * by->vect2.z
                  + in->vect3.z * by->vect3.z);
}



/* Extract quaternion from an orthonormal matrix (triad).
   Procedure extracts the quaternion parameters from the rotation
   matrix using the Spurrier's algorithm. */
void
W_extract_quaternion_from_triad (W_triad_3d_t *rotm, W_quaternion_t *q)
{
  int imax = 0;
  double q0 = 0, q1 = 0, q2 = 0, q3 = 0; 
  double rmtrace, rm11, rm22, rm33, maxt;

  rm11    = rotm->vect1.x;
  rm22    = rotm->vect2.y;
  rm33    = rotm->vect3.z;
  rmtrace = rm11 + rm22 + rm33;
  maxt    = rmtrace;
  if (rm11 > maxt) { maxt  = rm11; imax  = 1; }
  if (rm22 > maxt) { maxt  = rm22; imax  = 2; }
  if (rm33 > maxt) { maxt  = rm33; imax  = 3; }

  if        (imax == 0) {
    q0 = 0.5 * sqrt(1 + rmtrace);
    q1 = (rotm->vect2.z - rotm->vect3.y) * 0.25 / q0;
    q2 = (rotm->vect3.x - rotm->vect1.z) * 0.25 / q0;
    q3 = (rotm->vect1.y - rotm->vect2.x) * 0.25 / q0;
  } else if (imax == 1) {
    q1 = sqrt(0.5 * rm11 + 0.25 * (1 - rmtrace));
    q0 = (rotm->vect2.z - rotm->vect3.y) * 0.25 / q1;
    q2 = (rotm->vect2.x + rotm->vect1.y) * 0.25 / q1;
    q3 = (rotm->vect1.z + rotm->vect3.x) * 0.25 / q1;
    if (q0 < 0) { q0 = -q0; q1 = -q1; q2 = -q2; q3 = -q3; }
  } else if (imax == 2) {
    q2 = sqrt(0.5 * rm22 + 0.25 *(1 - rmtrace));
    q0 = (rotm->vect3.x - rotm->vect1.z) * 0.25 / q2;
    q1 = (rotm->vect2.x + rotm->vect1.y) * 0.25 / q2;
    q3 = (rotm->vect2.z + rotm->vect3.y) * 0.25 / q2;
    if (q0 < 0) { q0 = -q0; q1 = -q1; q2 = -q2; q3 = -q3; }
  } else if (imax == 3) {
    q3 = sqrt(0.5 * rm33 + 0.25 * (1 - rmtrace));
    q0 = (rotm->vect1.y - rotm->vect2.x) * 0.25 / q3;
    q2 = (rotm->vect2.z + rotm->vect3.y) * 0.25 / q3;
    q1 = (rotm->vect1.z + rotm->vect3.x) * 0.25 / q3;
    if (q0 < 0) { q0 = -q0; q1 = -q1; q2 = -q2; q3 = -q3; }
  }
  q->q0 = q0;
  q->q1 = q1;
  q->q2 = q2;
  q->q3 = q3;
}


/*
   Extract orthonormal matrix (triad) from a quaternion.
 */
void
W_extract_triad_from_quaternion (W_quaternion_t *q, W_triad_3d_t *rotm)
{
  double rotvn, q0, q1, q2, q3; 
  double q12, q13, q23, q00, q10, q20, q30;

  q0    = q->q0;
  q1    = q->q1;
  q2    = q->q2;
  q3    = q->q3;
  rotvn = sqrt(q1*q1 + q2*q2 + q3*q3);
  if (rotvn <= 0) {
    rotm->vect1.x = 1;
    rotm->vect2.y = 1;
    rotm->vect3.z = 1;
    rotm->vect2.x = 0;         
    rotm->vect1.y = 0;        
    rotm->vect3.x = 0;        
    rotm->vect1.z = 0;        
    rotm->vect3.y = 0;         
    rotm->vect2.z = 0;         
  } else {
    q12    = q1*q2;
    q13    = q1*q3;
    q23    = q2*q3;
    q10    = q0*q1;
    q20    = q0*q2;
    q30    = q0*q3;
    q00    = q0*q0;

    rotm->vect1.x = 2 * (q00 + q1*q1) - 1;
    rotm->vect2.y = 2 * (q00 + q2*q2) - 1;
    rotm->vect3.z = 2 * (q00 + q3*q3) - 1;
    rotm->vect2.x = 2 * (q12 - q30);
    rotm->vect1.y = 2 * (q12 + q30);
    rotm->vect3.x = 2 * (q13 + q20);
    rotm->vect1.z = 2 * (q13 - q20);
    rotm->vect3.y = 2 * (q23 - q10);
    rotm->vect2.z = 2 * (q23 + q10);
  }
}

/*
   Extract rotation vector from an orthonormal matrix (triad).
 */
void
W_extract_rot_vector_from_triad(W_triad_3d_t *rotm, W_coords_3d_t *rotv)
{
  int imax = 0;
  double rotvn, q0 = 0, q1 = 0, q2 = 0, q3 = 0; 
  double rmtrace, rm11, rm22, rm33, maxt, qq;

  rm11    = rotm->vect1.x;
  rm22    = rotm->vect2.y;
  rm33    = rotm->vect3.z;
  rmtrace = rm11 + rm22 + rm33;
  maxt    = rmtrace;
  if (rm11 > maxt) { maxt  = rm11; imax  = 1; }
  if (rm22 > maxt) { maxt  = rm22; imax  = 2; }
  if (rm33 > maxt) { maxt  = rm33; imax  = 3; }

  if        (imax == 0) {
    q0 = 0.5 * sqrt(1 + rmtrace);
    q1 = (rotm->vect2.z - rotm->vect3.y) * 0.25 / q0;
    q2 = (rotm->vect3.x - rotm->vect1.z) * 0.25 / q0;
    q3 = (rotm->vect1.y - rotm->vect2.x) * 0.25 / q0;
  } else if (imax == 1) {
    q1 = sqrt(0.5 * rm11 + 0.25 * (1 - rmtrace));
    q0 = (rotm->vect2.z - rotm->vect3.y) * 0.25 / q1;
    q2 = (rotm->vect2.x + rotm->vect1.y) * 0.25 / q1;
    q3 = (rotm->vect1.z + rotm->vect3.x) * 0.25 / q1;
    if (q0 < 0) { q0 = -q0; q1 = -q1; q2 = -q2; q3 = -q3; }
  } else if (imax == 2) {
    q2 = sqrt(0.5 * rm22 + 0.25 *(1 - rmtrace));
    q0 = (rotm->vect3.x - rotm->vect1.z) * 0.25 / q2;
    q1 = (rotm->vect2.x + rotm->vect1.y) * 0.25 / q2;
    q3 = (rotm->vect2.z + rotm->vect3.y) * 0.25 / q2;
    if (q0 < 0) { q0 = -q0; q1 = -q1; q2 = -q2; q3 = -q3; }
  } else if (imax == 3) {
    q3 = sqrt(0.5 * rm33 + 0.25 * (1 - rmtrace));
    q0 = (rotm->vect1.y - rotm->vect2.x) * 0.25 / q3;
    q2 = (rotm->vect2.z + rotm->vect3.y) * 0.25 / q3;
    q1 = (rotm->vect1.z + rotm->vect3.x) * 0.25 / q3;
    if (q0 < 0) { q0 = -q0; q1 = -q1; q2 = -q2; q3 = -q3; }
  }
  qq = sqrt(q1*q1 + q2*q2 + q3*q3);
  if (qq <= 0) {
    rotv->x = rotv->y = rotv->z = 0;      
  } else {
    if (qq < 0.5)
      rotvn = 2 * asin(qq); /* this is more precise for smaller angles */
    else 
      rotvn = 2 * acos(q0); /* this is more precise than sin for rotvn ~ PI */
    rotv->x = rotvn * q1 / qq;
    rotv->y = rotvn * q2 / qq;
    rotv->z = rotvn * q3 / qq;
  }
}

/*
   Extract triad from a rotation vector.
 */
void
W_extract_triad_from_rot_vector(W_coords_3d_t *rotv, W_triad_3d_t *rotm)
{
  double rotvn, q0, q1, q2, q3, rotvn2; 
  double q12, q13, q23, q00, q10, q20, q30;

  q1    = rotv->x;
  q2    = rotv->y;
  q3    = rotv->z;
  rotvn = sqrt(q1*q1 + q2*q2 + q3*q3);
  if (rotvn <= 0) {
    rotm->vect1.x = 1;
    rotm->vect2.y = 1;
    rotm->vect3.z = 1;
    rotm->vect2.x = 0;         
    rotm->vect1.y = 0;        
    rotm->vect3.x = 0;        
    rotm->vect1.z = 0;        
    rotm->vect3.y = 0;         
    rotm->vect2.z = 0;         
  } else {
    rotvn2 = rotvn / 2;
    q0     = cos(rotvn2);
    q1     = q1 / rotvn * sin(rotvn2);
    q2     = q2 / rotvn * sin(rotvn2);
    q3     = q3 / rotvn * sin(rotvn2);
    q12    = q1*q2;
    q13    = q1*q3;
    q23    = q2*q3;
    q10    = q0*q1;
    q20    = q0*q2;
    q30    = q0*q3;
    q00    = q0*q0;

    rotm->vect1.x = 2 * (q00 + q1*q1) - 1;
    rotm->vect2.y = 2 * (q00 + q2*q2) - 1;
    rotm->vect3.z = 2 * (q00 + q3*q3) - 1;
    rotm->vect2.x = 2 * (q12 - q30);
    rotm->vect1.y = 2 * (q12 + q30);
    rotm->vect3.x = 2 * (q13 + q20);
    rotm->vect1.z = 2 * (q13 - q20);
    rotm->vect3.y = 2 * (q23 - q10);
    rotm->vect2.z = 2 * (q23 + q10);
  }
}


#ifdef UNIX
/*
   Does file exist?
 */
W_bool_t /* TRUE if file exists */
W_file_exists (char *f)
{
  struct stat statbuf;

  return (stat(f, &statbuf) == 0);
}
#endif

 
/*
   Does file have a specified suffix?
 */
W_bool_t
W_has_suffix (char *f, char *s /* suffix */)
{
  int lf, ls, i;

  if (!f || !s)
    return FALSE;
  
  lf = strlen((f)); ls = min(strlen((s)), lf);
  for (i = 0; i < ls; i++)
    if (toupper((f)[lf-i]) != toupper((s)[ls-i]))
      return FALSE;
  return TRUE;
}




/*
   Get date and time as a string.
 */
void 
W_get_date_and_time (char *return_string /* needs to hold at least 64 characters */
                     )
{
  sprintf (return_string, "%s", ckit_get_date_and_time ());
}


  /* Compute value from a table. It is a table of pairs of values: 
   The first is the abscissa (e.g. time) and the second the ordinate 
   (e.g. the function value). The function W_eval_ordinate_at_abscissa 
   interpolates linearly for values between the abscissa nodal values. 
   Note, that the value of the table for abscissa < pairs[0].abscissa 
   is UNDEFINED. The value for abscissa < pairs[num_of_entries-1].abscissa 
 is given by the slope within the last interval by extrapolation. */

W_float_t
W_eval_ordinate_at_abscissa (W_index_t num_of_entries,
                             W_abscissa_ordinate_pair_t *pairs,
                             W_float_t abscissa)
{
  W_float_t dA, dO;
  W_index_t i, L = 0;

  assert(num_of_entries > 0 && pairs != NULL);
  
  if (num_of_entries <= 1)
    return pairs[0].ordinate;
  
  dA = pairs[1].abscissa - pairs[0].abscissa;
  dO = pairs[1].ordinate - pairs[0].ordinate;

  i = 1;
  while (i < num_of_entries) {
    L = i;
    if (abscissa <= pairs[i].abscissa)
      break;
    i++;
  }
  dA = pairs[L].abscissa - pairs[L-1].abscissa;
  dO = pairs[L].ordinate - pairs[L-1].ordinate;

  if (dA == 0)
    return (pairs[L].ordinate + pairs[L-1].ordinate) / 2;
  else
    return pairs[L-1].ordinate + (abscissa - pairs[L-1].abscissa) * dO / dA;
}


int
pv(char *name, W_float_t *v, int row, int col)
{
  int i;

  if (name != NULL)
    fprintf(stderr, "%s:\n", name);
  else
    fprintf(stderr, "Unknown:\n");
  for (i = 0; i < row*col; i++)
    fprintf(stderr, " %g%c", v[i], ((i+1)/col)*col == i+1? '\n': ';');
  fprintf(stderr, "\n");
  return 1;
}

#ifdef NEED_STAT_MALLOC
void *
W_stat_malloc_statistics(FILE *fp)
{
  memory_stats(fp, "Memory statistics:");
}
#endif

static long currently_alloced = 0;

caddr_t
logging_malloc(char *file, int line, char *type, int bytes)
{
  currently_alloced += bytes;
  fprintf(stderr, "M: %s (l.%d), %s (%d bytes), T: %ld\n",
          file, line, type, bytes, currently_alloced);
  return malloc(bytes);
}


caddr_t
logging_calloc(char *file, int line, char *type, int nelms, int bytes)
{
  currently_alloced += nelms*bytes;
  fprintf(stderr,
          "C: %s (l.%d), %s (%d bytes/type, %d nelms, %d bytes), T: %ld\n",
          file, line, type, bytes, nelms, bytes*nelms, currently_alloced);
  return calloc(nelms, bytes);
}



caddr_t
logging_realloc(char *file, int line, char *type, caddr_t ptr,
                int nelms, int bytes)
{
  currently_alloced += nelms*bytes;
  fprintf(stderr,
          "R: %s (l.%d), %s (%d bytes/type, %d nelms, %d bytes), T: %ld\n",
          file, line, type, bytes, nelms, bytes*nelms, currently_alloced);
  return realloc(ptr, nelms*bytes);
}


/*
   Get real, user and system times.
 */
W_bool_t /* TRUE if arguments hold valid values */
W_get_times (double *real_time,
             double *usr_time,
             double *systm_time) /* Returns the real, user and system
                                      time for the process */
{
  return ckit_get_times (real_time, usr_time, systm_time);
}


/*
   Return pointer to a character buffer which can be used for printing.
   The buffer holds at least 1024 characters.
 */
char *
W_buf(void)
{
  static char buf[2048];
  return buf;
}

/*
   Generate name of a post-processing file.
 */
char *
W_postfile_name (double time, char *basename, int type, W_bool_t number_seq)
{
  char *ext;
  static int postfile_number = 0;

  if      (type == HEXVUE_POSTPROC_FILE)      ext = "hvu";
  else if (type == VE3D_POSTPROC_FILE)        ext = "ve3d";
  else if (type == TECPLOT_I_POSTPROC_FILE)   ext = "tpl";
  else if (type == HEXVUE_XTRA_POSTPROC_FILE) ext = "xvu";
  else                                        ext = "post";

  if (number_seq)
    sprintf(W_buf(), "%s%d.%s", basename, postfile_number, ext);
  else
    sprintf(W_buf(), "%s%g.%s", basename, time, ext);
  postfile_number++;
  return W_buf();
}




double
W_f_quartic_spline (double r)
{
  double absr = fabs(r);
  /* Quartic spline function */
  if (absr >= 1)
    return 0;
  else {
    double tmp;
    tmp = (1 + ((-6 + (8 - 3 * absr) * absr) * absr) * absr);
    return tmp > 0? tmp: 0;     /* catch round-off errors */
  }
}


double 
W_f_r_quartic_spline (double r)
{
  double absr = fabs(r);
  /* Quartic spline function derivative */
  if (absr >= 1) {
    return 0;
  } else {
    if (r > 0)
      return -12 * absr * (1 + absr * (-2 + absr));
    else
      return  12 * absr * (1 + absr * (-2 + absr));
  }
}



double
W_f_quadratic (double r)
{
  double absr = fabs(r);
  /* Quadratic function */
  if (absr >= 1)
    return 0;
  else {
    return (1 - absr * absr);
  }
}


double 
W_f_r_quadratic (double r)
{
  double absr = fabs(r);
  /* Quadratic function derivative */
  if (absr >= 1) {
    return 0;
  } else {
    return -(r + r);
  }
}


double
W_f_linear (double r)
{
  double absr = fabs(r);
  /* Quadratic function */
  if (absr >= 1)
    return 0;
  else {
    return (1 - absr);
  }
}


double 
W_f_r_linear (double r)
{
  double absr = fabs(r);
  /* Quadratic function derivative */
  if (absr >= 1) {
    return 0;
  } else {
    if (r < 0) return +1;
    else       return -1;
  }
}


double 
W_f_composite_quadratic (double x)
{
  double absx = fabs(x);
  if (absx >= 1)
    return 0;
  else {
    if (absx > 0.5) {
      return 2 * (1 - absx) * (1 - absx);
    } else {
      return  1 - 2 * absx * absx;
    }
  }
}

double 
W_f_r_composite_quadratic (double x)
{
  if (x >= 1 || x <= -1)
    return 0;
  else {
    if (x > 0.5) {
      return 4 * (x - 1);
    } else if (x < -0.5) {
      return 4 * (x + 1);
    } else {
      return -4 * x;
    }
  }
}


#define RUN_IOP(action)                                         \
   {                                                            \
     if (!action) {                                             \
       REPORT_ERROR (NULL, (char *)ckit_io_error_explanation ());     \
     }                                                          \
   }

static int
write_hexvue_hex_binary (ckit_iobuf_t IOB,
                         G3D_loc_t *p[8],
                         int num_of_values,
                         double *values[8]
                         )
{
  int j, i, n = num_of_values;
  int reqlen = n;
  float geom[3];
  static int buflen        = 0;
  static float *minv       = NULL;
  static float *maxv       = NULL;
  static unsigned char *lo = NULL;
  static unsigned char *hi = NULL;

  if (buflen < reqlen) {
    minv   = NEWP_REALLOC_OBJ_ARRAY (minv, reqlen, float);
    maxv   = NEWP_REALLOC_OBJ_ARRAY (maxv, reqlen, float);
    lo     = NEWP_REALLOC_OBJ_ARRAY (lo, reqlen, unsigned char);
    hi     = NEWP_REALLOC_OBJ_ARRAY (hi, reqlen, unsigned char);
    buflen = reqlen;
  }

  FOR (i, n) {
    minv[i] = FLT_MAX;
    maxv[i] = -FLT_MAX;
    FOR (j, 8) {
      minv[i] = min (minv[i], values[j][i]);
      maxv[i] = max (maxv[i], values[j][i]);
    }
  }
  
  RUN_IOP (ckit_io_put_line (IOB, "BHV"));
  RUN_IOP (ckit_io_put_data (IOB, minv, n * sizeof (float)));
  RUN_IOP (ckit_io_put_data (IOB, maxv, n * sizeof (float)));
  FOR (j, 8) {
    geom[0] = p[j]->x;
    geom[1] = p[j]->y;
    geom[2] = p[j]->z;
    RUN_IOP (ckit_io_put_data (IOB, geom, 3 * sizeof (float)));
    FOR (i, n) {
      ckit_io_convert_double_to_bytes (values[j][i],
                                       minv[i], maxv[i],
                                       &lo[i], &hi[i]);
    }      
    RUN_IOP (ckit_io_put_data (IOB, lo, n * sizeof (unsigned char)));
    RUN_IOP (ckit_io_put_data (IOB, hi, n * sizeof (unsigned char)));
  }

  return 1;
}


static int
write_hexvue_hex_compressed (ckit_iobuf_t IOB,
                             G3D_loc_t *p[8],
                             int num_of_values,
                             double *values[8]
                             )
{
  int j, i, n = num_of_values, k;
  int reqlen = n;
  float flbox[6];
  unsigned char geombytes[24];
  static int buflen        = 0;
  static float *minv       = NULL;
  static float *maxv       = NULL;
  static unsigned char *byte = NULL;
  G3D_box_t box;

  if (buflen < reqlen) {
    minv   = NEWP_REALLOC_OBJ_ARRAY (minv, reqlen, float);
    maxv   = NEWP_REALLOC_OBJ_ARRAY (maxv, reqlen, float);
    byte   = NEWP_REALLOC_OBJ_ARRAY (byte, reqlen, unsigned char);
    buflen = reqlen;
  }

  FOR (i, n) {
    minv[i] = FLT_MAX;
    maxv[i] = -FLT_MAX;
    FOR (j, 8) {
      minv[i] = min (minv[i], values[j][i]);
      maxv[i] = max (maxv[i], values[j][i]);
    }
  }

  G3D_INIT_BOX (box, (*p[0]));
  G3D_UPDT_BOX (box, (*p[1]));
  G3D_UPDT_BOX (box, (*p[2]));
  G3D_UPDT_BOX (box, (*p[3]));
  G3D_UPDT_BOX (box, (*p[4]));
  G3D_UPDT_BOX (box, (*p[5]));
  G3D_UPDT_BOX (box, (*p[6]));
  G3D_UPDT_BOX (box, (*p[7]));
  flbox[0] = box.lower.x;
  flbox[1] = box.lower.y;
  flbox[2] = box.lower.z;
  flbox[3] = box.upper.x;
  flbox[4] = box.upper.y;
  flbox[5] = box.upper.z;
  
  RUN_IOP (ckit_io_put_line (IOB, "CHV"));
  RUN_IOP (ckit_io_put_data (IOB, minv,  n * sizeof (float)));
  RUN_IOP (ckit_io_put_data (IOB, maxv,  n * sizeof (float)));
  RUN_IOP (ckit_io_put_data (IOB, flbox, 6 * sizeof (float)));
  k = 0;
  FOR (j, 8) {
    ckit_io_convert_double_to_byte (p[j]->x, box.lower.x, box.upper.x,
                                    &geombytes[k]); k++;
    ckit_io_convert_double_to_byte (p[j]->y, box.lower.y, box.upper.y,
                                    &geombytes[k]); k++;
    ckit_io_convert_double_to_byte (p[j]->z, box.lower.z, box.upper.z,
                                    &geombytes[k]); k++;
  }
  RUN_IOP (ckit_io_put_data (IOB, geombytes, 24));
  FOR (j, 8) {
    FOR (i, n) {
      ckit_io_convert_double_to_byte (values[j][i], minv[i], maxv[i],
                                      &byte[i]);
    }      
    RUN_IOP (ckit_io_put_data (IOB, byte, n * sizeof (unsigned char)));
  }

  return 1;
}




static int
write_hexvue_hex_ascii (ckit_iobuf_t IOB,
                        G3D_loc_t *p[8],
                        int num_of_values,
                        double *values[8]
                        )
{
  int j, i, n = num_of_values;
  char buf[128];
  
  RUN_IOP (ckit_io_put_line (IOB, "HV"));
  FOR (j, 8) {
    sprintf (buf, "%g %g %g", p[j]->x, p[j]->y, p[j]->z);
    RUN_IOP (ckit_io_put_line (IOB, buf));
    FOR (i, n) {
      sprintf (buf, "%g", values[j][i]);
      RUN_IOP (ckit_io_put_line (IOB, buf));
    }      
  }
 
  return 1;
}


/*
   Write HEXVUE-style record for post-processing for a single hexahedron.
 */
int
W_write_hexvue_hex (ckit_iobuf_t IOB,
                    G3D_loc_t *p[8],
                    int num_of_values,
                    double *values[8],
                    char *mode
                    )
{
  if (mode) {
    if        (strcmp (mode, "binary") == 0) {
      /* binary */
      return write_hexvue_hex_binary (IOB, p, num_of_values, values);
    } else if (strcmp (mode, "compressed") == 0) {
      /* compressed */
      return write_hexvue_hex_compressed (IOB, p, num_of_values, values);
    } else {
      /* ascii */
      return write_hexvue_hex_ascii (IOB, p, num_of_values, values);
    }
  } else {
    return write_hexvue_hex_compressed (IOB, p, num_of_values, values);
  }
}


static int
write_hexvue_tri3_ascii (ckit_iobuf_t IOB,
                        G3D_loc_t *p[3],
                        int num_of_values,
                        double *values[3]
                        )
{
  int j, i, n = num_of_values;
  char buf[128];
  
  RUN_IOP (ckit_io_put_line (IOB, "T3"));
  FOR (j, 3) {
    sprintf (buf, "%g %g %g", p[j]->x, p[j]->y, p[j]->z);
    RUN_IOP (ckit_io_put_line (IOB, buf));
    FOR (i, n) {
      sprintf (buf, "%g", values[j][i]);
      RUN_IOP (ckit_io_put_line (IOB, buf));
    }      
  }
 
  return 1;
}

static int
write_hexvue_tri3_binary (ckit_iobuf_t IOB,
                          G3D_loc_t *p[3],
                          int num_of_values,
                          double *values[3]
                          )
{
  int j, i, n = num_of_values;
  int reqlen = n;
  float geom[3];
  static int buflen        = 0;
  static float *minv       = NULL;
  static float *maxv       = NULL;
  static unsigned char *lo = NULL;
  static unsigned char *hi = NULL;

  if (buflen < reqlen) {
    minv   = NEWP_REALLOC_OBJ_ARRAY (minv, reqlen, float);
    maxv   = NEWP_REALLOC_OBJ_ARRAY (maxv, reqlen, float);
    lo     = NEWP_REALLOC_OBJ_ARRAY (lo, reqlen, unsigned char);
    hi     = NEWP_REALLOC_OBJ_ARRAY (hi, reqlen, unsigned char);
    buflen = reqlen;
  }

  FOR (i, n) {
    minv[i] = FLT_MAX;
    maxv[i] = -FLT_MAX;
    FOR (j, 3) {
      minv[i] = min (minv[i], values[j][i]);
      maxv[i] = max (maxv[i], values[j][i]);
    }
  }
  
  RUN_IOP (ckit_io_put_line (IOB, "BT3"));
  RUN_IOP (ckit_io_put_data (IOB, minv, n * sizeof (float)));
  RUN_IOP (ckit_io_put_data (IOB, maxv, n * sizeof (float)));
  FOR (j, 3) {
    geom[0] = p[j]->x;
    geom[1] = p[j]->y;
    geom[2] = p[j]->z;
    RUN_IOP (ckit_io_put_data (IOB, geom, 3 * sizeof (float)));
    FOR (i, n) {
      ckit_io_convert_double_to_bytes (values[j][i],
                                       minv[i], maxv[i],
                                       &lo[i], &hi[i]);
    }      
    RUN_IOP (ckit_io_put_data (IOB, lo, n * sizeof (unsigned char)));
    RUN_IOP (ckit_io_put_data (IOB, hi, n * sizeof (unsigned char)));
  }

  return 1;
}

/*
   Write HEXVUE-style record for post-processing for a single tri3 triangle.
 */
int
W_write_hexvue_tri3 (ckit_iobuf_t IOB,
                     G3D_loc_t *p[3],
                     int num_of_values,
                     double *values[3],
                     char *mode
                     )
{
  if (mode) {
    if        (strcmp (mode, "binary") == 0) {
      /* binary */
      return write_hexvue_tri3_binary (IOB, p, num_of_values, values);
    } else {
      return write_hexvue_tri3_ascii (IOB, p, num_of_values, values);
    }
  } else {
    return write_hexvue_tri3_ascii (IOB, p, num_of_values, values);
  }
}

static int
write_hexvue_tet4_ascii (ckit_iobuf_t IOB,
                        G3D_loc_t *p[4],
                        int num_of_values,
                        double *values[4]
                        )
{
  int j, i, n = num_of_values;
  char buf[128];
  
  RUN_IOP (ckit_io_put_line (IOB, "T4"));
  FOR (j, 4) {
    sprintf (buf, "%g %g %g", p[j]->x, p[j]->y, p[j]->z);
    RUN_IOP (ckit_io_put_line (IOB, buf));
    FOR (i, n) {
      sprintf (buf, "%g", values[j][i]);
      RUN_IOP (ckit_io_put_line (IOB, buf));
    }      
  }
 
  return 1;
}

static int
write_hexvue_tet4_binary (ckit_iobuf_t IOB,
                          G3D_loc_t *p[4],
                          int num_of_values,
                          double *values[4]
                          )
{
  int j, i, n = num_of_values;
  int reqlen = n;
  float geom[4];
  static int buflen        = 0;
  static float *minv       = NULL;
  static float *maxv       = NULL;
  static unsigned char *lo = NULL;
  static unsigned char *hi = NULL;

  if (buflen < reqlen) {
    minv   = NEWP_REALLOC_OBJ_ARRAY (minv, reqlen, float);
    maxv   = NEWP_REALLOC_OBJ_ARRAY (maxv, reqlen, float);
    lo     = NEWP_REALLOC_OBJ_ARRAY (lo, reqlen, unsigned char);
    hi     = NEWP_REALLOC_OBJ_ARRAY (hi, reqlen, unsigned char);
    buflen = reqlen;
  }

  FOR (i, n) {
    minv[i] = FLT_MAX;
    maxv[i] = -FLT_MAX;
    FOR (j, 4) {
      minv[i] = min (minv[i], values[j][i]);
      maxv[i] = max (maxv[i], values[j][i]);
    }
  }
  
  RUN_IOP (ckit_io_put_line (IOB, "BT4"));
  RUN_IOP (ckit_io_put_data (IOB, minv, n * sizeof (float)));
  RUN_IOP (ckit_io_put_data (IOB, maxv, n * sizeof (float)));
  FOR (j, 4) {
    geom[0] = p[j]->x;
    geom[1] = p[j]->y;
    geom[2] = p[j]->z;
    RUN_IOP (ckit_io_put_data (IOB, geom, 4 * sizeof (float)));
    FOR (i, n) {
      ckit_io_convert_double_to_bytes (values[j][i],
                                       minv[i], maxv[i],
                                       &lo[i], &hi[i]);
    }      
    RUN_IOP (ckit_io_put_data (IOB, lo, n * sizeof (unsigned char)));
    RUN_IOP (ckit_io_put_data (IOB, hi, n * sizeof (unsigned char)));
  }

  return 1;
}

/*
   Write HEXVUE-style record for post-processing for a single tet4 tetrahedron.
 */
int
W_write_hexvue_tet4 (ckit_iobuf_t IOB,
                     G3D_loc_t *p[4],
                     int num_of_values,
                     double *values[4],
                     char *mode
                     )
{
  if (mode) {
    if        (strcmp (mode, "binary") == 0) {
      /* binary */
      return write_hexvue_tet4_binary (IOB, p, num_of_values, values);
    } else {
      return write_hexvue_tet4_ascii (IOB, p, num_of_values, values);
    }
  } else {
    return write_hexvue_tet4_ascii (IOB, p, num_of_values, values);
  }
}



/*
   Compute the angle of fracture propagation.
   Compute the angle of fracture propagation using the maximum
   circumferential stress criterion.
   The coordinate system is as follows:
   e1 ... normal to the crack front, in crack plane
   e2 ... normal to both the crack front and the crack plane
   e3 ... tangent to the crack front

   K_II positive should yields negative angle, K_II negative yields
   a positive angle; in all cases the absolute value of theta is < PI/2.
 */
W_bool_t
W_compute_theta (double K_I, double K_II, double *theta)
{
  double K_I_over_K_II, det, sqrt_det, tp, tm, vp, vm;

  *theta = 0;
  if (K_II != 0) {
    K_I_over_K_II = K_I / K_II;
    if (fabs (K_I_over_K_II) > 10000) {
      return TRUE; /* K_II negligible */
    }
    det           = K_I_over_K_II * K_I_over_K_II + 8;
    sqrt_det      = sqrt (det);
    vp            = 0.25 * (K_I_over_K_II + sqrt_det);
    vm            = 0.25 * (K_I_over_K_II - sqrt_det);
    tp            = 2 * atan2 (vp, 1);
    if        (tp > M_PI_2) {
      tp -= M_PI;
    } else if (tp < -M_PI_2) {
      tp += M_PI;
    }
    tm            = 2 * atan2 (vm, 1);
    if        (tm > M_PI_2) {
      tm -= M_PI;
    } else if (tm < -M_PI_2) {
      tm += M_PI;
    }
  retry:
    if (K_II > 0) {
      if (tp < 0)
        if (tm >= 0)
          *theta = tp;
        else
          if (tp <= tm)
            *theta = tm;
          else
            *theta = tp;
      else
        if (tm <= 0)
          *theta = tm;
        else {
          tm = -tm;
          tp = -tp;
          goto retry;
        }
    } else {
      if (tp > 0)
        if (tm <= 0)
          *theta = tp;
        else
          if (tp >= tm)
            *theta = tm;
          else
            *theta = tp;
      else
        if (tm >= 0)
          *theta = tm;
        else {
          tm = -tm;
          tp = -tp;
          goto retry;
        }
    }
  } else {
    *theta = 0;
  }
  return TRUE;
}



static double
D_of_C (double C_d, double C_s, double C)
{
  double beta1, beta2;
  beta1 = sqrt (1 - C*C / (C_d*C_d));
  beta2 = sqrt (1 - C*C / (C_s*C_s));
  return 4 * beta1 * beta2 - (1 - beta2 * beta2) * (1 - beta2 * beta2);
}

/*
   Compute the Rayleigh wave speed.
 */
int
W_compute_Rayleigh_speed (double C_d, double C_s,
                          double percent_error,
                          double *C_R)
{
  double hi     = C_s;
  double lo     = 0;
  double cr     = C_s;
  double v, vhi, vlo;

  vhi = D_of_C (C_d, C_s, hi);
  vlo = D_of_C (C_d, C_s, lo);
  if (vhi * vlo > 0) {
    *C_R = -FLT_MAX;
    return 0;
  }
  
  while ((hi - lo) / 2 > percent_error * cr) {
    cr = (hi + lo) / 2;
    v  = D_of_C (C_d, C_s, cr);
    if (v > 0) {
      if (vhi > 0)
        hi = cr;
      else
        lo = cr;
    } else {
      if (vlo < 0)
        lo = cr;
      else
        hi = cr;
    }
    vhi    = D_of_C (C_d, C_s, hi);
    vlo    = D_of_C (C_d, C_s, lo);
  }
  *C_R = (hi+lo)/2;
  return 1;
}


/*
   Compute the dilatational wave speed.
 */
double
W_3D_dilatational_wave_speed (double E, double nu, double rho)
{
  return sqrt(E * (1 - nu) / (1 + nu) / (1 - 2*nu) / rho);
}


/*
   Compute the plane stress dilatational wave speed.
 */
double
W_plstress_dilatational_wave_speed (double E, double nu, double rho)
{
  double kappa = (3 - nu) / (1 + nu);
  double mu    = E / 2 / (1 + nu);
  return sqrt((kappa + 1) / (kappa - 1) * mu / rho);
}

/*
   Compute the plane strain dilatational wave speed.
 */
double
W_plstrain_dilatational_wave_speed (double E, double nu, double rho)
{
  double kappa = (3 - 4 * nu);
  double mu    = E / 2 / (1 + nu);
  return sqrt((kappa + 1) / (kappa - 1) * mu / rho);
}



/*
   Compute the shear wave speed.
 */
double
W_shear_wave_speed (double E, double nu, double rho)
{
  double mu    = E / 2 / (1 + nu);
  return sqrt(mu / rho);
}


/*
   Compute equivalent opening SIF.
 */
double
W_equiv_KI (double KI, /* opening SIF */
          double KII, /* shearing SIF */
          double theta /* angle in the forward/normal plane */
          ) 
{
  double th  = theta/2;
  double cth = cos (th);
  double st  = sin (theta);
  return KI * cth * cth * cth - 1.5 * KII * cth * st;
} 


W_bool_t
W_is_end_of_block (const char *s)
{
  int i = 0;
  while (s[i] != '\0') {
    if (s[i] == END_OF_BLOCK)
      return TRUE;
    else if (isspace (s[i]))
      i++;
    else
      return FALSE;
  }
  return FALSE;
}

/*
   Read and forget an input block (essentially skips the entire block).
 */
W_bool_t
W_read_and_discard_block (W_domain_t *d)
{
  const char *input_line;
  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return TRUE;
  }
  return FALSE;
}

#if TIMED_WASP
double
W_timer_get_time (W_timer_t *timer)
{
  if (timer->num_calls > 0)
    return timer->cummulative_time / timer->num_calls;
  else
    return 0;
}
#endif

#if TIMED_WASP
W_timer_t *
W_timer_new (void)
{
  W_timer_t *timer = w_new (W_timer_t);
  CHECK_MALLOC (timer);
  timer->num_calls = 0;
  timer->cummulative_time = 0;
  timer->timer = ckit_timer_new ();
  CHECK_MALLOC (timer->timer);
  return timer;
}
#endif

#if TIMED_WASP
void
W_timer_restart (W_timer_t *timer)
{
  timer->num_calls++;
  ckit_timer_restart (timer->timer);
}
#endif

#if TIMED_WASP
void
W_timer_accumulate (W_timer_t *timer)
{
  timer->cummulative_time += ckit_timer_current (timer->timer);
}
#endif

W_dict_t *
W_dict_new_dictionary (void)
{
  W_dict_t *dict = w_new (W_dict_t); CHECK_MALLOC (dict);
  dict->d     = dict_create_dictionary (); CHECK_MALLOC (dict->d);
  dict->diter = dict_get_iterator (); CHECK_MALLOC (dict->diter);
  return dict;
}

/*
  Iterate objects in dictionary.
  Iterate objects in WASP dictionary.  Pass current object as curr.
  W_dict_t dictionary uses dict_dictionary dict_iterator from Ckit.
 */
void * /* next object or NULL if at end */
W_dict_next_obj (W_dict_t *dict, /* dictionary */
                 void *curr /* current object or NULL if starting */
                 )
{
  if (dict->diter == NULL) {
    dict->diter = dict_get_iterator (); CHECK_MALLOC (dict->diter);
  } 
  if (curr == NULL) {
    dict_reset_iterator (dict->diter);
  }
  return dict_iter_objs (dict->d, dict->diter);
}

/*
  Find object in dictionary.
  Find object in WASP dictionary by index (positive integer).
 */
void * /* object or NULL if not found */
W_dict_find_by_idx (W_dict_t *dict, /* dictionary */
                    W_index_t idx /* index */
                    )
{
  return dict_find_obj_with_int_name (dict->d, idx);
}

void
W_dict_add (W_dict_t *dict, /* dictionary */
            void *obj, /* object */
            W_index_t idx /* index */
            )
{
  dict_add_obj_int_name (dict->d, obj, idx);
}

int 
W_dict_count(W_dict_t *dict)
{
  return dict_count(dict->d);
}
