#include "Wconfig.h"

#if defined(PVM_WASP_WORKER) 

#include "Wxf.h"
#include "Wdomain.h"
#include "Whistry.h"
#include "Welem.h"
#include "Wtfuncs.h"
#include "Wnode.h"
#include "Wloads.h"
#include "Wmats.h"
#include "Wvset.h"
#include "Wpvm.h"
#include "EID.h"

/* RCSHEADER "$Id: worker.c,v 1.1.1.1 2001/07/24 02:42:52 pkrysl Exp $" */

static   char pbuf[132];

W_float_t
WD_compute_dt(W_domain_t *d)
{
  int nrecv = 0;
  double dt = FLT_MAX;

  RECEIVE(MASTER_TID(d->pvm_worker), ANY_TAG);
  BUFINFO();
  if (BUFINFO_TAG() == TIME_STEP_REQUEST_TAG) {
    VL(num_of_elements, dt = min(dt, WE_suggested_time_step(d->elements[IX])));
    INITSEND();
    PACK(double, &dt);
    SEND(MASTER_TID(d->pvm_worker), TIME_STEP_ANSWER_TAG);
    RECEIVE(MASTER_TID(d->pvm_worker), TIME_STEP_ANSWER_TAG);
    UNPACK(double, &dt);
  } else if (BUFINFO_TAG() == TIME_STEP_ANSWER_TAG) {
    UNPACK(double, &dt);
  }
    
  if (options.verbose) {
    sprintf(pbuf, "Time step = %g", (double)dt);
    W_update_log(pbuf);
  }
  return dt;
}

void
assemble_mass_matrix(W_domain_t *d);
static void
send_interface_loads(W_domain_t *d);
static void
receive_interface_loads(W_domain_t *d);

void
WD_calc_eff_loads(W_domain_t *d)
{
  assert(d->effective_loads != NULL);
  if (d->mass_matrix == NULL)
    assemble_mass_matrix(d);

  /* Compute them as if the domain was isolated */
  VL(num_of_equations, d->effective_loads[IX] = 0);
  VL(num_of_loads, WL_assemble_load(d->loads[IX], d));
  REG_ELEM_LOOP(d, WE_calc_eff_loads(theELEM));
  VL(num_of_elements, WE_assemble_restoring_forces(d->elements[IX]));
  VL(num_of_nodes, WN_assemble_ang_vel_dep_load(d->nodes[IX], d));
  /* Adjust interface loads */
  send_interface_loads(d);
  receive_interface_loads(d);
}

static void
send_interface_loads(W_domain_t *d)
{
  W_pvm_worker_t *w = d->pvm_worker;
  unsigned long nin = w->num_of_interface_nodes;
  W_local_to_interface_map_t *map = w->map;
  double forces[6];
  int i;
  
  INITSEND();
  PACK(ulong, &nin);
  for (i = 1; i <= nin; i++) {
    PACK(ulong, &map[i].interface_num);
    WN_pack_interface_vector(WD_get_node_by_idx(d, map[i].local_num),
                             d->effective_loads, forces);
    PACK_ARRAY(double, forces, 6);
  }
  SEND(MASTER_TID(d->pvm_worker), EFFLOAD_LOCAL_TAG);
}


static unsigned long num_of_interface_packages = 0;
static W_interface_package_t *interface_loads = NULL;

static void
receive_interface_loads(W_domain_t *d)
{
  unsigned long nin;
  int i;
  W_pvm_worker_t *w = d->pvm_worker;
  W_local_to_interface_map_t *map = w->map;
  
  RECEIVE(MASTER_TID(w), EFFLOAD_GLOBAL_TAG);
  UNPACK(ulong, &nin);
  if (num_of_interface_packages < nin) {
    num_of_interface_packages = nin;
    interface_loads
      = w_new_array(num_of_interface_packages+1, W_interface_package_t);
    CHECK_MALLOC(interface_loads);
  }
  if (nin != num_of_interface_packages)
    REPORT_FATAL_ERROR("Change in interface connectivity");
  for (i = 1; i <= num_of_interface_packages; i++) {
    UNPACK_ARRAY(double, interface_loads[i].forces, 6);
  }
  for (i = 1; i <= w->num_of_interface_nodes; i++) {
    WN_unpack_interface_vector(WD_get_node_by_idx(d, map[i].local_num),
                               interface_loads[map[i].interface_num].forces,
                               d->effective_loads);
  }
}

#endif /* if defined(PVM_WASP_WORKER) */
