#include "Wconfig.h"

#ifdef EFG_WASP

#include "Wtypes.h"
#include "WEFG.h"
#include "WEFGsf.h"
#include "Wdomain.h"
#include "Wgaussp.h"
#include "Wxf.h"
#include "Wutil.h"
#include "Wcdi.h"
#include "WEFGutil.h"
#include "WEFGpriv.h"
#include "WEFGcr.h"
#include "WEFGtc.h"
#include "WEFGii.h"
#include "Wmat3d.h"
#include "Wnode.h"
#include "WEFGsf.h"
#include "WEFGss.h"
#include "Whistry.h"
#include "Werrh.h"

/* RCSHEADER "$Id: efgtc.c,v 1.6 2003/11/25 16:55:48 cvs Exp $" */

#define MAX_RELAXATION_CYCLES 1

/* Triangulated crack surface */

struct ts_edge_t;
struct ts_tri_t;

struct triangulated_crack_surf_t;

typedef struct ts_crack_tip_history_client_data_t  {
  W_index_t                          pntidx;
  struct triangulated_crack_surf_t  *tc;
  struct ts_vert_t                  *orig_vert;
  struct ts_vert_t                  *vert;
  W_index_t                          buffidx;
  struct  {
    W_float_t                time;
    W_vect_t                 displ_t;
    W_velocity_3d_t          velocity_t;
    W_acceleration_3d_t      acceleration_t;
    double                   stress[6];
    double                   pressure;
    double                   equiv_pl_def;
    double                   K_I_0; /* SIF for a stationary crack */
    double                   K_II_0; /* SIF for a stationary crack */
    double                   K_III_0; /* SIF for a stationary crack */
    double                   K_I_c; /* SIF for a moving crack */
    double                   K_II_c; /* SIF for a moving crack */
    double                   K_III_c; /* SIF for a moving crack */
    double                   K_I_equiv;
    double                   theta; /* propagation angle in local c.s. */
    double                   G_0; /* stationary energy release rate */
    double                   G_c;/* non-stationary energy release rate */
    double                   C; /* crack speed */
  }                                 buffer[MAXHISTBUFFLNGTH];
}               ts_crack_tip_history_client_data_t;



#ifdef X_XFACE
#define DRAW_VERT_NUMS 0
typedef struct ts_vert_grep_t  {
  GraphicObj *marker;
#if DRAW_VERT_NUMS
  GraphicObj *text;
#endif
}               ts_vert_grep_t;
#endif

typedef struct ts_vert_t  {
  int                                  id;
  W_loc_t                              loc;
  W_loc_t                              before_loc;
  int                                  is_on_front;
  W_vect_t                             advance_vector;
  double                               K_I_0;
  double                               K_II_0;
  double                               K_III_0;
  double                               K_I_c;
  double                               K_II_c;
  double                               K_III_c;
  double                               K_I_equiv;
  double                               theta;
  double                               G_0;
  double                               G_c;
  double                               C;
  struct ts_vert_t                    *source_vert;
  W_vect_t                             normal;
  W_vect_t                             tangent;
  W_vect_t                             forward;
  int                                  is_on_boundary;
  WEFG_b_face_t                       *on_face;
  LIST_OF (ts_edge_t) edges;
  WEFG_sf_tab_t                        sftab_pos_normal;
  WEFG_sf_tab_t                        sftab_neg_normal;
  W_mat_3d_state_p                     ms_pos_normal;
  W_mat_3d_state_p                     ms_neg_normal;
  WEFG_hex_volume_t                   *enclosing_hex;
  W_loc_t                              ploc;
  ts_crack_tip_history_client_data_t  *hist_cd;
#ifdef X_XFACE
  ts_vert_grep_t                       grep_pos_normal;
  ts_vert_grep_t                       grep_neg_normal;
#endif
}  ts_vert_t;



struct ts_tri_t;

/* Vertices are traversed ccw */
typedef struct ts_edge_t  {
  int                            id;
  ts_vert_t                     *vertex[2];
  struct ts_tri_t               *triangles[2];
#ifdef X_XFACE
  GraphicObj                    *line_pos_normal;
  GraphicObj                    *line_neg_normal;
#endif
}  ts_edge_t;

#define E_IS_ON_FRONT(e) \
 (( (e) ->vertex[0]->is_on_front) && ((e) ->vertex[1]->is_on_front)) 

#define THE_OTHER_VERTEX(e, v) \
 ((e) ->vertex[0] == v? (e) ->vertex[1]: (e) ->vertex[0]) 

#define THE_OTHER_TRI(e, t) \
 ((e) ->triangles[0] == t? (e) ->triangles[1]: (e) ->triangles[0]) 



/* Edges are listed ccw */
typedef struct ts_tri_t  {
  int                            id;
  ts_edge_t                     *edges[3];/* these are just cached */
  ts_vert_t                     *vertex[3]; 
  G3D_box_t                      bbox;
  W_vect_t                       normal; /* cached */
  W_bool_t                       unified_normal;
#ifdef X_XFACE
  GraphicObj                    *triangle_pos_normal;
  GraphicObj                    *triangle_neg_normal;
#endif
}  ts_tri_t;


typedef enum ts_from_strength_stress_type_enum  {
  AVERAGED_STRESS = 0,
  LOCAL_STRESS    = 1
}             ts_from_strength_stress_type_enum;


typedef struct triangulated_crack_surf_t  {
  WEFG_crack_surf_t                    *crack_surf;
  W_bool_t                              is_stationary;
  LIST_OF (ts_vert_t *) vertices;
  LIST_OF (ts_vert_t *) on_front_vertices;
  LIST_OF (ts_edge_t *) edges;
  LIST_OF (ts_tri_t *) triangles;
  ts_from_strength_stress_type_enum     from_strength_stress_type;
  double                                edge_length_lower_bound;
  double                                edge_length_upper_bound;
  double                                min_edge_length;
  double                                max_edge_length;
  double                                rayleigh_vel_frac;
  W_bool_t                              allow_relaxation;
  W_bool_t                              allow_adaptivity;
  W_bool_t                              allow_E_star;
  double                                jdom_hsize_forward_plus;
  double                                jdom_hsize_forward_minus;
  double                                jdom_hsize_tangent_plus;
  double                                jdom_hsize_tangent_minus;
  double                                jdom_hsize_normal_plus;
  double                                jdom_hsize_normal_minus;
  bbt_front_end_t                      *bbt;
  W_vect_t                              average_normal;
  W_bool_t                              advance_prescribed;
  char                                 *advance_velocity[3];
  W_bool_t                              ignore_surface_j_contrib;
  double                                dist_to_front_for_displ_ratio;
  int                                   subsplit_for_2x2x2_ipol;
  W_bool_t                              extrapolate_at_the_bdry;
  double                                trim_advv_mult;
  double                                tilt_damping;
  ts_vert_t                            *master_vert;
  W_bool_t                              grow_in_plane;
  G3D_plane_t                           motion_plane;
  W_bool_t                              mixed_eval_at_the_bdry;
  W_bool_t                              drive_by_vel_incr;
  double                                vel_incr_fraction;
}               triangulated_crack_surf_t;


static void
check_defs (triangulated_crack_surf_t *tc);


static W_bool_t 
segment_intersects_crack_surf (WEFG_eval_for_packet_t *efp,
                              G3D_segment_t *s, WEFG_node_t *efgn,
                              WEFG_crack_surf_t *cs,
                              double  tol);
static int 
wrap_around_point (WEFG_crack_surf_t *cs,
                  double atx, double aty, double atz ,
                  double rcx, double rcy, double rcz ,
                  double bx, double by, double bz ,
                  double *wapx, double *wapy, double *wapz);
static int
plane_generating_vector (WEFG_crack_surf_t *cs,
                        double atx, double aty, double atz ,
                        double rcx, double rcy, double rcz ,
                        double *bx, double *by, double *bz );
static WEFG_pt_pos_wrt_crack_surf_enum
node_pos_wrt_crack_surf (WEFG_crack_surf_t *cs,
                        WEFG_node_t *efgn,
                        double physical_tol);
#ifdef X_XFACE
static void 
modify_crack_surf_grep (WEFG_superelem_t *se, double displ_scale, WEFG_crack_surf_t *cs);
#endif
static void 
propagate (WEFG_crack_surf_t *cs);
static void
write_hexvue_postproc_file (WEFG_crack_surf_t *cs,
                           ckit_iobuf_t IOB, char *mode);
static void
write_tecplot_postproc_file (WEFG_crack_surf_t *cs,
                            ckit_iobuf_t IOB, char *mode,
                            void *optdata);
static void
write_ve3d_loclist (WEFG_crack_surf_t *cs, ckit_iobuf_t IOB);
static void
write_ve3d_polys (WEFG_crack_surf_t *cs, ckit_iobuf_t IOB);
static void
update_stress (WEFG_superelem_t *se, WEFG_crack_surf_t *cs);
static void *
new_sampler (WEFG_crack_surf_t *cs, W_index_t pntidx, W_loc_t *loc);
static void 
sampler_store (WEFG_crack_surf_t *cs,
	      W_solution_history_t *h,
	      void *data,
	      double time);
static W_bool_t 
hex_intersected_by_crack_surf (WEFG_crack_surf_t *cs, 
                              WEFG_hex_volume_t *h
                              );


static WEFG_crack_surf_methods_t methods =  {
  segment_intersects_crack_surf,
  plane_generating_vector,
  wrap_around_point,
  node_pos_wrt_crack_surf,
  propagate,
#ifdef X_XFACE
  modify_crack_surf_grep,
#endif
  write_hexvue_postproc_file,
  write_ve3d_loclist,
  write_ve3d_polys,
  write_tecplot_postproc_file,
  update_stress,
  new_sampler,
  sampler_store,
  hex_intersected_by_crack_surf
} ;


static void
compute_axes (ts_vert_t *v,
             triangulated_crack_surf_t *tc);
static void
read_V (triangulated_crack_surf_t *tc, W_domain_t *d);
static void
read_T (triangulated_crack_surf_t *tc, W_domain_t *d);
static ts_vert_t *
new_vert (triangulated_crack_surf_t *tc, int id, W_loc_t *loc);
static ts_edge_t *
new_edge (triangulated_crack_surf_t *tc, int id,
         ts_vert_t *v0, ts_vert_t *v1);
static ts_tri_t *
new_tri (triangulated_crack_surf_t *tc, int id,
        ts_vert_t *v1,
        ts_vert_t *v2,
        ts_vert_t *v3,
        ts_edge_t *e1,
        ts_edge_t *e2,
        ts_edge_t *e3);
static ts_vert_t *
find_vert (triangulated_crack_surf_t *tc, int id);
static ts_edge_t *
find_edge_by_verts (triangulated_crack_surf_t *tc,
                   ts_vert_t *v0,
                   ts_vert_t *v1);
static void
vert_loc (ts_vert_t *v,
         triangulated_crack_surf_t *tc, 
         W_loc_t *pos_normal_loc, W_loc_t *neg_normal_loc);
static void
vert_sftabs (ts_vert_t *v, triangulated_crack_surf_t *tc);
static void
v_attach_e (ts_vert_t *v, ts_edge_t *e);
static void
e_attach_t (ts_edge_t *e, ts_tri_t *t);
static int
find_edge_vertices (ts_tri_t *t, int edge,
                   ts_vert_t **v0, ts_vert_t **v1);
static int
edge_idx (ts_tri_t *t, ts_edge_t *e);
static ts_edge_t *
connecting_edge (ts_vert_t *v, ts_vert_t *nv);
static void
j_surface (WEFG_crack_surf_t *cs,
          ts_vert_t *v,
          double hsize_forward_plus,
          double hsize_forward_minus,
          double hsize_tangent_plus,
          double hsize_tangent_minus,
          double hsize_normal_plus,
          double hsize_normal_minus,
          WEFG_tensor_prod_comp_func func,
          WEFG_tensor_prod_comp_der_func func_der,
          double *J_forward_static,
          double *J_forward_dynamic,
          double *J_normal_static,
          double *J_normal_dynamic
          );
static void
j_setup (WEFG_crack_surf_t *cs,
        ts_vert_t *v,
        double hsize_forward_plus,
        double hsize_forward_minus,
        double hsize_tangent_plus,
        double hsize_tangent_minus,
        double hsize_normal_plus,
        double hsize_normal_minus,
        WEFG_tensor_prod_comp_func func,
        WEFG_tensor_prod_comp_der_func func_der
        );
static void
loc_curr_loc (W_loc_t *loc, 
              triangulated_crack_surf_t *tc,
              WEFG_eval_for_packet_t *efp,
              W_loc_t *curr_loc);
static void
loc_sftabs (W_loc_t *loc, triangulated_crack_surf_t *tc,
            WEFG_eval_for_packet_t *efp);
static void
rel_displ_behind_front (triangulated_crack_surf_t *tc,
                       ts_vert_t *v,
                       double *opening, double *shearing, double *tearing);



/*
   The triangles in the surface are inflated (expanded) by a small
   amount to make the crack surface impermeable.  The reason for this
   measure lies in some observations which indicated that the rays
   connecting the quadrature points with the nodes were able in some
 (probably denegerate) cases penetrate the crack surface.

   This expansion amount is specified by the following macro.
   */
#define EXPTRI 1.005

#define SHRINK_TRI_DATA(theSHRINK, fstV, scndV, thrdV) \
   {                                                         \
     double theCV;                                          \
     theCV   = (fstV); theCV += (scndV); theCV += (thrdV);  \
     theCV   *= 0.3333333333333333333333333333333333333333; \
 (fstV) = theCV + (theSHRINK) * ((fstV) - theCV);     \
 (scndV) = theCV + (theSHRINK) * ((scndV) - theCV);     \
 (thrdV) = theCV + (theSHRINK) * ((thrdV) - theCV);     \
   } 
#define SHRINK_TRI_LOC_COMP(shrf, l0, l1, l2, comp) \
  SHRINK_TRI_DATA (shrf, (l0) .comp, (l1) .comp, (l2) .comp) 
#define SHRINK_TRI_LOC(shrf, l0, l1, l2) \
   {                                             \
    SHRINK_TRI_LOC_COMP (shrf, l0, l1, l2, x);   \
    SHRINK_TRI_LOC_COMP (shrf, l0, l1, l2, y);   \
    SHRINK_TRI_LOC_COMP (shrf, l0, l1, l2, z);   \
  } 

/*
   Read crack surface.
   Read crack surface data from stream.
   See the User's manual for details on the syntax.
 */
WEFG_crack_surf_t * /* crack or NULL if something went wrong */
read_triangulated_crack_surf (WEFG_superelem_t *se, /* EFG superelement */
                              W_domain_t *d
                              ) 
{
  const char *input_line;
  WEFG_crack_surf_t *cs = NULL;
  triangulated_crack_surf_t *tc = NULL;
  int kp;
  const char *keyw;
#define IS_COMMAND(command) \
 (tokens_token_and_keyword_equiv (command, keyw)) 
  int master_vert_id = -1;

  while ((input_line = W_get_input_line (d)) ) {
    TOKENS_PARSE (input_line);
    if (TOKENS_HAVE_KEYWORD ("END")) {
      if (TOKENS_HAVE_KEYWORD ("TRIANGULATED")) {
        check_defs (tc);
	efg_enable_ip_storage_to_search_structure (se);
        if (master_vert_id > 0) {
          tc->master_vert = find_vert (tc, master_vert_id);
          if (tc->master_vert == NULL) {
            REPORT_FATAL_ERROR (se->domain->errh, "No such master vertex");
          } 
        } 
        return cs;
      }  else
        REPORT_FATAL_ERROR (se->domain->errh, "TRIANGULATED does not end by END TRIANGULATED");
    }  else  {
      keyw = TOKEN_STR (1);
      if (tc == NULL && (!IS_COMMAND ("ID")) ) {
        REPORT_FATAL_ERROR (se->domain->errh, "TRIANGULATED data must start by the ID command");
      } 
      if (IS_COMMAND ("ID")) {
        int id;
        id = TOKEN_INT (2);
        tc = w_new (triangulated_crack_surf_t);
        CHECK_MALLOC (tc);
        tc->is_stationary     = FALSE;
        if ((kp = tokens_have_keyword (d->parser, 1, "STATIONARY")) > 0) {
          tc->is_stationary   = TRUE;
        } 
        tc->vertices          = make_list ();
        tc->edges             = make_list ();
        tc->triangles         = make_list ();
        tc->on_front_vertices = NULL;
        if (TOKENS_HAVE_KEYWORD ("AVERAGED_STRESS")) {
          tc->from_strength_stress_type = LOCAL_STRESS; 
        }  else  {
          tc->from_strength_stress_type = LOCAL_STRESS;
        } 
        tc->min_edge_length          = FLT_MAX;
        tc->max_edge_length          = 0;
        tc->rayleigh_vel_frac        = 1.0;
        if ((kp = tokens_have_keyword (d->parser, 1, "RAYLEIGH_VEL_FRAC")) > 0) {
          tc->rayleigh_vel_frac = TOKEN_DBL (kp+1);
        } 
        tc->edge_length_lower_bound = 0;
        if ((kp = tokens_have_keyword (d->parser, 1, "EDGE_LENGTH_LOWER_BOUND")) > 0) {
          tc->edge_length_lower_bound = TOKEN_DBL (kp+1);
        } 
        tc->edge_length_upper_bound = 0;
        if ((kp = tokens_have_keyword (d->parser, 1, "EDGE_LENGTH_UPPER_BOUND")) > 0) {
          tc->edge_length_upper_bound = TOKEN_DBL (kp+1);
        } 
        tc->allow_relaxation         = FALSE;
        if ((kp = tokens_have_keyword (d->parser, 1, "ALLOW_RELAXATION")) > 0) {
          tc->allow_relaxation = TRUE;
        } 
        tc->allow_adaptivity         = FALSE;
        if ((kp = tokens_have_keyword (d->parser, 1, "ALLOW_ADAPTIVITY")) > 0) {
          tc->allow_adaptivity = TRUE;
        } 
        tc->allow_E_star         = FALSE;
        if ((kp = tokens_have_keyword (d->parser, 1, "ALLOW_E_STAR")) > 0) {
          tc->allow_E_star = TRUE;
        } 
	tc->jdom_hsize_forward_plus  =
	tc->jdom_hsize_forward_minus =
	tc->jdom_hsize_tangent_plus  =
	tc->jdom_hsize_tangent_minus =
	tc->jdom_hsize_normal_plus   =   
	tc->jdom_hsize_normal_minus  = 0;  
	if ((kp = tokens_have_keyword (d->parser, 1, "JDOM")) > 0) {
          tc->jdom_hsize_forward_plus  = TOKEN_DBL (kp+1);
          tc->jdom_hsize_forward_minus = TOKEN_DBL (kp+2);
          tc->jdom_hsize_tangent_plus  = TOKEN_DBL (kp+3);
          tc->jdom_hsize_tangent_minus = TOKEN_DBL (kp+4);
          tc->jdom_hsize_normal_plus   = TOKEN_DBL (kp+5);
          tc->jdom_hsize_normal_minus  = TOKEN_DBL (kp+6);
        } 
        tc->ignore_surface_j_contrib = FALSE;
        tc->subsplit_for_2x2x2_ipol  = 4;
        tc->extrapolate_at_the_bdry  = FALSE;
        tc->mixed_eval_at_the_bdry   = FALSE;
        tc->drive_by_vel_incr        = FALSE;
        tc->vel_incr_fraction        = 1000000;
        tc->advance_prescribed       = FALSE;
        tc->advance_velocity[0]      =
        tc->advance_velocity[1]      =
        tc->advance_velocity[2]      = NULL;
        tc->dist_to_front_for_displ_ratio = 0.1;
        tc->trim_advv_mult           = 2.5;
        tc->tilt_damping             = 0.5;
        tc->master_vert              = NULL;
        tc->grow_in_plane            = FALSE;
        tc->bbt = bbt_new_bbt ();
        tc->bbt->max_num_of_triangles_per_box = 50;
        cs = efgcr_new_crack_surf (se, id, tc, &methods);
        tc->crack_surf = cs;
      }  else if (IS_COMMAND ("ADVANCE")) {
        const char *s;
        tc->advance_prescribed = TRUE;
        s = TOKEN_STR (2);
        DUPLICATE_STRING (tc->advance_velocity[0], s);
        s = TOKEN_STR (3);
        DUPLICATE_STRING (tc->advance_velocity[1], s);
        s = TOKEN_STR (4);
        DUPLICATE_STRING (tc->advance_velocity[2], s);
      }  else if (IS_COMMAND ("IGNORE_SURFACE_J_CONTRIB")) {
        tc->ignore_surface_j_contrib = TRUE;
      }  else if (IS_COMMAND ("DIST_TO_FRONT_FOR_DISPL_RATIO")) {
        tc->dist_to_front_for_displ_ratio = TOKEN_DBL (2);
      }  else if (IS_COMMAND ("TRIM_ADVV_MULT")) {
        tc->trim_advv_mult = TOKEN_DBL (2);
      }  else if (IS_COMMAND ("MASTER_VERT")) {
        master_vert_id = TOKEN_INT (2);
      }  else if (IS_COMMAND ("JDOM")) {
        kp = 1;
        tc->jdom_hsize_forward_plus  = TOKEN_DBL (kp+1);
        tc->jdom_hsize_forward_minus = TOKEN_DBL (kp+2);
        tc->jdom_hsize_tangent_plus  = TOKEN_DBL (kp+3);
        tc->jdom_hsize_tangent_minus = TOKEN_DBL (kp+4);
        tc->jdom_hsize_normal_plus   = TOKEN_DBL (kp+5);
        tc->jdom_hsize_normal_minus  = TOKEN_DBL (kp+6);
      }  else if (IS_COMMAND ("GROW_IN_PLANE")) {
        G3D_vect_t n;
        tc->grow_in_plane = TRUE;
        n.x = TOKEN_DBL (2);
        n.y = TOKEN_DBL (3);
        n.z = TOKEN_DBL (4);
        if (!G3D_plane_consts_p_and_n (&tc->motion_plane, &n, &n)) 
          REPORT_FATAL_ERROR (se->domain->errh, "invalid plane");
      }  else if (IS_COMMAND ("ALLOW_ADAPTIVITY")) {
        tc->allow_adaptivity = TRUE;
      }  else if (IS_COMMAND ("ALLOW_RELAXATION")) {
        tc->allow_relaxation = TRUE;
      }  else if (IS_COMMAND ("EDGE_LENGTH_LOWER_BOUND")) {
        tc->edge_length_lower_bound = TOKEN_DBL (2);
      }  else if (IS_COMMAND ("TILT_DAMPING")) {
        tc->tilt_damping = TOKEN_DBL (2);
        if (tc->tilt_damping > 1 || tc->tilt_damping < 0) {
          REPORT_FATAL_ERROR (se->domain->errh, "Tilt damping should be in the interval <0;1>");
        } 
      }  else if (IS_COMMAND ("EDGE_LENGTH_UPPER_BOUND")) {
        tc->edge_length_upper_bound = TOKEN_DBL (2);
      }  else if (IS_COMMAND ("ALLOW_E_STAR")) {
        tc->allow_E_star = TRUE;
      }  else if (IS_COMMAND ("RAYLEIGH_VEL_FRAC")) {
        tc->rayleigh_vel_frac = TOKEN_DBL (2);
      }  else if (IS_COMMAND ("DRIVE_BY_VEL_INCR")) {
        tc->drive_by_vel_incr = TRUE;
      }  else if (IS_COMMAND ("VEL_INCR_FRACTION")) {
        tc->vel_incr_fraction = TOKEN_DBL (2);
        if (tc->vel_incr_fraction < 1) REPORT_FATAL_ERROR (se->domain->errh, "VEL_INCR_FRACTION < 1");
      }  else if (IS_COMMAND ("STATIONARY")) {
        tc->is_stationary   = TRUE;
      }  else if (IS_COMMAND ("SUBSPLIT_FOR_2X2X2_IPOL")) {
        tc->subsplit_for_2x2x2_ipol = TOKEN_INT (2);
      }  else if (IS_COMMAND ("EXTRAPOLATE_AT_THE_BDRY")) {
        tc->extrapolate_at_the_bdry = TRUE;
      }  else if (IS_COMMAND ("MIXED_EVAL_AT_THE_BDRY")) {
        tc->mixed_eval_at_the_bdry = TRUE;
        tc->extrapolate_at_the_bdry = TRUE;
      }  else if (IS_COMMAND ("V")) {
        if (tc == NULL) 
          REPORT_FATAL_ERROR (d->errh, "Start data with ID");
        read_V (tc, d);
      }  else if (IS_COMMAND ("T")) {
        if (tc == NULL) 
          REPORT_FATAL_ERROR (d->errh, "Start data with ID");
        read_T (tc, d);
      } 
    } 
  } 
  REPORT_FATAL_ERROR (d->errh, "TRIANGULATED does not end by END TRIANGULATED");
  return NULL;
} 



static int
find_hex_enclosing_loc (triangulated_crack_surf_t *tc, W_loc_t *loc,
                       WEFG_hex_volume_t **h, W_loc_t *ploc);


/*
   Read crack surface vertex.
 */
static void
read_V (triangulated_crack_surf_t *tc, /* triangulated crack instance */
        W_domain_t *d
        ) 
{
  ts_vert_t *v;
  int id;
  W_loc_t loc;

  id    = TOKEN_INT (2);
  loc.x = TOKEN_DBL (3);
  loc.y = TOKEN_DBL (4);
  loc.z = TOKEN_DBL (5);
  v = new_vert (tc, id, &loc);
  if (!find_hex_enclosing_loc (tc, &v->loc, &v->enclosing_hex, &v->ploc)) {
    sprintf (W_buf (), "No enclosing hex at [%g;%g;%g]",
            v->loc.x, v->loc.y, v->loc.z);
    REPORT_FATAL_ERROR (tc->crack_surf->se->domain->errh, W_buf ());
  } 
  v->ms_pos_normal = WM3D_get_mat_state (tc->crack_surf->se->mat, &v->loc);
  v->ms_neg_normal = WM3D_get_mat_state (tc->crack_surf->se->mat, &v->loc);
  if (TOKENS_HAVE_KEYWORD ("ON_BDRY")) {
    v->is_on_boundary = TRUE;
  } 
  if (TOKENS_HAVE_KEYWORD ("ON_FRONT")) {
    v->is_on_front = TRUE;
  } 
  add_to_tail (tc->vertices, v);
} 


/* new vertex. */
static ts_vert_t * /* new vertex */
new_vert (triangulated_crack_surf_t *tc /* instance of triangulated crack */,
          int id, /* vertex id */
          W_loc_t *loc /* location */
          ) 
{
  ts_vert_t *v;
  static int fallback_id = 1000000;
  W_vect_t zerovect =  {0,0,0} ;
  
  v = w_new (ts_vert_t); CHECK_MALLOC (v);
  
  v->id                 = (id > 0? id: ++fallback_id);
  v->loc                = *loc;
  v->is_on_front        = FALSE;
  v->advance_vector     = zerovect;
  v->K_I_0              = 0;
  v->K_II_0             = 0;
  v->K_III_0            = 0;
  v->K_I_c              = 0;
  v->K_II_c             = 0;
  v->K_III_c            = 0;
  v->K_I_equiv          = 0;
  v->G_0                = 0;
  v->G_c                = 0;
  v->theta              = 0;
  v->C                  = 0;
  v->source_vert        = NULL;
  v->is_on_boundary     = FALSE;
  v->on_face            = NULL;
  v->edges              = make_list ();
  INIT_SFTAB (v->sftab_pos_normal);
  INIT_SFTAB (v->sftab_neg_normal);
  v->ms_pos_normal      = NULL;
  v->ms_neg_normal      = NULL;
  v->enclosing_hex      = NULL;
  v->hist_cd            = NULL;
#ifdef X_XFACE
  v->grep_pos_normal.marker = NULL;
  v->grep_neg_normal.marker = NULL;
# if DRAW_VERT_NUMS
  v->grep_pos_normal.text   = NULL;
  v->grep_neg_normal.text   = NULL;
# endif
#endif

  return v;
} 
  





/* new edge.
 create new crack front edge connecting specified vertices.
 */
static ts_edge_t *
new_edge (triangulated_crack_surf_t *tc /* instance of triangulated crack */,
          int id, /* edge id */
          ts_vert_t *v0, /* first vertex */
          ts_vert_t *v1 /* last vertex */
          ) 
{
  ts_edge_t *e;
  static int fallback_id = 1000000;

  e = w_new (ts_edge_t); CHECK_MALLOC (e);
  
  e->id             = (id > 0? id: ++fallback_id);
  e->vertex[0]      = v0;
  if (e->vertex[0] == NULL) REPORT_FATAL_ERROR (tc->crack_surf->se->domain->errh, "Vertex not found");
  e->vertex[1]      = v1;
  if (e->vertex[1] == NULL) REPORT_FATAL_ERROR (tc->crack_surf->se->domain->errh, "Vertex not found");
  e->triangles[0]   = NULL;
  e->triangles[1]   = NULL;

  return e;
} 
  

static void
forget_edge (triangulated_crack_surf_t *tc /* instance of triangulated crack */,
             ts_edge_t *e) 
{
  delete_list_node (tc->edges, e);
} 
  



static void
tri_normal (ts_tri_t *t);

/*
   Add crack triangle to search structure.
   Add crack surface triangle to a search structure.  The BBT (Bounding-box tree) library is used.
 */
static void
add_to_bbt (ts_tri_t *t, /* triangle */
            triangulated_crack_surf_t *tc /* instance of triangulated crack */
            ) 
{ 
  G3D_tri_t tri;
  G3D_plane_t plane;
  
  VECMAC_2_VEC (tri.loc[0], =, t->vertex[0]->loc);
  VECMAC_2_VEC (tri.loc[1], =, t->vertex[1]->loc);
  VECMAC_2_VEC (tri.loc[2], =, t->vertex[2]->loc);
  if (G3D_plane_consts_3 (&plane, &tri.loc[0], &tri.loc[1], &tri.loc[2])) {
    SHRINK_TRI_LOC (EXPTRI, tri.loc[0], tri.loc[1], tri.loc[2]);
    bbt_add_tri (tc->bbt, t, &tri);
  } 
} 


/* new triangle.
 create new crack surface triangle.
 
 */
static ts_tri_t *
new_tri (triangulated_crack_surf_t *tc /* instance of triangulated crack */,
         int id, /* triangle id */
        ts_vert_t *v1, /* vertex */
        ts_vert_t *v2, /* vertex */
        ts_vert_t *v3, /* vertex */
        ts_edge_t *e1, /* edge */
        ts_edge_t *e2, /* edge */
        ts_edge_t *e3 /* edge */
         ) 
{
  ts_tri_t *t;
  static int fallback_id = 1000000;
#define APPROX_THE_SAME(p0, p1) \
 ((fabs ((p0) .x - (p1) .x) < 1.e-6) &&         \
 (fabs ((p0) .y - (p1) .y) < 1.e-6) &&         \
 (fabs ((p0) .z - (p1) .z) < 1.e-6)) 

  t = w_new (ts_tri_t); CHECK_MALLOC (t);
  
  t->id             = (id > 0? id: ++fallback_id);
  t->vertex[0]      = v1;
  t->vertex[1]      = v2;
  t->vertex[2]      = v3;
  if ((APPROX_THE_SAME (t->vertex[0]->loc, t->vertex[1]->loc)) ||
 (APPROX_THE_SAME (t->vertex[1]->loc, t->vertex[2]->loc)) ||
 (APPROX_THE_SAME (t->vertex[2]->loc, t->vertex[0]->loc)) ) {
    REPORT_WARNING (tc->crack_surf->se->domain->errh, "identical locations of vertices");
  } 

  t->edges[0] = e1;
  v_attach_e (t->vertex[0], t->edges[0]);
  v_attach_e (t->vertex[1], t->edges[0]);
  e_attach_t (t->edges[0], t);
  
  t->edges[1] = e2;
  v_attach_e (t->vertex[1], t->edges[1]);
  v_attach_e (t->vertex[2], t->edges[1]);
  e_attach_t (t->edges[1], t);

  t->edges[2] = e3;
  v_attach_e (t->vertex[2], t->edges[2]);
  v_attach_e (t->vertex[0], t->edges[2]);
  e_attach_t (t->edges[2], t);

  t->bbox.lower = t->bbox.upper = t->vertex[0]->loc; /* bounding box */
  t->bbox.lower.x = min (t->bbox.lower.x, t->vertex[1]->loc.x);
  t->bbox.lower.x = min (t->bbox.lower.x, t->vertex[2]->loc.x);
  t->bbox.lower.y = min (t->bbox.lower.y, t->vertex[1]->loc.y);
  t->bbox.lower.y = min (t->bbox.lower.y, t->vertex[2]->loc.y);
  t->bbox.lower.z = min (t->bbox.lower.z, t->vertex[1]->loc.z);
  t->bbox.lower.z = min (t->bbox.lower.z, t->vertex[2]->loc.z);
  t->bbox.upper.x = max (t->bbox.upper.x, t->vertex[1]->loc.x);
  t->bbox.upper.x = max (t->bbox.upper.x, t->vertex[2]->loc.x);
  t->bbox.upper.y = max (t->bbox.upper.y, t->vertex[1]->loc.y);
  t->bbox.upper.y = max (t->bbox.upper.y, t->vertex[2]->loc.y);
  t->bbox.upper.z = max (t->bbox.upper.z, t->vertex[1]->loc.z);
  t->bbox.upper.z = max (t->bbox.upper.z, t->vertex[2]->loc.z);

  tri_normal (t);
  VECMAC_2_VEC (tc->average_normal, +=, t->normal);
  t->unified_normal = FALSE;

#ifdef X_XFACE
  t->triangle_pos_normal = NULL;
  t->triangle_neg_normal = NULL;
#endif /* X_XFACE  */

  add_to_bbt (t, tc);

  return t;
} 
  

/* valid normal?
 has the crack surface triangle a valid normal?
The area of the triangle as given by the norm of the cross product of two edge vectors
is compared to an `ideal' area of triangle with a right angle between the two edges.
If the true area is too small, the normal is deemed invalid (unreliable).
 */
static int
has_valid_normal (ts_vert_t *v0,
                  ts_vert_t *v1,
                  ts_vert_t *v2
                  ) 
{
  double two_areas, comp_area;
  W_vect_t a, b, c= {0,0,0} ;
  double la, lb;

  VECMAC_3_VEC (a, =, v1->loc, -, v0->loc);
  VECMAC_3_VEC (b, =, v2->loc, -, v0->loc);
  VECMAC_CROSS (c, =, a, x, b);
  two_areas =   VECMAC_NORM (c);

  la = VECMAC_NORM (a);
  lb = VECMAC_NORM (b);

  comp_area = 2 * (la + lb) * (la + lb) / 4 / 100;

  if (two_areas < comp_area) 
    return FALSE;
  else
    return TRUE;
} 

/* compute normal.
 compute normal of triangle.
 if the triangle has a valid normal, it is computed and normalized;
 otherwise the triangle normal is approximated by looking at its neighbors
 and adopting an average of their normals.
 */
static void
tri_normal (ts_tri_t *t) 
{
  W_vect_t a, b;
  double nn;

  if (has_valid_normal (t->vertex[0], t->vertex[1], t->vertex[2])) {
    VECMAC_3_VEC (a, =, t->vertex[1]->loc, -, t->vertex[0]->loc);
    VECMAC_3_VEC (b, =, t->vertex[2]->loc, -, t->vertex[1]->loc);
    VECMAC_CROSS (t->normal, =, a, x, b);
    VECMAC_NORMALIZE (t->normal);
  }  else  {
    /* take the normal from one of its neighbors */
    ts_tri_t *nt;
    VECMAC_3_SCALAR (t->normal, =, 0, 0, 0);
    nt = THE_OTHER_TRI (t->edges[0], t);
    if (nt) {
      VECMAC_2_VEC (t->normal, +=, nt->normal);
    } 
    nt = THE_OTHER_TRI (t->edges[1], t);
    if (nt) {
      VECMAC_2_VEC (t->normal, +=, nt->normal);
    } 
    nt = THE_OTHER_TRI (t->edges[2], t);
    if (nt) {
      VECMAC_2_VEC (t->normal, +=, nt->normal);
    } 
    nn = VECMAC_NORM (t->normal);
    if (nn > 0) {
      VECMAC_NORMALIZE (t->normal);
    }  else  {
      REPORT_FATAL_ERROR (NULL, "No normal for a triangle");
    } 
  } 
} 
  

/* read crack surface triangle.
   Read a crack surface triangle from the input stream.
   See the User's manual for details.
 */
static void
read_T (triangulated_crack_surf_t *tc /* instance of triangulated crack */,
        W_domain_t *d
        ) 
{
  ts_tri_t *t;
  int id;
  int v1, v2, v3;
  ts_vert_t *tv1, *tv2, *tv3;
  ts_edge_t *te1, *te2, *te3;
  
  id    = TOKEN_INT (2);
  v1    = TOKEN_INT (3);
  v2    = TOKEN_INT (4);
  v3    = TOKEN_INT (5);

  tv1   = find_vert (tc, v1);
  if (tv1 == NULL) REPORT_FATAL_ERROR (d->errh, "No such vertex");
  tv2   = find_vert (tc, v2);
  if (tv2 == NULL) REPORT_FATAL_ERROR (d->errh, "No such vertex");
  tv3   = find_vert (tc, v3);
  if (tv3 == NULL) REPORT_FATAL_ERROR (d->errh, "No such vertex");

  te1   = find_edge_by_verts (tc, tv1, tv2);
  if (te1 == NULL) {
    te1 = new_edge (tc, 0, tv1, tv2);
    add_to_tail (tc->edges, te1);
  } 
  te2   = find_edge_by_verts (tc, tv2, tv3);
  if (te2 == NULL) {
    te2 = new_edge (tc, 0, tv2, tv3);
    add_to_tail (tc->edges, te2);
  } 
  te3   = find_edge_by_verts (tc, tv3, tv1);
  if (te3 == NULL) {
    te3 = new_edge (tc, 0, tv3, tv1);
    add_to_tail (tc->edges, te3);
  } 

  t     = new_tri (tc, id, tv1, tv2, tv3, te1, te2, te3);
  
  add_to_tail (tc->triangles, t);
} 


/* find connecting edge.
 find edge connecting given vertices.
 */
static ts_edge_t * /* edge or NULL */
connecting_edge (ts_vert_t *v, ts_vert_t *nv) 
{
  ts_edge_t *e;

  loop_through_list (v->edges, e, ts_edge_t *) {
    if (THE_OTHER_VERTEX (e, v) == nv) 
      return e;
  } 
  return NULL;
} 



static int
cmp_vert_id (int id, ts_vert_t *v);

static int
cmp_vert_id (int id, ts_vert_t *v) 
{
  return v->id == id;
} 

static ts_vert_t *
find_vert (triangulated_crack_surf_t *tc /* instance of triangulated crack */, int id) 
{
  return (ts_vert_t *) search_list (tc->vertices, id, cmp_vert_id);
} 



static double loctol = 0;

static int
cmp_vert_loc (W_loc_t *l, ts_vert_t *v);

static int
cmp_vert_loc (W_loc_t *l, ts_vert_t *v) 
{
  return ((fabs (l->x - v->loc.x) <= loctol) &&
	 (fabs (l->y - v->loc.y) <= loctol) &&
	 (fabs (l->z - v->loc.z) <= loctol));
} 

static ts_vert_t *
find_vert_by_loc (triangulated_crack_surf_t *tc /* instance of triangulated crack */,
                  W_loc_t *l) 
{
  return (ts_vert_t *) search_list (tc->vertices, l, cmp_vert_loc);
} 




#if 0
static int
cmp_edge_id (int id, ts_edge_t *v);

static int
cmp_edge_id (int id, ts_edge_t *e) 
{
  return e->id == id;
} 

static ts_edge_t *
find_edge (triangulated_crack_surf_t *tc /* instance of triangulated crack */, int id) 
{
  return (ts_edge_t *) search_list (tc->edges, id, cmp_edge_id);
} 
#endif



static int
cmp_edge_verts (ts_vert_t *v[2], ts_edge_t *e);

static int
cmp_edge_verts (ts_vert_t *v[2], ts_edge_t *e) 
{
  return ((e->vertex[0] == v[0] && e->vertex[1] == v[1]) 
          || (e->vertex[1] == v[0] && e->vertex[0] == v[1]));
} 

static ts_edge_t *
find_edge_by_verts (triangulated_crack_surf_t *tc /* instance of triangulated crack */,
                   ts_vert_t *v0,
                   ts_vert_t *v1) 
{
  ts_vert_t *v[2];
  v[0] = v0; v[1] = v1;
  return (ts_edge_t *) search_list (tc->edges, v, cmp_edge_verts);
} 

/*
   Segment intersects crack surface?
   Does a segment intersect the crack surface?

   The tolerance tol is normalized by the segment length --> eps. If intersection detected
 for |lambda| < eps or |1-lambda| < eps, the intersection point is considered
 to lie on the crack surface (lambda being the parameterization of the segment;
 lambda=0 corresponds to the evaluation point, lambda=1 corresponds to the node).

 */
static W_bool_t 
segment_intersects_crack_surf (WEFG_eval_for_packet_t *efp, /* evaluation packet */
                               G3D_segment_t *s, /* segment */
                               WEFG_node_t *efgn, /* EFG node */
                               WEFG_crack_surf_t *cs, /* crack surface */
                               double  tol /* tolerance; see writeup */
                               ) 
{
  triangulated_crack_surf_t *tc = (triangulated_crack_surf_t *) cs->data;
  G3D_vect_t ab;
  double lambda;
  double eps, ab_length, ctol;
  ts_tri_t *tri = NULL;
#define SEGBB(s, b) { G3D_INIT_BOX (b, s->start); G3D_UPDT_BOX (b, s->end); } 
#define BOXES_OVERLAP(r1, r2) G3D_BOXES_OVERLAP (r1, r2) 

  /* If a point of the segment lies on the crack surface,
     handle it specially:  Use the crack surface normal to do the check.
   */
  if (efp->crack_surf == cs) {
    double normal_ab_dot;
    if (efp->pt_wrt_crack_surf_classification == WEFG_AT_FRONT) 
      return FALSE;
    VECMAC_3_VEC (ab, =, s->end, -, s->start);
    normal_ab_dot = VECMAC_DOT (ab, efp->crack_surf_normal);
    if (efp->pt_wrt_crack_surf_classification == WEFG_POSITIVE_NORMAL) {
      if (normal_ab_dot > 0) 
        return FALSE;
      else
        return TRUE;
    } 
    if (efp->pt_wrt_crack_surf_classification == WEFG_NEGATIVE_NORMAL) {
      if (normal_ab_dot < 0) 
        return FALSE;
      else
        return TRUE;
    } 
  } 
  
  /* Segment start is the evaluation point, end is the EFG node */
  VECMAC_3_VEC (ab, =, s->end, -, s->start);

  ab_length = VECMAC_NORM (ab);
  if (ab_length > 0) 
    eps       = tol / ab_length;             /* gives d_lambda */
  else
    eps       = 0.00001;
  ctol      = tol / 10;

  tc->bbt->eps = ctol;
  tc->bbt->test_box_only = FALSE;
  if (bbt_intersects_a_tri (tc->bbt, &s->start, &s->end)) {
    lambda = tc->bbt->lambda;
    if (lambda >= eps && lambda <= 1-eps) {
      return TRUE;
    }  else if (fabs (lambda) < eps) { /* eval pt on the surface? */
      double normal_ab_dot;
      tri = (ts_tri_t *) tc->bbt->intersected_object;
      normal_ab_dot = VECMAC_DOT (ab, tri->normal);
      if (normal_ab_dot > 0) return FALSE;
      else                   return TRUE;
    }  else if (fabs (1 - lambda) < eps) { /* node on the surface? */
      if (lambda == 1) {
        W_loc_t rl;
        EFGN_REF_COORDS (efgn, rl.x, rl.y, rl.z);
        sprintf (W_buf (), "Node %d on the crack surface at  {%g;%g;%g} ",
 (int) efgn->id, rl.x, rl.y, rl.z);
        REPORT_WARNING (tc->crack_surf->se->domain->errh, W_buf ());
      }  else  {
        /* arbitrarily assigning to one side */
        return (lambda > 1);
      } 
    } 
  } 
  
  return FALSE;
} 


#ifdef X_XFACE

static EPixel red, yellow, white;
static int inited = 0;

static void
init_props (void) 
{
  int suc = 0;

  red = ColorGetPixelFromString ("red", &suc);
  if (!suc) REPORT_WARNING (NULL, "failed to get the requested color");
  yellow = ColorGetPixelFromString ("yellow", &suc);
  if (!suc) REPORT_WARNING (NULL, "failed to get the requested color");
  white = ColorGetPixelFromString ("white", &suc);
  if (!suc) REPORT_WARNING (NULL, "failed to get the requested color");
  
  inited = 1;
} 

static void
modify_v_grep (WEFG_superelem_t *se, double displ_scale,
               triangulated_crack_surf_t *tc /* instance of triangulated crack */,
              ts_vert_t *v);
static void
modify_e_grep (WEFG_superelem_t *se, double displ_scale,
               triangulated_crack_surf_t *tc /* instance of triangulated crack */,
              ts_edge_t *e);
static void
modify_t_grep (WEFG_superelem_t *se, double displ_scale,
               triangulated_crack_surf_t *tc /* instance of triangulated crack */,
              ts_tri_t *t);

static void 
modify_crack_surf_grep (WEFG_superelem_t *se, double displ_scale,
               WEFG_crack_surf_t *cs) 
{
  triangulated_crack_surf_t *tc = (triangulated_crack_surf_t *) cs->data;
  ts_vert_t *v;
  ts_edge_t *e;
  ts_tri_t *t;

  if (!inited) init_props ();
  loop_through_list (tc->vertices, v, ts_vert_t *) {
    modify_v_grep (se, displ_scale, tc, v);
  } 
  loop_through_list (tc->edges, e, ts_edge_t *) {
    modify_e_grep (se, displ_scale, tc, e);
  } 
  loop_through_list (tc->triangles, t, ts_tri_t *) {
    modify_t_grep (se, displ_scale, tc, t);
  } 
} 
#endif /* X_XFACE  */

#ifdef X_XFACE
static void
modify_v_grep (WEFG_superelem_t *se, double displ_scale,
               triangulated_crack_surf_t *tc /* instance of triangulated crack */,
              ts_vert_t *v) 
{
  W_loc_t ploc, nloc;
  WCRec ppnts[1] =  { {0,0,0} } , npnts[1] =  { {0,0,0} } ;
#if DRAW_VERT_NUMS
  char buf[32];
#define NEW_GREPS(which) \
   {                                                                           \
    if (v->grep_##which##_normal.marker == NULL) {                            \
      v->grep_##which##_normal.marker = CreateMarker3D (npnts);               \
      sprintf (buf, "%d", v->id);                                             \
      v->grep_##which##_normal.text   = CreateAnnText3D (npnts, buf);         \
      EASValsSetMType (CIRCLE_MARKER);                                        \
      EASValsSetMSize (10);                                                   \
      EASValsSetLayer (EFG_NODE_LAYER);                                       \
      EGWithMaskChangeAttributes (MTYPE_MASK | MSIZE_MASK | LAYER_MASK,       \
                                  v->grep_##which##_normal.marker);           \
      EASValsSetLayer (WX_ANNOT_LAYER);                                       \
      EGWithMaskChangeAttributes (LAYER_MASK,                                 \
                                  v->grep_##which##_normal.text);             \
      EMNoDrawAddGraphicsToModel (ESIModel (), v->grep_##which##_normal.marker); \
      EMNoDrawAddGraphicsToModel (ESIModel (), v->grep_##which##_normal.text);  \
    }                                                                          \
  } 
#else
#define NEW_GREPS(which) \
   {                                                                           \
    if (v->grep_##which##_normal.marker == NULL) {                            \
      v->grep_##which##_normal.marker = CreateMarker3D (npnts);               \
      EASValsSetMType (CIRCLE_MARKER);                                        \
      EASValsSetMSize (10);                                                   \
      EASValsSetLayer (EFG_NODE_LAYER);                                       \
      EGWithMaskChangeAttributes (MTYPE_MASK | MSIZE_MASK | LAYER_MASK,       \
                                  v->grep_##which##_normal.marker);           \
      EASValsSetLayer (WX_ANNOT_LAYER);                                       \
      EMNoDrawAddGraphicsToModel (ESIModel (), v->grep_##which##_normal.marker); \
    }                                                                          \
  } 
#endif

  NEW_GREPS (pos);
  NEW_GREPS (neg);
  
  if (v->is_on_front) {
    EASValsSetColor (red);
  }  else  {
    EASValsSetColor (white);
  } 
  EASValsSetMType (FILLED_CIRCLE_MARKER); 
  if (v->is_on_boundary) {
    EASValsSetMType (SQUARE_MARKER);
  } 
  EGWithMaskChangeAttributes (MTYPE_MASK | COLOR_MASK,
                              v->grep_pos_normal.marker);
  EGWithMaskChangeAttributes (MTYPE_MASK | COLOR_MASK,
                              v->grep_neg_normal.marker);
#if DRAW_VERT_NUMS
  EGWithMaskChangeAttributes (COLOR_MASK, v->grep_pos_normal.text);
  EGWithMaskChangeAttributes (COLOR_MASK, v->grep_neg_normal.text);
#endif

  vert_loc (v, tc, &ploc, &nloc);
  ppnts[0].x = (v->loc.x + displ_scale * (ploc.x - v->loc.x));
  ppnts[0].y = (v->loc.y + displ_scale * (ploc.y - v->loc.y));
  ppnts[0].z = (v->loc.z + displ_scale * (ploc.z - v->loc.z));
  npnts[0].x = (v->loc.x + displ_scale * (nloc.x - v->loc.x));
  npnts[0].y = (v->loc.y + displ_scale * (nloc.y - v->loc.y));
  npnts[0].z = (v->loc.z + displ_scale * (nloc.z - v->loc.z));

  EMEraseGraphics (ESIModel (), v->grep_pos_normal.marker);
  EMEraseGraphics (ESIModel (), v->grep_neg_normal.marker);
  EGModifyGraphicsGeometry (v->grep_pos_normal.marker, ppnts);
  EGModifyGraphicsGeometry (v->grep_neg_normal.marker, npnts);
  EMDrawGraphics (ESIModel (), v->grep_pos_normal.marker);
  EMDrawGraphics (ESIModel (), v->grep_neg_normal.marker);

#if DRAW_VERT_NUMS
  EMEraseGraphics (ESIModel (), v->grep_pos_normal.text);
  EMEraseGraphics (ESIModel (), v->grep_neg_normal.text);
  EGModifyGraphicsGeometry (v->grep_pos_normal.text, ppnts);
  EGModifyGraphicsGeometry (v->grep_neg_normal.text, npnts);
  EMDrawGraphics (ESIModel (), v->grep_pos_normal.text);
  EMDrawGraphics (ESIModel (), v->grep_neg_normal.text);
#endif
} 
#endif /* X_XFACE  */

#ifdef X_XFACE
static void
modify_e_grep (WEFG_superelem_t *se, double displ_scale,
               triangulated_crack_surf_t *tc /* instance of triangulated crack */,
              ts_edge_t *e) 
{
} 
#endif /* X_XFACE  */

#ifdef X_XFACE
static void
modify_t_grep (WEFG_superelem_t *se, double displ_scale,
               triangulated_crack_surf_t *tc /* instance of triangulated crack */,
              ts_tri_t *t) 
{
  WCRec ppnts[3] =  { {0,0,0} , {0,0,0} , {0,0,0} } ;
  WCRec npnts[3] =  { {0,0,0} , {0,0,0} , {0,0,0} } ;
  W_loc_t ploc, nloc;
  int i;
  
  if (t->triangle_pos_normal == NULL) {
    t->triangle_pos_normal = CreateTriangle3D (ppnts);
    EASValsSetColor (yellow);
    EASValsSetShrink (0.95);
    EASValsSetLayer (EFG_SURFACE_LAYER);
    EGWithMaskChangeAttributes (COLOR_MASK | LAYER_MASK | SHRINK_MASK,
                               t->triangle_pos_normal);
    EMNoDrawAddGraphicsToModel (ESIModel (), t->triangle_pos_normal);
  } 
  if (t->triangle_neg_normal == NULL) {
    t->triangle_neg_normal = CreateTriangle3D (npnts);
    EASValsSetColor (yellow);
    EASValsSetShrink (0.95);
    EASValsSetLayer (EFG_SURFACE_LAYER);
    EGWithMaskChangeAttributes (COLOR_MASK | LAYER_MASK | SHRINK_MASK,
                               t->triangle_neg_normal);
    EMNoDrawAddGraphicsToModel (ESIModel (), t->triangle_neg_normal);
  } 

  FOR (i, 3) {
    vert_loc (t->vertex[i], tc, &ploc, &nloc);
    ppnts[i].x = (t->vertex[i]->loc.x
                  + displ_scale * (ploc.x - t->vertex[i]->loc.x));
    ppnts[i].y = (t->vertex[i]->loc.y
                  + displ_scale * (ploc.y - t->vertex[i]->loc.y));
    ppnts[i].z = (t->vertex[i]->loc.z
                  + displ_scale * (ploc.z - t->vertex[i]->loc.z));
    npnts[i].x = (t->vertex[i]->loc.x
                  + displ_scale * (nloc.x - t->vertex[i]->loc.x));
    npnts[i].y = (t->vertex[i]->loc.y
                  + displ_scale * (nloc.y - t->vertex[i]->loc.y));
    npnts[i].z = (t->vertex[i]->loc.z
                  + displ_scale * (nloc.z - t->vertex[i]->loc.z));
  } 
  EMEraseGraphics (ESIModel (), t->triangle_pos_normal);
  EMEraseGraphics (ESIModel (), t->triangle_neg_normal);
  EGModifyGraphicsGeometry (t->triangle_pos_normal, ppnts);
  EGModifyGraphicsGeometry (t->triangle_neg_normal, npnts);
  EMDrawGraphics (ESIModel (), t->triangle_pos_normal);
  EMDrawGraphics (ESIModel (), t->triangle_neg_normal);
} 
#endif /* X_XFACE  */

static WEFG_pt_pos_wrt_crack_surf_enum
node_pos_wrt_crack_surf (WEFG_crack_surf_t *cs,
                        WEFG_node_t *efgn,
                        double physical_tol) 
{
  /* RAW */
  return WEFG_POS_WRT_SURF_UNSPEC;
} 


static void
vert_loc (ts_vert_t *v,
         triangulated_crack_surf_t *tc /* instance of triangulated crack */, 
         W_loc_t *pos_normal_loc, W_loc_t *neg_normal_loc) 
{
  double x, y, z;
  
  vert_sftabs (v, tc);

  efgsf_eval_curr_loc (&v->sftab_pos_normal, &x, &y, &z);
  pos_normal_loc->x = x; pos_normal_loc->y = y; pos_normal_loc->z = z;
  
  efgsf_eval_curr_loc (&v->sftab_neg_normal, &x, &y, &z);
  neg_normal_loc->x = x; neg_normal_loc->y = y; neg_normal_loc->z = z;
} 



static void
vert_vel (ts_vert_t *v,
         triangulated_crack_surf_t *tc /* instance of triangulated crack */, 
         W_loc_t *pos_normal_vel, W_loc_t *neg_normal_vel) 
{
  double x, y, z;
  
  vert_sftabs (v, tc);

  efgsf_eval_curr_vel (&v->sftab_pos_normal, &x, &y, &z);
  pos_normal_vel->x = x; pos_normal_vel->y = y; pos_normal_vel->z = z;
  
  efgsf_eval_curr_vel (&v->sftab_neg_normal, &x, &y, &z);
  neg_normal_vel->x = x; neg_normal_vel->y = y; neg_normal_vel->z = z;
} 



static void
vert_acc (ts_vert_t *v,
         triangulated_crack_surf_t *tc /* instance of triangulated crack */, 
         W_loc_t *pos_normal_acc, W_loc_t *neg_normal_acc) 
{
  double x, y, z;
  
  vert_sftabs (v, tc);

  efgsf_eval_curr_acc (&v->sftab_pos_normal, &x, &y, &z);
  pos_normal_acc->x = x; pos_normal_acc->y = y; pos_normal_acc->z = z;
  
  efgsf_eval_curr_acc (&v->sftab_neg_normal, &x, &y, &z);
  neg_normal_acc->x = x; neg_normal_acc->y = y; neg_normal_acc->z = z;
} 




/*
   Evaluate the shape function table.
   Evaluate the shape function table at the vertex.

   Shape function table is evaluated only if it is `stale', ie. only if
   its time stamp indicates it is outdated by crack growth.  Note that
   there is a table for both the plus and minus side of the crack eventhough
   the vertex may be at the front.

   Assumptions:

   Crack must not pass through an interface cell.  An error is reported in that case.

   Warning:

   If the shape function table could not be evaluated, a warning is reported
   with message coming from the ESFLib indicating the reasons, and the update
   of the table is skipped (old values are kept in the table).
 */
static void
vert_sftabs (ts_vert_t *v, /* vertex */
             triangulated_crack_surf_t *tc /* crack surface */
             ) 
{
  WEFG_eval_for_packet_t efp;
  SFL_sf_table_t sf;
  WEFG_hex_volume_t *eh;

  efp.se                = tc->crack_surf->se;
  efp.crack_surf        = tc->crack_surf;
  efp.crack_surf_normal = v->normal;


  if (! (   SFTAB_IS_STALE (v->sftab_pos_normal, efp.se) 
        || SFTAB_IS_STALE (v->sftab_neg_normal, efp.se)) ) 
    return;

  sf = efgsf_sf_table (efp.se);

  eh = v->enclosing_hex;
  if (eh == NULL) {
    if (!find_hex_enclosing_loc (tc, &v->loc, &v->enclosing_hex, &v->ploc)) {
      sprintf (W_buf (), "No enclosing hex at [%g;%g;%g]",
              v->loc.x, v->loc.y, v->loc.z);
      REPORT_FATAL_ERROR (tc->crack_surf->se->domain->errh, W_buf ());
    }  else  {
      eh = v->enclosing_hex;
    } 
  } 
  
  if (SFTAB_IS_EMPTY (v->sftab_neg_normal) ||
      SFTAB_DAMAGED_INSIDE_INV_REGION_ONLY (v->sftab_neg_normal,
                                            efp.se, &v->loc)) {
    if (v->is_on_front) {
      efp.pt_wrt_crack_surf_classification = WEFG_AT_FRONT;
    }  else  {
      efp.pt_wrt_crack_surf_classification = WEFG_NEGATIVE_NORMAL; } 
    if (!SFTAB_IS_EMPTY (v->sftab_neg_normal)) {
      USE_SFTAB_SEARCH_FUNC (efp.se, v->sftab_neg_normal);
    } 
    if (hex_is_interface_vol (eh)) {
      REPORT_FATAL_ERROR (tc->crack_surf->se->domain->errh, "Crack is not allowed to pass through interface");
    }  else  {
      if (!SFL_shape_func_at (sf, v->loc.x, v->loc.y, v->loc.z, &efp, NULL)) {
        REPORT_WARNING_START (tc->crack_surf->se->domain->errh);
        sprintf (W_buf (), "At  {%g;%g;%g} \n", v->loc.x, v->loc.y, v->loc.z);
        WEH_error_msg (tc->crack_surf->se->domain->errh, W_buf ());
        WEH_error_msg (tc->crack_surf->se->domain->errh,
                     (char *) ESFL3_explain_error (ESFL3_sf_lib (sf)) );
        sprintf (W_buf (), "# nodes %d\n",
                 ESFL3_num_of_entries (sf));
        WEH_error_msg (tc->crack_surf->se->domain->errh, W_buf ());
        sprintf (W_buf (), "Vertex is %s\n",
                 (efp.pt_wrt_crack_surf_classification == WEFG_AT_FRONT
                  ? "at the front"
                  : (efp.pt_wrt_crack_surf_classification == WEFG_NEGATIVE_NORMAL
                     ? "on the minus side"
                     : "on the plus side")) );
        WEH_error_msg (tc->crack_surf->se->domain->errh, W_buf ());
        sprintf (W_buf (), "normal  {%g;%g;%g} \n",
                 v->normal.x, v->normal.y, v->normal.z);
        WEH_error_msg (tc->crack_surf->se->domain->errh, W_buf ());
        REPORT_WARNING_END (tc->crack_surf->se->domain->errh);
        goto skip_update_neg;
      } 
    } 
    efgsf_create_sftab (efp.se, &v->sftab_neg_normal, sf, NULL, FALSE);
  skip_update_neg:
    USE_DEFAULT_SEARCH_FUNC (efp.se);
    UPDATE_EFG_SF_STAT (efp.se, SFTAB_NUM_OF_ENTRIES (v->sftab_neg_normal));
  }  /* if (SFTAB_INVALID (v->sftab_neg_normal,... */
  TOUCH_SFTAB (v->sftab_neg_normal, efp.se);
  
  if (SFTAB_IS_EMPTY (v->sftab_pos_normal) ||
      SFTAB_DAMAGED_INSIDE_INV_REGION_ONLY (v->sftab_pos_normal,
                                            efp.se, &v->loc)) {
    if (v->is_on_front) {
      efp.pt_wrt_crack_surf_classification = WEFG_AT_FRONT;
    }  else  {
      efp.pt_wrt_crack_surf_classification = WEFG_POSITIVE_NORMAL; } 
    if (!SFTAB_IS_EMPTY (v->sftab_pos_normal)) {
      USE_SFTAB_SEARCH_FUNC (efp.se, v->sftab_pos_normal);
    } 
    if (hex_is_interface_vol (eh)) {
      REPORT_FATAL_ERROR (tc->crack_surf->se->domain->errh,
                          "Crack is not allowed to pass through interface");
    }  else  {
      if (!SFL_shape_func_at (sf, v->loc.x, v->loc.y, v->loc.z, &efp, NULL)) {
        REPORT_WARNING_START (tc->crack_surf->se->domain->errh);
        sprintf (W_buf (), "At  {%g;%g;%g} \n", v->loc.x, v->loc.y, v->loc.z);
        WEH_error_msg (tc->crack_surf->se->domain->errh, W_buf ());
        WEH_error_msg (tc->crack_surf->se->domain->errh,
                       (char *) ESFL3_explain_error (ESFL3_sf_lib (sf)) );
        sprintf (W_buf (), "# nodes %d\n",
                 ESFL3_num_of_entries (sf));
        WEH_error_msg (tc->crack_surf->se->domain->errh, W_buf ());
        sprintf (W_buf (), "Vertex is %s\n",
                 (efp.pt_wrt_crack_surf_classification == WEFG_AT_FRONT
                  ? "at the front"
                  : (efp.pt_wrt_crack_surf_classification == WEFG_NEGATIVE_NORMAL
                     ? "on the minus side"
                     : "on the plus side")) );
        WEH_error_msg (tc->crack_surf->se->domain->errh, W_buf ());
        sprintf (W_buf (), "normal  {%g;%g;%g} \n",
                 v->normal.x, v->normal.y, v->normal.z);
        WEH_error_msg (tc->crack_surf->se->domain->errh, W_buf ());
        REPORT_WARNING_END (tc->crack_surf->se->domain->errh);
        goto skip_update_pos;
      } 
    } 
    efgsf_create_sftab (efp.se, &v->sftab_pos_normal, sf, NULL, FALSE);
  skip_update_pos:
    USE_DEFAULT_SEARCH_FUNC (efp.se);
    UPDATE_EFG_SF_STAT (efp.se, SFTAB_NUM_OF_ENTRIES (v->sftab_pos_normal));
  }  /* if (SFTAB_INVALID (v->sftab_pos_normal, */
  TOUCH_SFTAB (v->sftab_pos_normal, efp.se);
} 


static void
v_attach_e (ts_vert_t *v, ts_edge_t *e) 
{
  if (!in_list (v->edges, e)) {
    add_to_tail (v->edges, e);
  } 
} 



static void
v_detach_e (ts_vert_t *v, ts_edge_t *e) 
{
  delete_list_node (v->edges, e);
} 



static void
e_attach_t (ts_edge_t *e, ts_tri_t *t) 
{
  if (e->triangles[0] == t || e->triangles[1] == t) 
    REPORT_FATAL_ERROR (NULL, "attaching a triangle twice");
  if (e->triangles[0] == NULL) 
    e->triangles[0] = t;
  else if (e->triangles[1] == NULL) 
    e->triangles[1] = t;
  else
    REPORT_FATAL_ERROR (NULL, "attaching too many triangles");
} 



static void
e_detach_t (ts_edge_t *e, ts_tri_t *t) 
{
  if (e->triangles[0] == t) 
    e->triangles[0] = NULL;
  else if (e->triangles[1] == t) 
    e->triangles[1] = NULL;
} 




static void
forget_tri (triangulated_crack_surf_t *tc /* instance of triangulated crack */,
            ts_tri_t *t) 
{
  e_detach_t (t->edges[0], t);
  v_detach_e (t->vertex[0], t->edges[0]);
  v_detach_e (t->vertex[1], t->edges[0]);
  
  e_detach_t (t->edges[1], t);
  v_detach_e (t->vertex[1], t->edges[1]);
  v_detach_e (t->vertex[2], t->edges[1]);

  e_detach_t (t->edges[2], t);
  v_detach_e (t->vertex[2], t->edges[2]);
  v_detach_e (t->vertex[0], t->edges[2]);

  delete_list_node (tc->triangles, t);

#ifdef X_XFACE
  if (t->triangle_neg_normal) 
    EMDeleteGraphics (ESIModel (), t->triangle_neg_normal);
  if (t->triangle_pos_normal) 
    EMDeleteGraphics (ESIModel (), t->triangle_pos_normal);
#endif
} 


static void
unify_normals (triangulated_crack_surf_t *tc);
static void
update_min_max_edge_length (triangulated_crack_surf_t *tc, ts_edge_t *e);
static void
initial_normals (triangulated_crack_surf_t *tc);

/*
   Check definition of triangulated crack surface.

   Check that the triangulated crack surface has been defined
   properly.  In particular, report error if (i) no vertices have
   been defined on the front; (ii) each front vertex is connected
   to at most two other front vertices (we cannot handle branching cracks);
   (iii)

   As a side effect, the triangle normals are re-oriented if needed to make
   the orientation of all surface triangles on the crack consistent.  Also,
   the minimum and maximum edge lengths are computed, and the initial list
   of on-the-front vertices is created.
 */
static void
check_defs (triangulated_crack_surf_t *tc) 
{
  int num_of_vert_on_front = 0;
  ts_vert_t *v;
  ts_edge_t *e;
  int nn;

  /* Must have some vertices on the front */
  loop_through_list (tc->vertices, v, ts_vert_t *) {
    if (v->is_on_front) num_of_vert_on_front++;
  } 
  if (num_of_vert_on_front == 0) 
    REPORT_FATAL_ERROR (tc->crack_surf->se->domain->errh, "No vertices on the crack front");

  /* Each vertex is supposed to be connected to only two other vertices */
  /* on the front.  We cannot handle branching fronts yet. */
  loop_through_list (tc->vertices, v, ts_vert_t *) {
    if (v->is_on_front) {
      nn = 0; /* zero neighbors */
      loop_through_list (v->edges, e, ts_edge_t *) {
        if (THE_OTHER_VERTEX (e, v) ->is_on_front) nn++;
      } 
      if (nn > 2) REPORT_FATAL_ERROR (tc->crack_surf->se->domain->errh,
                                      "Too many neighbors on the front");
    } 
  } 
  
  /* Unify normals */ 
  unify_normals (tc);

  /* Find out the lengths of the edges */
  loop_through_list (tc->edges, e, ts_edge_t *) {
    update_min_max_edge_length (tc, e);
  } 

  /* Assign normals to initial surface vertices */
  initial_normals (tc);

  /* Create the initial list of on-the-front vertices */
  tc->on_front_vertices = make_list ();
  loop_through_list (tc->vertices, v, ts_vert_t *) {
    if (v->is_on_front) {
      add_to_tail (tc->on_front_vertices, v);
      compute_axes (v, tc);
    } 
  }  
} 


/*
   Compute vertex normals.
   Compute vertex normals for crack vertices.
   Error is reported if a normal cannot be evaluated for any vertex .
 */
static void
initial_normals (triangulated_crack_surf_t *tc) 
{
  ts_vert_t *v;
  ts_tri_t *t;
  double n;
  int i;

  loop_through_list (tc->vertices, v, ts_vert_t *) {
    VECMAC_3_SCALAR (v->normal, =, 0, 0, 0);
  } 
  
  loop_through_list (tc->triangles, t, ts_tri_t *) {
    FOR (i, 3) {
      VECMAC_2_VEC (t->vertex[i]->normal, +=, t->normal);
    } 
  } 
  
  loop_through_list (tc->vertices, v, ts_vert_t *) {
    n = VECMAC_NORM (v->normal);
    if (n <= 0) {
      sprintf (W_buf (), "No normal for a crack surface vertex %d",
               (int) v->id);
      REPORT_FATAL_ERROR (tc->crack_surf->se->domain->errh, W_buf ());
    } 
    VECMAC_NORMALIZE (v->normal);
  } 
} 


static void
update_min_max_edge_length (triangulated_crack_surf_t *tc /* instance of triangulated crack */, ts_edge_t *e) 
{
  double el;
  W_loc_t t;

  VECMAC_3_VEC (t, =, e->vertex[1]->loc, -, e->vertex[0]->loc);
  el = VECMAC_NORM (t);
  if (E_IS_ON_FRONT (e)) {
    if (tc->min_edge_length > 0.5 * el) {
      tc->min_edge_length = 0.5 * el;
    } 
    if (tc->max_edge_length < 2 * el) {
      tc->max_edge_length = 2 * el;
    } 
  } 
} 



static int
the_same_orientation (ts_edge_t *e, ts_tri_t *t0, ts_tri_t *t1) 
{
  int eix0 = edge_idx (t0, e);
  int eix1 = edge_idx (t1, e);
  ts_vert_t *v00, *v10;
  ts_vert_t *v01, *v11;

  find_edge_vertices (t0, eix0, &v00, &v10);
  find_edge_vertices (t1, eix1, &v01, &v11);

  return (v00 == v11 && v10 == v01);
} 

static void
reverse_normal (ts_tri_t *t) 
{
  VECMAC_3_SCALAR (t->normal, =, -t->normal.x, -t->normal.y, -t->normal.z);
} 


/*
   Unify normals.

   Re-orient normals of some triangles if needed
   to produce consistent surface orientation.
 */
static void
unify_normals (triangulated_crack_surf_t *tc) 
{
  ts_tri_t *t, *nt;
  int nskipped = 0, done;
  int i;

  t = (ts_tri_t *) get_list_head (tc->triangles);
  t->unified_normal = TRUE;

 again:
  nskipped = 0;
  loop_through_list (tc->triangles, t, ts_tri_t *) {
    if (!t->unified_normal) {
      done = FALSE;
      FOR (i, 3) {
        nt = THE_OTHER_TRI (t->edges[i], t);
        if (nt) {
          if (nt->unified_normal) {
            if (!the_same_orientation (t->edges[i], t, nt)) 
              reverse_normal (t);
            t->unified_normal = TRUE;
            done = TRUE;
            break;
          } 
        } 
      }  /* FOR (i, 3) */
      if (!done) 
        nskipped++;
    }  /* if (!t->unified_normal) */
  } 
  
  if (nskipped > 0) 
    goto again;
} 


/*
   Find previous and next front vertex.

   Find previous and next vertex on the crack front.  Which is previous
   and which is next is given by the orientation of the triangles.
   The vertices on a triangle are traversed ccw when looking against its normal.
 */
static int
find_prev_next_vertex_on_front (ts_vert_t *v, /* vertex */
                                triangulated_crack_surf_t *tc /* instance of triangulated crack */,
                                ts_vert_t **prevv, /* output: previous vertex on front */
                                ts_vert_t **nextv /* output: next vertex on front */
                                ) 
{
  ts_edge_t *e;
  int eix;
  ts_vert_t *v0, *v1, *tov;

  *prevv = NULL;
  *nextv = NULL;
  loop_through_list (v->edges, e, ts_edge_t *) {
    tov = THE_OTHER_VERTEX (e, v);
    if (tov->is_on_front) {
      if (e->triangles[0] && e->triangles[1]) {
        REPORT_FATAL_ERROR (tc->crack_surf->se->domain->errh,
                            "edge on front used by two triangles");
      } 
      if (e->triangles[0] == NULL) REPORT_FATAL_ERROR (tc->crack_surf->se->domain->errh,
                                                       "no triangle");
      eix = edge_idx (e->triangles[0], e);
      if (eix < 0) REPORT_FATAL_ERROR (tc->crack_surf->se->domain->errh,
                                       "edge not in triangle");
      find_edge_vertices (e->triangles[0], eix, &v0, &v1);
      if (v0 == v) *nextv = v1;
      if (v1 == v) *prevv = v0;
    } 
  } 
  return TRUE;
} 


static int
edge_idx (ts_tri_t *t, ts_edge_t *e) 
{
  if      (t->edges[0] == e) return 0;
  else if (t->edges[1] == e) return 1;
  else if (t->edges[2] == e) return 2;
  else                       return -1;
} 



static int
find_edge_vertices (ts_tri_t *t, int edge,
                   ts_vert_t **v0,
                    ts_vert_t **v1
                    ) 
{
  if (t->edges[edge]->vertex[0] == t->vertex[edge]) {
    *v0 = t->edges[edge]->vertex[0];
    *v1 = t->edges[edge]->vertex[1];
  }  else if (t->edges[edge]->vertex[1] == t->vertex[edge]) {
    *v0 = t->edges[edge]->vertex[1];
    *v1 = t->edges[edge]->vertex[0];
  } 
  return TRUE;
} 


static void
predict_advance (ts_vert_t *v,
                 ts_vert_t *prevv, ts_vert_t *nextv,
                 ts_edge_t *preve, ts_edge_t *nexte, 
                 triangulated_crack_surf_t *tc);


static void
add_changed_triangles_to_bbt (ts_vert_t *v, triangulated_crack_surf_t *tc) 
{
  ts_edge_t *e;
  int i;
  ts_tri_t *t;
  static LIST_OF (ts_tri_t *) conn_tris = NULL;

  if (conn_tris == NULL) conn_tris = make_list ();
  flush_list (conn_tris, NOT_NODES);

  loop_through_list (v->edges, e, ts_edge_t *) {
    FOR (i, 2) {
      t = e->triangles[i];
      if (!in_list (conn_tris, t)) 
        add_to_tail (conn_tris, t);
    } 
  } 

  loop_through_list (conn_tris, t, ts_tri_t *) {
    add_to_bbt (t, tc);
  } 
} 

static void
trim_advance_vector (ts_vert_t *v, /* cloned vertex */
                     triangulated_crack_surf_t *tc) ;
/*
   Trim the advance vector.
   If the input vertex has a source vertex, the advance vector of the source
   is trimmed by boundary faces.
   As a side-effect, the enclosing integration volume is found.
 */
static void
trim_advance (ts_vert_t *newv, /* clone of vertex whose advance vector should be trimmed */
              triangulated_crack_surf_t *tc /* triangulated crack */
              ) 
{
  if (newv->source_vert) 
    trim_advance_vector (newv, tc);

  if (newv->enclosing_hex == NULL) {
    if (!find_hex_enclosing_loc (tc, &newv->loc,
                                 &newv->enclosing_hex, &newv->ploc)) {
      sprintf (W_buf (), "No enclosing hex at [%g;%g;%g]",
              newv->loc.x, newv->loc.y, newv->loc.z);
      REPORT_FATAL_ERROR (tc->crack_surf->se->domain->errh, W_buf ());
    } 
  } 
} 

/*
   Trim advance vector.
   Trim advance vector of the source vertex of the vertex on input.
   Let us denote the advance vector of the source vertex as V.  
   The algorithm is as follows:
   (i) Find all the intersections of V with the boundary faces.
   (ii) If there are none, the algorithm terminates; otherwise the intersection
   closest to the source vertex is chosen, and the intersected
   face is returned.
   (iii) If an intersection has been found, the location of the cloned vertex
   is set to the location of the intersection, the vertex is marked
   as being on the boundary, and the face pointer is stored
   in the cloned vertex.

   Precondition:

   The vertex on input *must* have a source vertex.
 */
static void
trim_advance_vector (ts_vert_t *v, /* cloned vertex */
                     triangulated_crack_surf_t *tc) 
{
  G3D_segment_t s;
  W_loc_t ip;
  WEFG_b_face_t *f;
  double advv_mult = tc->trim_advv_mult;

  /* check for intersection of the advance vector with the boundary */
  s.start = v->source_vert->loc;
  VECMAC_2_VEC (s.end, =, v->loc);
  f = efg_find_intersection_with_bface (tc->crack_surf->se, &s,
                                        -0.001, advv_mult, &ip,
                                        WEFG_B_FACE_CLOSEST_TO_START);
  if (f) {
    v->loc             = ip;
    v->is_on_boundary  = TRUE;
    v->on_face         = f;
  } 
} 


/*
   Trim front edge.
   Trim crack front edge by the boundary faces.
   The intersections of the tangent vector emanating from the input vertex
   with the boundary are found.  Only intersection within the
   distance equal to the length of the edge are considered.  
 */
static void
trim_bdry_edge (ts_vert_t *v,
                ts_vert_t *prevv, ts_vert_t *nextv,
                triangulated_crack_surf_t *tc
                ) 
{
  G3D_segment_t s;
  WEFG_b_face_t *f;
  W_loc_t ipn, ipp;
  W_bool_t ipnvalid = FALSE, ippvalid = FALSE;
  double dp = FLT_MAX, dn = FLT_MAX;
  double sl, dist, rat;
  
  /* check for intersection of an edge on the front going from this
     vertex with the boundary.  
     */
  s.start = v->loc;
  VECMAC_3_VEC (s.end, =, v->loc, +, v->source_vert->tangent);
  sl = v3_distance (&s.start, &s.end);
  if (prevv && !prevv->is_on_boundary) {
    dist = v3_distance (&v->loc, &prevv->loc);
    rat  = dist / sl; rat = max (2, rat);
    f = efg_find_intersection_with_bface (tc->crack_surf->se, &s,
                                          -rat, rat, &ipp,
                                          WEFG_B_FACE_CLOSEST_TO_START);
    if (f) {
      ippvalid   = TRUE;
      dp         = v3_distance (&v->loc, &ipp);
      v->on_face = f;
    } 
  } 
  if (nextv && !nextv->is_on_boundary) {
    dist = v3_distance (&v->loc, &nextv->loc);
    rat  = dist / sl; rat = max (2, rat);
    f = efg_find_intersection_with_bface (tc->crack_surf->se, &s,
                                          -rat, rat, &ipn,
                                          WEFG_B_FACE_CLOSEST_TO_START);
    if (f) {
      ipnvalid   = TRUE;
      dn         = v3_distance (&v->loc, &ipn);
      v->on_face = f;
    } 
  } 
  if (ippvalid && ipnvalid) {
    v->is_on_boundary  = TRUE;
    if (dp < dn) { v->loc = ipp;
    }  else        { v->loc = ipn;} 
  }  else if (ippvalid && !ipnvalid) {
    v->loc             = ipp;
    v->is_on_boundary  = TRUE;
  }  else if (ipnvalid && !ippvalid) {
    v->loc             = ipn;
    v->is_on_boundary  = TRUE;
  } 
} 



/*
   Compute prescribed advance vector.
   Evaluate advance vector at a front vertex from a symbolic expression
   for each component of the crack speed which may involve time (t), time step
   (dt), and X, Y, Z.
 */
static void
advance_prescribed (ts_vert_t *v, /* front vertex */
                    triangulated_crack_surf_t *tc /* triangulated crack */
                    ) 
{
  WEFG_crack_surf_t *cs = tc->crack_surf;
  double t  = WD_time (cs->se->domain);
  double dt = WD_time_step (cs->se->domain);
  int stat;
  double f;
  double C2 = 0;
  
  expeval_set_value ("X", v->loc.x);
  expeval_set_value ("Y", v->loc.y);
  expeval_set_value ("Z", v->loc.z);
  expeval_set_value ("t", t);
  expeval_set_value ("dt", dt);
  
  stat = expeval_eval (tc->advance_velocity[0], &f);
  if (stat != EXPEVAL_E_OK) {
    sprintf (W_buf (), "%s", expeval_err_msgs[stat]);
    REPORT_ERROR (tc->crack_surf->se->domain->errh, W_buf ());
  } 
  C2 += f * f;
  v->advance_vector.x = f * dt;

  stat = expeval_eval (tc->advance_velocity[1], &f);
  if (stat != EXPEVAL_E_OK) {
    sprintf (W_buf (), "%s", expeval_err_msgs[stat]);
    REPORT_ERROR (tc->crack_surf->se->domain->errh, W_buf ());
  } 
  C2 += f * f;
  v->advance_vector.y = f * dt;

  stat = expeval_eval (tc->advance_velocity[2], &f);
  if (stat != EXPEVAL_E_OK) {
    sprintf (W_buf (), "%s", expeval_err_msgs[stat]);
    REPORT_ERROR (tc->crack_surf->se->domain->errh, W_buf ());
  } 
  C2 += f * f;
  v->advance_vector.z = f * dt;
  v->C = sqrt (C2);
} 

/*
   Project vertex on face.
   Project vertex on boundary face.
 */
static void
project_on_face (ts_vert_t *v /* vertex */
                 ) 
{
  double avl = VECMAC_NORM (v->advance_vector);
  double ncn, projadvn;
  W_vect_t fnormal, projadv, zerovect= {0,0,0} ;

  if (avl <= 0) return;
  
  /* project the advance vector on the face */
  efggc_face_normal (v->on_face, &fnormal);
  ncn = VECMAC_DOT (v->advance_vector, fnormal);
  VECMAC_3_VEC (projadv, =, v->advance_vector, -, ncn * fnormal);
  projadvn = VECMAC_NORM (projadv);
  if (projadvn < 0.2 * avl) {
    v->advance_vector = zerovect;
  }  else  {
    v->advance_vector = projadv;
#if 0
    VECMAC_NORMALIZE (v->advance_vector);
    VECMAC_SCALAR (v->advance_vector, *=, avl);
#endif
  } 
} 


/*
   Compute normal at vertex.
   The normal unit vector at a crack vertex is evaluated
   as the normalized average of normals of the connected triangles.
   If of none of the triangles has a  valid normal, the normal of
   the source vertex is adopted.
 */
static void
compute_normal (ts_vert_t *v /* vertex */
                ) 
{
  ts_edge_t *e;
  int i;
  ts_tri_t *t;
  W_vect_t n =  {0,0,0} ;
  double nn;

  loop_through_list (v->edges, e, ts_edge_t *) {
    FOR (i, 2) {
      t = e->triangles[i];
      if (t) {
        VECMAC_2_VEC (n, +=, t->normal);
      } 
    } 
  } 
  nn = VECMAC_NORM (n);
  if (nn <= 0) {
    if (v->source_vert) {
      n = v->source_vert->normal;
    }  else   {
      REPORT_FATAL_ERROR (NULL, "No valid normal available");
    } 
  }  
  VECMAC_NORMALIZE (n);
  v->normal = n;
} 


/*
   Advance vertex.
   The crack front vertex is propagated (cloned) by this algorithm:
   (i) The ideal motion increment vector is predicted.
   (ii) If the vertex is bound to a boundary face, the advance vector
   is projected on this face.
   (iii) If the advance vector is non-zero, (a) a new vertex is created (it is
   the clone of the vertex being `moved'); (b) the advance vector is trimmed
   by the boundary faces; (c) new edges and triangles are created as necessary;
   (d) regions of invalidated connectivities are created; 
   (e) the crack surface is marked as changed.
 */
static void
advance_vertex (ts_vert_t *v, /* vertex to clone */
                triangulated_crack_surf_t *tc /* triangulated crack surface */
                ) 
{
  ts_vert_t *prevv, *nextv, *newv;
  W_loc_t newloc;
  ts_edge_t *preve, *nexte, *newe, *newpe, *newne;
  ts_tri_t *t;
  G3D_tri_t tria;
#undef  UPD_INVALID_REG
#define UPD_INVALID_REG(l0, l1, l2)                                 \
  {                                                                 \
    tria.loc[0] = l0; tria.loc[1] = l1; tria.loc[2] = l2;           \
    SHRINK_TRI_LOC (EXPTRI, tria.loc[0], tria.loc[1], tria.loc[2]); \
    efg_add_invalid_region (tc->crack_surf->se, 1, &tria);          \
  } 

#define ADVANCE_NON_ZERO(v) ((v->advance_vector.x != 0 || v->advance_vector.y != 0 || v->advance_vector.z != 0)) 

  prevv = NULL;
  nextv = NULL;
  find_prev_next_vertex_on_front (v, tc, &prevv, &nextv);
  if (prevv) { preve = connecting_edge (prevv, v); } 
  else                { preve = NULL; } 
  if (nextv) { nexte = connecting_edge (nextv, v);} 
  else                { nexte = NULL; } 

  predict_advance (v, prevv, nextv, preve, nexte, tc);

  if (v->on_face) 
    project_on_face (v);

  if (ADVANCE_NON_ZERO (v)) {
    VECMAC_3_VEC (newloc, =, v->loc, +, v->advance_vector);
    newv = new_vert (tc, 0, &newloc);
    add_to_tail (tc->vertices, newv);
    newv->source_vert = v;
    newv->ms_neg_normal
      = efg_interpolate_mat_state (tc->crack_surf->se,
				  &newv->loc,
                                  tc->crack_surf->se->max_support_size / 10);
    if (newv->ms_neg_normal == NULL) REPORT_ERROR (tc->crack_surf->se->domain->errh,
                                                   "No material state");
    newv->ms_pos_normal
      = efg_interpolate_mat_state (tc->crack_surf->se,
				  &v->loc,
                                  tc->crack_surf->se->max_support_size / 10);
    if (newv->ms_pos_normal == NULL) REPORT_ERROR (tc->crack_surf->se->domain->errh,
                                                   "No material state");

    newv->is_on_front     = TRUE;
    newv->is_on_boundary  = v->is_on_boundary;
    newv->K_I_0           = v->K_I_0;
    newv->K_II_0          = v->K_II_0;
    newv->K_III_0         = v->K_III_0;
    newv->K_I_c           = v->K_I_c;
    newv->K_II_c          = v->K_II_c;
    newv->K_III_c         = v->K_III_c;
    newv->K_I_equiv       = v->K_I_equiv;
    newv->theta           = v->theta;
    newv->C               = v->C;
    newv->G_0             = v->G_0;
    newv->G_c             = v->G_c;
    
    trim_advance (newv, tc);

    newe  = new_edge (tc, 0, newv, v);
    add_to_tail (tc->edges, newe);

    if (prevv) {
      newpe = new_edge (tc, 0, prevv, newv); add_to_tail (tc->edges, newpe);
      t = new_tri (tc, 0, v, prevv, newv, preve, newpe, newe);
      add_to_head (tc->triangles, t);
      UPD_INVALID_REG (v->loc, prevv->loc, newv->loc);
    }  
    if (nextv) {
      newne = new_edge (tc, 0, nextv, newv); add_to_tail (tc->edges, newne);
      t = new_tri (tc, 0, v, newv, nextv, newe, newne, nexte);
      add_to_head (tc->triangles, t);
      UPD_INVALID_REG (v->loc, newv->loc, nextv->loc);
    } 
    v->is_on_front        = FALSE;
    STALE_SFTAB (v->sftab_pos_normal);
    STALE_SFTAB (v->sftab_neg_normal);
    compute_normal (newv);
    if (v->hist_cd) {
      newv->hist_cd    = v->hist_cd;
      v->hist_cd->vert = newv;
    } 
    if (tc->master_vert == v) tc->master_vert = newv;
    tc->crack_surf->changed_shape = TRUE;
  } 
#undef  UPD_INVALID_REG
} 

static void
eval_A_k (double C, double nu, double c_d, double c_s, double C_R,
          double C_R_fraction,
          double *A_I, double *A_II, double *A_III)
{
  double alpha_d, alpha_s, D;

  if (C > C_R_fraction * C_R) {
    C = C_R_fraction * C_R;
  }
  
  alpha_d = sqrt (1 - (C * C) / (c_d * c_d));
  alpha_s = sqrt (1 - (C * C) / (c_s * c_s));
  D       = 4 * alpha_d * alpha_s - (1 + alpha_s * alpha_s) * (1 + alpha_s * alpha_s);
  if (D > 0) {
    *A_I     = (C * C) * alpha_d / ((1 - nu) * c_s * c_s * D);
    *A_II    = (C * C) * alpha_s / ((1 - nu) * c_s * c_s * D);
    *A_III   = 1 / alpha_s; 
  } else {
    *A_I = *A_II = *A_III = 1;
  }
}


/*
   Evalute SIF's using interaction integrals.

   RAW
   First compute the integration volume sizes.  They may be either
   specified in input, or can be computed from the lengths of front edges
   and max support size.
   Then set up the coordinate system to use.  That may be either the
   local triad at the vertex (for a free vertex), or triad constrained
   by projection to a boundary face (for a boundary vertex).
   Finally call efgii_sifs_volume() to evaluate the integrals, scale
   by the line integrals of the virtual extension function, and compute
   the energy release rate.

   Note:

   J-integrals are not evaluated; they are arbitrarily set to zero.
   Static SIF's are set equal to the dynamic SIF's. 
 */
static void 
do_eval_ii_sifs (ts_vert_t *v, /* vertex */
                 ts_vert_t *prevv, /* previous vertex on front */
                 ts_vert_t *nextv, /* next vertex on front */
                 ts_edge_t *preve, /* edge connecting v and prevv */
                 ts_edge_t *nexte, /* edge connecting v and nextv */
                 triangulated_crack_surf_t *tc /* instance of triangulated crack */,
                 double E_star, /* effective Young's modulus */
                 double nu,
                 double mu, /* shear modulus */
                 double c_d,    /* dilatational wave speed */
                 double c_s,    /* shear wave speed */
                 double C_R, /* Rayleigh wave speed */
                 double C_R_fraction,
                 double C,      /* crack speed */
                 double *J1static, /* output: static J_1 */
                 double *J1dynamic, /* output: dynamic J_1 */
                 double *J2static, /* output: static J_2 */
                 double *J2dynamic, /* output: dynamic J_2 */
                 double *KI_static, /* output: static opening SIF */
                 double *KII_static, /* output: static shearing SIF */
                 double *KIII_static, /* output: static tearing SIF */
                 double *KI_dynamic, /* output: dynamic opening SIF */
                 double *KII_dynamic, /* output: dynamic shearing SIF */
                 double *KIII_dynamic, /* output: dynamic tearing SIF */
                 double *KI_equiv_static, /* output: equivalent SIF */
                 double *KI_equiv_dynamic, /* output: equivalent SIF */
                 double *theta, /* output: angle of tilt in the plane normal to the front;
                                 theta is measured from the forward vector */
                 double *G_static, /* output: energy release rate */
                 double *G_dynamic
                 )
{
  double tangent_plus, tangent_minus;
  double forward_plus, forward_minus;
  double normal_plus, normal_minus;
  double L, q_line_integral = 0;
  WEFG_tensor_prod_comp_func tensor_prod_comp_func = NULL;
  WEFG_tensor_prod_comp_der_func tensor_prod_comp_der_func = NULL;
  double distp, distn;
  W_loc_t d;
  W_vect_t forward, tangent, normal, cf={0,0,0}, fnormal;
  double theta_d;
  const W_bool_t DYNAMIC_II = TRUE;
  const W_bool_t STATIC_II = TRUE;
  double A_I, A_II, A_III;

  distp = 0;
  if (prevv) {
    VECMAC_3_VEC (d, =, v->loc, -, prevv->loc);
    distp = VECMAC_NORM (d);
  } 
  distn = 0;
  if (nextv) {
    VECMAC_3_VEC (d, =, v->loc, -, nextv->loc);
    distn = VECMAC_NORM (d);
  } 
  
  forward_minus = (tc->jdom_hsize_forward_minus > 0
                   ? tc->jdom_hsize_forward_minus
                   : tc->crack_surf->se->max_support_size);
  forward_plus  = (tc->jdom_hsize_forward_plus > 0
                   ? tc->jdom_hsize_forward_plus
                   : tc->crack_surf->se->max_support_size);
  if (tc->jdom_hsize_tangent_minus > 0) {
    tangent_minus = tc->jdom_hsize_tangent_minus;
  }  else if (distp > 0) { tangent_minus = distp;
  }  else  { tangent_minus = 0.5 * tc->crack_surf->se->max_support_size; } 
  if (tc->jdom_hsize_tangent_plus > 0) {
    tangent_plus = tc->jdom_hsize_tangent_plus;
  }  else if (distn > 0) { tangent_plus = distn;
  }  else  { tangent_plus = 0.5 * tc->crack_surf->se->max_support_size; } 
  normal_minus  = (tc->jdom_hsize_normal_minus > 0
                   ? tc->jdom_hsize_normal_minus
                   : tc->crack_surf->se->max_support_size);
  normal_plus   = (tc->jdom_hsize_normal_plus > 0
                   ? tc->jdom_hsize_normal_plus
                  : tc->crack_surf->se->max_support_size);
  L = 0;
  if (prevv) {
    L += tangent_minus;
  } 
  if (nextv) {
    L += tangent_plus;
  } 

  switch (WOPT_OPT (tc->crack_surf->se->domain->options, q_func_type)) {
  case W_QUARTIC_Q_FUNC:
    tensor_prod_comp_func       = W_f_quartic_spline;
    tensor_prod_comp_der_func   = W_f_r_quartic_spline;
    q_line_integral = 2. / 5. * L;
    break;
  case W_LINEAR_Q_FUNC:
    tensor_prod_comp_func       =  W_f_linear;
    tensor_prod_comp_der_func   = W_f_r_linear;
    q_line_integral = 1. / 2. * L;
    break;
  case W_QUADRATIC_Q_FUNC:
    tensor_prod_comp_func       = W_f_quadratic;
      tensor_prod_comp_der_func = W_f_r_quadratic;
    q_line_integral = 2. / 3. * L;
    break;
  case W_COMPOSITE_QUADRATIC_Q_FUNC:
    tensor_prod_comp_func       = W_f_composite_quadratic;
      tensor_prod_comp_der_func = W_f_r_composite_quadratic;
    q_line_integral = 1. / 2. * L;
    break;
  default:
    REPORT_ERROR (tc->crack_surf->se->domain->errh, "Unresolved virtual extension function");
    break;
  } 

  /* determine the local coordinate system */
  forward = v->forward; tangent = v->tangent; normal  = v->normal;
  if (v->is_on_boundary) {
    if (v->on_face) {
      efggc_face_normal (v->on_face, &fnormal);
      VECMAC_CROSS (cf, =, normal, x, fnormal);
      if (VECMAC_DOT (cf, forward) < 0) {
        VECMAC_2_VEC (cf, =, -cf);
      } 
      forward = cf;
      VECMAC_CROSS (tangent, =, normal, x, forward);
    } 
  } 
  
  efgii_sifs_volume (tc->crack_surf->se,
                     &v->loc, &forward, &tangent, &normal,
                     forward_plus, forward_minus,
                     tangent_plus, tangent_minus,
                     normal_plus, normal_minus,
                     tensor_prod_comp_func,
                     tensor_prod_comp_der_func,
                     E_star, tc->subsplit_for_2x2x2_ipol,
                     STATIC_II,
                     KI_static, KII_static, KIII_static,
                     KI_equiv_static, &theta_d
                     );
 
  *J1static  = *J1dynamic = *J2static  = *J2dynamic = 0;

  *KI_static       /= q_line_integral;
  *KII_static      /= q_line_integral;
  *KIII_static     /= q_line_integral;
  *KI_equiv_static /= q_line_integral;

  eval_A_k (C, nu, c_d, c_s, C_R, C_R_fraction, &A_I, &A_II, &A_III);
  *KI_dynamic   = *KI_static   / A_I;
  *KII_dynamic  = *KII_static  / A_II;
  *KIII_dynamic = *KIII_static / A_III;
  *KI_equiv_dynamic = W_equiv_KI (*KI_dynamic, *KII_dynamic, theta_d);

  *G_static
    = (( (*KI_equiv_static) * (*KI_equiv_static) / E_star) 
       + ((*KIII_static) * (*KIII_static) / 2 / mu) 
       );
  *G_dynamic
    = (( (*KI_equiv_dynamic) * (*KI_equiv_dynamic) / E_star) 
       + ((*KIII_dynamic) * (*KIII_dynamic) / 2 / mu) 
       );
  *theta = theta_d;
} 


static void
compute_E_star (triangulated_crack_surf_t *tc,
                       ts_vert_t *v,
                double E, double nu,
                       double *E_star);

/*
   Compute SIFS.
   Compute the SIF's at a front vertex.

   Algorithm:

   (i) Evaluate shape function table at the vertex.
   (ii) Compute the effective Young's modulus at the vertex.
   (iii) Evaluate the SIF's from interaction integrals.
   (iv) Store data in the vertex for subsequent use.
 */
static void
do_compute_sifs (ts_vert_t *v, /* front vertex */
                 ts_vert_t *prevv, /* previous vertex on the front (may be NULL) */
                 ts_vert_t *nextv,/* next vertex on the front (may be NULL) */
                 ts_edge_t *preve, /* edge connecting /v/ with the previous vertex */
                 ts_edge_t *nexte, /* edge connecting /v/ with the next vertex */
                 triangulated_crack_surf_t *tc /* triangulated crack */
                 ) 
{
  double G_0, G_c;
  double J1static, J1dynamic;
  double J2static, J2dynamic;
  double KI_static, KII_static, KIII_static;
  double KI_dynamic, KII_dynamic, KIII_dynamic;
  double KI_equiv_static, KI_equiv_dynamic, theta;
  double E, nu, rho, E_star, mu;
  double c_d, c_s, C_R;

  vert_sftabs (v, tc);

  j_setup (tc->crack_surf, v, 0, 0, 0, 0, 0, 0, 0, 0);   /* dummy setup... */

  if (!WM3D_get_mat_const (tc->crack_surf->se->mat,
                          MAT_3D_YOUNG_MODULUS, &E)) 
    REPORT_FATAL_ERROR (tc->crack_surf->se->domain->errh, "no shear modulus");
  if (!WM3D_get_mat_const (tc->crack_surf->se->mat,
                          MAT_3D_POISSON, &nu)) 
    REPORT_FATAL_ERROR (tc->crack_surf->se->domain->errh, "no Poisson ratio");
  if (!WM3D_get_mat_const (tc->crack_surf->se->mat,
                          MAT_3D_SHEAR_MODULUS, &mu)) 
    REPORT_FATAL_ERROR (tc->crack_surf->se->domain->errh, "no shear modulus");
  rho = WM3D_get_volume_mass (tc->crack_surf->se->mat);
  
  if (!WM3D_get_mat_const (tc->crack_surf->se->mat,
                           MAT_3D_RAYLEIGH_PSTRAIN_WAVE_SPEED, &C_R)) 
    REPORT_FATAL_ERROR (tc->crack_surf->se->domain->errh, "no Rayleigh wave speed");
  
  compute_E_star (tc, v, E, nu, &E_star);/*...to be able to compute */

  c_d = W_3D_dilatational_wave_speed (E, nu, rho);
  c_s = W_shear_wave_speed (E, nu, rho);
  do_eval_ii_sifs (v, prevv, nextv, preve, nexte, tc, E_star, nu, mu,
                   c_d, c_s, C_R, tc->rayleigh_vel_frac, v->C,
                   &J1static, &J1dynamic, &J2static, &J2dynamic,
                   &KI_static, &KII_static, &KIII_static,
                   &KI_dynamic, &KII_dynamic, &KIII_dynamic,
                   &KI_equiv_static, &KI_equiv_dynamic, &theta, &G_0, &G_c);
  /* SIF's for a dynamically loaded, stationary crack */
  v->K_I_0   = KI_static;
  v->K_II_0  = KII_static;
  v->K_III_0 = KIII_static;
  v->K_I_c   = KI_dynamic;
  v->K_II_c  = KII_dynamic;
  v->K_III_c = KIII_dynamic;
  /* Effective SIF */
  v->K_I_equiv = KI_equiv_static;
  /* Tilting angle */
  v->theta = (1 - tc->tilt_damping) * theta;
  /* Energy release rate for a dynamically loaded, stationary crack  */
  v->G_0 = G_0;
  /* Energy release rate for a moving crack is computed later; preset it
     to that of a stationary crack */
  v->G_c = G_c;
} 


/*
   Compute SIF's.
   Compute the stress intensity factors at a front vertex.
 */
static void
compute_sifs (ts_vert_t *v, /* vertex */
              triangulated_crack_surf_t *tc /* triangulated crack surface */
              ) 
{
  ts_vert_t *prevv, *nextv;
  ts_edge_t *preve, *nexte;

  prevv = NULL;
  nextv = NULL;
  find_prev_next_vertex_on_front (v, tc, &prevv, &nextv);
  if (prevv) { preve = connecting_edge (prevv, v); } 
  else                { preve = NULL; } 
  if (nextv) { nexte = connecting_edge (nextv, v);} 
  else                { nexte = NULL; } 

  do_compute_sifs (v, prevv, nextv, preve, nexte, tc);
} 


static void
extrapol_sifs (triangulated_crack_surf_t *tc /* instance of triangulated crack */,
               ts_vert_t *v,
               ts_vert_t *adjv
               ) 
{
  ts_vert_t *adjadjv, *prevv, *nextv;
  double K_I_0, K_II_0, K_III_0, K_I_equiv, G_0, G_c, theta;
  
  prevv = nextv = NULL;
  find_prev_next_vertex_on_front (adjv, tc, &prevv, &nextv);
  adjadjv = NULL;
  if (prevv == v) adjadjv = nextv;
  else if (nextv == v) adjadjv = prevv;
  if (adjadjv) {
    if (adjadjv->is_on_front) {
      double n[2];
      double xv, xadjadj;
      W_vect_t vect;
      VECMAC_3_VEC (vect, =, adjv->loc, -, v->loc);
      xv      = VECMAC_NORM (vect);
      VECMAC_3_VEC (vect, =, adjadjv->loc, -, adjv->loc);
      xadjadj = VECMAC_NORM (vect);
      n[0]    = (xv + xadjadj) / xadjadj;
      n[1]    = -xv / xadjadj;
      K_I_0     = n[0] * adjv->K_I_0       + n[1] * adjadjv->K_I_0;
      K_II_0    = n[0] * adjv->K_II_0      + n[1] * adjadjv->K_II_0;
      K_III_0   = n[0] * adjv->K_III_0     + n[1] * adjadjv->K_III_0;
      K_I_equiv = n[0] * adjv->K_I_equiv + n[1] * adjadjv->K_I_equiv;
      G_0       = n[0] * adjv->G_0       + n[1] * adjadjv->G_0;
      G_0       = max (v->G_0, 0); /* should be positive */
      G_c       = v->G_0;
      theta     = n[0] * adjv->theta     + n[1] * adjadjv->theta;
      if (tc->mixed_eval_at_the_bdry) {
        K_I_0     = (K_I_0 + v->K_I_0) / 2;
        K_II_0    = (K_II_0 + v->K_II_0) / 2;
        K_III_0   = (K_III_0 + v->K_III_0) / 2;
        K_I_equiv = (K_I_equiv + v->K_I_equiv) / 2;
        G_0       = (G_0 + v->G_0) / 2;
        G_c       = (G_c + v->G_c) / 2;
        theta     = (theta + v->theta) / 2;
      } 
      v->K_I_c     = v->K_I_0     = K_I_0;
      v->K_II_c    = v->K_II_0    = K_II_0;
      v->K_III_c   = v->K_III_0   = K_III_0;
      v->K_I_equiv = K_I_equiv;
      v->G_0       = G_0;
      v->G_c       = G_c;
      if (theta > -M_PI_2 && theta < M_PI_2) 
        v->theta = theta;
    } 
  }  else  {
    v->K_I_0       = adjv->K_I_0;
    v->K_II_0      = adjv->K_II_0;
    v->K_III_0     = adjv->K_III_0;
    v->K_I_c       = adjv->K_I_c;
    v->K_II_c      = adjv->K_II_c;
    v->K_III_c     = adjv->K_III_c;
    v->K_I_equiv = adjv->K_I_equiv;
    v->theta     = adjv->theta;
    v->G_0       = adjv->G_0;
    v->G_c       = adjv->G_c;
  } 
} 


static void
correct_sifs_for_boundary_vert (ts_vert_t *v,
                                triangulated_crack_surf_t *tc) 
{
  ts_vert_t *prevv, *nextv;
  ts_edge_t *preve, *nexte;

  prevv = NULL;
  nextv = NULL;
  find_prev_next_vertex_on_front (v, tc, &prevv, &nextv);
  if (prevv) { preve = connecting_edge (prevv, v); } 
  else                { preve = NULL; } 
  if (nextv) { nexte = connecting_edge (nextv, v);} 
  else                { nexte = NULL; } 

  if (prevv) {
    if (! (prevv->is_on_boundary && prevv->is_on_front)) {
      extrapol_sifs (tc, v, prevv);
    } 
  } 
  if (nextv) {
    if (! (nextv->is_on_boundary && nextv->is_on_front)) {
      extrapol_sifs (tc, v, nextv);
    } 
  } 
} 



static void
trim_edge (ts_vert_t *newv,
                   triangulated_crack_surf_t *tc) 
{
  ts_vert_t *prevv = NULL, *nextv = NULL;
  find_prev_next_vertex_on_front (newv, tc, &prevv, &nextv);

  if (newv->source_vert) 
    trim_bdry_edge (newv, prevv, nextv, tc);

  if (newv->enclosing_hex == NULL) {
    if (!find_hex_enclosing_loc (tc, &newv->loc,
                                 &newv->enclosing_hex, &newv->ploc)) {
      sprintf (W_buf (), "No enclosing hex at [%g;%g;%g]",
              newv->loc.x, newv->loc.y, newv->loc.z);
      REPORT_FATAL_ERROR (tc->crack_surf->se->domain->errh, W_buf ());
    } 
  } 
} 


static void
vertex_front_axes (ts_vert_t *v,
                   ts_vert_t *prevv, ts_vert_t *nextv,
                   ts_edge_t *preve, ts_edge_t *nexte, 
                   triangulated_crack_surf_t *tc);


/*
   Compute axes at front vertex.
   Compute local coordinate system at a crack front vertex.
   First, the the neighbors of the input vertex are found.
   Then, the triad is computed.
   Finally, the triad is corrected if plane crack growth is prescribed.
 */
static void
compute_axes (ts_vert_t *v,
             triangulated_crack_surf_t *tc) 
{
  ts_vert_t *prevv = NULL, *nextv = NULL;
  ts_edge_t *preve = NULL, *nexte = NULL;

  find_prev_next_vertex_on_front (v, tc, &prevv, &nextv);
  if (prevv == NULL && nextv == NULL) {
    sprintf (W_buf (), "Disconnected vertex %d (no previous or next vertex) ",
            v->id);
    REPORT_ERROR (tc->crack_surf->se->domain->errh, W_buf ());
  } 
  if (prevv) { preve = connecting_edge (prevv, v); } 
  if (nextv) { nexte = connecting_edge (nextv, v);} 

  vertex_front_axes (v, prevv, nextv, preve, nexte, tc);

  if (tc->grow_in_plane) {
    G3D_vect_t proj;
    double projn;
#define PROJECT(projectedv, planenormal, v) \
     {                                                           \
      double dot;                                               \
      dot = VECMAC_DOT (v, (planenormal));                      \
      VECMAC_3_VEC (projectedv, =, v, -, dot * (planenormal));  \
    } 
    PROJECT (proj, tc->motion_plane.n, v->forward);
    projn = VECMAC_NORM (proj);
    if (projn <= 0) {
      REPORT_ERROR (tc->crack_surf->se->domain->errh, "no projection");
    }  else  {
      VECMAC_NORMALIZE (proj);  VECMAC_2_VEC (v->forward, =, proj);
    } 
    if (VECMAC_DOT (v->tangent, tc->motion_plane.n) < 0) {
      VECMAC_2_VEC (v->tangent, =, -tc->motion_plane.n);
    }  else  {
      VECMAC_2_VEC (v->tangent, =, tc->motion_plane.n);
    } 
    VECMAC_CROSS (v->normal, =, v->forward, x, v->tangent);
    VECMAC_NORMALIZE (v->normal);
  } 
} 






static void
vertex_front_axes (ts_vert_t *v,
                  ts_vert_t *prevv, ts_vert_t *nextv,
                  ts_edge_t *preve, ts_edge_t *nexte, 
                  triangulated_crack_surf_t *tc);



static void
edge_axes (ts_edge_t *e, ts_vert_t *sv, ts_vert_t *ev,
          W_loc_t *normal, W_loc_t *tangent, W_loc_t *forward);

/*
   Evaluate vertex triad.
   The normal is computed first.  Then, tangent is evaluated as the average
   direction of the edges emanating from the input vertex.  Finally,
   the triad is completed by the forward vector.
 */
static void
vertex_front_axes (ts_vert_t *v, /* vertex */
                   ts_vert_t *prevv, /* previous vertex on the front (may be NULL) */
                   ts_vert_t *nextv, /* next vertex on the front (may be NULL) */
                   ts_edge_t *preve, /* edge connecting the previous vertex
                                        on the front (may be NULL) */
                   ts_edge_t *nexte,  /* edge connecting the next vertex
                                        on the front (may be NULL) */
                   triangulated_crack_surf_t *tc /* triangulated crack */
                   ) 
{
  W_vect_t zerovect =  {0,0,0} ;
  W_vect_t n, t, f;
  
  v->normal  = zerovect;
  v->tangent = zerovect;
  v->forward = zerovect;
  
  compute_normal (v);
  
  if (preve) {
    edge_axes (preve, prevv, v, &n, &t, &f);
    VECMAC_2_VEC (v->tangent, +=, t);
  } 
  if (nexte) {
    edge_axes (nexte, v, nextv, &n, &t, &f);
    VECMAC_2_VEC (v->tangent, +=, t);
  } 
  if (VECMAC_NORM (v->tangent) <= 0) 
    REPORT_ERROR (tc->crack_surf->se->domain->errh, "bad tangent");
  VECMAC_NORMALIZE (v->tangent);
  VECMAC_CROSS (v->forward, =, v->tangent, x, v->normal);
} 


/*
   Evaluate triad.
   Compute triad at the crack front edge.
   Start with tangent (given by the two vertices),
   then compute the normal by looking at the adjoining triangle.
   Finally, complete by the forward vector.
 */
static void
edge_axes (ts_edge_t *e, /* edge */
           ts_vert_t *sv, /* start vertex */
           ts_vert_t *ev, /* end vertex */
           W_loc_t *normal, /* output: normal */
           W_loc_t *tangent, /* output: tangent */
           W_loc_t *forward /* output: forward vector */
           ) 
{
  /*
    NOTE The neither of the vector is normalized; the longer the edge, the
   larger weight of the basis vectors.
   */
  W_vect_t n =  {0,0,0} ;
  double tn, nn;
  
  VECMAC_3_VEC ((*tangent), =, ev->loc, -, sv->loc);
  tn = VECMAC_NORM ((*tangent));
  if (tn <= 0) {
    REPORT_ERROR (NULL, "Tangent invalid");
  } 
  VECMAC_NORMALIZE ((*tangent));
  *normal = n;
  if (e->triangles[0]) {
    VECMAC_2_VEC ((*normal), +=, e->triangles[0]->normal);
  } 
  if (e->triangles[1]) {
    VECMAC_2_VEC ((*normal), +=, e->triangles[1]->normal);
  } 
  nn = VECMAC_NORM ((*normal));
  if (nn <= 0) {
    REPORT_ERROR (NULL, "Normal invalid");
  } 
  
  VECMAC_CROSS ((*forward), =, (*tangent), x, (*normal));
} 






static void
relax_edge_lengths (ts_vert_t *v, triangulated_crack_surf_t *tc);



static void
write_tri (triangulated_crack_surf_t *tc, ts_tri_t *t, ckit_iobuf_t IOB);
static void
write_cf_seg (triangulated_crack_surf_t *tc, ts_edge_t *e, ckit_iobuf_t IOB);


static void
write_hexvue_postproc_file (WEFG_crack_surf_t *cs,
                           ckit_iobuf_t IOB, char *mode) 
{
  triangulated_crack_surf_t *tc = (triangulated_crack_surf_t *) cs->data;
  ts_tri_t *t;
  ts_edge_t *e;

  ckit_io_put_line (IOB, "\n!");
  ckit_io_put_line (IOB, "CS");
  
  ckit_io_put_line (IOB, "CF");
  e = (ts_edge_t *) get_list_next (tc->edges, NULL);
  while (e) {
    if (E_IS_ON_FRONT (e)) 
      write_cf_seg (tc, e, IOB);
    e = (ts_edge_t *) get_list_next (tc->edges, e);
  } 
  ckit_io_put_line (IOB, "CFEND");

  t = (ts_tri_t *) get_list_next (tc->triangles, NULL);
  while (t) {
    write_tri (tc, t, IOB);
    t = (ts_tri_t *) get_list_next (tc->triangles, t);
  } 
  ckit_io_put_line (IOB, "CSEND");
} 

static void
V_get_stress (triangulated_crack_surf_t *tc,
	     ts_vert_t *v,
	     W_output_stress_type_enum stress_type,
	     double pos_normal_stress[3][3],
	     double neg_normal_stress[3][3]);
static void
V_get_eqpls (triangulated_crack_surf_t *tc,
	    ts_vert_t *v,
	    double *pos_normal_eqpls,
	    double *neg_normal_eqpls);


static void
write_tecplot_postproc_file (WEFG_crack_surf_t *cs,
                            ckit_iobuf_t IOB, char *mode,
                            void *optdata) 
{
  triangulated_crack_surf_t *tc = (triangulated_crack_surf_t *) cs->data;
  ts_vert_t *v;
  W_index_t *count = (W_index_t *) optdata;
  W_loc_t ploc, nloc;
  double pos_normal_stress[3][3], neg_normal_stress[3][3];
  double pos_normal_eqpls = 0, neg_normal_eqpls = 0;
  double pos_normal_vonm = 0, neg_normal_vonm = 0;

  v = (ts_vert_t *) get_list_next (tc->vertices, NULL);
  while (v) {
    sprintf (W_buf (), "%g %g %g ", v->loc.x, v->loc.y, v->loc.z);
    ckit_io_put_line_no_nl (IOB, W_buf ());
    if (WOPT_OPT (tc->crack_surf->se->domain->options, postfile_output_spec_deprec)
        & W_POSTFILE_DISPL_MASK) {
      vert_loc (v, tc, &ploc, &nloc);
    } 
    if (WOPT_OPT (tc->crack_surf->se->domain->options, postfile_output_spec_deprec)
        & W_POSTFILE_STRESS_MASK ||
        WOPT_OPT (tc->crack_surf->se->domain->options, postfile_output_spec_deprec)
        & W_POSTFILE_VON_MISES_MASK) {
      V_get_stress (tc, v, WOPT_OPT (tc->crack_surf->se->domain->options, output_stress_type),
                   pos_normal_stress, neg_normal_stress);
    } 
    if (WOPT_OPT (tc->crack_surf->se->domain->options, postfile_output_spec_deprec)
        & W_POSTFILE_EQPLS_MASK) {
      V_get_eqpls (tc, v, &pos_normal_eqpls, &neg_normal_eqpls);
    } 
    if (WOPT_OPT (tc->crack_surf->se->domain->options, postfile_output_spec_deprec)
        & W_POSTFILE_VON_MISES_MASK) {
      double means;
      means = (  pos_normal_stress[0][0]
	       + pos_normal_stress[1][1]
	       + pos_normal_stress[2][2]) / 3;
      pos_normal_stress[0][0] -= means;
      pos_normal_stress[1][1] -= means;
      pos_normal_stress[2][2] -= means;
      pos_normal_vonm
        = sqrt (1.5
               * (  pos_normal_stress[0][0]*pos_normal_stress[0][0]
                  + pos_normal_stress[0][1]*pos_normal_stress[0][1]
                  + pos_normal_stress[0][2]*pos_normal_stress[0][2]
                  + pos_normal_stress[1][0]*pos_normal_stress[1][0]
                  + pos_normal_stress[1][1]*pos_normal_stress[1][1]
                  + pos_normal_stress[1][2]*pos_normal_stress[1][2]
                  + pos_normal_stress[2][0]*pos_normal_stress[2][0]
                  + pos_normal_stress[2][1]*pos_normal_stress[2][1]
                  + pos_normal_stress[2][2]*pos_normal_stress[2][2]
                  ) 
               );
      means = (  neg_normal_stress[0][0]
               + neg_normal_stress[1][1]
               + neg_normal_stress[2][2]) / 3;
      neg_normal_stress[0][0] -= means;
      neg_normal_stress[1][1] -= means;
      neg_normal_stress[2][2] -= means;
      neg_normal_vonm
        = sqrt (1.5
               * (  neg_normal_stress[0][0]*neg_normal_stress[0][0]
                  + neg_normal_stress[0][1]*neg_normal_stress[0][1]
                  + neg_normal_stress[0][2]*neg_normal_stress[0][2]
                  + neg_normal_stress[1][0]*neg_normal_stress[1][0]
                  + neg_normal_stress[1][1]*neg_normal_stress[1][1]
                  + neg_normal_stress[1][2]*neg_normal_stress[1][2]
                  + neg_normal_stress[2][0]*neg_normal_stress[2][0]
                  + neg_normal_stress[2][1]*neg_normal_stress[2][1]
                  + neg_normal_stress[2][2]*neg_normal_stress[2][2]
                  ) 
               );
    } 
    /* on the plus side */
    sprintf (W_buf (), "%g %g %g ", v->loc.x, v->loc.y, v->loc.z);
    ckit_io_put_line_no_nl (IOB, W_buf ());
    if (WOPT_OPT (tc->crack_surf->se->domain->options, postfile_output_spec_deprec)
        & W_POSTFILE_DISPL_MASK) {
      
      sprintf (W_buf (), "%g %g %g ", 
              ploc.x - v->loc.x, ploc.y - v->loc.y, ploc.z - v->loc.z);
      ckit_io_put_line_no_nl (IOB, W_buf ());
    } 
    if (WOPT_OPT (tc->crack_surf->se->domain->options, postfile_output_spec_deprec)
        & W_POSTFILE_STRESS_MASK) {
      sprintf (W_buf (), "%g %g %g %g %g %g ",
               pos_normal_stress[0][0],
               pos_normal_stress[0][1],
               pos_normal_stress[0][2],
               pos_normal_stress[1][1],
               pos_normal_stress[1][2],
               pos_normal_stress[2][2]);
      ckit_io_put_line_no_nl (IOB, W_buf ());
    } 
    if (WOPT_OPT (tc->crack_surf->se->domain->options, postfile_output_spec_deprec)
        & W_POSTFILE_EQPLS_MASK) {
      sprintf (W_buf (), "%g ", pos_normal_eqpls);
      ckit_io_put_line_no_nl (IOB, W_buf ());
    } 
    if (WOPT_OPT (tc->crack_surf->se->domain->options, postfile_output_spec_deprec)
        & W_POSTFILE_VON_MISES_MASK) {
      sprintf (W_buf (), "%g ", pos_normal_vonm);
      ckit_io_put_line_no_nl (IOB, W_buf ());
    } 
    ckit_io_put_line (IOB, "");
    (*count) ++;
    /* on the minus side */
    sprintf (W_buf (), "%g %g %g ", v->loc.x, v->loc.y, v->loc.z);
    ckit_io_put_line_no_nl (IOB, W_buf ());
    if (WOPT_OPT (tc->crack_surf->se->domain->options, postfile_output_spec_deprec)
        & W_POSTFILE_DISPL_MASK) {
      sprintf (W_buf (), "%g %g %g ",
              nloc.x - v->loc.x, nloc.y - v->loc.y, nloc.z - v->loc.z);
      ckit_io_put_line_no_nl (IOB, W_buf ());
    } 
    if (WOPT_OPT (tc->crack_surf->se->domain->options, postfile_output_spec_deprec)
        & W_POSTFILE_STRESS_MASK) {
      sprintf (W_buf (), "%g %g %g %g %g %g ",
               neg_normal_stress[0][0],
               neg_normal_stress[0][1],
               neg_normal_stress[0][2],
               neg_normal_stress[1][1],
               neg_normal_stress[1][2],
               neg_normal_stress[2][2]);
      ckit_io_put_line_no_nl (IOB, W_buf ());
    } 
    if (WOPT_OPT (tc->crack_surf->se->domain->options, postfile_output_spec_deprec)
        & W_POSTFILE_EQPLS_MASK) {
      sprintf (W_buf (), "%g ", neg_normal_eqpls);
      ckit_io_put_line_no_nl (IOB, W_buf ());
    } 
    if (WOPT_OPT (tc->crack_surf->se->domain->options, postfile_output_spec_deprec)
        & W_POSTFILE_VON_MISES_MASK) {
      sprintf (W_buf (), "%g ", neg_normal_vonm);
      ckit_io_put_line_no_nl (IOB, W_buf ());
    } 
    (*count) ++;
    ckit_io_put_line (IOB, "");
    v = (ts_vert_t *) get_list_next (tc->vertices, v);
  } 
} 






static void
write_tri (triangulated_crack_surf_t *tc /* instance of triangulated crack */, ts_tri_t *t, ckit_iobuf_t IOB) 
{
  int i;
  W_loc_t ploc, nloc;
  double pos_normal_stress[3][3], neg_normal_stress[3][3];
  double pos_normal_eqpls, neg_normal_eqpls;
  
  ckit_io_put_line (IOB, "CP");
  ckit_io_put_line (IOB, "3");
  FOR (i, 3) {
    sprintf (W_buf (), "%g %g %g",
            t->vertex[i]->loc.x, t->vertex[i]->loc.y, t->vertex[i]->loc.z);
    ckit_io_put_line (IOB, W_buf ());
    if (WOPT_OPT (tc->crack_surf->se->domain->options, postfile_output_spec_deprec)
        & W_POSTFILE_DISPL_MASK) {
      vert_loc (t->vertex[i], tc, &ploc, &nloc);
      sprintf (W_buf (), "%g %g\n%g %g\n%g %g",
              ploc.x - t->vertex[i]->loc.x, nloc.x - t->vertex[i]->loc.x,
              ploc.y - t->vertex[i]->loc.y, nloc.y - t->vertex[i]->loc.y,
              ploc.z - t->vertex[i]->loc.z, nloc.z - t->vertex[i]->loc.z);
      ckit_io_put_line (IOB, W_buf ());
    } 
    if (WOPT_OPT (tc->crack_surf->se->domain->options, postfile_output_spec_deprec)
        & W_POSTFILE_STRESS_MASK) {
      V_get_stress (tc, t->vertex[i],
		   WOPT_OPT (tc->crack_surf->se->domain->options, output_stress_type), 
		   pos_normal_stress, neg_normal_stress);
      sprintf (W_buf (), "%g %g\n%g %g\n%g %g\n%g %g\n%g %g\n%g %g",
	      pos_normal_stress[0][0], neg_normal_stress[0][0],
              pos_normal_stress[0][1], neg_normal_stress[0][1],
              pos_normal_stress[0][2], neg_normal_stress[0][2],
              pos_normal_stress[1][1], neg_normal_stress[1][1],
              pos_normal_stress[1][2], neg_normal_stress[1][2],
              pos_normal_stress[2][2], neg_normal_stress[2][2]);
      ckit_io_put_line (IOB, W_buf ());
    } 

    if (WOPT_OPT (tc->crack_surf->se->domain->options, postfile_output_spec_deprec)
        & W_POSTFILE_EQPLS_MASK) {
      V_get_eqpls (tc, t->vertex[i],
		  &pos_normal_eqpls, &neg_normal_eqpls);
      sprintf (W_buf (), "%g %g", pos_normal_eqpls, neg_normal_eqpls);
      ckit_io_put_line (IOB, W_buf ());
    } 
    if (WOPT_OPT (tc->crack_surf->se->domain->options, postfile_output_spec_deprec)
        & W_POSTFILE_VON_MISES_MASK) {
      double means;
      V_get_stress (tc, t->vertex[i],
		   W_OUTPUT_STRESS_CAUCHY,
		   pos_normal_stress, neg_normal_stress);
      means = (  pos_normal_stress[0][0]
	       + pos_normal_stress[1][1]
	       + pos_normal_stress[2][2]) / 3;
      pos_normal_stress[0][0] -= means;
      pos_normal_stress[1][1] -= means;
      pos_normal_stress[2][2] -= means;
      sprintf (W_buf (), "%g ",
                sqrt (1.5
                     * (  pos_normal_stress[0][0]*pos_normal_stress[0][0]
			+ pos_normal_stress[0][1]*pos_normal_stress[0][1]
			+ pos_normal_stress[0][2]*pos_normal_stress[0][2]
                        + pos_normal_stress[1][0]*pos_normal_stress[1][0]
			+ pos_normal_stress[1][1]*pos_normal_stress[1][1]
			+ pos_normal_stress[1][2]*pos_normal_stress[1][2]
                        + pos_normal_stress[2][0]*pos_normal_stress[2][0]
			+ pos_normal_stress[2][1]*pos_normal_stress[2][1]
			+ pos_normal_stress[2][2]*pos_normal_stress[2][2]
                        ) 
                     ) 
	      );
      ckit_io_put_line_no_nl (IOB, W_buf ());
      means = (  neg_normal_stress[0][0]
	       + neg_normal_stress[1][1]
	       + neg_normal_stress[2][2]) / 3;
      neg_normal_stress[0][0] -= means;
      neg_normal_stress[1][1] -= means;
      neg_normal_stress[2][2] -= means;
      sprintf (W_buf (), "%g ",
                sqrt (1.5
                     * (  neg_normal_stress[0][0]*neg_normal_stress[0][0]
			+ neg_normal_stress[0][1]*neg_normal_stress[0][1]
			+ neg_normal_stress[0][2]*neg_normal_stress[0][2]
                        + neg_normal_stress[1][0]*neg_normal_stress[1][0]
			+ neg_normal_stress[1][1]*neg_normal_stress[1][1]
			+ neg_normal_stress[1][2]*neg_normal_stress[1][2]
                        + neg_normal_stress[2][0]*neg_normal_stress[2][0]
			+ neg_normal_stress[2][1]*neg_normal_stress[2][1]
			+ neg_normal_stress[2][2]*neg_normal_stress[2][2]
                        ) 
                     ) 
	      );
      ckit_io_put_line (IOB, W_buf ());
    } 
  } 
} 

static void
write_cf_seg (triangulated_crack_surf_t *tc /* instance of triangulated crack */, ts_edge_t *e, ckit_iobuf_t IOB) 
{
  int eix = edge_idx (e->triangles[0], e);
  ts_vert_t *v0, *v1;

  find_edge_vertices (e->triangles[0], eix, &v0, &v1);
  
  sprintf (W_buf (), "SEG %g %g %g %g %g %g",
           v0->loc.x, v0->loc.y, v0->loc.z,
           v1->loc.x, v1->loc.y, v1->loc.z);
  ckit_io_put_line (IOB, W_buf ());
} 


static void
do_eval_sifs (ts_vert_t *v,
             ts_vert_t *prevv, ts_vert_t *nextv,
             ts_edge_t *preve, ts_edge_t *nexte, 
             triangulated_crack_surf_t *tc /* instance of triangulated crack */,
             double *J1static, double *J1dynamic,
             double *J2static, double *J2dynamic,
             double *KIstatic,
             double *KIIstatic,
             double *KIIIstatic,
             double *KIdynamic,
             double *KIIdynamic,
             double *KIIIdynamic,
             double *KIequiv,
             double *theta,
             double *o_displ, double *s_displ, double *t_displ);




static void
eval_Ks (triangulated_crack_surf_t *tc,
        ts_vert_t *v,
        double J1static, double J2static, 
        double J1dynamic, double J2dynamic, 
        double *KIstatic, double *KIIstatic, double *KIIIstatic,
        double *KIdynamic, double *KIIdynamic, double *KIIIdynamic,
        double *o_displ, double *s_displ, double *t_displ);


static void
do_eval_sifs (ts_vert_t *v,
             ts_vert_t *prevv, ts_vert_t *nextv,
             ts_edge_t *preve, ts_edge_t *nexte, 
             triangulated_crack_surf_t *tc /* instance of triangulated crack */,
             double *J1static, double *J1dynamic,
             double *J2static, double *J2dynamic,
             double *KIstatic, double *KIIstatic, double *KIIIstatic,
             double *KIdynamic, double *KIIdynamic, double *KIIIdynamic,
             double *KIequiv,
             double *theta,
             double *o_displ, double *s_displ, double *t_displ) 
{
  double along_front;
  W_loc_t d;
  double distp = 0, distn = 0;
  double q_line_integral = 0;
  double J1_surf_static = 0, J1_surf_dynamic = 0;
  double J2_surf_static = 0, J2_surf_dynamic = 0;
  double along_front_prev, along_front_next;
  W_options_t *o = tc->crack_surf->se->domain->options;

  if (prevv) {
    VECMAC_3_VEC (d, =, v->loc, -, prevv->loc);
    distp = VECMAC_NORM (d);
  } 
  if (nextv) {
    VECMAC_3_VEC (d, =, v->loc, -, nextv->loc);
    distn = VECMAC_NORM (d);
  } 

  along_front = min (distn, distp);
  if (tc->jdom_hsize_forward_plus <= 0) 
    tc->jdom_hsize_forward_plus  = tc->crack_surf->se->max_support_size;
  if (tc->jdom_hsize_forward_minus <= 0) 
    tc->jdom_hsize_forward_minus = tc->crack_surf->se->max_support_size;
  if (tc->jdom_hsize_tangent_plus <= 0) 
    tc->jdom_hsize_tangent_plus  = along_front;
  if (tc->jdom_hsize_tangent_minus <= 0) 
    tc->jdom_hsize_tangent_minus = along_front;
  if (tc->jdom_hsize_normal_plus <= 0) 
    tc->jdom_hsize_normal_plus   = tc->crack_surf->se->max_support_size;
  if (tc->jdom_hsize_normal_minus <= 0) 
    tc->jdom_hsize_normal_minus  = tc->crack_surf->se->max_support_size;

  if (distp > 0) {
    along_front_prev = min (distp, tc->jdom_hsize_tangent_minus);
  }  else  {
    along_front_prev = (distp + distn) / 100;
  } 
  if (distn > 0) {
    along_front_next = min (distn, tc->jdom_hsize_tangent_plus);
  }  else  {
    along_front_next = (distp + distn) / 100;
  } 
  
  *J1static  = 0;
  *J1dynamic = 0;
  *J2static  = 0;
  *J2dynamic = 0;
    
  efgj_j_volume (tc->crack_surf->se,
                &v->loc, &v->forward, &v->tangent, &v->normal,
                tc->jdom_hsize_forward_plus,
                tc->jdom_hsize_forward_minus,
                tc->jdom_hsize_tangent_plus,
                tc->jdom_hsize_tangent_minus,
                tc->jdom_hsize_normal_plus,
                tc->jdom_hsize_normal_minus,
 (WOPT_OPT (o, q_func_type) == W_QUARTIC_Q_FUNC
                ? W_f_quartic_spline
                :
 (WOPT_OPT (o, q_func_type) == W_LINEAR_Q_FUNC
                ? W_f_linear
                :
 (WOPT_OPT (o, q_func_type) == W_COMPOSITE_QUADRATIC_Q_FUNC
                ? W_f_composite_quadratic
                :
 (WOPT_OPT (o, q_func_type) == W_QUADRATIC_Q_FUNC
                ? W_f_quadratic
                 :
                NULL)) )),
 (WOPT_OPT (o, q_func_type) == W_QUARTIC_Q_FUNC
                ? W_f_r_quartic_spline
                :
 (WOPT_OPT (o, q_func_type) == W_LINEAR_Q_FUNC
                ? W_f_r_linear
                :
 (WOPT_OPT (o, q_func_type) == W_COMPOSITE_QUADRATIC_Q_FUNC
                ? W_f_r_composite_quadratic
                :
 (WOPT_OPT (o, q_func_type) == W_QUADRATIC_Q_FUNC
                ? W_f_r_quadratic
                 :
                NULL)) )),
                J1static, J1dynamic,
                J2static, J2dynamic);
  switch (WOPT_OPT (o, q_func_type)) {
  case W_QUARTIC_Q_FUNC:
    q_line_integral = 2. / 5. * (along_front_next + along_front_prev);
    break;
  case W_LINEAR_Q_FUNC:
    q_line_integral = 1. / 2. * (along_front_next + along_front_prev);
    break;
  case W_QUADRATIC_Q_FUNC:
    q_line_integral = 2. / 3. * (along_front_next + along_front_prev);
    break;
  case W_COMPOSITE_QUADRATIC_Q_FUNC:
    q_line_integral = 1. / 2. * (along_front_next + along_front_prev);
    break;
  default:
    REPORT_ERROR (tc->crack_surf->se->domain->errh, "Unresolved");
    break;
  } 
  
  *J1static   = *J1static / q_line_integral;
  *J1dynamic  = *J1dynamic / q_line_integral;
  *J2static   = *J2static / q_line_integral;
  *J2dynamic  = *J2dynamic / q_line_integral;

  J1_surf_static = J1_surf_dynamic =
  J2_surf_static = J2_surf_dynamic = 0;

  j_setup (tc->crack_surf, v, 
          tc->jdom_hsize_forward_plus,
          tc->jdom_hsize_forward_minus,
          tc->jdom_hsize_tangent_plus,
          tc->jdom_hsize_tangent_minus,
          tc->jdom_hsize_normal_plus,
          tc->jdom_hsize_normal_minus,
 (WOPT_OPT (o, q_func_type) == W_QUARTIC_Q_FUNC
          ? W_f_quartic_spline
          :
 (WOPT_OPT (o, q_func_type) == W_LINEAR_Q_FUNC
          ? W_f_linear
          :
 (WOPT_OPT (o, q_func_type) == W_COMPOSITE_QUADRATIC_Q_FUNC
          ? W_f_composite_quadratic
          :
 (WOPT_OPT (o, q_func_type) == W_QUADRATIC_Q_FUNC
          ? W_f_quadratic
          :
          NULL)) )),
 (WOPT_OPT (o, q_func_type) == W_QUARTIC_Q_FUNC
          ? W_f_r_quartic_spline
          :
 (WOPT_OPT (o, q_func_type) == W_LINEAR_Q_FUNC
          ? W_f_r_linear
          :
 (WOPT_OPT (o, q_func_type) == W_COMPOSITE_QUADRATIC_Q_FUNC
          ? W_f_r_composite_quadratic
          :
 (WOPT_OPT (o, q_func_type) == W_QUADRATIC_Q_FUNC
          ? W_f_r_quadratic
          :
          NULL)) )) 
          );

  if (!tc->ignore_surface_j_contrib) {
    j_surface (tc->crack_surf, v, 
              tc->jdom_hsize_forward_plus,
              tc->jdom_hsize_forward_minus,
              tc->jdom_hsize_tangent_plus,
              tc->jdom_hsize_tangent_minus,
              tc->jdom_hsize_normal_plus,
              tc->jdom_hsize_normal_minus,
 (WOPT_OPT (o, q_func_type) == W_QUARTIC_Q_FUNC
               ? W_f_quartic_spline
               :
 (WOPT_OPT (o, q_func_type) == W_LINEAR_Q_FUNC
                ? W_f_linear
                :
 (WOPT_OPT (o, q_func_type) == W_COMPOSITE_QUADRATIC_Q_FUNC
                 ? W_f_composite_quadratic
                 :
 (WOPT_OPT (o, q_func_type) == W_QUADRATIC_Q_FUNC
                  ? W_f_quadratic
                  :
                  NULL)) )),
 (WOPT_OPT (o, q_func_type) == W_QUARTIC_Q_FUNC
               ? W_f_r_quartic_spline
               :
 (WOPT_OPT (o, q_func_type) == W_LINEAR_Q_FUNC
                ? W_f_r_linear
                :
 (WOPT_OPT (o, q_func_type) == W_COMPOSITE_QUADRATIC_Q_FUNC
                 ? W_f_r_composite_quadratic
                 :
 (WOPT_OPT (o, q_func_type) == W_QUADRATIC_Q_FUNC
                  ? W_f_r_quadratic
                  :
                  NULL)) )),
              &J1_surf_static, &J1_surf_dynamic,
              &J2_surf_static, &J2_surf_dynamic);

    *J1static  += J1_surf_static / q_line_integral;
    *J1dynamic += J1_surf_dynamic / q_line_integral;
    *J2static  += J2_surf_static / q_line_integral;
    *J2dynamic += J2_surf_dynamic / q_line_integral;
  } 

  eval_Ks (tc, v,
          *J1static, *J2static,
          *J1static + *J1dynamic, *J2static + *J2dynamic,
          KIstatic, KIIstatic, KIIIstatic,
          KIdynamic, KIIdynamic, KIIIdynamic,
          o_displ, s_displ, t_displ);
} 






static void
j_setup (WEFG_crack_surf_t *cs,
        ts_vert_t *v,
        double hsize_forward_plus,
        double hsize_forward_minus,
        double hsize_tangent_plus,
        double hsize_tangent_minus,
        double hsize_normal_plus,
        double hsize_normal_minus,
        WEFG_tensor_prod_comp_func func,
        WEFG_tensor_prod_comp_der_func func_der
        );







/*
   Update stress.
   Update stress at all points on the crack surface.
   Re-compute the shape function tables, if needed, and
   then update the material state.
 */
static void
update_stress (WEFG_superelem_t *se, /* EFG superelement */
               WEFG_crack_surf_t *cs /* generic crack */
               ) 
{
  triangulated_crack_surf_t *tc = (triangulated_crack_surf_t *) cs->data;
  ts_vert_t *v;
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  double P[3][3];
  W_mat_3d_stress_type_enum out_type;

  ip_ctx_data.se       = cs->se;
  ip_ctx.time          = WD_time (se->domain);
  ip_ctx.time_step     = WD_time_step (se->domain);
  ip_ctx.ctx_data      = &ip_ctx_data;
  ip_ctx.get_kinem_var = efgss_get_kinem_var;

  loop_through_list (tc->vertices, v, ts_vert_t *) {
    vert_sftabs (v, tc);
    ip_ctx_data.sftab    = v->sftab_pos_normal;
    EFGSS_AT_NEW_IP ();
    if (!WM3D_update_stress (se->mat, v->ms_pos_normal, W_MAT_STATE_UPDATE_FINAL,
			    &ip_ctx, MAT_3D_1ST_PK, P,
			    &out_type)) 
      REPORT_FATAL_ERROR (se->domain->errh, "Failed to compute stress");
    ip_ctx_data.sftab    = v->sftab_neg_normal;
    EFGSS_AT_NEW_IP ();
    if (!WM3D_update_stress (se->mat, v->ms_neg_normal, W_MAT_STATE_UPDATE_FINAL,
			    &ip_ctx, MAT_3D_1ST_PK, P,
			    &out_type)) 
      REPORT_FATAL_ERROR (se->domain->errh, "Failed to compute stress");
  } 
} 


static void
V_get_stress (triangulated_crack_surf_t *tc /* instance of triangulated crack */,
	     ts_vert_t *v,
	     W_output_stress_type_enum stress_type,
	     double pos_normal_stress[3][3],
	     double neg_normal_stress[3][3]) 
{
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  W_mat_3d_stress_type_enum out_type;

  vert_sftabs (v, tc);
  
  ip_ctx_data.se       = tc->crack_surf->se;
  ip_ctx.time          = WD_time (tc->crack_surf->se->domain);
  ip_ctx.time_step     = WD_time_step (tc->crack_surf->se->domain);
  ip_ctx.ctx_data      = &ip_ctx_data;
  ip_ctx.get_kinem_var = efgss_get_kinem_var;

  ip_ctx_data.sftab    = v->sftab_pos_normal;
  EFGSS_AT_NEW_IP ();
  if (stress_type == W_OUTPUT_STRESS_CAUCHY) {
    if (!WM3D_update_stress (tc->crack_surf->se->mat, v->ms_pos_normal, W_MAT_STATE_UPDATE_TEMP,
                         &ip_ctx, MAT_3D_CAUCHY,
			 pos_normal_stress,
			 &out_type)) 
      REPORT_FATAL_ERROR (tc->crack_surf->se->domain->errh, "Failed to compute stress");
    if (out_type != MAT_3D_CAUCHY) 
      REPORT_FATAL_ERROR (tc->crack_surf->se->domain->errh, "Stress type not available");
  }  else if (stress_type == W_OUTPUT_STRESS_2ND_PK) {
    if (!WM3D_update_stress (tc->crack_surf->se->mat,
			 v->ms_pos_normal, W_MAT_STATE_UPDATE_TEMP,
                         &ip_ctx, MAT_3D_2ND_PK,
			 pos_normal_stress,
			 &out_type)) 
      REPORT_FATAL_ERROR (tc->crack_surf->se->domain->errh, "Failed to compute stress");
    if (out_type != MAT_3D_2ND_PK) 
      REPORT_FATAL_ERROR (tc->crack_surf->se->domain->errh, "Stress type not available");
  } 

  ip_ctx_data.sftab    = v->sftab_neg_normal;
  EFGSS_AT_NEW_IP ();
  if (stress_type == W_OUTPUT_STRESS_CAUCHY) {
    if (!WM3D_update_stress (tc->crack_surf->se->mat, v->ms_neg_normal, W_MAT_STATE_UPDATE_TEMP,
                         &ip_ctx, MAT_3D_CAUCHY,
			 neg_normal_stress,
			 &out_type)) 
      REPORT_FATAL_ERROR (tc->crack_surf->se->domain->errh, "Failed to compute stress");
    if (out_type != MAT_3D_CAUCHY) 
      REPORT_FATAL_ERROR (tc->crack_surf->se->domain->errh, "Stress type not available");
  }  else if (stress_type == W_OUTPUT_STRESS_2ND_PK) {
    if (!WM3D_update_stress (tc->crack_surf->se->mat,
			 v->ms_neg_normal, W_MAT_STATE_UPDATE_TEMP,
                         &ip_ctx, MAT_3D_2ND_PK,
			 neg_normal_stress,
			 &out_type)) 
      REPORT_FATAL_ERROR (tc->crack_surf->se->domain->errh, "Failed to compute stress");
    if (out_type != MAT_3D_2ND_PK) 
      REPORT_FATAL_ERROR (tc->crack_surf->se->domain->errh, "Stress type not available");
  } 
} 



static void
V_get_eqpls (triangulated_crack_surf_t *tc /* instance of triangulated crack */,
	    ts_vert_t *v,
	    double *pos_normal_eqpls,
	    double *neg_normal_eqpls) 
{
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;

  vert_sftabs (v, tc);
  
  ip_ctx_data.se       = tc->crack_surf->se;
  ip_ctx.time          = WD_time (tc->crack_surf->se->domain);
  ip_ctx.time_step     = WD_time_step (tc->crack_surf->se->domain);
  ip_ctx.ctx_data      = &ip_ctx_data;
  ip_ctx.get_kinem_var = efgss_get_kinem_var;

  ip_ctx_data.sftab    = v->sftab_pos_normal;
  EFGSS_AT_NEW_IP ();
  if (!WM3D_get_mat_state_var (tc->crack_surf->se->mat, v->ms_pos_normal,
                              &ip_ctx, MAT_3D_EQUIV_PLASTIC_DEF,
			      pos_normal_eqpls)) 
    REPORT_FATAL_ERROR (tc->crack_surf->se->domain->errh, "Failed to retrieve equiv plastic def");

  ip_ctx_data.sftab    = v->sftab_neg_normal;
  EFGSS_AT_NEW_IP ();
  if (!WM3D_get_mat_state_var (tc->crack_surf->se->mat, v->ms_neg_normal,
                              &ip_ctx, MAT_3D_EQUIV_PLASTIC_DEF,
			      neg_normal_eqpls)) 
    REPORT_FATAL_ERROR (tc->crack_surf->se->domain->errh, "Failed to retrieve equiv plastic def");
} 


static void
write_ve3d_loclist (WEFG_crack_surf_t *cs, ckit_iobuf_t IOB) 
{
  triangulated_crack_surf_t *tc = (triangulated_crack_surf_t *) cs->data;
  ts_vert_t *v;

  sprintf (W_buf (), "LOCLIST \"triangulated.crack.surface:%d\"", cs->id);
  ckit_io_put_line (IOB, W_buf ());
  loop_through_list (tc->vertices, v, ts_vert_t *) {
    sprintf (W_buf (), "%d %g %g %g", v->id, v->loc.x, v->loc.y, v->loc.z);
    ckit_io_put_line (IOB, W_buf ());
  } 
  ckit_io_put_line (IOB, "END_LOCLIST");
} 


static void
write_ve3d_polys (WEFG_crack_surf_t *cs, ckit_iobuf_t IOB) 
{
  triangulated_crack_surf_t *tc = (triangulated_crack_surf_t *) cs->data;
  int i;
  ts_tri_t *t;
  
  sprintf (W_buf (), "BDRYPOLYS \"triangulated.crack.surface:%d\"", cs->id);
  ckit_io_put_line (IOB, W_buf ());
  loop_through_list (tc->triangles, t, ts_tri_t *) {
    ckit_io_put_line_no_nl (IOB, "T ");
    FOR (i, 3) {
      sprintf (W_buf (), "%g %g %g ", 
              t->vertex[i]->loc.x, 
              t->vertex[i]->loc.y, 
              t->vertex[i]->loc.z);
      ckit_io_put_line_no_nl (IOB, W_buf ());
    } 
    ckit_io_put_line (IOB, "");
  } 
  
  ckit_io_put_line (IOB, "END_BDRYPOLYS\n");
} 

/*
   Is the triangle distortion acceptable?
   Check whether all triangles incident on the vertex
   are of acceptable shape by verifying that all their
   normals point in the `same' direction.
 */
static int /* FALSE if one or more triangles not acceptable */
acceptable_triangle_distortion (ts_vert_t *v /* vertex */
                                ) 
{
  ts_edge_t *e;
  W_vect_t firstn;

  e = (ts_edge_t *) get_list_next (v->edges, NULL);
  if (e->triangles[0]) {
    firstn = e->triangles[0]->normal;
  }  else if (e->triangles[1]) {
    firstn = e->triangles[1]->normal;
  }  else  {
    REPORT_FATAL_ERROR (NULL, "No triangles on an edge");
  } 
  e = NULL;
  while ((e = (ts_edge_t *) get_list_next (v->edges, e)) ) {
    if (e->triangles[0]) {
      tri_normal (e->triangles[0]);
      if (VECMAC_DOT (firstn, e->triangles[0]->normal) < 0) 
        return FALSE;
    }  else if (e->triangles[1]) {
      tri_normal (e->triangles[1]);
      if (VECMAC_DOT (firstn, e->triangles[1]->normal) < 0) 
        return FALSE;
    } 
  } 
  return TRUE;
} 

/*
   Relax edge lengths.
   Relax lengths of front edges by iteratively moving vertices along the front.
   The motion is reverted if resulting triangles are too skewed.
 */
static void
relax_edge_lengths (ts_vert_t *v, triangulated_crack_surf_t *tc) 
{
  ts_edge_t *preve, *nexte;
  ts_vert_t *prevv, *nextv;
  int max_tries = 3, ntries;
  
  prevv = NULL;
  nextv = NULL;
  find_prev_next_vertex_on_front (v, tc, &prevv, &nextv);
  if (prevv) { preve = connecting_edge (prevv, v); } 
  else                { preve = NULL; } 
  if (nextv) { nexte = connecting_edge (nextv, v);} 
  else                { nexte = NULL; } 

  vertex_front_axes (v, prevv, nextv, preve, nexte, tc);
    
  if (preve && nexte) {
    W_loc_t oldloc = v->before_loc, pt_on_plane, plane_normal;
    G3D_vect_t suggloc;
    G3D_plane_t plane;
    double lambda;
    G3D_bool_t parallel;

    VECMAC_3_VEC (pt_on_plane, = ,
                  0.5 * prevv->before_loc, +, 0.5 * nextv->before_loc);
    VECMAC_3_VEC (plane_normal, =, nextv->before_loc, -, prevv->before_loc);
    G3D_plane_consts_p_and_n (&plane, &pt_on_plane, &plane_normal);
    G3D_intersect_plane_and_line (&suggloc, &lambda, &parallel,
                                  &v->before_loc, &v->tangent, &plane,
                                  0, FLT_MAX);
    v->loc = suggloc;
    ntries = 0;
    while (!acceptable_triangle_distortion (v) 
           && (ntries++ < max_tries)) {
      VECMAC_3_VEC (v->loc, =, oldloc, +, suggloc);
      VECMAC_SCALAR (v->loc, *=, 0.5);
      suggloc = v->loc;
    } 
    if (ntries == max_tries) {
      v->loc = oldloc;          /* failed to relax; back up */
    } 
  } 
} 


/* adjust vertex status.
   adjust status of boundary-bound vertex.
   If it has any neighbors which are not bound to the boundary, the vertex
   is assumed to be in a position to move in the next step.
*/
static void
adjust_vertex_status (ts_vert_t *v, /* front vertex */
                      triangulated_crack_surf_t *tc /* triangulated crack surface */
                      ) 
{
  ts_vert_t *prevv, *nextv;

  find_prev_next_vertex_on_front (v, tc, &prevv, &nextv);

  v->is_on_front = FALSE;       /* assume it needs to be stopped */
  if (prevv) 
    if (prevv->is_on_front && (!prevv->is_on_boundary)) 
      v->is_on_front = TRUE;    /* it can move after all */
  if (nextv) 
    if (nextv->is_on_front && (!nextv->is_on_boundary)) 
      v->is_on_front = TRUE;   /* it can move after all */
} 



static void
adjust_edge_lengths (ts_vert_t *v,
                     triangulated_crack_surf_t *tc);


static void
copy_master_data (ts_vert_t *master, ts_vert_t *slave) 
{
  slave->K_I_0     = master->K_I_0;
  slave->K_II_0    = master->K_II_0;
  slave->K_III_0   = master->K_III_0;
  slave->K_I_c     = master->K_I_c;
  slave->K_II_c    = master->K_II_c;
  slave->K_III_c   = master->K_III_c;
  slave->K_I_equiv = master->K_I_equiv;
  slave->theta     = master->theta;
  slave->G_0       = master->G_0;
  slave->G_c       = master->G_c;
} 

/*
   Propagate crack.
   Propagate crack by the following algorithm:
   (i) Find the vertices on the front, and compute local coordinate systems.
   (ii) If a master vertex has been specified in the input, drive the crack
   by adopting the stress intensity factors (SIF) at the master for all front vertices;
   otherwise compute SIF's for all front vertices.  If desired,
   extrapolate SIF's for vertices bound to the boundary.
   (iii) Advance front vertices.
   (iv) Trim edges emanating from vertices bound to the boundary.
   (v) For each front vertex which is bound to the boundary, adjust its status
   (set is_on_front if it is free to move).
   (vi) If desired, adapt front discretization to avoid too long edges.
 */
static void 
propagate (WEFG_crack_surf_t *cs /* generic crack surface */
           ) 
{
  triangulated_crack_surf_t *tc = (triangulated_crack_surf_t *) cs->data;
  ts_vert_t *v;
  int i;

  /* Re-create the list of on-the-front vertices */
  flush_list (tc->on_front_vertices, NOT_NODES);
  loop_through_list (tc->vertices, v, ts_vert_t *) {
    if (v->is_on_front) {
      add_to_tail (tc->on_front_vertices, v);
      compute_axes (v, tc);
    } 
  } 

  if (tc->master_vert) {
    compute_sifs (tc->master_vert, tc);
    loop_through_list (tc->on_front_vertices, v, ts_vert_t *) {
      copy_master_data (tc->master_vert, v);
    } 
  }  else  {
    loop_through_list (tc->on_front_vertices, v, ts_vert_t *) {
      compute_sifs (v, tc);
    } 
  } 

  if (tc->extrapolate_at_the_bdry) {
    loop_through_list (tc->on_front_vertices, v, ts_vert_t *) {
      if (v->is_on_boundary) 
        correct_sifs_for_boundary_vert (v, tc);
    } 
  } 
  
  loop_through_list (tc->on_front_vertices, v, ts_vert_t *) {
    advance_vertex (v, tc);
  } 

  loop_through_list (tc->vertices, v, ts_vert_t *) {
    if (v->is_on_front && v->is_on_boundary) {
      trim_edge (v, tc);
    } 
  } 

  loop_through_list (tc->vertices, v, ts_vert_t *) {
    if (v->is_on_front && v->is_on_boundary) 
      adjust_vertex_status (v, tc);
  } 

  if (tc->allow_relaxation) {
    loop_through_list (tc->vertices, v, ts_vert_t *) {
      if (v->is_on_front) v->before_loc = v->loc;
    } 
    for (i = 0; i < MAX_RELAXATION_CYCLES; i++) {
      loop_through_list (tc->vertices, v, ts_vert_t *) {
        if (v->is_on_front && (!v->is_on_boundary)) {
          relax_edge_lengths (v, tc);
        } 
      } 
    } 
  } 

  if (tc->allow_adaptivity) {
    flush_list (tc->on_front_vertices, NOT_NODES);
    loop_through_list (tc->vertices, v, ts_vert_t *) {
      if (v->is_on_front) {
        add_to_tail (tc->on_front_vertices, v);
        compute_axes (v, tc);
      } 
    } 
    loop_through_list (tc->on_front_vertices, v, ts_vert_t *) {
      if (v->is_on_front && (!v->is_on_boundary)) {
        adjust_edge_lengths (v, tc);
      } 
    } 
  } 
} 



static void *
new_sampler (WEFG_crack_surf_t *cs, W_index_t pntidx, W_loc_t *loc) 
{
  ts_crack_tip_history_client_data_t *cd;
  triangulated_crack_surf_t *tc = (triangulated_crack_surf_t *) cs->data;
  ts_vert_t *v;

  loctol = cs->se->tolerance;
  v = find_vert_by_loc (tc, loc);
  if (v == NULL) 
    return NULL;		/* no vertex at all at that location */
  if (!v->is_on_front) 
    return NULL;		/* vertex is not on front */

  cd = w_new (ts_crack_tip_history_client_data_t);
  CHECK_MALLOC (cd);

  cd->pntidx             = pntidx;
  cd->tc                 = tc;
  cd->orig_vert          = v;
  cd->orig_vert->hist_cd = cd;
  cd->vert               = v;
  cd->buffidx            = 0;

  return (void *) cd;
} 



static void
store_to_buffer (ts_crack_tip_history_client_data_t *cd,
		W_float_t time);
static void
flush_history (W_solution_history_t *h,
	      ts_crack_tip_history_client_data_t *cd);





static void
store (ts_crack_tip_history_client_data_t *cd,
       W_history_writer_t *hw,
       W_float_t time) 
{
  double S[3][3];
  double equiv_pl_def, pressure;
  W_loc_t x, v, a;
#define STORE(SUFFIX, data)                             \
  {                                                     \
    char name[64];                                      \
    sprintf (name, "ctp%d%s", cd->pntidx, SUFFIX);      \
    WHR_store (hw, name, time, data);                   \
  }

  vert_loc (cd->vert, cd->tc, &x, &x);
  vert_vel (cd->vert, cd->tc, &v, &v);
  vert_acc (cd->vert, cd->tc, &a, &a);

  STORE ("TDx", x.x - cd->orig_vert->loc.x);
  STORE ("TDy", x.y - cd->orig_vert->loc.y);
  STORE ("TDz", x.z - cd->orig_vert->loc.z);
  STORE ("TVx", v.x);
  STORE ("TVy", v.y);
  STORE ("TVz", v.z);
  STORE ("TAx", a.x);
  STORE ("TAy", a.y);
  STORE ("TAz", a.z);

  if (cd->tc->crack_surf->se->compute_stress_at_output_pts) {
    V_get_stress (cd->tc, cd->vert,
		 WOPT_OPT (hw->domain->options, output_stress_type),
		 S, S);
    pressure  = (S[0][0] + S[1][1] + S[2][2]) / 3;
    if        (WOPT_OPT (hw->domain->options, output_stress_type)
               == W_OUTPUT_STRESS_CAUCHY) {
      STORE ("sxx", S[0][0]);
      STORE ("sxy", S[0][1]);
      STORE ("sxz", S[0][2]);
      STORE ("syy", S[1][1]);
      STORE ("syz", S[1][2]);
      STORE ("szz", S[2][2]);
      STORE ("press", pressure);
    } else if (WOPT_OPT (hw->domain->options, output_stress_type)
               == W_OUTPUT_STRESS_2ND_PK) {
      STORE ("Sxx", S[0][0]);
      STORE ("Sxy", S[0][1]);
      STORE ("Sxz", S[0][2]);
      STORE ("Syy", S[1][1]);
      STORE ("Syz", S[1][2]);
      STORE ("Szz", S[2][2]);
      STORE ("Press", pressure);
    } else {
      REPORT_ERROR(hw->domain->errh, "Unknown stress type for output");
    }
  } 
  if (cd->tc->crack_surf->se->compute_equiv_pl_def_at_output_pts) {
    V_get_eqpls (cd->tc, cd->vert, &equiv_pl_def, &equiv_pl_def);
    STORE ("EqPlS", equiv_pl_def);
  } 

  /* these are quantities evaluated for each front vertex */
  STORE ("K_I_0", cd->vert->K_I_0);
  STORE ("K_II_0", cd->vert->K_II_0);
  STORE ("K_III_0", cd->vert->K_III_0);
  STORE ("K_I_c", cd->vert->K_I_c);
  STORE ("K_II_c", cd->vert->K_II_c);
  STORE ("K_III_c", cd->vert->K_III_c);
  STORE ("K_I_equiv", cd->vert->K_I_equiv);
  STORE ("theta", cd->vert->theta);
  STORE ("G_0", cd->vert->G_0);
  STORE ("G_c", cd->vert->G_c);
  STORE ("C", cd->vert->C);
} 




static void 
sampler_store (WEFG_crack_surf_t *cs,
	      W_solution_history_t *h,
	      void *data,
	      double time) 
{
  ts_crack_tip_history_client_data_t *cd =
 (ts_crack_tip_history_client_data_t *) data;

  store (cd, h->writer, time);
} 

/*
   Find the closest front point.
   Find the forward vector at the point on the crack front closest to the given location.
   Note: The front is composed of straight-line segments connecting the front vertices.
 */
static int /* TRUE if closest point forward vector on return is valid. */
nearest_front_point_forward (WEFG_crack_surf_t *cs, /* generic crack surface */
                             W_loc_t *pt, /* location */
                             W_loc_t *front_forward /* output: forward vector at
                                                       the closest front point
                                                     */
                             ) 
{
  triangulated_crack_surf_t *tc = (triangulated_crack_surf_t *) cs->data;
  ts_vert_t *v;
  ts_edge_t *preve, *nexte;
  ts_vert_t *prevv, *nextv;
  W_vect_t dir, xl;
  double dirn, m;
  double mindist = FLT_MAX;
  double dist;
  W_vect_t front_pt, n;
  int had_at_least_one = FALSE;
  W_vect_t forward, tangent, normal;

  loop_through_list (tc->on_front_vertices, v, ts_vert_t *) {
    prevv = NULL;
    nextv = NULL;
    find_prev_next_vertex_on_front (v, tc, &prevv, &nextv);
    if (prevv) { preve = connecting_edge (prevv, v); } 
    else                { preve = NULL; } 
    if (nextv) { nexte = connecting_edge (nextv, v);} 
    else                { nexte = NULL; } 
    if (prevv) {
      VECMAC_3_VEC (dir, =, v->loc, -, prevv->loc);
      dirn = VECMAC_NORM (dir);
      VECMAC_3_VEC (xl,  =, (*pt),    -, prevv->loc);
      m = VECMAC_DOT (xl, dir) / dirn;
      if (m >= -0.01 * dirn && m < 1.01 * dirn) {
        VECMAC_3_SCALAR (front_pt, =, prevv->loc.x + m * dir.x,
                                     prevv->loc.y + m * dir.y,
                                     prevv->loc.z + m * dir.z);
        VECMAC_3_VEC (n, =, front_pt, -, (*pt));
        dist = VECMAC_NORM (n);
        if (dist < mindist) {
          edge_axes (preve, prevv, v, &normal, &tangent, &forward);
          *front_forward   = forward;
          mindist          = dist;
          had_at_least_one = TRUE;
        } 
      }  /* if (m < dirn) */
    } 
    if (nextv) {
      VECMAC_3_VEC (dir, =, nextv->loc, -, v->loc);
      dirn = VECMAC_NORM (dir);
      VECMAC_3_VEC (xl,  =, (*pt),    -, v->loc);
      m = VECMAC_DOT (xl, dir) / dirn;
      if (m >= -0.01 * dirn && m < 1.01 * dirn) {
        VECMAC_3_SCALAR (front_pt, =, v->loc.x + m * dir.x,
                                     v->loc.y + m * dir.y,
                                     v->loc.z + m * dir.z);
        VECMAC_3_VEC (n, =, front_pt, -, (*pt));
        dist = VECMAC_NORM (n);
        if (dist < mindist) {
          edge_axes (nexte, v, nextv, &normal, &tangent, &forward);
          *front_forward   = forward;
          mindist          = dist;
          had_at_least_one = TRUE;
        } 
      }  /* if (m < dirn) */
    } 
  } 
  return had_at_least_one;
} 

/*
   Compute the plane generating vector.
   Compute the plane generating vector for the node on input.
   The plane generating vector is the forward vector at the crack
   front point closest to the node.
 */
static int
plane_generating_vector (WEFG_crack_surf_t *cs, /* generic crack */
                         double atx, /* evaluation point coordinate */
                         double aty, /* evaluation point coordinate */
                         double atz, /* evaluation point coordinate */
                         double rcx, /* EFG node coordinate */
                         double rcy, /* EFG node coordinate */
                         double rcz, /* EFG node coordinate */
                         double *bx, /* output: plane-generating vector coordinate */
                         double *by, /* output: plane-generating vector coordinate */
                         double *bz /* output: plane-generating vector coordinate */
                         ) 
{
  W_vect_t nodeloc, frontptforward;

  VECMAC_3_SCALAR (nodeloc, =, rcx, rcy, rcz);
  if (!nearest_front_point_forward (cs, &nodeloc, &frontptforward)) {
    REPORT_ERROR (cs->se->domain->errh, "No plane generating vector");
    return 0;
  }  else  {
    *bx = frontptforward.x;
    *by = frontptforward.y;
    *bz = frontptforward.z;
    return 1;
  } 
} 




/*
   Compute the wrap-around point.
   Compute the wrap-around point using the plane generating vector.
   Denote the plane given by the node and the plane-generating vector as /P/.
   The wrap-around point is the intersection of /P/ with the crack front
   closest to the evaluation point /at/.

   bugs:

   If there is no intersection of the front with the plane /P/, the wrap-around
   point is {inf,inf,inf}.  Is this the correct solution?
 */
static int /* always TRUE; see the BUGS section */
wrap_around_point (WEFG_crack_surf_t *cs, /* generic crack */
                        double atx, /* evaluation point coordinate */
                         double aty, /* evaluation point coordinate */
                         double atz, /* evaluation point coordinate */
                        double rcx, /* EFG node coordinate */
                         double rcy, /* EFG node coordinate */
                         double rcz, /* EFG node coordinate */
                        double bx, /* plane-generating vector coordinate */
                         double by, /* plane-generating vector coordinate */
                         double bz, /* plane-generating vector coordinate */
                        double *wapx, /* wrap-around point coordinate */
                         double *wapy, /* wrap-around point coordinate */
                         double *wapz /* wrap-around point coordinate */
                   ) 
{
  triangulated_crack_surf_t *tc = (triangulated_crack_surf_t *) cs->data;
  G3D_plane_t plane;
  W_vect_t at, nodeloc, t;
  ts_vert_t *v;
  ts_edge_t *preve, *nexte;
  ts_vert_t *prevv, *nextv;
  double dist, mindist = FLT_MAX;
  G3D_vect_t s;
  double lambda;
  G3D_bool_t parallel;
  G3D_vect_t ab, p;

  /* RAW is this a graceful way to handle invalid points? */
  *wapx = FLT_MAX, *wapy = FLT_MAX, *wapz = FLT_MAX; /* invalid point */
  
  VECMAC_3_SCALAR (at,      =, atx, aty, atz);
  VECMAC_3_SCALAR (nodeloc, =, rcx, rcy, rcz);
  VECMAC_3_SCALAR (t,       =, atx + bx, aty + by, atz + bz);
  G3D_plane_consts_3_alt (&plane, &at, &nodeloc, &t);

  loop_through_list (tc->on_front_vertices, v, ts_vert_t *) {
    prevv = NULL;
    nextv = NULL;
    find_prev_next_vertex_on_front (v, tc, &prevv, &nextv);
    if (prevv) { preve = connecting_edge (prevv, v); } 
    else                { preve = NULL; } 
    if (nextv) { nexte = connecting_edge (nextv, v);} 
    else                { nexte = NULL; } 
    if (prevv) {
      VECMAC_3_VEC (ab, =, v->loc, -, prevv->loc);
      G3D_intersect_plane_and_line (&s, &lambda, &parallel, &prevv->loc,
                                   &ab, &plane, 0, FLT_MAX);
      if (lambda >= -0.01 && lambda <= 1.01) {
        VECMAC_3_VEC (p,  =, nodeloc, -, s);
        dist = VECMAC_NORM (p);
        if (dist < mindist) {
          *wapx = s.x, *wapy = s.y, *wapz = s.z;
          mindist          = dist;
        } 
      }  
    } 
    if (nextv) {
      VECMAC_3_VEC (ab, =, nextv->loc, -, v->loc);
      G3D_intersect_plane_and_line (&s, &lambda, &parallel, &v->loc,
                                   &ab, &plane, 0, FLT_MAX);
      if (lambda >= -0.01 && lambda <= 1.01) {
        VECMAC_3_VEC (p,  =, nodeloc, -, s);
        dist = VECMAC_NORM (p);
        if (dist < mindist) {
          *wapx = s.x, *wapy = s.y, *wapz = s.z;
          mindist          = dist;
        } 
      }  
    } 
  } 
  return TRUE;
} 




typedef struct j_integral_opts_t  {
  WEFG_superelem_t               *se;
  W_loc_t                         at;
  W_vect_t                        forward;
  W_vect_t                        tangent;
  W_vect_t                        normal;
  double                          hsize_Xbar_plus;	/* forward */
  double                          hsize_Xbar_minus;	/* forward */
  double                          hsize_Ybar_plus;	/* tangent */
  double                          hsize_Ybar_minus;	/* tangent */
  double                          hsize_Zbar_plus;	/* normal */
  double                          hsize_Zbar_minus;	/* normal */
  double                          transf_mat[3][3];
  W_loc_t                         currpt;
  W_loc_t                         currpt_bar;
  double                          q;
  double                          rho;
  WEFG_tensor_prod_comp_func      f;
  WEFG_tensor_prod_comp_der_func  f_r;
  double                          j_xbar_static;
  double                          j_xbar_dynamic;
  double                          j_zbar_static;
  double                          j_zbar_dynamic;
}               j_integral_opts_t;

static j_integral_opts_t jopts;

static void
q_func (void);
static void
j_surface_contrib (ts_tri_t *tri, triangulated_crack_surf_t *tc);
#define TRANSF_V_3(VP_, A_, V_) \
  {                                                                      \
   VP_.x  = (A_[0][0]*V_.x + A_[1][0]*V_.y + A_[2][0]*V_.z);            \
   VP_.y  = (A_[0][1]*V_.x + A_[1][1]*V_.y + A_[2][1]*V_.z);            \
   VP_.z  = (A_[0][2]*V_.x + A_[1][2]*V_.y + A_[2][2]*V_.z);            \
} 

static void
j_setup (WEFG_crack_surf_t *cs,
        ts_vert_t *v,
        double hsize_forward_plus,
        double hsize_forward_minus,
        double hsize_tangent_plus,
        double hsize_tangent_minus,
        double hsize_normal_plus,
        double hsize_normal_minus,
        WEFG_tensor_prod_comp_func func,
        WEFG_tensor_prod_comp_der_func func_der
        ) 
{
  W_loc_t origin= {0,0,0} ;
  
  jopts.se               = cs->se;
  jopts.at               = v->loc;
  jopts.forward          = v->forward;
  jopts.tangent          = v->tangent;
  jopts.normal           = v->normal;
  jopts.hsize_Xbar_plus  = hsize_forward_plus;
  jopts.hsize_Ybar_plus  = hsize_tangent_plus;
  jopts.hsize_Zbar_plus  = hsize_normal_plus;
  jopts.hsize_Xbar_minus = hsize_forward_minus;
  jopts.hsize_Ybar_minus = hsize_tangent_minus;
  jopts.hsize_Zbar_minus = hsize_normal_minus;
  jopts.transf_mat[0][0] = jopts.forward.x;
  jopts.transf_mat[1][0] = jopts.forward.y;
  jopts.transf_mat[2][0] = jopts.forward.z;
  jopts.transf_mat[0][1] = jopts.tangent.x;
  jopts.transf_mat[1][1] = jopts.tangent.y;
  jopts.transf_mat[2][1] = jopts.tangent.z;
  jopts.transf_mat[0][2] = jopts.normal.x;
  jopts.transf_mat[1][2] = jopts.normal.y;
  jopts.transf_mat[2][2] = jopts.normal.z;
  jopts.j_xbar_static    = 0;
  jopts.j_xbar_dynamic   = 0;
  jopts.j_zbar_static    = 0;
  jopts.j_zbar_dynamic   = 0;
  jopts.currpt_bar       = origin;
  jopts.rho              = WM3D_get_volume_mass (jopts.se->mat);
  jopts.f                = func;
  jopts.f_r              = func_der;
} 


/*
   Compute the effective moduli.
   Compute the effective elasticity moduli by modifying the
   virgin Young's modulus as proposed by Cherepanov.
 */
static void
compute_E_star (triangulated_crack_surf_t *tc /* instance of triangulated crack */,
                       ts_vert_t *v, /* vertex */
                       double E, double nu,
                       double *E_star /* effective Young's modulus */
                       ) 
{
  double E_plstrain;
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  double displ_grad[3][3], tmp[3][3];
  
  E_plstrain = E / (1 - nu*nu);

  if (tc->allow_E_star) {
    ip_ctx_data.se       = tc->crack_surf->se;
    ip_ctx.time          = WD_time (tc->crack_surf->se->domain);
    ip_ctx.time_step     = WD_time_step (tc->crack_surf->se->domain);
    ip_ctx.ctx_data      = &ip_ctx_data;
    ip_ctx.get_kinem_var = efgss_get_kinem_var;
    
    vert_sftabs (v, tc);
    ip_ctx_data.sftab = v->sftab_pos_normal;/*doesn't matter, is on front*/
    if (! (ip_ctx.get_kinem_var) (&ip_ctx, MAT_3D_DEF_GRAD, displ_grad)) {
      REPORT_ERROR (tc->crack_surf->se->domain->errh, "No displacement gradient");
      return;
    } 
    displ_grad[0][0] -= 1;
    displ_grad[1][1] -= 1;
    displ_grad[2][2] -= 1;
    
    MAT3DUTIL_MULT_ATB_3X3 (tmp, /* = */ jopts.transf_mat, /* * */ displ_grad);
    MAT3DUTIL_MULT_AB_3X3 (displ_grad, /* = */ tmp, /* * */ jopts.transf_mat);
    
    if (displ_grad[0][0] + displ_grad[2][2] != 0) {
      *E_star = E * (1 / (1 - nu*nu) +
                     (nu / (1 + nu) * displ_grad[1][1] /
                      (displ_grad[0][0] + displ_grad[2][2]) 
                      ) 
                     );
      if (*E_star < E) *E_star = E;
      if (*E_star > E_plstrain) *E_star = E_plstrain;
    }  else  {
      *E_star = E_plstrain;
    } 
  }  else  { /* use just the plane strain value */
    *E_star = E_plstrain;
  } 
} 


static void
j_surface (WEFG_crack_surf_t *cs,
          ts_vert_t *v,
          double hsize_forward_plus,
          double hsize_forward_minus,
          double hsize_tangent_plus,
          double hsize_tangent_minus,
          double hsize_normal_plus,
          double hsize_normal_minus,
          WEFG_tensor_prod_comp_func func,
          WEFG_tensor_prod_comp_der_func func_der,
          double *J_forward_static,
          double *J_forward_dynamic,
          double *J_normal_static,
          double *J_normal_dynamic
          ) 
{
  ts_tri_t *tri;
  W_loc_t origin= {0,0,0} ;
  triangulated_crack_surf_t *tc = (triangulated_crack_surf_t *) cs->data;
  
  jopts.se               = cs->se;
  jopts.at               = v->loc;
  jopts.forward          = v->forward;
  jopts.tangent          = v->tangent;
  jopts.normal           = v->normal;
  jopts.hsize_Xbar_plus  = hsize_forward_plus;
  jopts.hsize_Ybar_plus  = hsize_tangent_plus;
  jopts.hsize_Zbar_plus  = hsize_normal_plus;
  jopts.hsize_Xbar_minus = hsize_forward_minus;
  jopts.hsize_Ybar_minus = hsize_tangent_minus;
  jopts.hsize_Zbar_minus = hsize_normal_minus;
  jopts.transf_mat[0][0] = jopts.forward.x;
  jopts.transf_mat[1][0] = jopts.forward.y;
  jopts.transf_mat[2][0] = jopts.forward.z;
  jopts.transf_mat[0][1] = jopts.tangent.x;
  jopts.transf_mat[1][1] = jopts.tangent.y;
  jopts.transf_mat[2][1] = jopts.tangent.z;
  jopts.transf_mat[0][2] = jopts.normal.x;
  jopts.transf_mat[1][2] = jopts.normal.y;
  jopts.transf_mat[2][2] = jopts.normal.z;
  jopts.j_xbar_static    = 0;
  jopts.j_xbar_dynamic   = 0;
  jopts.j_zbar_static    = 0;
  jopts.j_zbar_dynamic   = 0;
  jopts.currpt_bar       = origin;
  jopts.rho              = WM3D_get_volume_mass (jopts.se->mat);
  jopts.f                = func;
  jopts.f_r              = func_der;

  loop_through_list (tc->triangles, tri, ts_tri_t *) {
    j_surface_contrib (tri, tc);
  } 
  *J_forward_static  = jopts.j_xbar_static;
  *J_forward_dynamic = jopts.j_xbar_dynamic;
  *J_normal_static   = jopts.j_zbar_static;
  *J_normal_dynamic  = jopts.j_zbar_dynamic;
} 



static void 
energy_densities (ts_vert_t *v, triangulated_crack_surf_t *tc,
                 double *Wpos, double *Wneg,
                 double *Kpos, double *Kneg);
static double 
tri_area (W_loc_t *p0, W_loc_t *p1, W_loc_t *p2);


static void
j_surface_contrib (ts_tri_t *tri, triangulated_crack_surf_t *tc) 
{
  W_loc_t d;
  double qf[3], qn[3];
  int i;
  double Wpos[3], Wneg[3], Kpos[3], Kneg[3];
  double triareathird;
  W_loc_t normal;
  
  triareathird = tri_area (&tri->vertex[0]->loc,
                           &tri->vertex[1]->loc,
                           &tri->vertex[2]->loc) / 3;
  FOR (i, 3) {
    VECMAC_2_VEC (jopts.currpt, =, tri->vertex[i]->loc);
    VECMAC_3_VEC (d, =, jopts.currpt, -, jopts.at);
    jopts.currpt_bar.x = VECMAC_DOT (jopts.forward, d);
    jopts.currpt_bar.y = VECMAC_DOT (jopts.tangent, d);
    jopts.currpt_bar.z = VECMAC_DOT (jopts.normal,  d);
    q_func ();
    TRANSF_V_3 (normal, jopts.transf_mat, tri->vertex[i]->normal);
    qf[i] = normal.x * jopts.q;
    qn[i] = normal.z * jopts.q;
    energy_densities (tri->vertex[i], tc,
                      &Wpos[i], &Wneg[i], &Kpos[i], &Kneg[i]);
  } 
  jopts.j_xbar_static +=
    triareathird * 
      (qf[0] * (Wpos[0] - Wneg[0]) + 
       qf[1] * (Wpos[1] - Wneg[1]) +
       qf[2] * (Wpos[2] - Wneg[2]));
  jopts.j_zbar_static +=
    triareathird *
      (qn[0] * (Wpos[0] - Wneg[0]) + 
       qn[1] * (Wpos[1] - Wneg[1]) +
       qn[2] * (Wpos[2] - Wneg[2]));
  jopts.j_xbar_dynamic += 
    triareathird * 
      (qf[0] * (Kpos[0] - Kneg[0]) + 
       qf[1] * (Kpos[1] - Kneg[1]) +
       qf[2] * (Kpos[2] - Kneg[2]));
  jopts.j_zbar_dynamic += 
    triareathird * 
      (qn[0] * (Kpos[0] - Kneg[0]) + 
       qn[1] * (Kpos[1] - Kneg[1]) +
       qn[2] * (Kpos[2] - Kneg[2]));
} 



static double 
tri_area (W_loc_t *p0, W_loc_t *p1, W_loc_t *p2) 
{
  W_vect_t v0, v1, cp= {0,0,0} ;
  VECMAC_3_VEC (v0, =, (*p1), -, (*p0));
  VECMAC_3_VEC (v1, =, (*p2), -, (*p0));
  VECMAC_CROSS (cp, =, v0, x, v1);
  return 0.5 * (VECMAC_NORM (cp));
} 


static void 
energy_densities (ts_vert_t *v,
                  triangulated_crack_surf_t *tc /* instance of triangulated crack */,
                  double *Wpos, double *Wneg,
                  double *Kpos, double *Kneg) 
{
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  W_vect_t vel;

  vert_sftabs (v, tc);
  
  ip_ctx_data.se       = tc->crack_surf->se;
  ip_ctx.time          = WD_time (tc->crack_surf->se->domain);
  ip_ctx.time_step     = WD_time_step (tc->crack_surf->se->domain);
  ip_ctx.ctx_data      = &ip_ctx_data;
  ip_ctx.get_kinem_var = efgss_get_kinem_var;

  ip_ctx_data.sftab    = v->sftab_pos_normal;
  EFGSS_AT_NEW_IP ();
  if (!WM3D_get_mat_state_var (tc->crack_surf->se->mat, v->ms_pos_normal,
			      &ip_ctx, MAT_3D_DEFORM_ENERGY_DENSITY, Wpos)) {
    REPORT_ERROR (tc->crack_surf->se->domain->errh, "Failed to retrieve deformation energy");
    return;
  } 

  ip_ctx_data.sftab    = v->sftab_neg_normal;
  EFGSS_AT_NEW_IP ();
  if (!WM3D_get_mat_state_var (tc->crack_surf->se->mat, v->ms_neg_normal,
			      &ip_ctx, MAT_3D_DEFORM_ENERGY_DENSITY, Wneg)) {
    REPORT_ERROR (tc->crack_surf->se->domain->errh, "Failed to retrieve deformation energy");
    return;
  } 

  efgsf_eval_curr_vel (&v->sftab_pos_normal, &vel.x, &vel.y, &vel.z);
  *Kpos = 0.5 * jopts.rho * VECMAC_DOT (vel, vel);
  efgsf_eval_curr_vel (&v->sftab_neg_normal, &vel.x, &vel.y, &vel.z);
  *Kneg = 0.5 * jopts.rho * VECMAC_DOT (vel, vel);
} 


static void
q_func (void) 
{
  double xs = (jopts.currpt_bar.x > 0
               ? jopts.hsize_Xbar_plus
               : jopts.hsize_Xbar_minus);
  double ys = (jopts.currpt_bar.y > 0
               ? jopts.hsize_Ybar_plus
               : jopts.hsize_Ybar_minus);
  double zs = (jopts.currpt_bar.z > 0
               ? jopts.hsize_Zbar_plus
               : jopts.hsize_Zbar_minus);
  double xbar = jopts.currpt_bar.x / xs;
  double ybar = jopts.currpt_bar.y / ys;
  double zbar = jopts.currpt_bar.z / zs;
  
  jopts.q = (jopts.f (xbar) * jopts.f (ybar) * jopts.f (zbar));
} 

static void
compute_ks (double J1, double J2, double E_star,
           double opening, double shearing, double tearing,
           double *KI, double *KII);

static void
eval_Ks (triangulated_crack_surf_t *tc /* instance of triangulated crack */,
        ts_vert_t *v,
        double J1static, double J2static, 
        double J1dynamic, double J2dynamic, 
        double *KIstatic, double *KIIstatic, double *KIIIstatic,
        double *KIdynamic, double *KIIdynamic, double *KIIIdynamic,
        double *o_displ, double *s_displ, double *t_displ) 
{
  double E, nu, E_star, E_plstrain;
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  double displ_grad[3][3], tmp[3][3];
  
  if (!WM3D_get_mat_const (tc->crack_surf->se->mat,
                          MAT_3D_YOUNG_MODULUS, &E)) 
    REPORT_FATAL_ERROR (tc->crack_surf->se->domain->errh, "no shear modulus");
  if (!WM3D_get_mat_const (tc->crack_surf->se->mat,
                          MAT_3D_POISSON, &nu)) 
    REPORT_FATAL_ERROR (tc->crack_surf->se->domain->errh, "no Poisson ratio");

  ip_ctx_data.se       = tc->crack_surf->se;
  ip_ctx.time          = WD_time (tc->crack_surf->se->domain);
  ip_ctx.time_step     = WD_time_step (tc->crack_surf->se->domain);
  ip_ctx.ctx_data      = &ip_ctx_data;
  ip_ctx.get_kinem_var = efgss_get_kinem_var;

  vert_sftabs (v, tc);
  ip_ctx_data.sftab    = v->sftab_pos_normal; /*does not matter; is on front */
  if (! (ip_ctx.get_kinem_var) (&ip_ctx, MAT_3D_DEF_GRAD, displ_grad)) {
    REPORT_ERROR (tc->crack_surf->se->domain->errh, "No displacement gradient");
    return;
  } 
  displ_grad[0][0] -= 1;
  displ_grad[1][1] -= 1;
  displ_grad[2][2] -= 1;
  
  MAT3DUTIL_MULT_ATB_3X3 (tmp, /* = */ jopts.transf_mat, /* * */ displ_grad);
  MAT3DUTIL_MULT_AB_3X3 (displ_grad, /* = */ tmp, /* * */ jopts.transf_mat);

  E_plstrain = E / (1 - nu*nu);
  if (displ_grad[0][0] + displ_grad[2][2] != 0) {
    E_star = E * (1 / (1 - nu*nu) +
 (nu / (1 + nu) * displ_grad[1][1] /
 (displ_grad[0][0] + displ_grad[2][2]) 
                   ) 
                  );
    if (E_star < 0 || E_star > E_plstrain) E_star = E_plstrain;
  }  else  {
    E_star = E_plstrain;
  } 

  rel_displ_behind_front (tc, v, o_displ, s_displ, t_displ);

  compute_ks (J1static, J2static, E_star, *o_displ, *s_displ, *t_displ,
             KIstatic, KIIstatic);
  compute_ks (J1dynamic, J2dynamic, E_star, *o_displ, *s_displ, *t_displ,
             KIdynamic, KIIdynamic);
} 


static void
compute_ks (double J1, double J2, double E_star,
           double opening, double shearing, double tearing,
           double *KI, double *KII) 
{
  double a, b, sqrt_E_star, sqrta, sqrtb;
  double root1, root2;

  sqrt_E_star = sqrt (E_star);

  if (J1 < 0) J1 = 0;           /* cannot be negative */
  a = J1 - J2;
  b = J1 + J2;
  if (a < 0) a = 0;    /* cannot be negative */
  if (b < 0) b = 0;    /* cannot be negative */
  
  sqrta = sqrt (a);
  sqrtb = sqrt (b);
  
  root1  = 0.5 * sqrt_E_star * fabs (sqrta + sqrtb);
  root2  = 0.5 * sqrt_E_star * fabs (sqrta - sqrtb);
  if (fabs (opening) > fabs (shearing)) {
    *KI  = max (root1, root2);
    *KII = min (root1, root2);
  }  else  {
    *KI  = min (root1, root2);
    *KII = max (root1, root2);
  } 
  if (opening < 0) *KI = - (*KI);
  if (shearing < 0) *KII = - (*KII);
} 


static WEFG_sf_tab_t sftab;
static int first = TRUE;

/* evaluate shape functions.
   evaluate shape function table at a point
 */
static void
loc_sftabs (W_loc_t *loc, /* location */
            triangulated_crack_surf_t *tc /* instance of triangulated crack */,
            WEFG_eval_for_packet_t *efp /* evaluation packet */
            ) 
{
  SFL_sf_table_t sf;
  WEFG_hex_volume_t *eh;
  WEFG_if_data_t if_data;
  W_loc_t ploc;

  if (first) {
    INIT_SFTAB (sftab);
    first = FALSE;
  } 
  
  sf = efgsf_sf_table (efp->se);
  
  if (!find_hex_enclosing_loc (tc, loc, &eh, &ploc)) {
    find_hex_enclosing_loc (tc, loc, &eh, &ploc);
    REPORT_ERROR (tc->crack_surf->se->domain->errh, "no enclosing hex");
  } 
  
  USE_DEFAULT_SEARCH_FUNC (efp->se);
  if (hex_is_interface_vol (eh)) {
    if_data.h     = eh;
    if_data.xi    = ploc.x;
    if_data.eta   = ploc.y;
    if_data.theta = ploc.z;
    if (!SFL_shape_func_at (sf, loc->x, loc->y, loc->z,
                           efp, (SFL_generic_ptr_t) &if_data)) {
      SFL_ERROR (sf, loc->x, loc->y, loc->z, E_K_FATAL);
    } 
  }  else  {
    if (!SFL_shape_func_at (sf, loc->x, loc->y, loc->z, efp, NULL)) {
      SFL_ERROR (sf, loc->x, loc->y, loc->z, E_K_FATAL);
    } 
  } 
  efgsf_create_sftab (efp->se, &sftab, sf, NULL, FALSE);
  TOUCH_SFTAB (sftab, efp->se);
} 

/* evaluate current location of point.
   evaluate current location of point.
 */
static void
loc_curr_loc (W_loc_t *loc, /* location in reference coordinates */
              triangulated_crack_surf_t *tc /* instance of triangulated crack */,
              WEFG_eval_for_packet_t *efp, /* evaluation packet */
              W_loc_t *curr_loc /* current location */
              ) 
{
  double x, y, z;

  loc_sftabs (loc, tc, efp);

  efgsf_eval_curr_loc (&sftab, &x, &y, &z);
  curr_loc->x = x; curr_loc->y = y; curr_loc->z = z;
} 




/* find integration volume enclosing point.
   Find integration volume enclosing a given point.
 */
static int /* TRUE if point inside any cell; FALSE otherwise */
find_hex_enclosing_loc (triangulated_crack_surf_t *tc /* instance of triangulated crack */,
                        W_loc_t *loc, /* locaiton */
                        WEFG_hex_volume_t **h, /* output: integration cell */
                        W_loc_t *ploc) /* output: parametric coordinates of the input
                                        point in the hex cell */
{
  double tol = 0.0001;

retry:
  *h = efg_find_enclosing_hex (tc->crack_surf->se, loc, tol, ploc);
  if (*h == NULL) {
    if (tol < 1) {
      tol *= 10;
      goto retry;
    }  else  {
      return FALSE;
    } 
  } 
  return TRUE;
} 


static void
rel_displ_behind_front (triangulated_crack_surf_t *tc /* instance of triangulated crack */,
                       ts_vert_t *v, 
                       double *opening, double *shearing, double *tearing) 
{
  W_loc_t loc;
  W_vect_t b, d, ld;
  W_loc_t curr_loc_pos;
  W_loc_t curr_loc_neg;
  double rat = tc->dist_to_front_for_displ_ratio;
  double dist = rat * tc->crack_surf->se->max_support_size;
  WEFG_eval_for_packet_t efp;

  b = v->forward;
  VECMAC_SCALAR (b, *=, -dist);

  VECMAC_3_VEC (loc, =, v->loc, +, b);
  efp.se         = tc->crack_surf->se;
  efp.crack_surf = tc->crack_surf;
  efp.crack_surf_normal = v->normal;
  efp.pt_wrt_crack_surf_classification = WEFG_POSITIVE_NORMAL;
  loc_curr_loc (&loc, tc, &efp, &curr_loc_pos);
  efp.pt_wrt_crack_surf_classification = WEFG_NEGATIVE_NORMAL;
  loc_curr_loc (&loc, tc, &efp, &curr_loc_neg);

  VECMAC_3_VEC (d, =, curr_loc_pos, -, curr_loc_neg);
  TRANSF_V_3 (ld, jopts.transf_mat, d);
  *opening  = ld.z;
  *shearing = ld.x;
  *tearing  = ld.y;
} 


/*
   Is integration volume intersected by the crack?
   Is integration volume intersected by the crack?  Check for all hex edges.
 */
static W_bool_t 
hex_intersected_by_crack_surf (WEFG_crack_surf_t *cs, /* crack surface */
                              WEFG_hex_volume_t *h /* integration volume */
                              ) 
{
  triangulated_crack_surf_t *tc = (triangulated_crack_surf_t *) cs->data;
  G3D_segment_t s;
  G3D_box_t hbox;
  int result = FALSE;
  int edge;
  
  G3D_INIT_BOX (hbox, h->vertex[0]->loc);
  G3D_UPDT_BOX (hbox, h->vertex[1]->loc);
  G3D_UPDT_BOX (hbox, h->vertex[2]->loc);
  G3D_UPDT_BOX (hbox, h->vertex[3]->loc);
  G3D_UPDT_BOX (hbox, h->vertex[4]->loc);
  G3D_UPDT_BOX (hbox, h->vertex[5]->loc);
  G3D_UPDT_BOX (hbox, h->vertex[6]->loc);
  G3D_UPDT_BOX (hbox, h->vertex[7]->loc);

  tc->bbt->test_box_only = TRUE;
  s.start = hbox.lower;
  s.end   = hbox.upper;
  if (!bbt_intersects_a_tri (tc->bbt, &s.start, &s.end)) {
    result = FALSE;
    goto exit;
  } 
  
  tc->bbt->eps = 0;
  tc->bbt->test_box_only = FALSE;

  FOR (edge, 12) {
    s.start = h->vertex[g3dhex_vertex_on_edge[edge][0]]->loc;
    s.end   = h->vertex[g3dhex_vertex_on_edge[edge][1]]->loc;
    if (bbt_intersects_a_tri (tc->bbt, &s.start, &s.end)) {
      result = TRUE;
      goto exit;
    } 
  } 

 exit:
  tc->bbt->test_box_only = FALSE;
  return result;
} 



/* comput crack speed.
 compute crack speed from a quadratic equation (assumes crack speed independent
 dynamic fracture toughness)
 */
static double
C_from_G (double G, /* energy release rate */
          double G_cr, /* critical energy release rate */
          double C_R /* Rayleigh speed */
          ) 
{
  double a, b, c;
  double roots[2];
  a = 1;
  b = -2;
  c = 1 - G_cr / G;
  CkitQuadraticRoots (a, b, c, roots);

  return C_R * (min (roots[0], roots[1]));
} 


/* Predict advance of vertex.
   Predict advance of a crack front vertex.

   Algorithm:

   (i) Set the advance vector intially to zero.
   (ii) Evaluate shape function table at the vertex.
   (iii) Compute the effective Young's modulus at the vertex.
   (iv) If the equivalent opening SIF is negative, crack is closed; return immediately.
   (v) Evaluate material constants.
   (vi) If the crack speed is prescribed, move the vertex according
   to the prescribed values; otherwise if the energy release rate is larger
   than the critical energy release rate and crack is not required to remain stationary
   compute the crack speed, clamp it to a fraction of the Rayleigh speed, evaluate
   the tilt, and finally compute the vertex advance vector.
   (vii) Store data in the vertex for subsequent use.
 */
static void
predict_advance (ts_vert_t *v, /* vertex */
                 ts_vert_t *prevv, /* previous vertex on the front (or NULL) */
                 ts_vert_t *nextv, /* next vertex on the front (or NULL) */
                 ts_edge_t *preve, /* previous edge on the front (or NULL) */
                 ts_edge_t *nexte,  /* next edge on the front (or NULL) */
                 triangulated_crack_surf_t *tc /* instance of triangulated crack */
                 ) 
{
  double G_cr, dt, adv_length;
  W_vect_t zerovect =  {0,0,0} ;
  double dyn_fracture_toughness;
  double E, nu, rho, E_star, mu;
  double C_R, C;
  double kC;
  double c_d, c_s;

  /* assume no advance */
  v->advance_vector = zerovect;
  
  vert_sftabs (v, tc);

  j_setup (tc->crack_surf, v, 0, 0, 0, 0, 0, 0, 0, 0);   /* dummy setup... */

  if (!WM3D_get_mat_const (tc->crack_surf->se->mat,
                          MAT_3D_YOUNG_MODULUS, &E)) 
    REPORT_FATAL_ERROR (tc->crack_surf->se->domain->errh, "no elastic modulus");
  if (!WM3D_get_mat_const (tc->crack_surf->se->mat,
                          MAT_3D_POISSON, &nu)) 
    REPORT_FATAL_ERROR (tc->crack_surf->se->domain->errh, "no Poisson ratio");
  if (!WM3D_get_mat_const (tc->crack_surf->se->mat,
                          MAT_3D_SHEAR_MODULUS, &mu)) 
    REPORT_FATAL_ERROR (tc->crack_surf->se->domain->errh, "no shear modulus");
  rho = WM3D_get_volume_mass (tc->crack_surf->se->mat);
  
  c_d = W_3D_dilatational_wave_speed (E, nu, rho);
  c_s = W_shear_wave_speed (E, nu, rho);

  compute_E_star (tc, v, E, nu, &E_star);/*...to be able to compute E_star */

  if (v->K_I_equiv <= 0) /* no advance for closing crack */
    return;

  if (!WM3D_get_mat_const (tc->crack_surf->se->mat,
                           MAT_3D_DYN_FRACTURE_TOUGHNESS,
                           &dyn_fracture_toughness)) 
    REPORT_FATAL_ERROR (tc->crack_surf->se->domain->errh, "no fracture toughness");
  
  G_cr = dyn_fracture_toughness * dyn_fracture_toughness / E_star;
  
  if (!WM3D_get_mat_const (tc->crack_surf->se->mat,
                           MAT_3D_RAYLEIGH_PSTRAIN_WAVE_SPEED, &C_R)) 
    REPORT_FATAL_ERROR (tc->crack_surf->se->domain->errh, "no Rayleigh wave speed");
  
  if (tc->advance_prescribed && (! (tc->is_stationary)) ) {
    advance_prescribed (v, tc);
  }  else if (! (tc->is_stationary)) {
    dt  = WD_time_step (tc->crack_surf->se->domain);
    C = v->C;
    if (tc->drive_by_vel_incr) {
      C += C_R * tc->rayleigh_vel_frac * (v->G_c / G_cr - 1) * dt * tc->vel_incr_fraction;
    } else {
      if (v->G_c > G_cr)
        C = C_R * tc->rayleigh_vel_frac * (1 - G_cr / v->G_c);
      else
        C = 0;
      printf ("C/C_R = %g\n", C / C_R);
    }
    if (C > 0) {
      double costh, sinth;
      /* Do not allow higher speed than a fraction of C_R */
      if (C > C_R * tc->rayleigh_vel_frac) 
        C = C_R * tc->rayleigh_vel_frac;
      adv_length = C * dt; 
      costh = cos (v->theta);
      sinth = sin (v->theta);
      if (costh < 0) {
        sprintf (W_buf (), "Bad theta=%g (K_I=%g, K_II=%g) ",
                 v->theta, v->K_I_0, v->K_II_0);
        REPORT_FATAL_ERROR (tc->crack_surf->se->domain->errh, W_buf ());
      } 
      v3_comb (&v->forward, &v->normal, &v->advance_vector, costh, sinth);
      VECMAC_NORMALIZE (v->advance_vector);
      VECMAC_SCALAR (v->advance_vector, *=, adv_length);
    } /* if (C > 0) */ else {
      C = 0;
    }
    v->C = C;
  }
} 


/* third vertex in a triangle.
 find the third vertex in the triangle
 */
static ts_vert_t * /* third vertex */
remaining_vert (ts_tri_t *t, /* triangle */
                ts_vert_t *v0, /* vertex */
                ts_vert_t *v1 /* vertex */
                ) 
{
  int have[3] =  {0,0,0} ;
  if (t->vertex[0] == v0 || t->vertex[0] == v1) have[0] = TRUE;
  if (t->vertex[1] == v0 || t->vertex[1] == v1) have[1] = TRUE;
  if (t->vertex[2] == v0 || t->vertex[2] == v1) have[2] = TRUE;
  if (!have[0]) 
    return t->vertex[0];
  else if (!have[1]) 
    return t->vertex[1];
  else if (!have[2]) 
    return t->vertex[2];
  else
    return NULL;
} 
  


  

#if 1
/*
   Adapt crack front.
   Adjust crack front discretization by subdividing too long edges in two
   (with corresponding split of adjacent triangles).
 */
static void
adjust_edge_lengths (ts_vert_t *v, /* front vertex */
                     triangulated_crack_surf_t *tc /* instance of triangulated crack */
                     ) 
{
  ts_vert_t *prevv, *nextv, *fes, *fee, *fem, *rv, *v0, *v1, *v2;
  W_loc_t newloc;
  ts_edge_t *preve, *nexte, *fe, *me, *newfe, *e0, *e1, *e2;
  ts_tri_t *t, *oldt;
  W_vect_t ev;
  double el;

  prevv = NULL;
  nextv = NULL;
  find_prev_next_vertex_on_front (v, tc, &prevv, &nextv);
  if (prevv) { preve = connecting_edge (prevv, v); } 
  else                { preve = NULL; } 
  if (nextv) { nexte = connecting_edge (nextv, v);} 
  else                { nexte = NULL; } 

#define TWO_FOR_ONE() \
   {                                                                      \
    oldt = fe->triangles[0];                                             \
    VECMAC_3_VEC (ev, =, fe->vertex[1]->loc, -, fe->vertex[0]->loc);     \
    el = VECMAC_NORM (ev);                                               \
    if (el >= tc->edge_length_upper_bound) {                             \
      rv  = remaining_vert (oldt, fes, fee);                             \
      e0  = connecting_edge (fes, rv);                                   \
      e1  = connecting_edge (fee, rv);                                   \
      forget_tri (tc, oldt);                                             \
      forget_edge (tc, fe);                                              \
      v_attach_e (rv, e0); v_attach_e (fes, e0);                         \
      v_attach_e (rv, e1); v_attach_e (fee, e1);                         \
                                                                         \
      VECMAC_3_VEC (newloc, =, 0.5 * fes->loc, +, 0.5 * fee->loc);       \
      fem = new_vert (tc, 0, &newloc);  add_to_tail (tc->vertices, fem); \
      fem->is_on_front = TRUE;                                           \
      fem->ms_neg_normal                                                 \
        = efg_interpolate_mat_state                                      \
 (tc->crack_surf->se,                                           \
           &fem->loc,                                                    \
           tc->crack_surf->se->max_support_size / 10);                   \
      if (fem->ms_neg_normal == NULL) REPORT_ERROR (tc->crack_surf->se->domain->errh, "No material state"); \
      fem->ms_pos_normal                                                 \
        = efg_interpolate_mat_state                                      \
 (tc->crack_surf->se,                                           \
           &fem->loc,                                                    \
           tc->crack_surf->se->max_support_size / 10);                   \
      if (fem->ms_pos_normal == NULL) REPORT_ERROR (tc->crack_surf->se->domain->errh, "No material state"); \
      me    = new_edge (tc, 0, rv, fem); add_to_tail (tc->edges, me);    \
      v_attach_e (fem, me); v_attach_e (rv, me);                         \
                                                                         \
      newfe = new_edge (tc, 0, fes, fem); add_to_tail (tc->edges, newfe); \
      v_attach_e (fes, newfe); v_attach_e (fem, newfe);                  \
      v0 = rv; v1 = fes; v2 = fem;                                       \
      e0 = connecting_edge (v0, v1);                                     \
      e1 = connecting_edge (v1, v2);                                     \
      e2 = connecting_edge (v2, v0);                                     \
      t   = new_tri (tc, 0, v0, v1, v2, e0, e1, e2);                     \
      add_to_head (tc->triangles, t);                                    \
                                                                         \
      newfe = new_edge (tc, 0, fee, fem); add_to_tail (tc->edges, newfe); \
      v_attach_e (fee, newfe); v_attach_e (fem, newfe);                  \
      v0 = rv; v1 = fem; v2 = fee;                                       \
      e0 = connecting_edge (v0, v1);                                     \
      e1 = connecting_edge (v1, v2);                                     \
      e2 = connecting_edge (v2, v0);                                     \
      t   = new_tri (tc, 0, v0, v1, v2, e0, e1, e2);                     \
      add_to_head (tc->triangles, t);                                    \
    }                                                                     \
  } 
  
  
  if (preve) {
    fe  = preve;
    fes = prevv;
    fee = v;
    TWO_FOR_ONE ();
  } 
  if (nexte) {
    fe  = nexte;
    fes = v;
    fee = nextv;
    TWO_FOR_ONE ();
  } 
} 
#else

#endif



#endif /* EFG_WASP  */



